Type: Package
Package: officer
Title: Manipulation of Microsoft Word and PowerPoint Documents
Version: 0.7.5
Authors@R: c(
    person("David", "Gohel", , "david.gohel@ardata.fr", role = c("aut", "cre")),
    person("Stefan", "Moog", , "moogs@gmx.de", role = "aut"),
    person("Mark", "Heckmann", , "heckmann.mark@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0002-0736-7417")),
    person("ArData", role = "cph"),
    person("Frank", "Hangler", , "frank@plotandscatter.com", role = "ctb",
           comment = "function body_replace_all_text"),
    person("Liz", "Sander", , "lsander@civisanalytics.com", role = "ctb",
           comment = "several documentation fixes"),
    person("Anton", "Victorson", , "anton@victorson.se", role = "ctb",
           comment = "fixes xml structures"),
    person("Jon", "Calder", , "jonmcalder@gmail.com", role = "ctb",
           comment = "update vignettes"),
    person("John", "Harrold", , "john.m.harrold@gmail.com", role = "ctb",
           comment = "function annotate_base"),
    person("John", "Muschelli", , "muschellij2@gmail.com", role = "ctb",
           comment = "google doc compatibility"),
    person("Bill", "Denney", , "wdenney@humanpredictions.com", role = "ctb",
           comment = c(ORCID = "0000-0002-5759-428X", "function as.matrix.rpptx")),
    person("Nikolai", "Beck", , "beck.nikolai@gmail.com", role = "ctb",
           comment = "set speaker notes for .pptx documents"),
    person("Greg", "Leleu", , "gregoire.leleu@gmail.com", role = "ctb",
           comment = "fields functionality in ppt"),
    person("Majid", "Eismann", role = "ctb"),
    person("Wahiduzzaman", "Khan", role = "ctb",
           comment = "vectorization of remove_slide"),
    person("Hongyuan", "Jia", , "hongyuanjia@cqust.edu.cn", role = "ctb",
           comment = c(ORCID = "0000-0002-0075-8183")),
    person("Michael", "Stackhouse", , "mike.stackhouse@atorusresearch.com", role = "ctb")
  )
Description: Access and manipulate 'Microsoft Word', 'RTF' and 'Microsoft
    PowerPoint' documents from R.  The package focuses on tabular and
    graphical reporting from R; it also provides two functions that let
    users get document content into data objects. A set of functions lets
    add and remove images, tables and paragraphs of text in new or
    existing documents.  The package does not require any installation of
    Microsoft products to be able to write Microsoft files.
License: MIT + file LICENSE
URL: https://ardata-fr.github.io/officeverse/,
        https://davidgohel.github.io/officer/
BugReports: https://github.com/davidgohel/officer/issues
Imports: cli, dplyr, graphics, grDevices, openssl, R6, ragg, stats,
        tidyr, utils, uuid, xml2 (>= 1.1.0), zip (>= 2.1.0)
Suggests: devEMF, doconv (>= 0.3.0), gdtools, ggplot2, knitr, magick,
        rmarkdown, rsvg, mschart, testthat, withr
Encoding: UTF-8
Collate: 'core_properties.R' 'custom_properties.R' 'defunct.R'
        'dev-utils.R' 'docx_add.R' 'docx_comments.R' 'docx_cursor.R'
        'docx_embed_font.R' 'docx_part.R' 'docx_replace.R'
        'docx_section.R' 'docx_settings.R' 'docx_styles.R'
        'docx_utils_funs.R' 'empty_content.R' 'formatting_properties.R'
        'fortify_docx.R' 'fortify_pptx.R' 'knitr_utils.R' 'officer.R'
        'ooxml.R' 'ooxml_block_objects.R' 'ooxml_run_objects.R'
        'openxml_content_type.R' 'openxml_document.R' 'pack_folder.R'
        'ph_location.R' 'post-proc.R' 'ppt_class_dir_collection.R'
        'ppt_classes.R' 'ppt_notes.R' 'ppt_ph_dedupe_layout.R'
        'ppt_ph_manipulate.R' 'ppt_ph_rename_layout.R'
        'ppt_ph_with_methods.R' 'pptx_informations.R'
        'pptx_layout_helper.R' 'pptx_matrix.R' 'utils.R'
        'pptx_slide_manip.R' 'read_docx.R' 'docx_write.R'
        'read_docx_styles.R' 'read_pptx.R' 'read_xlsx.R'
        'relationship.R' 'rtf.R' 'shape_properties.R' 'shorcuts.R'
        'docx_append_context.R' 'utils-xml.R' 'deprecated.R' 'zzz.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-15 12:39:25 UTC; davidgohel
Author: David Gohel [aut, cre],
  Stefan Moog [aut],
  Mark Heckmann [aut] (ORCID: <https://orcid.org/0000-0002-0736-7417>),
  ArData [cph],
  Frank Hangler [ctb] (function body_replace_all_text),
  Liz Sander [ctb] (several documentation fixes),
  Anton Victorson [ctb] (fixes xml structures),
  Jon Calder [ctb] (update vignettes),
  John Harrold [ctb] (function annotate_base),
  John Muschelli [ctb] (google doc compatibility),
  Bill Denney [ctb] (ORCID: <https://orcid.org/0000-0002-5759-428X>,
    function as.matrix.rpptx),
  Nikolai Beck [ctb] (set speaker notes for .pptx documents),
  Greg Leleu [ctb] (fields functionality in ppt),
  Majid Eismann [ctb],
  Wahiduzzaman Khan [ctb] (vectorization of remove_slide),
  Hongyuan Jia [ctb] (ORCID: <https://orcid.org/0000-0002-0075-8183>),
  Michael Stackhouse [ctb]
Maintainer: David Gohel <david.gohel@ardata.fr>
Repository: CRAN
Date/Publication: 2026-05-15 14:10:02 UTC
Built: R 4.5.2; ; 2026-05-15 18:16:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",app_properties)
S3method("[",core_properties)
S3method("[",custom_properties)
S3method("[<-",app_properties)
S3method("[<-",core_properties)
S3method("[<-",custom_properties)
S3method(as.character,fp_tab)
S3method(as.character,fp_tabs)
S3method(as.character,officer_cursor)
S3method(as.data.frame,external_img)
S3method(as.data.frame,fpar)
S3method(as.matrix,rpptx)
S3method(body_add,block_caption)
S3method(body_add,block_list)
S3method(body_add,block_list_items)
S3method(body_add,block_pour_docx)
S3method(body_add,block_section)
S3method(body_add,block_toc)
S3method(body_add,character)
S3method(body_add,data.frame)
S3method(body_add,external_img)
S3method(body_add,factor)
S3method(body_add,fpar)
S3method(body_add,gg)
S3method(body_add,numeric)
S3method(body_add,plot_instr)
S3method(body_add,run_columnbreak)
S3method(body_add,run_pagebreak)
S3method(dim,external_img)
S3method(format,fp_cell)
S3method(format,fp_par)
S3method(format,fp_text)
S3method(fortify_location,location_fullsize)
S3method(fortify_location,location_id)
S3method(fortify_location,location_label)
S3method(fortify_location,location_left)
S3method(fortify_location,location_manual)
S3method(fortify_location,location_right)
S3method(fortify_location,location_template)
S3method(fortify_location,location_type)
S3method(length,rdocx)
S3method(length,rpptx)
S3method(length,rxlsx)
S3method(ph_from_location,location_label)
S3method(ph_from_location,location_type)
S3method(print,block_caption)
S3method(print,block_list_items)
S3method(print,block_pour_docx)
S3method(print,block_section)
S3method(print,block_table)
S3method(print,block_toc)
S3method(print,fp_border)
S3method(print,fp_cell)
S3method(print,fp_par)
S3method(print,fp_text)
S3method(print,ftext)
S3method(print,layout_info)
S3method(print,rdocx)
S3method(print,rpptx)
S3method(print,rtf)
S3method(print,run_wordtext)
S3method(print,rxlsx)
S3method(print,sp_line)
S3method(print,sp_lineend)
S3method(print,unordered_list)
S3method(rtf_add,block_list)
S3method(rtf_add,block_section)
S3method(rtf_add,block_toc)
S3method(rtf_add,character)
S3method(rtf_add,double)
S3method(rtf_add,factor)
S3method(rtf_add,fpar)
S3method(rtf_add,gg)
S3method(rtf_add,plot_instr)
S3method(set_notes,block_list)
S3method(set_notes,character)
S3method(sheet_add_drawing,external_img)
S3method(sheet_add_drawing,gg)
S3method(sheet_write_data,block_list)
S3method(sheet_write_data,character)
S3method(sheet_write_data,data.frame)
S3method(sheet_write_data,default)
S3method(sheet_write_data,fpar)
S3method(to_html,block_list)
S3method(to_html,external_img)
S3method(to_html,fpar)
S3method(to_html,ftext)
S3method(to_html,hyperlink_ftext)
S3method(to_pml,block_list)
S3method(to_pml,block_list_items)
S3method(to_pml,block_table)
S3method(to_pml,external_img)
S3method(to_pml,fpar)
S3method(to_pml,ftext)
S3method(to_pml,hyperlink_ftext)
S3method(to_pml,sp_line)
S3method(to_pml,table_conditional_formatting)
S3method(to_pml,unordered_list)
S3method(to_rtf,block_list)
S3method(to_rtf,block_section)
S3method(to_rtf,block_toc)
S3method(to_rtf,default)
S3method(to_rtf,external_img)
S3method(to_rtf,floating_external_img)
S3method(to_rtf,fpar)
S3method(to_rtf,ftext)
S3method(to_rtf,hyperlink_ftext)
S3method(to_rtf,page_mar)
S3method(to_rtf,page_size)
S3method(to_rtf,prop_section)
S3method(to_rtf,run_autonum)
S3method(to_rtf,run_columnbreak)
S3method(to_rtf,run_linebreak)
S3method(to_rtf,run_pagebreak)
S3method(to_rtf,run_reference)
S3method(to_rtf,run_tab)
S3method(to_rtf,run_word_field)
S3method(to_rtf,section_columns)
S3method(to_wml,block_caption)
S3method(to_wml,block_gg)
S3method(to_wml,block_list)
S3method(to_wml,block_list_items)
S3method(to_wml,block_pour_docx)
S3method(to_wml,block_section)
S3method(to_wml,block_table)
S3method(to_wml,block_toc)
S3method(to_wml,external_img)
S3method(to_wml,floating_external_img)
S3method(to_wml,fp_cell)
S3method(to_wml,fp_par)
S3method(to_wml,fp_tab)
S3method(to_wml,fp_tabs)
S3method(to_wml,fp_text)
S3method(to_wml,fpar)
S3method(to_wml,ftext)
S3method(to_wml,hyperlink_ftext)
S3method(to_wml,page_mar)
S3method(to_wml,page_size)
S3method(to_wml,prop_section)
S3method(to_wml,prop_table)
S3method(to_wml,run_autonum)
S3method(to_wml,run_bookmark)
S3method(to_wml,run_columnbreak)
S3method(to_wml,run_comment)
S3method(to_wml,run_footnote)
S3method(to_wml,run_footnoteref)
S3method(to_wml,run_linebreak)
S3method(to_wml,run_pagebreak)
S3method(to_wml,run_reference)
S3method(to_wml,run_tab)
S3method(to_wml,run_word_field)
S3method(to_wml,run_wordtext)
S3method(to_wml,section_columns)
S3method(to_wml,table_colwidths)
S3method(to_wml,table_conditional_formatting)
S3method(to_wml,table_layout)
S3method(to_wml,table_stylenames)
S3method(to_wml,table_width)
S3method(update,fp_border)
S3method(update,fp_cell)
S3method(update,fp_par)
S3method(update,fp_text)
S3method(update,fpar)
S3method(update,sp_line)
S3method(update,sp_lineend)
export("layout_rename_ph_labels<-")
export("slide_visible<-")
export(add_sheet)
export(add_slide)
export(annotate_base)
export(as_base64)
export(base64_to_image)
export(block_caption)
export(block_gg)
export(block_list)
export(block_list_items)
export(block_pour_docx)
export(block_section)
export(block_table)
export(block_toc)
export(body_add)
export(body_add_blocks)
export(body_add_break)
export(body_add_caption)
export(body_add_docx)
export(body_add_fpar)
export(body_add_gg)
export(body_add_img)
export(body_add_list)
export(body_add_par)
export(body_add_plot)
export(body_add_table)
export(body_add_toc)
export(body_add_xml)
export(body_append_start_context)
export(body_append_stop_context)
export(body_bookmark)
export(body_comment)
export(body_end_block_section)
export(body_end_section_columns)
export(body_end_section_columns_landscape)
export(body_end_section_continuous)
export(body_end_section_landscape)
export(body_end_section_portrait)
export(body_import_docx)
export(body_remove)
export(body_replace_all_text)
export(body_replace_gg_at_bkm)
export(body_replace_img_at_bkm)
export(body_replace_plot_at_bkm)
export(body_replace_text_at_bkm)
export(body_set_default_section)
export(change_styles)
export(color_scheme)
export(cursor_backward)
export(cursor_begin)
export(cursor_bookmark)
export(cursor_end)
export(cursor_forward)
export(cursor_reach)
export(cursor_reach_index)
export(cursor_reach_test)
export(doc_properties)
export(docx_body_relationship)
export(docx_body_xml)
export(docx_bookmarks)
export(docx_comments)
export(docx_current_block_xml)
export(docx_dim)
export(docx_embed_font)
export(docx_reference_img)
export(docx_set_character_style)
export(docx_set_paragraph_style)
export(docx_set_settings)
export(docx_show_chunk)
export(docx_summary)
export(empty_content)
export(external_img)
export(floating_external_img)
export(footers_replace_all_text)
export(footers_replace_img_at_bkm)
export(footers_replace_text_at_bkm)
export(fortify_location)
export(fp_border)
export(fp_cell)
export(fp_par)
export(fp_par_lite)
export(fp_tab)
export(fp_tabs)
export(fp_text)
export(fp_text_lite)
export(fpar)
export(from_base64)
export(ftext)
export(get_reference_value)
export(headers_replace_all_text)
export(headers_replace_img_at_bkm)
export(headers_replace_text_at_bkm)
export(hyperlink_ftext)
export(image_to_base64)
export(layout_dedupe_ph_labels)
export(layout_default)
export(layout_properties)
export(layout_rename_ph_labels)
export(layout_summary)
export(list_item)
export(media_extract)
export(move_slide)
export(notes_location_label)
export(notes_location_type)
export(officer_url_encode)
export(on_slide)
export(ooxml_chart_uris)
export(open_file)
export(opts_current_table)
export(pack_folder)
export(page_mar)
export(page_size)
export(ph_hyperlink)
export(ph_location)
export(ph_location_fullsize)
export(ph_location_id)
export(ph_location_label)
export(ph_location_left)
export(ph_location_right)
export(ph_location_template)
export(ph_location_type)
export(ph_remove)
export(ph_slidelink)
export(ph_with)
export(ph_with.Date)
export(ph_with.block_list)
export(ph_with.block_list_items)
export(ph_with.character)
export(ph_with.data.frame)
export(ph_with.empty_content)
export(ph_with.external_img)
export(ph_with.factor)
export(ph_with.fpar)
export(ph_with.gg)
export(ph_with.logical)
export(ph_with.numeric)
export(ph_with.plot_instr)
export(ph_with.unordered_list)
export(ph_with.xml_document)
export(phs_with)
export(plot_in_png)
export(plot_instr)
export(plot_layout_properties)
export(pptx_summary)
export(prop_section)
export(prop_table)
export(read_docx)
export(read_pptx)
export(read_xlsx)
export(remove_slide)
export(resolve_location)
export(rtf_add)
export(rtf_doc)
export(rtf_set_paragraph_style)
export(rtf_styles_info)
export(run_autonum)
export(run_bookmark)
export(run_columnbreak)
export(run_comment)
export(run_footnote)
export(run_footnoteref)
export(run_linebreak)
export(run_pagebreak)
export(run_reference)
export(run_tab)
export(run_word_field)
export(run_wordtext)
export(sanitize_images)
export(section_columns)
export(set_autonum_bookmark)
export(set_doc_properties)
export(set_notes)
export(shape_properties_tags)
export(sheet_add_drawing)
export(sheet_names)
export(sheet_remove)
export(sheet_select)
export(sheet_write_data)
export(shortcuts)
export(slide_size)
export(slide_summary)
export(slide_visible)
export(solid_fill)
export(sp_line)
export(sp_lineend)
export(str_encode_to_rtf)
export(styles_info)
export(table_colwidths)
export(table_conditional_formatting)
export(table_layout)
export(table_stylenames)
export(table_width)
export(to_html)
export(to_pml)
export(to_rtf)
export(to_wml)
export(unordered_list)
export(unpack_folder)
export(uuid_generate)
export(wml_link_images)
export(write_elements_to_context)
export(xlsx_drawing)
export(xlsx_styles)
import(graphics)
importFrom(R6,R6Class)
importFrom(cli,cli_abort)
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,arrange)
importFrom(dplyr,as_tibble)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,consecutive_id)
importFrom(dplyr,filter)
importFrom(dplyr,first)
importFrom(dplyr,group_by)
importFrom(dplyr,if_else)
importFrom(dplyr,inner_join)
importFrom(dplyr,lag)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,relocate)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,syms)
importFrom(grDevices,col2rgb)
importFrom(grDevices,dev.off)
importFrom(grDevices,rgb)
importFrom(graphics,box)
importFrom(graphics,plot)
importFrom(graphics,rect)
importFrom(graphics,text)
importFrom(openssl,base64_decode)
importFrom(openssl,base64_encode)
importFrom(openssl,sha1)
importFrom(ragg,agg_png)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(tidyr,complete)
importFrom(utils,compareVersion)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,packageVersion)
importFrom(utils,str)
importFrom(utils,tail)
importFrom(uuid,UUIDgenerate)
importFrom(xml2,"xml_attr<-")
importFrom(xml2,"xml_name<-")
importFrom(xml2,"xml_text<-")
importFrom(xml2,as_list)
importFrom(xml2,as_xml_document)
importFrom(xml2,read_xml)
importFrom(xml2,write_xml)
importFrom(xml2,xml_add_child)
importFrom(xml2,xml_add_parent)
importFrom(xml2,xml_add_sibling)
importFrom(xml2,xml_attr)
importFrom(xml2,xml_attrs)
importFrom(xml2,xml_child)
importFrom(xml2,xml_children)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_find_first)
importFrom(xml2,xml_has_attr)
importFrom(xml2,xml_length)
importFrom(xml2,xml_missing)
importFrom(xml2,xml_name)
importFrom(xml2,xml_ns)
importFrom(xml2,xml_parent)
importFrom(xml2,xml_path)
importFrom(xml2,xml_remove)
importFrom(xml2,xml_replace)
importFrom(xml2,xml_set_attr)
importFrom(xml2,xml_set_attrs)
importFrom(xml2,xml_set_text)
importFrom(xml2,xml_text)
importFrom(zip,unzip)
importFrom(zip,zip)

officer R package
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

> Make corporate reporting with minimum hassle

[![R build
status](https://github.com/davidgohel/officer/workflows/R-CMD-check/badge.svg)](https://github.com/davidgohel/officer/actions)
[![version](https://www.r-pkg.org/badges/version/officer)](https://CRAN.R-project.org/package=officer)
[![codecov test
coverage](https://codecov.io/gh/davidgohel/officer/branch/main/graph/badge.svg)](https://app.codecov.io/gh/davidgohel/officer)

The officer package lets R users manipulate Word (`.docx`) and
PowerPoint (`*.pptx`) documents. In short, one can add images, tables
and text into documents from R. An initial document can be provided;
contents, styles and properties of the original document will then be
available. It also supports the writing of ‘RTF’ documents and basic
manipulation of Excel (‘.xlsx’) workbooks.

## Ressources

The help pages are in a bookdown located at:

<https://ardata-fr.github.io/officeverse/>

Manuals are available at:

<https://davidgohel.github.io/officer/>.

## Word documents

The `read_docx()` function will read an initial Word document (an empty
one by default) and lets you modify its content later.

The package provides functions to add R outputs into a Word document:

- images: produce your plot in png or emf files and add them into the
  document, as a whole paragraph or inside a paragraph.
- tables: add data.frames as tables, format is defined by the associated
  Word table style.
- text: add text as paragraphs or inside an existing paragraph, format
  is defined by the associated Word paragraph and text styles.
- field codes: add Word field codes inside paragraphs. Field codes is an
  old feature of MS Word to create calculated elements such as tables of
  contents, automatic numbering and hyperlinks.

File generation is performed with the `print` function.

### import Word document in a data.frame

The function `docx_summary()` reads and imports content of a Word
document into a data.frame. The function handles paragraphs, tables and
section breaks. The function `docx_comments()` reads comments of a Word
document and organise the results into a data.frame.

## PowerPoint documents

The function `read_pptx()` will read an initial PowerPoint document (an
empty one by default) and let you modify its content later.

The package provides functions to add R outputs into existing or new
PowerPoint slides:

- images: produce your plot in png or emf files and add them in a slide.
- tables: add data.frames as tables, format is defined by the associated
  PowerPoint table style.
- text: add text as paragraphs or inside an existing paragraph, format
  is defined in the corresponding layout of the slide.

In a PowerPoint document, one can set a slide as selected and reach a
particular shape (and remove it or add text).

File generation is performed with the `print()` function.

### import PowerPoint document in a data.frame

The `pptx_summary()` function reads and imports content of a PowerPoint
document into a data.frame. The function handles paragraphs, tables and
images.

## Extensions

### Tables and package `flextable`

The package
[flextable](https://ardata-fr.github.io/flextable-book/index.html)
brings a full API to produce nice tables and use them with packages
officer and rmarkdown.

### Vector graphics with package `rvg`

The package [rvg](https://github.com/davidgohel/rvg) brings an API to
produce nice vector graphics that can be embedded in PowerPoint
documents or Excel workbooks with `officer`.

### Native office charts with package `mschart`

The package [mschart](https://github.com/ardata-fr/mschart) combined
with `officer` can produce native office charts in PowerPoint and Word
documents.

### Advance Word documents with R Markdown with package `officedown`

The package [officedown](https://ardata-fr.github.io/officeverse/)
facilitates the formatting of Microsoft Word documents produced by R
Markdown documents.

## Installation

You can get the development version from GitHub:

``` r
devtools::install_github("davidgohel/officer")
```

Or the latest version on CRAN:

``` r
install.packages("officer")
```

## Getting help

If you have questions about how to use the package, visit Stack
Overflow’s `officer` tag and post your question there. I usually read
them and answer when possible.

## Contributing to the package

### Code of Conduct

Anyone getting involved in this package agrees to our [Code of
Conduct](https://github.com/davidgohel/officer/blob/main/CONDUCT.md).

### Bug reports

Bugs are reported through [GitHub
issues](https://github.com/davidgohel/officer/issues/new/choose). The
issue form requires a reproducible example (a
[reprex](https://reprex.tidyverse.org/)) and the output of
`sessionInfo()`; reports missing these elements are closed without
further response. The more reproducible your report, the more time can
go into investigating the bug rather than reconstructing it.

For usage questions (“how do I do X with officer”), please read
<https://ardata-fr.github.io/officeverse/> or
<https://davidgohel.github.io/officer/> or use [Stack
Overflow](https://stackoverflow.com/questions/tagged/officer) with the
`officer` tag instead.

### Pull requests

Before writing any code, please open an issue describing the change you
have in mind and the reason for it. The proposal needs to be discussed
and explicitly validated by the maintainer before work starts; pull
requests for features that have not gone through this step will not be
considered, no matter how well crafted. This isn’t meant to discourage
contributions, only to make sure the time you invest goes into a change
that will actually land.

Pull requests are accepted on a case-by-case basis: once a proposal has
been agreed in an issue, the maintainer will invite you to open a PR for
that specific change. Unsolicited PRs that bypass this step will be
closed.

This policy keeps the package coherent: officer’s internals are shared
with several downstream packages (flextable, officedown, rvg), and
changes that look local often have wider implications. Aligning on the
design in an issue first saves time on both sides.

When invited to open a PR, please include:

- the new function(s) with code and roxygen tags, including runnable
  examples
- documentation updates (vignette section if relevant)
- corresponding tests under `tests/testthat`

Run `rhub::check_for_cran()` locally before submitting; CI will run the
full check suite on the PR.

officer R package
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

> Make corporate reporting with minimum hassle

[![R build
status](https://github.com/davidgohel/officer/workflows/R-CMD-check/badge.svg)](https://github.com/davidgohel/officer/actions)
[![version](https://www.r-pkg.org/badges/version/officer)](https://CRAN.R-project.org/package=officer)
[![codecov test
coverage](https://codecov.io/gh/davidgohel/officer/branch/master/graph/badge.svg)](https://app.codecov.io/gh/davidgohel/officer)

The officer package lets R users manipulate Word (`.docx`) and
PowerPoint (`*.pptx`) documents. In short, one can add images, tables
and text into documents from R. An initial document can be provided;
contents, styles and properties of the original document will then be
available. It also supports the writing of ‘RTF’ documents.

## Ressources

The help pages are in a bookdown located at:

<https://ardata-fr.github.io/officeverse/>

Manuals are available at:

<https://davidgohel.github.io/officer/>.

## Word documents

The `read_docx()` function will read an initial Word document (an empty
one by default) and lets you modify its content later.

The package provides functions to add R outputs into a Word document:

- images: produce your plot in png or emf files and add them into the
  document, as a whole paragraph or inside a paragraph.
- tables: add data.frames as tables, format is defined by the associated
  Word table style.
- text: add text as paragraphs or inside an existing paragraph, format
  is defined by the associated Word paragraph and text styles.
- field codes: add Word field codes inside paragraphs. Field codes is an
  old feature of MS Word to create calculated elements such as tables of
  contents, automatic numbering and hyperlinks.

File generation is performed with the `print` function.

### import Word document in a data.frame

The function `docx_summary()` reads and imports content of a Word
document into a data.frame. The function handles paragraphs, tables and
section breaks. The function `docx_comments()` reads comments of a Word
document and organise the results into a data.frame.

## PowerPoint documents

The function `read_pptx()` will read an initial PowerPoint document (an
empty one by default) and let you modify its content later.

The package provides functions to add R outputs into existing or new
PowerPoint slides:

- images: produce your plot in png or emf files and add them in a slide.
- tables: add data.frames as tables, format is defined by the associated
  PowerPoint table style.
- text: add text as paragraphs or inside an existing paragraph, format
  is defined in the corresponding layout of the slide.

In a PowerPoint document, one can set a slide as selected and reach a
particular shape (and remove it or add text).

File generation is performed with the `print()` function.

### import PowerPoint document in a data.frame

The `pptx_summary()` function reads and imports content of a PowerPoint
document into a data.frame. The function handles paragraphs, tables and
images.

## Extensions

### Tables and package `flextable`

The package
[flextable](https://ardata-fr.github.io/flextable-book/index.html)
brings a full API to produce nice tables and use them with packages
officer and rmarkdown.

### Vector graphics with package `rvg`

The package [rvg](https://github.com/davidgohel/rvg) brings an API to
produce nice vector graphics that can be embedded in PowerPoint
documents or Excel workbooks with `officer`.

### Native office charts with package `mschart`

The package [mschart](https://github.com/ardata-fr/mschart) combined
with `officer` can produce native office charts in PowerPoint and Word
documents.

### Advance Word documents with R Markdown with package `officedown`

The package [officedown](https://ardata-fr.github.io/officeverse/)
facilitates the formatting of Microsoft Word documents produced by R
Markdown documents.

## Installation

You can get the development version from GitHub:

``` r
devtools::install_github("davidgohel/officer")
```

Or the latest version on CRAN:

``` r
install.packages("officer")
```

## Getting help

If you have questions about how to use the package, visit Stack
Overflow’s `officer` tag and post your question there. I usually read
them and answer when possible.

## Contributing to the package

### Code of Conduct

Anyone getting involved in this package agrees to our [Code of
Conduct](https://github.com/davidgohel/officer/blob/master/CONDUCT.md).

### Bug reports

When you file a [bug
report](https://github.com/davidgohel/officer/discussions/categories/q-a),
please spend some time making it easy for me to follow and reproduce.
The more time you spend on making the bug report coherent, the more time
I can dedicate to investigate the bug as opposed to the bug report.

### Contributing to the package development

A great way to start is to contribute an example or improve the
documentation.

If you want to submit a Pull Request to integrate functions of yours,
please provide:

- the new function(s) with code and roxygen tags (with examples)
- a new section in the appropriate vignette that describes how to use
  the new function
- add corresponding tests in directory `tests/testthat`.

By using rhub (run `rhub::check_for_cran()`), you will see if everything
is ok. When submitted, the PR will be evaluated automatically on travis
and appveyor and you will be able to see if something broke.

#' @export
#' @title Add continuous section
#' @description Section break starts the new section on the same page. This
#' type of section break is often used to change the number of columns
#' without starting a new page.
#' @param x an rdocx object
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#' str2 <- "Aenean venenatis varius elit et fermentum vivamus vehicula."
#' str2 <- rep(str2, 5)
#' str2 <- paste(str2, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = "Default section", style = "heading 1")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str2, style = "Normal")
#' doc_1 <- body_end_section_continuous(doc_1)
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_continuous <- function(x) {
  bs <- block_section(prop_section(type = "continuous"))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}

#' @export
#' @title Add landscape section
#' @description A section with landscape orientation is added to the document.
#' @param x an rdocx object
#' @param w,h page width, page height (in inches)
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_landscape(doc_1)
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_landscape <- function(x, w = 16838 / 1440, h = 11906 / 1440) {
  bs <- block_section(prop_section(
    page_size = page_size(width = w, height = h, orient = "landscape"),
    type = "oddPage"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}

#' @export
#' @title Add portrait section
#' @description A section with portrait orientation is added to the document.
#' @param x an rdocx object
#' @param w,h page width, page height (in inches)
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_portrait(doc_1)
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_portrait <- function(x, w = 16838 / 1440, h = 11906 / 1440) {
  bs <- block_section(prop_section(
    page_size = page_size(width = w, height = h, orient = "portrait"),
    type = "oddPage"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}

#' @export
#' @title Add multi columns section
#' @description A section with multiple columns is added to the
#' document.
#'
#' You may prefer to use [body_end_block_section()] that is
#' more flexible.
#' @param x an rdocx object
#' @inheritParams section_columns
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_columns(doc_1)
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_columns <- function(
  x,
  widths = c(2.5, 2.5),
  space = .25,
  sep = FALSE
) {
  bs <- block_section(prop_section(
    section_columns = section_columns(
      widths = widths,
      space = space,
      sep = sep
    ),
    type = "continuous"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}


#' @export
#' @title Add a landscape multi columns section
#' @description A landscape section with multiple columns is added to the document.
#' @param x an rdocx object
#' @inheritParams section_columns
#' @param w,h page width, page height (in inches)
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_columns_landscape(doc_1, widths = c(6, 2))
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_columns_landscape <- function(
  x,
  widths = c(2.5, 2.5),
  space = .25,
  sep = FALSE,
  w = 16838 / 1440,
  h = 11906 / 1440
) {
  bs <- block_section(prop_section(
    page_size = page_size(width = w, height = h, orient = "landscape"),
    section_columns = section_columns(
      widths = widths,
      space = space,
      sep = sep
    ),
    type = "oddPage"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}
#' @export
#' @title Add any section
#' @description Add a section to the document. You can
#' define any section with a [block_section] object. All other
#' `body_end_section_*` are specialized, this one is highly flexible
#' but it's up to the user to define the section properties.
#' @param x an rdocx object
#' @param value a [block_section] object
#' @examples
#' library(officer)
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 20)
#' str1 <- paste(str1, collapse = " ")
#'
#' ps <- prop_section(
#'   page_size = page_size(orient = "landscape"),
#'   page_margins = page_mar(top = 2),
#'   type = "continuous"
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#'
#' doc_1 <- body_end_block_section(doc_1, block_section(ps))
#'
#' doc_1 <- body_add_par(doc_1, value = str1, style = "centered")
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
#' @section Illustrations:
#'
#' \if{html}{\figure{body_end_block_section_doc_1.png}{options: style="width:80\%;"}}
body_end_block_section <- function(x, value) {
  stopifnot(inherits(value, "block_section"))
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml(x = x, str = xml_elt, pos = "after")
}

#' @export
#' @title Define Default Section
#' @description Define default section of the document. You can
#' define section propeerties (page size, orientation, ...) with a [prop_section] object.
#' @param x an rdocx object
#' @param value a [prop_section] object
#' @family functions for Word sections
#' @example inst/examples/example_body_set_default_section.R
#' @section Illustrations:
#'
#' \if{html}{\figure{body_set_default_section_doc_1.png}{options: style="width:80\%;"}}
body_set_default_section <- function(x, value) {
  stopifnot(inherits(value, "prop_section"))
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  z <- xml_find_first(docx_body_xml(x), "w:body/w:sectPr")
  xml_replace(z, as_xml_document(xml_elt))

  x
}


# utils ----
process_sections_content <- function(x, xml_str) {
  hof_summary <- extract_hof(x, xml_str)

  relationships <- docx_body_relationship(x)

  for (i in seq_len(nrow(hof_summary))) {
    # write file
    xml_doc <- as_xml_document(hof_summary[[i, "content_section"]])
    write_xml(
      xml_doc,
      file = file.path(x$package_dir, "word", hof_summary[[i, "file_name"]])
    )

    relationships$add(
      id = hof_summary[[i, "rid"]],
      type = paste0(
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/",
        hof_summary[[i, "type"]]
      ),
      target = hof_summary[[i, "file_name"]]
    )
    x$content_type$add_override(
      setNames(
        paste0(
          "application/vnd.openxmlformats-officedocument.wordprocessingml.",
          hof_summary[[i, "type"]],
          "+xml"
        ),
        paste0("/word/", hof_summary[[i, "file_name"]])
      )
    )
  }

  xml_str[hof_summary$reference_position] <- sprintf(
    xml_str[hof_summary$reference_position],
    hof_summary[["rid"]]
  )
  # xml_str <- gsub(" w:officer=\"true\"",  "", xml_str)

  # drop added sections contents from document xml
  lines_index_to_drop <- mapply(
    function(start, stop, str) {
      start:stop
    },
    start = hof_summary$start,
    stop = hof_summary$stop,
    SIMPLIFY = FALSE
  )
  lines_index_to_drop <- unlist(lines_index_to_drop)
  if (length(lines_index_to_drop) > 0) {
    xml_str <- xml_str[-lines_index_to_drop]
  }

  xml_str
}

guess_and_set_even_and_odd_headers <- function(x, xml_str) {
  test_even <- any(grepl(
    "(headerReference|footerReference)([^>]+)(w:type=\"even\")",
    xml_str
  ))
  if (test_even) {
    x$settings$even_and_odd_headers <- TRUE
  } else {
    x$settings$even_and_odd_headers <- FALSE
  }
  x
}

section_dimensions <- function(node) {
  section_obj <- as_list(node)

  landscape <- FALSE
  if (
    !is.null(attr(section_obj$pgSz, "orient")) &&
      attr(section_obj$pgSz, "orient") == "landscape"
  ) {
    landscape <- TRUE
  }

  h_ref <- as.integer(attr(section_obj$pgSz, "h"))
  w_ref <- as.integer(attr(section_obj$pgSz, "w"))

  mar_t <- as.integer(attr(section_obj$pgMar, "top"))
  mar_b <- as.integer(attr(section_obj$pgMar, "bottom"))
  mar_r <- as.integer(attr(section_obj$pgMar, "right"))
  mar_l <- as.integer(attr(section_obj$pgMar, "left"))
  mar_h <- as.integer(attr(section_obj$pgMar, "header"))
  mar_f <- as.integer(attr(section_obj$pgMar, "footer"))

  list(
    page = c("width" = w_ref, "height" = h_ref),
    landscape = landscape,
    margins = c(
      top = mar_t,
      bottom = mar_b,
      left = mar_l,
      right = mar_r,
      header = mar_h,
      footer = mar_f
    )
  )
}

hof_next_file_index <- function(x, type = "header") {
  pattern <- paste0("^(", type, ")([0-9]+)(\\.xml)$")
  files <- list.files(
    path = file.path(x$package_dir, "word"),
    pattern = pattern
  )
  empty_index <- 0L
  m <- regexpr("[0-9]+", files)
  index_chars <- regmatches(files, m)
  index_ints <- as.integer(index_chars)
  index <- max(c(empty_index, index_ints), na.rm = TRUE)
  index + 1L
}


extract_hof <- function(x, xml_str) {
  # init next_rel_id that will be the first rid
  relationships <- docx_body_relationship(x)
  next_rel_id <- relationships$get_next_id()

  next_footer_file_index <- hof_next_file_index(x, type = "footer")
  next_header_file_index <- hof_next_file_index(x, type = "header")

  # footers extraction
  starts <- grep("<w:ftr", xml_str, fixed = TRUE)
  stops <- grep("</w:ftr>", xml_str, fixed = TRUE)

  content_section_chrs <- mapply(
    function(start, stop, str) {
      z <- str[start:stop]
      paste(z, collapse = "")
    },
    start = starts,
    stop = stops,
    MoreArgs = list(str = xml_str),
    SIMPLIFY = FALSE
  )
  content_section_chrs <- unlist(content_section_chrs)
  file_indexes <- seq(
    from = next_footer_file_index,
    along.with = content_section_chrs
  )
  file_names <- sprintf("%s%.0f.xml", "footer", file_indexes)

  rids <- sprintf(
    "rId%.0f",
    seq(from = next_rel_id, along.with = content_section_chrs)
  )

  reference_positions <- grep(
    "<w:footerReference w:officer",
    xml_str,
    fixed = TRUE
  )

  footer_sections <- data.frame(
    content_section = content_section_chrs,
    file_name = file_names,
    rid = rids,
    type = rep("footer", length(rids)),
    start = starts,
    stop = stops,
    reference_position = reference_positions,
    stringsAsFactors = FALSE
  )

  # headers extraction
  starts <- grep("<w:hdr", xml_str, fixed = TRUE)
  stops <- grep("</w:hdr>", xml_str, fixed = TRUE)

  content_section_chrs <- mapply(
    function(start, stop, str) {
      z <- str[start:stop]
      paste(z, collapse = "")
    },
    start = starts,
    stop = stops,
    MoreArgs = list(str = xml_str),
    SIMPLIFY = FALSE
  )
  content_section_chrs <- unlist(content_section_chrs)
  file_indexes <- seq(
    from = next_header_file_index,
    along.with = content_section_chrs
  )
  file_names <- sprintf("%s%.0f.xml", "header", file_indexes)

  rids <- sprintf(
    "rId%.0f",
    seq(
      from = next_rel_id + nrow(footer_sections),
      along.with = content_section_chrs
    )
  )

  reference_positions <- grep(
    "<w:headerReference w:officer",
    xml_str,
    fixed = TRUE
  )

  header_sections <- data.frame(
    content_section = content_section_chrs,
    file_name = file_names,
    rid = rids,
    type = rep("header", length(rids)),
    start = starts,
    stop = stops,
    reference_position = reference_positions,
    stringsAsFactors = FALSE
  )

  rbind(header_sections, footer_sections)
}


copy_header_references_everywhere <- function(x, xml_str) {
  xml_body <- docx_body_xml(x)

  all_ref <- xml_find_all(
    xml_body,
    "/w:document/w:body/w:sectPr/w:headerReference"
  )
  all_ref <- as.character(all_ref)
  all_ref <- paste0(all_ref, collapse = "")
  m <- regexpr("(?=</w:sectPr)", xml_str, perl = TRUE)

  default_sect_pos <- tail(which(m > -1), n = 1)
  attr(m, "match.length")[default_sect_pos] <- -1
  m[default_sect_pos] <- -1

  regmatches(xml_str, m) <- all_ref

  xml_str
}

copy_footer_references_everywhere <- function(x, xml_str) {
  xml_body <- docx_body_xml(x)

  all_ref <- xml_find_all(
    xml_body,
    "/w:document/w:body/w:sectPr/w:footerReference"
  )
  all_ref <- as.character(all_ref)
  all_ref <- paste0(all_ref, collapse = "")
  m <- regexpr("(?=</w:sectPr)", xml_str, perl = TRUE)

  default_sect_pos <- tail(which(m > -1), n = 1)
  attr(m, "match.length")[default_sect_pos] <- -1
  m[default_sect_pos] <- -1

  regmatches(xml_str, m) <- all_ref

  xml_str
}

#' @export
#' @title Read 'Word' styles
#' @description Read Word styles and get results in
#' a data.frame.
#' @param x an rdocx object
#' @param type,is_default subsets for types (i.e. paragraph) and
#' default style (when `is_default` is TRUE or FALSE)
#' @examples
#' x <- read_docx()
#' styles_info(x)
#' styles_info(x, type = "paragraph", is_default = TRUE)
#' @family functions for Word document informations
styles_info <- function(
  x,
  type = c("paragraph", "character", "table", "numbering"),
  is_default = c(TRUE, FALSE)
) {
  styles <- x$styles
  styles <- styles[
    styles$style_type %in% type & styles$is_default %in% is_default,
  ]
  styles
}


#' @export
#' @title Add or replace paragraph style in a Word document
#' @description The function lets you add or replace a Word paragraph style.
#' @param x an rdocx object
#' @param style_id a unique style identifier for Word.
#' @param style_name a unique label associated with the style identifier.
#' This label is the name of the style when Word edit the document.
#' @param base_on the style name used as base style
#' @param fp_p paragraph formatting properties, see [fp_par()].
#' @param fp_t default text formatting properties. This is used as
#' text formatting properties, see [fp_text()]. If NULL (default), the
#' paragraph will used the default text formatting properties (defined by
#' the `base_on` argument).
#' @examples
#' library(officer)
#'
#' doc <- read_docx()
#'
#' doc <- docx_set_paragraph_style(
#'   doc,
#'   style_id = "rightaligned",
#'   style_name = "Explicit label",
#'   fp_p = fp_par(text.align = "right", padding = 20),
#'   fp_t = fp_text_lite(
#'     bold = TRUE,
#'     shading.color = "#FD34F0",
#'     color = "white")
#' )
#'
#' doc <- body_add_par(doc,
#'   value = "This is a test",
#'   style = "Explicit label")
#'
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_file
docx_set_paragraph_style <- function(
  x,
  style_id,
  style_name,
  base_on = "Normal",
  fp_p = fp_par(),
  fp_t = NULL
) {
  styles_file <- file.path(x$package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)

  if (grepl("[^a-zA-Z0-9\\-]+", style_id)) {
    stop("`style_id` should only contain '-', numbers and ascii characters.")
  }

  node_styles <- xml_find_first(doc, "/w:styles")

  fp_p$word_style <- NULL

  if (!is.null(fp_t)) {
    fp_t_xml <- rpr_wml(fp_t)
  } else {
    fp_t_xml <- ""
  }
  base_on <- get_style_id(data = x$styles, style = base_on, type = "paragraph")

  xml_code <- paste0(
    sprintf(
      "<w:style xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:type=\"paragraph\" w:customStyle=\"1\" w:styleId=\"%s\">",
      style_id
    ),
    sprintf("<w:name w:val=\"%s\"/>", style_name),
    sprintf("<w:basedOn w:val=\"%s\"/>", base_on),
    ppr_wml(fp_p),
    fp_t_xml,
    "</w:style>"
  )

  node_style <- xml_child(
    node_styles,
    sprintf("w:style[@w:styleId='%s']", style_id)
  )
  if (inherits(node_style, "xml_missing")) {
    xml_add_child(node_styles, as_xml_document(xml_code))
  } else {
    xml_replace(node_style, as_xml_document(xml_code))
  }

  write_xml(doc, file = styles_file)
  styles <- read_docx_styles(x$package_dir)
  x$styles <- styles

  x
}

#' @export
#' @title Add character style in a Word document
#' @description The function lets you add or modify Word character styles.
#' @param x an rdocx object
#' @param style_id a unique style identifier for Word.
#' @param style_name a unique label associated with the style identifier.
#' This label is the name of the style when Word edit the document.
#' @param base_on the character style name used as base style
#' @param fp_t Text formatting properties, see [fp_text()].
#' @examples
#' library(officer)
#' doc <- read_docx()
#'
#' doc <- docx_set_character_style(
#'   doc,
#'   style_id = "newcharstyle",
#'   style_name = "label for char style",
#'   base_on = "Default Paragraph Font",
#'   fp_text_lite(
#'     shading.color = "red",
#'     color = "white")
#' )
#' paragraph <- fpar(
#'   run_wordtext("hello",
#'     style_id = "newcharstyle"))
#'
#' doc <- body_add_fpar(doc, value = paragraph)
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_file
docx_set_character_style <- function(
  x,
  style_id,
  style_name,
  base_on,
  fp_t = fp_text_lite()
) {
  styles_file <- file.path(x$package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)
  node_styles <- xml_find_first(doc, "/w:styles")

  if (grepl("[^a-zA-Z0-9\\-]+", style_id)) {
    stop("`style_id` should only contain '-', numbers and ascii characters.")
  }

  base_on <- get_style_id(data = x$styles, style = base_on, type = "character")

  xml_code <- paste0(
    sprintf(
      "<w:style xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:type=\"character\" w:customStyle=\"1\" w:styleId=\"%s\">",
      style_id
    ),
    sprintf("<w:name w:val=\"%s\"/>", style_name),
    sprintf("<w:basedOn w:val=\"%s\"/>", base_on),
    rpr_wml(fp_t),
    "</w:style>"
  )

  node_style <- xml_child(
    node_styles,
    sprintf("w:style[@w:styleId='%s']", style_id)
  )
  if (inherits(node_style, "xml_missing")) {
    xml_add_child(node_styles, as_xml_document(xml_code))
  } else {
    xml_replace(node_style, as_xml_document(xml_code))
  }

  write_xml(doc, file = styles_file)
  styles <- read_docx_styles(x$package_dir)
  x$styles <- styles

  x
}


#' @export
#' @title Replace styles in a 'Word' Document
#' @description Replace styles with others in a 'Word' document. This function
#' can be used for paragraph, run/character and table styles.
#' @param x an rdocx object
#' @param mapstyles a named list, names are the replacement style,
#' content (as a character vector) are the styles to be replaced.
#' Use [styles_info()] to display available styles.
#' @examples
#' # creating a sample docx so that we can illustrate how
#' # to change styles
#' doc_1 <- read_docx()
#'
#' doc_1 <- body_add_par(doc_1, "A title", style = "heading 1")
#' doc_1 <- body_add_par(doc_1, "Another title", style = "heading 2")
#' doc_1 <- body_add_par(doc_1, "Hello world!", style = "Normal")
#' file <- print(doc_1, target = tempfile(fileext = ".docx"))
#'
#' # now we can illustrate how
#' # to change styles with `change_styles`
#' doc_2 <- read_docx(path = file)
#' mapstyles <- list(
#'   "centered" = c("Normal", "heading 2"),
#'   "strong" = "Default Paragraph Font"
#' )
#' doc_2 <- change_styles(doc_2, mapstyles = mapstyles)
#' print(doc_2, target = tempfile(fileext = ".docx"))
change_styles <- function(x, mapstyles) {
  if (is.null(mapstyles) || length(mapstyles) < 1) {
    return(x)
  }

  table_styles <- styles_info(x, type = c("paragraph", "character", "table"))

  from_styles <- unique(as.character(unlist(mapstyles)))
  to_styles <- unique(names(mapstyles))

  if (any(is.na(mfrom <- match(from_styles, table_styles$style_name)))) {
    stop(
      "could not find style ",
      paste0(shQuote(from_styles[is.na(mfrom)]), collapse = ", "),
      ".",
      call. = FALSE
    )
  }
  if (any(is.na(mto <- match(to_styles, table_styles$style_name)))) {
    stop(
      "could not find style ",
      paste0(shQuote(to_styles[is.na(mto)]), collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  mapping <- mapply(
    function(from, to) {
      id_to <- which(table_styles$style_name %in% to)
      id_to <- table_styles$style_id[id_to]

      id_from <- which(table_styles$style_name %in% from)
      types <- substring(table_styles$style_type[id_from], first = 1, last = 1)
      types[types %in% "c"] <- "r"
      types[types %in% "t"] <- "tbl"
      id_from <- table_styles$style_id[id_from]

      data.frame(
        from = id_from,
        to = rep(id_to, length(from)),
        types = types,
        stringsAsFactors = FALSE
      )
    },
    mapstyles,
    names(mapstyles),
    SIMPLIFY = FALSE
  )

  mapping <- do.call(rbind, mapping)
  row.names(mapping) <- NULL

  for (i in seq_len(nrow(mapping))) {
    all_nodes <- xml_find_all(
      x$doc_obj$get(),
      sprintf("//w:%sStyle[@w:val='%s']", mapping$types[i], mapping$from[i])
    )
    xml_attr(all_nodes, "w:val") <- rep(mapping$to[i], length(all_nodes))
  }

  x
}

#' @export
#' @title  Write a 'Word' File
#' @description
#' `print.rdocx()` is the essential output function for creating Word files
#' with officer. It takes an `rdocx` object (created with [read_docx()] and
#' populated with content) and writes it to disk as a `.docx` file.
#'
#' This function performs all necessary post-processing operations before
#' writing the file.
#'
#' The function is typically called at the end of your document creation
#' workflow, after all content has been added with `body_add_*()` functions.
#'
#' @param x an `rdocx` object created with [read_docx()]
#' @param target path to the `.docx` file to write. The file will be created
#' or overwritten if it already exists. If `NULL` and `preview = FALSE`, the
#' function returns `NULL` without writing a file.
#' @param copy_header_refs,copy_footer_refs logical, default is FALSE.
#' If TRUE, copy the references to the header and footer in each section
#' of the body of the document. This parameter is experimental and may change
#' in a future version.
#' @param preview Save `x` to a temporary file and open it (default `FALSE`).
#' When `TRUE`, the document is saved to a temporary location and opened with
#' the system's default application for `.docx` files, useful for quick previewing
#' during development.
#' @param ... unused
#' @return The full path to the created `.docx` file (invisibly). This allows
#' chaining operations or capturing the output path for further use.
#' @example inst/examples/example_print_docx.R
#' @seealso Create a 'Word' document object with [read_docx()], add content with
#' functions [body_add_par()], [body_add_plot()],
#' [body_add_table()], change settings with [docx_set_settings()], set
#' properties with [set_doc_properties()],  read 'Word' styles with
#' [styles_info()].
print.rdocx <- function(
  x,
  target = NULL,
  copy_header_refs = FALSE,
  copy_footer_refs = FALSE,
  preview = FALSE,
  ...
) {
  if (preview) {
    file <- tempfile(fileext = ".docx")
    print.rdocx(
      x,
      target = file,
      copy_header_refs = copy_header_refs,
      copy_footer_refs = copy_footer_refs,
      preview = FALSE,
      ...
    )
    open_file(file)
    return(invisible(file))
  }
  if (is.null(target)) {
    cat("rdocx document with", length(x), "element(s)\n")
    cat("\n* styles:\n")

    style_names <- styles_info(x)
    style_sample <- style_names$style_type
    names(style_sample) <- style_names$style_name
    print(style_sample)
    return(invisible())
  }

  if (!grepl(x = target, pattern = "\\.(docx)$", ignore.case = TRUE)) {
    stop(target, " should have '.docx' extension.")
  }

  if (is_windows() && is_doc_open(target)) {
    stop(target, " is open. To write to this document, please, close it.")
  }

  # write the xml to a tempfile in a formatted way so that grepl is easy
  xml_str <- xml_document_to_chrs(x$doc_obj$get())
  xml_str <- process_sections_content(x, xml_str)
  xml_str <- process_footnotes_content(
    x$doc_obj,
    footnotes = x$footnotes,
    xml_str
  )
  xml_str <- process_comments_content(x$doc_obj, comments = x$comments, xml_str)
  xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
  xml_str <- fix_hyperlink_refs_in_wml(xml_str, x$doc_obj)
  xml_str <- fix_img_refs_in_wml(
    xml_str,
    x$doc_obj,
    x$doc_obj$relationship(),
    x$package_dir
  )
  xml_str <- fix_svg_refs_in_wml(
    xml_str,
    x$doc_obj,
    x$doc_obj$relationship(),
    x$package_dir
  )
  # make all id unique for document
  xml_str <- fix_empty_ids_in_wml(xml_str)
  if (copy_header_refs) {
    xml_str <- copy_header_references_everywhere(x, xml_str = xml_str)
  }
  if (copy_footer_refs) {
    xml_str <- copy_footer_references_everywhere(x, xml_str = xml_str)
  }

  x <- replace_xml_body_from_chr(x = x, xml_str = xml_str)

  process_docx_poured(
    doc_obj = x$doc_obj,
    relationships = x$doc_obj$relationship(),
    content_type = x$content_type,
    package_dir = x$package_dir
  )

  x$headers <- update_hf_list(
    part_list = x$headers,
    type = "header",
    package_dir = x$package_dir
  )
  x$footers <- update_hf_list(
    part_list = x$footers,
    type = "footer",
    package_dir = x$package_dir
  )

  for (header in x$headers) {
    xml_str <- xml_document_to_chrs(header$get())
    xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
    xml_str <- fix_empty_ids_in_wml(xml_str)
    xml_str <- fix_hyperlink_refs_in_wml(xml_str, header)
    xml_str <- fix_img_refs_in_wml(
      xml_str,
      header,
      header$relationship(),
      x$package_dir
    )
    xml_str <- fix_svg_refs_in_wml(
      xml_str,
      header,
      header$relationship(),
      x$package_dir
    )
    tf_xml <- tempfile(fileext = ".txt")
    writeLines(xml_str, tf_xml, useBytes = TRUE)
    header$replace_xml(tf_xml)
  }
  for (footer in x$footers) {
    xml_str <- xml_document_to_chrs(footer$get())
    xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
    xml_str <- fix_empty_ids_in_wml(xml_str)
    xml_str <- fix_hyperlink_refs_in_wml(xml_str, footer)
    xml_str <- fix_img_refs_in_wml(
      xml_str,
      footer,
      footer$relationship(),
      x$package_dir
    )
    xml_str <- fix_svg_refs_in_wml(
      xml_str,
      footer,
      footer$relationship(),
      x$package_dir
    )

    tf_xml <- tempfile(fileext = ".txt")
    writeLines(xml_str, tf_xml, useBytes = TRUE)
    footer$replace_xml(tf_xml)
  }
  if (TRUE) {
    xml_str <- xml_document_to_chrs(x$footnotes$get())
    xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
    xml_str <- fix_empty_ids_in_wml(xml_str)
    xml_str <- fix_hyperlink_refs_in_wml(xml_str, x$footnotes)
    xml_str <- fix_img_refs_in_wml(
      xml_str,
      x$footnotes,
      x$footnotes$relationship(),
      x$package_dir
    )
    xml_str <- fix_svg_refs_in_wml(
      xml_str,
      x$footnotes,
      x$footnotes$relationship(),
      x$package_dir
    )

    tf_xml <- tempfile(fileext = ".txt")
    writeLines(xml_str, tf_xml, useBytes = TRUE)
    x$footnotes$replace_xml(tf_xml)
  }
  if (TRUE) {
    xml_str <- xml_document_to_chrs(x$comments$get())
    xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
    xml_str <- fix_empty_ids_in_wml(xml_str)
    xml_str <- fix_hyperlink_refs_in_wml(xml_str, x$comments)
    xml_str <- fix_img_refs_in_wml(
      xml_str,
      x$comments,
      x$comments$relationship(),
      x$package_dir
    )
    xml_str <- fix_svg_refs_in_wml(
      xml_str,
      x$comments,
      x$comments$relationship(),
      x$package_dir
    )

    tf_xml <- tempfile(fileext = ".txt")
    writeLines(xml_str, tf_xml, useBytes = TRUE)
    x$comments$replace_xml(tf_xml)
  }

  int_id <- 1 # unique id identifier
  # make all id unique for footnote
  int_id <- correct_id(x$comments$get(), int_id)

  body <- xml_find_first(x$doc_obj$get(), "w:body")

  # If body is not ending with an sectPr, create a continuous one append it
  if (!xml_name(xml_child(body, search = xml_length(body))) %in% "sectPr") {
    str <- paste0(
      "<w:sectPr xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">",
      "<w:type w:val=\"continuous\"/></w:sectPr>"
    )
    xml_add_child(body, as_xml_document(str))
  }

  for (header in x$headers) {
    header$save()
  }
  for (footer in x$footers) {
    footer$save()
  }

  x$doc_obj$save()
  x$content_type$save()
  x$footnotes$save()
  x$comments$save()

  x$rel$write(file.path(x$package_dir, "_rels", ".rels"))
  write_docx_settings(x)

  # save doc properties
  if (nrow(x$doc_properties$data) > 0) {
    x$doc_properties["modified", "value"] <- format(
      Sys.time(),
      "%Y-%m-%dT%H:%M:%SZ"
    )
    x$doc_properties["lastModifiedBy", "value"] <- Sys.getenv("USER")
    write_core_properties(x$doc_properties, x$package_dir)
  }
  if (nrow(x$doc_properties_custom$data) > 0) {
    write_custom_properties(x$doc_properties_custom, x$package_dir)
  }
  if (!is.null(x$app_properties)) {
    write_app_properties(x$app_properties, x$package_dir)
  }
  x <- sanitize_images(x, warn_user = FALSE)
  invisible(pack_folder(folder = x$package_dir, target = target))
}

props_to_ph_location <- function(props) {
  if (nrow(props) > 1) {
    cli::cli_alert_warning("More than one placeholder selected.")
  }
  props <- props[, c(
    "offx",
    "offy",
    "cx",
    "cy",
    "ph_label",
    "ph",
    "type",
    "fld_id",
    "fld_type",
    "rotation"
  )]
  names(props) <- c(
    "left",
    "top",
    "width",
    "height",
    "ph_label",
    "ph",
    "type",
    "fld_id",
    "fld_type",
    "rotation"
  )
  as_ph_location(props)
}


# id is deprecated and replaced by type_idx. Will be removed soon
get_ph_loc <- function(
  x,
  layout,
  master,
  type,
  type_idx = NULL,
  position_right,
  position_top,
  id = NULL,
  ph_id = NULL
) {
  props <- layout_properties(x, layout = layout, master = master)

  if (!is.null(ph_id)) {
    ids <- sort(stats::na.omit(as.numeric(props$id)))
    if (length(ids) <= 20) {
      .all_ids_switch <- c("x" = "Available ids:  {.val {ids}}.") # only if few ids
    } else {
      .all_ids_switch <- NULL
    }
    if (!ph_id %in% ids) {
      cli::cli_abort(
        c(
          "{.arg id} {.val {ph_id}} does not exist.",
          .all_ids_switch,
          "i" = cli::col_grey(
            "see column {.val id} in {.code layout_properties(..., '{layout}', '{master}')}"
          )
        ),
        call = NULL
      )
    }
    props <- props[props$id == ph_id, , drop = FALSE]
    return(props_to_ph_location(props))
  }

  types_on_layout <- unique(props$type)
  props <- props[props$type %in% type, , drop = FALSE]
  nr <- nrow(props)
  if (nr < 1) {
    cli::cli_abort(
      c(
        "Found no placeholder of type {.val {type}} on layout {.val {layout}}.",
        "x" = "Available types are {.val {types_on_layout}}",
        "i" = cli::col_grey(
          "see {.code layout_properties(x, '{layout}', '{master}')}"
        )
      ),
      call = NULL
    )
  }

  # id and type_idx are both used for now. 'id' is deprecated. The following code block can be removed in the future.
  if (!is.null(id)) {
    if (!id %in% 1L:nr) {
      cli::cli_abort(
        c(
          "{.arg id} is out of range.",
          "x" = "Must be between {.val {1L}} and {.val {nr}} for ph type {.val {type}}.",
          "i" = cli::col_grey(
            "see {.code layout_properties(x, '{layout}', '{master}')} for all phs with type '{type}'"
          )
        ),
        call = NULL
      )
    }
    # the ordering of 'type_idx' (top->bottom, left-righ) is different than for the 'id' arg (index
    # along the id colomn). Here, we restore the old ordering, to avoid a breaking change.
    props <- props[order(props$type, as.integer(props$id)), ] # set order for type idx. Removing the line would result in the default layout properties order, i.e., top->bottom left->right.
    props$.id <- stats::ave(
      props$type,
      props$master_name,
      props$name,
      props$type,
      FUN = seq_along
    )
    props <- props[props$.id == id, , drop = FALSE]
    return(props_to_ph_location(props))
  }

  if (!is.null(type_idx)) {
    if (!type_idx %in% props$type_idx) {
      cli::cli_abort(
        c(
          "{.arg type_idx} is out of range.",
          "x" = "Must be between {.val {1L}} and {.val {max(props$type_idx)}} for ph type {.val {type}}.",
          "i" = cli::col_grey(
            "see {.code layout_properties(..., layout = '{layout}', master = '{master}')} for indexes of type '{type}'"
          )
        ),
        call = NULL
      )
    }
    props <- props[props$type_idx == type_idx, , drop = FALSE]
    return(props_to_ph_location(props))
  }

  if (position_right) {
    props <- props[props$offx + 0.0001 > max(props$offx), ]
  } else {
    props <- props[props$offx - 0.0001 < min(props$offx), ]
  }
  if (position_top) {
    props <- props[props$offy - 0.0001 < min(props$offy), ]
  } else {
    props <- props[props$offy + 0.0001 > max(props$offy), ]
  }
  props_to_ph_location(props)
}


as_ph_location <- function(x, ...) {
  if (!is.data.frame(x)) {
    cli::cli_abort(
      c(
        "{.arg x} must be a data frame.",
        "x" = "You provided {.cls {class(x)[1]}} instead."
      )
    )
  }
  ref_names <- c(
    "width",
    "height",
    "left",
    "top",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )
  if (!all(is.element(ref_names, names(x)))) {
    stop(
      "missing column values:",
      paste0(setdiff(ref_names, names(x)), collapse = ",")
    )
  }
  out <- x[ref_names]
  as.list(out)
}


is_ph_location <- function(x) {
  inherits(x, "location_num") || inherits(x, "location_str")
}


#' @export
#' @title Eval a location on the current slide
#' @description Eval a shape location against the current slide.
#' This function is to be used to add custom openxml code. A
#' list is returned, it contains informations width, height, left
#' and top positions and other informations necessary to add a
#' content on a slide.
#' @param x a location for a placeholder.
#' @param doc an rpptx object
#' @param ... unused arguments
#' @example inst/examples/example_fortify_location.R
#' @seealso [ph_location()], [ph_with()]
#' @family functions for officer extensions
#' @keywords internal
fortify_location <- function(x, doc, ...) {
  UseMethod("fortify_location")
}

# _________________ ----
# main ----

#' @export
#' @title Location for a placeholder from scratch
#' @description The function will return a list that complies with
#' expected format for argument `location` of function [ph_with()].
#' @param left,top,width,height place holder coordinates
#' in inches.
#' @param newlabel a label for the placeholder. See section details.
#' @param bg background color
#' @param rotation rotation angle
#' @param ln a [sp_line()] object specifying the outline style.
#' @param geom shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html
#' @param ... unused arguments
#' @family functions for placeholder location
#' @details
#' The location of the bounding box associated to a placeholder
#' within a slide is specified with the left top coordinate,
#' the width and the height. These are defined in inches:
#'
#' \describe{
#'   \item{left}{left coordinate of the bounding box}
#'   \item{top}{top coordinate of the bounding box}
#'   \item{width}{width of the bounding box}
#'   \item{height}{height of the bounding box}
#' }
#'
#' In addition to these attributes, a label can be
#' associated with the shape. Shapes, text boxes, images and other objects
#' will be identified with that label in the *Selection Pane* of PowerPoint.
#' This label can then be reused by other functions such as `ph_location_label()`.
#' It can be set with argument `newlabel`.
#' @example inst/examples/example_ph_location.R
ph_location <- function(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  bg = NULL,
  rotation = NULL,
  ln = NULL,
  geom = NULL,
  ...
) {
  # remove name in case of named vector entry, e.g. ph_location(top = x["top])
  left <- unname(left)
  top <- unname(top)
  width <- unname(width)
  height <- unname(height)
  x <- list(
    left = left,
    top = top,
    width = width,
    height = height,
    ph_label = newlabel,
    ph = NA_character_,
    bg = bg,
    rotation = rotation,
    ln = ln,
    geom = geom,
    fld_type = NA_character_,
    fld_id = NA_character_
  )
  class(x) <- c("location_manual", "location_str")
  x
}


#' @export
fortify_location.location_manual <- function(x, doc, ...) {
  x
}

#' @title Location for a placeholder based on a template
#' @description The function will return a list that complies with
#' expected format for argument `location` of function
#' [ph_with()]. A placeholder will be used as template
#' and its positions will be updated with values `left`, `top`, `width`, `height`.
#' @param left,top,width,height place holder coordinates
#' in inches.
#' @param newlabel a label for the placeholder. See section details.
#' @param type placeholder type to look for in the slide layout, one
#' of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.
#' It will be used as a template placeholder.
#' @param id index of the placeholder template. If two body placeholder, there can be
#' two different index: 1 and 2 for the first and second body placeholders defined
#' in the layout.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_template.R
#' @export
ph_location_template <- function(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  type = NULL,
  id = 1,
  ...
) {
  x <- list(
    left = left,
    top = top,
    width = width,
    height = height,
    ph_label = newlabel,
    ph = NA_character_,
    type = type,
    id = id
  )

  class(x) <- c("location_template", "location_str")
  x
}
#' @export
fortify_location.location_template <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  if (!is.null(x$type)) {
    ph <- slide$get_xfrm(type = x$type, index = x$id)$ph
  } else {
    ph <- sprintf('<p:ph type="%s"/>', "body")
  }
  x <- ph_location(
    left = x$left,
    top = x$top,
    width = x$width,
    height = x$height,
    label = x$ph_label
  )
  x$ph <- ph
  fortify_location.location_manual(x)
}


#' @export
#' @title Location of a placeholder based on a type
#' @description The function will use the type name of the placeholder (e.g. body, title),
#' the layout name and few other criterias to find the corresponding location.
#' @param type placeholder type to look for in the slide layout, one
#' of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.
#' @param position_right the parameter is used when a selection with above
#' parameters does not provide a unique position (for example
#' layout 'Two Content' contains two element of type 'body').
#' If `TRUE`, the element the most on the right side will be selected,
#' otherwise the element the most on the left side will be selected.
#' @param position_top same than `position_right` but applied
#' to top versus bottom.
#' @param type_idx Type index of the placeholder. If there is more than one
#' placeholder of a type (e.g., `body`), the type index can be supplied to uniquely
#' identify a ph. The index is a running number starting at 1. It is assigned by
#' placeholder position  (top -> bottom, left -> right). See [plot_layout_properties()]
#' for details. If `idx` argument is used, `position_right` and `position_top`
#' are ignored.
#' @param id (**DEPRECATED, use `type_idx` instead**) Index of the placeholder.
#' If two body placeholder, there can be two different index: 1 and 2 for the
#' first and second body placeholders defined in the layout. If this argument
#' is used, `position_right` and `position_top` will be ignored.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_type.R
ph_location_type <- function(
  type = "body",
  type_idx = NULL,
  position_right = TRUE,
  position_top = TRUE,
  newlabel = NULL,
  id = NULL,
  ...
) {
  # the following two warnings can be deleted after the deprecated id arg is removed.
  if (!is.null(id) && !is.null(type_idx)) {
    cli::cli_warn("{.arg id} is ignored if {.arg type_idx} is provided ")
  }
  if (!is.null(id) && is.null(type_idx)) {
    cli::cli_warn(
      c(
        "!" = "The {.arg id} argument in {.fn ph_location_type} is deprecated as of {.pkg officer} 0.6.7.",
        "i" = "Please use the {.arg type_idx} argument instead.",
        "x" = cli::col_red(
          "Caution: new index logic in {.arg type_idx} (see docs)."
        )
      )
    )
  }

  ph_types <- c(
    "ctrTitle",
    "subTitle",
    "dt",
    "ftr",
    "sldNum",
    "title",
    "body",
    "pic",
    "chart",
    "tbl",
    "dgm",
    "media",
    "clipArt"
  )
  if (!type %in% ph_types) {
    cli::cli_abort(
      c(
        "type {.val {type}} is unknown.",
        "x" = "Must be one of {.or {.val {ph_types}}}"
      ),
      call = NULL
    )
  }
  x <- list(
    type = type,
    type_idx = type_idx,
    position_right = position_right,
    position_top = position_top,
    id = id,
    label = newlabel
  )
  class(x) <- c("location_type", "location_str")
  x
}


#' @export
fortify_location.location_type <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()
  args <- list(...)

  layout <- ifelse(is.null(args$layout), unique(xfrm$name), args$layout)
  master <- ifelse(is.null(args$master), unique(xfrm$master_name), args$master)

  # to avoid a breaking change, the deprecated id is passed along.
  # As type_idx uses a different index order than id, this is necessary until the id arg is removed.
  out <- get_ph_loc(
    doc,
    layout = layout,
    master = master,
    type = x$type,
    position_right = x$position_right,
    position_top = x$position_top,
    type_idx = x$type_idx,
    id = x$id,
    ph_id = NULL # id is deprecated and will be removed soon
  )
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a named placeholder
#' @description The function will use the label of a placeholder
#' to find the corresponding location.
#' @param ph_label placeholder label of the used layout. It can be read in PowerPoint or
#' with function [layout_properties()] in column `ph_label`.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_label.R
ph_location_label <- function(ph_label, newlabel = NULL, ...) {
  x <- list(ph_label = unname(ph_label), label = unname(newlabel))
  class(x) <- c("location_label", "location_str")
  x
}


#' @export
fortify_location.location_label <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()

  layout <- unique(xfrm$name)
  master <- unique(xfrm$master_name)

  props_all <- layout_properties(doc, layout = layout, master = master)

  if (nrow(props_all) == 0) {
    cli::cli_abort(
      c(
        "ph label {.val {x$ph_label}} does not exist",
        "x" = "The layout has no placeholders!"
      ),
      call = NULL
    )
  }

  props <- props_all[props_all$ph_label %in% x$ph_label, , drop = FALSE]

  if (nrow(props) < 1) {
    cli::cli_abort(
      c(
        "ph label {.val {x$ph_label}} does not exist",
        "x" = "layout {.val {layout}} has ph labels: {.val {props_all$ph_label}}",
        "i" = "See {.fn plot_layout_properties} for details"
      ),
      call = NULL
    )
  }

  if (nrow(props) > 1) {
    cli::cli_abort(
      c(
        "Placeholder {.val {x$ph_label}} label in layout {.val {layout}} is duplicated.",
        "x" = "A layout's ph labels must be unique.",
        "i" = "{.fn layout_dedupe_ph_labels} helps handling duplicates."
      ),
      call = NULL
    )
  }

  props <- props[, c(
    "offx",
    "offy",
    "cx",
    "cy",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )]
  names(props) <- c(
    "left",
    "top",
    "width",
    "height",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )
  row.names(props) <- NULL
  out <- as_ph_location(props)
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a full size element
#' @description The function will return the location corresponding
#' to a full size display.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @example inst/examples/example_ph_location_fullsize.R
ph_location_fullsize <- function(newlabel = "", ...) {
  x <- list(label = newlabel)
  class(x) <- c("location_fullsize", "location_str")
  x
}

#' @export
fortify_location.location_fullsize <- function(x, doc, ...) {
  layout_data <- slide_size(doc)
  layout_data$left <- 0L
  layout_data$top <- 0L
  if (!is.null(x$label)) {
    layout_data$ph_label <- x$label
  }
  layout_data$ph <- NA_character_
  layout_data$type <- "body"
  layout_data$rotation <- 0L
  layout_data$fld_id <- NA_character_
  layout_data$fld_type <- NA_character_

  as_ph_location(as.data.frame(layout_data, stringsAsFactors = FALSE))
}

#' @export
#' @title Location of a left body element
#' @description The function will return the location corresponding
#' to a left bounding box. The function assume the layout 'Two Content'
#' is existing. This is an helper function, if you don't have a layout
#' named 'Two Content', use [ph_location_type()] and set arguments
#' to your specific needs.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @example inst/examples/example_ph_location_left.R
ph_location_left <- function(newlabel = NULL, ...) {
  x <- list(label = newlabel)
  class(x) <- c("location_left", "location_str")
  x
}

#' @export
fortify_location.location_left <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()

  args <- list(...)
  master <- if (is.null(args$master)) unique(xfrm$master_name) else args$master
  out <- get_ph_loc(
    doc,
    layout = "Two Content",
    master = master,
    type = "body",
    position_right = FALSE,
    position_top = TRUE
  )
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a right body element
#' @description The function will return the location corresponding
#' to a right bounding box. The function assume the layout 'Two Content'
#' is existing. This is an helper function, if you don't have a layout
#' named 'Two Content', use [ph_location_type()] and set arguments
#' to your specific needs.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @example inst/examples/example_ph_location_left.R
ph_location_right <- function(newlabel = NULL, ...) {
  x <- list(label = newlabel)
  class(x) <- c("location_right", "location_str")
  x
}


#' @export
fortify_location.location_right <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()

  args <- list(...)
  master <- ifelse(is.null(args$master), unique(xfrm$master_name), args$master)
  out <- get_ph_loc(
    doc,
    layout = "Two Content",
    master = master,
    type = "body",
    position_right = TRUE,
    position_top = TRUE
  )
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a placeholder based on its id
#' @description Each placeholder has an id (a low integer value). The ids are unique across a single
#' layout. The function uses the placeholder's id to reference it. Different from a ph label,
#' the id is auto-assigned by PowerPoint and cannot be modified by the user.
#' Use [layout_properties()] (column `id`) and [plot_layout_properties()] (upper right
#' corner, in green) to find a placeholder's id.
#'
#' @param id placeholder id.
#' @param newlabel a new label to associate with the placeholder.
#' @param ... not used.
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_id.R
ph_location_id <- function(id, newlabel = NULL, ...) {
  ph_id <- id # for disambiguation, store initial value

  if (length(ph_id) > 1) {
    cli::cli_abort(
      c(
        "{.arg id} must be {cli::style_underline('one')} number",
        "x" = "Found more than one entry: {.val {ph_id}}"
      )
    )
  }
  if (is.null(ph_id) || is.na(ph_id) || length(ph_id) == 0) {
    cli::cli_abort("{.arg id} must be a positive number")
  }
  if (!is.integer(ph_id)) {
    ph_id <- suppressWarnings(as.integer(ph_id))
    if (is.na(ph_id)) {
      cli::cli_abort(
        c(
          "Cannot convert {.val {id}} to integer",
          "x" = "{.arg id} must be a number, you provided class {.cls {class(id)[1]}}"
        )
      )
    }
  }
  if (ph_id < 1) {
    cli::cli_abort(
      c(
        "{.arg id} must be a {cli::style_underline('positive')} number",
        "x" = "Found {.val {ph_id}}"
      )
    )
  }
  x <- list(
    type = NULL,
    type_idx = NULL,
    position_right = NULL,
    position_right = NULL,
    position_top = NULL,
    id = NULL,
    ph_id = ph_id,
    label = newlabel
  )
  class(x) <- c("location_id", "location_num")
  x
}


#' @export
fortify_location.location_id <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()
  args <- list(...)

  layout <- ifelse(is.null(args$layout), unique(xfrm$name), args$layout)
  master <- ifelse(is.null(args$master), unique(xfrm$master_name), args$master)
  out <- get_ph_loc(doc, layout = layout, master = master, ph_id = x$ph_id)
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}

# _________________ ----
# resolve ----

#' @export
#' @keywords internal
#' @title Resolve short form location
#' @description
#' Convert short form location format, a numeric or a string (e.g. `"body [1]"`),
#' into its corresponding location object. Under the hood, we parse the short form location and
#' call the corresponding function from the `ph_location_*` family. Note that short forms may not
#' cover all function from the `ph_location_*` and offer less customization.
#'
#' @section Short forms:
#' The following location short forms are implemented. The corresponding call of the function from the
#' `ph_location_*` family is displayed on the right.
#'
#' | **Short form** | **Description**                                             | **Location function**           |
#' |----------------|-------------------------------------------------------------|---------------------------------|
#' | `"left"`       | Keyword string                                              | `ph_location_left()`            |
#' | `"right"`      | Keyword string                                              | `ph_location_right()`           |
#' | `"fullsize"`   | Keyword string                                              | `ph_location_fullsize()`        |
#' | `"body [1]"`   | String: type + index in brackets (`1` if omitted)           | `ph_location_type("body", 1)`   |
#' | `"my_label"`   | Any string not matching a keyword or type                   | `ph_location_label("my_label")` |
#' | `2`            | Length 1 integer                                            | `ph_location_id(2)`             |
#' | `c(0,0,4,5)`   | Length 4 numeric, optionally named, `c(top=0, left=0, ...)` | `ph_location(0, 0, 4, 5)`       |
#'
#' @example inst/examples/example_resolve_location.R
resolve_location <- function(x) {
  if (is_ph_location(x)) {
    return(x)
  }
  if (is.null(x) || (is.vector(x) && is.na(x)[1])) {
    raise_location_value_error(x)
  }
  if (is.numeric(x)) {
    loc <- resolve_location_from_numeric(x)
  } else if (is.character(x)) {
    loc <- resolve_location_from_character(x)
  } else {
    cli::cli_abort(
      call = NULL,
      "Cannot resolve class {.cls {class(x)[1]}} into a location",
      "x" = "Must be a character or numeric vector."
    )
  }
  loc
}


raise_location_value_error <- function(x) {
  cli::cli_abort(
    call = NULL,
    c(
      "Incorrect value for {.arg location}",
      "x" = "Must be a vector (character or numeric) or a ph_location object"
    )
  )
}


resolve_location_from_numeric <- function(x) {
  # length 1 integer => ph_location_id()
  # length 4 numeric => ph_location()
  len <- length(x)
  if (len == 1) {
    if (!is_integerish(x)) {
      cli::cli_abort(
        c(
          "{.arg location} is a length 1 {.cls {class(x)[1]}}: {.val {x}}",
          "x" = "If a length 1 numeric, {.arg location} requires {.cls integer}"
        ),
        call = NULL
      )
    }
    location <- resolve_ph_location_id(x)
  } else if (len == 4) {
    location <- resolve_ph_location(x)
  } else {
    cli::cli_abort(
      c(
        "{.arg location} has incorrect length.",
        "x" = "Numeric vector passed to {.arg location} must have length 1 or 4"
      ),
      call = NULL
    )
  }
  location
}


resolve_ph_location_id <- function(x) {
  if (x < 0) {
    cli::cli_abort(
      c(
        "{.arg location} is negative.",
        "x" = "Integers passed to {.arg location} must be positive"
      ),
      call = NULL
    )
  }
  ph_location_id(id = x)
}


# checks named numeric vector with position info [for ph_location_position()].
# c(left =, top =, width =, height = )
#
fortify_named_location_position <- function(x) {
  if (!is_named(x)) {
    cli::cli_abort(
      c(
        "Some vector elements have no names",
        "x" = "{.arg x} must be a named vector"
      )
    )
  }
  args <- names(x)
  expected <- c("left", "top", "width", "height")
  matched <- pmatch(args, expected, duplicates.ok = TRUE)
  nms_new <- ifelse(is.na(matched), NA, expected[matched])

  # unknown position
  i_na <- is.na(nms_new)
  if (any(i_na)) {
    cli::cli_abort(
      c(
        "Found {sum(i_na)} unknown name{?s} in {.arg location}: {.val {args[i_na]}}",
        "x" = "{.arg location} understands {.val {expected}}",
        "i" = cli::col_silver("Partial name matching is supported")
      ),
      call = NULL
    )
  }
  # duplicate position
  ii_dupes <- duplicated(nms_new)
  if (any(ii_dupes)) {
    cli::cli_abort(
      c(
        "Duplicate entries in {.arg location}: {.val {unique(nms_new[ii_dupes])}}",
        "x" = "Each name in {.arg location} must be unique",
        "i" = cli::col_silver("Partial name matching is supported")
      ),
      call = NULL
    )
  }
  setNames(x, nms_new)
}


resolve_ph_location <- function(x) {
  if (is.null(names(x))) {
    names(x) <- c("left", "top", "width", "height") # same order as in ph_location() !
  }
  x <- fortify_named_location_position(x)
  do.call(ph_location, as.list(x))
}


resolve_location_from_character <- function(x) {
  if (is.character(x) && length(x) > 1) {
    cli::cli_abort(
      c(
        "{.arg location} has incorrect length.",
        "x" = "Character vector passed to {.arg location} must have length 1"
      ),
      call = NULL
    )
  }
  # cases:
  # - keyword left, => ph_location_left()
  #           right => ph_location_right()
  #           fullsize => ph_location_fullsize()
  # - type: body, title, ctrTitle, subTitle, dt, ftr, sldNum
  #         => ph_location_type()
  # - label: <any> => ph_location_label()
  if (x == "left") {
    location <- ph_location_left()
  } else if (x == "right") {
    location <- ph_location_right()
  } else if (x == "fullsize") {
    location <- ph_location_fullsize()
  } else if (has_ph_type_format(x)) {
    location <- do.call(ph_location_type, get_ph_type_info(x))
  } else {
    location <- ph_location_label(x)
  }
  location
}

# matches pattern "type [type_idx ]",
# e.g. "body", "body[1]", ""body [1]", "body    [ 1 ]" => all identical
.ph_type_pattern <- "^(body|title|ctrTitle|subTitle|dt|ftr|sldNum)\\s*(\\[\\s*\\d+\\s*\\])?$"

has_ph_type_format <- function(x) {
  grepl(.ph_type_pattern, trimws(x))
}


# extract type name and idx from squared brackets
# e.g. "body [1]" => list(type = "body", type_idx = 1)
get_ph_type_info <- function(x) {
  x <- trimws(x)
  matches <- regexec(.ph_type_pattern, x)
  extracted <- regmatches(x, matches)[[1]]
  l <- setNames(as.list(extracted), c("input", "type", "type_idx"))
  l$type_idx <- ifelse(l$type_idx == "", "[1]", l$type_idx) # if brackets with index is missing, "[1]" is used
  l$type_idx <- extract_integers(l$type_idx)
  l
}


# extract integers from text, e.g. "[1]" => 1
extract_integers <- function(x) {
  matches <- gregexpr("\\d+", x)
  v <- unlist(regmatches(x, matches))
  as.numeric(v)
}

test_that("add wrong arguments", {
  x <- read_pptx()
  expect_error(add_slide(x, "Title and XXXX", "Office Theme"), fixed = TRUE)
  expect_error(add_slide(x, "Title and Content", "Office XXXX"), fixed = TRUE)
})


test_that("default layout", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  expect_warning(
    add_slide(x),
    regexp = "Calling `add_slide()` without specifying a `layout` is deprecated",
    fixed = TRUE
  )
  expect_no_error(expect_warning(add_slide(x)))

  x <- layout_default(x, "Title Slide")
  expect_no_error(add_slide(x))
  expect_no_error(add_slide(x, layout = "Two Content"))
})


test_that("master is inferred", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  x <- add_slide(x, "Title Slide")
  x <- add_slide(x, "Title and Content")
  expect_equal(length(x), 2)

  file <- test_path("docs_dir", "test-layout-unique-and-dupe.pptx")
  x <- read_pptx(file)
  x <- add_slide(x, "Title Slide", "Master_1")
  x <- add_slide(x, "Title Slide") # name is unique so master=NULL works
  expect_error(
    add_slide(x, "Title and Content"),
    regex = "Layout \"Title and Content\" exists in more than one master"
  )
  x <- add_slide(x, "Title and Content", "Master_1")
  expect_equal(length(x), 3)
})


test_that("snapshot simple elements into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Two Content", master = "Office Theme")
  doc <- ph_with(doc, c(1L, 2L), location = ph_location_left())
  doc <- ph_with(
    doc,
    as.factor(c("rhhh", "vvvlllooo")),
    location = ph_location_right()
  )
  doc <- add_slide(doc, layout = "Two Content", master = "Office Theme")
  doc <- ph_with(doc, c(pi, pi / 2), location = ph_location_left())
  doc <- ph_with(doc, c(TRUE, FALSE), location = ph_location_right())
  expect_snapshot_doc(x = doc, name = "pptx-add-simple", engine = "testthat")
})


test_that("snapshot date into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)

  my_date <- as.Date("2025-09-02")

  x <- read_pptx()
  x <- add_slide(x, layout = "Two Content", master = "Office Theme")
  x <- ph_with(x, my_date, ph_location_type("dt"))
  x <- ph_with(x, my_date, ph_location_type("ftr"), date_format = "%d.%m.%Y")

  options(officer.date_format = "%D")
  x <- ph_with(x, my_date, ph_location_type("body", type_idx = 1))

  options(officer.date_format = NA) # NA should be okay as well
  expect_no_error({
    x <- ph_with(x, my_date, ph_location_type("body", type_idx = 2))
  })

  expect_snapshot_doc(x = x, name = "pptx-add-date", engine = "testthat")
})

test_that("add factor into placeholder", {
  my_factor <- factor(
    c("Level A", "Level B", "Level C"),
    levels = c("Level A", "Level B", "Level C")
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, my_factor, location = ph_location_type(type = "body"))

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 1)
  expect_equal(sm$text, "Level ALevel BLevel C")
})

test_that("add date into placeholder", {
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(
    doc,
    as.Date("2024-12-31", "%Y-%m-%d"),
    location = ph_location_type(type = "body")
  )

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 1)
  expect_equal(sm$text, "2024-12-31")
})

test_that("add unordered_list into placeholder", {
  my_ul <- unordered_list(
    level_list = c(1, 2, 1),
    str_list = c("Item 1", "Subitem", "Item 2")
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, my_ul, location = ph_location_type(type = "body"))

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 1)
  expect_equal(sm$text, "Item 1SubitemItem 2")
})

test_that("add plot_instr with custom dimensions", {
  anyplot <- plot_instr(code = {
    barplot(1:5, col = 2:6)
  })

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(
    doc,
    anyplot,
    location = ph_location(left = 1, top = 1, width = 4, height = 3)
  )

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 1)
  expect_equal(sm$cx, 4)
  expect_equal(sm$cy, 3)
})


test_that("snapshot ggplot into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not_installed("ggplot2")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  require(ggplot2)
  local_edition(3L)
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content")
  gg_plot <- ggplot(data = iris) +
    geom_point(
      mapping = aes(Sepal.Length, Petal.Length),
      size = 3
    ) +
    theme_minimal()
  doc <- ph_with(
    x = doc,
    value = gg_plot,
    location = ph_location_type(type = "body"),
    bg = "transparent"
  )
  doc <- ph_with(
    x = doc,
    value = "graphic title",
    location = ph_location_type(type = "title")
  )
  expect_snapshot_doc(x = doc, name = "pptx-add-ggplot2", engine = "testthat")
})


test_that("snapshot base plot into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  anyplot <- plot_instr(code = {
    barplot(1:5, col = 2:6)
  })
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(
    doc,
    anyplot,
    location = ph_location_fullsize(),
    bg = "#00000066",
    pointsize = 12
  )
  expect_snapshot_doc(x = doc, name = "pptx-add-barplot", engine = "testthat")
})


test_that("add unordered_list into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  ul1 <- unordered_list(
    level_list = c(0, 1, 1, 0, 0, 1, 1),
    str_list = c(
      "List1",
      "Item 1",
      "Item 2",
      "",
      "List 2",
      "Option A",
      "Option B"
    )
  )

  ul2 <- unordered_list(
    level_list = c(0, 1, 2, 0, 0, 1, 2) + 1,
    str_list = c(
      "List1",
      "Item 1",
      "Item 2",
      "",
      "List 2",
      "Option A",
      "Option B"
    ),
    style = fp_text_lite(color = "gray25")
  )

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(doc, ul1, location = ph_location_type())
  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(doc, ul2, location = ph_location_type())
  expect_snapshot_doc(x = doc, name = "pptx-add-ul", engine = "testthat")
})


test_that("add block_list into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  fpt_blue_bold <- fp_text_lite(color = "#006699", bold = TRUE)
  fpt_red_italic <- fp_text_lite(color = "#C32900", italic = TRUE)
  value <- block_list(
    fpar(ftext("hello world", fpt_blue_bold)),
    fpar(
      ftext("hello", fpt_blue_bold),
      " ",
      ftext("world", fpt_red_italic)
    ),
    fpar(
      ftext("hello world", fpt_red_italic)
    )
  )

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(
    doc,
    value = value,
    location = ph_location_type(),
    level_list = c(1, 2, 3)
  )
  expect_snapshot_doc(x = doc, name = "pptx-add-blocklist", engine = "testthat")
})


test_that("add formatted par into placeholder", {
  bold_face <- shortcuts$fp_bold(font.size = 30)
  bold_redface <- update(bold_face, color = "red")

  fpar_ <- fpar(
    ftext("Hello ", prop = bold_face),
    ftext("World", prop = bold_redface),
    ftext(", how are you?", prop = bold_face)
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, fpar_, location = ph_location_type(type = "body"))

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 1)
  expect_equal(sm[1, ]$text, "Hello World, how are you?")

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  cols <- xml_attr(xml_find_all(xmldoc, "//a:rPr/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("000000", "FF0000", "000000"))
  expect_equal(xml_attr(xml_find_all(xmldoc, "//a:rPr"), "b"), rep("1", 3))
})


test_that("add xml into placeholder", {
  xml_str <- "<p:sp xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\"><p:nvSpPr><p:cNvPr id=\"\" name=\"\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr><p:ph type=\"title\"/></p:nvPr></p:nvSpPr><p:spPr/>\n<p:txBody><a:bodyPr/><a:lstStyle/><a:p><a:r><a:rPr/><a:t>Hello world 1</a:t></a:r></a:p></p:txBody></p:sp>"
  library(xml2)
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(
    doc,
    value = as_xml_document(xml_str),
    location = ph_location_type(type = "body")
  )
  doc <- ph_with(
    doc,
    value = as_xml_document(xml_str),
    location = ph_location(left = 1, top = 1, width = 3, height = 3)
  )
  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)
  expect_equal(sm[1, ]$text, "Hello world 1")
  expect_equal(sm[2, ]$text, "Hello world 1")
})


test_that("slidelink shape", {
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Un titre 1", location = ph_location_type(type = "title"))
  doc <- ph_with(doc, "text 1", location = ph_location_type(type = "body"))
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Un titre 2", location = ph_location_type(type = "title"))
  doc <- on_slide(doc, index = 1)

  doc <- ph_slidelink(doc, type = "body", slide_index = 2)

  rel_df <- doc$slide$get_slide(1)$rel_df()

  slide_filename <- doc$slide$get_metadata()$name[2]

  expect_true(slide_filename %in% rel_df$target)
  row_num_ <- which(
    is.na(rel_df$target_mode) & rel_df$target %in% slide_filename
  )

  rid <- rel_df[row_num_, "id"]
  xpath_ <- sprintf("//p:sp[p:nvSpPr/p:cNvPr/a:hlinkClick/@r:id='%s']", rid)
  node_ <- xml_find_first(doc$slide$get_slide(1)$get(), xpath_)
  expect_false(inherits(node_, "xml_missing"))
})

test_that("hyperlink shape", {
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(
    x = doc,
    location = ph_location_type(type = "title"),
    value = "Un titre 1"
  )
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(
    x = doc,
    location = ph_location_type(type = "title"),
    value = "Un titre 2"
  )
  doc <- on_slide(doc, 1)
  doc <- ph_hyperlink(
    x = doc,
    type = "title",
    href = "https://cran.r-project.org"
  )
  outputfile <- tempfile(fileext = ".pptx")
  print(doc, target = outputfile)

  doc <- read_pptx(outputfile)
  rel_df <- doc$slide$get_slide(1)$rel_df()

  expect_true("https://cran.r-project.org" %in% rel_df$target)
  row_num_ <- which(
    !is.na(rel_df$target_mode) & rel_df$target %in% "https://cran.r-project.org"
  )

  rid <- rel_df[row_num_, "id"]
  xpath_ <- sprintf("//p:sp[p:nvSpPr/p:cNvPr/a:hlinkClick/@r:id='%s']", rid)
  node_ <- xml_find_first(doc$slide$get_slide(1)$get(), xpath_)
  expect_false(inherits(node_, "xml_missing"))
})

test_that("hyperlink image", {
  img.file <- file.path(R.home(component = "doc"), "html", "logo.jpg")

  x <- read_pptx()
  x <- add_slide(x, "Title and Content")
  x <- ph_with(
    x,
    value = external_img(img.file),
    location = ph_location(newlabel = "logo")
  )
  x <- ph_hyperlink(
    x = x,
    ph_label = "logo",
    href = "https://cran.r-project.org"
  )
  outputfile <- tempfile(fileext = ".pptx")
  print(x, target = outputfile)

  doc <- read_pptx(outputfile)
  rel_df <- doc$slide$get_slide(1)$rel_df()

  expect_true("https://cran.r-project.org" %in% rel_df$target)
})

test_that("img dims in pptx", {
  skip_on_os("windows")
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(
    doc,
    value = external_img(img.file),
    location = ph_location(
      left = 1,
      top = 1,
      height = 1.06,
      width = 1.39
    )
  )
  sm <- slide_summary(doc)

  expect_equal(nrow(sm), 1)
  expect_equal(sm$cx, 1.39)
  expect_equal(sm$cy, 1.06)
  expect_equal(sm$offx, 1)
  expect_equal(sm$offy, 1)

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(
    doc,
    value = external_img(img.file),
    location = ph_location(
      left = 1,
      top = 1,
      height = 1.06,
      width = 1.39
    ),
    use_loc_size = TRUE
  )
  sm <- slide_summary(doc)

  expect_equal(nrow(sm), 1)
  expect_equal(sm$cx, 1.39, tolerance = .01)
  expect_equal(sm$cy, 1.06, tolerance = .01)
  expect_equal(sm$offx, 1)
  expect_equal(sm$offy, 1)
})

test_that("empty_content in pptx", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(
    x = doc,
    value = empty_content(),
    location = ph_location(
      left = 0,
      top = 0,
      width = 2,
      height = 3,
      bg = "black"
    )
  )

  expect_equal(slide_summary(doc)$offy, 0)
  expect_equal(slide_summary(doc)$offx, 0)
  expect_equal(slide_summary(doc)$cy, 3)
  expect_equal(slide_summary(doc)$cx, 2)
})


unlink("*.pptx")

test_that("add wrong arguments", {
  doc <- read_pptx()
  expect_error(add_slide(doc, "Title and blah", "Office Theme"), fixed = TRUE)
  expect_error(add_slide(doc, "Title and Content", "Office Tddheme"), fixed = TRUE)
})


test_that("add simple elements into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Two Content", master = "Office Theme")
  doc <- ph_with(doc, c(1L, 2L), location = ph_location_left())
  doc <- ph_with(doc, as.factor(c("rhhh", "vvvlllooo")), location = ph_location_right())
  doc <- add_slide(doc, layout = "Two Content", master = "Office Theme")
  doc <- ph_with(doc, c(pi, pi / 2), location = ph_location_left())
  doc <- ph_with(doc, c(TRUE, FALSE), location = ph_location_right())
  expect_snapshot_doc(x = doc, name = "pptx-add-simple", engine = "testthat")
})


test_that("add ggplot into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not_installed("ggplot2")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  require(ggplot2)
  local_edition(3L)
  doc <- read_pptx()
  doc <- add_slide(doc)
  gg_plot <- ggplot(data = iris) +
    geom_point(
      mapping = aes(Sepal.Length, Petal.Length),
      size = 3
    ) +
    theme_minimal()
  doc <- ph_with(
    x = doc, value = gg_plot,
    location = ph_location_type(type = "body"),
    bg = "transparent"
  )
  doc <- ph_with(
    x = doc, value = "graphic title",
    location = ph_location_type(type = "title")
  )
  expect_snapshot_doc(x = doc, name = "pptx-add-ggplot2", engine = "testthat")
})


test_that("add base plot into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  anyplot <- plot_instr(code = {
    barplot(1:5, col = 2:6)
  })
  doc <- read_pptx()
  doc <- add_slide(doc)
  doc <- ph_with(
    doc, anyplot,
    location = ph_location_fullsize(),
    bg = "#00000066", pointsize = 12
  )
  expect_snapshot_doc(x = doc, name = "pptx-add-barplot", engine = "testthat")
})


test_that("add unordered_list into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  ul1 <- unordered_list(
    level_list = c(0, 1, 1, 0, 0, 1, 1),
    str_list = c("List1", "Item 1", "Item 2", "", "List 2", "Option A", "Option B")
  )

  ul2 <- unordered_list(
    level_list = c(0, 1, 2, 0, 0, 1, 2) + 1,
    str_list = c("List1", "Item 1", "Item 2", "", "List 2", "Option A", "Option B"),
    style = fp_text_lite(color = "gray25")
  )

  doc <- read_pptx()
  doc <- add_slide(doc)
  doc <- ph_with(doc, ul1, location = ph_location_type())
  doc <- add_slide(doc)
  doc <- ph_with(doc, ul2, location = ph_location_type())
  expect_snapshot_doc(x = doc, name = "pptx-add-ul", engine = "testthat")
})


test_that("add block_list into placeholder", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  fpt_blue_bold <- fp_text_lite(color = "#006699", bold = TRUE)
  fpt_red_italic <- fp_text_lite(color = "#C32900", italic = TRUE)
  value <- block_list(
    fpar(ftext("hello world", fpt_blue_bold)),
    fpar(
      ftext("hello", fpt_blue_bold), " ",
      ftext("world", fpt_red_italic)
    ),
    fpar(
      ftext("hello world", fpt_red_italic)
    )
  )

  doc <- read_pptx()
  doc <- add_slide(doc)
  doc <- ph_with(doc,
    value = value, location = ph_location_type(),
    level_list = c(1, 2, 3)
  )
  expect_snapshot_doc(x = doc, name = "pptx-add-blocklist", engine = "testthat")
})


test_that("add formatted par into placeholder", {
  bold_face <- shortcuts$fp_bold(font.size = 30)
  bold_redface <- update(bold_face, color = "red")

  fpar_ <- fpar(
    ftext("Hello ", prop = bold_face),
    ftext("World", prop = bold_redface),
    ftext(", how are you?", prop = bold_face)
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, fpar_, location = ph_location_type(type = "body"))

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 1)
  expect_equal(sm[1, ]$text, "Hello World, how are you?")

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  cols <- xml_attr(xml_find_all(xmldoc, "//a:rPr/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("000000", "FF0000", "000000"))
  expect_equal(xml_attr(xml_find_all(xmldoc, "//a:rPr"), "b"), rep("1", 3))
})


test_that("add xml into placeholder", {
  xml_str <- "<p:sp xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\"><p:nvSpPr><p:cNvPr id=\"\" name=\"\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr><p:ph type=\"title\"/></p:nvPr></p:nvSpPr><p:spPr/>\n<p:txBody><a:bodyPr/><a:lstStyle/><a:p><a:r><a:rPr/><a:t>Hello world 1</a:t></a:r></a:p></p:txBody></p:sp>"
  library(xml2)
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(doc, value = as_xml_document(xml_str), location = ph_location_type(type = "body"))
  doc <- ph_with(doc, value = as_xml_document(xml_str), location = ph_location(left = 1, top = 1, width = 3, height = 3))
  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)
  expect_equal(sm[1, ]$text, "Hello world 1")
  expect_equal(sm[2, ]$text, "Hello world 1")
})


test_that("slidelink shape", {
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Un titre 1", location = ph_location_type(type = "title"))
  doc <- ph_with(doc, "text 1", location = ph_location_type(type = "body"))
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Un titre 2", location = ph_location_type(type = "title"))
  doc <- on_slide(doc, index = 1)

  doc <- ph_slidelink(doc, type = "body", slide_index = 2)

  rel_df <- doc$slide$get_slide(1)$rel_df()

  slide_filename <- doc$slide$get_metadata()$name[2]

  expect_true(slide_filename %in% rel_df$target)
  row_num_ <- which(is.na(rel_df$target_mode) & rel_df$target %in% slide_filename)

  rid <- rel_df[row_num_, "id"]
  xpath_ <- sprintf("//p:sp[p:nvSpPr/p:cNvPr/a:hlinkClick/@r:id='%s']", rid)
  node_ <- xml_find_first(doc$slide$get_slide(1)$get(), xpath_)
  expect_false(inherits(node_, "xml_missing"))
})

test_that("hyperlink shape", {
  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(x = doc, location = ph_location_type(type = "title"), value = "Un titre 1")
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(x = doc, location = ph_location_type(type = "title"), value = "Un titre 2")
  doc <- on_slide(doc, 1)
  doc <- ph_hyperlink(x = doc, type = "title", href = "https://cran.r-project.org")
  outputfile <- tempfile(fileext = ".pptx")
  print(doc, target = outputfile)

  doc <- read_pptx(outputfile)
  rel_df <- doc$slide$get_slide(1)$rel_df()

  expect_true("https://cran.r-project.org" %in% rel_df$target)
  row_num_ <- which(!is.na(rel_df$target_mode) & rel_df$target %in% "https://cran.r-project.org")

  rid <- rel_df[row_num_, "id"]
  xpath_ <- sprintf("//p:sp[p:nvSpPr/p:cNvPr/a:hlinkClick/@r:id='%s']", rid)
  node_ <- xml_find_first(doc$slide$get_slide(1)$get(), xpath_)
  expect_false(inherits(node_, "xml_missing"))
})

test_that("hyperlink image", {
  img.file <- file.path(R.home(component = "doc"), "html", "logo.jpg")

  x <- read_pptx()
  x <- add_slide(x)
  x <- ph_with(x, value = external_img(img.file), location = ph_location(newlabel = "logo"))
  x <- ph_hyperlink(x = x, ph_label = "logo", href = "https://cran.r-project.org")
  outputfile <- tempfile(fileext = ".pptx")
  print(x, target = outputfile)

  doc <- read_pptx(outputfile)
  rel_df <- doc$slide$get_slide(1)$rel_df()

  expect_true("https://cran.r-project.org" %in% rel_df$target)
})

test_that("img dims in pptx", {
  skip_on_os("windows")
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(doc,
    value = external_img(img.file),
    location = ph_location(
      left = 1, top = 1,
      height = 1.06, width = 1.39
    )
  )
  sm <- slide_summary(doc)

  expect_equal(nrow(sm), 1)
  expect_equal(sm$cx, 1.39)
  expect_equal(sm$cy, 1.06)
  expect_equal(sm$offx, 1)
  expect_equal(sm$offy, 1)

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(doc,
    value = external_img(img.file),
    location = ph_location(
      left = 1, top = 1,
      height = 1.06, width = 1.39
    ),
    use_loc_size = TRUE
  )
  sm <- slide_summary(doc)

  expect_equal(nrow(sm), 1)
  expect_equal(sm$cx, 1.39, tolerance = .01)
  expect_equal(sm$cy, 1.06, tolerance = .01)
  expect_equal(sm$offx, 1)
  expect_equal(sm$offy, 1)
})

test_that("empty_content in pptx", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(
    x = doc, value = empty_content(),
    location = ph_location(
      left = 0, top = 0,
      width = 2, height = 3, bg = "black"
    )
  )

  expect_equal(slide_summary(doc)$offy, 0)
  expect_equal(slide_summary(doc)$offx, 0)
  expect_equal(slide_summary(doc)$cy, 3)
  expect_equal(slide_summary(doc)$cx, 2)
})


test_that("pptx ph locations", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")

  doc <- ph_with(
    x = doc, value = "left",
    location = ph_location_left()
  )
  doc <- ph_with(
    x = doc, value = "right",
    location = ph_location_right()
  )
  doc <- ph_with(
    x = doc, value = "title",
    location = ph_location_type(type = "title")
  )
  doc <- ph_with(
    x = doc, value = "fullsize",
    location = ph_location_fullsize()
  )
  doc <- ph_with(
    x = doc, value = "from title",
    location = ph_location_template(
      left = 1, width = 2, height = 1, top = 4,
      type = "title", newlabel = "newlabel"
    )
  )

  layouts_info <- layout_properties(doc)


  title_xfrm <- layouts_info[layouts_info$name %in% "Two Content" &
    layouts_info$type %in% "title", c("offx", "offy", "cx", "cy")]
  side_xfrm <- layouts_info[layouts_info$name %in% "Two Content" &
    layouts_info$type %in% "body", c("offx", "offy", "cx", "cy")]
  full_xfrm <- as.data.frame(slide_size(doc))
  names(full_xfrm) <- c("cx", "cy")
  full_xfrm <- cbind(data.frame(offx = 0L, offy = 0L), full_xfrm)
  from_title_xfrm <- data.frame(offx = 1, offy = 4, cx = 2, cy = 1)
  theorical_xfrm <- rbind(
    side_xfrm,
    title_xfrm,
    full_xfrm,
    from_title_xfrm
  )

  all_xfrm <- xml_find_all(
    x = doc$slide$get_slide(1)$get(),
    xpath = "/p:sld/p:cSld/p:spTree/p:sp/p:spPr/a:xfrm"
  )
  offx <- xml_attr(xml_child(all_xfrm, "a:off"), "x")
  offx <- as.integer(offx) / 914400
  offy <- xml_attr(xml_child(all_xfrm, "a:off"), "y")
  offy <- as.integer(offy) / 914400
  cx <- xml_attr(xml_child(all_xfrm, "a:ext"), "cx")
  cx <- as.integer(cx) / 914400
  cy <- xml_attr(xml_child(all_xfrm, "a:ext"), "cy")
  cy <- as.integer(cy) / 914400

  observed_xfrm <- data.frame(offx = offx, offy = offy, cx = cx, cy = cy)
  expect_equivalent(observed_xfrm, theorical_xfrm)
})


test_that("pptx ph_location_type", {
  opts <- options(cli.num_colors = 1) # suppress colors for easier error message check
  on.exit(options(opts))

  x <- read_pptx()
  x <- add_slide(x, "Two Content")

  expect_no_error({
    ph_with(x, "correct ph type id", ph_location_type("body", type_idx = 1))
  })

  expect_warning({
    ph_with(x, "cannot supply id AND type_idx", ph_location_type("body", type_idx = 1, id = 1))
  }, regexp = "`id` is ignored if `type_idx` is provided", fixed = TRUE)

  expect_warning({
    ph_with(x, "id still working with warning to avoid breaking change", ph_location_type("body", id = 1))
  }, regexp = "The `id` argument in `ph_location_type()` is deprecated", fixed = TRUE)

  expect_error({
      ph_with(x, "out of range type id", ph_location_type("body", type_idx = 3)) # 3 does not exists => no error or warning
  }, regexp = "`type_idx` is out of range.", fixed = TRUE)

  expect_error({
    expect_warning({
    ph_with(x, "out of range type id", ph_location_type("body", id = 3)) # 3 does not exists => no error or warning
    }, regexp = " The `id` argument in `ph_location_type()` is deprecated", fixed = TRUE)
  }, regexp = "`id` is out of range.", fixed = TRUE)

  expect_error({
    ph_with(x, "type okay but not available in layout", ph_location_type("tbl")) # tbl not on layout
  }, regexp = "Found no placeholder of type", fixed = TRUE)

  expect_error({
    ph_with(x, "xxx is unknown type", ph_location_type("xxx"))
  }, regexp = 'type "xxx" is unknown', fixed = TRUE)

  expect_no_error({ # for complete coverage
    ph_with(x, " ph type position_right", ph_location_type("body", position_right = TRUE))
  })
})


test_that("pptx ph_location_id", {
  opts <- options(cli.num_colors = 1) # no colors for easier error message check
  on.exit(options(opts))

  # direct errors
  error_exp <- "`id` must be one number"
  expect_error(ph_location_id(id = 1:2), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = -1:1), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = c("A", "B")), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = c(NA, NA)), regex = error_exp, fixed = TRUE)

  error_exp <- "`id` must be a positive number"
  expect_error(ph_location_id(id = NULL), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = NA), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = NaN), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = character(0)), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = integer(0)), regex = error_exp, fixed = TRUE)

  expect_error(ph_location_id(id = "A"), regex = 'Cannot convert "A" to integer', fixed = TRUE)
  expect_error(ph_location_id(id = ""), regex = 'Cannot convert "" to integer', fixed = TRUE)
  expect_error(ph_location_id(id = Inf), regex = "Cannot convert Inf to integer", fixed = TRUE)
  expect_error(ph_location_id(id = -Inf), regex = "Cannot convert -Inf to integer", fixed = TRUE)

  error_exp <- "`id` must be a positive number"
  expect_error(ph_location_id(id = 0), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = -1), regex = error_exp, fixed = TRUE)

  # downstream errors
  x <- read_pptx()
  x <- add_slide(x, "Comparison")

  expect_error(
    {
      ph_with(x, "id does not exist", ph_location_id(id = 1000))
    },
    "`id` 1000 does not exist",
    fixed = TRUE
  )

  # test for correct results
  expect_no_error({
    ids <- layout_properties(x, "Comparison")$id
    for (id in ids) {
      ph_with(x, paste("text:", id), ph_location_id(id, newlabel = paste("newlabel:", id)))
    }
  })
  nodes <- xml_find_all(
    x = x$slide$get_slide(1)$get(),
    xpath = "/p:sld/p:cSld/p:spTree/p:sp"
  )
  # text inside phs
  expect_true(all(xml_text(nodes) == paste("text:", ids)))
  # assigned shape names
  all_nvpr <- xml_find_all(nodes, "./p:nvSpPr/p:cNvPr")
  expect_true(all(xml_attr(all_nvpr, "name") == paste("newlabel:", ids)))
})


test_that("pptx ph labels", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")

  doc <- ph_with(
    x = doc, value = "elephant",
    location = ph_location_type(newlabel = "label1")
  )
  doc <- ph_with(
    x = doc, value = "elephant",
    location = ph_location_label(
      ph_label = "Date Placeholder 3",
      newlabel = "label2"
    )
  )
  doc <- ph_with(
    x = doc, value = "elephant",
    location = ph_location_left(newlabel = "label3")
  )
  doc <- ph_with(
    x = doc, value = "elephant",
    location = ph_location_right(newlabel = "label4")
  )

  all_nvpr <- xml_find_all(
    x = doc$slide$get_slide(1)$get(),
    xpath = "/p:sld/p:cSld/p:spTree/p:sp/p:nvSpPr/p:cNvPr"
  )
  expect_equal(
    xml_attr(all_nvpr, "name"),
    paste0("label", 1:4)
  )

  expect_error({
    doc <- ph_with(
      x = doc, value = "error if label does not exist",
      location = ph_location_label(ph_label = "xxx")
    )
  })
})



test_that("as_ph_location", {
  ref_names <- c("width", "height", "left", "top", "ph_label", "ph", "type", "rotation", "fld_id", "fld_type")
  l <- replicate(length(ref_names), "dummy", simplify = FALSE)
  df <- as.data.frame(l)
  names(df) <- ref_names

  expect_no_error({
    as_ph_location(df)
  })

  expect_error({
    as_ph_location(df[, -(1:2)])
  }, regexp = "missing column values:width,height", fixed = TRUE)

  expect_error({
    as_ph_location("wrong class supplied")
  }, regexp = "`x` must be a data frame", fixed = TRUE)
})


test_that("get_ph_loc", {
  x <- read_pptx()
  get_ph_loc(x, "Comparison", "Office Theme", type =  "body",
             position_right = TRUE, position_top = FALSE)

})


unlink("*.pptx")

test_that("pptx ph locations", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")

  doc <- ph_with(
    x = doc,
    value = "left",
    location = ph_location_left()
  )
  doc <- ph_with(
    x = doc,
    value = "right",
    location = ph_location_right()
  )
  doc <- ph_with(
    x = doc,
    value = "title",
    location = ph_location_type(type = "title")
  )
  doc <- ph_with(
    x = doc,
    value = "fullsize",
    location = ph_location_fullsize()
  )
  doc <- ph_with(
    x = doc,
    value = "from title",
    location = ph_location_template(
      left = 1,
      width = 2,
      height = 1,
      top = 4,
      type = "title",
      newlabel = "newlabel"
    )
  )

  layouts_info <- layout_properties(doc)

  title_xfrm <- layouts_info[
    layouts_info$name %in% "Two Content" & layouts_info$type %in% "title",
    c("offx", "offy", "cx", "cy")
  ]
  side_xfrm <- layouts_info[
    layouts_info$name %in% "Two Content" & layouts_info$type %in% "body",
    c("offx", "offy", "cx", "cy")
  ]
  full_xfrm <- as.data.frame(slide_size(doc))
  names(full_xfrm) <- c("cx", "cy")
  full_xfrm <- cbind(data.frame(offx = 0L, offy = 0L), full_xfrm)
  from_title_xfrm <- data.frame(offx = 1, offy = 4, cx = 2, cy = 1)
  theorical_xfrm <- rbind(
    side_xfrm,
    title_xfrm,
    full_xfrm,
    from_title_xfrm
  )

  all_xfrm <- xml_find_all(
    x = doc$slide$get_slide(1)$get(),
    xpath = "/p:sld/p:cSld/p:spTree/p:sp/p:spPr/a:xfrm"
  )
  offx <- xml_attr(xml_child(all_xfrm, "a:off"), "x")
  offx <- as.integer(offx) / 914400
  offy <- xml_attr(xml_child(all_xfrm, "a:off"), "y")
  offy <- as.integer(offy) / 914400
  cx <- xml_attr(xml_child(all_xfrm, "a:ext"), "cx")
  cx <- as.integer(cx) / 914400
  cy <- xml_attr(xml_child(all_xfrm, "a:ext"), "cy")
  cy <- as.integer(cy) / 914400

  observed_xfrm <- data.frame(offx = offx, offy = offy, cx = cx, cy = cy)
  expect_equivalent(observed_xfrm, theorical_xfrm)
})


test_that("pptx ph_location_type", {
  opts <- options(cli.num_colors = 1) # suppress colors for easier error message check
  on.exit(options(opts))

  x <- read_pptx()
  x <- add_slide(x, "Two Content")

  expect_no_error({
    ph_with(x, "correct ph type id", ph_location_type("body", type_idx = 1))
  })

  expect_warning(
    {
      ph_with(
        x,
        "cannot supply id AND type_idx",
        ph_location_type("body", type_idx = 1, id = 1)
      )
    },
    regexp = "`id` is ignored if `type_idx` is provided",
    fixed = TRUE
  )

  expect_warning(
    {
      ph_with(
        x,
        "id still working with warning to avoid breaking change",
        ph_location_type("body", id = 1)
      )
    },
    regexp = "The `id` argument in `ph_location_type()` is deprecated",
    fixed = TRUE
  )

  expect_error(
    {
      ph_with(x, "out of range type id", ph_location_type("body", type_idx = 3)) # 3 does not exists => no error or warning
    },
    regexp = "`type_idx` is out of range.",
    fixed = TRUE
  )

  expect_error(
    {
      expect_warning(
        {
          ph_with(x, "out of range type id", ph_location_type("body", id = 3)) # 3 does not exists => no error or warning
        },
        regexp = " The `id` argument in `ph_location_type()` is deprecated",
        fixed = TRUE
      )
    },
    regexp = "`id` is out of range.",
    fixed = TRUE
  )

  expect_error(
    {
      ph_with(
        x,
        "type okay but not available in layout",
        ph_location_type("tbl")
      ) # tbl not on layout
    },
    regexp = "Found no placeholder of type",
    fixed = TRUE
  )

  expect_error(
    {
      ph_with(x, "xxx is unknown type", ph_location_type("xxx"))
    },
    regexp = 'type "xxx" is unknown',
    fixed = TRUE
  )

  expect_no_error({
    # for complete coverage
    ph_with(
      x,
      " ph type position_right",
      ph_location_type("body", position_right = TRUE)
    )
  })
})


test_that("pptx ph_location_id", {
  opts <- options(cli.num_colors = 1) # no colors for easier error message check
  on.exit(options(opts))

  # direct errors
  error_exp <- "`id` must be one number"
  expect_error(ph_location_id(id = 1:2), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = -1:1), regex = error_exp, fixed = TRUE)
  expect_error(
    ph_location_id(id = c("A", "B")),
    regex = error_exp,
    fixed = TRUE
  )
  expect_error(ph_location_id(id = c(NA, NA)), regex = error_exp, fixed = TRUE)

  error_exp <- "`id` must be a positive number"
  expect_error(ph_location_id(id = NULL), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = NA), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = NaN), regex = error_exp, fixed = TRUE)
  expect_error(
    ph_location_id(id = character(0)),
    regex = error_exp,
    fixed = TRUE
  )
  expect_error(ph_location_id(id = integer(0)), regex = error_exp, fixed = TRUE)

  expect_error(
    ph_location_id(id = "A"),
    regex = 'Cannot convert "A" to integer',
    fixed = TRUE
  )
  expect_error(
    ph_location_id(id = ""),
    regex = 'Cannot convert "" to integer',
    fixed = TRUE
  )
  expect_error(
    ph_location_id(id = Inf),
    regex = "Cannot convert Inf to integer",
    fixed = TRUE
  )
  expect_error(
    ph_location_id(id = -Inf),
    regex = "Cannot convert -Inf to integer",
    fixed = TRUE
  )

  error_exp <- "`id` must be a positive number"
  expect_error(ph_location_id(id = 0), regex = error_exp, fixed = TRUE)
  expect_error(ph_location_id(id = -1), regex = error_exp, fixed = TRUE)

  # downstream errors
  x <- read_pptx()
  x <- add_slide(x, "Comparison")

  expect_error(
    {
      ph_with(x, "id does not exist", ph_location_id(id = 1000))
    },
    "`id` 1000 does not exist",
    fixed = TRUE
  )

  # test for correct results
  expect_no_error({
    ids <- layout_properties(x, "Comparison")$id
    for (id in ids) {
      ph_with(
        x,
        paste("text:", id),
        ph_location_id(id, newlabel = paste("newlabel:", id))
      )
    }
  })
  nodes <- xml_find_all(
    x = x$slide$get_slide(1)$get(),
    xpath = "/p:sld/p:cSld/p:spTree/p:sp"
  )
  # text inside phs
  expect_true(all(xml_text(nodes) == paste("text:", ids)))
  # assigned shape names
  all_nvpr <- xml_find_all(nodes, "./p:nvSpPr/p:cNvPr")
  expect_true(all(xml_attr(all_nvpr, "name") == paste("newlabel:", ids)))
})


test_that("pptx ph labels", {
  opts <- options(cli.num_colors = 1) # no colors for easier error message check
  on.exit(options(opts))

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")

  doc <- ph_with(
    x = doc,
    value = "elephant",
    location = ph_location_type(newlabel = "label1")
  )
  doc <- ph_with(
    x = doc,
    value = "elephant",
    location = ph_location_label(
      ph_label = "Date Placeholder 3",
      newlabel = "label2"
    )
  )
  doc <- ph_with(
    x = doc,
    value = "elephant",
    location = ph_location_left(newlabel = "label3")
  )
  doc <- ph_with(
    x = doc,
    value = "elephant",
    location = ph_location_right(newlabel = "label4")
  )

  all_nvpr <- xml_find_all(
    x = doc$slide$get_slide(1)$get(),
    xpath = "/p:sld/p:cSld/p:spTree/p:sp/p:nvSpPr/p:cNvPr"
  )
  expect_equal(
    xml_attr(all_nvpr, "name"),
    paste0("label", 1:4)
  )

  expect_error(
    {
      doc <- ph_with(
        x = doc,
        value = "error if label does not exist",
        location = ph_location_label(ph_label = "xxx")
      )
    },
    regexp = 'ph label "xxx" does not exist'
  )

  # edge case: layout without any placeholders
  file <- test_path("docs_dir/test-no-placeholders-in-layout.pptx")
  x <- read_pptx(file)
  expect_error(
    {
      x |> add_slide("layout_without_placeholders", xxx = "abc")
    },
    regex = "The layout has no placeholders!"
  )

  expect_error(
    {
      x <- x |> add_slide("layout_without_placeholders")
      x |> phs_with(xxx = "abc")
    },
    regex = "The layout has no placeholders!"
  )

  # edge case: duplicate labels
  expect_error(
    {
      file <- test_path("docs_dir", "test-pptx-dedupe-ph.pptx")
      x <- read_pptx(file)
      x <- x |> add_slide("2x2-dupes")
      x |> phs_with(Content = "xxx")
    },
    regex = 'Placeholder "Content" label in layout "2x2-dupes" is duplicated.'
  )
})


test_that("as_ph_location", {
  ref_names <- c(
    "width",
    "height",
    "left",
    "top",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )
  l <- replicate(length(ref_names), "dummy", simplify = FALSE)
  df <- as.data.frame(l)
  names(df) <- ref_names

  expect_no_error({
    as_ph_location(df)
  })

  expect_error(
    {
      as_ph_location(df[, -(1:2)])
    },
    regexp = "missing column values:width,height",
    fixed = TRUE
  )

  expect_error(
    {
      as_ph_location("wrong class supplied")
    },
    regexp = "`x` must be a data frame",
    fixed = TRUE
  )
})


test_that("get_ph_loc", {
  x <- read_pptx()
  expect_no_error({
    get_ph_loc(
      x,
      "Comparison",
      "Office Theme",
      type = "body",
      position_right = TRUE,
      position_top = FALSE
    )
  })
})


test_that("short-form locations", {
  opts <- options(cli.num_colors = 1) # no colors for easier error message check
  on.exit(options(opts))

  # helpers
  v2 <- v <- c(top = 0, left = 0.1, width = 5.2, height = 4)
  names(v2) <- substr(names(v2), 1, 1)
  expect_equal(fortify_named_location_position(v), v)
  expect_equal(fortify_named_location_position(v2), v)

  names(v2)[1] <- NA
  expect_error(
    fortify_named_location_position(v2),
    regexp = "Some vector elements have no names",
    fixed = TRUE
  )
  expect_error(
    fortify_named_location_position(unname(v)),
    regexp = "Some vector elements have no names",
    fixed = TRUE
  )

  res <- has_ph_type_format(c(
    "body",
    "body[1]",
    "body [1]",
    "  body    [ 1 ]  "
  ))
  expect_true(all(res))
  res <- has_ph_type_format(c(
    "body[]",
    "body  []",
    "body [a]",
    "body [ a1]",
    "body [1a]",
    "body [1",
    "body 1]",
    "unknown [1]"
  ))
  expect_true(all(!res))

  # incorrect input values/class
  err_msg <- "Must be a vector (character or numeric) or a ph_location object"
  expect_error(resolve_location(NA_integer_), regex = err_msg, fixed = TRUE)
  expect_error(resolve_location(NA_character_), regex = err_msg, fixed = TRUE)
  expect_error(resolve_location(NA_real_), regex = err_msg, fixed = TRUE)
  expect_error(resolve_location(NA), regex = err_msg, fixed = TRUE)
  expect_error(resolve_location(NULL), regex = err_msg, fixed = TRUE)
  expect_error(
    resolve_location(mtcars),
    regex = "Cannot resolve class <data.frame> into a location",
    fixed = TRUE
  )

  # numeric input
  v <- c(top = 0, left = 0, width = 5)
  expect_error(
    resolve_location(v),
    regex = "`location` has incorrect length.",
    fixed = TRUE
  )
  expect_error(
    resolve_location(unname(v)),
    regexp = "`location` has incorrect length.",
    fixed = TRUE
  )

  ## numeric: single integer
  v <- c("dummy" = 1)
  loc_1 <- resolve_location(v)
  loc_2 <- resolve_location(unname(v))
  loc_3 <- ph_location_id(v)
  loc_4 <- resolve_location(loc_3) # unchanged
  expect_equal(loc_1, loc_2)
  expect_equal(loc_1, loc_3)
  expect_equal(loc_1, loc_4)

  expect_error(
    resolve_location(1.1),
    regex = "If a length 1 numeric, `location` requires <integer>",
    fixed = TRUE
  )
  expect_error(
    resolve_location(-1),
    regex = "Integers passed to `location` must be positive",
    fixed = TRUE
  )

  ## numeric: (named) position vector
  v2 <- v <- c(left = 1, top = 2, width = 3.3, height = 4.4)
  names(v2) <- substr(names(v2), 1, 1)
  loc_1 <- resolve_location(unname(v))
  loc_2 <- resolve_location(v)
  loc_3 <- resolve_location(v2) # partial name matching
  loc_4 <- resolve_location(rev(v)) # order does not matter
  loc_5 <- ph_location(
    left = v["left"],
    top = v["top"],
    width = v["width"],
    height = v["height"]
  )
  loc_6 <- resolve_location(loc_5) # unchanged
  expect_equal(loc_1, loc_2)
  expect_equal(loc_1, loc_3)
  expect_equal(loc_1, loc_4)
  expect_equal(loc_1, loc_5)
  expect_equal(loc_1, loc_6)

  v <- c(top = 1, left = 2, width = 3, xxx = 10)
  expect_error(
    resolve_location(v),
    regex = 'Found 1 unknown name in `location`: "xxx"',
    fixed = TRUE
  )
  v <- c(top = 1, top = 2, width = 3, width = 4)
  expect_error(
    resolve_location(v),
    regex = 'Duplicate entries in `location`: "top" and "width"',
    fixed = TRUE
  )

  # character input
  v <- c("a", "b")
  expect_error(
    resolve_location(v),
    regex = "Character vector passed to `location` must have length 1",
    fixed = TRUE
  )

  ## keywords
  expect_equal(resolve_location("left"), ph_location_left())
  expect_equal(resolve_location("right"), ph_location_right())
  expect_equal(resolve_location("fullsize"), ph_location_fullsize())

  # type + type index
  types <- c("body", "title", "ctrTitle", "subTitle", "dt", "ftr", "sldNum")
  for (type in types) {
    loc <- ph_location_type(type, 2)
    expect_equal(resolve_location(paste0(type, "[2]")), loc)
    expect_equal(resolve_location(paste0(type, " [2]")), loc)
    expect_equal(resolve_location(paste0("  ", type, "  [2]  ")), loc)
    expect_equal(resolve_location(paste0("  ", type, "  [ 2  ]  ")), loc)
  }

  ## character: label if not keyword or type
  expect_equal(resolve_location("< a label>"), ph_location_label("< a label>"))
  expect_equal(
    resolve_location(c(xxx = "< a label>")),
    ph_location_label("< a label>")
  )
  expect_equal(resolve_location("body[]"), ph_location_label("body[]"))
  expect_equal(resolve_location("left "), ph_location_label("left "))
  expect_equal(resolve_location(" left"), ph_location_label(" left"))
  expect_equal(resolve_location(" left "), ph_location_label(" left "))

  # full example
  expect_no_error({
    x <- read_pptx()
    x <- add_slide(x, "Title Slide")
    x <- ph_with(x, "A title", "Title 1") # label
    x <- ph_with(x, "A subtitle", 3) # id
    x <- ph_with(x, "A date", "dt[1]") # type + index
    x <- ph_with(x, "A left text", "left") # keyword
    x <- ph_with(x, "More content", c(5, .5, 9, 2)) # numeric vector
  })
})


get_shapetree <- function(x, slide_idx = NULL) {
  stop_if_not_rpptx(x)
  slide_idx <- slide_idx %||% x$cursor
  xml_node <- x$slide$get_slide(slide_idx)$get()
  xml2::xml_child(xml_node, "*/p:spTree")
}


get_shapetrees <- function(x, slide_idx = NULL) {
  stop_if_not_rpptx(x)
  slide_idx <- slide_idx %||% seq_len(length(x))
  lapply(slide_idx, function(idx) get_shapetree(x, idx))
}


# all slide's shapetrees as a string and shape's UUIDs removed
# used to check if created slides are identical.
get_shapetrees_string <- function(x, slide_idx = NULL) {
  stop_if_not_rpptx(x)
  sp_tree <- get_shapetrees(x, slide_idx = slide_idx)
  sp_tree_chr <- vapply(sp_tree, paste, character(1))
  s <- paste(sp_tree_chr, collapse = " ")
  gsub(
    "[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}",
    "xxx",
    s
  ) # delete shape's UUIDs
}


test_that("ph_with, phs_with, add_slide equality", {
  x1 <- read_pptx()
  x1 <- add_slide(x1, "Two Content")

  x2 <- read_pptx()
  x2 <- add_slide(x2, "Two Content")
  x2 <- phs_with(x2) # empty case, no changes to rpptx

  st_1 <- get_shapetrees_string(x1)
  st_2 <- get_shapetrees_string(x2)

  expect_equal(st_1, st_2)

  x1 <- read_pptx()
  x1 <- add_slide(x1, "Two Content")
  x1 <- ph_with(x1, "A title", "Title 1")
  x1 <- ph_with(x1, "Jan. 26, 2025", "dt")
  x1 <- ph_with(x1, "Body text", "body [2]")
  x1 <- ph_with(x1, "Footer", 6)

  x2 <- read_pptx()
  x2 <- add_slide(x2, "Two Content")
  x2 <- phs_with(
    x2,
    `Title 1` = "A title",
    dt = "Jan. 26, 2025",
    `body[2]` = "Body text",
    `6` = "Footer"
  )

  x3 <- read_pptx()
  x3 <- add_slide(
    x3,
    "Two Content",
    `Title 1` = "A title",
    dt = "Jan. 26, 2025",
    `body[2]` = "Body text",
    `6` = "Footer"
  )

  st_1 <- get_shapetrees_string(x1)
  st_2 <- get_shapetrees_string(x2)
  st_3 <- get_shapetrees_string(x3)

  expect_equal(st_1, st_2)
  expect_equal(st_1, st_3)
})


test_that("phs_with .slide_idx arg", {
  x1 <- read_pptx()
  x1 <- add_slide(x1, "Two Content")
  x1 <- ph_with(x1, "Footer", "ftr")
  x1 <- ph_with(x1, "Jan. 26, 2025", "dt")
  x1 <- add_slide(x1, "Two Content")
  x1 <- ph_with(x1, "Footer", "ftr")
  x1 <- ph_with(x1, "Jan. 26, 2025", "dt")

  x2 <- read_pptx()
  x2 <- add_slide(x2, "Two Content")
  x2 <- add_slide(x2, "Two Content")
  x2 <- phs_with(x2, ftr = "Footer", dt = "Jan. 26, 2025", .slide_idx = 1:2)

  x3 <- read_pptx()
  x3 <- add_slide(x3, "Two Content")
  x3 <- add_slide(x3, "Two Content")
  x3 <- phs_with(x3, ftr = "Footer", dt = "Jan. 26, 2025", .slide_idx = "all")

  st_1 <- get_shapetrees_string(x1)
  st_2 <- get_shapetrees_string(x2)
  st_3 <- get_shapetrees_string(x3)

  expect_equal(st_1, st_2)
  expect_equal(st_1, st_3)
})


test_that("add_slide, phs_with: ... and .dots", {
  opts <- options(testthat.use_colours = FALSE)
  on.exit(opts)

  # errors
  x0 <- read_pptx()
  x0 <- add_slide(x0, "Two Content")
  expect_error(
    phs_with(x0, "A title", dt = "Jan. 26, 2025"),
    regexp = "Missing key in `...`"
  )
  expect_error(
    phs_with(x0, .dots = list("A title", dt = "Jan. 26, 2025")),
    regexp = "Missing names in `.dots`"
  )
  expect_error(
    add_slide(
      x0,
      "Two Content",
      master = "Office Theme",
      "A title",
      dt = "Jan. 26, 2025"
    ),
    regexp = "Missing key in `...`"
  )
  expect_error(
    add_slide(x0, "Two Content", .dots = list("A title", dt = "Jan. 26, 2025")),
    regexp = "Missing names in `.dots`"
  )

  # functionality
  x1 <- read_pptx()
  x1 <- add_slide(x1, "Two Content")
  expect_no_error(phs_with(x1)) # empty case
  x1 <- phs_with(
    x1,
    `Title 1` = "A title",
    dt = "Jan. 26, 2025",
    `body[2]` = "Body text",
    `6` = "Footer"
  )

  x2 <- read_pptx()
  x2 <- add_slide(x2, "Two Content")
  x2 <- phs_with(
    x2,
    .dots = list(
      `Title 1` = "A title",
      dt = "Jan. 26, 2025",
      `body[2]` = "Body text",
      `6` = "Footer"
    )
  )

  x3 <- read_pptx()
  x3 <- add_slide(x3, "Two Content")
  x3 <- phs_with(
    x3,
    `Title 1` = "A title",
    dt = "Jan. 26, 2025",
    .dots = list(`body[2]` = "Body text", `6` = "Footer")
  )

  x4 <- read_pptx()
  x4 <- add_slide(
    x4,
    "Two Content",
    `Title 1` = "A title",
    dt = "Jan. 26, 2025",
    .dots = list(`body[2]` = "Body text", `6` = "Footer")
  )

  x5 <- read_pptx()
  x5 <- add_slide(
    x5,
    "Two Content",
    .dots = list(
      `Title 1` = "A title",
      dt = "Jan. 26, 2025",
      `body[2]` = "Body text",
      `6` = "Footer"
    )
  )

  st_1 <- get_shapetrees_string(x1)
  st_2 <- get_shapetrees_string(x2)
  st_3 <- get_shapetrees_string(x3)
  st_4 <- get_shapetrees_string(x4)
  st_5 <- get_shapetrees_string(x5)

  expect_equal(st_1, st_2)
  expect_equal(st_1, st_3)
  expect_equal(st_1, st_4)
  expect_equal(st_1, st_5)
})

library(officer)

doc <- read_docx()
doc <- body_add_par(doc, value = "blah blah blah", style = "Normal")

z <- body_append_start_context(doc)

for (i in seq_len(50)) {
  write_elements_to_context(
    context = z,
    fpar(
      "Hello World, ",
      i,
      fp_p = fp_par(word_style = "heading 1")
    ),
    fpar(run_pagebreak())
  )
}
doc <- body_append_stop_context(z)


print(doc, target = tempfile(fileext = ".docx"))

library(officer)

resolve_location("left")
resolve_location("right")
resolve_location("fullsize")
resolve_location("body")
resolve_location("body [1]")
resolve_location("<some label>")
resolve_location(2)
resolve_location(c(0, 0, 4, 5))

library(officer)

# This example demonstrates how to create
# an small document -----

## Create a new Word document
doc <- read_docx()
doc <- body_add_par(doc, "hello world")
## Save the document
output_file <- print(doc, target = tempfile(fileext = ".docx"))

# preview mode: save to temp file and open locally ----
## Not run:
# print(doc, preview = TRUE)

path <- system.file("doc_examples/example.pptx", package = "officer")
x <- read_pptx(path)

slide_visible(x) # get slide visibilities

x <- slide_visible(x, hide = 1:2) # hide slides 1 and 2
x <- slide_visible(x, show = 1:2) # make slides 1 and 2 visible
x <- slide_visible(x, show = 1:2, hide = 3)

slide_visible(x) <- FALSE # hide all slides
slide_visible(x) <- c(TRUE, FALSE, TRUE) # set each slide separately
slide_visible(x) <- c(TRUE, FALSE) # warns that rhs values are recycled

slide_visible(x)[2] <- TRUE # set 2nd slide to visible
slide_visible(x)[c(1, 3)] <- FALSE # 1st and 3rd slide
slide_visible(x)[c(1, 3)] <- c(FALSE, FALSE) # identical

# This example demonstrates how to create a document
# with text, formatted paragraphs, tables, and plots ----
# organized in sections

# Create a new Word document
doc <- read_docx()

# Add main title
doc <- body_add_par(doc, "Annual Sales Report", style = "heading 1")

# Add introduction with formatted text using fpar
intro_text <- fpar(
  "This report presents the ",
  ftext(
    "quarterly sales analysis",
    fp_text_lite(bold = TRUE, color = "#C32900")
  ),
  " for the fiscal year. The following sections provide detailed insights ",
  "into our performance metrics and trends."
)
doc <- body_add_fpar(doc, intro_text)

## Section 1: Sales Data ----
doc <- body_add_par(doc, "Sales Performance", style = "heading 2")

# Add descriptive text
doc <- body_add_par(
  doc,
  "The table below summarizes sales data across different product categories:"
)

# Create and add a data frame as a table
sales_data <- data.frame(
  Quarter = c("Q1", "Q2", "Q3", "Q4"),
  Revenue = c(125000, 142000, 156000, 178000),
  Units = c(1250, 1420, 1560, 1780),
  Growth = c("5%", "13.6%", "9.9%", "14.1%"),
  stringsAsFactors = FALSE
)
doc <- body_add_table(doc, value = sales_data, style = "table_template")

# Add commentary with multiple formatted text elements
comment <- fpar(
  "Key finding: ",
  ftext(
    "Q4 showed the strongest performance",
    fp_text_lite(bold = TRUE, font.size = 11)
  ),
  " with a ",
  ftext("14.1% growth rate", fp_text_lite(color = "#006699", bold = TRUE)),
  " compared to the previous quarter."
)
doc <- body_add_fpar(doc, comment)

## Section 2: Visualizations ----
doc <- body_add_par(doc, "Revenue Trends", style = "heading 2")

# Add explanatory text
doc <- body_add_par(
  doc,
  "Figure 1 illustrates the quarterly revenue progression throughout the year."
)

# Create a plot showing revenue trends
revenue_plot <- plot_instr({
  quarters <- c("Q1", "Q2", "Q3", "Q4")
  revenue <- c(125000, 142000, 156000, 178000)
  barplot(
    revenue,
    names.arg = quarters,
    col = "#4472C4",
    border = NA,
    main = "Quarterly Revenue",
    ylab = "Revenue ($)",
    xlab = "Quarter",
    ylim = c(0, 200000)
  )
  grid(nx = NA, ny = NULL, col = "gray90", lty = 1)
})
doc <- body_add_plot(doc, revenue_plot, width = 6, height = 4)

# Add another section with a different plot
doc <- body_add_par(doc, "Sales Distribution Analysis", style = "heading 2")

# Add context for the second plot
analysis_intro <- fpar(
  "The distribution analysis below shows the ",
  ftext("variability in daily sales", fp_text_lite(italic = TRUE)),
  " across all quarters. This helps identify patterns and outliers in our sales data."
)
doc <- body_add_fpar(doc, analysis_intro)

# Create a density plot
distribution_plot <- plot_instr({
  # Simulate daily sales data
  set.seed(123)
  daily_sales <- c(
    rnorm(90, mean = 1400, sd = 200), # Q1-Q3
    rnorm(30, mean = 2000, sd = 250) # Q4 (higher mean)
  )
  plot(
    density(daily_sales),
    main = "Distribution of Daily Sales",
    xlab = "Daily Sales (Units)",
    ylab = "Density",
    col = "#C32900",
    lwd = 2
  )
  polygon(density(daily_sales), col = rgb(0.76, 0.16, 0, 0.2), border = NA)
  abline(v = mean(daily_sales), col = "#006699", lwd = 2, lty = 2)
  legend("topright", legend = "Mean", col = "#006699", lty = 2, lwd = 2)
})
doc <- body_add_plot(doc, distribution_plot, width = 6, height = 4)

# Add concluding remarks
doc <- body_add_par(doc, "Conclusion", style = "heading 2")
conclusion <- fpar(
  "The analysis demonstrates ",
  ftext("consistent growth", fp_text_lite(bold = TRUE, color = "#006699")),
  " throughout the year, with particularly strong performance in Q4. ",
  "This trend suggests effective market strategies and increasing customer demand."
)
doc <- body_add_fpar(doc, conclusion)

# Save the document
comprehensive_file <- print(doc, target = tempfile(fileext = ".docx"))

#' Change ph labels in a layout
#'
#' There are two versions of the function. The first takes a set of key-value pairs to rename the
#' ph labels. The second uses a right hand side (rhs) assignment to specify the new ph labels.
#' See section *Details*. \cr\cr
#' _NB:_ You can also rename ph labels directly in PowerPoint. Open the master template view
#' (`Alt` + `F10`) and go to `Home` > `Arrange` > `Selection Pane`.
#'
#' @details
#' * Note the difference between the terms `id` and `index`. Both can be found in the output of
#' [layout_properties()]. The unique ph `id` is found in column `id`. The `index` refers to the
#' index of the data frame row.
#' * In a right hand side (rhs) label assignment (`<- new_labels`), there are two ways to
#' optionally specify a subset of phs to rename. In both cases, the length of the rhs vector
#' (the new labels) must match the length of the id or index:
#'   1. use the `id` argument to specify ph ids to rename: `layout_rename_ph_labels(..., id = 2:3) <- new_labels`
#'   2. use an `index` in squared brackets: `layout_rename_ph_labels(...)[1:2] <- new_labels`
#'
#' @export
#' @rdname layout_rename_ph_labels
#' @param x An `rpptx` object.
#' @param layout Layout name or index. Index is the row index of [layout_summary()].
#' @param master Name of master. Only required if the layout name is not unique across masters.
#' @param ... Comma separated list of key-value pairs to rename phs. Either reference a ph via its label
#' (`"old label"` = `"new label"`) or its unique id (`"id"` = `"new label"`).
#' @param .dots Provide a named list or vector of key-value pairs to rename phs
#' (`list("old label"` = `"new label"`).
#' @param id Unique placeholder id (see column `id` in [layout_properties()] or [plot_layout_properties()]).
#' @param value Not relevant for user. A pure technical necessity for rhs assignments.
#' @return Vector of renamed ph labels.
#' @example inst/examples/example_layout_rename_ph_labels.R
#'
layout_rename_ph_labels <- function(
  x,
  layout,
  master = NULL,
  ...,
  .dots = NULL
) {
  stop_if_not_rpptx(x, "x")
  dots <- list(...)
  dots <- c(dots, .dots)
  if (length(dots) > 0 && !is_named(dots)) {
    cli::cli_abort(
      c(
        "Unnamed arguments are not allowed.",
        "x" = "Arguments {.arg ...} and {.arg .dots} both require key value pairs."
      ),
      call = NULL
    )
  }

  l <- get_layout(x, layout, master)
  lp <- layout_properties(x, l$layout_name, l$master_name)
  if (length(dots) == 0) {
    return(lp$ph_label)
  }
  df_renames <- .rename_df_from_dots(lp, dots)
  .set_ph_labels(l, df_renames)
  reload_slidelayouts(x)

  lp <- layout_properties(x, l$layout_name, l$master_name)
  invisible(lp$ph_label)
}


#' @export
#' @rdname layout_rename_ph_labels
`layout_rename_ph_labels<-` <- function(
  x,
  layout,
  master = NULL,
  id = NULL,
  value
) {
  l <- get_layout(x, layout, master)
  lp <- layout_properties(x, l$layout_name, l$master_name)

  if (!is.null(id)) {
    if (length(id) != length(value)) {
      cli::cli_abort(
        c(
          "{.arg id} and rhs vector must have the same length",
          "x" = "Number of ids ({.val {length(id)}}) and assigned values ({.val {length(value)}}) differ"
        )
      )
    }
    wrong_ids <- setdiff(id, lp$id)
    n_wrong <- length(wrong_ids)
    if (n_wrong > 0) {
      cli::cli_abort(c(
        "{cli::qty(n_wrong)} {?This/These} id{?s} {?does/do} not exist: {.val {wrong_ids}}",
        "x" = "Choose one of: {.val {lp$id}}",
        "i" = cli::col_grey(
          "Also see {.code plot_layout_properties(..., '{l$layout_name}', '{l$master_nam}')}"
        )
      ))
    }
    .idx <- match(id, lp$id) # user might enter ids in arbitrary order
    lp$ph_label[.idx] <- value
    value <- lp$ph_label
  }
  names(value) <- lp$id
  df_renames <- .rename_df_from_dots(lp, value)
  .set_ph_labels(l, df_renames)
  reload_slidelayouts(x)
}


# heuristic: if a number, then treat as ph_id
.detect_ph_id <- function(x) {
  suppressWarnings({
    # avoid character to NA warning
    !is.na(as.numeric(x)) # nchar(x) == 1 &
  })
}


# create data frame with: ph_id, ph_label, ph_label_new as a
# basis for subsequent renaming
#
# CAVEAT: the ph order in layout_properties() (i.e. get_xfrm_data()) is reference for the user.
# Using the 'slide_layout' object's xfrm() method does not yield the same ph order!
# We need to guarantee a proper match here.
#
.rename_df_from_dots <- function(lp, dots) {
  lp <- lp[, c("id", "ph_label")]
  label_old <- names(dots)
  label_new <- as.character(dots)
  is_id <- .detect_ph_id(label_old)
  is_label <- !is_id

  # warn if renaming a duplicate label
  ii <- duplicated(lp$ph_label)
  dupes <- lp$ph_label[ii]
  dupes_used <- intersect(label_old, dupes)
  n_dupes_used <- length(dupes_used)
  if (n_dupes_used > 0) {
    cli::cli_warn(
      c(
        "When renaming a label with duplicates, only the first occurrence is renamed.",
        "x" = "Renaming {n_dupes_used} ph label{?s} with duplicates: {.val {dupes_used}}"
      ),
      call = NULL
    )
  }

  # check for duplicate renames
  is_dupe <- duplicated(label_old)
  if (any(is_dupe)) {
    dupes <- unique(label_old[is_dupe])
    n_dupes <- length(dupes)
    cli::cli_abort(
      c(
        "Each id or label must only have one rename entry only.",
        "x" = "Found {n_dupes} duplicate id{?s}/label{?s} to rename: {.val {dupes}}"
      ),
      call = NULL
    )
  }

  # match by label and check for unknown labels
  label_old_ <- label_old[is_label]
  row_idx_label <- match(label_old_, table = lp$ph_label)
  i_wrong <- is.na(row_idx_label)
  n_wrong <- sum(i_wrong)
  if (n_wrong > 0) {
    cli::cli_abort(
      c(
        "Can't rename labels that don't exist.",
        "x" = "{cli::qty(n_wrong)}{?This label does/These labels do} not exist: {.val {label_old_[i_wrong]}}"
      ),
      call = NULL
    )
  }

  # match by id and check for unknown ids
  id_old_ <- label_old[is_id]
  row_idx_id <- match(id_old_, table = lp$id)
  i_wrong <- is.na(row_idx_id)
  n_wrong <- sum(i_wrong)
  if (n_wrong > 0) {
    cli::cli_abort(
      c(
        "Can't rename ids that don't exist.",
        "x" = "{cli::qty(n_wrong)}{?This id does/These ids do} not exist: {.val {id_old_[i_wrong]}}"
      ),
      call = NULL
    )
  }

  # check for collision between label and id
  idx_collision <- intersect(row_idx_label, row_idx_id)
  n_collision <- length(idx_collision)
  if (n_collision > 0) {
    df <- lp[idx_collision, c("id", "ph_label")]
    pairs <- paste(df$ph_label, "<-->", df$id)
    cli::cli_abort(
      c(
        "Either specify the label {.emph OR} the id of the ph to rename, not both.",
        "x" = "These labels and ids collide: {.val {pairs}}"
      ),
      call = NULL
    )
  }

  lp$ph_label_new <- NA
  lp$ph_label_new[row_idx_label] <- label_new[is_label]
  lp$ph_label_new[row_idx_id] <- label_new[is_id]
  lp[!is.na(lp$ph_label_new), , drop = FALSE]
}


.set_ph_labels <- function(l, df_renames) {
  if (!inherits(l, "layout_info")) {
    cli::cli_abort(
      c(
        "{.arg l} must a a {.cls layout_info} object",
        "x" = "Got {.cls {class(l)[1]}} instead"
      ),
      call = NULL
    )
  }
  layout_xml <- l$slide_layout$get()
  for (i in seq_len(nrow(df_renames))) {
    cnvpr_node <- xml2::xml_find_first(
      layout_xml,
      sprintf("p:cSld/p:spTree/*/p:nvSpPr/p:cNvPr[@id='%s']", df_renames$id[i])
    )
    xml2::xml_set_attr(cnvpr_node, "name", df_renames$ph_label_new[i])
  }
  l$slide_layout$save() # persist changes in slide layout xml file
}

test_that("ph_labels with same order as in layout_properties()", {
  x <- read_pptx()
  l1 <- layout_rename_ph_labels(x, "Comparison")
  l2 <- layout_properties(x, "Comparison")$ph_label
  expect_equal(l1, l2)
})


test_that("incorrect inputs are detected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layout <- "Comparison"

  # unnamed args in renaming (dots)
  error_msg <- "Unnamed arguments are not allowed."
  expect_error(layout_rename_ph_labels(x, layout, NULL, "xxxx"), error_msg)
  expect_error(
    layout_rename_ph_labels(x, layout, .dots = list("xxxx")),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, NULL, "xxx", .dots = list(a = "xxxx")),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, "xxx" = "a", .dots = list("xxxx")),
    error_msg
  )

  # unknown labels
  error_msg <- "Can't rename labels that don't exist."
  expect_error(layout_rename_ph_labels(x, layout, "xxxx" = "..."), error_msg)
  expect_error(
    layout_rename_ph_labels(x, layout, "xxxx" = "...", "yyy" = "..."),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, .dots = list("xxxx" = "...")),
    error_msg
  )

  # unknown ids
  error_msg <- "Can't rename ids that don't exist."
  expect_error(layout_rename_ph_labels(x, layout, "1" = "..."), error_msg)
  expect_error(
    layout_rename_ph_labels(x, layout, "1" = "...", "0" = "..."),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, .dots = list("1" = "...")),
    error_msg
  )

  # duplicate rename entries
  error_msg <- "Each id or label must only have one rename entry only."
  expect_error(
    layout_rename_ph_labels(x, layout, "Title 1" = "a", "Title 1" = "b"),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, "2" = "a", "2" = "b"),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(
      x,
      layout,
      .dots = list("Title 1" = "a", "Title 1" = "b")
    ),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, "2" = "a", .dots = list("2" = "b")),
    error_msg
  )

  # label and id collision
  error_msg <- "Either specify the label OR the id of the ph to rename, not both."
  expect_error(
    layout_rename_ph_labels(x, layout, "Title 1" = "a", "2" = "b"),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(x, layout, "Title 1" = "a", "2" = "b"),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(
      x,
      layout,
      .dots = list("Title 1" = "a", "2" = "b")
    ),
    error_msg
  )
  expect_error(
    layout_rename_ph_labels(
      x,
      layout,
      "Date Placeholder 6" = "a",
      "7" = "b",
      .dots = list("Title 1" = "a", "2" = "b")
    ),
    error_msg
  )
})


test_that("ph renaming works as expected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layout <- "Comparison"

  # rename using key-value pairs: 'old label' = 'new label' or 'id' = 'new label'
  layout_rename_ph_labels(x, layout, "Title 1" = "LABEL MATCHED") # label matching
  layout_rename_ph_labels(x, layout, "3" = "ID MATCHED") # id matching
  layout_rename_ph_labels(
    x,
    layout,
    "Date Placeholder 6" = "DATE",
    "8" = "FOOTER"
  ) # label and id
  layout_properties(x, layout)$ph_label

  x <- read_pptx()
  idx <- c(1, 2, 6, 7)
  l <- list(
    "Date Placeholder 6" = "idx_6",
    "8" = "idx_7",
    "Title 1" = "idx_1",
    "3" = "idx_2"
  ) # as list
  layout_rename_ph_labels(x, layout, .dots = l)
  expect_equal(layout_properties(x, layout)$ph_label[idx], paste0("idx_", idx))

  x <- read_pptx()
  l <- c(
    "Date Placeholder 6" = "idx_6",
    "8" = "idx_7",
    "Title 1" = "idx_1",
    "3" = "idx_2"
  ) # as vector
  layout_rename_ph_labels(x, layout, .dots = l)
  expect_equal(layout_properties(x, layout)$ph_label[idx], paste0("idx_", idx))

  x <- read_pptx()
  l <- list("Date Placeholder 6" = "idx_6", "3" = "idx_2")
  layout_rename_ph_labels(
    x,
    layout,
    "8" = "idx_7",
    "Title 1" = "idx_1",
    .dots = l
  ) # mix ... and .dots
  expect_equal(layout_properties(x, layout)$ph_label[idx], paste0("idx_", idx))

  # rename via rhs assignment and ph index (not id!)
  x <- read_pptx()
  rhs <- LETTERS[1:8]
  layout_rename_ph_labels(x, layout) <- rhs
  expect_equal(layout_properties(x, layout)$ph_label, rhs)

  rhs <- paste("CHANGED", 1:3)
  ph_label_check <- layout_properties(x, layout)$ph_label
  ph_label_check[1:3] <- rhs
  layout_rename_ph_labels(x, layout)[1:3] <- rhs
  expect_equal(layout_properties(x, layout)$ph_label, ph_label_check)

  # rename via rhs assignment and ph id (not index)
  lp_old <- ph_label_check <- layout_properties(x, layout)
  ids <- c(2, 4, 5)
  idx <- match(ids, lp_old$id) # row in layout properties
  rhs <- paste("ID =", ids)
  ph_label_check <- lp_old$ph_label
  ph_label_check[idx] <- rhs
  layout_rename_ph_labels(x, layout, id = ids) <- rhs
  lp_new <- layout_properties(x, layout)
  expect_equal(lp_new$ph_label, ph_label_check)
})


test_that("renaming duplicate labels replaces 1st occurrence only", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  file <- test_path("docs_dir", "test-pptx-dedupe-ph.pptx")
  x <- read_pptx(file)

  # rename first label occurrence only and issue warning (1 duped label)
  layout <- "2-dupes"
  ph_label_check <- layout_properties(x, layout)$ph_label
  idx <- which(ph_label_check == "Content 7") # exists twice
  new_value <- "xxxx"
  warn_msg <- "When renaming a label with duplicates, only the first occurrence is renamed."
  expect_warning(
    layout_rename_ph_labels(x, layout, "Content 7" = new_value),
    warn_msg
  )
  ph_label_new <- layout_properties(x, layout)$ph_label
  ph_label_check[idx[1]] <- new_value # only first occurrence is replaced
  expect_equal(ph_label_check, ph_label_new)

  # rename first label occurrence only and issue warning (2 duped labels)
  layout <- "2x2-dupes"
  ph_label_check <- layout_properties(x, layout)$ph_label
  ii <- which(duplicated(ph_label_check, fromLast = TRUE)) # index of 1st occurrence
  dupes <- ph_label_check[ii]
  vals <- LETTERS[seq_along(dupes)]
  names(vals) <- dupes
  warn_msg_1 <- "When renaming a label with duplicates, only the first occurrence is renamed."
  warn_msg_2 <- "Renaming 2 ph labels with duplicates"
  warn_msg <- paste0(warn_msg_1, ".*", warn_msg_2)
  expect_warning(layout_rename_ph_labels(x, layout, .dots = vals), warn_msg)
  ph_label_check[ii] <- vals
  ph_label_new <- layout_properties(x, layout)$ph_label
  expect_equal(ph_label_check, ph_label_new)
})

test_that("ph_labels with same order as in layout_properties()", {
  x <- read_pptx()
  l1 <- layout_rename_ph_labels(x, "Comparison")
  l2 <- layout_properties(x, "Comparison")$ph_label
  expect_equal(l1, l2)
})


test_that("incorrect inputs are detected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layout <- "Comparison"

  # unnamed args in renaming (dots)
  error_msg <- "Unnamed arguments are not allowed."
  expect_error(layout_rename_ph_labels(x, layout, NULL, "xxxx"), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, .dots = list("xxxx")), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, NULL, "xxx", .dots = list(a = "xxxx")), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "xxx" = "a", .dots = list("xxxx")), error_msg)

  # unknown labels
  error_msg <- "Can't rename labels that don't exist."
  expect_error(layout_rename_ph_labels(x, layout, "xxxx" = "..."), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "xxxx" = "...", "yyy" = "..."), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, .dots = list("xxxx" = "...")), error_msg)

  # unknown ids
  error_msg <- "Can't rename ids that don't exist."
  expect_error(layout_rename_ph_labels(x, layout, "1" = "..."), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "1" = "...", "0" = "..."), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, .dots = list("1" = "...")), error_msg)

  # duplicate rename entries
  error_msg <- "Each id or label must only have one rename entry only."
  expect_error(layout_rename_ph_labels(x, layout, "Title 1" = "a", "Title 1" = "b"), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "2" = "a", "2" = "b"), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, .dots = list("Title 1" = "a", "Title 1" = "b")), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "2" = "a", .dots = list("2" = "b")), error_msg)

  # label and id collision
  error_msg <- "Either specify the label OR the id of the ph to rename, not both."
  expect_error(layout_rename_ph_labels(x, layout, "Title 1" = "a", "2" = "b"), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "Title 1" = "a", "2" = "b"), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, .dots = list("Title 1" = "a", "2" = "b")), error_msg)
  expect_error(layout_rename_ph_labels(x, layout, "Date Placeholder 6" = "a", "7" = "b", .dots = list("Title 1" = "a", "2" = "b")), error_msg)
})


test_that("ph renaming works as expected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layout <- "Comparison"

  # rename using key-value pairs: 'old label' = 'new label' or 'id' = 'new label'
  layout_rename_ph_labels(x, layout, "Title 1" = "LABEL MATCHED") # label matching
  layout_rename_ph_labels(x, layout, "3" = "ID MATCHED") # id matching
  layout_rename_ph_labels(x, layout, "Date Placeholder 6" = "DATE", "8" = "FOOTER") # label and id
  layout_properties(x, layout)$ph_label

  x <- read_pptx()
  idx <- c(1, 2, 6, 7)
  l <- list("Date Placeholder 6" = "idx_6", "8" = "idx_7", "Title 1" = "idx_1", "3" = "idx_2") # as list
  layout_rename_ph_labels(x, layout, .dots = l)
  expect_equal(layout_properties(x, layout)$ph_label[idx], paste0("idx_", idx))

  x <- read_pptx()
  l <- c("Date Placeholder 6" = "idx_6", "8" = "idx_7", "Title 1" = "idx_1", "3" = "idx_2") # as vector
  layout_rename_ph_labels(x, layout, .dots = l)
  expect_equal(layout_properties(x, layout)$ph_label[idx], paste0("idx_", idx))

  x <- read_pptx()
  l <- list("Date Placeholder 6" = "idx_6", "3" = "idx_2")
  layout_rename_ph_labels(x, layout, "8" = "idx_7", "Title 1" = "idx_1", .dots = l) # mix ... and .dots
  expect_equal(layout_properties(x, layout)$ph_label[idx], paste0("idx_", idx))

  # rename via rhs assignment and ph index (not id!)
  x <- read_pptx()
  rhs <- LETTERS[1:8]
  layout_rename_ph_labels(x, layout) <- rhs
  expect_equal(layout_properties(x, layout)$ph_label, rhs)

  rhs <- paste("CHANGED", 1:3)
  ph_label_check <- layout_properties(x, layout)$ph_label
  ph_label_check[1:3] <- rhs
  layout_rename_ph_labels(x, layout)[1:3] <- rhs
  expect_equal(layout_properties(x, layout)$ph_label, ph_label_check)

  # rename via rhs assignment and ph id (not index)
  lp_old <- ph_label_check <- layout_properties(x, layout)
  ids <- c(2, 4, 5)
  idx <- match(ids, lp_old$id) # row in layout properties
  rhs <- paste("ID =", ids)
  ph_label_check <- lp_old$ph_label
  ph_label_check[idx] <- rhs
  layout_rename_ph_labels(x, layout, id = ids) <- rhs
  lp_new <- layout_properties(x, layout)
  expect_equal(lp_new$ph_label, ph_label_check)
})


test_that("renaming duplicate labels replaces 1st occurrence only", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  file <- test_path("docs_dir", "test-pptx-dedupe-ph.pptx")
  x <- read_pptx(file)

  # rename first label occurrence only and issue warning (1 duped label)
  layout <- "2-dupes"
  ph_label_check <- layout_properties(x, layout)$ph_label
  idx <- which(ph_label_check == "Content 7") # exists twice
  new_value <- "xxxx"
  warn_msg <- "When renaming a label with duplicates, only the first occurrence is renamed."
  expect_warning(layout_rename_ph_labels(x, layout, "Content 7" = new_value), warn_msg)
  ph_label_new <- layout_properties(x, layout)$ph_label
  ph_label_check[idx[1]] <- new_value # only first occurrence is replaced
  expect_equal(ph_label_check, ph_label_new)

  # rename first label occurrence only and issue warning (2 duped labels)
  layout <- "2x2-dupes"
  ph_label_check <- layout_properties(x, layout)$ph_label
  ii <- which(duplicated(ph_label_check, fromLast = TRUE)) # index of 1st occurrence
  dupes <- ph_label_check[ii]
  vals <- LETTERS[seq_along(dupes)]
  names(vals) <- dupes
  warn_msg_1 <- "When renaming a label with duplicates, only the first occurrence is renamed."
  warn_msg_2 <- "Renaming 2 ph labels with duplicates"
  warn_msg <- paste0(warn_msg_1, ".*", warn_msg_2)
  expect_warning(layout_rename_ph_labels(x, layout, .dots = vals), warn_msg)
  ph_label_check[ii] <- vals
  ph_label_new <- layout_properties(x, layout)$ph_label
  expect_equal(ph_label_check, ph_label_new)
})

# worksheets ------------------------------------------------------------

worksheets <- R6Class(
  "worksheets",
  inherit = openxml_document,

  public = list(
    initialize = function(path) {
      super$initialize(character(0))
      private$package_dir <- path
      presentation_filename <- file.path(path, "xl", "workbook.xml")
      self$feed(presentation_filename)

      slide_df <- self$get_sheets_df()
      private$sheet_id <- slide_df$sheet_id
      private$sheet_rid <- slide_df$rid
      private$sheet_name <- slide_df$name
    },

    view_on_sheet = function(sheet) {
      sheet_id <- self$get_sheet_id(sheet)
      wb_view <- xml_find_first(self$get(), "d1:bookViews/d1:workbookView")
      xml_attr(wb_view, "activeTab") <- sheet_id - 1
      self$save()
    },

    add_sheet = function(target, label) {
      private$rels_doc$add(
        id = paste0("rId", private$rels_doc$get_next_id()),
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet",
        target = target
      )
      rels <- private$rels_doc$get_data()
      rid <- rels[rels$target %in% target, "id"]

      ids <- private$sheet_id
      new_id <- max(ids) + 1

      private$sheet_id <- c(private$sheet_id, new_id)
      private$sheet_rid <- c(private$sheet_rid, rid)
      private$sheet_name <- c(private$sheet_name, label)

      children_ <- xml_children(self$get())
      sheets_id <- which(sapply(children_, function(x) xml_name(x) == "sheets"))
      xml_list <- xml_children(children_[[sheets_id]])

      xml_elt <- paste(
        sprintf(
          "<sheet name=\"%s\" sheetId=\"%.0f\" r:id=\"%s\"/>",
          htmlEscapeCopy(private$sheet_name),
          private$sheet_id,
          private$sheet_rid
        ),
        collapse = ""
      )
      xml_elt <- paste0(
        "<sheets xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
        xml_elt,
        "</sheets>"
      )
      xml_elt <- as_xml_document(xml_elt)

      if (!inherits(xml_list, "xml_missing")) {
        xml_replace(children_[[sheets_id]], xml_elt)
      } else {
        stop("could not find sheets entity")
      }

      self
    },

    get_new_sheetname = function() {
      sheet_dir <- file.path(private$package_dir, "xl/worksheets")
      if (!file.exists(sheet_dir)) {
        dir.create(
          file.path(sheet_dir, "_rels"),
          showWarnings = FALSE,
          recursive = TRUE
        )
      }

      sheet_files <- basename(list.files(sheet_dir, pattern = "\\.xml$"))
      sheet_name <- "sheet1.xml"
      if (length(sheet_files)) {
        slide_index <- as.integer(gsub(
          "^(sheet)([0-9]+)(\\.xml)$",
          "\\2",
          sheet_files
        ))
        sheet_name <- gsub(
          pattern = "[0-9]+",
          replacement = max(slide_index) + 1,
          sheet_name
        )
      }
      sheet_name
    },

    sheet_names = function() {
      private$sheet_name
    },

    get_sheet_id = function(name) {
      sheets_df <- self$get_sheets_df()
      bool_name_in_list <- sheets_df$name %in% name
      n_matches <- sum(bool_name_in_list, na.rm = TRUE)
      if (n_matches < 1) {
        stop("could not find ", shQuote(name), " sheet", call. = FALSE)
      }
      sheets_df$sheet_id[bool_name_in_list]
    },

    get_sheets_df = function() {
      children_ <- xml_children(self$get())
      sheets_id <- which(sapply(children_, function(x) xml_name(x) == "sheets"))
      sheet_nodes <- xml_children(children_[[sheets_id]])
      data.frame(
        stringsAsFactors = FALSE,
        name = xml_attr(sheet_nodes, "name"),
        sheet_id = as.integer(xml_attr(sheet_nodes, "sheetId")),
        rid = xml_attr(sheet_nodes, "id")
      )
    }
  ),
  private = list(
    sheet_id = NULL,
    sheet_rid = NULL,
    sheet_name = NULL,
    package_dir = NULL
  )
)

# sheet ------------------------------------------------------------
sheet <- R6Class(
  "sheet",
  inherit = openxml_document,
  public = list(
    feed = function(file) {
      private$filename <- file
      private$doc <- read_xml(file)

      private$rels_filename <- file.path(
        dirname(file),
        "_rels",
        paste0(basename(file), ".rels")
      )

      if (file.exists(private$rels_filename)) {
        private$rels_doc <- relationship$new()$feed_from_xml(
          private$rels_filename
        )
      } else {
        new_rel <- relationship$new()
        new_rel$write(private$rels_filename)
        private$rels_doc <- new_rel
      }
      self
    }
  )
)


# dir_sheets ----
dir_sheet <- R6Class(
  "dir_sheet",
  public = list(
    initialize = function(x) {
      private$package_dir <- x$package_dir
      private$sheets_path <- "xl/worksheets"
      self$update()
    },

    update = function() {
      dir_ <- file.path(private$package_dir, private$sheets_path)
      filenames <- list.files(
        path = dir_,
        pattern = "\\.xml$",
        full.names = TRUE
      )

      private$collection <- lapply(
        filenames,
        function(x, path) {
          sheet$new(path)$feed(x)
        },
        private$sheets_path
      )

      names(private$collection) <- sapply(private$collection, function(x) {
        x$name()
      })
      self
    },

    get_sheet_list = function() {
      dir_ <- file.path(private$package_dir, private$sheets_path)
      filenames <- list.files(
        path = dir_,
        pattern = "\\.xml$",
        full.names = TRUE
      )
      sheet_index <- seq_along(filenames)
      if (length(filenames)) {
        filenames <- basename(filenames)
        sheet_index <- as.integer(gsub(
          "^(sheet)([0-9]+)(\\.xml)$",
          "\\2",
          filenames
        ))
        filenames <- filenames[order(sheet_index)]
      }
      filenames
    },

    get_sheet = function(id) {
      l_ <- self$length()
      if (is.null(id) || !between(id, 1, l_)) {
        stop("unvalid id ", id, " (", l_, " sheet(s))", call. = FALSE)
      }
      sheet_files <- self$get_sheet_list()
      index <- which(names(private$collection) == sheet_files[id])
      private$collection[[index]]
    },

    length = function() {
      length(private$collection)
    }
  ),
  private = list(
    collection = NULL,
    package_dir = NULL,
    sheets_path = NULL
  )
)


# read_xlsx ----
#' @export
#' @title Create an 'Excel' document object
#' @description Read and import an xlsx file as an R object
#' representing the document. This function is experimental.
#' @param path path to the xlsx file to use as base document.
#' @param x an rxlsx object
#' @examples
#' read_xlsx()
read_xlsx <- function(path = NULL) {
  if (!is.null(path) && !file.exists(path)) {
    stop("could not find file ", shQuote(path), call. = FALSE)
  }

  if (is.null(path)) {
    path <- system.file(package = "officer", "template/template.xlsx")
  }

  if (!grepl("\\.xlsx$", path, ignore.case = TRUE)) {
    stop("read_xlsx only support xlsx files", call. = FALSE)
  }

  package_dir <- tempfile()
  unpack_folder(file = path, folder = package_dir)

  obj <- structure(
    list(package_dir = package_dir),
    .Names = c("package_dir"),
    class = "rxlsx"
  )

  obj$content_type <- content_type$new(package_dir)
  obj$worksheets <- worksheets$new(package_dir)
  obj$sheets <- dir_sheet$new(obj)
  obj$core_properties <- read_core_properties(obj$package_dir)

  obj
}

#' @export
#' @title Add a sheet
#' @description Add a sheet into an xlsx worksheet.
#' @param x rxlsx object
#' @param label sheet label
#' @examples
#' my_ws <- read_xlsx()
#' my_pres <- add_sheet(my_ws, label = "new sheet")
add_sheet <- function(x, label) {
  if (label %in% x$worksheets$sheet_names()) {
    stop("sheet ", shQuote(label), " already exist")
  }

  new_slidename <- x$worksheets$get_new_sheetname()

  xml_file <- file.path(x$package_dir, "xl/worksheets", new_slidename)

  template_file <- system.file(package = "officer", "template/sheet.xml")
  file.copy(template_file, xml_file, copy.mode = FALSE)

  rel_filename <- file.path(
    dirname(xml_file),
    "_rels",
    paste0(basename(xml_file), ".rels")
  )
  dir.create(dirname(rel_filename), showWarnings = FALSE)
  template_rel_file <- system.file(
    package = "officer",
    "template/sheet.xml.rels"
  )
  file.copy(template_rel_file, rel_filename, copy.mode = FALSE)

  # update presentation elements
  x$worksheets$add_sheet(
    target = file.path("worksheets", new_slidename),
    label = label
  )

  partname <- file.path("/xl/worksheets", new_slidename)
  override <- setNames(
    "application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml",
    partname
  )
  x$content_type$add_override(value = override)

  x$sheets$update()

  sheet_select(x, sheet = label)
}

#' @export
#' @rdname read_xlsx
length.rxlsx <- function(x) {
  x$sheets$length()
}

#' @export
#' @title Select sheet
#' @description Set a particular sheet selected when workbook will be
#' edited.
#' @param x rxlsx object
#' @param sheet sheet name
#' @examples
#' my_ws <- read_xlsx()
#' my_pres <- add_sheet(my_ws, label = "new sheet")
#' my_pres <- sheet_select(my_ws, sheet = "new sheet")
#' print(my_ws, target = tempfile(fileext = ".xlsx") )
sheet_select <- function(x, sheet) {
  x$worksheets$view_on_sheet(sheet)
  x
}

#' @export
#' @param target path to the xlsx file to write
#' @param ... unused
#' @rdname read_xlsx
#' @examples
#' x <- read_xlsx()
#' print(x, target = tempfile(fileext = ".xlsx"))
print.rxlsx <- function(x, target = NULL, ...) {
  if (is.null(target)) {
    cat("xlsx document with", length(x), "sheet(s):\n")
    print(x$worksheets$sheet_names())
    return(invisible())
  }

  if (!grepl(x = target, pattern = "\\.(xlsx)$", ignore.case = TRUE)) {
    stop(target, " should have '.xlsx' extension.")
  }

  x$worksheets$save()
  x$content_type$save()

  x$core_properties['modified', 'value'] <- format(
    Sys.time(),
    "%Y-%m-%dT%H:%M:%SZ"
  )
  x$core_properties['lastModifiedBy', 'value'] <- Sys.getenv("USER")
  write_core_properties(x$core_properties, x$package_dir)

  invisible(pack_folder(folder = x$package_dir, target = target))
}

test_that("console printing", {
  x <- read_xlsx()
  expect_output(print(x), regexp = "^xlsx document with 1 sheet")
})

test_that("check extention and print document", {
  x <- read_xlsx()
  print(x, target = "print.xlsx")
  expect_true(file.exists("print.xlsx"))
  expect_error(print(x, target = "print.xlsxxxx"))
})


unlink("*.xlsx")

test_that("fp_text - print", {
  fp <- fp_text(font.size = 10)
  expect_output(print(fp))
})

test_that("fp_text - update", {
  fp <- fp_text(font.size = 10)

  fp <- update(fp, font.size = 20)
  expect_equal(fp$font.size, 20)
  fp <- update(fp, color = "red")
  expect_equal(fp$color, "red")
  fp <- update(fp, font.family = "Time New Roman")
  expect_equal(fp$font.family, "Time New Roman")
  fp <- update(fp, vertical.align = "superscript")
  expect_equal(fp$vertical.align, "superscript")
  fp <- update(fp, shading.color = "yellow")
  expect_equal(fp$shading.color, "yellow")
})

test_that("fp_text - print", {
  fp <- fp_text(font.size = 10)
  expect_output(print(fp))
})

test_that("fp_text - update", {

  fp <- fp_text(font.size = 10)

  fp <- update(fp, font.size = 20)
  expect_equal(fp$font.size, 20)
  fp <- update(fp, color = "red")
  expect_equal(fp$color, "red")
  fp <- update(fp, font.family = "Time New Roman")
  expect_equal(fp$font.family, "Time New Roman")
  fp <- update(fp, vertical.align = "superscript")
  expect_equal(fp$vertical.align, "superscript")
  fp <- update(fp, shading.color = "yellow")
  expect_equal(fp$shading.color, "yellow")
})

test_that("create and manipulate sheet", {
  doc <- read_xlsx()
  doc <- add_sheet(doc, label = "sheet1")
  doc <- sheet_select(doc, sheet = "sheet1")

  expect_true("sheet1" %in% doc$worksheets$sheet_names())

  xml_sheets <- list.files(
    file.path(doc$package_dir, "xl", "worksheets"),
    pattern = "\\.xml$"
  )
  expect_equal(length(doc), 2)
  expect_equal(xml_sheets, c("sheet1.xml", "sheet2.xml"))

  sheet_id <- doc$worksheets$get_sheet_id("sheet1")
  wb_view <- xml_find_first(
    doc$worksheets$get(),
    "d1:bookViews/d1:workbookView"
  )
  expect_equal(as.integer(xml_attr(wb_view, "activeTab")), sheet_id - 1)
})

test_that("settings works", {
  x <- read_docx()
  x <- docx_set_settings(
    x = x,
    zoom = 1,
    default_tab_stop = .5,
    hyphenation_zone = .25,
    decimal_symbol = ".",
    list_separator = ";",
    compatibility_mode = "15",
    even_and_odd_headers = TRUE,
    auto_hyphenation = FALSE
  )
  file <- print(x, target = tempfile(fileext = ".docx"))
  x <- read_docx(path = file)
  expect_equal(x$settings$zoom, 1)
  expect_equal(x$settings$list_separator, ";")
  expect_true(x$settings$even_and_odd_headers)
})

test_that("settings preserves existing XML elements", {
  template <- system.file(
    "doc_examples",
    "example.docx",
    package = "officer"
  )
  x <- read_docx(template)

  # read original settings.xml
  settings_before <- xml2::read_xml(
    file.path(x$package_dir, "word", "settings.xml")
  )
  tags_before <- xml2::xml_name(xml2::xml_children(settings_before))

  x <- docx_set_settings(x, zoom = 2)
  file <- print(x, target = tempfile(fileext = ".docx"))

  unpack_dir <- tempfile()
  unpack_folder(file, unpack_dir)
  settings_after <- xml2::read_xml(
    file.path(unpack_dir, "word", "settings.xml")
  )
  tags_after <- xml2::xml_name(xml2::xml_children(settings_after))

  # all original tags should still be present
  expect_true(all(tags_before %in% tags_after))

  # zoom should be updated
  zoom_node <- xml2::xml_child(settings_after, "w:zoom")
  expect_equal(xml2::xml_attr(zoom_node, "percent"), "200")
})

test_that("docx_embed_font embeds font files", {
  gdtools::register_liberationsans()
  sysfonts <- gdtools::sys_fonts()
  libsans <- sysfonts[sysfonts$family == "Liberation Sans", ]

  x <- read_docx()
  x <- docx_embed_font(
    x,
    font_family = "Liberation Sans",
    regular = libsans$path[libsans$style == "Regular"],
    bold = libsans$path[libsans$style == "Bold"]
  )
  file <- print(x, target = tempfile(fileext = ".docx"))

  unpack_dir <- tempfile()
  unpack_folder(file, unpack_dir)

  # odttf files created
  odttfs <- list.files(
    file.path(unpack_dir, "word", "fonts"),
    pattern = "[.]odttf$"
  )
  expect_equal(length(odttfs), 2L)

  # fontTable.xml has embed entries
  ft <- xml2::read_xml(
    file.path(unpack_dir, "word", "fontTable.xml")
  )
  ns <- c(w = "http://schemas.openxmlformats.org/wordprocessingml/2006/main")
  font_node <- xml2::xml_find_first(
    ft,
    "w:font[@w:name='Liberation Sans']",
    ns = ns
  )
  expect_false(inherits(font_node, "xml_missing"))
  embed_reg <- xml2::xml_child(font_node, "w:embedRegular")
  embed_bold <- xml2::xml_child(font_node, "w:embedBold")
  expect_false(inherits(embed_reg, "xml_missing"))
  expect_false(inherits(embed_bold, "xml_missing"))

  # fontTable.xml.rels has relationships
  rels <- xml2::read_xml(
    file.path(unpack_dir, "word", "_rels", "fontTable.xml.rels")
  )
  rel_nodes <- xml2::xml_find_all(
    rels,
    "d1:Relationship",
    ns = xml2::xml_ns(rels)
  )
  expect_equal(length(rel_nodes), 2L)

  # settings.xml has embedTrueTypeFonts
  settings <- xml2::read_xml(
    file.path(unpack_dir, "word", "settings.xml")
  )
  embed_node <- xml2::xml_child(settings, "w:embedTrueTypeFonts")
  expect_false(inherits(embed_node, "xml_missing"))
})

test_that("docx_embed_font auto-detects font files", {
  gdtools::register_liberationsans()
  x <- read_docx()
  expect_message(
    x <- docx_embed_font(x, font_family = "Liberation Sans"),
    "Liberation Sans"
  )
  file <- print(x, target = tempfile(fileext = ".docx"))
  unpack_dir <- tempfile()
  unpack_folder(file, unpack_dir)

  odttfs <- list.files(
    file.path(unpack_dir, "word", "fonts"),
    pattern = "[.]odttf$"
  )
  expect_gte(length(odttfs), 1L)

  # silent mode
  x2 <- read_docx()
  expect_no_message(
    suppressMessages(
      docx_embed_font(x2, font_family = "Liberation Sans")
    )
  )
})

test_that("docx_embed_font auto-detect errors on unknown family", {
  x <- read_docx()
  expect_error(
    docx_embed_font(x, font_family = "NonExistentFont12345"),
    "not found"
  )
})

test_that("docx_embed_font validates inputs", {
  x <- read_docx()
  expect_error(
    docx_embed_font(x, "Test", regular = "nonexistent.ttf"),
    "does not exist"
  )
  expect_error(
    docx_embed_font("not_rdocx", "Test", regular = tempfile()),
    "rdocx"
  )
})

#' @export
#' @title Create a 'PowerPoint' document object
#' @description Read and import a pptx file as an R object
#' representing the document.
#'
#' The function is called `read_pptx` because it allows you to initialize an
#' object of class `rpptx` from an existing PowerPoint file. Content will be
#' added to the existing presentation. By default, an empty document is used.
#'
#' @param path path to the pptx file to use as base document. `potx` file are supported.
#' @section master layouts and slide layouts:
#' `read_pptx()` uses a PowerPoint file as the initial document. This is the original
#' PowerPoint document where all slide layouts, placeholders for shapes and
#' styles come from. Major points to be aware of are:
#'
#'   * Slide layouts are relative to a master layout. A document can contain one or
#' more master layouts; a master layout can contain one or more slide layouts.
#' * A slide layout inherits design properties from its master layout but some
#' properties can be overwritten.
#' * Designs and formatting properties of layouts and shapes (placeholders in a
#'                                                            layout) are defined within the initial document. There is no R function to
#' modify these values - they must be defined in the initial document.
#' @examples
#' read_pptx()
#' @seealso [print.rpptx()], [add_slide()], [plot_layout_properties()], [ph_with()]
read_pptx <- function(path = NULL) {
  if (!is.null(path) && !file.exists(path)) {
    stop("could not find file ", shQuote(path), call. = FALSE)
  }

  if (is.null(path)) {
    path <- system.file(package = "officer", "template/template.pptx")
  }

  if (!grepl("\\.(pptx|potx)$", path, ignore.case = TRUE)) {
    stop("read_pptx only support pptx files", call. = FALSE)
  }

  package_dir <- tempfile()
  unpack_folder(file = path, folder = package_dir)

  obj <- list(package_dir = package_dir)

  obj$table_styles <- read_table_style(package_dir)

  obj$presentation <- presentation$new(package_dir)

  obj$masterLayouts <- dir_master$new(
    package_dir,
    slide_master$new("ppt/slideMasters")
  )

  obj$slideLayouts <- dir_layout$new(
    package_dir,
    master_metadata = obj$masterLayouts$get_metadata(),
    master_xfrm = obj$masterLayouts$xfrm()
  )

  obj$slide <- dir_slide$new(package_dir, obj$slideLayouts$get_xfrm_data())
  obj$notesSlide <- dir_notesSlide$new(package_dir)
  obj$notesMaster <- dir_notesMaster$new(
    package_dir,
    slide_master$new("ppt/notesMasters")
  )
  obj$content_type <- content_type$new(package_dir)
  obj$core_properties <- read_core_properties(package_dir)
  obj$doc_properties_custom <- read_custom_properties(package_dir)

  obj$rel <- relationship$new()
  obj$rel$feed_from_xml(file.path(package_dir, "_rels", ".rels"))

  obj$cursor = obj$slide$length()
  obj$layout_default <- list(layout = NA, master = NA)

  class(obj) <- "rpptx"
  obj
}


read_table_style <- function(path) {
  file <- file.path(path, "ppt/tableStyles.xml")
  if (!file.exists(file)) {
    warning("tableStyles.xml file does not exist in PPTX")
    return(NULL)
  }
  doc <- read_xml(file)
  nodes <- xml_find_all(doc, "//a:tblStyleLst")
  data.frame(
    def = xml_attr(nodes, "def"),
    styleName = xml_attr(nodes, "styleName"),
    stringsAsFactors = FALSE
  )
}


#' @title Write a 'PowerPoint' file.
#' @description Create a 'PowerPoint' file from an `rpptx` object (created by [read_pptx()]).
#' @param x an `rpptx` object.
#' @param target path to the .pptx file to write. If `target` is `NULL` (default), the `rpptx` object is printed to
#'   the console.
#' @param preview Save `x` to a temporary file and open it (default `FALSE`).
#' @param ... unused.
#' @returns If preview is `TRUE`, returns the temp file path invisibly.
#' @examples
#' # write an rpptx object to a .pptx file ----
#' file <- tempfile(fileext = ".pptx")
#' x <- read_pptx() # empty presentation, has no slides yet
#' print(x, target = file)
#'
#' # preview mode: save to temp file and open locally ----
#' \dontrun{
#' print(x, preview = TRUE)
#' }
#' @export
#' @seealso [read_pptx()]
print.rpptx <- function(x, target = NULL, preview = FALSE, ...) {
  if (preview) {
    file <- tempfile(fileext = ".pptx")
    print.rpptx(x, target = file, preview = FALSE)
    open_file(file)
    return(invisible(file))
  }
  if (is.null(target)) {
    cli::cli_text("pptx document with {.val {length(x)}} slide{?s}")
    cli::cli_text("Available layouts and their associated master(s):")
    df <- as.data.frame(layout_summary(x))
    if (has_layout_default(x)) {
      cli::cli_text(cli::col_grey("(*) = Default layout"))
      la <- get_layout_default(x)
      i <- which(df$layout == la$layout & df$master == la$master)
      df$" " <- " " # empty column name
      df$" "[i] <- "*"
    }
    print(df)
    return(invisible())
  }

  if (!grepl(x = target, pattern = "\\.(pptx)$", ignore.case = TRUE)) {
    cli::cli_abort("{.val {target}} should have '.pptx' extension.")
  }

  if (is_windows() && is_doc_open(target)) {
    cli::cli_abort(
      "{.val {target}} is open. To write to this document, please, close it."
    )
  }
  x <- pptx_fortify_slides(x)
  x$rel$write(file.path(x$package_dir, "_rels", ".rels"))

  # viewProps - drop lastView if available
  viewProps <- read_xml(file.path(x$package_dir, "ppt/viewProps.xml"))
  if (!is.na(xml_attr(viewProps, "lastView"))) {
    xml_attr(viewProps, "lastView") <- NULL
  }
  write_xml(viewProps, file.path(x$package_dir, "ppt/viewProps.xml"))

  drop_templatenode_from_app(x$package_dir)

  x$presentation$save()
  x$content_type$save()
  x$slide$save_slides()
  x$notesSlide$save_slides()

  x$core_properties["modified", "value"] <- format(
    Sys.time(),
    "%Y-%m-%dT%H:%M:%SZ"
  )
  x$core_properties["lastModifiedBy", "value"] <- Sys.getenv("USER")
  write_core_properties(x$core_properties, x$package_dir)

  if (nrow(x$doc_properties_custom$data) > 0) {
    write_custom_properties(x$doc_properties_custom, x$package_dir)
  }

  x <- sanitize_images(x, warn_user = FALSE)
  invisible(pack_folder(folder = x$package_dir, target = target))
}

test_that("missing tableStyles file in pptx from figshare", {
  testthat::skip_if_offline()
  testthat::skip_on_os("windows")

  destfile <- tempfile(fileext = ".pptx")
  on.exit(file.remove(destfile), add = TRUE)

  download.file(
    "https://ndownloader.figshare.com/files/16252631",
    destfile = destfile
  )

  expect_warning({
    x <- read_pptx(destfile)
  })
})

test_that("as.matrix.rpptx", {
  example_pptx <- test_path("docs_dir/table-complex.pptx")
  doc <- read_pptx(path = example_pptx)
  # Single table, without fill
  expect_equal(
    as.matrix(doc),
    matrix(
      c(
        "Header 1 A",
        NA,
        "B",
        "B",
        "C",
        "",
        "",
        "",
        "",
        "tutu",
        "Header 2",
        "12.23blah blah",
        "1.23",
        "9.0",
        "6",
        "",
        "",
        "",
        "",
        NA,
        "Header 3",
        NA,
        "blah blah blah",
        "Salut",
        "Hello",
        "sisi",
        NA,
        NA,
        NA,
        NA
      ),
      nrow = 10,
      ncol = 3
    )
  )
  # Single table, with fill
  expect_equal(
    as.matrix(doc, span = "fill"),
    matrix(
      c(
        "Header 1 A",
        "Header 1 A",
        "B",
        "B",
        "C",
        "",
        "",
        "",
        "",
        "tutu",
        "Header 2",
        "12.23blah blah",
        "1.23",
        "9.0",
        "6",
        "",
        "",
        "",
        "",
        "tutu",
        "Header 3",
        "12.23blah blah",
        "blah blah blah",
        "Salut",
        "Hello",
        "sisi",
        "sisi",
        "sisi",
        "sisi",
        "tutu"
      ),
      nrow = 10,
      ncol = 3
    )
  )
  # Single table, specify slide_id and id gets the first table
  expect_equal(
    as.matrix(doc, slide_id = 1, id = 18),
    as.matrix(doc)
  )
  # All tables, without fill
  expect_equal(
    as.matrix(doc, slide_id = NULL),
    list(
      "1" = list(
        "18" = matrix(
          c(
            "Header 1 A",
            NA,
            "B",
            "B",
            "C",
            "",
            "",
            "",
            "",
            "tutu",
            "Header 2",
            "12.23blah blah",
            "1.23",
            "9.0",
            "6",
            "",
            "",
            "",
            "",
            NA,
            "Header 3",
            NA,
            "blah blah blah",
            "Salut",
            "Hello",
            "sisi",
            NA,
            NA,
            NA,
            NA
          ),
          nrow = 10,
          ncol = 3
        )
      )
    )
  )
})

destfile = tempfile(fileext = ".pptx")
download.file("https://ndownloader.figshare.com/files/16252631",
              destfile = destfile)

test_that("missing tableStyles file in pptx from figshare", {
  testthat::skip_on_os("windows")
  expect_warning({
    x <- read_pptx(destfile)
  })

})

file.remove(destfile)

test_that("as.matrix.rpptx", {
  example_pptx <- test_path("docs_dir/table-complex.pptx")
  doc <- read_pptx(path = example_pptx)
  # Single table, without fill
  expect_equal(
    as.matrix(doc),
    matrix(
      c("Header 1 A", NA, "B", "B", "C", "", "", "", "", 
        "tutu", "Header 2", "12.23blah blah", "1.23", "9.0", "6", "", 
        "", "", "", NA, "Header 3", NA, "blah blah blah", "Salut", "Hello", 
        "sisi", NA, NA, NA, NA),
      nrow=10, ncol=3
    )
  )
  # Single table, with fill
  expect_equal(
    as.matrix(doc, span="fill"),
    matrix(
      c("Header 1 A", "Header 1 A", "B", "B", "C", "", "", 
        "", "", "tutu", "Header 2", "12.23blah blah", "1.23", "9.0", 
        "6", "", "", "", "", "tutu", "Header 3", "12.23blah blah", "blah blah blah", 
        "Salut", "Hello", "sisi", "sisi", "sisi", "sisi", "tutu"),
      nrow=10, ncol=3
    )
  )
  # Single table, specify slide_id and id gets the first table
  expect_equal(
    as.matrix(doc, slide_id=1, id=18),
    as.matrix(doc)
  )
  # All tables, without fill
  expect_equal(
    as.matrix(doc, slide_id=NULL),
    list("1"=list("18"=matrix(
      c("Header 1 A", NA, "B", "B", "C", "", "", "", "", 
        "tutu", "Header 2", "12.23blah blah", "1.23", "9.0", "6", "", 
        "", "", "", NA, "Header 3", NA, "blah blah blah", "Salut", "Hello", 
        "sisi", NA, NA, NA, NA),
      nrow=10, ncol=3
    )))
  )
})

check_dep <- function() {
  if (!requireNamespace("rmarkdown")) {
    stop("package rmarkdown is required to use this function", call. = FALSE)
  }
  if (!requireNamespace("knitr")) {
    stop("package knitr is required to use this function", call. = FALSE)
  }
}

get_default_pandoc_data_file <- function(format = "pptx") {
  outfile <- tempfile(fileext = paste0(".", format))

  pandoc_exec <- rmarkdown::pandoc_exec()
  if (.Platform$OS.type %in% "windows") {
    pandoc_exec <- paste0(pandoc_exec, ".exe")
  }

  if (!rmarkdown::pandoc_available() || !file.exists(pandoc_exec)) {
    # Use officer template when no pandoc - stupid fallback as if o pandoc, no result
    outfile <- system.file(
      package = "officer",
      "template",
      paste0("template.", format)
    )
  } else {
    # note in that case outfile will be removed when session will end
    ref_doc <- paste0("reference.", format)
    system2(
      pandoc_exec,
      args = c("--print-default-data-file", ref_doc),
      stdout = outfile
    )
  }

  return(outfile)
}

#' @export
#' @title Get the document being used as a template
#' @description Get filename of the document
#' being used as a template in an R Markdown document
#' rendered as HTML, PowerPoint presentation or Word document. It requires
#' packages rmarkdown >= 1.10.14 and knitr.
#' @param format document format, one of 'pptx', 'docx' or 'html'
#' @return a name file
#' @importFrom utils compareVersion packageVersion
#' @family functions for officer extensions
#' @keywords internal
get_reference_value <- function(format = NULL) {
  if (!is.null(format) && length(format) != 1) {
    stop("format must be a scalar character")
  }

  check_dep()

  if (
    compareVersion(as.character(packageVersion("rmarkdown")), "1.10.14") < 0
  ) {
    stop("package rmarkdown >= 1.10.14 is required to use this function")
  }

  if (is.null(format)) {
    if (grepl("docx", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "docx"
    } else if (grepl("pptx", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "pptx"
    } else if (grepl("html", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "html"
    } else if (grepl("latex", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "latex"
    } else {
      stop("Unable to determine the format that should be used")
    }
  }
  if (!format %in% c("pptx", "docx", "html")) {
    stop("format must be have value 'docx', 'pptx' or 'html'.")
  }

  output.dir <- knitr::opts_knit$get("output.dir")
  if (is.null(output.dir)) {
    output.dir <- getwd()
  }

  pandoc_args <- knitr::opts_knit$get("rmarkdown.pandoc.args")

  rd <- grep("--reference-doc", pandoc_args)
  if (length(rd)) {
    reference_data <- pandoc_args[rd + 1]
    if (!file.exists(reference_data)) {
      reference_data <- file.path(output.dir, reference_data)
    }
  } else {
    reference_data <- get_default_pandoc_data_file(format = format)
  }
  return(normalizePath(reference_data, winslash = "/"))
}


knitr_opts_current <- function(x, default = FALSE) {
  check_dep()
  x <- knitr::opts_current$get(x)
  if (is.null(x)) {
    x <- default
  }
  x
}

#' @export
#' @title Table options in a 'knitr' context
#' @description Get options for table rendering in a 'knitr'
#' context. It should not be used by the end user, but
#' its documentation should be read as a place where
#' table options are documented when 'knitr' is used.
#'
#' The function is a utility to facilitate the retrieval of table
#' options supported by the 'flextable', 'officedown' and 'officer' packages.
#'
#' These options should be set with `knitr::opts_chunk$set()`.
#' The names and expected values are listed in the following sections.
#'
#' @section knitr chunk options for table captions:
#'
#' | **label**                                        |    **name**     | **value**  |
#' |:-------------------------------------------------|:---------------:|:----------:|
#' | caption id/bookmark                              | tab.id          |    NULL    |
#' | caption                                          | tab.cap         |    NULL    |
#' | display table caption on top of the table or not | tab.topcaption  |    TRUE    |
#' | caption table sequence identifier.               | tab.lp          |   "tab:"   |
#'
#' @section knitr chunk options for Word table captions:
#'
#' | **label**                                               |    **name**     | **value**                 |
#' |:--------------------------------------------------------|:---------------:|:-------------------------:|
#' | Word stylename to use for table captions.               | tab.cap.style   |            NULL           |
#' | prefix for numbering chunk (default to   "Table ").     | tab.cap.pre     |           Table           |
#' | suffix for numbering chunk (default to   ": ").         | tab.cap.sep     |            " :"           |
#' | title number depth                                      | tab.cap.tnd     |              0            |
#' | separator to use between title number and table number. | tab.cap.tns     |             "-"           |
#' | caption prefix formatting properties                    | tab.cap.fp_text | fp_text_lite(bold = TRUE) |
#'
#' @section knitr chunk options for Word tables:
#'
#' | **label**                                                     |       **name**      | **value**  |
#' |:--------------------------------------------------------------|:-------------------:|:----------:|
#' | the Word stylename to use for tables                          | tab.style           |    NULL    |
#' | autofit' or 'fixed' algorithm.                                | tab.layout          |  "autofit" |
#' | value of the preferred width of the table in percent (base 1).| tab.width           |      1     |
#' | Alternative title text                                        | tab.alt.title       |    NULL    |
#' | Alternative description text                                  | tab.alt.description |    NULL    |
#'
#' @section knitr chunk options for data.frame with officedown:
#'
#' | **label**                                                     |    **name**     | **value**  |
#' |:--------------------------------------------------------------|:---------------:|:----------:|
#' | apply or remove formatting from the first row in the table    | first_row       |    TRUE    |
#' | apply or remove formatting from the first column in the table | first_column    |    FALSE   |
#' | apply or remove formatting from the last row in the table     | last_row        |    FALSE   |
#' | apply or remove formatting from the last column in the table  | last_column     |    FALSE   |
#' | don't display odd and even rows                               | no_hband        |    TRUE    |
#' | don't display odd and even columns                            | no_vband        |    TRUE    |
#'
#' @return a list
#'
#' @section returned elements:
#'
#' * cap.style (default: NULL)
#' * cap.pre (default: "Table ")
#' * cap.sep (default: ":")
#' * cap.tnd (default: 0)
#' * cap.tns (default: "-")
#' * cap.fp_text (default: `fp_text_lite(bold = TRUE)`)
#' * id (default: NULL)
#' * cap (default: NULL)
#' * alt.title (default: NULL)
#' * alt.description (default: NULL)
#' * topcaption (default: TRUE)
#' * style (default: NULL)
#' * tab.lp (default: "tab:")
#' * table_layout (default: "autofit")
#' * table_width (default: 1)
#' * first_row (default: TRUE)
#' * first_column (default: FALSE)
#' * last_row (default: FALSE)
#' * last_column (default: FALSE)
#' * no_hband (default: TRUE)
#' * no_vband (default: TRUE)
#' @family functions for officer extensions
#' @keywords internal
opts_current_table <- function() {
  is_bookdown <- isTRUE(knitr::opts_knit$get('bookdown.internal.label'))

  if (is_bookdown) {
    tab.lp <- knitr_opts_current("tab.lp", default = "tab:")
  } else {
    tab.lp <- knitr_opts_current("tab.lp", default = NULL)
  }
  tab.topcaption <- knitr_opts_current("tab.topcaption", default = TRUE)

  tab.cap.pre <- knitr_opts_current("tab.cap.pre", default = "Table")
  tab.cap.sep <- knitr_opts_current("tab.cap.sep", default = ":")

  tab.cap.pre <- paste0(tab.cap.pre, " ")
  tab.cap.sep <- paste0(tab.cap.sep, " ")

  tab.cap <- knitr_opts_current("tab.cap", default = NULL)

  if (is_bookdown) {
    tab.id <- knitr_opts_current("label", default = NULL)
    tab.id <- knitr_opts_current("tab.id", default = tab.id)
  } else {
    tab.id <- knitr_opts_current("tab.id", default = NULL)
  }

  tab.cap.style <- knitr_opts_current("tab.cap.style", default = NULL)
  tab.alt.title <- knitr_opts_current("tab.alt.title", default = NULL)
  tab.alt.description <- knitr_opts_current(
    "tab.alt.description",
    default = NULL
  )

  tab.cap.tnd <- knitr_opts_current("tab.cap.tnd", default = 0)
  tab.cap.tns <- knitr_opts_current("tab.cap.tns", default = "-")
  tab.cap.fp_text <- knitr_opts_current(
    "tab.cap.fp_text",
    default = fp_text_lite(bold = TRUE)
  )

  tab.style <- knitr_opts_current("tab.style", default = NULL)
  tab.layout <- knitr_opts_current("tab.layout", default = "autofit")
  tab.width <- knitr_opts_current("tab.width", default = 1)

  first_row <- knitr_opts_current("first_row", default = TRUE)
  first_column <- knitr_opts_current("first_column", default = FALSE)
  last_row <- knitr_opts_current("last_row", default = FALSE)
  last_column <- knitr_opts_current("last_column", default = FALSE)
  no_hband <- knitr_opts_current("no_hband", default = TRUE)
  no_vband <- knitr_opts_current("no_vband", default = TRUE)

  list(
    cap.style = tab.cap.style,
    cap.pre = tab.cap.pre,
    cap.sep = tab.cap.sep,
    cap.tnd = tab.cap.tnd,
    cap.tns = tab.cap.tns,
    cap.fp_text = tab.cap.fp_text,
    id = tab.id,
    alt.title = tab.alt.title,
    alt.description = tab.alt.description,
    topcaption = tab.topcaption,
    cap = tab.cap,
    style = tab.style,
    tab.lp = tab.lp,
    table_layout = table_layout(type = tab.layout),
    table_width = table_width(width = tab.width, unit = "pct"),
    first_row = first_row,
    first_column = first_column,
    last_row = last_row,
    last_column = last_column,
    no_hband = no_hband,
    no_vband = no_vband
  )
}

test_that("get_default_pandoc_data_file works as expected", {
  skip_if_not_installed("knitr")
  skip_if_not_installed("rmarkdown")
  skip_if_not(rmarkdown::pandoc_available())

  pptx_pandoc_template <- get_default_pandoc_data_file(
    format = "pptx"
  )
  expect_true(file.exists(pptx_pandoc_template))
  layout_summary_data <- layout_summary(read_pptx(pptx_pandoc_template))
  expect_gt(nrow(layout_summary_data), 0L)
  expect_contains(layout_summary_data$layout, "Title and Content")

  docx_pandoc_template <- get_default_pandoc_data_file(
    format = "docx"
  )
  expect_true(file.exists(docx_pandoc_template))
  doc <- read_docx(docx_pandoc_template)
  styles <- styles_info(doc)
  expect_contains(styles$style_name, "Normal")
  expect_contains(styles$style_name, "Author")
})

test_that("get_reference_value works as expected for docx", {
  skip_if_not_installed("knitr")
  skip_if_not_installed("rmarkdown")
  skip_if_not(rmarkdown::pandoc_available())

  tmp_file <- tempfile(fileext = ".Rmd")
  template_docx <- system.file(package = "officer", "template/template.docx")
  cat(
    "
```{r, include=FALSE}
library(officer)
knitr::knit_hooks$set(dummy = function(before, options, envir) {
  if (!before) {
    template_docx <- get_reference_value(format = \"docx\")
    z <- read_docx(template_docx)
    styles_data <- styles_info(z, type = \"paragraph\")[c(\"style_name\")]

    return(paste(\"```{=openxml}\", to_wml(block_table(styles_data, header = FALSE)), \"```\", sep = \"\n\"))
  }
})
```

```{r, dummy=TRUE}
```

",
    file = tmp_file
  )

  out_file <- tempfile(fileext = ".docx")

  rmarkdown::render(
    input = tmp_file,
    output_format = rmarkdown::word_document(reference_docx = template_docx),
    output_file = out_file,
    quiet = TRUE
  )

  z <- read_docx(out_file)
  content_summary <- docx_summary(z)

  styles <- styles_info(read_docx(), type = "paragraph")

  expect_equal(content_summary$text, styles$style_name)
})

test_that("get_reference_value works as expected for pptx", {
  skip_if_not_installed("knitr")
  skip_if_not_installed("rmarkdown")
  skip_if_not(rmarkdown::pandoc_available())

  pptx_pandoc_template <- get_default_pandoc_data_file(
    format = "pptx"
  )
  template_pptx <- tempfile(fileext = ".pptx")
  file.copy(pptx_pandoc_template, template_pptx)

  str <- "

```{r, include=FALSE}
library(officer)
knitr::knit_hooks$set(dummy = function(before, options, envir) {
  if (!before) {
    template_docx <- get_reference_value(format = \"pptx\")
    saveRDS(template_docx, \"~/dump.RDS\")
    return(\"\")
  }
})
```

```{r, dummy=TRUE}
```

"

  tmp_file <- tempfile(fileext = ".Rmd")
  cat(str, file = tmp_file)

  out_file <- tempfile(fileext = ".pptx")

  rmarkdown::render(
    input = tmp_file,
    output_format = rmarkdown::powerpoint_presentation(
      reference_doc = template_pptx
    ),
    output_file = out_file,
    quiet = TRUE
  )

  expect_equal(basename(readRDS("~/dump.RDS")), basename(template_pptx))
  unlink("~/dump.RDS")
})

test_that("opts_current_table works as expected", {
  skip_if_not_installed("knitr")
  skip_if_not_installed("rmarkdown")
  skip_if_not(rmarkdown::pandoc_available())

  new_dir <- tempfile()
  dir.create(new_dir, showWarnings = FALSE)
  rmd_fp <- file.path(new_dir, "knitr-utils.Rmd")
  generated_rds <- tempfile(fileext = ".RDS")
  file.copy("docs_dir/knitr-utils.Rmd", to = rmd_fp)
  rmarkdown::render(
    input = rmd_fp,
    envir = parent.frame(),
    params = list(
      rds_path = generated_rds
    ),
    quiet = TRUE
  )
  x <- readRDS(generated_rds)
  expect_equal(x$cap.pre, "Tableau ")
  expect_equal(x$cap.style, "Normal")
  expect_equal(x$cap.sep, " : ")
  expect_equal(x$alt.title, "alt title")
  expect_equal(x$alt.description, "alt description")
  expect_equal(x$cap, "Coucou")
})

# check properties helpers ----

check_set_geom <- function(x) {
  # http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html
  geom_types <- c(
    "line",
    "lineInv",
    "triangle",
    "rtTriangle",
    "rect",
    "diamond",
    "parallelogram",
    "trapezoid",
    "nonIsoscelesTrapezoid",
    "pentagon",
    "hexagon",
    "heptagon",
    "octagon",
    "decagon",
    "dodecagon",
    "star4",
    "star5",
    "star6",
    "star7",
    "star8",
    "star10",
    "star12",
    "star16",
    "star24",
    "star32",
    "roundRect",
    "round1Rect",
    "round2SameRect",
    "round2DiagRect",
    "snipRoundRect",
    "snip1Rect",
    "snip2SameRect",
    "snip2DiagRect",
    "plaque",
    "ellipse",
    "teardrop",
    "homePlate",
    "chevron",
    "pieWedge",
    "pie",
    "blockArc",
    "donut",
    "noSmoking",
    "rightArrow",
    "leftArrow",
    "upArrow",
    "downArrow",
    "stripedRightArrow",
    "notchedRightArrow",
    "bentUpArrow",
    "leftRightArrow",
    "upDownArrow",
    "leftUpArrow",
    "leftRightUpArrow",
    "quadArrow",
    "leftArrowCallout",
    "rightArrowCallout",
    "upArrowCallout",
    "downArrowCallout",
    "leftRightArrowCallout",
    "upDownArrowCallout",
    "quadArrowCallout",
    "bentArrow",
    "uturnArrow",
    "circularArrow",
    "leftCircularArrow",
    "leftRightCircularArrow",
    "curvedRightArrow",
    "curvedLeftArrow",
    "curvedUpArrow",
    "curvedDownArrow",
    "swooshArrow",
    "cube",
    "can",
    "lightningBolt",
    "heart",
    "sun",
    "moon",
    "smileyFace",
    "irregularSeal1",
    "irregularSeal2",
    "foldedCorner",
    "bevel",
    "frame",
    "halfFrame",
    "corner",
    "diagStripe",
    "chord",
    "arc",
    "leftBracket",
    "rightBracket",
    "leftBrace",
    "rightBrace",
    "bracketPair",
    "bracePair",
    "straightConnector1",
    "bentConnector2",
    "bentConnector3",
    "bentConnector4",
    "bentConnector5",
    "curvedConnector2",
    "curvedConnector3",
    "curvedConnector4",
    "curvedConnector5",
    "callout1",
    "callout2",
    "callout3",
    "accentCallout1",
    "accentCallout2",
    "accentCallout3",
    "borderCallout1",
    "borderCallout2",
    "borderCallout3",
    "accentBorderCallout1",
    "accentBorderCallout2",
    "accentBorderCallout3",
    "wedgeRectCallout",
    "wedgeRoundRectCallout",
    "wedgeEllipseCallout",
    "cloudCallout",
    "cloud",
    "ribbon",
    "ribbon2",
    "ellipseRibbon",
    "ellipseRibbon2",
    "leftRightRibbon",
    "verticalScroll",
    "horizontalScroll",
    "wave",
    "doubleWave",
    "plus",
    "flowChartProcess",
    "flowChartDecision",
    "flowChartInputOutput",
    "flowChartPredefinedProcess",
    "flowChartInternalStorage",
    "flowChartDocument",
    "flowChartMultidocument",
    "flowChartTerminator",
    "flowChartPreparation",
    "flowChartManualInput",
    "flowChartManualOperation",
    "flowChartConnector",
    "flowChartPunchedCard",
    "flowChartPunchedTape",
    "flowChartSummingJunction",
    "flowChartOr",
    "flowChartCollate",
    "flowChartSort",
    "flowChartExtract",
    "flowChartMerge",
    "flowChartOfflineStorage",
    "flowChartOnlineStorage",
    "flowChartMagneticTape",
    "flowChartMagneticDisk",
    "flowChartMagneticDrum",
    "flowChartDisplay",
    "flowChartDelay",
    "flowChartAlternateProcess",
    "flowChartOffpageConnector",
    "actionButtonBlank",
    "actionButtonHome",
    "actionButtonHelp",
    "actionButtonInformation",
    "actionButtonForwardNext",
    "actionButtonBackPrevious",
    "actionButtonEnd",
    "actionButtonBeginning",
    "actionButtonReturn",
    "actionButtonDocument",
    "actionButtonSound",
    "actionButtonMovie",
    "gear6",
    "gear9",
    "funnel",
    "mathPlus",
    "mathMinus",
    "mathMultiply",
    "mathDivide",
    "mathEqual",
    "mathNotEqual",
    "cornerTabs",
    "squareTabs",
    "plaqueTabs",
    "chartX",
    "chartStar",
    "chartPlus"
  )

  if (!x %in% geom_types) {
    stop(
      "'",
      x,
      "' must be a valid geometry\n\n",
      "A valid geometry has to be one of ",
      paste0(paste0("'", geom_types, "'"), collapse = ", "),
      "."
    )
  } else {
    return(x)
  }
}

# line ----

#' @title Line properties
#'
#' @description Create a `sp_line` object that describes
#' line properties.
#'
#' @param color line color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param lwd line width (in point) - 0 or positive integer value.
#' @param lty single character value specifying the line type.
#' Expected value is one of the following : default `'solid'`
#' or `'dot'` or `'dash'` or `'lgDash'`
#' or `'dashDot'` or `'lgDashDot'` or `'lgDashDotDot'`
#' or `'sysDash'` or `'sysDot'` or `'sysDashDot'`
#' or `'sysDashDotDot'`.
#' @param linecmpd single character value specifying the compound line type.
#' Expected value is one of the following : default `'sng'`
#' or `'dbl'` or `'tri'` or `'thinThick'`
#' or `'thickThin'`
#' @param lineend single character value specifying the line end style
#' Expected value is one of the following : default `'rnd'`
#' or `'sq'` or `'flat'`
#' @param linejoin single character value specifying the line join style
#' Expected value is one of the following : default `'round'`
#' or `'bevel'` or `'miter'`
#' @param headend a `sp_lineend` object specifying line head end style
#' @param tailend a `sp_lineend` object specifying line tail end style
#' @return a `sp_line` object
#' @example inst/examples/example_sp_line.R
#' @family functions for defining shape properties
#' @seealso [sp_lineend]
#' @export
sp_line <- function(
  color = "transparent",
  lwd = 1,
  lty = "solid",
  linecmpd = "sng",
  lineend = "rnd",
  linejoin = "round",
  headend = sp_lineend(type = "none"),
  tailend = sp_lineend(type = "none")
) {
  out <- list()
  out <- check_set_color(out, color)
  out <- check_set_numeric(out, lwd)
  out <- check_set_choice(
    obj = out,
    value = lty,
    choices = c(
      "solid",
      "dot",
      "dash",
      "lgDash",
      "dashDot",
      "lgDashDot",
      "lgDashDotDot",
      "sysDash",
      "sysDot",
      "sysDashDot",
      "sysDashDotDot"
    )
  )
  out <- check_set_choice(
    obj = out,
    value = linecmpd,
    choices = c("sng", "dbl", "tri", "thinThick", "thickThin")
  )
  out <- check_set_choice(
    obj = out,
    value = lineend,
    choices = c("rnd", "sq", "flat")
  )
  out <- check_set_choice(
    obj = out,
    value = linejoin,
    choices = c("round", "bevel", "miter")
  )
  out <- check_set_class(obj = out, value = headend, cl = "sp_lineend")
  out <- check_set_class(obj = out, value = tailend, cl = "sp_lineend")

  class(out) <- c("sp_line")

  out
}

#' @param x,object `sp_line` object
#' @param ... further arguments - not used
#' @rdname sp_line
#' @export
print.sp_line = function(x, ...) {
  out <- data.frame(
    color = x$color,
    lwd = x$lwd,
    lty = x$lty,
    linecmpd = x$linecmpd,
    lineend = x$lineend,
    linejoin = x$linejoin,
    headend = unclass(x$headend),
    tailend = unclass(x$tailend),
    stringsAsFactors = FALSE
  )
  print(out)
  invisible()
}

#' @rdname sp_line
#' @export
update.sp_line <- function(
  object,
  color,
  lwd,
  lty,
  linecmpd,
  lineend,
  linejoin,
  headend,
  tailend,
  ...
) {
  if (!missing(color)) {
    object <- check_set_color(object, color)
  }
  if (!missing(lwd)) {
    object <- check_set_numeric(object, lwd)
  }
  if (!missing(lty)) {
    object <- check_set_choice(
      obj = object,
      value = lty,
      choices = c(
        "solid",
        "dot",
        "dash",
        "lgDash",
        "dashDot",
        "lgDashDot",
        "lgDashDotDot",
        "sysDash",
        "sysDot",
        "sysDashDot",
        "sysDashDotDot"
      )
    )
  }
  if (!missing(linecmpd)) {
    object <- check_set_choice(
      obj = object,
      value = linecmpd,
      choices = c("sng", "dbl", "tri", "thinThick", "thickThin")
    )
  }
  if (!missing(lineend)) {
    object <- check_set_choice(
      obj = object,
      value = lineend,
      choices = c("rnd", "sq", "flat")
    )
  }

  if (!missing(linejoin)) {
    object <- check_set_choice(
      obj = object,
      value = linejoin,
      choices = c("round", "bevel", "miter")
    )
  }

  if (!missing(headend)) {
    object <- check_set_class(obj = object, value = headend, cl = "sp_lineend")
  }
  if (!missing(tailend)) {
    object <- check_set_class(obj = object, value = tailend, cl = "sp_lineend")
  }

  object
}

#' @title Line end properties
#'
#' @description Create a `sp_lineend` object that describes
#' line end properties.
#'
#' @param type single character value specifying the line end type.
#' Expected value is one of the following : default `'none'`
#' or `'triangle'` or `'stealth'` or `'diamond'`
#' or `'oval'` or `'arrow'`
#' @param width single character value specifying the line end width
#' Expected value is one of the following : default `'sm'`
#' or `'med'` or `'lg'`
#' @param length single character value specifying the line end length
#' Expected value is one of the following : default `'sm'`
#' or `'med'` or `'lg'`
#' @return a `sp_lineend` object
#' @example inst/examples/example_sp_lineend.R
#' @family functions for defining shape properties
#' @seealso [sp_line]
#' @export
sp_lineend <- function(type = "none", width = "med", length = "med") {
  out <- list()
  out <- check_set_choice(
    obj = out,
    value = type,
    choices = c("none", "triangle", "stealth", "diamond", "oval", "arrow")
  )
  out <- check_set_choice(
    obj = out,
    value = width,
    choices = c("sm", "med", "lg")
  )
  out <- check_set_choice(
    obj = out,
    value = length,
    choices = c("sm", "med", "lg")
  )

  class(out) <- c("sp_lineend")

  out
}

#' @param x,object `sp_lineend` object
#' @param ... further arguments - not used
#' @rdname sp_lineend
#' @export
print.sp_lineend = function(x, ...) {
  out <- data.frame(
    type = x$type,
    width = x$width,
    length = x$length
  )
  print(out)
  invisible()
}

#' @rdname sp_lineend
#' @export
update.sp_lineend <- function(object, type, width, length, ...) {
  if (!missing(type)) {
    object <- check_set_choice(
      obj = object,
      value = type,
      choices = c("none", "triangle", "stealth", "diamond", "oval", "arrow")
    )
  }
  if (!missing(width)) {
    object <- check_set_choice(
      obj = object,
      value = width,
      choices = c("sm", "med", "lg")
    )
  }
  if (!missing(length)) {
    object <- check_set_choice(
      obj = object,
      value = length,
      choices = c("sm", "med", "lg")
    )
  }

  object
}

test_that("solid_fill: builds a valid <a:solidFill> with srgbClr and alpha", {
  out <- solid_fill("red")
  expect_match(out, "^<a:solidFill>")
  expect_match(out, "</a:solidFill>$")
  expect_match(out, "<a:srgbClr val=\"FF0000\">")
  expect_match(out, "<a:alpha val=\"100000\"/>")
})

test_that("solid_fill: accepts standard '#hex' color strings", {
  expect_match(solid_fill("#3366CC"), "<a:srgbClr val=\"3366CC\">")
})

test_that("solid_fill: encodes alpha channel for semi-transparent colors", {
  half <- grDevices::rgb(1, 0, 0, alpha = 0.5)
  out <- solid_fill(half)
  # 0.5 round-trips through 8-bit (128/255) -> ~50196
  expect_match(out, "<a:alpha val=\"5[0-9]{4}\"/>")
})

test_that("solid_fill: 'transparent' produces alpha=0", {
  out <- solid_fill("transparent")
  expect_match(out, "<a:alpha val=\"0\"/>")
})

test_that("solid_fill: invalid color errors", {
  expect_error(solid_fill("not-a-color"))
})

test_that("to_pml.sp_line: emits <a:ln w=...> with solid fill and dash", {
  ln <- sp_line(color = "blue", lwd = 2, lty = "dash")
  out <- to_pml(ln)
  expect_match(out, "^<a:ln w=\"25400\"") # 2pt = 25400 EMU
  expect_match(out, "<a:srgbClr val=\"0000FF\">")
  expect_match(out, "<a:prstDash val=\"dash\"/>")
})

test_that("to_pml.sp_line: transparent or zero-width line emits <a:noFill/>", {
  ln <- sp_line(color = "transparent")
  out <- to_pml(ln)
  expect_match(out, "<a:noFill/>")
})

test_that("to_pml.sp_line: respects linecmpd, lineend, linejoin", {
  ln <- sp_line(
    color = "black",
    lwd = 1,
    linecmpd = "dbl",
    lineend = "sq",
    linejoin = "miter"
  )
  out <- to_pml(ln)
  expect_match(out, "cap=\"sq\"")
  expect_match(out, "cmpd=\"dbl\"")
  expect_match(out, "<a:miter/>")
})

test_that("to_pml.sp_line: head/tail line ends are emitted", {
  ln <- sp_line(
    color = "black",
    lwd = 1,
    headend = sp_lineend(type = "arrow", width = "lg", length = "med"),
    tailend = sp_lineend(type = "triangle")
  )
  out <- to_pml(ln)
  expect_match(out, "<a:headEnd type=\"arrow\" w=\"lg\" len=\"med\"/>")
  expect_match(out, "<a:tailEnd type=\"triangle\"")
})

test_that("add outline to placeholder", {
  sp_line_ <- list(
    sp_line(color = "#FFFF00", lwd = 4.5, lty = "solid", linecmpd = "sng"),
    sp_line(color = "black", lwd = 9, lty = "dash", linecmpd = "dbl")
  )
  width <- 4
  height <- 3
  top <- 7.5 / 2 - height / 2
  left <- c(1, 10 - width - 1)

  loc_ <- lapply(
    seq_along(sp_line_),
    function(x) {
      ph_location(
        left = left[x],
        top = top,
        width = width,
        height = height,
        bg = "orange",
        ln = sp_line_[[x]]
      )
    }
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Box with solid outline", location = loc_[[1]])
  doc <- ph_with(doc, "Box with dashed double outline", location = loc_[[2]])

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)
  expect_equal(
    sm$text,
    c("Box with solid outline", "Box with dashed double outline")
  )

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  cols <- xml_attr(xml_find_all(xmldoc, "//a:ln/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("FFFF00", "000000"))
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:ln"), "w"),
    as.character(c(4.5, 9) * 12700)
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:ln"), "cmpd"),
    c("sng", "dbl")
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:ln/a:prstDash"), "val"),
    c("solid", "dash")
  )
})

test_that("add geometry to placeholder", {
  sp_line_ <- sp_line(
    color = "#FFFF00",
    lwd = 4.5,
    lty = "solid",
    linecmpd = "sng"
  )

  width <- 4
  height <- 3
  top <- 7.5 / 2 - height / 2
  left <- c(1, 10 - width - 1)

  loc_ <- list(
    ph_location(
      left = left[1],
      top = top,
      width = width,
      height = height,
      bg = "orange",
      ln = sp_line_,
      geom = "roundRect"
    ),
    ph_location(
      left = left[2],
      top = top,
      width = width,
      height = height,
      bg = "steelblue",
      ln = sp_line_,
      geom = "trapezoid"
    )
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Round Rectangle", location = loc_[[1]])
  doc <- ph_with(doc, "Trapezoid", location = loc_[[2]])

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)
  expect_equal(sm$text, c("Round Rectangle", "Trapezoid"))

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  cols <- xml_attr(
    xml_find_all(xmldoc, "//p:spPr/a:solidFill/a:srgbClr"),
    "val"
  )
  expect_equal(cols, c("FFA500", "4682B4"))
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:prstGeom"), "prst"),
    c("roundRect", "trapezoid")
  )
  cols <- xml_attr(xml_find_all(xmldoc, "//a:ln/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("FFFF00", "FFFF00"))
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:ln"), "w"),
    as.character(c(4.5, 4.5) * 12700)
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:ln/a:prstDash"), "val"),
    c("solid", "solid")
  )
})

test_that("add line geometry with lineends", {
  headend <- sp_lineend(type = "oval", width = "sm", length = "sm")
  tailend <- sp_lineend(type = "triangle", width = "lg", length = "lg")

  sp_line1 <- sp_line(
    color = "steelblue",
    lwd = 4.5,
    lty = "solid",
    linecmpd = "sng",
    headend = headend,
    tailend = tailend
  )
  sp_line2 <- update(
    sp_line1,
    linecmpd = "dbl",
    headend = tailend,
    tailend = headend
  )

  width <- 4
  height <- 3
  top <- 7.5 / 2 - height / 2
  left <- c(1, 10 - width - 1)

  loc_ <- list(
    ph_location(
      left = left[1],
      top = top,
      width = width,
      height = height,
      ln = sp_line1,
      geom = "line"
    ),
    ph_location(
      left = left[2],
      top = top,
      width = width,
      height = height,
      ln = sp_line2,
      rotation = 180,
      geom = "line"
    )
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "", location = loc_[[1]])
  doc <- ph_with(doc, "", location = loc_[[2]])

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:prstGeom"), "prst"),
    c("line", "line")
  )
  cols <- xml_attr(xml_find_all(xmldoc, "//a:ln/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("4682B4", "4682B4"))
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:ln"), "cmpd"),
    c("sng", "dbl")
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:headEnd"), "type"),
    c("oval", "triangle")
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:tailEnd"), "type"),
    c("triangle", "oval")
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:headEnd"), "w"),
    c("sm", "lg")
  )
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//a:tailEnd"), "len"),
    c("lg", "sm")
  )
})

test_that("add outline to placeholder", {
  sp_line_ <- list(
    sp_line(color = "#FFFF00", lwd = 4.5, lty = "solid", linecmpd = "sng"),
    sp_line(color = "black", lwd = 9, lty = "dash", linecmpd = "dbl")
  )
  width <- 4; height <- 3; top <- 7.5 / 2 - height / 2
  left <- c(1, 10 - width - 1);

  loc_ <- lapply(seq_along(sp_line_), function(x) ph_location(left = left[x], top = top, width = width, height = height, bg = "orange", ln = sp_line_[[x]]))

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Box with solid outline", location = loc_[[1]])
  doc <- ph_with(doc, "Box with dashed double outline", location = loc_[[2]])

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)
  expect_equal(sm$text, c("Box with solid outline", "Box with dashed double outline"))

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  cols <- xml_attr(xml_find_all(xmldoc, "//a:ln/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("FFFF00", "000000") )
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:ln"), "w"), as.character(c(4.5, 9) * 12700))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:ln"), "cmpd"), c("sng", "dbl"))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:ln/a:prstDash"), "val"), c("solid", "dash"))
})

test_that("add geometry to placeholder", {
  sp_line_ <- sp_line(color = "#FFFF00", lwd = 4.5, lty = "solid", linecmpd = "sng")

  width <- 4; height <- 3; top <- 7.5 / 2 - height / 2
  left <- c(1, 10 - width - 1);

  loc_ <- list(
    ph_location(left = left[1], top = top, width = width, height = height, bg = "orange", ln = sp_line_, geom = "roundRect"),
    ph_location(left = left[2], top = top, width = width, height = height, bg = "steelblue", ln = sp_line_, geom = "trapezoid")
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "Round Rectangle", location = loc_[[1]])
  doc <- ph_with(doc, "Trapezoid", location = loc_[[2]])

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)
  expect_equal(sm$text, c("Round Rectangle", "Trapezoid"))

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  cols <- xml_attr(xml_find_all(xmldoc, "//p:spPr/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("FFA500", "4682B4") )
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:prstGeom"), "prst"), c("roundRect", "trapezoid"))
  cols <- xml_attr(xml_find_all(xmldoc, "//a:ln/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("FFFF00", "FFFF00") )
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:ln"), "w"), as.character(c(4.5, 4.5) * 12700))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:ln/a:prstDash"), "val"), c("solid", "solid"))
})

test_that("add line geometry with lineends", {

  headend <- sp_lineend(type = "oval", width = "sm", length = "sm")
  tailend <- sp_lineend(type = "triangle", width = "lg", length = "lg")

  sp_line1 <- sp_line(color = "steelblue", lwd = 4.5, lty = "solid", linecmpd = "sng", headend = headend, tailend = tailend)
  sp_line2 <- update(sp_line1, linecmpd = "dbl", headend = tailend, tailend = headend)

  width <- 4; height <- 3
  top <- 7.5 / 2 - height / 2; left <- c(1, 10 - width - 1)

  loc_ <- list(
    ph_location(left = left[1], top = top, width = width, height = height, ln = sp_line1, geom = "line"),
    ph_location(left = left[2], top = top, width = width, height = height, ln = sp_line2, rotation = 180, geom = "line")
  )

  doc <- read_pptx()
  doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
  doc <- ph_with(doc, "", location = loc_[[1]])
  doc <- ph_with(doc, "", location = loc_[[2]])

  sm <- slide_summary(doc)
  expect_equal(nrow(sm), 2)

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:prstGeom"), "prst"), c("line", "line"))
  cols <- xml_attr(xml_find_all(xmldoc, "//a:ln/a:solidFill/a:srgbClr"), "val")
  expect_equal(cols, c("4682B4", "4682B4") )
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:ln"), "cmpd"), c("sng", "dbl"))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:headEnd"), "type"), c("oval", "triangle"))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:tailEnd"), "type"), c("triangle", "oval"))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:headEnd"), "w"), c("sm", "lg"))
  expect_equal( xml_attr(xml_find_all(xmldoc, "//a:tailEnd"), "len"), c("lg", "sm"))
})

# utils ----
is_scalar_character <- function(x) {
  is.character(x) && length(x) == 1
}
is_scalar_logical <- function(x) {
  is.logical(x) && length(x) == 1
}


# functions ----

#' @export
#' @title Replace text at a bookmark location
#' @description Replace text content enclosed in a bookmark
#' with different text. A bookmark will be considered as valid if enclosing words
#' within a paragraph; i.e., a bookmark along two or more paragraphs is invalid,
#' a bookmark set on a whole paragraph is also invalid, but bookmarking few words
#' inside a paragraph is valid.
#' @param x a docx device
#' @param bookmark bookmark id
#' @param value the replacement string, of type character
#' @example inst/examples/example_body_replace_text_at_bkm.R
body_replace_text_at_bkm <- function(x, bookmark, value) {
  stopifnot(is_scalar_character(value), is_scalar_character(bookmark))
  xml_replace_text_at_bkm(
    node = x$doc_obj$get(),
    bookmark = bookmark,
    value = value
  )
  x
}


#' @export
#' @rdname body_replace_text_at_bkm
body_replace_img_at_bkm <- function(x, bookmark, value) {
  stopifnot(
    inherits(x, "rdocx"),
    is_scalar_character(bookmark),
    inherits(value, "external_img")
  )
  docxpart_replace_img_at_bkm(
    node = x$doc_obj$get(),
    bookmark = bookmark,
    value = value
  )
  x
}

xml_replace_text_at_bkm <- function(node, bookmark, value) {
  text <- enc2utf8(value)
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", bookmark)
  bm_start <- xml_find_first(node, xpath_)
  if (inherits(bm_start, "xml_missing")) {
    return(FALSE)
  }

  str_ <- sprintf(
    "//w:bookmarkStart[@w:name='%s']/following-sibling::w:r",
    bookmark
  )
  following_start <- sapply(xml_find_all(node, str_), xml_path)
  str_ <- sprintf(
    "//w:bookmarkEnd[@w:id='%s']/preceding-sibling::w:r",
    xml_attr(bm_start, "id")
  )
  preceding_end <- sapply(xml_find_all(node, str_), xml_path)

  match_path <- base::intersect(following_start, preceding_end)
  if (length(match_path) < 1) {
    stop(
      "could not find any bookmark ",
      bookmark,
      " located INSIDE a single paragraph"
    )
  }

  run_nodes <- xml_find_all(node, paste0(match_path, collapse = "|"))

  for (node in run_nodes[setdiff(seq_along(run_nodes), 1)]) {
    xml_remove(node)
  }

  xml_text(run_nodes[[1]]) <- text
  TRUE
}

docxpart_replace_img_at_bkm <- function(node, bookmark, value) {
  stopifnot(is_scalar_character(bookmark))
  stopifnot(inherits(value, "external_img"))

  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", bookmark)
  bm_start <- xml_find_first(node, xpath_)
  if (inherits(bm_start, "xml_missing")) {
    stop("cannot find bookmark ", shQuote(bookmark), call. = FALSE)
  }

  str_ <- sprintf(
    "//w:bookmarkStart[@w:name='%s']/following-sibling::w:r",
    bookmark
  )
  following_start <- sapply(xml_find_all(node, str_), xml_path)
  str_ <- sprintf(
    "//w:bookmarkEnd[@w:id='%s']/preceding-sibling::w:r",
    xml_attr(bm_start, "id")
  )
  preceding_end <- sapply(xml_find_all(node, str_), xml_path)

  match_path <- base::intersect(following_start, preceding_end)
  if (length(match_path) < 1) {
    stop(
      "could not find any bookmark ",
      bookmark,
      " located INSIDE a single paragraph"
    )
  }

  out <- to_wml(value, add_ns = TRUE)

  run_nodes <- xml_find_all(node, paste0(match_path, collapse = "|"))
  for (node in run_nodes[setdiff(seq_along(run_nodes), 1)]) {
    xml_remove(node)
  }
  xml_replace(run_nodes[[1]], as_xml_document(out))
}

#' @export
#' @rdname body_replace_text_at_bkm
headers_replace_text_at_bkm <- function(x, bookmark, value) {
  stopifnot(is_scalar_character(value), is_scalar_character(bookmark))
  for (header in x$headers) {
    xml_replace_text_at_bkm(
      node = header$get(),
      bookmark = bookmark,
      value = value
    )
  }
  x
}

#' @export
#' @rdname body_replace_text_at_bkm
headers_replace_img_at_bkm <- function(x, bookmark, value) {
  for (header in x$headers) {
    docxpart_replace_img_at_bkm(
      node = header$get(),
      bookmark = bookmark,
      value = value
    )
  }
  x
}


#' @export
#' @rdname body_replace_text_at_bkm
footers_replace_text_at_bkm <- function(x, bookmark, value) {
  stopifnot(is_scalar_character(value), is_scalar_character(bookmark))
  for (footer in x$footers) {
    xml_replace_text_at_bkm(
      node = footer$get(),
      bookmark = bookmark,
      value = value
    )
  }
  x
}

#' @export
#' @rdname body_replace_text_at_bkm
footers_replace_img_at_bkm <- function(x, bookmark, value) {
  for (footer in x$footers) {
    docxpart_replace_img_at_bkm(
      node = footer$get(),
      bookmark = bookmark,
      value = value
    )
  }
  x
}

#' @export
#' @title Add plots at bookmark location in a 'Word' document
#' @description
#' Use these functions if you want to replace a paragraph containing
#' a bookmark with a 'ggplot' or a base plot.
#' @param value a ggplot object for body_replace_gg_at_bkm() or a set plot instructions
#' body_replace_plot_at_bkm(), see plot_instr().
#' @param bookmark bookmark id
#' @param keep Should the bookmark be preserved? Defaults to `FALSE`,
#' i.e.the bookmark will be lost as a side effect.
#' @inheritParams body_add_gg
#'
#' @example inst/examples/example_body_replace_gg_at_bkm.R
body_replace_gg_at_bkm <- function(
  x,
  bookmark,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  keep = FALSE,
  ...
) {
  x <- cursor_bookmark(x, bookmark)
  x <- body_add_gg(
    x = x,
    value = value,
    width = width,
    height = height,
    res = res,
    style = style,
    scale = scale,
    pos = "on",
    ...
  )
  if (keep) {
    x <- body_bookmark(x, bookmark)
  }

  x
}

#' @export
#' @rdname body_replace_gg_at_bkm
body_replace_plot_at_bkm <- function(
  x,
  bookmark,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  keep = FALSE,
  ...
) {
  x <- cursor_bookmark(x, bookmark)
  x <- body_add_plot(
    x = x,
    value = value,
    width = width,
    height = height,
    res = res,
    style = style,
    pos = "on",
    ...
  )
  if (keep) {
    x <- body_bookmark(x, bookmark)
  }

  x
}

#' @export
#' @title Replace text anywhere in the document
#' @description Replace text anywhere in the document, or at a cursor.
#'
#' Replace all occurrences of old_value with new_value. This method
#' uses [grepl()]/[gsub()] for pattern matching; you may
#' supply arguments as required (and therefore use [regex()] features)
#' using the optional `...` argument.
#'
#' Note that by default, grepl/gsub will use `fixed=FALSE`, which means
#' that `old_value` and `new_value` will be interepreted as regular
#' expressions.
#'
#' **Chunking of text**
#'
#' Note that the behind-the-scenes representation of text in a Word document is
#' frequently not what you might expect! Sometimes a paragraph of text is broken
#' up (or "chunked") into several "runs," as a result of style changes, pauses
#' in text entry, later revisions and edits, etc. If you have not styled the
#' text, and have entered it in an "all-at-once" fashion, e.g. by pasting it or
#' by outputing it programmatically into your Word document, then this will
#' likely not be a problem. If you are working with a manually-edited document,
#' however, this can lead to unexpected failures to find text.
#'
#' You can use the officer function [docx_show_chunk()] to
#' show how the paragraph of text at the current cursor has been chunked into
#' runs, and what text is in each chunk. This can help troubleshoot unexpected
#' failures to find text.
#' @seealso [grepl()], [regex()], [docx_show_chunk()]
#' @author Frank Hangler, \email{frank@plotandscatter.com}
#' @param x a docx device
#' @param old_value the value to replace
#' @param new_value the value to replace it with
#' @param only_at_cursor if `TRUE`, only search-and-replace at the current
#' cursor; if `FALSE` (default), search-and-replace in the entire document
#' (this can be slow on large documents!)
#' @param warn warn if `old_value` could not be found.
#' @param ... optional arguments to grepl/gsub (e.g. `fixed=TRUE`)
#' @example inst/examples/example_body_replace_all_text.R
body_replace_all_text <- function(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
) {
  stopifnot(
    is_scalar_character(old_value),
    is_scalar_character(new_value),
    is_scalar_logical(only_at_cursor)
  )

  oldValue <- enc2utf8(old_value)
  newValue <- enc2utf8(new_value)

  replacement_count <- 0

  base_node <- if (only_at_cursor) {
    docx_current_block_xml(x)
  } else {
    docx_body_xml(x)
  }

  # For each matching text node...
  for (text_node in xml_find_all(base_node, ".//w:t")) {
    # ...if it contains the oldValue...
    if (grepl(oldValue, xml_text(text_node), ...)) {
      replacement_count <- replacement_count + 1
      # Replace the node text with the newValue.
      xml_text(text_node) <- gsub(oldValue, newValue, xml_text(text_node), ...)
    }
  }

  # Alert the user if no replacements were made.
  if (replacement_count == 0 && warn) {
    search_zone_text <- if (only_at_cursor) {
      "at the cursor."
    } else {
      "in the document."
    }
    warning("Found 0 instances of '", oldValue, "' ", search_zone_text)
  }

  x
}

#' @export
#' @title Show underlying text tag structure
#' @description Show the structure of text tags at the current cursor. This is
#' most useful when trying to troubleshoot search-and-replace functionality
#' using [body_replace_all_text()].
#' @seealso [body_replace_all_text()]
#' @param x a docx device
docx_show_chunk <- function(x) {
  cursor_elt <- docx_current_block_xml(x)
  text_nodes <- xml_find_all(cursor_elt, ".//w:t")
  msg <- paste0(length(text_nodes), " text nodes found at this cursor.")
  msg_detail <- ""
  for (text_node in text_nodes) {
    msg_detail <- paste0(
      msg_detail,
      paste0("\n  <w:t>: '", xml_text(text_node), "'")
    )
  }
  message(paste(msg, msg_detail))
  invisible(x)
}


#' @export
#' @rdname body_replace_all_text
#' @section header_replace_all_text:
#' Replacements will be performed in each header of all sections.
headers_replace_all_text <- function(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
) {
  stopifnot(
    is_scalar_character(old_value),
    is_scalar_character(new_value),
    is_scalar_logical(only_at_cursor)
  )

  oldValue <- enc2utf8(old_value)
  newValue <- enc2utf8(new_value)

  for (header in x$headers) {
    replacement_count <- 0

    base_node <- header$get()

    # For each matching text node...
    for (text_node in xml_find_all(base_node, ".//w:t")) {
      # ...if it contains the oldValue...
      if (grepl(oldValue, xml_text(text_node), ...)) {
        replacement_count <- replacement_count + 1
        # Replace the node text with the newValue.
        xml_text(text_node) <- gsub(
          oldValue,
          newValue,
          xml_text(text_node),
          ...
        )
      }
    }

    # Alert the user if no replacements were made.
    if (replacement_count == 0 && warn) {
      search_zone_text <- if (only_at_cursor) {
        "at the cursor."
      } else {
        "in the document."
      }
      warning("Found 0 instances of '", oldValue, "' ", search_zone_text)
    }
  }

  x
}
#' @export
#' @rdname body_replace_all_text
#' @section header_replace_all_text:
#' Replacements will be performed in each footer of all sections.
footers_replace_all_text <- function(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
) {
  stopifnot(
    is_scalar_character(old_value),
    is_scalar_character(new_value),
    is_scalar_logical(only_at_cursor)
  )

  oldValue <- enc2utf8(old_value)
  newValue <- enc2utf8(new_value)

  for (footer in x$footers) {
    replacement_count <- 0

    base_node <- footer$get()

    # For each matching text node...
    for (text_node in xml_find_all(base_node, ".//w:t")) {
      # ...if it contains the oldValue...
      if (grepl(oldValue, xml_text(text_node), ...)) {
        replacement_count <- replacement_count + 1
        # Replace the node text with the newValue.
        xml_text(text_node) <- gsub(
          oldValue,
          newValue,
          xml_text(text_node),
          ...
        )
      }
    }

    # Alert the user if no replacements were made.
    if (replacement_count == 0 && warn) {
      search_zone_text <- if (only_at_cursor) {
        "at the cursor."
      } else {
        "in the document."
      }
      warning("Found 0 instances of '", oldValue, "' ", search_zone_text)
    }
  }

  x
}

test_that("replace bkm with text in body", {
  doc <- read_docx()
  fp <- fpar(
    run_bookmark(
      ftext(
        "centered text"
      ),
      bkm = "text_to_replace"
    ),
    ftext(
      text = ". How are you",
      prop = fp_text(
        color = NA,
        font.size = NA,
        bold = TRUE,
        italic = NA,
        underlined = NA,
        font.family = NA_character_,
        cs.family = NA_character_,
        eastasia.family = NA_character_,
        hansi.family = NA_character_,
        shading.color = NA_character_
      )
    )
  )
  doc <- body_add_fpar(doc, value = fp, style = "centered")

  xmldoc <- doc$doc_obj$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "text_to_replace")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  doc <- body_replace_text_at_bkm(doc, "text_to_replace", "not left aligned")
  xmldoc <- doc$doc_obj$get()
  newtext <- xml_find_first(xmldoc, "w:body/w:p[1]/w:r")
  expect_equal(xml_text(newtext), "not left aligned")

  lasttext <- xml_find_first(xmldoc, "w:body/w:p[1]/w:r[2]")
  expect_equal(xml_text(lasttext), ". How are you")
})

test_that("replace bkm with images in header/footer", {
  template <- system.file(package = "officer", "doc_examples/example.docx")
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")

  doc <- read_docx(path = template)
  doc <- headers_replace_img_at_bkm(
    x = doc,
    bookmark = "bmk_header",
    value = external_img(src = img.file, width = .53, height = .7)
  )
  doc <- footers_replace_img_at_bkm(
    x = doc,
    bookmark = "bmk_footer",
    value = external_img(src = img.file, width = .53, height = .7)
  )
  print(doc, target = "test_replace_img.docx")

  doc <- read_docx(path = "test_replace_img.docx")

  xmldoc <- doc$headers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "bmk_header")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  blip <- xml_find_first(
    xmldoc,
    "//w:p/w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip"
  )
  expect_equal(xml_attr(blip, "embed"), "rId1")

  xmldoc <- doc$footers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "bmk_footer")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  blip <- xml_find_first(
    xmldoc,
    "//w:p/w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip"
  )
  expect_equal(xml_attr(blip, "embed"), "rId1")
})

test_that("replace bkm in headers and footers", {
  doc <- read_docx(path = "docs_dir/table-complex.docx")
  doc <- headers_replace_text_at_bkm(doc, "hello1", "salut")
  doc <- footers_replace_text_at_bkm(doc, "hello2", "salut")

  xmldoc <- doc$headers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "hello1")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))
  newtext <- xml_find_first(xmldoc, "/w:hdr/w:p[1]/w:r")
  expect_equal(xml_text(newtext), "salut")

  xmldoc <- doc$footers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "hello2")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))
  newtext <- xml_find_first(xmldoc, "/w:ftr/w:p[1]/w:r[3]")
  expect_equal(xml_text(newtext), "salut")
})

test_that("replace text at bkm with a ggplot", {
  testthat::skip_if_not(requireNamespace("ggplot2", quietly = TRUE))
  library("ggplot2")

  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))

  doc <- read_docx()
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot")
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot_with_bkm")

  xmldoc <- doc$doc_obj$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  doc <- body_replace_gg_at_bkm(x = doc, bookmark = "plot", value = gg_plot)

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false(inherits(child_, "xml_missing"))

  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_true(inherits(bm_start, "xml_missing"))

  # Keep bookmark
  doc <- body_replace_gg_at_bkm(
    x = doc,
    bookmark = "plot_with_bkm",
    value = gg_plot,
    keep = TRUE
  )
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot_with_bkm")
  bm_start <- xml_find_first(doc$doc_obj$get(), xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false(inherits(child_, "xml_missing"))
})

test_that("replace text at bkm with a plot", {
  doc <- read_docx()
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot")
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot_with_bkm")

  xmldoc <- doc$doc_obj$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  doc <- body_replace_plot_at_bkm(
    doc,
    bookmark = "plot",
    value = plot_instr(code = {
      barplot(1:5, col = 2:6)
    })
  )

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false(inherits(child_, "xml_missing"))

  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_true(inherits(bm_start, "xml_missing"))

  # Keep bookmark
  doc <- body_replace_plot_at_bkm(
    doc,
    bookmark = "plot_with_bkm",
    value = plot_instr(code = {
      barplot(1:5, col = 2:6)
    }),
    keep = TRUE
  )
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot_with_bkm")
  bm_start <- xml_find_first(doc$doc_obj$get(), xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false(inherits(child_, "xml_missing"))
})

test_that("docx replace text", {
  doc <- read_docx()
  doc <- body_add_par(doc, "Placeholder one")
  doc <- body_add_par(doc, "Placeholder two")
  doc <- body_replace_all_text(
    doc,
    old_value = "placeholder",
    new_value = "new",
    only_at_cursor = FALSE,
    ignore.case = TRUE
  )
  xmldoc <- doc$doc_obj$get()
  expect_equal(xml_text(xml_find_all(xmldoc, "//w:p")), c("new one", "new two"))
})

test_that("docx replace all text", {
  doc <- read_docx(path = "docs_dir/table-complex.docx")

  doc <- headers_replace_all_text(
    doc,
    old_value = "hello",
    new_value = "salut",
    only_at_cursor = FALSE,
    ignore.case = TRUE
  )
  doc <- footers_replace_all_text(
    doc,
    old_value = "hello",
    new_value = "salut",
    only_at_cursor = FALSE,
    ignore.case = TRUE
  )

  xmldoc <- doc$headers[[1]]$get()
  expect_equal(xml_text(xml_find_all(xmldoc, "//w:p")), "salut world")

  xmldoc <- doc$footers[[1]]$get()
  expect_equal(xml_text(xml_find_all(xmldoc, "//w:p")), "world salut")
})

unlink("*.docx", force = TRUE)

test_that("replace bkm with text in body", {
  doc <- read_docx()
  fp <- fpar(
    run_bookmark(
      ftext(
        "centered text"
      ),
      bkm = "text_to_replace"
    ),
    ftext(
      text = ". How are you",
      prop = fp_text(
        color = NA, font.size = NA, bold = TRUE, italic = NA,
        underlined = NA, font.family = NA_character_,
        cs.family = NA_character_, eastasia.family = NA_character_,
        hansi.family = NA_character_, shading.color = NA_character_
      )
    )
  )
  doc <- body_add_fpar(doc, value = fp, style = "centered")

  xmldoc <- doc$doc_obj$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "text_to_replace")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  doc <- body_replace_text_at_bkm(doc, "text_to_replace", "not left aligned")
  xmldoc <- doc$doc_obj$get()
  newtext <- xml_find_first(xmldoc, "w:body/w:p[1]/w:r")
  expect_equal(xml_text(newtext), "not left aligned")

  lasttext <- xml_find_first(xmldoc, "w:body/w:p[1]/w:r[2]")
  expect_equal(xml_text(lasttext), ". How are you")
})

test_that("replace bkm with images in header/footer", {
  template <- system.file(package = "officer", "doc_examples/example.docx")
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")

  doc <- read_docx(path = template)
  doc <- headers_replace_img_at_bkm(
    x = doc, bookmark = "bmk_header",
    value = external_img(src = img.file, width = .53, height = .7)
  )
  doc <- footers_replace_img_at_bkm(
    x = doc, bookmark = "bmk_footer",
    value = external_img(src = img.file, width = .53, height = .7)
  )
  print(doc, target = "test_replace_img.docx")

  doc <- read_docx(path = "test_replace_img.docx")

  xmldoc <- doc$headers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "bmk_header")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  blip <- xml_find_first(xmldoc, "//w:p/w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip")
  expect_equal(xml_attr(blip, "embed"), "rId1")

  xmldoc <- doc$footers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "bmk_footer")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  blip <- xml_find_first(xmldoc, "//w:p/w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip")
  expect_equal(xml_attr(blip, "embed"), "rId1")
})

test_that("replace bkm in headers and footers", {
  doc <- read_docx(path = "docs_dir/table-complex.docx")
  doc <- headers_replace_text_at_bkm(doc, "hello1", "salut")
  doc <- footers_replace_text_at_bkm(doc, "hello2", "salut")

  xmldoc <- doc$headers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "hello1")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))
  newtext <- xml_find_first(xmldoc, "/w:hdr/w:p[1]/w:r")
  expect_equal(xml_text(newtext), "salut")

  xmldoc <- doc$footers[[1]]$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "hello2")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))
  newtext <- xml_find_first(xmldoc, "/w:ftr/w:p[1]/w:r[3]")
  expect_equal(xml_text(newtext), "salut")
})

test_that("replace text at bkm with a ggplot", {
  testthat::skip_if_not(requireNamespace("ggplot2", quietly = TRUE))
  library("ggplot2")

  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))

  doc <- read_docx()
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot")
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot_with_bkm")

  xmldoc <- doc$doc_obj$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  doc <- body_replace_gg_at_bkm(x = doc, bookmark = "plot", value = gg_plot)

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false( inherits(child_, "xml_missing") )

  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_true(inherits(bm_start, "xml_missing"))

  # Keep bookmark
  doc <- body_replace_gg_at_bkm(x = doc, bookmark = "plot_with_bkm",
                                value = gg_plot, keep = TRUE)
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot_with_bkm")
  bm_start <- xml_find_first(doc$doc_obj$get(), xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false( inherits(child_, "xml_missing") )
})

test_that("replace text at bkm with a plot", {
  doc <- read_docx()
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot")
  doc <- body_add_par(doc, "text_to_replace")
  doc <- body_bookmark(doc, "plot_with_bkm")

  xmldoc <- doc$doc_obj$get()
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot")
  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  doc <- body_replace_plot_at_bkm(
    doc,
    bookmark = "plot",
    value = plot_instr(code = { barplot(1:5, col = 2:6) })
  )

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false( inherits(child_, "xml_missing") )

  bm_start <- xml_find_first(xmldoc, xpath_)
  expect_true(inherits(bm_start, "xml_missing"))

  # Keep bookmark
  doc <- body_replace_plot_at_bkm(
    doc,
    bookmark = "plot_with_bkm",
    value = plot_instr(code = { barplot(1:5, col = 2:6) }),
    keep = TRUE
  )
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", "plot_with_bkm")
  bm_start <- xml_find_first(doc$doc_obj$get(), xpath_)
  expect_false(inherits(bm_start, "xml_missing"))

  node <- docx_current_block_xml(doc)
  child_ <- xml_child(node, "w:r/w:drawing")
  expect_false( inherits(child_, "xml_missing") )
})

test_that("docx replace text", {
  doc <- read_docx()
  doc <- body_add_par(doc, "Placeholder one")
  doc <- body_add_par(doc, "Placeholder two")
  doc <- body_replace_all_text(doc,
    old_value = "placeholder", new_value = "new",
    only_at_cursor = FALSE, ignore.case = TRUE
  )
  xmldoc <- doc$doc_obj$get()
  expect_equal(xml_text(xml_find_all(xmldoc, "//w:p")), c("new one", "new two"))
})

test_that("docx replace all text", {
  doc <- read_docx(path = "docs_dir/table-complex.docx")

  doc <- headers_replace_all_text(doc,
    old_value = "hello", new_value = "salut",
    only_at_cursor = FALSE, ignore.case = TRUE
  )
  doc <- footers_replace_all_text(doc,
    old_value = "hello", new_value = "salut",
    only_at_cursor = FALSE, ignore.case = TRUE
  )

  xmldoc <- doc$headers[[1]]$get()
  expect_equal(xml_text(xml_find_all(xmldoc, "//w:p")), "salut world")

  xmldoc <- doc$footers[[1]]$get()
  expect_equal(xml_text(xml_find_all(xmldoc, "//w:p")), "world salut")
})

unlink("*.docx", force = TRUE)

# Using a custom template ----
# This example shows how to start from an existing template
# instead of creating a blank document

# Get the path to a landscape template included in the package
template <- system.file(package = "officer", "doc_examples", "landscape.docx")

# Create a document based on the template
# The document will inherit the template's styles and page settings
doc_2 <- read_docx(path = template)

# Add a section with a table
doc_2 <- body_add_par(doc_2, "Motor Trend Car Data", style = "heading 2")
doc_2 <- body_add_table(doc_2, value = head(mtcars))

# Add a section with a plot
doc_2 <- body_add_par(doc_2, "Sales Distribution", style = "heading 2")
doc_2 <- body_add_plot(doc_2, distribution_plot)

# Save the document
docx_file_output <- print(doc_2, target = tempfile(fileext = ".docx"))

# generics -----

#' @export
#' @title Convert officer objects to WordprocessingML
#' @description Convert an object made with package officer
#' to WordprocessingML. The result is a string.
#' @param x object
#' @param add_ns should namespace be added to the top tag
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @keywords internal
to_wml <- function(x, add_ns = FALSE, ...) {
  UseMethod("to_wml")
}

#' @export
#' @title Convert officer objects to PresentationML
#' @description Convert an object made with package officer
#' to PresentationML. The result is a string.
#' @param x object
#' @param add_ns should namespace be added to the top tag
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @keywords internal
to_pml <- function(x, add_ns = FALSE, ...) {
  UseMethod("to_pml")
}

#' @export
#' @title Convert officer objects to HTML
#' @description Convert an object made with package officer
#' to HTML. The result is a string.
#' @param x object
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @keywords internal
to_html <- function(x, ...) {
  UseMethod("to_html")
}


# bookmark -----

bookmark <- function(id, str) {
  new_id <- uuid_generate()
  bm_start_str <- sprintf(
    "<w:bookmarkStart w:id=\"%s\" w:name=\"%s\"/>",
    new_id,
    id
  )
  bm_start_end <- sprintf("<w:bookmarkEnd w:id=\"%s\"/>", new_id)
  paste0(bm_start_str, str, bm_start_end)
}

# ftext -----
#' @export
#' @title Formatted chunk of text
#' @description Format a chunk of text with text formatting properties (bold, color, ...).
#' The function allows you to create pieces of text formatted the way you want.
#' @param text text value, a single character value
#' @param prop formatting text properties returned by [fp_text]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @section usage:
#' You can use this function in conjunction with [fpar] to create paragraphs
#' consisting of differently formatted text parts. You can also use this
#' function as an *r chunk* in an R Markdown document made with package
#' officedown.
#' @examples
#' ftext("hello", fp_text())
#'
#' properties1 <- fp_text(color = "red")
#' properties2 <- fp_text(bold = TRUE, shading.color = "yellow")
#' ftext1 <- ftext("hello", properties1)
#' ftext2 <- ftext("World", properties2)
#' paragraph <- fpar(ftext1, " ", ftext2)
#'
#' x <- read_docx()
#' x <- body_add(x, paragraph)
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [fp_text]
#' @family run functions for reporting
ftext <- function(text, prop = NULL) {
  if (is.character(text)) {
    value <- enc2utf8(text)
  } else {
    value <- formatC(text)
  }

  out <- list(value = value, pr = prop)
  class(out) <- c("ftext", "cot", "run")
  out
}

#' @export
print.ftext <- function(x, ...) {
  cat("text: ", x$value, "\nformat:\n", sep = "")
  print(x$pr)
}

#' @export
to_wml.ftext <- function(x, add_ns = FALSE, ...) {
  out <- paste0(
    "<w:r>",
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:t xml:space=\"preserve\">",
    htmlEscapeCopy(x$value),
    "</w:t></w:r>"
  )
  out
}

#' @export
to_pml.ftext <- function(x, add_ns = FALSE, ...) {
  open_tag <- ar_ns_no
  if (add_ns) {
    open_tag <- ar_ns_yes
  }
  paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_pml(x$pr),
    "<a:t>",
    htmlEscapeCopy(x$value),
    "</a:t></a:r>"
  )
}

#' @export
to_html.ftext <- function(x, ...) {
  sprintf(
    "<span style=\"%s\">%s</span>",
    if (!is.null(x$pr)) rpr_css(x$pr) else "",
    htmlEscapeCopy(x$value)
  )
}


# Word chunk ----
#' @export
#' @title Word chunk of text with a style
#' @description Format a chunk of text associated with a 'Word' character style.
#' The style is defined with its unique identifer.
#' @param text text value, a single character value
#' @param style_id 'Word' unique style identifier associated with the style to use.
#' @examples
#' run1 <- run_wordtext("hello", "DefaultParagraphFont")
#' paragraph <- fpar(run1)
#'
#' x <- read_docx()
#' x <- body_add_fpar(x, paragraph)
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [ftext()]
#' @family run functions for reporting
run_wordtext <- function(text, style_id = NULL) {
  if (is.character(text)) {
    value <- enc2utf8(text)
  } else {
    value <- formatC(text)
  }

  out <- list(value = value, style_id = style_id)
  class(out) <- c("run_wordtext", "cot", "run")
  out
}

#' @export
print.run_wordtext <- function(x, ...) {
  cat("text: `", x$value, "`\nword style id: `", x$style_id, "`\n", sep = "")
}
#
#' @export
to_wml.run_wordtext <- function(x, add_ns = FALSE, ...) {
  out <- paste0(
    "<w:r>",
    "<w:rPr>",
    sprintf("<w:rStyle w:val=\"%s\"/>", x$style_id),
    "</w:rPr>",
    "<w:t xml:space=\"preserve\">",
    htmlEscapeCopy(x$value),
    "</w:t></w:r>"
  )
  out
}


# Word computed field ----
#' @export
#' @title 'Word' computed field
#' @description Create a 'Word' computed field.
#' @note
#' In the previous version, this function was called `run_seqfield`
#' but the name was wrong and should have been `run_word_field`.
#' @inheritSection ftext usage
#' @param field Value for a "Word Computed Field" as a string.
#' @param seqfield deprecated in favor of `field`.
#' @param prop formatting text properties returned by [fp_text].
#' @examples
#' run_word_field(field = "PAGE  \\* MERGEFORMAT")
#' run_word_field(field = "Date \\@ \"MMMM d yyyy\"")
#' @family run functions for reporting
#' @family Word computed fields
run_word_field <- function(field, prop = NULL, seqfield = NULL) {
  if (!is.null(seqfield)) {
    field <- seqfield
    message("`seqfield` argument is deprecated in favor of `field`")
  }
  z <- list(
    field = field,
    pr = prop
  )
  class(z) <- c("run_word_field", "run")
  z
}

#' @export
#' @rdname run_word_field
run_seqfield <- run_word_field

#' @export
to_wml.run_word_field <- function(x, add_ns = FALSE, ...) {
  pr <- if (!is.null(x$pr)) rpr_wml(x$pr) else "<w:rPr/>"

  xml_elt_1 <- paste0(
    wr_ns_yes,
    pr,
    "<w:fldChar w:fldCharType=\"begin\" w:dirty=\"true\"/>",
    "</w:r>"
  )
  xml_elt_2 <- paste0(
    wr_ns_yes,
    pr,
    sprintf(
      "<w:instrText xml:space=\"preserve\" w:dirty=\"true\">%s</w:instrText>",
      x$field
    ),
    "</w:r>"
  )
  xml_elt_3 <- paste0(
    wr_ns_yes,
    pr,
    "<w:fldChar w:fldCharType=\"end\" w:dirty=\"true\"/>",
    "</w:r>"
  )
  out <- paste0(xml_elt_1, xml_elt_2, xml_elt_3)

  out
}

# autonum ----

#' @export
#' @title Auto number
#' @description Create an autonumbered chunk, i.e. a string
#' representation of a sequence, each item will be numbered.
#' These runs can also be bookmarked and be used later for
#' cross references.
#' @inheritSection ftext usage
#' @param seq_id sequence identifier
#' @param pre_label,post_label text to add before and after number
#' @param bkm bookmark id to associate with autonumber run. If NULL, no bookmark
#' is added. Value can only be made of alpha numeric characters, ':', -' and '_'.
#' @param bkm_all if TRUE, the bookmark will be set on the whole string, if
#' FALSE, the bookmark will be set on the number only. Default to FALSE.
#' As an effect when a reference to this bookmark is used, the text can
#' be like "Table 1" or "1" (pre_label is not included in the referenced
#' text).
#' @param prop formatting text properties returned by [fp_text].
#' @param start_at If not NULL, it must be a positive integer, it
#' specifies the new number to use, at which number the auto
#' numbering will restart.
#' @param tnd *title number depth*, a positive integer (only applies if positive)
#' that specify the depth (or heading of level *depth*) to use for prefixing
#' the caption number with this last reference number. For example, setting
#' `tnd=2` will generate numbered captions like '4.3-2' (figure 2 of
#' chapter 4.3).
#' @param tns separator to use between title number and table
#' number. Default is "-".
#' @examples
#' run_autonum()
#' run_autonum(seq_id = "fig", pre_label = "fig. ")
#' run_autonum(seq_id = "tab", pre_label = "Table ", bkm = "anytable")
#' run_autonum(
#'   seq_id = "tab", pre_label = "Table ", bkm = "anytable",
#'   tnd = 2, tns = " "
#' )
#' @family run functions for reporting
#' @family Word computed fields
run_autonum <- function(
  seq_id = "table",
  pre_label = "Table ",
  post_label = ": ",
  bkm = NULL,
  bkm_all = FALSE,
  prop = NULL,
  start_at = NULL,
  tnd = 0,
  tns = "-"
) {
  bkm <- check_bookmark_id(bkm)

  stopifnot(
    !is.null(tnd),
    !is.null(tns),
    is.numeric(tnd),
    sign(tnd) > -1,
    is.character(tns),
    is.null(start_at) || is.numeric(start_at),
    inherits(prop, "fp_text") || is.null(prop)
  )
  z <- list(
    seq_id = seq_id,
    pre_label = pre_label,
    post_label = post_label,
    bookmark = bkm,
    bookmark_all = bkm_all,
    pr = prop,
    start_at = start_at,
    tnd = tnd,
    tns = tns
  )
  class(z) <- c("run_autonum", "run")

  z
}

#' @export
#' @title Update bookmark of an autonumber run
#' @description This function lets recycling a object
#' made by [run_autonum()] by changing the bookmark value. This
#' is useful to avoid calling `run_autonum()` several times
#' because of many tables.
#' @param x an object of class [run_autonum()]
#' @param bkm bookmark id to associate with autonumber run.
#' Value can only be made of alpha numeric characters, ':', -' and '_'.
#' @examples
#' z <- run_autonum(
#'   seq_id = "tab", pre_label = "Table ",
#'   bkm = "anytable"
#' )
#' set_autonum_bookmark(z, bkm = "anothertable")
#' @seealso [run_autonum()]
set_autonum_bookmark <- function(x, bkm = NULL) {
  stopifnot(inherits(x, "run_autonum"))
  x$bookmark <- bkm
  x
}

#' @export
to_wml.run_autonum <- function(x, add_ns = FALSE, ...) {
  if (is.null(x$seq_id)) {
    return("")
  }

  pr <- if (!is.null(x$pr)) rpr_wml(x$pr) else "<w:rPr/>"

  run_str_pre <- sprintf(
    "<w:r>%s<w:t xml:space=\"preserve\">%s</w:t></w:r>",
    pr,
    x$pre_label
  )
  run_str_post <- sprintf(
    "<w:r>%s<w:t xml:space=\"preserve\">%s</w:t></w:r>",
    pr,
    x$post_label
  )

  seq_str <- paste0("SEQ ", x$seq_id, " \u005C* Arabic")
  if (!is.null(x$start_at) && is.numeric(x$start_at)) {
    seq_str <- paste(seq_str, "\\r", as.integer(x$start_at))
  }

  sqf <- run_word_field(field = seq_str, prop = x$pr)
  sf_str <- to_wml(sqf)

  if (x$tnd > 0) {
    z <- paste0(
      to_wml(run_word_field(
        field = paste0("STYLEREF ", x$tnd, " \\r"),
        prop = x$pr
      )),
      to_wml(ftext(x$tns, prop = x$pr))
    )
    sf_str <- paste0(z, sf_str)
  }

  if (!is.null(x$bookmark)) {
    if (x$bookmark_all) {
      out <- paste0(
        bookmark(
          id = x$bookmark,
          str = paste0(run_str_pre, sf_str)
        ),
        run_str_post
      )
    } else {
      sf_str <- bookmark(id = x$bookmark, sf_str)
      out <- paste0(run_str_pre, sf_str, run_str_post)
    }
  } else {
    out <- paste0(run_str_pre, sf_str, run_str_post)
  }

  out
}

# reference ----

#' @export
#' @title Cross reference
#' @description Create a representation of a reference
#' @param id reference id, a string
#' @param prop formatting text properties returned by [fp_text].
#' @inheritSection ftext usage
#' @examples
#' run_reference("a_ref")
#' @family run functions for reporting
#' @family Word computed fields
run_reference <- function(id, prop = NULL) {
  z <- paste0(" REF ", id, " \\h ")

  z <- list(
    id = id,
    pr = prop
  )
  class(z) <- c("run_reference", "run")

  z
}

#' @export
to_wml.run_reference <- function(x, add_ns = FALSE, ...) {
  out <- to_wml(run_word_field(
    field = paste0(" REF ", x$id, " \\h "),
    prop = x$pr
  ))
  out
}

# hyperlink text ----

#' @export
#' @title Formatted chunk of text with hyperlink
#' @description Format a chunk of text with text formatting properties (bold, color, ...),
#' the chunk is associated with an hyperlink.
#' @param text text value, a single character value
#' @param prop formatting text properties returned by [fp_text]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @param href URL value
#' @inheritSection ftext usage
#' @examples
#' ft <- fp_text(font.size = 12, bold = TRUE)
#' hyperlink_ftext(
#'   href = "https://cran.r-project.org/index.html",
#'   text = "some text", prop = ft
#' )
#' @family run functions for reporting
hyperlink_ftext <- function(text, prop = NULL, href) {
  if (is.character(text)) {
    value <- enc2utf8(text)
  } else {
    value <- formatC(text)
  }

  out <- list(value = value, pr = prop, href = href)
  class(out) <- c("hyperlink_ftext", "cot", "run")
  out
}

#' @export
to_wml.hyperlink_ftext <- function(x, add_ns = FALSE, ...) {
  out <- paste0(
    "<w:r>",
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:t xml:space=\"preserve\">",
    x$value,
    "</w:t></w:r>"
  )
  paste0(
    "<w:hyperlink r:id=\"",
    officer_url_encode(x$href),
    "\">",
    out,
    "</w:hyperlink>"
  )
}

#' @export
to_pml.hyperlink_ftext <- function(x, add_ns = FALSE, ...) {
  open_tag <- ar_ns_no
  if (add_ns) {
    open_tag <- ar_ns_yes
  }
  str_ <- sprintf("<a:hlinkClick r:id=\"%s\"/>", officer_url_encode(x$href))

  if (!is.null(x$pr)) {
    rpr_pml_ <- rpr_pml(x$pr)
    m <- regexpr(">", rpr_pml_)
    regmatches(rpr_pml_, m) <- paste0(">", str_)
  } else {
    rpr_pml_ <- paste0("<a:rPr>", str_, "</a:rPr>")
  }

  paste0(
    open_tag,
    rpr_pml_,
    "<a:t>",
    htmlEscapeCopy(x$value),
    "</a:t></a:r>"
  )
}

#' @export
to_html.hyperlink_ftext <- function(x, ...) {
  if (!is.null(x$pr)) {
    rpr_css <- paste0(" style=\"", rpr_css(x$pr), "\"")
  } else {
    rpr_css <- ""
  }

  sprintf("<a href=\"%s\"><span%s></span></a>", x$href, rpr_css)
}


# bookmark ----

#' @export
#' @title Bookmark for 'Word'
#' @description Add a bookmark on a run object.
#' @param bkm bookmark id to associate with run. Value can only be
#' made of alpha numeric characters, '-' and '_'.
#' @param run a run object, made with a call to one of
#' the "run functions for reporting".
#' @inheritSection ftext usage
#' @examples
#' ft <- fp_text(font.size = 12, bold = TRUE)
#' run_bookmark("par1", ftext("some text", ft))
#' @family run functions for reporting
run_bookmark <- function(bkm, run) {
  all_run <- FALSE

  if (inherits(run, "run")) {
    run <- list(run)
  }

  if (is.list(run) && !inherits(run, "run")) {
    all_run <- all(vapply(run, inherits, FUN.VALUE = FALSE, what = "run"))
  }

  if (!all_run) {
    stop(
      "`run` must be a run object (ftext for example) or a list of run objects."
    )
  }

  bkm <- check_bookmark_id(bkm)

  z <- list(id = bkm, run = run)
  class(z) <- c("run_bookmark", "run")
  z
}

#' @export
to_wml.run_bookmark <- function(x, add_ns = FALSE, ...) {
  runs <- lapply(x$run, to_wml, add_ns = add_ns, ...)
  runs <- do.call(paste0, runs)
  bookmark(id = x$id, str = runs)
}

# breaks ----

#' @export
#' @title Page break for 'Word'
#' @description Object representing a page break for a Word document.
#' @inheritSection ftext usage
#' @examples
#' fp_t <- fp_text(font.size = 12, bold = TRUE)
#' an_fpar <- fpar("let's add a break page", run_pagebreak(), ftext("and blah blah!", fp_t))
#'
#' x <- read_docx()
#' x <- body_add(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_pagebreak <- function() {
  z <- list()
  class(z) <- c("run_pagebreak", "run")
  z
}

#' @export
to_wml.run_pagebreak <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:br w:type=\"page\"/></w:r>"
  out
}

#' @export
#' @title Column break for 'Word'
#' @description Create a representation of a column break.
#' @inheritSection ftext usage
#' @examples
#' run_columnbreak()
#' @family run functions for reporting
run_columnbreak <- function() {
  z <- list()
  class(z) <- c("run_columnbreak", "run")

  z
}

#' @export
to_wml.run_columnbreak <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:br w:type=\"column\"/></w:r>"
  out
}

#' @export
#' @title Page break for 'Word'
#' @description Object representing a line break
#' for a Word document. The result must be used
#' within a call to [fpar].
#' @inheritSection ftext usage
#' @examples
#' fp_t <- fp_text(font.size = 12, bold = TRUE)
#' an_fpar <- fpar("let's add a line break", run_linebreak(), ftext("and blah blah!", fp_t))
#'
#' x <- read_docx()
#' x <- body_add(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_linebreak <- function() {
  z <- list()
  class(z) <- c("run_linebreak", "run")

  z
}

#' @export
to_wml.run_linebreak <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:br w:type=\"textWrapping\"/></w:r>"
  out
}

#' @export
#' @title Tab for 'Word'
#' @description Object representing a tab
#' in a Word document. The result must be used
#' within a call to [fpar]. It will only have
#' effects in Word output.
#'
#' Tabulation marks settings can be defined with [fp_tabs()] in
#' paragraph settings defined with [fp_par()].
#' @inheritSection ftext usage
#' @examples
#' z <- fp_tabs(
#'   fp_tab(pos = 0.5, style = "decimal"),
#'   fp_tab(pos = 1.5, style = "decimal")
#' )
#' par1 <- fpar(
#'   run_tab(), ftext("88."),
#'   run_tab(), ftext("987.45"),
#'   fp_p = fp_par(
#'     tabs = z
#'   )
#' )
#' par2 <- fpar(
#'   run_tab(), ftext("8."),
#'   run_tab(), ftext("670987.45"),
#'   fp_p = fp_par(
#'     tabs = z
#'   )
#' )
#' x <- read_docx()
#' x <- body_add(x, par1)
#' x <- body_add(x, par2)
#' print(x, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_tab <- function() {
  z <- list()
  class(z) <- c("run_tab", "run")

  z
}

#' @export
to_wml.run_tab <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:tab/></w:r>"
  out
}

# sections ----
inch_to_tweep <- function(x) round(x * 72 * 20, digits = 0)

#' @export
#' @title Page size object
#' @description The function creates a representation of the dimensions of a page. The
#' dimensions are defined by length, width and orientation. If the orientation is
#' in landscape mode then the length becomes the width and the width becomes the length.
#' @param width,height page width, page height, default to A4 format If NULL
#' the value will be ignored and Word will use the default value.
#' @param orient page orientation, either 'landscape', either 'portrait'.
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @examples
#' page_size(orient = "landscape")
#' @family functions for section definition
page_size <- function(
  width = 11906 / 1440,
  height = 16838 / 1440,
  orient = "portrait",
  unit = "in"
) {
  is_null_width <- is.null(width)
  is_null_height <- is.null(height)

  if (!is_null_width) {
    width <- convin(unit = unit, x = width)
  }
  if (!is_null_height) {
    height <- convin(unit = unit, x = height)
  }

  if (is_null_width) {
    w <- NULL
  } else {
    if (orient %in% "portrait") {
      w <- min(c(height, width))
    } else {
      w <- max(c(height, width))
    }
  }

  if (is_null_height) {
    h <- NULL
  } else {
    if (orient %in% "portrait") {
      h <- max(c(height, width))
    } else {
      h <- min(c(height, width))
    }
  }

  z <- list(width = w, height = h, orient = orient)
  class(z) <- c("page_size")
  z
}

#' @export
to_wml.page_size <- function(x, add_ns = FALSE, ...) {
  if (!is.null(x$width) && !is.null(x$height)) {
    out <- sprintf(
      "<w:pgSz w:h=\"%.0f\" w:w=\"%.0f\" w:orient=\"%s\"/>",
      inch_to_tweep(x$height),
      inch_to_tweep(x$width),
      x$orient
    )
  } else {
    out <- sprintf(
      "<w:pgSz w:orient=\"%s\"/>",
      x$orient
    )
  }

  out
}

#' @export
#' @title Section columns
#' @description The function creates a representation of the columns of a section.
#' @param widths columns widths in inches. If 3 values, 3 columns
#' will be produced.
#' @param space space in inches between columns.
#' @param sep if TRUE a line is separating columns.
#' @examples
#' section_columns()
#' @family functions for section definition
section_columns <- function(widths = c(2.5, 2.5), space = .25, sep = FALSE) {
  if (length(widths) < 2) {
    stop("length of widths should be at least 2")
  }

  z <- list(widths = widths, space = space, sep = sep)
  class(z) <- c("section_columns")
  z
}
#' @export
to_wml.section_columns <- function(x, add_ns = FALSE, ...) {
  widths <- x$widths * 20 * 72
  space <- x$space * 20 * 72

  columns_str_all_but_last <- sprintf(
    "<w:col w:w=\"%.0f\" w:space=\"%.0f\"/>",
    widths[-length(widths)],
    space
  )
  columns_str_last <- sprintf(
    "<w:col w:w=\"%.0f\"/>",
    widths[length(widths)]
  )
  columns_str <- c(columns_str_all_but_last, columns_str_last)

  sprintf(
    "<w:cols w:num=\"%.0f\" w:sep=\"%.0f\" w:space=\"%.0f\" w:equalWidth=\"0\">%s</w:cols>",
    length(widths),
    as.integer(x$sep),
    space,
    paste0(columns_str, collapse = "")
  )
}

#' @export
#' @title Page margins object
#' @description
#' Define margins for each page of a section.
#'
#' The function creates a representation of the dimensions of a page. The
#' dimensions are defined by length, width and orientation. If the orientation
#' is in landscape mode then the length becomes the width and the width becomes
#' the length.
#' @param bottom,top distance (in inches) between the bottom/top of the text
#' margin and the bottom/top of the page. The text is placed at the greater of
#' the value of this attribute and the extent of the header/footer text. A
#' negative value indicates that the content should be measured from the
#' bottom/top of the page regardless of the footer/header, and so will overlap
#' the footer/header. For example, `header=-0.5, bottom=1` means that the footer
#' must start one inch from the bottom of the page and the main document text
#' must start a half inch from the bottom of the page. In this case, the text
#' and footer overlap since bottom is negative.
#' @param footer distance (in inches) from the bottom edge of the page to the
#' bottom edge of the footer.
#' @param header distance (in inches) from the top edge of the page to the top
#' edge of the header.
#' @param left,right distance (in inches) from the left/right edge of the page
#' to the left/right edge of the text.
#' @param gutter page gutter (in inches).
#' @examples
#' page_mar()
#' @family functions for section definition
page_mar <- function(
  bottom = 1417 / 1440,
  top = 1417 / 1440,
  right = 1417 / 1440,
  left = 1417 / 1440,
  header = 708 / 1440,
  footer = 708 / 1440,
  gutter = 0 / 1440
) {
  z <- list(
    header = header,
    bottom = bottom,
    top = top,
    right = right,
    left = left,
    footer = footer,
    gutter = gutter
  )
  class(z) <- c("page_mar")
  z
}
#' @export
to_wml.page_mar <- function(x, add_ns = FALSE, ...) {
  out <- sprintf(
    "<w:pgMar w:header=\"%.0f\" w:bottom=\"%.0f\" w:top=\"%.0f\" w:right=\"%.0f\" w:left=\"%.0f\" w:footer=\"%.0f\" w:gutter=\"%.0f\"/>",
    inch_to_tweep(x$header),
    inch_to_tweep(x$bottom),
    inch_to_tweep(x$top),
    inch_to_tweep(x$right),
    inch_to_tweep(x$left),
    inch_to_tweep(x$footer),
    inch_to_tweep(x$gutter)
  )

  out
}

docx_section_type <- c(
  "continuous",
  "evenPage",
  "nextColumn",
  "nextPage",
  "oddPage"
)

#' @export
#' @title Section properties
#' @description A section is a grouping of blocks (ie. paragraphs and tables)
#' that have a set of properties that define pages on which the text will appear.
#'
#' A Section properties object stores information about page composition,
#' such as page size, page orientation, borders and margins.
#'
#' **Important**: When creating multiple sections in a document, it is strongly
#' recommended to use the **same `page_margins` object** for all sections to avoid
#' unwanted page breaks. Changing page margins between sections can cause Word to
#' insert automatic page breaks, even when using `type = "continuous"`. To ensure
#' consistent behavior, create a single `page_mar()` object and reuse it across
#' all `prop_section()` calls. See the examples in [body_end_block_section()] which
#' demonstrate this best practice.
#'
#' @param page_size page dimensions, an object generated with function [page_size].
#' @param page_margins page margins, an object generated with function [page_mar].
#' It is recommended to use the same margins object across all sections to prevent
#' unintended page breaks.
#' @param type Section type. It defines how the contents of the section will be
#' placed relative to the previous section. Available types are "continuous"
#' (begins the section on the next paragraph), "evenPage" (begins on the next
#' even-numbered page), "nextColumn" (begins on the next column on the page),
#' "nextPage" (begins on the following page), "oddPage" (begins on the next
#' odd-numbered page).
#' @param section_columns section columns, an object generated with function
#' [section_columns]. Use NULL (default value) for no content.
#' @param header_default content as a [block_list()] for the default page header.
#' Use NULL (default value) for no content.
#' @param header_even content as a [block_list()] for the even page header.
#' Use NULL (default value) for no content.
#' @param header_first content as a [block_list()] for the first page header.
#' Use NULL (default value) for no content.
#' @param footer_default content as a [block_list()] for the default page footer.
#' Use NULL (default value) for no content.
#' @param footer_even content as a [block_list()] for the even page footer. Use
#' NULL (default value) for no content.
#' @param footer_first content as a [block_list()] for the default page footer.
#' Use NULL (default value) for no content.
#' @example inst/examples/example_prop_section.R
#' @seealso [block_section]
#' @family functions for section definition
#' @section Illustrations:
#'
#' \if{html}{\figure{prop_section_doc_1.png}{options: style="width:80\%;"}}
prop_section <- function(
  page_size = NULL,
  page_margins = NULL,
  type = "continuous",
  section_columns = NULL,
  header_default = NULL,
  header_even = NULL,
  header_first = NULL,
  footer_default = NULL,
  footer_even = NULL,
  footer_first = NULL
) {
  z <- list()

  if (is.null(page_size)) {
    page_size <- page_size()
  }
  if (is.null(page_margins)) {
    page_margins <- page_mar()
  }

  if (!is.null(header_default) && !inherits(header_default, "block_list")) {
    stop(
      "header_default must be NULL or an object generated by `block_list()`."
    )
  }
  z$header_default <- header_default

  if (!is.null(header_even) && !inherits(header_even, "block_list")) {
    stop("header_even must be NULL or an object generated by `block_list()`.")
  }
  z$header_even <- header_even

  if (!is.null(header_first) && !inherits(header_first, "block_list")) {
    stop("header_first must be NULL or an object generated by `block_list()`.")
  }
  z$header_first <- header_first

  if (!is.null(footer_default) && !inherits(footer_default, "block_list")) {
    stop(
      "footer_default must be NULL or an object generated by `block_list()`."
    )
  }
  z$footer_default <- footer_default

  if (!is.null(footer_even) && !inherits(footer_even, "block_list")) {
    stop("footer_even must be NULL or an object generated by `block_list()`.")
  }
  z$footer_even <- footer_even

  if (!is.null(footer_first) && !inherits(footer_first, "block_list")) {
    stop("footer_first must be NULL or an object generated by `block_list()`.")
  }
  z$footer_first <- footer_first

  if (!is.null(type) && !type %in% docx_section_type) {
    stop(
      "type must be one of ",
      paste(shQuote(docx_section_type), collapse = ", ")
    )
  }
  z$type <- type

  if (!is.null(page_size) && !inherits(page_size, "page_size")) {
    stop("page_size must be a page_size object")
  }
  z$page_size <- page_size

  if (!is.null(page_margins) && !inherits(page_margins, "page_mar")) {
    stop("page_margins must be a page_mar object")
  }
  z$page_margins <- page_margins

  if (
    !is.null(section_columns) && !inherits(section_columns, "section_columns")
  ) {
    stop("section_columns must be a section_columns object")
  }
  z$section_columns <- section_columns

  z <- Filter(function(x) !is.null(x), z)
  class(z) <- c("prop_section", "prop")

  z
}

#' @export
to_wml.prop_section <- function(x, add_ns = FALSE, ...) {
  paste0(
    "<w:sectPr",
    if (add_ns) {
      " xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\""
    },
    ">",
    if (!is.null(x$header_default)) {
      paste0(
        "<w:headerReference w:officer=\"true\" r:id=\"%s\" w:type=\"default\">",
        "<w:hdr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$header_default, add_ns = add_ns),
        "</w:hdr>",
        "</w:headerReference>"
      )
    },
    if (!is.null(x$header_even)) {
      paste0(
        "<w:headerReference w:officer=\"true\" r:id=\"%s\" w:type=\"even\">",
        "<w:hdr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$header_even, add_ns = add_ns),
        "</w:hdr>",
        "</w:headerReference>"
      )
    },
    if (!is.null(x$header_first)) {
      paste0(
        "<w:headerReference w:officer=\"true\" r:id=\"%s\" w:type=\"first\">",
        "<w:hdr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$header_first, add_ns = add_ns),
        "</w:hdr>",
        "</w:headerReference>"
      )
    },
    if (!is.null(x$footer_default)) {
      paste0(
        "<w:footerReference w:officer=\"true\" r:id=\"%s\" w:type=\"default\">",
        "<w:ftr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$footer_default, add_ns = add_ns),
        "</w:ftr>",
        "</w:footerReference>"
      )
    },
    if (!is.null(x$footer_even)) {
      paste0(
        "<w:footerReference w:officer=\"true\" r:id=\"%s\" w:type=\"even\">",
        "<w:ftr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$footer_even, add_ns = add_ns),
        "</w:ftr>",
        "</w:footerReference>"
      )
    },
    if (!is.null(x$footer_first)) {
      paste0(
        "<w:footerReference w:officer=\"true\" r:id=\"%s\" w:type=\"first\">",
        "<w:ftr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$footer_first, add_ns = add_ns),
        "</w:ftr>",
        "</w:footerReference>"
      )
    },
    if (!is.null(x$page_margins)) to_wml(x$page_margins, add_ns = add_ns),
    if (!is.null(x$page_size)) to_wml(x$page_size, add_ns = add_ns),
    if (!is.null(x$header_first) || !is.null(x$footer_first)) "<w:titlePg/>",
    if (!is.null(x$type)) paste0("<w:type w:val=\"", x$type, "\"/>"),
    if (!is.null(x$section_columns)) {
      to_wml(x$section_columns, add_ns = add_ns)
    } else {
      "<w:cols/>"
    },
    "</w:sectPr>"
  )
}


# external_img ----
#' @export
#' @title External image
#' @description Wraps an image in an object that can then be embedded
#' in a PowerPoint slide or within a Word paragraph.
#'
#' The image is added as a shape in PowerPoint (it is not possible to mix text and
#' images in a PowerPoint form). With a Word document, the image will be
#' added inside a paragraph.
#' @param src image file path
#' @param width,height size of the image file. It can be ignored
#' if parameter `guess_size=TRUE`, see parameter `guess_size`.
#' @param guess_size If package 'magick' is installed, this option
#' can be used (set it to `TRUE`). The images will be read and
#' width and height will be guessed.
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param alt alternative text for images
#' @inheritSection ftext usage
#' @examples
#' # wrap r logo with external_img ----
#' srcfile <- file.path(R.home("doc"), "html", "logo.jpg")
#' extimg <- external_img(
#'   src = srcfile, height = 1.06 / 2,
#'   width = 1.39 / 2
#' )
#'
#' # pptx example ----
#' doc <- read_pptx()
#' doc <- add_slide(doc, "Title and Content")
#' doc <- ph_with(
#'   x = doc, value = extimg,
#'   location = ph_location_type(type = "body"),
#'   use_loc_size = FALSE
#' )
#' print(doc, target = tempfile(fileext = ".pptx"))
#'
#' fp_t <- fp_text(font.size = 20, color = "red")
#' an_fpar <- fpar(extimg, ftext(" is cool!", fp_t))
#'
#' # docx example ----
#' x <- read_docx()
#' x <- body_add(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [ph_with], [body_add], [fpar]
#' @family run functions for reporting
external_img <- function(
  src,
  width = .5,
  height = .2,
  unit = "in",
  guess_size = FALSE,
  alt = ""
) {
  # note: should it be vectorized
  check_src <- all(grepl("^rId", src)) || all(file.exists(src))
  if (!check_src) {
    stop(
      "src must be a string starting with 'rId' or an existing image filename"
    )
  }

  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (length(src) > 1) {
    if (length(width) == 1) {
      width <- rep(width, length(src))
    }
    if (length(height) == 1) height <- rep(height, length(src))
  }

  if (guess_size) {
    if (!requireNamespace("magick", quietly = TRUE)) {
      stop("package magick is required when using `guess_size` option.")
    }
    sizes <- lapply(src, function(x) {
      if (grepl("\\.svg$", x)) {
        z <- magick::image_read_svg(x)
      } else {
        z <- magick::image_read(x)
      }
      z <- magick::image_data(z)
      attr(z, "dim")[-1]
    })
    sizes <- do.call(rbind, sizes)
    width <- sizes[, 1] / 72
    height <- sizes[, 2] / 72
  }

  class(src) <- c("external_img", "cot", "run")
  attr(src, "dims") <- list(width = width, height = height)
  attr(src, "alt") <- alt
  src
}


#' @export
dim.external_img <- function(x) {
  x <- attr(x, "dims")
  data.frame(width = x$width, height = x$height)
}


#' @export
as.data.frame.external_img <- function(x, ...) {
  dimx <- attr(x, "dims")
  data.frame(
    path = as.character(x),
    width = dimx$width,
    height = dimx$height,
    alt = attr(x, "alt"),
    stringsAsFactors = FALSE
  )
}

temp_blipfill <- function(value, ns = "p") {
  svg_src <- NULL
  if (grepl("\\.svg", value)) {
    if (!requireNamespace("rsvg")) {
      stop("package 'rsvg' is required to convert svg file to rasters.")
    }

    svg_src <- tempfile(fileext = ".svg")
    file.copy(as.character(value), to = svg_src)

    img_src <- tempfile(fileext = ".png")
    rsvg::rsvg_png(svg_src, file = img_src)
  } else {
    img_src <- tempfile(
      fileext = gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", as.character(value))
    )
    file.copy(as.character(value), to = img_src)
  }

  if (!is.null(svg_src)) {
    blipfill <- paste0(
      "<",
      ns,
      ":blipFill>",
      sprintf("<a:blip r:embed=\"%s\">", img_src),
      "<a:extLst>",
      "<a:ext uri=\"{96DAC541-7B7A-43D3-8B79-37D633B846F1}\">",
      sprintf(
        "<asvg:svgBlip r:embed=\"%s\" xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\"/>",
        svg_src
      ),
      "</a:ext>",
      "</a:extLst>",
      "</a:blip>",
      "<a:stretch><a:fillRect/></a:stretch>",
      "</",
      ns,
      ":blipFill>"
    )
  } else {
    blipfill <- paste0(
      "<",
      ns,
      ":blipFill>",
      sprintf("<a:blip r:embed=\"%s\"/>", img_src),
      "<a:stretch><a:fillRect/></a:stretch>",
      "</",
      ns,
      ":blipFill>"
    )
  }
  blipfill
}

#' @export
to_wml.external_img <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  blipfill <- temp_blipfill(x, ns = "pic")

  paste0(
    open_tag,
    "<w:rPr/><w:drawing xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\"><wp:inline distT=\"0\" distB=\"0\" distL=\"0\" distR=\"0\">",
    sprintf(
      "<wp:extent cx=\"%.0f\" cy=\"%.0f\"/>",
      width * 12700 * 72,
      height * 12700 * 72
    ),
    "<wp:docPr id=\"\" name=\"\" descr=\"",
    attr(x, "alt"),
    "\"/>",
    "<wp:cNvGraphicFramePr><a:graphicFrameLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/></wp:cNvGraphicFramePr>",
    "<a:graphic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"><a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/picture\"><pic:pic xmlns:pic=\"http://schemas.openxmlformats.org/drawingml/2006/picture\">",
    "<pic:nvPicPr>",
    "<pic:cNvPr id=\"\" name=\"\"/>",
    "<pic:cNvPicPr><a:picLocks noChangeAspect=\"1\" noChangeArrowheads=\"1\"/>",
    "</pic:cNvPicPr></pic:nvPicPr>",
    blipfill,
    "<pic:spPr bwMode=\"auto\"><a:xfrm><a:off x=\"0\" y=\"0\"/>",
    sprintf(
      "<a:ext cx=\"%.0f\" cy=\"%.0f\"/></a:xfrm><a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom><a:noFill/></pic:spPr>",
      width * 12700 * 72,
      height * 12700 * 72
    ),
    "</pic:pic></a:graphicData></a:graphic></wp:inline></w:drawing></w:r>"
  )
}

#' @export
to_pml.external_img <- function(
  x,
  add_ns = FALSE,
  left = 0,
  top = 0,
  width = 3,
  height = 3,
  bg = "transparent",
  rot = 0,
  ph = "<p:ph/>",
  label = "",
  ln = sp_line(lwd = 0, linecmpd = "solid", lineend = "rnd"),
  ...
) {
  if (!is.null(bg) && !is.color(bg)) {
    stop("bg must be a valid color.", call. = FALSE)
  }

  bg_str <- solid_fill_pml(bg)
  ln_str <- ln_pml(ln)

  xfrm_str <- a_xfrm_str(
    left = left,
    top = top,
    width = width,
    height = height,
    rot = rot
  )
  if (is.null(ph) || is.na(ph)) {
    ph <- "<p:ph/>"
  }
  blipfill <- temp_blipfill(x, ns = "p")
  id <- uuid_generate()
  str <- "
<p:pic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">
  <p:nvPicPr>
    <p:cNvPr id=\"%s\" name=\"%s\" descr=\"%s\"/>
    <p:cNvPicPr><a:picLocks noGrp=\"1\"/></p:cNvPicPr>
    <p:nvPr>%s</p:nvPr>
  </p:nvPicPr>
  %s
  <p:spPr>%s<a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom>%s%s</p:spPr>
</p:pic>
"
  sprintf(
    str,
    id,
    label,
    attr(x, "alt"),
    ph,
    blipfill,
    xfrm_str,
    bg_str,
    ln_str
  )
}


#' @export
to_html.external_img <- function(x, ...) {
  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height
  b64_str <- image_to_base64(as.character(x))
  sprintf(
    "<img style=\"vertical-align:middle;width:%.0fpx;height:%.0fpx;\" src=\"%s\" alt=\"%s\"/>",
    width * 72,
    height * 72,
    b64_str,
    attr(x, "alt")
  )
}

# floating_external_img ----
#' @export
#' @title Floating external image
#' @description Wraps an image in an object that can be embedded as a floating
#' image in a 'Word' document. Unlike [external_img()], which creates inline images,
#' this function creates floating images that can be positioned anywhere on the page
#' and allow text wrapping around them.
#'
#' @param src image file path
#' @param width,height size of the image file. It can be ignored
#' if parameter `guess_size=TRUE`, see parameter `guess_size`.
#' @param pos_x,pos_y horizontal and vertical position of the image relative to the anchor point
#' @param pos_h_from horizontal positioning reference point, one of "margin", "page", "column", "character"
#' @param pos_v_from vertical positioning reference point, one of "margin", "page", "paragraph", "line"
#' @param wrap_type text wrapping type, one of "square", "topAndBottom", "through", "tight", "none"
#' @param wrap_side which side text wraps around, one of "bothSides", "left", "right", "largest"
#' @param wrap_dist_top,wrap_dist_bottom,wrap_dist_left,wrap_dist_right distance between image and text (in inches)
#' @param guess_size If package 'magick' is installed, this option
#' can be used (set it to `TRUE`). The images will be read and
#' width and height will be guessed.
#' @param unit unit for width, height, pos_x and pos_y, one of "in", "cm", "mm".
#' @param alt alternative text for images
#' @inheritSection ftext usage
#' @examples
#' library(officer)
#' srcfile <- file.path(R.home("doc"), "html", "logo.jpg")
#' floatimg <- floating_external_img(
#'   src = srcfile, height = 1.06 / 2, width = 1.39 / 2,
#'   pos_x = 0, pos_y = 0,
#'   pos_h_from = "margin", pos_v_from = "margin"
#' )
#'
#' text <- paste0(
#'   " is a floating image in a ",
#'   paste0(rep("very ", 30), collapse = ""),
#'   " long text!"
#' )
#'
#' # docx example ----
#' x <- read_docx()
#' fp_t <- fp_text(font.size = 20, color = "red")
#' an_fpar <- fpar(floatimg, ftext(text, fp_t))
#' x <- body_add_fpar(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#'
#' # rtf example ----
#' rtf_doc <- rtf_doc()
#' rtf_doc <- rtf_add(rtf_doc, an_fpar)
#' print(rtf_doc, target = tempfile(fileext = ".rtf"))
#' @seealso [external_img], [body_add], [fpar], [rtf_doc], [rtf_add]
#' @family run functions for reporting
floating_external_img <- function(
  src,
  width = .5,
  height = .2,
  pos_x = 0,
  pos_y = 0,
  pos_h_from = "margin",
  pos_v_from = "margin",
  wrap_type = "square",
  wrap_side = "bothSides",
  wrap_dist_top = 0,
  wrap_dist_bottom = 0,
  wrap_dist_left = 0.125,
  wrap_dist_right = 0.125,
  unit = "in",
  guess_size = FALSE,
  alt = ""
) {
  # note: should it be vectorized
  check_src <- all(grepl("^rId", src)) || all(file.exists(src))
  if (!check_src) {
    stop(
      "src must be a string starting with 'rId' or an existing image filename"
    )
  }

  # Validate positioning parameters
  pos_h_from <- match.arg(
    pos_h_from,
    c("margin", "page", "column", "character")
  )
  pos_v_from <- match.arg(pos_v_from, c("margin", "page", "paragraph", "line"))
  wrap_type <- match.arg(
    wrap_type,
    c("square", "topAndBottom", "through", "tight", "none")
  )
  wrap_side <- match.arg(wrap_side, c("bothSides", "left", "right", "largest"))

  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)
  pos_x <- convin(unit = unit, x = pos_x)
  pos_y <- convin(unit = unit, x = pos_y)

  if (length(src) > 1) {
    if (length(width) == 1) {
      width <- rep(width, length(src))
    }
    if (length(height) == 1) {
      height <- rep(height, length(src))
    }
    if (length(pos_x) == 1) {
      pos_x <- rep(pos_x, length(src))
    }
    if (length(pos_y) == 1) pos_y <- rep(pos_y, length(src))
  }

  if (guess_size) {
    if (!requireNamespace("magick", quietly = TRUE)) {
      stop("package magick is required when using `guess_size` option.")
    }
    sizes <- lapply(src, function(x) {
      if (grepl("\\.svg$", x)) {
        z <- magick::image_read_svg(x)
      } else {
        z <- magick::image_read(x)
      }
      z <- magick::image_data(z)
      attr(z, "dim")[-1]
    })
    sizes <- do.call(rbind, sizes)
    width <- sizes[, 1] / 72
    height <- sizes[, 2] / 72
  }

  class(src) <- c("floating_external_img", "external_img", "cot", "run")
  attr(src, "dims") <- list(width = width, height = height)
  attr(src, "pos") <- list(
    x = pos_x,
    y = pos_y,
    h_from = pos_h_from,
    v_from = pos_v_from
  )
  attr(src, "wrap") <- list(
    type = wrap_type,
    side = wrap_side,
    dist_top = wrap_dist_top,
    dist_bottom = wrap_dist_bottom,
    dist_left = wrap_dist_left,
    dist_right = wrap_dist_right
  )
  attr(src, "alt") <- alt
  src
}

#' @export
to_wml.floating_external_img <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  pos <- attr(x, "pos")
  pos_x <- pos$x
  pos_y <- pos$y
  pos_h_from <- pos$h_from
  pos_v_from <- pos$v_from

  wrap <- attr(x, "wrap")
  wrap_type <- wrap$type
  wrap_side <- wrap$side
  wrap_dist_top <- wrap$dist_top
  wrap_dist_bottom <- wrap$dist_bottom
  wrap_dist_left <- wrap$dist_left
  wrap_dist_right <- wrap$dist_right

  blipfill <- temp_blipfill(x, ns = "pic")

  # Convert dimensions to EMUs (English Metric Units: 1 inch = 914400 EMUs)
  cx <- sprintf("%.0f", width * 12700 * 72)
  cy <- sprintf("%.0f", height * 12700 * 72)

  # Convert positions to EMUs
  pos_x_emu <- sprintf("%.0f", pos_x * 12700 * 72)
  pos_y_emu <- sprintf("%.0f", pos_y * 12700 * 72)

  # Convert wrap distances to EMUs
  dist_t_emu <- sprintf("%.0f", wrap_dist_top * 914400)
  dist_b_emu <- sprintf("%.0f", wrap_dist_bottom * 914400)
  dist_l_emu <- sprintf("%.0f", wrap_dist_left * 914400)
  dist_r_emu <- sprintf("%.0f", wrap_dist_right * 914400)

  # Build wrapping element
  wrap_element <- switch(
    wrap_type,
    "square" = sprintf("<wp:wrapSquare wrapText=\"%s\"/>", wrap_side),
    "topAndBottom" = "<wp:wrapTopAndBottom/>",
    "through" = sprintf(
      "<wp:wrapThrough wrapText=\"%s\"><wp:wrapPolygon edited=\"0\"><wp:start x=\"0\" y=\"0\"/></wp:wrapPolygon></wp:wrapThrough>",
      wrap_side
    ),
    "tight" = sprintf(
      "<wp:wrapTight wrapText=\"%s\"><wp:wrapPolygon edited=\"0\"><wp:start x=\"0\" y=\"0\"/></wp:wrapPolygon></wp:wrapTight>",
      wrap_side
    ),
    "none" = "<wp:wrapNone/>"
  )

  paste0(
    open_tag,
    "<w:rPr/><w:drawing xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\">",
    sprintf(
      "<wp:anchor distT=\"%s\" distB=\"%s\" distL=\"%s\" distR=\"%s\" simplePos=\"0\" relativeHeight=\"251658240\" behindDoc=\"0\" locked=\"0\" layoutInCell=\"1\" allowOverlap=\"1\">",
      dist_t_emu,
      dist_b_emu,
      dist_l_emu,
      dist_r_emu
    ),
    "<wp:simplePos x=\"0\" y=\"0\"/>",
    sprintf(
      "<wp:positionH relativeFrom=\"%s\"><wp:posOffset>%s</wp:posOffset></wp:positionH>",
      pos_h_from,
      pos_x_emu
    ),
    sprintf(
      "<wp:positionV relativeFrom=\"%s\"><wp:posOffset>%s</wp:posOffset></wp:positionV>",
      pos_v_from,
      pos_y_emu
    ),
    sprintf("<wp:extent cx=\"%s\" cy=\"%s\"/>", cx, cy),
    "<wp:effectExtent l=\"0\" t=\"0\" r=\"0\" b=\"0\"/>",
    wrap_element,
    "<wp:docPr id=\"\" name=\"\" descr=\"",
    attr(x, "alt"),
    "\"/>",
    "<wp:cNvGraphicFramePr><a:graphicFrameLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/></wp:cNvGraphicFramePr>",
    "<a:graphic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"><a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/picture\"><pic:pic xmlns:pic=\"http://schemas.openxmlformats.org/drawingml/2006/picture\">",
    "<pic:nvPicPr>",
    "<pic:cNvPr id=\"\" name=\"\"/>",
    "<pic:cNvPicPr/>",
    "</pic:nvPicPr>",
    blipfill,
    "<pic:spPr><a:xfrm><a:off x=\"0\" y=\"0\"/>",
    sprintf("<a:ext cx=\"%s\" cy=\"%s\"/>", cx, cy),
    "</a:xfrm><a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom></pic:spPr>",
    "</pic:pic></a:graphicData></a:graphic>",
    "</wp:anchor></w:drawing></w:r>"
  )
}

# run_footnoteref ----
#' @export
#' @title Word footnote reference
#' @description Wraps a footnote reference in an object that can then be inserted
#' as a run/chunk with [fpar()] or within an R Markdown document.
#' @param prop formatting text properties returned by
#' [fp_text_lite()] or [fp_text()]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @family run functions for reporting
#' @examples
#' run_footnoteref()
#' to_wml(run_footnoteref())
run_footnoteref <- function(prop = NULL) {
  z <- list(pr = prop)
  class(z) <- c("run_footnoteref", "run")
  z
}
#' @export
to_wml.run_footnoteref <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  str <- paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:footnoteRef/></w:r>"
  )

  str
}

# run_footnote ----
#' @export
#' @title Footnote for 'Word'
#' @description Wraps a footnote in an object that can then be inserted
#' as a run/chunk with [fpar()] or within an R Markdown document.
#' @param x a set of blocks to be used as footnote content returned by
#'   function [block_list()].
#' @param prop formatting text properties returned by
#' [fp_text_lite()] or [fp_text()]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @family run functions for reporting
#' @examples
#' library(officer)
#'
#' fp_bold <- fp_text_lite(bold = TRUE)
#' fp_refnote <- fp_text_lite(vertical.align = "superscript")
#'
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#' bl <- block_list(
#'   fpar(ftext("hello", fp_bold)),
#'   fpar(
#'     ftext("hello world", fp_bold),
#'     external_img(src = img.file, height = 1.06, width = 1.39)
#'   )
#' )
#'
#' a_par <- fpar(
#'   "this paragraph contains a note ",
#'   run_footnote(x = bl, prop = fp_refnote),
#'   "."
#' )
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, value = a_par, style = "Normal")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
run_footnote <- function(x, prop = NULL) {
  z <- list(footnote = x, pr = prop)
  class(z) <- c("run_footnote", "run")
  z
}
#' @export
to_wml.run_footnote <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  x$footnote[[1]]$chunks <- append(
    list(run_footnoteref(x$pr)),
    x$footnote[[1]]$chunks
  )
  blocks <- sapply(x$footnote, to_wml)
  blocks <- paste(blocks, collapse = "")

  base_ns <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\""
  footnote_xml <- paste0(
    "<w:footnote ",
    base_ns,
    " w:id=\"%s\">",
    blocks,
    "</w:footnote>"
  )

  footnote_ref_xml <- paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:footnoteReference w:officer=\"true\" w:id=\"%s\">",
    footnote_xml,
    "</w:footnoteReference>",
    "</w:r>"
  )

  footnote_ref_xml
}

# run_comment ----
#' @export
#' @title Comment for 'Word'
#' @description Add a comment on a run object.
#' @param cmt a set of blocks to be used as comment content returned by
#'   function [block_list()].
#' the "run functions for reporting".
#' @param author comment author.
#' @param date comment date
#' @param initials comment initials
#' @param prop formatting text properties returned by
#' [fp_text_lite()] or [fp_text()]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @param run a run object, made with a call to one of
#' @examples
#' fp_bold <- fp_text_lite(bold = TRUE)
#' fp_red <- fp_text_lite(color = "red")
#'
#' bl <- block_list(
#'   fpar(ftext("Comment multiple words.", fp_bold)),
#'   fpar(
#'     ftext("Second line.", fp_red)
#'   )
#' )
#'
#' comment1 <- run_comment(
#'   cmt = bl,
#'   run = ftext("with a comment"),
#'   author = "Author Me",
#'   date = Sys.Date(),
#'   initials = "AM"
#' )
#' par1 <- fpar("A paragraph ", comment1)
#'
#' bl <- block_list(
#'   fpar(ftext("Comment a paragraph."))
#' )
#'
#' comment2 <- run_comment(
#'   cmt = bl, run = ftext("A commented paragraph"),
#'   author = "Author You",
#'   date = Sys.Date(),
#'   initials = "AY"
#' )
#' par2 <- fpar(comment2)
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, value = par1, style = "Normal")
#' doc <- body_add_fpar(doc, value = par2, style = "Normal")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_comment <- function(
  cmt,
  run = ftext(""),
  author = "",
  date = "",
  initials = "",
  prop = NULL
) {
  all_run <- FALSE

  if (inherits(run, "run")) {
    run <- list(run)
  }

  if (is.list(run) && !inherits(run, "run")) {
    all_run <- all(vapply(run, inherits, FUN.VALUE = FALSE, what = "run"))
  }

  if (!all_run) {
    stop(
      "`run` must be a run object (ftext for example) or a list of run objects."
    )
  }

  z <- list(
    comment = cmt,
    run = run,
    author = author,
    date = date,
    initials = initials,
    pr = prop
  )
  class(z) <- c("run_comment", "run")
  z
}

#' @export
to_wml.run_comment <- function(x, add_ns = FALSE, ...) {
  runs <- lapply(x$run, to_wml, add_ns = add_ns, ...)
  runs <- do.call(paste0, runs)

  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  blocks <- sapply(x$comment, to_wml)
  blocks <- paste(blocks, collapse = "")

  id <- basename(tempfile(pattern = "comment"))
  base_ns <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\""
  comment_xml <- paste0(
    sprintf(
      "<w:comment %s  w:id=\"%s\" w:author=\"%s\" w:date=\"%s\" w:initials=\"%s\">",
      base_ns,
      id,
      x$author,
      x$date,
      x$initials
    ),
    blocks,
    "</w:comment>"
  )

  comment_ref_xml <- paste0(
    "<w:commentRangeStart w:officer=\"true\" w:id=\"",
    id,
    "\"/>",
    runs,
    "<w:commentRangeEnd w:officer=\"true\" w:id=\"",
    id,
    "\"/>",
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:commentReference w:officer=\"true\" w:id=\"",
    id,
    "\">",
    comment_xml,
    "</w:commentReference>",
    "</w:r>"
  )

  comment_ref_xml
}

test_that("to_rtf works with default strings and ftext", {
  expect_equal(to_rtf(NULL), "")

  str <- "Default string"
  expect_equal(to_rtf(str), str)

  properties2 <- fp_text(bold = TRUE, shading.color = "yellow")
  ft <- ftext("Some text", properties2)
  expect_equal(
    to_rtf(ft),
    "%font:Arial%\\b\\fs20%ftcolor:black% %ftshading:yellow%Some text\\highlight0"
  )
})

test_that("to_rtf works with fpar and external images", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")

  bold_face <- shortcuts$fp_bold(font.size = 12)
  bold_redface <- update(bold_face, color = "red")
  fpar_1 <- fpar(
    "Hello World, ",
    ftext("how ", prop = bold_redface),
    external_img(src = img.file, height = 1.06 / 2, width = 1.39 / 2),
    ftext(" you?", prop = bold_face)
  )
  expect_true(grepl("Hello World, ", to_rtf(fpar_1)))
})

test_that("to_rtf works with run_word_field, run_pagebreak, run_columnbreak, and run_linebreak", {
  out <- to_rtf(run_word_field(field = "PAGE  \\* MERGEFORMAT"))

  expect_true(grepl("PAGE", out))

  fp_t <- fp_text(font.size = 12, bold = TRUE)
  an_fpar <- fpar(
    "let's add a break page",
    run_pagebreak(),
    ftext("and blah blah!", fp_t)
  )

  expect_true(grepl(" let's add a break page", to_rtf(an_fpar)))
  expect_true(grepl("and blah blah", to_rtf(an_fpar)))

  expect_true(grepl("column", to_rtf(run_columnbreak())))

  fp_t <- fp_text(font.size = 12, bold = TRUE)
  an_fpar <- fpar(
    "let's add a line break",
    run_linebreak(),
    ftext("and blah blah!", fp_t)
  )

  expect_true(grepl(" let's add a line break", to_rtf(an_fpar)))
  expect_true(grepl("and blah blah", to_rtf(an_fpar)))
})

test_that("to_rtf works with hyperlinks and block_list", {
  ft <- fp_text(font.size = 12, bold = TRUE)
  ft <- hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "some text",
    prop = ft
  )

  expect_true(grepl("HYPERLINK", to_rtf(ft)))

  fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
  fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)

  value <- block_list(
    fpar(ftext("hello world\\t", fpt_blue_bold)),
    fpar(
      ftext("hello", fpt_blue_bold),
      " ",
      ftext("world", fpt_red_italic)
    ),
    fpar(
      ftext("hello world", fpt_red_italic)
    )
  )

  expect_true(grepl("C32900", to_rtf(value)))
  expect_true(grepl("006699", to_rtf(value)))
  expect_true(grepl("hello world", to_rtf(value)))
})

test_that("to_rtf works for run_autonum", {
  ra <- run_autonum(
    seq_id = "tab",
    pre_label = "Table ",
    bkm = "anytable",
    tnd = 2,
    tns = " "
  )

  expect_true(grepl("bkmkstart anytable", to_rtf(ra)))
  expect_true(grepl("SEQ tab", to_rtf(ra)))

  rr <- run_reference("a_ref")

  expect_true(grepl("REF a_ref ", to_rtf(rr)))

  ps <- prop_section(
    page_size = page_size(orient = "landscape"),
    page_margins = page_mar(top = 2),
    type = "continuous"
  )
  bs <- block_section(ps)

  expect_true(grepl("lndscpsxn", to_rtf(bs)))
  expect_true(grepl("pghsxn11906", to_rtf(bs)))
  expect_true(grepl("margt2880", to_rtf(bs))) # top = 2

  expect_true(grepl("margb1417", to_rtf(page_mar())))
  expect_true(grepl("colw3600", to_rtf(section_columns())))
  expect_true(grepl("pgwsxn16838", to_rtf(page_size(orient = "landscape"))))
  expect_true(grepl("lndscpsxn", to_rtf(page_size(orient = "landscape"))))
})

test_that("tcpr_rtf is retrieved by format", {
  obj <- fp_cell(margin = 1)
  fp_c <- update(obj, margin.bottom = 5)

  expect_true(grepl("clvertalc", format(fp_c, "rtf")))
})

test_that("to_rtf with fp_par_lite", {
  x <- fpar("hello", fp_p = fp_par_lite())
  expect_equal(to_rtf(x), "{hello\\par}")
  x <- fpar("hello", fp_p = fp_par_lite(text.align = "center"))
  expect_equal(to_rtf(x), "{\\pard\\qc hello\\par}")
})

test_that("border_rtf", {
  x <- fp_border()
  expect_equal(
    border_rtf(x, side = "right"),
    "\\clbrdrr\\brdrs\\brdrw20%ftlinecolor:black%"
  )
  expect_equal(
    border_rtf(x, side = "left"),
    "\\clbrdrl\\brdrs\\brdrw20%ftlinecolor:black%"
  )
  expect_equal(
    border_rtf(x, side = "top"),
    "\\clbrdrt\\brdrs\\brdrw20%ftlinecolor:black%"
  )
  expect_equal(
    border_rtf(x, side = "bottom"),
    "\\clbrdrb\\brdrs\\brdrw20%ftlinecolor:black%"
  )
})

test_that("fp_text_lite and rtf", {
  x <- fp_text_lite()
  expect_equal(format(x, type = "rtf"), "")
  x <- fp_text_lite(vertical.align = "superscript", font.size = 12)
  expect_equal(format(x, type = "rtf"), "\\super\\up6\\fs24 ")
  x <- fp_text_lite(vertical.align = "subscript", font.size = 12)
  expect_equal(format(x, type = "rtf"), "\\sub\\dn6\\fs24 ")
})

txt_lorem <- "Tellus nunc facilisis, quam tempor tempor. Purus lectus eros metus turpis mattis platea praesent sed. Sed ac, interdum tincidunt auctor mi tempor a pulvinar ut massa. Accumsan id fusce at platea! Magna tincidunt parturient ultricies netus vestibulum libero ex ornare in molestie, risus. Tempor nibh mollis pharetra vitae mauris nec cum et curabitur, lorem quam semper, lectus. Rutrum in in at eu non, diam. Eu fermentum ac donec eleifend. Condimentum neque etiam sit euismod mollis sollicitudin dictumst."
img.file <- file.path(R.home("doc"), "html", "logo.jpg")

bl_header_first <- block_list(
  fpar(
    ftext("header_first "),
    external_img(src = img.file, height = 1.06 / 2, width = 1.39 / 2)
  )
)

bl_header_even <- block_list(fpar(ftext("header_even")))
bl_header_odd <- block_list(fpar(ftext("header_odd")))
bl_footer_first <- block_list(fpar(ftext("footer_first")))
bl_footer_even <- block_list(fpar(ftext("footer_even")))
bl_footer_odd <- block_list(fpar(ftext("footer_odd")))

x <- read_docx()

for (i in 1:40) {
  x <- body_add_par(x, value = txt_lorem)
}


x <- body_set_default_section(
  x,
  value = prop_section(
    page_margins = page_mar(bottom = 2),
    page_size = page_size(orient = "portrait"),
    header_first = bl_header_first,
    footer_first = bl_footer_first,
    header_default = bl_header_odd,
    header_even = bl_header_even,
    footer_default = bl_footer_odd,
    footer_even = bl_footer_even
  )
)
filename <- print(x, target = tempfile(fileext = ".docx"))

test_that("add header and footer in docx", {
  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)header$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 3)
  if (nrow(subset_rel) > 0) {
    pattern <- "^(header)([0-9]+)(\\.xml)$"
    expect_true(all(grepl(pattern, subset_rel$target)))
    nodes_def_sectpr <- xml_find_all(
      docx_body_xml(x),
      "//w:body/w:sectPr/w:headerReference"
    )
    expect_length(nodes_def_sectpr, 3)
    nodes_def_sectpr <- xml_find_all(
      docx_body_xml(x),
      "//w:body/w:sectPr/w:footerReference"
    )
    expect_length(nodes_def_sectpr, 3)

    first_str <- readLines(file.path(
      x$package_dir,
      "word/_rels/header3.xml.rels"
    ))
    expect_true(any(grepl("\\.jpg", first_str)))
  }
  nodes_first_sectpr <- xml_find_first(
    docx_body_xml(x),
    "//w:body/w:sectPr/w:headerReference[@w:type='first']"
  )
  expect_false(inherits(nodes_first_sectpr, "xml_missing"))

  id <- xml_attr(nodes_first_sectpr, "id")
  target <- subset_rel$target[subset_rel$id %in% id]

  xml_doc <- read_xml(file.path(x$package_dir, "word", target))
  expect_equal(xml_text(xml_doc), "header_first ")

  sect_title_node <- xml_find_first(
    docx_body_xml(x),
    "//w:body/w:sectPr/w:titlePg"
  )
  expect_false(inherits(sect_title_node, "xml_missing"))
})


library(officer)
txt_lorem <- "Tellus nunc facilisis, quam tempor tempor. Purus lectus eros metus turpis mattis platea praesent sed. Sed ac, interdum tincidunt auctor mi tempor a pulvinar ut massa. Accumsan id fusce at platea! Magna tincidunt parturient ultricies netus vestibulum libero ex ornare in molestie, risus. Tempor nibh mollis pharetra vitae mauris nec cum et curabitur, lorem quam semper, lectus. Rutrum in in at eu non, diam. Eu fermentum ac donec eleifend. Condimentum neque etiam sit euismod mollis sollicitudin dictumst."

x <- read_docx()

for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}

x <- body_end_block_section(
  x,
  value = block_section(prop_section(
    page_size = page_size(orient = "portrait")
  ))
)

for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}

x <- body_end_block_section(
  x,
  value = block_section(prop_section(
    page_size = page_size(orient = "landscape")
  ))
)

for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}

x <- body_set_default_section(
  x,
  value = prop_section(
    page_margins = page_mar(),
    page_size = page_size(orient = "portrait")
  )
)
filename <- print(x, target = tempfile(fileext = ".docx"))


test_that("add docx landscape and portrait page dims", {
  x <- read_docx(path = filename)
  node_types <- xml_find_all(
    docx_body_xml(x),
    "//w:sectPr/w:pgSz"
  )
  types <- xml_attr(node_types, "orient")
  expect_equal(types, c("portrait", "landscape", "portrait"))

  h <- xml_attr(node_types, "h")
  expect_equal(h, c("16838", "11906", "16838"))
  w <- xml_attr(node_types, "w")
  expect_equal(w, c("11906", "16838", "11906"))
})

img.file <- file.path(R.home("doc"), "html", "logo.jpg")
fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)
bl <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(
    ftext("hello", fpt_blue_bold),
    " ",
    ftext("world", fpt_red_italic)
  ),
  fpar(
    ftext("hello world", fpt_red_italic),
    external_img(
      src = img.file,
      height = 1.06,
      width = 1.39
    )
  )
)

anyplot <- plot_instr(code = {
  col <- c(
    "#440154FF",
    "#443A83FF",
    "#31688EFF",
    "#21908CFF",
    "#35B779FF",
    "#8FD744FF",
    "#FDE725FF"
  )
  barplot(1:7, col = col, yaxt = "n")
})

test_that("rtf_add works with text, paragraphs, and plots (ggplot2 too)", {
  def_text <- fp_text_lite(color = "#006943", bold = TRUE)
  center_par <- fp_par(text.align = "left", padding = 1, line_spacing = 1.3)

  np <- fp_par(line_spacing = 1.4, padding = 3, )
  fpt_def <- fp_text(
    font.size = 11,
    italic = TRUE,
    bold = TRUE,
    underline = TRUE
  )

  doc <- rtf_doc(normal_par = np, normal_chunk = fpt_def)

  expect_identical(doc$normal_par, np)
  expect_identical(doc$normal_chunk, fpt_def)
  expect_identical(doc$content, list())

  doc <- rtf_add(
    x = doc,
    value = fpar(
      ftext("how are you?", prop = def_text),
      fp_p = fp_par(text.align = "center")
    )
  )

  expect_identical(
    doc$content[[1]]$chunks[[1]],
    ftext("how are you?", prop = def_text)
  )
  expect_identical(doc$content[[1]]$fp_p, fp_par(text.align = "center"))

  a_paragraph <- fpar(
    ftext("Here is a date: ", prop = def_text),
    run_word_field(field = "Date \\@ \"MMMM d yyyy\""),
    fp_p = center_par
  )
  doc <- rtf_add(
    x = doc,
    value = block_list(
      a_paragraph,
      a_paragraph,
      a_paragraph
    )
  )

  expect_identical(doc$content[[4]]$chunks, a_paragraph$chunks)

  if (require("ggplot2")) {
    gg <- gg_plot <- ggplot(data = iris) +
      geom_point(mapping = aes(Sepal.Length, Petal.Length))
    doc <- rtf_add(doc, gg, width = 3, height = 4, ppr = center_par)

    expect_true(grepl("\\.png", doc$content[[5]]$chunks[[1]]))
    expect_identical(
      attr(doc$content[[5]]$chunks[[1]], "dims"),
      list(width = 3, height = 4)
    )
  }
  anyplot <- plot_instr(code = {
    barplot(1:5, col = 2:6)
  })

  doc <- rtf_add(doc, anyplot, width = 5, height = 4, ppr = center_par)
  expect_true(grepl("\\.png", doc$content[[6]]$chunks[[1]]))
  expect_identical(
    attr(doc$content[[6]]$chunks[[1]], "dims"),
    list(width = 5, height = 4)
  )

  expect_s3_class(doc, "rtf")

  expect_identical(
    capture.output(print.rtf(doc)),
    "rtf document with 6 element(s)"
  )

  bl <- block_list(
    fpar(ftext("hello world\\t", fpt_blue_bold)),
    fpar(
      ftext("hello", fpt_blue_bold),
      " ",
      ftext("world", fpt_red_italic)
    ),
    fpar(
      ftext("hello world", fpt_red_italic)
    )
  )

  expect_silent({
    doc <- rtf_add(doc, bl)
  })

  ps <- prop_section(
    page_size = page_size(orient = "landscape"),
    page_margins = page_mar(top = 2),
    type = "continuous"
  )
  bs <- block_section(ps)

  expect_silent({
    doc <- rtf_add(doc, bs)
  })
  expect_silent({
    doc <- rtf_add(doc, "a character")
  })
  expect_silent({
    doc <- rtf_add(doc, factor("a factor"))
  })
  expect_silent({
    doc <- rtf_add(doc, 1.1)
  })

  outfile <- print(doc, target = tempfile(fileext = ".rtf"))
  expect_true(file.exists(outfile))
})

test_that("body_add_table", {
  x <- read_docx()
  x <- body_add_table(x, value = iris, style = "table_template")
  node <- docx_current_block_xml(x)
  expect_equal(xml_name(node), "tbl")
})

test_that("wml structure", {
  x <- read_docx()
  x <- body_add_table(x, value = iris, style = "table_template")
  node <- docx_current_block_xml(x)
  expect_length(xml_find_all(node, xpath = "w:tr[w:trPr/w:tblHeader]"), 1)
  expect_length(
    xml_find_all(node, xpath = "w:tr[not(w:trPr/w:tblHeader)]"),
    nrow(iris)
  )
  expect_length(
    xml_find_all(
      node,
      xpath = "w:tr[not(w:trPr/w:tblHeader)]/w:tc/w:p/w:r/w:t"
    ),
    ncol(iris) * nrow(iris)
  )
})

test_that("block_caption", {
  run_num <- run_autonum(
    seq_id = "tab",
    pre_label = "tab. ",
    bkm = "mtcars_table"
  )

  expect_output(
    print(block_caption("mtcars table", style = "Normal")),
    "caption \\[autonum off\\]: mtcars table"
  )
  expect_output(
    print(block_caption("mtcars table", style = "Normal", autonum = run_num)),
    "caption \\[autonum on\\]: mtcars table"
  )

  caption <- block_caption("mtcars table", autonum = run_num)
  expect_equal(caption$style, "Normal")

  expect_match(
    to_wml(caption, knitting = TRUE),
    "::: \\{custom-style=\"Normal\"\\}"
  )
  expect_match(to_wml(caption, knitting = TRUE), "mtcars table\\n:::")
  expect_match(to_wml(caption, knitting = FALSE), "^<w:p>")
  expect_match(to_wml(caption, knitting = FALSE), "</w:p>$")
  expect_match(to_wml(caption, knitting = FALSE), "mtcars table</w:t>")
})

test_that("names stay as is", {
  df <- data.frame(
    "hello coco" = c(1, 2),
    value = c("одно значение", "еще значение"),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )
  x <- read_docx()
  x <- body_add_table(x, value = df, style = "table_template")
  node <- docx_current_block_xml(x)
  first_text <- xml_find_first(
    node,
    xpath = "w:tr[w:trPr/w:tblHeader]/w:tc/w:p/w:r/w:t"
  )
  first_text <- xml_text(first_text)
  expect_equal(first_text, "hello coco")
})

test_that("add comments", {
  fp_bold <- fp_text_lite(bold = TRUE)
  fp_red <- fp_text_lite(color = "red")

  bl <- block_list(
    fpar(ftext("Comment multiple words.", fp_bold)),
    fpar(
      ftext("Second line.", fp_red)
    )
  )

  a_par <- fpar(
    "This paragraph contains",
    run_comment(
      cmt = bl,
      run = ftext("a comment."),
      author = "Author Me",
      date = "2023-06-01"
    )
  )

  b_par <- fpar(
    run_comment(
      cmt = block_list(
        fpar(
          ftext("Comment a paragraph.")
        )
      ),
      run = ftext("This paragraph contains another comment."),
      author = "Author You",
      date = "2023-06-01",
      initials = "OM"
    )
  )
  doc <- read_docx()
  doc <- body_add_fpar(doc, value = a_par, style = "Normal")
  doc <- body_add_fpar(doc, value = b_par, style = "Normal")

  docx_file <- print(doc, target = tempfile(fileext = ".docx"))
  docx_dir <- tempfile()
  unpack_folder(docx_file, docx_dir)

  doc <- read_xml(file.path(docx_dir, "word/comments.xml"))
  comment1 <- xml_find_first(doc, "w:comment[@w:id='0']")
  comment2 <- xml_find_first(doc, "w:comment[@w:id='1']")

  expect_false(inherits(comment1, "xml_missing"))
  expect_false(inherits(comment2, "xml_missing"))

  expect_length(xml_children(comment1), 2)
  expect_length(xml_children(comment2), 1)
})

test_that("docx_comments accounts for multiple comments in a paragraph", {
  multi_comment_par <- fpar(
    "This paragraph",
    run_comment(
      cmt = block_list(
        fpar(
          ftext("First Comment.")
        )
      ),
      run = ftext("contains"),
      author = "Author Me",
      date = "2023-06-01"
    ),
    "multiple",
    run_comment(
      cmt = block_list(
        fpar(
          ftext("Second Comment.")
        )
      ),
      run = ftext("comments"),
      author = "Author Me",
      date = "2023-06-01"
    )
  )

  doc <- read_docx()
  doc <- body_add_fpar(doc, value = multi_comment_par, style = "Normal")
  docx_file <- print(doc, target = tempfile(fileext = ".docx"))

  comments <- docx_comments(read_docx(docx_file))

  expect_equal(nrow(comments), 2)

  expect_equal(
    comments$text,
    list("First Comment.", "Second Comment.")
  )
  expect_equal(
    comments$commented_text,
    list("contains", "comments")
  )
})

test_that("docx_comments", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  # Doc includes 16 comments
  expect_equal(nrow(comments), 16)
  # No NAs in "commented_text"
  expect_true(
    all(
      vapply(
        comments[["commented_text"]],
        function(x) all(!is.na(x)),
        FUN.VALUE = logical(1)
      )
    )
  )
  # Accounts for empty comments or multi line comments
  expect_true(
    all(
      lengths(comments[["text"]][-c(1, 4)]) == 1
    )
  )
  ## Comment 1 has 2 lines
  expect_equal(
    length(comments[["text"]][[1]]),
    2
  )
  ## Comment 4 is empty
  expect_identical(
    comments[["text"]][[4]],
    character(0)
  )
})

test_that("docx_comments accounts for comments spanning no or multiple paragraphs", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  expect_true(
    all(
      lengths(comments[["para_id"]][-c(5, 6)]) == 1
    )
  )
  ## Comment 5 spans no paragraph
  expect_identical(
    comments[["para_id"]][[5]],
    character(0)
  )
  expect_equal(
    paste(comments[["commented_text"]][[5]], collapse = " "),
    ""
  )
  ## Comment 6 spans 2 paragraphs
  expect_equal(
    length(comments[["para_id"]][[6]]),
    2
  )
  expect_equal(
    paste(comments[["commented_text"]][[6]], collapse = " "),
    "a comment … … which spans multiple paragraphs."
  )
})

test_that("docx_comments accounts for comments spanning no or multiple runs", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  expect_true(
    all(
      lengths(comments[["commented_text"]][-c(5, 6, 7, 15)]) == 1
    )
  )
  ## Comment 5 spans no run
  expect_identical(
    comments[["commented_text"]][[5]],
    character(0)
  )
  expect_equal(
    paste(comments[["commented_text"]][[5]], collapse = " "),
    ""
  )
  ## Comment 6 spans 2 runs as it spans 2 paragraphs
  expect_equal(
    length(comments[["commented_text"]][[6]]),
    2
  )
  expect_equal(
    paste(comments[["commented_text"]][[6]], collapse = " "),
    "a comment … … which spans multiple paragraphs."
  )
  ## Comment 7 spans 3 runs.
  expect_equal(
    length(comments[["commented_text"]][[7]]),
    3
  )
  expect_equal(
    paste(comments[["commented_text"]][[7]], collapse = ""),
    "spanning multiple runs."
  )
  ## Comment 15 spans 3 runs because of an inner comment
  expect_equal(
    length(comments[["commented_text"]][[15]]),
    3
  )
  expect_equal(
    paste(comments[["commented_text"]][[15]], collapse = ""),
    "This paragraph contains two nested comments."
  )
})

test_that("docx_comments accounts for replies", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  # Make "unique" id based on commented text and paragraph id
  comments$unique_id <- paste(
    comments$para_id,
    comments$commented_text,
    sep = "."
  )
  comments$unique_id <- factor(
    comments$unique_id,
    levels = unique(comments$unique_id)
  )
  comments$unique_id <- as.integer(comments$unique_id)

  comments_split <- split(
    comments,
    comments$unique_id
  )

  # Accouting for replies we have only 12 comments
  expect_equal(
    length(comments_split),
    12
  )
  expect_equal(
    vapply(
      comments_split[-c(8, 9, 10)],
      nrow,
      FUN.VALUE = integer(1),
      USE.NAMES = FALSE
    ),
    rep(1, 9)
  )
  ## 8 has two replies
  expect_equal(
    nrow(comments_split[[8]]),
    3
  )
  expect_equal(
    unique(unlist(comments[8:10, "commented_text"])),
    comments[["commented_text"]][[8]]
  )
  expect_equal(
    unique(unlist(comments[8:10, "para_id"])),
    comments[["para_id"]][[8]]
  )
  ## 9 and 10 have one reply each
  expect_equal(
    nrow(comments_split[[9]]),
    2
  )
  expect_equal(
    unique(unlist(comments[11:12, "commented_text"])),
    comments[["commented_text"]][[11]]
  )
  expect_equal(
    unique(unlist(comments[11:12, "para_id"])),
    comments[["para_id"]][[11]]
  )
  expect_equal(
    nrow(comments_split[[10]]),
    2
  )
  expect_equal(
    unique(unlist(comments[13:14, "commented_text"])),
    comments[["commented_text"]][[13]]
  )
  expect_equal(
    unique(unlist(comments[13:14, "para_id"])),
    comments[["para_id"]][[13]]
  )
})

test_that("docx_comments accounts for nested comments", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  ## Outer Comment 15 spans 3 runs
  expect_equal(
    length(comments[["commented_text"]][[15]]),
    3
  )
  expect_equal(
    paste(comments[["commented_text"]][[15]], collapse = ""),
    "This paragraph contains two nested comments."
  )
  expect_equal(
    paste(comments[["text"]][[15]], collapse = ""),
    "Outer Comment."
  )

  ## Inner Comment 16 spans 1 run
  expect_equal(
    length(comments[["commented_text"]][[16]]),
    1
  )
  expect_equal(
    paste(comments[["commented_text"]][[16]], collapse = ""),
    "contains two "
  )
  expect_equal(
    paste(comments[["text"]][[16]], collapse = ""),
    "Inner Comment."
  )
})

source("utils.R")

get_wml_node <- function(x) {
  xml_ <- to_wml(x)
  read_xml(wml_str(xml_))
}
get_pml_node <- function(x) {
  xml_ <- pml_str(to_pml(x))
  read_xml(xml_)
}

test_that("fpar content", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  fp <- fpar(ftext("toto", fp_text(font.size = 9)), "titi")
  fp <- update(fp, fp_t = fp_text(bold = TRUE, font.size = 9))
  df <- as.data.frame(fp)
  expect_equal(df$value, c("toto", "titi"))
  expect_equal(df$size, c(9, 9))
  expect_equal(df$bold, c(FALSE, TRUE))
  expect_equal(df$italic, c(FALSE, FALSE))
})


test_that("fpar wml", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  node <- get_wml_node(fp)

  expect_equal(xml_text(node), "tototiti")
  expect_equal(xml_text(xml_find_all(node, "//w:r/w:t")), c("toto", "titi"))
  val <- xml_attr(xml_find_all(node, "//w:pPr/w:jc"), "val")
  expect_equal(val, "left")

  fp <- update(fp, fp_p = fp_par(text.align = "center"))
  node <- get_wml_node(fp)
  val <- xml_find_all(node, "//w:pPr/w:jc")
  val <- xml_attr(val, "val")
  expect_equal(val, "center")

  fp <- fpar(
    ftext("pi", fp_text()),
    ftext(" = ", fp_text(italic = TRUE)),
    pi
  )
  fp <- update(fp, fp_t = fp_text(bold = TRUE))
  node <- get_wml_node(fp)
  italic <- xml_find_all(node, "//w:r/w:rPr")
  italic <- xml_child(italic, "w:i")
  italic <- xml_attr(italic, "val") %in% "true"

  bold <- xml_find_all(node, "//w:r/w:rPr")
  bold <- xml_child(bold, "w:b")
  bold <- xml_attr(bold, "val") %in% "true"

  expect_equal(italic, c(FALSE, TRUE, FALSE))
  expect_equal(bold, c(FALSE, FALSE, TRUE))
})


test_that("fpar pml", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  node <- get_pml_node(fp)

  expect_equal(xml_text(node), "tototiti")
  expect_equal(xml_text(xml_find_all(node, "//a:r/a:t")), c("toto", "titi"))
  expect_equal(xml_attr(xml_find_all(node, "//a:r/a:rPr"), "b"), c("1", "0"))
  expect_equal(xml_attr(xml_find_all(node, "//a:r/a:rPr"), "i"), c("0", "1"))
  val <- xml_attr(xml_find_all(node, "//a:p/a:pPr"), "algn")
  expect_equal(val, "l")

  fp <- update(fp, fp_p = fp_par(text.align = "center"))
  node <- get_pml_node(fp)
  val <- xml_attr(xml_find_all(node, "//a:p/a:pPr"), "algn")
  expect_equal(val, "ctr")
})


test_that("fpar css", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  node <- read_xml(to_html(fp))

  expect_equal(xml_text(node), "tototiti")
  expect_equal(xml_text(xml_find_all(node, "//span")), c("toto", "titi"))
  bold <- xml_attr(xml_find_all(node, "//span"), "style")
  bold <- has_css_attr(bold, atname = "font-weight", value = "bold")
  expect_equal(bold, c(TRUE, FALSE))

  italic <- xml_attr(xml_find_all(node, "//span"), "style")
  italic <- has_css_attr(italic, atname = "font-style", value = "italic")
  expect_equal(italic, c(FALSE, TRUE))

  val <- xml_attr(xml_find_all(node, "//p"), "style")
  val <- has_css_attr(val, atname = "text-align", value = "left")
  expect_true(val)

  fp <- update(fp, fp_p = fp_par(text.align = "center"))
  node <- read_xml(to_html(fp))
  val <- xml_attr(xml_find_all(node, "//p"), "style")
  val <- has_css_attr(val, atname = "text-align", value = "center")
  expect_true(val)
})

test_that("add footnotes", {
  fp_bold <- fp_text_lite(bold = TRUE)
  fp_refnote <- fp_text_lite(vertical.align = "superscript")

  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  bl <- block_list(
    fpar(ftext("hello", fp_bold)),
    fpar(
      ftext("hello world", fp_bold),
      external_img(src = img.file, height = 1.06, width = 1.39)
    )
  )

  a_par <- fpar(
    "this paragraph contains a note ",
    run_footnote(x = bl, prop = fp_refnote),
    "."
  )
  b_par <- fpar(
    "this paragraph contains another note ",
    run_footnote(
      x = block_list(
        fpar(ftext("Salut", fp_text_lite()))
      ),
      prop = fp_refnote
    ),
    "."
  )

  doc <- read_docx()
  doc <- body_add_fpar(doc, value = a_par, style = "Normal")
  doc <- body_add_fpar(doc, value = b_par, style = "Normal")

  docx_file <- print(doc, target = tempfile(fileext = ".docx"))
  docx_dir <- tempfile()
  unpack_folder(docx_file, docx_dir)

  doc <- read_xml(file.path(docx_dir, "word/footnotes.xml"))
  footnote1 <- xml_find_first(doc, "w:footnote[@w:id='1']")
  footnote2 <- xml_find_first(doc, "w:footnote[@w:id='2']")

  expect_false(inherits(footnote1, "xml_missing"))
  expect_false(inherits(footnote2, "xml_missing"))

  expect_length(xml_children(footnote1), 2)
  expect_length(xml_children(footnote2), 1)
})

txt_lorem <- "Tellus nunc facilisis, quam tempor tempor. Purus lectus eros metus turpis mattis platea praesent sed. Sed ac, interdum tincidunt auctor mi tempor a pulvinar ut massa. Accumsan id fusce at platea! Magna tincidunt parturient ultricies netus vestibulum libero ex ornare in molestie, risus. Tempor nibh mollis pharetra vitae mauris nec cum et curabitur, lorem quam semper, lectus. Rutrum in in at eu non, diam. Eu fermentum ac donec eleifend. Condimentum neque etiam sit euismod mollis sollicitudin dictumst."
img.file <- file.path( R.home("doc"), "html", "logo.jpg" )

value_first <- block_list(fpar(ftext("hello first "), external_img(src = img.file, height = 1.06/2, width = 1.39/2)))
value_even <- block_list(fpar(ftext("hello even")))
value_default <- block_list(fpar(ftext("hello default")))
footer_first <- block_list(fpar(ftext("bye first")))
footer_even <- block_list(fpar(ftext("bye even")))
footer_default <- block_list(fpar(ftext("bye default")))

ps <- prop_section(
  header_default = value_default, footer_default = footer_default,
  header_first = value_first, footer_first = footer_first,
  header_even = value_even, footer_even = footer_even
)
x <- read_docx()
for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}
x <- body_end_block_section(x, value = block_section(ps))
for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}
x <- body_set_default_section(
  x,
  value = prop_section(
    page_margins = page_mar(bottom = 2),
    page_size = page_size(width = 21 / 2.54, height = 29.7 / 2.54, orient = "portrait"),
    header_default = block_list(fpar(ftext("yok yok on top"))),
    header_even = block_list(fpar(ftext("yok yok on even top"))),
    footer_default = block_list(fpar(ftext("Chewbacca on bottom"))),
    footer_even = block_list(fpar(ftext("Chewbacca on even bottom")))
  )
)
filename <- print(x, target = tempfile(fileext = ".docx"))

test_that("add header and footer in docx", {
  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)header$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 5)
  if (nrow(subset_rel) > 0) {
    pattern <- "^(header)([0-9]+)(\\.xml)$"
    expect_true(all(grepl(pattern, subset_rel$target)))
    nodes_def_sectpr <- xml_find_all(docx_body_xml(x), "//w:body/w:sectPr/w:headerReference")
    expect_length(nodes_def_sectpr, 2)
    nodes_def_sectpr <- xml_find_all(docx_body_xml(x), "//w:body/w:sectPr/w:footerReference")
    expect_length(nodes_def_sectpr, 2)

    first_str <- readLines(file.path(x$package_dir, "word/_rels/header5.xml.rels"))
    expect_true(any(grepl("\\.jpg", first_str)))
  }
})

test_that("margin and sizes are copied", {
  x <- read_docx(path = filename)
  node_def_sectpr <- xml_find_first(docx_body_xml(x), "//w:body/w:sectPr")
  nodes_sectpr <- xml_find_all(docx_body_xml(x), "//w:pPr/w:sectPr")
  pgMar <- as.character(xml_child(node_def_sectpr, "w:pgMar"))
  pgSz <- as.character(xml_child(node_def_sectpr, "w:pgSz"))
  for (i in seq_along(nodes_sectpr)) {
    expect_equal(as.character(xml_child(nodes_sectpr[[i]], "w:pgMar")), pgMar)
    expect_equal(as.character(xml_child(nodes_sectpr[[i]], "w:pgSz")), pgSz)
  }
})

img.file <- file.path(R.home("doc"), "html", "logo.jpg")
fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)
bl <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(
    ftext("hello", fpt_blue_bold), " ",
    ftext("world", fpt_red_italic)
  ),
  fpar(
    ftext("hello world", fpt_red_italic),
    external_img(
      src = img.file, height = 1.06, width = 1.39
    )
  )
)

anyplot <- plot_instr(code = {
  col <- c(
    "#440154FF", "#443A83FF", "#31688EFF",
    "#21908CFF", "#35B779FF", "#8FD744FF", "#FDE725FF"
  )
  barplot(1:7, col = col, yaxt = "n")
})

test_that("rtf_add works with text, paragraphs, and plots (ggplot2 too)", {
  def_text <- fp_text_lite(color = "#006943", bold = TRUE)
  center_par <- fp_par(text.align = "left", padding = 1, line_spacing = 1.3)

  np <- fp_par(line_spacing = 1.4, padding = 3, )
  fpt_def <- fp_text(font.size = 11, italic = TRUE, bold = TRUE, underline = TRUE)

  doc <- rtf_doc(normal_par = np, normal_chunk = fpt_def)

  expect_identical(doc$normal_par, np)
  expect_identical(doc$normal_chunk, fpt_def)
  expect_identical(doc$content, list())

  doc <- rtf_add(
    x = doc,
    value = fpar(
      ftext("how are you?", prop = def_text),
      fp_p = fp_par(text.align = "center")
    )
  )

  expect_identical(doc$content[[1]]$chunks[[1]], ftext("how are you?", prop = def_text))
  expect_identical(doc$content[[1]]$fp_p, fp_par(text.align = "center"))

  a_paragraph <- fpar(
    ftext("Here is a date: ", prop = def_text),
    run_word_field(field = "Date \\@ \"MMMM d yyyy\""),
    fp_p = center_par
  )
  doc <- rtf_add(
    x = doc,
    value = block_list(
      a_paragraph,
      a_paragraph,
      a_paragraph
    )
  )

  expect_identical(doc$content[[4]]$chunks, a_paragraph$chunks)

  if (require("ggplot2")) {
    gg <- gg_plot <- ggplot(data = iris) +
      geom_point(mapping = aes(Sepal.Length, Petal.Length))
    doc <- rtf_add(doc, gg,
      width = 3, height = 4,
      ppr = center_par
    )

    expect_true(grepl("\\.png", doc$content[[5]]$chunks[[1]]))
    expect_identical(attr(doc$content[[5]]$chunks[[1]], "dims"), list(width = 3, height = 4))
  }
  anyplot <- plot_instr(code = {
    barplot(1:5, col = 2:6)
  })

  doc <- rtf_add(doc, anyplot,
    width = 5, height = 4,
    ppr = center_par
  )
  expect_true(grepl("\\.png", doc$content[[6]]$chunks[[1]]))
  expect_identical(attr(doc$content[[6]]$chunks[[1]], "dims"), list(width = 5, height = 4))

  expect_s3_class(doc, "rtf")

  expect_identical(capture.output(print.rtf(doc)), "rtf document with 6 element(s)")

  bl <- block_list(
    fpar(ftext("hello world\\t", fpt_blue_bold)),
    fpar(
      ftext("hello", fpt_blue_bold), " ",
      ftext("world", fpt_red_italic)
    ),
    fpar(
      ftext("hello world", fpt_red_italic)
    )
  )

  expect_silent({doc <- rtf_add(doc, bl)})

  ps <- prop_section(
    page_size = page_size(orient = "landscape"),
    page_margins = page_mar(top = 2),
    type = "continuous"
  )
  bs <- block_section(ps)

  expect_silent({doc <- rtf_add(doc, bs)})
  expect_silent({doc <- rtf_add(doc, "a character")})
  expect_silent({doc <- rtf_add(doc, factor("a factor"))})
  expect_silent({doc <- rtf_add(doc, 1.1)})

  outfile <- print(doc, target = tempfile(fileext = ".rtf"))
  expect_true(file.exists(outfile))
})

test_that("body_add_table", {
  x <- read_docx()
  x <- body_add_table(x, value = iris, style = "table_template")
  node <- docx_current_block_xml(x)
  expect_equal(xml_name(node), "tbl")
})

test_that("wml structure", {
  x <- read_docx()
  x <- body_add_table(x, value = iris, style = "table_template")
  node <- docx_current_block_xml(x)
  expect_length(xml_find_all(node, xpath = "w:tr[w:trPr/w:tblHeader]"), 1)
  expect_length(xml_find_all(node, xpath = "w:tr[not(w:trPr/w:tblHeader)]"), nrow(iris))
  expect_length(xml_find_all(node, xpath = "w:tr[not(w:trPr/w:tblHeader)]/w:tc/w:p/w:r/w:t"), ncol(iris) * nrow(iris))
})

test_that("block_caption", {

  run_num <- run_autonum(seq_id = "tab", pre_label = "tab. ",
                         bkm = "mtcars_table")

  expect_output(print(block_caption("mtcars table", style = "Normal")),
                "caption \\[autonum off\\]: mtcars table")
  expect_output(print(block_caption("mtcars table", style = "Normal", autonum = run_num)),
                "caption \\[autonum on\\]: mtcars table")

  caption <- block_caption("mtcars table", autonum = run_num)
  expect_equal(caption$style, "Normal")

  expect_match(to_wml(caption, knitting = TRUE), "::: \\{custom-style=\"Normal\"\\}")
  expect_match(to_wml(caption, knitting = TRUE), "mtcars table\\n:::")
  expect_match(to_wml(caption, knitting = FALSE), "^<w:p>")
  expect_match(to_wml(caption, knitting = FALSE), "</w:p>$")
  expect_match(to_wml(caption, knitting = FALSE), "mtcars table</w:t>")
})

test_that("names stay as is", {
  df <- data.frame(
    "hello coco" = c(1, 2), value = c("одно значение", "еще значение"),
    check.names = FALSE, stringsAsFactors = FALSE
  )
  x <- read_docx()
  x <- body_add_table(x, value = df, style = "table_template")
  node <- docx_current_block_xml(x)
  first_text <- xml_find_first(node, xpath = "w:tr[w:trPr/w:tblHeader]/w:tc/w:p/w:r/w:t")
  first_text <- xml_text(first_text)
  expect_equal(first_text, "hello coco")
})

test_that("add comments", {
  fp_bold <- fp_text_lite(bold = TRUE)
  fp_red <- fp_text_lite(color = "red")

  bl <- block_list(
    fpar(ftext("Comment multiple words.", fp_bold)),
    fpar(
      ftext("Second line.", fp_red)
    )
  )

  a_par <- fpar(
    "This paragraph contains",
    run_comment(
      cmt = bl,
      run = ftext("a comment."),
      author = "Author Me",
      date = "2023-06-01"
    )
  )

  b_par <- fpar(
    run_comment(
      cmt = block_list(
        fpar(
          ftext("Comment a paragraph.")
        )
      ),
      run = ftext("This paragraph contains another comment."),
      author = "Author You",
      date = "2023-06-01",
      initials = "OM"
    )
  )
  doc <- read_docx()
  doc <- body_add_fpar(doc, value = a_par, style = "Normal")
  doc <- body_add_fpar(doc, value = b_par, style = "Normal")

  docx_file <- print(doc, target = tempfile(fileext = ".docx"))
  docx_dir <- tempfile()
  unpack_folder(docx_file, docx_dir)

  doc <- read_xml(file.path(docx_dir, "word/comments.xml"))
  comment1 <- xml_find_first(doc, "w:comment[@w:id='0']")
  comment2 <- xml_find_first(doc, "w:comment[@w:id='1']")

  expect_false(inherits(comment1, "xml_missing"))
  expect_false(inherits(comment2, "xml_missing"))

  expect_length(xml_children(comment1), 2)
  expect_length(xml_children(comment2), 1)
})

test_that("docx_comments accounts for multiple comments in a paragraph", {
  multi_comment_par <- fpar(
    "This paragraph",
    run_comment(
      cmt = block_list(
        fpar(
          ftext("First Comment.")
        )
      ),
      run = ftext("contains"),
      author = "Author Me",
      date = "2023-06-01"
    ),
    "multiple",
    run_comment(
      cmt = block_list(
        fpar(
          ftext("Second Comment.")
        )
      ),
      run = ftext("comments"),
      author = "Author Me",
      date = "2023-06-01"
    )
  )

  doc <- read_docx()
  doc <- body_add_fpar(doc, value = multi_comment_par, style = "Normal")
  docx_file <- print(doc, target = tempfile(fileext = ".docx"))

  comments <- docx_comments(read_docx(docx_file))

  expect_equal(nrow(comments), 2)

  expect_equal(
    comments$text,
    list("First Comment.", "Second Comment.")
  )
  expect_equal(
    comments$commented_text,
    list("contains", "comments")
  )
})

test_that("docx_comments", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  # Doc includes 16 comments
  expect_equal(nrow(comments), 16)
  # No NAs in "commented_text"
  expect_true(
    all(
      vapply(
        comments[["commented_text"]],
        function(x) all(!is.na(x)),
        FUN.VALUE = logical(1)
      )
    )
  )
  # Accounts for empty comments or multi line comments
  expect_true(
    all(
      lengths(comments[["text"]][-c(1, 4)]) == 1
    )
  )
  ## Comment 1 has 2 lines
  expect_equal(
    length(comments[["text"]][[1]]), 2
  )
  ## Comment 4 is empty
  expect_identical(
    comments[["text"]][[4]], character(0)
  )
})

test_that("docx_comments accounts for comments spanning no or multiple paragraphs", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  expect_true(
    all(
      lengths(comments[["para_id"]][-c(5, 6)]) == 1
    )
  )
  ## Comment 5 spans no paragraph
  expect_identical(
    comments[["para_id"]][[5]], character(0)
  )
  expect_equal(
    paste(comments[["commented_text"]][[5]], collapse = " "),
    ""
  )
  ## Comment 6 spans 2 paragraphs
  expect_equal(
    length(comments[["para_id"]][[6]]), 2
  )
  expect_equal(
    paste(comments[["commented_text"]][[6]], collapse = " "),
    "a comment … … which spans multiple paragraphs."
  )
})

test_that("docx_comments accounts for comments spanning no or multiple runs", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  expect_true(
    all(
      lengths(comments[["commented_text"]][-c(5, 6, 7, 15)]) == 1
    )
  )
  ## Comment 5 spans no run
  expect_identical(
    comments[["commented_text"]][[5]], character(0)
  )
  expect_equal(
    paste(comments[["commented_text"]][[5]], collapse = " "),
    ""
  )
  ## Comment 6 spans 2 runs as it spans 2 paragraphs
  expect_equal(
    length(comments[["commented_text"]][[6]]), 2
  )
  expect_equal(
    paste(comments[["commented_text"]][[6]], collapse = " "),
    "a comment … … which spans multiple paragraphs."
  )
  ## Comment 7 spans 3 runs.
  expect_equal(
    length(comments[["commented_text"]][[7]]), 3
  )
  expect_equal(
    paste(comments[["commented_text"]][[7]], collapse = ""),
    "spanning multiple runs."
  )
  ## Comment 15 spans 3 runs because of an inner comment
  expect_equal(
    length(comments[["commented_text"]][[15]]), 3
  )
  expect_equal(
    paste(comments[["commented_text"]][[15]], collapse = ""),
    "This paragraph contains two nested comments."
  )
})

test_that("docx_comments accounts for replies", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  # Make "unique" id based on commented text and paragraph id
  comments$unique_id <- paste(
    comments$para_id, comments$commented_text,
    sep = "."
  )
  comments$unique_id <- factor(
    comments$unique_id,
    levels = unique(comments$unique_id)
  )
  comments$unique_id <- as.integer(comments$unique_id)

  comments_split <- split(
    comments,
    comments$unique_id
  )

  # Accouting for replies we have only 12 comments
  expect_equal(
    length(comments_split),
    12
  )
  expect_equal(
    vapply(
      comments_split[-c(8, 9, 10)], nrow,
      FUN.VALUE = integer(1), USE.NAMES = FALSE
    ),
    rep(1, 9)
  )
  ## 8 has two replies
  expect_equal(
    nrow(comments_split[[8]]),
    3
  )
  expect_equal(
    unique(unlist(comments[8:10, "commented_text"])),
    comments[["commented_text"]][[8]]
  )
  expect_equal(
    unique(unlist(comments[8:10, "para_id"])),
    comments[["para_id"]][[8]]
  )
  ## 9 and 10 have one reply each
  expect_equal(
    nrow(comments_split[[9]]),
    2
  )
  expect_equal(
    unique(unlist(comments[11:12, "commented_text"])),
    comments[["commented_text"]][[11]]
  )
  expect_equal(
    unique(unlist(comments[11:12, "para_id"])),
    comments[["para_id"]][[11]]
  )
  expect_equal(
    nrow(comments_split[[10]]),
    2
  )
  expect_equal(
    unique(unlist(comments[13:14, "commented_text"])),
    comments[["commented_text"]][[13]]
  )
  expect_equal(
    unique(unlist(comments[13:14, "para_id"])),
    comments[["para_id"]][[13]]
  )
})

test_that("docx_comments accounts for nested comments", {
  example_docx <- "docs_dir/test-docx_comments.docx"
  doc <- read_docx(path = example_docx)

  comments <- docx_comments(doc)

  ## Outer Comment 15 spans 3 runs
  expect_equal(
    length(comments[["commented_text"]][[15]]), 3
  )
  expect_equal(
    paste(comments[["commented_text"]][[15]], collapse = ""),
    "This paragraph contains two nested comments."
  )
  expect_equal(
    paste(comments[["text"]][[15]], collapse = ""),
    "Outer Comment."
  )

  ## Inner Comment 16 spans 1 run
  expect_equal(
    length(comments[["commented_text"]][[16]]), 1
  )
  expect_equal(
    paste(comments[["commented_text"]][[16]], collapse = ""),
    "contains two "
  )
  expect_equal(
    paste(comments[["text"]][[16]], collapse = ""),
    "Inner Comment."
  )
})

source("utils.R")

get_wml_node <- function(x){
  xml_ <- to_wml(x)
  read_xml( wml_str(xml_) )
}
get_pml_node <- function(x){
  xml_ <- pml_str(to_pml(x))
  read_xml( xml_ )
}

test_that("fpar content", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
    )
  fp <- fpar( ftext("toto", fp_text(font.size = 9)), "titi" )
  fp <- update(fp, fp_t = fp_text(bold = TRUE, font.size = 9))
  df <- as.data.frame(fp)
  expect_equal(df$value, c("toto", "titi") )
  expect_equal(df$size, c(9, 9) )
  expect_equal(df$bold, c(FALSE, TRUE) )
  expect_equal(df$italic, c(FALSE, FALSE) )
})


test_that("fpar wml", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  node <- get_wml_node(fp)

  expect_equal(xml_text(node), "tototiti")
  expect_equal(xml_text(xml_find_all(node, "//w:r/w:t")), c("toto", "titi") )
  val <- xml_attr(xml_find_all(node, "//w:pPr/w:jc"), "val")
  expect_equal(val, "left")

  fp <- update(fp, fp_p = fp_par(text.align = "center"))
  node <- get_wml_node(fp)
  val <- xml_find_all(node, "//w:pPr/w:jc")
  val <- xml_attr(val, "val")
  expect_equal(val, "center")

  fp <- fpar(
    ftext("pi", fp_text()),
    ftext(" = ", fp_text(italic = TRUE)),
    pi
  )
  fp <- update(fp, fp_t = fp_text(bold = TRUE))
  node <- get_wml_node(fp)
  italic <- xml_find_all(node, "//w:r/w:rPr")
  italic <- xml_child(italic, "w:i")
  italic <- xml_attr(italic, "val") %in% "true"

  bold <- xml_find_all(node, "//w:r/w:rPr")
  bold <- xml_child(bold, "w:b")
  bold <- xml_attr(bold, "val") %in% "true"

  expect_equal(italic, c(FALSE, TRUE, FALSE) )
  expect_equal(bold, c(FALSE, FALSE, TRUE) )
})



test_that("fpar pml", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  node <- get_pml_node(fp)

  expect_equal(xml_text(node), "tototiti")
  expect_equal(xml_text(xml_find_all(node, "//a:r/a:t")), c("toto", "titi") )
  expect_equal(xml_attr(xml_find_all(node, "//a:r/a:rPr"), "b"), c("1", "0") )
  expect_equal(xml_attr(xml_find_all(node, "//a:r/a:rPr"), "i"), c("0", "1") )
  val <- xml_attr(xml_find_all(node, "//a:p/a:pPr"), "algn")
  expect_equal(val, "l")


  fp <- update(fp, fp_p = fp_par(text.align = "center"))
  node <- get_pml_node(fp)
  val <- xml_attr(xml_find_all(node, "//a:p/a:pPr"), "algn")
  expect_equal(val, "ctr")
})



test_that("fpar css", {
  fp <- fpar(
    ftext("toto", shortcuts$fp_bold()),
    ftext("titi", shortcuts$fp_italic())
  )
  node <- read_xml( to_html(fp) )

  expect_equal(xml_text(node), "tototiti")
  expect_equal(xml_text(xml_find_all(node, "//span")), c("toto", "titi") )
  bold <- xml_attr(xml_find_all(node, "//span"), "style")
  bold <- has_css_attr(bold, atname = "font-weight", value = "bold")
  expect_equal(bold, c(TRUE, FALSE) )

  italic <- xml_attr(xml_find_all(node, "//span"), "style")
  italic <- has_css_attr(italic, atname = "font-style", value = "italic")
  expect_equal(italic, c(FALSE, TRUE) )

  val <- xml_attr(xml_find_all(node, "//p"), "style")
  val <- has_css_attr(val, atname = "text-align", value = "left")
  expect_true(val)


  fp <- update(fp, fp_p = fp_par(text.align = "center"))
  node <- read_xml( to_html(fp) )
  val <- xml_attr(xml_find_all(node, "//p"), "style")
  val <- has_css_attr(val, atname = "text-align", value = "center")
  expect_true(val)
})

test_that("add footnotes", {

  fp_bold <- fp_text_lite(bold = TRUE)
  fp_refnote <- fp_text_lite(vertical.align = "superscript")

  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  bl <- block_list(
    fpar(ftext("hello", fp_bold)),
    fpar(
      ftext("hello world", fp_bold),
      external_img(src = img.file, height = 1.06, width = 1.39)
    )
  )

  a_par <- fpar(
    "this paragraph contains a note ",
    run_footnote(x = bl, prop = fp_refnote),
    "."
  )
  b_par <- fpar(
    "this paragraph contains another note ",
    run_footnote(
      x = block_list(
        fpar(ftext("Salut", fp_text_lite()))
      ),
      prop = fp_refnote),
    "."
  )

  doc <- read_docx()
  doc <- body_add_fpar(doc, value = a_par, style = "Normal")
  doc <- body_add_fpar(doc, value = b_par, style = "Normal")

  docx_file <- print(doc, target = tempfile(fileext = ".docx"))
  docx_dir <- tempfile()
  unpack_folder(docx_file, docx_dir)

  doc <- read_xml(file.path(docx_dir, "word/footnotes.xml"))
  footnote1 <- xml_find_first(doc, "w:footnote[@w:id='1']")
  footnote2 <- xml_find_first(doc, "w:footnote[@w:id='2']")

  expect_false( inherits(footnote1, "xml_missing") )
  expect_false( inherits(footnote2, "xml_missing") )

  expect_length(xml_children(footnote1), 2)
  expect_length(xml_children(footnote2), 1)

})

source("utils.R")

get_wml_node <- function(x) {
  xml_ <- to_wml(x)
  read_xml(wml_str(xml_))
}


test_that("run_autonum wml", {
  z <- expand.grid(
    pre_label = c("Table ", "Tableau "),
    post_label = c(": ", " : "),
    stringsAsFactors = FALSE
  )
  z$bkm <- c("bkm1", "bkm2", "bkm3", "null")
  z$bkm_all <- rep(c(TRUE, FALSE), 2)
  z$seq_id <- c("tab", "tbl", "fig", "plt")
  runs <- mapply(
    .pre = z$pre_label,
    .post = z$post_label,
    .bkm = z$bkm,
    .bkm_all = z$bkm_all,
    .seq_id = z$seq_id,
    FUN = function(.pre, .post, .bkm, .bkm_all, .seq_id) {
      z <- run_autonum(
        pre_label = .pre,
        seq_id = .seq_id,
        post_label = .post,
        bkm = if (.bkm %in% "null") NULL else .bkm,
        bkm_all = .bkm_all
      )
      to_wml(z)
    },
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )

  bkm_pattern <- "<w\\:bookmarkStart w\\:id=\"[0-9a-z\\-]+\" w\\:name=\"%s\"/>"
  expect_match(object = runs[[1]], regexp = sprintf(bkm_pattern, z$bkm[1]))
  expect_match(object = runs[[2]], regexp = sprintf(bkm_pattern, z$bkm[2]))
  expect_match(object = runs[[3]], regexp = sprintf(bkm_pattern, z$bkm[3]))
  expect_no_match(object = runs[[4]], regexp = sprintf(bkm_pattern, z$bkm[4]))

  expect_match(
    object = runs[[3]],
    regexp = sprintf(paste0("^", bkm_pattern), z$bkm[3])
  )
  expect_no_match(
    object = runs[[4]],
    regexp = sprintf(paste0("^", bkm_pattern), z$bkm[4])
  )

  node <- read_xml(wml_str(runs[[1]]))
  expect_equal(xml_text(xml_child(node, "w:r")), "Table ")
  expect_equal(xml_text(xml_child(node, xml_length(node))), ": ")
  node <- read_xml(wml_str(runs[[4]]))
  expect_equal(xml_text(xml_child(node, "w:r")), "Tableau ")
  expect_equal(xml_text(xml_child(node, xml_length(node))), " : ")

  seq_pattern <- "<w\\:instrText xml\\:space=\"preserve\" w\\:dirty=\"true\">SEQ %s"
  expect_match(object = runs[[1]], regexp = sprintf(seq_pattern, z$seq_id[1]))
  expect_match(object = runs[[2]], regexp = sprintf(seq_pattern, z$seq_id[2]))
  expect_match(object = runs[[3]], regexp = sprintf(seq_pattern, z$seq_id[3]))
  expect_match(object = runs[[4]], regexp = sprintf(seq_pattern, z$seq_id[4]))

  z <- run_autonum(seq_id = "tbl", start_at = 2, tnd = 2, tns = ".")
  node <- read_xml(wml_str(to_wml(z)))
  expect_equal(
    xml_text(node),
    "Table STYLEREF 2 \\r.SEQ tbl \\* Arabic \\r 2: "
  )
  z <- run_autonum(seq_id = "tbl", tnd = 1, tns = "-")
  node <- read_xml(wml_str(to_wml(z)))
  expect_equal(xml_text(node), "Table STYLEREF 1 \\r-SEQ tbl \\* Arabic: ")
})

test_that("to_rtf works with default strings and ftext", {
  expect_equal(to_rtf(NULL), "")

  str <- "Default string"
  expect_equal(to_rtf(str), str)

  properties2 <- fp_text(bold = TRUE, shading.color = "yellow")
  ft <- ftext("Some text", properties2)
  expect_equal(
    to_rtf(ft),
    "%font:Arial%\\b\\fs20%ftcolor:black% %ftshading:yellow%Some text\\highlight0"
  )
})

test_that("to_rtf works with fpar and external images", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")

  bold_face <- shortcuts$fp_bold(font.size = 12)
  bold_redface <- update(bold_face, color = "red")
  fpar_1 <- fpar(
    "Hello World, ",
    ftext("how ", prop = bold_redface),
    external_img(src = img.file, height = 1.06 / 2, width = 1.39 / 2),
    ftext(" you?", prop = bold_face)
  )
  expect_true(grepl("Hello World, ", to_rtf(fpar_1)))
})

test_that("to_rtf works with run_word_field, run_pagebreak, run_columnbreak, and run_linebreak", {
  out <- to_rtf(run_word_field(field = "PAGE  \\* MERGEFORMAT"))

  expect_true(grepl("PAGE", out))

  fp_t <- fp_text(font.size = 12, bold = TRUE)
  an_fpar <- fpar(
    "let's add a break page",
    run_pagebreak(),
    ftext("and blah blah!", fp_t)
  )

  expect_true(grepl(" let's add a break page", to_rtf(an_fpar)))
  expect_true(grepl("and blah blah", to_rtf(an_fpar)))

  expect_true(grepl("column", to_rtf(run_columnbreak())))

  fp_t <- fp_text(font.size = 12, bold = TRUE)
  an_fpar <- fpar(
    "let's add a line break",
    run_linebreak(),
    ftext("and blah blah!", fp_t)
  )

  expect_true(grepl(" let's add a line break", to_rtf(an_fpar)))
  expect_true(grepl("and blah blah", to_rtf(an_fpar)))
})

test_that("to_rtf works with hyperlinks and block_list", {
  ft <- fp_text(font.size = 12, bold = TRUE)
  ft <- hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "some text",
    prop = ft
  )

  expect_true(grepl("HYPERLINK", to_rtf(ft)))

  fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
  fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)

  value <- block_list(
    fpar(ftext("hello world\\t", fpt_blue_bold)),
    fpar(
      ftext("hello", fpt_blue_bold),
      " ",
      ftext("world", fpt_red_italic)
    ),
    fpar(
      ftext("hello world", fpt_red_italic)
    )
  )

  expect_true(grepl("C32900", to_rtf(value)))
  expect_true(grepl("006699", to_rtf(value)))
  expect_true(grepl("hello world", to_rtf(value)))
})

test_that("to_rtf works for run_autonum", {
  ra <- run_autonum(
    seq_id = "tab",
    pre_label = "Table ",
    bkm = "anytable",
    tnd = 2,
    tns = " "
  )

  expect_true(grepl("bkmkstart anytable", to_rtf(ra)))
  expect_true(grepl("SEQ tab", to_rtf(ra)))

  rr <- run_reference("a_ref")

  expect_true(grepl("REF a_ref ", to_rtf(rr)))

  ps <- prop_section(
    page_size = page_size(orient = "landscape"),
    page_margins = page_mar(top = 2),
    type = "continuous"
  )
  bs <- block_section(ps)

  expect_true(grepl("lndscpsxn", to_rtf(bs)))
  expect_true(grepl("pghsxn11906", to_rtf(bs)))
  expect_true(grepl("margt2880", to_rtf(bs))) # top = 2

  expect_true(grepl("margb1417", to_rtf(page_mar())))
  expect_true(grepl("colw3600", to_rtf(section_columns())))
  expect_true(grepl("pgwsxn16838", to_rtf(page_size(orient = "landscape"))))
  expect_true(grepl("lndscpsxn", to_rtf(page_size(orient = "landscape"))))
})

test_that("tcpr_rtf is retrieved by format", {
  obj <- fp_cell(margin = 1)
  fp_c <- update(obj, margin.bottom = 5)

  expect_true(grepl("clvertalc", format(fp_c, "rtf")))
})

test_that("to_rtf with fp_par_lite", {
  x <- fpar("hello", fp_p = fp_par_lite())
  expect_equal(to_rtf(x), "{hello\\par}")
  x <- fpar("hello", fp_p = fp_par_lite(text.align = "center"))
  expect_equal(to_rtf(x), "{\\pard\\qc hello\\par}")
})

test_that("border_rtf", {
  x <- fp_border()
  expect_equal(
    border_rtf(x, side = "right"),
    "\\clbrdrr\\brdrs\\brdrw20%ftlinecolor:black%"
  )
  expect_equal(
    border_rtf(x, side = "left"),
    "\\clbrdrl\\brdrs\\brdrw20%ftlinecolor:black%"
  )
  expect_equal(
    border_rtf(x, side = "top"),
    "\\clbrdrt\\brdrs\\brdrw20%ftlinecolor:black%"
  )
  expect_equal(
    border_rtf(x, side = "bottom"),
    "\\clbrdrb\\brdrs\\brdrw20%ftlinecolor:black%"
  )
})

test_that("fp_text_lite and rtf", {
  x <- fp_text_lite()
  expect_equal(format(x, type = "rtf"), "")
  x <- fp_text_lite(vertical.align = "superscript", font.size = 12)
  expect_equal(format(x, type = "rtf"), "\\super\\up6\\fs24 ")
  x <- fp_text_lite(vertical.align = "subscript", font.size = 12)
  expect_equal(format(x, type = "rtf"), "\\sub\\dn6\\fs24 ")
})

test_that("to_rtf.section_columns emits per-column widths (#726)", {
  x <- section_columns(widths = c(2, 3), space = 0.25)
  out <- officer:::to_rtf.section_columns(x)
  expect_equal(
    out,
    "\\cols2\\colsx360\\colno1\\colw2880\\colsr360\\colno2\\colw4320"
  )

  # sep = TRUE adds \linebetcol just after the global \colsx
  x2 <- section_columns(widths = c(2, 3), space = 0.25, sep = TRUE)
  expect_equal(
    officer:::to_rtf.section_columns(x2),
    "\\cols2\\colsx360\\linebetcol\\colno1\\colw2880\\colsr360\\colno2\\colw4320"
  )

  # three columns: two \colsr entries (one per non-last column)
  x3 <- section_columns(widths = c(1, 2, 3), space = 0.5)
  out3 <- officer:::to_rtf.section_columns(x3)
  expect_match(out3, "\\\\colno1\\\\colw1440\\\\colsr720")
  expect_match(out3, "\\\\colno2\\\\colw2880\\\\colsr720")
  expect_match(out3, "\\\\colno3\\\\colw4320(?!\\\\colsr)", perl = TRUE)
})

test_that("block_section emits \\sect before section properties (#726)", {
  bs <- block_section(prop_section(
    section_columns = section_columns(widths = c(2, 3), space = 0.25)
  ))
  out <- to_rtf(bs)
  # block_section() configures the section that follows: \sect closes
  # the previous section, then the property keywords apply to the new
  # one.
  expect_match(out, "^\\\\sect\\\\sectd")
  expect_match(out, "\\\\sectd.*\\\\cols2\\\\colsx360")
})

test_that("run_columnbreak emits a delimiter after \\column (#726)", {
  # The trailing space prevents "\columnRight" from being parsed as a
  # single unknown control word when text follows directly.
  expect_equal(to_rtf(run_columnbreak()), "\\column ")
})

test_that("each block_section emits its own {\\header}{\\footer} with resolved colors", {
  hf <- function(label, color) {
    block_list(fpar(ftext(label, fp_text_lite(color = color))))
  }
  default_section <- prop_section(
    header_default = hf("DEFAULT-H", "#006699"),
    footer_default = hf("DEFAULT-F", "#666666")
  )
  sec_b <- block_section(prop_section(
    type = "nextPage",
    header_default = hf("SECB-H", "#aa0000"),
    footer_default = hf("SECB-F", "#00aa00")
  ))

  doc <- rtf_doc(def_sec = default_section)
  doc <- rtf_add(doc, fpar(ftext("body of default section")))
  doc <- rtf_add(doc, sec_b)
  doc <- rtf_add(doc, fpar(ftext("body of section B")))

  target <- tempfile(fileext = ".rtf")
  print(doc, target = target)
  rtf <- paste(readLines(target, warn = FALSE), collapse = "\n")

  # both sections have header and footer groups
  expect_equal(length(gregexpr("\\{\\\\header", rtf)[[1]]), 2L)
  expect_equal(length(gregexpr("\\{\\\\footer", rtf)[[1]]), 2L)
  expect_match(rtf, "DEFAULT-H")
  expect_match(rtf, "SECB-H")
  expect_match(rtf, "SECB-F")
  # colors used only in headers/footers are registered and resolved (no raw placeholders)
  expect_false(grepl("%ftcolor:", rtf, fixed = TRUE))
  # default section header sits before the \sect that closes that section
  default_h_pos <- regexpr("DEFAULT-H", rtf)
  secb_props_pos <- regexpr("\\\\sect\\\\sectd", rtf)
  expect_true(default_h_pos < secb_props_pos)
})

source("utils.R")

get_wml_node <- function(x){
  xml_ <- to_wml(x)
  read_xml( wml_str(xml_) )
}


test_that("run_autonum wml", {

  z <- expand.grid(
    pre_label = c("Table ", "Tableau "),
    post_label = c(": ", " : "),
    stringsAsFactors = FALSE
  )
  z$bkm <- c("bkm1", "bkm2", "bkm3", "null")
  z$bkm_all <- rep(c(TRUE, FALSE), 2)
  z$seq_id <- c("tab", "tbl", "fig", "plt")
  runs <- mapply(.pre = z$pre_label,
         .post = z$post_label,
         .bkm = z$bkm,
         .bkm_all = z$bkm_all,
         .seq_id = z$seq_id,
         FUN = function(.pre, .post, .bkm, .bkm_all, .seq_id) {
           z <- run_autonum(
             pre_label = .pre,
             seq_id = .seq_id,
             post_label = .post,
             bkm = if (.bkm %in% "null") NULL else .bkm,
             bkm_all = .bkm_all)
           to_wml(z)
         }, SIMPLIFY = FALSE,
         USE.NAMES = FALSE)

  bkm_pattern <- "<w\\:bookmarkStart w\\:id=\"[0-9a-z\\-]+\" w\\:name=\"%s\"/>"
  expect_match(object = runs[[1]], regexp = sprintf(bkm_pattern, z$bkm[1]))
  expect_match(object = runs[[2]], regexp = sprintf(bkm_pattern, z$bkm[2]))
  expect_match(object = runs[[3]], regexp = sprintf(bkm_pattern, z$bkm[3]))
  expect_no_match(object = runs[[4]], regexp = sprintf(bkm_pattern, z$bkm[4]))

  expect_match(object = runs[[3]], regexp = sprintf(paste0("^", bkm_pattern), z$bkm[3]))
  expect_no_match(object = runs[[4]], regexp = sprintf(paste0("^", bkm_pattern), z$bkm[4]))

  node <- read_xml(wml_str(runs[[1]]))
  expect_equal(xml_text(xml_child(node, "w:r")), "Table ")
  expect_equal(xml_text(xml_child(node, xml_length(node))), ": ")
  node <- read_xml(wml_str(runs[[4]]))
  expect_equal(xml_text(xml_child(node, "w:r")), "Tableau ")
  expect_equal(xml_text(xml_child(node, xml_length(node))), " : ")

  seq_pattern <- "<w\\:instrText xml\\:space=\"preserve\" w\\:dirty=\"true\">SEQ %s"
  expect_match(object = runs[[1]], regexp = sprintf(seq_pattern, z$seq_id[1]))
  expect_match(object = runs[[2]], regexp = sprintf(seq_pattern, z$seq_id[2]))
  expect_match(object = runs[[3]], regexp = sprintf(seq_pattern, z$seq_id[3]))
  expect_match(object = runs[[4]], regexp = sprintf(seq_pattern, z$seq_id[4]))

  z <- run_autonum(seq_id = "tbl", start_at = 2, tnd = 2, tns = ".")
  node <- read_xml(wml_str(to_wml(z)))
  expect_equal(xml_text(node), "Table STYLEREF 2 \\r.SEQ tbl \\* Arabic \\r 2: ")
  z <- run_autonum(seq_id = "tbl", tnd = 1, tns = "-")
  node <- read_xml(wml_str(to_wml(z)))
  expect_equal(xml_text(node), "Table STYLEREF 1 \\r-SEQ tbl \\* Arabic: ")
})

fp_bold <- shortcuts$fp_bold()

test_that("block_list structure", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  bl <- block_list(
    fpar(ftext("hello world", fp_bold)),
    fpar(
      ftext("hello", fp_bold),
      external_img(src = img.file, height = 1.06, width = 1.39)
    )
  )

  expect_length(bl, 2)
  expect_length(bl[[2]]$chunks, 2)

  expect_is(bl[[2]]$chunks[[1]], "ftext")
  expect_is(bl[[2]]$chunks[[2]], "external_img")
  expect_is(bl[[2]]$chunks[[1]], "cot")
  expect_is(bl[[2]]$chunks[[1]], "run")
  expect_is(bl[[2]]$chunks[[2]], "cot")
})

fp_bold <- shortcuts$fp_bold()

test_that("block_list structure", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  bl <- block_list(
    fpar(ftext("hello world", fp_bold)),
    fpar(
      ftext("hello", fp_bold),
      external_img(src = img.file, height = 1.06, width = 1.39)
    )
  )

  expect_length(bl, 2)
  expect_length(bl[[2]]$chunks, 2)

  expect_is(bl[[2]]$chunks[[1]], "ftext")
  expect_is(bl[[2]]$chunks[[2]], "external_img")
  expect_is(bl[[2]]$chunks[[1]], "cot")
  expect_is(bl[[2]]$chunks[[1]], "run")
  expect_is(bl[[2]]$chunks[[2]], "cot")
})

# block_list_items / list_item ----

test_that("list_item and block_list_items constructors", {
  li <- list_item("hello", level = 2)
  expect_s3_class(li, "list_item")
  expect_s3_class(li$fpar, "fpar")
  expect_equal(li$level, 2L)

  li2 <- list_item(
    fpar(ftext("bold", fp_text(bold = TRUE))),
    level = 1
  )
  expect_s3_class(li2$fpar$chunks[[1]], "ftext")

  items <- block_list_items(
    list_item(fpar("a"), level = 1),
    list_item(fpar("b"), level = 2),
    list_type = "decimal"
  )
  expect_s3_class(items, "block_list_items")
  expect_equal(items$list_type, "decimal")
  expect_length(items$items, 2L)

  expect_error(
    block_list_items("not a list_item"),
    "list_item"
  )
  expect_error(
    block_list_items(list_item("a"), list_type = "invalid"),
    "should be one of"
  )
})

test_that("to_wml.block_list_items produces correct markers and levels", {
  items_bullet <- block_list_items(
    list_item(fpar("Item 1"), level = 1),
    list_item(fpar("Sub-item"), level = 2),
    list_type = "bullet"
  )
  wml <- to_wml(items_bullet)
  expect_true(grepl("officer-list-bullet-", wml))
  expect_true(grepl("w:ilvl w:val=\"0\"", wml))
  expect_true(grepl("w:ilvl w:val=\"1\"", wml))
  expect_false(grepl("w:ind w:left=\"0\"", wml))

  items_decimal <- block_list_items(
    list_item(fpar("First"), level = 1),
    list_item(fpar("Sub-first"), level = 2),
    list_type = "decimal"
  )
  wml2 <- to_wml(items_decimal)
  expect_true(grepl("officer-list-decimal-", wml2))
  expect_true(grepl("w:ilvl w:val=\"1\"", wml2))
})

test_that("to_pml.block_list_items produces correct bullet/number markup", {
  items_bullet <- block_list_items(
    list_item(fpar("Item 1"), level = 1),
    list_item(fpar("Sub-item"), level = 2),
    list_type = "bullet"
  )
  pml <- to_pml(items_bullet)
  expect_true(grepl("a:buChar", pml))
  expect_false(grepl("a:buNone", pml))
  expect_true(grepl("lvl=\"1\"", pml))

  items_decimal <- block_list_items(
    list_item(fpar("First"), level = 1),
    list_item(fpar("Sub-first"), level = 2),
    list_type = "decimal"
  )
  pml2 <- to_pml(items_decimal)
  expect_true(grepl("a:buAutoNum", pml2))
  expect_false(grepl("a:buNone", pml2))
  expect_true(grepl("lvl=\"1\"", pml2))
})

test_that("block_list_items Word round-trip resolves markers", {
  doc <- read_docx()
  bullets <- block_list_items(
    list_item(
      fpar(ftext("Red", fp_text(color = "red"))),
      level = 1
    ),
    list_item(fpar("Sub"), level = 2),
    list_type = "bullet"
  )
  numbers <- block_list_items(
    list_item(fpar("First"), level = 1),
    list_item(fpar("Sub-first"), level = 2),
    list_type = "decimal"
  )
  doc <- body_add(doc, bullets)
  doc <- body_add(doc, numbers)
  out <- print(doc, target = tempfile(fileext = ".docx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)

  doc_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "document.xml")
    ),
    collapse = ""
  )
  expect_false(grepl("officer-list", doc_xml))

  num_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "numbering.xml")
    ),
    collapse = ""
  )
  expect_true(grepl("w:val=\"bullet\"", num_xml))
  expect_true(grepl("w:val=\"decimal\"", num_xml))
})

test_that("block_list_items works with write_elements_to_context", {
  doc <- read_docx()
  bullets <- block_list_items(
    list_item(fpar("Bullet 1"), level = 1),
    list_item(fpar("Bullet 2"), level = 2),
    list_type = "bullet"
  )
  numbers <- block_list_items(
    list_item(fpar("Number 1"), level = 1),
    list_item(fpar("Number 2"), level = 1),
    list_type = "decimal"
  )
  ctx <- body_append_start_context(doc)
  ctx <- write_elements_to_context(ctx, bullets, numbers)
  doc <- body_append_stop_context(ctx)
  out <- print(doc, target = tempfile(fileext = ".docx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)

  doc_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "document.xml")
    ),
    collapse = ""
  )
  expect_false(grepl("officer-list", doc_xml))
})

test_that("body_add_list adds bullet and numbered lists", {
  doc <- read_docx()
  bullets <- block_list_items(
    list_item(fpar("Item 1"), level = 1),
    list_item(fpar("Sub-item"), level = 2),
    list_type = "bullet"
  )
  numbers <- block_list_items(
    list_item(fpar("First"), level = 1),
    list_item(fpar("Second"), level = 1),
    list_type = "decimal"
  )
  doc <- body_add_list(doc, items = bullets)
  doc <- body_add_list(doc, items = numbers)
  out <- print(doc, target = tempfile(fileext = ".docx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)

  doc_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "document.xml")
    ),
    collapse = ""
  )
  expect_false(grepl("officer-list", doc_xml))
  expect_true(grepl("w:numPr", doc_xml))

  num_xml <- paste(
    readLines(
      file.path(unpack_dir, "word", "numbering.xml")
    ),
    collapse = ""
  )
  expect_true(grepl("w:val=\"bullet\"", num_xml))
  expect_true(grepl("w:val=\"decimal\"", num_xml))
})

test_that("block_list_items PowerPoint output is valid", {
  items <- block_list_items(
    list_item(fpar("Item 1"), level = 1),
    list_item(fpar("Sub-item"), level = 2),
    list_type = "bullet"
  )
  ppt <- read_pptx()
  ppt <- add_slide(ppt, layout = "Title and Content")
  ppt <- ph_with(ppt, items, location = ph_location_type(type = "body"))
  out <- print(ppt, target = tempfile(fileext = ".pptx"))
  expect_true(file.exists(out))
  expect_gt(file.info(out)$size, 0)
})

library(officer)

x <- read_pptx()
layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
layout_properties(x = x, master = "Office Theme")
layout_properties(x = x, layout = "Two Content")
layout_properties(x = x)

# this name will be used to print the file
# change it to "youfile.pptx" to write the pptx
# file in your working directory.
fileout <- tempfile(fileext = ".pptx")

doc_1 <- read_pptx()
sz <- slide_size(doc_1)

# add text and a table ----
doc_1 <- add_slide(doc_1, layout = "Two Content", master = "Office Theme")
doc_1 <- ph_with(
  x = doc_1,
  value = c("Table cars"),
  location = ph_location_type(type = "title")
)
doc_1 <- ph_with(
  x = doc_1,
  value = names(cars),
  location = ph_location_left()
)
doc_1 <- ph_with(
  x = doc_1,
  value = cars,
  location = ph_location_right()
)
doc_1 <- ph_with(
  x = doc_1,
  value = Sys.Date(),
  location = ph_location_type("dt")
)

# add a base plot ----
anyplot <- plot_instr(code = {
  col <- c(
    "#440154FF",
    "#443A83FF",
    "#31688EFF",
    "#21908CFF",
    "#35B779FF",
    "#8FD744FF",
    "#FDE725FF"
  )
  barplot(1:7, col = col, yaxt = "n")
})

doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  doc_1,
  anyplot,
  location = ph_location_fullsize(),
  bg = "#006699"
)

# add a ggplot2 plot ----
if (require("ggplot2")) {
  doc_1 <- add_slide(doc_1, "Title and Content")
  gg_plot <- ggplot(data = iris) +
    geom_point(
      mapping = aes(Sepal.Length, Petal.Length),
      size = 3
    ) +
    theme_minimal()
  doc_1 <- ph_with(
    x = doc_1,
    value = gg_plot,
    location = ph_location_type(type = "body"),
    bg = "transparent"
  )
  doc_1 <- ph_with(
    x = doc_1,
    value = "graphic title",
    location = ph_location_type(type = "title")
  )
}

# add a external images ----
doc_1 <- add_slide(doc_1, layout = "Title and Content", master = "Office Theme")
doc_1 <- ph_with(
  x = doc_1,
  value = empty_content(),
  location = ph_location(
    left = 0,
    top = 0,
    width = sz$width,
    height = sz$height,
    bg = "black"
  )
)

svg_file <- file.path(R.home(component = "doc"), "html/Rlogo.svg")
if (require("rsvg")) {
  doc_1 <- ph_with(
    x = doc_1,
    value = "External images",
    location = ph_location_type(type = "title")
  )
  doc_1 <- ph_with(
    x = doc_1,
    external_img(svg_file, 100 / 72, 76 / 72),
    location = ph_location_right(),
    use_loc_size = FALSE
  )
  doc_1 <- ph_with(
    x = doc_1,
    external_img(svg_file),
    location = ph_location_left(),
    use_loc_size = TRUE
  )
}

# add a block_list ----
dummy_text <- readLines(system.file(
  package = "officer",
  "doc_examples/text.txt"
))
fp_1 <- fp_text(bold = TRUE, color = "pink", font.size = 0)
fp_2 <- fp_text(bold = TRUE, font.size = 0)
fp_3 <- fp_text(italic = TRUE, color = "red", font.size = 0)
bl <- block_list(
  fpar(ftext("hello world", fp_1)),
  fpar(
    ftext("hello", fp_2),
    ftext("hello", fp_3)
  ),
  dummy_text
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = bl,
  location = ph_location_type(type = "body")
)


# fpar ------
fpt <- fp_text(
  bold = TRUE,
  font.family = "Bradley Hand",
  font.size = 150,
  color = "#F5595B"
)
hw <- fpar(
  ftext("hello ", fpt),
  hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "cran",
    prop = fpt
  )
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = hw,
  location = ph_location_type(type = "body")
)
# unordered_list ----
ul <- unordered_list(
  level_list = c(1, 2, 2, 3, 3, 1),
  str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
  style = fp_text(color = "red", font.size = 0)
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = ul,
  location = ph_location_type()
)

print(doc_1, target = fileout)


# Example using short-form locations ----
x <- read_pptx()
x <- add_slide(x, "Title Slide")
x <- ph_with(x, "A title", "Title 1") # label
x <- ph_with(x, "A subtitle", 3) # id
x <- ph_with(x, "A left text", "left") # keyword
x <- ph_with(x, Sys.Date(), "dt[1]") # type + index
x <- ph_with(x, "More content", c(5, .5, 5, 2)) # numeric vector (left, top, width, heigh)
# \dontrun{
# print(x, preview = TRUE) # opens file locally
# }

read_core_properties <- function(package_dir) {
  filename <- file.path(package_dir, "docProps/core.xml")
  if (!file.exists(filename)) {
    # docx and pptx files exported from Google Docs or Slides do not have properties set. Use one from the template to get by
    warning(
      "No properties found. Using properties from template file in officer."
    )
    filename <- system.file(package = "officer", "template/core.xml")
  }
  doc <- read_xml(filename)
  ns_ <- xml_ns(doc)

  all_ <- xml_find_all(doc, "/cp:coreProperties/*")

  if (length(all_) < 1) {
    out <- list(
      data = structure(
        character(0),
        .Dim = c(0L, 4L),
        .Dimnames = list(
          NULL,
          c("ns", "name", "attrs", "value")
        )
      ),
      ns = c(
        cp = "http://schemas.openxmlformats.org/package/2006/metadata/core-properties",
        dc = "http://purl.org/dc/elements/1.1/",
        dcmitype = "http://purl.org/dc/dcmitype/",
        dcterms = "http://purl.org/dc/terms/",
        xsi = "http://www.w3.org/2001/XMLSchema-instance"
      )
    )
    return(out)
  }

  names_ <- sapply(all_, xml_name, ns = ns_)
  names_ <- strsplit(names_, ":")

  # concat all attrs in single chars
  attrs <- sapply(xml_attrs(all_), function(ns_) {
    paste0('xsi:', names(ns_), '=\"', ns_, '\"', collapse = " ")
  })
  attrs <- ifelse(sapply(xml_attrs(all_), length) < 1, "", paste0(" ", attrs))

  propnames <- sapply(names_, function(x) x[2])

  props <- matrix(
    c(sapply(names_, function(x) x[1]), propnames, attrs, xml_text(all_)),
    ncol = 4
  )
  colnames(props) <- c("ns", "name", "attrs", "value")
  rownames(props) <- propnames
  z <- list(data = props, ns = unclass(ns_))
  class(z) <- "core_properties"
  z
}

default_cp_tags <- list(
  title = list(ns = "dc", attrs = NA_character_),
  subject = list(ns = "dc", attrs = NA_character_),
  creator = list(ns = "dc", attrs = NA_character_),
  keywords = list(ns = "cp", attrs = NA_character_),
  description = list(ns = "dc", attrs = NA_character_),
  lastModifiedBy = list(ns = "cp", attrs = NA_character_),
  revision = list(ns = "cp", attrs = NA_character_),
  created = list(ns = "dcterms", attrs = 'xsi:type="dcterms:W3CDTF"'),
  modified = list(ns = "dcterms", attrs = 'xsi:type="dcterms:W3CDTF"'),
  category = list(ns = "cp", attrs = NA_character_)
)


#' @export
`[<-.core_properties` <- function(x, i, j, value) {
  if (!i %in% x$data[, "name"]) {
    attrs <- ifelse(
      is.na(default_cp_tags[[i]]$attrs),
      "",
      paste0(" ", default_cp_tags[[i]]$attrs)
    )
    new <- matrix(
      c(default_cp_tags[[i]]$ns, i, attrs, value),
      nrow = 1,
      dimnames = list(i, colnames(x$data))
    )
    x$data <- rbind(x$data, new)
  } else {
    x$data[i, j] <- value
  }
  x
}


#' @export
`[.core_properties` <- function(x, i, j) {
  x$data[i, j]
}


write_core_properties <- function(core_matrix, package_dir) {
  ns_ <- core_matrix$ns
  core_matrix <- core_matrix$data
  if (!is.matrix(core_matrix)) {
    stop("core_properties should be stored in a character matrix.")
  }

  ns_ <- paste0('xmlns:', names(ns_), '=\"', ns_, '\"', collapse = " ")
  xml_ <- paste0(
    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
    '\n',
    '<cp:coreProperties ',
    ns_,
    '>'
  )
  properties <- sprintf(
    "<%s:%s%s>%s</%s:%s>",
    core_matrix[, "ns"],
    core_matrix[, "name"],
    core_matrix[, "attrs"],
    htmlEscapeCopy(core_matrix[, "value"]),
    core_matrix[, "ns"],
    core_matrix[, "name"]
  )
  xml_ <- paste0(
    xml_,
    paste0(properties, collapse = ""),
    "</cp:coreProperties>"
  )
  props_dir = file.path(package_dir, "docProps")
  dir.create(props_dir, recursive = TRUE, showWarnings = FALSE)
  filename <- file.path(props_dir, "core.xml")
  writeLines(enc2utf8(xml_), filename, useBytes = TRUE)
  invisible()
}

drop_templatenode_from_app <- function(package_dir) {
  file <- file.path(package_dir, "docProps", "app.xml")
  if (!file.exists(file)) {
    return()
  }
  doc <- read_xml(file)
  node_template <- xml_child(doc, "d1:Template")
  if (!inherits(node_template, "xml_missing")) {
    xml_remove(node_template)
    write_xml(doc, file)
  }
}


# app.xml properties ----

read_app_properties <- function(package_dir) {
  file <- file.path(package_dir, "docProps", "app.xml")
  if (!file.exists(file)) {
    return(NULL)
  }
  doc <- read_xml(file)
  ns <- xml_ns(doc)

  hyperlink_base_node <- xml_find_first(doc, "//d1:HyperlinkBase", ns = ns)
  hyperlink_base <- if (inherits(hyperlink_base_node, "xml_missing")) {
    NA_character_
  } else {
    xml_text(hyperlink_base_node)
  }

  company_node <- xml_find_first(doc, "//d1:Company", ns = ns)
  company <- if (inherits(company_node, "xml_missing")) {
    NA_character_
  } else {
    xml_text(company_node)
  }

  z <- list(
    data = data.frame(
      name = c("HyperlinkBase", "Company"),
      value = c(hyperlink_base, company),
      stringsAsFactors = FALSE
    ),
    file = file
  )
  class(z) <- "app_properties"
  z
}


#' @export
`[<-.app_properties` <- function(x, i, j, value) {
  if (!i %in% x$data$name) {
    new_row <- data.frame(name = i, value = value, stringsAsFactors = FALSE)
    x$data <- rbind(x$data, new_row)
  } else {
    x$data[x$data$name == i, j] <- value
  }
  x
}


#' @export
`[.app_properties` <- function(x, i, j) {
  x$data[x$data$name == i, j]
}

#' @importFrom xml2 xml_set_text
write_app_properties <- function(app_props, package_dir) {
  if (is.null(app_props)) {
    return(invisible())
  }

  file <- file.path(package_dir, "docProps", "app.xml")
  if (!file.exists(file)) {
    return(invisible())
  }

  doc <- read_xml(file)
  ns <- xml_ns(doc)

  for (i in seq_len(nrow(app_props$data))) {
    prop_name <- app_props$data$name[i]
    prop_value <- app_props$data$value[i]

    if (is.na(prop_value) || prop_value == "") {
      next
    }

    xpath <- paste0("//d1:", prop_name)
    node <- xml_find_first(doc, xpath, ns = ns)
    if (inherits(node, "xml_missing")) {
      new_node <- read_xml(sprintf(
        "<%s>%s</%s>",
        prop_name,
        htmlEscapeCopy(prop_value),
        prop_name
      ))
      xml_add_child(doc, new_node)
    } else {
      xml_set_text(node, prop_value)
    }
  }

  write_xml(doc, file)
  invisible()
}

x <- read_pptx()

layout_summary(x) # available layouts

x <- add_slide(x, layout = "Two Content")

x <- layout_default(x, "Title Slide") # set default layout for `add_slide()`
x <- add_slide(x) # uses default layout

# use `...` to fill placeholders when adding slide
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)

is_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}
is_scalar_datetime <- function(x) {
  inherits(x, "POSIXt") && length(x) == 1 && !is.na(x)
}


#' @export
#' @title Number of blocks inside an rdocx object
#' @description Return the number of blocks inside an rdocx object.
#' This number also include the default section definition of a
#' Word document - default Word section is an uninvisible element.
#' @param x an rdocx object
#' @examples
#' # how many elements are there in an new document produced
#' # with the default template.
#' length(read_docx())
#' @family functions for Word document informations
length.rdocx <- function(x) {
  xml_length(xml_child(x$doc_obj$get(), "w:body"))
}


#' @export
#' @title Read document properties
#' @description Read Word or PowerPoint document properties
#' and get results in a data.frame.
#' @param x an `rdocx` or `rpptx` object
#' @examples
#' x <- read_docx()
#' doc_properties(x)
#' @return a data.frame
#' @family functions for Word document informations
#' @family functions for reading presentation information
doc_properties <- function(x) {
  if (inherits(x, "rdocx")) {
    cp <- x$doc_properties
  } else if (inherits(x, "rpptx") || inherits(x, "rxlsx")) {
    cp <- x$core_properties
  } else {
    stop("x should be a rpptx or a rdocx or a rxlsx object.")
  }

  properties_custom <- x$doc_properties_custom

  out_custom <- data.frame(
    tag = properties_custom[, "name"],
    value = properties_custom[, "value"],
    stringsAsFactors = FALSE
  )
  out <- data.frame(
    tag = cp[, "name"],
    value = cp[, "value"],
    stringsAsFactors = FALSE
  )
  out <- rbind(out, out_custom)

  # Add app properties for rdocx
  if (inherits(x, "rdocx") && !is.null(x$app_properties)) {
    app_props <- x$app_properties$data
    app_props <- app_props[!is.na(app_props$value), , drop = FALSE]
    if (nrow(app_props) > 0) {
      out_app <- data.frame(
        tag = app_props$name,
        value = app_props$value,
        stringsAsFactors = FALSE
      )
      out <- rbind(out, out_app)
    }
  }

  row.names(out) <- NULL
  out
}

#' @export
#' @title Set document properties
#' @description set Word or PowerPoint document properties. These are not visible
#' in the document but are available as metadata of the document.
#'
#' Any character property can be added as a document property.
#' It provides an easy way to insert arbitrary fields. Given the challenges
#' that can be encountered with find-and-replace in word with officer, the
#' use of document fields and quick text fields provides a much more robust
#' approach to automatic document generation from R.
#' @note
#' The "last modified" and "last modified by" fields will be automatically be updated
#' when the file is written.
#' @param x an rdocx or rpptx object
#' @param title,subject,creator,description text fields
#' @param created a date object
#' @param hyperlink_base a string specifying the base URL for relative hyperlinks
#' in the document (only for rdocx).
#' @param ... named arguments (names are field names), each element is a single
#' character value specifying value associated with the corresponding field name.
#' These pairs of *key-value* are added as custom properties. If a value is
#' `NULL` or `NA`, the corresponding field is set to '' in the document properties.
#' @param values a named list (names are field names), each element is a single
#' character value specifying value associated with the corresponding field name.
#' If `values` is provided, argument `...` will be ignored.
#' @examples
#' x <- read_docx()
#' x <- set_doc_properties(x, title = "title",
#'   subject = "document subject", creator = "Me me me",
#'   description = "this document is empty",
#'   created = Sys.time(),
#'   yoyo = "yok yok",
#'   glop = "pas glop")
#' x <- doc_properties(x)
#' @family functions for Word document informations
set_doc_properties <- function(
  x,
  title = NULL,
  subject = NULL,
  creator = NULL,
  description = NULL,
  created = NULL,
  hyperlink_base = NULL,
  ...,
  values = NULL
) {
  if (inherits(x, "rdocx")) {
    cp <- x$doc_properties
  } else if (inherits(x, "rpptx")) {
    cp <- x$core_properties
  } else {
    stop("x should be a rpptx or rdocx object.")
  }

  if (!is.null(title)) {
    if (!is_string(title)) {
      cli::cli_warn(
        c(
          "!" = "The value for property 'title' is not a string.",
          "i" = "It will be set to '' in the document properties"
        )
      )
      title <- ""
    }
    cp["title", "value"] <- title
  }

  if (!is.null(subject)) {
    if (!is_string(subject)) {
      cli::cli_warn(
        c(
          "!" = "The value for property 'subject' is not a string.",
          "i" = "It will be set to '' in the document properties"
        )
      )
      subject <- ""
    }
    cp["subject", "value"] <- subject
  }

  if (!is.null(creator)) {
    if (!is_string(creator)) {
      cli::cli_warn(
        c(
          "!" = "The value for property 'creator' is not a string.",
          "i" = "It will be set to '' in the document properties"
        )
      )
      creator <- ""
    }
    cp["creator", "value"] <- creator
  }

  if (!is.null(description)) {
    if (!is_string(description)) {
      cli::cli_warn(
        c(
          "!" = "The value for property 'description' is not a string.",
          "i" = "It will be set to '' in the document properties"
        )
      )
      description <- ""
    }
    cp["description", "value"] <- description
  }

  if (!is.null(created)) {
    if (!is_scalar_datetime(created)) {
      cli::cli_warn(
        c(
          "!" = "The value for property 'created' is not a date-time.",
          "i" = "It will not be set in the document properties"
        )
      )
    } else {
      cp["created", "value"] <- format(created, "%Y-%m-%dT%H:%M:%SZ")
    }
  }

  if (!is.null(hyperlink_base)) {
    if (!inherits(x, "rdocx")) {
      cli::cli_warn(
        c(
          "!" = "{.arg hyperlink_base} is only supported for Word documents.",
          "i" = "It will be ignored."
        )
      )
    } else if (!is_string(hyperlink_base)) {
      cli::cli_warn(
        c(
          "!" = "The value for property {.val hyperlink_base} is not a string.",
          "i" = "It will not be set in the document properties."
        )
      )
    } else {
      if (is.null(x$app_properties)) {
        cli::cli_warn(
          c(
            "!" = "No app.xml file found in the document.",
            "i" = "{.arg hyperlink_base} will not be set."
          )
        )
      } else {
        x$app_properties["HyperlinkBase", "value"] <- hyperlink_base
      }
    }
  }

  if (is.null(values)) {
    values <- list(...)
  }

  if (length(values) > 0) {
    x$content_type$add_override(
      setNames(
        "application/vnd.openxmlformats-officedocument.custom-properties+xml",
        "/docProps/custom.xml"
      )
    )
    x$rel$add(
      id = paste0("rId", x$rel$get_next_id()),
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/custom-properties",
      target = "docProps/custom.xml"
    )

    custom_props <- x$doc_properties_custom
    for (i in seq_along(values)) {
      .value. <- ""
      if (
        !is.null(values[[i]]) &&
          length(values[[i]]) == 1 &&
          !is.na(values[[i]])
      ) {
        .value. <- format(values[[i]])
        .value. <- enc2utf8(.value.)
      } else if (
        !is.null(values[[i]]) &&
          length(values[[i]]) == 1 &&
          is.na(values[[i]])
      ) {
        .value. <- ""
      } else if (is.null(values[[i]])) {
        .value. <- ""
      } else if (length(values[[i]]) != 1) {
        .value. <- ""
        cli::cli_warn(
          c(
            "!" = "The value for property '{names(values)[i]}' is not a single character value.",
            "i" = "It will be set to '' in the document properties"
          )
        )
      }
      custom_props[names(values)[i], "value"] <- .value.
    }
    x$doc_properties_custom <- custom_props
  }

  if (inherits(x, "rdocx")) {
    x$doc_properties <- cp
  } else {
    x$core_properties <- cp
  }

  x
}


#' @export
#' @title List Word bookmarks
#' @description List bookmarks id that can be found in a
#' 'Word' document.
#' @param x an `rdocx` object
#' @examples
#' library(officer)
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
#' doc_1 <- body_bookmark(doc_1, "text_to_replace_1")
#' doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
#' doc_1 <- body_bookmark(doc_1, "text_to_replace_2")
#'
#' docx_bookmarks(doc_1)
#'
#' docx_bookmarks(read_docx())
#' @family functions for Word document informations
docx_bookmarks <- function(x) {
  stopifnot(inherits(x, "rdocx"))

  doc_ <- xml_find_all(x$doc_obj$get(), "//w:bookmarkStart[@w:name]")
  setdiff(xml_attr(doc_, "name"), "_GoBack")
}

#' @export
#' @title 'Word' page layout
#' @description Get page width, page height and margins (in inches). The return values
#' are those corresponding to the section where the cursor is.
#' @param x an `rdocx` object
#' @examples
#' docx_dim(read_docx())
#' @family functions for Word document informations
docx_dim <- function(x) {
  cursor <- as.character(x$officer_cursor)
  if (is.na(cursor)) {
    next_section <- xml_find_first(
      x$doc_obj$get(),
      "/w:document/w:body/w:sectPr"
    )
  } else {
    xpath_ <- paste0(
      file.path(cursor, "following-sibling::w:sectPr"),
      "|",
      file.path(cursor, "following-sibling::w:p/w:pPr/w:sectPr"),
      "|",
      "//w:sectPr"
    )
    next_section <- xml_find_first(x$doc_obj$get(), xpath_)
  }

  sd <- section_dimensions(next_section)
  sd$page <- sd$page / (20 * 72)
  sd$margins <- sd$margins / (20 * 72)
  sd
}

# set and remove the default layout
x <- read_pptx()
layout_default(x) # no defaults
x <- layout_default(x, "Title and Content") # set default
layout_default(x)
x <- add_slide(x) # new slide with default layout
x <- layout_default(x, NULL) # remove default
layout_default(x) # no defaults

# use when repeatedly adding slides with same layout
x <- read_pptx()
x <- layout_default(x, "Title and Content")
x <- add_slide(x, title = "1. Slide", body = "Some content")
x <- add_slide(x, title = "2. Slide", body = "Some more content")
x <- add_slide(x, title = "3. Slide", body = "Even more content")

library(officer)

# create a template ----
doc <- read_docx()
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "Hello text to replace")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "Hello text to replace")
doc <- body_add_par(doc, "blah blah blah")
template_file <- print(
  x = doc,
  target = tempfile(fileext = ".docx")
)

# replace all pars containing "to replace" ----
doc <- read_docx(path = template_file)
while (cursor_reach_test(doc, "to replace")) {
  doc <- cursor_reach(doc, "to replace")

  doc <- body_add_fpar(
    x = doc,
    pos = "on",
    value = fpar(
      "Here is a link: ",
      hyperlink_ftext(
        text = "yopyop",
        href = "https://cran.r-project.org/"
      )
    )
  )
}

doc <- cursor_end(doc)
doc <- body_add_par(doc, "Yap yap yap yap...")

result_file <- print(
  x = doc,
  target = tempfile(fileext = ".docx")
)

# cursor_bookmark ----

doc <- read_docx()
doc <- body_add_par(doc, "centered text", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
doc <- body_add_par(doc, "A title", style = "heading 1")
doc <- body_add_par(doc, "Hello world!", style = "Normal")
doc <- cursor_bookmark(doc, "text_to_replace")
doc <- body_add_table(doc, value = iris, style = "table_template")

print(doc, target = tempfile(fileext = ".docx"))

library(officer)

# use key-value format to fill phs
x <- read_pptx()
x <- add_slide(x, "Two Content")
x <- phs_with(
  x,
  `Title 1` = "A title", # ph label
  dt = Sys.Date(), # ph type
  `body[2]` = "Body 2", # ph type + type index
  left = "Left side", # ph keyword
  `6` = "Footer" # ph index
)

# reuse ph content via the .dots arg
x <- read_pptx()
my_ph_list <- list(`6` = "Footer", dt = Sys.Date())
x <- add_slide(x, "Two Content")
x <- phs_with(
  x,
  `Title 1` = "Title A",
  `body[2]` = "Body A",
  .dots = my_ph_list
)
x <- add_slide(x, "Two Content")
x <- phs_with(
  x,
  `Title 1` = "Title B",
  `body[2]` = "Body B",
  .dots = my_ph_list
)

# use the .slide_idx arg to select which slide(s) to process
x <- read_pptx()
x <- add_slide(x, "Two Content")
x <- add_slide(x, "Two Content")
x <- phs_with(x, `6` = "Footer", dt = Sys.Date(), .slide_idx = 1:2)

# run to open temp pptx file locally
# \dontrun{
# print(x, preview = TRUE)
# }

library(officer)

if (require("ggplot2")) {
  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))

  doc <- read_docx()
  doc <- body_add_par(doc, "insert_plot_here")
  doc <- body_bookmark(doc, "plot")
  doc <- body_replace_gg_at_bkm(doc, bookmark = "plot", value = gg_plot)
  print(doc, target = tempfile(fileext = ".docx"))
}
doc <- read_docx()
doc <- body_add_par(doc, "insert_plot_here")
doc <- body_bookmark(doc, "plot")
if (capabilities(what = "png")) {
  doc <- body_replace_plot_at_bkm(
    doc,
    bookmark = "plot",
    value = plot_instr(
      code = {
        barplot(1:5, col = 2:6)
      }
    )
  )
}
print(doc, target = tempfile(fileext = ".docx"))

# presentation ------------------------------------------------------------

presentation <- R6Class(
  "presentation",
  inherit = openxml_document,

  public = list(
    initialize = function(package_dir) {
      super$initialize(character(0))
      presentation_filename <- file.path(package_dir, "ppt", "presentation.xml")
      self$feed(presentation_filename)

      slide_df <- private$get_slide_df()
      private$slide_id <- slide_df$id
      private$slide_rid <- slide_df$rid
    },

    add_slide = function(target) {
      private$rels_doc$add(
        id = paste0("rId", private$rels_doc$get_next_id()),
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide",
        target = target
      )
      rels <- private$rels_doc$get_data()
      rid <- rels[rels$target %in% target, "id"]

      ids <- private$slide_id
      if (length(ids) < 1) {
        new_id <- 256
      } else {
        new_id <- max(ids) + 1
      }

      private$slide_id <- c(private$slide_id, new_id)
      private$slide_rid <- c(private$slide_rid, rid)

      private$update_xml()

      self
    },
    slide_data = function() {
      rel_df <- self$rel_df()
      rel_df <- select(
        .data = rel_df,
        all_of(setNames(c("id", "target"), c("slide_rid", "target")))
      )

      ref <- data.frame(
        slide_id = private$slide_id,
        slide_rid = private$slide_rid,
        stringsAsFactors = FALSE
      )

      left_join(ref, rel_df, by = "slide_rid")
    },

    move_slide = function(from, to) {
      slide_list <- self$slide_data()

      ord <- seq_len(nrow(slide_list))
      if (from < to) {
        to <- to + 1
      }
      ord[ord >= to] <- ord[ord >= to] + 1L
      ord[from] <- to
      slide_list <- slide_list[order(ord), , drop = FALSE]
      private$slide_id <- slide_list$slide_id
      private$slide_rid <- slide_list$slide_rid

      private$update_xml()

      self
    },

    remove_slide = function(target) {
      reldf <- self$rel_df()
      id <- which(basename(reldf$target) %in% basename(target))
      rid <- reldf$id[id]
      private$rels_doc$remove(target = target)

      dropid <- which(private$slide_rid %in% rid)

      private$slide_id <- private$slide_id[-dropid]
      private$slide_rid <- private$slide_rid[-dropid]

      private$update_xml()

      self
    }
  ),
  private = list(
    slide_id = NULL,
    slide_rid = NULL,

    get_slide_df = function() {
      nodes <- xml_find_all(private$doc, "//p:sldIdLst/p:sldId")
      id <- as.integer(xml_attr(nodes, "id", ns = xml_ns(private$doc)))
      rid <- xml_attr(nodes, "r:id", ns = xml_ns(private$doc))
      data.frame(id = id, rid = rid, stringsAsFactors = FALSE)
    },
    update_xml = function() {
      xml_list <- xml_find_first(private$doc, "//p:sldIdLst")
      xml_elt <- paste(
        sprintf(
          "<p:sldId id=\"%.0f\" r:id=\"%s\"/>",
          private$slide_id,
          private$slide_rid
        ),
        collapse = ""
      )

      xml_elt <- paste0(
        "<p:sldIdLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
        xml_elt,
        "</p:sldIdLst>"
      )
      xml_elt <- as_xml_document(xml_elt)

      if (!inherits(xml_list, "xml_missing")) {
        xml_replace(xml_list, xml_elt)
      } else {
        ## needs to be after all MasterIdLst elements. placing it before sldSz seems to be the safest option.
        xml_add_sibling(
          xml_find_first(private$doc, "//p:sldSz"),
          xml_elt,
          .where = "before"
        )
      }

      self
    }
  )
)


# slide master ------------------------------------------------------------
slide_master <- R6Class(
  "slide_master",
  inherit = openxml_document,
  public = list(
    name = function() {
      theme_ <- private$theme_file()
      root <- gsub(paste0(self$dir_name(), "$"), "", dirname(private$filename))
      xml_attr(read_xml(file.path(root, theme_)), "name")
    },

    colors = function() {
      theme_ <- private$theme_file()
      root <- gsub(paste0(self$dir_name(), "$"), "", dirname(private$filename))

      doc <- read_xml(file.path(root, theme_))
      read_theme_colors(doc, self$name())
    },

    xfrm = function() {
      nodeset <- xml_find_all(
        self$get(),
        as_xpath_content_sel("p:cSld/p:spTree/")
      )
      read_xfrm(nodeset, self$file_name(), self$name())
    }
  ),
  private = list(
    theme_file = function() {
      data <- self$rel_df()
      theme_file <- data$target[basename(data$type) == "theme"]
      file.path("ppt/theme", basename(theme_file))
    }
  )
)

# slide_layout ------------------------------------------------------------
slide_layout <- R6Class(
  "slide_layout",
  inherit = openxml_document,
  public = list(
    get_metadata = function() {
      rels <- self$rel_df()
      rels <- rels[basename(rels$type) == "slideMaster", ]
      data.frame(
        stringsAsFactors = FALSE,
        name = self$name(),
        filename = self$file_name(),
        master_file = rels$target
      )
    },
    xfrm = function() {
      rels <- self$rel_df()
      rels <- rels[basename(rels$type) == "slideMaster", ]

      nodeset <- xml_find_all(
        self$get(),
        as_xpath_content_sel("p:cSld/p:spTree/")
      )
      data <- read_xfrm(nodeset, self$file_name(), self$name())
      if (nrow(data)) {
        data$master_file <- basename(rels$target)
      } else {
        data$master_file <- character(0)
      }
      data
    },
    write_template = function(new_file) {
      path <- system.file(package = "officer", "template/slide.xml")

      rel_filename <- file.path(
        dirname(new_file),
        "_rels",
        paste0(basename(new_file), ".rels")
      )

      newrel <- relationship$new()$add(
        id = "rId1",
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slideLayout",
        target = file.path("../slideLayouts", basename(self$file_name()))
      )
      newrel$write(path = rel_filename)
      file.copy(path, to = new_file, copy.mode = FALSE)
      self
    },

    name = function() {
      csld <- xml_find_first(self$get(), "//p:cSld")
      xml_attr(csld, "name")
    }
  )
)

# slide ------------------------------------------------------------
slide <- R6Class(
  "slide",
  inherit = openxml_document,
  public = list(
    feed = function(file) {
      super$feed(file)
      slide_info <- private$rels_doc$get_data()
      slide_info <- slide_info[basename(slide_info$type) == "slideLayout", ]
      private$layout_file <- basename(slide_info$target)
      self
    },

    set_layout_xfrm = function(xfrm_ref) {
      private$element_data <- xfrm_ref[xfrm_ref$file == private$layout_file, ]
      self
    },
    reference_img = function(src, dir_name) {
      src <- unique(src)
      private$rels_doc$add_img(src, root_target = "../media")
      dir.create(dir_name, recursive = TRUE, showWarnings = FALSE)
      file.copy(from = src, to = file.path(dir_name, basename(src)))
      self
    },

    reference_slide = function(slide_file) {
      rel_dat <- private$rels_doc$get_data()

      if (!slide_file %in% rel_dat$target) {
        next_id <- private$rels_doc$get_next_id()
        private$rels_doc$add(
          id = sprintf("rId%.0f", next_id),
          type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide",
          target = slide_file
        )
      }
      self
    },

    reference_hyperlink = function(href) {
      rel_dat <- private$rels_doc$get_data()

      if (!href %in% rel_dat$target) {
        next_id <- private$rels_doc$get_next_id()
        private$rels_doc$add(
          id = sprintf("rId%.0f", next_id),
          type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink",
          target = href,
          target_mode = "External"
        )
      }
      self
    },

    get_xfrm = function(type = NULL, index = 1) {
      out <- private$element_data
      if (!is.null(type)) {
        if (type %in% out$type) {
          type_matches <- which(out$type == type)
          if (index <= length(type_matches)) {
            id <- type_matches[index]
          } else {
            stop(
              type,
              " can only have ",
              length(type_matches),
              " element(s) but index is set to ",
              index
            )
          }
        } else {
          stop("type ", type, " is not available in the slide layout")
        }
        out <- out[id, ]
      }
      out
    },

    layout_name = function() {
      private$layout_file
    },

    get_metadata = function() {
      rels <- self$rel_df()
      rels <- rels[basename(rels$type) == "slideLayout", ]
      data.frame(
        stringsAsFactors = FALSE,
        name = self$name(),
        filename = self$file_name(),
        layout_file = rels$target
      )
    }
  ),
  private = list(
    layout_file = NULL,
    element_data = NULL
  )
)


# notesSlide ------------------------------------------------------------
notesSlide <- R6Class(
  "notesSlide",
  inherit = openxml_document,
  public = list(
    feed = function(file) {
      super$feed(file)
      self
    },

    fortify_id = function() {
      cnvpr <- xml_find_all(private$doc, "//p:cNvPr")
      for (i in seq_along(cnvpr)) {
        xml_attr(cnvpr[[i]], "id") <- i
      }
      self
    },

    get_metadata = function() {
      rels <- self$rel_df()
    }
  )
)

library(officer)

# ph_location_type demo ----

loc_title <- ph_location_type(type = "title")
loc_footer <- ph_location_type(type = "ftr")
loc_dt <- ph_location_type(type = "dt")
loc_slidenum <- ph_location_type(type = "sldNum")
loc_body <- ph_location_type(type = "body")


doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(x = doc, "Un titre", location = loc_title)
doc <- ph_with(x = doc, "pied de page", location = loc_footer)
doc <- ph_with(x = doc, format(Sys.Date()), location = loc_dt)
doc <- ph_with(x = doc, "slide 1", location = loc_slidenum)
doc <- ph_with(x = doc, letters[1:10], location = loc_body)

loc_subtitle <- ph_location_type(type = "subTitle")
loc_ctrtitle <- ph_location_type(type = "ctrTitle")
doc <- add_slide(doc, layout = "Title Slide")
doc <- ph_with(x = doc, "Un sous titre", location = loc_subtitle)
doc <- ph_with(x = doc, "Un titre", location = loc_ctrtitle)

fileout <- tempfile(fileext = ".pptx")
print(doc, target = fileout)

library(officer)

txt_lorem <- rep(
  "Purus lectus eros metus turpis mattis platea praesent sed. ",
  50
)
txt_lorem <- paste0(txt_lorem, collapse = "")

header_first <- block_list(fpar(ftext("text for first page header")))
header_even <- block_list(fpar(ftext("text for even page header")))
header_default <- block_list(fpar(ftext("text for default page header")))
footer_first <- block_list(fpar(ftext("text for first page footer")))
footer_even <- block_list(fpar(ftext("text for even page footer")))
footer_default <- block_list(fpar(ftext("text for default page footer")))

ps <- prop_section(
  header_default = header_default,
  footer_default = footer_default,
  header_first = header_first,
  footer_first = footer_first,
  header_even = header_even,
  footer_even = footer_even
)

x <- read_docx()

x <- docx_set_settings(
  x = x,
  zoom = 2,
  list_separator = ",",
  even_and_odd_headers = TRUE
)

for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}
x <- body_set_default_section(
  x,
  value = ps
)
print(x, target = tempfile(fileext = ".docx"))

library(officer)

# example file from the package
file_input <- system.file(
  package = "officer",
  "doc_examples/example.docx"
)

# create a new rdocx document
x <- read_docx()

# import content from file_input
x <- body_import_docx(
  x = x,
  src = file_input,
  # style mapping for paragraphs and tables
  par_style_mapping = list(
    "Normal" = c("List Paragraph")
  ),
  tbl_style_mapping = list(
    "Normal Table" = "Light Shading"
  )
)

# Create temporary file
tf <- tempfile(fileext = ".docx")
# write to file
print(x, target = tf)

library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Comparison")
plot_layout_properties(doc, "Comparison")

doc <- ph_with(doc, "The Title", location = ph_location_id(id = 2)) # title
doc <- ph_with(doc, "Left Header", location = ph_location_id(id = 3)) # left header
doc <- ph_with(doc, "Left Content", location = ph_location_id(id = 4)) # left content
doc <- ph_with(doc, "The Footer", location = ph_location_id(id = 8)) # footer

file <- tempfile(fileext = ".pptx")
print(doc, file)

## file.show(file) # may not work on your system

#' @export
#' @rdname cursor
#' @title Set cursor in a 'Word' document
#' @description A set of functions is available to manipulate
#' the position of a virtual cursor. This cursor will be used when
#' inserting, deleting or updating elements in the document.
#' @section cursor_begin:
#' Set the cursor at the beginning of the document, on the first element
#' of the document (usually a paragraph or a table).
#' @param x a docx device
#' @example inst/examples/example_cursor.R
cursor_begin <- function(x) {
  if (length(x$officer_cursor$nodes_names) > 0L) {
    x$officer_cursor$which <- 1L
  } else {
    x$officer_cursor$which <- 0L
  }
  x
}

#' @rdname cursor
#' @param id bookmark id
#' @section cursor_bookmark:
#' Set the cursor at a bookmark that has previously been set.
#' @export
cursor_bookmark <- function(x, id) {
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", id)
  bm_start <- xml_find_first(x$doc_obj$get(), xpath_)

  if (inherits(bm_start, "xml_missing")) {
    stop("cannot find bookmark ", shQuote(id), call. = FALSE)
  }

  bm_id <- xml_attr(bm_start, "id")

  nodes_with_text <- xml_find_all(
    x$doc_obj$get(),
    "/w:document/w:body/*|/w:ftr/*|/w:hdr/*"
  )
  test_start <- sapply(nodes_with_text, function(node) {
    expr <- sprintf("/descendant::w:bookmarkStart[@w:id='%s']", bm_id)
    match_node <- xml_child(node, expr)
    !inherits(match_node, "xml_missing")
  })
  if (!any(test_start)) {
    stop(
      "bookmark ",
      shQuote(id),
      " has not been found in the document",
      call. = FALSE
    )
  }

  test_end <- sapply(nodes_with_text, function(node) {
    expr <- sprintf("/descendant::w:bookmarkEnd[@w:id='%s']", bm_id)
    match_node <- xml_child(node, expr)
    !inherits(match_node, "xml_missing")
  })

  on_same_par <- test_start == test_end
  if (!all(on_same_par)) {
    stop(
      "bookmark ",
      shQuote(id),
      " does not end in the same paragraph (or is on the whole paragraph)",
      call. = FALSE
    )
  }

  x$officer_cursor$which <- which(test_start)[1]

  x
}

#' @export
#' @rdname cursor
#' @section cursor_end:
#' Set the cursor at the end of the document, on the last element
#' of the document.
cursor_end <- function(x) {
  x$officer_cursor$which <- length(x$officer_cursor$nodes_names)
  x
}

#' @export
#' @rdname cursor
#' @param index element index in the document
#' @section cursor_reach_index:
#' Set the cursor at a specific index position in the document.
cursor_reach_index <- function(x, index) {
  l_ <- length(x$officer_cursor$nodes_names)
  if (l_ < 1) {
    stop("document contains no element", call. = FALSE)
  }
  if (!between(index, 1, l_)) {
    stop("invalid index ", index, " (", l_, " element(s))", call. = FALSE)
  }
  x$officer_cursor$which <- index
  x
}

#' @export
#' @rdname cursor
#' @param keyword keyword to look for as a regular expression
#' @param fixed logical. If TRUE, pattern is a string to be matched as is.
#' @section cursor_reach:
#' Set the cursor on the first element of the document
#' that contains text specified in argument `keyword`.
#' The argument `keyword` is a regexpr pattern.
cursor_reach <- function(x, keyword, fixed = FALSE) {
  nodes_with_text <- xml_find_all(
    x$doc_obj$get(),
    "/w:document/w:body/*|/w:ftr/*|/w:hdr/*"
  )

  if (length(nodes_with_text) < 1) {
    stop("no text found in the document", call. = FALSE)
  }

  text_ <- xml_text(nodes_with_text)
  test_ <- grepl(pattern = keyword, x = text_, fixed = fixed)
  if (!any(test_)) {
    stop(keyword, " has not been found in the document", call. = FALSE)
  }
  x$officer_cursor$which <- which(test_)[1]
  x
}

#' @export
#' @rdname cursor
#' @param keyword keyword to look for as a regular expression
#' @section cursor_reach_test:
#' Test if an expression has a match in the document
#' that contains text specified in argument `keyword`.
#' The argument `keyword` is a regexpr pattern.
cursor_reach_test <- function(x, keyword) {
  nodes_with_text <- xml_find_all(
    x$doc_obj$get(),
    "/w:document/w:body/*|/w:ftr/*|/w:hdr/*"
  )

  if (length(nodes_with_text) < 1) {
    stop("no text found in the document", call. = FALSE)
  }

  text_ <- xml_text(nodes_with_text)
  test_ <- grepl(pattern = keyword, x = text_)
  any(test_)
}

#' @export
#' @rdname cursor
#' @section cursor_forward:
#' Move the cursor forward, it increments the cursor in the document.
cursor_forward <- function(x) {
  x$officer_cursor$which <- min(c(
    length(x$officer_cursor$nodes_names),
    x$officer_cursor$which + 1L
  ))
  x
}

#' @export
#' @rdname cursor
#' @section cursor_backward:
#' Move the cursor backward, it decrements the cursor in the document.
cursor_backward <- function(x) {
  x$officer_cursor$which <- max(c(0L, x$officer_cursor$which - 1L))
  x
}

# officer docx cursor ----
officer_cursor <- function(node) {
  nodes <- xml_find_all(node, "/w:document/w:body/*")
  nodes_names <- xml_name(nodes)
  nodes_names <- nodes_names[!nodes_names %in% "sectPr"]
  x <- list(
    nodes_names = nodes_names,
    which = length(nodes_names)
  )
  class(x) <- "officer_cursor"
  x
}


#' @export
as.character.officer_cursor <- function(x, ...) {
  if (length(x$nodes_names) < 1) {
    return(NA_character_)
  }
  paste0("/w:document/w:body/*[", x$which, "]")
}

## xml on cursor ----
ooxml_on_cursor <- function(x, node) {
  if (length(x$nodes_names) < 1) {
    return(NULL)
  }
  node <- xml_find_first(node, as.character(x))
  if (inherits(node, "xml_missing")) {
    stop("cursor does not correspond to any node", call. = FALSE)
  }
  node
}


## cursor feed ----
cursor_append <- function(x, what) {
  x$nodes_names <- c(x$nodes_names, what)
  x$which <- x$which + 1L
  x
}
cursor_add_after <- function(x, what) {
  seq_left <- seq_along(x$which)
  set_left <- x$nodes_names[seq_left]
  set_right <- x$nodes_names[-seq_left]
  x$nodes_names <- c(set_left, what, set_right)
  x$which <- x$which + 1L
  x
}

cursor_add_before <- function(x, what) {
  seq_left <- seq_along(x$which - 1)
  set_left <- x$nodes_names[seq_left]
  set_right <- x$nodes_names[-seq_left]
  x$nodes_names <- c(set_left, what, set_right)
  x
}
cursor_replace_nodename <- function(x, what) {
  x$nodes_names[x$which] <- what
  x
}
cursor_delete <- function(x, what) {
  stopifnot(x$nodes_names[x$which] == what)
  x$nodes_names <- x$nodes_names[-x$which]
  if (x$which > length(x$nodes_names)) {
    x$which <- x$which - 1L
  }
  x
}

test_that("default template", {
  x <- read_docx()
  expect_equal(length(x), 1)
  expect_true(file.exists(x$package_dir))
})

test_that("docx dim", {
  x <- read_docx()
  dims <- docx_dim(x)

  expect_equal(names(dims), c("page", "landscape", "margins"))
  expect_length(dims$page, 2)
  expect_length(dims$landscape, 1)
  expect_length(dims$margins, 6)

  expect_equal(
    dims$page,
    c(width = 8.263889, height = 11.694444),
    tolerance = .001
  )

  ps <- prop_section(
    page_size = page_size(
      width = 8.263889,
      height = 11.694444,
      orient = "landscape"
    ),
    page_margins = page_mar(top = 2),
    type = "oddPage"
  )
  x <- read_docx()
  x <- body_add_par(x = x, value = "paragraph 1", style = "Normal")
  x <- body_end_block_section(x, block_section(ps))
  x <- cursor_begin(x)
  dims <- docx_dim(x)
  expect_equivalent(
    object = dims,
    expected = list(
      page = c(width = 11.694444, height = 8.263889),
      landscape = TRUE,
      margins = c(
        top = 2,
        bottom = 1417 / 1440,
        left = 1417 / 1440,
        right = 1417 / 1440,
        header = 708 / 1440,
        footer = 708 / 1440
      )
    ),
    tolerance = .00001
  )
})

test_that("list bookmarks", {
  template_file <- system.file(package = "officer", "doc_examples/example.docx")
  x <- read_docx(path = template_file)
  bookmarks <- docx_bookmarks(x)

  expect_equal(bookmarks, c("bmk_1", "bmk_2"))
})

test_that("console printing", {
  x <- read_docx()
  x <- body_add_par(x, "Hello world", style = "Normal")
  expect_output(print(x), "rdocx document with")
})

test_that("check extention and print document", {
  x <- read_docx()
  outfile <- print(x, target = tempfile(fileext = ".docx"))
  expect_true(file.exists(outfile))

  expect_error(print(x, target = tempfile(fileext = ".docxxxx")))
})

test_that("style is read from document", {
  x <- read_docx()
  expect_silent({
    x <- body_add_par(x = x, value = "paragraph 1", style = "Normal")
  })

  expect_error({
    x <- body_add_par(x = x, value = "paragraph 1", style = "blahblah")
  })
})


test_that("id are sequentially defined", {
  doc <- read_docx()
  any_img <- FALSE
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  if (file.exists(img.file)) {
    doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)
    any_img <- TRUE
  }
  tmp_file <- print(doc, target = tempfile(fileext = ".docx"))
  skip_if_not(any_img)

  pack_dir <- tempfile(pattern = "dir")
  unpack_folder(file = tmp_file, folder = pack_dir)

  all_ids <- read_xml(x = file.path(pack_dir, "word/document.xml"))
  all_ids <- xml_find_all(all_ids, "//*[@id]")
  all_ids <- xml_attr(all_ids, "id")

  expect_equal(length(unique(all_ids)), length(all_ids))
  expect_true(all(grepl("[0-9]+", all_ids)))

  ids <- as.integer(all_ids)
  expect_true(all(diff(ids) == 1))
})


test_that("cursor behavior", {
  doc <- read_docx()
  doc <- body_add_par(doc, "paragraph 1", style = "Normal")
  doc <- body_add_par(doc, "paragraph 2", style = "Normal")
  doc <- body_add_par(doc, "paragraph 3", style = "Normal")
  doc <- body_bookmark(doc, "bkm1")
  doc <- body_add_par(doc, "paragraph 4", style = "Normal")
  doc <- body_add_par(doc, "paragraph 5", style = "Normal")
  doc <- body_add_par(doc, "paragraph 6", style = "Normal")
  doc <- body_add_par(doc, "paragraph 7", style = "Normal")
  doc <- cursor_begin(doc)
  init_file <- print(doc, target = tempfile(fileext = ".docx"))

  doc <- read_docx(path = init_file)
  doc <- cursor_begin(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 1")
  doc <- cursor_forward(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 2")
  doc <- cursor_end(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 7")
  doc <- cursor_backward(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 6")
  doc <- cursor_reach(doc, keyword = "paragraph 5")
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 5")
  doc <- cursor_bookmark(doc, "bkm1")
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 3")
})

test_that("cursor and position", {
  doc <- read_docx()
  doc <- body_add_par(doc, "paragraph 1", style = "Normal")
  doc <- body_add_par(doc, "paragraph 2", style = "Normal")
  doc <- cursor_begin(doc)
  doc <- body_add_par(doc, "new 1", style = "Normal", pos = "before")
  doc <- cursor_forward(doc)
  doc <- body_add_par(doc, "new 2", style = "Normal")

  ds_ <- docx_summary(doc)

  expect_equal(ds_$text, c("new 1", "paragraph 1", "new 2", "paragraph 2"))

  # cursor_reach_index
  doc <- cursor_reach_index(doc, 2)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 1")
  doc <- cursor_reach_index(doc, 4)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 2")
  expect_error(cursor_reach_index(doc, 0), "invalid index")
  expect_error(cursor_reach_index(doc, 5), "invalid index")

  doc <- body_remove(doc)
  doc <- body_remove(doc)
  ds_ <- docx_summary(doc)
  expect_equal(ds_$text, c("new 1", "paragraph 1"))
  doc <- read_docx()
  expect_warning(
    body_remove(doc),
    "There is nothing left to remove in the document"
  )
})

test_that("cursor and replacement", {
  doc <- read_docx()
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "Hello text to replace")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "Hello text to replace")
  doc <- body_add_par(doc, "blah blah blah")
  template_file <- print(
    x = doc,
    target = tempfile(fileext = ".docx")
  )

  doc <- read_docx(path = template_file)
  while (cursor_reach_test(doc, "to replace")) {
    doc <- cursor_reach(doc, "to replace")

    doc <- body_add_fpar(
      x = doc,
      pos = "on",
      value = fpar(
        "Here is a link: ",
        hyperlink_ftext(
          text = "yopyop",
          href = "https://cran.r-project.org/"
        )
      )
    )
  }

  doc <- cursor_end(doc)
  doc <- body_add_par(doc, "Yap yap yap yap...")
  expect_equal(
    xml_text(xml_find_all(docx_body_xml(doc), "//w:p")),
    c(
      "blah blah blah",
      "blah blah blah",
      "blah blah blah",
      "Here is a link: yopyop",
      "blah blah blah",
      "blah blah blah",
      "blah blah blah",
      "Here is a link: yopyop",
      "blah blah blah",
      "Yap yap yap yap..."
    )
  )
})

to_docx <- function(docx, value, title, subtitle) {
  # Add table
  pre_label <- seq_id <- "Table"
  docx <- body_add_table(docx, value = value, align_table = "center")

  # Add title above table
  run_num <- run_autonum(seq_id = seq_id, pre_label = pre_label)
  title <- block_caption(title, style = "Image Caption", autonum = run_num)
  docx <- body_add_caption(docx, title, pos = "before")
  # Add subtitle above table
  subtitle <- fpar(ftext(subtitle, prop = fp_text(color = "red")))
  docx <- body_add_fpar(docx, subtitle, style = "Normal", pos = "after")
  docx <- cursor_end(docx)

  invisible(docx)
}


test_that("cursor before table", {
  docx <- read_docx()
  tab <- head(mtcars[1:3])

  docx <- to_docx(docx, tab, "Title 1", "Subtitle 1")
  docx <- to_docx(docx, tab, "Title 2", "Subtitle 2")

  nodes_body <- xml_find_all(docx_body_xml(docx), "//w:body/*")

  expect_equal(
    xml_name(nodes_body),
    c("p", "p", "tbl", "p", "p", "tbl", "sectPr")
  )
})

test_that("autoHyphenation", {
  x <- read_docx()
  x$settings$auto_hyphenation <- TRUE
  new_file <- print(x, target = tempfile(fileext = ".docx"))
  new_dir <- unpack_folder(file = new_file, folder = tempfile(pattern = "dir"))
  xml_settings_file <- file.path(new_dir, "word/settings.xml")
  xml_settings <- read_xml(xml_settings_file)
  autoHyphenation_node <- xml_find_first(xml_settings, "//w:autoHyphenation")
  expect_false(inherits(autoHyphenation_node, "xml_missing"))
})

test_that("default template", {
  x <- read_docx()
  expect_equal(length(x), 1)
  expect_true(file.exists(x$package_dir))
})

test_that("docx dim", {
  x <- read_docx()
  dims <- docx_dim(x)

  expect_equal(names(dims), c("page", "landscape", "margins"))
  expect_length(dims$page, 2)
  expect_length(dims$landscape, 1)
  expect_length(dims$margins, 6)

  expect_equal(dims$page, c(width = 8.263889, height = 11.694444), tolerance = .001)

  ps <- prop_section(
    page_size = page_size(width = 8.263889, height = 11.694444, orient = "landscape"),
    page_margins = page_mar(top = 2),
    type = "oddPage"
  )
  x <- read_docx()
  x <- body_add_par(x = x, value = "paragraph 1", style = "Normal")
  x <- body_end_block_section(x, block_section(ps))
  x <- cursor_begin(x)
  dims <- docx_dim(x)
  expect_equivalent(
    object = dims,
    expected = list(
      page = c(width = 11.694444, height = 8.263889),
      landscape = TRUE,
      margins = c(top = 2, bottom = 1, left = 1,
                  right = 1, header = 0.5, footer = 0.5)),
    tolerance = .00001)
})

test_that("list bookmarks", {
  template_file <- system.file(package = "officer", "doc_examples/example.docx")
  x <- read_docx(path = template_file)
  bookmarks <- docx_bookmarks(x)

  expect_equal(bookmarks, c("bmk_1", "bmk_2"))
})

test_that("console printing", {
  x <- read_docx()
  x <- body_add_par(x, "Hello world", style = "Normal")
  expect_output(print(x), "docx document with")
})

test_that("check extention and print document", {
  x <- read_docx()
  outfile <- print(x, target = tempfile(fileext = ".docx"))
  expect_true(file.exists(outfile))

  expect_error(print(x, target = tempfile(fileext = ".docxxxx")))
})

test_that("style is read from document", {
  x <- read_docx()
  expect_silent({
    x <- body_add_par(x = x, value = "paragraph 1", style = "Normal")
  })

  expect_error({
    x <- body_add_par(x = x, value = "paragraph 1", style = "blahblah")
  })
})


test_that("id are sequentially defined", {
  doc <- read_docx()
  any_img <- FALSE
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  if (file.exists(img.file)) {
    doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)
    any_img <- TRUE
  }
  tmp_file <- print(doc, target = tempfile(fileext = ".docx"))
  skip_if_not(any_img)

  pack_dir <- tempfile(pattern = "dir")
  unpack_folder(file = tmp_file, folder = pack_dir)

  all_ids <- read_xml(x = file.path(pack_dir, "word/document.xml"))
  all_ids <- xml_find_all(all_ids, "//*[@id]")
  all_ids <- xml_attr(all_ids, "id")

  expect_equal(length(unique(all_ids)), length(all_ids))
  expect_true(all(grepl("[0-9]+", all_ids)))

  ids <- as.integer(all_ids)
  expect_true(all(diff(ids) == 1))
})


test_that("cursor behavior", {
  doc <- read_docx()
  doc <- body_add_par(doc, "paragraph 1", style = "Normal")
  doc <- body_add_par(doc, "paragraph 2", style = "Normal")
  doc <- body_add_par(doc, "paragraph 3", style = "Normal")
  doc <- body_bookmark(doc, "bkm1")
  doc <- body_add_par(doc, "paragraph 4", style = "Normal")
  doc <- body_add_par(doc, "paragraph 5", style = "Normal")
  doc <- body_add_par(doc, "paragraph 6", style = "Normal")
  doc <- body_add_par(doc, "paragraph 7", style = "Normal")
  doc <- cursor_begin(doc)
  init_file <- print(doc, target = tempfile(fileext = ".docx"))

  doc <- read_docx(path = init_file)
  doc <- cursor_begin(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 1")
  doc <- cursor_forward(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 2")
  doc <- cursor_end(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 7")
  doc <- cursor_backward(doc)
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 6")
  doc <- cursor_reach(doc, keyword = "paragraph 5")
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 5")
  doc <- cursor_bookmark(doc, "bkm1")
  current_node <- docx_current_block_xml(doc)
  expect_equal(xml_text(current_node), "paragraph 3")
})

test_that("cursor and position", {
  doc <- read_docx()
  doc <- body_add_par(doc, "paragraph 1", style = "Normal")
  doc <- body_add_par(doc, "paragraph 2", style = "Normal")
  doc <- cursor_begin(doc)
  doc <- body_add_par(doc, "new 1", style = "Normal", pos = "before")
  doc <- cursor_forward(doc)
  doc <- body_add_par(doc, "new 2", style = "Normal")

  ds_ <- docx_summary(doc)

  expect_equal(ds_$text, c("new 1", "paragraph 1", "new 2", "paragraph 2"))
  doc <- body_remove(doc)
  doc <- body_remove(doc)
  ds_ <- docx_summary(doc)
  expect_equal(ds_$text, c("new 1", "paragraph 1"))
  doc <- read_docx()
  expect_warning(body_remove(doc), "There is nothing left to remove in the document")
})

test_that("cursor and replacement", {
  doc <- read_docx()
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "Hello text to replace")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "blah blah blah")
  doc <- body_add_par(doc, "Hello text to replace")
  doc <- body_add_par(doc, "blah blah blah")
  template_file <- print(
    x = doc,
    target = tempfile(fileext = ".docx")
  )

  doc <- read_docx(path = template_file)
  while (cursor_reach_test(doc, "to replace")) {
    doc <- cursor_reach(doc, "to replace")

    doc <- body_add_fpar(
      x = doc,
      pos = "on",
      value = fpar(
        "Here is a link: ",
        hyperlink_ftext(
          text = "yopyop",
          href = "https://cran.r-project.org/"
        )
      )
    )
  }

  doc <- cursor_end(doc)
  doc <- body_add_par(doc, "Yap yap yap yap...")
  expect_equal(
    xml_text(xml_find_all(docx_body_xml(doc), "//w:p")),
    c(
      "blah blah blah", "blah blah blah", "blah blah blah", "Here is a link: yopyop",
      "blah blah blah", "blah blah blah", "blah blah blah", "Here is a link: yopyop",
      "blah blah blah", "Yap yap yap yap..."
    )
  )
})

to_docx <- function(docx, value, title, subtitle) {
  # Add table
  pre_label <- seq_id <- "Table"
  docx <- body_add_table(docx, value = value, align_table = "center")

  # Add title above table
  run_num <- run_autonum(seq_id = seq_id, pre_label = pre_label)
  title <- block_caption(title, style = "Image Caption", autonum = run_num)
  docx <- body_add_caption(docx, title, pos = "before")
  # Add subtitle above table
  subtitle <- fpar(ftext(subtitle, prop = fp_text(color = "red")))
  docx <- body_add_fpar(docx, subtitle, style = "Normal", pos = "after")
  docx <- cursor_end(docx)

  invisible(docx)
}


test_that("cursor before table", {
  docx <- read_docx()
  tab <- head(mtcars[1:3])

  docx <- to_docx(docx, tab, "Title 1", "Subtitle 1")
  docx <- to_docx(docx, tab, "Title 2", "Subtitle 2")

  nodes_body <- xml_find_all(docx_body_xml(docx), "//w:body/*")

  expect_equal(
    xml_name(nodes_body),
    c("p", "p", "tbl", "p", "p", "tbl", "sectPr")
  )
})

#' @importFrom xml2 xml_attr<- xml_name<- xml_text<- as_list as_xml_document read_xml
#' write_xml xml_add_child xml_add_parent xml_add_sibling xml_attr xml_attrs xml_child
#' xml_children xml_find_all xml_find_first xml_length xml_missing xml_name xml_ns
#' xml_path xml_remove xml_replace xml_set_attr xml_set_attrs xml_text
#' @importFrom stats setNames

read_xfrm <- function(nodeset, file, name) {
  if (length(nodeset) < 1) {
    return(data.frame(
      stringsAsFactors = FALSE,
      type = character(0),
      id = character(0),
      ph_label = character(0),
      ph = character(0),
      file = character(0),
      offx = integer(0),
      offy = integer(0),
      cx = integer(0),
      cy = integer(0),
      rotation = integer(0),
      name = character(0),
      fld_id = character(0),
      fld_type = character(0)
    ))
  }

  ph <- xml_child(nodeset, "p:nvSpPr/p:nvPr/p:ph")
  type <- xml_attr(ph, "type")
  type[is.na(type)] <- "body"
  id <- xml_attr(xml_child(nodeset, "/p:cNvPr"), "id")
  label <- xml_attr(xml_child(nodeset, "/p:cNvPr"), "name")

  off <- xml_child(nodeset, "p:spPr/a:xfrm/a:off")
  ext <- xml_child(nodeset, "p:spPr/a:xfrm/a:ext")
  rot <- xml_child(nodeset, "p:spPr/a:xfrm")

  fld_id <- xml_attr(xml_child(nodeset, "/p:txBody/a:p/a:fld"), "id")
  fld_type <- xml_attr(xml_child(nodeset, "/p:txBody/a:p/a:fld"), "type")

  data.frame(
    stringsAsFactors = FALSE,
    type = type,
    id = id,
    ph_label = label,
    ph = as.character(ph),
    file = basename(file),
    offx = as.integer(xml_attr(off, "x")),
    offy = as.integer(xml_attr(off, "y")),
    cx = as.integer(xml_attr(ext, "cx")),
    cy = as.integer(xml_attr(ext, "cy")),
    rotation = as.integer(xml_attr(rot, "rot")),
    fld_id,
    fld_type,
    name = name
  )
}


fortify_master_xfrm <- function(master_xfrm) {
  master_xfrm <- as.data.frame(master_xfrm)
  has_type <- grepl("type=", master_xfrm$ph)
  master_xfrm <- master_xfrm[has_type, ]
  if (nrow(master_xfrm) > 0) {
    # see #597
    list_xfrm <- split(master_xfrm, master_xfrm$file)
    list_xfrm <- lapply(list_xfrm, function(x) {
      x[!duplicated(x$type), , drop = FALSE]
    })
    master_xfrm <- do.call("rbind", list_xfrm)
  }

  tmp_names <- names(master_xfrm)

  old_ <- c("offx", "offy", "cx", "cy", "fld_id", "fld_type", "name")
  new_ <- c(
    "offx_ref",
    "offy_ref",
    "cx_ref",
    "cy_ref",
    "fld_id_ref",
    "fld_type_ref",
    "master_name"
  )
  tmp_names[match(old_, tmp_names)] <- new_
  names(master_xfrm) <- tmp_names
  master_xfrm$id <- NULL
  master_xfrm$ph <- NULL
  master_xfrm$ph_label <- NULL
  master_xfrm$rotation <- NULL

  master_xfrm
}


xfrmize <- function(slide_xfrm, master_xfrm) {
  x <- as.data.frame(slide_xfrm)

  master_ref <- unique(data.frame(
    file = master_xfrm$file,
    master_name = master_xfrm$name,
    stringsAsFactors = FALSE
  ))
  master_xfrm <- fortify_master_xfrm(master_xfrm)

  slide_key_id <- paste0(x$master_file, x$type)
  master_key_id <- paste0(master_xfrm$file, master_xfrm$type)

  slide_xfrm_no_match <- x[!slide_key_id %in% master_key_id, ]
  slide_xfrm_no_match <- merge(
    slide_xfrm_no_match,
    master_ref,
    by.x = "master_file",
    by.y = "file",
    all.x = TRUE,
    all.y = FALSE
  )

  x <- merge(
    x,
    master_xfrm,
    by.x = c("master_file", "type"),
    by.y = c("file", "type"),
    all = FALSE
  )
  x$offx <- ifelse(!is.finite(x$offx), x$offx_ref, x$offx)
  x$offy <- ifelse(!is.finite(x$offy), x$offy_ref, x$offy)
  x$cx <- ifelse(!is.finite(x$cx), x$cx_ref, x$cx)
  x$cy <- ifelse(!is.finite(x$cy), x$cy_ref, x$cy)
  x$offx_ref <- NULL
  x$offy_ref <- NULL
  x$cx_ref <- NULL
  x$cy_ref <- NULL
  x$fld_id_ref <- NULL
  x$fld_type_ref <- NULL

  x <- rbind(x, slide_xfrm_no_match, stringsAsFactors = FALSE)

  i_master <- get_file_index(x$master_file)
  i_layout <- get_file_index(x$file)
  x <- x[order(i_master, i_layout, x$offy, x$offx), , drop = FALSE] # intuitive sorting: top -> bottom, left -> right
  x <- x[!(is.na(x$offx) | is.na(x$offy) | is.na(x$cx) | is.na(x$cy)), ]

  x$type_idx <- stats::ave(
    x$type,
    x$master_file,
    x$file,
    x$type,
    FUN = seq_along
  )
  x$type_idx <- as.numeric(x$type_idx) # NB: ave returns character

  x$id <- as.integer(x$id)

  rownames(x) <- NULL # prevent meaningless rownames
  x
}


read_theme_colors <- function(doc, theme) {
  nodes <- xml_find_all(doc, "//a:clrScheme/*")

  names_ <- xml_name(nodes)
  col_types_ <- xml_name(xml_children(nodes))
  vals <- xml_attr(xml_children(nodes), "val")
  last_colors_ <- xml_attr(xml_children(nodes), "lastClr")
  vals <- ifelse(
    col_types_ == "srgbClr",
    paste0("#", vals),
    paste0("#", last_colors_)
  )
  data.frame(
    stringsAsFactors = FALSE,
    name = names_,
    type = col_types_,
    value = vals,
    theme = theme
  )
}


characterise_df <- function(x) {
  names(x) <- htmlEscapeCopy(names(x))
  x <- lapply(x, function(x) {
    if (is.character(x)) {
      x
    } else if (is.factor(x)) {
      as.character(x)
    } else {
      gsub("(^ | $)+", "", format(x))
    }
  })
  data.frame(x, stringsAsFactors = FALSE, check.names = FALSE)
}


xpath_content_selector <- "*[self::p:cxnSp or self::p:sp or self::p:graphicFrame or self::p:grpSp or self::p:pic]"

as_xpath_content_sel <- function(prefix) {
  paste0(prefix, xpath_content_selector)
}


between <- function(x, left, right) {
  x >= left & x <= right
}


rbind_match_columns <- function(list_df) {
  col <- unique(unlist(lapply(list_df, colnames)))
  x <- Filter(function(x) nrow(x) > 0, list_df)
  x <- lapply(
    x,
    function(x, col) {
      x[, setdiff(col, colnames(x))] <- NA
      x
    },
    col = col
  )
  do.call(rbind, x)
}

set_row_span <- function(row_details) {
  row_details$first[!row_details$first & !row_details$row_merge] <- TRUE
  row_details$row_merge <- NULL

  row_details <- split(row_details, row_details$cell_id)

  row_details <- mapply(
    function(dat) {
      rowspan_values_at_breaks <- rle(cumsum(dat$first))$lengths
      rowspan_pos_at_breaks <- which(dat$first)
      dat$row_span <- 0L
      dat$row_span[rowspan_pos_at_breaks] <- rowspan_values_at_breaks
      dat
    },
    row_details,
    SIMPLIFY = FALSE
  )
  row_details <- rbind_match_columns(row_details)
  row_details$first <- NULL
  row_details
}


#' @importFrom grDevices col2rgb rgb
is.color = function(x) {
  # http://stackoverflow.com/a/13290832/3315962
  out = sapply(x, function(x) {
    tryCatch(is.matrix(col2rgb(x)), error = function(e) F)
  })

  nout <- names(out)
  if (!is.null(nout) && any(is.na(nout))) {
    out[is.na(nout)] = FALSE
  }

  out
}

correct_id <- function(doc, int_id) {
  all_uid <- xml_find_all(doc, "//*[@id]")
  for (z in seq_along(all_uid)) {
    if (!grepl("[^0-9]", xml_attr(all_uid[[z]], "id"))) {
      xml_attr(all_uid[[z]], "id") <- int_id
      int_id <- int_id + 1
    }
  }
  int_id
}


check_bookmark_id <- function(bkm) {
  if (!is.null(bkm)) {
    invalid_bkm <- is.character(bkm) &&
      length(bkm) == 1 &&
      nchar(bkm) > 0 &&
      grepl("[^:[:alnum:]_-]+", bkm)
    if (invalid_bkm) {
      stop(
        "bkm [",
        bkm,
        "] should only contain alphanumeric characters, ':', '-' and '_'.",
        call. = FALSE
      )
    }
  }
  bkm
}

is_windows <- function() {
  "windows" %in% .Platform$OS.type
}

is_doc_open <- function(file) {
  # The function checks if the `file` is open (a.k.a. is being edited).
  # This function is valid on Windows operating system only.
  suppressWarnings(file.exists(file) && !file.rename(from = file, to = file))
}


# Extract trailing numeric index in .xml filename
#
# Useful to for slideMaster and slideLayout .xml files.
#
# Examples:
#   files <- c("slideLayout1.xml", "slideLayout2.xml", "slideLayout10.xml")
#   get_file_index(files)
#
get_file_index <- function(file) {
  x <- sub(
    pattern = ".+?(\\d+).xml$",
    replacement = "\\1",
    x = basename(file),
    ignore.case = TRUE
  )
  as.numeric(x)
}


# Sort xml filenames by trailing numeric index
#
# Useful to for slideMaster and slideLayout xml files.
#
# Examples:
#   files <- c("slideLayout1.xml", "slideLayout2.xml", "slideLayout12.xml")
#   sort_vec_by_index(files)  # => order corresponding to trailing index
#   sort(files)           # => incorrect lexicographical ordering
#
sort_vec_by_index <- function(x) {
  indexes <- get_file_index(x)
  x[order(indexes)]
}


# Sort dataframe column by trailing index
#
# df: A dataframe
# ...: columsn to sort by, comma separated
#
# Examples:
# df <- data.frame(
#   a = paste0("file_", rep(3:1, each = 2), ".xml"),
#   b = paste0("file_", rep(3:1, 2), ".xml")
# )
# sort_dataframe_by_index(df, "a", "b")
# sort_dataframe_by_index(df, "b", "a")
#
sort_dataframe_by_index <- function(df, ...) {
  sort_columns <- c(...)
  l <- lapply(sort_columns, function(.col) {
    get_file_index(df[[.col]])
  })
  df[do.call(order, l), , drop = FALSE]
}


# rename dataframe columns
#
# Examples:
#   df_rename(mtcars, c("mpg", "cyl"), c("A", "B"))
#
df_rename <- function(df, old, new) {
  .nms <- names(df)
  .nms[match(old, .nms)] <- new
  stats::setNames(df, .nms)
}


# replacement for stopifnot() with nicer user feedback
stop_if_not_class <- function(x, class, arg = NULL) {
  check <- inherits(x, what = class)
  if (!check) {
    msg_arg <- ifelse(
      is.null(arg),
      "Incorrect input.",
      "Incorrect input for {.arg {arg}}"
    )
    cli::cli_abort(
      c(
        msg_arg,
        "x" = "Expected {.cls {class}} but got {.cls {class(x)[1]}}"
      ),
      call = NULL
    )
  }
}


stop_if_not_rpptx <- function(x, arg = NULL) {
  stop_if_not_class(x, "rpptx", arg)
}


stop_if_not_integerish <- function(x, arg = NULL) {
  check <- is_integerish(x)
  if (!check) {
    msg_arg <- ifelse(
      is.null(arg),
      "Incorrect input.",
      "Incorrect input for {.arg {arg}}"
    )
    cli::cli_abort(
      c(
        msg_arg,
        "x" = "Expected integerish values but got {.cls {class(x)[1]}}"
      ),
      call = NULL
    )
  }
}


#' Ensure valid slide indexes
#'
#' @param x An `rpptx` object.
#' @param idx Slide indexes.
#' @param arg Name of argument to use in error message (optional).
#' @param call Environment to display in error message. Defaults to caller env.
#'   Set `NULL` to suppress (see [cli::cli_abort]).
#' @keywords internal
stop_if_not_in_slide_range <- function(
  x,
  idx,
  arg = NULL,
  call = parent.frame()
) {
  stop_if_not_rpptx(x)
  stop_if_not_integerish(idx)

  n_slides <- length(x)
  idx_available <- seq_len(n_slides)
  idx_outside <- setdiff(idx, idx_available)
  n_outside <- length(idx_outside)

  if (n_outside == 0) {
    return(invisible(NULL))
  }
  argname <- ifelse(is.null(arg), "", "of {.arg {arg}} ")
  part_1 <- paste0(
    "{n_outside} index{?es} ",
    argname,
    "outside slide range: {.val {idx_outside}}"
  )
  part_2 <- ifelse(
    n_slides == 0,
    "Presentation has no slides!",
    "Slide indexes must be in the range [{min(idx_available)}..{max(idx_available)}]"
  )
  cli::cli_abort(c(part_1, "x" = part_2), call = call)
}


check_unit <- function(unit, choices, several.ok = FALSE) {
  if (!several.ok && length(unit) != 1) {
    cli::cli_abort(
      c(
        "{.arg unit} is not length 1.",
        "x" = "{.arg unit} must be {.emph a string}."
      )
    )
  }
  if (!unit %in% choices) {
    cli::cli_abort(
      c(
        "{.arg unit} should be one of {.or {choices}}.",
        "x" = "{.arg unit} was {.emph {unit}\"}."
      )
    )
  }

  unit
}


#' Update a vector or list of named defaults values
#'
#' Helper to update named default values. Takes care of most common cases and
#' returns sensible error messages.
#'
#' @param x A (named) vector or list to update the default values with.
#' @param default A *named* vector or list with default values.
#' @param argname Name of the arg. Used in error message, defaults to `"x"`.
#' @param default_if_null Returns defaults values if `x` is `NULL`? (default is `TRUE`)
#' @param partial Enable partial name matching= (default `TRUE`).
#' @param as_list Return a list (default `TRUE`) or a named vector?
#' @return List of update default values.
#' @noRd
#' @examples
#' defaults <- c(aa = 1, bb = 2, cc = 3)
#' update_named_defaults(c(a = 99, b = 0), defaults) # partial name match
#' update_named_defaults(c(a = 99, b = 0), defaults, as_list = FALSE) # return as vector
#' update_named_defaults(3:1, defaults) # match by position
#' update_named_defaults(2, defaults)
#' update_named_defaults(NULL, defaults, default_if_null = TRUE)
#'
update_named_defaults <- function(
  x,
  default,
  argname = "x",
  default_if_null = TRUE,
  partial = TRUE,
  as_list = TRUE
) {
  if (default_if_null && is.null(x)) {
    res <- as.list(default)
    if (!as_list) {
      res <- unlist(res)
    }
    return(res)
  }

  x <- as.list(x)
  default <- as.list(default)

  if (!is_named(default)) {
    cli::cli_abort(
      c(
        "Some default vector elements have no names",
        "x" = "{.arg default} must be a named vector"
      ),
      call = NULL
    )
  }
  x <- as.list(x)
  len_x <- length(x)
  len_default <- length(default)
  names_default <- names(default)
  if (len_x > len_default) {
    cli::cli_abort(
      c(
        "Length of {.arg {argname}} ({.val {len_x}}) exceeds length of {.arg default} ({.val {len_default}})",
        "x" = "Length of {.arg x} must be smaller or equal to the length of {.arg default}"
      ),
      call = NULL
    )
  }

  # unnamed case => convert to named case
  if (!is_named(x)) {
    if (len_x == 1) {
      x <- rep(x, len_default)
      len_x <- length(x)
    }
    if (len_x != len_default) {
      cli::cli_abort(
        c(
          "{.arg {argname}} has incorrect length ({len_x})",
          "x" = "If {.arg {argname}} has no names, it must be length 1 or the length of {.arg default} ({len_default})"
        ),
        call = NULL
      )
    }
    names(x) <- names_default
  }

  # named case
  if (partial) {
    # => partial name matching
    matched <- pmatch(names(x), names_default, duplicates.ok = TRUE)
  } else {
    # exact name matching
    matched <- match(names(x), names_default, nomatch = NA)
  }
  nms_new <- ifelse(is.na(matched), NA, names_default[matched])
  i_na <- is.na(nms_new)

  if (any(i_na)) {
    msg_partial <- ifelse(
      partial,
      "Partial name matching is supported",
      "Partial name matching is not enabled"
    )
    cli::cli_abort(
      c(
        "Found {sum(i_na)} unknown name{?s} in {.arg {argname}}: {.val {names(x)[i_na]}}",
        "x" = "{.arg {argname}} understands {.val {names_default}}",
        "i" = cli::col_silver(msg_partial)
      ),
      call = NULL
    )
  }
  # duplicate position
  ii_dupes <- duplicated(nms_new)
  if (any(ii_dupes)) {
    cli::cli_abort(
      c(
        "Duplicate entries in {.arg location}: {.val {unique(nms_new[ii_dupes])}}",
        "x" = "Each name in {.arg location} must be unique",
        "i" = cli::col_silver("Partial name matching is supported")
      ),
      call = NULL
    )
  }
  x <- setNames(x, nms_new)
  res <- utils::modifyList(x = default, val = as.list(x), keep.null = TRUE)
  if (!as_list) {
    res <- unlist(res)
  }
  res
}


# htmlEscapeCopy ----

htmlEscapeCopy <- local({
  .htmlSpecials <- list(
    `&` = '&amp;',
    `<` = '&lt;',
    `>` = '&gt;'
  )
  .htmlSpecialsPattern <- paste(names(.htmlSpecials), collapse = '|')
  .htmlSpecialsAttrib <- c(
    .htmlSpecials,
    `'` = '&#39;',
    `"` = '&quot;',
    `\r` = '&#13;',
    `\n` = '&#10;'
  )
  .htmlSpecialsPatternAttrib <- paste(
    names(.htmlSpecialsAttrib),
    collapse = '|'
  )
  function(text, attribute = FALSE) {
    pattern <- if (attribute) {
      .htmlSpecialsPatternAttrib
    } else {
      .htmlSpecialsPattern
    }
    text <- enc2utf8(as.character(text))
    # Short circuit in the common case that there's nothing to escape
    if (!any(grepl(pattern, text, useBytes = TRUE))) {
      return(text)
    }
    specials <- if (attribute) {
      .htmlSpecialsAttrib
    } else {
      .htmlSpecials
    }
    for (chr in names(specials)) {
      text <- gsub(chr, specials[[chr]], text, fixed = TRUE, useBytes = TRUE)
    }
    Encoding(text) <- "UTF-8"
    return(text)
  }
})


# metric units -----------------------------------------------
#
cm_to_inches <- function(x) {
  x / 2.54
}

mm_to_inches <- function(x) {
  x / 25.4
}

convin <- function(unit, x) {
  unit <- match.arg(unit, choices = c("in", "cm", "mm"), several.ok = FALSE)
  if (!identical("in", unit)) {
    x <- do.call(paste0(unit, "_to_inches"), list(x = x))
  }
  x
}


# from rlang pkg  ------------------------------------------

is_named <- function(x) {
  nms <- names(x)
  if (is.null(nms)) {
    return(FALSE)
  }
  if (any(detect_void_name(nms))) {
    return(FALSE)
  }
  TRUE
}


detect_void_name <- function(x) {
  x == "" | is.na(x)
}


# is_integerish(1)
# is_integerish(1.0)
# is_integerish(c(1.0, 2.0))
is_integerish <- function(x) {
  ii <- all(is.numeric(x) | is.integer(x))
  jj <- all(x == as.integer(x))
  ii && jj
}


# coalesce for R
`%||%` <- function(l, r) {
  if (is.null(l)) r else l
}


# file ops  ------------------------------------------

#' Opens a file locally
#'
#' Opening a file locally requires a compatible application to be installed
#' (e.g., MS Office or LibreOffice for .pptx or .docx files).
#'
#' @details
#' *NB:* Function is a small wrapper around [utils::browseURL()] to have a more
#' suitable function name.
#'
#' @param path File path.
#' @export
#' @examples
#' x <- read_pptx()
#' x <- add_slide(x, "Title Slide", ctrTitle = "My Title")
#' file <- print(x, tempfile(fileext = ".pptx"))
#' \dontrun{
#' open_file(file)}
open_file <- function(path) {
  path <- normalizePath(path, winslash = "/", mustWork = TRUE)
  utils::browseURL(path)
}

test_that("trailing file index extraction / sorting", {
  files <- c(
    "slideLayout1.xml",
    "slideLayout11.xml",
    "slideLayout2.xml",
    "slideLayout10.xml"
  )

  expect_equal(get_file_index(files), c(1, 11, 2, 10))

  expect_equal(
    sort_vec_by_index(files),
    c(
      "slideLayout1.xml",
      "slideLayout2.xml",
      "slideLayout10.xml",
      "slideLayout11.xml"
    )
  )

  df <- data.frame(file1 = files, file2 = rev(files))
  a <- sort_dataframe_by_index(df, "file1")
  b <- sort_dataframe_by_index(df, "file2")
  expect_true(all(a == rev(b)))
})


test_that("misc", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  df <- df_rename(mtcars, c("mpg", "cyl"), c("A", "B"))
  expect_true(all(names(df)[1:2] == c("A", "B")))

  expect_true(is_integerish(1))
  expect_true(is_integerish(1:3))
  expect_true(is_integerish(1.0))
  expect_true(is_integerish(c(1.0, 2.0)))
  expect_false(is_integerish(1.00001))
  expect_false(is_integerish(c(1.1, 2.0)))
  expect_false(is_integerish(TRUE))
  expect_false(is_integerish(FALSE))
  expect_error(
    expect_warning(stop_if_not_integerish(LETTERS)),
    regex = "Expected integerish values but got <character>"
  )
  expect_error(
    stop_if_not_integerish(c(1.1, 1.2)),
    regex = "Expected integerish values but got <numeric>"
  )
})


test_that("stop_if_not_in_slide_range", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  expect_error(
    regex = "Presentation has no slides",
    stop_if_not_in_slide_range(x, 1)
  )

  x <- add_slide(x, "Title and Content")
  x <- add_slide(x, "Title and Content")
  expect_no_error(stop_if_not_in_slide_range(x, 1:2))
  expect_error(
    regex = "1 index outside slide range",
    stop_if_not_in_slide_range(x, -1)
  )
  expect_error(
    regex = "2 indexes of `my_arg` outside slide range",
    stop_if_not_in_slide_range(x, 3:4, arg = "my_arg")
  )
  expect_error(
    regex = "Expected integerish values but got <numeric>",
    stop_if_not_in_slide_range(x, 3.1)
  )

  foo <- function() {
    stop_if_not_in_slide_range(x, 3:4)
  }
  error_text <- tryCatch(foo(), error = paste)
  grepl("^Error in `foo()`:", error_text, fixed = TRUE)

  foo <- function() {
    stop_if_not_in_slide_range(x, 3:4, call = NULL)
  }
  error_text <- tryCatch(foo(), error = paste)
  grepl("^Error:", error_text, fixed = TRUE)
})


test_that("update_named_defaults", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  # errors
  expect_error(
    update_named_defaults(x = 3, default = 1:3),
    "Some default vector elements have no names"
  )
  expect_error(
    update_named_defaults(x = 1:2, default = c(a = 1, b = 2, c = 3)),
    "`x` has incorrect length (2)",
    fixed = TRUE
  )
  expect_error(
    update_named_defaults(x = 1:2, default = c(a = 1)),
    "Length of `x` (2) exceeds length of `default` (1)",
    fixed = TRUE
  )
  expect_error(
    update_named_defaults(
      x = c(d = 3, b = 2, c = 1),
      default = c(aa = 1, bb = 2, cc = 3)
    ),
    'Found 1 unknown name in `x`: "d"',
    fixed = TRUE
  )
  expect_error(
    update_named_defaults(
      x = c(b = 3, b = 2, c = 1),
      default = c(aa = 1, bb = 2, cc = 3)
    ),
    'Duplicate entries in `location`: "bb"',
    fixed = TRUE
  )
  expect_error(
    update_named_defaults(
      x = c(b = 3, b = 2, c = 1),
      default = c(aa = 1, bb = 2, cc = 3),
      argname = "xxx",
      partial = FALSE
    ),
    'Found 3 unknown names in `xxx`: "b", "b", and "c"',
    fixed = TRUE
  )

  # working
  expect_equal(
    update_named_defaults(x = 2, default = c(a = 1, b = 2, c = 3)),
    list(a = 2, b = 2, c = 2)
  )
  expect_equal(
    update_named_defaults(x = 3:1, default = c(a = 1, b = 2, c = 3)),
    list(a = 3, b = 2, c = 1)
  )
  expect_equal(
    update_named_defaults(x = NULL, default = c(a = 1, b = 2, c = 3)),
    list(a = 1, b = 2, c = 3)
  )
  expect_equal(
    update_named_defaults(
      x = NULL,
      default = list(a = 1, b = 2, c = 3),
      as_list = FALSE
    ),
    c(a = 1, b = 2, c = 3)
  )
  expect_equal(
    update_named_defaults(
      x = 3:1,
      default = c(a = 1, b = 2, c = 3),
      as_list = FALSE
    ),
    c(a = 3, b = 2, c = 1)
  )
  expect_equal(
    update_named_defaults(
      x = c(a = 3, b = 2, c = 1),
      default = c(aa = 1, bb = 2, cc = 3)
    ),
    list(aa = 3, bb = 2, cc = 1)
  )
  expect_equal(
    update_named_defaults(x = c(a = 99), default = c(aa = 1, bb = 2, cc = 3)),
    list(aa = 99, bb = 2, cc = 3)
  )
})

test_that("trailing file index extraction / sorting", {
  files <- c("slideLayout1.xml", "slideLayout11.xml", "slideLayout2.xml", "slideLayout10.xml")

  expect_equal(get_file_index(files), c(1, 11, 2, 10))

  expect_equal(sort_vec_by_index(files), c("slideLayout1.xml", "slideLayout2.xml", "slideLayout10.xml", "slideLayout11.xml"))

  df <- data.frame(file1 = files, file2 = rev(files))
  a <- sort_dataframe_by_index(df, "file1")
  b <- sort_dataframe_by_index(df, "file2")
  expect_true(all(a == rev(b)))
})


test_that("misc", {
  df <- df_rename(mtcars, c("mpg", "cyl"), c("A", "B"))
  expect_true(all(names(df)[1:2] == c("A", "B")))
})

# caption ----

#' @export
#' @title Caption block
#' @description Create a representation of a
#' caption that can be used for cross reference.
#' @param label a scalar character representing label to display
#' @param style paragraph style name
#' @param autonum an object generated with function [run_autonum]
#' @examples
#' library(officer)
#'
#' run_num <- run_autonum(seq_id = "tab", pre_label = "tab. ",
#'   bkm = "mtcars_table")
#' caption <- block_caption("mtcars table",
#'   style = "Normal",
#'   autonum = run_num
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add(doc_1, "A title", style = "heading 1")
#' doc_1 <- body_add(doc_1, "Hello world!", style = "Normal")
#' doc_1 <- body_add(doc_1, caption)
#' doc_1 <- body_add(doc_1, mtcars, style = "table_template")
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family block functions for reporting
block_caption <- function(label, style = NULL, autonum = NULL) {
  if (is.null(style)) {
    style <- "Normal"
  }

  z <- list(
    label = label,
    autonum = autonum,
    style = style
  )

  class(z) <- c("block_caption", "block")
  z
}

#' @export
print.block_caption <- function(x, ...) {
  if (is.null(x$autonum)) {
    auton <- "[autonum off]"
  } else {
    auton <- "[autonum on]"
  }
  cat("caption ", auton, ": ", x$label, "\n", sep = "")
}

to_wml_block_caption_officer <- function(x, add_ns = FALSE) {
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  autonum <- ""
  if (!is.null(x$autonum)) {
    autonum <- to_wml(x$autonum)
  }

  run_str <- sprintf(
    "<w:r><w:t xml:space=\"preserve\">%s</w:t></w:r>",
    htmlEscapeCopy(x$label)
  )
  run_str <- paste0(autonum, run_str)

  out <- sprintf(
    "%s<w:pPr><w:pStyle w:pstlname=\"%s\"/></w:pPr>%s</w:p>",
    open_tag,
    x$style,
    run_str
  )

  out
}
to_wml_block_caption_pandoc <- function(x, bookdown_id = NULL) {
  if (is.null(x$label)) {
    return("")
  }

  autonum <- ""
  if (!is.null(x$autonum)) {
    autonum <- paste("`", to_wml(x$autonum), "`{=openxml}", sep = "")
  }

  run_str <- paste0(autonum, htmlEscapeCopy(x$label))

  paste0(
    if (!is.null(x$style)) paste0("\n\n::: {custom-style=\"", x$style, "\"}"),
    "\n\n",
    # "<caption>\n\n",
    if (!is.null(bookdown_id)) bookdown_id,
    run_str,
    # "\n\n</caption>",
    if (!is.null(x$style)) paste0("\n:::\n"),
    "\n\n"
  )
}

#' @export
to_wml.block_caption <- function(x, add_ns = FALSE, knitting = FALSE, ...) {
  if (knitting) {
    to_wml_block_caption_pandoc(x, bookdown_id = list(...)$bookdown_id)
  } else {
    to_wml_block_caption_officer(x, add_ns = add_ns)
  }
}


# toc ----

#' @export
#' @title Table of content for 'Word'
#' @description Create a representation of a table
#' of content for Word documents.
#' @param level max title level of the table
#' @param style optional. If not NULL, its value is used as style in the
#' document that will be used to build entries of the TOC.
#' @param seq_id optional. If not NULL, its value is used as sequence
#' identifier in the document that will be used to build entries of the
#' TOC. See also [run_autonum()] to specify a sequence identifier.
#' @param separator unused, no effect
#' @examples
#' block_toc(level = 2)
#' block_toc(style = "Table Caption")
#' @family block functions for reporting
block_toc <- function(level = 3, style = NULL, seq_id = NULL, separator = ";") {
  z <- list(
    level = level,
    style = style,
    seq_id = seq_id,
    separator = separator
  )
  class(z) <- c("block_toc", "block")

  z
}

#' @export
print.block_toc <- function(x, ...) {
  if (is.null(x$style) && is.null(x$seq_id)) {
    cat("TOC - max level: ", x$level, "\n", sep = "")
  } else if (!is.null(x$style)) {
    cat("TOC for style: ", x$style, "\n", sep = "")
  } else if (!is.null(x$seq_id)) {
    cat("TOC for seq identifier: ", x$seq_id, "\n", sep = "")
  }
}

#' @export
to_wml.block_toc <- function(x, add_ns = FALSE, ...) {
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  if (is.null(x$style) && is.null(x$seq_id)) {
    out <- paste0(
      open_tag,
      "<w:pPr/>",
      to_wml(
        run_word_field(
          field = sprintf(
            "TOC \\o \"1-%.0f\" \\h \\z \\u",
            x$level
          )
        )
      ),
      "</w:p>"
    )
  } else if (!is.null(x$style)) {
    out <- paste0(
      open_tag,
      "<w:pPr/>",
      to_wml(
        run_word_field(
          field = sprintf(
            "TOC \\h \\z \\t \"%s\"",
            x$style
          )
        )
      ),
      "</w:p>"
    )
  } else if (!is.null(x$seq_id)) {
    out <- paste0(
      open_tag,
      "<w:pPr/>",
      to_wml(
        run_word_field(
          field = sprintf("TOC \\h \\z \\c \"%s\"", x$seq_id)
        )
      ),
      "</w:p>"
    )
  }

  out
}

# pour_docx ----

#' @export
#' @title External Word document placeholder
#' @description Pour the content of a docx file in the resulting docx
#' from an 'R Markdown' document.
#' @param file external docx file path
#' @examples
#' library(officer)
#' docx <- tempfile(fileext = ".docx")
#' doc <- read_docx()
#' doc <- body_add(doc, iris[1:20,], style = "table_template")
#' print(doc, target = docx)
#'
#' target <- tempfile(fileext = ".docx")
#' doc_1 <- read_docx()
#' doc_1 <- body_add(doc_1, block_pour_docx(docx))
#' print(doc_1, target = target)
#' @family block functions for reporting
block_pour_docx <- function(file) {
  if (!file.exists(file)) {
    stop("file {", file, "} does not exist.", call. = FALSE)
  }
  if (!grepl("\\.docx$", file, ignore.case = TRUE)) {
    stop("file {", file, "} is not a docx file.", call. = FALSE)
  }
  if (grepl("&", file, ignore.case = TRUE)) {
    stop(
      "file path {",
      file,
      "} contains '&', please rename your file.",
      call. = FALSE
    )
  }
  if (grepl(" ", basename(file), ignore.case = TRUE)) {
    stop(
      "file path {",
      basename(file),
      "} contains ' ', please rename your file.",
      call. = FALSE
    )
  }

  z <- list(file = file)
  class(z) <- c("block_pour_docx", "block")
  z
}

#' @export
print.block_pour_docx <- function(x, ...) {
  cat("Pour docx: ", x$file, "\n", sep = "")
}

#' @export
to_wml.block_pour_docx <- function(x, add_ns = FALSE, ...) {
  ns_str <- ""
  if (add_ns) {
    ns_str <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" "
  }
  out <- paste0("<w:altChunk ", ns_str, "r:id=\"", x$file, "\"/>")
  out
}


# section ----

#' @export
#' @title Section for 'Word'
#' @description Create a representation of a section.
#'
#' A section affects preceding paragraphs or tables; i.e.
#' a section starts at the end of the previous section (or the beginning of
#' the document if no preceding section exists), and stops where the
#' section is declared.
#'
#' When a new landscape section is needed, it is recommended to add a block_section
#' with `type = "continuous"`, to add the content to be appened in the new section
#' and finally to add a block_section with `page_size = page_size(orient = "landscape")`.
#' @param property section properties defined with function [prop_section]
#' @examples
#' ps <- prop_section(
#'   page_size = page_size(orient = "landscape"),
#'   page_margins = page_mar(top = 2),
#'   type = "continuous"
#' )
#' block_section(ps)
#' @family block functions for reporting
block_section <- function(property) {
  z <- list(
    property = property
  )
  class(z) <- c("block_section", "block")

  z
}

#' @export
print.block_section <- function(x, ...) {
  cat("----- end of section: ", "\n", sep = "")
}

#' @export
to_wml.block_section <- function(x, add_ns = FALSE, ...) {
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  out <- paste0(open_tag, "<w:pPr>", to_wml(x$property), "</w:pPr></w:p>")

  out
}


# table properties ----

#' @export
#' @title Table conditional formatting
#' @description Tables can be conditionally formatted based on few properties as
#' whether the content is in the first row, last row, first column, or last
#' column, or whether the rows or columns are to be banded.
#' @param first_row,last_row apply or remove formatting from the first or last row in the table.
#' @param first_column,last_column apply or remove formatting from the first or last column in the table.
#' @param no_hband,no_vband don't display odd and even rows or columns with
#' alternating shading for ease of reading.
#' @note
#' You must define a format for first_row, first_column and other properties
#' if you need to use them. The format is defined in a docx template.
#' @examples
#' table_conditional_formatting(first_row = TRUE, first_column = TRUE)
#' @family functions for table definition
table_conditional_formatting <- function(
  first_row = TRUE,
  first_column = FALSE,
  last_row = FALSE,
  last_column = FALSE,
  no_hband = FALSE,
  no_vband = TRUE
) {
  z <- list(
    first_row = first_row,
    first_column = first_column,
    last_row = last_row,
    last_column = last_column,
    no_hband = no_hband,
    no_vband = no_vband
  )
  class(z) <- c("table_conditional_formatting")
  z
}

#' @export
to_wml.table_conditional_formatting <- function(x, add_ns = FALSE, ...) {
  paste0(
    "<w:tblLook w:firstRow=\"",
    as.integer(x$first_row),
    "\" w:lastRow=\"",
    as.integer(x$last_row),
    "\" w:firstColumn=\"",
    as.integer(x$first_column),
    "\" w:lastColumn=\"",
    as.integer(x$last_column),
    "\" w:noHBand=\"",
    as.integer(x$no_hband),
    "\" w:noVBand=\"",
    as.integer(x$no_vband),
    "\"/>"
  )
}

#' @export
to_pml.table_conditional_formatting <- function(x, add_ns = FALSE, ...) {
  expr_ <- paste0(
    " firstRow=\"%.0f\" lastRow=\"%.0f\"",
    " firstColumn=\"%.0f\" lastColumn=\"%.0f\"",
    " bandRow=\"%.0f\" bandCol=\"%.0f\""
  )
  sprintf(
    expr_,
    x$first_row,
    x$last_row,
    x$first_column,
    x$last_column,
    !x$no_hband,
    !x$no_vband
  )
}

table_layout_types <- c("autofit", "fixed")


#' @export
#' @title Algorithm for table layout
#' @description When a table is displayed in a document, it can
#' either be displayed using a fixed width or autofit layout algorithm:
#'
#' * fixed: uses fixed widths for columns. The width of the table is not
#' changed regardless of the contents of the cells.
#' * autofit: uses the contents of each cell and the table width to
#' determine the final column widths.
#' @param type 'autofit' or 'fixed' algorithm. Default to 'autofit'.
#' @family functions for table definition
table_layout <- function(type = "autofit") {
  if (!type %in% table_layout_types) {
    stop(
      "type must be one of ",
      paste(table_layout_types, collapse = ", "),
      "."
    )
  }

  z <- list(type = type)
  class(z) <- "table_layout"
  z
}

#' @export
to_wml.table_layout <- function(x, add_ns = FALSE, ...) {
  sprintf("<w:tblLayout w:type=\"%s\"/>", x$type)
}

table_layout_width_units <- c("in", "pct")

#' @export
#' @title Preferred width for a table
#' @description Define the preferred width for a table.
#' @section Word:
#' All widths in a table are considered preferred because widths of
#' columns can conflict and the table layout rules can require a
#' preference to be overridden.
#' @param width value of the preferred width of the table.
#' @param unit unit of the width. Possible values are 'in' (inches) and 'pct' (percent)
#' @family functions for table definition
table_width <- function(width = 1, unit = "pct") {
  if (!unit %in% table_layout_width_units) {
    stop(
      "unit must be one of ",
      paste(table_layout_width_units, collapse = ", "),
      "."
    )
  }

  z <- list(width = width, unit = unit)
  class(z) <- "table_width"
  z
}
#' @export
to_wml.table_width <- function(x, add_ns = FALSE, ...) {
  if (x$unit %in% "pct") {
    tbl_width <- sprintf(
      "<w:tblW w:type=\"pct\" w:w=\"%.0f\"/>",
      x$width * 5000
    )
  } else {
    tbl_width <- sprintf(
      "<w:tblW w:type=\"dxa\" w:w=\"%.0f\"/>",
      x$width * 1440
    )
  }
  tbl_width
}

#' @export
#' @title Column widths of a table
#' @description The function defines the size of each column of a table.
#' @param widths Column widths expressed in inches.
#' @family functions for table definition
table_colwidths <- function(widths = NULL) {
  z <- list(widths = widths)
  class(z) <- "table_colwidths"
  z
}

#' @export
#' @title Paragraph styles for columns
#' @description The function defines the paragraph styles for columns.
#' @param stylenames a named character vector, names are column names, values are
#' paragraph styles associated with each column. If a column is not
#' specified, default value 'Normal' is used.
#' Another form is as a named list, the list names are the styles
#' and the contents are column names to be formatted with the
#' corresponding style.
#' @family functions for table definition
#' @examples
#' library(officer)
#'
#' stylenames <- c(
#'   vs = "centered", am = "centered",
#'   gear = "centered", carb = "centered"
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_table(doc_1,
#'   value = mtcars, style = "table_template",
#'   stylenames = table_stylenames(stylenames = stylenames)
#' )
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#'
#'
#' stylenames <- list(
#'   "centered" = c("vs", "am", "gear", "carb")
#' )
#'
#' doc_2 <- read_docx()
#' doc_2 <- body_add_table(doc_2,
#'   value = mtcars, style = "table_template",
#'   stylenames = table_stylenames(stylenames = stylenames)
#' )
#'
#' print(doc_2, target = tempfile(fileext = ".docx"))
table_stylenames <- function(stylenames = list()) {
  if (length(stylenames) > 0 && is.null(attr(stylenames, "names"))) {
    stop("stylenames should have names")
  }

  if (length(stylenames) > 0 && is.list(stylenames)) {
    .l <- vapply(stylenames, length, FUN.VALUE = 0L)
    zz <- inverse.rle(
      structure(
        list(
          lengths = .l,
          values = names(stylenames)
        ),
        class = "rle"
      )
    )
    names(zz) <- as.character(unlist(stylenames))
    stylenames <- as.list(zz)
  } else if (is.character(stylenames)) {
    stylenames <- as.list(stylenames)
  }

  z <- list(stylenames = stylenames)
  class(z) <- "table_stylenames"
  z
}

#' @export
#' @importFrom utils modifyList
to_wml.table_stylenames <- function(x, add_ns = FALSE, dat, ...) {
  stylenames <- rep("Normal", ncol(dat))
  names(stylenames) <- colnames(dat)
  stylenames <- as.list(stylenames)
  # restrict to only existing cols
  x$stylenames <- x$stylenames[names(x$stylenames) %in% colnames(dat)]
  stylenames <- modifyList(stylenames, val = x$stylenames)
  stylenames <- lapply(stylenames, function(x) {
    sprintf("<w:pStyle w:pstlname=\"%s\"/>", x)
  })
  stylenames
}

#' @export
to_wml.table_colwidths <- function(x, add_ns = FALSE, ...) {
  if (length(x$widths) < 1) {
    return("")
  }
  grid_col_str <- sprintf("<w:gridCol w:w=\"%.0f\"/>", x$widths * 1440)
  grid_col_str <- paste(grid_col_str, collapse = "")
  paste0("<w:tblGrid>", grid_col_str, "</w:tblGrid>")
}

#' @export
#' @title Table properties
#' @description Define table properties such as fixed or autofit layout,
#' table width in the document, eventually column widths.
#' @param style table style to be used to format table
#' @param layout layout defined by [table_layout()],
#' @param width table width in the document defined by [table_width()]
#' @param stylenames columns styles defined by [table_stylenames()]
#' @param colwidths column widths defined by [table_colwidths()]
#' @param align table alignment (one of left, center or right)
#' @param tcf conditional formatting settings defined by [table_conditional_formatting()]
#' @param word_title alternative text for Word table (used as title of the table)
#' @param word_description alternative text for Word table (used as description of the table)
#' @examples
#' prop_table()
#' to_wml(prop_table())
#' @family functions for table definition
prop_table <- function(
  style = NA_character_,
  layout = table_layout(),
  width = table_width(),
  stylenames = table_stylenames(),
  colwidths = table_colwidths(),
  tcf = table_conditional_formatting(),
  align = "center",
  word_title = NULL,
  word_description = NULL
) {
  z <- list(
    style = style,
    layout = layout,
    width = width,
    colsizes = colwidths,
    stylenames = stylenames,
    tcf = tcf,
    align = align,
    word_title = word_title,
    word_description = word_description
  )
  class(z) <- c("prop_table")
  z
}

#' @export
to_wml.prop_table <- function(x, add_ns = FALSE, base_document = NULL, ...) {
  style <- NA_character_
  if (!is.null(x$style) && !is.na(x$style)) {
    if (is.null(x$style) && !is.null(base_document$default_styles$table)) {
      style <- base_document$default_styles$table
    } else {
      style <- x$style
    }
  }

  tbl_layout <- to_wml(x$layout, add_ns = add_ns)

  width <- ""
  if (!is.null(x$width) && "autofit" %in% x$layout$type) {
    width <- to_wml(x$width, add_ns = add_ns)
  }

  colwidths <- to_wml(x$colsizes, add_ns = add_ns)
  tcf <- to_wml(x$tcf, add_ns = add_ns)
  paste0(
    "<w:tblPr>",
    if (!is.null(x$word_title)) {
      paste0("<w:tblCaption w:val=\"", htmlEscapeCopy(x$word_title), "\"/>")
    },
    if (!is.null(x$word_description)) {
      paste0(
        "<w:tblDescription w:val=\"",
        htmlEscapeCopy(x$word_description),
        "\"/>"
      )
    },
    if (!is.na(style)) paste0("<w:tblStyle w:tstlname=\"", style, "\"/>"),
    tbl_layout,
    sprintf("<w:jc w:val=\"%s\"/>", x$align),
    width,
    tcf,
    "</w:tblPr>",
    if (x$layout$type %in% "fixed") colwidths
  )
}


# table ----
table_docx <- function(
  x,
  header,
  style_id,
  properties,
  alignment = NULL,
  add_ns = FALSE,
  base_document = base_document
) {
  open_tag <- tbl_ns_no
  if (add_ns) {
    open_tag <- tbl_ns_yes
  }

  str <- paste0(
    open_tag,
    to_wml(properties, add_ns = add_ns, base_document = base_document)
  )

  stylenames <- to_wml(
    properties$stylenames,
    base_document = base_document,
    dat = x
  )
  stylenames <- unlist(stylenames)
  stylenames <- as.character(stylenames)

  if (is.null(alignment)) {
    alignment <- rep("", ncol(x))
  } else {
    alignment <- match.arg(
      alignment,
      c("left", "right", "center"),
      several.ok = TRUE
    )
    if (length(alignment) < ncol(x)) {
      alignment <- rep(alignment, length.out = ncol(x))
    }
    alignment <- sprintf("<w:jc w:val=\"%s\"/>", alignment)
  }

  header_str <- character(length = 0L)
  if (header) {
    header_str <- paste0(
      "<w:tr><w:trPr><w:tblHeader/></w:trPr>",
      paste0(
        "<w:tc><w:p>",
        sprintf("<w:pPr>%s%s</w:pPr>", stylenames, alignment),
        "<w:r><w:t>",
        htmlEscapeCopy(enc2utf8(colnames(x))),
        "</w:t></w:r></w:p></w:tc>",
        collapse = ""
      ),
      "</w:tr>"
    )
  }
  as_tc <- function(x, align, stylenames) {
    paste0(
      "<w:tc><w:p>",
      sprintf("<w:pPr>%s%s</w:pPr>", stylenames, align),
      "<w:r><w:t>",
      htmlEscapeCopy(enc2utf8(x)),
      "</w:t></w:r></w:p></w:tc>"
    )
  }
  z <- mapply(as_tc, x, alignment, stylenames, SIMPLIFY = FALSE)
  z <- do.call(paste0, z)
  z <- paste0("<w:tr>", z, "</w:tr>", collapse = "")

  paste0(str, header_str, z, "</w:tbl>")
}

table_pptx <- function(
  x,
  style_id,
  col_width,
  row_height,
  tcf,
  header = TRUE,
  alignment = NULL
) {
  str <- paste0(
    "<a:tbl>",
    sprintf("<a:tblPr %s>", to_pml(tcf)),
    sprintf("<a:tableStyleId>%s</a:tableStyleId>", style_id),
    "</a:tblPr>",
    "<a:tblGrid>",
    paste0(
      sprintf("<a:gridCol w=\"%.0f\"/>", rep(col_width, length(x))),
      collapse = ""
    ),
    "</a:tblGrid>"
  )

  as_tc <- function(x, align) {
    paste0(
      "<a:tc><a:txBody><a:bodyPr/><a:lstStyle/><a:p>",
      "<a:pPr algn=\"",
      align,
      "\"/>",
      "<a:r><a:t>",
      htmlEscapeCopy(enc2utf8(x)),
      "</a:t></a:r></a:p></a:txBody></a:tc>"
    )
  }

  if (is.null(alignment)) {
    alignment <- rep("r", ncol(x))
  } else {
    alignment <- match.arg(alignment, c("l", "r", "ctr"), several.ok = TRUE)
  }

  header_str <- character(length = 0L)
  if (header) {
    header_str <- paste0(
      sprintf("<a:tr h=\"%.0f\">", row_height),
      paste0(as_tc(colnames(x), align = alignment), collapse = ""),
      "</a:tr>"
    )
  }

  z <- mapply(as_tc, x, alignment, SIMPLIFY = FALSE)
  z <- do.call(paste0, z)
  z <- paste0(
    sprintf("<a:tr h=\"%.0f\">", row_height),
    z,
    "</a:tr>",
    collapse = ""
  )

  z <- paste0(str, header_str, z, "</a:tbl>")
  z <- paste0(
    "<a:graphic>",
    "<a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/table\">",
    z,
    "</a:graphicData>",
    "</a:graphic>"
  )
}


#' @export
#' @title Table block
#' @description Create a representation of a table
#' @param x a data.frame to add as a table
#' @param header display header if TRUE
#' @param properties table properties, see [prop_table()].
#' Table properties are not handled identically between Word and PowerPoint
#' output format. They are fully supported with Word but for PowerPoint (which
#' does not handle as many things as Word for tables), only conditional
#' formatting properties are supported.
#' @param alignment alignment for each columns, 'l' for left, 'r' for right
#' and 'c' for center. Default to NULL.
#' @examples
#' block_table(x = head(iris))
#'
#' block_table(x = mtcars, header = TRUE,
#'   properties = prop_table(
#'     tcf = table_conditional_formatting(
#'       first_row = TRUE, first_column = TRUE)
#'   ))
#' @family block functions for reporting
#' @seealso [prop_table()]
block_table <- function(
  x,
  header = TRUE,
  properties = prop_table(),
  alignment = NULL
) {
  stopifnot(is.data.frame(x))
  if (inherits(x, "tbl_df")) {
    x <- as.data.frame(
      x,
      check.names = FALSE,
      stringsAsFactors = FALSE
    )
  }

  z <- list(
    x = x,
    header = header,
    properties = properties,
    alignment = alignment
  )
  class(z) <- c("block_table", "block")

  z
}

#' @export
#' @importFrom utils str
print.block_table <- function(x, ...) {
  str(x$x)
}

#' @export
to_wml.block_table <- function(x, add_ns = FALSE, base_document = NULL, ...) {
  value <- characterise_df(x$x)

  out <- table_docx(
    x = value,
    header = x$header,
    properties = x$properties,
    alignment = x$alignment,
    base_document = base_document,
    add_ns = add_ns
  )

  out
}

#' @export
to_pml.block_table <- function(
  x,
  add_ns = FALSE,
  left = 0,
  top = 0,
  width = 3,
  height = 3,
  bg = "transparent",
  rot = 0,
  label = "",
  ph = "<p:ph/>",
  ...
) {
  if (!is.null(bg) && !is.color(bg)) {
    stop("bg must be a valid color.", call. = FALSE)
  }

  bg_str <- solid_fill_pml(bg)

  xfrm_str <- p_xfrm_str(
    left = left,
    top = top,
    width = width,
    height = height,
    rot = rot
  )
  if (is.null(ph) || is.na(ph)) {
    ph = "<p:ph/>"
  }
  value <- characterise_df(x$x)
  value_str <- table_pptx(
    value,
    style_id = x$properties$style,
    alignment = x$alignment,
    col_width = as.integer((width / ncol(x$x)) * 914400),
    row_height = as.integer((height / nrow(x$x)) * 914400),
    tcf = x$properties$tcf,
    header = x$header
  )

  id <- uuid_generate()
  str <- paste0(
    "<p:graphicFrame xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
    "<p:nvGraphicFramePr>",
    sprintf("<p:cNvPr id=\"%s\" name=\"%s\"/>", id, label),
    "<p:cNvGraphicFramePr><a:graphicFrameLocks noGrp=\"1\"/></p:cNvGraphicFramePr>",
    sprintf("<p:nvPr>%s</p:nvPr>", ph),
    "</p:nvGraphicFramePr>",
    xfrm_str,
    bg_str,
    value_str,
    "</p:graphicFrame>"
  )
  str
}

# fpar ----

#' @export
#' @title Formatted paragraph
#' @description Create a paragraph representation by concatenating
#' formatted text or images. The result can be inserted in a Word document
#' or a PowerPoint presentation and can also be inserted in a [block_list()]
#' call.
#'
#' All its arguments will be concatenated to create a paragraph where chunks of
#' text and images are associated with formatting properties.
#'
#' `fpar()` supports [ftext()], [external_img()], `run_*()` functions
#' (i.e. [run_autonum()], [run_word_field()]) when output is Word, and simple strings.
#'
#' Default text and paragraph formatting properties can also be modified
#' with function `update()`.
#'
#' @param ... cot objects ([ftext()], [external_img()])
#' @param fp_p paragraph formatting properties, see [fp_par()]
#' @param fp_t default text formatting properties. This is used as
#' text formatting properties when simple text is provided as argument,
#' see [fp_text()].
#' @param values a list of cot objects. If provided, argument `...` will be ignored.
#' @param object fpar object
#' @examples
#' fpar(ftext("hello", shortcuts$fp_bold()))
#'
#' # mix text and image -----
#' img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
#'
#' bold_face <- shortcuts$fp_bold(font.size = 12)
#' bold_redface <- update(bold_face, color = "red")
#' fpar_1 <- fpar(
#'   "Hello World, ",
#'   ftext("how ", prop = bold_redface ),
#'   external_img(src = img.file, height = 1.06/2, width = 1.39/2),
#'   ftext(" you?", prop = bold_face ) )
#' fpar_1
#'
#' img_in_par <- fpar(
#'   external_img(src = img.file, height = 1.06/2, width = 1.39/2),
#'   fp_p = fp_par(text.align = "center") )
#' @family block functions for reporting
#' @seealso [block_list()], [body_add_fpar()], [ph_with()]
fpar <- function(
  ...,
  fp_p = fp_par(word_style = NA_character_),
  fp_t = fp_text_lite(),
  values = NULL
) {
  out <- list()

  if (is.null(values)) {
    values <- list(...)
  }

  out$chunks <- values
  out$fp_p <- fp_p
  out$fp_t <- fp_t
  class(out) <- c("fpar", "block")
  out
}

#' @export
#' @rdname fpar
#' @importFrom stats update
update.fpar <- function(object, fp_p = NULL, fp_t = NULL, ...) {
  if (!is.null(fp_p)) {
    object$fp_p <- fp_p
  }
  if (!is.null(fp_t)) {
    object$fp_t <- fp_t
  }

  object
}


fortify_fpar <- function(x) {
  lapply(x$chunks, function(chk) {
    if (!inherits(chk, c("cot", "run"))) {
      chk <- ftext(text = format(chk), prop = x$fp_t)
    }
    chk
  })
}


#' @export
as.data.frame.fpar <- function(x, ...) {
  chks <- fortify_fpar(x)
  chks <- chks[sapply(chks, function(x) inherits(x, "ftext"))]
  chks <- mapply(
    function(x) {
      data.frame(
        value = x$value,
        size = x$pr$font.size,
        bold = x$pr$bold,
        italic = x$pr$italic,
        font.family = x$pr$font.family,
        stringsAsFactors = FALSE
      )
    },
    chks,
    SIMPLIFY = FALSE
  )
  rbind_match_columns(chks)
}


#' @export
to_wml.fpar <- function(x, add_ns = FALSE, style_id = NULL, ...) {
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }
  if (is.null(style_id)) {
    par_style <- ppr_wml(x$fp_p)
  } else {
    par_style <- paste0(
      "<w:pPr><w:pStyle w:val=\"",
      style_id,
      "\"/></w:pPr>"
    )
  }

  chks <- fortify_fpar(x)
  z <- lapply(chks, to_wml)
  z$collapse <- ""
  z <- do.call(paste0, z)
  paste0(open_tag, par_style, z, "</w:p>")
}

#' @export
to_pml.fpar <- function(x, add_ns = FALSE, ...) {
  open_tag <- ap_ns_no
  if (add_ns) {
    open_tag <- ap_ns_yes
  }

  par_style <- ppr_pml(x$fp_p)
  chks <- fortify_fpar(x)
  z <- lapply(chks, to_pml)
  z$collapse <- ""
  z <- do.call(paste0, z)
  paste0(open_tag, par_style, z, "</a:p>")
}


#' @export
to_html.fpar <- function(x, add_ns = FALSE, ...) {
  par_style <- ppr_css(x$fp_p)
  chks <- fortify_fpar(x)
  z <- lapply(chks, to_html)
  z$collapse <- ""
  z <- do.call(paste0, z)
  sprintf("<p style=\"%s\">%s</p>", par_style, z)
}

# block_list -----

#' @export
#' @title List of blocks
#' @description A list of blocks can be used to gather
#' several blocks (paragraphs, tables, ...) into a single
#' object. The result can be added into a Word document or a
#' PowerPoint presentation.
#' @param ... a list of blocks. When output is only for
#' Word, objects of class [external_img()] can
#' also be used in fpar construction to mix text and images
#' in a single paragraph. Supported objects are:
#' [block_caption()], [block_pour_docx()], [block_section()],
#' [block_table()], [block_toc()], [fpar()], [plot_instr()].
#' @examples
#' # block list ------
#'
#' img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
#' fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
#' fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)
#'
#'
#' ## This can be only be used in a MS word output as pptx does
#' ## not support paragraphs made of text and images.
#' ## (actually it can be used but image will not appear in the
#' ## pptx output)
#' value <- block_list(
#'   fpar(ftext("hello world", fpt_blue_bold)),
#'   fpar(ftext("hello", fpt_blue_bold), " ",
#'        ftext("world", fpt_red_italic)),
#'   fpar(
#'     ftext("hello world", fpt_red_italic),
#'           external_img(
#'             src = img.file, height = 1.06, width = 1.39)))
#' value
#'
#' doc <- read_docx()
#' doc <- body_add(doc, value)
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#'
#' value <- block_list(
#'   fpar(ftext("hello world", fpt_blue_bold)),
#'   fpar(ftext("hello", fpt_blue_bold), " ",
#'        ftext("world", fpt_red_italic)),
#'   fpar(
#'     ftext("blah blah blah", fpt_red_italic)))
#' value
#'
#' doc <- read_pptx()
#' doc <- add_slide(doc, "Title and Content")
#' doc <- ph_with(doc, value, location = ph_location_type(type = "body"))
#' print(doc, target = tempfile(fileext = ".pptx"))
#' @seealso [ph_with()], [body_add_blocks()], [fpar()]
#' @family block functions for reporting
block_list <- function(...) {
  x <- list(...)
  z <- list()
  for (i in x) {
    if (inherits(i, "block")) {
      z <- append(z, list(i))
    } else if (inherits(i, "flextable")) {
      z <- append(z, list(i))
    } else if (is.character(i)) {
      z <- append(z, lapply(i, fpar))
    }
  }
  class(z) <- c("block_list", "block")
  z
}

#' @export
to_wml.block_list <- function(x, add_ns = FALSE, ...) {
  out <- character(length(x))
  for (i in seq_along(x)) {
    out[i] <- to_wml(x[[i]], add_ns = add_ns)
  }

  paste0(out, collapse = "")
}

#' @export
to_pml.block_list <- function(x, add_ns = FALSE, ...) {
  pars <- sapply(x, to_pml)
  pars <- paste0(pars, collapse = "")
  pars
}

#' @export
to_html.block_list <- function(x, add_ns = FALSE, ...) {
  str <- vapply(x, to_html, NA_character_)
  paste0(str, collapse = "")
}

# unordered list ----
#' @export
#' @title Unordered list
#' @description unordered list of text for PowerPoint
#' presentations. Each text is associated with
#' a hierarchy level.
#' @param str_list list of strings to be included in the object
#' @param level_list list of levels for hierarchy structure. Use
#' 0 for 'no bullet', 1 for level 1, 2 for level 2 and so on.
#' @param style text style, a `fp_text` object list or a
#' single `fp_text` objects. Use `fp_text(font.size = 0, ...)` to
#' inherit from default sizes of the presentation.
#' @examples
#' unordered_list(
#'   level_list = c(1, 2, 2, 3, 3, 1),
#'   str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
#'   style = fp_text(color = "red", font.size = 0)
#' )
#'
#' unordered_list(
#'   level_list = c(1, 2, 1),
#'   str_list = c("Level1", "Level2", "Level1"),
#'   style = list(
#'     fp_text(color = "red", font.size = 0),
#'     fp_text(color = "pink", font.size = 0),
#'     fp_text(color = "orange", font.size = 0)
#'   )
#' )
#' @seealso [ph_with()]
#' @family block functions for reporting
unordered_list <- function(
  str_list = character(0),
  level_list = integer(0),
  style = NULL
) {
  stopifnot(is.character(str_list))
  stopifnot(is.numeric(level_list))

  if (length(str_list) != length(level_list) & length(str_list) > 0) {
    stop("str_list and level_list have different lenghts.")
  }

  if (!is.null(style)) {
    if (inherits(style, "fp_text")) {
      style <- lapply(seq_len(length(str_list)), function(x) style)
    }
  }
  x <- list(
    str = str_list,
    lvl = level_list,
    style = style
  )
  class(x) <- "unordered_list"
  x
}
#' @export
#' @noRd
print.unordered_list <- function(x, ...) {
  print(data.frame(str = x$str, lvl = x$lvl, stringsAsFactors = FALSE))
  invisible()
}

#' @export
to_pml.unordered_list <- function(x, add_ns = FALSE, ...) {
  open_tag <- ap_ns_no
  if (add_ns) {
    open_tag <- ap_ns_yes
  }
  if (!is.null(x$style)) {
    style_str <- sapply(x$style, format, type = "pml")
    style_str <- rep_len(style_str, length.out = length(x$str))
  } else {
    style_str <- rep("<a:rPr/>", length(x$str))
  }
  tmpl <- "%s<a:pPr%s>%s</a:pPr><a:r>%s<a:t>%s</a:t></a:r></a:p>"
  lvl <- sprintf(" lvl=\"%.0f\"", x$lvl - 1)
  lvl <- ifelse(x$lvl > 1, lvl, "")
  bu_none <- ifelse(x$lvl < 1, "<a:buNone/>", "")
  p <- sprintf(tmpl, open_tag, lvl, bu_none, style_str, htmlEscapeCopy(x$str))
  p <- paste(p, collapse = "")
  p
}

# plot_instr -----
#' @title Wrap plot instructions for png plotting in Powerpoint or Word
#' @description A simple wrapper to capture
#' plot instructions that will be executed and copied in a document. It produces
#' an object of class 'plot_instr' with a corresponding method [ph_with()] and
#' [body_add_plot()].
#'
#' The function enable usage of any R plot with argument `code`. Wrap your code
#' between curly bracket if more than a single expression.
#'
#' @param code plotting instructions
#' @examples
#' # plot_instr demo ----
#'
#' anyplot <- plot_instr(code = {
#'   barplot(1:5, col = 2:6)
#'   })
#'
#' doc <- read_docx()
#' doc <- body_add(doc, anyplot, width = 5, height = 4)
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#'
#' doc <- read_pptx()
#' doc <- add_slide(doc, "Title and Content")
#' doc <- ph_with(
#'   doc, anyplot,
#'   location = ph_location_fullsize(),
#'   bg = "#00000066", pointsize = 12)
#' print(doc, target = tempfile(fileext = ".pptx"))
#' @export
#' @import graphics
#' @seealso [ph_with()], [body_add_plot()]
#' @family block functions for reporting
plot_instr <- function(code) {
  out <- list()
  out$code <- substitute(code)
  class(out) <- "plot_instr"
  return(out)
}

# ggplot -----

#' @export
#' @title 'ggplot' block
#' @description A simple wrapper to add a 'ggplot' object as a png in a document.
#' It produces an object of class 'block_gg' with a corresponding method [to_wml()]
#' that can be used to convert the object to a WordML string.
#' @param value 'ggplot' object
#' @param fp_p paragraph formatting properties, see [fp_par()]
#' @param width,height image size in units expressed by the unit argument.
#' Defaults to "in"ches.
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
#' @param unit One of the following units in which the width and height
#' arguments are expressed: "in", "cm" or "mm".
#' @family block functions for reporting
#' @examples
#' library(officer)
#' if (require("ggplot2")) {
#'   set.seed(2)
#'   doc <- read_docx()
#'
#'   z <- body_append_start_context(doc)
#'
#'   for (i in seq_len(3)) {
#'     df <- data.frame(x = runif(10), y = runif(10))
#'     gg <- ggplot(df, aes(x = x, y = y)) + geom_point()
#'
#'     write_elements_to_context(
#'       context = z,
#'       block_gg(
#'         value = gg
#'       )
#'     )
#'   }
#'
#'   doc <- body_append_stop_context(z)
#'
#'   print(doc, target = tempfile(fileext = ".docx"))
#' }
block_gg <- function(
  value,
  fp_p = fp_par(),
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  unit = "in"
) {
  if (!inherits(value, "gg")) {
    cli::cli_abort("Argument value is not a ggplot object.")
  }

  unit <- check_unit(unit, c("in", "cm", "mm"))

  out <- list(
    gg = value,
    fp_p = fp_p,
    width = width,
    height = height,
    res = res,
    scale = scale,
    unit = unit
  )

  class(out) <- c("block_gg", "block")
  return(out)
}

#' @export
to_wml.block_gg <- function(x, add_ns = FALSE, ...) {
  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = x$width,
    height = x$height,
    units = x$unit,
    res = x$res,
    scaling = x$scale,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      print(x$gg)
    },
    finally = {
      dev.off()
    }
  )
  fp <- fpar(
    external_img(
      src = file,
      width = x$width,
      height = x$height,
      unit = x$unit
    ),
    fp_p = x$fp_p
  )
  to_wml(fp, add_ns = add_ns)
}

# check properties helpers ----

check_spread_integer <- function(obj, value, dest) {
  varname <- as.character(substitute(value))
  if (is.numeric(value) && length(value) == 1 && value >= 0) {
    for (i in dest) {
      obj[[i]] <- as.integer(value)
    }
  } else if (length(value) == 1 && is.na(value)) {
    for (i in dest) {
      obj[[varname]] <- value
    }
  } else {
    stop(varname, " must be a positive integer scalar.", call. = FALSE)
  }
  obj
}

check_set_color <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (!is.color(value) && !is.na(value)) {
    stop(varname, " must be a valid color.", call. = FALSE)
  } else {
    obj[[varname]] <- value
  }
  obj
}


check_set_border <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (!inherits(value, "fp_border")) {
    stop(varname, " must be a fp_border object.", call. = FALSE)
  } else {
    obj[[varname]] <- value
  }
  obj
}

check_spread_border <- function(obj, value, dest) {
  varname <- as.character(substitute(value))
  if (!inherits(value, "fp_border")) {
    stop(varname, " must be a fp_border object.", call. = FALSE)
  }
  for (i in dest) {
    obj[[i]] <- value
  }
  obj
}

check_set_integer <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.numeric(value) && length(value) == 1 && value >= 0)) {
    obj[[varname]] <- as.integer(value)
  } else {
    stop(varname, " must be a positive integer scalar.", call. = FALSE)
  }
  obj
}

check_set_numeric <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.numeric(value) && length(value) == 1 && value >= 0)) {
    obj[[varname]] <- as.double(value)
  } else {
    stop(varname, " must be a positive numeric scalar.", call. = FALSE)
  }
  obj
}


check_set_bool <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.logical(value) && length(value) == 1)) {
    obj[[varname]] <- value
  } else {
    stop(varname, " must be a boolean", call. = FALSE)
  }
  obj
}
check_set_chr <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.character(value) && length(value) == 1)) {
    obj[[varname]] <- value
  } else {
    stop(varname, " must be a string", call. = FALSE)
  }
  obj
}

check_set_choice <- function(obj, value, choices) {
  varname <- as.character(substitute(value))

  if (is.na(value)) {
    obj[[varname]] <- value
  } else {
    if (is.character(value) && length(value) == 1) {
      if (!value %in% choices) {
        stop(
          varname,
          " must be one of ",
          paste(shQuote(choices), collapse = ", "),
          call. = FALSE
        )
      }
      obj[[varname]] <- value
    } else {
      stop(varname, " must be a character scalar.", call. = FALSE)
    }
  }

  obj
}

check_set_class <- function(obj, value, cl) {
  varname <- as.character(substitute(value))
  cl_str <- sprintf(" must be a %s object.", cl)
  if (!inherits(value, cl)) {
    stop(varname, cl_str, call. = FALSE)
  } else {
    obj[[varname]] <- value
  }
  obj
}

default_rpr <- data.frame(
  stringsAsFactors = FALSE,
  font.size = NA_integer_,
  bold = as.logical(NA),
  italic = as.logical(NA),
  underlined = as.logical(NA),
  strike = as.logical(NA),
  color = NA_character_,
  font.family = NA_character_,
  bold.cs = as.logical(NA),
  font.size.cs = NA_integer_,
  vertical.align = NA_character_,
  shading.color = NA_character_,
  hansi.family = NA_character_,
  eastasia.family = NA_character_,
  cs.family = NA_character_,
  lang.val = NA_character_,
  lang.eastasia = NA_character_,
  lang.bidi = NA_character_
)

is_positive_numeric <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || (is.numeric(x) && length(x) == 1 && x >= 0)
  if (!test) {
    stop(varname, " must be a positive numeric scalar.", call. = FALSE)
  }
  test
}
is_bool <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || (is.logical(x) && length(x) == 1)
  if (!test) {
    stop(varname, " must be a boolean.", call. = FALSE)
  }
  test
}
is_color <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || is.color(x)
  if (!test) {
    stop(varname, " must be a valid color.", call. = FALSE)
  }
  test
}
is_character <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || (is.character(x) && length(x) == 1)
  if (!test) {
    stop(varname, " must be a string", call. = FALSE)
  }
  test
}
# fp_text ----
#' @title Text formatting properties
#' @description Create an `fp_text` object that describes
#' text formatting properties.
#'
#' @param color font color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param font.size font size (in point) - 0 or positive integer value.
#' @param bold is bold
#' @param italic is italic
#' @param underlined is underlined
#' @param strike is strikethrough
#' @param font.family single character value. Specifies the font to
#' be used to format characters in the Unicode range (U+0000-U+007F).
#' @param cs.family optional font to be used to format
#' characters in a complex script Unicode range. For example, Arabic
#' text might be displayed using the "Arial Unicode MS" font.
#' @param eastasia.family optional font to be used to
#' format characters in an East Asian Unicode range. For example,
#' Japanese text might be displayed using the "MS Mincho" font.
#' @param hansi.family optional. Specifies the font to be used to format
#' characters in a Unicode range which does not fall into one of the
#' other categories.
#' @param vertical.align single character value specifying font vertical alignments.
#' Expected value is one of the following : default `'baseline'`
#' or `'subscript'` or `'superscript'`
#' @param shading.color shading color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @return a `fp_text` object
#' @examples
#' fp_text()
#' fp_text(color = "red")
#' fp_text(bold = TRUE, shading.color = "yellow")
#' @family functions for defining formatting properties
#' @seealso [ftext()], [fpar()]
#' @export
fp_text <- function(
  color = "black",
  font.size = 10,
  bold = FALSE,
  italic = FALSE,
  underlined = FALSE,
  strike = FALSE,
  font.family = "Arial",
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  vertical.align = "baseline",
  shading.color = "transparent"
) {
  out <- default_rpr
  if (is_positive_numeric(font.size)) {
    out$font.size <- font.size
    out$font.size.cs <- font.size
  }
  if (is_bool(bold)) {
    out$bold <- bold
    out$bold.cs <- bold
  }
  if (is_bool(italic)) {
    out$italic <- italic
  }
  if (is_bool(underlined)) {
    out$underlined <- underlined
  }
  if (is_bool(strike)) {
    out$strike <- strike
  }
  if (is_color(color)) {
    out$color <- color
  }
  if (is_character(font.family)) {
    out$font.family <- font.family
  }

  if (is.null(cs.family)) {
    cs.family <- font.family
  }
  if (is.null(eastasia.family)) {
    eastasia.family <- font.family
  }
  if (is.null(hansi.family)) {
    hansi.family <- font.family
  }
  if (is_character(cs.family)) {
    out$cs.family <- cs.family
  }
  if (is_character(eastasia.family)) {
    out$eastasia.family <- eastasia.family
  }
  if (is_character(hansi.family)) {
    out$hansi.family <- hansi.family
  }

  out <- check_set_choice(
    obj = out,
    value = vertical.align,
    choices = c("subscript", "superscript", "baseline")
  )
  out <- check_set_color(out, shading.color)

  class(out) <- "fp_text"

  out
}


#' @rdname fp_text
#' @description Function `fp_text_lite()` is generating properties
#' with only entries for the parameters users provided. The
#' undefined properties will inherit from the default settings.
#' @export
fp_text_lite <- function(
  color = NA,
  font.size = NA,
  font.family = NA,
  cs.family = NA,
  eastasia.family = NA,
  hansi.family = NA,
  bold = NA,
  italic = NA,
  underlined = NA,
  strike = NA,
  vertical.align = "baseline",
  shading.color = NA
) {
  fp_text(
    color = color,
    font.size = font.size,
    bold = bold,
    italic = italic,
    underlined = underlined,
    strike = strike,
    font.family = font.family,
    cs.family = cs.family,
    eastasia.family = eastasia.family,
    hansi.family = hansi.family,
    vertical.align = vertical.align,
    shading.color = shading.color
  )
}

#' @rdname fp_text
#' @param format format type, wml for MS word, pml for
#' MS PowerPoint and html.
#' @param type output type - one of 'wml', 'pml', 'html', 'rtf'.
#' @export
format.fp_text <- function(x, type = "wml", ...) {
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("wml", "pml", "html", "rtf"))

  if (type == "wml") {
    rpr_wml(x)
  } else if (type == "pml") {
    rpr_pml(x)
  } else if (type == "html") {
    rpr_css(x)
  } else if (type == "rtf") {
    rpr_rtf(x)
  } else {
    stop("unimplemented type")
  }
}

#' @export
to_wml.fp_text <- function(x, add_ns = FALSE, ...) {
  format(x, type = "wml")
}

#' @param x `fp_text` object
#' @examples
#' print(fp_text(color = "red", font.size = 12))
#' @rdname fp_text
#' @export
print.fp_text <- function(x, ...) {
  out <- data.frame(
    font.size = as.double(x$font.size),
    italic = x$italic,
    bold = x$bold,
    underlined = x$underlined,
    strike = x$strike,
    color = x$color,
    shading = x$shading.color,
    fontname = x$font.family,
    fontname_cs = x$cs.family,
    fontname_eastasia = x$eastasia.family,
    fontname.hansi = x$hansi.family,
    vertical_align = x$vertical.align,
    stringsAsFactors = FALSE
  )
  print(out)
  invisible()
}


#' @param object `fp_text` object to modify
#' @param ... further arguments - not used
#' @rdname fp_text
#' @export
update.fp_text <- function(
  object,
  color,
  font.size,
  bold,
  italic,
  underlined,
  strike,
  font.family,
  cs.family,
  eastasia.family,
  hansi.family,
  vertical.align,
  shading.color,
  ...
) {
  if (!missing(font.size)) {
    object <- check_set_numeric(obj = object, font.size)
  }
  if (!missing(bold)) {
    object <- check_set_bool(obj = object, bold)
  }
  if (!missing(italic)) {
    object <- check_set_bool(obj = object, italic)
  }
  if (!missing(underlined)) {
    object <- check_set_bool(obj = object, underlined)
  }
  if (!missing(strike)) {
    object <- check_set_bool(obj = object, strike)
  }
  if (!missing(color)) {
    object <- check_set_color(object, color)
  }
  if (!missing(font.family)) {
    object <- check_set_chr(object, font.family)
  }
  if (!missing(cs.family)) {
    object <- check_set_chr(object, cs.family)
  }
  if (!missing(eastasia.family)) {
    object <- check_set_chr(object, eastasia.family)
  }
  if (!missing(hansi.family)) {
    object <- check_set_chr(object, hansi.family)
  }
  if (!missing(vertical.align)) {
    object <- check_set_choice(
      obj = object,
      value = vertical.align,
      choices = c("subscript", "superscript", "baseline")
    )
  }
  if (!missing(shading.color)) {
    object <- check_set_color(object, shading.color)
  }

  object
}

# fp_border ----
border_styles <- c(
  "nil",
  "none",
  "solid",
  "single",
  "thick",
  "double",
  "dotted",
  "dashed",
  "dotDash",
  "dotDotDash",
  "triple",
  "thinThickSmallGap",
  "thickThinSmallGap",
  "thinThickThinSmallGap",
  "thinThickMediumGap",
  "thickThinMediumGap",
  "thinThickThinMediumGap",
  "thinThickLargeGap",
  "thickThinLargeGap",
  "thinThickThinLargeGap",
  "wave",
  "doubleWave",
  "dashSmallGap",
  "dashDotStroked",
  "threeDEmboss",
  "ridge",
  "threeDEngrave",
  "groove",
  "outset",
  "inset"
)

#' @title Border properties object
#'
#' @description create a border properties object.
#'
#' @param color border color - single character value (e.g. "#000000" or "black")
#' @param style border style - single character value : See Details for supported
#'  border styles.
#' @param width border width - an integer value : 0>= value
#' @details For Word output the following border styles are supported:
#'
#' * "none" or "nil" - No Border
#' * "solid" or "single" - Single Line Border
#' * "thick" - Single Line Border
#' * "double" - Double Line Border
#' * "dotted" - Dotted Line Border
#' * "dashed" - Dashed Line Border
#' * "dotDash" - Dot Dash Line Border
#' * "dotDotDash" - Dot Dot Dash Line Border
#' * "triple" - Triple Line Border
#' * "thinThickSmallGap" - Thin, Thick Line Border
#' * "thickThinSmallGap" - Thick, Thin Line Border
#' * "thinThickThinSmallGap" - Thin, Thick, Thin Line Border
#' * "thinThickMediumGap" - Thin, Thick Line Border
#' * "thickThinMediumGap" - Thick, Thin Line Border
#' * "thinThickThinMediumGap" - Thin, Thick, Thin Line Border
#' * "thinThickLargeGap" - Thin, Thick Line Border
#' * "thickThinLargeGap" - Thick, Thin Line Border
#' * "thinThickThinLargeGap" - Thin, Thick, Thin Line Border
#' * "wave" - Wavy Line Border
#' * "doubleWave" - Double Wave Line Border
#' * "dashSmallGap" - Dashed Line Border
#' * "dashDotStroked" - Dash Dot Strokes Line Border
#' * "threeDEmboss" or "ridge" - 3D Embossed Line Border
#' * "threeDEngrave" or "groove" - 3D Engraved Line Border
#' * "outset" - Outset Line Border
#' * "inset" - Inset Line Border
#'
#' For HTML output only a limited amount of border styles are supported:
#'
#' * "none" or "nil" - No Border
#' * "solid" or "single" - Single Line Border
#' * "double" - Double Line Border
#' * "dotted" - Dotted Line Border
#' * "dashed" - Dashed Line Border
#' * "threeDEmboss" or "ridge" - 3D Embossed Line Border
#' * "threeDEngrave" or "groove" - 3D Engraved Line Border
#' * "outset" - Outset Line Border
#' * "inset" - Inset Line Border
#'
#' Non-supported Word border styles will default to "solid".
#' @examples
#' fp_border()
#' fp_border(color = "orange", style = "solid", width = 1)
#' fp_border(color = "gray", style = "dotted", width = 1)
#' @export
#' @family functions for defining formatting properties
fp_border <- function(color = "black", style = "solid", width = 1) {
  out <- list()
  out <- check_set_numeric(obj = out, width)
  out <- check_set_color(out, color)
  out <- check_set_choice(
    obj = out,
    style,
    choices = border_styles
  )

  class(out) <- "fp_border"
  out
}

#' @param object fp_border object
#' @param ... further arguments - not used
#' @rdname fp_border
#' @examples
#'
#' # modify object ------
#' border <- fp_border()
#' update(border, style = "dotted", width = 3)
#' @export
update.fp_border <- function(object, color, style, width, ...) {
  if (!missing(color)) {
    object <- check_set_color(object, color)
  }

  if (!missing(width)) {
    object <- check_set_integer(obj = object, width)
  }

  if (!missing(style)) {
    object <- check_set_choice(obj = object, style, choices = border_styles)
  }

  object
}

#' @export
print.fp_border <- function(x, ...) {
  msg <- paste0(
    "line: color: ",
    x$color,
    ", width: ",
    x$width,
    ", style: ",
    x$style,
    "\n"
  )
  cat(msg)
  invisible()
}

# tabs ----
#' @export
#' @title Tabulation mark properties object
#'
#' @description create a tabulation mark properties setting object for Word
#' or RTF. Results can be used as arguments of [fp_tabs()].
#'
#' Once tabulation marks settings are defined, tabulation marks can
#' be added with [run_tab()] inside a call to [fpar()] or
#' with `\t` within 'flextable' content.
#'
#' @param pos Specifies the position of the tab stop (in inches).
#' @param style style of the tab. Possible values are:
#' "decimal", "left", "right" or "center".
#' @examples
#' fp_tab(pos = 0.4, style = "decimal")
#' fp_tab(pos = 1, style = "right")
#' @family functions for defining formatting properties
fp_tab <- function(pos, style = "decimal") {
  out <- list()
  out <- check_set_numeric(obj = out, pos)
  out <- check_set_choice(
    obj = out,
    style,
    choices = c(
      "decimal",
      "left",
      "right",
      "center"
    )
  )

  class(out) <- "fp_tab"
  out
}

#' @export
to_wml.fp_tab <- function(x, add_ns = FALSE, ...) {
  sprintf(
    "<w:tab w:val=\"%s\" w:leader=\"none\" w:pos=\"%.0f\"/>",
    x$style,
    inch_to_tweep(x$pos)
  )
}
#' @export
as.character.fp_tab <- function(x, ...) {
  paste(x$style, x$pos, sep = "_")
}


#' @export
#' @title Tabs properties object
#'
#' @description create a set of tabulation mark properties object for Word or RTF.
#' Results can be used as arguments `tabs` of [fp_par()] and will only have
#' effects in Word or RTF outputs.
#'
#' Once a set of tabulation marks settings is defined, tabulation marks can
#' be added with [run_tab()] inside a call to [fpar()] or
#' with `\t` within 'flextable' content.
#' @param ... [fp_tab] objects
#' @examples
#' z <- fp_tabs(
#'   fp_tab(pos = 0.4, style = "decimal"),
#'   fp_tab(pos = 1, style = "decimal")
#' )
#' fpar(
#'   run_tab(), ftext("88."),
#'   run_tab(), ftext("987.45"),
#'   fp_p = fp_par(
#'     tabs = z
#'   )
#' )
#' @family functions for defining formatting properties
fp_tabs <- function(...) {
  out <- list(...)
  if (length(out) > 0) {
    are_fp_tab <- sapply(out, inherits, "fp_tab")
    if (!all(are_fp_tab)) {
      stop("Function `fp_tabs()` only accept `fp_tab` objects as arguments.")
    }
  }

  class(out) <- "fp_tabs"
  out
}

#' @export
to_wml.fp_tabs <- function(x, add_ns = FALSE, ...) {
  if (length(x) > 0) {
    paste0("<w:tabs>", do.call(paste0, lapply(x, to_wml)), "</w:tabs>")
  } else {
    ""
  }
}

#' @export
as.character.fp_tabs <- function(x, ...) {
  if (length(x) > 0) {
    z <- lapply(x, as.character)
    z$sep <- "&"
    do.call(paste, z)
  } else {
    NA_character_
  }
}

# fp_par -----
#' @title Paragraph formatting properties
#'
#' @description Create a `fp_par` object that describes
#' paragraph formatting properties.
#'
#' @param text.align text alignment - a single character value, expected value
#' is one of 'left', 'right', 'center', 'justify'.
#' @param padding.bottom,padding.top,padding.left,padding.right paragraph paddings - 0 or positive integer value.
#' @param padding paragraph paddings - 0 or positive integer value. Argument `padding` overwrites
#' arguments `padding.bottom`, `padding.top`, `padding.left`, `padding.right`.
#' @param line_spacing line spacing, 1 is single line spacing, 2 is double line spacing.
#' @param border shortcut for all borders.
#' @param border.bottom,border.left,border.top,border.right [fp_border()] for
#' borders. overwrite other border properties.
#' @param shading.color shading color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param keep_with_next a scalar logical. Specifies that the paragraph (or at least part of it) should be rendered
#' on the same page as the next paragraph when possible.
#' @param tabs NULL (default) for no tabulation marks setting
#' or an object returned by [fp_tabs()]. Note this can only have effect with Word
#' or RTF outputs.
#' @param word_style Word paragraph style name
#' @return a `fp_par` object
#' @examples
#' fp_par(text.align = "center", padding = 5)
#' @export
#' @family functions for defining formatting properties
#' @seealso [fpar]
fp_par <- function(
  text.align = "left",
  padding = 0,
  line_spacing = 1,
  border = fp_border(width = 0),
  padding.bottom,
  padding.top,
  padding.left,
  padding.right,
  border.bottom,
  border.left,
  border.top,
  border.right,
  shading.color = "transparent",
  keep_with_next = FALSE,
  tabs = NULL,
  word_style = "Normal"
) {
  out <- list()

  out <- check_set_color(out, shading.color)
  out <- check_set_choice(
    obj = out,
    value = text.align,
    choices = c("left", "right", "center", "justify")
  )
  # padding checking
  out <- check_spread_integer(
    out,
    padding,
    c(
      "padding.bottom",
      "padding.top",
      "padding.left",
      "padding.right"
    )
  )
  if (!missing(padding.bottom)) {
    out <- check_set_numeric(obj = out, padding.bottom)
  }
  if (!missing(padding.left)) {
    out <- check_set_numeric(obj = out, padding.left)
  }
  if (!missing(padding.top)) {
    out <- check_set_numeric(obj = out, padding.top)
  }
  if (!missing(padding.right)) {
    out <- check_set_numeric(obj = out, padding.right)
  }

  out <- check_set_numeric(obj = out, line_spacing)

  # border checking
  if (!is.null(border) && !isFALSE(border)) {
    out <- check_spread_border(
      obj = out,
      border,
      dest = c(
        "border.bottom",
        "border.top",
        "border.left",
        "border.right"
      )
    )
  }

  if (!missing(border.top) && !isFALSE(border.top)) {
    out <- check_set_border(obj = out, border.top)
  }
  if (!missing(border.bottom) && !isFALSE(border.bottom)) {
    out <- check_set_border(obj = out, border.bottom)
  }
  if (!missing(border.left) && !isFALSE(border.left)) {
    out <- check_set_border(obj = out, border.left)
  }
  if (!missing(border.right) && !isFALSE(border.right)) {
    out <- check_set_border(obj = out, border.right)
  }

  if (!is.null(tabs)) {
    out[["tabs"]] <- tabs
  }

  out <- check_set_chr(obj = out, word_style)

  out$keep_with_next <- keep_with_next
  class(out) <- "fp_par"

  out
}


#' @rdname fp_par
#' @description Function `fp_par_lite()` is generating properties
#' with only entries for the parameters users provided. The
#' undefined properties will inherit from the default settings.
#' @export
fp_par_lite <- function(
  text.align = NA,
  padding = NA,
  line_spacing = NA,
  border = FALSE,
  padding.bottom = NA,
  padding.top = NA,
  padding.left = NA,
  padding.right = NA,
  border.bottom = FALSE,
  border.left = FALSE,
  border.top = FALSE,
  border.right = FALSE,
  shading.color = NA,
  keep_with_next = NA,
  tabs = FALSE,
  word_style = NA
) {
  if (isFALSE(tabs)) {
    tabs <- NULL
  }
  fp_par(
    text.align = text.align,
    padding = padding,
    line_spacing = line_spacing,
    border = border,
    padding.bottom = padding.bottom,
    padding.top = padding.top,
    padding.left = padding.left,
    padding.right = padding.right,
    border.bottom = border.bottom,
    border.left = border.left,
    border.top = border.top,
    border.right = border.right,
    shading.color = shading.color,
    keep_with_next = keep_with_next,
    tabs = tabs,
    word_style = word_style
  )
}


#' @export
#' @importFrom grDevices col2rgb
format.fp_par <- function(x, type = "wml", ...) {
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("wml", "pml", "html", "rtf"))

  if (type == "wml") {
    ppr_wml(x)
  } else if (type == "pml") {
    ppr_pml(x)
  } else if (type == "html") {
    ppr_css(x)
  } else if (type == "rtf") {
    ppr_rtf(x)
  } else {
    stop("unimplemented")
  }
}
#' @export
to_wml.fp_par <- function(x, add_ns = FALSE, ...) {
  format(x, type = "wml")
}

#' @param x,object `fp_par` object
#' @param ... further arguments - not used
#' @rdname fp_par
#' @export
print.fp_par <- function(x, ...) {
  out <- data.frame(
    text.align = as.character(x$text.align),
    padding.top = as.character(x$padding.top),
    padding.bottom = as.character(x$padding.bottom),
    padding.left = as.character(x$padding.left),
    padding.right = as.character(x$padding.right),
    shading.color = as.character(x$shading.color)
  )
  out <- as.data.frame(t(out))
  names(out) <- "values"
  print(out)
  if (
    !is.null(x$border.top) &&
      !isFALSE(x$border.top) &&
      !is.null(x$border.bottom) &&
      !isFALSE(x$border.bottom) &&
      !is.null(x$border.left) &&
      !isFALSE(x$border.left) &&
      !is.null(x$border.right) &&
      !isFALSE(x$border.right)
  ) {
    cat("borders:\n")
    borders <- rbind(
      as.data.frame(unclass(x$border.top)),
      as.data.frame(unclass(x$border.bottom)),
      as.data.frame(unclass(x$border.left)),
      as.data.frame(unclass(x$border.right))
    )
    row.names(borders) <- c("top", "bottom", "left", "right")
    print(borders)
  } else {
    cat("no borders!\n")
  }
}


#' @rdname fp_par
#' @examples
#' obj <- fp_par(text.align = "center", padding = 1)
#' update(obj, padding.bottom = 5)
#' @export
update.fp_par <- function(
  object,
  text.align,
  padding,
  border,
  padding.bottom,
  padding.top,
  padding.left,
  padding.right,
  border.bottom,
  border.left,
  border.top,
  border.right,
  shading.color,
  keep_with_next,
  word_style,
  ...
) {
  if (!missing(text.align)) {
    object <- check_set_choice(
      obj = object,
      value = text.align,
      choices = c("left", "right", "center", "justify")
    )
  }
  if (!missing(word_style)) {
    object <- check_set_chr(obj = object, word_style)
  }

  # padding checking
  if (!missing(padding)) {
    object <- check_spread_integer(
      object,
      padding,
      c(
        "padding.bottom",
        "padding.top",
        "padding.left",
        "padding.right"
      )
    )
  }
  if (!missing(padding.bottom)) {
    object <- check_set_integer(obj = object, padding.bottom)
  }
  if (!missing(padding.left)) {
    object <- check_set_integer(obj = object, padding.left)
  }
  if (!missing(padding.top)) {
    object <- check_set_integer(obj = object, padding.top)
  }
  if (!missing(padding.right)) {
    object <- check_set_integer(obj = object, padding.right)
  }

  # border checking
  if (!missing(border)) {
    object <- check_spread_border(
      obj = object,
      border,
      dest = c(
        "border.bottom",
        "border.top",
        "border.left",
        "border.right"
      )
    )
  }
  if (!missing(border.top)) {
    object <- check_set_border(obj = object, border.top)
  }
  if (!missing(border.bottom)) {
    object <- check_set_border(obj = object, border.bottom)
  }
  if (!missing(border.left)) {
    object <- check_set_border(obj = object, border.left)
  }
  if (!missing(border.right)) {
    object <- check_set_border(obj = object, border.right)
  }

  if (!missing(shading.color)) {
    object <- check_set_color(object, shading.color)
  }
  if (!missing(keep_with_next)) {
    object <- check_set_bool(object, keep_with_next)
  }

  object
}

# fp_cell ----

vertical.align.styles <- c("top", "center", "bottom")
text.directions <- c("lrtb", "tbrl", "btlr")

#' @title Cell formatting properties
#'
#' @description Create a `fp_cell` object that describes cell formatting properties.
#'
#' @param border shortcut for all borders.
#' @param border.bottom,border.left,border.top,border.right [fp_border()] for borders.
#' @param vertical.align cell content vertical alignment - a single character value,
#' expected value is one of "center" or "top" or "bottom"
#' @param margin shortcut for all margins.
#' @param margin.bottom,margin.top,margin.left,margin.right cell margins - 0 or positive integer value.
#' @param background.color cell background color - a single character value specifying a
#' valid color (e.g. "#000000" or "black").
#' @param text.direction cell text rotation - a single character value, expected
#' value is one of "lrtb", "tbrl", "btlr".
#' @param rowspan specify how many rows the cell is spanned over
#' @param colspan specify how many columns the cell is spanned over
#' @export
#' @family functions for defining formatting properties
fp_cell <- function(
  border = fp_border(width = 0),
  border.bottom,
  border.left,
  border.top,
  border.right,
  vertical.align = "center",
  margin = 0,
  margin.bottom,
  margin.top,
  margin.left,
  margin.right,
  background.color = "transparent",
  text.direction = "lrtb",
  rowspan = 1,
  colspan = 1
) {
  out <- list()

  # border checking
  out <- check_spread_border(
    obj = out,
    border,
    dest = c(
      "border.bottom",
      "border.top",
      "border.left",
      "border.right"
    )
  )
  if (!missing(border.top)) {
    out <- check_set_border(obj = out, border.top)
  }
  if (!missing(border.bottom)) {
    out <- check_set_border(obj = out, border.bottom)
  }
  if (!missing(border.left)) {
    out <- check_set_border(obj = out, border.left)
  }
  if (!missing(border.right)) {
    out <- check_set_border(obj = out, border.right)
  }

  # background-color checking
  out <- check_set_color(out, background.color)

  out <- check_set_choice(
    obj = out,
    value = vertical.align,
    choices = vertical.align.styles
  )
  out <- check_set_choice(
    obj = out,
    value = text.direction,
    choices = text.directions
  )

  # margin checking
  out <- check_spread_integer(
    out,
    margin,
    c(
      "margin.bottom",
      "margin.top",
      "margin.left",
      "margin.right"
    )
  )

  if (!missing(margin.bottom)) {
    out <- check_set_integer(obj = out, margin.bottom)
  }
  if (!missing(margin.left)) {
    out <- check_set_integer(obj = out, margin.left)
  }
  if (!missing(margin.top)) {
    out <- check_set_integer(obj = out, margin.top)
  }
  if (!missing(margin.right)) {
    out <- check_set_integer(obj = out, margin.right)
  }

  out <- check_set_integer(obj = out, rowspan)
  out <- check_set_integer(obj = out, colspan)

  class(out) <- "fp_cell"
  out
}


#' @export
#' @rdname fp_cell
#' @param x,object `fp_cell` object
#' @param type output type - one of 'wml', 'pml', 'html', 'rtf'.
#' @param ... further arguments - not used
format.fp_cell <- function(x, type = "wml", ...) {
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("wml", "pml", "html", "rtf"))

  if (type == "wml") {
    tcpr_wml(x)
  } else if (type == "pml") {
    tcpr_pml(x)
  } else if (type == "html") {
    tcpr_css(x)
  } else if (type == "rtf") {
    tcpr_rtf(x)
  } else {
    stop("unimplemented")
  }
}

#' @export
to_wml.fp_cell <- function(x, add_ns = FALSE, ...) {
  format(x, type = "wml")
}

#' @export
#' @rdname fp_cell
print.fp_cell <- function(x, ...) {
  cat(format(x, type = "html"))
}


#' @rdname fp_cell
#' @examples
#' obj <- fp_cell(margin = 1)
#' update(obj, margin.bottom = 5)
#' @export
update.fp_cell <- function(
  object,
  border,
  border.bottom,
  border.left,
  border.top,
  border.right,
  vertical.align,
  margin = 0,
  margin.bottom,
  margin.top,
  margin.left,
  margin.right,
  background.color,
  text.direction,
  rowspan = 1,
  colspan = 1,
  ...
) {
  if (!missing(border)) {
    object <- check_spread_border(
      obj = object,
      border,
      dest = c(
        "border.bottom",
        "border.top",
        "border.left",
        "border.right"
      )
    )
  }
  if (!missing(border.top)) {
    object <- check_set_border(obj = object, border.top)
  }
  if (!missing(border.bottom)) {
    object <- check_set_border(obj = object, border.bottom)
  }
  if (!missing(border.left)) {
    object <- check_set_border(obj = object, border.left)
  }
  if (!missing(border.right)) {
    object <- check_set_border(obj = object, border.right)
  }

  # background-color checking
  if (!missing(background.color)) {
    object <- check_set_color(object, background.color)
  }

  if (!missing(vertical.align)) {
    object <- check_set_choice(
      obj = object,
      value = vertical.align,
      choices = vertical.align.styles
    )
  }
  if (!missing(text.direction)) {
    object <- check_set_choice(
      obj = object,
      value = text.direction,
      choices = text.directions
    )
  }

  # margin checking
  if (!missing(margin)) {
    object <- check_spread_integer(
      object,
      margin,
      c(
        "margin.bottom",
        "margin.top",
        "margin.left",
        "margin.right"
      )
    )
  }

  if (!missing(margin.bottom)) {
    object <- check_set_integer(obj = object, margin.bottom)
  }
  if (!missing(margin.left)) {
    object <- check_set_integer(obj = object, margin.left)
  }
  if (!missing(margin.top)) {
    object <- check_set_integer(obj = object, margin.top)
  }
  if (!missing(margin.right)) {
    object <- check_set_integer(obj = object, margin.right)
  }

  if (!missing(rowspan)) {
    object <- check_set_integer(obj = object, rowspan)
  }
  if (!missing(colspan)) {
    object <- check_set_integer(obj = object, colspan)
  }

  object
}

source("utils.R")

test_that("fp_border", {
  expect_error(fp_border(width = -5), "width must be a positive numeric scalar")
  expect_error(fp_border(color = "glop"), "color must be a valid color")
  expect_error(fp_border(style = "glop"), "style must be one of")
  x <- fp_border(color = "red", style = "dashed", width = 5)
  x <- update(x, color = "yellow", style = "solid", width = 2)
  expect_equal(x$color, "yellow")
  expect_equal(x$style, "solid")
  expect_equal(x$width, 2)
})

pml_border_node <- function(x, side = "L") {
  x <- fp_cell(border = x)
  xml_ <- pml_str(format(x, type = "pml"))
  doc <- read_xml(xml_)
  xml_find_first(doc, paste0("//a:tcPr/a:ln", side))
}


test_that("pml fp_border", {
  node <- pml_border_node(fp_border(width = 2, color = "#00FF0099"))

  col <- xml_attr(xml_child(node, "/a:solidFill/a:srgbClr"), "val")
  expect_equal(col, "00FF00")

  alpha <- xml_attr(xml_child(node, "/a:solidFill/a:srgbClr/a:alpha"), "val")
  expect_equal(alpha, "60000")

  type <- xml_attr(xml_child(node, "/a:prstDash"), "val")
  expect_equal(type, "solid")
  node <- pml_border_node(fp_border(style = "dashed"))
  type <- xml_attr(xml_child(node, "/a:prstDash"), "val")
  expect_equal(type, "sysDash")
  node <- pml_border_node(fp_border(style = "dotted", width = 2))
  type <- xml_attr(xml_child(node, "/a:prstDash"), "val")
  expect_equal(type, "sysDot")

  width <- as.integer(xml_attr(node, "w"))
  expect_equal(width, 12700 * 2)

  node <- pml_border_node(fp_border(color = "transparent"))
  col <- xml_child(node, "a:noFill")
  expect_false(inherits(col, "xml_missing"))
})

test_that("print fp_border", {
  x <- fp_border()
  expect_output(print(x))
})


wml_border_node <- function(x, side = "bottom") {
  x <- fp_cell(border = x)
  xml_ <- wml_str(format(x, type = "wml"))
  doc <- read_xml(xml_)
  xml_find_first(doc, paste0("//w:tcPr/w:tcBorders/w:", side))
}

test_that("wml fp_border", {
  node <- wml_border_node(fp_border(width = 2, color = "#00FF00"))

  col <- xml_attr(node, "color")
  expect_equal(col, "00FF00")

  sz <- xml_attr(node, "sz")
  expect_equal(sz, "16")

  node <- wml_border_node(fp_border(style = "dotted"))
  expect_equal(xml_attr(node, "val"), "dotted")
  node <- wml_border_node(fp_border(style = "solid"))
  expect_equal(xml_attr(node, "val"), "single")
  node <- wml_border_node(fp_border(style = "dashed"))
  expect_equal(xml_attr(node, "val"), "dashed")

  node <- wml_border_node(fp_border(style = "nil"))
  expect_equal(xml_attr(node, "val"), "nil")
  node <- wml_border_node(fp_border(style = "none"))
  expect_equal(xml_attr(node, "val"), "none")
  node <- wml_border_node(fp_border(style = "single"))
  expect_equal(xml_attr(node, "val"), "single")
  node <- wml_border_node(fp_border(style = "thick"))
  expect_equal(xml_attr(node, "val"), "thick")
  node <- wml_border_node(fp_border(style = "double"))
  expect_equal(xml_attr(node, "val"), "double")
  node <- wml_border_node(fp_border(style = "dotDash"))
  expect_equal(xml_attr(node, "val"), "dotDash")
  node <- wml_border_node(fp_border(style = "dotDotDash"))
  expect_equal(xml_attr(node, "val"), "dotDotDash")
  node <- wml_border_node(fp_border(style = "triple"))
  expect_equal(xml_attr(node, "val"), "triple")
  node <- wml_border_node(fp_border(style = "thinThickSmallGap"))
  expect_equal(xml_attr(node, "val"), "thinThickSmallGap")
  node <- wml_border_node(fp_border(style = "thickThinSmallGap"))
  expect_equal(xml_attr(node, "val"), "thickThinSmallGap")
  node <- wml_border_node(fp_border(style = "thinThickThinSmallGap"))
  expect_equal(xml_attr(node, "val"), "thinThickThinSmallGap")
  node <- wml_border_node(fp_border(style = "thinThickMediumGap"))
  expect_equal(xml_attr(node, "val"), "thinThickMediumGap")
  node <- wml_border_node(fp_border(style = "thickThinMediumGap"))
  expect_equal(xml_attr(node, "val"), "thickThinMediumGap")
  node <- wml_border_node(fp_border(style = "thinThickThinMediumGap"))
  expect_equal(xml_attr(node, "val"), "thinThickThinMediumGap")
  node <- wml_border_node(fp_border(style = "thinThickLargeGap"))
  expect_equal(xml_attr(node, "val"), "thinThickLargeGap")
  node <- wml_border_node(fp_border(style = "thickThinLargeGap"))
  expect_equal(xml_attr(node, "val"), "thickThinLargeGap")
  node <- wml_border_node(fp_border(style = "thinThickThinLargeGap"))
  expect_equal(xml_attr(node, "val"), "thinThickThinLargeGap")
  node <- wml_border_node(fp_border(style = "wave"))
  expect_equal(xml_attr(node, "val"), "wave")
  node <- wml_border_node(fp_border(style = "doubleWave"))
  expect_equal(xml_attr(node, "val"), "doubleWave")
  node <- wml_border_node(fp_border(style = "dashSmallGap"))
  expect_equal(xml_attr(node, "val"), "dashSmallGap")
  node <- wml_border_node(fp_border(style = "dashDotStroked"))
  expect_equal(xml_attr(node, "val"), "dashDotStroked")
  node <- wml_border_node(fp_border(style = "threeDEmboss"))
  expect_equal(xml_attr(node, "val"), "threeDEmboss")
  node <- wml_border_node(fp_border(style = "threeDEngrave"))
  expect_equal(xml_attr(node, "val"), "threeDEngrave")
  node <- wml_border_node(fp_border(style = "ridge"))
  expect_equal(xml_attr(node, "val"), "threeDEmboss")
  node <- wml_border_node(fp_border(style = "groove"))
  expect_equal(xml_attr(node, "val"), "threeDEngrave")
  node <- wml_border_node(fp_border(style = "outset"))
  expect_equal(xml_attr(node, "val"), "outset")
  node <- wml_border_node(fp_border(style = "inset"))
  expect_equal(xml_attr(node, "val"), "inset")
})


regexp_border <- function(width, style, color, x) {
  css <- format(fp_par(border = x), type = "html")
  reg <- sprintf("border-bottom: %.02fpt %s %s", width, style, color)
  grepl(reg, css)
}

test_that("css fp_border", {
  col <- "#00FF00"
  x <- fp_border(width = 2, color = col, style = "solid")
  rb <- regexp_border(2, "solid", "rgba\\(0,255,0,1.00\\)", x)
  expect_true(rb)

  col <- "#00FF0099"
  x <- fp_border(width = 4, color = col, style = "single")
  rb <- regexp_border(4, "solid", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "dashed")
  rb <- regexp_border(4, "dashed", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "dotted")
  rb <- regexp_border(4, "dotted", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "double")
  rb <- regexp_border(4, "double", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "inset")
  rb <- regexp_border(4, "inset", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "outset")
  rb <- regexp_border(4, "outset", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "none")
  rb <- regexp_border(4, "none", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "nil")
  rb <- regexp_border(4, "none", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "threeDEmboss")
  rb <- regexp_border(4, "ridge", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "ridge")
  rb <- regexp_border(4, "ridge", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "threeDEngrave")
  rb <- regexp_border(4, "groove", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(width = 4, color = col, style = "groove")
  rb <- regexp_border(4, "groove", "rgba\\(0,255,0,0.60\\)", x)
  expect_true(rb)

  x <- fp_border(color = "transparent")
  rb <- regexp_border(1, "solid", "transparent", x)
  expect_true(rb)
})

source("utils.R")

test_that("fp_par", {
  expect_error(fp_par(text.align = "glop"), "must be one of")
  expect_error(
    fp_par(padding = -2),
    "padding must be a positive integer scalar"
  )
  expect_error(fp_par(border = fp_text()), "border must be a fp_border object")
  expect_error(
    fp_par(shading.color = "#340F2A78d"),
    "shading\\.color must be a valid color"
  )
  x <- fp_par()
  x <- update(
    x,
    padding = 3,
    text.align = "center",
    shading.color = "#340F2A78",
    border = fp_border(color = "red")
  )
  expect_equal(x$shading.color, "#340F2A78")
  expect_equal(x$padding.bottom, 3)
  expect_equal(x$padding.left, 3)
  expect_equal(x$border.bottom, fp_border(color = "red"))
})


test_that("print fp_par", {
  x <- fp_par()
  expect_output(print(x))
})

get_ppr <- function(x) {
  str <- format(x, type = "pml")
  doc <- as_xml_document(pml_str(str))
  xml_find_first(doc, "a:pPr")
}

test_that("format pml fp_par", {
  x <- fp_par(text.align = "left")
  node <- get_ppr(x)
  expect_equal(xml_attr(node, "algn"), "l")
  x <- update(x, text.align = "center")
  node <- get_ppr(x)
  expect_equal(xml_attr(node, "algn"), "ctr")
  x <- update(x, text.align = "right")
  node <- get_ppr(x)
  expect_equal(xml_attr(node, "algn"), "r")
  x <- update(x, padding = 3, padding.bottom = 0)
  node <- get_ppr(x)
  expect_equal(xml_attr(xml_child(node, "a:spcBef/a:spcPts"), "val"), "300")
  expect_equal(xml_attr(xml_child(node, "a:spcAft/a:spcPts"), "val"), "0")
  expect_equal(xml_attr(node, "marL"), as.character(12700 * 3))
  expect_equal(xml_attr(node, "marR"), as.character(12700 * 3))
})

test_that("format pml fp_par_lite", {
  x <- fp_par_lite()
  expect_equal(ppr_pml(x), "<a:pPr><a:buNone/></a:pPr>")
  x <- fp_par_lite(text.align = "left")
  expect_equal(ppr_pml(x), "<a:pPr algn=\"l\"><a:buNone/></a:pPr>")
})

test_that("format wml fp_par_lite", {
  x <- fp_par_lite()
  expect_equal(ppr_wml(x), "<w:pPr></w:pPr>")
  x <- fp_par_lite(text.align = "left")
  expect_equal(ppr_wml(x), "<w:pPr><w:jc w:val=\"left\"/></w:pPr>")
})


test_that("format css fp_par", {
  x <- fp_par(text.align = "left")
  expect_true(has_css_attr(x, "text-align", "left"))
  x <- update(x, text.align = "center")
  expect_true(has_css_attr(x, "text-align", "center"))
  x <- update(x, text.align = "right")
  expect_true(has_css_attr(x, "text-align", "right"))
  x <- update(x, padding = 3, padding.bottom = 0)
  expect_true(has_css_attr(x, "padding-top", "3pt"))
  expect_true(has_css_attr(x, "padding-bottom", "0pt"))
  expect_true(has_css_attr(x, "padding-right", "3pt"))
  expect_true(has_css_attr(x, "padding-left", "3pt"))

  x <- update(x, shading.color = "#00FF0099")
  expect_true(has_css_color(x, "background-color", "rgba\\(0,255,0,0.60\\)"))
})


is_align <- function(x, align) {
  xml_ <- format(x, type = "wml")
  doc <- read_xml(wml_str(xml_))
  val <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:jc"), "val")
  val == align
}


test_that("wml text align", {
  x <- fp_par(text.align = "center")
  expect_true(is_align(x, align = "center"))

  x <- update(x, text.align = "left")
  expect_true(is_align(x, align = "left"))

  x <- update(x, text.align = "right")
  expect_true(is_align(x, align = "right"))

  x <- update(x, text.align = "justify")
  expect_true(is_align(x, align = "both"))
})


get_padding <- function(x, align) {
  xml_ <- format(x, type = "wml")
  doc <- read_xml(wml_str(xml_))
  after <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:spacing"), "after")
  before <- xml_attr(
    xml_find_first(doc, "/w:document/w:pPr/w:spacing"),
    "before"
  )
  left <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:ind"), "left")
  right <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:ind"), "right")
  out <- as.integer(c(after, before, left, right)) / 20
  names(out) <- c("bottom", "top", "left", "right")
  out
}

test_that("wml padding", {
  x <- fp_par(padding = 2)
  expect_equal(
    get_padding(x),
    c("bottom" = 2, "top" = 2, "left" = 2, "right" = 2)
  )

  x <- fp_par(padding = 2, padding.top = 5)
  expect_equal(
    get_padding(x),
    c("bottom" = 2, "top" = 5, "left" = 2, "right" = 2)
  )

  x <- fp_par(padding = 2, padding.bottom = 5)
  expect_equal(
    get_padding(x),
    c("bottom" = 5, "top" = 2, "left" = 2, "right" = 2)
  )
})

test_that("wml shading.color", {
  x <- fp_par(shading.color = "#FF0000")

  xml_ <- format(x, type = "wml")
  doc <- read_xml(wml_str(xml_))
  shd <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:shd"), "fill")
  expect_equal(shd, "FF0000")
})

test_that("fp_par_lite", {
  wml <- format(fp_par_lite(), type = "wml")
  expect_equal(wml, "<w:pPr></w:pPr>")
  wml <- format(fp_par_lite(word_style = "dummy"), type = "wml")
  expect_equal(wml, "<w:pPr><w:pStyle w:pstlname=\"dummy\"/></w:pPr>")
  wml <- format(fp_par_lite(text.align = "center"), type = "wml")
  expect_equal(wml, "<w:pPr><w:jc w:val=\"center\"/></w:pPr>")
  wml <- format(
    fp_par_lite(word_style = "dummy", text.align = "center"),
    type = "wml"
  )
  expect_equal(
    wml,
    "<w:pPr><w:pStyle w:pstlname=\"dummy\"/><w:jc w:val=\"center\"/></w:pPr>"
  )
})

source("utils.R")

test_that("fp_border", {
  expect_error( fp_border(width = -5), "width must be a positive numeric scalar" )
  expect_error( fp_border(color = "glop"), "color must be a valid color" )
  expect_error( fp_border(style = "glop"), "style must be one of" )
  x <- fp_border(color = "red", style = "dashed", width = 5)
  x <- update(x, color = "yellow", style = "solid", width = 2)
  expect_equal(x$color, "yellow")
  expect_equal(x$style, "solid")
  expect_equal(x$width, 2)


})

pml_border_node <- function(x, side = "L"){
  x <- fp_cell(border = x )
  xml_ <- pml_str(format(x, type = "pml"))
  doc <- read_xml( xml_ )
  xml_find_first(doc, paste0("//a:tcPr/a:ln", side))
}



test_that("pml fp_border", {

  node <- pml_border_node(fp_border(width = 2, color = "#00FF0099"))

  col <- xml_attr(xml_child(node, "/a:solidFill/a:srgbClr"), "val")
  expect_equal(col, "00FF00")

  alpha <- xml_attr(xml_child(node, "/a:solidFill/a:srgbClr/a:alpha"), "val")
  expect_equal(alpha, "60000")

  type <- xml_attr(xml_child(node, "/a:prstDash"),"val")
  expect_equal(type, "solid")
  node <- pml_border_node(fp_border(style = "dashed"))
  type <- xml_attr(xml_child(node, "/a:prstDash"), "val")
  expect_equal(type, "sysDash")
  node <- pml_border_node(fp_border(style = "dotted", width = 2))
  type <- xml_attr(xml_child(node, "/a:prstDash"), "val")
  expect_equal(type, "sysDot")

  width <- as.integer(xml_attr(node, "w"))
  expect_equal(width, 12700 * 2)

  node <- pml_border_node(fp_border(color = "transparent"))
  col <- xml_child(node, "a:noFill")
  expect_false(inherits(col, "xml_missing"))
})


wml_border_node <- function(x, side = "bottom"){
  x <- fp_cell(border = x )
  xml_ <- wml_str(format(x, type = "wml"))
  doc <- read_xml( xml_ )
  xml_find_first(doc, paste0("//w:tcPr/w:tcBorders/w:", side))
}

test_that("wml fp_border", {

  node <- wml_border_node(fp_border(width = 2, color = "#00FF00"))

  col <- xml_attr(node, "color")
  expect_equal(col, "00FF00")

  sz <- xml_attr(node, "sz")
  expect_equal(sz, "16")

  node <- wml_border_node(fp_border(style = "dotted"))
  expect_equal(xml_attr(node, "val"), "dotted")
  node <- wml_border_node(fp_border(style = "solid"))
  expect_equal(xml_attr(node, "val"), "single")
  node <- wml_border_node(fp_border(style = "dashed"))
  expect_equal(xml_attr(node, "val"), "dashed")

  node <- wml_border_node(fp_border(style = "nil"))
  expect_equal(xml_attr(node, "val"), "nil")
  node <- wml_border_node(fp_border(style = "none"))
  expect_equal(xml_attr(node, "val"), "none")
  node <- wml_border_node(fp_border(style = "single"))
  expect_equal(xml_attr(node, "val"), "single")
  node <- wml_border_node(fp_border(style = "thick"))
  expect_equal(xml_attr(node, "val"), "thick")
  node <- wml_border_node(fp_border(style = "double"))
  expect_equal(xml_attr(node, "val"), "double")
  node <- wml_border_node(fp_border(style = "dotDash"))
  expect_equal(xml_attr(node, "val"), "dotDash")
  node <- wml_border_node(fp_border(style = "dotDotDash"))
  expect_equal(xml_attr(node, "val"), "dotDotDash")
  node <- wml_border_node(fp_border(style = "triple"))
  expect_equal(xml_attr(node, "val"), "triple")
  node <- wml_border_node(fp_border(style = "thinThickSmallGap"))
  expect_equal(xml_attr(node, "val"), "thinThickSmallGap")
  node <- wml_border_node(fp_border(style = "thickThinSmallGap"))
  expect_equal(xml_attr(node, "val"), "thickThinSmallGap")
  node <- wml_border_node(fp_border(style = "thinThickThinSmallGap"))
  expect_equal(xml_attr(node, "val"), "thinThickThinSmallGap")
  node <- wml_border_node(fp_border(style = "thinThickMediumGap"))
  expect_equal(xml_attr(node, "val"), "thinThickMediumGap")
  node <- wml_border_node(fp_border(style = "thickThinMediumGap"))
  expect_equal(xml_attr(node, "val"), "thickThinMediumGap")
  node <- wml_border_node(fp_border(style = "thinThickThinMediumGap"))
  expect_equal(xml_attr(node, "val"), "thinThickThinMediumGap")
  node <- wml_border_node(fp_border(style = "thinThickLargeGap"))
  expect_equal(xml_attr(node, "val"), "thinThickLargeGap")
  node <- wml_border_node(fp_border(style = "thickThinLargeGap"))
  expect_equal(xml_attr(node, "val"), "thickThinLargeGap")
  node <- wml_border_node(fp_border(style = "thinThickThinLargeGap"))
  expect_equal(xml_attr(node, "val"), "thinThickThinLargeGap")
  node <- wml_border_node(fp_border(style = "wave"))
  expect_equal(xml_attr(node, "val"), "wave")
  node <- wml_border_node(fp_border(style = "doubleWave"))
  expect_equal(xml_attr(node, "val"), "doubleWave")
  node <- wml_border_node(fp_border(style = "dashSmallGap"))
  expect_equal(xml_attr(node, "val"), "dashSmallGap")
  node <- wml_border_node(fp_border(style = "dashDotStroked"))
  expect_equal(xml_attr(node, "val"), "dashDotStroked")
  node <- wml_border_node(fp_border(style = "threeDEmboss"))
  expect_equal(xml_attr(node, "val"), "threeDEmboss")
  node <- wml_border_node(fp_border(style = "threeDEngrave"))
  expect_equal(xml_attr(node, "val"), "threeDEngrave")
  node <- wml_border_node(fp_border(style = "ridge"))
  expect_equal(xml_attr(node, "val"), "threeDEmboss")
  node <- wml_border_node(fp_border(style = "groove"))
  expect_equal(xml_attr(node, "val"), "threeDEngrave")
  node <- wml_border_node(fp_border(style = "outset"))
  expect_equal(xml_attr(node, "val"), "outset")
  node <- wml_border_node(fp_border(style = "inset"))
  expect_equal(xml_attr(node, "val"), "inset")
})


regexp_border <- function(width, style, color, x){
  css <- format(fp_par(border = x), type = "html")
  reg <- sprintf("border-bottom: %.02fpt %s %s", width, style, color)
  grepl(reg, css)
}

test_that("css fp_border", {

  col <- "#00FF00"
  x <- fp_border(width = 2, color = col, style = "solid")
  rb <- regexp_border(2, "solid", "rgba\\(0,255,0,1.00\\)", x)
  expect_true( rb )

  col <- "#00FF0099"
  x <- fp_border(width = 4, color = col, style = "single")
  rb <- regexp_border(4, "solid", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "dashed")
  rb <- regexp_border(4, "dashed", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "dotted")
  rb <- regexp_border(4, "dotted", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "double")
  rb <- regexp_border(4, "double", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "inset")
  rb <- regexp_border(4, "inset", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "outset")
  rb <- regexp_border(4, "outset", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "none")
  rb <- regexp_border(4, "none", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "nil")
  rb <- regexp_border(4, "none", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "threeDEmboss")
  rb <- regexp_border(4, "ridge", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "ridge")
  rb <- regexp_border(4, "ridge", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "threeDEngrave")
  rb <- regexp_border(4, "groove", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(width = 4, color = col, style = "groove")
  rb <- regexp_border(4, "groove", "rgba\\(0,255,0,0.60\\)", x)
  expect_true( rb )

  x <- fp_border(color = "transparent")
  rb <- regexp_border(1, "solid", "transparent", x)
  expect_true( rb )
})

source("utils.R")

test_that("fp_par", {
  expect_error(fp_par(text.align = "glop"), "must be one of")
  expect_error(
    fp_par(padding = -2),
    "padding must be a positive integer scalar"
  )
  expect_error(fp_par(border = fp_text()), "border must be a fp_border object")
  expect_error(
    fp_par(shading.color = "#340F2A78d"),
    "shading\\.color must be a valid color"
  )
  x <- fp_par()
  x <- update(
    x,
    padding = 3,
    text.align = "center",
    shading.color = "#340F2A78",
    border = fp_border(color = "red")
  )
  expect_equal(x$shading.color, "#340F2A78")
  expect_equal(x$padding.bottom, 3)
  expect_equal(x$padding.left, 3)
  expect_equal(x$border.bottom, fp_border(color = "red"))
})


test_that("print fp_par", {
  x <- fp_par()
  expect_output(print(x))
})

get_ppr <- function(x) {
  str <- format(x, type = "pml")
  doc <- as_xml_document(pml_str(str))
  xml_find_first(doc, "a:pPr")
}

test_that("format pml fp_par", {
  x <- fp_par(text.align = "left")
  node <- get_ppr(x)
  expect_equal(xml_attr(node, "algn"), "l")
  x <- update(x, text.align = "center")
  node <- get_ppr(x)
  expect_equal(xml_attr(node, "algn"), "ctr")
  x <- update(x, text.align = "right")
  node <- get_ppr(x)
  expect_equal(xml_attr(node, "algn"), "r")
  x <- update(x, padding = 3, padding.bottom = 0)
  node <- get_ppr(x)
  expect_equal(xml_attr(xml_child(node, "a:spcBef/a:spcPts"), "val"), "300")
  expect_equal(xml_attr(xml_child(node, "a:spcAft/a:spcPts"), "val"), "0")
  expect_equal(xml_attr(node, "marL"), as.character(12700 * 3))
  expect_equal(xml_attr(node, "marR"), as.character(12700 * 3))
})

test_that("format pml fp_par_lite", {
  x <- fp_par_lite()
  expect_equal(ppr_pml(x), "<a:pPr><a:buNone/></a:pPr>")
  x <- fp_par_lite(text.align = "left")
  expect_equal(ppr_pml(x), "<a:pPr algn=\"l\"><a:buNone/></a:pPr>")
})

test_that("format wml fp_par_lite", {
  x <- fp_par_lite()
  expect_equal(ppr_wml(x), "<w:pPr></w:pPr>")
  x <- fp_par_lite(text.align = "left")
  expect_equal(ppr_wml(x), "<w:pPr><w:jc w:val=\"left\"/></w:pPr>")
})


test_that("format css fp_par", {
  x <- fp_par(text.align = "left")
  expect_true(has_css_attr(x, "text-align", "left"))
  x <- update(x, text.align = "center")
  expect_true(has_css_attr(x, "text-align", "center"))
  x <- update(x, text.align = "right")
  expect_true(has_css_attr(x, "text-align", "right"))
  x <- update(x, padding = 3, padding.bottom = 0)
  expect_true(has_css_attr(x, "padding-top", "3pt"))
  expect_true(has_css_attr(x, "padding-bottom", "0pt"))
  expect_true(has_css_attr(x, "padding-right", "3pt"))
  expect_true(has_css_attr(x, "padding-left", "3pt"))

  x <- update(x, shading.color = "#00FF0099")
  expect_true(has_css_color(x, "background-color", "rgba\\(0,255,0,0.60\\)"))
})


is_align <- function(x, align) {
  xml_ <- format(x, type = "wml")
  doc <- read_xml(wml_str(xml_))
  val <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:jc"), "val")
  val == align
}


test_that("wml text align", {
  x <- fp_par(text.align = "center")
  expect_true(is_align(x, align = "center"))

  x <- update(x, text.align = "left")
  expect_true(is_align(x, align = "left"))

  x <- update(x, text.align = "right")
  expect_true(is_align(x, align = "right"))

  x <- update(x, text.align = "justify")
  expect_true(is_align(x, align = "both"))
})


get_padding <- function(x, align) {
  xml_ <- format(x, type = "wml")
  doc <- read_xml(wml_str(xml_))
  after <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:spacing"), "after")
  before <- xml_attr(
    xml_find_first(doc, "/w:document/w:pPr/w:spacing"),
    "before"
  )
  left <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:ind"), "left")
  right <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:ind"), "right")
  out <- as.integer(c(after, before, left, right)) / 20
  names(out) <- c("bottom", "top", "left", "right")
  out
}

test_that("wml padding", {
  x <- fp_par(padding = 2)
  expect_equal(
    get_padding(x),
    c("bottom" = 2, "top" = 2, "left" = 2, "right" = 2)
  )

  x <- fp_par(padding = 2, padding.top = 5)
  expect_equal(
    get_padding(x),
    c("bottom" = 2, "top" = 5, "left" = 2, "right" = 2)
  )

  x <- fp_par(padding = 2, padding.bottom = 5)
  expect_equal(
    get_padding(x),
    c("bottom" = 5, "top" = 2, "left" = 2, "right" = 2)
  )
})

test_that("wml first_line / hanging indent", {
  # default: neither first_line nor hanging -> legacy firstLine="0"
  x <- fp_par()
  wml <- format(x, type = "wml")
  expect_match(wml, "w:firstLine=\"0\" w:firstLineChars=\"0\"", fixed = TRUE)

  # hanging > 0 -> w:hanging
  x <- fp_par(padding.left = 40, hanging = 20)
  wml <- format(x, type = "wml")
  expect_match(wml, "w:hanging=\"400\"", fixed = TRUE)
  expect_false(grepl("w:firstLine", wml))

  # first_line > 0 -> w:firstLine
  x <- fp_par(padding.left = 40, first_line = 15)
  wml <- format(x, type = "wml")
  expect_match(wml, "w:firstLine=\"300\"", fixed = TRUE)
  expect_false(grepl("w:hanging", wml))

  # hanging wins if both are provided
  x <- fp_par(padding.left = 40, first_line = 15, hanging = 20)
  wml <- format(x, type = "wml")
  expect_match(wml, "w:hanging=\"400\"", fixed = TRUE)
  expect_false(grepl("w:firstLine", wml))
})

test_that("pml first_line / hanging indent", {
  x <- fp_par(hanging = 20)
  pml <- format(x, type = "pml")
  expect_match(pml, "indent=\"-254000\"", fixed = TRUE)

  x <- fp_par(first_line = 15)
  pml <- format(x, type = "pml")
  expect_match(pml, "indent=\"190500\"", fixed = TRUE)

  # default: no indent attribute
  x <- fp_par()
  pml <- format(x, type = "pml")
  expect_false(grepl("indent=", pml))
})

test_that("css first_line / hanging indent", {
  x <- fp_par(hanging = 20)
  css <- format(x, type = "html")
  expect_match(css, "text-indent:-20pt;", fixed = TRUE)

  x <- fp_par(first_line = 15)
  css <- format(x, type = "html")
  expect_match(css, "text-indent:15pt;", fixed = TRUE)

  x <- fp_par()
  css <- format(x, type = "html")
  expect_false(grepl("text-indent:", css))
})

test_that("rtf first_line / hanging indent", {
  x <- fp_par(padding.left = 40, hanging = 20)
  rtf <- format(x, type = "rtf")
  expect_match(rtf, "\\fi-400", fixed = TRUE)

  x <- fp_par(padding.left = 40, first_line = 15)
  rtf <- format(x, type = "rtf")
  expect_match(rtf, "\\fi300", fixed = TRUE)

  # default (fp_par() triggers padding.left=0 via padding=0 default)
  x <- fp_par()
  rtf <- format(x, type = "rtf")
  expect_match(rtf, "\\fi0", fixed = TRUE)
})

test_that("validate_indent rejects bad inputs", {
  expect_error(fp_par(hanging = "a"), "single numeric")
  expect_error(fp_par(first_line = c(1, 2)), "single numeric")
  # NA is explicitly OK (default)
  expect_silent(fp_par(hanging = NA))
})

test_that("wml shading.color", {
  x <- fp_par(shading.color = "#FF0000")

  xml_ <- format(x, type = "wml")
  doc <- read_xml(wml_str(xml_))
  shd <- xml_attr(xml_find_first(doc, "/w:document/w:pPr/w:shd"), "fill")
  expect_equal(shd, "FF0000")
})

test_that("fp_par_lite", {
  wml <- format(fp_par_lite(), type = "wml")
  expect_equal(wml, "<w:pPr></w:pPr>")
  wml <- format(fp_par_lite(word_style = "dummy"), type = "wml")
  expect_equal(wml, "<w:pPr><w:pStyle w:pstlname=\"dummy\"/></w:pPr>")
  wml <- format(fp_par_lite(text.align = "center"), type = "wml")
  expect_equal(wml, "<w:pPr><w:jc w:val=\"center\"/></w:pPr>")
  wml <- format(
    fp_par_lite(word_style = "dummy", text.align = "center"),
    type = "wml"
  )
  expect_equal(
    wml,
    "<w:pPr><w:pStyle w:pstlname=\"dummy\"/><w:jc w:val=\"center\"/></w:pPr>"
  )
})

test_that("fp_tab", {
  tab <- fp_tab(pos = 0.4, style = "decimal")
  expect_equal(
    to_wml(tab),
    "<w:tab w:val=\"decimal\" w:leader=\"none\" w:pos=\"576\"/>"
  )
  expect_equal(rtf_fp_tab(tab), "\\tx576\\tqdec")
  tab <- fp_tab(pos = 0.4, style = "left")
  expect_equal(
    to_wml(tab),
    "<w:tab w:val=\"left\" w:leader=\"none\" w:pos=\"576\"/>"
  )
  expect_equal(rtf_fp_tab(tab), "\\tx576")
  tab <- fp_tab(pos = 0.4, style = "right")
  expect_equal(
    to_wml(tab),
    "<w:tab w:val=\"right\" w:leader=\"none\" w:pos=\"576\"/>"
  )
  expect_equal(rtf_fp_tab(tab), "\\tx576\\tqr")
  tab <- fp_tab(pos = 0.4, style = "center")
  expect_equal(
    to_wml(tab),
    "<w:tab w:val=\"center\" w:leader=\"none\" w:pos=\"576\"/>"
  )
  expect_equal(rtf_fp_tab(tab), "\\tx576\\tqc")
  expect_error(fp_tab(pos = 0.4, style = "blah"))
})


test_that("fp_tabs", {
  z <- fp_tabs(
    fp_tab(pos = 0.4, style = "decimal"),
    fp_tab(pos = 1, style = "decimal")
  )
  expect_equal(
    to_wml(z),
    "<w:tabs><w:tab w:val=\"decimal\" w:leader=\"none\" w:pos=\"576\"/><w:tab w:val=\"decimal\" w:leader=\"none\" w:pos=\"1440\"/></w:tabs>"
  )
})

library(officer)

sp_lineend()
sp_lineend(type = "triangle")
sp_lineend(type = "arrow", width = "lg", length = "lg")
print(sp_lineend(type = "triangle", width = "lg"))
obj <- sp_lineend(type = "triangle", width = "lg")
update(obj, type = "arrow")

#' @export
#' @title Empty block for 'PowerPoint'
#' @description Create an empty object to include as an empty placeholder shape in a
#' presentation. This comes in handy when presentation are updated
#' through R, but a user still wants to add some comments in this new content.
#'
#' Empty content also works with layout fields (slide number and date) to preserve them:
#' they are included on the slide and keep being updated by PowerPoint, i.e. update to the
#' when the slide number when the slide moves in the deck, update to the date.
#'
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' doc <- read_pptx()
#' doc <- add_slide(doc, layout = "Two Content",
#'   master = "Office Theme")
#' doc <- ph_with(x = doc, value = empty_content(),
#'  location = ph_location_type(type = "title") )
#'
#' doc <- add_slide(doc, "Title and Content")
#' # add slide number as a computer field
#' doc <- ph_with(
#'   x = doc, value = empty_content(),
#'   location = ph_location_type(type = "sldNum"))
#'
#' print(doc, target = fileout )
#' @seealso [ph_with()], [body_add_blocks()]
empty_content <- function() {
  x <- list()
  class(x) <- "empty_content"
  x
}

#' Change ph labels in a layout
#'
#' There are two versions of the function. The first takes a set of key-value pairs to rename the
#' ph labels. The second uses a right hand side (rhs) assignment to specify the new ph labels.
#' See section *Details*. \cr\cr
#' _NB:_ You can also rename ph labels directly in PowerPoint. Open the master template view
#' (`Alt` + `F10`) and go to `Home` > `Arrange` > `Selection Pane`.
#'
#' @details
#' * Note the difference between the terms `id` and `index`. Both can be found in the output of
#' [layout_properties()]. The unique ph `id` is found in column `id`. The `index` refers to the
#' index of the data frame row.
#' * In a right hand side (rhs) label assignment (`<- new_labels`), there are two ways to
#' optionally specify a subset of phs to rename. In both cases, the length of the rhs vector
#' (the new labels) must match the length of the id or index:
#'   1. use the `id` argument to specify ph ids to rename: `layout_rename_ph_labels(..., id = 2:3) <- new_labels`
#'   2. use an `index` in squared brackets: `layout_rename_ph_labels(...)[1:2] <- new_labels`
#'
#' @export
#' @rdname layout_rename_ph_labels
#' @param x An `rpptx` object.
#' @param layout Layout name or index. Index is the row index of [layout_summary()].
#' @param master Name of master. Only required if the layout name is not unique across masters.
#' @param ... Comma separated list of key-value pairs to rename phs. Either reference a ph via its label
#' (`"old label"` = `"new label"`) or its unique id (`"id"` = `"new label"`).
#' @param .dots Provide a named list or vector of key-value pairs to rename phs
#' (`list("old label"` = `"new label"`).
#' @param id Unique placeholder id (see column `id` in [layout_properties()] or [plot_layout_properties()]).
#' @param value Not relevant for user. A pure technical necessity for rhs assignments.
#' @return Vector of renamed ph labels.
#' @example inst/examples/example_layout_rename_ph_labels.R
#'
layout_rename_ph_labels <- function(x, layout, master = NULL, ..., .dots = NULL) {
  stop_if_not_rpptx(x, "x")
  dots <- list(...)
  dots <- c(dots, .dots)
  if (length(dots) > 0 && !is_named(dots)) {
    cli::cli_abort(
      c("Unnamed arguments are not allowed.",
        "x" = "Arguments {.arg ...} and {.arg .dots} both require key value pairs."
      ),
      call = NULL
    )
  }

  l <- get_layout(x, layout, master)
  lp <- layout_properties(x, l$layout_name, l$master_name)
  if (length(dots) == 0) {
    return(lp$ph_label)
  }
  df_renames <- .rename_df_from_dots(lp, dots)
  .set_ph_labels(l, df_renames)
  reload_slidelayouts(x)

  lp <- layout_properties(x, l$layout_name, l$master_name)
  invisible(lp$ph_label)
}


#' @export
#' @rdname layout_rename_ph_labels
`layout_rename_ph_labels<-` <- function(x, layout, master = NULL, id = NULL, value) {
  l <- get_layout(x, layout, master)
  lp <- layout_properties(x, l$layout_name, l$master_name)

  if (!is.null(id)) {
    if (length(id) != length(value)) {
      cli::cli_abort(
        c("{.arg id} and rhs vector must have the same length",
          "x" = "Number of ids ({.val {length(id)}}) and assigned values ({.val {length(value)}}) differ"
        )
      )
    }
    wrong_ids <- setdiff(id, lp$id)
    n_wrong <- length(wrong_ids)
    if (n_wrong > 0) {
      cli::cli_abort(c(
        "{cli::qty(n_wrong)} {?This/These} id{?s} {?does/do} not exist: {.val {wrong_ids}}",
        "x" = "Choose one of: {.val {lp$id}}",
        "i" = cli::col_grey("Also see {.code plot_layout_properties(..., '{l$layout_name}', '{l$master_nam}')}")
      ))
    }
    .idx <- match(id, lp$id) # user might enter ids in arbitrary order
    lp$ph_label[.idx] <- value
    value <- lp$ph_label
  }
  names(value) <- lp$id
  df_renames <- .rename_df_from_dots(lp, value)
  .set_ph_labels(l, df_renames)
  reload_slidelayouts(x)
}


# heuristic: if a number, then treat as ph_id
.detect_ph_id <- function(x) {
  suppressWarnings({ # avoid character to NA warning
    !is.na(as.numeric(x)) # nchar(x) == 1 &
  })
}


# create data frame with: ph_id, ph_label, ph_label_new as a
# basis for subsequent renaming
#
# CAVEAT: the ph order in layout_properties() (i.e. get_xfrm_data()) is reference for the user.
# Using the 'slide_layout' object's xfrm() method does not yield the same ph order!
# We need to guarantee a proper match here.
#
.rename_df_from_dots <- function(lp, dots) {
  lp <- lp[, c("id", "ph_label")]
  label_old <- names(dots)
  label_new <- as.character(dots)
  is_id <- .detect_ph_id(label_old)
  is_label <- !is_id

  # warn if renaming a duplicate label
  ii <- duplicated(lp$ph_label)
  dupes <- lp$ph_label[ii]
  dupes_used <- intersect(label_old, dupes)
  n_dupes_used <- length(dupes_used)
  if (n_dupes_used > 0) {
    cli::cli_warn(c(
      "When renaming a label with duplicates, only the first occurrence is renamed.",
      "x" = "Renaming {n_dupes_used} ph label{?s} with duplicates: {.val {dupes_used}}"
    ), call = NULL)
  }

  # check for duplicate renames
  is_dupe <- duplicated(label_old)
  if (any(is_dupe)) {
    dupes <- unique(label_old[is_dupe])
    n_dupes <- length(dupes)
    cli::cli_abort(c(
      "Each id or label must only have one rename entry only.",
      "x" = "Found {n_dupes} duplicate id{?s}/label{?s} to rename: {.val {dupes}}"
    ), call = NULL)
  }

  # match by label and check for unknown labels
  label_old_ <- label_old[is_label]
  row_idx_label <- match(label_old_, table = lp$ph_label)
  i_wrong <- is.na(row_idx_label)
  n_wrong <- sum(i_wrong)
  if (n_wrong > 0) {
    cli::cli_abort(c(
      "Can't rename labels that don't exist.",
      "x" = "{cli::qty(n_wrong)}{?This label does/These labels do} not exist: {.val {label_old_[i_wrong]}}"
    ), call = NULL)
  }

  # match by id and check for unknown ids
  id_old_ <- label_old[is_id]
  row_idx_id <- match(id_old_, table = lp$id)
  i_wrong <- is.na(row_idx_id)
  n_wrong <- sum(i_wrong)
  if (n_wrong > 0) {
    cli::cli_abort(c(
      "Can't rename ids that don't exist.",
      "x" = "{cli::qty(n_wrong)}{?This id does/These ids do} not exist: {.val {id_old_[i_wrong]}}"
    ), call = NULL)
  }

  # check for collision between label and id
  idx_collision <- intersect(row_idx_label, row_idx_id)
  n_collision <- length(idx_collision)
  if (n_collision > 0) {
    df <- lp[idx_collision, c("id", "ph_label")]
    pairs <- paste(df$ph_label, "<-->", df$id)
    cli::cli_abort(c(
      "Either specify the label {.emph OR} the id of the ph to rename, not both.",
      "x" = "These labels and ids collide: {.val {pairs}}"
    ), call = NULL)
  }

  lp$ph_label_new <- NA
  lp$ph_label_new[row_idx_label] <- label_new[is_label]
  lp$ph_label_new[row_idx_id] <- label_new[is_id]
  lp[!is.na(lp$ph_label_new), , drop = FALSE]
}


.set_ph_labels <- function(l, df_renames) {
  if (!inherits(l, "layout_info")) {
    cli::cli_abort(
      c("{.arg l} must a a {.cls layout_info} object",
        "x" = "Got {.cls {class(l)[1]}} instead"
      ),
      call = NULL
    )
  }
  layout_xml <- l$slide_layout$get()
  for (i in seq_len(nrow(df_renames))) {
    cnvpr_node <- xml2::xml_find_first(layout_xml, sprintf("p:cSld/p:spTree/*/p:nvSpPr/p:cNvPr[@id='%s']", df_renames$id[i]))
    xml2::xml_set_attr(cnvpr_node, "name", df_renames$ph_label_new[i])
  }
  l$slide_layout$save() # persist changes in slide layout xml file
}

# to_rtf ----
#' @export
#' @title Convert officer objects to RTF
#' @description Convert an object made with package officer
#' to RTF.
#' @param x object to convert to RTF. Supported
#' objects are:
#' - [ftext()]
#' - [external_img()]
#' - [run_autonum()]
#' - [run_columnbreak()]
#' - [run_linebreak()]
#' - [run_word_field()]
#' - [run_reference()]
#' - [run_pagebreak()]
#' - [hyperlink_ftext()]
#' - [block_list()]
#' - [fpar()]
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @return a string containing the RTF code
#' @keywords internal
to_rtf <- function(x, ...) {
  UseMethod("to_rtf")
}

## to_rtf blocks ----
#' @export
to_rtf.fpar <- function(x, ...) {
  chks <- fortify_fpar(x)
  rtf_chunks <- lapply(chks, to_rtf)
  rtf_chunks$collapse <- ""
  rtf_chunks <- do.call(paste0, rtf_chunks)
  paste0("{", ppr_rtf(x$fp_p), rtf_chunks, "\\par}")
}

#' @export
to_rtf.default <- function(x, ...) {
  if (is.null(x)) {
    return("")
  }
  as.character(x)
}

#' @export
to_rtf.block_list <- function(x, ...) {
  rtf_blocks <- lapply(x, to_rtf)
  rtf_blocks$collapse <- ""
  do.call(paste0, rtf_blocks)
}


## to_rtf chunks ----

#' @export
to_rtf.ftext <- function(x, ...) {
  rtf_fp_strings <- ""
  value <- str_encode_to_rtf(x$value)
  if (!is.null(x$pr)) {
    rtf_fp_strings <- rpr_rtf(x$pr)
    if (!is.na(x$pr$shading.color)) {
      value <- paste0(
        "%ftshading:",
        x$pr$shading.color,
        "%",
        value,
        "\\highlight0"
      )
    }
  }

  paste0(rtf_fp_strings, value)
}

#' @export
to_rtf.run_tab <- function(x, ...) {
  "\\tab "
}


#' @export
to_rtf.external_img <- function(x, ...) {
  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  if (!grepl("\\.png$", x)) {
    imgpath <- tempfile(fileext = ".png")
    img <- magick::image_read(x)
    magick::image_write(img, path = imgpath)
  } else {
    imgpath <- x
  }

  # https://stat.ethz.ch/pipermail/r-help/2017-September/449253.html
  dat <- readBin(imgpath, what = "raw", size = 1, endian = "little", n = 1e+8)
  dat <- paste(dat, collapse = "")

  rtf_str <- sprintf(
    "{\\pict\\pngblip\\picwgoal%.0f\\pichgoal%.0f ",
    width * 1440,
    height * 1440
  )
  paste(rtf_str, dat, "\n}", sep = "")
}


#' @export
to_rtf.floating_external_img <- function(x, ...) {
  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  pos <- attr(x, "pos")
  pos_x <- pos$x
  pos_y <- pos$y
  pos_h_from <- pos$h_from
  pos_v_from <- pos$v_from

  wrap <- attr(x, "wrap")
  wrap_type <- wrap$type
  wrap_side <- wrap$side
  wrap_dist_top <- wrap$dist_top
  wrap_dist_bottom <- wrap$dist_bottom
  wrap_dist_left <- wrap$dist_left
  wrap_dist_right <- wrap$dist_right

  if (!grepl("\\.png$", x)) {
    imgpath <- tempfile(fileext = ".png")
    img <- magick::image_read(x)
    magick::image_write(img, path = imgpath)
  } else {
    imgpath <- x
  }

  # Read image data
  dat <- readBin(imgpath, what = "raw", size = 1, endian = "little", n = 1e+8)
  dat <- paste(dat, collapse = "")

  # RTF units: twips (1/1440 inch)
  width_twips <- width * 1440
  height_twips <- height * 1440
  pos_x_twips <- pos_x * 1440
  pos_y_twips <- pos_y * 1440

  # EMU units for wrap distances (1 EMU = 1/914400 inch, so multiply inches by 914400)
  # But for RTF {\sp{\sn property}{\sv value}}, wrap distances are in EMUs
  dist_t_emus <- wrap_dist_top * 914400
  dist_b_emus <- wrap_dist_bottom * 914400
  dist_l_emus <- wrap_dist_left * 914400
  dist_r_emus <- wrap_dist_right * 914400

  # RTF wrap codes for \shpwr:
  # 1 = wrap around top and bottom
  # 2 = wrap around shape
  # 3 = none (as if shape isn't present)
  # 4 = wrap tightly
  # 5 = wrap text through shape
  wrap_code <- switch(
    wrap_type,
    "square" = "\\shpwr2",
    "tight" = "\\shpwr4",
    "topAndBottom" = "\\shpwr1",
    "through" = "\\shpwr5",
    "none" = "\\shpwr3",
    "\\shpwr2" # Default to square wrap
  )

  # RTF wrap side codes for \shpwrk:
  # 0 = wrap both sides
  # 1 = wrap left side only
  # 2 = wrap right side only
  # 3 = wrap only on largest side
  wrap_side_code <- switch(
    wrap_side,
    "bothSides" = "\\shpwrk0",
    "left" = "\\shpwrk1",
    "right" = "\\shpwrk2",
    "largest" = "\\shpwrk3",
    "\\shpwrk0" # Default
  )

  # RTF positioning codes
  # Horizontal positioning relative to:
  # \shpbxpage = page, \shpbxmargin = margin, \shpbxcolumn = column
  pos_h_code <- switch(
    pos_h_from,
    "page" = "\\shpbxpage",
    "margin" = "\\shpbxmargin",
    "column" = "\\shpbxcolumn",
    "character" = "\\shpbxcolumn",
    "\\shpbxmargin" # Default
  )

  # Vertical positioning relative to:
  # \shpbypage = page, \shpbymargin = margin, \shpbypara = paragraph
  pos_v_code <- switch(
    pos_v_from,
    "page" = "\\shpbypage",
    "margin" = "\\shpbymargin",
    "paragraph" = "\\shpbypara",
    "line" = "\\shpbypara",
    "\\shpbymargin" # Default
  )

  # posrelh property: 0=margin, 1=page, 2=column
  posrelh <- switch(
    pos_h_from,
    "margin" = 0,
    "page" = 1,
    "column" = 2,
    "character" = 2,
    0 # Default to margin
  )

  # posrelv property: 0=margin, 1=page, 2=paragraph
  posrelv <- switch(
    pos_v_from,
    "margin" = 0,
    "page" = 1,
    "paragraph" = 2,
    "line" = 2,
    0 # Default to margin
  )

  # Build the RTF shape structure following Word's format:
  # {\shp \shpleft... \shptop... \shpright... \shpbottom... \shpfhdr0
  #       \shpbx... \shpbxignore \shpby... \shpbyignore
  #       \shpwr... \shpwrk... \shpfblwtxt... \shpz... \shplid...
  #       {\*\shpinst {\sp{\sn prop}{\sv val}}... \par}}}

  rtf_str <- paste0(
    "{\\shp",
    sprintf(
      "\\shpleft%.0f\\shptop%.0f\\shpright%.0f\\shpbottom%.0f",
      pos_x_twips,
      pos_y_twips,
      pos_x_twips + width_twips,
      pos_y_twips + height_twips
    ),
    "\\shpfhdr0", # Not in header/footer
    pos_h_code,
    "\\shpbxignore", # Horizontal positioning
    pos_v_code,
    "\\shpbyignore", # Vertical positioning
    wrap_code,
    wrap_side_code, # Wrap type and side
    "\\shpfblwtxt0", # Image in front of text
    "\\shpz0", # Z-order
    "\\shplid1027" # Shape ID
  )

  # Start the shape properties ({\*\shpinst ...})
  rtf_str <- paste0(rtf_str, "{\\*\\shpinst")

  # Add shape properties following {\sp{\sn PropertyName}{\sv PropertyValue}} format
  # shapeType: 75 = picture frame
  rtf_str <- paste0(rtf_str, "{\\sp{\\sn shapeType}{\\sv 75}}")

  # pib: Binary picture data
  rtf_str <- paste0(
    rtf_str,
    sprintf(
      "{\\sp{\\sn pib}{\\sv {\\pict\\pngblip\\picwgoal%.0f\\pichgoal%.0f %s}}}",
      width_twips,
      height_twips,
      dat
    )
  )

  # Add wrap distances (in EMUs)
  if (dist_l_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dxWrapDistLeft}{\\sv %.0f}}", dist_l_emus)
    )
  }
  if (dist_r_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dxWrapDistRight}{\\sv %.0f}}", dist_r_emus)
    )
  }
  if (dist_t_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dyWrapDistTop}{\\sv %.0f}}", dist_t_emus)
    )
  }
  if (dist_b_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dyWrapDistBottom}{\\sv %.0f}}", dist_b_emus)
    )
  }

  # Add positioning properties (posrelh, posrelv are not in the Word example,
  # but keeping them won't hurt for compatibility with newer readers)
  # Note: These are commented out since the Word example doesn't include them
  # rtf_str <- paste0(rtf_str, sprintf("{\\sp{\\sn posrelh}{\\sv %d}}", posrelh))
  # rtf_str <- paste0(rtf_str, sprintf("{\\sp{\\sn posrelv}{\\sv %d}}", posrelv))

  # fBehindDocument: 0 = in front of text, 1 = behind text
  # Note: This is redundant with \shpfblwtxt0 above, but Word includes it
  # rtf_str <- paste0(rtf_str, "{\\sp{\\sn fBehindDocument}{\\sv 0}}")

  # Close the shape structure: \par}}} closes {\*\shpinst and {\shp
  paste0(rtf_str, "\\par}}}")
}


#' @export
to_rtf.run_word_field <- function(x, ...) {
  rtf_fp <- ""
  if (!is.null(x$pr)) {
    rtf_fp <- rpr_rtf(x$pr)
  }
  sprintf("%s{\\field{\\*\\fldinst %s}}", rtf_fp, x$field)
}

#' @export
to_rtf.run_pagebreak <- function(x, ...) {
  "\\page"
}
#' @export
to_rtf.run_columnbreak <- function(x, ...) {
  "\\column"
}
#' @export
to_rtf.run_linebreak <- function(x, ...) {
  "\\line"
}

#' @export
to_rtf.hyperlink_ftext <- function(x, ...) {
  rtf_fp <- ""
  if (!is.null(x$pr)) {
    rtf_fp <- rpr_rtf(x$pr)
  }
  sprintf(
    "%s{\\field{\\*\\fldinst HYPERLINK \"%s\"}{\\fldrslt %s}}",
    rtf_fp,
    officer_url_encode(x$href),
    str_encode_to_rtf(x$value)
  )
}
rtf_bookmark <- function(id, str) {
  bm_start_str <- sprintf("{\\*\\bkmkstart %s}", id)
  bm_start_end <- sprintf("{\\*\\bkmkend %s}", id)
  paste0(bm_start_str, str, bm_start_end)
}

#' @export
to_rtf.run_autonum <- function(x, ...) {
  if (is.null(x$seq_id)) {
    return("")
  }

  run_str_pre <- to_rtf(ftext(x$pre_label, prop = x$pr))
  run_str_post <- to_rtf(ftext(x$post_label, prop = x$pr))

  seq_str <- paste0("SEQ ", x$seq_id, " \\\\* Arabic")
  if (!is.null(x$start_at) && is.numeric(x$start_at)) {
    seq_str <- paste(seq_str, "\\r", as.integer(x$start_at))
  }

  sqf <- run_word_field(field = seq_str, prop = x$pr)
  sf_str <- to_rtf(sqf)

  if (x$tnd > 0) {
    z <- paste0(
      to_rtf(run_word_field(
        field = paste0("STYLEREF ", x$tnd, " \\r"),
        prop = x$pr
      )),
      to_rtf(ftext(x$tns, prop = x$pr))
    )
    sf_str <- paste0(z, sf_str)
  }

  if (!is.null(x$bookmark)) {
    if (x$bookmark_all) {
      out <- paste0(
        rtf_bookmark(id = x$bookmark, str = paste0(run_str_pre, sf_str)),
        run_str_post
      )
    } else {
      sf_str <- rtf_bookmark(id = x$bookmark, sf_str)
      out <- paste0(run_str_pre, sf_str, run_str_post)
    }
  } else {
    out <- paste0(run_str_pre, sf_str, run_str_post)
  }

  out
}

#' @export
to_rtf.run_reference <- function(x, ...) {
  out <- to_rtf(run_word_field(
    field = paste0(" REF ", x$id, " \\\\h "),
    prop = x$pr
  ))
  out
}

## to_rtf sections ----

#' @export
to_rtf.block_section <- function(x, ...) {
  paste0("\\sect", to_rtf(x$property))
}

#' @export
to_rtf.prop_section <- function(x, ...) {
  rtf_type <- if (is.null(x$type)) {
    ""
  } else if (x$type %in% "nextPage") {
    "\\sbkpage"
  } else if (x$type %in% "oddPage") {
    "\\sbkodd"
  } else if (x$type %in% "evenPage") {
    "\\sbkeven"
  } else {
    "\\sbknone"
  }

  hfr_type <- guess_hfr_type(x)
  extra_rtf <- ""
  str_h <- ""
  str_f <- ""
  str_h_even <- ""
  str_f_even <- ""
  str_h_first <- ""
  str_f_first <- ""

  if ("none" %in% hfr_type) {
    hfr_str <- ""
  } else if ("default" %in% hfr_type) {
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\header ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footer ", to_rtf(x$footer_default), "}")
    }
    hfr_str <- paste0(str_h, str_f)
  } else if ("evenodd" %in% hfr_type) {
    extra_rtf <- "\\facingp"
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\headerl ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$header_even)) {
      str_h_even <- paste0("{\\headerr ", to_rtf(x$header_even), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footerl ", to_rtf(x$footer_default), "}")
    }
    if (!is.null(x$footer_even)) {
      str_f_even <- paste0("{\\footerr ", to_rtf(x$footer_even), "}")
    }
    hfr_str <- paste0(
      str_h,
      str_h_even,
      str_f,
      str_f_even
    )
  } else if ("evenoddfirst" %in% hfr_type) {
    extra_rtf <- "\\facingp\\titlepg"
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\headerl ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$header_even)) {
      str_h_even <- paste0("{\\headerr ", to_rtf(x$header_even), "}")
    }
    if (!is.null(x$header_first)) {
      str_h_first <- paste0("{\\headerf ", to_rtf(x$header_first), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footerl ", to_rtf(x$footer_default), "}")
    }
    if (!is.null(x$footer_even)) {
      str_f_even <- paste0("{\\footerr ", to_rtf(x$footer_even), "}")
    }
    if (!is.null(x$footer_first)) {
      str_f_first <- paste0("{\\footerf ", to_rtf(x$footer_first), "}")
    }
    hfr_str <- paste0(
      str_h,
      str_h_even,
      str_h_first,
      str_f,
      str_f_even,
      str_f_first
    )
  } else {
    extra_rtf <- "\\titlepg"
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\headerl ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$header_first)) {
      str_h_first <- paste0("{\\headerf ", to_rtf(x$header_first), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footerl ", to_rtf(x$footer_default), "}")
    }
    if (!is.null(x$footer_first)) {
      str_f_first <- paste0("{\\footerf ", to_rtf(x$footer_first), "}")
    }
    hfr_str <- paste0(
      str_h,
      str_h_first,
      str_f,
      str_f_first
    )
  }

  rtf_str <- paste0(
    if (!is.null(x$page_margins)) to_rtf(x$page_margins),
    if (!is.null(x$page_size)) to_rtf(x$page_size),
    rtf_type,
    if (!is.null(x$section_columns)) to_rtf(x$section_columns),
    extra_rtf
  )

  # sections def and hdr/ftr content are separated in rtf
  attr(rtf_str, "hfr") <- hfr_str

  rtf_str
}

guess_hfr_type <- function(x) {
  has_default <- (!is.null(x$header_default) || !is.null(x$footer_default))
  has_even <- (!is.null(x$header_even) || !is.null(x$footer_even))
  has_first <- (!is.null(x$header_first) || !is.null(x$footer_first))

  if (has_default && !has_even && !has_first) {
    "default"
  } else if (has_default && has_even && !has_first) {
    "evenodd"
  } else if (has_default && has_even && has_first) {
    "evenoddfirst"
  } else if (has_default && !has_even && has_first) {
    "defaultfirst"
  } else {
    "none"
  }
}

#' @export
to_rtf.page_mar <- function(x, ...) {
  sprintf(
    "\\margb%.0f\\margt%.0f\\margr%.0f\\margl%.0f\\gutter%.0f\\headery%.0f\\footery%.0f",
    inch_to_tweep(x$bottom),
    inch_to_tweep(x$top),
    inch_to_tweep(x$right),
    inch_to_tweep(x$left),
    inch_to_tweep(x$gutter),
    inch_to_tweep(x$header),
    inch_to_tweep(x$footer)
  )
}

#' @export
to_rtf.page_size <- function(x, ...) {
  if (!is.null(x$width) && !is.null(x$height)) {
    out <- sprintf(
      "\\pghsxn%.0f\\pgwsxn%.0f%s",
      inch_to_tweep(x$height),
      inch_to_tweep(x$width),
      if ("landscape" %in% x$orient) {
        "\\lndscpsxn"
      } else {
        ""
      }
    )
  } else {
    out <- if ("landscape" %in% x$orient) {
      "\\lndscpsxn"
    } else {
      ""
    }
  }

  out
}

#' @export
to_rtf.section_columns <- function(x, ...) {
  widths <- x$widths * 20 * 72
  space <- x$space * 20 * 72

  columns_str <- sprintf("\\colw%.0f\\colsx%.0f", widths[length(widths)], space)

  linebetcol <- ""
  if (x$sep) {
    linebetcol <- "\\linebetcol"
  }

  sprintf(
    "\\cols%.0f%s%s",
    length(widths),
    linebetcol,
    columns_str
  )
}

# RTF management ----
font_table <- function(families) {
  data.frame(
    family = families,
    index = seq_along(families) - 1,
    match = sprintf("%%font:%s%%", families),
    stringsAsFactors = FALSE
  )
}
color_table <- function(colors) {
  data.frame(
    color = colors,
    index = seq_along(colors),
    match_ftcolor = sprintf("%%ftcolor:%s%%", colors),
    match_ftshading = sprintf("%%ftshading:%s%%", colors),
    match_ftlinecolor = sprintf("%%ftlinecolor:%s%%", colors),
    match_ftbgcolor = sprintf("%%ftbgcolor:%s%%", colors),
    stringsAsFactors = FALSE
  )
}

ppr_rtf <- function(x) {
  text_align_ <- ""
  if (!is.na(x$text.align)) {
    if ("left" %in% x$text.align) {
      text_align_ <- "\\ql"
    } else if ("center" %in% x$text.align) {
      text_align_ <- "\\qc"
    } else if ("right" %in% x$text.align) {
      text_align_ <- "\\qr"
    } else if ("justified" %in% x$text.align) {
      text_align_ <- "\\qj"
    }
  }

  tabs <- ""
  if (!is.null(x$tabs) && length(x$tabs) > 0 && !isFALSE(x$tabs)) {
    tabs <- paste0(sapply(x$tabs, rtf_fp_tab), collapse = "")
  }

  keep_with_next <- ""
  if (isTRUE(x$keep_with_next)) {
    keep_with_next <- "\\keepn"
  }
  leftright_padding <- ""
  if (!is.na(x$padding.left) && !is.na(x$padding.right)) {
    leftright_padding <- sprintf(
      "\\fi0\\li%.0f\\ri%.0f",
      x$padding.left * 20,
      x$padding.right * 20
    )
  }
  topbot_spacing <- ""
  if (!is.na(x$padding.bottom) && !is.na(x$padding.top)) {
    topbot_spacing <- sprintf(
      "\\sb%.0f\\sa%.0f",
      x$padding.bottom * 20,
      x$padding.top * 20
    )
  }

  line_spacing <- ""
  if (!is.na(x$line_spacing)) {
    line_spacing <- sprintf("\\sl%.0f\\slmult1", 240 * x$line_spacing)
  }

  out <- paste0(
    line_spacing,
    text_align_,
    tabs,
    keep_with_next,
    topbot_spacing,
    leftright_padding
  )
  if (nchar(out) > 0) {
    out <- paste0("\\pard", out, " ")
  }
  out
}

rpr_rtf <- function(x) {
  out <- ""

  if (!is.na(x$font.family)) {
    out <- paste0(out, "%font:", x$font.family, "%")
  }

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, "\\i")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, "\\b")
    }
  }
  if (!is.na(x$underlined)) {
    if (x$underlined) {
      out <- paste0(out, "\\ul")
    }
  }
  if (!is.na(x$strike)) {
    if (x$strike) {
      out <- paste0(out, "\\strike")
    }
  }

  if (x$vertical.align == "superscript" && !is.na(x$font.size)) {
    out <- paste0(out, sprintf("\\super\\up%.0f", round(x$font.size / 2)))
  } else if (x$vertical.align == "subscript" && !is.na(x$font.size)) {
    out <- paste0(out, sprintf("\\sub\\dn%.0f", round(x$font.size / 2)))
  }

  if (!is.na(x$font.size)) {
    out <- paste0(
      out,
      rtf_fontsize(x$font.size)
    )
  }

  if (!is.na(x$color)) {
    out <- paste0(
      out,
      paste0("%ftcolor:", x$color, "%")
    )
  }

  if (nchar(out) > 0) {
    out <- paste0(out, " ")
  }
  out
}

border_rtf <- function(x, side) {
  x$style[x$style %in% "solid"] <- "single"
  if (!x$style %in% c("dotted", "dashed", "single")) {
    x$style <- "single"
  }
  if (x$width < 0.0001 || is_transparent(x$color)) {
    return("")
  }

  linestyle <- c(
    "dotted" = "\\brdrdot",
    "dashed" = "\\brdrdash",
    "single" = "\\brdrs"
  )

  style_ <- linestyle[x$style]
  width_ <- paste0("\\brdrw", round(x$width * 20))
  color_ <- paste0("%ftlinecolor:", x$color, "%")

  paste0(
    "\\clbrdr",
    substr(side, 1, 1),
    style_,
    width_,
    color_
  )
}

tcpr_rtf <- function(x) {
  if (colalpha(x$background.color) > 0) {
    background.color <- paste0("%ftbgcolor:", x$background.color, "%")
  } else {
    background.color <- ""
  }
  vertalign <- c(
    "center" = "\\clvertalc",
    "top" = "\\clvertalt",
    "bottom" = "\\clvertalb"
  )
  vertical.align <- vertalign[x$vertical.align]

  text.direction <- ""
  text.direction[x$text.direction %in% "btlr"] <- "\\stextflow2"
  text.direction[x$text.direction %in% "tbrl"] <- "\\stextflow3"

  bb <- border_rtf(x$border.bottom, side = "bottom")
  bt <- border_rtf(x$border.top, side = "top")
  bl <- border_rtf(x$border.left, side = "left")
  br <- border_rtf(x$border.right, side = "right")

  rowspan <- ""
  if (x$rowspan > 1) {
    rowspan <- "\\clmgf"
  } else if (x$rowspan < 1) {
    rowspan <- "\\clmrg"
  }
  colspan <- ""
  if (x$colspan < 1) {
    colspan <- "\\clvmrg"
  } else {
    colspan <- "\\clvmgf"
  }

  paste0(
    bb,
    bt,
    bl,
    br,
    text.direction,
    rowspan,
    colspan,
    vertical.align,
    background.color
  )
}

# main ----
#' @export
#' @title Create an RTF document object
#' @description Creation of the object representing an
#' RTF document which can then receive contents with
#' the [rtf_add()] function and be written to a file with
#' the `print(x, target="doc.rtf")` function.
#' @param def_sec a [block_section] object used to defined default section.
#' @param normal_par an object generated by [fp_par()]
#' @param normal_chunk an object generated by [fp_text()]
#' @examples
#' rtf_doc(normal_par = fp_par(padding = 3))
#' @seealso [read_docx()], [print.rtf()], [rtf_add()]
#' @return an object of class `rtf` representing an
#' empty RTF document.
rtf_doc <- function(
  def_sec = prop_section(),
  normal_par = fp_par(),
  normal_chunk = fp_text(font.family = "Arial", font.size = 11)
) {
  styles <- data.frame(
    style_name = "Normal",
    style_id = 1L,
    type = "paragraph",
    rtf = rtf_par_style(
      fp_p = normal_par,
      fp_t = normal_chunk
    ),
    stringsAsFactors = FALSE
  )

  x <- list(
    content = list(),
    default_section = def_sec,
    styles = styles,
    normal_par = normal_par,
    normal_chunk = normal_chunk
  )
  class(x) <- "rtf"
  x
}

#' @export
#' @title Add content into an RTF document
#' @description This function add 'officer' objects into an RTF document.
#' Values are added as new paragraphs. See section 'Methods (by class)'
#' that list supported objects.
#' @param x rtf object, created by [rtf_doc()].
#' @param value object to add in the document. Supported objects
#' are vectors, graphics, block of formatted paragraphs. Use package
#' 'flextable' to add tables.
#' @param ... further arguments passed to or from other methods. When
#' adding a `ggplot` object or [plot_instr], these arguments will be used
#' by png function. See section 'Methods' to see what arguments can be used.
#' @example inst/examples/example_rtf_add.R
rtf_add <- function(x, value, ...) {
  UseMethod("rtf_add", value)
}


#' @export
#' @describeIn rtf_add add a new section definition
rtf_add.block_section <- function(x, value, ...) {
  x$content[[length(x$content) + 1]] <- value
  x
}

#' @export
#' @describeIn rtf_add add characters as new paragraphs
rtf_add.character <- function(x, value, ...) {
  values <- lapply(
    X = str_encode_to_rtf(value),
    function(z) {
      fpar(ftext(z, prop = x$normal_chunk), fp_p = x$normal_par)
    }
  )

  x$content <- append(x$content, values)
  x
}

#' @export
#' @describeIn rtf_add add a factor vector as new paragraphs
rtf_add.factor <- function(x, value, ...) {
  values <- lapply(
    X = str_encode_to_rtf(as.character(value)),
    function(z) {
      fpar(ftext(z, prop = x$normal_chunk), fp_p = x$normal_par)
    }
  )

  x$content <- append(x$content, values)
  x
}

#' @export
#' @describeIn rtf_add add a double vector as new paragraphs
#' @param formatter function used to format the numerical values
rtf_add.double <- function(x, value, formatter = formatC, ...) {
  values <- lapply(
    X = value,
    function(z) {
      fpar(ftext(formatC(z), prop = x$normal_chunk), fp_p = x$normal_par)
    }
  )

  x$content <- append(x$content, values)
  x
}

#' @export
#' @describeIn rtf_add add an [fpar()]
rtf_add.fpar <- function(x, value, ...) {
  x$content[[length(x$content) + 1]] <- value
  x
}

#' @export
#' @describeIn rtf_add add an [block_list()]
rtf_add.block_list <- function(x, value, ...) {
  x$content <- append(x$content, as.list(value))
  x
}


#' @export
#' @describeIn rtf_add add a ggplot2
#' @param width height in inches
#' @param height height in inches
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
#' @param ppr [fp_par()] to apply to paragraph.
rtf_add.gg <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  ppr = fp_par(text.align = "center"),
  ...
) {
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      print(value)
    },
    finally = {
      dev.off()
    }
  )

  value <- fpar(
    external_img(
      src = file,
      width = width,
      height = height,
      guess_size = FALSE
    ),
    fp_p = ppr
  )

  x$content[[length(x$content) + 1]] <- value
  x
}
#' @export
#' @describeIn rtf_add add a [plot_instr()] object
rtf_add.plot_instr <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  ppr = fp_par(text.align = "center"),
  ...
) {
  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )

  value <- fpar(
    external_img(
      src = file,
      width = width,
      height = height,
      guess_size = FALSE
    ),
    fp_p = ppr
  )

  x$content[[length(x$content) + 1]] <- value
  x
}

#' @title Write an 'RTF' File
#' @description Write the RTF object and its content to a file.
#' @param x an 'rtf' object created with [rtf_doc()]
#' @param target path to the RTF file to write
#' @param ... unused
#' @examples
#' # write a rdocx object in a rtf file ----
#' doc <- rtf_doc()
#' print(doc, target = tempfile(fileext = ".rtf"))
#' @export
#' @seealso [rtf_doc()]
print.rtf <- function(x, target = NULL, ...) {
  if (is.null(target)) {
    cat("rtf document with", length(x$content), "element(s)\n")
    return(invisible())
  }

  rtf_content <- sapply(x$content, to_rtf)
  rtf_ss <- rtf_stylesheet(x)
  txt_to_scan <- c(rtf_content, rtf_ss)
  m <- gregexec(pattern = "%font:[a-zA-Z ]+%", text = txt_to_scan)
  fonts <- unique(unlist(regmatches(txt_to_scan, m)))
  fonts <- gsub("(^%font:|%$)", "", fonts)
  family_table <- font_table(families = fonts)

  m <- gregexec(pattern = "\\%ftcolor\\:[a-zA-Z #0-9]+\\%", text = txt_to_scan)
  ftcolor <- unlist(regmatches(txt_to_scan, m))
  m <- gregexec(
    pattern = "\\%ftshading\\:[a-zA-Z #0-9]+\\%",
    text = txt_to_scan
  )
  ftshading <- unlist(regmatches(txt_to_scan, m))
  m <- gregexec(
    pattern = "\\%ftlinecolor\\:[a-zA-Z #0-9]+\\%",
    text = txt_to_scan
  )
  ftlinecolor <- unlist(regmatches(txt_to_scan, m))
  m <- gregexec(
    pattern = "\\%ftbgcolor\\:[a-zA-Z #0-9]+\\%",
    text = txt_to_scan
  )
  ftbgcolor <- unlist(regmatches(txt_to_scan, m))

  supported_highlight <- c(
    "black",
    "blue",
    "cyan",
    "green",
    "magenta",
    "red",
    "yellow",
    "darkblue",
    "darkcyan",
    "darkgreen",
    "darkmagenta",
    "darkred",
    "#8B8000",
    "darkgrey",
    "lightgray"
  )

  colors <- unique(c(
    ftcolor,
    ftshading,
    ftlinecolor,
    ftbgcolor,
    supported_highlight
  ))
  colors <- gsub(
    "(^%(ftcolor|ftshading|ftlinecolor|ftbgcolor):|%$)",
    "",
    colors
  )
  color_table <- color_table(colors = colors)

  header <- c(
    "{\\rtf1\\deff0",
    rtfize_font_table(family_table),
    rtfize_color_table(color_table)
  )

  default_sect <- to_rtf(x$default_section)
  all_strings <- c(
    header,
    rtf_ss,
    default_sect,
    rtf_content,
    attr(default_sect, "hfr"),
    "}"
  )

  all_strings <- fix_font_ref(all_strings, family_table)
  all_strings <- fix_font_color(all_strings, color_table)
  all_strings <- fix_line_color(all_strings, color_table)
  all_strings <- fix_bg_color(all_strings, color_table)
  all_strings <- fix_font_shading(all_strings, color_table)

  writeLines(all_strings, target, useBytes = TRUE)
  target
}

#' @export
#' @title Encode UTF8 string to RTF
#' @description Convert strings to RTF valid codes.
#' @param z character vector to be converted
#' @family functions for officer extensions
#' @return character vector of results encoded to RTF
#' @keywords internal
#' @examples
#' str_encode_to_rtf("Hello World")
str_encode_to_rtf <- function(z) {
  # https://stackoverflow.com/questions/1368020/how-to-output-unicode-string-to-rtf-using-c
  char_list <- strsplit(z, "")
  int_list <- lapply(z, utf8ToInt)
  rtf_strings <- mapply(
    function(charv, intv) {
      is_tobeprotected <- charv %in% c("\\", "{", "}")
      # is_std_ansi <- intv < 128
      # is_ext_ansi <- !is_std_ansi & intv < 256
      is_ext_ansi <- intv < 128
      is_unic2b <- !is_ext_ansi & intv < 32768
      is_unic4b <- !is_ext_ansi & !is_unic2b
      charv[is_unic2b] <- paste0("\\uc1\\u", intv[is_unic2b], "?")
      charv[is_unic4b] <- paste0("\\uc1\\u", intv[is_unic4b] - 65536, "?")
      paste0(charv, collapse = "")
    },
    char_list,
    int_list,
    SIMPLIFY = TRUE
  )
  as.character(rtf_strings)
}


# tools ----
rtf_fontsize <- function(x) {
  size <- formatC(x * 2, format = "f", digits = 0)
  z <- sprintf("\\fs%s", size)
  z[is.na(x)] <- ""
  z
}
rtf_color <- function(x, color_table, mark = "cf") {
  index <- color_table$index[match(x, color_table$color)]
  z <- sprintf("\\%s%.0f", mark, index)
  z[is.na(index)] <- ""
  z
}

rtfize_font_table <- function(x) {
  z <- sprintf("{\\f%.0f %s;}", x$index, x$family)
  z <- paste0(z, collapse = "")
  paste0("{\\fonttbl", z, "}")
}

rtf_color_code <- function(color) {
  color <- as.vector(col2rgb(color, alpha = FALSE))
  sprintf(
    "\\red%.0f\\green%.0f\\blue%.0f;",
    color[1],
    color[2],
    color[3]
  )
}

rtf_fp_tab <- function(x) {
  tab_style <- "\\tqdec"
  if (x$style %in% "left") {
    tab_style <- ""
  } else if (x$style %in% "center") {
    tab_style <- "\\tqc"
  } else if (x$style %in% "right") {
    tab_style <- "\\tqr"
  }

  sprintf(
    "\\tx%.0f%s",
    inch_to_tweep(x$pos),
    tab_style
  )
}

rtfize_color_table <- function(x) {
  z <- vapply(x$color, rtf_color_code, "")
  z <- paste0(z, collapse = "")
  paste0("{\\colortbl;", z, "}")
}

fix_font_ref <- function(x, font_tbl) {
  for (i in seq_len(nrow(font_tbl))) {
    m <- gregexec(pattern = font_tbl$match[i], text = x)
    regmatches(x, m) <- paste0("\\f", font_tbl$index[i])
  }
  x
}
fix_font_color <- function(x, color_tbl) {
  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftcolor[i], text = x)
    regmatches(x, m) <- paste0("\\cf", color_tbl$index[i])
  }
  x
}
fix_line_color <- function(x, color_tbl) {
  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftlinecolor[i], text = x)
    regmatches(x, m) <- paste0("\\brdrcf", color_tbl$index[i])
  }
  x
}
fix_bg_color <- function(x, color_tbl) {
  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftbgcolor[i], text = x)
    regmatches(x, m) <- paste0("\\clcbpat", color_tbl$index[i])
  }
  x
}
fix_font_shading <- function(x, color_tbl) {
  supported_highlight <- c(
    "black",
    "blue",
    "cyan",
    "green",
    "magenta",
    "red",
    "yellow",
    "darkblue",
    "darkcyan",
    "darkgreen",
    "darkmagenta",
    "darkred",
    "#8B8000",
    "darkgrey",
    "lightgray"
  )

  supported_highlight <-
    data.frame(
      colors = supported_highlight,
      index = seq_along(supported_highlight),
      match_ftshading = sprintf("%%ftshading:%s%%", supported_highlight)
    )

  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftshading[i], text = x)
    matches <- Filter(function(x) x > -1, m)
    if (length(matches) > 0) {
      used_shading <- unique(unlist(regmatches(x, m)))
      if (
        !used_shading %in%
          c(supported_highlight$match_ftshading, "%ftshading:transparent%")
      ) {
        stop(
          "chunk highlight color (",
          used_shading,
          ") in RTF can only have values: ",
          paste0(shQuote(supported_highlight$colors), collapse = ", ")
        )
      }
    }

    regmatches(x, m) <- sprintf("\\highlight%.0f", color_tbl$index[i])
  }
  x <- gsub("%ftshading:transparent%", "", x, fixed = TRUE)
  x
}

rtf_par_style <- function(fp_p = fp_par(), fp_t = NULL) {
  if (!is.null(fp_t)) {
    fp_t_rtf <- rpr_rtf(fp_t)
  } else {
    fp_t_rtf <- ""
  }

  paste0(ppr_rtf(fp_p), fp_t_rtf)
}

# Not used in {officer}
rtf_set_paragraph_style <- function(
  x,
  style_name,
  fp_p = fp_par(),
  fp_t = NULL
) {
  index <- which(x$styles$style_name %in% style_name)
  style_id <- if (length(index) < 1) {
    style_id <- nrow(x$styles) + 1L
  } else {
    style_id <- index
  }
  new_style <- data.frame(
    style_name = style_name,
    style_id = style_id,
    type = "paragraph",
    rtf = rtf_par_style(fp_p = fp_p, fp_t = fp_t),
    stringsAsFactors = FALSE
  )

  if (length(index) < 1) {
    x$styles <- rbind(x$styles, new_style)
  } else {
    x$styles[index, "rtf"] <- new_style$rtf
  }

  x
}

rtf_stylesheet <- function(x) {
  out <- character(nrow(x$styles))
  for (i in seq_len(nrow(x$styles))) {
    out[i] <- sprintf(
      "\n{\\f%.0f\\s%.0f%s %s;}",
      i,
      i,
      x$styles$rtf[i],
      x$styles$style_name[i]
    )
  }
  paste0(
    "{\\stylesheet ",
    paste0(out, collapse = ""),
    "}"
  )
}


# check for gregexec -----
if (!"gregexec" %in% getNamespaceExports("base")) {
  # copied from R source, grep.R
  gregexec <- function(
    pattern,
    text,
    ignore.case = FALSE,
    perl = FALSE,
    fixed = FALSE,
    useBytes = FALSE
  ) {
    if (is.factor(text) && length(levels(text)) < length(text)) {
      out <- gregexec(
        pattern,
        c(levels(text), NA_character_),
        ignore.case,
        perl,
        fixed,
        useBytes
      )
      outna <- out[length(out)]
      out <- out[text]
      out[is.na(text)] <- outna
      return(out)
    }

    dat <- gregexpr(
      pattern = pattern,
      text = text,
      ignore.case = ignore.case,
      fixed = fixed,
      useBytes = useBytes,
      perl = perl
    )
    if (perl && !fixed) {
      ## Perl generates match data, so use that
      capt.attr <- c("capture.start", "capture.length", "capture.names")
      process <- function(x) {
        if (anyNA(x) || any(x < 0)) {
          y <- x
        } else {
          ## Interleave matches with captures
          y <- t(cbind(x, attr(x, "capture.start")))
          attributes(y)[names(attributes(x))] <- attributes(x)
          ml <- t(cbind(attr(x, "match.length"), attr(x, "capture.length")))
          nm <- attr(x, "capture.names")
          ## Remove empty names that `gregexpr` returns
          dimnames(ml) <- dimnames(y) <-
            if (any(nzchar(nm))) list(c("", nm), NULL)
          attr(y, "match.length") <- ml
          y
        }
        attributes(y)[capt.attr] <- NULL
        y
      }
      lapply(dat, process)
    } else {
      ## For TRE or fixed we must compute the match data ourselves
      m1 <- lapply(
        regmatches(text, dat),
        regexec,
        pattern = pattern,
        ignore.case = ignore.case,
        perl = perl,
        fixed = fixed,
        useBytes = useBytes
      )
      mlen <- lengths(m1)
      res <- vector("list", length(m1))
      im <- mlen > 0
      res[!im] <- dat[!im] # -1, NA
      res[im] <- Map(
        function(outer, inner) {
          tmp <- do.call(cbind, inner)
          attributes(tmp)[names(attributes(inner))] <- attributes(inner)
          attr(tmp, "match.length") <-
            do.call(cbind, lapply(inner, `attr`, "match.length"))
          # useBytes/index.type should be same for all so use outer vals
          attr(tmp, "useBytes") <- attr(outer, "useBytes")
          attr(tmp, "index.type") <- attr(outer, "index.type")
          tmp + rep(outer - 1L, each = nrow(tmp))
        },
        dat[im],
        m1[im]
      )
      res
    }
  }
}

library(officer)

# Example 1: Mixing different section layouts ----
# This example demonstrates how to create a document with multiple sections,
# each with different page orientations and column layouts

# Define a landscape section with single column
# This is useful for wide tables or charts
landscape_one_column <- block_section(
  prop_section(
    page_size = page_size(orient = "landscape"),
    type = "continuous"
  )
)

# Define a landscape section with two columns
# Useful for text-heavy content in landscape mode (e.g., newsletters)
landscape_two_columns <- block_section(
  prop_section(
    page_size = page_size(orient = "landscape"),
    type = "continuous",
    section_columns = section_columns(widths = c(4.75, 4.75))
  )
)

# Create a new document
doc_1 <- read_docx()

# Section 1: Landscape single column for wide table ----
# Add a title for the first section
doc_1 <- body_add_par(doc_1, "Wide Table Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "This table is displayed in landscape orientation to accommodate all columns."
)

# Add a wide table with multiple columns
doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")

# End the landscape single-column section
doc_1 <- body_end_block_section(doc_1, value = landscape_one_column)

# Section 2: Landscape two columns for text content ----
# Add a title for the two-column section
doc_1 <- body_add_par(doc_1, "Two-Column Text Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "The following text flows across two columns in landscape orientation."
)

# Add text content that will flow across two columns
doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 50), collapse = " "))

# End the landscape two-column section
doc_1 <- body_end_block_section(doc_1, value = landscape_two_columns)

# Section 3: Return to portrait orientation ----
# After ending the previous sections, we're back to portrait (default)
doc_1 <- body_add_par(doc_1, "Portrait Table Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "This section returns to portrait orientation with a taller table."
)

# Add a longer table in portrait orientation
doc_1 <- body_add_table(doc_1, value = mtcars[1:25, ], style = "table_template")

# Save the document
output_file_1 <- tempfile(fileext = ".docx")
print(doc_1, target = output_file_1)


# Example 2: Different headers and footers (first, even, odd pages) ----
# This example demonstrates the complete header/footer system with:
# - Different header/footer for the first page
# - Different header/footer for even pages (left-side pages in duplex printing)
# - Default header/footer for odd pages (right-side pages)

# Create sample text to generate multiple pages
lorem_text <- paste(
  rep("Purus lectus eros metus turpis mattis platea praesent sed. ", 50),
  collapse = ""
)

# Define content for FIRST page header
# Typically used for title pages or cover pages
header_first <- block_list(
  fpar(
    ftext(
      "First Page Header - Title Page",
      fp_text_lite(bold = TRUE, color = "#4472C4", font.size = 14)
    ),
    fp_p = fp_par(
      text.align = "center",
      padding.bottom = 12,
      border.bottom = fp_border(color = "#4472C4", width = 2)
    )
  )
)

# Define content for EVEN pages header (left-side pages when printed)
# In duplex printing, this appears on the left side
header_even <- block_list(
  fpar(
    ftext("Chapter Title", fp_text_lite(italic = TRUE, font.size = 10)),
    fp_p = fp_par(text.align = "left")
  )
)

# Define content for DEFAULT pages header (odd pages/right-side)
# In duplex printing, this appears on the right side
header_default <- block_list(
  fpar(
    ftext("Document Title", fp_text_lite(italic = TRUE, font.size = 10)),
    fp_p = fp_par(text.align = "right")
  )
)

# Define content for FIRST page footer
footer_first <- block_list(
  fpar(
    ftext(
      "Company Name - Confidential",
      fp_text_lite(font.size = 9, color = "#666666")
    ),
    fp_p = fp_par(text.align = "center")
  )
)

# Define content for EVEN pages footer (includes page number on left)
footer_even <- block_list(
  fpar(
    run_word_field(field = "PAGE", prop = fp_text_lite(font.size = 9)),
    ftext(" | Document Name", fp_text_lite(font.size = 9)),
    fp_p = fp_par(
      text.align = "left",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Define content for DEFAULT pages footer (includes page number on right)
footer_default <- block_list(
  fpar(
    ftext("Document Name | ", fp_text_lite(font.size = 9)),
    run_word_field(field = "PAGE", prop = fp_text_lite(font.size = 9)),
    fp_p = fp_par(
      text.align = "right",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Create section properties with all header/footer variants
# When all three are defined (first, even, default), Word will use:
# - header_first/footer_first for page 1
# - header_even/footer_even for pages 2, 4, 6, etc.
# - header_default/footer_default for pages 3, 5, 7, etc.
section_with_all_hf <- prop_section(
  header_default = header_default,
  footer_default = footer_default,
  header_first = header_first,
  footer_first = footer_first,
  header_even = header_even,
  footer_even = footer_even
)

# Create a new document
doc_2 <- read_docx()

# Add enough content to create multiple pages
# This will demonstrate how the different headers/footers appear
for (i in 1:20) {
  doc_2 <- body_add_par(doc_2, paste0("Paragraph ", i, ": ", lorem_text))
}

# Apply the section properties with all header/footer configurations
doc_2 <- body_set_default_section(doc_2, value = section_with_all_hf)

# Save the document
# Open this document and scroll through pages to see:
# - Page 1: Special first page header/footer
# - Page 2: Even page header/footer with page number on left
# - Page 3: Odd page (default) header/footer with page number on right
# - And so on...
output_file_2 <- tempfile(fileext = ".docx")
print(doc_2, target = output_file_2)

#' @export
#' @title Create a 'PowerPoint' document object
#' @description Read and import a pptx file as an R object
#' representing the document.
#'
#' The function is called `read_pptx` because it allows you to initialize an
#' object of class `rpptx` from an existing PowerPoint file. Content will be
#' added to the existing presentation. By default, an empty document is used.
#'
#' @param path path to the pptx file to use as base document. `potx` file are supported.
#' @section master layouts and slide layouts:
#' `read_pptx()` uses a PowerPoint file as the initial document. This is the original
#' PowerPoint document where all slide layouts, placeholders for shapes and
#' styles come from. Major points to be aware of are:
#'
#'   * Slide layouts are relative to a master layout. A document can contain one or
#' more master layouts; a master layout can contain one or more slide layouts.
#' * A slide layout inherits design properties from its master layout but some
#' properties can be overwritten.
#' * Designs and formatting properties of layouts and shapes (placeholders in a
#'                                                            layout) are defined within the initial document. There is no R function to
#' modify these values - they must be defined in the initial document.
#' @examples
#' read_pptx()
#' @seealso [print.rpptx()], [add_slide()], [plot_layout_properties()], [ph_with()]
read_pptx <- function( path = NULL ){

  if( !is.null(path) && !file.exists(path))
    stop("could not find file ", shQuote(path), call. = FALSE)

  if( is.null(path) )
    path <- system.file(package = "officer", "template/template.pptx")

  if(!grepl("\\.(pptx|potx)$", path, ignore.case = TRUE)){
    stop("read_pptx only support pptx files", call. = FALSE)
  }

  package_dir <- tempfile()
  unpack_folder( file = path, folder = package_dir )

  obj <- list(package_dir = package_dir)


  obj$table_styles <- read_table_style(package_dir)

  obj$presentation <- presentation$new(package_dir)

  obj$masterLayouts <- dir_master$new(package_dir, slide_master$new("ppt/slideMasters") )

  obj$slideLayouts <- dir_layout$new( package_dir,
                                      master_metadata = obj$masterLayouts$get_metadata(),
                                      master_xfrm = obj$masterLayouts$xfrm() )

  obj$slide <- dir_slide$new( package_dir, obj$slideLayouts$get_xfrm_data() )
  obj$notesSlide <- dir_notesSlide$new(package_dir)
  obj$notesMaster <- dir_notesMaster$new(package_dir, slide_master$new("ppt/notesMasters"))
  obj$content_type <- content_type$new( package_dir )
  obj$core_properties <- read_core_properties(package_dir)
  obj$doc_properties_custom <- read_custom_properties(package_dir)

  obj$rel <- relationship$new()
  obj$rel$feed_from_xml(file.path(package_dir, "_rels", ".rels"))


  obj$cursor = obj$slide$length()
  class(obj) <- "rpptx"
  obj
}

read_table_style <- function(path){
  file <- file.path(path, "ppt/tableStyles.xml")
  if (!file.exists(file)) {
    warning("tableStyles.xml file does not exist in PPTX")
    return(NULL)
  }
  doc <- read_xml(file)
  nodes <- xml_find_all(doc, "//a:tblStyleLst")
  data.frame(def = xml_attr(nodes, "def"),
             styleName = xml_attr(nodes, "styleName"),
             stringsAsFactors = FALSE )
}

#' @title Write a 'PowerPoint' file.
#' @description Write a 'PowerPoint' file
#' with an object of class 'rpptx' (created with
#' [read_pptx()]).
#' @param x an rpptx object
#' @param target path to the pptx file to write
#' @param ... unused
#' @examples
#' # write a rdocx object in a pptx file ----
#' file <- tempfile(fileext = ".pptx")
#' doc <- read_pptx()
#' print(doc, target = file)
#' @export
#' @seealso \code{\link{read_pptx}}
print.rpptx <- function(x, target = NULL, ...){

  if( is.null( target) ){
    cat("pptx document with", length(x), "slide(s)\n")
    cat("Available layouts and their associated master(s) are:\n")
    print(as.data.frame( layout_summary(x)) )
    return(invisible())
  }

  if( !grepl(x = target, pattern = "\\.(pptx)$", ignore.case = TRUE) )
    stop(target , " should have '.pptx' extension.")

  if (is_windows() && is_doc_open(target)) {
    stop(target, " is open. To write to this document, please, close it.")
  }
  x <- pptx_fortify_slides(x)
  x$rel$write(file.path(x$package_dir, "_rels", ".rels"))

  # viewProps - drop lastView if available
  viewProps <- read_xml(file.path(x$package_dir, "ppt/viewProps.xml"))
  if (!is.na(xml_attr(viewProps, "lastView"))) {
    xml_attr(viewProps, "lastView") <- NULL
  }
  write_xml(viewProps, file.path(x$package_dir, "ppt/viewProps.xml"))

  drop_templatenode_from_app(x$package_dir)

  x$presentation$save()
  x$content_type$save()

  x$slide$save_slides()

  x$notesSlide$save_slides()

  x$core_properties['modified','value'] <- format( Sys.time(), "%Y-%m-%dT%H:%M:%SZ")
  x$core_properties['lastModifiedBy','value'] <- Sys.getenv("USER")
  write_core_properties(x$core_properties, x$package_dir)

  if(nrow(x$doc_properties_custom$data) >0 ){
    write_custom_properties(x$doc_properties_custom, x$package_dir)
  }

  invisible(pack_folder(folder = x$package_dir, target = target ))
}

#' @importFrom xml2 xml_children
read_custom_properties <- function(package_dir) {
  filename <- file.path(package_dir, "docProps/custom.xml")
  if (!file.exists(filename)) {
    filename <- system.file(package = "officer", "template/custom.xml")
  }
  doc <- read_xml(filename)
  all_children <- xml_children(doc)

  pid_values <- vapply(all_children, xml_attr, NA_character_, "pid")
  name_values <- vapply(all_children, xml_attr, NA_character_, "name")
  properties_types <- vapply(
    all_children,
    function(x) {
      xml_name(xml_child(x, 1))
    },
    NA_character_
  )

  # this complicated thing is there because some unidentifed
  # softwares fill this fields with invalid xml - so we have
  # to consider it will not be correctly rewritten and will
  # make an invalid document
  value_values <- vapply(
    all_children,
    function(x) {
      as.character(xml_child(x, 1))
    },
    NA_character_
  )
  pat1 <- sprintf("<vt\\:(%s)/>", paste0(properties_types, collapse = "|"))
  pat2 <- sprintf("<vt\\:(%s)>", paste0(properties_types, collapse = "|"))
  pat3 <- sprintf("</vt\\:(%s)>", paste0(properties_types, collapse = "|"))
  value_values <- gsub(pat1, "", value_values)
  value_values <- gsub(pat2, "", value_values)
  value_values <- gsub(pat3, "", value_values)
  str <- c(pid_values, name_values, properties_types, value_values)
  z <- matrix(
    str,
    ncol = 4,
    dimnames = list(NULL, c("pid", "name", "type", "value"))
  )
  z <- list(data = z)
  class(z) <- "custom_properties"
  z
}


#' @export
`[<-.custom_properties` <- function(x, i, j, value) {
  if (!i %in% x$data[, "name"]) {
    if (nrow(x$data) < 1) {
      pid <- 2
    } else {
      pid <- max(as.integer(x$data[, "pid"])) + 1L
    }
    new <- matrix(c(as.character(pid), i, "lpwstr", value), ncol = 4)
    x$data <- rbind(x$data, new)
  } else {
    x$data[x$data[, "name"] %in% i, j] <- value
  }
  x
}


#' @export
`[.custom_properties` <- function(x, i, j) {
  if (nrow(x$data) < 1) {
    return(character())
  }
  if (missing(i)) {
    x$data[, j]
  } else {
    x$data[x$data[, "name"] %in% i, j]
  }
}

write_custom_properties <- function(custom_props, package_dir) {
  xml_props <- sprintf(
    "<property fmtid=\"{D5CDD505-2E9C-101B-9397-08002B2CF9AE}\" pid=\"%s\" name=\"%s\"><vt:%s>%s</vt:%s></property>",
    custom_props$data[, 1],
    custom_props$data[, 2],
    custom_props$data[, 3],
    custom_props$data[, 4],
    custom_props$data[, 3]
  )

  xml_ <- c(
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
    "<Properties ",
    "xmlns=\"http://schemas.openxmlformats.org/officeDocument/2006/custom-properties\" ",
    "xmlns:vt=\"http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes\">",
    xml_props,
    "</Properties>"
  )
  props_dir <- file.path(package_dir, "docProps")
  dir.create(props_dir, recursive = TRUE, showWarnings = FALSE)
  filename <- file.path(props_dir, "custom.xml")
  writeLines(enc2utf8(xml_), filename, useBytes = TRUE)
  invisible()
}

check_dep <- function(){
  if( !requireNamespace("rmarkdown") )
    stop("package rmarkdown is required to use this function", call. = FALSE)
  if( !requireNamespace("knitr") )
    stop("package knitr is required to use this function", call. = FALSE)
}

get_default_pandoc_data_file <- function(format = "pptx") {
  outfile <- tempfile(fileext = paste0(".", format))

  pandoc_exec <- rmarkdown::pandoc_exec()
  if(.Platform$OS.type %in% "windows"){
    pandoc_exec <- paste0(pandoc_exec, ".exe")
  }

  if(!rmarkdown::pandoc_available() || !file.exists(pandoc_exec)){
    # Use officer template when no pandoc - stupid fallback as if o pandoc, no result
    outfile <- system.file(package = "officer", "template", paste0("template.", format))
  } else {
    # note in that case outfile will be removed when session will end
    ref_doc <- paste0("reference.", format)
    system2(pandoc_exec,
            args = c("--print-default-data-file", ref_doc),
            stdout = outfile)
  }

  return(outfile)
}

#' @export
#' @title Get the document being used as a template
#' @description Get filename of the document
#' being used as a template in an R Markdown document
#' rendered as HTML, PowerPoint presentation or Word document. It requires
#' packages rmarkdown >= 1.10.14 and knitr.
#' @param format document format, one of 'pptx', 'docx' or 'html'
#' @return a name file
#' @importFrom utils compareVersion packageVersion
#' @family functions for officer extensions
#' @keywords internal
get_reference_value <- function(format = NULL) {
  if (!is.null(format) && length(format) != 1) {
    stop("format must be a scalar character")
  }

  check_dep()

  if (compareVersion(as.character(packageVersion("rmarkdown")), "1.10.14") < 0) {
    stop("package rmarkdown >= 1.10.14 is required to use this function")
  }

  if (is.null(format)) {
    if (grepl("docx", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "docx"
    } else if (grepl("pptx", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "pptx"
    } else if (grepl("html", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "html"
    } else if (grepl("latex", knitr::opts_knit$get("rmarkdown.pandoc.to"))) {
      format <- "latex"
    } else {
      stop("Unable to determine the format that should be used")
    }
  }
  if (!format %in% c("pptx", "docx", "html")) {
    stop("format must be have value 'docx', 'pptx' or 'html'.")
  }

  output.dir <- knitr::opts_knit$get("output.dir")
  if (is.null(output.dir)) {
    output.dir <- getwd()
  }

  pandoc_args <- knitr::opts_knit$get("rmarkdown.pandoc.args")

  rd <- grep("--reference-doc", pandoc_args)
  if (length(rd)) {
    reference_data <- pandoc_args[rd + 1]
    if (!file.exists(reference_data)) {
      reference_data <- file.path(output.dir, reference_data)
    }
  } else {
    reference_data <- get_default_pandoc_data_file(format = format)
  }
  return(normalizePath(reference_data, winslash = "/"))
}


knitr_opts_current <- function(x, default = FALSE){
  check_dep()
  x <- knitr::opts_current$get(x)
  if(is.null(x)) x <- default
  x
}

#' @export
#' @title Table options in a 'knitr' context
#' @description Get options for table rendering in a 'knitr'
#' context. It should not be used by the end user, but
#' its documentation should be read as a place where
#' table options are documented when 'knitr' is used.
#'
#' The function is a utility to facilitate the retrieval of table
#' options supported by the 'flextable', 'officedown' and 'officer' packages.
#'
#' These options should be set with `knitr::opts_chunk$set()`.
#' The names and expected values are listed in the following sections.
#'
#' @section knitr chunk options for table captions:
#'
#' | **label**                                        |    **name**     | **value**  |
#' |:-------------------------------------------------|:---------------:|:----------:|
#' | caption id/bookmark                              | tab.id          |    NULL    |
#' | caption                                          | tab.cap         |    NULL    |
#' | display table caption on top of the table or not | tab.topcaption  |    TRUE    |
#' | caption table sequence identifier.               | tab.lp          |   "tab:"   |
#'
#' @section knitr chunk options for Word table captions:
#'
#' | **label**                                               |    **name**     | **value**                 |
#' |:--------------------------------------------------------|:---------------:|:-------------------------:|
#' | Word stylename to use for table captions.               | tab.cap.style   |            NULL           |
#' | prefix for numbering chunk (default to   "Table ").     | tab.cap.pre     |           Table           |
#' | suffix for numbering chunk (default to   ": ").         | tab.cap.sep     |            " :"           |
#' | title number depth                                      | tab.cap.tnd     |              0            |
#' | separator to use between title number and table number. | tab.cap.tns     |             "-"           |
#' | caption prefix formatting properties                    | tab.cap.fp_text | fp_text_lite(bold = TRUE) |
#'
#' @section knitr chunk options for Word tables:
#'
#' | **label**                                                     |       **name**      | **value**  |
#' |:--------------------------------------------------------------|:-------------------:|:----------:|
#' | the Word stylename to use for tables                          | tab.style           |    NULL    |
#' | autofit' or 'fixed' algorithm.                                | tab.layout          |  "autofit" |
#' | value of the preferred width of the table in percent (base 1).| tab.width           |      1     |
#' | Alternative title text                                        | tab.alt.title       |    NULL    |
#' | Alternative description text                                  | tab.alt.description |    NULL    |
#'
#' @section knitr chunk options for data.frame with officedown:
#'
#' | **label**                                                     |    **name**     | **value**  |
#' |:--------------------------------------------------------------|:---------------:|:----------:|
#' | apply or remove formatting from the first row in the table    | first_row       |    TRUE    |
#' | apply or remove formatting from the first column in the table | first_column    |    FALSE   |
#' | apply or remove formatting from the last row in the table     | last_row        |    FALSE   |
#' | apply or remove formatting from the last column in the table  | last_column     |    FALSE   |
#' | don't display odd and even rows                               | no_hband        |    TRUE    |
#' | don't display odd and even columns                            | no_vband        |    TRUE    |
#'
#' @return a list
#'
#' @section returned elements:
#'
#' * cap.style (default: NULL)
#' * cap.pre (default: "Table ")
#' * cap.sep (default: ":")
#' * cap.tnd (default: 0)
#' * cap.tns (default: "-")
#' * cap.fp_text (default: `fp_text_lite(bold = TRUE)`)
#' * id (default: NULL)
#' * cap (default: NULL)
#' * alt.title (default: NULL)
#' * alt.description (default: NULL)
#' * topcaption (default: TRUE)
#' * style (default: NULL)
#' * tab.lp (default: "tab:")
#' * table_layout (default: "autofit")
#' * table_width (default: 1)
#' * first_row (default: TRUE)
#' * first_column (default: FALSE)
#' * last_row (default: FALSE)
#' * last_column (default: FALSE)
#' * no_hband (default: TRUE)
#' * no_vband (default: TRUE)
#' @family functions for officer extensions
#' @keywords internal
opts_current_table <- function() {

  is_bookdown <- isTRUE(knitr::opts_knit$get('bookdown.internal.label'))

  if (is_bookdown) {
    tab.lp <- knitr_opts_current("tab.lp", default = "tab:")
  } else {
    tab.lp <- knitr_opts_current("tab.lp", default = NULL)
  }
  tab.topcaption <- knitr_opts_current("tab.topcaption", default = TRUE)

  tab.cap.pre <- knitr_opts_current("tab.cap.pre", default = "Table")
  tab.cap.sep <- knitr_opts_current("tab.cap.sep", default = ":")

  tab.cap.pre <- paste0(tab.cap.pre, " ")
  tab.cap.sep <- paste0(tab.cap.sep, " ")

  tab.cap <- knitr_opts_current("tab.cap", default = NULL)

  if (is_bookdown) {
    tab.id <- knitr_opts_current("label", default = NULL)
    tab.id <- knitr_opts_current("tab.id", default = tab.id)
  } else {
    tab.id <- knitr_opts_current("tab.id", default = NULL)
  }

  tab.cap.style <- knitr_opts_current("tab.cap.style", default = NULL)
  tab.alt.title <- knitr_opts_current("tab.alt.title", default = NULL)
  tab.alt.description <- knitr_opts_current("tab.alt.description", default = NULL)

  tab.cap.tnd <- knitr_opts_current("tab.cap.tnd", default = 0)
  tab.cap.tns <- knitr_opts_current("tab.cap.tns", default = "-")
  tab.cap.fp_text <- knitr_opts_current("tab.cap.fp_text", default = fp_text_lite(bold = TRUE))

  tab.style <- knitr_opts_current("tab.style", default = NULL)
  tab.layout <- knitr_opts_current("tab.layout", default = "autofit")
  tab.width <- knitr_opts_current("tab.width", default = 1)

  first_row <- knitr_opts_current("first_row", default = TRUE)
  first_column <- knitr_opts_current("first_column", default = FALSE)
  last_row <- knitr_opts_current("last_row", default = FALSE)
  last_column <- knitr_opts_current("last_column", default = FALSE)
  no_hband <- knitr_opts_current("no_hband", default = TRUE)
  no_vband <- knitr_opts_current("no_vband", default = TRUE)

  list(
    cap.style = tab.cap.style,
    cap.pre = tab.cap.pre,
    cap.sep = tab.cap.sep,
    cap.tnd = tab.cap.tnd,
    cap.tns = tab.cap.tns,
    cap.fp_text = tab.cap.fp_text,
    id = tab.id,
    alt.title = tab.alt.title,
    alt.description = tab.alt.description,
    topcaption = tab.topcaption,
    cap = tab.cap,
    style = tab.style,
    tab.lp = tab.lp,
    table_layout = table_layout(type = tab.layout),
    table_width = table_width(width = tab.width, unit = "pct"),
    first_row = first_row,
    first_column = first_column,
    last_row = last_row,
    last_column = last_column,
    no_hband = no_hband,
    no_vband = no_vband
  )
}

#' @name officer-defunct
#' @title Defunct Functions in Package officer
#' @export
#' @rdname officer-defunct
#' @details `docx_reference_img()` is not usefull anymore.
#' @param ... unused arguments
#' @keywords internal
docx_reference_img <- function(...){
  .Defunct(msg = "The `docx_reference_img()` function is no longer useful.")
}

#' @export
#' @title Fast Append context to a Word document
#' @description
#' This function is used to append content to a Word document
#' in a fast way.
#'
#' It does not use the XML tree of the document
#' neither the cursor that is responsible for increasing the
#' performance of Word document generation when looping
#' over a large number of elements.
#'
#' This function must be used with the `write_elements_to_context()`
#' and `body_append_stop_context()` functions:
#'
#' 1. `body_append_start_context()` creates a context and returns
#'    a list with the context and the file connection.
#' 2. `write_elements_to_context()` writes the elements to the context
#'   file connection.
#' 3. `body_append_stop_context()` closes the file connection and
#'   replaces the XML in the document with the new XML.
#' @param x an rdocx object
#' @param additional_ns a named character vector of additional XML namespaces
#' to be added to the root node of the document. The names of the vector
#' are the namespace prefixes and the values are the namespace URIs.
#'
#' This argument is useful when the elements to be added to the document require
#' additional namespaces that are not already present in the document and not
#' part of the xml generated by `to_wml()`. Simple users are not expected
#' to use this argument. It is mainly intended for developers of officer extensions.
#' @return `body_append_start_context()` returns a list representing the context
#' that contains:
#' - `doc`: the original document object
#' - `file_con`: the file connection to the context
#' - `file_path`: the path to the context file
#' - `final_str`: the final XML string to be appended to the document
#' later when calling `body_append_stop_context()`.
#'
#' This object should not be modified by the user but instead
#' passed to `write_elements_to_context()` and `body_append_stop_context()`.
#'
#' `write_elements_to_context()` returns the context object.
#'
#' `body_append_stop_context()` returns the `rdocx` object with the
#' cursor position set to the end of the document.
#' @importFrom utils tail
#' @rdname body_append_context
#' @example inst/examples/example_body_append_context.R
#' @family functions for adding content
body_append_start_context <- function(x, additional_ns = character()) {
  if (!is.character(additional_ns)) {
    cli::cli_abort("`additional_ns` must be a named character vector.")
  }
  if (length(additional_ns) > 0 && is.null(names(additional_ns))) {
    cli::cli_abort("`additional_ns` must be a named character vector.")
  }
  if (!inherits(x, "rdocx")) {
    cli::cli_abort("`x` must be an rdocx object.")
  }

  current_xml_file <- tempfile(fileext = ".txt")
  xml_doc <- docx_body_xml(x)
  current_xml_str <- as.character(xml_doc)
  writeLines(current_xml_str, current_xml_file)
  current_xml_str_lines <- readLines(current_xml_file)
  unlink(current_xml_file)

  section_start <- grepl("<w:sectPr", current_xml_str_lines)
  default_section_pos <- tail(which(section_start), n = 1)

  if (length(additional_ns) > 0) {
    add_ns_str <- sprintf(
      " xmlns:%s=\"%s\"",
      names(additional_ns),
      additional_ns
    )
    add_ns_str <- paste0(add_ns_str, collapse = "")
    current_xml_str_lines[2] <- gsub(
      "<w:document",
      paste0("<w:document", add_ns_str),
      current_xml_str_lines[2],
      fixed = TRUE
    )
  }

  output_file <- tempfile(fileext = ".txt")
  con <- file(output_file, open = "w", encoding = "UTF-8")
  cat(
    current_xml_str_lines[seq_len(default_section_pos - 1)],
    sep = "\n",
    file = con,
    append = TRUE
  )

  out <- list(
    doc = x,
    file_con = con,
    file_path = output_file,
    final_str = current_xml_str_lines[
      default_section_pos:length(current_xml_str_lines)
    ]
  )
  class(out) <- "rdocx_append_context"
  out
}

#' @rdname body_append_context
#' @param ... elements to be written to the context. These can be
#' paragraphs, tables, images, etc. The elements should have an associated
#' `to_wml()` method that converts them to WML format.
#' @param context the context object created by `body_append_start_context()`.
#' @export
write_elements_to_context <- function(context, ...) {
  objects <- list(...)
  for (obj in objects) {
    cat(
      to_wml(obj, add_ns = TRUE),
      sep = "\n",
      file = context$file_con,
      append = TRUE
    )
  }
  context
}

#' @rdname body_append_context
#' @export
body_append_stop_context <- function(context) {
  cat(context$final_str, sep = "\n", file = context$file_con, append = TRUE)
  close(context$file_con)
  context$doc$doc_obj$replace_xml(context$file_path)
  unlink(context$file_path)
  x <- context$doc
  x$officer_cursor <- officer_cursor(x$doc_obj$get())
  x
}

# utils ----
is_scalar_character <- function( x ) {
  is.character(x) && length(x) == 1
}
is_scalar_logical <- function( x ) {
  is.logical(x) && length(x) == 1
}



# functions ----


#' @export
#' @title Replace text at a bookmark location
#' @description Replace text content enclosed in a bookmark
#' with different text. A bookmark will be considered as valid if enclosing words
#' within a paragraph; i.e., a bookmark along two or more paragraphs is invalid,
#' a bookmark set on a whole paragraph is also invalid, but bookmarking few words
#' inside a paragraph is valid.
#' @param x a docx device
#' @param bookmark bookmark id
#' @param value the replacement string, of type character
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "a paragraph to replace", style = "centered")
#' doc <- body_bookmark(doc, "text_to_replace")
#' doc <- body_replace_text_at_bkm(doc, "text_to_replace", "new text")
body_replace_text_at_bkm <- function( x, bookmark, value ){
  stopifnot(is_scalar_character(value), is_scalar_character(bookmark))
  xml_replace_text_at_bkm(node = x$doc_obj$get(), bookmark = bookmark, value = value)
  x
}


#' @export
#' @rdname body_replace_text_at_bkm
#' @examples
#'
#'
#' # demo usage of bookmark and images ----
#' template <- system.file(package = "officer", "doc_examples/example.docx")
#'
#' img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
#'
#' doc <- read_docx(path = template)
#' doc <- headers_replace_img_at_bkm(x = doc, bookmark = "bmk_header",
#'                                   value = external_img(src = img.file, width = .53, height = .7))
#' doc <- footers_replace_img_at_bkm(x = doc, bookmark = "bmk_footer",
#'                                   value = external_img(src = img.file, width = .53, height = .7))
#' print(doc, target = tempfile(fileext = ".docx"))
#'
body_replace_img_at_bkm <- function( x, bookmark, value ){
  stopifnot(inherits(x, "rdocx"),
            is_scalar_character(bookmark),
            inherits(value, "external_img"))
  docxpart_replace_img_at_bkm(node = x$doc_obj$get(), bookmark = bookmark, value = value)
  x
}

xml_replace_text_at_bkm <- function(node, bookmark, value){

  text <- enc2utf8(value)
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", bookmark)
  bm_start <- xml_find_first(node, xpath_)
  if( inherits(bm_start, "xml_missing") ) {
    return(FALSE)
  }

  str_ <- sprintf("//w:bookmarkStart[@w:name='%s']/following-sibling::w:r", bookmark )
  following_start <- sapply( xml_find_all(node, str_), xml_path )
  str_ <- sprintf("//w:bookmarkEnd[@w:id='%s']/preceding-sibling::w:r", xml_attr(bm_start, "id") )
  preceding_end <- sapply( xml_find_all(node, str_), xml_path )

  match_path <- base::intersect(following_start, preceding_end)
  if( length(match_path) < 1 )
    stop("could not find any bookmark ", bookmark, " located INSIDE a single paragraph" )

  run_nodes <- xml_find_all(node, paste0( match_path, collapse = "|" ) )

  for(node in run_nodes[setdiff(seq_along(run_nodes), 1)])
    xml_remove(node)

  xml_text(run_nodes[[1]] ) <- text
  TRUE
}

docxpart_replace_img_at_bkm <- function(node, bookmark, value) {
  stopifnot(is_scalar_character(bookmark))
  stopifnot(inherits(value, "external_img"))

  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", bookmark)
  bm_start <- xml_find_first(node, xpath_)
  if (inherits(bm_start, "xml_missing")) {
    stop("cannot find bookmark ", shQuote(bookmark), call. = FALSE)
  }

  str_ <- sprintf("//w:bookmarkStart[@w:name='%s']/following-sibling::w:r", bookmark)
  following_start <- sapply(xml_find_all(node, str_), xml_path)
  str_ <- sprintf("//w:bookmarkEnd[@w:id='%s']/preceding-sibling::w:r", xml_attr(bm_start, "id"))
  preceding_end <- sapply(xml_find_all(node, str_), xml_path)

  match_path <- base::intersect(following_start, preceding_end)
  if (length(match_path) < 1) {
    stop("could not find any bookmark ", bookmark, " located INSIDE a single paragraph")
  }

  out <- to_wml(value, add_ns = TRUE)

  run_nodes <- xml_find_all(node, paste0(match_path, collapse = "|"))
  for (node in run_nodes[setdiff(seq_along(run_nodes), 1)]) {
    xml_remove(node)
  }
  xml_replace(run_nodes[[1]], as_xml_document(out))
}

#' @export
#' @rdname body_replace_text_at_bkm
headers_replace_text_at_bkm <- function( x, bookmark, value ){
  stopifnot(is_scalar_character(value), is_scalar_character(bookmark))
  for(header in x$headers){
    xml_replace_text_at_bkm(node = header$get(), bookmark = bookmark, value = value)
  }
  x
}

#' @export
#' @rdname body_replace_text_at_bkm
headers_replace_img_at_bkm <- function( x, bookmark, value ){
  for(header in x$headers){
    docxpart_replace_img_at_bkm(node = header$get(), bookmark = bookmark, value = value)
  }
  x
}



#' @export
#' @rdname body_replace_text_at_bkm
footers_replace_text_at_bkm <- function( x, bookmark, value ){
  stopifnot(is_scalar_character(value), is_scalar_character(bookmark))
  for(footer in x$footers){
    xml_replace_text_at_bkm(node = footer$get(), bookmark = bookmark, value = value)
  }
  x
}

#' @export
#' @rdname body_replace_text_at_bkm
footers_replace_img_at_bkm <- function( x, bookmark, value ){
  for(footer in x$footers){
    docxpart_replace_img_at_bkm(node = footer$get(), bookmark = bookmark, value = value)
  }
  x
}

#' @export
#' @title Add plots at bookmark location in a 'Word' document
#' @description
#' Use these functions if you want to replace a paragraph containing
#' a bookmark with a 'ggplot' or a base plot.
#' @param value a ggplot object for body_replace_gg_at_bkm() or a set plot instructions
#' body_replace_plot_at_bkm(), see plot_instr().
#' @param bookmark bookmark id
#' @param keep Should the bookmark be preserved? Defaults to `FALSE`,
#' i.e.the bookmark will be lost as a side effect.
#' @inheritParams body_add_gg
#'
#' @examples
#' if (require("ggplot2")) {
#'   gg_plot <- ggplot(data = iris) +
#'     geom_point(mapping = aes(Sepal.Length, Petal.Length))
#'
#'   doc <- read_docx()
#'   doc <- body_add_par(doc, "insert_plot_here")
#'   doc <- body_bookmark(doc, "plot")
#'   doc <- body_replace_gg_at_bkm(doc, bookmark = "plot", value = gg_plot)
#'   print(doc, target = tempfile(fileext = ".docx"))
#' }
body_replace_gg_at_bkm <- function(x, bookmark, value, width = 6, height = 5,
                                   res = 300, style = "Normal", scale = 1,
                                   keep = FALSE, ...) {
  x <- cursor_bookmark(x, bookmark)
  x <- body_add_gg(
    x = x, value = value, width = width, height = height,
    res = res, style = style, scale = scale, pos = "on", ...
  )
  if (keep) {
    x <- body_bookmark(x, bookmark)
  }

  x
}

#' @export
#' @rdname body_replace_gg_at_bkm
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "insert_plot_here")
#' doc <- body_bookmark(doc, "plot")
#' if (capabilities(what = "png")) {
#'   doc <- body_replace_plot_at_bkm(
#'     doc, bookmark = "plot",
#'     value = plot_instr(
#'       code = {
#'         barplot(1:5, col = 2:6)
#'       }
#'     )
#'   )
#' }
#' print(doc, target = tempfile(fileext = ".docx"))
body_replace_plot_at_bkm <- function(x, bookmark, value, width = 6, height = 5,
                                   res = 300, style = "Normal",
                                   keep = FALSE, ...) {
  x <- cursor_bookmark(x, bookmark)
  x <- body_add_plot(
    x = x, value = value, width = width, height = height,
    res = res, style = style, pos = "on", ...
  )
  if (keep) {
    x <- body_bookmark(x, bookmark)
  }

  x
}

#' @export
#' @title Replace text anywhere in the document
#' @description Replace text anywhere in the document, or at a cursor.
#'
#' Replace all occurrences of old_value with new_value. This method
#' uses \code{\link{grepl}}/\code{\link{gsub}} for pattern matching; you may
#' supply arguments as required (and therefore use \code{\link{regex}} features)
#' using the optional \code{...} argument.
#'
#' Note that by default, grepl/gsub will use \code{fixed=FALSE}, which means
#' that \code{old_value} and \code{new_value} will be interepreted as regular
#' expressions.
#'
#' \strong{Chunking of text}
#'
#' Note that the behind-the-scenes representation of text in a Word document is
#' frequently not what you might expect! Sometimes a paragraph of text is broken
#' up (or "chunked") into several "runs," as a result of style changes, pauses
#' in text entry, later revisions and edits, etc. If you have not styled the
#' text, and have entered it in an "all-at-once" fashion, e.g. by pasting it or
#' by outputing it programmatically into your Word document, then this will
#' likely not be a problem. If you are working with a manually-edited document,
#' however, this can lead to unexpected failures to find text.
#'
#' You can use the officer function \code{\link{docx_show_chunk}} to
#' show how the paragraph of text at the current cursor has been chunked into
#' runs, and what text is in each chunk. This can help troubleshoot unexpected
#' failures to find text.
#' @seealso \code{\link{grep}}, \code{\link{regex}}, \code{\link{docx_show_chunk}}
#' @author Frank Hangler, \email{frank@plotandscatter.com}
#' @param x a docx device
#' @param old_value the value to replace
#' @param new_value the value to replace it with
#' @param only_at_cursor if \code{TRUE}, only search-and-replace at the current
#' cursor; if \code{FALSE} (default), search-and-replace in the entire document
#' (this can be slow on large documents!)
#' @param warn warn if \code{old_value} could not be found.
#' @param ... optional arguments to grepl/gsub (e.g. \code{fixed=TRUE})
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "Placeholder one")
#' doc <- body_add_par(doc, "Placeholder two")
#'
#' # Show text chunk at cursor
#' docx_show_chunk(doc)  # Output is 'Placeholder two'
#'
#' # Simple search-and-replace at current cursor, with regex turned off
#' doc <- body_replace_all_text(doc, old_value = "Placeholder",
#'   new_value = "new", only_at_cursor = TRUE, fixed = TRUE)
#' docx_show_chunk(doc)  # Output is 'new two'
#'
#' # Do the same, but in the entire document and ignoring case
#' doc <- body_replace_all_text(doc, old_value = "placeholder",
#'   new_value = "new", only_at_cursor=FALSE, ignore.case = TRUE)
#' doc <- cursor_backward(doc)
#' docx_show_chunk(doc) # Output is 'new one'
#'
#' # Use regex : replace all words starting with "n" with the word "example"
#' doc <- body_replace_all_text(doc, "\\bn.*?\\b", "example")
#' docx_show_chunk(doc) # Output is 'example one'
body_replace_all_text <- function( x, old_value, new_value,
                                   only_at_cursor = FALSE,
                                   warn = TRUE, ... ){
  stopifnot(is_scalar_character(old_value),
            is_scalar_character(new_value),
            is_scalar_logical(only_at_cursor))

  oldValue <- enc2utf8(old_value)
  newValue <- enc2utf8(new_value)

  replacement_count <- 0

  base_node <- if (only_at_cursor) {
    docx_current_block_xml(x)
  } else {
    docx_body_xml(x)
  }

  # For each matching text node...
  for (text_node in xml_find_all(base_node, ".//w:t")) {
    # ...if it contains the oldValue...
    if (grepl(oldValue, xml_text(text_node), ...)) {
      replacement_count <- replacement_count + 1
      # Replace the node text with the newValue.
      xml_text(text_node) <- gsub(oldValue, newValue, xml_text(text_node), ...)
    }
  }

  # Alert the user if no replacements were made.
  if (replacement_count == 0 && warn) {
    search_zone_text <- if (only_at_cursor) "at the cursor." else "in the document."
    warning("Found 0 instances of '", oldValue, "' ", search_zone_text)
  }

  x
}

#' @export
#' @title Show underlying text tag structure
#' @description Show the structure of text tags at the current cursor. This is
#' most useful when trying to troubleshoot search-and-replace functionality
#' using \code{\link{body_replace_all_text}}.
#' @seealso \code{\link{body_replace_all_text}}
#' @param x a docx device
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "Placeholder one")
#' doc <- body_add_par(doc, "Placeholder two")
#'
#' # Show text chunk at cursor
#' docx_show_chunk(doc)  # Output is 'Placeholder two'
docx_show_chunk <- function( x ){
  cursor_elt <- docx_current_block_xml(x)
  text_nodes <- xml_find_all(cursor_elt, ".//w:t")
  msg <- paste0(length(text_nodes), " text nodes found at this cursor.")
  msg_detail <- ""
  for (text_node in text_nodes) {
    msg_detail <- paste0( msg_detail,
                          paste0("\n  <w:t>: '",
                                 xml_text(text_node), "'") )
  }
  message(paste(msg, msg_detail))
  invisible(x)
}




#' @export
#' @rdname body_replace_all_text
#' @section header_replace_all_text:
#' Replacements will be performed in each header of all sections.
headers_replace_all_text <- function( x, old_value, new_value, only_at_cursor = FALSE, warn = TRUE,  ... ){
  stopifnot(is_scalar_character(old_value),
            is_scalar_character(new_value),
            is_scalar_logical(only_at_cursor))

  oldValue <- enc2utf8(old_value)
  newValue <- enc2utf8(new_value)

  for(header in x$headers){

    replacement_count <- 0

    base_node <- header$get()

    # For each matching text node...
    for (text_node in xml_find_all(base_node, ".//w:t")) {
      # ...if it contains the oldValue...
      if (grepl(oldValue, xml_text(text_node), ...)) {
        replacement_count <- replacement_count + 1
        # Replace the node text with the newValue.
        xml_text(text_node) <- gsub(oldValue, newValue, xml_text(text_node), ...)
      }
    }

    # Alert the user if no replacements were made.
    if (replacement_count == 0 && warn) {
      search_zone_text <- if (only_at_cursor) "at the cursor." else "in the document."
      warning("Found 0 instances of '", oldValue, "' ", search_zone_text)
    }

  }

  x
}
#' @export
#' @rdname body_replace_all_text
#' @section header_replace_all_text:
#' Replacements will be performed in each footer of all sections.
footers_replace_all_text <- function( x, old_value, new_value, only_at_cursor = FALSE, warn = TRUE, ... ){
  stopifnot(is_scalar_character(old_value),
            is_scalar_character(new_value),
            is_scalar_logical(only_at_cursor))

  oldValue <- enc2utf8(old_value)
  newValue <- enc2utf8(new_value)

  for(footer in x$footers){
    replacement_count <- 0

    base_node <- footer$get()

    # For each matching text node...
    for (text_node in xml_find_all(base_node, ".//w:t")) {
      # ...if it contains the oldValue...
      if (grepl(oldValue, xml_text(text_node), ...)) {
        replacement_count <- replacement_count + 1
        # Replace the node text with the newValue.
        xml_text(text_node) <- gsub(oldValue, newValue, xml_text(text_node), ...)
      }
    }

    # Alert the user if no replacements were made.
    if (replacement_count == 0 && warn) {
      search_zone_text <- if (only_at_cursor) "at the cursor." else "in the document."
      warning("Found 0 instances of '", oldValue, "' ", search_zone_text)
    }
  }

  x
}

#' @export
#' @title Remove unused media from a document
#' @description The function will scan the media
#' directory and delete images that are not used
#' anymore. This function is to be used when images
#' have been replaced many times.
#' @param x \code{rdocx} or \code{rpptx} object
#' @keywords internal
sanitize_images <- function(x){

  rel_files <- list.files(x$package_dir, pattern = "\\.xml.rels$", recursive = TRUE, full.names = TRUE)

  image_files <- lapply(rel_files,
                        function(x){
                          zz <- read_xml(x)
                          rels <- xml_children(zz)
                          rels <- rels[xml_attr(rels, "Type") %in% "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"]
                          xml_attr(rels, "Target")
                        })
  image_files <- unique(unlist(image_files))

  if(inherits(x, "rdocx")){
    base_doc <- file.path(x$package_dir, "word")
    existing_img <- list.files(
      file.path(base_doc, "media"),
      pattern = "\\.(png|jpg|jpeg|eps|emf)$",
      ignore.case = TRUE,
      recursive = TRUE, full.names = TRUE)
    existing_img <- gsub(paste0(base_doc, "/"), "", existing_img, fixed = TRUE)
    unlink(file.path(base_doc,
                     setdiff(image_files, existing_img)
                     ), force = TRUE)
  } else if(inherits(x, "rpptx")){
    base_doc <- file.path(x$package_dir, "ppt")
    existing_img <- list.files(
      file.path(base_doc, "media"),
      pattern = "\\.(png|jpg|jpeg|eps|emf)$",
      ignore.case = TRUE,
      recursive = TRUE, full.names = TRUE)
    existing_img <- gsub(paste0(base_doc, "/"), "",
                         existing_img, fixed = TRUE)
    unlink(file.path(base_doc, setdiff(image_files, existing_img)), force = TRUE)
  }
  x
}

get_all_attrs <- function(nodeset) {
  # extract all children per node
  children <- lapply(nodeset, xml_children)
  # extract all node names per child
  name <- lapply(children, xml_name)
  # extract all attributes per child
  attr <- lapply(children, xml_attrs)
  # set attribute names
  for (i in seq_along(attr)) {
    names(attr[[i]]) <- name[[i]]
  }
  attr
}

get_attr_val <- function(lst, node, attr = NULL) {
  if (is.null(attr)) {
    attr <- "val"
  }
  if (!length(lst)) {
    return(NA_character_)
  }

  vapply(lst, FUN.VALUE = "", function(node_attr) {
    a <- node_attr[[node]][attr]
    if (is.null(a)) NA_character_ else a
  })
}

# a faster way for constructing a data.frame
as_data_frame <- function(x, ...) {
  stopifnot(length(unique(lengths(x))) == 1L)
  structure(x, ..., row.names = seq_along(x[[1L]]), class = "data.frame")
}

read_rpr <- function(x) {
  node_rpr <- xml_child(x, "w:rPr")

  # check if empty
  not_empty <- lengths(node_rpr) > 0L

  # all elements to extract
  elem <- c(
    "font.size",
    "bold",
    "italic",
    "underlined",
    "color",
    "font.family",
    "vertical.align",
    "shading.color",
    "hansi.family",
    "eastasia.family",
    "cs.family",
    "bold.cs",
    "font.size.cs",
    "lang.val",
    "lang.eastasia",
    "lang.bidi"
  )

  # init results
  rpr <- replicate(length(elem), list(rep(NA_character_, length(node_rpr))))
  names(rpr) <- elem

  # directly return if all are empty
  if (!any(not_empty)) {
    return(as_data_frame(rpr))
  }

  # further reduce to only work on non-empty nodes and extract all attributes of
  # every child
  rpr_attr <- get_all_attrs(node_rpr[not_empty])

  # replace with actual attribute values
  rpr[["font.size"]][not_empty] <- get_attr_val(rpr_attr, "sz")
  rpr[["bold"]][not_empty] <- get_attr_val(rpr_attr, "b")
  rpr[["italic"]][not_empty] <- get_attr_val(rpr_attr, "i")
  rpr[["underlined"]][not_empty] <- get_attr_val(rpr_attr, "u")
  rpr[["color"]][not_empty] <- get_attr_val(rpr_attr, "color")
  rpr[["font.family"]][not_empty] <- get_attr_val(rpr_attr, "rFonts", "ascii")

  rpr[["vertical.align"]][not_empty] <- get_attr_val(rpr_attr, "vertAlign")
  rpr[["shading.color"]][not_empty] <- get_attr_val(rpr_attr, "shd", "fill")

  rpr[["hansi.family"]][not_empty] <- get_attr_val(rpr_attr, "rFonts", "hAnsi")
  rpr[["eastasia.family"]][not_empty] <- get_attr_val(
    rpr_attr,
    "rFonts",
    "eastAsia"
  )
  rpr[["cs.family"]][not_empty] <- get_attr_val(rpr_attr, "rFonts", "cs")
  rpr[["bold.cs"]][not_empty] <- get_attr_val(rpr_attr, "bCs")
  rpr[["font.size.cs"]][not_empty] <- get_attr_val(rpr_attr, "szCs")

  rpr[["lang.val"]][not_empty] <- get_attr_val(rpr_attr, "lang")
  rpr[["lang.eastasia"]][not_empty] <- get_attr_val(
    rpr_attr,
    "lang",
    "eastAsia"
  )
  rpr[["lang.bidi"]][not_empty] <- get_attr_val(rpr_attr, "lang", "bidi")

  as_data_frame(rpr)
}

read_ppr <- function(x) {
  node_ppr <- xml_child(x, "w:pPr")

  # check if empty
  not_empty <- lengths(node_ppr) > 0L

  # all elements to extract
  elem <- c(
    "align",
    "keep_next",
    "line_spacing",
    "padding.bottom",
    "padding.top",
    "padding.left",
    "padding.right",
    "shading.color.par",
    "border.bottom.width",
    "border.bottom.color",
    "border.bottom.style",
    "border.top.width",
    "border.top.color",
    "border.top.style",
    "border.left.width",
    "border.left.color",
    "border.left.style",
    "border.right.width",
    "border.right.color",
    "border.right.style"
  )

  # init results
  ppr <- replicate(length(elem), list(rep(NA_character_, length(node_ppr))))
  names(ppr) <- elem

  # init values for non-character columns
  ppr[paste0(
    "border.",
    c("bottom", "top", "left", "right"),
    ".width"
  )] <- list(rep(NA_integer_, length(node_ppr)))
  ppr["line_spacing"] <- list(rep(NA_real_, length(node_ppr)))
  ppr["keep_next"] <- list(rep(FALSE, length(node_ppr)))

  # directly return if all are empty
  if (!any(not_empty)) {
    return(as_data_frame(ppr))
  }

  # further reduce to only work on non-empty nodes and extract all attributes of
  # every child
  ppr_attr <- get_all_attrs(node_ppr[not_empty])

  # NOTE: 'w:pBdr' has to be treated specially since it is nested inside 'w:pPr'
  bdr <- xml_child(node_ppr[not_empty], "w:pBdr")
  not_empty_bdr <- lengths(bdr) > 0L
  bdr_attr <- get_all_attrs(bdr[not_empty_bdr])

  # replace with actual attribute values
  ppr[["align"]][not_empty] <- get_attr_val(ppr_attr, "jc")
  ppr[["keep_next"]][not_empty] <- vapply(
    ppr_attr,
    function(attrs) "keepNext" %in% names(attrs),
    logical(1L)
  )
  ppr[["line_spacing"]][not_empty] <- as.integer(get_attr_val(
    ppr_attr,
    "spacing",
    "line"
  )) /
    240L

  ppr[["padding.bottom"]][not_empty] <- get_attr_val(
    ppr_attr,
    "spacing",
    "after"
  )
  ppr[["padding.top"]][not_empty] <- get_attr_val(ppr_attr, "spacing", "before")
  ppr[["padding.left"]][not_empty] <- get_attr_val(ppr_attr, "ind", "left")
  ppr[["padding.right"]][not_empty] <- get_attr_val(ppr_attr, "ind", "right")
  ppr[["shading.color.par"]][not_empty] <- get_attr_val(ppr_attr, "shd", "fill")

  ppr[["border.bottom.width"]][not_empty][
    not_empty_bdr
  ] <- as.integer(get_attr_val(bdr_attr, "bottom", "sz")) / 8L
  ppr[["border.bottom.color"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "bottom",
    "color"
  )
  ppr[["border.bottom.style"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "bottom",
    "val"
  )

  ppr[["border.top.width"]][not_empty][
    not_empty_bdr
  ] <- as.integer(get_attr_val(bdr_attr, "top", "sz")) / 8L
  ppr[["border.top.color"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "top",
    "color"
  )
  ppr[["border.top.style"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "top",
    "val"
  )

  ppr[["border.left.width"]][not_empty][
    not_empty_bdr
  ] <- as.integer(get_attr_val(bdr_attr, "left", "sz")) / 8L
  ppr[["border.left.color"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "left",
    "color"
  )
  ppr[["border.left.style"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "left",
    "val"
  )

  ppr[["border.right.width"]][not_empty][
    not_empty_bdr
  ] <- as.integer(get_attr_val(bdr_attr, "right", "sz")) / 8L
  ppr[["border.right.color"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "right",
    "color"
  )
  ppr[["border.right.style"]][not_empty][not_empty_bdr] <- get_attr_val(
    bdr_attr,
    "right",
    "val"
  )

  as_data_frame(ppr)
}

read_docx_styles <- function(package_dir) {
  styles_file <- file.path(package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)

  all_styles <- xml_find_all(doc, "/w:styles/w:style")

  ppr <- read_ppr(all_styles)
  rpr <- read_rpr(all_styles)

  main <- data.frame(
    stringsAsFactors = FALSE,
    style_type = xml_attr(all_styles, "type"),
    style_id = xml_attr(all_styles, "styleId"),
    style_name = xml_attr(xml_child(all_styles, "w:name"), "val"),
    base_on = xml_attr(xml_child(all_styles, "w:basedOn"), "val"),
    is_custom = xml_attr(all_styles, "customStyle") %in% "1",
    is_default = xml_attr(all_styles, "default") %in% "1"
  )
  out <- cbind(main, ppr, rpr)
  out
}

get_style_id <- function(data, style, type) {
  ref <- data[data$style_type == type, ]

  if (!style %in% ref$style_name) {
    t_ <- shQuote(ref$style_name, type = "sh")
    t_ <- paste(t_, collapse = ", ")
    t_ <- paste0("c(", t_, ")")
    stop(
      "could not match any style named ",
      shQuote(style, type = "sh"),
      " in ",
      t_,
      call. = FALSE
    )
  }
  ref$style_id[ref$style_name == style]
}

# generics -----

#' @export
#' @title Convert officer objects to WordprocessingML
#' @description Convert an object made with package officer
#' to WordprocessingML. The result is a string.
#' @param x object
#' @param add_ns should namespace be added to the top tag
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @keywords internal
to_wml <- function(x, add_ns = FALSE, ...) {
  UseMethod("to_wml")
}

#' @export
#' @title Convert officer objects to PresentationML
#' @description Convert an object made with package officer
#' to PresentationML. The result is a string.
#' @param x object
#' @param add_ns should namespace be added to the top tag
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @keywords internal
to_pml <- function(x, add_ns = FALSE, ...) {
  UseMethod("to_pml")
}

#' @export
#' @title Convert officer objects to HTML
#' @description Convert an object made with package officer
#' to HTML. The result is a string.
#' @param x object
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @keywords internal
to_html <- function(x, ...) {
  UseMethod("to_html")
}


# bookmark -----

bookmark <- function(id, str) {
  new_id <- uuid_generate()
  bm_start_str <- sprintf("<w:bookmarkStart w:id=\"%s\" w:name=\"%s\"/>", new_id, id)
  bm_start_end <- sprintf("<w:bookmarkEnd w:id=\"%s\"/>", new_id)
  paste0(bm_start_str, str, bm_start_end)
}

# ftext -----
#' @export
#' @title Formatted chunk of text
#' @description Format a chunk of text with text formatting properties (bold, color, ...).
#' The function allows you to create pieces of text formatted the way you want.
#' @param text text value, a single character value
#' @param prop formatting text properties returned by [fp_text]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @section usage:
#' You can use this function in conjunction with [fpar] to create paragraphs
#' consisting of differently formatted text parts. You can also use this
#' function as an *r chunk* in an R Markdown document made with package
#' officedown.
#' @examples
#' ftext("hello", fp_text())
#'
#' properties1 <- fp_text(color = "red")
#' properties2 <- fp_text(bold = TRUE, shading.color = "yellow")
#' ftext1 <- ftext("hello", properties1)
#' ftext2 <- ftext("World", properties2)
#' paragraph <- fpar(ftext1, " ", ftext2)
#'
#' x <- read_docx()
#' x <- body_add(x, paragraph)
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [fp_text]
#' @family run functions for reporting
ftext <- function(text, prop = NULL) {
  if (is.character(text)) {
    value <- enc2utf8(text)
  } else {
    value <- formatC(text)
  }

  out <- list(value = value, pr = prop)
  class(out) <- c("ftext", "cot", "run")
  out
}

#' @export
print.ftext <- function(x, ...) {
  cat("text: ", x$value, "\nformat:\n", sep = "")
  print(x$pr)
}

#' @export
to_wml.ftext <- function(x, add_ns = FALSE, ...) {
  out <- paste0(
    "<w:r>", if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:t xml:space=\"preserve\">",
    htmlEscapeCopy(x$value), "</w:t></w:r>"
  )
  out
}

#' @export
to_pml.ftext <- function(x, add_ns = FALSE, ...) {
  open_tag <- ar_ns_no
  if (add_ns) {
    open_tag <- ar_ns_yes
  }
  paste0(
    open_tag, if (!is.null(x$pr)) rpr_pml(x$pr),
    "<a:t>", htmlEscapeCopy(x$value), "</a:t></a:r>"
  )
}

#' @export
to_html.ftext <- function(x, ...) {
  sprintf("<span style=\"%s\">%s</span>", if (!is.null(x$pr)) rpr_css(x$pr) else "", htmlEscapeCopy(x$value))
}


# Word chunk ----
#' @export
#' @title Word chunk of text with a style
#' @description Format a chunk of text associated with a 'Word' character style.
#' The style is defined with its unique identifer.
#' @param text text value, a single character value
#' @param style_id 'Word' unique style identifier associated with the style to use.
#' @examples
#' run1 <- run_wordtext("hello", "DefaultParagraphFont")
#' paragraph <- fpar(run1)
#'
#' x <- read_docx()
#' x <- body_add_fpar(x, paragraph)
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [ftext()]
#' @family run functions for reporting
run_wordtext <- function(text, style_id = NULL) {
  if (is.character(text)) {
    value <- enc2utf8(text)
  } else {
    value <- formatC(text)
  }

  out <- list(value = value, style_id = style_id)
  class(out) <- c("run_wordtext", "cot", "run")
  out
}

#' @export
print.run_wordtext <- function(x, ...) {
  cat("text: `", x$value, "`\nword style id: `", x$style_id, "`\n", sep = "")
}
#
#' @export
to_wml.run_wordtext <- function(x, add_ns = FALSE, ...) {
  out <- paste0(
    "<w:r>",
    "<w:rPr>",
    sprintf("<w:rStyle w:val=\"%s\"/>", x$style_id),
    "</w:rPr>",
    "<w:t xml:space=\"preserve\">",
    htmlEscapeCopy(x$value), "</w:t></w:r>"
  )
  out
}


# Word computed field ----
#' @export
#' @title 'Word' computed field
#' @description Create a 'Word' computed field.
#' @note
#' In the previous version, this function was called `run_seqfield`
#' but the name was wrong and should have been `run_word_field`.
#' @inheritSection ftext usage
#' @param field Value for a "Word Computed Field" as a string.
#' @param seqfield deprecated in favor of `field`.
#' @param prop formatting text properties returned by [fp_text].
#' @examples
#' run_word_field(field = "PAGE  \\* MERGEFORMAT")
#' run_word_field(field = "Date \\@ \"MMMM d yyyy\"")
#' @family run functions for reporting
#' @family Word computed fields
run_word_field <- function(field, prop = NULL, seqfield = NULL) {
  if (!is.null(seqfield)) {
    field <- seqfield
    message("`seqfield` argument is deprecated in favor of `field`")
  }
  z <- list(
    field = field, pr = prop
  )
  class(z) <- c("run_word_field", "run")
  z
}

#' @export
#' @rdname run_word_field
run_seqfield <- run_word_field

#' @export
to_wml.run_word_field <- function(x, add_ns = FALSE, ...) {
  pr <- if (!is.null(x$pr)) rpr_wml(x$pr) else "<w:rPr/>"

  xml_elt_1 <- paste0(
    wr_ns_yes,
    pr,
    "<w:fldChar w:fldCharType=\"begin\" w:dirty=\"true\"/>",
    "</w:r>"
  )
  xml_elt_2 <- paste0(
    wr_ns_yes,
    pr,
    sprintf("<w:instrText xml:space=\"preserve\" w:dirty=\"true\">%s</w:instrText>", x$field),
    "</w:r>"
  )
  xml_elt_3 <- paste0(
    wr_ns_yes,
    pr,
    "<w:fldChar w:fldCharType=\"end\" w:dirty=\"true\"/>",
    "</w:r>"
  )
  out <- paste0(xml_elt_1, xml_elt_2, xml_elt_3)

  out
}

# autonum ----

#' @export
#' @title Auto number
#' @description Create an autonumbered chunk, i.e. a string
#' representation of a sequence, each item will be numbered.
#' These runs can also be bookmarked and be used later for
#' cross references.
#' @inheritSection ftext usage
#' @param seq_id sequence identifier
#' @param pre_label,post_label text to add before and after number
#' @param bkm bookmark id to associate with autonumber run. If NULL, no bookmark
#' is added. Value can only be made of alpha numeric characters, ':', -' and '_'.
#' @param bkm_all if TRUE, the bookmark will be set on the whole string, if
#' FALSE, the bookmark will be set on the number only. Default to FALSE.
#' As an effect when a reference to this bookmark is used, the text can
#' be like "Table 1" or "1" (pre_label is not included in the referenced
#' text).
#' @param prop formatting text properties returned by [fp_text].
#' @param start_at If not NULL, it must be a positive integer, it
#' specifies the new number to use, at which number the auto
#' numbering will restart.
#' @param tnd *title number depth*, a positive integer (only applies if positive)
#' that specify the depth (or heading of level *depth*) to use for prefixing
#' the caption number with this last reference number. For example, setting
#' `tnd=2` will generate numbered captions like '4.3-2' (figure 2 of
#' chapter 4.3).
#' @param tns separator to use between title number and table
#' number. Default is "-".
#' @examples
#' run_autonum()
#' run_autonum(seq_id = "fig", pre_label = "fig. ")
#' run_autonum(seq_id = "tab", pre_label = "Table ", bkm = "anytable")
#' run_autonum(
#'   seq_id = "tab", pre_label = "Table ", bkm = "anytable",
#'   tnd = 2, tns = " "
#' )
#' @family run functions for reporting
#' @family Word computed fields
run_autonum <- function(seq_id = "table", pre_label = "Table ", post_label = ": ",
                        bkm = NULL, bkm_all = FALSE, prop = NULL,
                        start_at = NULL,
                        tnd = 0, tns = "-") {
  bkm <- check_bookmark_id(bkm)

  stopifnot(
    !is.null(tnd),
    !is.null(tns),
    is.numeric(tnd),
    sign(tnd) > -1,
    is.character(tns),
    is.null(start_at) || is.numeric(start_at),
    inherits(prop, "fp_text") || is.null(prop)
  )
  z <- list(
    seq_id = seq_id,
    pre_label = pre_label,
    post_label = post_label,
    bookmark = bkm,
    bookmark_all = bkm_all,
    pr = prop,
    start_at = start_at,
    tnd = tnd, tns = tns
  )
  class(z) <- c("run_autonum", "run")

  z
}

#' @export
#' @title Update bookmark of an autonumber run
#' @description This function lets recycling a object
#' made by [run_autonum()] by changing the bookmark value. This
#' is useful to avoid calling `run_autonum()` several times
#' because of many tables.
#' @param x an object of class [run_autonum()]
#' @param bkm bookmark id to associate with autonumber run.
#' Value can only be made of alpha numeric characters, ':', -' and '_'.
#' @examples
#' z <- run_autonum(
#'   seq_id = "tab", pre_label = "Table ",
#'   bkm = "anytable"
#' )
#' set_autonum_bookmark(z, bkm = "anothertable")
#' @seealso [run_autonum()]
set_autonum_bookmark <- function(x, bkm = NULL) {
  stopifnot(inherits(x, "run_autonum"))
  x$bookmark <- bkm
  x
}

#' @export
to_wml.run_autonum <- function(x, add_ns = FALSE, ...) {
  if (is.null(x$seq_id)) {
    return("")
  }

  pr <- if (!is.null(x$pr)) rpr_wml(x$pr) else "<w:rPr/>"

  run_str_pre <- sprintf("<w:r>%s<w:t xml:space=\"preserve\">%s</w:t></w:r>", pr, x$pre_label)
  run_str_post <- sprintf("<w:r>%s<w:t xml:space=\"preserve\">%s</w:t></w:r>", pr, x$post_label)

  seq_str <- paste0("SEQ ", x$seq_id, " \u005C* Arabic")
  if (!is.null(x$start_at) && is.numeric(x$start_at)) {
    seq_str <- paste(seq_str, "\\r", as.integer(x$start_at))
  }

  sqf <- run_word_field(field = seq_str, prop = x$pr)
  sf_str <- to_wml(sqf)

  if (x$tnd > 0) {
    z <- paste0(
      to_wml(run_word_field(field = paste0("STYLEREF ", x$tnd, " \\r"), prop = x$pr)),
      to_wml(ftext(x$tns, prop = x$pr))
    )
    sf_str <- paste0(z, sf_str)
  }


  if (!is.null(x$bookmark)) {
    if (x$bookmark_all) {
      out <- paste0(
        bookmark(
          id = x$bookmark,
          str = paste0(run_str_pre, sf_str)
        ),
        run_str_post
      )
    } else {
      sf_str <- bookmark(id = x$bookmark, sf_str)
      out <- paste0(run_str_pre, sf_str, run_str_post)
    }
  } else {
    out <- paste0(run_str_pre, sf_str, run_str_post)
  }

  out
}

# reference ----

#' @export
#' @title Cross reference
#' @description Create a representation of a reference
#' @param id reference id, a string
#' @param prop formatting text properties returned by [fp_text].
#' @inheritSection ftext usage
#' @examples
#' run_reference("a_ref")
#' @family run functions for reporting
#' @family Word computed fields
run_reference <- function(id, prop = NULL) {
  z <- paste0(" REF ", id, " \\h ")

  z <- list(
    id = id, pr = prop
  )
  class(z) <- c("run_reference", "run")

  z
}

#' @export
to_wml.run_reference <- function(x, add_ns = FALSE, ...) {
  out <- to_wml(run_word_field(field = paste0(" REF ", x$id, " \\h "), prop = x$pr))
  out
}

# hyperlink text ----

#' @export
#' @title Formatted chunk of text with hyperlink
#' @description Format a chunk of text with text formatting properties (bold, color, ...),
#' the chunk is associated with an hyperlink.
#' @param text text value, a single character value
#' @param prop formatting text properties returned by [fp_text]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @param href URL value
#' @inheritSection ftext usage
#' @examples
#' ft <- fp_text(font.size = 12, bold = TRUE)
#' hyperlink_ftext(
#'   href = "https://cran.r-project.org/index.html",
#'   text = "some text", prop = ft
#' )
#' @family run functions for reporting
hyperlink_ftext <- function(text, prop = NULL, href) {
  if (is.character(text)) {
    value <- enc2utf8(text)
  } else {
    value <- formatC(text)
  }

  out <- list(value = value, pr = prop, href = href)
  class(out) <- c("hyperlink_ftext", "cot", "run")
  out
}

#' @export
to_wml.hyperlink_ftext <- function(x, add_ns = FALSE, ...) {
  out <- paste0(
    "<w:r>", if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:t xml:space=\"preserve\">",
    x$value, "</w:t></w:r>"
  )
  paste0("<w:hyperlink r:id=\"", officer_url_encode(x$href), "\">", out, "</w:hyperlink>")
}

#' @export
to_pml.hyperlink_ftext <- function(x, add_ns = FALSE, ...) {
  open_tag <- ar_ns_no
  if (add_ns) {
    open_tag <- ar_ns_yes
  }
  str_ <- sprintf("<a:hlinkClick r:id=\"%s\"/>", officer_url_encode(x$href))

  if (!is.null(x$pr)) {
    rpr_pml_ <- rpr_pml(x$pr)
    m <- regexpr(">", rpr_pml_)
    regmatches(rpr_pml_, m) <- paste0(">", str_)
  } else {
    rpr_pml_ <- paste0("<a:rPr>", str_, "</a:rPr>")
  }

  paste0(
    open_tag, rpr_pml_,
    "<a:t>", htmlEscapeCopy(x$value), "</a:t></a:r>"
  )
}

#' @export
to_html.hyperlink_ftext <- function(x, ...) {
  if (!is.null(x$pr)) {
    rpr_css <- paste0(" style=\"", rpr_css(x$pr), "\"")
  } else {
    rpr_css <- ""
  }

  sprintf("<a href=\"%s\"><span%s></span></a>", x$href, rpr_css)
}


# bookmark ----

#' @export
#' @title Bookmark for 'Word'
#' @description Add a bookmark on a run object.
#' @param bkm bookmark id to associate with run. Value can only be
#' made of alpha numeric characters, '-' and '_'.
#' @param run a run object, made with a call to one of
#' the "run functions for reporting".
#' @inheritSection ftext usage
#' @examples
#' ft <- fp_text(font.size = 12, bold = TRUE)
#' run_bookmark("par1", ftext("some text", ft))
#' @family run functions for reporting
run_bookmark <- function(bkm, run) {
  all_run <- FALSE

  if (inherits(run, "run")) {
    run <- list(run)
  }

  if (is.list(run) && !inherits(run, "run")) {
    all_run <- all(vapply(run, inherits, FUN.VALUE = FALSE, what = "run"))
  }


  if (!all_run) {
    stop("`run` must be a run object (ftext for example) or a list of run objects.")
  }

  bkm <- check_bookmark_id(bkm)

  z <- list(id = bkm, run = run)
  class(z) <- c("run_bookmark", "run")
  z
}

#' @export
to_wml.run_bookmark <- function(x, add_ns = FALSE, ...) {
  runs <- lapply(x$run, to_wml, add_ns = add_ns, ...)
  runs <- do.call(paste0, runs)
  bookmark(id = x$id, str = runs)
}

# breaks ----

#' @export
#' @title Page break for 'Word'
#' @description Object representing a page break for a Word document.
#' @inheritSection ftext usage
#' @examples
#' fp_t <- fp_text(font.size = 12, bold = TRUE)
#' an_fpar <- fpar("let's add a break page", run_pagebreak(), ftext("and blah blah!", fp_t))
#'
#' x <- read_docx()
#' x <- body_add(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_pagebreak <- function() {
  z <- list()
  class(z) <- c("run_pagebreak", "run")
  z
}

#' @export
to_wml.run_pagebreak <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:br w:type=\"page\"/></w:r>"
  out
}

#' @export
#' @title Column break for 'Word'
#' @description Create a representation of a column break.
#' @inheritSection ftext usage
#' @examples
#' run_columnbreak()
#' @family run functions for reporting
run_columnbreak <- function() {
  z <- list()
  class(z) <- c("run_columnbreak", "run")

  z
}

#' @export
to_wml.run_columnbreak <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:br w:type=\"column\"/></w:r>"
  out
}

#' @export
#' @title Page break for 'Word'
#' @description Object representing a line break
#' for a Word document. The result must be used
#' within a call to [fpar].
#' @inheritSection ftext usage
#' @examples
#' fp_t <- fp_text(font.size = 12, bold = TRUE)
#' an_fpar <- fpar("let's add a line break", run_linebreak(), ftext("and blah blah!", fp_t))
#'
#' x <- read_docx()
#' x <- body_add(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_linebreak <- function() {
  z <- list()
  class(z) <- c("run_linebreak", "run")

  z
}

#' @export
to_wml.run_linebreak <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:br w:type=\"textWrapping\"/></w:r>"
  out
}

#' @export
#' @title Tab for 'Word'
#' @description Object representing a tab
#' in a Word document. The result must be used
#' within a call to [fpar]. It will only have
#' effects in Word output.
#'
#' Tabulation marks settings can be defined with [fp_tabs()] in
#' paragraph settings defined with [fp_par()].
#' @inheritSection ftext usage
#' @examples
#' z <- fp_tabs(
#'   fp_tab(pos = 0.5, style = "decimal"),
#'   fp_tab(pos = 1.5, style = "decimal")
#' )
#' par1 <- fpar(
#'   run_tab(), ftext("88."),
#'   run_tab(), ftext("987.45"),
#'   fp_p = fp_par(
#'     tabs = z
#'   )
#' )
#' par2 <- fpar(
#'   run_tab(), ftext("8."),
#'   run_tab(), ftext("670987.45"),
#'   fp_p = fp_par(
#'     tabs = z
#'   )
#' )
#' x <- read_docx()
#' x <- body_add(x, par1)
#' x <- body_add(x, par2)
#' print(x, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_tab <- function() {
  z <- list()
  class(z) <- c("run_tab", "run")

  z
}

#' @export
to_wml.run_tab <- function(x, add_ns = FALSE, ...) {
  out <- "<w:r><w:tab/></w:r>"
  out
}

# sections ----
inch_to_tweep <- function(x) round(x * 72 * 20, digits = 0)

#' @export
#' @title Page size object
#' @description The function creates a representation of the dimensions of a page. The
#' dimensions are defined by length, width and orientation. If the orientation is
#' in landscape mode then the length becomes the width and the width becomes the length.
#' @param width,height page width, page height (in inches).
#' @param orient page orientation, either 'landscape', either 'portrait'.
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @examples
#' page_size(orient = "landscape")
#' @family functions for section definition
page_size <- function(width = 21 / 2.54, height = 29.7 / 2.54, orient = "portrait", unit = "in") {

  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (orient %in% "portrait") {
    h <- max(c(height, width))
    w <- min(c(height, width))
  } else {
    h <- min(c(height, width))
    w <- max(c(height, width))
  }

  z <- list(width = w, height = h, orient = orient)
  class(z) <- c("page_size")
  z
}

#' @export
to_wml.page_size <- function(x, add_ns = FALSE, ...) {
  out <- sprintf(
    "<w:pgSz w:h=\"%.0f\" w:w=\"%.0f\" w:orient=\"%s\"/>",
    inch_to_tweep(x$height), inch_to_tweep(x$width), x$orient
  )
  out
}

#' @export
#' @title Section columns
#' @description The function creates a representation of the columns of a section.
#' @param widths columns widths in inches. If 3 values, 3 columns
#' will be produced.
#' @param space space in inches between columns.
#' @param sep if TRUE a line is separating columns.
#' @examples
#' section_columns()
#' @family functions for section definition
section_columns <- function(widths = c(2.5, 2.5), space = .25, sep = FALSE) {
  if (length(widths) < 2) {
    stop("length of widths should be at least 2")
  }

  z <- list(widths = widths, space = space, sep = sep)
  class(z) <- c("section_columns")
  z
}
#' @export
to_wml.section_columns <- function(x, add_ns = FALSE, ...) {
  widths <- x$widths * 20 * 72
  space <- x$space * 20 * 72

  columns_str_all_but_last <- sprintf(
    "<w:col w:w=\"%.0f\" w:space=\"%.0f\"/>",
    widths[-length(widths)], space
  )
  columns_str_last <- sprintf(
    "<w:col w:w=\"%.0f\"/>",
    widths[length(widths)]
  )
  columns_str <- c(columns_str_all_but_last, columns_str_last)

  sprintf(
    "<w:cols w:num=\"%.0f\" w:sep=\"%.0f\" w:space=\"%.0f\" w:equalWidth=\"0\">%s</w:cols>",
    length(widths), as.integer(x$sep),
    space, paste0(columns_str, collapse = "")
  )
}

#' @export
#' @title Page margins object
#' @description The margins for each page of a sectionThe function creates a representation of the dimensions of a page. The
#' dimensions are defined by length, width and orientation. If the orientation is
#' in landscape mode then the length becomes the width and the width becomes the length.
#' @param bottom,top distance (in inches) between the bottom/top of the text margin and the bottom/top of the page. The text is placed at the greater of the value of this attribute and the extent of the header/footer text. A negative value indicates that the content should be measured from the bottom/topp of the page regardless of the footer/header, and so will overlap the footer/header. For example, `header=-0.5, bottom=1` means that the footer must start one inch from the bottom of the page and the main document text must start a half inch from the bottom of the page. In this case, the text and footer overlap since bottom is negative.
#' @param footer distance (in inches) from the bottom edge of the page to the bottom edge of the footer.
#' @param header distance (in inches) from the top edge of the page to the top edge of the header.
#' @param left,right distance (in inches) from the left/right edge of the page to the left/right edge of the text.
#' @param gutter page gutter (in inches).
#' @examples
#' page_mar()
#' @family functions for section definition
page_mar <- function(bottom = 1, top = 1, right = 1, left = 1,
                     header = 0.5, footer = 0.5, gutter = 0.5) {
  z <- list(header = header, bottom = bottom, top = top, right = right, left = left, footer = footer, gutter = gutter)
  class(z) <- c("page_mar")
  z
}
#' @export
to_wml.page_mar <- function(x, add_ns = FALSE, ...) {
  out <- sprintf(
    "<w:pgMar w:header=\"%.0f\" w:bottom=\"%.0f\" w:top=\"%.0f\" w:right=\"%.0f\" w:left=\"%.0f\" w:footer=\"%.0f\" w:gutter=\"%.0f\"/>",
    inch_to_tweep(x$header), inch_to_tweep(x$bottom), inch_to_tweep(x$top),
    inch_to_tweep(x$right), inch_to_tweep(x$left), inch_to_tweep(x$footer),
    inch_to_tweep(x$gutter)
  )

  out
}

docx_section_type <- c("continuous", "evenPage", "nextColumn", "nextPage", "oddPage")

#' @export
#' @title Section properties
#' @description A section is a grouping of blocks (ie. paragraphs and tables)
#' that have a set of properties that define pages on which the text will appear.
#'
#' A Section properties object stores information about page composition,
#' such as page size, page orientation, borders and margins.
#' @param page_size page dimensions, an object generated with function [page_size].
#' @param page_margins page margins, an object generated with function [page_mar].
#' @param type Section type. It defines how the contents of the section will be
#' placed relative to the previous section. Available types are "continuous"
#' (begins the section on the next paragraph), "evenPage" (begins on the next
#' even-numbered page), "nextColumn" (begins on the next column on the page),
#' "nextPage" (begins on the following page), "oddPage" (begins on the next
#' odd-numbered page).
#' @param section_columns section columns, an object generated with function [section_columns]. Use NULL (default value) for no content.
#' @param header_default content as a [block_list()] for the default page header. Use NULL (default value) for no content.
#' @param header_even content as a [block_list()] for the even page header. Use NULL (default value) for no content.
#' @param header_first content as a [block_list()] for the first page header. Use NULL (default value) for no content.
#' @param footer_default content as a [block_list()] for the default page footer. Use NULL (default value) for no content.
#' @param footer_even content as a [block_list()] for the even page footer. Use NULL (default value) for no content.
#' @param footer_first content as a [block_list()] for the default page footer. Use NULL (default value) for no content.
#' @examples
#' library(officer)
#'
#' landscape_one_column <- block_section(
#'   prop_section(
#'     page_size = page_size(orient = "landscape"), type = "continuous"
#'   )
#' )
#' landscape_two_columns <- block_section(
#'   prop_section(
#'     page_size = page_size(orient = "landscape"), type = "continuous",
#'     section_columns = section_columns(widths = c(4.75, 4.75))
#'   )
#' )
#'
#' doc_1 <- read_docx()
#' # there starts section with landscape_one_column
#' doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")
#' doc_1 <- body_end_block_section(doc_1, value = landscape_one_column)
#' # there stops section with landscape_one_column
#'
#'
#' # there starts section with landscape_two_columns
#' doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 50), collapse = " "))
#' doc_1 <- body_end_block_section(doc_1, value = landscape_two_columns)
#' # there stops section with landscape_two_columns
#'
#' doc_1 <- body_add_table(doc_1, value = mtcars[1:25, ], style = "table_template")
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#'
#'
#' # an example with headers and footers -----
#' txt_lorem <- rep(
#'   "Purus lectus eros metus turpis mattis platea praesent sed. ",
#'   50
#' )
#' txt_lorem <- paste0(txt_lorem, collapse = "")
#'
#' header_first <- block_list(fpar(ftext("text for first page header")))
#' header_even <- block_list(fpar(ftext("text for even page header")))
#' header_default <- block_list(fpar(ftext("text for default page header")))
#' footer_first <- block_list(fpar(ftext("text for first page footer")))
#' footer_even <- block_list(fpar(ftext("text for even page footer")))
#' footer_default <- block_list(fpar(ftext("text for default page footer")))
#'
#' ps <- prop_section(
#'   header_default = header_default, footer_default = footer_default,
#'   header_first = header_first, footer_first = footer_first,
#'   header_even = header_even, footer_even = footer_even
#' )
#' x <- read_docx()
#' for (i in 1:20) {
#'   x <- body_add_par(x, value = txt_lorem)
#' }
#' x <- body_set_default_section(
#'   x,
#'   value = ps
#' )
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [block_section]
#' @family functions for section definition
#' @section Illustrations:
#'
#' \if{html}{\figure{prop_section_doc_1.png}{options: width=80\%}}
prop_section <- function(page_size = NULL, page_margins = NULL,
                         type = NULL, section_columns = NULL,
                         header_default = NULL, header_even = NULL, header_first = NULL,
                         footer_default = NULL, footer_even = NULL, footer_first = NULL) {
  z <- list()

  if (!is.null(header_default) && !inherits(header_default, "block_list")) {
    stop("header_default must be NULL or an object generated by `block_list()`.")
  }
  z$header_default <- header_default

  if (!is.null(header_even) && !inherits(header_even, "block_list")) {
    stop("header_even must be NULL or an object generated by `block_list()`.")
  }
  z$header_even <- header_even

  if (!is.null(header_first) && !inherits(header_first, "block_list")) {
    stop("header_first must be NULL or an object generated by `block_list()`.")
  }
  z$header_first <- header_first

  if (!is.null(footer_default) && !inherits(footer_default, "block_list")) {
    stop("footer_default must be NULL or an object generated by `block_list()`.")
  }
  z$footer_default <- footer_default

  if (!is.null(footer_even) && !inherits(footer_even, "block_list")) {
    stop("footer_even must be NULL or an object generated by `block_list()`.")
  }
  z$footer_even <- footer_even

  if (!is.null(footer_first) && !inherits(footer_first, "block_list")) {
    stop("footer_first must be NULL or an object generated by `block_list()`.")
  }
  z$footer_first <- footer_first

  if (!is.null(type) && !type %in% docx_section_type) {
    stop("type must be one of ", paste(shQuote(docx_section_type), collapse = ", "))
  }
  z$type <- type

  if (!is.null(page_size) && !inherits(page_size, "page_size")) {
    stop("page_size must be a page_size object")
  }
  z$page_size <- page_size

  if (!is.null(page_margins) && !inherits(page_margins, "page_mar")) {
    stop("page_margins must be a page_mar object")
  }
  z$page_margins <- page_margins

  if (!is.null(section_columns) && !inherits(section_columns, "section_columns")) {
    stop("section_columns must be a section_columns object")
  }
  z$section_columns <- section_columns

  z <- Filter(function(x) !is.null(x), z)
  class(z) <- c("prop_section", "prop")

  z
}

#' @export
to_wml.prop_section <- function(x, add_ns = FALSE, ...) {
  paste0(
    "<w:sectPr w:officer=\"true\"",
    if (add_ns) " xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\"",
    ">",
    if (!is.null(x$header_default)) {
      paste0(
        "<w:headerReference w:type=\"default\">",
        "<w:hdr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$header_default),
        "</w:hdr>",
        "</w:headerReference>"
      )
    },
    if (!is.null(x$header_even)) {
      paste0(
        "<w:headerReference w:type=\"even\">",
        "<w:hdr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$header_even),
        "</w:hdr>",
        "</w:headerReference>"
      )
    },
    if (!is.null(x$header_first)) {
      paste0(
        "<w:headerReference w:type=\"first\">",
        "<w:hdr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$header_first),
        "</w:hdr>",
        "</w:headerReference>"
      )
    },
    if (!is.null(x$footer_default)) {
      paste0(
        "<w:footerReference w:type=\"default\">",
        "<w:ftr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$footer_default),
        "</w:ftr>",
        "</w:footerReference>"
      )
    },
    if (!is.null(x$footer_even)) {
      paste0(
        "<w:footerReference w:type=\"even\">",
        "<w:ftr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$footer_even),
        "</w:ftr>",
        "</w:footerReference>"
      )
    },
    if (!is.null(x$footer_first)) {
      paste0(
        "<w:footerReference w:type=\"first\">",
        "<w:ftr xmlns:wpc=\"http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas\" xmlns:cx=\"http://schemas.microsoft.com/office/drawing/2014/chartex\" xmlns:cx1=\"http://schemas.microsoft.com/office/drawing/2015/9/8/chartex\" xmlns:cx2=\"http://schemas.microsoft.com/office/drawing/2015/10/21/chartex\" xmlns:cx3=\"http://schemas.microsoft.com/office/drawing/2016/5/9/chartex\" xmlns:cx4=\"http://schemas.microsoft.com/office/drawing/2016/5/10/chartex\" xmlns:cx5=\"http://schemas.microsoft.com/office/drawing/2016/5/11/chartex\" xmlns:cx6=\"http://schemas.microsoft.com/office/drawing/2016/5/12/chartex\" xmlns:cx7=\"http://schemas.microsoft.com/office/drawing/2016/5/13/chartex\" xmlns:cx8=\"http://schemas.microsoft.com/office/drawing/2016/5/14/chartex\" xmlns:mc=\"http://schemas.openxmlformats.org/markup-compatibility/2006\" xmlns:aink=\"http://schemas.microsoft.com/office/drawing/2016/ink\" xmlns:am3d=\"http://schemas.microsoft.com/office/drawing/2017/model3d\" xmlns:o=\"urn:schemas-microsoft-com:office:office\" xmlns:oel=\"http://schemas.microsoft.com/office/2019/extlst\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:m=\"http://schemas.openxmlformats.org/officeDocument/2006/math\" xmlns:v=\"urn:schemas-microsoft-com:vml\" xmlns:wp14=\"http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:w10=\"urn:schemas-microsoft-com:office:word\" xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\" xmlns:w15=\"http://schemas.microsoft.com/office/word/2012/wordml\" xmlns:w16cex=\"http://schemas.microsoft.com/office/word/2018/wordml/cex\" xmlns:w16cid=\"http://schemas.microsoft.com/office/word/2016/wordml/cid\" xmlns:w16=\"http://schemas.microsoft.com/office/word/2018/wordml\" xmlns:w16sdtdh=\"http://schemas.microsoft.com/office/word/2020/wordml/sdtdatahash\" xmlns:w16se=\"http://schemas.microsoft.com/office/word/2015/wordml/symex\" xmlns:wpg=\"http://schemas.microsoft.com/office/word/2010/wordprocessingGroup\" xmlns:wpi=\"http://schemas.microsoft.com/office/word/2010/wordprocessingInk\" xmlns:wne=\"http://schemas.microsoft.com/office/word/2006/wordml\" xmlns:wps=\"http://schemas.microsoft.com/office/word/2010/wordprocessingShape\" mc:Ignorable=\"w14 w15 w16se w16cid w16 w16cex w16sdtdh wp14\">",
        to_wml(x$footer_first),
        "</w:ftr>",
        "</w:footerReference>"
      )
    },
    if (!is.null(x$page_margins)) to_wml(x$page_margins),
    if (!is.null(x$page_size)) to_wml(x$page_size),
    if (!is.null(x$type)) paste0("<w:type w:val=\"", x$type, "\"/>"),
    if (!is.null(x$section_columns)) to_wml(x$section_columns) else "<w:cols/>",
    "</w:sectPr>"
  )
}


# external_img ----
#' @export
#' @title External image
#' @description Wraps an image in an object that can then be embedded
#' in a PowerPoint slide or within a Word paragraph.
#'
#' The image is added as a shape in PowerPoint (it is not possible to mix text and
#' images in a PowerPoint form). With a Word document, the image will be
#' added inside a paragraph.
#' @param src image file path
#' @param width,height size of the image file. It can be ignored
#' if parameter `guess_size=TRUE`, see parameter `guess_size`.
#' @param guess_size If package 'magick' is installed, this option
#' can be used (set it to `TRUE`). The images will be read and
#' width and height will be guessed.
#' @param unit unit for width and height, one of "in", "cm", "mm".
#' @param alt alternative text for images
#' @inheritSection ftext usage
#' @examples
#' # wrap r logo with external_img ----
#' srcfile <- file.path(R.home("doc"), "html", "logo.jpg")
#' extimg <- external_img(
#'   src = srcfile, height = 1.06 / 2,
#'   width = 1.39 / 2
#' )
#'
#' # pptx example ----
#' doc <- read_pptx()
#' doc <- add_slide(doc)
#' doc <- ph_with(
#'   x = doc, value = extimg,
#'   location = ph_location_type(type = "body"),
#'   use_loc_size = FALSE
#' )
#' print(doc, target = tempfile(fileext = ".pptx"))
#'
#' fp_t <- fp_text(font.size = 20, color = "red")
#' an_fpar <- fpar(extimg, ftext(" is cool!", fp_t))
#'
#' # docx example ----
#' x <- read_docx()
#' x <- body_add(x, an_fpar)
#' print(x, target = tempfile(fileext = ".docx"))
#' @seealso [ph_with], [body_add], [fpar]
#' @family run functions for reporting
external_img <- function(src, width = .5, height = .2, unit = "in", guess_size = FALSE, alt = "") {
  # note: should it be vectorized
  check_src <- all(grepl("^rId", src)) || all(file.exists(src))
  if (!check_src) {
    stop("src must be a string starting with 'rId' or an existing image filename")
  }

  width <- convin(unit = unit, x = width)
  height <- convin(unit = unit, x = height)

  if (length(src) > 1) {
    if (length(width) == 1) width <- rep(width, length(src))
    if (length(height) == 1) height <- rep(height, length(src))
  }

  if (guess_size) {
    if (!requireNamespace("magick", quietly = TRUE)) {
      stop("package magick is required when using `guess_size` option.")
    }
    sizes <- lapply(src, function(x) {
      if (grepl("\\.svg$", x)) {
        z <- magick::image_read_svg(x)
      } else {
        z <- magick::image_read(x)
      }
      z <- magick::image_data(z)
      attr(z, "dim")[-1]
    })
    sizes <- do.call(rbind, sizes)
    width <- sizes[, 1] / 72
    height <- sizes[, 2] / 72
  }

  class(src) <- c("external_img", "cot", "run")
  attr(src, "dims") <- list(width = width, height = height)
  attr(src, "alt") <- alt
  src
}


#' @export
dim.external_img <- function(x) {
  x <- attr(x, "dims")
  data.frame(width = x$width, height = x$height)
}


#' @export
as.data.frame.external_img <- function(x, ...) {
  dimx <- attr(x, "dims")
  data.frame(path = as.character(x), width = dimx$width, height = dimx$height, alt = attr(x, "alt"), stringsAsFactors = FALSE)
}

temp_blipfill <- function(value, ns = "p") {
  svg_src <- NULL
  if (grepl("\\.svg", value)) {
    if (!requireNamespace("rsvg")) {
      stop("package 'rsvg' is required to convert svg file to rasters.")
    }

    svg_src <- tempfile(fileext = ".svg")
    file.copy(as.character(value), to = svg_src)

    img_src <- tempfile(fileext = ".png")
    rsvg::rsvg_png(svg_src, file = img_src)
  } else {
    img_src <- tempfile(fileext = gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", as.character(value)))
    file.copy(as.character(value), to = img_src)
  }

  if (!is.null(svg_src)) {
    blipfill <- paste0(
      "<", ns, ":blipFill>",
      sprintf("<a:blip cstate=\"print\" r:embed=\"%s\">", img_src),
      "<a:extLst>",
      "<a:ext uri=\"{96DAC541-7B7A-43D3-8B79-37D633B846F1}\">",
      sprintf("<asvg:svgBlip r:embed=\"%s\" xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\"/>", svg_src),
      "</a:ext>",
      "</a:extLst>",
      "</a:blip>",
      "<a:stretch><a:fillRect/></a:stretch>",
      "</", ns, ":blipFill>"
    )
  } else {
    blipfill <- paste0(
      "<", ns, ":blipFill>",
      sprintf("<a:blip cstate=\"print\" r:embed=\"%s\"/>", img_src),
      "<a:stretch><a:fillRect/></a:stretch>",
      "</", ns, ":blipFill>"
    )
  }
  blipfill
}

#' @export
to_wml.external_img <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  blipfill <- temp_blipfill(x, ns = "pic")

  paste0(
    open_tag,
    "<w:rPr/><w:drawing xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\"><wp:inline distT=\"0\" distB=\"0\" distL=\"0\" distR=\"0\">",
    sprintf("<wp:extent cx=\"%.0f\" cy=\"%.0f\"/>", width * 12700 * 72, height * 12700 * 72),
    "<wp:docPr id=\"\" name=\"\" descr=\"", attr(x, "alt"), "\"/>",
    "<wp:cNvGraphicFramePr><a:graphicFrameLocks xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" noChangeAspect=\"1\"/></wp:cNvGraphicFramePr>",
    "<a:graphic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"><a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/picture\"><pic:pic xmlns:pic=\"http://schemas.openxmlformats.org/drawingml/2006/picture\">",
    "<pic:nvPicPr>",
    "<pic:cNvPr id=\"\" name=\"\"/>",
    "<pic:cNvPicPr><a:picLocks noChangeAspect=\"1\" noChangeArrowheads=\"1\"/>",
    "</pic:cNvPicPr></pic:nvPicPr>",
    blipfill,
    "<pic:spPr bwMode=\"auto\"><a:xfrm><a:off x=\"0\" y=\"0\"/>",
    sprintf("<a:ext cx=\"%.0f\" cy=\"%.0f\"/></a:xfrm><a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom><a:noFill/></pic:spPr>", width * 12700, height * 12700),
    "</pic:pic></a:graphicData></a:graphic></wp:inline></w:drawing></w:r>"
  )
}


#' @export
to_pml.external_img <- function(x, add_ns = FALSE,
                                left = 0, top = 0,
                                width = 3, height = 3,
                                bg = "transparent",
                                rot = 0,
                                ph = "<p:ph/>",
                                label = "",
                                ln = sp_line(lwd = 0, linecmpd = "solid", lineend = "rnd"),
                                ...) {
  if (!is.null(bg) && !is.color(bg)) {
    stop("bg must be a valid color.", call. = FALSE)
  }

  bg_str <- solid_fill_pml(bg)
  ln_str <- ln_pml(ln)

  xfrm_str <- a_xfrm_str(left = left, top = top, width = width, height = height, rot = rot)
  if (is.null(ph) || is.na(ph)) {
    ph <- "<p:ph/>"
  }
  blipfill <- temp_blipfill(x, ns = "p")
  id <- uuid_generate()
  str <- "
<p:pic xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">
  <p:nvPicPr>
    <p:cNvPr id=\"%s\" name=\"%s\" descr=\"%s\"/>
    <p:cNvPicPr><a:picLocks noGrp=\"1\"/></p:cNvPicPr>
    <p:nvPr>%s</p:nvPr>
  </p:nvPicPr>
  %s
  <p:spPr>%s<a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom>%s%s</p:spPr>
</p:pic>
"
  sprintf(str, id, label, attr(x, "alt"), ph, blipfill, xfrm_str, bg_str, ln_str)
}



#' @export
to_html.external_img <- function(x, ...) {
  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height
  b64_str <- image_to_base64(as.character(x))
  sprintf("<img style=\"vertical-align:middle;width:%.0fpx;height:%.0fpx;\" src=\"%s\" alt=\"%s\"/>", width * 72, height * 72, b64_str, attr(x, "alt"))
}

# run_footnoteref ----
#' @export
#' @title Word footnote reference
#' @description Wraps a footnote reference in an object that can then be inserted
#' as a run/chunk with [fpar()] or within an R Markdown document.
#' @param prop formatting text properties returned by
#' [fp_text_lite()] or [fp_text()]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @family run functions for reporting
#' @examples
#' run_footnoteref()
#' to_wml(run_footnoteref())
run_footnoteref <- function(prop = NULL) {
  z <- list(pr = prop)
  class(z) <- c("run_footnoteref", "run")
  z
}
#' @export
to_wml.run_footnoteref <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  str <- paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:footnoteRef/></w:r>"
  )

  str
}

# run_footnote ----
#' @export
#' @title Footnote for 'Word'
#' @description Wraps a footnote in an object that can then be inserted
#' as a run/chunk with [fpar()] or within an R Markdown document.
#' @param x a set of blocks to be used as footnote content returned by
#'   function [block_list()].
#' @param prop formatting text properties returned by
#' [fp_text_lite()] or [fp_text()]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @family run functions for reporting
#' @examples
#' library(officer)
#'
#' fp_bold <- fp_text_lite(bold = TRUE)
#' fp_refnote <- fp_text_lite(vertical.align = "superscript")
#'
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#' bl <- block_list(
#'   fpar(ftext("hello", fp_bold)),
#'   fpar(
#'     ftext("hello world", fp_bold),
#'     external_img(src = img.file, height = 1.06, width = 1.39)
#'   )
#' )
#'
#' a_par <- fpar(
#'   "this paragraph contains a note ",
#'   run_footnote(x = bl, prop = fp_refnote),
#'   "."
#' )
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, value = a_par, style = "Normal")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
run_footnote <- function(x, prop = NULL) {
  z <- list(footnote = x, pr = prop)
  class(z) <- c("run_footnote", "run")
  z
}
#' @export
to_wml.run_footnote <- function(x, add_ns = FALSE, ...) {
  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  x$footnote[[1]]$chunks <- append(list(run_footnoteref(x$pr)), x$footnote[[1]]$chunks)
  blocks <- sapply(x$footnote, to_wml)
  blocks <- paste(blocks, collapse = "")

  id <- basename(tempfile(pattern = "footnote"))
  base_ns <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\""
  footnote_xml <- paste0(
    "<w:footnote ", base_ns, " w:id=\"",
    id,
    "\">",
    blocks, "</w:footnote>"
  )

  footnote_ref_xml <- paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:footnoteReference w:id=\"",
    id,
    "\">",
    footnote_xml,
    "</w:footnoteReference>",
    "</w:r>"
  )

  footnote_ref_xml
}

# run_comment ----
#' @export
#' @title Comment for 'Word'
#' @description Add a comment on a run object.
#' @param cmt a set of blocks to be used as comment content returned by
#'   function [block_list()].
#' the "run functions for reporting".
#' @param author comment author.
#' @param date comment date
#' @param initials comment initials
#' @param prop formatting text properties returned by
#' [fp_text_lite()] or [fp_text()]. It also can be NULL in
#' which case, no formatting is defined (the default is applied).
#' @param run a run object, made with a call to one of
#' @examples
#' fp_bold <- fp_text_lite(bold = TRUE)
#' fp_red <- fp_text_lite(color = "red")
#'
#' bl <- block_list(
#'   fpar(ftext("Comment multiple words.", fp_bold)),
#'   fpar(
#'     ftext("Second line.", fp_red)
#'   )
#' )
#'
#' comment1 <- run_comment(
#'   cmt = bl,
#'   run = ftext("with a comment"),
#'   author = "Author Me",
#'   date = Sys.Date(),
#'   initials = "AM"
#' )
#' par1 <- fpar("A paragraph ", comment1)
#'
#' bl <- block_list(
#'   fpar(ftext("Comment a paragraph."))
#' )
#'
#' comment2 <- run_comment(
#'   cmt = bl, run = ftext("A commented paragraph"),
#'   author = "Author You",
#'   date = Sys.Date(),
#'   initials = "AY"
#' )
#' par2 <- fpar(comment2)
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, value = par1, style = "Normal")
#' doc <- body_add_fpar(doc, value = par2, style = "Normal")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family run functions for reporting
run_comment <- function(cmt, run = ftext(""), author = "", date = "", initials = "", prop = NULL) {
  all_run <- FALSE

  if (inherits(run, "run")) {
    run <- list(run)
  }

  if (is.list(run) && !inherits(run, "run")) {
    all_run <- all(vapply(run, inherits, FUN.VALUE = FALSE, what = "run"))
  }

  if (!all_run) {
    stop("`run` must be a run object (ftext for example) or a list of run objects.")
  }

  z <- list(comment = cmt, run = run, author = author, date = date, initials = initials, pr = prop)
  class(z) <- c("run_comment", "run")
  z
}

#' @export
to_wml.run_comment <- function(x, add_ns = FALSE, ...) {
  runs <- lapply(x$run, to_wml, add_ns = add_ns, ...)
  runs <- do.call(paste0, runs)

  open_tag <- wr_ns_no
  if (add_ns) {
    open_tag <- wr_ns_yes
  }

  blocks <- sapply(x$comment, to_wml)
  blocks <- paste(blocks, collapse = "")

  id <- basename(tempfile(pattern = "comment"))
  base_ns <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\""
  comment_xml <- paste0(
    sprintf(
      "<w:comment %s  w:id=\"%s\" w:author=\"%s\" w:date=\"%s\" w:initials=\"%s\">",
      base_ns, id, x$author, x$date, x$initials
    ),
    blocks,
    "</w:comment>"
  )

  comment_ref_xml <- paste0(
    "<w:commentRangeStart w:id=\"", id, "\"/>",
    runs,
    "<w:commentRangeEnd w:id=\"", id, "\"/>",
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:commentReference w:id=\"",
    id,
    "\">",
    comment_xml,
    "</w:commentReference>",
    "</w:r>"
  )

  comment_ref_xml
}

#' @title Write a ggplot Object to PNG File
#'
#' @description Renders a ggplot object to a PNG file using ragg for high-quality output.
#'
#' @param ggobj A ggplot object to render
#' @param width Numeric, width of the output image
#' @param height Numeric, height of the output image
#' @param res Numeric, resolution in DPI (default 200)
#' @param units Character, units for width and height ("in", "cm", "mm", "px") (default "in")
#' @param pointsize Integer, The default pointsize of the device in pt
#' @param scaling scaling factor to apply
#' @param path Character, output file path. If NULL, a temporary file is created (default NULL)
#'
#' @return Character, the path to the created PNG file
#'
#' @importFrom ragg agg_png
#' @importFrom grDevices dev.off
#' @examples
#' plot_in_png(
#'   code = {
#'     barplot(1:10)
#'   },
#'   width = 5,
#'   height = 4,
#'   res = 72,
#'   units = "in"
#' )
#' @export
#' @keywords internal
plot_in_png <- function(
  ggobj = NULL,
  code = NULL,
  width,
  height,
  res = 200,
  units = "in",
  pointsize = 11,
  scaling = 1,
  path = NULL
) {
  if (is.null(path)) {
    path <- tempfile(fileext = ".png")
  }

  if (!is.null(ggobj)) {
    code <- str2lang("print(ggobj)")
  }

  agg_png(
    filename = path,
    width = width,
    height = height,
    units = units,
    background = "transparent",
    res = res,
    pointsize = pointsize,
    scaling = scaling
  )
  tryCatch(
    {
      eval(code)
    },
    finally = dev.off()
  )
  path
}

#' @importFrom openssl base64_encode
#' @export
#' @title Encode Character Vector to Base64
#' @description
#' Encodes one or more elements of a character vector into Base64 format.
#' @param x A character vector. NA values are preserved.
#' @return A character vector of Base64-encoded strings, same length as `x`.
#' @examples
#' as_base64(letters)
#' as_base64(c("hello", NA, "world"))
#' @keywords internal
as_base64 <- function(x) {
  if (!is.character(x)) {
    stop("'x' must be a character vector.")
  }
  z <- vapply(
    x,
    function(elem) {
      if (is.na(elem)) {
        NA_character_
      } else {
        base64_encode(charToRaw(elem))
      }
    },
    NA_character_
  )
  unname(z)
}

#' @importFrom openssl base64_decode
#' @export
#' @title Decode Base64 Vector to Character
#' @description
#' Decodes one or more Base64-encoded elements back into their original character form.
#' @param x A character vector of Base64-encoded strings. NA values are preserved.
#' @return A character vector of decoded strings, same length as `x`.
#' @examples
#' z <- as_base64(c("hello", "world"))
#' from_base64(z)
#' @keywords internal
from_base64 <- function(x) {
  if (!is.character(x)) {
    stop("'x' must be a character vector of Base64 strings.")
  }
  z <- vapply(
    x,
    function(elem) {
      if (is.na(elem)) {
        NA_character_
      } else {
        raw <- tryCatch(
          {
            b64_str <- base64_decode(elem)
            if (all(nchar(b64_str) < 1)) {
              stop("empty result")
            }
            b64_str
          },
          error = function(e) {
            stop(
              "Failed to decode Base64 element: '",
              elem,
              "'.",
              call. = FALSE
            )
          }
        )
        rawToChar(raw)
      }
    },
    NA_character_
  )
  unname(z)
}


#' Convert Data URIs to PNG Files
#'
#' @description Decodes base64-encoded data URIs and writes them to PNG files.
#'
#' @param data_uri Character, a data URI character vector starting with "data:image/png;base64,"
#' @param output_files Character, paths to the output PNG files
#'
#' @return Character, the paths to the created PNG files
#'
#' @importFrom openssl base64_decode
#' @examples
#' rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
#' base64_str <- image_to_base64(rlogo)
#' base64_to_image(
#'   data_uri = base64_str,
#'   output_files = tempfile(fileext = ".jpeg")
#' )
#' @export
#' @keywords internal
base64_to_image <- function(data_uri, output_files) {
  for (i in seq_along(data_uri)) {
    base64_part <- sub("^data:image/[^;]+;base64,", "", data_uri[[i]])
    raw_data <- base64_decode(base64_part)
    writeBin(raw_data, output_files[[i]])
  }

  output_files
}

mime_type <- function(paths) {
  result <- character(length(paths))
  pattern <- "\\.(png|gif|jpg|jpeg|svg|tiff|pdf|webp)$"
  m <- regexpr(pattern = pattern, text = paths)
  result[attr(m, "match.length") > -1] <- regmatches(paths, m)
  result <- gsub("\\.jpg", ".jpeg", result)
  result <- gsub("\\.svg", ".svg+xml", result)
  result <- gsub("^\\.{1}", "", result)
  prefix <- ifelse(result %in% "pdf", "application", "image")
  result <- paste(prefix, result, sep = "/")
  result[attr(m, "match.length") < 0] <- NA_character_
  result
}

#' @importFrom openssl base64_encode
#' @export
#' @title Images to base64
#' @description encodes images into base64 strings.
#' @param filepaths file names.
#' @keywords internal
#' @examples
#' rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
#' base64_str <- image_to_base64(rlogo)
image_to_base64 <- function(filepaths) {
  mimes <- mime_type(paths = filepaths)
  if (any(is.na(mimes))) {
    cli::cli_abort(
      paste0(
        "Unknown image(s) format: ",
        cli::ansi_collapse(
          basename(filepaths)[is.na(mimes)],
          trunc = 5
        )
      )
    )
  }

  if (any(!file.exists(filepaths))) {
    cli::cli_abort(
      paste0(
        "File(s) not found: ",
        cli::ansi_collapse(
          basename(filepaths)[!file.exists(filepaths)],
          trunc = 5
        )
      )
    )
  }

  base64_lst <- mapply(
    FUN = function(filepath, mime) {
      dat <- readBin(
        filepath,
        what = "raw",
        size = 1,
        endian = "little",
        n = file.info(filepath)$size
      )
      base64_str <- base64_encode(bin = dat)
      base64_str <- sprintf("data:%s;base64,%s", mime, base64_str)
      base64_str
    },
    filepath = filepaths,
    mime = mimes,
    SIMPLIFY = FALSE
  )
  unname(unlist(base64_lst))
}

#' @importFrom uuid UUIDgenerate
#' @export
#' @title generates unique identifiers
#' @description generates unique identifiers based
#' on [uuid::UUIDgenerate()].
#' @param n integer, number of unique identifiers to generate.
#' @param ... arguments sent to [uuid::UUIDgenerate()]
#' @keywords internal
#' @examples
#' uuid_generate(n = 5)
uuid_generate <- function(n = 1, ...) {
  UUIDgenerate(n = n, ...)
}

.url_special_chars <- list(
  `&` = "&amp;",
  `<` = "&lt;",
  `>` = "&gt;",
  `'` = "&#39;",
  `"` = "&quot;",
  ` ` = "&nbsp;"
)

#' @export
#' @title officer url encoder
#' @description encode url so that it can be easily
#' decoded when 'officer' write a file to the disk.
#' @param x a character vector of URL
#' @keywords internal
#' @examples
#' officer_url_encode("https://cran.r-project.org/")
officer_url_encode <- function(x) {
  for (chr in names(.url_special_chars)) {
    x <- gsub(chr, .url_special_chars[[chr]], x, fixed = TRUE, useBytes = TRUE)
  }
  Encoding(x) <- "UTF-8"
  x
}

officer_url_decode <- function(x) {
  for (chr in rev(names(.url_special_chars))) {
    x <- gsub(.url_special_chars[[chr]], chr, x, fixed = TRUE, useBytes = TRUE)
  }
  Encoding(x) <- "UTF-8"
  x
}

img.file <- file.path(R.home(component = "doc"), "html", "logo.jpg")
svg_file <- file.path(R.home(component = "doc"), "html", "Rlogo.svg")

ext_img <- external_img(img.file)
ext_svg <- external_img(svg_file)

test_that("add image in HTML", {
  expect_match(
    to_html(ext_svg),
    "<img style=\"vertical-align:middle;width:36px;height:14px;\" src=\"data:image/svg\\+xml;base64,"
  )
  expect_match(
    to_html(ext_img),
    "<img style=\"vertical-align:middle;width:36px;height:14px;\" src=\"data:image/jpeg;base64,"
  )
})

test_that("add image in docx", {
  x <- read_docx()
  x <- body_add_fpar(x, fpar(ext_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)
    node_blip <- xml_find_first(body, "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})

pic <- file.path(R.home("doc"), "html", "logo.jpg")
base_dir <- tempfile()
file1 <- file.path(base_dir, "dir1", "logo1.jpg")
file2 <- file.path(base_dir, "dir2", "logo1.jpg")
file3 <- file.path(base_dir, "dir2", "logo2.jpg")
dir.create(file.path(base_dir, "dir1"), recursive = TRUE)
dir.create(file.path(base_dir, "dir2"), recursive = TRUE)
file.copy(pic, file1)
file.copy(pic, file2)
file.copy(pic, file3)

test_that("add multiple images in docx", {
  x <- read_docx()
  x <- body_add_img(x, src = file1, width = 1, height = 1)
  x <- body_add_img(x, src = file2, width = 1, height = 1)
  x <- body_add_img(x, src = file3, width = 1, height = 1)
  docx_file <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = docx_file)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)
    node_blip <- xml_find_all(body, "//a:blip")
    expect_length(node_blip, 3)
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})

test_that("add image in pptx", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content")
  x <- ph_with(x, ext_img, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = filename)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    node_blip <- xml_find_first(slide$get(), "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})


test_that("add multiple images in pptx", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content")
  x <- ph_with(
    x = x,
    value = external_img(src = file1),
    location = ph_location(left = 0),
    use_loc_size = FALSE
  )
  x <- ph_with(
    x = x,
    value = external_img(src = file2),
    location = ph_location(left = 3),
    use_loc_size = FALSE
  )
  x <- ph_with(
    x = x,
    value = external_img(src = file3),
    location = ph_location(left = 6),
    use_loc_size = FALSE
  )
  pptx_file <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = pptx_file)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()

  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    body <- slide$get()
    node_blip <- xml_find_all(body, "//a:blip")
    expect_length(node_blip, 3)
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})

test_that("add svg in docx", {
  skip_if_not_installed("rsvg")
  x <- read_docx()
  x <- body_add_fpar(x, fpar(ext_svg))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 2)
  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)
    node_blip <- xml_find_first(body, "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
    node_svgblip <- xml_find_first(body, "//asvg:svgBlip")
    expect_false(inherits(node_svgblip, "xml_missing"))
    expect_true(all(xml_attr(node_svgblip, "embed") %in% subset_rel$id))
  }
  new_file <- print(x, target = tempfile(fileext = ".docx"))
  new_folder <- unpack_folder(new_file, tempfile())
  media_files <- list.files(file.path(new_folder, "word", "media"))
  expect_length(media_files, 2)
})

test_that("add svg in pptx", {
  skip_if_not_installed("rsvg")
  x <- read_pptx()
  x <- add_slide(x, "Title and Content")
  x <- ph_with(x, ext_svg, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = filename)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 2)
  if (nrow(subset_rel) > 0) {
    node_blip <- xml_find_first(slide$get(), "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
    node_svgblip <- xml_find_first(slide$get(), "//asvg:svgBlip")
    expect_false(inherits(node_svgblip, "xml_missing"))
    expect_true(all(xml_attr(node_svgblip, "embed") %in% subset_rel$id))
  }
})

test_that("file size does not inflate with identical images", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  doc <- read_docx()
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)
  file1 <- print(doc, target = tempfile(fileext = ".docx"))
  doc <- read_docx(path = file1)
  doc <- body_remove(doc)
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)
  file2 <- print(doc, target = tempfile(fileext = ".docx"))
  expect_equal(file.size(file1), file.size(file2), tolerance = 10)
})


# docx floating image tests ----

test_that("add floating image in docx with default params", {
  float_img <- floating_external_img(
    img.file,
    width = 2,
    height = 1.5,
    pos_x = 1,
    pos_y = 2
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)

  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)

    # Check that anchor element exists (not inline)
    node_anchor <- xml_find_first(body, "//wp:anchor")
    expect_false(inherits(node_anchor, "xml_missing"))

    # Check inline does NOT exist
    node_inline <- xml_find_first(body, "//wp:inline")
    expect_true(inherits(node_inline, "xml_missing"))

    # Check image reference
    node_blip <- xml_find_first(body, "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))

    # Check default wrap distances (0, 0, 0.125, 0.125 inches)
    # 0.125 inches = 114300 EMUs
    expect_equal(xml_attr(node_anchor, "distT"), "0")
    expect_equal(xml_attr(node_anchor, "distB"), "0")
    expect_equal(xml_attr(node_anchor, "distL"), "114300")
    expect_equal(xml_attr(node_anchor, "distR"), "114300")

    # Check default positioning (margin)
    node_pos_h <- xml_find_first(body, "//wp:positionH")
    expect_equal(xml_attr(node_pos_h, "relativeFrom"), "margin")

    node_pos_v <- xml_find_first(body, "//wp:positionV")
    expect_equal(xml_attr(node_pos_v, "relativeFrom"), "margin")

    # Check positions (1 inch = 914400 EMUs, 2 inches = 1828800 EMUs)
    pos_x_offset <- xml_text(xml_find_first(node_pos_h, "wp:posOffset"))
    pos_y_offset <- xml_text(xml_find_first(node_pos_v, "wp:posOffset"))
    expect_equal(pos_x_offset, "914400") # 1 inch
    expect_equal(pos_y_offset, "1828800") # 2 inches

    # Check default wrap (square, bothSides)
    node_wrap <- xml_find_first(body, "//wp:wrapSquare")
    expect_false(inherits(node_wrap, "xml_missing"))
    expect_equal(xml_attr(node_wrap, "wrapText"), "bothSides")

    # Check dimensions (2 inches = 1828800 EMUs, 1.5 inches = 1371600 EMUs)
    node_extent <- xml_find_first(body, "//wp:extent")
    expect_equal(xml_attr(node_extent, "cx"), "1828800") # width
    expect_equal(xml_attr(node_extent, "cy"), "1371600") # height
  }
})


test_that("add floating image with custom positioning", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 0.5,
    pos_y = 1.5,
    pos_h_from = "page",
    pos_v_from = "paragraph"
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  body <- docx_body_xml(x)

  # Check positioning reference
  node_pos_h <- xml_find_first(body, "//wp:positionH")
  expect_equal(xml_attr(node_pos_h, "relativeFrom"), "page")

  node_pos_v <- xml_find_first(body, "//wp:positionV")
  expect_equal(xml_attr(node_pos_v, "relativeFrom"), "paragraph")

  # Check position values (0.5 inch = 457200 EMUs, 1.5 inches = 1371600 EMUs)
  pos_x_offset <- xml_text(xml_find_first(node_pos_h, "wp:posOffset"))
  pos_y_offset <- xml_text(xml_find_first(node_pos_v, "wp:posOffset"))
  expect_equal(pos_x_offset, "457200")
  expect_equal(pos_y_offset, "1371600")
})


test_that("add floating image with custom wrapping", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    wrap_type = "tight",
    wrap_side = "left"
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  body <- docx_body_xml(x)

  # Check wrap type
  node_wrap <- xml_find_first(body, "//wp:wrapTight")
  expect_false(inherits(node_wrap, "xml_missing"))
  expect_equal(xml_attr(node_wrap, "wrapText"), "left")
})


test_that("add floating image with wrap topAndBottom", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    wrap_type = "topAndBottom"
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  body <- docx_body_xml(x)

  # Check wrap type
  node_wrap <- xml_find_first(body, "//wp:wrapTopAndBottom")
  expect_false(inherits(node_wrap, "xml_missing"))
})


test_that("add floating image with custom distances", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    wrap_dist_top = 0.1,
    wrap_dist_bottom = 0.2,
    wrap_dist_left = 0.3,
    wrap_dist_right = 0.4
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  body <- docx_body_xml(x)
  node_anchor <- xml_find_first(body, "//wp:anchor")

  # Check custom distances
  # 0.1 inch = 91440 EMUs
  # 0.2 inch = 182880 EMUs
  # 0.3 inch = 274320 EMUs
  # 0.4 inch = 365760 EMUs
  expect_equal(xml_attr(node_anchor, "distT"), "91440")
  expect_equal(xml_attr(node_anchor, "distB"), "182880")
  expect_equal(xml_attr(node_anchor, "distL"), "274320")
  expect_equal(xml_attr(node_anchor, "distR"), "365760")
})


test_that("add floating image with wrap none", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    wrap_type = "none"
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  body <- docx_body_xml(x)

  # Check wrap type
  node_wrap <- xml_find_first(body, "//wp:wrapNone")
  expect_false(inherits(node_wrap, "xml_missing"))
})


test_that("add floating image with all custom params", {
  float_img <- floating_external_img(
    img.file,
    width = 2.5,
    height = 1.8,
    pos_x = 0.75,
    pos_y = 1.25,
    pos_h_from = "column",
    pos_v_from = "line",
    wrap_type = "through",
    wrap_side = "right",
    wrap_dist_top = 0.05,
    wrap_dist_bottom = 0.15,
    wrap_dist_left = 0.25,
    wrap_dist_right = 0.35
  )

  x <- read_docx()
  x <- body_add_fpar(x, fpar(float_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  body <- docx_body_xml(x)
  node_anchor <- xml_find_first(body, "//wp:anchor")

  # Check positioning
  node_pos_h <- xml_find_first(body, "//wp:positionH")
  expect_equal(xml_attr(node_pos_h, "relativeFrom"), "column")
  pos_x_offset <- xml_text(xml_find_first(node_pos_h, "wp:posOffset"))
  expect_equal(pos_x_offset, "685800") # 0.75 inch

  node_pos_v <- xml_find_first(body, "//wp:positionV")
  expect_equal(xml_attr(node_pos_v, "relativeFrom"), "line")
  pos_y_offset <- xml_text(xml_find_first(node_pos_v, "wp:posOffset"))
  expect_equal(pos_y_offset, "1143000") # 1.25 inches

  # Check wrapping
  node_wrap <- xml_find_first(body, "//wp:wrapThrough")
  expect_false(inherits(node_wrap, "xml_missing"))
  expect_equal(xml_attr(node_wrap, "wrapText"), "right")

  # Check distances
  expect_equal(xml_attr(node_anchor, "distT"), "45720") # 0.05 inch
  expect_equal(xml_attr(node_anchor, "distB"), "137160") # 0.15 inch
  expect_equal(xml_attr(node_anchor, "distL"), "228600") # 0.25 inch
  expect_equal(xml_attr(node_anchor, "distR"), "320040") # 0.35 inch

  # Check dimensions (2.5 inches = 2286000 EMUs, 1.8 inches = 1645920 EMUs)
  node_extent <- xml_find_first(body, "//wp:extent")
  expect_equal(xml_attr(node_extent, "cx"), "2286000")
  expect_equal(xml_attr(node_extent, "cy"), "1645920")
})


# rtf floating image tests ----

test_that("add floating image in RTF with default params", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 0.75,
    pos_x = 0.5,
    pos_y = 1
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check shape structure exists
  expect_true(grepl("\\{\\\\shp", rtf_text))
  expect_true(grepl("\\{\\\\\\*\\\\shpinst", rtf_text))

  # Check shapeType = 75 (picture frame)
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn shapeType\\}\\{\\\\sv 75\\}\\}",
    rtf_text
  ))

  # Check position (0.5 inch = 720 twips, 1 inch = 1440 twips)
  expect_true(grepl("\\\\shpleft720", rtf_text))
  expect_true(grepl("\\\\shptop1440", rtf_text))
  expect_true(grepl("\\\\shpright2160", rtf_text)) # left (720) + width (1440)
  expect_true(grepl("\\\\shpbottom2520", rtf_text)) # top (1440) + height (1080)

  # Check default positioning (margin)
  expect_true(grepl("\\\\shpbxmargin", rtf_text))
  expect_true(grepl("\\\\shpbymargin", rtf_text))

  # Check default wrap (square, both sides)
  expect_true(grepl("\\\\shpwr2", rtf_text))
  expect_true(grepl("\\\\shpwrk0", rtf_text))

  # Check image in front of text
  expect_true(grepl("\\\\shpfblwtxt0", rtf_text))

  # Check picture data exists
  expect_true(grepl("\\{\\\\sp\\{\\\\sn pib\\}", rtf_text))
  expect_true(grepl("\\{\\\\pict\\\\pngblip", rtf_text))

  # Check structure ends correctly
  expect_true(grepl("\\\\par\\}\\}\\}", rtf_text))
})


test_that("add floating image in RTF with custom positioning", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 1.5,
    pos_y = 2.5,
    pos_h_from = "page",
    pos_v_from = "paragraph"
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check position (1.5 inches = 2160 twips, 2.5 inches = 3600 twips)
  expect_true(grepl("\\\\shpleft2160", rtf_text))
  expect_true(grepl("\\\\shptop3600", rtf_text))

  # Check positioning reference
  expect_true(grepl("\\\\shpbxpage", rtf_text))
  expect_true(grepl("\\\\shpbypara", rtf_text))
})


test_that("add floating image in RTF with tight wrap left", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 1,
    pos_y = 1,
    wrap_type = "tight",
    wrap_side = "left"
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check wrap type (tight = 4)
  expect_true(grepl("\\\\shpwr4", rtf_text))

  # Check wrap side (left = 1)
  expect_true(grepl("\\\\shpwrk1", rtf_text))
})


test_that("add floating image in RTF with topAndBottom wrap", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 1,
    pos_y = 1,
    wrap_type = "topAndBottom"
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check wrap type (topAndBottom = 1)
  expect_true(grepl("\\\\shpwr1", rtf_text))
})


test_that("add floating image in RTF with through wrap right", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 1,
    pos_y = 1,
    wrap_type = "through",
    wrap_side = "right"
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check wrap type (through = 5)
  expect_true(grepl("\\\\shpwr5", rtf_text))

  # Check wrap side (right = 2)
  expect_true(grepl("\\\\shpwrk2", rtf_text))
})


test_that("add floating image in RTF with none wrap", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 1,
    pos_y = 1,
    wrap_type = "none"
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check wrap type (none = 3)
  expect_true(grepl("\\\\shpwr3", rtf_text))
})


test_that("add floating image in RTF with custom wrap distances", {
  float_img <- floating_external_img(
    img.file,
    width = 1,
    height = 1,
    pos_x = 1,
    pos_y = 1,
    wrap_dist_top = 0.1,
    wrap_dist_bottom = 0.2,
    wrap_dist_left = 0.15,
    wrap_dist_right = 0.25
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check wrap distances in EMUs
  # 0.1 inch = 91440 EMUs
  # 0.2 inch = 182880 EMUs
  # 0.15 inch = 137160 EMUs
  # 0.25 inch = 228600 EMUs
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dxWrapDistLeft\\}\\{\\\\sv 137160\\}\\}",
    rtf_text
  ))
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dxWrapDistRight\\}\\{\\\\sv 228600\\}\\}",
    rtf_text
  ))
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dyWrapDistTop\\}\\{\\\\sv 91440\\}\\}",
    rtf_text
  ))
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dyWrapDistBottom\\}\\{\\\\sv 182880\\}\\}",
    rtf_text
  ))
})


test_that("add floating image in RTF with all custom params", {
  float_img <- floating_external_img(
    img.file,
    width = 2,
    height = 1.5,
    pos_x = 0.75,
    pos_y = 1.25,
    pos_h_from = "column",
    pos_v_from = "page",
    wrap_type = "square",
    wrap_side = "largest",
    wrap_dist_top = 0.05,
    wrap_dist_bottom = 0.1,
    wrap_dist_left = 0.15,
    wrap_dist_right = 0.2
  )

  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(float_img))
  rtf_file <- print(doc, target = tempfile(fileext = ".rtf"))

  rtf_content <- readLines(rtf_file, warn = FALSE)
  rtf_text <- paste(rtf_content, collapse = "")

  # Check position (0.75 inch = 1080 twips, 1.25 inch = 1800 twips)
  expect_true(grepl("\\\\shpleft1080", rtf_text))
  expect_true(grepl("\\\\shptop1800", rtf_text))
  # Right = left + width = 1080 + 2880 = 3960
  # Bottom = top + height = 1800 + 2160 = 3960
  expect_true(grepl("\\\\shpright3960", rtf_text))
  expect_true(grepl("\\\\shpbottom3960", rtf_text))

  # Check positioning reference
  expect_true(grepl("\\\\shpbxcolumn", rtf_text))
  expect_true(grepl("\\\\shpbypage", rtf_text))

  # Check wrap type (square = 2) and side (largest = 3)
  expect_true(grepl("\\\\shpwr2", rtf_text))
  expect_true(grepl("\\\\shpwrk3", rtf_text))

  # Check wrap distances
  # 0.05 inch = 45720 EMUs
  # 0.1 inch = 91440 EMUs
  # 0.15 inch = 137160 EMUs
  # 0.2 inch = 182880 EMUs
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dyWrapDistTop\\}\\{\\\\sv 45720\\}\\}",
    rtf_text
  ))
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dyWrapDistBottom\\}\\{\\\\sv 91440\\}\\}",
    rtf_text
  ))
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dxWrapDistLeft\\}\\{\\\\sv 137160\\}\\}",
    rtf_text
  ))
  expect_true(grepl(
    "\\{\\\\sp\\{\\\\sn dxWrapDistRight\\}\\{\\\\sv 182880\\}\\}",
    rtf_text
  ))

  # Check picture dimensions in twips (2 inch = 2880 twips, 1.5 inch = 2160 twips)
  expect_true(grepl("\\\\picwgoal2880", rtf_text))
  expect_true(grepl("\\\\pichgoal2160", rtf_text))
})


# plot_in_png tests ----

test_that("plot_in_png with ggplot object", {
  skip_if_not_installed("ggplot2")

  gg <- ggplot2::ggplot(mtcars, ggplot2::aes(x = mpg, y = hp)) +
    ggplot2::geom_point()

  png_file <- plot_in_png(
    ggobj = gg,
    width = 5,
    height = 4,
    res = 72,
    units = "in"
  )

  expect_true(file.exists(png_file))
  expect_match(png_file, "\\.png$")
  expect_true(file.size(png_file) > 0)
})

test_that("plot_in_png with code expression", {
  png_file <- plot_in_png(
    code = {
      plot(1:10, 1:10)
    },
    width = 5,
    height = 4,
    res = 72,
    units = "in"
  )

  expect_true(file.exists(png_file))
  expect_match(png_file, "\\.png$")
  expect_true(file.size(png_file) > 0)
})

test_that("plot_in_png with custom path", {
  custom_path <- tempfile(fileext = ".png")

  png_file <- plot_in_png(
    code = {
      barplot(1:5)
    },
    width = 4,
    height = 3,
    res = 96,
    units = "in",
    path = custom_path
  )

  expect_equal(png_file, custom_path)
  expect_true(file.exists(custom_path))
})


# as_base64 and from_base64 tests ----

test_that("as_base64 with multiple values", {
  input <- c("hello", "world", "test")
  result <- as_base64(input)

  expect_type(result, "character")
  expect_length(result, 3)
  expect_false(any(is.na(result)))
})

test_that("as_base64 with NA values", {
  input <- c("hello", NA_character_, "world")
  result <- as_base64(input)

  expect_length(result, 3)
  expect_equal(result[1], as_base64("hello"))
  expect_true(is.na(result[2]))
  expect_equal(result[3], as_base64("world"))
})

test_that("as_base64 with invalid input", {
  expect_error(as_base64(123), "'x' must be a character vector")
  expect_error(as_base64(list("a", "b")), "'x' must be a character vector")
})

test_that("from_base64 with multiple values", {
  original <- c("hello", "world", "test")
  encoded <- as_base64(original)
  decoded <- from_base64(encoded)

  expect_equal(decoded, original)
})

test_that("from_base64 with NA values", {
  encoded <- c(as_base64("hello"), NA_character_, as_base64("world"))
  result <- from_base64(encoded)

  expect_length(result, 3)
  expect_equal(result[1], "hello")
  expect_true(is.na(result[2]))
  expect_equal(result[3], "world")
})

test_that("from_base64 with invalid input type", {
  expect_error(from_base64(123), "'x' must be a character vector")
})

test_that("from_base64 with invalid base64 string", {
  expect_error(
    from_base64("not_valid_base64!!!"),
    "Failed to decode Base64 element"
  )
})


# base64_to_image tests ----

test_that("base64_to_image converts data URI to image file", {
  img1 <- file.path(R.home("doc"), "html", "logo.jpg")
  img2 <- file.path(R.home("doc"), "html", "Rlogo.svg")
  base64_str <- image_to_base64(c(img1, img2))

  output_files <- c(
    tempfile(fileext = ".jpg"),
    tempfile(fileext = ".svg")
  )
  result <- base64_to_image(base64_str, output_files = output_files)

  expect_equal(result, output_files)
  expect_true(all(file.exists(output_files)))
  expect_true(all(file.size(output_files) > 0))
})


# image_to_base64 error handling tests ----

test_that("image_to_base64 with multiple files", {
  img1 <- file.path(R.home("doc"), "html", "logo.jpg")
  img2 <- file.path(R.home("doc"), "html", "Rlogo.svg")

  result <- image_to_base64(c(img1, img2))

  expect_type(result, "character")
  expect_length(result, 2)
  expect_match(result[1], "^data:image/jpeg;base64,")
  expect_match(result[2], "^data:image/svg\\+xml;base64,")
})

test_that("image_to_base64 with unknown format", {
  temp_file <- tempfile(fileext = ".xyz")
  writeLines("test", temp_file)

  expect_error(
    image_to_base64(temp_file),
    "Unknown image\\(s\\) format"
  )
})

test_that("image_to_base64 with non-existent file", {
  fake_file <- tempfile(fileext = ".png")

  expect_error(
    image_to_base64(fake_file),
    "File\\(s\\) not found"
  )
})

test_that("image_to_base64 with multiple non-existent files", {
  fake1 <- tempfile(fileext = ".png")
  fake2 <- tempfile(fileext = ".jpg")

  expect_error(
    image_to_base64(c(fake1, fake2)),
    "File\\(s\\) not found"
  )
})

library(officer)

x <- read_pptx()
x <- add_slide(x, "Title and Content")
x <- ph_with(x, "Hello world 1", location = ph_location_type())
x <- add_slide(x, "Title and Content")
x <- ph_with(x, "Hello world 2", location = ph_location_type())
x <- move_slide(x, index = 1, to = 2)

complete_styles_mapping <- function(
  styles_id_from,
  style_mapping,
  styles_info_tbl_to,
  styles_info_tbl_from,
  style_type = "paragraph",
  document_part_label
) {
  if (length(styles_id_from) < 1L) {
    return(style_mapping)
  }

  style_data_to <- styles_info_tbl_to[
    styles_info_tbl_to$style_type %in% style_type,
  ]
  style_data_from <- styles_info_tbl_from[
    styles_info_tbl_from$style_type %in% style_type,
  ]

  def_name_to <- head(style_data_to$style_name[style_data_to$is_default], n = 1)

  styles_name_from <- style_data_from$style_name[
    match(
      styles_id_from,
      style_data_from$style_id,
      nomatch = head(which(style_data_from$is_default), n = 1)
    )
  ]

  missing_styles <- setdiff(styles_name_from, unlist(style_mapping))

  if (length(missing_styles) > 0) {
    possible_match <- style_data_to$style_name[
      style_data_to$style_name %in% missing_styles
    ]
    for (amatch in possible_match) {
      if (is.null(style_mapping[[amatch]])) {
        style_mapping[[amatch]] <- character()
      }
      style_mapping[[amatch]] <- append(style_mapping[[amatch]], missing_styles)
    }
    missing_styles <- setdiff(missing_styles, possible_match)
  }

  if (length(missing_styles) > 0) {
    missing_styles_cli <- style_data_from[
      style_data_from$style_name %in% missing_styles,
      c("style_name", "style_id")
    ]
    missing_styles_cli <- paste0(
      "Style ",
      shQuote(missing_styles_cli$style_name),
      " with id {.code ",
      missing_styles_cli$style_id,
      "} could not be mapped"
    )

    names(missing_styles_cli) <- rep("*", length(missing_styles_cli))
    cli::cli_warn(
      c(
        "!" = "Style(s) mapping(s) for '{style_type}s' are missing in the {document_part_label} of the document:",
        missing_styles_cli,
        "i" = "Style name '{def_name_to}' is used instead."
      )
    )

    if (is.null(style_mapping[[def_name_to]])) {
      style_mapping[[def_name_to]] <- character()
    }
    style_mapping[[def_name_to]] <- append(
      style_mapping[[def_name_to]],
      missing_styles
    )
  }

  style_mapping
}

style_mapping_fortify <- function(
  style_mapping,
  styles_ids,
  styles_info_tbl_from,
  styles_info_tbl_to,
  style_type = "paragraph",
  document_part_label
) {
  style_mapping <- complete_styles_mapping(
    style_mapping = style_mapping,
    styles_id_from = styles_ids,
    styles_info_tbl_to = styles_info_tbl_to,
    styles_info_tbl_from = styles_info_tbl_from,
    style_type = style_type,
    document_part_label = document_part_label
  )

  style_mapping <- mapply(
    function(name_from, name_to) {
      data.frame(
        name_from = name_from,
        name_to = rep(name_to, length(name_from))
      )
    },
    name_from = style_mapping,
    name_to = names(style_mapping),
    SIMPLIFY = FALSE
  )
  style_mapping <- do.call(rbind, style_mapping)
  row.names(style_mapping) <- NULL
  if (is.null(style_mapping)) {
    style_mapping <- data.frame(
      name_from = character(0),
      name_to = character(0)
    )
  }

  sty_par_info_from <- styles_info_tbl_from[
    styles_info_tbl_from$style_type %in% style_type,
    c("style_id", "style_name")
  ]
  sty_par_info_to <- styles_info_tbl_to[
    styles_info_tbl_to$style_type %in% style_type,
    c("style_id", "style_name")
  ]

  style_mapping_par <- merge(
    x = style_mapping,
    y = sty_par_info_to,
    by.x = "name_to",
    by.y = "style_name",
    all.x = FALSE,
    all.y = FALSE
  )
  names(style_mapping_par)[names(style_mapping_par) %in% "style_id"] <- "id_to"
  style_mapping_par <- merge(
    x = style_mapping_par,
    y = sty_par_info_from,
    by.x = "name_from",
    by.y = "style_name",
    all.x = TRUE,
    all.y = FALSE
  )
  names(style_mapping_par)[
    names(style_mapping_par) %in% "style_id"
  ] <- "id_from"
  style_mapping_par <- style_mapping_par[c("id_from", "id_to")]

  style_mapping_par
}

numberings_append_xml <- function(file_numbering_from, file_numbering_to) {
  numbering_from <- read_xml(file_numbering_from)
  abstractnum_from <- xml_find_all(numbering_from, "w:abstractNum")
  num_from <- xml_find_all(numbering_from, "w:num")

  numbering_to <- read_xml(file_numbering_to)
  abstractnum_to <- xml_find_all(numbering_to, "w:abstractNum")
  num_to <- xml_find_all(numbering_to, "w:num")

  used_num_id_to <- xml_attr(num_to, "numId")
  used_num_id_to <- as.integer(used_num_id_to)
  used_num_id_to_next <- max(used_num_id_to) + 1L

  abst_num_id_to <- xml_attr(abstractnum_to, "abstractNumId")
  abst_num_id_to <- as.integer(abst_num_id_to)
  abst_num_id_to_next <- max(abst_num_id_to) + 1L

  used_num_id_from <- xml_attr(num_from, "numId")
  used_num_id_from <- as.integer(used_num_id_from)
  used_num_id_from_next <- max(used_num_id_from) + 1L

  abst_num_id_from <- xml_attr(abstractnum_from, "abstractNumId")
  abst_num_id_from <- as.integer(abst_num_id_from)
  abst_num_id_from_next <- max(abst_num_id_from) + 1L

  for (id_from in abst_num_id_from) {
    abstractNumFrom <- xml_find_first(
      numbering_from,
      sprintf(
        "w:abstractNum[@w:abstractNumId='%s']",
        as.character(id_from)
      )
    )
    xml_attr(abstractNumFrom, "w:abstractNumId") <- as.character(
      abst_num_id_to_next
    )

    usedAbsNumFrom <- xml_find_all(
      numbering_from,
      sprintf(
        "w:num/w:abstractNumId[@w:val='%s']",
        as.character(id_from)
      )
    )
    xml_attr(usedAbsNumFrom, "w:val") <- as.character(abst_num_id_to_next)
    abst_num_id_to_next <- abst_num_id_to_next + 1L
  }

  mapping_from <- data.frame(
    from = used_num_id_from,
    to = seq(used_num_id_to_next, along.with = used_num_id_from, by = 1L)
  )

  for (i in seq_len(nrow(mapping_from))) {
    id_from <- mapping_from$from[i]
    id_to <- mapping_from$to[i]
    usedNumFrom <- xml_find_all(
      numbering_from,
      sprintf("w:num[@w:numId='%s']", as.character(id_from))
    )
    xml_attr(usedNumFrom, "w:numId") <- as.character(id_to)
  }

  for (node in abstractnum_from) {
    xml_add_sibling(
      xml_find_first(numbering_to, "w:abstractNum"),
      node
    )
  }
  for (node in num_from) {
    xml_add_sibling(
      xml_find_first(numbering_to, "w:num"),
      node
    )
  }

  write_xml(
    numbering_to,
    file = file_numbering_to
  )

  mapping_from
}

# docx_part -----
docx_part <- R6Class(
  "docx_part",
  inherit = openxml_document,
  public = list(
    initialize = function(path, main_file, cursor, body_xpath) {
      super$initialize("word")
      private$package_dir <- path
      private$body_xpath <- body_xpath
      super$feed(file.path(private$package_dir, "word", main_file))
      private$cursor <- cursor
    },
    length = function() {
      xml_length(xml_find_first(self$get(), private$body_xpath))
    },
    patch_wml = function(
      package_dir,
      styles_info_tbl_from,
      styles_info_tbl_to,
      numbering_mapping,
      par_style_mapping = list(),
      run_style_mapping = list(),
      tbl_style_mapping = list(),
      additional_ns = character(0),
      prepend_chunks_on_styles = list()
    ) {
      doc_str <- self$encode_wml_str(additional_ns = additional_ns)
      document_rels <- self$rel_df()

      # images processing -----
      doc_from_img <- document_rels[basename(document_rels$type) %in% "image", ]
      for (i in seq_len(nrow(doc_from_img))) {
        fileext <- paste0(".", tools::file_ext(doc_from_img$target[i]))
        new_file <- tempfile(fileext = fileext)
        file.copy(
          from = file.path(
            package_dir,
            "word/media",
            basename(doc_from_img$target[i])
          ),
          to = new_file,
          overwrite = TRUE
        )
        pat <- "r:embed=\"%s\""
        pat <- sprintf(pat, doc_from_img$id[i])
        m <- gregexpr(pat, doc_str)
        regmatches(doc_str, m) <- sprintf(
          "r:embed=\"%s\"",
          new_file
        )
      }

      # external links processing -----
      doc_from_hl <- document_rels[
        basename(document_rels$type) %in% "hyperlink",
      ]
      for (i in seq_len(nrow(doc_from_hl))) {
        pat <- "r:id=\"%s\""
        pat <- sprintf(pat, doc_from_hl$id[i])
        m <- gregexpr(pat, doc_str)
        regmatches(doc_str, m) <- sprintf("r:id=\"%s\"", doc_from_hl$target[i])
      }

      # numberings processing -----
      for (i in seq_len(nrow(numbering_mapping))) {
        id_from <- numbering_mapping$from[i]
        id_to <- numbering_mapping$to[i]
        m <- gregexpr(
          sprintf("<w:numId w:val=\"%s\"/>", as.character(id_from)),
          doc_str,
          fixed = TRUE
        )
        regmatches(doc_str, m) <- sprintf(
          "<w:numId w:val=\"%s\"/>",
          as.character(id_to)
        )
      }

      sty_par_info_to <- styles_info_tbl_to[
        styles_info_tbl_to$style_type %in% "paragraph",
      ]
      sty_par_info_from <- styles_info_tbl_from[
        styles_info_tbl_from$style_type %in% "paragraph",
      ]

      # append chunks when specific styles are found -----
      for (style in names(prepend_chunks_on_styles)) {
        style_id <- sty_par_info_from$style_id[
          sty_par_info_from$style_name %in% style
        ]
        # find all paragraphs with this style
        match_pstyle <- grep(
          sprintf("w:pStyle w:val=\"%s\"", style_id),
          doc_str
        )
        # find all </w:pPr> after each match
        match_end_ppr <- grep("</w:pPr>", doc_str)
        for (par_i in match_pstyle) {
          # find next </w:pPr>
          current_match_end_ppr <- match_end_ppr[match_end_ppr > par_i]
          current_match_end_ppr <- head(current_match_end_ppr, n = 1)
          # prepend chunk
          if (length(current_match_end_ppr) == 1) {
            doc_str[current_match_end_ppr] <- paste0(
              doc_str[current_match_end_ppr],
              to_wml(prepend_chunks_on_styles[[style]])
            )
          }
        }
      }

      # par styles processing -----
      m <- gregexpr("w:pStyle w:val=\"[[:alnum:]]+\"", doc_str)
      zz <- regmatches(doc_str, m)
      zz <- unlist(zz)
      p_styles <- gsub("w:pStyle w:val=\"([[:alnum:]]+)\"", "\\1", zz)
      p_styles <- unique(p_styles)
      p_styles <- setdiff(p_styles, sty_par_info_to$style_id)

      mapping_styles <- style_mapping_fortify(
        styles_ids = p_styles,
        style_mapping = par_style_mapping,
        styles_info_tbl_from = styles_info_tbl_from,
        styles_info_tbl_to = styles_info_tbl_to,
        style_type = "paragraph",
        document_part_label = self$document_part_label()
      )
      for (i in seq_len(nrow(mapping_styles))) {
        doc_str <- gsub(
          paste0("w:pStyle w:val=\"", mapping_styles$id_from[i], "\""),
          sprintf("w:pStyle w:val=\"%s\"", mapping_styles$id_to[i]),
          doc_str
        )
      }

      # runs/characters styles processing -----
      sty_chr_info_to <- styles_info_tbl_to[
        styles_info_tbl_to$style_type %in% "character",
      ]

      m <- gregexpr("w:rStyle w:val=\"[[:alnum:]]+\"", doc_str)
      zz <- regmatches(doc_str, m)
      zz <- unlist(zz)
      r_styles <- gsub("w:rStyle w:val=\"([[:alnum:]]+)\"", "\\1", zz)
      r_styles <- unique(r_styles)
      r_styles <- setdiff(r_styles, sty_chr_info_to$style_id)

      mapping_styles <- style_mapping_fortify(
        styles_ids = r_styles,
        style_mapping = run_style_mapping,
        styles_info_tbl_from = styles_info_tbl_from,
        styles_info_tbl_to = styles_info_tbl_to,
        style_type = "character",
        document_part_label = self$document_part_label()
      )
      for (i in seq_len(nrow(mapping_styles))) {
        doc_str <- gsub(
          paste0("w:rStyle w:val=\"", mapping_styles$id_from[i], "\""),
          sprintf("w:rStyle w:val=\"%s\"", mapping_styles$id_to[i]),
          doc_str
        )
      }

      # tables styles processing -----
      sty_tab_info_to <- styles_info_tbl_to[
        styles_info_tbl_to$style_type %in% "table",
      ]

      m <- gregexpr("w:tblStyle w:val=\"[[:alnum:]]+\"", doc_str)
      zz <- regmatches(doc_str, m)
      zz <- unlist(zz)
      tbl_styles <- gsub("w:tblStyle w:val=\"([[:alnum:]]+)\"", "\\1", zz)
      tbl_styles <- unique(tbl_styles)
      tbl_styles <- setdiff(tbl_styles, sty_tab_info_to$style_id)
      mapping_styles <- style_mapping_fortify(
        styles_ids = tbl_styles,
        style_mapping = tbl_style_mapping,
        styles_info_tbl_from = styles_info_tbl_from,
        styles_info_tbl_to = styles_info_tbl_to,
        style_type = "table",
        document_part_label = self$document_part_label()
      )
      for (i in seq_len(nrow(mapping_styles))) {
        doc_str <- gsub(
          paste0("w:tblStyle w:val=\"", mapping_styles$id_from[i], "\""),
          sprintf("w:tblStyle w:val=\"%s\"", mapping_styles$id_to[i]),
          doc_str
        )
      }
      doc_str
    }
  ),
  private = list(
    package_dir = NULL,
    cursor = NULL,
    body_xpath = NULL
  )
)

# body_part -----
body_part <- R6Class(
  "body_part",
  inherit = docx_part,
  public = list(
    document_part_label = function() {
      "body"
    },
    encode_wml_str = function(additional_ns) {
      body <- self$get()
      body <- xml_find_first(body, "w:body")

      # sections are removed
      xml_remove(xml_find_all(body, "//w:sectPr"))

      # comments are removed
      xml_remove(xml_find_all(body, "//w:commentRangeStart"))
      xml_remove(xml_find_all(body, "//w:commentRangeEnd"))
      xml_remove(xml_find_all(body, "//w:commentReference"))

      chr_body <- xml_document_to_chrs(body)
      chr_body
    }
  )
)


# footnotes_part -----
footnotes_part <- R6Class(
  "footnotes_part",
  inherit = docx_part,
  public = list(
    document_part_label = function() {
      "footnotes"
    },
    encode_wml_str = function(additional_ns) {
      footnotes <- self$get()
      footnotes <- xml_find_all(footnotes, "w:footnote[not(@w:type)]")

      chr_footnotes <- vapply(
        footnotes,
        function(x) {
          str <- sapply(xml_children(x), as.character)
          paste0(str, collapse = "")
        },
        FUN.VALUE = ""
      )
      add_ns_str <- sprintf(
        " xmlns:%s=\"%s\"",
        names(additional_ns),
        additional_ns
      )
      add_ns_str <- paste0(add_ns_str, collapse = "")

      chr_footnotes <- paste0(
        "<w:footnoteReference w:officer=\"true\" w:id=\"%s\">",
        "<w:footnote ",
        add_ns_str,
        " xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:id=\"%s\">",
        chr_footnotes,
        "</w:footnote>",
        "</w:footnoteReference>"
      )
      names(chr_footnotes) <- xml_attr(footnotes, "id")

      chr_footnotes
    }
  )
)

#' @export
#' @importFrom zip zip
#' @title compress a folder
#' @description compress a folder to a target file. The
#' function returns the complete path to target file.
#' @param folder folder to compress
#' @param target path of the archive to create
#' @keywords internal
pack_folder <- function(folder, target) {
  target <- absolute_path(target)
  dir_fi <- dirname(target)

  if (!file.exists(dir_fi)) {
    stop("directory ", shQuote(dir_fi), " does not exist.", call. = FALSE)
  } else if (file.access(dir_fi) < 0) {
    stop("can not write to directory ", shQuote(dir_fi), call. = FALSE)
  } else if (file.exists(target) && file.access(target) < 0) {
    stop(shQuote(target), " already exists and is not writable", call. = FALSE)
  } else if (!file.exists(target)) {
    old_warn <- getOption("warn")
    options(warn = -1)
    x <- tryCatch(
      {
        cat("", file = target)
        TRUE
      },
      error = function(e) FALSE,
      finally = unlink(target, force = TRUE)
    )
    options(warn = old_warn)
    if (!x) {
      stop(
        shQuote(target),
        " cannot be written, please check your permissions.",
        call. = FALSE
      )
    }
  }

  curr_wd <- getwd()
  setwd(folder)
  tryCatch(
    zip::zipr(
      zipfile = target,
      include_directories = FALSE,
      files = list.files(path = ".", all.files = FALSE),
      recurse = TRUE
    ),
    error = function(e) {
      stop("Could not write ", shQuote(target), " [", e$message, "]")
    },
    finally = {
      setwd(curr_wd)
    }
  )

  target
}

#' @export
#' @importFrom zip unzip
#' @title Extract files from a zip file
#' @description Extract files from a zip file to a folder. The
#' function returns the complete path to destination folder.
#' @param file path of the archive to unzip
#' @param folder folder to create
#' @keywords internal
unpack_folder <- function(file, folder) {
  stopifnot(file.exists(file))

  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", file)

  # force deletion if already existing
  unlink(folder, recursive = TRUE, force = TRUE)

  if (l10n_info()$`UTF-8`) {
    zip::unzip(zipfile = file, exdir = folder)
  } else {
    officer_wd_folder <- getOption("officer_wd_folder")
    if (is.null(officer_wd_folder) || !dir.exists(officer_wd_folder)) {
      officer_wd_folder <- tempdir()
    }
    # unable to unzip a file with accent when on windows
    newfile <- tempfile(tmpdir = officer_wd_folder, fileext = file_type)
    file.copy(from = file, to = newfile)
    zip::unzip(zipfile = newfile, exdir = folder)
    unlink(newfile, force = TRUE)
  }

  absolute_path(folder)
}

absolute_path <- function(x) {
  if (length(x) != 1L) {
    stop("'x' must be a single character string")
  }
  epath <- path.expand(x)

  if (file.exists(epath)) {
    epath <- normalizePath(epath, "/", mustWork = TRUE)
  } else {
    if (!dir.exists(dirname(epath))) {
      stop("directory of ", x, " does not exist.", call. = FALSE)
    }
    cat("", file = epath)
    epath <- normalizePath(epath, "/", mustWork = TRUE)
    unlink(epath)
  }
  epath
}

test_that("pack_folder does not use full path", {
  dir_ <- tempfile()
  dir.create(dir_)
  file <- tempfile(tmpdir = dir_)
  cat("test", file = file)
  pack_folder(dir_, target = "test.zip")

  dir_ <- tempfile()
  unpack_folder(file = "test.zip", dir_)
  expect_equal(list.files(dir_), basename(file))
  unlink("test.zip", force = TRUE)
})

test_that("pack_folder behavior", {
  dir_ <- tempfile()
  dir.create(dir_)
  file <- tempfile(tmpdir = dir_)
  cat("test", file = file)

  expect_error(pack_folder(dir_, target = "dummy_dir/test.zip"))
})

test_that("pack_folder does not use full path", {
  dir_ <- tempfile()
  dir.create(dir_)
  file <- tempfile(tmpdir = dir_)
  cat("test", file = file)
  pack_folder(dir_, target = "test.zip")

  dir_ <- tempfile()
  unpack_folder(file = "test.zip", dir_)
  expect_equal(list.files(dir_), basename(file))
  unlink("test.zip", force = TRUE)
})

test_that("pack_folder behavior", {
  dir_ <- tempfile()
  dir.create(dir_)
  file <- tempfile(tmpdir = dir_)
  cat("test", file = file)

  expect_error(pack_folder(dir_, target = "dummy_dir/test.zip"))
})

# presentation ------------------------------------------------------------

presentation <- R6Class(
  "presentation",
  inherit = openxml_document,

  public = list(

    initialize = function( package_dir ) {
      super$initialize(character(0))
      presentation_filename <- file.path(package_dir, "ppt", "presentation.xml")
      self$feed(presentation_filename)

      slide_df <- private$get_slide_df()
      private$slide_id <- slide_df$id
      private$slide_rid <- slide_df$rid

    },

    add_slide = function(target){


      private$rels_doc$add(id = paste0("rId", private$rels_doc$get_next_id() ),
                           type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide",
                           target = target )
      rels <- private$rels_doc$get_data()
      rid <- rels[rels$target %in% target,"id"]


      ids <- private$slide_id
      if( length( ids ) < 1 )
        new_id <- 256
      else new_id <- max(ids) + 1

      private$slide_id <- c( private$slide_id, new_id)
      private$slide_rid <- c( private$slide_rid, rid)

      private$update_xml()

      self
    },
    slide_data = function(){
      rel_df <- self$rel_df()
      rel_df <- rel_df[, c("id", "target")]
      names(rel_df) <- c("slide_rid", "target")
      ref <- data.frame(slide_id = private$slide_id,
                 slide_rid = private$slide_rid,
                 stringsAsFactors = FALSE)
      base::merge(x = ref, y = rel_df, sort = FALSE,
                  by = "slide_rid", all.x = TRUE, all.y = FALSE)
    },

    move_slide = function(from, to){
      slide_list <- self$slide_data()

      ord <- seq_len(nrow(slide_list))
      if( from < to ) to <- to+1
      ord[ord >= to] <- ord[ord >= to] + 1L
      ord[from] <- to
      slide_list <- slide_list[order(ord),,drop = FALSE]
      private$slide_id <- slide_list$slide_id
      private$slide_rid <- slide_list$slide_rid

      private$update_xml()

      self
    },

    remove_slide = function(target){

      reldf <- self$rel_df()
      id <- which( basename(reldf$target) %in% basename(target)  )
      rid <- reldf$id[id]
      private$rels_doc$remove(target = target )

      dropid <- which( private$slide_rid %in% rid )

      private$slide_id <- private$slide_id[-dropid]
      private$slide_rid <- private$slide_rid[-dropid]

      private$update_xml()

      self
    }

  ),
  private = list(

    slide_id = NULL,
    slide_rid = NULL,

    get_slide_df = function() {
      nodes <- xml_find_all(private$doc, "//p:sldIdLst/p:sldId")
      id <- as.integer( xml_attr(nodes, "id", ns = xml_ns(private$doc)) )
      rid <- xml_attr(nodes, "r:id", ns = xml_ns(private$doc))
      data.frame(id = id, rid = rid, stringsAsFactors = FALSE)
    },
    update_xml = function(){
      xml_list <- xml_find_first(private$doc, "//p:sldIdLst")
      xml_elt <- paste(
        sprintf("<p:sldId id=\"%.0f\" r:id=\"%s\"/>", private$slide_id, private$slide_rid),
        collapse = "" )

      xml_elt <- paste0("<p:sldIdLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
                        xml_elt,
                        "</p:sldIdLst>")
      xml_elt <- as_xml_document(xml_elt)

      if( !inherits(xml_list, "xml_missing")){
        xml_replace(xml_list, xml_elt)
      } else{ ## needs to be after all MasterIdLst elements. placing it before sldSz seems to be the safest option.
        xml_add_sibling(xml_find_first(private$doc, "//p:sldSz"), xml_elt, .where = "before")
      }

      self
    }
  )
)




# slide master ------------------------------------------------------------
slide_master <- R6Class(
  "slide_master",
  inherit = openxml_document,
  public = list(

    name = function(){
      theme_ <- private$theme_file()
      root <- gsub( paste0(self$dir_name(), "$"), "", dirname( private$filename ) )
      xml_attr(read_xml(file.path( root,theme_)), "name")
    },

    colors = function(){
      theme_ <- private$theme_file()
      root <- gsub( paste0(self$dir_name(), "$"), "", dirname( private$filename ) )

      doc <- read_xml(file.path( root,theme_))
      read_theme_colors( doc, self$name() )
    },

    xfrm = function(){
      nodeset <- xml_find_all( self$get(), as_xpath_content_sel("p:cSld/p:spTree/") )
      read_xfrm(nodeset, self$file_name(), self$name())
    }


  ),
  private = list(

    theme_file = function(){
      data <- self$rel_df()
      theme_file <- data$target[basename(data$type) == "theme"]
      file.path( "ppt/theme", basename(theme_file) )
    }

  )

)

# slide_layout ------------------------------------------------------------
slide_layout <- R6Class(
  "slide_layout",
  inherit = openxml_document,
  public = list(

    get_metadata = function( ){
      rels <- self$rel_df()
      rels <- rels[basename( rels$type ) == "slideMaster", ]
      data.frame(stringsAsFactors = FALSE, name = self$name(), filename = self$file_name(), master_file = rels$target)
    },
    xfrm = function(){
      rels <- self$rel_df()
      rels <- rels[basename( rels$type ) == "slideMaster", ]

      nodeset <- xml_find_all( self$get(), as_xpath_content_sel("p:cSld/p:spTree/"))
      data <- read_xfrm(nodeset, self$file_name(), self$name())
      if( nrow(data))
        data$master_file <- basename(rels$target)
      else data$master_file <- character(0)
      data
    },
    write_template = function(new_file){

      path <- system.file(package = "officer", "template/slide.xml")

      rel_filename <- file.path(
        dirname(new_file), "_rels",
        paste0(basename(new_file), ".rels") )

      newrel <- relationship$new()$add(
        id = "rId1", type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slideLayout",
        target = file.path("../slideLayouts", basename(self$file_name())) )
      newrel$write(path = rel_filename)
      file.copy(path, to = new_file)
      self
    },

    name = function(){
      csld <- xml_find_first(self$get(), "//p:cSld")
      xml_attr(csld, "name")
    }


  )

)

# slide ------------------------------------------------------------
slide <- R6Class(
  "slide",
  inherit = openxml_document,
  public = list(

    feed = function( file ) {
      super$feed(file)
      slide_info <- private$rels_doc$get_data()
      slide_info <- slide_info[basename(slide_info$type) == "slideLayout",]
      private$layout_file <- basename( slide_info$target )
      self
    },

    set_layout_xfrm = function(xfrm_ref){
      private$element_data <- xfrm_ref[xfrm_ref$file == private$layout_file,]
      self
    },
    reference_img = function(src, dir_name){
      src <- unique( src )
      private$rels_doc$add_img(src, root_target = "../media")
      dir.create(dir_name, recursive = TRUE, showWarnings = FALSE)
      file.copy(from = src, to = file.path(dir_name, basename(src)))
      self
    },

    reference_slide = function(slide_file){

      rel_dat <- private$rels_doc$get_data()

      if( !slide_file %in% rel_dat$target ){
        next_id <- private$rels_doc$get_next_id()
        private$rels_doc$add(id = sprintf("rId%.0f", next_id),
                             type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide",
                             target = slide_file )
      }
      self
    },

    reference_hyperlink = function(href){

      rel_dat <- private$rels_doc$get_data()

      if( !href %in% rel_dat$target ){
        next_id <- private$rels_doc$get_next_id()
        private$rels_doc$add(id = sprintf("rId%.0f", next_id),
                             type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink",
                             target = href, target_mode = "External" )
      }
      self
    },

    get_xfrm = function(type = NULL, index = 1){
      out <- private$element_data
      if( !is.null(type) ){
        if( type %in% out$type ){
          type_matches <- which( out$type == type )
          if( index <= length(type_matches) )
            id <- type_matches[index]
          else stop(type, " can only have ", length(type_matches), " element(s) but index is set to ", index)

        } else stop("type ", type, " is not available in the slide layout")
        out <- out[id, ]
      }
      out
    },

    layout_name = function(){
      private$layout_file
    },

    get_metadata = function( ){
      rels <- self$rel_df()
      rels <- rels[basename( rels$type ) == "slideLayout", ]
      data.frame(stringsAsFactors = FALSE, name = self$name(), filename = self$file_name(), layout_file = rels$target)
    }

  ),
  private = list(
    layout_file = NULL,
    element_data = NULL
  )

)


# notesSlide ------------------------------------------------------------
notesSlide <- R6Class(
  "notesSlide",
  inherit = openxml_document,
  public = list(

    feed = function( file ) {
      super$feed(file)
      self
    },

    fortify_id = function(){
      cnvpr <- xml_find_all(private$doc, "//p:cNvPr")
      for(i in seq_along(cnvpr))
        xml_attr( cnvpr[[i]], "id") <- i
      self
    },

    get_metadata = function( ){
      rels <- self$rel_df()
    }

  )

)

#' Default layout for new slides
#'
#' Set or remove the default layout used when calling `add_slide()`.
#'
#' @param x An `rpptx` object.
#' @param layout Layout name. If `NULL` (default), removes the default layout.
#' @param master Name of master. Only required if layout name is not unique across masters.
#' @param as_list If `TRUE`, return a list with layout and master instead of the `rpptx` object.
#' @export
#' @example inst/examples/example_layout_default.R
#' @seealso [add_slide()]
#' @return The `rpptx` object.
layout_default <- function(x, layout = NULL, master = NULL, as_list = FALSE) {
  stop_if_not_rpptx(x)

  if (as_list) {
    return(x$layout_default)
  }

  if (is.null(layout) || is.na(layout)) {
    return(init_layout_default(x))
  }

  la <- get_layout(x, layout, master, layout_by_id = FALSE)
  x$layout_default <- list(layout = la$layout_name, master = la$master_name)
  x
}


init_layout_default <- function(x) {
  stop_if_not_rpptx(x)
  x$layout_default <- list(layout = NA, master = NA)
  x
}


get_layout_default <- function(x) {
  stop_if_not_rpptx(x)
  x$layout_default
}


has_layout_default <- function(x) {
  stop_if_not_rpptx(x)
  la <- get_layout_default(x)
  !is.null(la) && !identical(la, list(layout = NA, master = NA))
}


# check if layout exists
layout_exists <- function(x, layout) {
  stop_if_not_rpptx(x)
  layout %in% x$slideLayouts$names()
}

#' Layout selection helper
#'
#' Select a layout by name or index. The master name is inferred and only required
#' for disambiguation in case the layout name is not unique across masters.
#'
#' @param x An `rpptx` object.
#' @param layout Layout name or index. Index refers to the row index of the [layout_summary()]
#' output.
#' @param master Name of master. Only required if layout name is not unique across masters.
#' @return A `<layout_info>` object, i.e. a list with the entries `index`, `layout_name`,
#' `layout_file`, `master_name`, `master_file`, and `slide_layout`.
#' @param layout_by_id Allow layout index instead of name? (default is `TRUE`)
#' @param get_first If layout exists in multiple master, return first occurence (default `FALSE`).
#' @keywords internal
get_layout <- function(
  x,
  layout = NULL,
  master = NULL,
  layout_by_id = TRUE,
  get_first = FALSE
) {
  stop_if_not_rpptx(x, "x")

  if (!layout_by_id && is.numeric(layout)) {
    cli::cli_abort(
      c(
        "{.arg layout} must be {.cls character}",
        "x" = "Got class {.cls {class(layout)[1]}} instead"
      )
    )
  }
  if (!(is.numeric(layout) || is.character(layout))) {
    cli::cli_abort(
      c(
        "{.arg layout} must be {.cls numeric} or {.cls character}",
        "x" = "Got class {.cls {class(layout)[1]}} instead"
      )
    )
  }
  if (length(layout) != 1) {
    cli::cli_abort(
      c(
        "{.arg layout} is not length 1",
        "x" = "{.arg layout} must be {.emph one} layout name or index."
      )
    )
  }
  df <- x$slideLayouts$get_metadata()
  names(df)[2:3] <- c("layout_name", "layout_file") # consistent naming
  n_layouts <- nrow(df)
  df$index <- seq_len(n_layouts)

  if (n_layouts == 0) {
    cli::cli_alert_danger("No layouts available.")
    return(NULL)
  }

  if (is.numeric(layout)) {
    res <- get_row_by_index(df, layout)
  } else {
    res <- get_row_by_name(df, layout, master, get_first = get_first)
  }
  l <- as.list(res)
  slide_layout <- x$slideLayouts$collection_get(l$layout_file)
  l <- c(l, slide_layout = slide_layout)
  l <- l[c(
    "index",
    "layout_name",
    "layout_file",
    "master_name",
    "master_file",
    "slide_layout"
  )] # nice order
  class(l) <- c("layout_info", "list")
  l
}


is_layout_info <- function(x) {
  inherits(x, "layout_info")
}


#' @export
print.layout_info <- function(x, ...) {
  cli::cli_h3("{.cls layout_info} object")
  str(utils::head(x, -1), give.attr = FALSE, no.list = TRUE)
  cat(" $ slide_layout: 'R6' <slide_layout>")
}


get_row_by_index <- function(df, layout) {
  index <- layout
  if (!index %in% df$index) {
    cli::cli_abort(
      c(
        "Layout index out of bounds.",
        "x" = "Index must be between {.val {1}} and {.val {nrow(df)}}.",
        "i" = cli::col_grey("See row indexes in {.fn layout_summary}")
      ),
      call = NULL
    )
  }
  df[index, ]
}


# select layout by name
# get_first: get first occurence if layout exists in mutiple masters
get_row_by_name <- function(df, layout, master, get_first = FALSE) {
  if (!is.null(master)) {
    masters <- unique(df$master_name)
    if (!master %in% masters) {
      cli::cli_abort(
        c(
          "master {.val {master}} does not exist.",
          "i" = "See {.fn layout_summary} for available masters."
        ),
        call = NULL
      )
    }
    df <- df[df$master_name == master, ]
  }

  df <- df[df$layout_name == layout, ]
  if (nrow(df) == 0) {
    msg <- ifelse(
      is.null(master),
      "Layout {.val {layout}} does not exist",
      "Layout {.val {layout}} does not exist in master {.val {master}}"
    )
    cli::cli_abort(
      c(msg, "i" = "See {.fn layout_summary} for available layouts."),
      call = NULL
    )
  }

  if (get_first) {
    df <- df[1L, , drop = FALSE]
  }

  if (nrow(df) > 1) {
    cli::cli_abort(
      c(
        "Layout {.val {layout}} exists in more than one master",
        "x" = "Please specify the master name in arg {.arg master} for disambiguation"
      ),
      call = NULL
    )
  }
  df
}


# get <layout_info> object for slide layout
get_slide_layout <- function(x, slide_idx) {
  stop_if_not_rpptx(x)
  if (length(x) == 0) {
    cli::cli_abort(
      c(
        "Presentation does not have any slides yet",
        "x" = "Can only get the layout for an existing slides",
        "i" = "You can add a slide using {.fn add_slide}"
      ),
      call = NULL
    )
  }
  ensure_slide_index_exists(x, slide_idx)
  df <- x$slide$get_xfrm()[[slide_idx]]
  layout <- unique(df$name)
  master <- unique(df$master_name)
  get_layout(x, layout, master)
}


# get <layout_info> object for layout of current slide
get_layout_for_current_slide <- function(x) {
  get_slide_layout(x, x$cursor)
}

test_that("get_layout works as expected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  # all layouts unique
  x <- read_pptx()

  expect_error(get_layout(x, "Title Slide", "xxx"), 'master "xxx" does not exist')
  expect_error(get_layout(x, "xxx", "Office Theme"), 'Layout "xxx" does not exist in master "Office Theme"')

  expect_error(get_layout(x, 0), "Layout index out of bounds.")
  expect_error(get_layout(x, 10), "Layout index out of bounds.")
  expect_error(get_layout(x, Inf), "Layout index out of bounds.")

  expect_no_error(l1 <- get_layout(x, "Title Slide"))
  expect_s3_class(l1, "layout_info")
  expect_no_error(l2 <- get_layout(x, "Title Slide", "Office Theme"))
  expect_s3_class(l2, "layout_info")

  expect_no_error(la <- get_layout(x, 1))
  expect_equal(la$index, 1)

  # same layout in several masters
  file <- test_path("docs_dir", "test-three-identical-masters.pptx")
  x <- read_pptx(file)

  expect_error(get_layout(x, "xxx", NULL), 'Layout "xxx" does not exist')
  expect_error(get_layout(x, "xxx", "Master_1"), 'Layout "xxx" does not exist in master "Master_1"')
  expect_error(get_layout(x, "Title Slide"), "Layout exists in more than one master")
  expect_error(get_layout(x, "Title Slide", "xxx"), 'master "xxx" does not exist')

  expect_no_error(la <- get_layout(x, "Title Slide", "Master_1"))
  expect_s3_class(la, "layout_info")

  expect_no_error(la <- get_layout(x, 1))
  expect_equal(la$index, 1)
})


test_that("incorrect inputs are detected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layout <- "Comparison"

  expect_error(get_layout("no rpptx object", layout), "Incorrect input for `x`")

  # incorrect layout arg input
  error_msg <- "`layout` must be <numeric> or <character>"
  expect_error(get_layout(x, NA), error_msg)
  expect_error(get_layout(x, NULL), error_msg)
  expect_error(get_layout(x, mtcars), error_msg)

  # out of bound index
  error_msg <- "Layout index out of bounds"
  expect_error(get_layout(x, 0), error_msg)
  expect_error(get_layout(x, -1), error_msg)
  expect_error(get_layout(x, 8), error_msg)
  expect_error(get_layout(x, Inf), error_msg)

  # non existing layout
  expect_error(get_layout(x, "xxx"), 'Layout "xxx" does not exist.')
  expect_error(get_layout(x, ""), 'Layout "" does not exist.')

  # not exactly one layout
  error_msg <- "`layout` is not length 1"
  expect_error(get_layout(x, c("a", "b")), error_msg)
  expect_error(get_layout(x, 1:2), error_msg)
  expect_error(get_layout(x, integer()), error_msg)
  expect_error(get_layout(x, character()), error_msg)

  # layout not unique
  file <- test_path("docs_dir", "test-three-identical-masters.pptx")
  x <- read_pptx(file)
  expect_error(get_layout(x, "Title Slide"), "Layout exists in more than one master")
})


test_that("<layout_info> prints correctly", {
  x <- read_pptx()
  layout <- "Comparison"
  l <- get_layout(x, layout)
  out <- capture.output(print(l))
  expect_equal(length(out), length(l))
})


test_that("get layout from slide", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()

  # fails if no slides exist
  expect_error(get_slide_layout(x, 0), "Presentation does not have any slides yet")
  expect_error(get_layout_for_current_slide(x), "Presentation does not have any slides yet")

  # detect correct slide layout
  layout <- "Comparison"
  x <- add_slide(x, layout)

  expect_error(get_slide_layout(x, 0), "Slide index 0 is out of range")
  expect_error(get_slide_layout(x, 2), "Slide index 2 is out of range")

  error_msg <- "`slide_idx` must be <numeric>"
  expect_error(get_slide_layout(x, "1"), error_msg)
  expect_error(get_slide_layout(x, NA), error_msg)
  expect_error(get_slide_layout(x, NULL), error_msg)

  expect_no_error(get_slide_layout(x, 1))
  expect_no_error(get_layout_for_current_slide(x))

  la_slide <- get_slide_layout(x, 1)
  la_current <- get_layout_for_current_slide(x)
  la_reference <- get_layout(x, layout)
  expect_identical(la_current, la_reference)
  expect_identical(la_slide, la_reference)

  layout <- "Title Slide"
  x <- add_slide(x, layout)
  la_slide <- get_slide_layout(x, 2)
  la_current <- get_layout_for_current_slide(x)
  la_reference <- get_layout(x, layout)
  expect_identical(la_current, la_reference)
  expect_identical(la_slide, la_reference)
})

# get_layout ----

test_that("get_layout works as expected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  # all layouts unique
  x <- read_pptx()

  expect_error(
    get_layout(x, "Title Slide", "xxx"),
    'master "xxx" does not exist'
  )
  expect_error(
    get_layout(x, "xxx", "Office Theme"),
    'Layout "xxx" does not exist in master "Office Theme"'
  )

  expect_error(get_layout(x, 0), "Layout index out of bounds.")
  expect_error(get_layout(x, 10), "Layout index out of bounds.")
  expect_error(get_layout(x, Inf), "Layout index out of bounds.")

  expect_no_error(l1 <- get_layout(x, "Title Slide"))
  expect_s3_class(l1, "layout_info")
  expect_no_error(l2 <- get_layout(x, "Title Slide", "Office Theme"))
  expect_s3_class(l2, "layout_info")

  expect_no_error(la <- get_layout(x, 1))
  expect_equal(la$index, 1)
  expect_error(
    get_layout(x, 1, layout_by_id = FALSE),
    regexp = "`layout` must be <character>"
  )

  # same layout in several masters
  file <- test_path("docs_dir", "test-three-identical-masters.pptx")
  x <- read_pptx(file)

  expect_error(get_layout(x, "xxx", NULL), 'Layout "xxx" does not exist')
  expect_error(
    get_layout(x, "xxx", "Master_1"),
    'Layout "xxx" does not exist in master "Master_1"'
  )
  expect_error(
    get_layout(x, "Title Slide"),
    "Layout \"Title Slide\" exists in more than one master"
  )
  expect_error(
    get_layout(x, "Title Slide", "xxx"),
    'master "xxx" does not exist'
  )

  expect_no_error(la <- get_layout(x, "Title Slide", "Master_1"))
  expect_s3_class(la, "layout_info")

  expect_no_error(la <- get_layout(x, 1))
  expect_equal(la$index, 1)
})


test_that("is_layout_info", {
  x <- read_pptx()
  la <- get_layout(x, "Title Slide")
  expect_true(is_layout_info(la))
  expect_false(is_layout_info(NULL))
  expect_false(is_layout_info(1))
  expect_false(is_layout_info(list()))
})


test_that("incorrect inputs are detected", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layout <- "Comparison"

  expect_error(get_layout("no rpptx object", layout), "Incorrect input for `x`")

  # incorrect layout arg input
  error_msg <- "`layout` must be <numeric> or <character>"
  expect_error(get_layout(x, NA), error_msg)
  expect_error(get_layout(x, NULL), error_msg)
  expect_error(get_layout(x, mtcars), error_msg)

  # out of bound index
  error_msg <- "Layout index out of bounds"
  expect_error(get_layout(x, 0), error_msg)
  expect_error(get_layout(x, -1), error_msg)
  expect_error(get_layout(x, 8), error_msg)
  expect_error(get_layout(x, Inf), error_msg)

  # non existing layout
  expect_error(get_layout(x, "xxx"), 'Layout "xxx" does not exist.')
  expect_error(get_layout(x, ""), 'Layout "" does not exist.')

  # not exactly one layout
  error_msg <- "`layout` is not length 1"
  expect_error(get_layout(x, c("a", "b")), error_msg)
  expect_error(get_layout(x, 1:2), error_msg)
  expect_error(get_layout(x, integer()), error_msg)
  expect_error(get_layout(x, character()), error_msg)

  # layout not unique
  file <- test_path("docs_dir", "test-three-identical-masters.pptx")
  x <- read_pptx(file)
  expect_error(
    get_layout(x, "Title Slide"),
    "Layout \"Title Slide\" exists in more than one master"
  )
})


test_that("<layout_info> prints correctly", {
  x <- read_pptx()
  layout <- "Comparison"
  l <- get_layout(x, layout)
  out <- capture.output(print(l))
  expect_equal(length(out), length(l))
})


test_that("get layout from slide", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()

  # fails if no slides exist
  expect_error(
    get_slide_layout(x, 0),
    "Presentation does not have any slides yet"
  )
  expect_error(
    get_layout_for_current_slide(x),
    "Presentation does not have any slides yet"
  )

  # detect correct slide layout
  layout <- "Comparison"
  x <- add_slide(x, layout)

  expect_error(get_slide_layout(x, 0), "Slide index 0 is out of range")
  expect_error(get_slide_layout(x, 2), "Slide index 2 is out of range")

  error_msg <- "`slide_idx` must be <numeric>"
  expect_error(get_slide_layout(x, "1"), error_msg)
  expect_error(get_slide_layout(x, NA), error_msg)
  expect_error(get_slide_layout(x, NULL), error_msg)

  expect_no_error(get_slide_layout(x, 1))
  expect_no_error(get_layout_for_current_slide(x))

  la_slide <- get_slide_layout(x, 1)
  la_current <- get_layout_for_current_slide(x)
  la_reference <- get_layout(x, layout)
  expect_identical(la_current, la_reference)
  expect_identical(la_slide, la_reference)

  layout <- "Title Slide"
  x <- add_slide(x, layout)
  la_slide <- get_slide_layout(x, 2)
  la_current <- get_layout_for_current_slide(x)
  la_reference <- get_layout(x, layout)
  expect_identical(la_current, la_reference)
  expect_identical(la_slide, la_reference)
})


# layout_default ----

test_that("layout_default", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  expect_false(has_layout_default(x))
  la_default <- list(layout = NA, master = NA)
  expect_identical(x$layout_default, la_default)
  expect_identical(get_layout_default(x), la_default)
  expect_identical(layout_default(x, as_list = TRUE), la_default)

  layout <- "Title Slide"
  x <- layout_default(x, layout)
  expect_true(has_layout_default(x))
  expect_identical(
    x$layout_default,
    list(layout = layout, master = "Office Theme")
  )

  x <- layout_default(x, NA)
  expect_false(has_layout_default(x))
  expect_identical(get_layout_default(x), list(layout = NA, master = NA))

  expect_error(layout_default(x, "xxx"), 'Layout "xxx" does not exist')
})

#' @title Get comments in a Word document as a data.frame
#' @description return a data.frame representing the comments in a Word document.
#' @param x an rdocx object
#' @details
#' Each row of the returned data frame contains data for one comment. The
#'    columns contain the following information:
#' * "comment_id" - unique comment id
#' * "author" - name of the comment author
#' * "initials" - initials of the comment author
#' * "date" - timestamp of the comment
#' * "text" - a list column of characters containing the comment text. Elements can
#'    be vectors of length > 1 if a comment contains multiple paragraphs,
#'    blocks or runs or of length 0 if the comment is empty.
#' * "para_id" - a list column of characters containing the parent paragraph IDs.
#'    Elememts can be vectors of length > 1 if a comment spans multiple paragraphs
#'    or of length 0 if the comment has no parent paragraph.
#' * "commented_text" - a list column of characters containing the
#'    commented text. Elements can be vectors of length > 1 if a comment
#'    spans multiple paragraphs or runs or of length 0 if the commented text is empty.
#' @example inst/examples/example_docx_comments.R
#' @export
docx_comments <- function(x) {
  stopifnot(inherits(x, "rdocx"))

  comment_ids <- xml_attr(
    xml_find_all(
      x$doc_obj$get(),
      "/w:document/w:body//*[self::w:commentRangeStart]"
    ),
    "id"
  )

  comment_text_runs <- lapply(comment_ids, function(id) {
    xml_find_all(
      x$doc_obj$get(),
      paste0(
        "/w:document/w:body//*[self::w:r[w:t and",
        "preceding::w:commentRangeStart[@w:id=\'",
        id,
        "\']",
        " and ",
        "following::w:commentRangeEnd[@w:id=\'",
        id,
        "\']]]"
      )
    )
  })

  data <- data.frame(
    comment_id = comment_ids
  )
  # Add parent paragraph id
  data$para_id <- lapply(
    comment_text_runs,
    function(x) xml_attr(xml_parent(x), "paraId")
  )
  data$commented_text <- lapply(comment_text_runs, xml_text)

  comments <- xml_find_all(x$comments$get(), "//w:comments/w:comment")

  out <- data.frame(
    stringsAsFactors = FALSE,
    comment_id = xml_attr(comments, "id"),
    author = xml_attr(comments, "author"),
    initials = xml_attr(comments, "initials"),
    date = xml_attr(comments, "date")
  )

  out$text <- lapply(
    comments,
    function(x) xml_text(xml_find_all(x, "w:p/w:r/w:t"))
  )

  data <- merge(out, data, by = "comment_id", all.x = TRUE)
  data[order(as.integer(data$comment_id)), ]
}

x <- read_pptx()

# select layout explicitly
plot_layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
plot_layout_properties(x = x, layout = "Title Slide") # no master needed if layout name unique
plot_layout_properties(x = x, layout = 1) # use layout index instead of name

# plot default layout if one is set
x <- layout_default(x, "Title and Content")
plot_layout_properties(x)

# plot current slide's layout (default if no layout is passed)
x <- add_slide(x, "Title Slide")
plot_layout_properties(x)

# specify which slide's layout to plot by index
plot_layout_properties(x, slide_idx = 1)

# change appearance: what to show, font size, legend etc.
plot_layout_properties(
  x,
  layout = "Two Content",
  title = FALSE,
  type = FALSE,
  id = FALSE
)
plot_layout_properties(x, layout = 4, cex = c(labels = .8, id = .7, type = .7))
plot_layout_properties(x, 1, legend = TRUE)

#' @export
#' @rdname cursor
#' @title Set cursor in a 'Word' document
#' @description A set of functions is available to manipulate
#' the position of a virtual cursor. This cursor will be used when
#' inserting, deleting or updating elements in the document.
#' @section cursor_begin:
#' Set the cursor at the beginning of the document, on the first element
#' of the document (usually a paragraph or a table).
#' @param x a docx device
#' @examples
#' library(officer)
#'
#' # create a template ----
#' doc <- read_docx()
#' doc <- body_add_par(doc, "blah blah blah")
#' doc <- body_add_par(doc, "blah blah blah")
#' doc <- body_add_par(doc, "blah blah blah")
#' doc <- body_add_par(doc, "Hello text to replace")
#' doc <- body_add_par(doc, "blah blah blah")
#' doc <- body_add_par(doc, "blah blah blah")
#' doc <- body_add_par(doc, "blah blah blah")
#' doc <- body_add_par(doc, "Hello text to replace")
#' doc <- body_add_par(doc, "blah blah blah")
#' template_file <- print(
#'   x = doc,
#'   target = tempfile(fileext = ".docx")
#' )
#'
#' # replace all pars containing "to replace" ----
#' doc <- read_docx(path = template_file)
#' while (cursor_reach_test(doc, "to replace")) {
#'   doc <- cursor_reach(doc, "to replace")
#'
#'   doc <- body_add_fpar(
#'     x = doc,
#'     pos = "on",
#'     value = fpar(
#'       "Here is a link: ",
#'       hyperlink_ftext(
#'         text = "yopyop",
#'         href = "https://cran.r-project.org/"
#'       )
#'     )
#'   )
#' }
#'
#' doc <- cursor_end(doc)
#' doc <- body_add_par(doc, "Yap yap yap yap...")
#'
#' result_file <- print(
#'   x = doc,
#'   target = tempfile(fileext = ".docx")
#' )
cursor_begin <- function(x) {
  if (length(x$officer_cursor$nodes_names) > 0L) {
    x$officer_cursor$which <- 1L
  } else {
    x$officer_cursor$which <- 0L
  }
  x
}

#' @rdname cursor
#' @param id bookmark id
#' @section cursor_bookmark:
#' Set the cursor at a bookmark that has previously been set.
#' @export
#' @examples
#'
#' # cursor_bookmark ----
#'
#' doc <- read_docx()
#' doc <- body_add_par(doc, "centered text", style = "centered")
#' doc <- body_bookmark(doc, "text_to_replace")
#' doc <- body_add_par(doc, "A title", style = "heading 1")
#' doc <- body_add_par(doc, "Hello world!", style = "Normal")
#' doc <- cursor_bookmark(doc, "text_to_replace")
#' doc <- body_add_table(doc, value = iris, style = "table_template")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
cursor_bookmark <- function(x, id) {
  xpath_ <- sprintf("//w:bookmarkStart[@w:name='%s']", id)
  bm_start <- xml_find_first(x$doc_obj$get(), xpath_)

  if (inherits(bm_start, "xml_missing")) {
    stop("cannot find bookmark ", shQuote(id), call. = FALSE)
  }

  bm_id <- xml_attr(bm_start, "id")

  nodes_with_text <- xml_find_all(
    x$doc_obj$get(),
    "/w:document/w:body/*|/w:ftr/*|/w:hdr/*"
  )
  test_start <- sapply(nodes_with_text, function(node) {
    expr <- sprintf("/descendant::w:bookmarkStart[@w:id='%s']", bm_id)
    match_node <- xml_child(node, expr)
    !inherits(match_node, "xml_missing")
  })
  if (!any(test_start)) {
    stop("bookmark ", shQuote(id), " has not been found in the document", call. = FALSE)
  }

  test_end <- sapply(nodes_with_text, function(node) {
    expr <- sprintf("/descendant::w:bookmarkEnd[@w:id='%s']", bm_id)
    match_node <- xml_child(node, expr)
    !inherits(match_node, "xml_missing")
  })

  on_same_par <- test_start == test_end
  if (!all(on_same_par)) {
    stop("bookmark ", shQuote(id), " does not end in the same paragraph (or is on the whole paragraph)", call. = FALSE)
  }

  x$officer_cursor$which <- which(test_start)[1]

  x
}

#' @export
#' @rdname cursor
#' @section cursor_end:
#' Set the cursor at the end of the document, on the last element
#' of the document.
cursor_end <- function(x) {
  x$officer_cursor$which <- length(x$officer_cursor$nodes_names)
  x
}

#' @export
#' @rdname cursor
#' @param keyword keyword to look for as a regular expression
#' @param fixed logical. If TRUE, pattern is a string to be matched as is.
#' @section cursor_reach:
#' Set the cursor on the first element of the document
#' that contains text specified in argument \code{keyword}.
#' The argument \code{keyword} is a regexpr pattern.
cursor_reach <- function(x, keyword, fixed = FALSE) {
  nodes_with_text <- xml_find_all(
    x$doc_obj$get(),
    "/w:document/w:body/*|/w:ftr/*|/w:hdr/*"
  )

  if (length(nodes_with_text) < 1) {
    stop("no text found in the document", call. = FALSE)
  }

  text_ <- xml_text(nodes_with_text)
  test_ <- grepl(pattern = keyword, x = text_, fixed = fixed)
  if (!any(test_)) {
    stop(keyword, " has not been found in the document", call. = FALSE)
  }
  x$officer_cursor$which <- which(test_)[1]
  x
}

#' @export
#' @rdname cursor
#' @param keyword keyword to look for as a regular expression
#' @section cursor_reach_test:
#' Test if an expression has a match in the document
#' that contains text specified in argument \code{keyword}.
#' The argument \code{keyword} is a regexpr pattern.
cursor_reach_test <- function(x, keyword) {
  nodes_with_text <- xml_find_all(
    x$doc_obj$get(),
    "/w:document/w:body/*|/w:ftr/*|/w:hdr/*"
  )

  if (length(nodes_with_text) < 1) {
    stop("no text found in the document", call. = FALSE)
  }

  text_ <- xml_text(nodes_with_text)
  test_ <- grepl(pattern = keyword, x = text_)
  any(test_)
}

#' @export
#' @rdname cursor
#' @section cursor_forward:
#' Move the cursor forward, it increments the cursor in the document.
cursor_forward <- function(x) {
  x$officer_cursor$which <- min(c(length(x$officer_cursor$nodes_names), x$officer_cursor$which + 1L))
  x
}

#' @export
#' @rdname cursor
#' @section cursor_backward:
#' Move the cursor backward, it decrements the cursor in the document.
cursor_backward <- function(x) {
  x$officer_cursor$which <- max(c(0L, x$officer_cursor$which - 1L))
  x
}

# officer docx cursor ----
officer_cursor <- function(node) {
  nodes <- xml_find_all(node, "/w:document/w:body/*")
  nodes_names <- xml_name(nodes)
  nodes_names <- nodes_names[!nodes_names %in% "sectPr"]
  x <- list(
    nodes_names = nodes_names,
    which = length(nodes_names)
  )
  class(x) <- "officer_cursor"
  x
}


#' @export
as.character.officer_cursor <- function(x, ...) {
  if (length(x$nodes_names) < 1) {
    return(NA_character_)
  }
  paste0("/w:document/w:body/*[", x$which, "]")
}

## xml on cursor ----
ooxml_on_cursor <- function(x, node) {
  if (length(x$nodes_names) < 1) {
    return(NULL)
  }
  node <- xml_find_first(node, as.character(x))
  if (inherits(node, "xml_missing")) {
    stop("cursor does not correspond to any node", call. = FALSE)
  }
  node
}


## cursor feed ----
cursor_append <- function(x, what) {
  x$nodes_names <- c(x$nodes_names, what)
  x$which <- x$which + 1L
  x
}
cursor_add_after <- function(x, what) {
  seq_left <- seq_along(x$which)
  set_left <- x$nodes_names[seq_left]
  set_right <- x$nodes_names[-seq_left]
  x$nodes_names <- c(set_left, what, set_right)
  x$which <- x$which + 1L
  x
}

cursor_add_before <- function(x, what) {
  seq_left <- seq_along(x$which - 1)
  set_left <- x$nodes_names[seq_left]
  set_right <- x$nodes_names[-seq_left]
  x$nodes_names <- c(set_left, what, set_right)
  x
}
cursor_replace_nodename <- function(x, what) {
  x$nodes_names[x$which] <- what
  x
}
cursor_delete <- function(x, what) {
  stopifnot(x$nodes_names[x$which] == what)
  x$nodes_names <- x$nodes_names[-x$which]
  if (x$which > length(x$nodes_names)) {
    x$which <- x$which - 1L
  }
  x
}

#' @export
#' @title Add continuous section
#' @description Section break starts the new section on the same page. This
#' type of section break is often used to change the number of columns
#' without starting a new page.
#' @param x an rdocx object
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#' str2 <- "Aenean venenatis varius elit et fermentum vivamus vehicula."
#' str2 <- rep(str2, 5)
#' str2 <- paste(str2, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = "Default section", style = "heading 1")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str2, style = "Normal")
#' doc_1 <- body_end_section_continuous(doc_1)
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_continuous <- function(x) {
  bs <- block_section(prop_section(type = "continuous"))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}

#' @export
#' @title Add landscape section
#' @description A section with landscape orientation is added to the document.
#' @param x an rdocx object
#' @param w,h page width, page height (in inches)
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_landscape(doc_1)
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_landscape <- function(x, w = 16838 / 1440, h = 11906 / 1440) {
  bs <- block_section(prop_section(
    page_size = page_size(width = w, height = h, orient = "landscape"),
    type = "oddPage"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}

#' @export
#' @title Add portrait section
#' @description A section with portrait orientation is added to the document.
#' @param x an rdocx object
#' @param w,h page width, page height (in inches)
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_portrait(doc_1)
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_portrait <- function(x, w = 16838 / 1440, h = 11906 / 1440) {
  bs <- block_section(prop_section(
    page_size = page_size(width = w, height = h, orient = "portrait"),
    type = "oddPage"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}

#' @export
#' @title Add multi columns section
#' @description A section with multiple columns is added to the
#' document.
#'
#' You may prefer to use [body_end_block_section()] that is
#' more flexible.
#' @param x an rdocx object
#' @inheritParams section_columns
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_columns(doc_1)
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_columns <- function(
  x,
  widths = c(2.5, 2.5),
  space = 0.25,
  sep = FALSE
) {
  bs <- block_section(prop_section(
    section_columns = section_columns(
      widths = widths,
      space = space,
      sep = sep
    ),
    type = "continuous"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}


#' @export
#' @title Add a landscape multi columns section
#' @description A landscape section with multiple columns is added to the document.
#' @param x an rdocx object
#' @inheritParams section_columns
#' @param w,h page width, page height (in inches)
#' @examples
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 5)
#' str1 <- paste(str1, collapse = " ")
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_end_section_columns_landscape(doc_1, widths = c(6, 2))
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
body_end_section_columns_landscape <- function(
  x,
  widths = c(2.5, 2.5),
  space = 0.25,
  sep = FALSE,
  w = 16838 / 1440,
  h = 11906 / 1440
) {
  bs <- block_section(prop_section(
    page_size = page_size(width = w, height = h, orient = "landscape"),
    section_columns = section_columns(
      widths = widths,
      space = space,
      sep = sep
    ),
    type = "oddPage"
  ))
  str <- to_wml(bs, add_ns = TRUE)
  body_add_xml(x, str = str, pos = "after")
}
#' @export
#' @title Add any section
#' @description Add a section to the document. You can
#' define any section with a [block_section] object. All other
#' `body_end_section_*` are specialized, this one is highly flexible
#' but it's up to the user to define the section properties.
#'
#' @section Section model in Word:
#' A `block_section` added with `body_end_block_section()` applies to the
#' content that **precedes** the call: it closes the section that holds the
#' previous paragraphs / tables and inherits any layout (orientation, columns,
#' margins, headers / footers) defined by the `block_section`. The function
#' name reflects this: it marks the *end* of a section.
#'
#' Typical pattern: add the content, then close it with the section that
#' should layout it.
#'
#' ```r
#' doc <- read_docx() |>
#'   body_add_par("This paragraph is in landscape orientation.") |>
#'   body_end_block_section(block_section(prop_section(
#'     page_size = page_size(orient = "landscape")
#'   )))
#' ```
#'
#' The default section of the document (defined by the template or by
#' [body_set_default_section()]) closes any content added after the last
#' `body_end_block_section()` call.
#'
#' The RTF output uses the opposite model: `rtf_add(block_section(...))`
#' applies to the content that *follows* the call. See [rtf_add()].
#'
#' @param x an rdocx object
#' @param value a [block_section] object
#' @examples
#' library(officer)
#' str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
#' str1 <- rep(str1, 20)
#' str1 <- paste(str1, collapse = " ")
#'
#' ps <- prop_section(
#'   page_size = page_size(orient = "landscape"),
#'   page_margins = page_mar(top = 2),
#'   type = "continuous"
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#' doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
#'
#' doc_1 <- body_end_block_section(doc_1, block_section(ps))
#'
#' doc_1 <- body_add_par(doc_1, value = str1, style = "centered")
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for Word sections
#' @section Illustrations:
#'
#' \if{html}{\figure{body_end_block_section_doc_1.png}{options: style="width:80\%;"}}
body_end_block_section <- function(x, value) {
  stopifnot(inherits(value, "block_section"))
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml(x = x, str = xml_elt, pos = "after")
}

#' @export
#' @title Define Default Section
#' @description Define default section of the document. You can
#' define section propeerties (page size, orientation, ...) with a [prop_section] object.
#' @param x an rdocx object
#' @param value a [prop_section] object
#' @family functions for Word sections
#' @example inst/examples/example_body_set_default_section.R
#' @section Illustrations:
#'
#' \if{html}{\figure{body_set_default_section_doc_1.png}{options: style="width:80\%;"}}
body_set_default_section <- function(x, value) {
  stopifnot(inherits(value, "prop_section"))
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  z <- xml_find_first(docx_body_xml(x), "w:body/w:sectPr")
  xml_replace(z, as_xml_document(xml_elt))

  x
}


# utils ----
process_sections_content <- function(x, xml_str) {
  hof_summary <- extract_hof(x, xml_str)

  relationships <- docx_body_relationship(x)

  for (i in seq_len(nrow(hof_summary))) {
    # write file
    xml_doc <- as_xml_document(hof_summary[[i, "content_section"]])
    write_xml(
      xml_doc,
      file = file.path(x$package_dir, "word", hof_summary[[i, "file_name"]])
    )

    relationships$add(
      id = hof_summary[[i, "rid"]],
      type = paste0(
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/",
        hof_summary[[i, "type"]]
      ),
      target = hof_summary[[i, "file_name"]]
    )
    x$content_type$add_override(
      setNames(
        paste0(
          "application/vnd.openxmlformats-officedocument.wordprocessingml.",
          hof_summary[[i, "type"]],
          "+xml"
        ),
        paste0("/word/", hof_summary[[i, "file_name"]])
      )
    )
  }

  xml_str[hof_summary$reference_position] <- sprintf(
    xml_str[hof_summary$reference_position],
    hof_summary[["rid"]]
  )
  # xml_str <- gsub(" w:officer=\"true\"",  "", xml_str)

  # drop added sections contents from document xml
  lines_index_to_drop <- mapply(
    function(start, stop, str) {
      start:stop
    },
    start = hof_summary$start,
    stop = hof_summary$stop,
    SIMPLIFY = FALSE
  )
  lines_index_to_drop <- unlist(lines_index_to_drop)
  if (length(lines_index_to_drop) > 0) {
    xml_str <- xml_str[-lines_index_to_drop]
  }

  xml_str
}

section_dimensions <- function(node) {
  section_obj <- as_list(node)

  landscape <- FALSE
  if (
    !is.null(attr(section_obj$pgSz, "orient")) &&
      attr(section_obj$pgSz, "orient") == "landscape"
  ) {
    landscape <- TRUE
  }

  h_ref <- as.integer(attr(section_obj$pgSz, "h"))
  w_ref <- as.integer(attr(section_obj$pgSz, "w"))

  mar_t <- as.integer(attr(section_obj$pgMar, "top"))
  mar_b <- as.integer(attr(section_obj$pgMar, "bottom"))
  mar_r <- as.integer(attr(section_obj$pgMar, "right"))
  mar_l <- as.integer(attr(section_obj$pgMar, "left"))
  mar_h <- as.integer(attr(section_obj$pgMar, "header"))
  mar_f <- as.integer(attr(section_obj$pgMar, "footer"))

  list(
    page = c("width" = w_ref, "height" = h_ref),
    landscape = landscape,
    margins = c(
      top = mar_t,
      bottom = mar_b,
      left = mar_l,
      right = mar_r,
      header = mar_h,
      footer = mar_f
    )
  )
}

hof_next_file_index <- function(x, type = "header") {
  pattern <- paste0("^(", type, ")([0-9]+)(\\.xml)$")
  files <- list.files(
    path = file.path(x$package_dir, "word"),
    pattern = pattern
  )
  empty_index <- 0L
  m <- regexpr("[0-9]+", files)
  index_chars <- regmatches(files, m)
  index_ints <- as.integer(index_chars)
  index <- max(c(empty_index, index_ints), na.rm = TRUE)
  index + 1L
}


extract_hof <- function(x, xml_str) {
  # init next_rel_id that will be the first rid
  relationships <- docx_body_relationship(x)
  next_rel_id <- relationships$get_next_id()

  next_footer_file_index <- hof_next_file_index(x, type = "footer")
  next_header_file_index <- hof_next_file_index(x, type = "header")

  # footers extraction
  starts <- grep("<w:ftr", xml_str, fixed = TRUE)
  stops <- grep("</w:ftr>", xml_str, fixed = TRUE)

  content_section_chrs <- mapply(
    function(start, stop, str) {
      z <- str[start:stop]
      paste(z, collapse = "")
    },
    start = starts,
    stop = stops,
    MoreArgs = list(str = xml_str),
    SIMPLIFY = FALSE
  )
  content_section_chrs <- unlist(content_section_chrs)
  file_indexes <- seq(
    from = next_footer_file_index,
    along.with = content_section_chrs
  )
  file_names <- sprintf("%s%.0f.xml", "footer", file_indexes)

  rids <- sprintf(
    "rId%.0f",
    seq(from = next_rel_id, along.with = content_section_chrs)
  )

  reference_positions <- grep(
    "<w:footerReference w:officer",
    xml_str,
    fixed = TRUE
  )

  footer_sections <- data.frame(
    content_section = content_section_chrs,
    file_name = file_names,
    rid = rids,
    type = rep("footer", length(rids)),
    start = starts,
    stop = stops,
    reference_position = reference_positions,
    stringsAsFactors = FALSE
  )

  # headers extraction
  starts <- grep("<w:hdr", xml_str, fixed = TRUE)
  stops <- grep("</w:hdr>", xml_str, fixed = TRUE)

  content_section_chrs <- mapply(
    function(start, stop, str) {
      z <- str[start:stop]
      paste(z, collapse = "")
    },
    start = starts,
    stop = stops,
    MoreArgs = list(str = xml_str),
    SIMPLIFY = FALSE
  )
  content_section_chrs <- unlist(content_section_chrs)
  file_indexes <- seq(
    from = next_header_file_index,
    along.with = content_section_chrs
  )
  file_names <- sprintf("%s%.0f.xml", "header", file_indexes)

  rids <- sprintf(
    "rId%.0f",
    seq(
      from = next_rel_id + nrow(footer_sections),
      along.with = content_section_chrs
    )
  )

  reference_positions <- grep(
    "<w:headerReference w:officer",
    xml_str,
    fixed = TRUE
  )

  header_sections <- data.frame(
    content_section = content_section_chrs,
    file_name = file_names,
    rid = rids,
    type = rep("header", length(rids)),
    start = starts,
    stop = stops,
    reference_position = reference_positions,
    stringsAsFactors = FALSE
  )

  rbind(header_sections, footer_sections)
}


copy_header_references_everywhere <- function(x, xml_str) {
  xml_body <- docx_body_xml(x)

  all_ref <- xml_find_all(
    xml_body,
    "/w:document/w:body/w:sectPr/w:headerReference"
  )
  all_ref <- as.character(all_ref)
  all_ref <- paste0(all_ref, collapse = "")
  m <- regexpr("(?=</w:sectPr)", xml_str, perl = TRUE)

  default_sect_pos <- tail(which(m > -1), n = 1)
  attr(m, "match.length")[default_sect_pos] <- -1
  m[default_sect_pos] <- -1

  regmatches(xml_str, m) <- all_ref

  xml_str
}

copy_footer_references_everywhere <- function(x, xml_str) {
  xml_body <- docx_body_xml(x)

  all_ref <- xml_find_all(
    xml_body,
    "/w:document/w:body/w:sectPr/w:footerReference"
  )
  all_ref <- as.character(all_ref)
  all_ref <- paste0(all_ref, collapse = "")
  m <- regexpr("(?=</w:sectPr)", xml_str, perl = TRUE)

  default_sect_pos <- tail(which(m > -1), n = 1)
  attr(m, "match.length")[default_sect_pos] <- -1
  m[default_sect_pos] <- -1

  regmatches(xml_str, m) <- all_ref

  xml_str
}

#' @title shortcuts for formatting properties
#'
#' @description
#' Shortcuts for [fp_text()], [fp_par()], [fp_cell()] and [fp_border()].
#' @name shortcuts
#' @examples
#' shortcuts$fp_bold()
#' shortcuts$fp_italic()
#' shortcuts$b_null()
NULL

#' @rdname shortcuts
#' @format NULL
#' @docType NULL
#' @keywords NULL
#' @export
shortcuts <- list(
  fp_bold = function(...) {
    fp_text(bold = TRUE, ...)
  },
  fp_italic = function(...) {
    fp_text(italic = TRUE, ...)
  },
  b_null = function(...) {
    fp_border(width = 0, ...)
  }
)

#' @export
#' @title append a footnote
#' @description this function is deprecated.
#' @param x an rdocx object
#' @param ... unused
#' @note This function is now deprecated because it is not
#' efficient and make users write complex code. Use instead [fpar()] to build
#' formatted paragraphs.
#' @keywords internal
slip_in_footnote <- function(x, ...){
  message("`slip_in_footnote()` is deprecated, it is replaced by `run_footnote()`. Pausing for 10 seconds.")
  Sys.sleep(10)
  x
}

test_that("Defunct functions gives the right messages", {
  expect_error(
    slip_in_seqfield(),
    "run_word_field"
  )
  expect_error(
    slip_in_column_break(),
    "run_columnbreak"
  )
  expect_error(
    slip_in_xml(),
    "fpar"
  )
  expect_error(
    slip_in_text(),
    "fpar"
  )
  expect_error(
    slip_in_footnote(),
    "run_footnote"
  )
})

#' @export
#' @title Read 'Word' styles
#' @description Read Word styles and get results in
#' a data.frame.
#' @param x an rdocx object
#' @param type,is_default subsets for types (i.e. paragraph) and
#' default style (when `is_default` is TRUE or FALSE)
#' @examples
#' x <- read_docx()
#' styles_info(x)
#' styles_info(x, type = "paragraph", is_default = TRUE)
#' @family functions for Word document informations
styles_info <- function(
  x,
  type = c("paragraph", "character", "table", "numbering"),
  is_default = c(TRUE, FALSE)
) {
  styles <- x$styles
  styles <- styles[
    styles$style_type %in% type & styles$is_default %in% is_default,
  ]
  styles
}


#' @export
#' @title Add or replace paragraph style in a Word document
#' @description The function lets you add or replace a Word paragraph style.
#' @param x an rdocx object
#' @param style_id a unique style identifier for Word.
#' @param style_name a unique label associated with the style identifier.
#' This label is the name of the style when Word edit the document.
#' @param base_on the style name used as base style
#' @param fp_p paragraph formatting properties, see [fp_par()].
#' @param fp_t default text formatting properties. This is used as
#' text formatting properties, see [fp_text()]. If NULL (default), the
#' paragraph will used the default text formatting properties (defined by
#' the `base_on` argument).
#' @examples
#' library(officer)
#'
#' doc <- read_docx()
#'
#' doc <- docx_set_paragraph_style(
#'   doc,
#'   style_id = "rightaligned",
#'   style_name = "Explicit label",
#'   fp_p = fp_par(text.align = "right", padding = 20),
#'   fp_t = fp_text_lite(
#'     bold = TRUE,
#'     shading.color = "#FD34F0",
#'     color = "white")
#' )
#'
#' doc <- body_add_par(doc,
#'   value = "This is a test",
#'   style = "Explicit label")
#'
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_file
docx_set_paragraph_style <- function(
  x,
  style_id,
  style_name,
  base_on = "Normal",
  fp_p = fp_par(),
  fp_t = NULL
) {
  styles_file <- file.path(x$package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)

  if (grepl("[^a-zA-Z0-9\\-]+", style_id)) {
    stop("`style_id` should only contain '-', numbers and ascii characters.")
  }

  node_styles <- xml_find_first(doc, "/w:styles")

  fp_p$word_style <- NULL

  if (!is.null(fp_t)) {
    fp_t_xml <- rpr_wml(fp_t)
  } else {
    fp_t_xml <- ""
  }
  base_on <- get_style_id(data = x$styles, style = base_on, type = "paragraph")

  xml_code <- paste0(
    sprintf(
      "<w:style xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:type=\"paragraph\" w:customStyle=\"1\" w:styleId=\"%s\">",
      style_id
    ),
    sprintf("<w:name w:val=\"%s\"/>", style_name),
    sprintf("<w:basedOn w:val=\"%s\"/>", base_on),
    ppr_wml(fp_p),
    fp_t_xml,
    "</w:style>"
  )

  node_style <- xml_child(
    node_styles,
    sprintf("w:style[@w:styleId='%s']", style_id)
  )
  if (inherits(node_style, "xml_missing")) {
    xml_add_child(node_styles, as_xml_document(xml_code))
  } else {
    xml_replace(node_style, as_xml_document(xml_code))
  }

  write_xml(doc, file = styles_file)
  styles <- read_docx_styles(x$package_dir)
  x$styles <- styles

  x
}

#' @export
#' @title Add character style in a Word document
#' @description The function lets you add or modify Word character styles.
#' @param x an rdocx object
#' @param style_id a unique style identifier for Word.
#' @param style_name a unique label associated with the style identifier.
#' This label is the name of the style when Word edit the document.
#' @param base_on the character style name used as base style
#' @param fp_t Text formatting properties, see [fp_text()].
#' @examples
#' library(officer)
#' doc <- read_docx()
#'
#' doc <- docx_set_character_style(
#'   doc,
#'   style_id = "newcharstyle",
#'   style_name = "label for char style",
#'   base_on = "Default Paragraph Font",
#'   fp_text_lite(
#'     shading.color = "red",
#'     color = "white")
#' )
#' paragraph <- fpar(
#'   run_wordtext("hello",
#'     style_id = "newcharstyle"))
#'
#' doc <- body_add_fpar(doc, value = paragraph)
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_file
docx_set_character_style <- function(
  x,
  style_id,
  style_name,
  base_on,
  fp_t = fp_text_lite()
) {
  styles_file <- file.path(x$package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)
  node_styles <- xml_find_first(doc, "/w:styles")

  if (grepl("[^a-zA-Z0-9\\-]+", style_id)) {
    stop("`style_id` should only contain '-', numbers and ascii characters.")
  }

  base_on <- get_style_id(data = x$styles, style = base_on, type = "character")

  xml_code <- paste0(
    sprintf(
      "<w:style xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:type=\"character\" w:customStyle=\"1\" w:styleId=\"%s\">",
      style_id
    ),
    sprintf("<w:name w:val=\"%s\"/>", style_name),
    sprintf("<w:basedOn w:val=\"%s\"/>", base_on),
    rpr_wml(fp_t),
    "</w:style>"
  )

  node_style <- xml_child(
    node_styles,
    sprintf("w:style[@w:styleId='%s']", style_id)
  )
  if (inherits(node_style, "xml_missing")) {
    xml_add_child(node_styles, as_xml_document(xml_code))
  } else {
    xml_replace(node_style, as_xml_document(xml_code))
  }

  write_xml(doc, file = styles_file)
  styles <- read_docx_styles(x$package_dir)
  x$styles <- styles

  x
}


#' @export
#' @title Replace styles in a 'Word' Document
#' @description Replace styles with others in a 'Word' document. This function
#' can be used for paragraph, run/character and table styles.
#' @param x an rdocx object
#' @param mapstyles a named list, names are the replacement style,
#' content (as a character vector) are the styles to be replaced.
#' Use [styles_info()] to display available styles.
#' @examples
#' # creating a sample docx so that we can illustrate how
#' # to change styles
#' doc_1 <- read_docx()
#'
#' doc_1 <- body_add_par(doc_1, "A title", style = "heading 1")
#' doc_1 <- body_add_par(doc_1, "Another title", style = "heading 2")
#' doc_1 <- body_add_par(doc_1, "Hello world!", style = "Normal")
#' file <- print(doc_1, target = tempfile(fileext = ".docx"))
#'
#' # now we can illustrate how
#' # to change styles with `change_styles`
#' doc_2 <- read_docx(path = file)
#' mapstyles <- list(
#'   "centered" = c("Normal", "heading 2"),
#'   "strong" = "Default Paragraph Font"
#' )
#' doc_2 <- change_styles(doc_2, mapstyles = mapstyles)
#' print(doc_2, target = tempfile(fileext = ".docx"))
change_styles <- function(x, mapstyles) {
  if (is.null(mapstyles) || length(mapstyles) < 1) {
    return(x)
  }

  table_styles <- styles_info(x, type = c("paragraph", "character", "table"))

  from_styles <- unique(as.character(unlist(mapstyles)))
  to_styles <- unique(names(mapstyles))

  mfrom <- match(from_styles, table_styles$style_name)
  if (anyNA(mfrom)) {
    stop(
      "could not find style ",
      paste0(shQuote(from_styles[is.na(mfrom)]), collapse = ", "),
      ".",
      call. = FALSE
    )
  }
  mto <- match(to_styles, table_styles$style_name)
  if (anyNA(mto)) {
    stop(
      "could not find style ",
      paste0(shQuote(to_styles[is.na(mto)]), collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  mapping <- mapply(
    function(from, to) {
      id_to <- which(table_styles$style_name %in% to)
      id_to <- table_styles$style_id[id_to]

      id_from <- which(table_styles$style_name %in% from)
      types <- substring(table_styles$style_type[id_from], first = 1, last = 1)
      types[types %in% "c"] <- "r"
      types[types %in% "t"] <- "tbl"
      id_from <- table_styles$style_id[id_from]

      data.frame(
        from = id_from,
        to = rep(id_to, length(from)),
        types = types,
        stringsAsFactors = FALSE
      )
    },
    mapstyles,
    names(mapstyles),
    SIMPLIFY = FALSE
  )

  mapping <- do.call(rbind, mapping)
  row.names(mapping) <- NULL

  for (i in seq_len(nrow(mapping))) {
    all_nodes <- xml_find_all(
      x$doc_obj$get(),
      sprintf("//w:%sStyle[@w:val='%s']", mapping$types[i], mapping$from[i])
    )
    xml_attr(all_nodes, "w:val") <- rep(mapping$to[i], length(all_nodes))
  }

  x
}

x <- read_pptx()

# INFO -------------

# Returns layout's ph_labels by default in same order as layout_properties()
layout_rename_ph_labels(x, "Comparison")
layout_properties(x, "Comparison")$ph_label


# BASICS -----------
#
# HINT: run `plot_layout_properties(x, "Comparison")` to see how labels change

# rename using key-value pairs: 'old label' = 'new label' or 'id' = 'new label'
layout_rename_ph_labels(x, "Comparison", "Title 1" = "LABEL MATCHED") # label matching
layout_rename_ph_labels(x, "Comparison", "3" = "ID MATCHED") # id matching
layout_rename_ph_labels(
  x,
  "Comparison",
  "Date Placeholder 6" = "DATE",
  "8" = "FOOTER"
) # label, id

# rename using a named list and the .dots arg
renames <- list("Content Placeholder 3" = "CONTENT_1", "6" = "CONTENT_2")
layout_rename_ph_labels(x, "Comparison", .dots = renames)

# rename via rhs assignment and optional index (not id!)
layout_rename_ph_labels(x, "Comparison") <- LETTERS[1:8]
layout_rename_ph_labels(x, "Comparison")[1:3] <- paste("CHANGED", 1:3)

# rename via rhs assignment and ph id (not index)
layout_rename_ph_labels(x, "Comparison", id = c(2, 4)) <- paste("ID =", c(2, 4))


# MORE ------------

# make all labels lower case
labels <- layout_rename_ph_labels(x, "Comparison")
layout_rename_ph_labels(x, "Comparison") <- tolower(labels)

# rename all labels to type [type_idx]
lp <- layout_properties(x, "Comparison")
layout_rename_ph_labels(x, "Comparison") <- paste0(
  lp$type,
  " [",
  lp$type_idx,
  "]"
)

# rename duplicated placeholders (see also `layout_dedupe_ph_labels()`)
file <- system.file("doc_examples", "ph_dupes.pptx", package = "officer")
x <- read_pptx(file)
lp <- layout_properties(x, "2-dupes")
idx <- which(lp$ph_label == "Content 7") # exists twice
layout_rename_ph_labels(x, "2-dupes")[idx] <- paste("DUPLICATE", seq_along(idx))

# warning: in case of duped labels only the first occurrence is renamed
x <- read_pptx(file)
layout_rename_ph_labels(x, "2-dupes", "Content 7" = "new label")

#' @export
#' @title Set 'Microsoft Word' Document Settings
#' @description
#' Set various settings of a 'Microsoft Word' document generated with 'officer'.
#' Options include:
#' - zoom factor (default view in Word),
#' - default tab stop,
#' - hyphenation zone,
#' - decimal symbol,
#' - list separator (see details below),
#' - compatibility mode,
#' - even and odd headers management (see details below),
#' - and auto hyphenation activation.
#'
#' @details
#' - `even_and_odd_headers`: If TRUE, 'Microsoft Word' will use different
#' headers for odd and even pages ("Different Odd & Even Pages" feature in Word).
#' This is useful for professional documents or reports that require alternating
#' page layouts.
#'
#' - `list_separator`: Sets the character used by 'Microsoft Word' to separate
#' items in lists (for example, when inserting tables or lists in Word). This
#' parameter affects how 'Microsoft Word' handles data import/export (CSV, etc.) and can be
#' adapted to language or local conventions (e.g., ";" for French, "," for English).
#'
#' @seealso [read_docx()]
#' @param x an rdocx object
#' @param zoom zoom factor, default is 1 (100%)
#' @param default_tab_stop default tab stop in inches, default is 0.5
#' @param hyphenation_zone hyphenation zone in inches, default is 0.25
#' @param decimal_symbol decimal symbol, default is "."
#' @param list_separator list separator, default is ";". Sets the separator
#' used by Word for lists (see details).
#' @param compatibility_mode compatibility mode, default is "15"
#' @param even_and_odd_headers whether to use different headers for even and
#' odd pages, default is FALSE. Enables the "Different Odd and Even Pages"
#' feature in 'Microsoft Word'.
#' @param auto_hyphenation whether to enable auto hyphenation, default is FALSE.
#' @param unit unit for `default_tab_stop` and `hyphenation_zone`, one of "in", "cm", "mm".
#' @example inst/examples/example_docx_set_settings.R
docx_set_settings <- function(
  x,
  zoom = 1,
  default_tab_stop = .5,
  hyphenation_zone = .25,
  decimal_symbol = ".",
  list_separator = ";",
  compatibility_mode = "15",
  even_and_odd_headers = FALSE,
  auto_hyphenation = FALSE,
  unit = "in"
) {
  settings <- docx_settings(
    zoom = zoom,
    default_tab_stop = convin(unit = unit, x = default_tab_stop),
    hyphenation_zone = convin(unit = unit, x = hyphenation_zone),
    decimal_symbol = decimal_symbol,
    list_separator = list_separator,
    even_and_odd_headers = even_and_odd_headers,
    auto_hyphenation = auto_hyphenation,
    compatibility_mode = compatibility_mode
  )
  x$settings <- settings
  x
}

docx_settings <- function(
  zoom = 1,
  default_tab_stop = .5,
  hyphenation_zone = .25,
  decimal_symbol = ".",
  list_separator = ";",
  compatibility_mode = "15",
  even_and_odd_headers = FALSE,
  auto_hyphenation = FALSE
) {
  x <- list(
    zoom = zoom,
    default_tab_stop = default_tab_stop,
    hyphenation_zone = hyphenation_zone,
    decimal_symbol = decimal_symbol,
    list_separator = list_separator,
    even_and_odd_headers = even_and_odd_headers,
    auto_hyphenation = auto_hyphenation,
    compatibility_mode = compatibility_mode
  )
  class(x) <- "docx_settings"
  x
}

update_docx_settings_from_file <- function(x, file) {
  if (is.null(file)) {
    cli::cli_abort(
      "File settings for Word document is null."
    )
  }
  if (!file.exists(file)) {
    cli::cli_abort(
      "File settings for Word document does not exists {.file {file}}."
    )
  }

  node_doc <- read_xml(file)

  node_zoom <- xml_child(node_doc, "w:zoom")
  if (!inherits(node_zoom, "xml_missing")) {
    x$zoom <- as.integer(xml_attr(node_zoom, "percent")) / 100
  }

  node_tab_stop <- xml_child(node_doc, "w:defaultTabStop")
  if (!inherits(node_tab_stop, "xml_missing")) {
    x$default_tab_stop <- as.integer(xml_attr(node_tab_stop, "val")) /
      1440
  }

  node_hyphenation_zone <- xml_child(node_doc, "w:hyphenationZone")
  if (!inherits(node_hyphenation_zone, "xml_missing")) {
    x$hyphenation_zone <- as.integer(xml_attr(
      node_hyphenation_zone,
      "val"
    )) /
      1440
  }

  node_auto_hyphenation <- xml_child(node_doc, "w:autoHyphenation")
  if (!inherits(node_auto_hyphenation, "xml_missing")) {
    x$auto_hyphenation <- TRUE
  }

  node_decimal_symbol <- xml_child(node_doc, "w:decimalSymbol")
  if (!inherits(node_decimal_symbol, "xml_missing")) {
    x$decimal_symbol <- xml_attr(node_decimal_symbol, "val")
  }

  node_list_separator <- xml_child(node_doc, "w:listSeparator")
  if (!inherits(node_list_separator, "xml_missing")) {
    x$list_separator <- xml_attr(node_list_separator, "val")
  }

  node_evenodd_headers <- xml_child(node_doc, "w:evenAndOddHeaders")
  if (!inherits(node_evenodd_headers, "xml_missing")) {
    x$even_and_odd_headers <- TRUE
  } else {
    x$even_and_odd_headers <- FALSE
  }

  x
}

write_docx_settings <- function(x) {
  settings <- x$settings
  str <- paste0(
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n",
    "<w:settings xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\">",
    sprintf("<w:zoom w:percent=\"%.0f\"/>", settings$zoom * 100),
    sprintf(
      "<w:defaultTabStop w:val=\"%.0f\"/>",
      settings$default_tab_stop * 1440
    ),
    if (settings$auto_hyphenation) sprintf("<w:autoHyphenation/>"),
    sprintf(
      "<w:hyphenationZone w:val=\"%.0f\"/>",
      settings$hyphenation_zone * 1440
    ),
    sprintf(
      "<w:compat><w:compatSetting w:name=\"compatibilityMode\" w:uri=\"http://schemas.microsoft.com/office/word\" w:val=\"%s\"/></w:compat>",
      settings$compatibility_mode
    ),
    sprintf("<w:decimalSymbol w:val=\"%s\"/>", settings$decimal_symbol),
    sprintf("<w:listSeparator w:val=\"%s\"/>", settings$list_separator),
    if (settings$even_and_odd_headers) {
      "<w:evenAndOddHeaders w:val=\"1\"/>"
    } else {
      "<w:evenAndOddHeaders w:val=\"0\"/>"
    },
    "</w:settings>"
  )
  file <- file.path(x$package_dir, "word", "settings.xml")
  writeLines(str, file, useBytes = TRUE)
  TRUE
}

test_that("settings works", {
  x <- read_docx()
  x <- docx_set_settings(
    x = x,
    zoom = 1,
    default_tab_stop = .5,
    hyphenation_zone = .25,
    decimal_symbol = ".",
    list_separator = ";",
    compatibility_mode = "15",
    even_and_odd_headers = TRUE,
    auto_hyphenation = FALSE
  )
  file <- print(x, target = tempfile(fileext = ".docx"))
  x <- read_docx(path = file)
  expect_equal(x$settings$zoom, 1)
  expect_equal(x$settings$list_separator, ";")
  expect_true(x$settings$even_and_odd_headers)
})

#' @export
#' @title  Write a 'Word' File
#' @description
#' `print.rdocx()` is the essential output function for creating Word files
#' with officer. It takes an `rdocx` object (created with [read_docx()] and
#' populated with content) and writes it to disk as a `.docx` file.
#'
#' This function performs all necessary post-processing operations before
#' writing the file.
#'
#' The function is typically called at the end of your document creation
#' workflow, after all content has been added with `body_add_*()` functions.
#'
#' @param x an `rdocx` object created with [read_docx()]
#' @param target path to the `.docx` file to write. The file will be created
#' or overwritten if it already exists. If `NULL` and `preview = FALSE`, the
#' function returns `NULL` without writing a file.
#' @param copy_header_refs,copy_footer_refs logical, default is FALSE.
#' If TRUE, copy the references to the header and footer in each section
#' of the body of the document. This parameter is experimental and may change
#' in a future version.
#' @param preview Save `x` to a temporary file and open it (default `FALSE`).
#' When `TRUE`, the document is saved to a temporary location and opened with
#' the system's default application for `.docx` files, useful for quick previewing
#' during development.
#' @param ... unused
#' @return The full path to the created `.docx` file (invisibly). This allows
#' chaining operations or capturing the output path for further use.
#' @example inst/examples/example_print_docx.R
#' @seealso Create a 'Word' document object with [read_docx()], add content with
#' functions [body_add_par()], [body_add_plot()],
#' [body_add_table()], change settings with [docx_set_settings()], set
#' properties with [set_doc_properties()],  read 'Word' styles with
#' [styles_info()].
print.rdocx <- function(
  x,
  target = NULL,
  copy_header_refs = FALSE,
  copy_footer_refs = FALSE,
  preview = FALSE,
  ...
) {
  if (preview) {
    file <- tempfile(fileext = ".docx")
    print.rdocx(
      x,
      target = file,
      copy_header_refs = copy_header_refs,
      copy_footer_refs = copy_footer_refs,
      preview = FALSE,
      ...
    )
    open_file(file)
    return(invisible(file))
  }
  if (is.null(target)) {
    cat("rdocx document with", length(x), "element(s)\n")
    cat("\n* styles:\n")

    style_names <- styles_info(x)
    style_sample <- style_names$style_type
    names(style_sample) <- style_names$style_name
    print(style_sample)
    return(invisible())
  }

  if (!grepl(x = target, pattern = "\\.(docx)$", ignore.case = TRUE)) {
    stop(target, " should have '.docx' extension.")
  }

  if (is_windows() && is_doc_open(target)) {
    stop(target, " is open. To write to this document, please, close it.")
  }

  # write the xml to a tempfile in a formatted way so that grepl is easy
  xml_str <- xml_document_to_chrs(x$doc_obj$get())
  xml_str <- process_sections_content(x, xml_str)
  xml_str <- process_footnotes_content(
    x$doc_obj,
    footnotes = x$footnotes,
    xml_str
  )
  xml_str <- process_comments_content(x$doc_obj, comments = x$comments, xml_str)
  xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
  xml_str <- fix_hyperlink_refs_in_wml(xml_str, x$doc_obj)
  xml_str <- fix_img_refs_in_wml(
    xml_str,
    x$doc_obj,
    x$doc_obj$relationship(),
    x$package_dir
  )
  xml_str <- fix_svg_refs_in_wml(
    xml_str,
    x$doc_obj,
    x$doc_obj$relationship(),
    x$package_dir
  )
  # make all id unique for document
  xml_str <- fix_empty_ids_in_wml(xml_str)
  # resolve list markers to real numId values
  xml_str <- process_list_markers(xml_str, x$package_dir)
  if (copy_header_refs) {
    xml_str <- copy_header_references_everywhere(x, xml_str = xml_str)
  }
  if (copy_footer_refs) {
    xml_str <- copy_footer_references_everywhere(x, xml_str = xml_str)
  }

  x <- replace_xml_body_from_chr(x = x, xml_str = xml_str)

  process_docx_poured(
    doc_obj = x$doc_obj,
    relationships = x$doc_obj$relationship(),
    content_type = x$content_type,
    package_dir = x$package_dir
  )

  x$headers <- update_hf_list(
    part_list = x$headers,
    type = "header",
    package_dir = x$package_dir
  )
  x$footers <- update_hf_list(
    part_list = x$footers,
    type = "footer",
    package_dir = x$package_dir
  )

  for (header in x$headers) {
    xml_str <- xml_document_to_chrs(header$get())
    xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
    xml_str <- fix_empty_ids_in_wml(xml_str)
    xml_str <- fix_hyperlink_refs_in_wml(xml_str, header)
    xml_str <- fix_img_refs_in_wml(
      xml_str,
      header,
      header$relationship(),
      x$package_dir
    )
    xml_str <- fix_svg_refs_in_wml(
      xml_str,
      header,
      header$relationship(),
      x$package_dir
    )
    tf_xml <- tempfile(fileext = ".txt")
    writeLines(xml_str, tf_xml, useBytes = TRUE)
    header$replace_xml(tf_xml)
  }
  for (footer in x$footers) {
    xml_str <- xml_document_to_chrs(footer$get())
    xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
    xml_str <- fix_empty_ids_in_wml(xml_str)
    xml_str <- fix_hyperlink_refs_in_wml(xml_str, footer)
    xml_str <- fix_img_refs_in_wml(
      xml_str,
      footer,
      footer$relationship(),
      x$package_dir
    )
    xml_str <- fix_svg_refs_in_wml(
      xml_str,
      footer,
      footer$relationship(),
      x$package_dir
    )

    tf_xml <- tempfile(fileext = ".txt")
    writeLines(xml_str, tf_xml, useBytes = TRUE)
    footer$replace_xml(tf_xml)
  }

  xml_str <- xml_document_to_chrs(x$footnotes$get())
  xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
  xml_str <- fix_empty_ids_in_wml(xml_str)
  xml_str <- fix_hyperlink_refs_in_wml(xml_str, x$footnotes)
  xml_str <- fix_img_refs_in_wml(
    xml_str,
    x$footnotes,
    x$footnotes$relationship(),
    x$package_dir
  )
  xml_str <- fix_svg_refs_in_wml(
    xml_str,
    x$footnotes,
    x$footnotes$relationship(),
    x$package_dir
  )

  tf_xml <- tempfile(fileext = ".txt")
  writeLines(xml_str, tf_xml, useBytes = TRUE)
  x$footnotes$replace_xml(tf_xml)

  xml_str <- xml_document_to_chrs(x$comments$get())
  xml_str <- convert_custom_styles_in_wml(xml_str, x$styles)
  xml_str <- fix_empty_ids_in_wml(xml_str)
  xml_str <- fix_hyperlink_refs_in_wml(xml_str, x$comments)
  xml_str <- fix_img_refs_in_wml(
    xml_str,
    x$comments,
    x$comments$relationship(),
    x$package_dir
  )
  xml_str <- fix_svg_refs_in_wml(
    xml_str,
    x$comments,
    x$comments$relationship(),
    x$package_dir
  )

  tf_xml <- tempfile(fileext = ".txt")
  writeLines(xml_str, tf_xml, useBytes = TRUE)
  x$comments$replace_xml(tf_xml)

  int_id <- 1 # unique id identifier
  # make all id unique for footnote
  int_id <- correct_id(x$comments$get(), int_id)

  body <- xml_find_first(x$doc_obj$get(), "w:body")

  # If body is not ending with an sectPr, create a continuous one append it
  if (!xml_name(xml_child(body, search = xml_length(body))) %in% "sectPr") {
    str <- paste0(
      "<w:sectPr xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">",
      "<w:type w:val=\"continuous\"/></w:sectPr>"
    )
    xml_add_child(body, as_xml_document(str))
  }

  for (header in x$headers) {
    header$save()
  }
  for (footer in x$footers) {
    footer$save()
  }

  x$doc_obj$save()
  x$content_type$save()
  x$footnotes$save()
  x$comments$save()

  x$rel$write(file.path(x$package_dir, "_rels", ".rels"))
  write_docx_settings(x)

  # save doc properties
  if (nrow(x$doc_properties$data) > 0) {
    x$doc_properties["modified", "value"] <- format(
      Sys.time(),
      "%Y-%m-%dT%H:%M:%SZ"
    )
    x$doc_properties["lastModifiedBy", "value"] <- Sys.getenv("USER")
    write_core_properties(x$doc_properties, x$package_dir)
  }
  if (nrow(x$doc_properties_custom$data) > 0) {
    write_custom_properties(x$doc_properties_custom, x$package_dir)
  }
  if (!is.null(x$app_properties)) {
    write_app_properties(x$app_properties, x$package_dir)
  }
  x <- sanitize_images(x, warn_user = FALSE)
  invisible(pack_folder(folder = x$package_dir, target = target))
}

# caption ----

#' @export
#' @title Caption block
#' @description Create a representation of a
#' caption that can be used for cross reference.
#' @param label a scalar character representing label to display
#' @param style paragraph style name
#' @param autonum an object generated with function [run_autonum]
#' @examples
#' library(officer)
#'
#' run_num <- run_autonum(seq_id = "tab", pre_label = "tab. ",
#'   bkm = "mtcars_table")
#' caption <- block_caption("mtcars table",
#'   style = "Normal",
#'   autonum = run_num
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add(doc_1, "A title", style = "heading 1")
#' doc_1 <- body_add(doc_1, "Hello world!", style = "Normal")
#' doc_1 <- body_add(doc_1, caption)
#' doc_1 <- body_add(doc_1, mtcars, style = "table_template")
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family block functions for reporting
block_caption <- function(label, style = NULL, autonum = NULL) {

  if (is.null(style)) {
    style <- "Normal"
  }

  z <- list(
    label = label,
    autonum = autonum,
    style = style
  )

  class(z) <- c("block_caption", "block")
  z
}

#' @export
print.block_caption <- function(x, ...) {
  if (is.null(x$autonum)) {
    auton <- "[autonum off]"
  } else {
    auton <- "[autonum on]"
  }
  cat("caption ", auton, ": ", x$label, "\n", sep = "")
}

to_wml_block_caption_officer <- function(x, add_ns = FALSE){

  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  autonum <- ""
  if (!is.null(x$autonum)) {
    autonum <- to_wml(x$autonum)
  }

  run_str <- sprintf("<w:r><w:t xml:space=\"preserve\">%s</w:t></w:r>", htmlEscapeCopy(x$label))
  run_str <- paste0(autonum, run_str)

  out <- sprintf(
    "%s<w:pPr><w:pStyle w:stlname=\"%s\"/></w:pPr>%s</w:p>",
    open_tag, x$style, run_str
  )

  out
}
to_wml_block_caption_pandoc <- function(x, bookdown_id = NULL){

  if(is.null(x$label)) return("")

  autonum <- ""
  if (!is.null(x$autonum)) {
    autonum <- paste("`", to_wml(x$autonum), "`{=openxml}", sep = "")
  }

  run_str <- paste0(autonum, htmlEscapeCopy(x$label))

  paste0(
    if (!is.null(x$style)) paste0("\n\n::: {custom-style=\"", x$style, "\"}"),
    "\n\n",
    # "<caption>\n\n",
    if (!is.null(bookdown_id)) bookdown_id,
    run_str,
    # "\n\n</caption>",
    if (!is.null(x$style)) paste0("\n:::\n"),
    "\n\n"
  )
}

#' @export
to_wml.block_caption <- function(x, add_ns = FALSE, knitting = FALSE, ...) {
  if(knitting)
    to_wml_block_caption_pandoc(x, bookdown_id = list(...)$bookdown_id)
  else
    to_wml_block_caption_officer(x, add_ns = add_ns)
}


# toc ----

#' @export
#' @title Table of content for 'Word'
#' @description Create a representation of a table
#' of content for Word documents.
#' @param level max title level of the table
#' @param style optional. If not NULL, its value is used as style in the
#' document that will be used to build entries of the TOC.
#' @param seq_id optional. If not NULL, its value is used as sequence
#' identifier in the document that will be used to build entries of the
#' TOC. See also [run_autonum()] to specify a sequence identifier.
#' @param separator optional. Some configurations need "," (i.e. from Canada) separator instead of ";"
#' @examples
#' block_toc(level = 2)
#' block_toc(style = "Table Caption")
#' @family block functions for reporting
block_toc <- function(level = 3, style = NULL, seq_id = NULL, separator = ";") {
  z <- list(
    level = level, style = style, seq_id = seq_id, separator = separator
  )
  class(z) <- c("block_toc", "block")

  z
}

#' @export
print.block_toc <- function(x, ...) {
  if (is.null(x$style) && is.null(x$seq_id)) {
    cat("TOC - max level: ", x$level, "\n", sep = "")
  } else if (!is.null(x$style)) {
    cat("TOC for style: ", x$style, "\n", sep = "")
  } else if (!is.null(x$seq_id)) {
    cat("TOC for seq identifier: ", x$seq_id, "\n", sep = "")
  }
}

#' @export
to_wml.block_toc <- function(x, add_ns = FALSE, ...) {

  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }


  if(is.null(x$style) && is.null(x$seq_id)) {
    out <- paste0(
      open_tag,
      "<w:pPr/>",
      to_wml(
        run_word_field(
          field = sprintf(
            "TOC \\o \"1-%.0f\" \\h \\z \\u",
            x$level
          )
        )
      ),
      "</w:p>"
    )
  } else if(!is.null(x$style)) {
    out <- paste0(
      open_tag,
      "<w:pPr/>",
      to_wml(
        run_word_field(
          field = sprintf(
            "TOC \\h \\z \\t \"%s%s1\"",
            x$style, x$separator
          )
        )
      ),
      "</w:p>"
    )
  } else if(!is.null(x$seq_id)) {
    out <- paste0(
      open_tag,
      "<w:pPr/>",
      to_wml(
        run_word_field(
          field = sprintf("TOC \\h \\z \\c \"%s\"", x$seq_id)
        )
      ),
      "</w:p>"
    )
  }



  out
}

# pour_docx ----

#' @export
#' @title External Word document placeholder
#' @description Pour the content of a docx file in the resulting docx
#' from an 'R Markdown' document.
#' @param file external docx file path
#' @examples
#' library(officer)
#' docx <- tempfile(fileext = ".docx")
#' doc <- read_docx()
#' doc <- body_add(doc, iris[1:20,], style = "table_template")
#' print(doc, target = docx)
#'
#' target <- tempfile(fileext = ".docx")
#' doc_1 <- read_docx()
#' doc_1 <- body_add(doc_1, block_pour_docx(docx))
#' print(doc_1, target = target)
#' @family block functions for reporting
block_pour_docx <- function(file){
  if(!file.exists(file)){
    stop("file {", file, "} does not exist.", call. = FALSE)
  }
  if(!grepl("\\.docx$", file, ignore.case = TRUE)){
    stop("file {", file, "} is not a docx file.", call. = FALSE)
  }
  if(grepl("&", file, ignore.case = TRUE)){
    stop("file path {", file, "} contains '&', please rename your file.", call. = FALSE)
  }
  if(grepl(" ", basename(file), ignore.case = TRUE)){
    stop("file path {", basename(file), "} contains ' ', please rename your file.", call. = FALSE)
  }

  z <- list(file = file)
  class(z) <- c("block_pour_docx", "block")
  z
}

#' @export
print.block_pour_docx <- function(x, ...) {
  cat("Pour docx: ", x$file, "\n", sep = "")
}

#' @export
to_wml.block_pour_docx <- function(x, add_ns = FALSE, ...) {
  ns_str <- ""
  if (add_ns) {
    ns_str <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" "
  }
  out <- paste0("<w:altChunk ", ns_str,
                    "r:id=\"", x$file, "\"/>")
  out
}



# section ----

#' @export
#' @title Section for 'Word'
#' @description Create a representation of a section.
#'
#' A section affects preceding paragraphs or tables; i.e.
#' a section starts at the end of the previous section (or the beginning of
#' the document if no preceding section exists), and stops where the
#' section is declared.
#'
#' When a new landscape section is needed, it is recommended to add a block_section
#' with `type = "continuous"`, to add the content to be appened in the new section
#' and finally to add a block_section with `page_size = page_size(orient = "landscape")`.
#' @param property section properties defined with function [prop_section]
#' @examples
#' ps <- prop_section(
#'   page_size = page_size(orient = "landscape"),
#'   page_margins = page_mar(top = 2),
#'   type = "continuous"
#' )
#' block_section(ps)
#' @family block functions for reporting
block_section <- function(property) {
  z <- list(
    property = property
  )
  class(z) <- c("block_section", "block")

  z
}

#' @export
print.block_section <- function(x, ...) {
  cat("----- end of section: ", "\n", sep = "")
}

#' @export
to_wml.block_section <- function(x, add_ns = FALSE, ...) {
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  out <- paste0(open_tag,
    "<w:pPr>",
    to_wml(x$property),
    "</w:pPr></w:p>"
  )

  out
}


# table properties ----


#' @export
#' @title Table conditional formatting
#' @description Tables can be conditionally formatted based on few properties as
#' whether the content is in the first row, last row, first column, or last
#' column, or whether the rows or columns are to be banded.
#' @param first_row,last_row apply or remove formatting from the first or last row in the table.
#' @param first_column,last_column apply or remove formatting from the first or last column in the table.
#' @param no_hband,no_vband don't display odd and even rows or columns with
#' alternating shading for ease of reading.
#' @note
#' You must define a format for first_row, first_column and other properties
#' if you need to use them. The format is defined in a docx template.
#' @examples
#' table_conditional_formatting(first_row = TRUE, first_column = TRUE)
#' @family functions for table definition
table_conditional_formatting <- function(
  first_row = TRUE, first_column = FALSE,
  last_row = FALSE, last_column = FALSE,
  no_hband = FALSE, no_vband = TRUE){

  z <- list(first_row = first_row, first_column = first_column,
            last_row = last_row, last_column = last_column,
            no_hband = no_hband, no_vband = no_vband)
  class(z) <- c("table_conditional_formatting")
  z
}

#' @export
to_wml.table_conditional_formatting <- function(x, add_ns = FALSE, ...) {
  paste0("<w:tblLook w:firstRow=\"", as.integer(x$first_row),
         "\" w:lastRow=\"", as.integer(x$last_row),
         "\" w:firstColumn=\"", as.integer(x$first_column),
         "\" w:lastColumn=\"", as.integer(x$last_column),
         "\" w:noHBand=\"", as.integer(x$no_hband),
         "\" w:noVBand=\"", as.integer(x$no_vband), "\"/>")
}

#' @export
to_pml.table_conditional_formatting <- function(x, add_ns = FALSE, ...){
  expr_ <- paste0(" firstRow=\"%.0f\" lastRow=\"%.0f\"",
         " firstColumn=\"%.0f\" lastColumn=\"%.0f\"",
         " bandRow=\"%.0f\" bandCol=\"%.0f\""
         )
  sprintf(expr_,
          x$first_row, x$last_row,
          x$first_column, x$last_column,
          !x$no_hband, !x$no_vband)
}

table_layout_types <- c("autofit", "fixed")


#' @export
#' @title Algorithm for table layout
#' @description When a table is displayed in a document, it can
#' either be displayed using a fixed width or autofit layout algorithm:
#'
#' * fixed: uses fixed widths for columns. The width of the table is not
#' changed regardless of the contents of the cells.
#' * autofit: uses the contents of each cell and the table width to
#' determine the final column widths.
#' @param type 'autofit' or 'fixed' algorithm. Default to 'autofit'.
#' @family functions for table definition
table_layout <- function(type = "autofit"){

  if(!type %in% table_layout_types){
    stop("type must be one of ", paste(table_layout_types, collapse = ", "), ".")
  }

  z <- list(type = type)
  class(z) <- "table_layout"
  z
}

#' @export
to_wml.table_layout <- function(x, add_ns = FALSE, ...) {
  sprintf("<w:tblLayout w:type=\"%s\"/>", x$type)
}

table_layout_width_units <- c("in", "pct")

#' @export
#' @title Preferred width for a table
#' @description Define the preferred width for a table.
#' @section Word:
#' All widths in a table are considered preferred because widths of
#' columns can conflict and the table layout rules can require a
#' preference to be overridden.
#' @param width value of the preferred width of the table.
#' @param unit unit of the width. Possible values are 'in' (inches) and 'pct' (percent)
#' @family functions for table definition
table_width <- function(width = 1, unit = "pct"){
  if(!unit %in% table_layout_width_units){
    stop("unit must be one of ", paste(table_layout_width_units, collapse = ", "), ".")
  }

  z <- list(width = width, unit = unit)
  class(z) <- "table_width"
  z

}
#' @export
to_wml.table_width <- function(x, add_ns = FALSE, ...) {
  if(x$unit %in% "pct"){
    tbl_width <- sprintf("<w:tblW w:type=\"pct\" w:w=\"%.0f\"/>",
                         x$width * 5000)
  } else {
    tbl_width <- sprintf("<w:tblW w:type=\"dxa\" w:w=\"%.0f\"/>",
                         x$width * 1440)
  }
  tbl_width
}

#' @export
#' @title Column widths of a table
#' @description The function defines the size of each column of a table.
#' @param widths Column widths expressed in inches.
#' @family functions for table definition
table_colwidths <- function(widths = NULL){
  z <- list(widths = widths)
  class(z) <- "table_colwidths"
  z
}

#' @export
#' @title Paragraph styles for columns
#' @description The function defines the paragraph styles for columns.
#' @param stylenames a named character vector, names are column names, values are
#' paragraph styles associated with each column. If a column is not
#' specified, default value 'Normal' is used.
#' Another form is as a named list, the list names are the styles
#' and the contents are column names to be formatted with the
#' corresponding style.
#' @family functions for table definition
#' @examples
#' library(officer)
#'
#' stylenames <- c(
#'   vs = "centered", am = "centered",
#'   gear = "centered", carb = "centered"
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_table(doc_1,
#'   value = mtcars, style = "table_template",
#'   stylenames = table_stylenames(stylenames = stylenames)
#' )
#'
#' print(doc_1, target = tempfile(fileext = ".docx"))
#'
#'
#' stylenames <- list(
#'   "centered" = c("vs", "am", "gear", "carb")
#' )
#'
#' doc_2 <- read_docx()
#' doc_2 <- body_add_table(doc_2,
#'   value = mtcars, style = "table_template",
#'   stylenames = table_stylenames(stylenames = stylenames)
#' )
#'
#' print(doc_2, target = tempfile(fileext = ".docx"))
table_stylenames <- function(stylenames = list()){

  if( length(stylenames) > 0 && is.null(attr(stylenames, "names")) ){
    stop("stylenames should have names")
  }

  if( length(stylenames) > 0 && is.list(stylenames) ){
    .l <- vapply(stylenames, length, FUN.VALUE = 0L)
    zz <- inverse.rle(
      structure(list(
        lengths = .l,
        values = names(stylenames)),
        class = "rle")
      )
    names(zz) <- as.character(unlist(stylenames))
    stylenames <- as.list(zz)
  } else if(is.character(stylenames)){
    stylenames <- as.list(stylenames)
  }

  z <- list(stylenames = stylenames)
  class(z) <- "table_stylenames"
  z
}

#' @export
#' @importFrom utils modifyList
to_wml.table_stylenames <- function(x, add_ns = FALSE, dat, ...) {

  stylenames <- rep("Normal", ncol(dat))
  names(stylenames) <- colnames(dat)
  stylenames <- as.list(stylenames)
  # restrict to only existing cols
  x$stylenames <- x$stylenames[names(x$stylenames) %in% colnames(dat)]
  stylenames <- modifyList(stylenames, val = x$stylenames)
  stylenames <- lapply(stylenames, function(x){
    sprintf("<w:pStyle w:stlname=\"%s\"/>", x)
  })
  stylenames
}

#' @export
to_wml.table_colwidths <- function(x, add_ns = FALSE, ...) {
  if(length(x$widths) < 1) return("")
  grid_col_str <- sprintf("<w:gridCol w:w=\"%.0f\"/>", x$widths * 1440)
  grid_col_str <- paste(grid_col_str, collapse = "")
  paste0("<w:tblGrid>", grid_col_str, "</w:tblGrid>")
}

#' @export
#' @title Table properties
#' @description Define table properties such as fixed or autofit layout,
#' table width in the document, eventually column widths.
#' @param style table style to be used to format table
#' @param layout layout defined by [table_layout()],
#' @param width table width in the document defined by [table_width()]
#' @param stylenames columns styles defined by [table_stylenames()]
#' @param colwidths column widths defined by [table_colwidths()]
#' @param align table alignment (one of left, center or right)
#' @param tcf conditional formatting settings defined by [table_conditional_formatting()]
#' @param word_title alternative text for Word table (used as title of the table)
#' @param word_description alternative text for Word table (used as description of the table)
#' @examples
#' prop_table()
#' to_wml(prop_table())
#' @family functions for table definition
prop_table <- function(style = NA_character_, layout = table_layout(),
                       width = table_width(),
                       stylenames = table_stylenames(),
                       colwidths = table_colwidths(),
                       tcf = table_conditional_formatting(),
                       align = "center",
                       word_title = NULL,
                       word_description = NULL){


  z <- list(
    style = style,
    layout = layout,
    width = width,
    colsizes = colwidths,
    stylenames = stylenames,
    tcf = tcf, align = align,
    word_title = word_title,
    word_description = word_description
  )
  class(z) <- c("prop_table")
  z
}

#' @export
to_wml.prop_table <- function(x, add_ns = FALSE, base_document = NULL, ...) {

  style <- NA_character_
  if(!is.null(x$style) && !is.na(x$style)){
    if (is.null(x$style) && !is.null(base_document$default_styles$table)) {
      style <- base_document$default_styles$table
    } else {
      style <- x$style
    }
  }

  tbl_layout <- to_wml(x$layout, add_ns= add_ns)

  width <- ""
  if(!is.null(x$width) && "autofit" %in% x$layout$type)
    width <- to_wml(x$width, add_ns= add_ns)

  colwidths <- to_wml(x$colsizes, add_ns= add_ns)
  tcf <- to_wml(x$tcf, add_ns= add_ns)
  paste0("<w:tblPr>",
         if(!is.null(x$word_title)) paste0("<w:tblCaption w:val=\"", htmlEscapeCopy(x$word_title), "\"/>"),
         if(!is.null(x$word_description)) paste0("<w:tblDescription w:val=\"", htmlEscapeCopy(x$word_description), "\"/>"),
         if(!is.na(style)) paste0("<w:tblStyle w:stlname=\"", style, "\"/>"),
         tbl_layout,
         sprintf( "<w:jc w:val=\"%s\"/>", x$align ),
         width, tcf,
         "</w:tblPr>",
         if(x$layout$type %in% "fixed") colwidths
         )

}


# table ----
table_docx <- function(x, header, style_id,
                       properties, alignment = NULL, add_ns = FALSE,
                       base_document = base_document) {
  open_tag <- tbl_ns_no
  if (add_ns) {
    open_tag <- tbl_ns_yes
  }

  str <- paste0(
    open_tag,
    to_wml(properties, add_ns = add_ns, base_document = base_document)
  )

  stylenames <- to_wml(properties$stylenames, base_document = base_document, dat = x)
  stylenames <- unlist(stylenames)
  stylenames <- as.character(stylenames)

  if(is.null(alignment)){
    alignment <- rep("", ncol(x))
  } else{
    alignment <- match.arg(alignment, c("left", "right", "center"), several.ok = TRUE )
    if(length(alignment) < ncol(x)){
      alignment <- rep(alignment, length.out = ncol(x) )
    }
    alignment <- sprintf("<w:jc w:val=\"%s\"/>", alignment)
  }

  header_str <- character(length = 0L)
  if (header) {

    header_str <- paste0(
      "<w:tr><w:trPr><w:tblHeader/></w:trPr>",
      paste0("<w:tc><w:p>",
             sprintf("<w:pPr>%s%s</w:pPr>", stylenames, alignment),
             "<w:r><w:t>",
        htmlEscapeCopy(enc2utf8(colnames(x))),
        "</w:t></w:r></w:p></w:tc>",
        collapse = ""
      ),
      "</w:tr>"
    )
  }
  as_tc <- function(x, align, stylenames) {
    paste0(
      "<w:tc><w:p>",
      sprintf("<w:pPr>%s%s</w:pPr>", stylenames, align),
      "<w:r><w:t>",
      htmlEscapeCopy(enc2utf8(x)),
      "</w:t></w:r></w:p></w:tc>"
    )
  }
  z <- mapply(as_tc, x, alignment, stylenames, SIMPLIFY = FALSE)
  z <- do.call(paste0, z)
  z <- paste0("<w:tr>", z, "</w:tr>", collapse = "")

  paste0(str, header_str, z, "</w:tbl>")
}

table_pptx <- function(x, style_id, col_width, row_height,
                       tcf, header = TRUE, alignment = NULL ){
  str <- paste0("<a:tbl>",
                sprintf("<a:tblPr %s>", to_pml(tcf)),
                sprintf("<a:tableStyleId>%s</a:tableStyleId>", style_id),
                "</a:tblPr>",
                "<a:tblGrid>",
                paste0(sprintf("<a:gridCol w=\"%.0f\"/>",
                               rep(col_width, length(x))),
                       collapse = ""),
                "</a:tblGrid>")

  as_tc <- function(x, align) {
    paste0("<a:tc><a:txBody><a:bodyPr/><a:lstStyle/><a:p>",
           "<a:pPr algn=\"", align, "\"/>",
           "<a:r><a:t>",
           htmlEscapeCopy(enc2utf8(x)),
           "</a:t></a:r></a:p></a:txBody></a:tc>"
    )
  }

  if(is.null(alignment)){
    alignment <- rep("r", ncol(x))
  } else{
    alignment <- match.arg(alignment, c("l", "r", "ctr"), several.ok = TRUE )
  }

  header_str <- character(length = 0L)
  if( header ){
    header_str  <- paste0(
      sprintf("<a:tr h=\"%.0f\">", row_height),
      paste0(as_tc(colnames(x), align = alignment), collapse = ""),
      "</a:tr>"
    )
  }

  z <- mapply(as_tc, x, alignment, SIMPLIFY = FALSE)
  z <- do.call(paste0, z)
  z <- paste0(sprintf("<a:tr h=\"%.0f\">", row_height), z, "</a:tr>", collapse = "")

  z <- paste0(str, header_str, z, "</a:tbl>")
  z <- paste0(
    "<a:graphic>",
    "<a:graphicData uri=\"http://schemas.openxmlformats.org/drawingml/2006/table\">",
    z, "</a:graphicData>", "</a:graphic>")
}


#' @export
#' @title Table block
#' @description Create a representation of a table
#' @param x a data.frame to add as a table
#' @param header display header if TRUE
#' @param properties table properties, see [prop_table()].
#' Table properties are not handled identically between Word and PowerPoint
#' output format. They are fully supported with Word but for PowerPoint (which
#' does not handle as many things as Word for tables), only conditional
#' formatting properties are supported.
#' @param alignment alignment for each columns, 'l' for left, 'r' for right
#' and 'c' for center. Default to NULL.
#' @examples
#' block_table(x = head(iris))
#'
#' block_table(x = mtcars, header = TRUE,
#'   properties = prop_table(
#'     tcf = table_conditional_formatting(
#'       first_row = TRUE, first_column = TRUE)
#'   ))
#' @family block functions for reporting
#' @seealso [prop_table()]
block_table <- function(x, header = TRUE, properties = prop_table(), alignment = NULL) {

  stopifnot(is.data.frame(x))
  if(inherits(x, "tbl_df"))
    x <- as.data.frame(
      x, check.names = FALSE, stringsAsFactors = FALSE )

  z <- list(
    x = x,
    header = header,
    properties = properties,
    alignment = alignment
  )
  class(z) <- c("block_table", "block")

  z
}

#' @export
#' @importFrom utils str
print.block_table <- function(x, ...) {
  str(x$x)
}

#' @export
to_wml.block_table <- function(x, add_ns = FALSE, base_document = NULL, ...) {

  value <- characterise_df(x$x)

  out <- table_docx(
    x = value, header = x$header,
    properties = x$properties,
    alignment = x$alignment,
    base_document = base_document,
    add_ns = add_ns
  )

  out
}

#' @export
to_pml.block_table <- function(x, add_ns = FALSE,
                               left = 0, top = 0, width = 3, height = 3,
                               bg = "transparent", rot = 0, label = "", ph = "<p:ph/>", ...){

  if( !is.null(bg) && !is.color( bg ) )
    stop("bg must be a valid color.", call. = FALSE )

  bg_str <- solid_fill_pml(bg)

  xfrm_str <- p_xfrm_str(left = left, top = top, width = width, height = height, rot = rot)
  if( is.null(ph) || is.na(ph)){
    ph = "<p:ph/>"
  }
  value <- characterise_df(x$x)
  value_str <- table_pptx(value, style_id = x$properties$style,
                          alignment = x$alignment,
                        col_width = as.integer((width/ncol(x$x))*914400),
                        row_height = as.integer((height/nrow(x$x))*914400),
                        tcf = x$properties$tcf,
                        header = x$header )

  id <- uuid_generate()
  str <- paste0("<p:graphicFrame xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
                "<p:nvGraphicFramePr>",
                sprintf("<p:cNvPr id=\"%s\" name=\"%s\"/>", id, label),
                "<p:cNvGraphicFramePr><a:graphicFrameLocks noGrp=\"1\"/></p:cNvGraphicFramePr>",
                sprintf("<p:nvPr>%s</p:nvPr>", ph),
                "</p:nvGraphicFramePr>",
                xfrm_str,
                bg_str,
                value_str,
                "</p:graphicFrame>")
  str
}

# fpar ----

#' @export
#' @title Formatted paragraph
#' @description Create a paragraph representation by concatenating
#' formatted text or images. The result can be inserted in a Word document
#' or a PowerPoint presentation and can also be inserted in a [block_list()]
#' call.
#'
#' All its arguments will be concatenated to create a paragraph where chunks of
#' text and images are associated with formatting properties.
#'
#' \code{fpar} supports [ftext()], [external_img()], \code{run_*} functions
#' (i.e. [run_autonum()], [run_word_field()]) when output is Word, and simple strings.
#'
#' Default text and paragraph formatting properties can also be modified
#' with function `update()`.
#'
#' @param ... cot objects ([ftext()], [external_img()])
#' @param fp_p paragraph formatting properties, see [fp_par()]
#' @param fp_t default text formatting properties. This is used as
#' text formatting properties when simple text is provided as argument,
#' see [fp_text()].
#' @param values a list of cot objects. If provided, argument \code{...} will be ignored.
#' @param object fpar object
#' @examples
#' fpar(ftext("hello", shortcuts$fp_bold()))
#'
#' # mix text and image -----
#' img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
#'
#' bold_face <- shortcuts$fp_bold(font.size = 12)
#' bold_redface <- update(bold_face, color = "red")
#' fpar_1 <- fpar(
#'   "Hello World, ",
#'   ftext("how ", prop = bold_redface ),
#'   external_img(src = img.file, height = 1.06/2, width = 1.39/2),
#'   ftext(" you?", prop = bold_face ) )
#' fpar_1
#'
#' img_in_par <- fpar(
#'   external_img(src = img.file, height = 1.06/2, width = 1.39/2),
#'   fp_p = fp_par(text.align = "center") )
#' @family block functions for reporting
#' @seealso [block_list()], [body_add_fpar()], [ph_with()]
fpar <- function( ..., fp_p = fp_par(), fp_t = fp_text_lite(), values = NULL) {
  out <- list()

  if( is.null(values)){
    values <- list(...)
  }

  out$chunks <- values
  out$fp_p <- fp_p
  out$fp_t <- fp_t
  class(out) <- c("fpar", "block")
  out
}

#' @export
#' @rdname fpar
#' @importFrom stats update
update.fpar <- function (object, fp_p = NULL, fp_t = NULL, ...){

  if(!is.null(fp_p)){
    object$fp_p <- fp_p
  }
  if(!is.null(fp_t)){
    object$fp_t <- fp_t
  }

  object
}


fortify_fpar <- function(x){
  lapply(x$chunks, function(chk) {
    if( !inherits(chk, c("cot", "run")) ){
      chk <- ftext(text = format(chk), prop = x$fp_t )
    }
    chk
  })
}


#' @export
as.data.frame.fpar <- function( x, ...){
  chks <- fortify_fpar(x)
  chks <- chks[sapply(chks, function(x) inherits(x, "ftext"))]
  chks <- mapply(function(x){
    data.frame(value = x$value, size = x$pr$font.size,
               bold = x$pr$bold, italic = x$pr$italic,
               font.family = x$pr$font.family, stringsAsFactors = FALSE )
  }, chks, SIMPLIFY = FALSE)
  rbind_match_columns(chks)
}


#' @export
to_wml.fpar <- function(x, add_ns = FALSE, style_id = NULL, ...) {

  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }
  if(is.null(style_id)){
    par_style <- ppr_wml(x$fp_p)
  } else par_style <- paste0(
    "<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>")

  chks <- fortify_fpar(x)
  z <- lapply(chks, to_wml)
  z$collapse <- ""
  z <- do.call(paste0, z)
  paste0(open_tag, par_style, z, "</w:p>")
}

#' @export
to_pml.fpar <- function(x, add_ns = FALSE, ...) {

  open_tag <- ap_ns_no
  if (add_ns) {
    open_tag <- ap_ns_yes
  }

  par_style <- ppr_pml(x$fp_p)
  chks <- fortify_fpar(x)
  z <- lapply(chks, to_pml)
  z$collapse <- ""
  z <- do.call(paste0, z)
  paste0(open_tag, par_style, z, "</a:p>")
}


#' @export
to_html.fpar <- function(x, add_ns = FALSE, ...) {
  par_style <- ppr_css(x$fp_p)
  chks <- fortify_fpar(x)
  z <- lapply(chks, to_html)
  z$collapse <- ""
  z <- do.call(paste0, z)
  sprintf("<p style=\"%s\">%s</p>", par_style, z)
}

# block_list -----

#' @export
#' @title List of blocks
#' @description A list of blocks can be used to gather
#' several blocks (paragraphs, tables, ...) into a single
#' object. The result can be added into a Word document or a
#' PowerPoint presentation.
#' @param ... a list of blocks. When output is only for
#' Word, objects of class [external_img()] can
#' also be used in fpar construction to mix text and images
#' in a single paragraph. Supported objects are:
#' [block_caption()], [block_pour_docx()], [block_section()],
#' [block_table()], [block_toc()], [fpar()], [plot_instr()].
#' @examples
#' # block list ------
#'
#' img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
#' fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
#' fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)
#'
#'
#' ## This can be only be used in a MS word output as pptx does
#' ## not support paragraphs made of text and images.
#' ## (actually it can be used but image will not appear in the
#' ## pptx output)
#' value <- block_list(
#'   fpar(ftext("hello world", fpt_blue_bold)),
#'   fpar(ftext("hello", fpt_blue_bold), " ",
#'        ftext("world", fpt_red_italic)),
#'   fpar(
#'     ftext("hello world", fpt_red_italic),
#'           external_img(
#'             src = img.file, height = 1.06, width = 1.39)))
#' value
#'
#' doc <- read_docx()
#' doc <- body_add(doc, value)
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#'
#' value <- block_list(
#'   fpar(ftext("hello world", fpt_blue_bold)),
#'   fpar(ftext("hello", fpt_blue_bold), " ",
#'        ftext("world", fpt_red_italic)),
#'   fpar(
#'     ftext("blah blah blah", fpt_red_italic)))
#' value
#'
#' doc <- read_pptx()
#' doc <- add_slide(doc)
#' doc <- ph_with(doc, value, location = ph_location_type(type = "body"))
#' print(doc, target = tempfile(fileext = ".pptx"))
#' @seealso [ph_with()], [body_add_blocks()], [fpar()]
#' @family block functions for reporting
block_list <- function(...){
  x <- list(...)
  z <- list()
  for(i in x){
    if(inherits(i, "block")) {
      z <- append(z, list(i))
    } else if(inherits(i, "flextable")){
      z <- append(z, list(i))
    } else if(is.character(i)){
      z <- append(z, lapply(i, fpar))
    }
  }
  class(z) <- c("block_list", "block")
  z
}

#' @export
to_wml.block_list <- function(x, add_ns = FALSE, ...) {
  out <- character(length(x))
  for(i in seq_along(x) ){
    out[i] <- to_wml(x[[i]], add_ns = add_ns)
  }

  paste0(out, collapse = "")
}

#' @export
to_pml.block_list <- function(x, add_ns = FALSE, ...) {
  pars <- sapply(x, to_pml)
  pars <- paste0(pars, collapse = "")
  pars
}

#' @export
to_html.block_list <- function(x, add_ns = FALSE, ...) {
  str <- vapply(x, to_html, NA_character_)
  paste0(str, collapse = "")
}

# unordered list ----
#' @export
#' @title Unordered list
#' @description unordered list of text for PowerPoint
#' presentations. Each text is associated with
#' a hierarchy level.
#' @param str_list list of strings to be included in the object
#' @param level_list list of levels for hierarchy structure. Use
#' 0 for 'no bullet', 1 for level 1, 2 for level 2 and so on.
#' @param style text style, a \code{fp_text} object list or a
#' single \code{fp_text} objects. Use \code{fp_text(font.size = 0, ...)} to
#' inherit from default sizes of the presentation.
#' @examples
#' unordered_list(
#' level_list = c(1, 2, 2, 3, 3, 1),
#' str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
#' style = fp_text(color = "red", font.size = 0) )
#' unordered_list(
#' level_list = c(1, 2, 1),
#' str_list = c("Level1", "Level2", "Level1"),
#' style = list(
#'   fp_text(color = "red", font.size = 0),
#'   fp_text(color = "pink", font.size = 0),
#'   fp_text(color = "orange", font.size = 0)
#'   ))
#' @seealso \code{\link{ph_with}}
#' @family block functions for reporting
unordered_list <- function(str_list = character(0), level_list = integer(0), style = NULL){
  stopifnot(is.character(str_list))
  stopifnot(is.numeric(level_list))

  if (length(str_list) != length(level_list) & length(str_list) > 0) {
    stop("str_list and level_list have different lenghts.")
  }

  if( !is.null(style)){
    if( inherits(style, "fp_text") )
      style <- lapply(seq_len(length(str_list)), function(x) style )
  }
  x <- list(
    str = str_list,
    lvl = level_list,
    style = style
  )
  class(x) <- "unordered_list"
  x
}
#' @export
#' @noRd
print.unordered_list <- function(x, ...){
  print(data.frame(str = x$str,
                   lvl = x$lvl,
                   stringsAsFactors = FALSE))
  invisible()
}

#' @export
to_pml.unordered_list <- function(x, add_ns = FALSE, ...) {

  open_tag <- ap_ns_no
  if (add_ns) {
    open_tag <- ap_ns_yes
  }
  if( !is.null(x$style)){
    style_str <- sapply(x$style, format, type = "pml")
    style_str <- rep_len(style_str, length.out = length(x$str))
  } else style_str <- rep("<a:rPr/>", length(x$str))
  tmpl <- "%s<a:pPr%s>%s</a:pPr><a:r>%s<a:t>%s</a:t></a:r></a:p>"
  lvl <- sprintf(" lvl=\"%.0f\"", x$lvl - 1)
  lvl <- ifelse(x$lvl > 1, lvl, "")
  bu_none <- ifelse(x$lvl < 1, "<a:buNone/>", "")
  p <- sprintf(tmpl, open_tag, lvl, bu_none, style_str, htmlEscapeCopy(x$str) )
  p <- paste(p, collapse = "")
  p
}

# plot_instr -----
#' @title Wrap plot instructions for png plotting in Powerpoint or Word
#' @description A simple wrapper to capture
#' plot instructions that will be executed and copied in a document. It produces
#' an object of class 'plot_instr' with a corresponding method [ph_with()] and
#' [body_add_plot()].
#'
#' The function enable usage of any R plot with argument `code`. Wrap your code
#' between curly bracket if more than a single expression.
#'
#' @param code plotting instructions
#' @examples
#' # plot_instr demo ----
#'
#' anyplot <- plot_instr(code = {
#'   barplot(1:5, col = 2:6)
#'   })
#'
#' doc <- read_docx()
#' doc <- body_add(doc, anyplot, width = 5, height = 4)
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#'
#' doc <- read_pptx()
#' doc <- add_slide(doc)
#' doc <- ph_with(
#'   doc, anyplot,
#'   location = ph_location_fullsize(),
#'   bg = "#00000066", pointsize = 12)
#' print(doc, target = tempfile(fileext = ".pptx"))
#' @export
#' @import graphics
#' @seealso [ph_with()], [body_add_plot()]
#' @family block functions for reporting
plot_instr <- function(code) {
  out <- list()
  out$code <- substitute(code)
  class(out) <- "plot_instr"
  return(out)
}

library(officer)

def_text <- fp_text_lite(color = "#006699", bold = TRUE)
center_par <- fp_par(text.align = "center", padding = 3)

doc <- rtf_doc(
  normal_par = fp_par(line_spacing = 1.4, padding = 3)
)

doc <- rtf_add(
  x = doc,
  value = fpar(
    ftext("how are you?", prop = def_text),
    fp_p = fp_par(text.align = "center")
  )
)

a_paragraph <- fpar(
  ftext("Here is a date: ", prop = def_text),
  run_word_field(field = "Date \\@ \"MMMM d yyyy\""),
  fp_p = center_par
)
doc <- rtf_add(
  x = doc,
  value = block_list(
    a_paragraph,
    a_paragraph,
    a_paragraph
  )
)

if (require("ggplot2")) {
  gg <- gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))
  doc <- rtf_add(doc, gg, width = 3, height = 4, ppr = center_par)
}
anyplot <- plot_instr(code = {
  barplot(1:5, col = 2:6)
})
doc <- rtf_add(doc, anyplot, width = 5, height = 4, ppr = center_par)

print(doc, target = tempfile(fileext = ".rtf"))

props_to_ph_location <- function(props) {
  if (nrow(props) > 1) {
    cli::cli_alert_warning("More than one placeholder selected.")
  }
  props <- props[, c(
    "offx",
    "offy",
    "cx",
    "cy",
    "ph_label",
    "ph",
    "type",
    "fld_id",
    "fld_type",
    "rotation"
  )]
  names(props) <- c(
    "left",
    "top",
    "width",
    "height",
    "ph_label",
    "ph",
    "type",
    "fld_id",
    "fld_type",
    "rotation"
  )
  as_ph_location(props)
}


# id is deprecated and replaced by type_idx. Will be removed soon
get_ph_loc <- function(
  x,
  layout,
  master,
  type,
  type_idx = NULL,
  position_right,
  position_top,
  id = NULL,
  ph_id = NULL
) {
  props <- layout_properties(x, layout = layout, master = master)

  if (!is.null(ph_id)) {
    ids <- sort(stats::na.omit(as.numeric(props$id)))
    if (length(ids) <= 20) {
      .all_ids_switch <- c("x" = "Available ids:  {.val {ids}}.") # only if few ids
    } else {
      .all_ids_switch <- NULL
    }
    if (!ph_id %in% ids) {
      cli::cli_abort(
        c(
          "{.arg id} {.val {ph_id}} does not exist.",
          .all_ids_switch,
          "i" = cli::col_grey(
            "see column {.val id} in {.code layout_properties(..., '{layout}', '{master}')}"
          )
        ),
        call = NULL
      )
    }
    props <- props[props$id == ph_id, , drop = FALSE]
    return(props_to_ph_location(props))
  }

  types_on_layout <- unique(props$type)
  props <- props[props$type %in% type, , drop = FALSE]
  nr <- nrow(props)
  if (nr < 1) {
    cli::cli_abort(
      c(
        "Found no placeholder of type {.val {type}} on layout {.val {layout}}.",
        "x" = "Available types are {.val {types_on_layout}}",
        "i" = cli::col_grey(
          "see {.code layout_properties(x, '{layout}', '{master}')}"
        )
      ),
      call = NULL
    )
  }

  # id and type_idx are both used for now. 'id' is deprecated. The following code block can be removed in the future.
  if (!is.null(id)) {
    if (!id %in% 1L:nr) {
      cli::cli_abort(
        c(
          "{.arg id} is out of range.",
          "x" = "Must be between {.val {1L}} and {.val {nr}} for ph type {.val {type}}.",
          "i" = cli::col_grey(
            "see {.code layout_properties(x, '{layout}', '{master}')} for all phs with type '{type}'"
          )
        ),
        call = NULL
      )
    }
    # the ordering of 'type_idx' (top->bottom, left-righ) is different than for the 'id' arg (index
    # along the id colomn). Here, we restore the old ordering, to avoid a breaking change.
    # set order for type idx. Removing the line would result in the default
    # layout properties order, i.e., top->bottom left->right.
    props <- arrange(props, .data$type, as.integer(.data$id))
    props$.id <- stats::ave(
      props$type,
      props$master_name,
      props$name,
      props$type,
      FUN = seq_along
    )
    props <- props[props$.id == id, , drop = FALSE]
    return(props_to_ph_location(props))
  }

  if (!is.null(type_idx)) {
    if (!type_idx %in% props$type_idx) {
      cli::cli_abort(
        c(
          "{.arg type_idx} is out of range.",
          "x" = "Must be between {.val {1L}} and {.val {max(props$type_idx)}} for ph type {.val {type}}.",
          "i" = cli::col_grey(
            "see {.code layout_properties(..., layout = '{layout}', master = '{master}')} for indexes of type '{type}'"
          )
        ),
        call = NULL
      )
    }
    props <- props[props$type_idx == type_idx, , drop = FALSE]
    return(props_to_ph_location(props))
  }

  if (position_right) {
    props <- props[props$offx + 0.0001 > max(props$offx), ]
  } else {
    props <- props[props$offx - 0.0001 < min(props$offx), ]
  }
  if (position_top) {
    props <- props[props$offy - 0.0001 < min(props$offy), ]
  } else {
    props <- props[props$offy + 0.0001 > max(props$offy), ]
  }
  props_to_ph_location(props)
}


as_ph_location <- function(x, ...) {
  if (!is.data.frame(x)) {
    cli::cli_abort(
      c(
        "{.arg x} must be a data frame.",
        "x" = "You provided {.cls {class(x)[1]}} instead."
      )
    )
  }
  ref_names <- c(
    "width",
    "height",
    "left",
    "top",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )
  if (!all(is.element(ref_names, names(x)))) {
    stop(
      "missing column values:",
      paste0(setdiff(ref_names, names(x)), collapse = ",")
    )
  }
  out <- x[ref_names]
  as.list(out)
}


is_ph_location <- function(x) {
  inherits(x, "location_num") || inherits(x, "location_str")
}


#' @export
#' @title Eval a location on the current slide
#' @description Eval a shape location against the current slide.
#' This function is to be used to add custom openxml code. A
#' list is returned, it contains informations width, height, left
#' and top positions and other informations necessary to add a
#' content on a slide.
#' @param x a location for a placeholder.
#' @param doc an rpptx object
#' @param ... unused arguments
#' @example inst/examples/example_fortify_location.R
#' @seealso [ph_location()], [ph_with()]
#' @family functions for officer extensions
#' @keywords internal
fortify_location <- function(x, doc, ...) {
  UseMethod("fortify_location")
}

# _________________ ----
# main ----

#' @export
#' @title Location for a placeholder from scratch
#' @description The function will return a list that complies with
#' expected format for argument `location` of function [ph_with()].
#' @param left,top,width,height place holder coordinates
#' in inches.
#' @param newlabel a label for the placeholder. See section details.
#' @param bg background color
#' @param rotation rotation angle
#' @param ln a [sp_line()] object specifying the outline style.
#' @param geom shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html
#' @param ... unused arguments
#' @family functions for placeholder location
#' @details
#' The location of the bounding box associated to a placeholder
#' within a slide is specified with the left top coordinate,
#' the width and the height. These are defined in inches:
#'
#' \describe{
#'   \item{left}{left coordinate of the bounding box}
#'   \item{top}{top coordinate of the bounding box}
#'   \item{width}{width of the bounding box}
#'   \item{height}{height of the bounding box}
#' }
#'
#' In addition to these attributes, a label can be
#' associated with the shape. Shapes, text boxes, images and other objects
#' will be identified with that label in the *Selection Pane* of PowerPoint.
#' This label can then be reused by other functions such as `ph_location_label()`.
#' It can be set with argument `newlabel`.
#' @example inst/examples/example_ph_location.R
ph_location <- function(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  bg = NULL,
  rotation = NULL,
  ln = NULL,
  geom = NULL,
  ...
) {
  # remove name in case of named vector entry, e.g. ph_location(top = x["top])
  left <- unname(left)
  top <- unname(top)
  width <- unname(width)
  height <- unname(height)
  x <- list(
    left = left,
    top = top,
    width = width,
    height = height,
    ph_label = newlabel,
    ph = NA_character_,
    bg = bg,
    rotation = rotation,
    ln = ln,
    geom = geom,
    fld_type = NA_character_,
    fld_id = NA_character_
  )
  class(x) <- c("location_manual", "location_str")
  x
}


#' @export
fortify_location.location_manual <- function(x, doc, ...) {
  x
}

#' @title Location for a placeholder based on a template
#' @description The function will return a list that complies with
#' expected format for argument `location` of function
#' [ph_with()]. A placeholder will be used as template
#' and its positions will be updated with values `left`, `top`, `width`, `height`.
#' @param left,top,width,height place holder coordinates
#' in inches.
#' @param newlabel a label for the placeholder. See section details.
#' @param type placeholder type to look for in the slide layout, one
#' of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.
#' It will be used as a template placeholder.
#' @param id index of the placeholder template. If two body placeholder, there can be
#' two different index: 1 and 2 for the first and second body placeholders defined
#' in the layout.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_template.R
#' @export
ph_location_template <- function(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  type = NULL,
  id = 1,
  ...
) {
  x <- list(
    left = left,
    top = top,
    width = width,
    height = height,
    ph_label = newlabel,
    ph = NA_character_,
    type = type,
    id = id
  )

  class(x) <- c("location_template", "location_str")
  x
}
#' @export
fortify_location.location_template <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  if (!is.null(x$type)) {
    ph <- slide$get_xfrm(type = x$type, index = x$id)$ph
  } else {
    ph <- sprintf('<p:ph type="%s"/>', "body")
  }
  x <- ph_location(
    left = x$left,
    top = x$top,
    width = x$width,
    height = x$height,
    label = x$ph_label
  )
  x$ph <- ph
  fortify_location.location_manual(x)
}


#' @export
#' @title Location of a placeholder based on a type
#' @description The function will use the type name of the placeholder (e.g. body, title),
#' the layout name and few other criterias to find the corresponding location.
#' @param type placeholder type to look for in the slide layout, one
#' of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.
#' @param position_right the parameter is used when a selection with above
#' parameters does not provide a unique position (for example
#' layout 'Two Content' contains two element of type 'body').
#' If `TRUE`, the element the most on the right side will be selected,
#' otherwise the element the most on the left side will be selected.
#' @param position_top same than `position_right` but applied
#' to top versus bottom.
#' @param type_idx Type index of the placeholder. If there is more than one
#' placeholder of a type (e.g., `body`), the type index can be supplied to uniquely
#' identify a ph. The index is a running number starting at 1. It is assigned by
#' placeholder position  (top -> bottom, left -> right). See [plot_layout_properties()]
#' for details. If `idx` argument is used, `position_right` and `position_top`
#' are ignored.
#' @param id (**DEPRECATED, use `type_idx` instead**) Index of the placeholder.
#' If two body placeholder, there can be two different index: 1 and 2 for the
#' first and second body placeholders defined in the layout. If this argument
#' is used, `position_right` and `position_top` will be ignored.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_type.R
ph_location_type <- function(
  type = "body",
  type_idx = NULL,
  position_right = TRUE,
  position_top = TRUE,
  newlabel = NULL,
  id = NULL,
  ...
) {
  # the following two warnings can be deleted after the deprecated id arg is removed.
  if (!is.null(id) && !is.null(type_idx)) {
    cli::cli_warn("{.arg id} is ignored if {.arg type_idx} is provided ")
  }
  if (!is.null(id) && is.null(type_idx)) {
    cli::cli_warn(
      c(
        "!" = "The {.arg id} argument in {.fn ph_location_type} is deprecated as of {.pkg officer} 0.6.7.",
        "i" = "Please use the {.arg type_idx} argument instead.",
        "x" = cli::col_red(
          "Caution: new index logic in {.arg type_idx} (see docs)."
        )
      )
    )
  }

  ph_types <- c(
    "ctrTitle",
    "subTitle",
    "dt",
    "ftr",
    "sldNum",
    "title",
    "body",
    "pic",
    "chart",
    "tbl",
    "dgm",
    "media",
    "clipArt"
  )
  if (!type %in% ph_types) {
    cli::cli_abort(
      c(
        "type {.val {type}} is unknown.",
        "x" = "Must be one of {.or {.val {ph_types}}}"
      ),
      call = NULL
    )
  }
  x <- list(
    type = type,
    type_idx = type_idx,
    position_right = position_right,
    position_top = position_top,
    id = id,
    label = newlabel
  )
  class(x) <- c("location_type", "location_str")
  x
}


#' @export
fortify_location.location_type <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()
  args <- list(...)

  layout <- ifelse(is.null(args$layout), unique(xfrm$name), args$layout)
  master <- ifelse(is.null(args$master), unique(xfrm$master_name), args$master)

  # to avoid a breaking change, the deprecated id is passed along.
  # As type_idx uses a different index order than id, this is necessary until the id arg is removed.
  out <- get_ph_loc(
    doc,
    layout = layout,
    master = master,
    type = x$type,
    position_right = x$position_right,
    position_top = x$position_top,
    type_idx = x$type_idx,
    id = x$id,
    ph_id = NULL # id is deprecated and will be removed soon
  )
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a named placeholder
#' @description The function will use the label of a placeholder
#' to find the corresponding location.
#' @param ph_label placeholder label of the used layout. It can be read in PowerPoint or
#' with function [layout_properties()] in column `ph_label`.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_label.R
ph_location_label <- function(ph_label, newlabel = NULL, ...) {
  x <- list(ph_label = unname(ph_label), label = unname(newlabel))
  class(x) <- c("location_label", "location_str")
  x
}


#' @export
fortify_location.location_label <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()

  layout <- unique(xfrm$name)
  master <- unique(xfrm$master_name)

  props_all <- layout_properties(doc, layout = layout, master = master)

  if (nrow(props_all) == 0) {
    cli::cli_abort(
      c(
        "ph label {.val {x$ph_label}} does not exist",
        "x" = "The layout has no placeholders!"
      ),
      call = NULL
    )
  }

  props <- props_all[props_all$ph_label %in% x$ph_label, , drop = FALSE]

  if (nrow(props) < 1) {
    cli::cli_abort(
      c(
        "ph label {.val {x$ph_label}} does not exist",
        "x" = "layout {.val {layout}} has ph labels: {.val {props_all$ph_label}}",
        "i" = "See {.fn plot_layout_properties} for details"
      ),
      call = NULL
    )
  }

  if (nrow(props) > 1) {
    cli::cli_abort(
      c(
        "Placeholder {.val {x$ph_label}} label in layout {.val {layout}} is duplicated.",
        "x" = "A layout's ph labels must be unique.",
        "i" = "{.fn layout_dedupe_ph_labels} helps handling duplicates."
      ),
      call = NULL
    )
  }

  props <- props[, c(
    "offx",
    "offy",
    "cx",
    "cy",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )]
  names(props) <- c(
    "left",
    "top",
    "width",
    "height",
    "ph_label",
    "ph",
    "type",
    "rotation",
    "fld_id",
    "fld_type"
  )
  row.names(props) <- NULL
  out <- as_ph_location(props)
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a full size element
#' @description The function will return the location corresponding
#' to a full size display.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @example inst/examples/example_ph_location_fullsize.R
ph_location_fullsize <- function(newlabel = "", ...) {
  x <- list(label = newlabel)
  class(x) <- c("location_fullsize", "location_str")
  x
}

#' @export
fortify_location.location_fullsize <- function(x, doc, ...) {
  layout_data <- slide_size(doc)
  layout_data$left <- 0L
  layout_data$top <- 0L
  if (!is.null(x$label)) {
    layout_data$ph_label <- x$label
  }
  layout_data$ph <- NA_character_
  layout_data$type <- "body"
  layout_data$rotation <- 0L
  layout_data$fld_id <- NA_character_
  layout_data$fld_type <- NA_character_

  as_ph_location(as.data.frame(layout_data, stringsAsFactors = FALSE))
}

#' @export
#' @title Location of a left body element
#' @description The function will return the location corresponding
#' to a left bounding box. The function assume the layout 'Two Content'
#' is existing. This is an helper function, if you don't have a layout
#' named 'Two Content', use [ph_location_type()] and set arguments
#' to your specific needs.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @example inst/examples/example_ph_location_left.R
ph_location_left <- function(newlabel = NULL, ...) {
  x <- list(label = newlabel)
  class(x) <- c("location_left", "location_str")
  x
}

#' @export
fortify_location.location_left <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()

  args <- list(...)
  master <- if (is.null(args$master)) unique(xfrm$master_name) else args$master
  out <- get_ph_loc(
    doc,
    layout = "Two Content",
    master = master,
    type = "body",
    position_right = FALSE,
    position_top = TRUE
  )
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a right body element
#' @description The function will return the location corresponding
#' to a right bounding box. The function assume the layout 'Two Content'
#' is existing. This is an helper function, if you don't have a layout
#' named 'Two Content', use [ph_location_type()] and set arguments
#' to your specific needs.
#' @param newlabel a label to associate with the placeholder.
#' @param ... unused arguments
#' @family functions for placeholder location
#' @example inst/examples/example_ph_location_left.R
ph_location_right <- function(newlabel = NULL, ...) {
  x <- list(label = newlabel)
  class(x) <- c("location_right", "location_str")
  x
}


#' @export
fortify_location.location_right <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()

  args <- list(...)
  master <- ifelse(is.null(args$master), unique(xfrm$master_name), args$master)
  out <- get_ph_loc(
    doc,
    layout = "Two Content",
    master = master,
    type = "body",
    position_right = TRUE,
    position_top = TRUE
  )
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}


#' @export
#' @title Location of a placeholder based on its id
#' @description Each placeholder has an id (a low integer value). The ids are unique across a single
#' layout. The function uses the placeholder's id to reference it. Different from a ph label,
#' the id is auto-assigned by PowerPoint and cannot be modified by the user.
#' Use [layout_properties()] (column `id`) and [plot_layout_properties()] (upper right
#' corner, in green) to find a placeholder's id.
#'
#' @param id placeholder id.
#' @param newlabel a new label to associate with the placeholder.
#' @param ... not used.
#' @family functions for placeholder location
#' @inherit ph_location details
#' @example inst/examples/example_ph_location_id.R
ph_location_id <- function(id, newlabel = NULL, ...) {
  ph_id <- id # for disambiguation, store initial value

  if (length(ph_id) > 1) {
    cli::cli_abort(
      c(
        "{.arg id} must be {cli::style_underline('one')} number",
        "x" = "Found more than one entry: {.val {ph_id}}"
      )
    )
  }
  if (is.null(ph_id) || is.na(ph_id) || length(ph_id) == 0) {
    cli::cli_abort("{.arg id} must be a positive number")
  }
  if (!is.integer(ph_id)) {
    ph_id <- suppressWarnings(as.integer(ph_id))
    if (is.na(ph_id)) {
      cli::cli_abort(
        c(
          "Cannot convert {.val {id}} to integer",
          "x" = "{.arg id} must be a number, you provided class {.cls {class(id)[1]}}"
        )
      )
    }
  }
  if (ph_id < 1) {
    cli::cli_abort(
      c(
        "{.arg id} must be a {cli::style_underline('positive')} number",
        "x" = "Found {.val {ph_id}}"
      )
    )
  }
  x <- list(
    type = NULL,
    type_idx = NULL,
    position_right = NULL,
    position_top = NULL,
    id = NULL,
    ph_id = ph_id,
    label = newlabel
  )
  class(x) <- c("location_id", "location_num")
  x
}


#' @export
fortify_location.location_id <- function(x, doc, ...) {
  slide <- doc$slide$get_slide(doc$cursor)
  xfrm <- slide$get_xfrm()
  args <- list(...)

  layout <- ifelse(is.null(args$layout), unique(xfrm$name), args$layout)
  master <- ifelse(is.null(args$master), unique(xfrm$master_name), args$master)
  out <- get_ph_loc(doc, layout = layout, master = master, ph_id = x$ph_id)
  if (!is.null(x$label)) {
    out$ph_label <- x$label
  }
  out
}

# _________________ ----
# resolve ----

#' @export
#' @keywords internal
#' @title Resolve short form location
#' @description
#' Convert short form location format, a numeric or a string (e.g. `"body [1]"`),
#' into its corresponding location object. Under the hood, we parse the short form location and
#' call the corresponding function from the `ph_location_*` family. Note that short forms may not
#' cover all function from the `ph_location_*` and offer less customization.
#'
#' @section Short forms:
#' The following location short forms are implemented. The corresponding call of the function from the
#' `ph_location_*` family is displayed on the right.
#'
#' | **Short form** | **Description**                                             | **Location function**           |
#' |----------------|-------------------------------------------------------------|---------------------------------|
#' | `"left"`       | Keyword string                                              | `ph_location_left()`            |
#' | `"right"`      | Keyword string                                              | `ph_location_right()`           |
#' | `"fullsize"`   | Keyword string                                              | `ph_location_fullsize()`        |
#' | `"body [1]"`   | String: type + index in brackets (`1` if omitted)           | `ph_location_type("body", 1)`   |
#' | `"my_label"`   | Any string not matching a keyword or type                   | `ph_location_label("my_label")` |
#' | `2`            | Length 1 integer                                            | `ph_location_id(2)`             |
#' | `c(0,0,4,5)`   | Length 4 numeric, optionally named, `c(top=0, left=0, ...)` | `ph_location(0, 0, 4, 5)`       |
#'
#' @example inst/examples/example_resolve_location.R
resolve_location <- function(x) {
  if (is_ph_location(x)) {
    return(x)
  }
  if (is.null(x) || (is.vector(x) && is.na(x)[1])) {
    raise_location_value_error(x)
  }
  if (is.numeric(x)) {
    loc <- resolve_location_from_numeric(x)
  } else if (is.character(x)) {
    loc <- resolve_location_from_character(x)
  } else {
    cli::cli_abort(
      call = NULL,
      "Cannot resolve class {.cls {class(x)[1]}} into a location",
      "x" = "Must be a character or numeric vector."
    )
  }
  loc
}


raise_location_value_error <- function(x) {
  cli::cli_abort(
    call = NULL,
    c(
      "Incorrect value for {.arg location}",
      "x" = "Must be a vector (character or numeric) or a ph_location object"
    )
  )
}


resolve_location_from_numeric <- function(x) {
  # length 1 integer => ph_location_id()
  # length 4 numeric => ph_location()
  len <- length(x)
  if (len == 1) {
    if (!is_integerish(x)) {
      cli::cli_abort(
        c(
          "{.arg location} is a length 1 {.cls {class(x)[1]}}: {.val {x}}",
          "x" = "If a length 1 numeric, {.arg location} requires {.cls integer}"
        ),
        call = NULL
      )
    }
    location <- resolve_ph_location_id(x)
  } else if (len == 4) {
    location <- resolve_ph_location(x)
  } else {
    cli::cli_abort(
      c(
        "{.arg location} has incorrect length.",
        "x" = "Numeric vector passed to {.arg location} must have length 1 or 4"
      ),
      call = NULL
    )
  }
  location
}


resolve_ph_location_id <- function(x) {
  if (x < 0) {
    cli::cli_abort(
      c(
        "{.arg location} is negative.",
        "x" = "Integers passed to {.arg location} must be positive"
      ),
      call = NULL
    )
  }
  ph_location_id(id = x)
}


# checks named numeric vector with position info [for ph_location_position()].
# c(left =, top =, width =, height = )
#
fortify_named_location_position <- function(x) {
  if (!is_named(x)) {
    cli::cli_abort(
      c(
        "Some vector elements have no names",
        "x" = "{.arg x} must be a named vector"
      )
    )
  }
  args <- names(x)
  expected <- c("left", "top", "width", "height")
  matched <- pmatch(args, expected, duplicates.ok = TRUE)
  nms_new <- ifelse(is.na(matched), NA, expected[matched])

  # unknown position
  i_na <- is.na(nms_new)
  if (any(i_na)) {
    cli::cli_abort(
      c(
        "Found {sum(i_na)} unknown name{?s} in {.arg location}: {.val {args[i_na]}}",
        "x" = "{.arg location} understands {.val {expected}}",
        "i" = cli::col_silver("Partial name matching is supported")
      ),
      call = NULL
    )
  }
  # duplicate position
  ii_dupes <- duplicated(nms_new)
  if (any(ii_dupes)) {
    cli::cli_abort(
      c(
        "Duplicate entries in {.arg location}: {.val {unique(nms_new[ii_dupes])}}",
        "x" = "Each name in {.arg location} must be unique",
        "i" = cli::col_silver("Partial name matching is supported")
      ),
      call = NULL
    )
  }
  setNames(x, nms_new)
}


resolve_ph_location <- function(x) {
  if (is.null(names(x))) {
    names(x) <- c("left", "top", "width", "height") # same order as in ph_location() !
  }
  x <- fortify_named_location_position(x)
  do.call(ph_location, as.list(x))
}


resolve_location_from_character <- function(x) {
  if (is.character(x) && length(x) > 1) {
    cli::cli_abort(
      c(
        "{.arg location} has incorrect length.",
        "x" = "Character vector passed to {.arg location} must have length 1"
      ),
      call = NULL
    )
  }
  # cases:
  # - keyword left, => ph_location_left()
  #           right => ph_location_right()
  #           fullsize => ph_location_fullsize()
  # - type: body, title, ctrTitle, subTitle, dt, ftr, sldNum
  #         => ph_location_type()
  # - label: <any> => ph_location_label()
  if (x == "left") {
    location <- ph_location_left()
  } else if (x == "right") {
    location <- ph_location_right()
  } else if (x == "fullsize") {
    location <- ph_location_fullsize()
  } else if (has_ph_type_format(x)) {
    location <- do.call(ph_location_type, get_ph_type_info(x))
  } else {
    location <- ph_location_label(x)
  }
  location
}

# matches pattern "type [type_idx ]",
# e.g. "body", "body[1]", ""body [1]", "body    [ 1 ]" => all identical
.ph_type_pattern <- "^(body|title|ctrTitle|subTitle|dt|ftr|sldNum)\\s*(\\[\\s*\\d+\\s*\\])?$"

has_ph_type_format <- function(x) {
  grepl(.ph_type_pattern, trimws(x))
}


# extract type name and idx from squared brackets
# e.g. "body [1]" => list(type = "body", type_idx = 1)
get_ph_type_info <- function(x) {
  x <- trimws(x)
  matches <- regexec(.ph_type_pattern, x)
  extracted <- regmatches(x, matches)[[1]]
  l <- setNames(as.list(extracted), c("input", "type", "type_idx"))
  l$type_idx <- ifelse(l$type_idx == "", "[1]", l$type_idx) # if brackets with index is missing, "[1]" is used
  l$type_idx <- extract_integers(l$type_idx)
  l
}


# extract integers from text, e.g. "[1]" => 1
extract_integers <- function(x) {
  matches <- gregexpr("\\d+", x)
  v <- unlist(regmatches(x, matches))
  as.numeric(v)
}

# check properties helpers ----

check_spread_integer <- function(obj, value, dest) {
  varname <- as.character(substitute(value))
  if (is.numeric(value) && length(value) == 1 && value >= 0) {
    for (i in dest) {
      obj[[i]] <- as.integer(value)
    }
  } else {
    stop(varname, " must be a positive integer scalar.", call. = FALSE)
  }
  obj
}

check_set_color <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (!is.color(value) && !is.na(value)) {
    stop(varname, " must be a valid color.", call. = FALSE)
  } else {
    obj[[varname]] <- value
  }
  obj
}

check_set_pic <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (!grepl(pattern = "^rId[0-9]+", value)) {
    stop(varname, " must be a valid reference id: ", value, call. = FALSE)
  }
  obj[[varname]] <- value
  obj
}
check_set_file <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (!file.exists(value)) {
    stop(varname, " must be a valid filename.", call. = FALSE)
  }
  obj[[varname]] <- value
  obj
}
check_set_border <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (!inherits(value, "fp_border")) {
    stop(varname, " must be a fp_border object.", call. = FALSE)
  } else {
    obj[[varname]] <- value
  }
  obj
}

check_spread_border <- function(obj, value, dest) {
  varname <- as.character(substitute(value))
  if (!inherits(value, "fp_border")) {
    stop(varname, " must be a fp_border object.", call. = FALSE)
  }
  for (i in dest) {
    obj[[i]] <- value
  }
  obj
}

check_set_integer <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.numeric(value) && length(value) == 1 && value >= 0)) {
    obj[[varname]] <- as.integer(value)
  } else {
    stop(varname, " must be a positive integer scalar.", call. = FALSE)
  }
  obj
}

check_set_numeric <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.numeric(value) && length(value) == 1 && value >= 0)) {
    obj[[varname]] <- as.double(value)
  } else {
    stop(varname, " must be a positive numeric scalar.", call. = FALSE)
  }
  obj
}



check_set_bool <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.logical(value) && length(value) == 1)) {
    obj[[varname]] <- value
  } else {
    stop(varname, " must be a boolean", call. = FALSE)
  }
  obj
}
check_set_chr <- function(obj, value) {
  varname <- as.character(substitute(value))
  if (is.na(value) || (is.character(value) && length(value) == 1)) {
    obj[[varname]] <- value
  } else {
    stop(varname, " must be a string", call. = FALSE)
  }
  obj
}

check_set_choice <- function(obj, value, choices) {
  varname <- as.character(substitute(value))

  if (is.na(value)) {
    obj[[varname]] <- value
  } else {
    if (is.character(value) && length(value) == 1) {
      if (!value %in% choices) {
        stop(varname, " must be one of ",
          paste(shQuote(choices), collapse = ", "),
          call. = FALSE
        )
      }
      obj[[varname]] <- value
    } else {
      stop(varname, " must be a character scalar.", call. = FALSE)
    }
  }

  obj
}

check_set_class <- function(obj, value, cl) {
  varname <- as.character(substitute(value))
  cl_str <- sprintf(" must be a %s object.", cl)
  if (!inherits(value, cl)) {
    stop(varname, cl_str, call. = FALSE)
  } else {
    obj[[varname]] <- value
  }
  obj
}

default_rpr <- data.frame(
  stringsAsFactors = FALSE,
  font.size = NA_integer_,
  bold = as.logical(NA),
  italic = as.logical(NA),
  underlined = as.logical(NA),
  color = NA_character_,
  font.family = NA_character_,
  bold.cs = as.logical(NA),
  font.size.cs = NA_integer_,
  vertical.align = NA_character_,
  shading.color = NA_character_,
  hansi.family = NA_character_,
  eastasia.family = NA_character_,
  cs.family = NA_character_,
  lang.val = NA_character_,
  lang.eastasia = NA_character_,
  lang.bidi = NA_character_
)

is_positive_numeric <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || (is.numeric(x) && length(x) == 1 && x >= 0)
  if (!test) {
    stop(varname, " must be a positive numeric scalar.", call. = FALSE)
  }
  test
}
is_bool <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || (is.logical(x) && length(x) == 1)
  if (!test) {
    stop(varname, " must be a boolean.", call. = FALSE)
  }
  test
}
is_color <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || is.color(x)
  if (!test) {
    stop(varname, " must be a valid color.", call. = FALSE)
  }
  test
}
is_character <- function(x) {
  varname <- as.character(substitute(x))
  test <- is.na(x) || (is.character(x) && length(x) == 1)
  if (!test) {
    stop(varname, " must be a string", call. = FALSE)
  }
  test
}
# fp_text ----
#' @title Text formatting properties
#'
#' @description Create a \code{fp_text} object that describes
#' text formatting properties.
#'
#' @param color font color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param font.size font size (in point) - 0 or positive integer value.
#' @param bold is bold
#' @param italic is italic
#' @param underlined is underlined
#' @param font.family single character value. Specifies the font to
#' be used to format characters in the Unicode range (U+0000-U+007F).
#' @param cs.family optional font to be used to format
#' characters in a complex script Unicode range. For example, Arabic
#' text might be displayed using the "Arial Unicode MS" font.
#' @param eastasia.family optional font to be used to
#' format characters in an East Asian Unicode range. For example,
#' Japanese text might be displayed using the "MS Mincho" font.
#' @param hansi.family optional. Specifies the font to be used to format
#' characters in a Unicode range which does not fall into one of the
#' other categories.
#' @param vertical.align single character value specifying font vertical alignments.
#' Expected value is one of the following : default \code{'baseline'}
#' or \code{'subscript'} or \code{'superscript'}
#' @param shading.color shading color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @return a \code{fp_text} object
#' @examples
#' fp_text()
#' fp_text(color = "red")
#' fp_text(bold = TRUE, shading.color = "yellow")
#' @family functions for defining formatting properties
#' @seealso [ftext], [fpar]
#' @export
fp_text <- function(color = "black", font.size = 10,
                    bold = FALSE, italic = FALSE, underlined = FALSE,
                    font.family = "Arial",
                    cs.family = NULL, eastasia.family = NULL, hansi.family = NULL,
                    vertical.align = "baseline",
                    shading.color = "transparent") {
  out <- default_rpr
  if (is_positive_numeric(font.size)) {
    out$font.size <- font.size
    out$font.size.cs <- font.size
  }
  if (is_bool(bold)) {
    out$bold <- bold
    out$bold.cs <- bold
  }
  if (is_bool(italic)) {
    out$italic <- italic
  }
  if (is_bool(underlined)) {
    out$underlined <- underlined
  }
  if (is_color(color)) {
    out$color <- color
  }
  if (is_character(font.family)) {
    out$font.family <- font.family
  }

  if (is.null(cs.family)) cs.family <- font.family
  if (is.null(eastasia.family)) eastasia.family <- font.family
  if (is.null(hansi.family)) hansi.family <- font.family
  if (is_character(cs.family)) {
    out$cs.family <- cs.family
  }
  if (is_character(eastasia.family)) {
    out$eastasia.family <- eastasia.family
  }
  if (is_character(hansi.family)) {
    out$hansi.family <- hansi.family
  }


  out <- check_set_choice(
    obj = out, value = vertical.align,
    choices = c("subscript", "superscript", "baseline")
  )
  out <- check_set_color(out, shading.color)

  class(out) <- "fp_text"

  out
}


#' @rdname fp_text
#' @description Function `fp_text_lite()` is generating properties
#' with only entries for the parameters users provided. The
#' undefined properties will inherit from the default settings.
#' @export
fp_text_lite <- function(
    color = NA, font.size = NA,
    font.family = NA, cs.family = NA, eastasia.family = NA, hansi.family = NA,
    bold = NA, italic = NA, underlined = NA,
    vertical.align = "baseline", shading.color = NA) {
  fp_text(
    color = color, font.size = font.size,
    bold = bold, italic = italic, underlined = underlined,
    font.family = font.family, cs.family = cs.family, eastasia.family = eastasia.family, hansi.family = hansi.family,
    vertical.align = vertical.align, shading.color = shading.color
  )
}

#' @rdname fp_text
#' @param format format type, wml for MS word, pml for
#' MS PowerPoint and html.
#' @param type output type - one of 'wml', 'pml', 'html', 'rtf'.
#' @export
format.fp_text <- function(x, type = "wml", ...) {
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("wml", "pml", "html", "rtf"))

  if (type == "wml") {
    rpr_wml(x)
  } else if (type == "pml") {
    rpr_pml(x)
  } else if (type == "html") {
    rpr_css(x)
  } else if (type == "rtf") {
    rpr_rtf(x)
  } else {
    stop("unimplemented type")
  }
}

#' @export
to_wml.fp_text <- function(x, add_ns = FALSE, ...) {
  format(x, type = "wml")
}

#' @param x \code{fp_text} object
#' @examples
#' print(fp_text(color = "red", font.size = 12))
#' @rdname fp_text
#' @export
print.fp_text <- function(x, ...) {
  out <- data.frame(
    font.size = as.double(x$font.size),
    italic = x$italic,
    bold = x$bold,
    underlined = x$underlined,
    color = x$color,
    shading = x$shading.color,
    fontname = x$font.family,
    fontname_cs = x$cs.family,
    fontname_eastasia = x$eastasia.family,
    fontname.hansi = x$hansi.family,
    vertical_align = x$vertical.align, stringsAsFactors = FALSE
  )
  print(out)
  invisible()
}


#' @param object \code{fp_text} object to modify
#' @param ... further arguments - not used
#' @rdname fp_text
#' @export
update.fp_text <- function(object, color, font.size,
                           bold, italic, underlined,
                           font.family, cs.family, eastasia.family, hansi.family,
                           vertical.align, shading.color, ...) {
  if (!missing(font.size)) {
    object <- check_set_numeric(obj = object, font.size)
  }
  if (!missing(bold)) {
    object <- check_set_bool(obj = object, bold)
  }
  if (!missing(italic)) {
    object <- check_set_bool(obj = object, italic)
  }
  if (!missing(underlined)) {
    object <- check_set_bool(obj = object, underlined)
  }
  if (!missing(color)) {
    object <- check_set_color(object, color)
  }
  if (!missing(font.family)) {
    object <- check_set_chr(object, font.family)
  }
  if (!missing(cs.family)) {
    object <- check_set_chr(object, cs.family)
  }
  if (!missing(eastasia.family)) {
    object <- check_set_chr(object, eastasia.family)
  }
  if (!missing(hansi.family)) {
    object <- check_set_chr(object, hansi.family)
  }
  if (!missing(vertical.align)) {
    object <- check_set_choice(
      obj = object, value = vertical.align,
      choices = c("subscript", "superscript", "baseline")
    )
  }
  if (!missing(shading.color)) {
    object <- check_set_color(object, shading.color)
  }

  object
}

# fp_border ----
border_styles <- c(
  "nil", "none", "solid", "single", "thick", "double", "dotted", "dashed",
  "dotDash", "dotDotDash", "triple", "thinThickSmallGap", "thickThinSmallGap",
  "thinThickThinSmallGap", "thinThickMediumGap", "thickThinMediumGap",
  "thinThickThinMediumGap", "thinThickLargeGap", "thickThinLargeGap",
  "thinThickThinLargeGap", "wave", "doubleWave", "dashSmallGap",
  "dashDotStroked", "threeDEmboss", "ridge", "threeDEngrave", "groove",
  "outset", "inset"
)

#' @title Border properties object
#'
#' @description create a border properties object.
#'
#' @param color border color - single character value (e.g. "#000000" or "black")
#' @param style border style - single character value : See Details for supported
#'  border styles.
#' @param width border width - an integer value : 0>= value
#' @details For Word output the following border styles are supported:
#'
#' * "none" or "nil" - No Border
#' * "solid" or "single" - Single Line Border
#' * "thick" - Single Line Border
#' * "double" - Double Line Border
#' * "dotted" - Dotted Line Border
#' * "dashed" - Dashed Line Border
#' * "dotDash" - Dot Dash Line Border
#' * "dotDotDash" - Dot Dot Dash Line Border
#' * "triple" - Triple Line Border
#' * "thinThickSmallGap" - Thin, Thick Line Border
#' * "thickThinSmallGap" - Thick, Thin Line Border
#' * "thinThickThinSmallGap" - Thin, Thick, Thin Line Border
#' * "thinThickMediumGap" - Thin, Thick Line Border
#' * "thickThinMediumGap" - Thick, Thin Line Border
#' * "thinThickThinMediumGap" - Thin, Thick, Thin Line Border
#' * "thinThickLargeGap" - Thin, Thick Line Border
#' * "thickThinLargeGap" - Thick, Thin Line Border
#' * "thinThickThinLargeGap" - Thin, Thick, Thin Line Border
#' * "wave" - Wavy Line Border
#' * "doubleWave" - Double Wave Line Border
#' * "dashSmallGap" - Dashed Line Border
#' * "dashDotStroked" - Dash Dot Strokes Line Border
#' * "threeDEmboss" or "ridge" - 3D Embossed Line Border
#' * "threeDEngrave" or "groove" - 3D Engraved Line Border
#' * "outset" - Outset Line Border
#' * "inset" - Inset Line Border
#'
#' For HTML output only a limited amount of border styles are supported:
#'
#' * "none" or "nil" - No Border
#' * "solid" or "single" - Single Line Border
#' * "double" - Double Line Border
#' * "dotted" - Dotted Line Border
#' * "dashed" - Dashed Line Border
#' * "threeDEmboss" or "ridge" - 3D Embossed Line Border
#' * "threeDEngrave" or "groove" - 3D Engraved Line Border
#' * "outset" - Outset Line Border
#' * "inset" - Inset Line Border
#'
#' Non-supported Word border styles will default to "solid".
#' @examples
#' fp_border()
#' fp_border(color = "orange", style = "solid", width = 1)
#' fp_border(color = "gray", style = "dotted", width = 1)
#' @export
#' @family functions for defining formatting properties
fp_border <- function(color = "black", style = "solid", width = 1) {
  out <- list()
  out <- check_set_numeric(obj = out, width)
  out <- check_set_color(out, color)
  out <- check_set_choice(
    obj = out, style,
    choices = border_styles
  )

  class(out) <- "fp_border"
  out
}

#' @param object fp_border object
#' @param ... further arguments - not used
#' @rdname fp_border
#' @examples
#'
#' # modify object ------
#' border <- fp_border()
#' update(border, style = "dotted", width = 3)
#' @export
update.fp_border <- function(object, color, style, width, ...) {
  if (!missing(color)) {
    object <- check_set_color(object, color)
  }

  if (!missing(width)) {
    object <- check_set_integer(obj = object, width)
  }

  if (!missing(style)) {
    object <- check_set_choice(obj = object, style, choices = border_styles)
  }

  object
}

#' @export
print.fp_border <- function(x, ...) {
  msg <- paste0("line: color: ", x$color, ", width: ", x$width, ", style: ", x$style, "\n")
  cat(msg)
  invisible()
}

# tabs ----
#' @export
#' @title Tabulation mark properties object
#'
#' @description create a tabulation mark properties setting object for Word
#' or RTF. Results can be used as arguments of [fp_tabs()].
#'
#' Once tabulation marks settings are defined, tabulation marks can
#' be added with [run_tab()] inside a call to [fpar()] or
#' with `\t` within 'flextable' content.
#'
#' @param pos Specifies the position of the tab stop (in inches).
#' @param style style of the tab. Possible values are:
#' "decimal", "left", "right" or "center".
#' @examples
#' fp_tab(pos = 0.4, style = "decimal")
#' fp_tab(pos = 1, style = "right")
#' @family functions for defining formatting properties
fp_tab <- function(pos, style = "decimal") {
  out <- list()
  out <- check_set_numeric(obj = out, pos)
  out <- check_set_choice(
    obj = out, style,
    choices = c(
      "decimal", "left", "right", "center"
    )
  )

  class(out) <- "fp_tab"
  out
}

#' @export
to_wml.fp_tab <- function(x, add_ns = FALSE, ...) {
  sprintf(
    "<w:tab w:val=\"%s\" w:leader=\"none\" w:pos=\"%.0f\"/>",
    x$style, inch_to_tweep(x$pos)
  )
}
#' @export
as.character.fp_tab <- function(x, ...) {
  paste(x$style, x$pos, sep = "_")
}


#' @export
#' @title Tabs properties object
#'
#' @description create a set of tabulation mark properties object for Word or RTF.
#' Results can be used as arguments `tabs` of [fp_par()] and will only have
#' effects in Word or RTF outputs.
#'
#' Once a set of tabulation marks settings is defined, tabulation marks can
#' be added with [run_tab()] inside a call to [fpar()] or
#' with `\t` within 'flextable' content.
#' @param ... [fp_tab] objects
#' @examples
#' z <- fp_tabs(
#'   fp_tab(pos = 0.4, style = "decimal"),
#'   fp_tab(pos = 1, style = "decimal")
#' )
#' fpar(
#'   run_tab(), ftext("88."),
#'   run_tab(), ftext("987.45"),
#'   fp_p = fp_par(
#'     tabs = z
#'   )
#' )
#' @family functions for defining formatting properties
fp_tabs <- function(...) {
  out <- list(...)
  if (length(out) > 0) {
    are_fp_tab <- sapply(out, inherits, "fp_tab")
    if (!all(are_fp_tab)) {
      stop("Function `fp_tabs()` only accept `fp_tab` objects as arguments.")
    }
  }

  class(out) <- "fp_tabs"
  out
}

#' @export
to_wml.fp_tabs <- function(x, add_ns = FALSE, ...) {
  if (length(x) > 0) {
    paste0("<w:tabs>", do.call(paste0, lapply(x, to_wml)), "</w:tabs>")
  } else {
    ""
  }
}

#' @export
as.character.fp_tabs <- function(x, ...) {
  if (length(x) > 0) {
    z <- lapply(x, as.character)
    z$sep <- "&"
    do.call(paste, z)
  } else {
    NA_character_
  }
}

# fp_par -----
#' @title Paragraph formatting properties
#'
#' @description Create a \code{fp_par} object that describes
#' paragraph formatting properties.
#'
#' @param text.align text alignment - a single character value, expected value
#' is one of 'left', 'right', 'center', 'justify'.
#' @param padding.bottom,padding.top,padding.left,padding.right paragraph paddings - 0 or positive integer value.
#' @param padding paragraph paddings - 0 or positive integer value. Argument \code{padding} overwrites
#' arguments \code{padding.bottom}, \code{padding.top}, \code{padding.left}, \code{padding.right}.
#' @param line_spacing line spacing, 1 is single line spacing, 2 is double line spacing.
#' @param border shortcut for all borders.
#' @param border.bottom,border.left,border.top,border.right \code{\link{fp_border}} for
#' borders. overwrite other border properties.
#' @param shading.color shading color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param keep_with_next a scalar logical. Specifies that the paragraph (or at least part of it) should be rendered
#' on the same page as the next paragraph when possible.
#' @param tabs NULL (default) for no tabulation marks setting
#' or an object returned by [fp_tabs()]. Note this can only have effect with Word
#' or RTF outputs.
#' @param word_style Word paragraph style name
#' @return a \code{fp_par} object
#' @examples
#' fp_par(text.align = "center", padding = 5)
#' @export
#' @family functions for defining formatting properties
#' @seealso [fpar]
fp_par <- function(text.align = "left",
                   padding = 0,
                   line_spacing = 1,
                   border = fp_border(width = 0),
                   padding.bottom, padding.top,
                   padding.left, padding.right,
                   border.bottom, border.left,
                   border.top, border.right,
                   shading.color = "transparent",
                   keep_with_next = FALSE,
                   tabs = NULL,
                   word_style = "Normal") {
  out <- list()

  out <- check_set_color(out, shading.color)
  out <- check_set_choice(
    obj = out, value = text.align,
    choices = c("left", "right", "center", "justify")
  )
  # padding checking
  out <- check_spread_integer(
    out, padding,
    c(
      "padding.bottom", "padding.top",
      "padding.left", "padding.right"
    )
  )
  if (!missing(padding.bottom)) {
    out <- check_set_numeric(obj = out, padding.bottom)
  }
  if (!missing(padding.left)) {
    out <- check_set_numeric(obj = out, padding.left)
  }
  if (!missing(padding.top)) {
    out <- check_set_numeric(obj = out, padding.top)
  }
  if (!missing(padding.right)) {
    out <- check_set_numeric(obj = out, padding.right)
  }

  out <- check_set_numeric(obj = out, line_spacing)

  # border checking
  out <- check_spread_border(
    obj = out, border,
    dest = c(
      "border.bottom", "border.top",
      "border.left", "border.right"
    )
  )

  if (!missing(border.top)) {
    out <- check_set_border(obj = out, border.top)
  }
  if (!missing(border.bottom)) {
    out <- check_set_border(obj = out, border.bottom)
  }
  if (!missing(border.left)) {
    out <- check_set_border(obj = out, border.left)
  }
  if (!missing(border.right)) {
    out <- check_set_border(obj = out, border.right)
  }

  if (!is.null(tabs)) {
    out[["tabs"]] <- tabs
  }

  out <- check_set_chr(obj = out, word_style)

  out$keep_with_next <- keep_with_next
  class(out) <- "fp_par"

  out
}

#' @export
#' @importFrom grDevices col2rgb
format.fp_par <- function(x, type = "wml", ...) {
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("wml", "pml", "html", "rtf"))

  if (type == "wml") {
    ppr_wml(x)
  } else if (type == "pml") {
    ppr_pml(x)
  } else if (type == "html") {
    ppr_css(x)
  } else if (type == "rtf") {
    ppr_rtf(x)
  } else {
    stop("unimplemented")
  }
}
#' @export
to_wml.fp_par <- function(x, add_ns = FALSE, ...) {
  format(x, type = "wml")
}

#' @param x,object \code{fp_par} object
#' @param ... further arguments - not used
#' @rdname fp_par
#' @export
print.fp_par <- function(x, ...) {
  out <- data.frame(
    text.align = as.character(x$text.align),
    padding.top = as.character(x$padding.top),
    padding.bottom = as.character(x$padding.bottom),
    padding.left = as.character(x$padding.left),
    padding.right = as.character(x$padding.right),
    shading.color = as.character(x$shading.color)
  )
  out <- as.data.frame(t(out))
  names(out) <- "values"
  print(out)
  cat("borders:\n")
  borders <- rbind(
    as.data.frame(unclass(x$border.top)),
    as.data.frame(unclass(x$border.bottom)),
    as.data.frame(unclass(x$border.left)),
    as.data.frame(unclass(x$border.right))
  )
  row.names(borders) <- c("top", "bottom", "left", "right")
  print(borders)
}


#' @rdname fp_par
#' @examples
#' obj <- fp_par(text.align = "center", padding = 1)
#' update(obj, padding.bottom = 5)
#' @export
update.fp_par <- function(object, text.align, padding, border,
                          padding.bottom, padding.top, padding.left, padding.right,
                          border.bottom, border.left, border.top, border.right,
                          shading.color, keep_with_next, word_style, ...) {
  if (!missing(text.align)) {
    object <- check_set_choice(
      obj = object, value = text.align,
      choices = c("left", "right", "center", "justify")
    )
  }
  if (!missing(word_style)) {
    object <- check_set_chr(obj = object, word_style)
  }

  # padding checking
  if (!missing(padding)) {
    object <- check_spread_integer(
      object, padding,
      c(
        "padding.bottom", "padding.top",
        "padding.left", "padding.right"
      )
    )
  }
  if (!missing(padding.bottom)) {
    object <- check_set_integer(obj = object, padding.bottom)
  }
  if (!missing(padding.left)) {
    object <- check_set_integer(obj = object, padding.left)
  }
  if (!missing(padding.top)) {
    object <- check_set_integer(obj = object, padding.top)
  }
  if (!missing(padding.right)) {
    object <- check_set_integer(obj = object, padding.right)
  }

  # border checking
  if (!missing(border)) {
    object <- check_spread_border(
      obj = object, border,
      dest = c(
        "border.bottom", "border.top",
        "border.left", "border.right"
      )
    )
  }
  if (!missing(border.top)) {
    object <- check_set_border(obj = object, border.top)
  }
  if (!missing(border.bottom)) {
    object <- check_set_border(obj = object, border.bottom)
  }
  if (!missing(border.left)) {
    object <- check_set_border(obj = object, border.left)
  }
  if (!missing(border.right)) {
    object <- check_set_border(obj = object, border.right)
  }

  if (!missing(shading.color)) {
    object <- check_set_color(object, shading.color)
  }
  if (!missing(keep_with_next)) {
    object <- check_set_bool(object, keep_with_next)
  }

  object
}

# fp_cell ----

vertical.align.styles <- c("top", "center", "bottom")
text.directions <- c("lrtb", "tbrl", "btlr")

#' @title Cell formatting properties
#'
#' @description Create a \code{fp_cell} object that describes cell formatting properties.
#'
#' @param border shortcut for all borders.
#' @param border.bottom,border.left,border.top,border.right \code{\link{fp_border}} for borders.
#' @param vertical.align cell content vertical alignment - a single character value,
#' expected value is one of "center" or "top" or "bottom"
#' @param margin shortcut for all margins.
#' @param margin.bottom,margin.top,margin.left,margin.right cell margins - 0 or positive integer value.
#' @param background.color cell background color - a single character value specifying a
#' valid color (e.g. "#000000" or "black").
#' @param text.direction cell text rotation - a single character value, expected
#' value is one of "lrtb", "tbrl", "btlr".
#' @param rowspan specify how many rows the cell is spanned over
#' @param colspan specify how many columns the cell is spanned over
#' @export
#' @family functions for defining formatting properties
fp_cell <- function(border = fp_border(width = 0),
                    border.bottom, border.left, border.top, border.right,
                    vertical.align = "center",
                    margin = 0,
                    margin.bottom, margin.top, margin.left, margin.right,
                    background.color = "transparent",
                    text.direction = "lrtb",
                    rowspan = 1,
                    colspan = 1) {
  out <- list()

  # border checking
  out <- check_spread_border(
    obj = out, border,
    dest = c(
      "border.bottom", "border.top",
      "border.left", "border.right"
    )
  )
  if (!missing(border.top)) {
    out <- check_set_border(obj = out, border.top)
  }
  if (!missing(border.bottom)) {
    out <- check_set_border(obj = out, border.bottom)
  }
  if (!missing(border.left)) {
    out <- check_set_border(obj = out, border.left)
  }
  if (!missing(border.right)) {
    out <- check_set_border(obj = out, border.right)
  }

  # background-color checking
  out <- check_set_color(out, background.color)

  out <- check_set_choice(
    obj = out, value = vertical.align,
    choices = vertical.align.styles
  )
  out <- check_set_choice(
    obj = out, value = text.direction,
    choices = text.directions
  )

  # margin checking
  out <- check_spread_integer(
    out, margin,
    c(
      "margin.bottom", "margin.top",
      "margin.left", "margin.right"
    )
  )

  if (!missing(margin.bottom)) {
    out <- check_set_integer(obj = out, margin.bottom)
  }
  if (!missing(margin.left)) {
    out <- check_set_integer(obj = out, margin.left)
  }
  if (!missing(margin.top)) {
    out <- check_set_integer(obj = out, margin.top)
  }
  if (!missing(margin.right)) {
    out <- check_set_integer(obj = out, margin.right)
  }

  out <- check_set_integer(obj = out, rowspan)
  out <- check_set_integer(obj = out, colspan)

  class(out) <- "fp_cell"
  out
}


#' @export
#' @rdname fp_cell
#' @param x,object \code{fp_cell} object
#' @param type output type - one of 'wml', 'pml', 'html', 'rtf'.
#' @param ... further arguments - not used
format.fp_cell <- function(x, type = "wml", ...) {
  stopifnot(length(type) == 1)
  stopifnot(type %in% c("wml", "pml", "html", "rtf"))

  if (type == "wml") {
    tcpr_wml(x)
  } else if (type == "pml") {
    tcpr_pml(x)
  } else if (type == "html") {
    tcpr_css(x)
  } else if (type == "rtf") {
    tcpr_rtf(x)
  } else {
    stop("unimplemented")
  }
}

#' @export
to_wml.fp_cell <- function(x, add_ns = FALSE, ...) {
  format(x, type = "wml")
}

#' @export
#' @rdname fp_cell
print.fp_cell <- function(x, ...) {
  cat(format(x, type = "html"))
}



#' @rdname fp_cell
#' @examples
#' obj <- fp_cell(margin = 1)
#' update(obj, margin.bottom = 5)
#' @export
update.fp_cell <- function(object, border,
                           border.bottom, border.left, border.top, border.right,
                           vertical.align, margin = 0,
                           margin.bottom, margin.top, margin.left, margin.right,
                           background.color,
                           text.direction,
                           rowspan = 1,
                           colspan = 1, ...) {
  if (!missing(border)) {
    object <- check_spread_border(
      obj = object, border,
      dest = c(
        "border.bottom", "border.top",
        "border.left", "border.right"
      )
    )
  }
  if (!missing(border.top)) {
    object <- check_set_border(obj = object, border.top)
  }
  if (!missing(border.bottom)) {
    object <- check_set_border(obj = object, border.bottom)
  }
  if (!missing(border.left)) {
    object <- check_set_border(obj = object, border.left)
  }
  if (!missing(border.right)) {
    object <- check_set_border(obj = object, border.right)
  }

  # background-color checking
  if (!missing(background.color)) {
    object <- check_set_color(object, background.color)
  }

  if (!missing(vertical.align)) {
    object <- check_set_choice(
      obj = object, value = vertical.align,
      choices = vertical.align.styles
    )
  }
  if (!missing(text.direction)) {
    object <- check_set_choice(
      obj = object, value = text.direction,
      choices = text.directions
    )
  }

  # margin checking
  if (!missing(margin)) {
    object <- check_spread_integer(
      object, margin,
      c(
        "margin.bottom", "margin.top",
        "margin.left", "margin.right"
      )
    )
  }

  if (!missing(margin.bottom)) {
    object <- check_set_integer(obj = object, margin.bottom)
  }
  if (!missing(margin.left)) {
    object <- check_set_integer(obj = object, margin.left)
  }
  if (!missing(margin.top)) {
    object <- check_set_integer(obj = object, margin.top)
  }
  if (!missing(margin.right)) {
    object <- check_set_integer(obj = object, margin.right)
  }

  if (!missing(rowspan)) {
    object <- check_set_integer(obj = object, rowspan)
  }
  if (!missing(colspan)) {
    object <- check_set_integer(obj = object, colspan)
  }

  object
}

xml_document_to_chrs <- function(xml_doc) {
  # Write the XML document to a temporary file
  tf <- tempfile(fileext = ".xml")
  write_xml(xml_doc, options = c("format"), file = tf)
  # Read the XML file as a character vector
  xml_str <- readLines(tf, warn = FALSE)
  xml_str <- gsub("^[[:blank:]]+", "", xml_str)
  # Prepend newlines to the XML string
  # where the opening tags are found
  m <- gregexpr("(?=<[^/])", xml_str, perl = TRUE)
  regmatches(xml_str, m) <- "\n"
  # The following can be replaced with gregexpr()
  # but there is minor performance gain with the following
  writeLines(xml_str, tf, useBytes = TRUE)
  xml_str <- readLines(tf, warn = FALSE)

  xml_str <- xml_str[xml_str != ""]
  xml_str
}

# replace body xml object with a new xml stored as a character vector
replace_xml_body_from_chr <- function(x, xml_str) {
  tf_xml <- tempfile(fileext = ".txt")
  writeLines(xml_str, tf_xml, useBytes = TRUE)
  x$doc_obj$replace_xml(tf_xml)
  x
}

#' @title Fix empty IDs in WML string
#' @description
#' This function corrects empty IDs found in a WordprocessingML (WML) string.
#' It ensures that all IDs are unique.
#' @param xml_str A character string containing WML with potentially empty embed IDs.
#' @return A character string (WML) where all empty IDs have been replaced
#' with unique IDs.
#' @noRd
fix_empty_ids_in_wml <- function(xml_str) {
  m <- regexpr(" id=\"\"", xml_str)
  id_values <- seq_len(sum(m > 0))
  regmatches(xml_str, m) <- sprintf(" id=\"%.0f\"", id_values)
  xml_str
}

#' @title Register images in relationships part of the Word package
#' @description
#' Ensures that each image referenced in the document is correctly declared in
#' the .rels file and assigned a unique relationship ID. Copies images if needed.
#' @param path_table data.frame containing a column named `path` that define the
#' image paths to be registered.
#' @param relationships A relationships object where the images will be registered.
#' @param package_dir The root directory of the unzipped Word package (used to store media).
#' @param media_dir The path within the package where image files are stored (default: "word/media").
#' @param media_rel_dir Relative path from the relationships part to the image (default: "media").
#' @noRd
register_images_in_relationships <- function(
  path_table,
  relationships,
  package_dir,
  media_dir = "word/media",
  media_rel_dir = "media"
) {
  path_table$id <- rep(NA_character_, nrow(path_table))
  for (i in seq_len(nrow(path_table))) {
    path <- path_table[[i, "path"]]
    dest_basename <- fake_newname(path)
    img_path <- file.path(package_dir, media_dir)
    if (!file.exists(file.path(img_path, dest_basename))) {
      dir.create(img_path, recursive = TRUE, showWarnings = FALSE)
      file.copy(from = path, to = file.path(img_path, dest_basename))
    }
    if (
      !file.path(media_rel_dir, dest_basename) %in%
        relationships$get_data()$target
    ) {
      rid <- sprintf("rId%.0f", relationships$get_next_id())
      relationships$add(
        id = rid,
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image",
        target = file.path(media_rel_dir, dest_basename)
      )
    } else {
      reldf <- relationships$get_data()
      rid <- reldf$id[basename(reldf$target) %in% dest_basename]
    }
    path_table$id[i] <- rid
  }
  path_table
}

#' @title Fix <a:blip> image references in WML
#' @description
#' This function replaces custom `<a:blip>` tags used by the officer package
#' with standard `<a:blip>` XML nodes, enabling proper image embedding
#' in a WordprocessingML (WML) string. Each embedded image is registered
#' in the relationships file using a unique `rId`. If this function is not executed,
#' the resulting Word document may be corrupted or display missing images.
#' @param xml_str A character string containing WML (WordprocessingML XML) where <a:blip> tags must be replaced.
#' @param doc_obj An `xml_document` object corresponding to the part being processed (usually the main document or a header/footer).
#' @param relationships A `relationship` object used to track and register media references.
#' @param package_dir The root directory of the unzipped Word package (used to store media files).
#' @param media_dir Path within the package where image files are stored. Defaults to `"word/media"`.
#' @param media_rel_dir Relative path from the `.rels` file to the image folder. Defaults to `"media"`.
#' @return A character string (WML) with all <a:blip> nodes replaced by valid <a:blip> image references using registered `rId`s.
#' @noRd
fix_img_refs_in_wml <- function(
  xml_str,
  doc_obj,
  relationships,
  package_dir,
  media_dir = "word/media",
  media_rel_dir = "media"
) {
  has_match <- grepl("<a:blip r:embed=\"[^\"]+\"", xml_str) &
    !grepl("<a:blip r:embed=\"rId[0-9]+\"", xml_str)

  if (!any(has_match)) {
    return(xml_str)
  }

  img_nodes_chr <- xml_str[has_match]

  path_values <- gsub(
    "<a:blip r:embed=\"([^\"]+)\".*",
    "\\1",
    img_nodes_chr
  )
  path_values <- unique(path_values)

  path_table <- data.frame(
    path = path_values,
    stringsAsFactors = FALSE
  )

  path_table <- register_images_in_relationships(
    path_table = path_table,
    relationships = relationships,
    package_dir = package_dir,
    media_dir = media_dir,
    media_rel_dir = media_rel_dir
  )
  for (i in seq_len(nrow(path_table))) {
    path <- path_table[[i, "path"]]
    rid <- path_table[[i, "id"]]
    xml_str <- gsub(
      sprintf("<a:blip r:embed=\"%s\"", path),
      sprintf("<a:blip r:embed=\"%s\"", rid),
      xml_str,
      fixed = TRUE
    )
  }
  xml_str
}


#' @title Fix <asvg:svgBlip> image references in WML
#' @description
#' This function replaces custom `<asvg:svgBlip>` tags used by the officer package
#' with standard `<asvg:svgBlip>` XML nodes, enabling proper SVG embedding
#' in a WordprocessingML (WML) string. Each embedded SVG is registered
#' in the relationships file using a unique `rId`. If this function is not executed,
#' the resulting Word document may be corrupted or display missing images.
#'
#' The function also ensures that fallback images are used when SVG rendering
#' is not supported by the target application (e.g., Word for Windows), again,
#' this is about setting the correct unique `rId`.
#' @param xml_str A character string containing WML (WordprocessingML XML) where <p:svg> tags must be replaced.
#' @param doc_obj An `xml_document` object corresponding to the part being processed (usually the main document or a header/footer).
#' @param relationships A `relationship` object used to track and register media references.
#' @param package_dir The root directory of the unzipped Word package (used to store media files).
#' @param media_dir Path within the package where image files are stored. Defaults to `"word/media"`.
#' @param media_rel_dir Relative path from the `.rels` file to the image folder. Defaults to `"media"`.
#' @return A character string (WML) with all <asvg:svgBlip> nodes replaced by valid <asvg:svgBlip> image references using registered `rId`s.
#' @noRd
fix_svg_refs_in_wml <- function(
  xml_str,
  doc_obj,
  relationships,
  package_dir,
  media_dir = "word/media",
  media_rel_dir = "media"
) {
  has_match <- grepl(
    "<asvg:svgBlip xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\" r:embed=\"[^\"]+\"/>",
    xml_str
  ) &
    !grepl(
      "<asvg:svgBlip xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\" r:embed=\"rId[0-9]+\"/>",
      xml_str
    )
  if (!any(has_match)) {
    return(xml_str)
  }

  img_nodes_chr <- xml_str[has_match]
  svg_path_values <- gsub(
    "<asvg:svgBlip xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\" r:embed=\"([^\"]+)\"/>",
    "\\1",
    img_nodes_chr
  )
  svg_path_values <- unique(svg_path_values)

  path_table <- data.frame(
    path = svg_path_values,
    stringsAsFactors = FALSE
  )

  path_table <- register_images_in_relationships(
    path_table = path_table,
    relationships = relationships,
    package_dir = package_dir,
    media_dir = media_dir,
    media_rel_dir = media_rel_dir
  )
  for (i in seq_len(nrow(path_table))) {
    path <- path_table[[i, "path"]]
    rid <- path_table[[i, "id"]]

    xml_str <- gsub(
      sprintf(
        "<asvg:svgBlip xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\" r:embed=\"%s\"/>",
        path
      ),
      sprintf(
        "<asvg:svgBlip xmlns:asvg=\"http://schemas.microsoft.com/office/drawing/2016/SVG/main\" r:embed=\"%s\"/>",
        rid
      ),
      xml_str,
      fixed = TRUE
    )
  }

  has_match <- grepl("<a:blip r:embed=\"[^\"]+\">", xml_str) &
    !grepl("<a:blip r:embed=\"rId[0-9]+\">", xml_str)
  img_nodes_chr <- xml_str[has_match]
  img_path_values <- gsub(
    "<a:blip r:embed=\"([^\"]+)\">",
    "\\1",
    img_nodes_chr
  )
  img_path_values <- unique(img_path_values)
  path_table <- data.frame(
    path = img_path_values,
    stringsAsFactors = FALSE
  )
  path_table <- register_images_in_relationships(
    path_table = path_table,
    relationships = relationships,
    package_dir = package_dir,
    media_dir = media_dir,
    media_rel_dir = media_rel_dir
  )
  for (i in seq_len(nrow(path_table))) {
    path <- path_table[[i, "path"]]
    rid <- path_table[[i, "id"]]
    xml_str <- gsub(
      sprintf("<a:blip r:embed=\"%s\">", path),
      sprintf("<a:blip r:embed=\"%s\">", rid),
      xml_str,
      fixed = TRUE
    )
  }
  xml_str
}

#' @title Fix <w:hyperlink> hyperlink references in WML
#' @description
#' This function replaces custom `<w:hyperlink>` tags (used internally by the officer package)
#' with standard OOXML `<w:hyperlink>` elements in a WordprocessingML (WML) string.
#' It ensures that all external hyperlinks are registered in the relationships file
#' with valid `rId` identifiers and updates the WML string accordingly.
#'
#' If this function is not executed, the resulting Word document may be corrupted
#' or contain invalid or non-functional hyperlinks.
#' @param xml_str A character string containing WML (WordprocessingML XML) with <w:hyperlink> tags to replace.
#' @param relationships A `relationship` object used to register each hyperlink target and assign unique `rId`s.
#' @return A character string (WML) with all <w:hyperlink> nodes replaced by valid <w:hyperlink> references using registered `rId`s.
#' @noRd
fix_hyperlink_refs_in_wml <- function(xml_str, doc_obj) {
  rel <- doc_obj$relationship()
  next_rel_id <- rel$get_next_id()

  has_match <- grepl("<w:hyperlink r:id=\"[^\"]+\"", xml_str) &
    !grepl("<w:hyperlink r:id=\"rId[0-9]+\"", xml_str)
  if (!any(has_match)) {
    return(xml_str)
  }

  hyperlink_nodes_chr <- xml_str[has_match]

  hyperlink_values <- gsub(
    ".*<w:hyperlink\\s+r:id=\"([^\"]+)\"[^>]*>.*",
    "\\1",
    hyperlink_nodes_chr
  )
  hyperlink_values <- unique(hyperlink_values)
  hyperlinks_table <- data.frame(
    url = hyperlink_values,
    id = seq(from = next_rel_id, along.with = hyperlink_values, by = 1L),
    stringsAsFactors = FALSE
  )

  for (i in seq_len(nrow(hyperlinks_table))) {
    url <- hyperlinks_table[[i, "url"]]
    rid <- sprintf("rId%.0f", hyperlinks_table[[i, "id"]])
    rel$add(
      id = rid,
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink",
      target = officer_url_decode(url),
      target_mode = "External"
    )
    xml_str <- gsub(
      sprintf("<w:hyperlink r:id=\"%s\"", url),
      sprintf("<w:hyperlink r:id=\"%s\"", rid),
      xml_str,
      fixed = TRUE
    )
  }
  xml_str
}

#' @title Convert custom style tags to OOXML style references
#' @description
#' This function parses a WordprocessingML (WML) string containing officer-specific
#' style tags and replaces them with standard OOXML-compliant `<w:pStyle>` nodes.
#'
#' This conversion is required for Word to correctly interpret and apply the styles
#' defined in the document's style part. Without this processing step, custom styles
#' will not be rendered or recognized.
#' @param xml_str A character string containing WML (WordprocessingML XML) with
#' officer custom style tags.
#' @param styles style table returned by `styles_info()`
#' @return A character string (WML) with all custom style tags replaced by standard OOXML style references.
#' @noRd
convert_custom_styles_in_wml <- function(xml_str, styles) {
  m <- gregexpr("w:(pstlname|tstlname)=\"([^\"]+)\"", xml_str)
  stylenames <- unique(unlist(regmatches(xml_str, m)))

  if (length(stylenames) < 1) {
    return(xml_str)
  }

  stylenames_types <- gsub(
    "w:(pstlname|tstlname)=\"([^\"]+)\"",
    "\\1",
    stylenames
  )
  stylenames <- gsub("w:(pstlname|tstlname)=\"([^\"]+)\"", "\\2", stylenames)

  if (!all(stylenames %in% styles$style_name)) {
    missing_styles <- paste0(
      shQuote(unique(setdiff(stylenames, styles$style_name))),
      collapse = ", "
    )
    stop("Some styles can not be found in the document: ", missing_styles)
  }

  styleids <- styles$style_id[match(stylenames, styles$style_name)]
  for (i in seq_along(stylenames)) {
    if ("pstlname" %in% stylenames_types[i]) {
      xml_str <- gsub(
        sprintf("w:pstlname=\"%s\"", stylenames[i]),
        sprintf("w:val=\"%s\"", styleids[i]),
        xml_str,
        fixed = TRUE
      )
    } else if ("tstlname" %in% stylenames_types[i]) {
      xml_str <- gsub(
        sprintf("w:tstlname=\"%s\"", stylenames[i]),
        sprintf("w:val=\"%s\"", styleids[i]),
        xml_str,
        fixed = TRUE
      )
    }
  }
  xml_str
}


#' @title Process footnote content written by officer
#' @description
#' This function processes a WordprocessingML (WML) string containing footnote content
#' written by officer. It ensures that all custom tags are converted into valid
#' OOXML equivalents by applying the appropriate fix/convert functions, i.e.
#' by copying the content of the footnote into the footnotes part of the document
#' and by setting the correct relationship ID (rId) for each footnote.
#'
#' This step is essential for ensuring that footnote content is compatible with
#' Word's rendering engine and avoids corruption of the document structure.
#'
#' @param x A character string containing WML representing footnote content.
#' @param footnotes An `xml_document` object representing the document footnotes being processed.
#' @param xml_str A character string containing WML (WordprocessingML XML) with
#' officer custom footnotes tags.
#' @return A character string (WML) converted into valid OOXML elements.
#' @noRd
process_footnotes_content <- function(x, footnotes, xml_str) {
  index <- length(xml_find_all(footnotes$get(), "w:footnote")) - 1

  footnotes_summary <- extract_footnotes(x, xml_str, index = index)
  if (nrow(footnotes_summary) < 1) {
    return(xml_str)
  }

  xml_str[footnotes_summary$reference_position] <- sprintf(
    xml_str[footnotes_summary$reference_position],
    footnotes_summary[["rid"]]
  )

  footnotes_summary$content_section <- sprintf(
    footnotes_summary$content_section,
    footnotes_summary$rid
  )

  relationships <- x$relationship()
  if (nrow(footnotes_summary) > 0) {
    next_id <- relationships$get_next_id()

    relationships$add(
      paste0("rId", next_id),
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/footnotes",
      target = "footnotes.xml"
    )
  }
  for (i in seq_len(nrow(footnotes_summary))) {
    rid <- footnotes_summary[[i, "rid"]]
    content_section <- footnotes_summary[[i, "content_section"]]

    newfootnote <- as_xml_document(content_section)
    xml_add_child(footnotes$get(), newfootnote)
  }

  # drop added sections contents from document xml
  lines_index_to_drop <- mapply(
    function(start, stop, str) {
      start:stop
    },
    start = footnotes_summary$start,
    stop = footnotes_summary$stop,
    SIMPLIFY = FALSE
  )
  lines_index_to_drop <- unlist(lines_index_to_drop)
  if (length(lines_index_to_drop) > 0) {
    xml_str <- xml_str[-lines_index_to_drop]
  }

  xml_str
}

#' @title Process comments content written by officer
#' @description
#' This function processes a WordprocessingML (WML) string containing comments content
#' written by officer. It ensures that all custom tags are converted into valid
#' OOXML equivalents by applying the appropriate fix/convert functions, i.e.
#' by copying the content of the comment into the comments part of the document
#' and by setting the correct relationship ID (rId) for each comment.
#'
#' This step is essential for ensuring that comments content is compatible with
#' Word's rendering engine and avoids corruption of the document structure.
#'
#' @param x A character string containing WML representing footnote content.
#' @param comments An `xml_document` object representing the document comments being processed.
#' @param xml_str A character string containing WML (WordprocessingML XML) with
#' officer custom comments tags.
#' @return A character string (WML) converted into valid OOXML elements.
#' @noRd
process_comments_content <- function(x, comments, xml_str) {
  index <- length(xml_find_all(comments$get(), "w:comment"))
  comments_summary <- extract_comments(x, xml_str, index)
  if (nrow(comments_summary) < 1) {
    return(xml_str)
  }
  relationships <- x$relationship()
  if (nrow(comments_summary) > 0) {
    next_id <- relationships$get_next_id()
    relationships$add(
      paste0("rId", next_id),
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/comments",
      target = "comments.xml"
    )
  }

  comments_summary$fake_id <- gsub(
    "<w:commentReference w:officer=\"true\" w:id=\"([^\"]+)\">",
    "\\1",
    xml_str[comments_summary$reference_position],
  )
  comments_summary$new_id <- seq(
    from = index,
    along.with = comments_summary$reference_position
  )
  comments_summary$new_id <- as.character(comments_summary$new_id)
  for (i in seq_len(nrow(comments_summary))) {
    content_section <- comments_summary[[i, "content_section"]]
    fake_id <- comments_summary[[i, "fake_id"]]
    new_id <- comments_summary[[i, "new_id"]]
    content_section <- gsub(
      fake_id,
      new_id,
      content_section,
      fixed = TRUE
    )
    newcomment <- as_xml_document(content_section)
    xml_add_child(comments$get(), newcomment)

    reference_position <- comments_summary[[i, "reference_position"]]
    range_start_position <- comments_summary[[i, "range_start_position"]]
    range_end_position <- comments_summary[[i, "range_end_position"]]
    xml_str[c(reference_position, range_start_position, range_end_position)] <-
      gsub(
        fake_id,
        new_id,
        xml_str[c(reference_position, range_start_position, range_end_position)]
      )
  }

  # drop added comments contents from document xml
  lines_index_to_drop <- mapply(
    function(start, stop, str) {
      start:stop
    },
    start = comments_summary$start,
    stop = comments_summary$stop,
    SIMPLIFY = FALSE
  )
  lines_index_to_drop <- unlist(lines_index_to_drop)
  if (length(lines_index_to_drop) > 0) {
    xml_str <- xml_str[-lines_index_to_drop]
  }

  xml_str
}

extract_comments <- function(x, xml_str, index) {
  # comments extraction
  starts <- grep("<w:comment ", xml_str, fixed = TRUE)
  stops <- grep("</w:comment>", xml_str, fixed = TRUE)

  reference_positions <- grep(
    "<w:commentReference w:officer=\"true\"",
    xml_str,
    fixed = FALSE
  )
  range_start_positions <- grep(
    "<w:commentRangeStart( xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\"){0,1} w:officer=\"true\"",
    xml_str,
    fixed = FALSE
  )
  range_end_positions <- grep(
    "<w:commentRangeEnd( xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\"){0,1} w:officer=\"true\"",
    xml_str,
    fixed = FALSE
  )

  starts <- starts[(starts - 1) %in% reference_positions]
  stops <- stops[(starts - 1) %in% reference_positions]

  content_chrs <- mapply(
    function(start, stop, str) {
      z <- str[start:stop]
      paste(z, collapse = "")
    },
    start = starts,
    stop = stops,
    MoreArgs = list(str = xml_str),
    SIMPLIFY = FALSE
  )
  content_chrs <- unlist(content_chrs) %||% character(0)

  ids <- sprintf(
    "%.0f",
    seq(from = index, along.with = content_chrs)
  )

  if (length(ids) < 1) {
    reference_positions <- integer(0)
    range_start_positions <- integer(0)
    range_end_positions <- integer(0)
  }

  comment_sections <- data.frame(
    content_section = content_chrs,
    id = ids,
    start = starts,
    stop = stops,
    reference_position = reference_positions,
    range_start_position = range_start_positions,
    range_end_position = range_end_positions,
    stringsAsFactors = FALSE
  )
  comment_sections
}
extract_footnotes <- function(x, xml_str, index) {
  # init next_rel_id that will be the first rid
  relationships <- x$relationship()
  next_rel_id <- relationships$get_next_id()

  # footers extraction
  starts <- grep("<w:footnote ", xml_str, fixed = TRUE)
  stops <- grep("</w:footnote>", xml_str, fixed = TRUE)

  reference_positions <- grep(
    "<w:footnoteReference w:officer=\"true\"",
    xml_str,
    fixed = TRUE
  )

  starts <- starts[(starts - 1) %in% reference_positions]
  stops <- stops[(starts - 1) %in% reference_positions]

  content_chrs <- mapply(
    function(start, stop, str) {
      z <- str[start:stop]
      paste(z, collapse = "")
    },
    start = starts,
    stop = stops,
    MoreArgs = list(str = xml_str),
    SIMPLIFY = FALSE
  )
  content_chrs <- unlist(content_chrs) %||% character(0)

  rids <- sprintf(
    "%.0f",
    seq(from = index, along.with = content_chrs)
  )

  if (length(rids) < 1) {
    reference_positions <- integer(0)
  }
  footnote_sections <- data.frame(
    content_section = content_chrs,
    rid = rids,
    start = starts,
    stop = stops,
    reference_position = reference_positions,
    stringsAsFactors = FALSE
  )
  footnote_sections
}

library(officer)

# Example 1: Creating a multi-column section ----
# This example demonstrates how to create a section with two columns,
# which is useful for newsletters, brochures, or academic papers

# Create sample text for demonstration
sample_text <- paste(
  "Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
  "Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
)
# Repeat text to fill columns
long_text <- paste(rep(sample_text, 5), collapse = " ")

# Need to use the same page margins each time
# a prop_section is created
page_margins <- page_mar()

# Create a new document
doc_1 <- read_docx()

# Add a title for the two-column section
doc_1 <- body_add_par(doc_1, "Newsletter Article", style = "heading 1")
doc_1 <- body_end_block_section(
  doc_1,
  block_section(
    prop_section(
      type = "continuous",
      page_margins = page_margins
    )
  )
)

# Define section properties for two columns
# - Two columns of equal width (3 inches each)
# - 0.5 inch spacing between columns
# - Continuous section (no page break)
two_column_section <- prop_section(
  type = "continuous",
  page_margins = page_margins,
  section_columns = section_columns(
    widths = c(3, 3),
    space = 0.5,
    sep = FALSE
  )
)

# Add content that will flow across two columns
doc_1 <- body_add_par(doc_1, "Introduction", style = "heading 2")
doc_1 <- body_add_fpar(
  doc_1,
  value = fpar(ftext(long_text), run_columnbreak()),
  style = "Normal"
)
doc_1 <- body_add_par(doc_1, "Main Content", style = "heading 2")
doc_1 <- body_add_par(doc_1, value = long_text, style = "Normal")

# End the two-column section
# Content after this point will return to single column layout
doc_1 <- body_end_block_section(doc_1, block_section(two_column_section))

# Add content in normal single-column layout
doc_1 <- body_add_par(doc_1, "Conclusion (Single Column)", style = "heading 1")
doc_1 <- body_add_par(doc_1, value = sample_text, style = "Normal")
default_sect_properties <- prop_section(
  type = "continuous",
  page_margins = page_margins
)
doc_1 <- body_set_default_section(doc_1, default_sect_properties)

# Save the document
output_file_1 <- tempfile(fileext = ".docx")
print(doc_1, target = output_file_1)


# Example 2: Section with custom header and footer ----
# This example shows how to create a specific section with its own
# header (containing R logo) and footer (with date), while other
# sections have no header/footer

# Get the path to R logo
img_path <- file.path(R.home("doc"), "html", "logo.jpg")

# Create header content with R logo aligned to the right
section_header <- block_list(
  fpar(
    external_img(src = img_path, height = 0.4, width = 0.4),
    ftext(
      " | Special Report Section",
      fp_text_lite(font.size = 10, color = "#4472C4")
    ),
    fp_p = fp_par(text.align = "right", padding.bottom = 6)
  )
)

# Create footer content with current date centered
section_footer <- block_list(
  fpar(
    ftext("Generated on: ", fp_text_lite(font.size = 9)),
    run_word_field(
      field = "Date \\@ \"MMMM d, yyyy\"",
      prop = fp_text_lite(font.size = 9, italic = TRUE)
    ),
    fp_p = fp_par(
      text.align = "center",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Define section properties with header and footer
# This section will have landscape orientation and custom margins
section_with_hf <- prop_section(
  page_size = page_size(orient = "landscape", width = 11, height = 8.5),
  page_margins = page_mar(
    bottom = 1,
    top = 1.2,
    right = 1,
    left = 1,
    header = 0.5,
    footer = 0.5
  ),
  type = "continuous",
  header_default = section_header,
  footer_default = section_footer
)

# Create a new document
doc_2 <- read_docx()

# Section 1: Normal section without header/footer
doc_2 <- body_add_par(doc_2, "Introduction", style = "heading 1")
doc_2 <- body_add_par(
  doc_2,
  "This is the introduction section with standard formatting."
)
doc_2 <- body_add_par(doc_2, value = sample_text, style = "Normal")

# Section 2: Special section WITH header and footer
# Add content for the special section
doc_2 <- body_add_par(doc_2, "Data Analysis Section", style = "heading 1")
doc_2 <- body_add_par(
  doc_2,
  "This section contains detailed analysis and is displayed in landscape format."
)

# Add a wide table that benefits from landscape orientation
analysis_data <- data.frame(
  Category = c("Sales", "Marketing", "R&D", "Operations"),
  Q1 = c(150, 85, 120, 95),
  Q2 = c(165, 92, 125, 102),
  Q3 = c(178, 88, 135, 108),
  Q4 = c(195, 98, 140, 115),
  Total = c(688, 363, 520, 420),
  Growth = c("13%", "7%", "8%", "10%")
)
doc_2 <- body_add_table(doc_2, value = analysis_data, style = "table_template")

# Add a chart
doc_2 <- body_add_par(doc_2, "Quarterly Trends", style = "heading 2")
trend_plot <- plot_instr({
  quarters <- paste0("Q", 1:4)
  sales <- c(150, 165, 178, 195)
  marketing <- c(85, 92, 88, 98)

  barplot(
    rbind(sales, marketing),
    beside = TRUE,
    names.arg = quarters,
    col = c("#4472C4", "#ED7D31"),
    border = NA,
    main = "Sales vs Marketing Budget",
    ylab = "Amount ($K)",
    xlab = "Quarter",
    legend.text = c("Sales", "Marketing"),
    args.legend = list(x = "topleft")
  )
})
doc_2 <- body_add_plot(doc_2, trend_plot, width = 7, height = 4)

# End the section with header/footer
# This applies the landscape orientation and header/footer to this section only
doc_2 <- body_end_block_section(doc_2, block_section(section_with_hf))

# Section 3: Return to normal section without header/footer
doc_2 <- body_add_par(doc_2, "Conclusions", style = "heading 1")
doc_2 <- body_add_par(
  doc_2,
  "This final section returns to portrait orientation without header or footer."
)
doc_2 <- body_add_par(doc_2, value = sample_text, style = "Normal")

# Save the document
output_file_2 <- tempfile(fileext = ".docx")
print(doc_2, target = output_file_2)

#' @export
#' @title Empty block for 'PowerPoint'
#' @description Create an empty object to include as an empty placeholder shape in a
#' presentation. This comes in handy when presentation are updated
#' through R, but a user still wants to add some comments in this new content.
#'
#' Empty content also works with layout fields (slide number and date) to preserve them:
#' they are included on the slide and keep being updated by PowerPoint, i.e. update to the
#' when the slide number when the slide moves in the deck, update to the date.
#'
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' doc <- read_pptx()
#' doc <- add_slide(doc, layout = "Two Content",
#'   master = "Office Theme")
#' doc <- ph_with(x = doc, value = empty_content(),
#'  location = ph_location_type(type = "title") )
#'
#' doc <- add_slide(doc)
#' # add slide number as a computer field
#' doc <- ph_with(
#'   x = doc, value = empty_content(),
#'   location = ph_location_type(type = "sldNum"))
#'
#' print(doc, target = fileout )
#' @seealso [ph_with()], [body_add_blocks()]
empty_content <- function(){
  x <- list()
  class(x) <- "empty_content"
  x
}

library(officer)

sp_line()
sp_line(color = "red", lwd = 2)
sp_line(lty = "dot", linecmpd = "dbl")
print(sp_line(color = "red", lwd = 2))
obj <- sp_line(color = "red", lwd = 2)
update(obj, linecmpd = "dbl")

#' @importFrom xml2 xml_children
read_custom_properties <- function(package_dir) {
  filename <- file.path(package_dir, "docProps/custom.xml")
  if (!file.exists(filename)) {
    filename <- system.file(package = "officer", "template/custom.xml")
  }
  doc <- read_xml(filename)
  all_children <- xml_children(doc)

  pid_values <- vapply(all_children, xml_attr, NA_character_, "pid")
  name_values <- vapply(all_children, xml_attr, NA_character_, "name")
  properties_types <- vapply(all_children, function(x) {
    xml_name(xml_child(x, 1))
  }, NA_character_)

  # this complicated thing is there because some unidentifed
  # softwares fill this fields with invalid xml - so we have
  # to consider it will not be correctly rewritten and will
  # make an invalid document
  value_values <- vapply(all_children, function(x) {
    as.character(xml_child(x, 1))
  }, NA_character_)
  pat1 <- sprintf("<vt\\:(%s)/>", paste0(properties_types, collapse = "|"))
  pat2 <- sprintf("<vt\\:(%s)>", paste0(properties_types, collapse = "|"))
  pat3 <- sprintf("</vt\\:(%s)>", paste0(properties_types, collapse = "|"))
  value_values <- gsub(pat1, "", value_values)
  value_values <- gsub(pat2, "", value_values)
  value_values <- gsub(pat3, "", value_values)
  str <- c(pid_values, name_values, properties_types, value_values)
  z <- matrix(str,
    ncol = 4,
    dimnames = list(NULL, c("pid", "name", "type", "value"))
  )
  z <- list(data = z)
  class(z) <- "custom_properties"
  z
}


#' @export
`[<-.custom_properties` <- function(x, i, j, value) {
  if (!i %in% x$data[, "name"]) {
    if (nrow(x$data) < 1) {
      pid <- 2
    } else {
      pid <- max(as.integer(x$data[, "pid"])) + 1L
    }
    new <- matrix(c(as.character(pid), i, "lpwstr", value), ncol = 4)
    x$data <- rbind(x$data, new)
  } else {
    x$data[x$data[, "name"] %in% i, j] <- value
  }
  x
}


#' @export
`[.custom_properties` <- function(x, i, j) {
  if (nrow(x$data) < 1) {
    return(character())
  }
  if (missing(i)) {
    x$data[, j]
  } else {
    x$data[x$data[, "name"] %in% i, j]
  }
}

write_custom_properties <- function(custom_props, package_dir) {
  xml_props <- sprintf(
    "<property fmtid=\"{D5CDD505-2E9C-101B-9397-08002B2CF9AE}\" pid=\"%s\" name=\"%s\"><vt:%s>%s</vt:%s></property>",
    custom_props$data[, 1],
    custom_props$data[, 2],
    custom_props$data[, 3],
    custom_props$data[, 4],
    custom_props$data[, 3]
  )

  xml_ <- c(
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
    "<Properties ",
    "xmlns=\"http://schemas.openxmlformats.org/officeDocument/2006/custom-properties\" ",
    "xmlns:vt=\"http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes\">",
    xml_props,
    "</Properties>"
  )
  props_dir <- file.path(package_dir, "docProps")
  dir.create(props_dir, recursive = TRUE, showWarnings = FALSE)
  filename <- file.path(props_dir, "custom.xml")
  writeLines(enc2utf8(xml_), filename, useBytes = TRUE)
  invisible()
}

# Adds notes master if not present. There can be at most one notesMaster in a presentation.
add_notesMaster <- function(x) {
  presentation_relations <- x$presentation$rel_df()
  if (!any(grepl("notesMaster$", presentation_relations$type))) {
    # copy notesMaster file from templates
    path <- system.file(package = "officer", "template/notesMaster.xml")
    xml_file <- file.path(x$package_dir, "ppt/notesMasters/notesMaster1.xml")
    dir.create(dirname(xml_file), showWarnings = FALSE, recursive = FALSE)
    file.copy(path, to = xml_file, overwrite = TRUE, copy.mode = FALSE)

    # copy theme file
    theme_file <- "theme1.xml"
    theme_files <- list.files(
      path = file.path(x$package_dir, "ppt/theme"),
      pattern = "\\.xml$",
      full.names = F
    )
    theme_index <- as.integer(gsub(
      "^(theme)([0-9]+)(\\.xml)$",
      "\\2",
      theme_files
    ))
    theme_file <- gsub(
      pattern = "[0-9]+",
      replacement = max(theme_index) + 1,
      theme_file
    )
    file.copy(
      file.path(x$package_dir, "ppt/theme/theme1.xml"),
      to = file.path(x$package_dir, "ppt/theme", theme_file),
      overwrite = TRUE,
      copy.mode = FALSE
    )

    # add relation to theme file
    newrel <- relationship$new()$add(
      id = "rId1",
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme",
      target = paste0("../theme/", theme_file)
    )
    newrel$write(
      path = file.path(
        dirname(xml_file),
        "_rels",
        paste0(basename(xml_file), ".rels")
      )
    )

    x$notesMaster$add_notesMaster(xml_file)

    # register xml files in content_type
    partname <- "/ppt/notesMasters/notesMaster1.xml"
    content_type <- setNames(
      "application/vnd.openxmlformats-officedocument.presentationml.notesMaster+xml",
      partname
    )
    x$content_type$add_override(content_type)

    partname <- paste0("/ppt/theme/", theme_file)
    content_type <- setNames(
      "application/vnd.openxmlformats-officedocument.theme+xml",
      partname
    )
    x$content_type$add_override(content_type)

    # add notesMaster relation to presentation
    prs_relations <- x$presentation$relationship()
    rId <- paste0("rId", prs_relations$get_next_id())
    prs_relations$add(
      id = rId,
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster",
      target = "notesMasters/notesMaster1.xml"
    )
    prs_relations$write(file.path(
      x$package_dir,
      "ppt/_rels",
      "presentation.xml.rels"
    ))

    xml_elt <- as_xml_document(paste0(
      "<p:notesMasterIdLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">
                                         <p:notesMasterId r:id=\"",
      rId,
      "\"/>
                                       </p:notesMasterIdLst>"
    ))
    xml_add_sibling(
      xml_find_first(x$presentation$get(), "//p:sldMasterIdLst"),
      xml_elt
    )
  }
  return(x)
}

# Create notesSlide for current slide if it doesn't exist.
# Returns name of notesSlide for current slide.
get_or_create_notesSlide <- function(x) {
  slide <- x$slide$get_slide(x$cursor)
  rels <- slide$rel_df()
  if (!any(grepl("notesSlide$", rels$type))) {
    # add notesMaster if it doesn't exist
    add_notesMaster(x)

    nslidename <- x$notesSlide$get_new_slidename()

    # copy slide template
    xml_file <- file.path(x$package_dir, "ppt/notesSlides", nslidename)
    path <- system.file(package = "officer", "template/notesSlide.xml")
    file.copy(path, to = xml_file, copy.mode = FALSE)

    # add relationships to notesMaster and slide
    rel_filename <- file.path(
      dirname(xml_file),
      "_rels",
      paste0(basename(xml_file), ".rels")
    )
    newrel <- relationship$new()$add(
      id = "rId1",
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster",
      target = "../notesMasters/notesMaster1.xml"
    )
    newrel$add(
      id = "rId2",
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide",
      target = file.path("../slides", slide$name())
    )
    newrel$write(path = rel_filename)

    # add relationship from slide to notesSlide
    rel <- slide$relationship()
    rel$add(
      id = paste0("rId", rel$get_next_id()),
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesSlide",
      target = file.path("../notesSlides", nslidename)
    )
    rel$write(file.path(
      dirname(slide$file_name()),
      "_rels",
      paste0(basename(slide$file_name()), ".rels")
    ))

    # update presentation elements
    x$content_type$add_notesSlide(
      partname = file.path("/ppt/notesSlides", nslidename)
    )

    # add new notesSlide to collection
    x$notesSlide$add_slide(xml_file)
  } else {
    nslidename <- basename(rels$target[grepl("notesSlide$", rels$type)])
  }
  return(nslidename)
}


#' @export
#' @title Location of a named placeholder for notes
#' @description The function will use the label of a placeholder
#' to find the corresponding location in the slide notes.
#' @param ph_label placeholder label of the used notes master
#' @param ... unused arguments
notes_location_label <- function(ph_label, ...) {
  x <- list(ph_label = ph_label)
  class(x) <- c("location_label", "location_str")
  x
}

#' @export
#' @title Location of a placeholder for notes
#' @description The function will use the type name of the placeholder (e.g. body, hdr),
#' to find the corresponding location.
#' @param type placeholder label of the used notes master
#' @param ... unused arguments
notes_location_type <- function(type = "body", ...) {
  ph_types <- c("ftr", "sldNum", "hdr", "body", "sldImg")
  if (!type %in% ph_types) {
    stop(
      "argument type ('",
      type,
      "') expected to be a value of ",
      paste0(shQuote(ph_types), collapse = ", "),
      "."
    )
  }
  x <- list(type = type)
  class(x) <- c("location_type", "location_str")
  x
}

ph_from_location <- function(loc, doc, ...) {
  UseMethod("ph_from_location", loc)
}


#' @export
ph_from_location.location_label <- function(loc, doc, ...) {
  xfrm <- doc$notesMaster$xfrm()
  location <- xfrm[xfrm$ph_label == loc$ph_label, ]
  if (nrow(location) < 1) {
    stop("No placeholder with label ", loc$ph_label, " found!")
  }

  id <- uuid_generate()
  str <- "<p:nvSpPr><p:cNvPr id=\"%s\" name=\"%s\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr>%s</p:nvPr></p:nvSpPr><p:spPr/>"
  new_ph <- sprintf(str, id, location$ph_label, location$ph)
  return(list(ph = new_ph, label = location$ph_label))
}


#' @export
ph_from_location.location_type <- function(loc, doc, ...) {
  xfrm <- doc$notesMaster$xfrm()
  location <- xfrm[xfrm$type == loc$type, ]
  if (nrow(location) < 1) {
    stop("No placeholder of type ", loc$type, " found!")
  }

  id <- uuid_generate()
  str <- "<p:nvSpPr><p:cNvPr id=\"%s\" name=\"%s\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr>%s</p:nvPr></p:nvSpPr><p:spPr/>"
  new_ph <- sprintf(str, id, location$ph_label[1], location$ph[1])
  return(list(ph = new_ph, label = location$ph_label[1]))
}


#' @export
#' @title Set notes for current slide
#' @description Set speaker notes for the current slide in a pptx presentation.
#' @param x an rpptx object
#' @param value text to be added to notes
#' @param location a placeholder location object.
#' It will be used to specify the location of the new shape. This location
#' can be defined with a call to one of the notes_ph functions. See
#' section "see also".
#' @param ... further arguments passed to or from other methods.
#' @examples
#' # this name will be used to print the file
#' # change it to "youfile.pptx" to write the pptx
#' # file in your working directory.
#' fileout <- tempfile(fileext = ".pptx")
#' fpt_blue_bold <- fp_text_lite(color = "#006699", bold = TRUE)
#' doc <- read_pptx()
#' # add a slide with some text ----
#' doc <- add_slide(doc, layout = "Title and Content")
#' doc <- ph_with(x = doc, value = "Slide Title 1",
#'    location = ph_location_type(type = "title") )
#' # set speaker notes for the slide ----
#' doc <- set_notes(doc, value = "This text will only be visible for the speaker.",
#'    location = notes_location_type("body"))
#'
#' # add a slide with some text ----
#' doc <- add_slide(doc, layout = "Title and Content")
#' doc <- ph_with(x = doc, value = "Slide Title 2",
#'    location = ph_location_type(type = "title") )
#' bl <- block_list(
#'   fpar(ftext("hello world", fpt_blue_bold)),
#'   fpar(ftext("Turlututu chapeau pointu", fpt_blue_bold))
#' )
#' doc <- set_notes(doc, value = bl,
#'    location = notes_location_type("body"))
#'
#' print(doc, target = fileout)
#'
#' @seealso [print.rpptx()], [read_pptx()], [add_slide()], [notes_location_label()], [notes_location_type()]
#' @family slide_manipulation
set_notes <- function(x, value, location, ...) {
  UseMethod("set_notes", value)
}


#' @export
#' @describeIn set_notes add a character vector to a place holder in the notes on the
#' current slide, values will be added as paragraphs.
set_notes.character <- function(x, value, location, ...) {
  # get or create notesSlide for current slide
  nslidename <- get_or_create_notesSlide(x)

  idx <- x$notesSlide$slide_index(nslidename)
  nSlide <- x$notesSlide$get_slide(idx)

  new_ph <- ph_from_location(location, x)

  # remove placeholder if already used
  xml_remove(xml_find_first(
    nSlide$get(),
    paste0("//p:spTree/p:sp[p:nvSpPr/p:cNvPr[@name='", new_ph$label, "']]")
  ))

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- as_xml_document(paste0(
    psp_ns_yes,
    new_ph$ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  ))

  xml_add_child(xml_find_first(nSlide$get(), "//p:spTree"), xml_elt)
  nSlide$fortify_id()

  return(x)
}


#' @export
#' @describeIn set_notes add a [block_list()] to a place holder in the notes on the
#' current slide.
set_notes.block_list <- function(x, value, location, ...) {
  # get or create notesSlide for current slide
  nslidename <- get_or_create_notesSlide(x)

  idx <- x$notesSlide$slide_index(nslidename)
  nSlide <- x$notesSlide$get_slide(idx)

  new_ph <- ph_from_location(location, x)

  # remove placeholder if already used
  xml_remove(xml_find_first(
    nSlide$get(),
    paste0("//p:spTree/p:sp[p:nvSpPr/p:cNvPr[@name='", new_ph$label, "']]")
  ))

  pars <- sapply(value, to_pml)
  pars <- paste0(pars, collapse = "")

  xml_elt <- as_xml_document(paste0(
    psp_ns_yes,
    new_ph$ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  ))

  xml_add_child(xml_find_first(nSlide$get(), "//p:spTree"), xml_elt)
  nSlide$fortify_id()

  return(x)
}

# worksheets ------------------------------------------------------------

worksheets <- R6Class(
  "worksheets",
  inherit = openxml_document,

  public = list(
    initialize = function(path) {
      super$initialize(character(0))
      private$package_dir <- path
      presentation_filename <- file.path(path, "xl", "workbook.xml")
      self$feed(presentation_filename)

      slide_df <- self$get_sheets_df()
      private$sheet_id <- slide_df$sheet_id
      private$sheet_rid <- slide_df$rid
      private$sheet_name <- slide_df$name
    },

    view_on_sheet = function(sheet) {
      sheet_id <- self$get_sheet_id(sheet)
      wb_view <- xml_find_first(self$get(), "d1:bookViews/d1:workbookView")
      xml_attr(wb_view, "activeTab") <- sheet_id - 1
      self$save()
    },

    add_sheet = function(target, label) {
      private$rels_doc$add(
        id = paste0("rId", private$rels_doc$get_next_id()),
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet",
        target = target
      )
      rels <- private$rels_doc$get_data()
      rid <- rels[rels$target %in% target, "id"]

      ids <- private$sheet_id
      new_id <- if (length(ids)) max(ids) + 1L else 1L

      private$sheet_id <- c(private$sheet_id, new_id)
      private$sheet_rid <- c(private$sheet_rid, rid)
      private$sheet_name <- c(private$sheet_name, label)

      children_ <- xml_children(self$get())
      sheets_id <- which(sapply(children_, function(x) xml_name(x) == "sheets"))
      xml_list <- xml_children(children_[[sheets_id]])

      xml_elt <- paste(
        sprintf(
          "<sheet name=\"%s\" sheetId=\"%.0f\" r:id=\"%s\"/>",
          htmlEscapeCopy(private$sheet_name),
          private$sheet_id,
          private$sheet_rid
        ),
        collapse = ""
      )
      xml_elt <- paste0(
        "<sheets xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
        xml_elt,
        "</sheets>"
      )
      xml_elt <- as_xml_document(xml_elt)

      if (!inherits(xml_list, "xml_missing")) {
        xml_replace(children_[[sheets_id]], xml_elt)
      } else {
        cli::cli_abort("Could not find sheets entity in workbook XML.")
      }

      self
    },

    get_new_sheetname = function() {
      sheet_dir <- file.path(private$package_dir, "xl/worksheets")
      if (!file.exists(sheet_dir)) {
        dir.create(
          file.path(sheet_dir, "_rels"),
          showWarnings = FALSE,
          recursive = TRUE
        )
      }

      sheet_files <- basename(list.files(sheet_dir, pattern = "\\.xml$"))
      sheet_name <- "sheet1.xml"
      if (length(sheet_files)) {
        slide_index <- as.integer(gsub(
          "^(sheet)([0-9]+)(\\.xml)$",
          "\\2",
          sheet_files
        ))
        sheet_name <- gsub(
          pattern = "[0-9]+",
          replacement = max(slide_index) + 1,
          sheet_name
        )
      }
      sheet_name
    },

    sheet_names = function() {
      private$sheet_name
    },

    get_sheet_id = function(name) {
      sheets_df <- self$get_sheets_df()
      bool_name_in_list <- sheets_df$name %in% name
      n_matches <- sum(bool_name_in_list, na.rm = TRUE)
      if (n_matches < 1) {
        cli::cli_abort("Could not find sheet {.val {name}}.")
      }
      sheets_df$sheet_id[bool_name_in_list]
    },

    get_sheets_df = function() {
      children_ <- xml_children(self$get())
      sheets_id <- which(sapply(children_, function(x) xml_name(x) == "sheets"))
      sheet_nodes <- xml_children(children_[[sheets_id]])
      data.frame(
        stringsAsFactors = FALSE,
        name = xml_attr(sheet_nodes, "name"),
        sheet_id = as.integer(xml_attr(sheet_nodes, "sheetId")),
        rid = xml_attr(sheet_nodes, "id")
      )
    },

    remove_sheet = function(label) {
      df <- self$get_sheets_df()
      idx <- which(df$name == label)
      if (!length(idx)) {
        cli::cli_abort("Sheet {.val {label}} not found.")
      }
      rid <- df$rid[idx]

      # drop relationship in workbook.xml.rels -- relationship$remove()
      # matches by basename(target), so look up target first
      rels_df <- private$rels_doc$get_data()
      target <- rels_df$target[rels_df$id == rid]
      if (length(target)) {
        private$rels_doc$remove(target[1])
      }

      # drop <sheet> entry in workbook.xml
      ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
      sheets_node <- xml_find_first(self$get(), "d1:sheets", ns = ns)
      if (inherits(sheets_node, "xml_missing")) {
        cli::cli_abort("No {.code <sheets>} element in workbook.xml.")
      }
      sheet_nodes <- xml_children(sheets_node)
      xml_remove(sheet_nodes[[idx]])

      # drop private state
      private$sheet_id <- private$sheet_id[-idx]
      private$sheet_rid <- private$sheet_rid[-idx]
      private$sheet_name <- private$sheet_name[-idx]

      self$save()
      self
    }
  ),
  private = list(
    sheet_id = NULL,
    sheet_rid = NULL,
    sheet_name = NULL,
    package_dir = NULL
  )
)

# sheet ------------------------------------------------------------
sheet <- R6Class(
  "sheet",
  inherit = openxml_document,
  public = list(
    feed = function(file) {
      private$filename <- file
      private$doc <- read_xml(file)

      private$rels_filename <- file.path(
        dirname(file),
        "_rels",
        paste0(basename(file), ".rels")
      )

      if (file.exists(private$rels_filename)) {
        private$rels_doc <- relationship$new()$feed_from_xml(
          private$rels_filename
        )
      } else {
        new_rel <- relationship$new()
        new_rel$write(private$rels_filename)
        private$rels_doc <- new_rel
      }
      self
    }
  )
)


# xlsx_drawing ----

#' @export
#' @title URIs for chart relationships in xlsx drawings
#' @description Returns the pair of URIs needed by
#' [xlsx_drawing] to embed a chart into a sheet, keeping the
#' relationship type and graphic URI consistent so Excel opens the
#' file without repair.
#' @param kind chart family. `"drawingml"` covers classic charts
#' (bar, line, pie, scatter, area, bubble, radar, stock).
#' `"chartex"` covers the newer family (boxplot, funnel, histogram,
#' pareto, sunburst, treemap, waterfall).
#' @return a named list with elements `rel_type` and `graphic_uri`.
#' @examples
#' ooxml_chart_uris("drawingml")
#' ooxml_chart_uris("chartex")
#' @keywords internal
ooxml_chart_uris <- function(kind = c("drawingml", "chartex")) {
  kind <- match.arg(kind)
  if (kind == "drawingml") {
    list(
      rel_type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/chart",
      graphic_uri = "http://schemas.openxmlformats.org/drawingml/2006/chart"
    )
  } else {
    list(
      rel_type = "http://schemas.microsoft.com/office/2014/relationships/chartEx",
      graphic_uri = "http://schemas.microsoft.com/office/drawing/2014/chartex"
    )
  }
}

# Parse an Excel cell reference (e.g. "C4", "AA12") into 0-based
# col/row indices used by OOXML <xdr:from>/<xdr:to>. Lowercase
# accepted. Errors on malformed input.
parse_cell_ref <- function(ref) {
  if (!is.character(ref) || length(ref) != 1L || is.na(ref) || !nzchar(ref)) {
    stop(
      "cell reference must be a single non-empty string like \"C4\"",
      call. = FALSE
    )
  }
  m <- regmatches(ref, regexec("^([A-Za-z]+)([0-9]+)$", ref))[[1]]
  if (length(m) != 3L) {
    stop("invalid cell reference: ", shQuote(ref), call. = FALSE)
  }
  letters_part <- toupper(m[2])
  row1 <- as.integer(m[3])
  if (row1 < 1L) {
    stop("invalid cell reference: ", shQuote(ref), call. = FALSE)
  }
  col1 <- 0L
  for (ch in strsplit(letters_part, "", fixed = TRUE)[[1]]) {
    col1 <- col1 * 26L + (utf8ToInt(ch) - utf8ToInt("A") + 1L)
  }
  list(col = col1 - 1L, row = row1 - 1L)
}

# Translate the user-facing `anchor` string of sheet_add_drawing methods
# into a (from, to) pair consumed by add_*_anchor R6 methods.
#   NULL       -> list(from = NULL, to = NULL)  (fixed-position)
#   "B2"       -> list(from = "B2", to = NULL)  (move only)
#   "B2:H20"   -> list(from = "B2", to = "H20") (move + size)
parse_anchor <- function(anchor) {
  if (is.null(anchor)) {
    return(list(from = NULL, to = NULL))
  }
  if (!is.character(anchor) || length(anchor) != 1L || !nzchar(anchor)) {
    stop(
      "`anchor` must be a single string like \"B2\" or \"B2:H20\"",
      call. = FALSE
    )
  }
  if (grepl(":", anchor, fixed = TRUE)) {
    parts <- strsplit(anchor, ":", fixed = TRUE)[[1]]
    if (length(parts) != 2L || !all(nzchar(parts))) {
      stop(
        "`anchor` range must be \"FROM:TO\" (e.g. \"B2:H20\")",
        call. = FALSE
      )
    }
    return(list(from = parts[1], to = parts[2]))
  }
  list(from = anchor, to = NULL)
}

# Wrap a drawing inner element (e.g. <xdr:graphicFrame>, <xdr:pic>) in
# one of the three OOXML anchor envelopes:
#   - <xdr:twoCellAnchor editAs="..."> when both `from` and `to` are
#     supplied (Excel default: "Move and size with cells")
#   - <xdr:oneCellAnchor> when only `from` is supplied (width/height
#     in inches define the size)
#   - <xdr:absoluteAnchor> otherwise (left/top/width/height in inches)
# `from`/`to` are Excel cell references like "B2", "H20".
build_anchor_xml <- function(
  inner_xml,
  from = NULL,
  to = NULL,
  edit_as = "twoCell",
  left = 1,
  top = 1,
  width = 6,
  height = 4
) {
  emu_per_in <- 914400
  ns_decl <- paste0(
    " xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\"",
    " xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"",
    " xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\""
  )
  if (!is.null(to) && is.null(from)) {
    stop("`to` requires `from` to be supplied as well", call. = FALSE)
  }
  if (!is.null(from) && !is.null(to)) {
    f <- parse_cell_ref(from)
    t <- parse_cell_ref(to)
    sprintf(
      paste0(
        "<xdr:twoCellAnchor editAs=\"%s\"",
        ns_decl,
        ">",
        "<xdr:from><xdr:col>%d</xdr:col><xdr:colOff>0</xdr:colOff>",
        "<xdr:row>%d</xdr:row><xdr:rowOff>0</xdr:rowOff></xdr:from>",
        "<xdr:to><xdr:col>%d</xdr:col><xdr:colOff>0</xdr:colOff>",
        "<xdr:row>%d</xdr:row><xdr:rowOff>0</xdr:rowOff></xdr:to>",
        "%s<xdr:clientData/></xdr:twoCellAnchor>"
      ),
      edit_as,
      f$col,
      f$row,
      t$col,
      t$row,
      inner_xml
    )
  } else if (!is.null(from)) {
    f <- parse_cell_ref(from)
    sprintf(
      paste0(
        "<xdr:oneCellAnchor",
        ns_decl,
        ">",
        "<xdr:from><xdr:col>%d</xdr:col><xdr:colOff>0</xdr:colOff>",
        "<xdr:row>%d</xdr:row><xdr:rowOff>0</xdr:rowOff></xdr:from>",
        "<xdr:ext cx=\"%.0f\" cy=\"%.0f\"/>",
        "%s<xdr:clientData/></xdr:oneCellAnchor>"
      ),
      f$col,
      f$row,
      width * emu_per_in,
      height * emu_per_in,
      inner_xml
    )
  } else {
    sprintf(
      paste0(
        "<xdr:absoluteAnchor",
        ns_decl,
        ">",
        "<xdr:pos x=\"%.0f\" y=\"%.0f\"/>",
        "<xdr:ext cx=\"%.0f\" cy=\"%.0f\"/>",
        "%s<xdr:clientData/></xdr:absoluteAnchor>"
      ),
      left * emu_per_in,
      top * emu_per_in,
      width * emu_per_in,
      height * emu_per_in,
      inner_xml
    )
  }
}

# Build the default <c:chart>/<cx:chart> fragment placed inside
# <a:graphicData> of a drawing anchor, dispatching on graphic_uri.
default_chart_inner <- function(graphic_uri, chart_rid) {
  is_chartex <- identical(
    graphic_uri,
    ooxml_chart_uris("chartex")$graphic_uri
  )
  sprintf(
    paste0(
      "<%1$s:chart",
      " xmlns:%1$s=\"%2$s\"",
      " xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\"",
      " r:id=\"%3$s\"/>"
    ),
    if (is_chartex) "cx" else "c",
    graphic_uri,
    chart_rid
  )
}

#' @export
#' @title Drawing manager for xlsx sheets
#' @description R6 class that manages a drawing file for an xlsx
#' sheet. Used internally by [sheet_add_drawing()] methods.
#' @param package_dir path to the unpacked xlsx directory
#' @param sheet_obj a sheet R6 object
#' @param content_type a content_type R6 object
#' @param chart_rid relationship id of the chart (returned by
#' `add_chart_rel()`). Ignored when `chart_inner` is supplied
#' explicitly.
#' @param chart_basename filename of the chart XML
#' @param left,top top-left position in inches (fixed-position mode)
#' @param width,height size in inches
#' @param from,to Excel cell references like `"C4"` (1-based,
#' case-insensitive)
#' @param edit_as how Excel should treat the drawing when rows or
#' columns are resized: `"twoCell"` (move and size with cells, the
#' default), `"oneCell"` (move only) or `"absolute"` (neither). Only
#' meaningful when both `from` and `to` are supplied.
#' @keywords internal
xlsx_drawing <- R6Class(
  "xlsx_drawing",
  inherit = openxml_document,

  public = list(
    #' @description Create or reuse a drawing for a sheet.
    initialize = function(package_dir, sheet_obj, content_type) {
      super$initialize("xl/drawings")
      private$package_dir <- package_dir

      drawings_dir <- file.path(package_dir, "xl", "drawings")
      dir.create(drawings_dir, recursive = TRUE, showWarnings = FALSE)
      rels_dir <- file.path(drawings_dir, "_rels")
      dir.create(rels_dir, recursive = TRUE, showWarnings = FALSE)

      # check if sheet already has a drawing
      sheet_rels <- sheet_obj$rel_df()
      drawing_rel <- sheet_rels[
        grepl("drawing$", sheet_rels$type),
        ,
        drop = FALSE
      ]

      if (nrow(drawing_rel) > 0) {
        # reuse existing drawing
        drawing_target <- drawing_rel$target[1L]
        drawing_file <- file.path(
          package_dir,
          "xl",
          "worksheets",
          drawing_target
        )
        self$feed(normalizePath(drawing_file, mustWork = TRUE))
      } else {
        # create new drawing
        drawing_name <- private$next_drawing_name(drawings_dir)
        drawing_file <- file.path(drawings_dir, drawing_name)

        drawing_xml <- paste0(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
          "<xdr:wsDr xmlns:xdr=\"http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing\"",
          " xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\"/>"
        )
        writeLines(drawing_xml, drawing_file, useBytes = TRUE)
        self$feed(drawing_file)

        # add relationship from sheet to drawing
        next_id <- sheet_obj$relationship()$get_next_id()
        rid <- paste0("rId", next_id)
        sheet_obj$relationship()$add(
          id = rid,
          type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/drawing",
          target = paste0("../drawings/", drawing_name)
        )

        # add <drawing r:id="rIdN"/> to sheet XML
        ns <- c(
          d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
        )
        existing_drawing <- xml_find_first(
          sheet_obj$get(),
          "d1:drawing",
          ns = ns
        )
        if (inherits(existing_drawing, "xml_missing")) {
          drawing_ref <- read_xml(sprintf(
            "<drawing xmlns=\"http://schemas.openxmlformats.org/spreadsheetml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:id=\"%s\"/>",
            rid
          ))
          # insert before extLst if present (OOXML element order)
          ext_lst <- xml_find_first(
            sheet_obj$get(),
            "d1:extLst",
            ns = ns
          )
          if (!inherits(ext_lst, "xml_missing")) {
            xml_add_sibling(ext_lst, drawing_ref, .where = "before")
          } else {
            xml_add_child(sheet_obj$get(), drawing_ref)
          }
        }
        sheet_obj$save()

        # content type
        partname <- paste0("/xl/drawings/", drawing_name)
        content_type$add_override(setNames(
          "application/vnd.openxmlformats-officedocument.drawing+xml",
          partname
        ))
      }
    },

    #' @description Place a chart on the sheet. The chart can be
    #'   positioned in three ways, matching Excel's own "Format
    #'   Picture" options:
    #'
    #'   * Move and size with cells (Excel's default): supply
    #'     `from` and `to` as cell references (e.g. `"B2"` and
    #'     `"H20"`). The chart spans those two cells and follows them
    #'     when rows/columns are inserted or resized.
    #'   * Move but don't size with cells: supply `from` only.
    #'     The chart anchors to that cell with the size set by
    #'     `width`/`height`.
    #'   * Fixed position: leave `from` and `to` `NULL`. The
    #'     chart is placed at `left`/`top` with size `width`/`height`,
    #'     independent of cell layout.
    #' @param graphic_uri chart family URI; defaults to the classic
    #'   chart family. Use [ooxml_chart_uris()] to obtain the value
    #'   for the chartEx family.
    #' @param chart_inner advanced. XML fragment referencing the chart
    #'   part. Leave `NULL` to let the method build the right reference
    #'   from `graphic_uri`.
    add_chart_anchor = function(
      chart_rid,
      left = 1,
      top = 1,
      width = 6,
      height = 4,
      from = NULL,
      to = NULL,
      edit_as = c("twoCell", "oneCell", "absolute"),
      graphic_uri = ooxml_chart_uris("drawingml")$graphic_uri,
      chart_inner = NULL
    ) {
      edit_as <- match.arg(edit_as)
      nv_id <- private$next_cNvPr_id()
      if (is.null(chart_inner)) {
        chart_inner <- default_chart_inner(graphic_uri, chart_rid)
      }
      frame_xml <- sprintf(
        paste0(
          "<xdr:graphicFrame macro=\"\">",
          "<xdr:nvGraphicFramePr>",
          "<xdr:cNvPr id=\"%d\" name=\"Chart %d\"/>",
          "<xdr:cNvGraphicFramePr/>",
          "</xdr:nvGraphicFramePr>",
          "<xdr:xfrm><a:off x=\"0\" y=\"0\"/><a:ext cx=\"0\" cy=\"0\"/></xdr:xfrm>",
          "<a:graphic><a:graphicData uri=\"%s\">%s</a:graphicData></a:graphic>",
          "</xdr:graphicFrame>"
        ),
        nv_id,
        nv_id,
        graphic_uri,
        chart_inner
      )
      anchor_xml <- build_anchor_xml(
        frame_xml,
        from = from,
        to = to,
        edit_as = edit_as,
        left = left,
        top = top,
        width = width,
        height = height
      )
      xml_add_child(self$get(), read_xml(anchor_xml))
      self$save()
      self
    },

    #' @description Register a chart part with the drawing. Returns
    #'   the relationship id, to be passed to `add_chart_anchor()`.
    #' @param rel_type chart family relationship type; defaults to the
    #'   classic chart family. Use [ooxml_chart_uris()] to obtain the
    #'   value for the chartEx family.
    add_chart_rel = function(
      chart_basename,
      rel_type = ooxml_chart_uris("drawingml")$rel_type
    ) {
      next_id <- self$relationship()$get_next_id()
      rid <- paste0("rId", next_id)
      self$relationship()$add(
        id = rid,
        type = rel_type,
        target = paste0("../charts/", chart_basename)
      )
      self$save()
      rid
    },

    #' @description Add a relationship from the drawing to a media image.
    #' @param image_basename filename (without directory) of the image
    #'   sitting in `xl/media/`.
    add_image_rel = function(image_basename) {
      next_id <- self$relationship()$get_next_id()
      rid <- paste0("rId", next_id)
      self$relationship()$add(
        id = rid,
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image",
        target = paste0("../media/", image_basename)
      )
      self$save()
      rid
    },

    #' @description Place an image on the sheet. Positioning works
    #'   the same way as `add_chart_anchor()`: supply `from` and `to`
    #'   to move and size with cells, `from` only to move (but not
    #'   resize) with cells, or leave both `NULL` for a fixed position.
    #' @param image_rid relationship id of the image (returned by
    #'   `add_image_rel()`)
    #' @param alt alternative text
    add_image_anchor = function(
      image_rid,
      left = 1,
      top = 1,
      width = 2,
      height = 2,
      from = NULL,
      to = NULL,
      edit_as = c("twoCell", "oneCell", "absolute"),
      alt = ""
    ) {
      edit_as <- match.arg(edit_as)
      nv_id <- private$next_cNvPr_id()
      pic_xml <- sprintf(
        paste0(
          "<xdr:pic>",
          "<xdr:nvPicPr>",
          "<xdr:cNvPr id=\"%d\" name=\"Picture %d\" descr=\"%s\"/>",
          "<xdr:cNvPicPr/>",
          "</xdr:nvPicPr>",
          "<xdr:blipFill>",
          "<a:blip r:embed=\"%s\"/>",
          "<a:stretch><a:fillRect/></a:stretch>",
          "</xdr:blipFill>",
          "<xdr:spPr>",
          "<a:xfrm><a:off x=\"0\" y=\"0\"/><a:ext cx=\"0\" cy=\"0\"/></a:xfrm>",
          "<a:prstGeom prst=\"rect\"><a:avLst/></a:prstGeom>",
          "</xdr:spPr>",
          "</xdr:pic>"
        ),
        nv_id,
        nv_id,
        htmlEscapeCopy(alt),
        image_rid
      )
      anchor_xml <- build_anchor_xml(
        pic_xml,
        from = from,
        to = to,
        edit_as = edit_as,
        left = left,
        top = top,
        width = width,
        height = height
      )
      xml_add_child(self$get(), read_xml(anchor_xml))
      self$save()
      self
    }
  ),

  private = list(
    package_dir = NULL,

    next_drawing_name = function(drawings_dir) {
      existing <- list.files(drawings_dir, pattern = "^drawing[0-9]+\\.xml$")
      if (length(existing) == 0L) {
        return("drawing1.xml")
      }
      nums <- as.integer(gsub("\\D", "", existing))
      paste0("drawing", max(nums) + 1L, ".xml")
    },

    next_cNvPr_id = function() {
      nodes <- xml_find_all(
        self$get(),
        "//xdr:cNvPr",
        ns = c(
          xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
        )
      )
      if (length(nodes) == 0L) {
        return(2L)
      }
      max(as.integer(xml_attr(nodes, "id"))) + 1L
    }
  )
)


# dir_sheets ----
dir_sheet <- R6Class(
  "dir_sheet",
  public = list(
    initialize = function(x) {
      private$package_dir <- x$package_dir
      private$sheets_path <- "xl/worksheets"
      self$update()
    },

    update = function() {
      dir_ <- file.path(private$package_dir, private$sheets_path)
      filenames <- list.files(
        path = dir_,
        pattern = "\\.xml$",
        full.names = TRUE
      )

      private$collection <- lapply(
        filenames,
        function(x, path) {
          sheet$new(path)$feed(x)
        },
        private$sheets_path
      )

      names(private$collection) <- sapply(private$collection, function(x) {
        x$name()
      })
      self
    },

    get_sheet_list = function() {
      dir_ <- file.path(private$package_dir, private$sheets_path)
      filenames <- list.files(
        path = dir_,
        pattern = "\\.xml$",
        full.names = TRUE
      )
      sheet_index <- seq_along(filenames)
      if (length(filenames)) {
        filenames <- basename(filenames)
        sheet_index <- as.integer(gsub(
          "^(sheet)([0-9]+)(\\.xml)$",
          "\\2",
          filenames
        ))
        filenames <- filenames[order(sheet_index)]
      }
      filenames
    },

    get_sheet = function(id) {
      l_ <- self$length()
      if (is.null(id) || !between(id, 1, l_)) {
        cli::cli_abort(
          "Invalid sheet id {.val {id}} ({l_} sheet(s) available)."
        )
      }
      sheet_files <- self$get_sheet_list()
      index <- which(names(private$collection) == sheet_files[id])
      private$collection[[index]]
    },

    length = function() {
      length(private$collection)
    }
  ),
  private = list(
    collection = NULL,
    package_dir = NULL,
    sheets_path = NULL
  )
)


# xlsx_styles ----
#' @export
#' @title Style manager for xlsx workbooks
#' @description R6 class that manages fonts, fills, borders, and
#' cell formats in an xlsx workbook's `styles.xml`. Used internally
#' by [sheet_write_data()] and available for extensions.
#' @param package_dir path to the unpacked xlsx directory
#' @param name font family name
#' @param size font size in points
#' @param bold logical, bold
#' @param italic logical, italic
#' @param underline logical, underline
#' @param color hex color string (6 chars, without `#`)
#' @param bg_color hex color string for fill background
#' @param top_style border style for top side
#' @param top_color border color for top side
#' @param bottom_style border style for bottom side
#' @param bottom_color border color for bottom side
#' @param left_style border style for left side
#' @param left_color border color for left side
#' @param right_style border style for right side
#' @param right_color border color for right side
#' @param font_id integer, font index (0-based)
#' @param fill_id integer, fill index (0-based)
#' @param border_id integer, border index (0-based)
#' @param num_fmt_id integer, number format id
#' @param halign horizontal alignment
#' @param valign vertical alignment
#' @param text_rotation text rotation angle (0-180)
#' @param wrap_text logical, enable text wrapping
#' @keywords internal
xlsx_styles <- R6Class(
  "xlsx_styles",

  public = list(
    #' @description Initialize styles from an xlsx package directory.
    initialize = function(package_dir) {
      private$file <- file.path(package_dir, "xl", "styles.xml")
      private$doc <- read_xml(private$file)
      private$ns <- c(
        d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
      )
      private$index_existing()
    },

    #' @description Get or create a font index.
    get_font_id = function(
      name = "Calibri",
      size = 11,
      bold = FALSE,
      italic = FALSE,
      underline = FALSE,
      color = "000000"
    ) {
      sig <- paste(
        name,
        size,
        as.integer(bold),
        as.integer(italic),
        as.integer(underline),
        color,
        sep = "|"
      )
      idx <- match(sig, private$font_sigs)
      if (!is.na(idx)) {
        return(idx - 1L)
      }

      xml_parts <- c(
        if (bold) "<b/>",
        if (italic) "<i/>",
        if (underline) "<u/>",
        sprintf("<sz val=\"%s\"/>", size),
        sprintf("<color rgb=\"FF%s\"/>", toupper(color)),
        sprintf("<name val=\"%s\"/>", htmlEscapeCopy(name))
      )
      font_xml <- sprintf(
        "<font xmlns=\"%s\">%s</font>",
        private$ssml,
        paste0(xml_parts, collapse = "")
      )
      private$add_to_collection("d1:fonts", font_xml, sig, "font_sigs")
    },

    #' @description Get or create a fill index.
    get_fill_id = function(bg_color = "FFFFFF") {
      sig <- toupper(bg_color)
      idx <- match(sig, private$fill_sigs)
      if (!is.na(idx)) {
        return(idx - 1L)
      }

      fill_xml <- sprintf(
        "<fill xmlns=\"%s\"><patternFill patternType=\"solid\"><fgColor rgb=\"FF%s\"/></patternFill></fill>",
        private$ssml,
        sig
      )
      private$add_to_collection("d1:fills", fill_xml, sig, "fill_sigs")
    },

    #' @description Get or create a border index.
    get_border_id = function(
      top_style = NULL,
      top_color = NULL,
      bottom_style = NULL,
      bottom_color = NULL,
      left_style = NULL,
      left_color = NULL,
      right_style = NULL,
      right_color = NULL
    ) {
      sig <- paste(
        top_style %||% "",
        top_color %||% "",
        bottom_style %||% "",
        bottom_color %||% "",
        left_style %||% "",
        left_color %||% "",
        right_style %||% "",
        right_color %||% "",
        sep = "|"
      )
      idx <- match(sig, private$border_sigs)
      if (!is.na(idx)) {
        return(idx - 1L)
      }

      make_side <- function(tag, style, color) {
        if (is.null(style) || style == "") {
          return(sprintf("<%s/>", tag))
        }
        sprintf(
          "<%s style=\"%s\"><color rgb=\"FF%s\"/></%s>",
          tag,
          style,
          toupper(color %||% "000000"),
          tag
        )
      }
      left_xml <- make_side("left", left_style, left_color)
      right_xml <- make_side("right", right_style, right_color)
      top_xml <- make_side("top", top_style, top_color)
      bottom_xml <- make_side("bottom", bottom_style, bottom_color)

      border_xml <- sprintf(
        "<border xmlns=\"%s\">%s%s%s%s<diagonal/></border>",
        private$ssml,
        left_xml,
        right_xml,
        top_xml,
        bottom_xml
      )
      private$add_to_collection("d1:borders", border_xml, sig, "border_sigs")
    },

    #' @description Get or create a cell format (xf) index.
    get_style_id = function(
      font_id = 0L,
      fill_id = 0L,
      border_id = 0L,
      num_fmt_id = 0L,
      halign = NA,
      valign = NA,
      text_rotation = 0L,
      wrap_text = FALSE
    ) {
      sig <- paste(
        font_id,
        fill_id,
        border_id,
        num_fmt_id,
        halign %||% "",
        valign %||% "",
        text_rotation,
        as.integer(wrap_text),
        sep = "|"
      )
      idx <- match(sig, private$xf_sigs)
      if (!is.na(idx)) {
        return(idx - 1L)
      }

      apply_parts <- c(
        if (font_id > 0L) " applyFont=\"1\"",
        if (fill_id > 0L) " applyFill=\"1\"",
        if (border_id > 0L) " applyBorder=\"1\"",
        if (num_fmt_id > 0L) " applyNumberFormat=\"1\""
      )

      has_align <- (!is.na(halign) && halign != "") ||
        (!is.na(valign) && valign != "") ||
        text_rotation > 0L ||
        wrap_text
      if (has_align) {
        apply_parts <- c(apply_parts, " applyAlignment=\"1\"")
      }

      align_xml <- ""
      if (has_align) {
        align_parts <- c(
          if (!is.na(halign) && halign != "") {
            sprintf(" horizontal=\"%s\"", halign)
          },
          if (!is.na(valign) && valign != "") {
            sprintf(" vertical=\"%s\"", valign)
          },
          if (text_rotation > 0L) {
            sprintf(" textRotation=\"%d\"", text_rotation)
          },
          if (wrap_text) " wrapText=\"1\""
        )
        align_xml <- sprintf(
          "<alignment%s/>",
          paste0(align_parts, collapse = "")
        )
      }

      xf_xml <- sprintf(
        "<xf xmlns=\"%s\" numFmtId=\"%d\" fontId=\"%d\" fillId=\"%d\" borderId=\"%d\" xfId=\"0\"%s>%s</xf>",
        private$ssml,
        num_fmt_id,
        font_id,
        fill_id,
        border_id,
        paste0(apply_parts, collapse = ""),
        align_xml
      )
      private$add_to_collection("d1:cellXfs", xf_xml, sig, "xf_sigs")
    },

    #' @description Get or create a cell format index for a number format.
    get_xf_id = function(num_fmt_id) {
      self$get_style_id(num_fmt_id = as.integer(num_fmt_id))
    },

    #' @description Save styles.xml to disk.
    save = function() {
      write_xml(private$doc, private$file)
    }
  ),

  private = list(
    file = NULL,
    doc = NULL,
    ns = NULL,
    ssml = "http://schemas.openxmlformats.org/spreadsheetml/2006/main",
    font_sigs = character(0),
    fill_sigs = character(0),
    border_sigs = character(0),
    xf_sigs = character(0),

    index_existing = function() {
      ns <- private$ns

      # fonts
      font_nodes <- xml_find_all(private$doc, "d1:fonts/d1:font", ns = ns)
      private$font_sigs <- vapply(
        font_nodes,
        function(node) {
          nm <- xml_text(xml_find_first(node, "d1:name/@val", ns = ns))
          sz <- xml_text(xml_find_first(node, "d1:sz/@val", ns = ns))
          b <- !inherits(xml_find_first(node, "d1:b", ns = ns), "xml_missing")
          i <- !inherits(xml_find_first(node, "d1:i", ns = ns), "xml_missing")
          u <- !inherits(xml_find_first(node, "d1:u", ns = ns), "xml_missing")
          col_node <- xml_find_first(node, "d1:color", ns = ns)
          col <- if (inherits(col_node, "xml_missing")) {
            "000000"
          } else {
            rgb <- xml_attr(col_node, "rgb")
            if (!is.na(rgb) && nchar(rgb) == 8) substring(rgb, 3) else "theme"
          }
          paste(
            nm,
            sz,
            as.integer(b),
            as.integer(i),
            as.integer(u),
            col,
            sep = "|"
          )
        },
        character(1L)
      )

      # fills
      fill_nodes <- xml_find_all(private$doc, "d1:fills/d1:fill", ns = ns)
      private$fill_sigs <- vapply(
        fill_nodes,
        function(node) {
          fg <- xml_find_first(node, "d1:patternFill/d1:fgColor", ns = ns)
          if (inherits(fg, "xml_missing")) {
            return("none")
          }
          rgb <- xml_attr(fg, "rgb")
          if (!is.na(rgb) && nchar(rgb) == 8) {
            toupper(substring(rgb, 3))
          } else {
            "none"
          }
        },
        character(1L)
      )

      # borders
      border_nodes <- xml_find_all(
        private$doc,
        "d1:borders/d1:border",
        ns = ns
      )
      private$border_sigs <- vapply(
        border_nodes,
        function(node) {
          get_side <- function(side) {
            s <- xml_find_first(node, paste0("d1:", side), ns = ns)
            style <- xml_attr(s, "style")
            col_node <- xml_find_first(s, "d1:color", ns = ns)
            col <- if (inherits(col_node, "xml_missing")) {
              ""
            } else {
              rgb <- xml_attr(col_node, "rgb")
              if (!is.na(rgb) && nchar(rgb) == 8) substring(rgb, 3) else ""
            }
            paste(if (is.na(style)) "" else style, col, sep = "|")
          }
          paste(
            get_side("top"),
            get_side("bottom"),
            get_side("left"),
            get_side("right"),
            sep = "|"
          )
        },
        character(1L)
      )

      # cellXfs
      xf_nodes <- xml_find_all(private$doc, "d1:cellXfs/d1:xf", ns = ns)
      private$xf_sigs <- vapply(
        xf_nodes,
        function(node) {
          fid <- xml_attr(node, "fontId") %||% "0"
          filid <- xml_attr(node, "fillId") %||% "0"
          bid <- xml_attr(node, "borderId") %||% "0"
          nfid <- xml_attr(node, "numFmtId") %||% "0"
          align <- xml_find_first(node, "d1:alignment", ns = ns)
          ha <- if (inherits(align, "xml_missing")) {
            ""
          } else {
            xml_attr(align, "horizontal") %||% ""
          }
          va <- if (inherits(align, "xml_missing")) {
            ""
          } else {
            xml_attr(align, "vertical") %||% ""
          }
          rot <- if (inherits(align, "xml_missing")) {
            "0"
          } else {
            xml_attr(align, "textRotation") %||% "0"
          }
          wrap <- if (inherits(align, "xml_missing")) {
            "0"
          } else {
            xml_attr(align, "wrapText") %||% "0"
          }
          paste(fid, filid, bid, nfid, ha, va, rot, wrap, sep = "|")
        },
        character(1L)
      )
    },

    add_to_collection = function(xpath, xml_str, sig, sig_field) {
      parent <- xml_find_first(private$doc, xpath, ns = private$ns)
      xml_add_child(parent, read_xml(xml_str))
      count <- length(private[[sig_field]]) + 1L
      xml_attr(parent, "count") <- as.character(count)
      private[[sig_field]] <- c(private[[sig_field]], sig)
      count - 1L # 0-based
    }
  )
)

# validation helpers ----

validate_sheet_name <- function(label) {
  if (is.null(label) || !nzchar(label)) {
    cli::cli_abort("Sheet name cannot be empty.")
  }
  if (nchar(label) > 31L) {
    cli::cli_abort(
      "Sheet name {.val {label}} exceeds 31 characters ({nchar(label)})."
    )
  }
  if (grepl("[\\]\\[*?/\\\\:]", label, perl = TRUE)) {
    cli::cli_abort(
      "Sheet name {.val {label}} contains invalid characters."
    )
  }
  if (grepl("^'|'$", label)) {
    cli::cli_abort(
      "Sheet name {.val {label}} cannot start or end with a single quote."
    )
  }
  invisible(label)
}

check_sheet_exists <- function(x, sheet) {
  available <- x$worksheets$sheet_names()
  if (!sheet %in% available) {
    cli::cli_abort(c(
      "Sheet {.val {sheet}} not found.",
      "i" = "Available sheets: {.val {available}}."
    ))
  }
  invisible(sheet)
}

# read_xlsx ----
#' @export
#' @title Create an 'Excel' document object
#' @description Read and import an xlsx file as an R object
#' representing the document. This function is experimental.
#' @param path path to the xlsx file to use as base document.
#' @param x an rxlsx object
#' @examples
#' read_xlsx()
read_xlsx <- function(path = NULL) {
  if (!is.null(path) && !file.exists(path)) {
    cli::cli_abort("Could not find file {.file {path}}.")
  }

  if (is.null(path)) {
    path <- system.file(package = "officer", "template/template.xlsx")
  }

  if (!grepl("\\.xlsx$", path, ignore.case = TRUE)) {
    cli::cli_abort("{.fun read_xlsx} only supports {.file .xlsx} files.")
  }

  package_dir <- tempfile()
  unpack_folder(file = path, folder = package_dir)

  obj <- structure(
    list(package_dir = package_dir),
    .Names = c("package_dir"),
    class = "rxlsx"
  )

  obj$content_type <- content_type$new(package_dir)
  obj$worksheets <- worksheets$new(package_dir)
  obj$sheets <- dir_sheet$new(obj)
  obj$styles <- xlsx_styles$new(package_dir)
  obj$core_properties <- read_core_properties(obj$package_dir)

  obj
}

#' @export
#' @title Add a sheet
#' @description Add a sheet into an xlsx worksheet.
#' @param x rxlsx object
#' @param label sheet label
#' @details
#' `read_xlsx()` returns a workbook that already contains one default
#' sheet shipped with the template (named `"Sheet1"` or `"Feuil1"`
#' depending on the locale). `add_sheet()` is purely additive: the
#' default sheet is kept as-is. Remove it explicitly with
#' [sheet_remove()] if it is not wanted.
#' @examples
#' my_ws <- read_xlsx()
#' my_pres <- add_sheet(my_ws, label = "new sheet")
add_sheet <- function(x, label) {
  validate_sheet_name(label)
  if (label %in% x$worksheets$sheet_names()) {
    cli::cli_abort("Sheet {.val {label}} already exists.")
  }

  new_slidename <- x$worksheets$get_new_sheetname()

  xml_file <- file.path(x$package_dir, "xl/worksheets", new_slidename)

  template_file <- system.file(package = "officer", "template/sheet.xml")
  file.copy(template_file, xml_file, copy.mode = FALSE)

  rel_filename <- file.path(
    dirname(xml_file),
    "_rels",
    paste0(basename(xml_file), ".rels")
  )
  dir.create(dirname(rel_filename), showWarnings = FALSE)
  template_rel_file <- system.file(
    package = "officer",
    "template/sheet.xml.rels"
  )
  file.copy(template_rel_file, rel_filename, copy.mode = FALSE)

  # update presentation elements
  x$worksheets$add_sheet(
    target = file.path("worksheets", new_slidename),
    label = label
  )

  partname <- file.path("/xl/worksheets", new_slidename)
  override <- setNames(
    "application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml",
    partname
  )
  x$content_type$add_override(value = override)

  x$sheets$update()

  sheet_select(x, sheet = label)
}

#' @export
#' @title Remove a sheet
#' @description Remove a sheet from an xlsx workbook, deleting the
#' worksheet XML, its relationship file and the content-type override.
#' @param x rxlsx object
#' @param sheet name of the sheet to remove
#' @return the rxlsx object (invisibly)
#' @examples
#' wb <- read_xlsx()
#' wb <- add_sheet(wb, "kept")
#' default_name <- sheet_names(wb)[1]
#' wb <- sheet_remove(wb, sheet = default_name)
#' print(wb, target = tempfile(fileext = ".xlsx"))
sheet_remove <- function(x, sheet) {
  stopifnot(inherits(x, "rxlsx"))
  check_sheet_exists(x, sheet)

  df <- x$worksheets$get_sheets_df()
  idx <- which(df$name == sheet)
  rid <- df$rid[idx]

  # find the sheet XML filename from workbook rels via the rel target
  sheet_files <- x$sheets$get_sheet_list()
  # target in workbook rels looks like "worksheets/sheetN.xml"
  # we derive the basename from the private state carried by worksheets
  sheet_target <- paste0("worksheets/", sheet_files)
  # match rid to position in private state
  pos <- match(rid, df$rid)
  sheet_basename <- sheet_files[pos]

  # remove files on disk
  xml_path <- file.path(x$package_dir, "xl/worksheets", sheet_basename)
  rels_path <- file.path(
    x$package_dir,
    "xl/worksheets/_rels",
    paste0(sheet_basename, ".rels")
  )
  if (file.exists(xml_path)) {
    file.remove(xml_path)
  }
  if (file.exists(rels_path)) {
    file.remove(rels_path)
  }

  # remove content-type override
  partname <- paste0("/xl/worksheets/", sheet_basename)
  x$content_type$remove_slide(partname)

  # update workbook XML: <sheet> entry, rels, private state
  x$worksheets$remove_sheet(sheet)

  # refresh the sheet collection
  x$sheets$update()

  invisible(x)
}

#' @export
#' @rdname read_xlsx
length.rxlsx <- function(x) {
  x$sheets$length()
}

#' @export
#' @title Sheet names of an xlsx workbook
#' @description Return the sheet names of an `rxlsx` object, in the
#' order they appear in the workbook.
#' @param x an rxlsx object (created by [read_xlsx()]).
#' @return A character vector of sheet names.
#' @examples
#' wb <- read_xlsx()
#' sheet_names(wb)
#'
#' wb <- add_sheet(wb, label = "new sheet")
#' sheet_names(wb)
sheet_names <- function(x) {
  stopifnot(inherits(x, "rxlsx"))
  x$worksheets$sheet_names()
}

#' @export
#' @title Select sheet
#' @description Set a particular sheet selected when workbook will be
#' edited.
#' @param x rxlsx object
#' @param sheet sheet name
#' @examples
#' my_ws <- read_xlsx()
#' my_pres <- add_sheet(my_ws, label = "new sheet")
#' my_pres <- sheet_select(my_ws, sheet = "new sheet")
#' print(my_ws, target = tempfile(fileext = ".xlsx") )
sheet_select <- function(x, sheet) {
  check_sheet_exists(x, sheet)
  x$worksheets$view_on_sheet(sheet)
  active_index <- which(x$worksheets$sheet_names() == sheet)
  ns <- "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
  for (i in seq_len(x$sheets$length())) {
    sheet_obj <- x$sheets$get_sheet(i)
    sv <- xml_find_first(
      sheet_obj$get(),
      "d1:sheetViews/d1:sheetView",
      ns = c(d1 = ns)
    )
    if (!inherits(sv, "xml_missing")) {
      if (i == active_index) {
        xml_attr(sv, "tabSelected") <- "1"
      } else {
        xml_attr(sv, "tabSelected") <- "0"
      }
    }
    sheet_obj$save()
  }
  x
}

#' @export
#' @title Write data to a sheet
#' @description Write a content into a sheet of an xlsx workbook.
#' Multiple calls can write to different positions on the same sheet.
#'
#' This is a generic function dispatching on `value`. Supported inputs:
#' - `data.frame`: written as a table (header in row `start_row`, data
#'   starting at `start_row + 1`).
#' - `character`: each element in its own cell. The `character` method
#'   accepts a `direction` argument (`"vertical"` -- default -- or
#'   `"horizontal"`) to stack elements in a column or a row.
#' - [fpar()]: richtext paragraph written into a single cell at
#'   `(start_row, start_col)`. Font, size, colour, bold, italic,
#'   underline, strikethrough and sub/superscript chunks are honoured.
#' - [block_list()]: one cell per `fpar` item. Also accepts
#'   `direction = "vertical"` (default) or `"horizontal"`.
#'
#' @param x rxlsx object
#' @param value a `data.frame`, `character` vector, [fpar()] or
#' [block_list()]
#' @param sheet sheet name (must already exist)
#' @param start_row row index where the header / first cell will be
#' written (default 1)
#' @param start_col column index where the first column / first cell
#' will be written (default 1)
#' @param ... method-specific arguments. In particular,
#' `direction = "vertical"` (default) or `"horizontal"` is honoured by
#' the `character` and `block_list` methods.
#' @example inst/examples/example-sheet_write_data.R
sheet_write_data <- function(
  x,
  value,
  sheet,
  start_row = 1L,
  start_col = 1L,
  ...
) {
  UseMethod("sheet_write_data", value)
}

#' @export
sheet_write_data.default <- function(x, value, sheet, ...) {
  cli::cli_abort(c(
    "No {.fun sheet_write_data} method for an object of class
     {.cls {class(value)[1]}}.",
    "i" = "Supported inputs: {.cls data.frame}, {.cls character},
           {.cls fpar}, {.cls block_list}."
  ))
}

#' @export
sheet_write_data.data.frame <- function(
  x,
  value,
  sheet,
  start_row = 1L,
  start_col = 1L,
  ...
) {
  stopifnot(inherits(x, "rxlsx"))
  check_sheet_exists(x, sheet)
  start_row <- as.integer(start_row)
  start_col <- as.integer(start_col)

  # resolve style IDs for date/datetime columns
  date_xf <- NULL
  datetime_xf <- NULL
  for (j in seq_len(ncol(value))) {
    col <- value[[j]]
    if (inherits(col, "POSIXct") && is.null(datetime_xf)) {
      datetime_xf <- x$styles$get_xf_id(22L)
    } else if (inherits(col, "Date") && is.null(date_xf)) {
      date_xf <- x$styles$get_xf_id(14L)
    }
  }

  new_cells <- df_to_cells(
    value,
    start_row,
    start_col,
    date_xf = date_xf,
    datetime_xf = datetime_xf
  )

  write_cells_into_sheet(x, sheet, new_cells)
}

# Shared plumbing: merge a named list of row XML strings (keyed by row
# number as character) into the target sheet's <sheetData>.
write_cells_into_sheet <- function(x, sheet, new_cells) {
  sheet_obj <- x$sheets$get_sheet(
    which(x$worksheets$sheet_names() == sheet)
  )
  sheet_doc <- sheet_obj$get()
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  existing_rows <- xml_find_all(sheet_doc, "d1:sheetData/d1:row", ns = ns)

  if (length(existing_rows) == 0L) {
    # fast path: no existing data, skip merge
    row_nums <- as.integer(names(new_cells))
    idx <- order(row_nums)
    row_strs <- sprintf(
      "<row r=\"%d\">%s</row>",
      row_nums[idx],
      unlist(new_cells[idx])
    )
  } else {
    # read existing rows
    existing_map <- list()
    for (row_node in existing_rows) {
      r <- xml_attr(row_node, "r")
      cell_nodes <- xml_find_all(row_node, "d1:c", ns = ns)
      cell_strs <- vapply(cell_nodes, as.character, character(1L))
      existing_map[[r]] <- paste0(cell_strs, collapse = "")
    }

    # merge: new rows overwrite or extend existing rows
    for (r in names(new_cells)) {
      if (is.null(existing_map[[r]])) {
        existing_map[[r]] <- new_cells[[r]]
      } else {
        old_doc <- read_xml(paste0("<row>", existing_map[[r]], "</row>"))
        new_doc <- read_xml(paste0("<row>", new_cells[[r]], "</row>"))
        old_cells <- xml_find_all(old_doc, "c")
        new_cells_nodes <- xml_find_all(new_doc, "c")
        old_refs <- xml_attr(old_cells, "r")
        new_refs <- xml_attr(new_cells_nodes, "r")
        keep <- !old_refs %in% new_refs
        parts <- c(
          vapply(old_cells[keep], as.character, character(1L)),
          vapply(new_cells_nodes, as.character, character(1L))
        )
        existing_map[[r]] <- paste0(parts, collapse = "")
      }
    }

    row_nums <- as.integer(names(existing_map))
    existing_map <- existing_map[order(row_nums)]
    row_nums <- sort(row_nums)
    row_strs <- sprintf(
      "<row r=\"%d\">%s</row>",
      row_nums,
      unlist(existing_map)
    )
  }

  # update sheetData in the existing DOM (preserves namespaces, etc.)
  sheet_data_node <- xml_find_first(sheet_doc, "d1:sheetData", ns = ns)
  new_sheet_data <- read_xml(paste0(
    "<sheetData xmlns=\"http://schemas.openxmlformats.org/spreadsheetml/2006/main\">",
    paste0(row_strs, collapse = ""),
    "</sheetData>"
  ))
  xml_replace(sheet_data_node, new_sheet_data)

  # remove dimension (Excel recalculates it)
  dim_node <- xml_find_first(sheet_doc, "d1:dimension", ns = ns)
  if (!inherits(dim_node, "xml_missing")) {
    xml_remove(dim_node)
  }

  sheet_file <- sheet_obj$file_name()
  write_xml(sheet_doc, sheet_file)
  sheet_obj$feed(sheet_file)

  x
}

# Convert a data.frame to a named list of cell XML strings, keyed by row number
df_to_cells <- function(
  data,
  start_row = 1L,
  start_col = 1L,
  date_xf = NULL,
  datetime_xf = NULL
) {
  nr <- nrow(data)
  nc <- ncol(data)
  col_indices <- seq(start_col, length.out = nc)
  col_letters <- int_to_col(col_indices)
  header_row <- start_row
  data_rows <- seq(start_row + 1L, length.out = nr)

  # header cells
  header_cells <- sprintf(
    "<c r=\"%s%d\" t=\"inlineStr\"><is><t>%s</t></is></c>",
    col_letters,
    header_row,
    htmlEscapeCopy(names(data))
  )

  # data cells: matrix (nr x nc)
  cell_matrix <- matrix(NA_character_, nrow = nr, ncol = nc)
  for (j in seq_len(nc)) {
    col <- data[[j]]
    refs <- sprintf("%s%d", col_letters[j], data_rows)
    if (inherits(col, "POSIXct") && !is.null(datetime_xf)) {
      num_val <- sprintf("%.10f", (as.numeric(col) / 86400) + 25569)
      cells <- sprintf(
        "<c r=\"%s\" s=\"%d\"><v>%s</v></c>",
        refs,
        datetime_xf,
        num_val
      )
      cells[is.na(col)] <- sprintf("<c r=\"%s\"/>", refs[is.na(col)])
    } else if (inherits(col, "Date") && !is.null(date_xf)) {
      num_val <- sprintf("%.17G", as.numeric(col) + 25569)
      cells <- sprintf(
        "<c r=\"%s\" s=\"%d\"><v>%s</v></c>",
        refs,
        date_xf,
        num_val
      )
      cells[is.na(col)] <- sprintf("<c r=\"%s\"/>", refs[is.na(col)])
    } else if (is.logical(col)) {
      cells <- sprintf(
        "<c r=\"%s\" t=\"b\"><v>%d</v></c>",
        refs,
        as.integer(col)
      )
      cells[is.na(col)] <- sprintf("<c r=\"%s\"/>", refs[is.na(col)])
    } else if (is.numeric(col)) {
      val_str <- sprintf("%.17G", col)
      cells <- sprintf("<c r=\"%s\"><v>%s</v></c>", refs, val_str)
      cells[is.na(col)] <- sprintf("<c r=\"%s\"/>", refs[is.na(col)])
    } else {
      col <- as.character(col)
      na_mask <- is.na(col)
      cells <- sprintf(
        "<c r=\"%s\" t=\"inlineStr\"><is><t>%s</t></is></c>",
        refs,
        htmlEscapeCopy(col)
      )
      cells[na_mask] <- sprintf("<c r=\"%s\"/>", refs[na_mask])
    }
    cell_matrix[, j] <- cells
  }

  # build named list: row number -> already-collapsed row strings
  row_cells <- do.call(paste0, as.data.frame(cell_matrix))
  result <- setNames(
    as.list(c(paste0(header_cells, collapse = ""), row_cells)),
    as.character(c(header_row, data_rows))
  )
  result
}

# --- Richtext cell helpers ------------------------------------------------
# Convert one `ftext`-style chunk (list with $value and $pr = fp_text)
# to a spreadsheetml <r><rPr>...</rPr><t xml:space="preserve">...</t></r>.
# Bare-string chunks from `fpar()` are normalised upstream via
# `fortify_fpar()`.
chunk_to_xlsx_run <- function(chunk) {
  pr <- chunk$pr
  rpr_parts <- character(0)
  if (isTRUE(pr$bold)) {
    rpr_parts <- c(rpr_parts, "<b/>")
  }
  if (isTRUE(pr$italic)) {
    rpr_parts <- c(rpr_parts, "<i/>")
  }
  if (isTRUE(pr$underlined)) {
    rpr_parts <- c(rpr_parts, "<u/>")
  }
  if (isTRUE(pr$strike)) {
    rpr_parts <- c(rpr_parts, "<strike/>")
  }
  if (!is.na(pr$font.size)) {
    rpr_parts <- c(rpr_parts, sprintf("<sz val=\"%g\"/>", pr$font.size))
  }
  if (!is.na(pr$color) && !identical(pr$color, "transparent")) {
    rpr_parts <- c(
      rpr_parts,
      sprintf("<color rgb=\"FF%s\"/>", hex_color(pr$color))
    )
  }
  if (!is.na(pr$font.family)) {
    rpr_parts <- c(
      rpr_parts,
      sprintf("<rFont val=\"%s\"/>", htmlEscapeCopy(pr$font.family))
    )
  }
  if (identical(pr$vertical.align, "superscript")) {
    rpr_parts <- c(rpr_parts, "<vertAlign val=\"superscript\"/>")
  } else if (identical(pr$vertical.align, "subscript")) {
    rpr_parts <- c(rpr_parts, "<vertAlign val=\"subscript\"/>")
  }

  rpr <- if (length(rpr_parts)) {
    paste0("<rPr>", paste0(rpr_parts, collapse = ""), "</rPr>")
  } else {
    ""
  }

  sprintf(
    "<r>%s<t xml:space=\"preserve\">%s</t></r>",
    rpr,
    htmlEscapeCopy(chunk$value)
  )
}

# Convert a fpar to the <is>...</is> body of an inline-string cell.
fpar_to_inlinestr <- function(fpar) {
  chunks <- fortify_fpar(fpar)
  if (!length(chunks)) {
    return("<is><t xml:space=\"preserve\"/></is>")
  }
  runs <- vapply(chunks, chunk_to_xlsx_run, character(1L))
  paste0("<is>", paste0(runs, collapse = ""), "</is>")
}

# Build the full <c r="..."> cell for a fpar at a given ref.
fpar_to_cell_xml <- function(fpar, ref) {
  sprintf(
    "<c r=\"%s\" t=\"inlineStr\">%s</c>",
    ref,
    fpar_to_inlinestr(fpar)
  )
}

# --- sheet_write_data methods: character, fpar, block_list ----------------

#' @export
sheet_write_data.character <- function(
  x,
  value,
  sheet,
  start_row = 1L,
  start_col = 1L,
  direction = c("vertical", "horizontal"),
  ...
) {
  stopifnot(inherits(x, "rxlsx"))
  check_sheet_exists(x, sheet)
  direction <- match.arg(direction)
  start_row <- as.integer(start_row)
  start_col <- as.integer(start_col)

  n <- length(value)
  if (direction == "vertical") {
    rows <- seq(start_row, length.out = n)
    cols <- rep(start_col, n)
  } else {
    rows <- rep(start_row, n)
    cols <- seq(start_col, length.out = n)
  }
  refs <- sprintf("%s%d", int_to_col(cols), rows)

  na_mask <- is.na(value)
  cells <- character(n)
  cells[!na_mask] <- sprintf(
    "<c r=\"%s\" t=\"inlineStr\"><is><t xml:space=\"preserve\">%s</t></is></c>",
    refs[!na_mask],
    htmlEscapeCopy(value[!na_mask])
  )
  cells[na_mask] <- sprintf("<c r=\"%s\"/>", refs[na_mask])

  # group by row
  new_cells <- tapply(cells, rows, paste0, collapse = "")
  new_cells <- as.list(new_cells)
  names(new_cells) <- as.character(unique(rows))

  write_cells_into_sheet(x, sheet, new_cells)
}

#' @export
sheet_write_data.fpar <- function(
  x,
  value,
  sheet,
  start_row = 1L,
  start_col = 1L,
  ...
) {
  stopifnot(inherits(x, "rxlsx"))
  check_sheet_exists(x, sheet)
  start_row <- as.integer(start_row)
  start_col <- as.integer(start_col)

  ref <- sprintf("%s%d", int_to_col(start_col), start_row)
  cell <- fpar_to_cell_xml(value, ref)

  new_cells <- setNames(list(cell), as.character(start_row))
  write_cells_into_sheet(x, sheet, new_cells)
}

#' @export
sheet_write_data.block_list <- function(
  x,
  value,
  sheet,
  start_row = 1L,
  start_col = 1L,
  direction = c("vertical", "horizontal"),
  ...
) {
  stopifnot(inherits(x, "rxlsx"))
  check_sheet_exists(x, sheet)
  direction <- match.arg(direction)
  start_row <- as.integer(start_row)
  start_col <- as.integer(start_col)

  # only fpar items are rendered -- skip non-fpar blocks with a warning
  is_fpar <- vapply(value, inherits, logical(1), what = "fpar")
  if (!all(is_fpar)) {
    cli::cli_warn(c(
      "{.fun sheet_write_data} can only render {.cls fpar} items in a
       {.cls block_list}; skipping {sum(!is_fpar)} non-fpar item{?s}."
    ))
  }
  fpars <- value[is_fpar]
  n <- length(fpars)
  if (n == 0L) {
    return(invisible(x))
  }

  if (direction == "vertical") {
    rows <- seq(start_row, length.out = n)
    cols <- rep(start_col, n)
  } else {
    rows <- rep(start_row, n)
    cols <- seq(start_col, length.out = n)
  }
  refs <- sprintf("%s%d", int_to_col(cols), rows)

  cells <- vapply(
    seq_len(n),
    function(i) {
      fpar_to_cell_xml(fpars[[i]], refs[i])
    },
    character(1L)
  )

  new_cells <- tapply(cells, rows, paste0, collapse = "")
  new_cells <- as.list(new_cells)
  names(new_cells) <- as.character(unique(rows))

  write_cells_into_sheet(x, sheet, new_cells)
}

#' @export
#' @title Add a drawing to an Excel sheet
#' @description Add a graphical element into a sheet of an xlsx workbook.
#' This is a generic function dispatching on `value`. Supported inputs
#' include:
#' - [external_img()]: an image file (PNG, JPEG, GIF, ...) copied into
#'   the workbook and placed on the sheet.
#' - `gg`: a ggplot object rendered to PNG via `ragg::agg_png()` and
#'   embedded via the `external_img` path. Extra arguments: `res`
#'   (default 300 ppi), `alt_text` (auto-detected via
#'   `ggplot2::get_alt_text()` when empty), `scale` (same semantics as
#'   `ggplot2::ggsave()`).
#'
#' The drawing can be positioned in three ways, matching Excel's own
#' "Format Picture" options:
#' - Fixed position (default): pass `left` / `top` and
#'   `width` / `height` in inches. The drawing stays put when rows or
#'   columns are inserted/resized.
#' - Move and size with cells (Excel's default): pass
#'   `anchor = "B2:H20"` (a cell range). The drawing spans those two
#'   cells and follows them.
#' - Move but don't size with cells: pass `anchor = "B2"`
#'   (a single cell). The drawing anchors to that cell, keeping the
#'   size set by `width` / `height`.
#'
#' Additional methods for `ms_chart` and `dml` are provided by the
#' extension packages 'mschart' and 'rvg'.
#'
#' Use [sheet_write_data()] to write data into the sheet before or
#' after adding a drawing.
#' @param x rxlsx object created by [read_xlsx()]
#' @param value object to add (dispatched to the appropriate method)
#' @param sheet sheet name (must already exist, see [add_sheet()])
#' @param ... method-specific arguments. Common across `external_img`
#'   and `gg` methods: `left` / `top` (fixed-position top-left in
#'   inches, default `(1, 1)`), `width` / `height` (size in inches),
#'   `anchor` (cell reference like `"B2"` or `"B2:H20"`, see
#'   description), `edit_as` (one of `"twoCell"`, `"oneCell"`,
#'   `"absolute"`; how Excel treats the drawing when rows/columns are
#'   resized — only meaningful with a cell range `anchor`). The `gg`
#'   method also accepts `res` (ppi, default 300), `alt_text`
#'   (auto-detected via `ggplot2::get_alt_text()` if empty) and
#'   `scale` (passed to [ragg::agg_png()]).
#' @return the rxlsx object (invisibly)
#' @examples
#' img <- system.file("extdata", "example.png", package = "officer")
#' if (nzchar(img) && file.exists(img)) {
#'   x <- read_xlsx()
#'   x <- add_sheet(x, label = "pics")
#'   x <- sheet_add_drawing(
#'     x, sheet = "pics",
#'     value = external_img(img, width = 2, height = 2)
#'   )
#'   print(x, target = tempfile(fileext = ".xlsx"))
#' }
#'
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   gg <- ggplot2::ggplot(iris, ggplot2::aes(Sepal.Length, Sepal.Width)) +
#'     ggplot2::geom_point()
#'   x <- read_xlsx()
#'   x <- add_sheet(x, label = "plots")
#'   x <- sheet_add_drawing(x, value = gg, sheet = "plots",
#'                          width = 4, height = 3)
#'   print(x, target = tempfile(fileext = ".xlsx"))
#' }
#' @seealso [read_xlsx()], [add_sheet()], [sheet_write_data()]
sheet_add_drawing <- function(x, value, sheet, ...) {
  UseMethod("sheet_add_drawing", value)
}

#' @export
#' @method sheet_add_drawing external_img
sheet_add_drawing.external_img <- function(
  x,
  value,
  sheet,
  left = 1,
  top = 1,
  width = NULL,
  height = NULL,
  anchor = NULL,
  edit_as = c("twoCell", "oneCell", "absolute"),
  ...
) {
  edit_as <- match.arg(edit_as)
  ft <- parse_anchor(anchor)
  stopifnot(inherits(x, "rxlsx"))
  check_sheet_exists(x, sheet)

  dims <- attr(value, "dims")
  if (is.null(width)) {
    width <- dims$width[1]
  }
  if (is.null(height)) {
    height <- dims$height[1]
  }
  alt <- attr(value, "alt") %||% ""

  src <- as.character(value)[1]
  if (!file.exists(src)) {
    cli::cli_abort("Image file {.path {src}} does not exist.")
  }

  # copy image to xl/media/ with a unique name (file.copy overwrites
  # if the destination already exists)
  media_dir <- file.path(x$package_dir, "xl", "media")
  dir.create(media_dir, showWarnings = FALSE, recursive = TRUE)
  ext <- tolower(tools::file_ext(src))
  if (!nzchar(ext)) {
    ext <- "png"
  }
  target_path <- tempfile(
    pattern = "img",
    tmpdir = media_dir,
    fileext = paste0(".", ext)
  )
  file.copy(src, target_path, overwrite = TRUE)

  sheet_obj <- x$sheets$get_sheet(
    which(x$worksheets$sheet_names() == sheet)
  )
  drawing <- xlsx_drawing$new(x$package_dir, sheet_obj, x$content_type)

  image_rid <- drawing$add_image_rel(basename(target_path))
  drawing$add_image_anchor(
    image_rid = image_rid,
    left = left,
    top = top,
    width = width,
    height = height,
    from = ft$from,
    to = ft$to,
    edit_as = edit_as,
    alt = alt
  )

  invisible(x)
}

#' @export
#' @method sheet_add_drawing gg
sheet_add_drawing.gg <- function(
  x,
  value,
  sheet,
  left = 1,
  top = 1,
  width = 6,
  height = 4,
  anchor = NULL,
  edit_as = c("twoCell", "oneCell", "absolute"),
  res = 300,
  alt_text = "",
  scale = 1,
  ...
) {
  edit_as <- match.arg(edit_as)
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    cli::cli_abort("The {.pkg ggplot2} package is required for this method.")
  }
  stopifnot(inherits(value, "gg"))

  file <- tempfile(fileext = ".png")
  ragg::agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  print(value)
  dev.off()
  on.exit(unlink(file), add = TRUE)

  if (is.null(alt_text) || !nzchar(alt_text)) {
    alt_text <- ggplot2::get_alt_text(value)
    if (is.null(alt_text)) alt_text <- ""
  }

  sheet_add_drawing(
    x,
    external_img(file, width = width, height = height, alt = alt_text),
    sheet = sheet,
    left = left,
    top = top,
    anchor = anchor,
    edit_as = edit_as
  )
}

# Convert integer column index to Excel letter (1=A, 2=B, ..., 27=AA)
int_to_col <- function(x) {
  vapply(
    x,
    function(n) {
      if (is.na(n)) {
        return(NA_character_)
      }
      col <- ""
      while (n > 0) {
        n <- n - 1L
        col <- paste0(LETTERS[n %% 26L + 1L], col)
        n <- n %/% 26L
      }
      col
    },
    character(1L)
  )
}

#' @export
#' @param target path to the xlsx file to write
#' @param ... unused
#' @rdname read_xlsx
#' @examples
#' x <- read_xlsx()
#' print(x, target = tempfile(fileext = ".xlsx"))
print.rxlsx <- function(x, target = NULL, ...) {
  if (is.null(target)) {
    cat("xlsx document with", length(x), "sheet(s):\n")
    print(x$worksheets$sheet_names())
    return(invisible())
  }

  if (!grepl(x = target, pattern = "\\.(xlsx)$", ignore.case = TRUE)) {
    cli::cli_abort("{.file {target}} should have {.file .xlsx} extension.")
  }

  x$worksheets$save()
  x$content_type$save()
  x$styles$save()

  x$core_properties['modified', 'value'] <- format(
    Sys.time(),
    "%Y-%m-%dT%H:%M:%SZ"
  )
  x$core_properties['lastModifiedBy', 'value'] <- Sys.getenv("USER")
  write_core_properties(x$core_properties, x$package_dir)

  invisible(pack_folder(folder = x$package_dir, target = target))
}

test_that("console printing", {
  x <- read_xlsx()
  expect_output(print(x), regexp = "^xlsx document with 1 sheet")
})

test_that("check extention and print document", {
  x <- read_xlsx()
  print(x, target = "print.xlsx")
  expect_true(file.exists("print.xlsx"))
  expect_error(print(x, target = "print.xlsxxxx"))
})


unlink("*.xlsx")

test_that("ooxml_chart_uris returns paired URIs for both kinds", {
  dml <- ooxml_chart_uris("drawingml")
  expect_named(dml, c("rel_type", "graphic_uri"))
  expect_match(dml$rel_type, "/officeDocument/2006/relationships/chart$")
  expect_match(dml$graphic_uri, "/drawingml/2006/chart$")

  cx <- ooxml_chart_uris("chartex")
  expect_match(cx$rel_type, "/2014/relationships/chartEx$")
  expect_match(cx$graphic_uri, "/drawing/2014/chartex$")

  expect_identical(ooxml_chart_uris(), dml)
  expect_error(ooxml_chart_uris("nope"))
})

test_that("default_chart_inner dispatches on graphic_uri", {
  u <- ooxml_chart_uris("drawingml")
  dml <- officer:::default_chart_inner(u$graphic_uri, "rId7")
  expect_match(dml, "^<c:chart ")
  expect_match(dml, "xmlns:c=\"[^\"]+/drawingml/2006/chart\"")
  expect_match(dml, "r:id=\"rId7\"")

  v <- ooxml_chart_uris("chartex")
  cx <- officer:::default_chart_inner(v$graphic_uri, "rId7")
  expect_match(cx, "^<cx:chart ")
  expect_match(cx, "xmlns:cx=\"[^\"]+/drawing/2014/chartex\"")
  expect_match(cx, "r:id=\"rId7\"")
})

test_that("parse_cell_ref handles A1-style refs", {
  expect_equal(officer:::parse_cell_ref("A1"), list(col = 0L, row = 0L))
  expect_equal(officer:::parse_cell_ref("C4"), list(col = 2L, row = 3L))
  expect_equal(officer:::parse_cell_ref("aa12"), list(col = 26L, row = 11L))
  expect_equal(officer:::parse_cell_ref("Z99"), list(col = 25L, row = 98L))
  expect_error(officer:::parse_cell_ref("3A"), "invalid")
  expect_error(officer:::parse_cell_ref(""), "non-empty")
  expect_error(officer:::parse_cell_ref(c("A1", "B2")), "single")
})

# Helper: get a fresh xlsx_drawing on sheet 1 of a freshly-written xlsx,
# and return the inspector that reads the drawing XML back.
fresh_drawing <- function() {
  tmp <- tempfile(fileext = ".xlsx")
  x <- read_xlsx()
  print(x, target = tmp)
  x2 <- read_xlsx(tmp)
  sheet_obj <- x2$sheets$get_sheet(1)
  drw <- xlsx_drawing$new(x2$package_dir, sheet_obj, x2$content_type)
  list(drw = drw, doc = function() drw$get())
}

test_that("add_chart_anchor emits twoCellAnchor with from + to", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_chart_rel("chart1.xml")
  fd$drw$add_chart_anchor(rid, from = "B2", to = "H20")
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  anchor <- xml2::xml_find_first(fd$doc(), "//xdr:twoCellAnchor", ns)
  expect_false(inherits(anchor, "xml_missing"))
  expect_equal(xml2::xml_attr(anchor, "editAs"), "twoCell")
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:from/xdr:col", ns)),
    "1"
  )
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:from/xdr:row", ns)),
    "1"
  )
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:to/xdr:col", ns)),
    "7"
  )
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:to/xdr:row", ns)),
    "19"
  )
})

test_that("add_chart_anchor honours edit_as", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_chart_rel("chart1.xml")
  fd$drw$add_chart_anchor(rid, from = "A1", to = "B2", edit_as = "absolute")
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  expect_equal(
    xml2::xml_attr(
      xml2::xml_find_first(fd$doc(), "//xdr:twoCellAnchor", ns),
      "editAs"
    ),
    "absolute"
  )
})

test_that("add_chart_anchor emits oneCellAnchor with from only", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_chart_rel("chart1.xml")
  fd$drw$add_chart_anchor(rid, from = "C4", width = 5, height = 3)
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  anchor <- xml2::xml_find_first(fd$doc(), "//xdr:oneCellAnchor", ns)
  expect_false(inherits(anchor, "xml_missing"))
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:from/xdr:col", ns)),
    "2"
  )
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:from/xdr:row", ns)),
    "3"
  )
  ext <- xml2::xml_find_first(anchor, "xdr:ext", ns)
  expect_equal(xml2::xml_attr(ext, "cx"), as.character(5 * 914400))
  expect_equal(xml2::xml_attr(ext, "cy"), as.character(3 * 914400))
})

test_that("add_chart_anchor keeps absoluteAnchor when from/to omitted", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_chart_rel("chart1.xml")
  fd$drw$add_chart_anchor(rid, left = 0.5, top = 0.5, width = 4, height = 3)
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  expect_false(inherits(
    xml2::xml_find_first(fd$doc(), "//xdr:absoluteAnchor", ns),
    "xml_missing"
  ))
})

test_that("add_chart_anchor errors on `to` without `from`", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_chart_rel("chart1.xml")
  expect_error(
    fd$drw$add_chart_anchor(rid, to = "B2"),
    "`to` requires `from`"
  )
})

test_that("add_image_anchor emits twoCellAnchor with from + to", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_image_rel("image1.png")
  fd$drw$add_image_anchor(rid, from = "B2", to = "F10")
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  anchor <- xml2::xml_find_first(fd$doc(), "//xdr:twoCellAnchor", ns)
  expect_false(inherits(anchor, "xml_missing"))
  pic <- xml2::xml_find_first(anchor, "xdr:pic", ns)
  expect_false(inherits(pic, "xml_missing"))
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:to/xdr:col", ns)),
    "5"
  )
})

test_that("add_image_anchor emits oneCellAnchor with from only", {
  fd <- fresh_drawing()
  rid <- fd$drw$add_image_rel("image1.png")
  fd$drw$add_image_anchor(rid, from = "C4", width = 2, height = 1.5)
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  expect_false(inherits(
    xml2::xml_find_first(fd$doc(), "//xdr:oneCellAnchor", ns),
    "xml_missing"
  ))
})

test_that("parse_anchor decodes user-facing anchor strings", {
  expect_equal(officer:::parse_anchor(NULL), list(from = NULL, to = NULL))
  expect_equal(officer:::parse_anchor("C4"), list(from = "C4", to = NULL))
  expect_equal(
    officer:::parse_anchor("B2:H20"),
    list(from = "B2", to = "H20")
  )
  expect_error(officer:::parse_anchor(42), "single string")
  expect_error(officer:::parse_anchor("B2:"), "FROM:TO")
  expect_error(officer:::parse_anchor(""), "FROM:TO|single string")
})

test_that("sheet_add_drawing(external_img, anchor = range) uses twoCellAnchor", {
  img <- system.file(
    "medias",
    "figures",
    "body_add_doc_1.png",
    package = "officer"
  )
  skip_if_not(nzchar(img) && file.exists(img))
  x <- read_xlsx()
  x <- add_sheet(x, label = "pics")
  x <- sheet_add_drawing(
    x,
    sheet = "pics",
    value = external_img(img, width = 2, height = 2),
    anchor = "B2:F10"
  )
  out <- tempfile(fileext = ".xlsx")
  print(x, target = out)
  pkg <- tempfile()
  dir.create(pkg)
  zip::unzip(out, exdir = pkg)
  drw <- list.files(
    file.path(pkg, "xl", "drawings"),
    pattern = "\\.xml$",
    full.names = TRUE
  )[1]
  doc <- xml2::read_xml(drw)
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  anchor <- xml2::xml_find_first(doc, "//xdr:twoCellAnchor", ns)
  expect_false(inherits(anchor, "xml_missing"))
  expect_equal(
    xml2::xml_text(xml2::xml_find_first(anchor, "xdr:to/xdr:col", ns)),
    "5"
  )
})

test_that("sheet_add_drawing(external_img, anchor = single cell) uses oneCellAnchor", {
  img <- system.file(
    "medias",
    "figures",
    "body_add_doc_1.png",
    package = "officer"
  )
  skip_if_not(nzchar(img) && file.exists(img))
  x <- read_xlsx()
  x <- add_sheet(x, label = "pics")
  x <- sheet_add_drawing(
    x,
    sheet = "pics",
    value = external_img(img, width = 2, height = 2),
    anchor = "C4"
  )
  out <- tempfile(fileext = ".xlsx")
  print(x, target = out)
  pkg <- tempfile()
  dir.create(pkg)
  zip::unzip(out, exdir = pkg)
  drw <- list.files(
    file.path(pkg, "xl", "drawings"),
    pattern = "\\.xml$",
    full.names = TRUE
  )[1]
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  expect_false(inherits(
    xml2::xml_find_first(xml2::read_xml(drw), "//xdr:oneCellAnchor", ns),
    "xml_missing"
  ))
})

test_that("sheet_add_drawing(external_img) without anchor stays absolute", {
  img <- system.file(
    "medias",
    "figures",
    "body_add_doc_1.png",
    package = "officer"
  )
  skip_if_not(nzchar(img) && file.exists(img))
  x <- read_xlsx()
  x <- add_sheet(x, label = "pics")
  x <- sheet_add_drawing(
    x,
    sheet = "pics",
    value = external_img(img, width = 2, height = 2)
  )
  out <- tempfile(fileext = ".xlsx")
  print(x, target = out)
  pkg <- tempfile()
  dir.create(pkg)
  zip::unzip(out, exdir = pkg)
  drw <- list.files(
    file.path(pkg, "xl", "drawings"),
    pattern = "\\.xml$",
    full.names = TRUE
  )[1]
  ns <- c(
    xdr = "http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing"
  )
  expect_false(inherits(
    xml2::xml_find_first(xml2::read_xml(drw), "//xdr:absoluteAnchor", ns),
    "xml_missing"
  ))
})

# Helper: fresh workbook without the template's default sheet, so that the
# first user-added sheet lives at xl/worksheets/sheet1.xml. Tests that don't
# care about the underlying file name use `read_xlsx()` directly.
new_xlsx <- function() {
  x <- read_xlsx()
  sheet_remove(x, sheet = sheet_names(x)[1])
}

test_that("add_sheet is additive and does not drop the default sheet", {
  doc <- read_xlsx()
  default_name <- sheet_names(doc)[1]
  doc <- add_sheet(doc, label = "sheet1")
  expect_setequal(sheet_names(doc), c(default_name, "sheet1"))
})

test_that("sheet_select sets the active tab", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "sheet1")
  doc <- sheet_select(doc, sheet = "sheet1")

  sheet_id <- doc$worksheets$get_sheet_id("sheet1")
  wb_view <- xml_find_first(
    doc$worksheets$get(),
    "d1:bookViews/d1:workbookView"
  )
  expect_equal(as.integer(xml_attr(wb_view, "activeTab")), sheet_id - 1)
})

test_that("sheet_remove removes sheet + files + content-type", {
  doc <- read_xlsx()
  default_name <- sheet_names(doc)[1]
  doc <- add_sheet(doc, label = "keep")
  expect_length(sheet_names(doc), 2)

  doc <- sheet_remove(doc, sheet = default_name)
  expect_equal(sheet_names(doc), "keep")

  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  remaining <- list.files(
    file.path(unpack_dir, "xl/worksheets"),
    pattern = "\\.xml$"
  )
  expect_length(remaining, 1)
})

test_that("sheet_select deselects other sheets", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "new")
  doc <- sheet_select(doc, sheet = "new")
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  sheet_files <- list.files(
    file.path(unpack_dir, "xl/worksheets"),
    pattern = "\\.xml$"
  )
  expect_length(sheet_files, 1)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  sheet_xml <- read_xml(file.path(unpack_dir, "xl/worksheets", sheet_files[1]))
  sv <- xml_find_first(sheet_xml, "d1:sheetViews/d1:sheetView", ns = ns)
  expect_equal(xml_attr(sv, "tabSelected"), "1")
})

test_that("sheet_write_data writes correct cells", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  doc <- sheet_write_data(doc, value = head(iris, 3), sheet = "data")
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )
  rows <- xml_find_all(sheet_xml, "d1:sheetData/d1:row", ns = ns)

  # header + 3 data rows
  expect_equal(length(rows), 4L)

  # header row has 5 columns
  header_cells <- xml_find_all(rows[[1]], "d1:c", ns = ns)
  expect_equal(length(header_cells), 5L)

  # first header is Sepal.Length
  first_val <- xml_text(xml_find_first(
    header_cells[[1]],
    "d1:is/d1:t",
    ns = ns
  ))
  expect_equal(first_val, "Sepal.Length")

  # numeric value in B2
  b2 <- xml_find_first(rows[[2]], "d1:c[@r='B2']/d1:v", ns = ns)
  expect_equal(xml_text(b2), "3.5")

  # text value in E2 (Species)
  e2 <- xml_find_first(rows[[2]], "d1:c[@r='E2']/d1:is/d1:t", ns = ns)
  expect_equal(xml_text(e2), "setosa")
})

test_that("sheet_write_data with start_row and start_col", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  doc <- sheet_write_data(
    doc,
    value = data.frame(a = 1:2, b = 3:4),
    sheet = "data",
    start_row = 5,
    start_col = 3
  )
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )

  # header at row 5, col C
  c5 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='5']/d1:c[@r='C5']",
    ns = ns
  )
  expect_false(inherits(c5, "xml_missing"))
  expect_equal(
    xml_text(xml_find_first(c5, "d1:is/d1:t", ns = ns)),
    "a"
  )

  # data at D6
  d6 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='6']/d1:c[@r='D6']/d1:v",
    ns = ns
  )
  expect_equal(xml_text(d6), "3")
})

test_that("sheet_write_data merges two datasets", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  doc <- sheet_write_data(
    doc,
    value = data.frame(x = 1:2),
    sheet = "data"
  )
  doc <- sheet_write_data(
    doc,
    value = data.frame(y = 10:11),
    sheet = "data",
    start_col = 3
  )
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )

  # row 1 has both headers
  row1_cells <- xml_find_all(
    sheet_xml,
    "d1:sheetData/d1:row[@r='1']/d1:c",
    ns = ns
  )
  expect_equal(length(row1_cells), 2L)

  # A2 has 1, C2 has 10
  a2 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='2']/d1:c[@r='A2']/d1:v",
    ns = ns
  )
  c2 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='2']/d1:c[@r='C2']/d1:v",
    ns = ns
  )
  expect_equal(xml_text(a2), "1")
  expect_equal(xml_text(c2), "10")
})

test_that("sheet_write_data handles NA", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  doc <- sheet_write_data(
    doc,
    value = data.frame(a = c(1, NA), b = c("x", NA)),
    sheet = "data"
  )
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )

  # A3 is empty (NA numeric)
  a3 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='3']/d1:c[@r='A3']",
    ns = ns
  )
  expect_false(inherits(a3, "xml_missing"))
  a3_val <- xml_find_first(a3, "d1:v", ns = ns)
  expect_true(inherits(a3_val, "xml_missing"))
})

test_that("sheet_write_data handles Date columns", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  df <- data.frame(
    d = as.Date(c("2024-01-15", "2024-06-30", NA)),
    x = 1:3
  )
  doc <- sheet_write_data(doc, value = df, sheet = "data")
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")

  # styles.xml has numFmtId 14 (date)
  styles_xml <- paste(
    readLines(
      file.path(unpack_dir, "xl/styles.xml"),
      warn = FALSE
    ),
    collapse = ""
  )
  expect_true(grepl("numFmtId=\"14\"", styles_xml))

  # date cells have s= attribute
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )
  a2 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='2']/d1:c[@r='A2']",
    ns = ns
  )
  expect_false(is.na(xml_attr(a2, "s")))

  # date value is Excel serial number (2024-01-15 = 45306)
  a2_val <- as.numeric(xml_text(xml_find_first(a2, "d1:v", ns = ns)))
  expect_equal(a2_val, as.numeric(as.Date("2024-01-15")) + 25569)

  # NA date produces empty cell
  a4 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='4']/d1:c[@r='A4']",
    ns = ns
  )
  a4_val <- xml_find_first(a4, "d1:v", ns = ns)
  expect_true(inherits(a4_val, "xml_missing"))
})

test_that("sheet_write_data handles POSIXct columns", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  df <- data.frame(
    dt = as.POSIXct(
      c("2024-01-15 10:30:00", "2024-06-30 14:00:00", NA),
      tz = "UTC"
    ),
    x = 1:3
  )
  doc <- sheet_write_data(doc, value = df, sheet = "data")
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")

  # styles.xml has numFmtId 22 (datetime)
  styles_xml <- paste(
    readLines(
      file.path(unpack_dir, "xl/styles.xml"),
      warn = FALSE
    ),
    collapse = ""
  )
  expect_true(grepl("numFmtId=\"22\"", styles_xml))

  # datetime cells have s= attribute
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )
  a2 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='2']/d1:c[@r='A2']",
    ns = ns
  )
  expect_false(is.na(xml_attr(a2, "s")))

  # value is a serial number > 45306 (date part) with fractional time
  a2_val <- as.numeric(xml_text(xml_find_first(a2, "d1:v", ns = ns)))
  expect_gt(a2_val, 45306)
  expect_true(a2_val != round(a2_val)) # has fractional part (time)

  # NA datetime produces empty cell
  a4 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='4']/d1:c[@r='A4']",
    ns = ns
  )
  a4_val <- xml_find_first(a4, "d1:v", ns = ns)
  expect_true(inherits(a4_val, "xml_missing"))
})

test_that("sheet_write_data handles logical columns", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "data")
  df <- data.frame(flag = c(TRUE, FALSE, NA))
  doc <- sheet_write_data(doc, value = df, sheet = "data")
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")
  sheet_xml <- read_xml(
    file.path(unpack_dir, "xl/worksheets/sheet1.xml")
  )

  # TRUE -> t="b", v=1
  a2 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='2']/d1:c[@r='A2']",
    ns = ns
  )
  expect_equal(xml_attr(a2, "t"), "b")
  expect_equal(xml_text(xml_find_first(a2, "d1:v", ns = ns)), "1")

  # FALSE -> t="b", v=0
  a3 <- xml_find_first(
    sheet_xml,
    "d1:sheetData/d1:row[@r='3']/d1:c[@r='A3']",
    ns = ns
  )
  expect_equal(xml_text(xml_find_first(a3, "d1:v", ns = ns)), "0")
})

test_that("sheet_add_drawing creates drawing infrastructure", {
  skip_if_not_installed("mschart", minimum_version = "0.4.2")
  library(mschart)

  my_chart <- ms_barchart(
    data = data.frame(
      x = c("A", "B"),
      y = c(1, 2),
      group = rep("s1", 2)
    ),
    x = "x",
    y = "y",
    group = "group"
  )

  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "chart")
  doc <- sheet_add_drawing(
    doc,
    value = my_chart,
    sheet = "chart"
  )
  out <- print(doc, target = tempfile(fileext = ".xlsx"))

  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  ns <- c(d1 = "http://schemas.openxmlformats.org/spreadsheetml/2006/main")

  # drawing file exists
  expect_true(
    file.exists(file.path(unpack_dir, "xl/drawings/drawing1.xml"))
  )

  # chart file exists
  chart_files <- list.files(
    file.path(unpack_dir, "xl/charts"),
    pattern = "^chart.*\\.xml$"
  )
  expect_gte(length(chart_files), 1L)

  # sheet has drawing ref before extLst
  sheet_str <- paste(
    readLines(
      file.path(unpack_dir, "xl/worksheets/sheet1.xml"),
      warn = FALSE
    ),
    collapse = ""
  )
  drawing_pos <- regexpr("<drawing", sheet_str)
  extlst_pos <- regexpr("<extLst", sheet_str)
  if (extlst_pos > 0) {
    expect_true(drawing_pos < extlst_pos)
  }

  # drawing rels points to chart
  drawing_rels <- paste(
    readLines(
      file.path(unpack_dir, "xl/drawings/_rels/drawing1.xml.rels"),
      warn = FALSE
    ),
    collapse = ""
  )
  expect_true(grepl("chart", drawing_rels))

  # no externalData in chart
  chart_str <- paste(
    readLines(
      file.path(unpack_dir, "xl/charts", chart_files[1]),
      warn = FALSE
    ),
    collapse = ""
  )
  expect_false(grepl("externalData", chart_str))
})

sheet_cells_xml <- function(doc, sheet_xml_name = "sheet1.xml") {
  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  paste(
    readLines(file.path(unpack_dir, "xl/worksheets", sheet_xml_name)),
    collapse = "\n"
  )
}

test_that("sheet_write_data dispatches on character (vertical)", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "chr")
  doc <- sheet_write_data(
    doc,
    value = c("A", "B", "C"),
    sheet = "chr",
    start_row = 2,
    start_col = 3
  )
  xml <- sheet_cells_xml(doc)
  expect_match(
    xml,
    "<c r=\"C2\" t=\"inlineStr\"><is><t[^>]*>A</t>",
    fixed = FALSE
  )
  expect_match(
    xml,
    "<c r=\"C3\" t=\"inlineStr\"><is><t[^>]*>B</t>",
    fixed = FALSE
  )
  expect_match(
    xml,
    "<c r=\"C4\" t=\"inlineStr\"><is><t[^>]*>C</t>",
    fixed = FALSE
  )
})

test_that("sheet_write_data dispatches on character (horizontal)", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "chr")
  doc <- sheet_write_data(
    doc,
    value = c("X", "Y", "Z"),
    sheet = "chr",
    start_row = 1,
    start_col = 1,
    direction = "horizontal"
  )
  xml <- sheet_cells_xml(doc)
  expect_match(xml, "r=\"A1\"[^>]*>[^<]*<is><t[^>]*>X</t>", fixed = FALSE)
  expect_match(xml, "r=\"B1\"[^>]*>[^<]*<is><t[^>]*>Y</t>", fixed = FALSE)
  expect_match(xml, "r=\"C1\"[^>]*>[^<]*<is><t[^>]*>Z</t>", fixed = FALSE)
})

test_that("sheet_write_data on fpar mixes bare strings and ftext", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "fp")
  f <- fpar(
    "plain ",
    ftext("bold", fp_text_lite(bold = TRUE)),
    " plain again"
  )
  doc <- sheet_write_data(
    doc,
    value = f,
    sheet = "fp",
    start_row = 1,
    start_col = 1
  )
  xml <- sheet_cells_xml(doc)
  expect_match(xml, "<c r=\"A1\" t=\"inlineStr\">", fixed = TRUE)
  # bare string runs have no <rPr>
  expect_match(xml, "<r><t xml:space=\"preserve\">plain </t></r>", fixed = TRUE)
  expect_match(xml, "<b/>", fixed = TRUE)
  expect_match(xml, ">bold</t>", fixed = TRUE)
  expect_match(
    xml,
    "<r><t xml:space=\"preserve\"> plain again</t></r>",
    fixed = TRUE
  )
})

test_that("sheet_write_data on fpar emits richtext runs", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "fp")
  f <- fpar(
    ftext("bold ", fp_text(bold = TRUE, color = "red")),
    ftext("italic ", fp_text(italic = TRUE)),
    ftext("under ", fp_text(underlined = TRUE)),
    ftext("strike ", fp_text(strike = TRUE)),
    ftext("H", fp_text()),
    ftext("2", fp_text(vertical.align = "subscript")),
    ftext("O", fp_text())
  )
  doc <- sheet_write_data(
    doc,
    value = f,
    sheet = "fp",
    start_row = 1,
    start_col = 1
  )
  xml <- sheet_cells_xml(doc)
  expect_match(xml, "<c r=\"A1\" t=\"inlineStr\">", fixed = TRUE)
  expect_match(xml, "<b/>", fixed = TRUE)
  expect_match(xml, "rgb=\"FFFF0000\"", fixed = TRUE)
  expect_match(xml, "<i/>", fixed = TRUE)
  expect_match(xml, "<u/>", fixed = TRUE)
  expect_match(xml, "<strike/>", fixed = TRUE)
  expect_match(xml, "<vertAlign val=\"subscript\"/>", fixed = TRUE)
})

test_that("xlsx_styles$get_font_id caches fonts by signature", {
  skip_if_not_installed("gdtools")
  skip_if_not(
    gdtools::font_family_exists("Arial"),
    "Arial font family is not available on this system"
  )

  doc <- read_xlsx()
  styles <- doc$styles

  # first call -> registers a new font, returns a 0-based index
  id1 <- styles$get_font_id(
    name = "Arial",
    size = 12,
    bold = TRUE,
    italic = FALSE,
    underline = FALSE,
    color = "FF0000"
  )
  expect_true(is.numeric(id1))
  expect_gte(id1, 0)

  # same signature -> same index (cache hit)
  id2 <- styles$get_font_id(
    name = "Arial",
    size = 12,
    bold = TRUE,
    italic = FALSE,
    underline = FALSE,
    color = "FF0000"
  )
  expect_equal(id1, id2)

  # different signature -> different index
  id3 <- styles$get_font_id(
    name = "Arial",
    size = 12,
    bold = FALSE,
    italic = TRUE,
    underline = TRUE,
    color = "00FF00"
  )
  expect_false(id3 == id1)

  # the new font XML is persisted in styles.xml on print
  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  styles_xml <- paste(
    readLines(file.path(unpack_dir, "xl/styles.xml")),
    collapse = "\n"
  )
  expect_match(styles_xml, "Arial", fixed = TRUE)
  expect_match(styles_xml, "<b/>", fixed = TRUE)
  expect_match(styles_xml, "<i/>", fixed = TRUE)
  expect_match(styles_xml, "<u/>", fixed = TRUE)
  expect_match(styles_xml, "FF00FF00", fixed = TRUE)
})

test_that("xlsx_styles$get_fill_id caches fills by signature", {
  doc <- read_xlsx()
  styles <- doc$styles

  id1 <- styles$get_fill_id(bg_color = "FF9900")
  expect_true(is.numeric(id1))
  expect_gte(id1, 0)

  # cache hit
  id2 <- styles$get_fill_id(bg_color = "FF9900")
  expect_equal(id1, id2)

  # different color -> different index
  id3 <- styles$get_fill_id(bg_color = "0066CC")
  expect_false(id3 == id1)

  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  styles_xml <- paste(
    readLines(file.path(unpack_dir, "xl/styles.xml")),
    collapse = "\n"
  )
  expect_match(styles_xml, "FFFF9900", fixed = TRUE)
  expect_match(styles_xml, "FF0066CC", fixed = TRUE)
  expect_match(styles_xml, "patternType=\"solid\"", fixed = TRUE)
})

test_that("xlsx_styles$get_border_id caches borders by signature", {
  doc <- read_xlsx()
  styles <- doc$styles

  id1 <- styles$get_border_id(
    top_style = "thin",
    top_color = "000000",
    bottom_style = "thick",
    bottom_color = "FF0000",
    left_style = "medium",
    left_color = "00FF00",
    right_style = "dashed",
    right_color = "0000FF"
  )
  expect_true(is.numeric(id1))
  expect_gte(id1, 0)

  # cache hit
  id2 <- styles$get_border_id(
    top_style = "thin",
    top_color = "000000",
    bottom_style = "thick",
    bottom_color = "FF0000",
    left_style = "medium",
    left_color = "00FF00",
    right_style = "dashed",
    right_color = "0000FF"
  )
  expect_equal(id1, id2)

  # different signature -> different id
  id3 <- styles$get_border_id(top_style = "thin", top_color = "000000")
  expect_false(id3 == id1)

  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)
  styles_xml <- paste(
    readLines(file.path(unpack_dir, "xl/styles.xml")),
    collapse = "\n"
  )
  # sides with styles and colours
  expect_match(styles_xml, "<top style=\"thin\"", fixed = TRUE)
  expect_match(styles_xml, "<bottom style=\"thick\"", fixed = TRUE)
  expect_match(styles_xml, "<left style=\"medium\"", fixed = TRUE)
  expect_match(styles_xml, "<right style=\"dashed\"", fixed = TRUE)
  expect_match(styles_xml, "FFFF0000", fixed = TRUE)
  # mandatory <diagonal/> empty placeholder
  expect_match(styles_xml, "<diagonal/>", fixed = TRUE)
})

test_that("sheet_write_data on block_list stacks fpars vertically", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "bl")
  bl <- block_list(
    fpar(ftext("line 1", fp_text(bold = TRUE))),
    fpar(ftext("line 2", fp_text(italic = TRUE)))
  )
  doc <- sheet_write_data(
    doc,
    value = bl,
    sheet = "bl",
    start_row = 5,
    start_col = 2
  )
  xml <- sheet_cells_xml(doc)
  expect_match(xml, "r=\"B5\"[^>]*t=\"inlineStr\"", fixed = FALSE)
  expect_match(xml, "r=\"B6\"[^>]*t=\"inlineStr\"", fixed = FALSE)
  expect_match(xml, "line 1", fixed = TRUE)
  expect_match(xml, "line 2", fixed = TRUE)
})

test_that("sheet_write_data default method errors on unsupported input", {
  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "x")
  expect_error(sheet_write_data(doc, value = 42, sheet = "x"), "method")
})

test_that("sheet_add_drawing.gg renders a ggplot as PNG and embeds it", {
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("ragg")

  gg <- ggplot2::ggplot(iris, ggplot2::aes(Sepal.Length, Sepal.Width)) +
    ggplot2::geom_point()

  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "plots")
  doc <- sheet_add_drawing(
    doc,
    value = gg,
    sheet = "plots",
    left = 1,
    top = 1,
    width = 4,
    height = 3
  )

  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)

  media <- list.files(file.path(unpack_dir, "xl/media"))
  expect_length(media, 1)
  expect_match(media[1], "\\.png$")

  drawings <- list.files(
    file.path(unpack_dir, "xl/drawings"),
    pattern = "\\.xml$"
  )
  expect_length(drawings, 1)
})

test_that("sheet_add_drawing(.external_img) embeds images", {
  img <- file.path(R.home("doc"), "html", "logo.jpg")
  skip_if_not(file.exists(img), message = "no sample image available")

  doc <- new_xlsx()
  doc <- add_sheet(doc, label = "pics")
  doc <- sheet_add_drawing(
    doc,
    sheet = "pics",
    value = external_img(img, width = 2, height = 1.5),
    left = 2,
    top = 1
  )
  doc <- sheet_add_drawing(
    doc,
    sheet = "pics",
    value = external_img(img, width = 3, height = 2),
    left = 2,
    top = 4
  )

  out <- print(doc, target = tempfile(fileext = ".xlsx"))
  unpack_dir <- tempfile()
  unpack_folder(out, unpack_dir)

  media <- list.files(file.path(unpack_dir, "xl/media"))
  expect_length(media, 2)

  drawings <- list.files(
    file.path(unpack_dir, "xl/drawings"),
    pattern = "\\.xml$"
  )
  expect_length(drawings, 1)

  drawing_xml <- paste(
    readLines(file.path(unpack_dir, "xl/drawings", drawings[1])),
    collapse = "\n"
  )
  # both anchors in the single drawing
  anchors <- regmatches(
    drawing_xml,
    gregexpr("<xdr:absoluteAnchor", drawing_xml, perl = TRUE)
  )[[1]]
  expect_length(anchors, 2)
})

library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, "Title", location = ph_location_type(type = "title"))
doc <- ph_with(
  doc,
  "Hello world",
  location = ph_location_template(top = 4, type = "title")
)
print(doc, target = tempfile(fileext = ".pptx"))

#' @export
#' @title Add a page break in a 'Word' document
#' @description Add a page break into an rdocx object
#' @param x an rdocx object
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @examples
#' doc <- read_docx()
#' doc <- body_add_break(doc)
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_break <- function(x, pos = "after") {
  str <- runs_to_p_wml(run_pagebreak(), add_ns = TRUE)
  body_add_xml(x = x, str = str, pos = pos)
}

#' @export
#' @title Add an image in a 'Word' document
#' @description Add an image into an rdocx object.
#' @inheritParams body_add_break
#' @param src image filename, the basename of the file must not contain any blank.
#' @param style paragraph style
#' @param width,height image size in units expressed by the unit argument.
#' Defaults to "in"ches.
#' @param unit One of the following units in which the width and height
#' arguments are expressed: "in", "cm" or "mm".
#' @example inst/examples/example_body_add_img.R
#' @family functions for adding content
body_add_img <- function(
  x,
  src,
  style = NULL,
  width,
  height,
  pos = "after",
  unit = "in"
) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  unit <- check_unit(unit, c("in", "cm", "mm"))

  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", src)

  if (file_type %in% ".svg") {
    if (!requireNamespace("rsvg")) {
      stop("package 'rsvg' is required to convert svg file to rasters.")
    }

    file <- tempfile(fileext = ".png")
    rsvg::rsvg_png(src, file = file)
    src <- file
    file_type <- ".png"
  }

  new_src <- tempfile(fileext = file_type)
  file.copy(src, to = new_src)

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  ext_img <- external_img(new_src, width = width, height = height, unit = unit)
  xml_elt <- runs_to_p_wml(ext_img, add_ns = TRUE, style_id = style_id)

  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add an external docx in a 'Word' document
#' @description Add content of a docx into an rdocx object.
#'
#' The function is using a 'Microsoft Word' feature: when the
#' document will be edited, the content of the file will be
#' inserted in the main document.
#'
#' This feature is unlikely to work as expected if the
#' resulting document is edited by another software. You
#' can use function [body_import_docx()] to import the content
#' as an alternative.
#'
#' The file is added when the method `print()` that
#' produces the final Word file is called, so don't remove
#' file defined with `src` before.
#' @inheritParams body_add_break
#' @param src docx filename, the path of the file must not contain
#' any '&' and the basename must not contain any space.
#' @examples
#' file1 <- tempfile(fileext = ".docx")
#' file2 <- tempfile(fileext = ".docx")
#' file3 <- tempfile(fileext = ".docx")
#' x <- read_docx()
#' x <- body_add_par(x, "hello world 1", style = "Normal")
#' print(x, target = file1)
#'
#' x <- read_docx()
#' x <- body_add_par(x, "hello world 2", style = "Normal")
#' print(x, target = file2)
#'
#' x <- read_docx(path = file1)
#' x <- body_add_break(x)
#' x <- body_add_docx(x, src = file2)
#' print(x, target = file3)
#' @export
#' @family functions for adding content
body_add_docx <- function(x, src, pos = "after") {
  if (grepl(" ", basename(src))) {
    stop("The basename of the file to be inserted cannot contain spaces.")
  }

  xml_elt <- to_wml(block_pour_docx(file = src), add_ns = TRUE)
  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @importFrom utils head
#' @export
#' @title Import an external docx in a 'Word' document
#' @description Import body content and footnotes of a Word document into an rdocx object.
#'
#' The function is similar to [body_add_docx()] but instead of adding
#' the content as an external object, the document is read and all its
#' content is appended to the target document.
#' @details
#' The following operations are performed when importing a document:
#'
#' - Numberings are copied from the source document to the target document.
#' - Styles are not copied. If styles in the source document do not exist
#' in the target document, the style specified in the `par_style_mapping`,
#' `run_style_mapping` and `tbl_style_mapping` arguments will be used instead.
#' If no mapping is provided, the default style will be used and a warning is emitted.
#' @inheritParams body_add_break
#' @param src path to docx file to import
#' @param par_style_mapping,run_style_mapping,tbl_style_mapping
#' Named lists describing how to remap styles from the source document (`src`)
#' to styles available in the target document `x`. For each list entry, the name
#' of the element is the target style (in `x`), and the value is a character
#' vector of style names from the source document that should be replaced by
#' this target style.
#'
#' - `par_style_mapping`: applies to paragraph styles.
#' - `run_style_mapping`: applies to character (run) styles.
#' - `tbl_style_mapping`: applies to table styles.
#'
#' Examples:
#'
#' ```r
#' par_style_mapping = list(
#'   "Normal"    = c("List Paragraph", "Body Text"),
#'   "heading 1" = "Heading 1"
#' )
#' run_style_mapping = list(
#'   "Emphasis"  = c("Emphasis", "Italic")
#' )
#' tbl_style_mapping = list(
#'   "Normal Table" = c("Light Shading")
#' )
#' ```
#'
#' Use [styles_info()] to inspect available styles and verify their names.
#' @param prepend_chunks_on_styles A named list of run chunks to prepend
#' to runs with specific styles. The names of the list are paragraph style names
#' and the values run chunks to prepend. The first motivation for this argument
#' is to allow prepending of runs in paragraphs with a defined style, for example
#' to add a [run_autonum()] with all image captions.
#' @example inst/examples/example_body_import_docx.R
#' @family functions for adding content
body_import_docx <- function(
  x,
  src,
  par_style_mapping = list(),
  run_style_mapping = list(),
  tbl_style_mapping = list(),
  prepend_chunks_on_styles = list()
) {
  doc_from <- read_docx(src)
  sty_info_from <- styles_info(doc_from)
  sty_info_to <- styles_info(x)

  num_mapping <- numberings_append_xml(
    file_numbering_from = file.path(doc_from$package_dir, "word/numbering.xml"),
    file_numbering_to = file.path(x$package_dir, "word/numbering.xml")
  )

  ns_from <- xml_ns(doc_from$footnotes$get())
  ns_to <- xml_ns(x$footnotes$get())
  footnotes_chr <- doc_from$footnotes$patch_wml(
    package_dir = doc_from$package_dir,
    styles_info_tbl_from = sty_info_from,
    styles_info_tbl_to = sty_info_to,
    numbering_mapping = num_mapping,
    par_style_mapping = par_style_mapping,
    run_style_mapping = run_style_mapping,
    tbl_style_mapping = tbl_style_mapping,
    additional_ns = ns_from[setdiff(names(ns_from), names(ns_to))]
  )

  ns_from <- xml_ns(doc_from$doc_obj$get())
  ns_to <- xml_ns(x$doc_obj$get())
  body_chr <- doc_from$doc_obj$patch_wml(
    package_dir = doc_from$package_dir,
    styles_info_tbl_from = sty_info_from,
    styles_info_tbl_to = sty_info_to,
    numbering_mapping = num_mapping,
    par_style_mapping = par_style_mapping,
    run_style_mapping = run_style_mapping,
    tbl_style_mapping = tbl_style_mapping,
    additional_ns = ns_from[setdiff(names(ns_from), names(ns_to))],
    prepend_chunks_on_styles = prepend_chunks_on_styles
  )

  for (id in names(footnotes_chr)) {
    pat <- "<w:footnoteReference w:id=\"%s\"/>"
    pat <- sprintf(pat, id)
    match_pos <- grep(pat, body_chr, fixed = TRUE)
    body_chr[match_pos] <- unname(footnotes_chr[id])
  }

  # drop starting and ending tags for w:body
  body_chr <- head(body_chr, -1)
  body_chr <- tail(body_chr, -1)

  z <- body_append_start_context(
    x,
    additional_ns = ns_from[setdiff(names(ns_from), names(ns_to))]
  )
  cat(body_chr, sep = "\n", file = z$file_con, append = TRUE)
  x <- body_append_stop_context(z)
  x
}

#' @export
#' @title Add a 'ggplot' in a 'Word' document
#' @description Add a ggplot as a png image into an rdocx object.
#' @inheritParams body_add_break
#' @param value ggplot object
#' @param style paragraph style
#' @param width,height plot size in units expressed by the unit argument.
#' Defaults to a width of 6 and a height of 5 "in"ches.
#' @param unit One of the following units in which the width and height
#' arguments are expressed: "in", "cm" or "mm".
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @param ... Arguments to be passed to png function.
#' @importFrom grDevices dev.off
#' @examples
#' if (require("ggplot2")) {
#'   doc <- read_docx()
#'
#'   gg_plot <- ggplot(data = iris) +
#'     geom_point(mapping = aes(Sepal.Length, Petal.Length))
#'
#'   if (capabilities(what = "png")) {
#'     doc <- body_add_gg(doc, value = gg_plot, style = "centered")
#'
#'     # Set the unit in which the width and height arguments are expressed
#'     doc <- body_add_gg(doc, value = gg_plot, style = "centered", unit = "cm")
#'   }
#'
#'   print(doc, target = tempfile(fileext = ".docx"))
#' }
#' @family functions for adding content
#' @importFrom ragg agg_png
body_add_gg <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  pos = "after",
  unit = "in",
  ...
) {
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  stopifnot(inherits(value, "gg"))

  args <- sys.call()
  if ("units" %in% names(args[-1])) {
    cli::cli_abort(
      c("Found a {.arg units} argument. Did you mean {.arg unit}?")
    )
  }

  unit <- check_unit(unit, c("in", "cm", "mm"))

  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    scaling = scale,
    units = unit,
    res = res,
    background = "transparent",
    ...
  )
  print(value)
  dev.off()
  on.exit(unlink(file))
  body_add_img(
    x,
    src = file,
    style = style,
    width = width,
    height = height,
    pos = pos,
    unit = unit
  )
}


#' @export
#' @title Add a list of blocks into a 'Word' document
#' @description Add a list of blocks produced by [block_list()] into
#' into an rdocx object.
#' @inheritParams body_add_break
#' @param blocks set of blocks to be used as footnote content returned by
#'   function [block_list()].
#' @examples
#' library(officer)
#'
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#'
#' bl <- block_list(
#'   fpar(ftext("hello", shortcuts$fp_bold(color = "red"))),
#'   fpar(
#'     ftext("hello world", shortcuts$fp_bold()),
#'     external_img(src = img.file, height = 1.06, width = 1.39),
#'     fp_p = fp_par(text.align = "center")
#'   )
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_blocks(doc_1, blocks = bl)
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_blocks <- function(x, blocks, pos = "after") {
  stopifnot(inherits(blocks, "block_list"))

  if (length(blocks) > 0) {
    pos_vector <- rep("after", length(blocks))
    pos_vector[1] <- pos
    for (i in seq_along(blocks)) {
      x <- body_add_xml(
        x,
        str = to_wml(blocks[[i]], add_ns = TRUE),
        pos = pos_vector[i]
      )
    }
  }

  x
}


#' @export
#' @title Add paragraphs of text in a 'Word' document
#' @description Add a paragraph of text into an rdocx object
#' @param x a docx device
#' @param value a character
#' @param style paragraph style name
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "A title", style = "heading 1")
#' doc <- body_add_par(doc, "Hello world!", style = "Normal")
#' doc <- body_add_par(doc, "centered text", style = "centered")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_par <- function(x, value, style = NULL, pos = "after") {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"",
    style_id,
    "\"/></w:pPr><w:r><w:t xml:space=\"preserve\">",
    htmlEscapeCopy(value),
    "</w:t></w:r></w:p>"
  )
  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add fpar in a 'Word' document
#' @description Add an [fpar()] (a formatted paragraph)
#' into an rdocx object.
#' @param x a docx device
#' @param value a character
#' @param style paragraph style. If NULL, paragraph settings from `fpar` will be used. If not
#' NULL, it must be a paragraph style name (located in the template
#' provided as `read_docx(path = ...)`); in that case, paragraph settings from `fpar` will be
#' ignored.
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @examples
#' bold_face <- shortcuts$fp_bold(font.size = 30)
#' bold_redface <- update(bold_face, color = "red")
#' fpar_ <- fpar(
#'   ftext("Hello ", prop = bold_face),
#'   ftext("World", prop = bold_redface),
#'   ftext(", how are you?", prop = bold_face)
#' )
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, fpar_)
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#' # a way of using fpar to center an image in a Word doc ----
#' rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
#' img_in_par <- fpar(
#'   external_img(src = rlogo, height = 1.06 / 2, width = 1.39 / 2),
#'   hyperlink_ftext(
#'     href = "https://cran.r-project.org/index.html",
#'     text = "cran", prop = bold_redface
#'   ),
#'   fp_p = fp_par(text.align = "center")
#' )
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, img_in_par)
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#' @seealso [fpar()]
#' @family functions for adding content
body_add_fpar <- function(x, value, style = NULL, pos = "after") {
  if (!is.null(style)) {
    style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")
  } else {
    style_id <- NULL
  }

  xml_elt <- to_wml(value, add_ns = TRUE, style_id = style_id)

  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add table in a 'Word' document
#' @description Add a table into an rdocx object.
#' @param x a docx device
#' @param value a data.frame to add as a table
#' @param style table style
#' @param pos where to add the new element relative to the cursor,
#' one of after", "before", "on".
#' @param header display header if TRUE
#' @param alignment columns alignement, argument length must match with columns length,
#' values must be "l" (left), "r" (right) or "c" (center).
#' @param align_table table alignment within document, value must be "left", "center" or "right"
#' @param stylenames columns styles defined by [table_stylenames()]
#' @param first_row Specifies that the first column conditional formatting should be
#' applied. Details for this and other conditional formatting options can be found
#' at http://officeopenxml.com/WPtblLook.php.
#' @param last_row Specifies that the first column conditional formatting should be applied.
#' @param first_column Specifies that the first column conditional formatting should
#' be applied.
#' @param last_column Specifies that the first column conditional formatting should be
#' applied.
#' @param no_hband Specifies that the first column conditional formatting should be applied.
#' @param no_vband Specifies that the first column conditional formatting should be applied.
#' @examples
#' doc <- read_docx()
#' doc <- body_add_table(doc, iris, style = "table_template")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_table <- function(
  x,
  value,
  style = NULL,
  pos = "after",
  header = TRUE,
  alignment = NULL,
  align_table = "center",
  stylenames = table_stylenames(),
  first_row = TRUE,
  first_column = FALSE,
  last_row = FALSE,
  last_column = FALSE,
  no_hband = FALSE,
  no_vband = TRUE
) {
  pt <- prop_table(
    style = style,
    layout = table_layout(),
    width = table_width(),
    stylenames = stylenames,
    tcf = table_conditional_formatting(
      first_row = first_row,
      first_column = first_column,
      last_row = last_row,
      last_column = last_column,
      no_hband = no_hband,
      no_vband = no_vband
    ),
    align = align_table
  )

  bt <- block_table(
    x = value,
    header = header,
    properties = pt,
    alignment = alignment
  )
  xml_elt <- to_wml(bt, add_ns = TRUE, base_document = x)
  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add table of content in a 'Word' document
#' @description Add a table of content into an rdocx object.
#' The TOC will be generated by Word, if the document is not
#' edited with Word (i.e. Libre Office) the TOC will not be generated.
#' @param x an rdocx object
#' @param level max title level of the table
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @param style optional. style in the document that will be used to build entries of the TOC.
#' @param separator unused, no effect
#' @examples
#' doc <- read_docx()
#' doc <- body_add_toc(doc)
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_toc <- function(
  x,
  level = 3,
  pos = "after",
  style = NULL,
  separator = ";"
) {
  bt <- block_toc(level = level, style = style)
  out <- to_wml(bt, add_ns = TRUE)
  body_add_xml(x = x, str = out, pos = pos)
}


#' @export
#' @title Add plot in a 'Word' document
#' @description Add a plot as a png image into an rdocx object.
#' @inheritParams body_add_break
#' @param value plot instructions, see [plot_instr()].
#' @param style paragraph style
#' @param width,height plot size in units expressed by the unit argument.
#' Defaults to a width of 6 and a height of 5 "in"ches.
#' @param unit One of the following units in which the width and height
#' arguments are expressed: "in", "cm" or "mm".
#' @param res resolution of the png image in ppi
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @param ... Arguments to be passed to png function.
#' @importFrom grDevices dev.off
#' @examples
#' doc <- read_docx()
#'
#' if (capabilities(what = "png")) {
#'   p <- plot_instr(
#'       code = {
#'         barplot(1:5, col = 2:6)
#'       }
#'     )
#'
#'   doc <- body_add_plot(doc, value = p, style = "centered")
#'
#'   # Set the unit in which the width and height arguments are expressed
#'   doc <- body_add_plot(doc, value = p, style = "centered", unit = "cm")
#' }
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_plot <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  pos = "after",
  unit = "in",
  ...
) {
  unit <- check_unit(unit, c("in", "cm", "mm"))

  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = unit,
    res = res,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )
  on.exit(unlink(file))
  body_add_img(
    x,
    src = file,
    style = style,
    width = width,
    height = height,
    unit = unit,
    pos = pos
  )
}


#' @export
#' @title Add Word caption in a 'Word' document
#' @description Add a Word caption into an rdocx object.
#' @param x an rdocx object
#' @param value an object returned by [block_caption()]
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @family functions for adding content
#' @examples
#' doc <- read_docx()
#'
#' if (capabilities(what = "png")) {
#'   doc <- body_add_plot(doc,
#'     value = plot_instr(
#'       code = {
#'         barplot(1:5, col = 2:6)
#'       }
#'     ),
#'     style = "centered"
#'   )
#' }
#' run_num <- run_autonum(
#'   seq_id = "fig", pre_label = "Figure ",
#'   bkm = "barplot"
#' )
#' caption <- block_caption("a barplot",
#'   style = "Normal",
#'   autonum = run_num
#' )
#' doc <- body_add_caption(doc, caption)
#' print(doc, target = tempfile(fileext = ".docx"))
body_add_caption <- function(x, value, pos = "after") {
  stopifnot(inherits(value, "block_caption"))
  if (!value$style %in% x$styles$style_name) {
    stop(
      "caption is using style ",
      shQuote(value$style),
      " that does not exist in the Word document."
    )
  }
  out <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml(x = x, str = out, pos = pos)
}


#' @export
#' @title Add an xml string as document element
#' @description Add an xml string as document element in the document. This function
#' is to be used to add custom openxml code.
#' @param x an rdocx object
#' @param str a wml string
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @keywords internal
body_add_xml <- function(x, str, pos = c("after", "before", "on")) {
  xml_elt <- as_xml_document(str)
  pos <- match.arg(pos)

  cursor_elt <- docx_current_block_xml(x)
  if (is.null(cursor_elt)) {
    xml_add_child(
      xml_find_first(x$doc_obj$get(), "/w:document/w:body"),
      xml_elt,
      .where = 0
    )
    .name <- xml_name(xml_elt)
    x$officer_cursor <- cursor_append(x$officer_cursor, .name)
  } else if (pos == "on") {
    xml_replace(cursor_elt, xml_elt)
    x$officer_cursor <- cursor_replace_nodename(
      x$officer_cursor,
      xml_name(xml_elt)
    )
  } else if (pos == "after") {
    xml_add_sibling(cursor_elt, xml_elt, .where = pos)
    x$officer_cursor <- cursor_add_after(x$officer_cursor, xml_name(xml_elt))
  } else {
    xml_add_sibling(cursor_elt, xml_elt, .where = pos)
    x$officer_cursor <- cursor_add_before(x$officer_cursor, xml_name(xml_elt))
  }
  x
}

body_add_xml2 <- function(x, str) {
  x <- cursor_end(x)
  xml_elt <- as_xml_document(str)
  cursor_elt <- docx_current_block_xml(x)
  if (is.null(cursor_elt)) {
    xml_add_child(
      xml_find_first(x$doc_obj$get(), "/w:document/w:body"),
      xml_elt,
      .where = 0
    )
    x$officer_cursor <- cursor_append(x$officer_cursor, xml_name(xml_elt))
  } else {
    xml_add_sibling(cursor_elt, xml_elt, .where = "after")
    x$officer_cursor <- cursor_append(x$officer_cursor, xml_name(cursor_elt))
  }

  x
}

#' @export
#' @title Add bookmark in a 'Word' document
#' @description Add a bookmark at the cursor location. The bookmark
#' is added on the first run of text in the current paragraph.
#' @param x an rdocx object
#' @param id bookmark name
#' @examples
#'
#' # cursor_bookmark ----
#'
#' doc <- read_docx()
#' doc <- body_add_par(doc, "centered text", style = "centered")
#' doc <- body_bookmark(doc, "text_to_replace")
body_bookmark <- function(x, id) {
  cursor_elt <- docx_current_block_xml(x)
  ns_ <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\""
  new_id <- uuid_generate()
  id <- check_bookmark_id(id)

  bm_start_str <- sprintf(
    "<w:bookmarkStart w:id=\"%s\" w:name=\"%s\" %s/>",
    new_id,
    id,
    ns_
  )
  bm_start_end <- sprintf("<w:bookmarkEnd %s w:id=\"%s\"/>", ns_, new_id)

  path_ <- paste0(xml_path(cursor_elt), "//w:r")

  node <- xml_find_first(x$doc_obj$get(), path_)
  xml_add_sibling(node, as_xml_document(bm_start_str), .where = "before")
  xml_add_sibling(node, as_xml_document(bm_start_end), .where = "after")

  x
}

#' @export
#' @title Remove an element in a 'Word' document
#' @description Remove element pointed by cursor from a
#' 'Word' document.
#' @param x an rdocx object
#' @examples
#' library(officer)
#'
#' str1 <- rep("Lorem ipsum dolor sit amet, consectetur adipiscing elit. ", 20)
#' str1 <- paste(str1, collapse = "")
#'
#' str2 <- "Drop that text"
#'
#' str3 <- rep("Aenean venenatis varius elit et fermentum vivamus vehicula. ", 20)
#' str3 <- paste(str3, collapse = "")
#'
#' my_doc <- read_docx()
#' my_doc <- body_add_par(my_doc, value = str1, style = "Normal")
#' my_doc <- body_add_par(my_doc, value = str2, style = "centered")
#' my_doc <- body_add_par(my_doc, value = str3, style = "Normal")
#'
#' new_doc_file <- print(my_doc,
#'   target = tempfile(fileext = ".docx")
#' )
#'
#' my_doc <- read_docx(path = new_doc_file)
#' my_doc <- cursor_reach(my_doc, keyword = "that text")
#' my_doc <- body_remove(my_doc)
#'
#' print(my_doc, target = tempfile(fileext = ".docx"))
body_remove <- function(x) {
  cursor_elt <- docx_current_block_xml(x)

  if (is.null(cursor_elt)) {
    warning("There is nothing left to remove in the document")
    return(x)
  }
  x$officer_cursor <- cursor_delete(x$officer_cursor, xml_name(cursor_elt))
  xml_remove(cursor_elt)
  x
}

#' @export
#' @title Add comment in a 'Word' document
#' @description Add a comment at the cursor location. The comment
#' is added on the first run of text in the current paragraph.
#' @param x an rdocx object
#' @param cmt a set of blocks to be used as comment content returned by
#'   function [block_list()].
#' @param author comment author.
#' @param date comment date
#' @param initials comment initials
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "Paragraph")
#' doc <- body_comment(doc, block_list("This is a comment."))
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_comments(read_docx(docx_file))
body_comment <- function(
  x,
  cmt = ftext(""),
  author = "",
  date = "",
  initials = ""
) {
  cursor_elt <- docx_current_block_xml(x)
  ns_ <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\""

  open_tag <- wr_ns_yes

  blocks <- sapply(cmt, to_wml)
  blocks <- paste(blocks, collapse = "")

  id <- basename(tempfile(pattern = "comment"))

  cmt_xml <- paste0(
    sprintf(
      "<w:comment %s  w:id=\"%s\" w:author=\"%s\" w:date=\"%s\" w:initials=\"%s\">",
      ns_,
      id,
      author,
      date,
      initials
    ),
    blocks,
    "</w:comment>"
  )

  cmt_start_str <- sprintf(
    "<w:commentRangeStart w:officer=\"true\" w:id=\"%s\" %s/>",
    id,
    ns_
  )
  cmt_start_end <- sprintf(
    "<w:commentRangeEnd w:officer=\"true\" w:id=\"%s\" %s/>",
    id,
    ns_
  )

  path_ <- paste0(xml_path(cursor_elt), "//w:r")

  cmt_ref_xml <- paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:commentReference w:officer=\"true\" w:id=\"",
    id,
    "\">",
    cmt_xml,
    "</w:commentReference>",
    "</w:r>"
  )

  path_ <- paste0(xml_path(cursor_elt), "//w:r")

  node <- xml_find_first(x$doc_obj$get(), path_)
  xml_add_sibling(node, as_xml_document(cmt_start_str), .where = "before")
  xml_add_sibling(node, as_xml_document(cmt_start_end), .where = "after")
  xml_add_sibling(node, as_xml_document(cmt_ref_xml), .where = "after")

  x
}

# body_add and methods -----
#' @export
#' @title Add content into a Word document
#' @description This function adds objects into a Word document. Values are added
#' as new paragraphs or tables.
#'
#' This function is experimental and will replace the `body_add_*` functions
#' later. For now it is only to be used for successive additions and cannot
#' be used in conjunction with the `body_add_*` functions.
#' @param x an rdocx object
#' @param value object to add in the document. Supported objects
#' are vectors, data.frame, graphics, block of formatted paragraphs,
#' unordered list of formatted paragraphs,
#' pretty tables with package flextable, 'Microsoft' charts with package mschart.
#' @param ... further arguments passed to or from other methods. When
#' adding a `ggplot` object or [plot_instr], these arguments will be used
#' by png function. See method signatures to see what arguments can be used.
#' @example inst/examples/example_body_add_1.R
#' @section Illustrations:
#'
#' \if{html}{\figure{body_add_doc_1.png}{options: style="width:70\%;"}}
#' @keywords internal
body_add <- function(x, value, ...) {
  UseMethod("body_add", value)
}


#' @export
#' @describeIn body_add add a character vector.
#' @param style paragraph style name. These names are available with function [styles_info]
#' and are the names of the Word styles defined in the base document (see
#' argument `path` from [read_docx]).
body_add.character <- function(x, value, style = NULL, ...) {
  if (!is.null(style)) {
    style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")
  } else {
    style_id <- NULL
  }

  runs <- paste0(
    "<w:r><w:t xml:space=\"preserve\">",
    htmlEscapeCopy(value),
    "</w:t></w:r>"
  )

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"",
    style_id,
    "\"/></w:pPr>",
    runs,
    "</w:p>"
  )
  for (str in xml_elt) {
    x <- body_add_xml2(x = x, str = str)
  }
  x
}

#' @export
#' @describeIn body_add add a numeric vector.
#' @param format_fun a function to be used to format values.
body_add.numeric <- function(
  x,
  value,
  style = NULL,
  format_fun = formatC,
  ...
) {
  value <- format_fun(value, ...)
  body_add(x, value = value, style = style, ...)
}

#' @export
#' @describeIn body_add add a factor vector.
body_add.factor <- function(
  x,
  value,
  style = NULL,
  format_fun = as.character,
  ...
) {
  value <- format_fun(value)
  body_add(x, value = value, style = style, ...)
}


#' @export
#' @describeIn body_add add a [fpar] object. These objects enable
#' the creation of formatted paragraphs made of formatted chunks of text.
body_add.fpar <- function(x, value, style = NULL, ...) {
  img_src <- sapply(value$chunks, function(x) {
    if (inherits(x, "external_img")) {
      as.character(x)
    } else {
      NA_character_
    }
  })
  img_src <- unique(img_src[!is.na(img_src)])

  if (!is.null(style)) {
    style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")
  } else {
    style_id <- NULL
  }

  xml_elt <- to_wml(value, add_ns = TRUE, style_id = style_id)

  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @param header display header if TRUE
#' @param tcf conditional formatting settings defined by [table_conditional_formatting()]
#' @param alignment columns alignement, argument length must match with columns length,
#' values must be "l" (left), "r" (right) or "c" (center).
#' @describeIn body_add add a data.frame object with [block_table()].
body_add.data.frame <- function(
  x,
  value,
  style = NULL,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
) {
  pt <- prop_table(
    style = style,
    layout = table_layout(),
    width = table_width(),
    tcf = tcf
  )

  bt <- block_table(
    x = value,
    header = header,
    properties = pt,
    alignment = alignment
  )
  xml_elt <- to_wml(bt, add_ns = TRUE, base_document = x)

  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add a [block_caption] object. These objects enable
#' the creation of set of formatted paragraphs made of formatted chunks of text.
body_add.block_caption <- function(x, value, ...) {
  if (!value$style %in% x$styles$style_name) {
    stop(
      "caption is using style ",
      shQuote(value$style),
      " that does not exist in the Word document."
    )
  }

  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add a [block_list] object. Use this function
#' to add a list of block elements (e.g. paragraphs, images, tables)
#' into a Word document in a more efficient way than with usual
#' `body_add_*` functions. This function will add several elements
#' in a faster way because the cursor is not calculated for
#' each iteraction over the elements, as a consequence the
#' function only append elements at the end of the document
#' and does not allow to insert elements at a specific position.
body_add.block_list <- function(x, value, ...) {
  x <- cursor_end(x)
  cursor_elt <- docx_current_block_xml(x)
  for (i in rev(seq_along(value))) {
    str <- to_wml(value[[i]], add_ns = TRUE)
    xml_elt <- as_xml_document(str)
    if (is.null(cursor_elt)) {
      xml_add_child(
        xml_find_first(x$doc_obj$get(), "/w:document/w:body"),
        xml_elt,
        .where = 0
      )
    } else {
      xml_add_sibling(cursor_elt, xml_elt, .where = "after")
    }
  }
  cursor_end(x)
}

#' @export
#' @describeIn body_add add a table of content (a [block_toc] object).
body_add.block_toc <- function(x, value, ...) {
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add an image (a [external_img] object).
body_add.external_img <- function(x, value, style = "Normal", ...) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", value)

  if (file_type %in% ".svg") {
    if (!requireNamespace("rsvg")) {
      stop("package 'rsvg' is required to convert svg file to rasters.")
    }

    file <- tempfile(fileext = ".png")
    rsvg::rsvg_png(src, file = file)
    value <- external_img(
      src = file,
      width = attr(value, "dims")$width,
      height = attr(value, "dims")$height
    )
    file_type <- ".png"
  }

  new_src <- tempfile(fileext = file_type)
  file.copy(value, to = new_src)
  value[1] <- new_src

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"",
    style_id,
    "\"/></w:pPr>",
    to_wml(value, add_ns = FALSE),
    "</w:p>"
  )

  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @describeIn body_add add a [run_pagebreak] object.
body_add.run_pagebreak <- function(x, value, style = NULL, ...) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"",
    style_id,
    "\"/></w:pPr>",
    to_wml(value, add_ns = FALSE),
    "</w:p>"
  )

  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @describeIn body_add add a [run_columnbreak] object.
body_add.run_columnbreak <- function(x, value, style = NULL, ...) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"",
    style_id,
    "\"/></w:pPr>",
    to_wml(value, add_ns = FALSE),
    "</w:p>"
  )

  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add a ggplot object.
#' @param width,height plot size in units expressed by the unit argument.
#' Defaults to a width of 6 and a height of 5 "in"ches.
#' @param unit One of the following units in which the width and height
#' arguments are expressed: "in", "cm" or "mm".
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
body_add.gg <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  unit = "in",
  ...
) {
  if (!requireNamespace("ggplot2")) {
    cli::cli_abort("package ggplot2 is required to use this function")
  }

  args <- sys.call()
  if ("units" %in% names(args[-1])) {
    cli::cli_abort(
      c("Found a {.arg units} argument. Did you mean {.arg unit}?")
    )
  }

  unit <- check_unit(unit, c("in", "cm", "mm"))

  bl <- block_gg(
    value = value,
    fp_p = fp_par_lite(word_style = style),
    width = width,
    height = height,
    res = res,
    scale = scale,
    unit = unit
  )
  xml_elt <- to_wml(bl, add_ns = TRUE)
  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @describeIn body_add add a base plot with a [plot_instr] object.
body_add.plot_instr <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  unit = "in",
  ...
) {
  unit <- check_unit(unit, c("in", "cm", "mm"))

  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = unit,
    res = res,
    scaling = 1,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )
  on.exit(unlink(file))

  value <- external_img(src = file, width = width, height = height, unit = unit)

  body_add(x, value, style = style)
}

#' @export
#' @describeIn body_add pour content of an external docx file with with a [block_pour_docx] object
body_add.block_pour_docx <- function(x, value, ...) {
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add ends a section with a [block_section] object
body_add.block_section <- function(x, value, ...) {
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}

getncheck <- function(x, str) {
  child_ <- xml_child(x, str)
  expect_false(inherits(child_, "xml_missing"))
  child_
}

plot_with_unit_and_check <- function(
  value,
  body_add_fun,
  width = 6,
  height = 5,
  ...
) {
  .in_to_emu <- 914400
  .cm_to_emu <- 360000
  .mm_to_emu <- 36000

  x <- read_docx()
  x <- body_add_fun(x, value, width = width, height = height, ...)
  x <- body_add_fun(x, value, unit = "in", width = width, height = height, ...)
  x <- body_add_fun(x, value, unit = "cm", width = width, height = height, ...)
  x <- body_add_fun(x, value, unit = "mm", width = width, height = height, ...)
  x <- cursor_end(x)
  node <- docx_current_block_xml(x)

  expect_equal(
    as.numeric(
      xml_attr(xml_find_all(node, "//wp:extent"), "cx")
    ) /
      c(.in_to_emu, .in_to_emu, .cm_to_emu, .mm_to_emu),
    rep(width, 4)
  )
  expect_equal(
    as.numeric(
      xml_attr(xml_find_all(node, "//wp:extent"), "cy")
    ) /
      c(.in_to_emu, .in_to_emu, .cm_to_emu, .mm_to_emu),
    rep(height, 4)
  )
  # Non valid unit
  expect_error(
    body_add_fun(x, value, unit = "px", ...)
  )
  # Has "units="
  expect_error(
    body_add_fun(x, value, units = "cm", ...)
  )
  # Has multiple units
  expect_error(
    body_add_fun(x, value, unit = c("cm", "in", "mm"), ...)
  )
}

test_that("body_add_break", {
  x <- read_docx()
  x <- body_add_break(x)

  node <- docx_current_block_xml(x)
  expect_is(xml_child(node, "/w:r/w:br"), "xml_node")
})


test_that("body_end_sections", {
  x <- read_docx()
  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_end_section_landscape(x)

  node <- docx_current_block_xml(x)
  expect_false(inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing"))
  ps <- xml_child(node, "w:pPr/w:sectPr/w:pgSz")
  expect_false(inherits(ps, "xml_missing"))
  expect_equal(xml_attr(ps, "orient"), "landscape")

  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_add_par(x, "paragraph 2", style = "Normal")
  x <- body_end_section_columns(x)

  outfile <- tempfile(fileext = ".docx")
  print(x, target = outfile)
  x <- read_docx(outfile)

  node <- docx_current_block_xml(x)
  expect_false(inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing"))
  sect <- xml_child(node, "w:pPr/w:sectPr")

  expect_false(inherits(sect, "xml_missing"))
  expect_false(inherits(xml_child(sect, "w:cols"), "xml_missing"))

  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_add_par(x, "paragraph 2", style = "Normal")
  x <- body_end_section_columns_landscape(x)

  node <- docx_current_block_xml(x)
  expect_false(inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing"))

  ps <- xml_child(node, "w:pPr/w:sectPr/w:pgSz")
  expect_false(inherits(ps, "xml_missing"))
  expect_equal(xml_attr(ps, "orient"), "landscape")

  sect <- xml_child(node, "w:pPr/w:sectPr")
  expect_false(inherits(sect, "xml_missing"))
  expect_false(inherits(xml_child(sect, "w:cols"), "xml_missing"))

  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_add_par(x, "paragraph 2", style = "Normal")
  x <- body_end_section_portrait(x)

  outfile <- tempfile(fileext = ".docx")
  print(x, target = outfile)

  x <- read_docx(outfile)
  node <- docx_current_block_xml(x)
  expect_false(inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing"))

  ps <- xml_child(node, "w:pPr/w:sectPr/w:pgSz")
  expect_false(inherits(ps, "xml_missing"))
  expect_equal(xml_attr(ps, "orient"), "portrait")
})


test_that("body_add_toc", {
  x <- read_docx()
  x <- body_add_par(x, "paragraph 1")
  x <- body_add_toc(x)

  node <- docx_current_block_xml(x)

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='begin']")

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='end']")

  child_ <- getncheck(node, "w:r/w:instrText")
  expect_equal(xml_text(child_), "TOC \\o \"1-3\" \\h \\z \\u")

  x <- body_add_toc(x, style = "Normal")
  node <- docx_current_block_xml(x)

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='begin']")

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='end']")

  child_ <- getncheck(node, "w:r/w:instrText")
  expect_equal(xml_text(child_), "TOC \\h \\z \\t \"Normal\"")

  expect_output(print(block_toc(level = 2)), "TOC - max level: 2")
  expect_output(
    print(block_toc(level = 2, style = "Normal")),
    "TOC for style: Normal"
  )
  expect_output(
    print(block_toc(level = 2, seq_id = "tab")),
    "TOC for seq identifier: tab"
  )

  expect_match(
    to_wml(block_toc(seq_id = "tab")),
    "TOC \\\\h \\\\z \\\\c \"tab\""
  )
  expect_match(
    to_wml(block_toc(style = "Normal")),
    "TOC \\\\h \\\\z \\\\t \"Normal\""
  )
  expect_match(
    to_wml(block_toc(level = 2)),
    "TOC \\\\o \"1-2\" \\\\h \\\\z \\\\u"
  )
})

test_that("body_add_img", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  x <- read_docx()
  x <- body_add_img(x, img.file, width = 2.5, height = 1.3)

  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:drawing")
})

test_that("body_add_img with units", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")

  plot_with_unit_and_check(img.file, body_add_img, height = 2.69, width = 3.53)
})

test_that("external_img add", {
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  x <- read_docx()
  x <- body_add_fpar(
    x = x,
    value = fpar(
      external_img(src = img.file, width = .3, height = .3)
    )
  )
  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:drawing")
})


test_that("ggplot add", {
  testthat::skip_if_not(requireNamespace("ggplot2", quietly = TRUE))
  library("ggplot2")

  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))
  x <- read_docx()
  x <- body_add_gg(x, value = gg_plot, style = "centered")
  x <- cursor_end(x)
  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:drawing")
})

test_that("ggplot add with unit", {
  testthat::skip_if_not(requireNamespace("ggplot2", quietly = TRUE))
  library("ggplot2")

  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))

  plot_with_unit_and_check(gg_plot, body_add_gg)
  plot_with_unit_and_check(gg_plot, body_add)
})

test_that("plot add with unit", {
  base_plot <- plot_instr(
    code = {
      barplot(1:5, col = 2:6)
    }
  )

  # Base plot errors for "mm" with default pointsize of 12.
  plot_with_unit_and_check(base_plot, body_add_plot, pointsize = 1)
  plot_with_unit_and_check(base_plot, body_add, pointsize = 1)
})

test_that("fpar add", {
  bold_face <- shortcuts$fp_bold(font.size = 20)
  bold_redface <- update(bold_face, color = "red")
  fpar_ <- fpar(
    ftext("This is a big ", prop = bold_face),
    ftext("text", prop = bold_redface)
  )
  fpar_ <- update(fpar_, fp_p = fp_par(text.align = "center"))
  x <- read_docx()
  x <- body_add_fpar(x, fpar_)

  node <- docx_current_block_xml(x)
  expect_equal(xml_text(node), "This is a big text")

  x <- read_docx()
  try(
    {
      x <- body_add_fpar(x, fpar_, style = "centered")
    },
    silent = TRUE
  )
  expect_is(x, "rdocx")
})
test_that("svg add", {
  skip_if_not_installed("rsvg")
  srcfile <- file.path(R.home("doc"), "html", "Rlogo.svg")
  x <- read_docx()
  x <- body_add_fpar(x, fpar(external_img(srcfile)))
  path <- print(x, target = tempfile(fileext = ".docx"))
  x <- read_docx(path = path)
  node <- docx_current_block_xml(x)
  reldf <- x$doc_obj$rel_df()
  relidsvg <- reldf[grepl("\\.svg$", reldf$target), "id"]
  relidpng <- reldf[grepl("\\.png$", reldf$target), "id"]
  node_blip <- xml_child(
    node,
    "w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip"
  )
  expect_equal(xml_attr(node_blip, "embed"), relidpng)
  node_svgblip <- xml_child(
    node,
    "w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip/a:extLst/a:ext/asvg:svgBlip"
  )
  expect_equal(xml_attr(node_svgblip, "embed"), relidsvg)
})

test_that("add docx into docx", {
  file_1 <- tempfile(fileext = ".docx")
  img.file <- file.path(R.home("doc"), "html", "logo.jpg")
  doc <- read_docx()
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)
  print(doc, target = file_1)

  file_2 <- tempfile(fileext = ".docx")
  final_doc <- read_docx()
  doc <- body_add_docx(x = doc, src = file_1)
  print(doc, target = file_2)

  new_dir <- tempfile()
  unpack_folder(file_2, folder = new_dir)

  doc_parts <- read_xml(file.path(new_dir, "[Content_Types].xml"))
  doc_parts <- xml_find_all(
    doc_parts,
    "d1:Override[@ContentType='application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml']"
  )
  doc_parts <- xml_attr(doc_parts, "PartName")
  doc_parts <- basename(doc_parts)
  expect_equal(
    doc_parts[grepl("\\.docx$", doc_parts)],
    list.files(file.path(new_dir, "word"), pattern = "\\.docx$")
  )
})

test_that("Add comment at cursor position", {
  fp_bold <- fp_text_lite(bold = TRUE)
  fp_red <- fp_text_lite(color = "red")

  doc <- read_docx()
  doc <- body_add_par(doc, "This is a first Paragraph.")
  doc <- body_comment(
    doc,
    cmt = block_list("Comment on first par."),
    author = "Proofreader",
    date = Sys.Date()
  )
  doc <- body_add_fpar(
    doc,
    fpar("This is a second Paragraph. ", "This is a third Paragraph."),
    style = "Normal"
  )

  doc <- body_comment(
    doc,
    cmt = block_list(
      fpar(ftext("Comment on second par ...", fp_bold)),
      fpar(
        ftext("... with a second line.", fp_red)
      )
    ),
    author = "Proofreader 2",
    date = Sys.Date()
  )

  docx_file <- print(doc, target = tempfile(fileext = ".docx"))
  docx_dir <- tempfile()
  unpack_folder(docx_file, docx_dir)

  doc <- read_xml(file.path(docx_dir, "word/comments.xml"))
  comment1 <- xml_find_first(doc, "w:comment[@w:id='0']")
  comment2 <- xml_find_first(doc, "w:comment[@w:id='1']")

  expect_false(inherits(comment1, "xml_missing"))
  expect_false(inherits(comment2, "xml_missing"))

  expect_length(xml_children(comment1), 1)
  expect_length(xml_children(comment2), 2)
})


test_that("body_append_context - add content into docx", {
  doc <- read_docx()
  z <- body_append_start_context(doc)
  for (i in seq_len(10)) {
    write_elements_to_context(
      context = z,
      fpar("Hello World, ", i, fp_p = fp_par(word_style = "Normal")),
      fpar(run_pagebreak())
    )
  }
  doc <- body_append_stop_context(z)
  all_added_pars <- xml_find_all(
    docx_body_xml(doc),
    "w:body/w:p"
  )
  expect_length(all_added_pars, 20)
})


test_that("body_add R objects", {
  doc <- read_docx()
  doc <- body_add(doc, letters[1:5])
  doc <- body_add(doc, iris$Species[1:5])
  doc <- body_add(doc, 1:5)
  doc <- body_add(doc, fpar("hello"), style = "Normal")
  doc <- body_add(doc, head(iris), style = "table_template")
  docx_xml <- docx_body_xml(doc)

  tf <- print(doc, target = tempfile(fileext = ".docx"))

  doc <- read_docx(path = tf)
  doc_summary <- docx_summary(doc)

  txt <- doc_summary$text
  ref <- c(
    "a",
    "b",
    "c",
    "d",
    "e",
    "setosa",
    "setosa",
    "setosa",
    "setosa",
    "setosa",
    "1",
    "2",
    "3",
    "4",
    "5",
    "hello",
    "Sepal.Length",
    "Sepal.Width",
    "Petal.Length",
    "Petal.Width",
    "Species",
    "5.1",
    "3.5",
    "1.4",
    "0.2",
    "setosa",
    "4.9",
    "3.0",
    "1.4",
    "0.2",
    "setosa",
    "4.7",
    "3.2",
    "1.3",
    "0.2",
    "setosa",
    "4.6",
    "3.1",
    "1.5",
    "0.2",
    "setosa",
    "5.0",
    "3.6",
    "1.4",
    "0.2",
    "setosa",
    "5.4",
    "3.9",
    "1.7",
    "0.4",
    "setosa"
  )
  expect_equal(txt, ref)

  table_style_name <- doc_summary[
    !is.na(doc_summary$table_index),
  ]$table_stylename
  table_style_name <- unique(table_style_name)
  expect_equal(table_style_name, "table_template")
})

getncheck <- function(x, str){
  child_ <- xml_child(x, str)
  expect_false( inherits(child_, "xml_missing") )
  child_
}

test_that("body_add_break", {
  x <- read_docx()
  x <- body_add_break(x)

  node <- docx_current_block_xml(x)
  expect_is( xml_child(node, "/w:r/w:br"), "xml_node" )
})


test_that("body_end_sections", {

  x <- read_docx()
  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_end_section_landscape(x)

  node <- docx_current_block_xml(x)
  expect_false( inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing") )
  ps <- xml_child(node, "w:pPr/w:sectPr/w:pgSz")
  expect_false( inherits(ps, "xml_missing") )
  expect_equal( xml_attr(ps, "orient"), "landscape")

  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_add_par(x, "paragraph 2", style = "Normal")
  x <- body_end_section_columns(x)

  outfile <- tempfile(fileext = ".docx")
  print(x, target = outfile)
  x <- read_docx(outfile)

  node <- docx_current_block_xml(x)
  expect_false( inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing") )
  sect <- xml_child(node, "w:pPr/w:sectPr")

  expect_false( inherits(sect, "xml_missing") )
  expect_false( inherits(xml_child(sect, "w:cols"), "xml_missing") )

  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_add_par(x, "paragraph 2", style = "Normal")
  x <- body_end_section_columns_landscape(x)

  node <- docx_current_block_xml(x)
  expect_false( inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing") )

  ps <- xml_child(node, "w:pPr/w:sectPr/w:pgSz")
  expect_false( inherits(ps, "xml_missing") )
  expect_equal( xml_attr(ps, "orient"), "landscape")

  sect <- xml_child(node, "w:pPr/w:sectPr")
  expect_false( inherits(sect, "xml_missing") )
  expect_false( inherits(xml_child(sect, "w:cols"), "xml_missing") )

  x <- body_add_par(x, "paragraph 1", style = "Normal")
  x <- body_add_par(x, "paragraph 2", style = "Normal")
  x <- body_end_section_portrait(x)

  outfile <- tempfile(fileext = ".docx")
  print(x, target = outfile)

  x <- read_docx(outfile)
  node <- docx_current_block_xml(x)
  expect_false( inherits(xml_child(node, "w:pPr/w:sectPr"), "xml_missing") )

  ps <- xml_child(node, "w:pPr/w:sectPr/w:pgSz")
  expect_false( inherits(ps, "xml_missing") )
  expect_equal( xml_attr(ps, "orient"), "portrait")
})


test_that("body_add_toc", {

  x <- read_docx()
  x <- body_add_par(x, "paragraph 1")
  x <- body_add_toc(x)

  node <- docx_current_block_xml(x)

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='begin']")

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='end']")

  child_ <- getncheck(node, "w:r/w:instrText")
  expect_equal( xml_text(child_), "TOC \\o \"1-3\" \\h \\z \\u" )


  x <- body_add_toc(x, style = "Normal")
  node <- docx_current_block_xml(x)

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='begin']")

  child_ <- getncheck(node, "w:r/w:fldChar[@w:fldCharType='end']")

  child_ <- getncheck(node, "w:r/w:instrText")
  expect_equal( xml_text(child_), "TOC \\h \\z \\t \"Normal;1\"" )

  expect_output(print(block_toc(level = 2)), "TOC - max level: 2")
  expect_output(print(block_toc(level = 2, style = "Normal")), "TOC for style: Normal")
  expect_output(print(block_toc(level = 2, seq_id = "tab")), "TOC for seq identifier: tab")

  expect_match(to_wml(block_toc(seq_id = "tab")), "TOC \\\\h \\\\z \\\\c \"tab\"")
  expect_match(to_wml(block_toc(style = "Normal")), "TOC \\\\h \\\\z \\\\t \"Normal;1\"")
  expect_match(to_wml(block_toc(level = 2)), "TOC \\\\o \"1-2\" \\\\h \\\\z \\\\u")
})

test_that("body_add_img", {

  img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
  x <- read_docx()
  x <- body_add_img(x, img.file, width=2.5, height=1.3)

  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:drawing")
})

test_that("external_img add", {
  img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
  x <- read_docx()
  x <- body_add_fpar(
    x = x,
    value = fpar(
      external_img(src = img.file, width = .3, height = .3)
    )
  )
  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:drawing")
})

test_that("ggplot add", {
  testthat::skip_if_not(requireNamespace("ggplot2", quietly = TRUE))
  library("ggplot2")

  gg_plot <- ggplot(data = iris ) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))
  x <- read_docx()
  x <- body_add_gg(x, value = gg_plot, style = "centered" )
  x <- cursor_end(x)
  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:drawing")
})

test_that("fpar add", {
  bold_face <- shortcuts$fp_bold(font.size = 20)
  bold_redface <- update(bold_face, color = "red")
  fpar_ <- fpar(ftext("This is a big ", prop = bold_face),
                ftext("text", prop = bold_redface ) )
  fpar_ <- update(fpar_, fp_p = fp_par(text.align = "center"))
  x <- read_docx()
  x <- body_add_fpar(x, fpar_)

  node <- docx_current_block_xml(x)
  expect_equal(xml_text(node), "This is a big text" )

  x <- read_docx()
  try({x <- body_add_fpar(x, fpar_, style = "centered")}, silent = TRUE)
  expect_is(x, "rdocx")

})
test_that("svg add", {
  skip_if_not_installed("rsvg")
  srcfile <- file.path( R.home("doc"), "html", "Rlogo.svg" )
  x <- read_docx()
  x <- body_add_fpar(x, fpar(external_img(srcfile)))
  path <- print(x, target = tempfile(fileext = ".docx"))
  x <- read_docx(path = path)
  node <- docx_current_block_xml(x)
  reldf <- x$doc_obj$rel_df()
  relidsvg <- reldf[grepl("\\.svg$", reldf$target), "id"]
  relidpng <- reldf[grepl("\\.png$", reldf$target), "id"]
  node_blip <- xml_child(node, "w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip")
  expect_equal(xml_attr(node_blip, "embed"), relidpng)
  node_svgblip <- xml_child(node, "w:r/w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip/a:extLst/a:ext/asvg:svgBlip")
  expect_equal(xml_attr(node_svgblip, "embed"), relidsvg)
})

test_that("add docx into docx", {

  img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
  doc <- read_docx()
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39 )
  print(doc, target = "external_file.docx")

  final_doc <- read_docx()
  doc <- body_add_docx(x = doc, src = "external_file.docx" )
  print(doc, target = "final.docx")

  new_dir <- tempfile()
  unpack_folder("final.docx", folder = new_dir)

  doc_parts <- read_xml(file.path(new_dir, "[Content_Types].xml"))
  doc_parts <- xml_find_all(doc_parts, "d1:Override[@ContentType='application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml']")
  doc_parts <- xml_attr(doc_parts, "PartName")
  doc_parts <- basename(doc_parts)
  expect_equal(doc_parts[grepl("\\.docx$", doc_parts)],
               list.files(file.path(new_dir, "word"), pattern = "\\.docx$") )
})

test_that("Add comment at cursor position", {
  fp_bold <- fp_text_lite(bold = TRUE)
  fp_red <- fp_text_lite(color = "red")

  doc <- read_docx()
  doc <- body_add_par(doc, "This is a first Paragraph.")
  doc <- body_comment(doc,
                      cmt = block_list("Comment on first par."),
                      author = "Proofreader",
                      date = Sys.Date()
  )
  doc <- body_add_fpar(
    doc,
    fpar("This is a second Paragraph. ", "This is a third Paragraph."),
    style = "Normal"
  )

  doc <- body_comment(doc,
                      cmt = block_list(
                        fpar(ftext("Comment on second par ...", fp_bold)),
                        fpar(
                          ftext("... with a second line.", fp_red)
                        )
                      ),
                      author = "Proofreader 2",
                      date = Sys.Date()
  )

  docx_file <- print(doc, target = tempfile(fileext = ".docx"))
  docx_dir <- tempfile()
  unpack_folder(docx_file, docx_dir)

  doc <- read_xml(file.path(docx_dir, "word/comments.xml"))
  comment1 <- xml_find_first(doc, "w:comment[@w:id='0']")
  comment2 <- xml_find_first(doc, "w:comment[@w:id='1']")

  expect_false(inherits(comment1, "xml_missing"))
  expect_false(inherits(comment2, "xml_missing"))

  expect_length(xml_children(comment1), 1)
  expect_length(xml_children(comment2), 2)
})

test_that("docx summary", {
  example_docx <- system.file(package = "officer", "doc_examples/example.docx")

  doc <- read_docx(path = example_docx)
  doc_data <- docx_summary(doc)
  table_data <- doc_data[!is.na(doc_data$table_index) & doc_data$is_header, ]
  expect_equal(table_data$text, c("Petals", "Internode", "Sepal", "Bract"))

  doc_data_pr <- docx_summary(doc)
  table_data <- doc_data_pr[!is.na(doc_data_pr$table_index), ]
  expect_equal(tail(doc_data_pr, 2)[["text"]], c("Note", "New line note"))

  doc <- read_docx()
  doc <- body_add_fpar(
    x = doc,
    value = fpar(run_word_field(field = "Date \\@ \"MMMM d yyyy\""))
  )
  x <- print(doc, tempfile(fileext = ".docx"))
  doc <- read_docx(x)
  expect_equal(docx_summary(doc)$text, "Date \\@ \"MMMM d yyyy\"")
  doc <- read_docx(x)
  expect_equal(nrow(docx_summary(doc, remove_fields = TRUE)), 0)
})

test_that("complex docx table", {
  doc <- read_docx(path = "docs_dir/table-complex.docx")
  doc_data <- docx_summary(doc)
  table_data <- doc_data[!is.na(doc_data$table_index), ]
  table_data <- table_data[order(table_data$row_id, table_data$cell_id), ]

  first_row <- table_data[table_data$row_id %in% 1, ]

  expect_equal(
    first_row$text,
    c(
      "Column head",
      "column head",
      "column head",
      "column head",
      "x",
      "y"
    )
  )
  expect_equal(as.integer(first_row$col_span), c(1, 1, 1, 3, 1, 1))
  expect_equal(as.integer(first_row$row_span), c(2, 2, 2, 1, 2, 2))
  expect_true(all(
    table_data$text[table_data$col_span < 1 | table_data$row_span < 1] %in%
      NA_character_
  ))
})

test_that("preserves non breaking hyphens", {
  doc <- read_docx()

  make_xml_elt <- function(x, y) {
    sprintf(
      paste0(
        officer:::wp_ns_yes,
        "<w:pPr><w:pStyle w:val=\"Normal\"/></w:pPr><w:r>",
        "<w:t xml:space=\"preserve\">%s</w:t>",
        "<w:noBreakHyphen/>",
        "<w:t xml:space=\"preserve\">%s</w:t>",
        "</w:r></w:p>"
      ),
      x,
      y
    )
  }
  doc <- officer:::body_add_xml(
    x = doc,
    str = make_xml_elt("Inspector", "General")
  )
  doc <- officer:::body_add_xml(
    x = doc,
    str = make_xml_elt("General", "Inspector")
  )
  docx_summary(doc)
  expect_equal(
    docx_summary(doc)[["text"]],
    c("Inspector\u002DGeneral", "General\u002DInspector")
  )
})

test_that("detailed summary", {
  doc <- read_docx()

  fpar_ <- fpar(
    ftext("Formatted ", prop = fp_text(bold = TRUE, color = "red")),
    ftext(
      "paragraph ",
      prop = fp_text(
        shading.color = "blue"
      )
    ),
    ftext(
      "with multiple runs.",
      prop = fp_text(italic = TRUE, font.size = 20, font.family = "Arial")
    )
  )

  doc <- body_add_fpar(doc, fpar_, style = "Normal")

  fpar_ <- fpar(
    "Unformatted ",
    "paragraph ",
    "with multiple runs."
  )

  doc <- body_add_fpar(doc, fpar_, style = "Normal")

  doc <- body_add_par(doc, "Single Run", style = "Normal")

  doc <- body_add_fpar(
    doc,
    fpar(
      "Single formatetd run ",
      fp_t = fp_text(bold = TRUE, color = "red")
    )
  )

  xml_elt <- paste0(
    officer:::wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"Normal\"/></w:pPr>",
    "<w:r><w:rPr></w:rPr><w:t>NA</w:t></w:r>",
    "<w:r><w:rPr><w:b/><w:i/></w:rPr><w:t>toggle</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"0\"/><w:i w:val=\"0\"/></w:rPr><w:t>0</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"1\"/><w:i w:val=\"1\"/></w:rPr><w:t>1</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"false\"/><w:i w:val=\"false\"/></w:rPr><w:t>false</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"true\"/><w:i w:val=\"true\"/></w:rPr><w:t>true</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"off\"/><w:i w:val=\"off\"/></w:rPr><w:t>off</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"on\"/><w:i w:val=\"on\"/></w:rPr><w:t>on</w:t></w:r>",
    "</w:p>"
  )

  doc <- officer:::body_add_xml(
    x = doc,
    str = xml_elt
  )

  # #662
  xml_elt <- paste0(
    officer:::wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"Normal\"/></w:pPr>",
    "<w:r><w:rPr><w:u/><w:sz/><w:szCs/><w:color/><w:shd/></w:rPr><w:t>no attributes</w:t></w:r>",
    "<w:r><w:rPr><w:shd w:val=\"clear\"/></w:rPr><w:t>shading wo color and fill</w:t></w:r>",
    "</w:p>"
  )

  doc <- officer:::body_add_xml(
    x = doc,
    str = xml_elt
  )

  doc_sum <- docx_summary(doc, detailed = TRUE)

  expect_true("run_content_text" %in% names(doc_sum))
  expect_equal(nrow(doc_sum), 18L)

  expect_true(is.logical(doc_sum$bold))
  expect_true(is.logical(doc_sum$italic))
  expect_true(is.integer(doc_sum$sz))
  expect_true(is.integer(doc_sum$sz_cs))
  expect_true(is.logical(doc_sum$underline))
  expect_true(is.character(doc_sum$color))
  expect_true(is.character(doc_sum$shading))
  expect_true(is.character(doc_sum$shading_color))
  expect_true(is.character(doc_sum$shading_fill))
})


test_that("pptx summary", {
  example_pptx <- system.file(package = "officer", "doc_examples/example.pptx")
  doc <- read_pptx(path = example_pptx)
  doc_data <- pptx_summary(doc)
  table_data <- subset(
    doc_data,
    content_type %in% "table cell" & row_id == 1 & slide_id == 1
  )
  expect_equal(table_data$text, c("Header 1 ", "Header 2", "Header 3"))

  table_data <- subset(
    doc_data,
    content_type %in% "table cell" & row_id == 4 & slide_id == 1
  )
  expect_equal(table_data$text, c("B", "9.0", "Salut"))

  example_pptx <- "docs_dir/table-complex.pptx"
  doc <- read_pptx(path = example_pptx)
  doc_data <- pptx_summary(doc)
  table_data <- subset(
    doc_data,
    content_type %in% "table cell" & cell_id == 3 & slide_id == 1
  )
  expect_equal(
    table_data$text,
    c(
      "Header 3",
      NA,
      "blah blah blah",
      "Salut",
      "Hello",
      "sisi",
      NA,
      NA,
      NA,
      NA
    )
  )
  expect_equal(table_data$row_span, c(1, 1, 1, 1, 1, 4, 0, 0, 0, 1))
  expect_equal(table_data$col_span, c(1, 0, 1, 1, 1, 1, 1, 1, 1, 0))

  slide2_data <- subset(doc_data, content_type %in% "paragraph" & slide_id == 2)
  expect_equal(slide2_data$text, c("coco", "line of text", "blah blah blah"))

  doc_data_pr <- pptx_summary(doc, preserve = TRUE)
  expect_equal(doc_data_pr[23, ][["text"]], "blah\n \nblah\n \nblah")
})


test_that("empty slide summary", {
  example_pptx <- "docs_dir/test_empty.pptx"
  doc <- read_pptx(path = example_pptx)
  run <- try(pptx_summary(doc), silent = TRUE)
  expect_false(inherits(run, "try-error"))
})

test_that("body_import_docx messages", {
  tf <- tempfile(fileext = ".docx")

  x <- read_docx()
  testthat::expect_warning(
    x <- body_import_docx(
      x = x,
      src = system.file(
        package = "officer",
        "doc_examples",
        "example.docx"
      )
    ),
    "Style(s) mapping(s) for 'paragraphs' are missing in the body of the document:",
    fixed = TRUE
  )

  x <- read_docx()
  testthat::expect_no_message(
    x <- body_import_docx(
      x = x,
      src = system.file(
        package = "officer",
        "doc_examples",
        "example.docx"
      ),
      par_style_mapping = list(
        "Normal" = c("List Paragraph")
      ),
      tbl_style_mapping = list(
        "Normal Table" = "Light Shading"
      )
    )
  )
  docx_summary(x)
})

test_that("body_import_docx copy the content", {
  tf <- tempfile(fileext = ".docx")
  fi <- system.file(
    package = "officer",
    "doc_examples",
    "example.docx"
  )
  x <- read_docx()
  x <- body_import_docx(
    x = x,
    src = fi,
    par_style_mapping = list(
      "Normal" = c("List Paragraph")
    ),
    tbl_style_mapping = list(
      "Normal Table" = "Light Shading"
    )
  )
  print(x, target = tf)

  old_doc <- read_docx(fi)
  old_doc_head <- docx_summary(old_doc)

  new_doc <- read_docx(tf)
  new_doc_head <- docx_summary(new_doc)

  expect_equal(
    head(new_doc_head)[c("doc_index", "text")],
    head(old_doc_head)[c("doc_index", "text")]
  )
  expect_equal(
    head(new_doc_head)$style_name,
    c("heading 1", NA, "heading 1", "Normal", "Normal", "Normal")
  )
})

test_that("body_import_docx preserves abstractNum and num correspondence", {
  tf <- tempfile(fileext = ".docx")
  template_file <- "docs_dir/officer-lists.docx"
  # Create front document
  front <- read_docx()
  front <- body_add_par(front, "Frontmatter", style = "heading 1")
  tf_front <- tempfile(fileext = ".docx")
  print(front, target = tf_front)

  # Merge with document containing lists
  merged <- read_docx(tf_front)
  merged <- cursor_end(merged)
  merged <- body_import_docx(
    merged,
    par_style_mapping = list(
      "Normal" = c("First Paragraph", "Compact")
    ),
    src = template_file
  )
  print(merged, target = tf)

  # Read the front document and extract numbering information
  front_doc <- read_docx(tf_front)
  numbering_file <- file.path(front_doc$package_dir, "word/numbering.xml")
  front_numbering_xml <- xml2::read_xml(numbering_file)
  # Read the list document and extract numbering information
  list_doc <- read_docx(template_file)
  numbering_file <- file.path(list_doc$package_dir, "word/numbering.xml")
  list_numbering_xml <- xml2::read_xml(numbering_file)
  # Read the merged document and extract numbering information
  merged_doc <- read_docx(tf)
  numbering_file <- file.path(merged_doc$package_dir, "word/numbering.xml")
  merged_numbering_xml <- xml2::read_xml(numbering_file)

  # Extract abstractNum IDs
  abstractnum_nodes <- xml2::xml_find_all(
    merged_numbering_xml,
    "//w:abstractNum"
  )
  abstractnum_ids <- xml2::xml_attr(abstractnum_nodes, "abstractNumId")

  # Extract num elements and their associated abstractNumId references
  num_nodes <- xml2::xml_find_all(merged_numbering_xml, "//w:num")
  num_ids <- xml2::xml_attr(num_nodes, "numId")

  # For each num, get the referenced abstractNumId
  abstractnum_refs <- sapply(num_nodes, function(node) {
    ref_node <- xml2::xml_find_first(node, "w:abstractNumId")
    xml2::xml_attr(ref_node, "val")
  })

  # Verify that all referenced abstractNumIds exist in the document
  expect_true(
    all(abstractnum_refs %in% abstractnum_ids),
    info = "All num elements should reference existing abstractNum definitions"
  )

  # Verify that abstractNum IDs are unique
  expect_equal(
    length(abstractnum_ids),
    length(unique(abstractnum_ids)),
    info = "abstractNum IDs should be unique"
  )

  # Extract abstractNum IDs from front document
  front_abstractnum_nodes <- xml2::xml_find_all(
    front_numbering_xml,
    "//w:abstractNum"
  )
  front_abstractnum_ids <- xml2::xml_attr(
    front_abstractnum_nodes,
    "abstractNumId"
  )

  # Extract num IDs from front document
  front_num_nodes <- xml2::xml_find_all(front_numbering_xml, "//w:num")
  front_num_ids <- xml2::xml_attr(front_num_nodes, "numId")

  # Extract abstractNum IDs from list document
  list_abstractnum_nodes <- xml2::xml_find_all(
    list_numbering_xml,
    "//w:abstractNum"
  )
  list_abstractnum_ids <- xml2::xml_attr(
    list_abstractnum_nodes,
    "abstractNumId"
  )

  # Extract num IDs from list document
  list_num_nodes <- xml2::xml_find_all(list_numbering_xml, "//w:num")
  list_num_ids <- xml2::xml_attr(list_num_nodes, "numId")

  # Verify that the number of abstractNum in merged document equals
  # the sum from front and list documents
  expect_equal(
    length(abstractnum_ids),
    length(front_abstractnum_ids) + length(list_abstractnum_ids),
    info = paste(
      "Merged document should have",
      length(front_abstractnum_ids) + length(list_abstractnum_ids),
      "abstractNum elements"
    )
  )

  # Verify that the number of num in merged document equals
  # the sum from front and list documents
  expect_equal(
    length(num_ids),
    length(front_num_ids) + length(list_num_ids),
    info = paste(
      "Merged document should have",
      length(front_num_ids) + length(list_num_ids),
      "num elements"
    )
  )

  # Verify that abstractNum IDs in merged document are sequential iterations
  # Convert IDs to integers
  front_abstractnum_ids_int <- as.integer(front_abstractnum_ids)
  list_abstractnum_ids_int <- as.integer(list_abstractnum_ids)
  merged_abstractnum_ids_int <- as.integer(abstractnum_ids)

  # The merged document should have:
  # - All IDs from front document (unchanged)
  # - All IDs from list document (renumbered to follow front IDs)
  max_front_abstractnum <- if (length(front_abstractnum_ids_int) > 0) {
    max(front_abstractnum_ids_int)
  } else {
    -1L
  }

  # Expected IDs: front IDs + list IDs renumbered starting after max front ID
  expected_abstractnum_ids <- c(
    front_abstractnum_ids_int,
    seq_along(list_abstractnum_ids_int) + max_front_abstractnum
  )

  expect_equal(
    sort(merged_abstractnum_ids_int),
    sort(expected_abstractnum_ids),
    info = "abstractNum IDs should be properly renumbered in merged document"
  )

  unlink(tf_front)
  unlink(tf)
})

test_that("docx summary", {
  example_docx <- system.file(package = "officer", "doc_examples/example.docx")

  doc <- read_docx(path = example_docx)
  doc_data <- docx_summary(doc)
  table_data <- subset(doc_data, content_type %in% "table cell" & is_header)
  expect_equal( table_data$text, c("Petals", "Internode", "Sepal", "Bract") )

  doc_data_pr <- docx_summary(doc, preserve = TRUE)
  expect_equal( doc_data_pr[28, ][["text"]], "Note\nNew line note" )

  doc <- read_docx()
  doc <- body_add_fpar(
    x = doc,
    value = fpar(run_word_field(field = "Date \\@ \"MMMM d yyyy\""))
  )
  x <- print(doc, tempfile(fileext = ".docx"))
  doc <- read_docx(x)
  expect_equal(docx_summary(doc)$text, "Date \\@ \"MMMM d yyyy\"")
  doc <- read_docx(x)
  expect_equal(docx_summary(doc, remove_fields = TRUE)$text, "")
})

test_that("complex docx table", {

  doc <- read_docx(path = "docs_dir/table-complex.docx" )
  doc_data <- docx_summary(doc)
  table_data <- doc_data[doc_data$content_type == "table cell", ]
  table_data <- table_data[order(table_data$row_id, table_data$cell_id),]

  first_row <- table_data[table_data$row_id %in% 1,]

  expect_equal( first_row$text, c("Column head", "column head", "column head", "column head",
                                  NA, NA, "x", "y") )
  expect_equal( first_row$col_span, c(1, 1, 1, 3, 0, 0, 1, 1) )
  expect_equal( first_row$row_span, c(2, 2, 2, 1, 1, 1, 2, 2) )
  expect_true( all( table_data$text[table_data$col_span < 1 | table_data$row_span < 1] %in% NA_character_ ) )
})

test_that("preserves non breaking hyphens", {
  doc <- read_docx()

  make_xml_elt <- function(x, y) {
    sprintf(
      paste0(
        officer:::wp_ns_yes,
        "<w:pPr><w:pStyle w:val=\"Normal\"/></w:pPr><w:r>",
        "<w:t xml:space=\"preserve\">%s</w:t>",
        "<w:noBreakHyphen/>",
        "<w:t xml:space=\"preserve\">%s</w:t>",
        "</w:r></w:p>"
      ), x, y
    )
  }
  doc <- officer:::body_add_xml(
    x = doc, str = make_xml_elt("Inspector", "General")
  )
  doc <- officer:::body_add_xml(
    x = doc, str = make_xml_elt("General", "Inspector")
  )
  expect_equal(
    docx_summary(doc)[["text"]],
    c("Inspector\u002DGeneral", "General\u002DInspector")
  )
})

test_that("detailed summary", {
  doc <- read_docx()

  fpar_ <- fpar(
    ftext("Formatted ", prop = fp_text(bold = TRUE, color = "red")),
    ftext("paragraph ", prop = fp_text(
      shading.color = "blue"
    )),
    ftext("with multiple runs.",
          prop = fp_text(italic = TRUE, font.size = 20, font.family = "Arial")
    )
  )

  doc <- body_add_fpar(doc, fpar_, style = "Normal")

  fpar_ <- fpar(
    "Unformatted ",
    "paragraph ",
    "with multiple runs."
  )

  doc <- body_add_fpar(doc, fpar_, style = "Normal")

  doc <- body_add_par(doc, "Single Run", style = "Normal")

  doc <- body_add_fpar(doc,
                       fpar(
                         "Single formatetd run ",
                         fp_t = fp_text(bold = TRUE, color = "red")
                       )
  )

  xml_elt <- paste0(
    officer:::wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"Normal\"/></w:pPr>",
    "<w:r><w:rPr></w:rPr><w:t>NA</w:t></w:r>",
    "<w:r><w:rPr><w:b/><w:i/></w:rPr><w:t>toggle</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"0\"/><w:i w:val=\"0\"/></w:rPr><w:t>0</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"1\"/><w:i w:val=\"1\"/></w:rPr><w:t>1</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"false\"/><w:i w:val=\"false\"/></w:rPr><w:t>false</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"true\"/><w:i w:val=\"true\"/></w:rPr><w:t>true</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"off\"/><w:i w:val=\"off\"/></w:rPr><w:t>off</w:t></w:r>",
    "<w:r><w:rPr><w:b w:val=\"on\"/><w:i w:val=\"on\"/></w:rPr><w:t>on</w:t></w:r>",
    "</w:p>"
  )

  doc <- officer:::body_add_xml(
    x = doc, str = xml_elt
  )

  doc_sum <- docx_summary(doc, detailed = TRUE)

  expect_true("run" %in% names(doc_sum))
  expect_type(doc_sum$run, "list")
  expect_equal(lengths(doc_sum$run), rep(11, 5))
  expect_equal(sapply(doc_sum$run, nrow), c(3, 3, 1, 1, 8))

  expect_true(all(sapply(doc_sum$run$bold, is.logical)))
  expect_true(all(sapply(doc_sum$run$italic, is.logical)))
  expect_true(all(sapply(doc_sum$run$sz, is.integer)))
  expect_true(all(sapply(doc_sum$run$szCs, is.integer)))
  expect_true(all(sapply(doc_sum$run$underline, is_character)))
  expect_true(all(sapply(doc_sum$run$color, is_character)))
  expect_true(all(sapply(doc_sum$run$shading, is_character)))
  expect_true(all(sapply(doc_sum$run$shading_color, is_character)))
  expect_true(all(sapply(doc_sum$run$shading_fill, is_character)))
})




test_that("pptx summary", {
  example_pptx <- system.file(package = "officer", "doc_examples/example.pptx")
  doc <- read_pptx(path = example_pptx)
  doc_data <- pptx_summary(doc)
  table_data <- subset(doc_data, content_type %in% "table cell" & row_id == 1 & slide_id == 1)
  expect_equal( table_data$text, c("Header 1 ", "Header 2", "Header 3") )

  table_data <- subset(doc_data, content_type %in% "table cell" & row_id == 4 & slide_id == 1)
  expect_equal( table_data$text, c("B", "9.0", "Salut") )

  example_pptx <- "docs_dir/table-complex.pptx"
  doc <- read_pptx(path = example_pptx)
  doc_data <- pptx_summary(doc)
  table_data <- subset(doc_data, content_type %in% "table cell" & cell_id == 3 & slide_id == 1)
  expect_equal( table_data$text,
                c("Header 3", NA, "blah blah blah", "Salut", "Hello", "sisi", NA, NA, NA, NA) )
  expect_equal( table_data$row_span,
                c(1, 1, 1, 1, 1, 4, 0, 0, 0, 1) )
  expect_equal( table_data$col_span,
                c(1, 0, 1, 1, 1, 1, 1, 1, 1, 0) )

  slide2_data <- subset(doc_data, content_type %in% "paragraph" & slide_id == 2)
  expect_equal( slide2_data$text, c("coco", "line of text", "blah blah blah") )

  doc_data_pr <- pptx_summary(doc, preserve = TRUE)
  expect_equal( doc_data_pr[23, ][["text"]], "blah\n \nblah\n \nblah" )
})


test_that("empty slide summary", {
  example_pptx <- "docs_dir/test_empty.pptx"
  doc <- read_pptx(path = example_pptx)
  run <- try(pptx_summary(doc), silent = TRUE)
  expect_false(inherits(run, "try-error"))
})

img.file <- file.path(R.home(component = "doc"), "html", "logo.jpg")
svg_file <- file.path(R.home(component = "doc"), "html", "Rlogo.svg")

ext_img <- external_img(img.file)
ext_svg <- external_img(svg_file)

test_that("add image in HTML", {
  expect_match(
    to_html(ext_svg),
    "<img style=\"vertical-align:middle;width:36px;height:14px;\" src=\"data:image/svg\\+xml;base64,"
  )
  expect_match(
    to_html(ext_img),
    "<img style=\"vertical-align:middle;width:36px;height:14px;\" src=\"data:image/jpeg;base64,"
  )
})

test_that("add image in docx", {
  x <- read_docx()
  x <- body_add_fpar(x, fpar(ext_img))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)
    node_blip <- xml_find_first(body, "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})

pic <- file.path(R.home("doc"), "html", "logo.jpg" )
base_dir <- tempfile()
file1 <- file.path(base_dir, "dir1", "logo1.jpg")
file2 <- file.path(base_dir, "dir2", "logo1.jpg")
file3 <- file.path(base_dir, "dir2", "logo2.jpg")
dir.create(file.path(base_dir, "dir1"), recursive = TRUE)
dir.create(file.path(base_dir, "dir2"), recursive = TRUE)
file.copy(pic, file1)
file.copy(pic, file2)
file.copy(pic, file3)

test_that("add multiple images in docx", {
  x <- read_docx()
  x <- body_add_img(x, src = file1, width = 1, height = 1)
  x <- body_add_img(x, src = file2, width = 1, height = 1)
  x <- body_add_img(x, src = file3, width = 1, height = 1)
  docx_file <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = docx_file)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)
    node_blip <- xml_find_all(body, "//a:blip")
    expect_length(node_blip, 3)
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})

test_that("add image in pptx", {
  x <- read_pptx()
  x <- add_slide(x)
  x <- ph_with(x, ext_img, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = filename)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {

    node_blip <- xml_find_first(slide$get(), "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})


test_that("add multiple images in pptx", {
  x <- read_pptx()
  x <- add_slide(x)
  x <- ph_with(x = x, value = external_img(src = file1),
               location = ph_location(left = 0),
               use_loc_size = FALSE)
  x <- ph_with(x = x, value = external_img(src = file2),
               location = ph_location(left = 3),
               use_loc_size = FALSE)
  x <- ph_with(x = x, value = external_img(src = file3),
               location = ph_location(left = 6),
               use_loc_size = FALSE)
  pptx_file <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = pptx_file)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()

  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 1)
  if (nrow(subset_rel) > 0) {
    body <- slide$get()
    node_blip <- xml_find_all(body, "//a:blip")
    expect_length(node_blip, 3)
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
  }
})

test_that("add svg in docx", {
  skip_if_not_installed("rsvg")
  x <- read_docx()
  x <- body_add_fpar(x, fpar(ext_svg))
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 2)
  if (nrow(subset_rel) > 0) {
    body <- docx_body_xml(x)
    node_blip <- xml_find_first(body, "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
    node_svgblip <- xml_find_first(body, "//asvg:svgBlip")
    expect_false(inherits(node_svgblip, "xml_missing"))
    expect_true(all(xml_attr(node_svgblip, "embed") %in% subset_rel$id))
  }
})

test_that("add svg in pptx", {
  skip_if_not_installed("rsvg")
  x <- read_pptx()
  x <- add_slide(x)
  x <- ph_with(x, ext_svg, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = filename)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()
  subset_rel <- rel_df[grepl("^http://schemas(.*)image$", rel_df$type), ]
  expect_true(nrow(subset_rel) == 2)
  if (nrow(subset_rel) > 0) {
    node_blip <- xml_find_first(slide$get(), "//a:blip")
    expect_false(inherits(node_blip, "xml_missing"))
    expect_true(all(xml_attr(node_blip, "embed") %in% subset_rel$id))
    node_svgblip <- xml_find_first(slide$get(), "//asvg:svgBlip")
    expect_false(inherits(node_svgblip, "xml_missing"))
    expect_true(all(xml_attr(node_svgblip, "embed") %in% subset_rel$id))
  }
})

test_that("file size does not inflate with identical images", {
  img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
  doc <- read_docx()
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39 )
  file1 <- print(doc, target = tempfile(fileext = ".docx"))
  doc <- read_docx(path = file1)
  doc <- body_remove(doc)
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39 )
  file2 <- print(doc, target = tempfile(fileext = ".docx"))
  expect_equal(file.size(file1), file.size(file2), tolerance = 10)
})

attr_chunk <- function(x) {
  if (!is.null(x) && length(x) > 0) {
    attribs <- paste0(names(x), "=", shQuote(x, type = "cmd"), collapse = " ")
    attribs <- paste0(" ", attribs)
  } else {
    attribs <- ""
  }
  attribs
}

# content_type -----
content_type <- R6Class(
  "content_type",
  public = list(
    initialize = function(package_dir) {
      private$filename <- file.path(package_dir, "[Content_Types].xml")

      doc <- read_xml(x = private$filename)
      ns <- xml_ns(doc)

      node_template <- xml_find_first(
        doc,
        "d1:Override[@ContentType='application/vnd.openxmlformats-officedocument.presentationml.template.main+xml']"
      )
      if (!inherits(node_template, "xml_missing")) {
        xml_attr(
          node_template,
          'ContentType'
        ) <- "application/vnd.openxmlformats-officedocument.presentationml.presentation.main+xml"
      }
      node_template <- xml_find_first(
        doc,
        "d1:Override[@ContentType='application/vnd.openxmlformats-officedocument.wordprocessingml.template.main+xml']"
      )
      if (!inherits(node_template, "xml_missing")) {
        xml_attr(
          node_template,
          'ContentType'
        ) <- "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml"
      }

      extension <- xml_find_all(
        doc,
        "//*[contains(local-name(), 'Default')]/@Extension",
        ns = ns
      )
      extension <- xml_text(extension)
      content_type <- xml_find_all(
        doc,
        "//*[contains(local-name(), 'Default')]/@ContentType",
        ns = ns
      )
      content_type <- xml_text(content_type)
      names(content_type) <- extension
      private$default <- content_type

      partname <- xml_find_all(
        doc,
        "//*[contains(local-name(), 'Override')]/@PartName",
        ns = ns
      )
      partname <- xml_text(partname)
      content_type <- xml_find_all(
        doc,
        "//*[contains(local-name(), 'Override')]/@ContentType",
        ns = ns
      )
      content_type <- xml_text(content_type)
      names(content_type) <- partname
      private$override <- content_type
    },

    add_slide = function(partname) {
      partname <- basename(partname)
      partname <- file.path("/ppt", "slides", partname)
      content_type <- setNames(
        "application/vnd.openxmlformats-officedocument.presentationml.slide+xml",
        partname
      )
      override <- c(private$override, content_type)
      private$override <- override
      self
    },
    add_notesSlide = function(partname) {
      partname <- basename(partname)
      partname <- file.path("/ppt", "notesSlides", partname)
      content_type <- setNames(
        "application/vnd.openxmlformats-officedocument.presentationml.notesSlide+xml",
        partname
      )
      override <- c(private$override, content_type)
      private$override <- override
      self
    },
    add_override = function(value) {
      if (!names(value) %in% names(private$override)) {
        override <- c(private$override, value)
        private$override <- override
      }
      self
    },
    remove_slide = function(partname) {
      id <- which(basename(names(private$override)) %in% basename(partname))
      private$override <- private$override[-id]
      self
    },
    add_ext = function(extension, type) {
      if (
        !type %in% private$default && !extension %in% names(private$default)
      ) {
        content_type <- setNames(type, extension)
        default <- c(private$default, content_type)
        private$default <- default
      }
      self
    },

    save = function() {
      self$add_ext(extension = "jpeg", type = "image/jpeg")
      self$add_ext(extension = "gif", type = "image/gif")
      self$add_ext(extension = "png", type = "image/png")
      self$add_ext(extension = "svg", type = "image/svg+xml")
      self$add_ext(extension = "bmp", type = "image/bmp")
      self$add_ext(extension = "emf", type = "image/x-emf")
      self$add_ext(extension = "wmf", type = "image/x-wmf")
      self$add_ext(extension = "tiff", type = "image/tiff")
      self$add_ext(extension = "pdf", type = "application/pdf")
      self$add_ext(extension = "jpg", type = "application/octet-stream")
      attribs <- attr_chunk(c(
        xmlns = "http://schemas.openxmlformats.org/package/2006/content-types"
      ))
      out <- paste0(
        "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
        "\n<Types",
        attribs,
        ">"
      )
      if (length(private$default) > 0) {
        default <- sprintf(
          "<Default Extension=\"%s\" ContentType=\"%s\"/>",
          names(private$default),
          private$default
        )
        default <- paste0(default, collapse = "")
        out <- paste0(out, default)
      }
      if (length(private$override) > 0) {
        override <- sprintf(
          "<Override PartName=\"%s\" ContentType=\"%s\"/>",
          names(private$override),
          private$override
        )
        override <- paste0(override, collapse = "")
        out <- paste0(out, override)
      }
      out <- paste0(out, "</Types>")
      cat(out, file = private$filename)
      self
    }
  ),
  private = list(
    filename = NULL,
    default = NULL,
    override = NULL
  )
)

get_all_attrs <- function(nodeset) {
    # extract all children per node
    children <- lapply(nodeset, xml_children)
    # extract all node names per child
    name <- lapply(children, xml_name)
    # extract all attributes per child
    attr <- lapply(children, xml_attrs)
    # set attribute names
    for (i in seq_along(attr)) {
      names(attr[[i]]) <- name[[i]]
    }
    attr
}

get_attr_val <- function(lst, node, attr = NULL) {
  if (is.null(attr)) attr <- "val"
  if (!length(lst)) return(NA_character_)

  vapply(lst, FUN.VALUE = "", function(node_attr) {
    a <- node_attr[[node]][attr]
    if (is.null(a)) NA_character_ else a
  })
}

# a faster way for constructing a data.frame
as_data_frame <- function(x, ...) {
  stopifnot(length(unique(lengths(x))) == 1L)
  structure(x, ..., row.names = seq_along(x[[1L]]), class = "data.frame")
}

read_rpr <- function(x) {
  node_rpr <- xml_child(x, "w:rPr")

  # check if empty
  not_empty <- lengths(node_rpr) > 0L

  # all elements to extract
  elem <- c(
    "font.size", "bold", "italic", "underlined", "color", "font.family",
    "vertical.align", "shading.color",
    "hansi.family", "eastasia.family", "cs.family", "bold.cs", "font.size.cs",
    "lang.val", "lang.eastasia", "lang.bidi"
  )

  # init results
  rpr <- replicate(length(elem), list(rep(NA_character_, length(node_rpr))))
  names(rpr) <- elem

  # directly return if all are empty
  if (!any(not_empty)) return(as_data_frame(rpr))

  # further reduce to only work on non-empty nodes and extract all attributes of
  # every child
  rpr_attr <- get_all_attrs(node_rpr[not_empty])

  # replace with actual attribute values
  rpr[["font.size"]][not_empty]       <- get_attr_val(rpr_attr, "sz")
  rpr[["bold"]][not_empty]            <- get_attr_val(rpr_attr, "b")
  rpr[["italic"]][not_empty]          <- get_attr_val(rpr_attr, "i")
  rpr[["underlined"]][not_empty]      <- get_attr_val(rpr_attr, "u")
  rpr[["color"]][not_empty]           <- get_attr_val(rpr_attr, "color")
  rpr[["font.family"]][not_empty]     <- get_attr_val(rpr_attr, "rFonts", "ascii")

  rpr[["vertical.align"]][not_empty]  <- get_attr_val(rpr_attr, "vertAlign")
  rpr[["shading.color"]][not_empty]   <- get_attr_val(rpr_attr, "shd", "fill")

  rpr[["hansi.family"]][not_empty]    <- get_attr_val(rpr_attr, "rFonts", "hAnsi")
  rpr[["eastasia.family"]][not_empty] <- get_attr_val(rpr_attr, "rFonts", "eastAsia")
  rpr[["cs.family"]][not_empty]       <- get_attr_val(rpr_attr, "rFonts", "cs")
  rpr[["bold.cs"]][not_empty]         <- get_attr_val(rpr_attr, "bCs")
  rpr[["font.size.cs"]][not_empty]    <- get_attr_val(rpr_attr, "szCs")

  rpr[["lang.val"]][not_empty]        <- get_attr_val(rpr_attr, "lang")
  rpr[["lang.eastasia"]][not_empty]   <- get_attr_val(rpr_attr, "lang", "eastAsia")
  rpr[["lang.bidi"]][not_empty]       <- get_attr_val(rpr_attr, "lang", "bidi")

  as_data_frame(rpr)
}

read_ppr <- function(x) {
  node_ppr <- xml_child(x, "w:pPr")

  # check if empty
  not_empty <- lengths(node_ppr) > 0L

  # all elements to extract
  elem <- c(
    "align", "keep_next", "line_spacing",
    "padding.bottom", "padding.top", "padding.left", "padding.right", "shading.color.par",
    "border.bottom.width", "border.bottom.color", "border.bottom.style",
    "border.top.width", "border.top.color", "border.top.style",
    "border.left.width", "border.left.color", "border.left.style",
    "border.right.width", "border.right.color", "border.right.style"
  )

  # init results
  ppr <- replicate(length(elem), list(rep(NA_character_, length(node_ppr))))
  names(ppr) <- elem

  # init values for non-character columns
  ppr[paste0("border.", c("bottom", "top", "left", "right"), ".width")] <- list(rep(NA_integer_, length(node_ppr)))
  ppr["line_spacing"] <- list(rep(NA_real_, length(node_ppr)))
  ppr["keep_next"] <- list(rep(FALSE, length(node_ppr)))

  # directly return if all are empty
  if (!any(not_empty)) return(as_data_frame(ppr))

  # further reduce to only work on non-empty nodes and extract all attributes of
  # every child
  ppr_attr <- get_all_attrs(node_ppr[not_empty])

  # NOTE: 'w:pBdr' has to be treated specially since it is nested inside 'w:pPr'
  bdr <- xml_child(node_ppr[not_empty], "w:pBdr")
  not_empty_bdr <- lengths(bdr) > 0L
  bdr_attr <- get_all_attrs(bdr[not_empty_bdr])

  # replace with actual attribute values
  ppr[["align"]][not_empty] <- get_attr_val(ppr_attr, "jc")
  ppr[["keep_next"]][not_empty] <- vapply(ppr_attr, function(attrs) "keepNext" %in% names(attrs), logical(1L))
  ppr[["line_spacing"]][not_empty] <- as.integer(get_attr_val(ppr_attr, "spacing", "line")) / 240L

  ppr[["padding.bottom"]][not_empty] <- get_attr_val(ppr_attr, "spacing", "after")
  ppr[["padding.top"]][not_empty] <- get_attr_val(ppr_attr, "spacing", "before")
  ppr[["padding.left"]][not_empty] <- get_attr_val(ppr_attr, "ind", "left")
  ppr[["padding.right"]][not_empty] <- get_attr_val(ppr_attr, "ind", "right")
  ppr[["shading.color.par"]][not_empty] <- get_attr_val(ppr_attr, "shd", "fill")

  ppr[["border.bottom.width"]][not_empty][not_empty_bdr] <- as.integer(get_attr_val(bdr_attr, "bottom", "sz")) / 8L
  ppr[["border.bottom.color"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "bottom", "color")
  ppr[["border.bottom.style"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "bottom", "val")

  ppr[["border.top.width"]][not_empty][not_empty_bdr] <- as.integer(get_attr_val(bdr_attr, "top", "sz")) / 8L
  ppr[["border.top.color"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "top", "color")
  ppr[["border.top.style"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "top", "val")

  ppr[["border.left.width"]][not_empty][not_empty_bdr] <- as.integer(get_attr_val(bdr_attr, "left", "sz")) / 8L
  ppr[["border.left.color"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "left", "color")
  ppr[["border.left.style"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "left", "val")

  ppr[["border.right.width"]][not_empty][not_empty_bdr] <- as.integer(get_attr_val(bdr_attr, "right", "sz")) / 8L
  ppr[["border.right.color"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "right", "color")
  ppr[["border.right.style"]][not_empty][not_empty_bdr] <- get_attr_val(bdr_attr, "right", "val")

  as_data_frame(ppr)
}

read_docx_styles <- function(package_dir) {
  styles_file <- file.path(package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)

  all_styles <- xml_find_all(doc, "/w:styles/w:style")

  ppr <- read_ppr(all_styles)
  rpr <- read_rpr(all_styles)

  main <- data.frame(stringsAsFactors = FALSE,
    style_type = xml_attr(all_styles, "type"),
    style_id = xml_attr(all_styles, "styleId"),
    style_name = xml_attr(xml_child(all_styles, "w:name"), "val"),
    base_on = xml_attr(xml_child(all_styles, "w:basedOn"), "val"),
    is_custom = xml_attr(all_styles, "customStyle") %in% "1",
    is_default = xml_attr(all_styles, "default") %in% "1"
  )
  out <- cbind(main, ppr, rpr)
  out
}

get_style_id <- function(data, style, type) {
  ref <- data[data$style_type == type, ]

  if (!style %in% ref$style_name) {
    t_ <- shQuote(ref$style_name, type = "sh")
    t_ <- paste(t_, collapse = ", ")
    t_ <- paste0("c(", t_, ")")
    stop("could not match any style named ", shQuote(style, type = "sh"), " in ", t_, call. = FALSE)
  }
  ref$style_id[ref$style_name == style]
}

# check properties helpers ----

check_set_geom <- function(x) {
  # http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html
  geom_types <- c(
    "line",
    "lineInv",
    "triangle",
    "rtTriangle",
    "rect",
    "diamond",
    "parallelogram",
    "trapezoid",
    "nonIsoscelesTrapezoid",
    "pentagon",
    "hexagon",
    "heptagon",
    "octagon",
    "decagon",
    "dodecagon",
    "star4",
    "star5",
    "star6",
    "star7",
    "star8",
    "star10",
    "star12",
    "star16",
    "star24",
    "star32",
    "roundRect",
    "round1Rect",
    "round2SameRect",
    "round2DiagRect",
    "snipRoundRect",
    "snip1Rect",
    "snip2SameRect",
    "snip2DiagRect",
    "plaque",
    "ellipse",
    "teardrop",
    "homePlate",
    "chevron",
    "pieWedge",
    "pie",
    "blockArc",
    "donut",
    "noSmoking",
    "rightArrow",
    "leftArrow",
    "upArrow",
    "downArrow",
    "stripedRightArrow",
    "notchedRightArrow",
    "bentUpArrow",
    "leftRightArrow",
    "upDownArrow",
    "leftUpArrow",
    "leftRightUpArrow",
    "quadArrow",
    "leftArrowCallout",
    "rightArrowCallout",
    "upArrowCallout",
    "downArrowCallout",
    "leftRightArrowCallout",
    "upDownArrowCallout",
    "quadArrowCallout",
    "bentArrow",
    "uturnArrow",
    "circularArrow",
    "leftCircularArrow",
    "leftRightCircularArrow",
    "curvedRightArrow",
    "curvedLeftArrow",
    "curvedUpArrow",
    "curvedDownArrow",
    "swooshArrow",
    "cube",
    "can",
    "lightningBolt",
    "heart",
    "sun",
    "moon",
    "smileyFace",
    "irregularSeal1",
    "irregularSeal2",
    "foldedCorner",
    "bevel",
    "frame",
    "halfFrame",
    "corner",
    "diagStripe",
    "chord",
    "arc",
    "leftBracket",
    "rightBracket",
    "leftBrace",
    "rightBrace",
    "bracketPair",
    "bracePair",
    "straightConnector1",
    "bentConnector2",
    "bentConnector3",
    "bentConnector4",
    "bentConnector5",
    "curvedConnector2",
    "curvedConnector3",
    "curvedConnector4",
    "curvedConnector5",
    "callout1",
    "callout2",
    "callout3",
    "accentCallout1",
    "accentCallout2",
    "accentCallout3",
    "borderCallout1",
    "borderCallout2",
    "borderCallout3",
    "accentBorderCallout1",
    "accentBorderCallout2",
    "accentBorderCallout3",
    "wedgeRectCallout",
    "wedgeRoundRectCallout",
    "wedgeEllipseCallout",
    "cloudCallout",
    "cloud",
    "ribbon",
    "ribbon2",
    "ellipseRibbon",
    "ellipseRibbon2",
    "leftRightRibbon",
    "verticalScroll",
    "horizontalScroll",
    "wave",
    "doubleWave",
    "plus",
    "flowChartProcess",
    "flowChartDecision",
    "flowChartInputOutput",
    "flowChartPredefinedProcess",
    "flowChartInternalStorage",
    "flowChartDocument",
    "flowChartMultidocument",
    "flowChartTerminator",
    "flowChartPreparation",
    "flowChartManualInput",
    "flowChartManualOperation",
    "flowChartConnector",
    "flowChartPunchedCard",
    "flowChartPunchedTape",
    "flowChartSummingJunction",
    "flowChartOr",
    "flowChartCollate",
    "flowChartSort",
    "flowChartExtract",
    "flowChartMerge",
    "flowChartOfflineStorage",
    "flowChartOnlineStorage",
    "flowChartMagneticTape",
    "flowChartMagneticDisk",
    "flowChartMagneticDrum",
    "flowChartDisplay",
    "flowChartDelay",
    "flowChartAlternateProcess",
    "flowChartOffpageConnector",
    "actionButtonBlank",
    "actionButtonHome",
    "actionButtonHelp",
    "actionButtonInformation",
    "actionButtonForwardNext",
    "actionButtonBackPrevious",
    "actionButtonEnd",
    "actionButtonBeginning",
    "actionButtonReturn",
    "actionButtonDocument",
    "actionButtonSound",
    "actionButtonMovie",
    "gear6",
    "gear9",
    "funnel",
    "mathPlus",
    "mathMinus",
    "mathMultiply",
    "mathDivide",
    "mathEqual",
    "mathNotEqual",
    "cornerTabs",
    "squareTabs",
    "plaqueTabs",
    "chartX",
    "chartStar",
    "chartPlus"
  )

  if (!x %in% geom_types) {
    stop(
      "'",
      x,
      "' must be a valid geometry\n\n",
      "A valid geometry has to be one of ",
      paste0(paste0("'", geom_types, "'"), collapse = ", "),
      "."
    )
  } else {
    return(x)
  }
}

# line ----

#' @title Line properties
#'
#' @description Create a `sp_line` object that describes
#' line properties.
#'
#' @param color line color - a single character value specifying
#' a valid color (e.g. "#000000" or "black").
#' @param lwd line width (in point) - 0 or positive integer value.
#' @param lty single character value specifying the line type.
#' Expected value is one of the following : default `'solid'`
#' or `'dot'` or `'dash'` or `'lgDash'`
#' or `'dashDot'` or `'lgDashDot'` or `'lgDashDotDot'`
#' or `'sysDash'` or `'sysDot'` or `'sysDashDot'`
#' or `'sysDashDotDot'`.
#' @param linecmpd single character value specifying the compound line type.
#' Expected value is one of the following : default `'sng'`
#' or `'dbl'` or `'tri'` or `'thinThick'`
#' or `'thickThin'`
#' @param lineend single character value specifying the line end style
#' Expected value is one of the following : default `'rnd'`
#' or `'sq'` or `'flat'`
#' @param linejoin single character value specifying the line join style
#' Expected value is one of the following : default `'round'`
#' or `'bevel'` or `'miter'`
#' @param headend a `sp_lineend` object specifying line head end style
#' @param tailend a `sp_lineend` object specifying line tail end style
#' @return a `sp_line` object
#' @example inst/examples/example_sp_line.R
#' @family functions for defining shape properties
#' @seealso [sp_lineend]
#' @export
sp_line <- function(
  color = "transparent",
  lwd = 1,
  lty = "solid",
  linecmpd = "sng",
  lineend = "rnd",
  linejoin = "round",
  headend = sp_lineend(type = "none"),
  tailend = sp_lineend(type = "none")
) {
  out <- list()
  out <- check_set_color(out, color)
  out <- check_set_numeric(out, lwd)
  out <- check_set_choice(
    obj = out,
    value = lty,
    choices = c(
      "solid",
      "dot",
      "dash",
      "lgDash",
      "dashDot",
      "lgDashDot",
      "lgDashDotDot",
      "sysDash",
      "sysDot",
      "sysDashDot",
      "sysDashDotDot"
    )
  )
  out <- check_set_choice(
    obj = out,
    value = linecmpd,
    choices = c("sng", "dbl", "tri", "thinThick", "thickThin")
  )
  out <- check_set_choice(
    obj = out,
    value = lineend,
    choices = c("rnd", "sq", "flat")
  )
  out <- check_set_choice(
    obj = out,
    value = linejoin,
    choices = c("round", "bevel", "miter")
  )
  out <- check_set_class(obj = out, value = headend, cl = "sp_lineend")
  out <- check_set_class(obj = out, value = tailend, cl = "sp_lineend")

  class(out) <- c("sp_line")

  out
}

#' @param x,object `sp_line` object
#' @param ... further arguments - not used
#' @rdname sp_line
#' @export
print.sp_line = function(x, ...) {
  out <- data.frame(
    color = x$color,
    lwd = x$lwd,
    lty = x$lty,
    linecmpd = x$linecmpd,
    lineend = x$lineend,
    linejoin = x$linejoin,
    headend = unclass(x$headend),
    tailend = unclass(x$tailend),
    stringsAsFactors = FALSE
  )
  print(out)
  invisible()
}

#' @rdname sp_line
#' @export
update.sp_line <- function(
  object,
  color,
  lwd,
  lty,
  linecmpd,
  lineend,
  linejoin,
  headend,
  tailend,
  ...
) {
  if (!missing(color)) {
    object <- check_set_color(object, color)
  }
  if (!missing(lwd)) {
    object <- check_set_numeric(object, lwd)
  }
  if (!missing(lty)) {
    object <- check_set_choice(
      obj = object,
      value = lty,
      choices = c(
        "solid",
        "dot",
        "dash",
        "lgDash",
        "dashDot",
        "lgDashDot",
        "lgDashDotDot",
        "sysDash",
        "sysDot",
        "sysDashDot",
        "sysDashDotDot"
      )
    )
  }
  if (!missing(linecmpd)) {
    object <- check_set_choice(
      obj = object,
      value = linecmpd,
      choices = c("sng", "dbl", "tri", "thinThick", "thickThin")
    )
  }
  if (!missing(lineend)) {
    object <- check_set_choice(
      obj = object,
      value = lineend,
      choices = c("rnd", "sq", "flat")
    )
  }

  if (!missing(linejoin)) {
    object <- check_set_choice(
      obj = object,
      value = linejoin,
      choices = c("round", "bevel", "miter")
    )
  }

  if (!missing(headend)) {
    object <- check_set_class(obj = object, value = headend, cl = "sp_lineend")
  }
  if (!missing(tailend)) {
    object <- check_set_class(obj = object, value = tailend, cl = "sp_lineend")
  }

  object
}

#' @export
#' @rdname sp_line
#' @param add_ns unused; kept for compatibility with the [to_pml()]
#' generic.
to_pml.sp_line <- function(x, add_ns = FALSE, ...) {
  ln_pml(x)
}

#' @title Line end properties
#'
#' @description Create a `sp_lineend` object that describes
#' line end properties.
#'
#' @param type single character value specifying the line end type.
#' Expected value is one of the following : default `'none'`
#' or `'triangle'` or `'stealth'` or `'diamond'`
#' or `'oval'` or `'arrow'`
#' @param width single character value specifying the line end width
#' Expected value is one of the following : default `'sm'`
#' or `'med'` or `'lg'`
#' @param length single character value specifying the line end length
#' Expected value is one of the following : default `'sm'`
#' or `'med'` or `'lg'`
#' @return a `sp_lineend` object
#' @example inst/examples/example_sp_lineend.R
#' @family functions for defining shape properties
#' @seealso [sp_line]
#' @export
sp_lineend <- function(type = "none", width = "med", length = "med") {
  out <- list()
  out <- check_set_choice(
    obj = out,
    value = type,
    choices = c("none", "triangle", "stealth", "diamond", "oval", "arrow")
  )
  out <- check_set_choice(
    obj = out,
    value = width,
    choices = c("sm", "med", "lg")
  )
  out <- check_set_choice(
    obj = out,
    value = length,
    choices = c("sm", "med", "lg")
  )

  class(out) <- c("sp_lineend")

  out
}

#' @param x,object `sp_lineend` object
#' @param ... further arguments - not used
#' @rdname sp_lineend
#' @export
print.sp_lineend = function(x, ...) {
  out <- data.frame(
    type = x$type,
    width = x$width,
    length = x$length
  )
  print(out)
  invisible()
}

#' @rdname sp_lineend
#' @export
update.sp_lineend <- function(object, type, width, length, ...) {
  if (!missing(type)) {
    object <- check_set_choice(
      obj = object,
      value = type,
      choices = c("none", "triangle", "stealth", "diamond", "oval", "arrow")
    )
  }
  if (!missing(width)) {
    object <- check_set_choice(
      obj = object,
      value = width,
      choices = c("sm", "med", "lg")
    )
  }
  if (!missing(length)) {
    object <- check_set_choice(
      obj = object,
      value = length,
      choices = c("sm", "med", "lg")
    )
  }

  object
}

# This example demonstrates the versatility of body_add()
# by showing how to add various content types to a Word document

# Create a new Word document
x <- read_docx()

# Add a heading for the table of contents section
x <- body_add(x, "Table of content", style = "heading 1")

# Insert an automatic table of contents
# This will list all headings in the document
x <- body_add(x, block_toc())

# Insert a page break to start content on a new page
x <- body_add(x, run_pagebreak())

# Add a main section heading
x <- body_add(x, "Iris Dataset Sample", style = "heading 1")

# Add a data.frame as a table
# The first 6 rows of the iris dataset will be displayed
x <- body_add(x, head(iris), style = "table_template")

# Add another section heading
x <- body_add(x, "Alphabetic List", style = "heading 1")

# Add a character vector as paragraphs
# Each letter will appear as a separate paragraph with "Normal" style
x <- body_add(x, letters, style = "Normal")

# Insert a continuous section break
# Content continues on the same page but section properties can change
x <- body_add(x, block_section(prop_section(type = "continuous")))

# Add a plot created with R base graphics
# plot_instr() captures the plotting code to generate the image
x <- body_add(x, plot_instr(code = barplot(1:5, col = 2:6)))

# Change to landscape orientation for the next section
# This creates a new section with landscape page layout
x <- body_add(
  x,
  block_section(prop_section(page_size = page_size(orient = "landscape")))
)

# Save the document to a file
output_file <- tempfile(fileext = ".docx")
print(x, target = output_file)

#' Detect and handle duplicate placeholder labels
#'
#' PowerPoint does not enforce unique placeholder labels in a layout.
#' Selecting a placeholder via its label using [ph_location_label] will throw
#' an error, if the label is not unique. [layout_dedupe_ph_labels] helps to detect,
#' rename, or delete duplicate placholder labels.
#'
#' @param x An `rpptx` object.
#' @param action Action to perform on duplicate placeholder labels. One of:
#' * `detect` (default) = show info on dupes only, make no changes
#' * `rename` = create unique labels. Labels are renamed by appending a sequential number
#'    separated by dot to duplicate labels. For example, `c("title", "title")` becomes `c("title.1", "title.2")`.
#' * `delete` = only keep one of the placeholders with a duplicate label
#' @param print_info Print action information (e.g. renamed placeholders) to console?
#'  Default is `FALSE`. Always `TRUE` for action `detect`.
#' @return A `rpptx` object (with modified placeholder labels).
#' @export
#' @examples
#' x <- read_pptx()
#' layout_dedupe_ph_labels(x)
#'
#' file <- system.file("doc_examples", "ph_dupes.pptx", package = "officer")
#' x <- read_pptx(file)
#' layout_dedupe_ph_labels(x)
#' layout_dedupe_ph_labels(x, "rename", print_info = TRUE)
#'
layout_dedupe_ph_labels <- function(x, action = "detect", print_info = FALSE) {
  if (!inherits(x, "rpptx")) {
    stop("'x' must be an 'rpptx' object", call. = FALSE)
  }
  action <- match.arg(action, c("detect", "rename", "delete"))
  layout_names <- x$slideLayouts$get_metadata()$filename
  xfrm_list <- lapply(
    layout_names,
    .dedupe_phs_in_layout,
    x = x,
    action = action
  )
  x <- reload_slidelayouts(x) # reinit slideLayouts to get processed ph labels [e.g. when calling x$slideLayouts$get_xfrm_data()]
  if (print_info | action == "detect") {
    .print_dedupe_info(x = x, xfrm_list = xfrm_list, action = action)
  }
  invisible(x)
}


# handle placeholder labels in a single layout
#
# layout_file: layout filename (e.g. "slideLayout1.xml").
# x: An `rpptx` object
#
# returns: Dataframe with placeholder info. Only needed for .print_dedupe_info()
.dedupe_phs_in_layout <- function(layout_file, x, action = "rename") {
  ph_label <- NULL
  if (!grepl("\\.xml$", layout_file, ignore.case = TRUE)) {
    stop("'layout_file' must be an .xml file", call. = FALSE)
  }
  action <- match.arg(action, c("detect", "rename", "delete"))
  layout <- x$slideLayouts$collection_get(layout_file)
  xfrm <- layout$xfrm()
  xfrm <- subset(
    xfrm,
    duplicated(ph_label) | duplicated(ph_label, fromLast = TRUE)
  )
  if (nrow(xfrm) == 0) {
    return()
  }
  xfrm <- transform(
    xfrm,
    ph_label_new = make_strings_unique(ph_label),
    delete_flag = duplicated(ph_label)
  ) # prepare once for all action types
  if (action == "detect") {
    return(xfrm) # no further action required
  } else if (action == "rename") {
    xfrm$delete_flag <- FALSE
  } else if (action == "delete") {
    xfrm$ph_label_new <- xfrm$ph_label
  }

  # rename label or delete ph shape
  layout_xml <- layout$get()
  for (i in 1L:nrow(xfrm)) {
    shape <- xml2::xml_find_first(
      layout_xml,
      sprintf("p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]", xfrm$id[i])
    )
    if (xfrm$delete_flag[i]) {
      xml2::xml_remove(shape)
    } else {
      nodes <- xml2::xml_find_first(shape, ".//p:cNvPr")
      xml2::xml_set_attr(nodes, "name", xfrm$ph_label_new[i])
    }
  }
  layout$save() # persist changes in slideout xml file
  xfrm
}


# reload slideLayouts (if layout XML in package_dir has changed)
reload_slidelayouts <- function(x) {
  x$slideLayouts$initialize(
    x$package_dir,
    master_metadata = x$masterLayouts$get_metadata(),
    master_xfrm = x$masterLayouts$xfrm()
  )
  x
}


# Create unique string by appending a sepatator and a number
# make_strings_unique(c("A", "B", "B", "C", "A"))
make_strings_unique <- function(x, sep = ".") {
  ii <- stats::ave(x, x, FUN = seq_along)
  paste0(x, sep, ii)
}


# helper mostly for testing
has_ph_dupes <- function(x) {
  if (!inherits(x, "rpptx")) {
    stop("'x' must be an 'rpptx' object", call. = FALSE)
  }
  xfrm <- x$slideLayouts$get_xfrm_data()
  dupes <- stats::aggregate(
    ph_label ~ master_name + name,
    data = xfrm,
    FUN = function(x) sum(duplicated(x)) > 0
  )
  any(dupes$ph_label)
}


# print info on what was done (if print_info = TRUE)
.print_dedupe_info <- function(x, xfrm_list, action) {
  .df_1 <- do.call(rbind, xfrm_list)
  if (is.null(.df_1)) {
    cat("No duplicate placeholder labels detected.")
    return(invisible(NULL))
  }
  .df_2 <- x$slideLayouts$get_xfrm_data()
  .df_2 <- unique(.df_2[, c("master_file", "master_name"), drop = FALSE])
  df <- merge(.df_1, .df_2, sort = FALSE)
  rownames(df) <- NULL
  df <- df[,
    c("master_name", "name", "ph_label", "ph_label_new", "delete_flag"),
    drop = FALSE
  ]
  colnames(df)[2] <- "layout_name"
  if (action == "detect") {
    cat("Placeholders with duplicate labels:\n")
    cat(cli::col_grey(
      "* 'ph_label_new' = new placeholder label for action = 'rename'\n"
    ))
    cat(cli::col_grey(
      "* 'delete_flag' = deleted placeholders for action = 'delete'\n"
    ))
  } else if (action == "rename") {
    df$delete_flag <- NULL
    cat("Renamed duplicate placeholder labels:\n")
    cat(cli::col_grey("* 'ph_label_new' = new placeholder label\n"))
  } else if (action == "delete") {
    df <- df[df$delete_flag, , drop = FALSE]
    df$ph_label_new <- NULL
    cat("Removed placeholders with duplicate labels:\n")
  }
  print(df)
}

test_that("incorrect input formats are detected", {
  expect_error(
    layout_dedupe_ph_labels("file.xxx"),
    regexp = "'x' must be an 'rpptx' object"
  )
  expect_error(
    has_ph_dupes("file.xxx"),
    regexp = "'x' must be an 'rpptx' object"
  )
  expect_error(
    .dedupe_phs_in_layout("file.xxx"),
    regexp = "'layout_file' must be an .xml file"
  )
})


test_that("handling ph dupes function works when there are none", {
  x <- read_pptx() # sample PPTX has no dupes
  expect_false(has_ph_dupes(x))
  . <- capture.output(expect_no_error({
    layout_dedupe_ph_labels(x, print_info = TRUE)
    layout_dedupe_ph_labels(x, action = "rename", print_info = TRUE)
    layout_dedupe_ph_labels(x, action = "delete", print_info = TRUE)
  }))
})


test_that("handling ph dupes works", {
  file_in <- test_path("docs_dir/test-pptx-dedupe-ph.pptx")

  # referencing a duplicate placeholder via its label should throw an error.
  # if this should change for some reason, the test fails as we would need to
  # check if deduplication is still relevant
  x <- read_pptx(file_in)
  x <- add_slide(x, layout = "2x2-dupes", master = "Master1")
  expect_no_error(ph_with(x, "abc", ph_location_label(ph_label = "Title 1")))
  expect_error(ph_with(x, "abc", ph_location_label(ph_label = "Content")))

  # action = detect
  x_det <- read_pptx(file_in)
  expect_true(has_ph_dupes(x_det))
  out <- capture.output({
    x_det <- layout_dedupe_ph_labels(x_det)
  })
  expect_true(any(grepl("Content 7.1", out)))
  expect_true(has_ph_dupes(x_det))

  # action = rename
  x_rename <- read_pptx(file_in)
  before <- x_rename$slideLayouts$get_xfrm_data()$ph_label
  out <- capture.output({
    x_rename <- layout_dedupe_ph_labels(
      x_rename,
      action = "rename",
      print_info = TRUE
    )
  })
  expect_true(any(grepl("Content 7", out)))
  expect_true(any(grepl("Content 7.1", out)))
  after <- x_rename$slideLayouts$get_xfrm_data()$ph_label
  expect_false(has_ph_dupes(x_rename))
  expect_true(any(before != after))
  expect_equal(length(before), length(after))

  # action = delete
  x_delete <- read_pptx(file_in)
  before <- x_delete$slideLayouts$get_xfrm_data()$ph_label
  out <- capture.output({
    x_delete <- layout_dedupe_ph_labels(
      x_delete,
      action = "delete",
      print_info = TRUE
    )
  })
  expect_true(any(grepl("Content 7", out)))
  after <- x_delete$slideLayouts$get_xfrm_data()$ph_label
  expect_false(has_ph_dupes(x_delete))
  expect_gt(length(before), length(after))

  # annotate base should work with deduped phs
  file <- tempfile(fileext = ".pptx")
  output_file <- tempfile(fileext = ".pptx")
  print(x_rename, target = file)
  expect_no_error(annotate_base(file, output_file = output_file))
  print(x_delete, target = file)
  expect_no_error(annotate_base(file, output_file = output_file))
})

test_that("incorrect input formats are detected", {
  expect_error(layout_dedupe_ph_labels("file.xxx"), regexp = "'x' must be an 'rpptx' object")
  expect_error(has_ph_dupes("file.xxx"), regexp = "'x' must be an 'rpptx' object")
  expect_error(.dedupe_phs_in_layout("file.xxx"), regexp = "'layout_file' must be an .xml file")
})



test_that("handling ph dupes function works when there are none", {
  x <- read_pptx() # sample PPTX has no dupes
  expect_false(has_ph_dupes(x))
  . <- capture.output(expect_no_error({
    layout_dedupe_ph_labels(x, print_info = TRUE)
    layout_dedupe_ph_labels(x, action = "rename", print_info = TRUE)
    layout_dedupe_ph_labels(x, action = "delete", print_info = TRUE)
  }))
})


test_that("handling ph dupes works", {
  file_in <- test_path("docs_dir/test-pptx-dedupe-ph.pptx")

  # referencing a duplicate placeholder via its label should throw an error.
  # if this should change for some reason, the test fails as we would need to
  # check if deduplication is still relevant
  x <- read_pptx(file_in)
  x <- add_slide(x, layout = "2x2-dupes", master = "Master1")
  expect_no_error(ph_with(x, "abc", ph_location_label(ph_label = "Title 1")))
  expect_error(ph_with(x, "abc", ph_location_label(ph_label = "Content")))

  # action = detect
  x_det <- read_pptx(file_in)
  expect_true(has_ph_dupes(x_det))
  out <- capture.output({
    x_det <- layout_dedupe_ph_labels(x_det)
  })
  expect_true(any(grepl("Content 7.1", out)))
  expect_true(has_ph_dupes(x_det))

  # action = rename
  x_rename <- read_pptx(file_in)
  before <- x_rename$slideLayouts$get_xfrm_data()$ph_label
  out <- capture.output({
    x_rename <- layout_dedupe_ph_labels(x_rename, action = "rename", print_info = TRUE)
  })
  expect_true(any(grepl("Content 7", out)))
  expect_true(any(grepl("Content 7.1", out)))
  after <- x_rename$slideLayouts$get_xfrm_data()$ph_label
  expect_false(has_ph_dupes(x_rename))
  expect_true(any(before != after))
  expect_equal(length(before), length(after))

  # action = delete
  x_delete <- read_pptx(file_in)
  before <- x_delete$slideLayouts$get_xfrm_data()$ph_label
  out <- capture.output({
    x_delete <- layout_dedupe_ph_labels(x_delete, action = "delete", print_info = TRUE)
  })
  expect_true(any(grepl("Content 7", out)))
  after <- x_delete$slideLayouts$get_xfrm_data()$ph_label
  expect_false(has_ph_dupes(x_delete))
  expect_gt(length(before), length(after))

  # annotate base should work with deduped phs
  file <- tempfile(fileext = ".pptx")
  output_file <- tempfile(fileext = ".pptx")
  print(x_rename, target = file)
  expect_no_error(annotate_base(file, output_file = output_file))
  print(x_delete, target = file)
  expect_no_error(annotate_base(file, output_file = output_file))
})

docx_part <- R6Class(
  "docx_part",
  inherit = openxml_document,
  public = list(

    initialize = function( path, main_file, cursor, body_xpath ) {
      super$initialize("word")
      private$package_dir <- path
      private$body_xpath <- body_xpath
      super$feed(file.path(private$package_dir, "word", main_file))
      private$cursor <- cursor
    },
    length = function( ){
      xml_length( xml_find_first(self$get(), private$body_xpath ) )
    }

  ),
  private = list(
    package_dir = NULL,
    cursor = NULL,
    body_xpath = NULL
  )

)

#' @title PowerPoint table to matrix
#' @description Convert the data in an a 'PowerPoint' table
#' to a matrix or all data to a list of matrices.
#' @param x The rpptx object to convert (as created by [officer::read_pptx()])
#' @param ... Ignored
#' @param slide_id The slide number to load from (NA indicates first slide with
#'   a table, NULL indicates all slides and all tables)
#' @param id The table ID to load from (ignored it `is.null(slide_id)`, NA
#'   indicates to load the first table from the `slide_id`)
#' @param span How should col_span/row_span values be handled? `NA` means
#'   to leave the value as `NA`, and `"fill"` means to fill matrix
#'   cells with the value.
#' @return A matrix with the data, or if `slide_id=NULL`, a list of
#'   matrices
#' @examples
#' library(officer)
#' pptx_file <- system.file(package="officer", "doc_examples", "example.pptx")
#' z <- read_pptx(pptx_file)
#' as.matrix(z, slide_id = NULL)
#' @export
as.matrix.rpptx <- function(
  x,
  ...,
  slide_id = NA_integer_,
  id = NA_character_,
  span = c(NA_character_, "fill")
) {
  span <- match.arg(span)
  d_summary_raw <- officer::pptx_summary(x)
  # filter to just table data
  d_summary_prep_table <- d_summary_raw[
    d_summary_raw$content_type %in% "table cell",
  ]
  stopifnot("No tables to extract" = nrow(d_summary_prep_table) > 0)
  d_summary_prep <-
    d_summary_prep_table[
      d_summary_prep_table$row_span > 0 & d_summary_prep_table$col_span > 0,
    ]
  stopifnot("All cells had col_span=0 or row_span=0" = nrow(d_summary_prep) > 0)
  if (is.null(slide_id)) {
    data_extract <- d_summary_prep
  } else {
    if (is.na(slide_id)) {
      slide_id <- min(d_summary_prep$slide_id)
      id <- min(d_summary_prep$id[d_summary_prep$slide_id == slide_id])
    }
    data_extract <- d_summary_prep[
      d_summary_prep$slide_id == slide_id & d_summary_prep$id == id,
    ]
    if (nrow(data_extract) == 0) {
      stop(
        "No data in slide_id=%s, id=%s",
        as.character(slide_id),
        as.character(id)
      )
    }
  }
  ret <- list()
  for (current_slide_id in unique(data_extract$slide_id)) {
    ret[[as.character(current_slide_id)]] <-
      as_matrix_rpptx_single_slide(
        data_extract[data_extract$slide_id == current_slide_id, ],
        span = span
      )
  }
  if (length(slide_id) == 1 && length(id) == 1) {
    # When the user is specific, return a matrix (otherwise, return the list of
    # lists of matrices)
    ret <- ret[[1]][[1]]
  }
  ret
}

as_matrix_rpptx_single_slide <- function(x, span = c(NA_character_, "fill")) {
  span <- match.arg(span)
  stopifnot(nrow(x) > 0)
  stopifnot(length(unique(x$slide_id)) == 1)
  ret <- list()
  for (current_id in unique(x$id)) {
    ret[[as.character(current_id)]] <-
      as_matrix_rpptx_single_table(x[x$id == current_id, ], span = span)
  }
  ret
}

as_matrix_rpptx_single_table <- function(x, span = c(NA_character_, "fill")) {
  span <- match.arg(span)
  stopifnot(nrow(x) > 0)
  stopifnot(length(unique(x$slide_id)) == 1)
  stopifnot(length(unique(x$id)) == 1)
  ret <- matrix(NA_character_, nrow = max(x$row_id), ncol = max(x$cell_id))
  for (current_data_idx in seq_len(nrow(x))) {
    if (span %in% "fill") {
      # fill the row/col span
      all_rows <- seq(
        x$row_id[current_data_idx],
        x$row_id[current_data_idx] + x$row_span[current_data_idx] - 1
      )
      all_cols <- seq(
        x$cell_id[current_data_idx],
        x$cell_id[current_data_idx] + x$col_span[current_data_idx] - 1
      )
    } else {
      # just the current cell
      all_rows <- x$row_id[current_data_idx]
      all_cols <- x$cell_id[current_data_idx]
    }
    for (current_col in all_cols) {
      # A loop or some other multiple assignment is needed (multiple indexing
      # does not work on rows and columns simultaneously for matrices)
      ret[all_rows, current_col] <- x$text[current_data_idx]
    }
  }
  ret
}

library(officer)

x <- read_pptx()
x <- add_slide(x, "Title and Content")
x <- remove_slide(x)

# Remove multiple slides at once
x <- read_pptx()
x <- add_slide(x, "Title and Content")
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)
x <- add_slide(x, "Title and Content")
x <- remove_slide(x, index = c(2, 4))
pptx_file <- print(x, target = tempfile(fileext = ".pptx"))
pptx_file

read_core_properties <- function(package_dir) {
  filename <- file.path(package_dir, "docProps/core.xml")
  if (!file.exists(filename)) {
    # docx and pptx files exported from Google Docs or Slides do not have properties set. Use one from the template to get by
    warning(
      "No properties found. Using properties from template file in officer."
    )
    filename <- system.file(package = "officer", "template/core.xml")
  }
  doc <- read_xml(filename)
  ns_ <- xml_ns(doc)

  all_ <- xml_find_all(doc, "/cp:coreProperties/*")

  if (length(all_) < 1) {
    out <- list(
      data = structure(
        character(0),
        .Dim = c(0L, 4L),
        .Dimnames = list(
          NULL,
          c("ns", "name", "attrs", "value")
        )
      ),
      ns = c(
        cp = "http://schemas.openxmlformats.org/package/2006/metadata/core-properties",
        dc = "http://purl.org/dc/elements/1.1/",
        dcmitype = "http://purl.org/dc/dcmitype/",
        dcterms = "http://purl.org/dc/terms/",
        xsi = "http://www.w3.org/2001/XMLSchema-instance"
      )
    )
    return(out)
  }

  names_ <- sapply(all_, xml_name, ns = ns_)
  names_ <- strsplit(names_, ":")

  # concat all attrs in single chars
  attrs <- sapply(xml_attrs(all_), function(ns_) {
    paste0('xsi:', names(ns_), '=\"', ns_, '\"', collapse = " ")
  })
  attrs <- ifelse(lengths(xml_attrs(all_)) < 1, "", paste0(" ", attrs))

  propnames <- sapply(names_, function(x) x[2])

  props <- matrix(
    c(sapply(names_, function(x) x[1]), propnames, attrs, xml_text(all_)),
    ncol = 4
  )
  colnames(props) <- c("ns", "name", "attrs", "value")
  rownames(props) <- propnames
  z <- list(data = props, ns = unclass(ns_))
  class(z) <- "core_properties"
  z
}

default_cp_tags <- list(
  title = list(ns = "dc", attrs = NA_character_),
  subject = list(ns = "dc", attrs = NA_character_),
  creator = list(ns = "dc", attrs = NA_character_),
  keywords = list(ns = "cp", attrs = NA_character_),
  description = list(ns = "dc", attrs = NA_character_),
  lastModifiedBy = list(ns = "cp", attrs = NA_character_),
  revision = list(ns = "cp", attrs = NA_character_),
  created = list(ns = "dcterms", attrs = 'xsi:type="dcterms:W3CDTF"'),
  modified = list(ns = "dcterms", attrs = 'xsi:type="dcterms:W3CDTF"'),
  category = list(ns = "cp", attrs = NA_character_)
)


#' @export
`[<-.core_properties` <- function(x, i, j, value) {
  if (!i %in% x$data[, "name"]) {
    attrs <- ifelse(
      is.na(default_cp_tags[[i]]$attrs),
      "",
      paste0(" ", default_cp_tags[[i]]$attrs)
    )
    new <- matrix(
      c(default_cp_tags[[i]]$ns, i, attrs, value),
      nrow = 1,
      dimnames = list(i, colnames(x$data))
    )
    x$data <- rbind(x$data, new)
  } else {
    x$data[i, j] <- value
  }
  x
}


#' @export
`[.core_properties` <- function(x, i, j) {
  x$data[i, j]
}


write_core_properties <- function(core_matrix, package_dir) {
  ns_ <- core_matrix$ns
  core_matrix <- core_matrix$data
  if (!is.matrix(core_matrix)) {
    stop("core_properties should be stored in a character matrix.")
  }

  ns_ <- paste0('xmlns:', names(ns_), '=\"', ns_, '\"', collapse = " ")
  xml_ <- paste0(
    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
    '\n',
    '<cp:coreProperties ',
    ns_,
    '>'
  )
  properties <- sprintf(
    "<%s:%s%s>%s</%s:%s>",
    core_matrix[, "ns"],
    core_matrix[, "name"],
    core_matrix[, "attrs"],
    htmlEscapeCopy(core_matrix[, "value"]),
    core_matrix[, "ns"],
    core_matrix[, "name"]
  )
  xml_ <- paste0(
    xml_,
    paste0(properties, collapse = ""),
    "</cp:coreProperties>"
  )
  props_dir = file.path(package_dir, "docProps")
  dir.create(props_dir, recursive = TRUE, showWarnings = FALSE)
  filename <- file.path(props_dir, "core.xml")
  writeLines(enc2utf8(xml_), filename, useBytes = TRUE)
  invisible()
}

drop_templatenode_from_app <- function(package_dir) {
  file <- file.path(package_dir, "docProps", "app.xml")
  if (!file.exists(file)) {
    return()
  }
  doc <- read_xml(file)
  node_template <- xml_child(doc, "d1:Template")
  if (!inherits(node_template, "xml_missing")) {
    xml_remove(node_template)
    write_xml(doc, file)
  }
}


# app.xml properties ----

read_app_properties <- function(package_dir) {
  file <- file.path(package_dir, "docProps", "app.xml")
  if (!file.exists(file)) {
    return(NULL)
  }
  doc <- read_xml(file)
  ns <- xml_ns(doc)

  hyperlink_base_node <- xml_find_first(doc, "//d1:HyperlinkBase", ns = ns)
  hyperlink_base <- if (inherits(hyperlink_base_node, "xml_missing")) {
    NA_character_
  } else {
    xml_text(hyperlink_base_node)
  }

  company_node <- xml_find_first(doc, "//d1:Company", ns = ns)
  company <- if (inherits(company_node, "xml_missing")) {
    NA_character_
  } else {
    xml_text(company_node)
  }

  z <- list(
    data = data.frame(
      name = c("HyperlinkBase", "Company"),
      value = c(hyperlink_base, company),
      stringsAsFactors = FALSE
    ),
    file = file
  )
  class(z) <- "app_properties"
  z
}


#' @export
`[<-.app_properties` <- function(x, i, j, value) {
  if (!i %in% x$data$name) {
    new_row <- data.frame(name = i, value = value, stringsAsFactors = FALSE)
    x$data <- rbind(x$data, new_row)
  } else {
    x$data[x$data$name == i, j] <- value
  }
  x
}


#' @export
`[.app_properties` <- function(x, i, j) {
  x$data[x$data$name == i, j]
}

#' @importFrom xml2 xml_set_text
write_app_properties <- function(app_props, package_dir) {
  if (is.null(app_props)) {
    return(invisible())
  }

  file <- file.path(package_dir, "docProps", "app.xml")
  if (!file.exists(file)) {
    return(invisible())
  }

  doc <- read_xml(file)
  ns <- xml_ns(doc)

  for (i in seq_len(nrow(app_props$data))) {
    prop_name <- app_props$data$name[i]
    prop_value <- app_props$data$value[i]

    if (is.na(prop_value) || prop_value == "") {
      next
    }

    xpath <- paste0("//d1:", prop_name)
    node <- xml_find_first(doc, xpath, ns = ns)
    if (inherits(node, "xml_missing")) {
      new_node <- read_xml(sprintf(
        "<%s>%s</%s>",
        prop_name,
        htmlEscapeCopy(prop_value),
        prop_name
      ))
      xml_add_child(doc, new_node)
    } else {
      xml_set_text(node, prop_value)
    }
  }

  write_xml(doc, file)
  invisible()
}

#' @export
#' @importFrom zip zip
#' @title compress a folder
#' @description compress a folder to a target file. The
#' function returns the complete path to target file.
#' @param folder folder to compress
#' @param target path of the archive to create
#' @keywords internal
pack_folder <- function( folder, target ){

  target <- absolute_path(target)
  dir_fi <- dirname(target)

  if( !file.exists(dir_fi) ){
    stop("directory ", shQuote(dir_fi), " does not exist.", call. = FALSE)
  } else if( file.access(dir_fi) < 0 ){
    stop("can not write to directory ", shQuote(dir_fi), call. = FALSE)
  } else if( file.exists(target) && file.access(target) < 0 ){
    stop(shQuote(target), " already exists and is not writable", call. = FALSE)
  } else if( !file.exists(target) ){
    old_warn <- getOption("warn")
    options(warn = -1)
    x <- tryCatch({cat("", file = target);TRUE}, error = function(e) FALSE, finally = unlink(target, force = TRUE) )
    options(warn = old_warn)
    if( !x )
      stop(shQuote(target), " cannot be written, please check your permissions.", call. = FALSE)
  }

  curr_wd <- getwd()
  setwd(folder)
  tryCatch(
    zip::zipr(zipfile = target, include_directories = FALSE,
                files = list.files(path = ".", all.files = FALSE), recurse = TRUE)
    , error = function(e) {
      stop("Could not write ", shQuote(target), " [", e$message, "]")
    },
    finally = {
      setwd(curr_wd)
    })

  target
}

#' @export
#' @importFrom zip unzip
#' @title Extract files from a zip file
#' @description Extract files from a zip file to a folder. The
#' function returns the complete path to destination folder.
#' @param file path of the archive to unzip
#' @param folder folder to create
#' @keywords internal
unpack_folder <- function(file, folder){

  stopifnot(file.exists(file))

  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", file)

  # force deletion if already existing
  unlink(folder, recursive = TRUE, force = TRUE)

  if( l10n_info()$`UTF-8` ){
    zip::unzip( zipfile = file, exdir = folder )
  } else {
    officer_wd_folder <- getOption("officer_wd_folder")
    if(is.null(officer_wd_folder) || !dir.exists(officer_wd_folder)){
      officer_wd_folder <- tempdir()
    }
    # unable to unzip a file with accent when on windows
    newfile <- tempfile(tmpdir = officer_wd_folder, fileext = file_type)
    file.copy(from = file, to = newfile)
    zip::unzip( zipfile = newfile, exdir = folder )
    unlink(newfile, force = TRUE)
  }

  absolute_path(folder)
}

absolute_path <- function(x){

  if (length(x) != 1L)
    stop("'x' must be a single character string")
  epath <- path.expand(x)

  if( file.exists(epath)){
    epath <- normalizePath(epath, "/", mustWork = TRUE)
  } else {
    if( !dir.exists(dirname(epath)) ){
      stop("directory of ", x, " does not exist.", call. = FALSE)
    }
    cat("", file = epath)
    epath <- normalizePath(epath, "/", mustWork = TRUE)
    unlink(epath)
  }
  epath
}

library(officer)

# ph_location_label demo ----

doc <- read_pptx()
doc <- add_slide(doc, layout = "Title and Content")

# all ph_label can be read here
layout_properties(doc, layout = "Title and Content")

doc <- ph_with(
  doc,
  head(iris),
  location = ph_location_label(ph_label = "Content Placeholder 2")
)
doc <- ph_with(
  doc,
  format(Sys.Date()),
  location = ph_location_label(ph_label = "Date Placeholder 3")
)
doc <- ph_with(
  doc,
  "This is a title",
  location = ph_location_label(ph_label = "Title 1")
)

print(doc, target = tempfile(fileext = ".pptx"))

#' Layout selection helper
#'
#' Select a layout by name or index. The master name is inferred and only required
#' for disambiguation in case the layout name is not unique across masters.
#'
#' @param x An `rpptx` object.
#' @param layout Layout name or index. Index refers to the row index of the [layout_summary()]
#' output.
#' @param master Name of master. Only required if layout name is not unique across masters.
#' @return A `<layout_info>` object, i.e. a list with the entries `index`, `layout_name`,
#' `layout_file`, `master_name`, `master_file`, and `slide_layout`.
#' @keywords internal
get_layout <- function(x, layout, master = NULL) {
  stop_if_not_rpptx(x, "x")
  if (!(is.numeric(layout) || is.character(layout))) {
    cli::cli_abort(
      c("{.arg layout} must be {.cls numeric} or {.cls character}",
        "x" = "Got class {.cls {class(layout)[1]}} instead"
      )
    )
  }
  if (length(layout) != 1) {
    cli::cli_abort(
      c("{.arg layout} is not length 1",
        "x" = "{.arg layout} must be {.emph one} layout name or index."
      )
    )
  }
  df <- x$slideLayouts$get_metadata()
  names(df)[2:3] <- c("layout_name", "layout_file") # consistent naming
  n_layouts <- nrow(df)
  df$index <- seq_len(n_layouts)

  if (n_layouts == 0) {
    cli::cli_alert_danger("No layouts available.")
    return(NULL)
  }

  if (is.numeric(layout)) {
    res <- get_row_by_index(df, layout)
  } else {
    res <- get_row_by_name(df, layout, master)
  }
  l <- as.list(res)
  slide_layout <- x$slideLayouts$collection_get(l$layout_file)
  l <- c(l, slide_layout = slide_layout)
  l <- l[c("index", "layout_name", "layout_file", "master_name", "master_file", "slide_layout")] # nice order
  class(l) <- c("layout_info", "list")
  l
}


#' @export
print.layout_info <- function(x, ...) {
  cli::cli_h3("{.cls layout_info} object")
  str(utils::head(x, -1), give.attr = FALSE, no.list = TRUE)
  cat(" $ slide_layout: 'R6' <slide_layout>")
}


get_row_by_index <- function(df, layout) {
  index <- layout
  if (!index %in% df$index) {
    cli::cli_abort(
      c("Layout index out of bounds.",
        "x" = "Index must be between {.val {1}} and {.val {nrow(df)}}.",
        "i" = cli::col_grey("See row indexes in {.fn layout_summary}")
      ),
      call = NULL
    )
  }
  df[index, ]
}


# select layout by name
get_row_by_name <- function(df, layout, master) {
  if (!is.null(master)) {
    masters <- unique(df$master_name)
    if (!master %in% masters) {
      cli::cli_abort(c(
        "master {.val {master}} does not exist.",
        "i" = "See {.fn layout_summary} for available masters."
      ), call = NULL)
    }
    df <- df[df$master_name == master, ]
  }

  df <- df[df$layout_name == layout, ]
  if (nrow(df) == 0) {
    msg <- ifelse(is.null(master),
      "Layout {.val {layout}} does not exist",
      "Layout {.val {layout}} does not exist in master {.val {master}}"
    )
    cli::cli_abort(c(msg, "i" = "See {.fn layout_summary} for available layouts."), call = NULL)
    return(NULL)
  }
  if (nrow(df) > 1) {
    cli::cli_abort(c(
      "Layout exists in more than one master",
      "x" = "Please specify the master name in arg {.arg master}"
    ), call = NULL)
  }
  df
}


# get <layout_info> object for slide layout
get_slide_layout <- function(x, slide_idx) {
  stop_if_not_rpptx(x)
  if (length(x) == 0) {
    cli::cli_abort(
      c("Presentation does not have any slides yet",
        "x" = "Can only get the layout for an existing slides",
        "i" = "You can add a slide using {.fn add_slide}")
      , call = NULL)
  }
  ensure_slide_index_exists(x, slide_idx)
  df <- x$slide$get_xfrm()[[slide_idx]]
  layout <- unique(df$name)
  master <- unique(df$master_name)
  get_layout(x, layout, master)
}


# get <layout_info> object for layout of current slide
get_layout_for_current_slide <- function(x) {
  get_slide_layout(x, x$cursor)
}

#' @title Get comments in a Word document as a data.frame
#' @description return a data.frame representing the comments in a Word document.
#' @param x an rdocx object
#' @details
#' Each row of the returned data frame contains data for one comment. The
#'    columns contain the following information:
#' * "comment_id" - unique comment id
#' * "author" - name of the comment author
#' * "initials" - initials of the comment author
#' * "date" - timestamp of the comment
#' * "text" - a list column of characters containing the comment text. Elements can
#'    be vectors of length > 1 if a comment contains multiple paragraphs,
#'    blocks or runs or of length 0 if the comment is empty.
#' * "para_id" - a list column of characters containing the parent paragraph IDs.
#'    Elememts can be vectors of length > 1 if a comment spans multiple paragraphs
#'    or of length 0 if the comment has no parent paragraph.
#' * "commented_text" - a list column of characters containing the
#'    commented text. Elememts can be vectors of length > 1 if a comment
#'    spans multiple paragraphs or runs or of length 0 if the commented text is empty.
#' @examples
#' bl <- block_list(
#'   fpar("Comment multiple words."),
#'   fpar("Second line")
#' )
#'
#' a_par <- fpar(
#'   "This paragraph contains",
#'   run_comment(
#'     cmt = bl,
#'     run = ftext("a comment."),
#'     author = "Author Me",
#'     date = "2023-06-01"
#'   )
#' )
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, value = a_par, style = "Normal")
#'
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#'
#' docx_comments(read_docx(docx_file))
#' @export
docx_comments <- function(x) {
  stopifnot(inherits(x, "rdocx"))

  comment_ids <- xml_attr(
    xml_find_all(
      x$doc_obj$get(), "/w:document/w:body//*[self::w:commentRangeStart]"
    ), "id"
  )

  comment_text_runs <- lapply(comment_ids, function(id) {
    xml_find_all(
      x$doc_obj$get(),
      paste0(
        "/w:document/w:body//*[self::w:r[w:t and",
        "preceding::w:commentRangeStart[@w:id=\'", id, "\']",
        " and ",
        "following::w:commentRangeEnd[@w:id=\'", id, "\']]]"
      )
    )
  })

  data <- data.frame(
    comment_id = comment_ids
  )
  # Add parent paragraph id
  data$para_id <- lapply(
    comment_text_runs,
    function(x) xml_attr(xml_parent(x), "paraId")
  )
  data$commented_text <- lapply(comment_text_runs, xml_text)

  comments <- xml_find_all(x$comments$get(), "//w:comments/w:comment")

  out <- data.frame(
    stringsAsFactors = FALSE,
    comment_id = xml_attr(comments, "id"),
    author = xml_attr(comments, "author"),
    initials = xml_attr(comments, "initials"),
    date = xml_attr(comments, "date")
  )

  out$text <- lapply(
    comments,
    function(x) xml_text(xml_find_all(x, "w:p/w:r/w:t"))
  )

  data <- merge(out, data, by = "comment_id", all.x = TRUE)
  data[order(as.integer(data$comment_id)), ]
}

library(officer)

doc <- read_docx()

img.file <- file.path(R.home("doc"), "html", "logo.jpg")
if (file.exists(img.file)) {
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)

  # Set the unit in which the width and height arguments are expressed
  doc <- body_add_img(
    x = doc,
    src = img.file,
    height = 2.69,
    width = 3.53,
    unit = "cm"
  )
}

print(doc, target = tempfile(fileext = ".docx"))

x <- read_pptx()

# select layout explicitly
plot_layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
plot_layout_properties(x = x, layout = "Title Slide") # no master needed if layout name unique
plot_layout_properties(x = x, layout = 1) # use layout index instead of name

# plot current slide's layout (default if no layout is passed)
x <- read_pptx()
x <- add_slide(x, "Title Slide")
plot_layout_properties(x)

# change appearance: what to show, font size, legend etc.
plot_layout_properties(x, layout = "Two Content", title = FALSE, type = FALSE, id = FALSE)
plot_layout_properties(x, layout = 4, cex = c(labels = .8, id = .7, type = .7))
plot_layout_properties(x, 1, legend = TRUE)

# tags with namespaces ----

wp_ns_yes <- "<w:p xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">"
wp_ns_no <- "<w:p>"

tbl_ns_yes <- "<w:tbl xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\">"
tbl_ns_no <- "<w:tbl>"

wr_ns_yes <- "<w:r xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">"
wr_ns_no <- "<w:r>"

ar_ns_yes <- "<a:r xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">"
ar_ns_no <- "<a:r>"

ap_ns_yes <- "<a:p xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">"
ap_ns_no <- "<a:p>"

psp_ns_yes <- "<p:sp xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">"
psp_ns_no <- "<p:sp>"

# utils -----
runs_to_p_wml <- function(..., add_ns = FALSE, style_id = NULL) {
  runs <- list(...)
  run_str <- lapply(runs, to_wml, add_ns = FALSE)
  run_str$collapse <- ""
  run_str <- do.call(paste0, run_str)
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  if (!is.null(style_id)) {
    ppr <- paste0("<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>")
  } else {
    ppr <- "<w:pPr/>"
  }
  out <- paste0(open_tag, ppr, run_str, "</w:p>")
  out
}


a_xfrm_str <- function(left = 0, top = 0, width = 3, height = 3, rot = 0) {
  start_tag <- "<a:xfrm>"
  if (!is.null(rot) && !is.na(rot) && rot != 0) {
    start_tag <- sprintf("<a:xfrm rot=\"%.0f\">", -rot * 60000)
  }

  xfrm_str <- paste0(
    start_tag,
    "<a:off x=\"%.0f\" y=\"%.0f\"/><a:ext cx=\"%.0f\" cy=\"%.0f\"/></a:xfrm>"
  )
  sprintf(
    xfrm_str,
    left * 914400,
    top * 914400,
    width * 914400,
    height * 914400
  )
}

p_xfrm_str <- function(left = 0, top = 0, width = 3, height = 3, rot = 0) {
  if (is.null(rot) || !is.finite(rot)) {
    rot <- 0
  }

  xfrm_str <- "<p:xfrm rot=\"%.0f\"><a:off x=\"%.0f\" y=\"%.0f\"/><a:ext cx=\"%.0f\" cy=\"%.0f\"/></p:xfrm>"
  sprintf(
    xfrm_str,
    -rot * 60000,
    left * 914400,
    top * 914400,
    width * 914400,
    height * 914400
  )
}


# colors ----

css_color <- function(color) {
  color <- as.vector(col2rgb(color, alpha = TRUE)) / c(1, 1, 1, 255)

  if (!(color[4] > 0)) {
    "transparent"
  } else {
    sprintf(
      "rgba(%.0f,%.0f,%.0f,%.2f)",
      color[1],
      color[2],
      color[3],
      color[4]
    )
  }
}
hex_color <- function(color) {
  color <- as.vector(col2rgb(color, alpha = TRUE)) / c(1, 1, 1, 255)
  sprintf(
    "%02X%02X%02X",
    color[1],
    color[2],
    color[3]
  )
}
colalpha <- function(x) {
  if ("transparent" %in% x) {
    return(0)
  }
  alpha <- as.vector(col2rgb(x, alpha = TRUE))[4] / 255
  alpha
}

is_transparent <- function(color) {
  !(colalpha(color) > 0)
}

solid_fill <- function(color) {
  sprintf(
    paste0(
      "<a:solidFill><a:srgbClr val=\"%s\">",
      "<a:alpha val=\"%.0f\"/>",
      "</a:srgbClr></a:solidFill>"
    ),
    hex_color(color),
    colalpha(color) * 100000
  )
}

solid_fill_pml <- function(bg) {
  bg_str <- ""
  if (!is.null(bg)) {
    bg_str <- solid_fill(bg)
  }
  bg_str
}

# geom ----

prst_geom_pml <- function(x) {
  geom_str <- ""
  if (!is.null(x)) {
    x <- check_set_geom(x)
    tagname <- paste0("a:prstGeom")
    geom_str <- sprintf("<%s prst=\"%s\"><a:avLst/></%s>", tagname, x, tagname)
  }
  geom_str
}

# line ----

ln_pml <- function(x) {
  ln_str <- ""
  if (!is.null(x)) {
    color_ <- ""
    if (is_transparent(x$color) || x$lwd < .001) {
      color_ <- "<a:noFill/>"
    } else {
      color_ <- solid_fill(x$color)
    }

    dash_ <- dash_pml(x$lty)
    join_ <- linejoin_pml(x$linejoin)
    head_ <- lineend_pml(x$headend, "head")
    tail_ <- lineend_pml(x$tailend, "tail")

    ln_str <- sprintf(
      paste0(
        "<a:ln w=\"%s\" cap=\"%s\" cmpd=\"%s\">",
        color_,
        dash_,
        join_,
        head_,
        tail_,
        "</a:ln>"
      ),
      12700 * x$lwd,
      x$lineend,
      x$linecmpd
    )
  }
  ln_str
}

lineend_pml <- function(x, side) {
  lineend_str <- ""
  if (!is.null(x)) {
    tagname <- paste0("a:", side, "End")
    lineend_str <- sprintf(
      "<%s type=\"%s\" w=\"%s\" len=\"%s\"/>",
      tagname,
      x$type,
      x$width,
      x$length
    )
  }
  lineend_str
}

dash_pml <- function(x) {
  dash_str <- ""
  if (!is.null(x)) {
    dash_str <- sprintf("<a:prstDash val=\"%s\"/>", x)
  }
  dash_str
}

linejoin_pml <- function(x) {
  linejoin_str <- ""
  if (!is.null(x)) {
    linejoin_str <- sprintf("<a:%s/>", x)
  }
  linejoin_str
}

# border ----

border_pml <- function(x, side) {
  tagname <- paste0("a:ln", side)

  width_ <- sprintf("w=\"%.0f\"", x$width * 12700)

  if (is_transparent(x$color) || x$width < .001) {
    color_ <- "<a:noFill/>"
  } else {
    color_ <- solid_fill(x$color)
  }

  if (!x$style %in% c("dotted", "dashed", "solid")) {
    x$style <- "solid"
  }
  if ("dotted" %in% x$style) {
    x$style <- "sysDot"
  } else if ("dashed" %in% x$style) {
    x$style <- "sysDash"
  }

  style_ <- sprintf("<a:prstDash val=\"%s\"/>", x$style)

  paste0(
    "<",
    tagname,
    " ",
    "algn=\"ctr\" cmpd=\"sng\" cap=\"flat\"",
    " ",
    width_,
    ">",
    color_,
    style_,
    "</",
    tagname,
    ">"
  )
}


border_wml <- function(x, side) {
  tagname <- paste0("w:", side)
  x$style[x$style %in% "solid"] <- "single"
  x$style[x$style %in% "ridge"] <- "threeDEmboss"
  x$style[x$style %in% "groove"] <- "threeDEngrave"
  if (!x$style %in% border_styles) {
    x$style <- "single"
  }
  if (x$width < 0.0001 || is_transparent(x$color)) {
    x$style <- "none"
  }

  style_ <- sprintf("w:val=\"%s\"", x$style)
  width_ <- sprintf("w:sz=\"%.0f\"", x$width * 8)
  color_ <- sprintf("w:color=\"%s\"", hex_color(x$color))

  paste0(
    "<",
    tagname,
    " ",
    style_,
    " ",
    width_,
    " ",
    "w:space=\"0\"",
    " ",
    color_,
    "/>"
  )
}
border_css <- function(x, side) {
  color_ <- css_color(x$color)
  if (!(x$width > 0)) {
    color_ <- "transparent"
  }

  width_ <- sprintf("%.02fpt", x$width)

  x$style[x$style %in% "threeDEmboss"] <- "ridge"
  x$style[x$style %in% "threeDEngrave"] <- "groove"
  x$style[x$style %in% "nil"] <- "none"

  if (
    !x$style %in%
      c(
        "dotted",
        "dashed",
        "solid",
        "double",
        "inset",
        "outset",
        "ridge",
        "groove",
        "none"
      )
  ) {
    x$style <- "solid"
  }
  paste0("border-", side, ": ", width_, " ", x$style, " ", color_, ";")
}

# ppr ----
ppr_pml <- function(x) {
  align <- ""
  if (!is.na(x$text.align)) {
    align <- " algn=\"r\""
    if (x$text.align == "left") {
      align <- " algn=\"l\""
    } else if (x$text.align == "center") {
      align <- " algn=\"ctr\""
    } else if (x$text.align == "justify") {
      align <- " algn=\"just\""
    }
  }

  leftright_padding <- ""
  if (!is.na(x$padding.left) && !is.na(x$padding.right)) {
    leftright_padding <- sprintf(
      " marL=\"%.0f\" marR=\"%.0f\"",
      x$padding.left * 12700,
      x$padding.right * 12700
    )
  }

  top_padding <- ""
  if (!is.na(x$padding.top)) {
    top_padding <- sprintf(
      "<a:spcBef><a:spcPts val=\"%.0f\" /></a:spcBef>",
      x$padding.top * 100
    )
  }
  bottom_padding <- ""
  if (!is.na(x$padding.bottom)) {
    bottom_padding <- sprintf(
      "<a:spcAft><a:spcPts val=\"%.0f\" /></a:spcAft>",
      x$padding.bottom * 100
    )
  }
  line_spacing <- ""
  if (!is.na(x$line_spacing)) {
    line_spacing <- sprintf(
      "<a:lnSpc><a:spcPct val=\"%.0f\"/></a:lnSpc>",
      x$line_spacing * 100000
    )
  }

  paste0(
    "<a:pPr",
    align,
    leftright_padding,
    ">",
    line_spacing,
    top_padding,
    bottom_padding,
    "<a:buNone/>",
    "</a:pPr>"
  )
}

ppr_css <- function(x) {
  text.align <- sprintf("text-align:%s;", x$text.align)
  borders <- paste0(
    border_css(x$border.bottom, "bottom"),
    border_css(x$border.top, "top"),
    border_css(x$border.left, "left"),
    border_css(x$border.right, "right")
  )

  paddings <- sprintf(
    "padding-top:%.0fpt;padding-bottom:%.0fpt;padding-left:%.0fpt;padding-right:%.0fpt;",
    x$padding.top,
    x$padding.bottom,
    x$padding.left,
    x$padding.right
  )
  ls <- formatC(
    x$line_spacing,
    format = "f",
    digits = 2,
    decimal.mark = ".",
    drop0trailing = TRUE
  )
  line_spacing <- sprintf("line-height: %s;", ls)

  shading.color <- sprintf("background-color:%s;", css_color(x$shading.color))

  paste0(
    "margin:0pt;",
    text.align,
    borders,
    paddings,
    line_spacing,
    shading.color
  )
}

ppr_wml <- function(x) {
  if ("justify" %in% x$text.align) {
    x$text.align <- "both"
  }
  pstyle <- ""
  if (!is.null(x$word_style) && !is.na(x$word_style)) {
    pstyle <- sprintf("<w:pStyle w:pstlname=\"%s\"/>", x$word_style)
  }
  tabs <- ""
  if (!is.null(x$tabs) && !isFALSE(x$tabs)) {
    tabs <- to_wml(x$tabs)
  }

  text_align_ <- ""
  if (!is.na(x$text.align)) {
    text_align_ <- sprintf("<w:jc w:val=\"%s\"/>", x$text.align)
  }

  keep_with_next <- ""
  if (isTRUE(x$keep_with_next)) {
    keep_with_next <- "<w:keepNext/>"
  }

  borders_ <- ""
  if (
    !is.null(x$border.top) &&
      !isFALSE(x$border.top) &&
      !is.null(x$border.bottom) &&
      !isFALSE(x$border.bottom) &&
      !is.null(x$border.left) &&
      !isFALSE(x$border.left) &&
      !is.null(x$border.right) &&
      !isFALSE(x$border.right)
  ) {
    borders_ <- paste0(
      "<w:pBdr>",
      border_wml(x$border.bottom, "bottom"),
      border_wml(x$border.top, "top"),
      border_wml(x$border.left, "left"),
      border_wml(x$border.right, "right"),
      "</w:pBdr>"
    )
  }

  leftright_padding <- ""
  if (!is.na(x$padding.left) && !is.na(x$padding.right)) {
    leftright_padding <- sprintf(
      "<w:ind w:left=\"%.0f\" w:right=\"%.0f\" w:firstLine=\"0\" w:firstLineChars=\"0\"/>",
      x$padding.left * 20,
      x$padding.right * 20
    )
  }

  topbot_spacing <- ""
  if (
    !is.na(x$padding.bottom) && !is.na(x$padding.top) && !is.na(x$line_spacing)
  ) {
    topbot_spacing <- sprintf(
      "<w:spacing w:after=\"%.0f\" w:before=\"%.0f\" w:line=\"%.0f\"/>",
      x$padding.bottom * 20,
      x$padding.top * 20,
      x$line_spacing * 240
    )
  }

  shading_ <- ""
  if (!is_transparent(x$shading.color)) {
    shading_ <- sprintf(
      "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"%s\"/>",
      hex_color(x$shading.color)
    )
  }

  paste0(
    "<w:pPr>",
    pstyle,
    text_align_,
    keep_with_next,
    borders_,
    tabs,
    topbot_spacing,
    leftright_padding,
    shading_,
    "</w:pPr>"
  )
}


# rpr ----

rpr_pml <- function(x) {
  if (is_transparent(x$color)) {
    return("")
  }

  out <- "<a:rPr cap=\"none\""

  if (!is.na(x$font.size)) {
    if (x$font.size > 0) {
      out <- paste0(out, sprintf(" sz=\"%.0f\"", x$font.size * 100))
    }
  }

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, " i=\"1\"")
    } else {
      out <- paste0(out, " i=\"0\"")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, " b=\"1\"")
    } else {
      out <- paste0(out, " b=\"0\"")
    }
  }

  if (!is.na(x$underlined)) {
    if (x$underlined) {
      out <- paste0(out, " u=\"sng\"")
    } else {
      out <- paste0(out, " u=\"none\"")
    }
  }

  has_strike <- !is.null(x$strike) && !is.na(x$strike) && x$strike
  if (has_strike) {
    out <- paste0(out, " strike=\"sngStrike\"")
  } else {
    out <- paste0(out, " strike=\"noStrike\"")
  }

  if (x$vertical.align == "superscript") {
    out <- paste0(out, " baseline=\"40000\"")
  } else if (x$vertical.align == "subscript") {
    out <- paste0(out, " baseline=\"-40000\"")
  }
  out <- paste0(out, ">")

  if (!is.na(x$color)) {
    out <- paste0(out, solid_fill(x$color))

    if (!is_transparent(x$shading.color)) {
      shad <- sprintf(
        paste0(
          "<a:highlight><a:srgbClr val=\"%s\">",
          "<a:alpha val=\"%.0f\"/>",
          "</a:srgbClr></a:highlight>"
        ),
        hex_color(x$shading.color),
        colalpha(x$shading.color) * 100000
      )
      out <- paste0(out, shad)
    }
  }

  out <- paste0(
    out,
    if (!is.na(x$font.family)) {
      sprintf("<a:latin typeface=\"%s\"/>", x$font.family)
    },
    if (!is.na(x$cs.family)) sprintf("<a:cs typeface=\"%s\"/>", x$cs.family),
    if (!is.na(x$eastasia.family)) {
      sprintf("<a:ea typeface=\"%s\"/>", x$eastasia.family)
    },
    if (!is.na(x$hansi.family)) {
      sprintf("<a:sym typeface=\"%s\"/>", x$hansi.family)
    }
  )

  out <- paste0(out, "</a:rPr>")
  out
}

rpr_wml <- function(x) {
  out <- paste0(
    "<w:rPr><w:rFonts",
    if (!is.na(x$font.family)) paste0(" w:ascii=\"", x$font.family, "\""),
    if (!is.na(x$hansi.family)) paste0(" w:hAnsi=\"", x$hansi.family, "\""),
    if (!is.na(x$eastasia.family)) {
      paste0(" w:eastAsia=\"", x$eastasia.family, "\"")
    },
    if (!is.na(x$cs.family)) paste0(" w:cs=\"", x$cs.family, "\""),
    "/>"
  )

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, "<w:i w:val=\"true\"/>")
    } else {
      out <- paste0(out, "<w:i w:val=\"false\"/>")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, "<w:b w:val=\"true\"/>")
    } else {
      out <- paste0(out, "<w:b w:val=\"false\"/>")
    }
  }
  if (!is.na(x$underlined)) {
    if (x$underlined) {
      out <- paste0(out, "<w:u w:val=\"single\"/>")
    } else {
      out <- paste0(out, "<w:u w:val=\"none\"/>")
    }
  }

  has_strike <- !is.null(x$strike) && !is.na(x$strike) && x$strike
  if (has_strike) {
    out <- paste0(out, "<w:strike w:val=\"true\"/>")
  } else {
    out <- paste0(out, "<w:strike w:val=\"false\"/>")
  }

  if (x$vertical.align == "superscript") {
    out <- paste0(out, "<w:vertAlign w:val=\"superscript\"/>")
  } else if (x$vertical.align == "subscript") {
    out <- paste0(out, "<w:vertAlign w:val=\"subscript\"/>")
  }

  if (!is.na(x$font.size)) {
    out <- paste0(
      out,
      sprintf(
        "<w:sz w:val=\"%.0f\"/><w:szCs w:val=\"%.0f\"/>",
        x$font.size * 2,
        x$font.size * 2
      )
    )
  }

  if (!is.na(x$color)) {
    out <- paste0(
      out,
      sprintf("<w:color w:val=\"%s\"/>", hex_color(x$color))
    )
  }

  if (!is.na(x$shading.color)) {
    if (is.na(x$color)) {
      stop("shading.color must be used **with** a color.")
    }
    if (!is_transparent(x$shading.color)) {
      out <- paste0(
        out,
        sprintf(
          "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"%s\"/>",
          hex_color(x$shading.color)
        )
      )
    }
    if (colalpha(x$color) < 1) {
      out <- paste0(
        out,
        sprintf(
          paste0(
            "<w14:textFill><w14:solidFill><w14:srgbClr val=\"%s\">",
            "<w14:alpha val=\"%.0f\"/>",
            "</w14:srgbClr></w14:solidFill></w14:textFill>"
          ),
          hex_color(x$color),
          colalpha(x$color) * 100000
        )
      )
    }
  }

  out <- paste0(out, "</w:rPr>")
  out
}

rpr_css <- function(x) {
  out <- ""

  if (!is.na(x$font.family)) {
    out <- paste0(out, sprintf("font-family:'%s';", x$font.family))
  }
  if (!is.na(x$color)) {
    out <- paste0(out, sprintf("color:%s;", css_color(x$color)))
  }
  if (!is.na(x$font.size)) {
    out <- paste0(out, sprintf("font-size:%0.1fpt;", x$font.size))
  }

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, "font-style:italic;")
    } else {
      out <- paste0(out, "font-style:normal;")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, "font-weight:bold;")
    } else {
      out <- paste0(out, "font-weight:normal;")
    }
  }

  has_underline <- !is.na(x$underlined) && x$underlined
  if (has_underline) {
    out <- paste0(out, "text-decoration:underline;")
  }

  has_strike <- !is.null(x$strike) && !is.na(x$strike) && x$strike
  if (has_strike) {
    out <- paste0(out, "text-decoration:line-through;")
  }

  if (!is.na(x$shading.color)) {
    if (!is_transparent(x$shading.color)) {
      out <- paste0(
        out,
        sprintf("background-color:%s;", css_color(x$shading.color))
      )
    } else {
      out <- paste0(out, "background-color:transparent;")
    }
  }

  if (x$vertical.align == "superscript") {
    out <- paste0(out, "vertical-align: super;")
  } else if (x$vertical.align == "subscript") {
    out <- paste0(out, "vertical-align: sub;")
  }

  out
}
# tcpr ----

tcpr_pml <- function(x) {
  text.direction <-
    if (x$text.direction %in% "btlr") {
      " vert=\"vert270\""
    } else if (x$text.direction %in% "tbrl") {
      " vert=\"vert\""
    } else {
      ""
    }

  vertical.align <-
    if (x$vertical.align %in% "center") {
      " anchor=\"ctr\""
    } else if (x$vertical.align %in% "top") {
      " anchor=\"t\""
    } else {
      " anchor=\"b\""
    }

  margins <- sprintf(
    " marB=\"%.0f\" marT=\"%.0f\" marR=\"%.0f\" marL=\"%.0f\"",
    x$margin.bottom * 12700,
    x$margin.top * 12700,
    x$margin.right * 12700,
    x$margin.left * 12700
  )

  background.color <- paste0(
    sprintf(
      "<a:solidFill><a:srgbClr val=\"%s\">",
      hex_color(x$background.color)
    ),
    sprintf("<a:alpha val=\"%.0f\"/>", colalpha(x$background.color) * 100000),
    "</a:srgbClr></a:solidFill>"
  )

  bb <- border_pml(x$border.bottom, side = "B")
  bt <- border_pml(x$border.top, side = "T")
  bl <- border_pml(x$border.left, side = "L")
  br <- border_pml(x$border.right, side = "R")

  pml_attrs <- paste0(text.direction, vertical.align, margins)
  paste0(
    "<a:tcPr",
    pml_attrs,
    ">",
    bl,
    br,
    bt,
    bb,
    background.color,
    "</a:tcPr>"
  )
}

tcpr_wml <- function(x) {
  background.color <- sprintf(
    "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"%s\"/>",
    hex_color(x$background.color)
  )
  vertical.align <- ifelse(
    x$vertical.align %in% c("center", "top"),
    sprintf("<w:vAlign w:val=\"%s\"/>", x$vertical.align),
    "<w:vAlign w:val=\"bottom\"/>"
  )
  text.direction <- ifelse(
    x$text.direction %in% "btlr",
    "<w:textDirection w:val=\"btLr\"/>",
    ifelse(
      x$text.direction %in% "tbrl",
      "<w:textDirection w:val=\"tbRl\"/>",
      ""
    )
  )

  bb <- border_wml(x$border.bottom, side = "bottom")
  bt <- border_wml(x$border.top, side = "top")
  bl <- border_wml(x$border.left, side = "left")
  br <- border_wml(x$border.right, side = "right")

  margin.bottom <- sprintf(
    "<w:bottom w:w=\"%.0f\" w:type=\"dxa\"/>",
    x$margin.bottom * 20
  )
  margin.top <- sprintf(
    "<w:top w:w=\"%.0f\" w:type=\"dxa\"/>",
    x$margin.top * 20
  )
  margin.left <- sprintf(
    "<w:left w:w=\"%.0f\" w:type=\"dxa\"/>",
    x$margin.left * 20
  )
  margin.right <- sprintf(
    "<w:right w:w=\"%.0f\" w:type=\"dxa\"/>",
    x$margin.right * 20
  )

  rowspan <- ""
  if (x$rowspan > 1) {
    rowspan <- paste0("<w:gridSpan w:val=\"", x$rowspan, "\"/>")
  }
  colspan <- ""
  if (x$colspan > 1) {
    colspan <- "<w:vMerge w:val=\"restart\"/>"
  } else if (x$colspan < 1) {
    colspan <- "<w:vMerge/>"
  }

  paste0(
    "<w:tcPr>",
    rowspan,
    colspan,
    "<w:tcBorders>",
    bb,
    bt,
    bl,
    br,
    "</w:tcBorders>",
    background.color,
    "<w:tcMar>",
    margin.top,
    margin.bottom,
    margin.left,
    margin.right,
    "</w:tcMar>",
    text.direction,
    vertical.align,
    "</w:tcPr>"
  )
}

css_px <- function(x, format = "%.0fpx") {
  ifelse(is.na(x), "inherit", ifelse(x < 0.001, "0", sprintf(format, x)))
}

tcpr_css <- function(x) {
  background.color <- ifelse(
    colalpha(x$background.color) > 0,
    sprintf(
      "background-clip: padding-box;background-color:%s;",
      css_color(x$background.color)
    ),
    "background-color:transparent;"
  )

  width <- ifelse(
    is.null(x$width) || is.na(x$width),
    "",
    sprintf("width:%s;", css_px(x$width * 72))
  )
  height <- ifelse(
    is.null(x$height) || is.na(x$height),
    "",
    sprintf("height:%s;", css_px(x$height * 72))
  )
  vertical.align <- ifelse(
    x$vertical.align %in% "center",
    "vertical-align:middle;",
    ifelse(
      x$vertical.align %in% "top",
      "vertical-align:top;",
      "vertical-align:bottom;"
    )
  )

  bb <- border_css(x$border.bottom, side = "bottom")
  bt <- border_css(x$border.top, side = "top")
  bl <- border_css(x$border.left, side = "left")
  br <- border_css(x$border.right, side = "right")

  margin.bottom <- sprintf(
    "margin-bottom:%s;",
    sprintf("%.0fpt", x$margin.bottom)
  )
  margin.top <- sprintf("margin-top:%s;", sprintf("%.0fpt", x$margin.top))
  margin.left <- sprintf("margin-left:%s;", sprintf("%.0fpt", x$margin.left))
  margin.right <- sprintf("margin-right:%s;", sprintf("%.0fpt", x$margin.right))

  paste0(
    width,
    height,
    background.color,
    vertical.align,
    bb,
    bt,
    bl,
    br,
    margin.bottom,
    margin.top,
    margin.left,
    margin.right
  )
}

#' @export
#' @title transform an xml string with images references
#' @description This function is useless now, as the processing of images
#' is automated when using [print.rdocx()].
#' @param x an rdocx object
#' @param str wml string
#' @family functions for officer extensions
#' @keywords internal
wml_link_images <- function(x, str) {
  str
}

# by capturing the path, we are making 'unique' new image names.
#' @importFrom openssl sha1
fake_newname <- function(filename) {
  which_files <- grepl("\\.[a-zA-Z0-0]+$", filename)
  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", filename[which_files])
  dest_basename <- sapply(filename[which_files], function(z) {
    as.character(sha1(file(z)))
  })
  dest_basename <- paste0(dest_basename, file_type)
  x <- filename
  x[which_files] <- dest_basename
  x
}

process_images <- function(
  doc_obj,
  relationships,
  package_dir,
  media_dir = "word/media",
  media_rel_dir = "media"
) {
  hl_nodes <- xml_find_all(
    doc_obj$get(),
    "//a:blip[@r:embed]|//asvg:svgBlip[@r:embed]",
    ns = c(
      "a" = "http://schemas.openxmlformats.org/drawingml/2006/main",
      "asvg" = "http://schemas.microsoft.com/office/drawing/2016/SVG/main",
      "r" = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
    )
  )
  which_to_add <- hl_nodes[!grepl("^rId[0-9]+$", xml_attr(hl_nodes, "embed"))]
  hl_ref <- unique(xml_attr(which_to_add, "embed"))
  for (i in seq_along(hl_ref)) {
    dest_basename <- fake_newname(hl_ref[i])
    img_path <- file.path(package_dir, media_dir)
    if (!file.exists(file.path(img_path, dest_basename))) {
      dir.create(img_path, recursive = TRUE, showWarnings = FALSE)
      file.copy(from = hl_ref[i], to = file.path(img_path, dest_basename))
    }
    if (
      !file.path(media_rel_dir, dest_basename) %in%
        relationships$get_data()$target
    ) {
      rid <- sprintf("rId%.0f", relationships$get_next_id())
      relationships$add(
        id = rid,
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image",
        target = file.path(media_rel_dir, dest_basename)
      )
    } else {
      reldf <- relationships$get_data()
      rid <- reldf$id[basename(reldf$target) %in% dest_basename]
    }
    which_match_id <- grepl(
      dest_basename,
      fake_newname(xml_attr(which_to_add, "embed")),
      fixed = TRUE
    )
    xml_attr(which_to_add[which_match_id], "r:embed") <- rep(
      rid,
      sum(which_match_id)
    )
  }
}

process_docx_poured <- function(
  doc_obj,
  relationships,
  content_type,
  package_dir,
  media_dir = "word"
) {
  hl_nodes <- xml_find_all(
    doc_obj$get(),
    "//w:altChunk[@r:id]",
    ns = c(
      "w" = "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
      "r" = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
    )
  )

  which_to_add <- hl_nodes[!grepl("^rId[0-9]+$", xml_attr(hl_nodes, "id"))]
  hl_ref <- unique(xml_attr(which_to_add, "id"))
  for (i in seq_along(hl_ref)) {
    rid <- sprintf("rId%.0f", relationships$get_next_id())

    file.copy(
      from = hl_ref[i],
      to = file.path(package_dir, media_dir, basename(hl_ref[i]))
    )

    relationships$add(
      id = rid,
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/aFChunk",
      target = basename(hl_ref[i])
    )
    content_type$add_override(
      setNames(
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml",
        paste0("/", media_dir, "/", basename(hl_ref[i]))
      )
    )

    which_match_id <- grepl(
      hl_ref[i],
      xml_attr(which_to_add, "id"),
      fixed = TRUE
    )
    xml_attr(which_to_add[which_match_id], "r:id") <- rep(
      rid,
      sum(which_match_id)
    )
  }
}

#' @importFrom xml2 xml_remove as_xml_document xml_parent xml_child
process_links <- function(doc_obj, type = "wml") {
  rel <- doc_obj$relationship()
  if ("wml" %in% type) {
    hl_nodes <- xml_find_all(doc_obj$get(), "//w:hyperlink[@r:id]")
  } else {
    hl_nodes <- xml_find_all(doc_obj$get(), "//a:hlinkClick[@r:id]")
  }
  which_to_add <- hl_nodes[!grepl("^rId[0-9]+$", xml_attr(hl_nodes, "id"))]
  hl_ref <- unique(xml_attr(which_to_add, "id"))
  for (i in seq_along(hl_ref)) {
    rid <- sprintf("rId%.0f", rel$get_next_id())

    rel$add(
      id = rid,
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink",
      target = officer_url_decode(hl_ref[i]),
      target_mode = "External"
    )

    which_match_id <- grepl(
      hl_ref[i],
      xml_attr(which_to_add, "id"),
      fixed = TRUE
    )
    xml_attr(which_to_add[which_match_id], "r:id") <- rep(
      rid,
      sum(which_match_id)
    )
  }
}


update_hf_list <- function(part_list = list(), type = "header", package_dir) {
  files <- list.files(
    path = file.path(package_dir, "word"),
    pattern = sprintf("^%s[0-9]*.xml$", type)
  )
  files <- files[!basename(files) %in% names(part_list)]
  if (type %in% "header") {
    cursor <- "/w:hdr/*[1]"
    body_xpath <- "/w:hdr"
  } else {
    cursor <- "/w:ftr/*[1]"
    body_xpath <- "/w:ftr"
  }

  new_list <- lapply(files, function(x) {
    docx_part$new(
      path = package_dir,
      main_file = x,
      cursor = cursor,
      body_xpath = body_xpath
    )
  })
  names(new_list) <- basename(files)
  append(part_list, new_list)
}

#' @export
#' @title Remove unused media from a document
#' @description The function will scan the media
#' directory and delete images that are not used
#' anymore. This function is to be used when images
#' have been replaced many times.
#' @param x `rdocx` or `rpptx` object
#' @param warn_user TRUE to make sure users are warned when
#' using this function that will be un-exported in a next
#' version
#' @keywords internal
sanitize_images <- function(x, warn_user = TRUE) {
  if (warn_user) {
    cli::cli_warn(
      c(
        "!" = "Function {.fn sanitize_images} will be removed soon. You should remove calls to {.fn sanitize_images}.",
        "i" = "Instead the function is run before saving the document to a file."
      )
    )
  }

  if (inherits(x, "rdocx")) {
    image_files <- c()
    all_docs <- append(x$headers, x$footers)
    all_docs[[length(all_docs) + 1]] <- x$doc_obj
    all_docs[[length(all_docs) + 1]] <- x$footnotes

    for (doc_part in all_docs) {
      suppressWarnings({
        blip_nodes <- xml_find_all(
          doc_part$get(),
          "//a:blip[contains(@r:embed, 'rId')]|//asvg:svgBlip[contains(@r:embed, 'rId')]",
          ns = c(
            "a" = "http://schemas.openxmlformats.org/drawingml/2006/main",
            "asvg" = "http://schemas.microsoft.com/office/drawing/2016/SVG/main",
            "r" = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
          )
        )
      })

      embed_list <- xml_attr(blip_nodes, "embed")
      embed_data <- filter(
        .data = doc_part$rel_df(),
        basename(.data$type) %in% "image",
        .data$id %in% embed_list
      )
      embed_data <- embed_data$target
      image_files[[length(image_files) + 1]] <- embed_data
    }

    image_files <- do.call(c, image_files)
    image_files <- unique(image_files)

    base_doc <- file.path(x$package_dir, "word")
    existing_img <- list.files(
      file.path(base_doc, "media"),
      pattern = "\\.(png|jpg|jpeg|eps|emf|svg)$",
      ignore.case = TRUE,
      recursive = TRUE,
      full.names = TRUE
    )

    existing_img <- gsub(paste0(base_doc, "/"), "", existing_img, fixed = TRUE)
    unlink(
      file.path(base_doc, setdiff(existing_img, image_files)),
      force = TRUE
    )

    # these iterations will remove removed images from relationships
    for (doc_part in all_docs) {
      rel <- doc_part$relationship()
      rel_data <- rel$get_data()
      rel_data <- rel_data[basename(rel_data$type) %in% "image", ]
      rel_data <- rel_data[!file.exists(file.path(base_doc, rel_data$target)), ]
      if (nrow(rel_data) > 0) {
        rel$remove(rel_data$target)
        doc_part$save()
      }
    }
  } else if (inherits(x, "rpptx")) {
    rel_files <- list.files(
      x$package_dir,
      pattern = "\\.xml.rels$",
      recursive = TRUE,
      full.names = TRUE
    )

    image_files <- lapply(rel_files, function(x) {
      zz <- read_xml(x)
      rels <- xml_children(zz)
      rels <- rels[
        xml_attr(rels, "Type") %in%
          "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"
      ]
      xml_attr(rels, "Target")
    })
    image_files <- unique(unlist(image_files))
    base_doc <- file.path(x$package_dir, "ppt")
    existing_img <- list.files(
      file.path(base_doc, "media"),
      pattern = "\\.(png|jpg|jpeg|eps|emf|svg)$",
      ignore.case = TRUE,
      recursive = TRUE,
      full.names = TRUE
    )
    existing_img <- gsub(
      paste0(base_doc, "/"),
      "../",
      existing_img,
      fixed = TRUE
    )
    unlink(
      file.path(base_doc, setdiff(existing_img, image_files)),
      force = TRUE
    )
  }
  x
}

#' @title shortcuts for formatting properties
#'
#' @description
#' Shortcuts for \code{fp_text}, \code{fp_par},
#' \code{fp_cell} and \code{fp_border}.
#' @name shortcuts
#' @examples
#' shortcuts$fp_bold()
#' shortcuts$fp_italic()
#' shortcuts$b_null()
NULL

#' @rdname shortcuts
#' @format NULL
#' @docType NULL
#' @keywords NULL
#' @export
shortcuts <- list(
  fp_bold = function(...) {fp_text( bold = TRUE, ... )},
  fp_italic = function( ... ) {fp_text( italic = TRUE, ... )},
  b_null = function( ... )  {fp_border( width = 0, ... )}
)

library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(
  doc,
  "Hello world",
  location = ph_location(width = 4, height = 3, newlabel = "hello")
)
print(doc, target = tempfile(fileext = ".pptx"))

# Set geometry and outline
doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
loc <- ph_location(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  bg = "steelblue",
  ln = sp_line(color = "red", lwd = 2.5),
  geom = "trapezoid"
)
doc <- ph_with(doc, "", loc = loc)
print(doc, target = tempfile(fileext = ".pptx"))

x <- read_pptx()

# INFO -------------

# Returns layout's ph_labels by default in same order as layout_properties()
layout_rename_ph_labels(x, "Comparison")
layout_properties(x, "Comparison")$ph_label


# BASICS -----------
#
# HINT: run `plot_layout_properties(x, "Comparison")` to see how labels change

# rename using key-value pairs: 'old label' = 'new label' or 'id' = 'new label'
layout_rename_ph_labels(x, "Comparison", "Title 1" = "LABEL MATCHED") # label matching
layout_rename_ph_labels(x, "Comparison", "3" = "ID MATCHED") # id matching
layout_rename_ph_labels(x, "Comparison", "Date Placeholder 6" = "DATE", "8" = "FOOTER") # label, id

# rename using a named list and the .dots arg
renames <- list("Content Placeholder 3" = "CONTENT_1", "6" = "CONTENT_2")
layout_rename_ph_labels(x, "Comparison", .dots = renames)

# rename via rhs assignment and optional index (not id!)
layout_rename_ph_labels(x, "Comparison") <- LETTERS[1:8]
layout_rename_ph_labels(x, "Comparison")[1:3] <- paste("CHANGED", 1:3)

# rename via rhs assignment and ph id (not index)
layout_rename_ph_labels(x, "Comparison", id = c(2, 4)) <- paste("ID =", c(2, 4))


# MORE ------------

# make all labels lower case
labels <- layout_rename_ph_labels(x, "Comparison")
layout_rename_ph_labels(x, "Comparison") <- tolower(labels)

# rename all labels to type [type_idx]
lp <- layout_properties(x, "Comparison")
layout_rename_ph_labels(x, "Comparison") <- paste0(lp$type, " [", lp$type_idx, "]")

# rename duplicated placeholders (see also `layout_dedupe_ph_labels()`)
file <- system.file("doc_examples", "ph_dupes.pptx", package = "officer")
x <- read_pptx(file)
lp <- layout_properties(x, "2-dupes")
idx <- which(lp$ph_label == "Content 7") # exists twice
layout_rename_ph_labels(x, "2-dupes")[idx] <- paste("DUPLICATE", seq_along(idx))

# warning: in case of duped labels only the first occurrence is renamed
x <- read_pptx(file)
layout_rename_ph_labels(x, "2-dupes", "Content 7" = "new label")

#' @title Manipulate Microsoft Word and PowerPoint Documents with 'officer'
#'
#' @description
#' The officer package facilitates access to and manipulation of
#' 'Microsoft Word' and 'Microsoft PowerPoint' documents from R.
#' It also supports the writing of 'RTF' documents.
#'
#' Examples of usage are:
#'
#' * Create Word documents with tables, titles, TOC and graphics
#' * Importation of Word and PowerPoint files into data objects
#' * Write updated content back to a PowerPoint presentation
#' * Clinical reporting automation
#' * Production of reports from a shiny application
#'
#' To start with officer, read about [read_docx()], [read_pptx()]
#' or [rtf_doc()].
#'
#' The package is also providing several objects that can be printed
#' in 'R Markdown' documents for advanced Word or PowerPoint
#' reporting as [run_autonum()] and [block_caption()].
#' @section Get Word content in a data.frame:
#'
#' While officer allows you to generate Word and PowerPoint documents,
#' an important feature is also the ability to read the content of existing
#' Word documents. Use [docx_summary()] to extract document content as a
#' structured data.frame, making it easy to analyze and process Word files
#' programmatically.
#'
#' @section Copy an officer object:
#'
#' 'officer' objects of class `rdocx` or `rpptx` use R6 classes with reference
#' semantics. **Assignment does NOT create a copy**:
#'
#' ```
#' pptx1 <- read_pptx()
#' pptx2 <- pptx1  # pptx2 is a reference to pptx1, not a copy!
#' ```
#'
#' If you need independent documents (e.g., in loops), read the template
#' each time:
#'
#' ```
#' for (i in 1:10) {
#'   doc <- read_docx("template.docx")  # Read fresh each iteration
#'   # ... modify doc ...
#'   print(doc, target = paste0("output_", i, ".docx"))
#' }
#' ```
#'
#' @seealso The user documentation: \url{https://ardata-fr.github.io/officeverse/} and
#' manuals \url{https://davidgohel.github.io/officer/}
#' @docType package
#' @name officer
"_PACKAGE"

#' @importFrom R6 R6Class
relationship <- R6Class(
  "relationship",
  public = list(
    initialize = function(
      id = character(0),
      type = character(0),
      target = character(0)
    ) {
      private$id <- id
      private$type <- type
      private$target <- target
      private$target_mode <- rep(NA_character_, length(target))
      private$ext_src <- character(length(id))
    },
    feed_from_xml = function(path) {
      doc <- read_xml(x = path)
      children <- xml_children(doc)
      ns <- xml_ns(doc)

      private$id <- c(
        private$id,
        vapply(children, xml_attr, NA_character_, attr = "Id", ns)
      )
      private$type <- c(
        private$type,
        vapply(children, xml_attr, NA_character_, attr = "Type", ns)
      )
      private$target <- c(
        private$target,
        vapply(children, xml_attr, NA_character_, attr = "Target", ns)
      )
      private$target_mode <- c(
        private$target_mode,
        vapply(children, xml_attr, NA_character_, attr = "TargetMode", ns)
      )
      private$ext_src <- c(private$ext_src, character(length(children)))
      self
    },
    write = function(path) {
      if (length(private$id)) {
        str <- paste0(
          "<Relationship Id=\"",
          private$id,
          "\" Type=\"",
          private$type,
          "\" Target=\"",
          htmlEscapeCopy(private$target),
          ifelse(
            is.na(private$target_mode),
            "",
            "\" TargetMode=\"External"
          ),
          "\"/>"
        )
      } else {
        str <- character(length = 0)
      }

      str <- c(
        "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
        "\n<Relationships  xmlns=\"http://schemas.openxmlformats.org/package/2006/relationships\">",
        str,
        "</Relationships>"
      )
      dir.create(dirname(path), showWarnings = FALSE, recursive = FALSE)
      writeLines(str, con = path, useBytes = TRUE)
      self
    },
    get_next_id = function() {
      max(c(0, private$get_int_id()), na.rm = TRUE) + 1
    },
    get_data = function() {
      data <- data.frame(
        id = character(0),
        int_id = integer(0),
        type = character(0),
        target = character(0),
        target_mode = character(0),
        ext_src = character(0),
        stringsAsFactors = FALSE
      )
      if (length(private$id)) {
        chr_stripped_id <- gsub("rId([0-9]+)", "\\1", private$id)
        test_id_as_num <- grepl("^rId[0-9]+$", private$id)
        if (any(!test_id_as_num)) {
          wrong_targets <- paste0(
            "invalid id(s) for {shQuote(private$target[!test_id_as_num])}"
          )
          names(wrong_targets) <- rep("*", length(wrong_targets))
          cli::cli_warn(
            c(
              "x" = "relationship file contains invalid id(s).",
              wrong_targets
            )
          )
        }
        data <- data.frame(
          id = private$id,
          int_id = as.integer(chr_stripped_id),
          type = private$type,
          target = private$target,
          target_mode = private$target_mode,
          ext_src = private$ext_src,
          stringsAsFactors = FALSE
        )

        data[order(data$id), ]
      }
      data
    },
    get_images_path = function() {
      is_img <- basename(private$type) %in% "image"
      targets <- private$target[is_img]
      names(targets) <- private$id[is_img]
      targets
    },
    add_img = function(src, root_target) {
      src <- setdiff(src, private$ext_src)
      if (!length(src)) {
        return(self)
      }

      if (any(grepl(" ", basename(src)))) {
        stop(
          paste(src, collapse = ", "),
          ": images with blanks in their basenames are not supported, please rename the file(s).",
          call. = FALSE
        )
      }

      last_id <- max(c(0, private$get_int_id()), na.rm = TRUE)

      id <- paste0("rId", seq_along(src) + last_id)
      type <- rep(
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image",
        length(src)
      )
      target <- file.path(root_target, basename(src))

      private$id <- c(private$id, id)
      private$type <- c(private$type, type)
      private$target <- c(private$target, target)
      private$target_mode <- c(private$target_mode, rep(NA, length(id)))
      private$ext_src <- c(private$ext_src, src)

      self
    },
    add_drawing = function(src, root_target) {
      src <- setdiff(src, private$ext_src)
      if (!length(src)) {
        return(self)
      }
      last_id <- max(c(0, private$get_int_id()), na.rm = TRUE)

      id <- paste0("rId", seq_along(src) + last_id)
      type <- rep(
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/drawing",
        length(src)
      )
      target <- file.path(root_target, basename(src))

      private$id <- unlist(c(private$id, id))
      private$type <- unlist(c(private$type, type))
      private$target <- unlist(c(private$target, target))
      private$target_mode <- unlist(c(private$target_mode, rep(NA, length(id))))
      private$ext_src <- unlist(c(private$ext_src, rep(NA, length(id))))

      self
    },
    add = function(id, type, target, target_mode = NA) {
      Encoding(target) <- "UTF-8"
      if (!target %in% private$target) {
        private$id <- c(private$id, id)
        private$type <- c(private$type, type)
        private$target <- c(private$target, target)
        private$target_mode <- c(private$target_mode, target_mode)
        private$ext_src <- c(private$ext_src, "")
      }
      self
    },
    remove = function(target) {
      id <- which(basename(private$target) %in% basename(target))
      private$id <- private$id[-id]
      private$type <- private$type[-id]
      private$target <- private$target[-id]
      private$target_mode <- private$target_mode[-id]
      private$ext_src <- private$ext_src[-id]
      self
    },
    show = function() {
      print(self$get_data())
    }
  ),
  private = list(
    id = NA,
    type = NA,
    target = NA,
    target_mode = NA,
    ext_src = NA,
    get_int_id = function() {
      if (length(private$id) > 0) {
        as.integer(gsub("rId([0-9]+)", "\\1", private$id))
      } else {
        integer(0)
      }
    }
  )
)

check_all_types <- function(dat) {
  expected_names <- c(
    "id",
    "int_id",
    "type",
    "target",
    "target_mode",
    "ext_src"
  )

  expect_is(dat, "data.frame")
  expect_equal(names(dat), expected_names)

  expect_is(dat$id, "character")
  expect_is(dat$int_id, "integer")
  expect_is(dat$type, "character")
  expect_is(dat$target, "character")
  expect_is(dat$target_mode, "character")
  expect_is(dat$ext_src, "character")
}

test_that("get_data", {
  rel <- officer:::relationship$new()
  dat <- rel$get_data()
  check_all_types(dat)

  rel <- officer:::relationship$new()
  rel$add_img(src = "test.png", root_target = "../media")
  dat <- rel$get_data()
  check_all_types(dat)

  rel <- officer:::relationship$new()
  rel$add_drawing(src = "test.xml", root_target = "../drawings")
  dat <- rel$get_data()
  check_all_types(dat)
})

check_all_types <- function(dat){
  expected_names <- c("id", "int_id", "type", "target", "target_mode", "ext_src")

  expect_is(dat, "data.frame" )
  expect_equal(names(dat), expected_names )

  expect_is(dat$id, "character" )
  expect_is(dat$int_id, "integer" )
  expect_is(dat$type, "character" )
  expect_is(dat$target, "character" )
  expect_is(dat$target_mode, "character" )
  expect_is(dat$ext_src, "character" )
}

test_that("get_data", {

  rel <- officer:::relationship$new()
  dat <- rel$get_data()
  check_all_types(dat)

  rel <- officer:::relationship$new()
  rel$add_img(src = "test.png", root_target = "../media")
  dat <- rel$get_data()
  check_all_types(dat)

  rel <- officer:::relationship$new()
  rel$add_drawing(src = "test.xml", root_target = "../drawings")
  dat <- rel$get_data()
  check_all_types(dat)
})

# Adds notes master if not present. There can be at most one notesMaster in a presentation.
add_notesMaster <- function(x) {

  presentation_relations <- x$presentation$rel_df()
  if (!any(grepl("notesMaster$", presentation_relations$type))) {

    # copy notesMaster file from templates
    path <- system.file(package = "officer", "template/notesMaster.xml")
    xml_file <- file.path(x$package_dir, "ppt/notesMasters/notesMaster1.xml")
    dir.create(dirname(xml_file), showWarnings = FALSE, recursive = FALSE)
    file.copy(path, to = xml_file,
              overwrite = TRUE,
              copy.mode = FALSE)

    # copy theme file
    theme_file <- "theme1.xml"
    theme_files <- list.files(path = file.path(x$package_dir, "ppt/theme"), pattern = "\\.xml$", full.names = F)
    theme_index <- as.integer(gsub("^(theme)([0-9]+)(\\.xml)$", "\\2", theme_files ))
    theme_file <- gsub(pattern = "[0-9]+", replacement = max(theme_index) + 1, theme_file)
    file.copy(file.path(x$package_dir, "ppt/theme/theme1.xml"),
              to = file.path(x$package_dir, "ppt/theme", theme_file),
              overwrite = TRUE,
              copy.mode = FALSE)

    # add relation to theme file
    newrel <- relationship$new()$add(id = "rId1",
                                     type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme",
                                     target = paste0("../theme/", theme_file))
    newrel$write(path = file.path(dirname(xml_file), "_rels",  paste0(basename(xml_file), ".rels")))

    x$notesMaster$add_notesMaster(xml_file)

    # register xml files in content_type
    partname <- "/ppt/notesMasters/notesMaster1.xml"
    content_type <- setNames("application/vnd.openxmlformats-officedocument.presentationml.notesMaster+xml", partname )
    x$content_type$add_override(content_type)

    partname <- paste0("/ppt/theme/", theme_file)
    content_type <- setNames("application/vnd.openxmlformats-officedocument.theme+xml", partname )
    x$content_type$add_override(content_type)

    # add notesMaster relation to presentation
    prs_relations <- x$presentation$relationship()
    rId <-  paste0("rId",prs_relations$get_next_id())
    prs_relations$add(id = rId,
                      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster",
                      target = "notesMasters/notesMaster1.xml")
    prs_relations$write(file.path(x$package_dir, "ppt/_rels", "presentation.xml.rels"))

    xml_elt <- as_xml_document(paste0("<p:notesMasterIdLst xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">
                                         <p:notesMasterId r:id=\"", rId, "\"/>
                                       </p:notesMasterIdLst>"))
    xml_add_sibling(xml_find_first(x$presentation$get(), "//p:sldMasterIdLst"), xml_elt)
  }
  return(x)
}

# Create notesSlide for current slide if it doesn't exist.
# Returns name of notesSlide for current slide.
get_or_create_notesSlide <- function(x) {

  slide <- x$slide$get_slide(x$cursor)
  rels <- slide$rel_df()
  if (!any(grepl("notesSlide$", rels$type))) {

    # add notesMaster if it doesn't exist
    add_notesMaster(x)

    nslidename <- x$notesSlide$get_new_slidename()

    # copy slide template
    xml_file <- file.path(x$package_dir, "ppt/notesSlides", nslidename)
    path <- system.file(package = "officer", "template/notesSlide.xml")
    file.copy(path, to = xml_file, copy.mode = FALSE)

    # add relationships to notesMaster and slide
    rel_filename <- file.path(
      dirname(xml_file), "_rels",
      paste0(basename(xml_file), ".rels"))
    newrel <- relationship$new()$add(
      id = "rId1",
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster",
      target = "../notesMasters/notesMaster1.xml")
    newrel$add(id = "rId2",
               type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide",
               target = file.path("../slides", slide$name()))
    newrel$write(path = rel_filename)

    # add relationship from slide to notesSlide
    rel <- slide$relationship()
    rel$add(id = paste0("rId", rel$get_next_id()),
            type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesSlide",
            target = file.path( "../notesSlides", nslidename) )
    rel$write(file.path(
      dirname(slide$file_name()), "_rels",
      paste0(basename(slide$file_name()), ".rels")))

    # update presentation elements
    x$content_type$add_notesSlide(partname = file.path( "/ppt/notesSlides", nslidename) )

    # add new notesSlide to collection
    x$notesSlide$add_slide(xml_file)
  } else {
    nslidename <- basename(rels$target[grepl("notesSlide$", rels$type)])
  }
  return(nslidename)
}



#' @export
#' @title Location of a named placeholder for notes
#' @description The function will use the label of a placeholder
#' to find the corresponding location in the slide notes.
#' @param ph_label placeholder label of the used notes master
#' @param ... unused arguments
notes_location_label <- function( ph_label, ...){
  x <- list(ph_label = ph_label)
  class(x) <- c("location_label", "location_str")
  x
}

#' @export
#' @title Location of a placeholder for notes
#' @description The function will use the type name of the placeholder (e.g. body, hdr),
#' to find the corresponding location.
#' @param type placeholder label of the used notes master
#' @param ... unused arguments
notes_location_type <- function( type = "body", ...){

  ph_types <- c("ftr", "sldNum", "hdr", "body", "sldImg")
  if(!type %in% ph_types){
    stop("argument type ('", type, "') expected to be a value of ",
         paste0(shQuote(ph_types), collapse = ", "), ".")
  }
  x <- list(type = type)
  class(x) <- c("location_type", "location_str")
  x
}

ph_from_location <- function(loc, doc, ...){
  UseMethod("ph_from_location", loc)
}


#' @export
ph_from_location.location_label <- function(loc, doc, ...) {
  xfrm <- doc$notesMaster$xfrm()
  location <- xfrm[xfrm$ph_label == loc$ph_label, ]
  if (nrow(location) < 1) stop("No placeholder with label ", loc$ph_label, " found!")

  id <- uuid_generate()
  str <- "<p:nvSpPr><p:cNvPr id=\"%s\" name=\"%s\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr>%s</p:nvPr></p:nvSpPr><p:spPr/>"
  new_ph <- sprintf(str, id, location$ph_label, location$ph)
  return(list(ph = new_ph, label = location$ph_label))
}


#' @export
ph_from_location.location_type <- function(loc, doc, ...) {
  xfrm <- doc$notesMaster$xfrm()
  location <- xfrm[xfrm$type == loc$type, ]
  if (nrow(location) < 1) stop("No placeholder of type ", loc$type, " found!")

  id <- uuid_generate()
  str <- "<p:nvSpPr><p:cNvPr id=\"%s\" name=\"%s\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr>%s</p:nvPr></p:nvSpPr><p:spPr/>"
  new_ph <- sprintf(str, id, location$ph_label[1], location$ph[1])
  return(list(ph = new_ph, label = location$ph_label[1]))
}

#' @export
#' @title Set notes for current slide
#' @description Set speaker notes for the current slide in a pptx presentation.
#' @param x an rpptx object
#' @param value text to be added to notes
#' @param location a placeholder location object.
#' It will be used to specify the location of the new shape. This location
#' can be defined with a call to one of the notes_ph functions. See
#' section \code{"see also"}.
#' @param ... further arguments passed to or from other methods.
#' @examples
#' # this name will be used to print the file
#' # change it to "youfile.pptx" to write the pptx
#' # file in your working directory.
#' fileout <- tempfile(fileext = ".pptx")
#' fpt_blue_bold <- fp_text_lite(color = "#006699", bold = TRUE)
#' doc <- read_pptx()
#' # add a slide with some text ----
#' doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
#' doc <- ph_with(x = doc, value = "Slide Title 1",
#'    location = ph_location_type(type = "title") )
#' # set speaker notes for the slide ----
#' doc <- set_notes(doc, value = "This text will only be visible for the speaker.",
#'    location = notes_location_type("body"))
#'
#' # add a slide with some text ----
#' doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
#' doc <- ph_with(x = doc, value = "Slide Title 2",
#'    location = ph_location_type(type = "title") )
#' bl <- block_list(
#'   fpar(ftext("hello world", fpt_blue_bold)),
#'   fpar(ftext("Turlututu chapeau pointu", fpt_blue_bold))
#' )
#' doc <- set_notes(doc, value = bl,
#'    location = notes_location_type("body"))
#'
#' print(doc, target = fileout)
#'
#' @seealso [print.rpptx()], [read_pptx()], [add_slide()], [notes_location_label()], [notes_location_type()]
#' @family functions slide manipulation
set_notes <- function(x, value, location, ...){
  UseMethod("set_notes", value)
}

#' @export
#' @describeIn set_notes add a character vector to a place holder in the notes on the
#' current slide, values will be added as paragraphs.
set_notes.character <- function( x, value, location, ... ){

  # get or create notesSlide for current slide
  nslidename <- get_or_create_notesSlide(x)

  idx <- x$notesSlide$slide_index(nslidename)
  nSlide <- x$notesSlide$get_slide(idx)

  new_ph <- ph_from_location(location, x)

  # remove placeholder if already used
  xml_remove(xml_find_first(nSlide$get(), paste0("//p:spTree/p:sp[p:nvSpPr/p:cNvPr[@name='", new_ph$label, "']]")))


  pars <- paste0("<a:p><a:r><a:rPr/><a:t>", htmlEscapeCopy(value), "</a:t></a:r></a:p>", collapse = "")
  xml_elt <- as_xml_document(paste0( psp_ns_yes, new_ph$ph,
                                     "<p:txBody><a:bodyPr/><a:lstStyle/>",
                                     pars, "</p:txBody></p:sp>" ))

  xml_add_child(xml_find_first(nSlide$get(), "//p:spTree"), xml_elt)
  nSlide$fortify_id()

  return(x)
}

#' @export
#' @describeIn set_notes add a [block_list()] to a place holder in the notes on the
#' current slide.
set_notes.block_list <- function( x, value, location, ... ){

  # get or create notesSlide for current slide
  nslidename <- get_or_create_notesSlide(x)

  idx <- x$notesSlide$slide_index(nslidename)
  nSlide <- x$notesSlide$get_slide(idx)

  new_ph <- ph_from_location(location, x)

  # remove placeholder if already used
  xml_remove(xml_find_first(nSlide$get(), paste0("//p:spTree/p:sp[p:nvSpPr/p:cNvPr[@name='", new_ph$label, "']]")))

  pars <- sapply(value, to_pml)
  pars <- paste0(pars, collapse = "")

  xml_elt <- as_xml_document(paste0( psp_ns_yes, new_ph$ph,
                                     "<p:txBody><a:bodyPr/><a:lstStyle/>",
                                     pars, "</p:txBody></p:sp>" ))

  xml_add_child(xml_find_first(nSlide$get(), "//p:spTree"), xml_elt)
  nSlide$fortify_id()

  return(x)
}

#' @export
#' @title Number of slides
#' @description Function `length` will return the number of slides.
#' @param x an rpptx object
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres, "Title and Content")
#' my_pres <- add_slide(my_pres, "Title and Content")
#' length(my_pres)
#' @family functions for reading presentation information
length.rpptx <- function(x) {
  x$slide$length()
}


#' @export
#' @title Slides width and height
#' @description Get the width and height of slides in inches as
#' a named vector.
#' @inheritParams length.rpptx
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres,
#'   layout = "Two Content", master = "Office Theme")
#' slide_size(my_pres)
#' @family functions for reading presentation information
slide_size <- function(x) {
  pres <- x$presentation$get()
  dimensions <- xml_attrs(xml_find_first(pres, "p:sldSz"))
  dimensions <- as.list(as.integer(dimensions[c("cx", "cy")]) / 914400)
  names(dimensions) <- c("width", "height")
  dimensions
}


#' @export
#' @title Presentation layouts summary
#' @description Get information about slide layouts and
#' master layouts into a data.frame. This function returns
#' a data.frame containing all layout and master names.
#' @inheritParams length.rpptx
#' @examples
#' my_pres <- read_pptx()
#' layout_summary ( x = my_pres )
#' @family functions for reading presentation information
layout_summary <- function(x) {
  data <- x$slideLayouts$get_metadata()
  data.frame(
    layout = data$name,
    master = data$master_name,
    stringsAsFactors = FALSE
  )
}


#' @export
#' @title Slide layout properties
#' @description Detailed information about the placeholders on the slide layouts (label, position, etc.).
#' See *Value* section below for more info.
#' @param x an `rpptx` object
#' @param layout slide layout name. If `NULL`, returns all layouts.
#' @param master master layout name where `layout` is located. If `NULL`, returns all masters.
#' @returns Returns a data frame with one row per placeholder and the following columns:
#' * `master_name`: Name of master (a `.pptx` file may have more than one)
#' * `name`: Name of layout
#' * `type`: Placeholder type
#' * `type_idx`: Running index for phs of the same type. Ordering by ph position
#'    (top -> bottom, left -> right)
#' * `id`: A unique placeholder id (assigned by PowerPoint automatically, starts at 2, potentially non-consecutive)
#' * `ph_label`: Placeholder label (can be set by the user in PowerPoint)
#' * `ph`: Placholder XML fragment (usually not needed)
#' * `offx`,`offy`: placeholder's distance from left and top edge (in inch)
#' * `cx`,`cy`: width and height of placeholder (in inch)
#' * `rotation`: rotation in degrees
#' * `fld_id` is generally stored as a hexadecimal or GUID value
#' * `fld_type`: a unique identifier for a particular field
#'
#' @example inst/examples/example_layout_properties.R
#' @family functions for reading presentation information
layout_properties <- function(x, layout = NULL, master = NULL) {
  data <- x$slideLayouts$get_xfrm_data()
  if (!is.null(layout) && !is.null(master)) {
    data <- data[data$name == layout & data$master_name %in% master, ]
  } else if (is.null(layout) && !is.null(master)) {
    data <- data[data$master_name %in% master, ]
  } else if (!is.null(layout) && is.null(master)) {
    data <- data[data$name == layout, ]
  }
  data <- data[, c(
    "master_name",
    "name",
    "type",
    "type_idx",
    "id",
    "ph_label",
    "ph",
    "offx",
    "offy",
    "cx",
    "cy",
    "rotation",
    "fld_id",
    "fld_type"
  )]
  data[["offx"]] <- data[["offx"]] / 914400
  data[["offy"]] <- data[["offy"]] / 914400
  data[["cx"]] <- data[["cx"]] / 914400
  data[["cy"]] <- data[["cy"]] / 914400
  data[["rotation"]] <- data[["rotation"]] / 60000

  rownames(data) <- NULL
  data
}


#' @export
#' @title Slide layout properties plot
#' @description Plot slide layout properties into corresponding placeholders.
#'  This can be useful to help visualize placeholders locations and identifiers.
#'  *All* information in the plot stems from the [layout_properties()] output.
#'  See *Details* section for more info.
#' @details
#' The plot contains all relevant information to reference a placeholder via the `ph_location_*`
#' function family:
#'
#' * `label`: ph label (red, center) to be used in [ph_location_label()].
#'    _NB_: The label can be assigned by the user in PowerPoint.
#' * `type[idx]`: ph type + type index in brackets (blue, upper left) to be used in [ph_location_type()].
#'    _NB_: The index is consecutive and is sorted by ph position (top -> bottom, left -> right).
#' * `id`: ph id (green, upper right) to be used in `ph_location_id()` (forthcoming).
#'    _NB_: The id is set by PowerPoint automatically and lack a meaningful order.
#'
#' @param x an `rpptx` object
#' @param layout slide layout name or numeric index (row index from [layout_summary()]. If `NULL` (default), it plots
#'   the current slide's layout or the default layout (if set and there are not slides yet).
#' @param master master layout name where `layout` is located. Can be omitted if layout is unambiguous.
#' @param slide_idx Numeric slide index (default `NULL`) to specify which slide’s layout should be plotted.
#' @param title if `TRUE` (default), adds a title with the layout and master name (latter in square brackets) at the top.
#' @param labels if `TRUE` (default), adds placeholder labels (centered in *red*).
#' @param type if `TRUE` (default), adds the placeholder type and its index (in square brackets)
#'   in the upper left corner (in *blue*).
#' @param id if `TRUE` (default), adds the placeholder's unique `id` (see column `id` from
#'  [layout_properties()]) in the upper right corner (in *green*).
#' @param cex List or vector to specify font size for `labels`, `type`, and `id`. Default is
#'   `c(labels = .5, type = .5, id = .5)`. See [graphics::text()] for details on how `cex` works.
#'   Matching by position and partial name matching is supported. A single numeric value will apply to all
#'   three parameters.
#' @param legend Add a legend to the plot (default `FALSE`).
#' @importFrom graphics plot rect text box
#' @family functions for reading presentation information
#' @example inst/examples/example_plot_layout_properties.R
#'
plot_layout_properties <- function(
  x,
  layout = NULL,
  master = NULL,
  slide_idx = NULL,
  labels = TRUE,
  title = TRUE,
  type = TRUE,
  id = TRUE,
  cex = c(labels = .5, type = .5, id = .5),
  legend = FALSE
) {
  stop_if_not_rpptx(x, "x")
  loffset <- ifelse(legend, 1, 0) # make space for legend at top
  old_par <- par(mar = c(2, 2, 1.5 + loffset, 0))
  on.exit(par(old_par))

  .cex <- update_named_defaults(
    cex,
    default = list(labels = .5, type = .5, id = .5),
    default_if_null = TRUE,
    argname = "cex"
  )
  if (.cex$labels <= 0) {
    labels <- FALSE
  }
  if (.cex$type <= 0) {
    type <- FALSE
  }
  if (.cex$id <= 0) {
    id <- FALSE
  }

  if (is.null(layout) && is.null(master) && length(x) == 0) {
    # fail or use default layout if set
    if (!has_layout_default(x)) {
      cli::cli_abort(
        c(
          "No {.arg layout} selected and no slides in presentation.",
          "x" = "Pass a layout name or index (see {.fn layout_summary})"
        )
      )
    }
    .ld <- get_layout_default(x)
    la <- get_layout(x, layout = .ld$layout, master = .ld$master)
    cli::cli_inform(c("i" = "Showing default layout: {.val {la$layout_name}}"))
  } else if (
    is.null(layout) && is.null(master) && length(x) > 0 && is.null(slide_idx)
  ) {
    # use current slides layout as default
    la <- get_layout_for_current_slide(x)
    cli::cli_inform(c(
      "i" = "Showing current slide's layout: {.val {la$layout_name}}"
    ))
  } else if (
    is.null(layout) && is.null(master) && length(x) > 0 && !is.null(slide_idx)
  ) {
    # use layout for selected slide
    la <- get_slide_layout(x, slide_idx = slide_idx)
    cli::cli_inform(c(
      "i" = "Showing layout of slide {.val {slide_idx}}: {.val {la$layout_name}}"
    ))
  } else {
    la <- get_layout(x, layout, master, layout_by_id = TRUE)
  }

  dat <- layout_properties(x, layout = la$layout_name, master = la$master_name)

  s <- slide_size(x)
  h <- s$height
  w <- s$width
  offx <- offy <- cx <- cy <- NULL # avoid R CMD CHECK problem
  list2env(dat[, c("offx", "offy", "cx", "cy")], environment()) # make available inside functions

  plot(
    x = c(0, w),
    y = -c(0, h),
    asp = 1,
    type = "n",
    axes = FALSE,
    xlab = NA,
    ylab = NA
  )
  rect(xleft = 0, xright = w, ybottom = 0, ytop = -h, border = "darkgrey")
  rect(xleft = offx, xright = offx + cx, ybottom = -offy, ytop = -(offy + cy))
  mtext("y [inch]", side = 2, line = 0, cex = .9, col = "darkgrey")
  mtext("x [inch]", side = 1, line = 0, cex = .9, col = "darkgrey")

  if (title) {
    title(
      main = paste0(la$layout_name, " [", la$master_name, "]"),
      line = 0 + loffset
    )
  }
  if (labels) {
    # centered
    text(
      x = offx + cx / 2,
      y = -(offy + cy / 2),
      labels = dat$ph_label,
      cex = .cex$labels,
      col = "red",
      adj = c(.5, 1)
    ) # adj-vert: avoid interference with type/id in small phs
  }
  if (type) {
    # upper left corner
    .type_info <- paste0(dat$type, " [", dat$type_idx, "]") # type + index in brackets
    text(
      x = offx,
      y = -offy,
      labels = .type_info,
      cex = .cex$type,
      col = "blue",
      adj = c(-.1, 1.2)
    )
  }
  if (id) {
    # upper right corner
    text(
      x = offx + cx,
      y = -offy,
      labels = dat$id,
      cex = .cex$id,
      col = "darkgreen",
      adj = c(1.3, 1.2)
    )
  }
  if (legend) {
    legend(
      x = w / 2,
      y = 0,
      x.intersp = 0.4,
      xjust = .5,
      yjust = 0,
      legend = c("type [type_idx]", "ph_label", "id"),
      fill = c("blue", "red", "darkgreen"),
      bty = "n",
      pt.cex = 1.2,
      cex = .7,
      text.width = NA,
      text.col = c("blue", "red", "darkgreen"),
      horiz = TRUE,
      xpd = TRUE
    )
  }
}


#' @export
#' @title Placeholder parameters annotation
#' @description generates a slide from each layout in the base document to
#' identify the placeholder indexes, types, names, master names and layout names.
#'
#' This is to be used when need to know what parameters should be used with
#' `ph_location*` calls. The parameters are printed in their corresponding shapes.
#'
#' Note that if there are duplicated `ph_label`, you should not use `ph_location_label()`.
#' Hint: You can dedupe labels using [layout_dedupe_ph_labels()].
#'
#' @param path path to the pptx file to use as base document or NULL to use the officer default
#' @param output_file filename to store the annotated powerpoint file or NULL to suppress generation
#' @return rpptx object of the annotated PowerPoint file
#' @examples
#' # To generate an anotation of the default base document with officer:
#' annotate_base(output_file = tempfile(fileext = ".pptx"))
#'
#' # To generate an annotation of the base document 'mydoc.pptx' and place the
#' # annotated output in 'mydoc_annotate.pptx'
#' # annotate_base(path = 'mydoc.pptx', output_file='mydoc_annotate.pptx')
#'
#' @family functions for reading presentation information
annotate_base <- function(path = NULL, output_file = 'annotated_layout.pptx') {
  ppt <- read_pptx(path = path)
  while (length(ppt) > 0) {
    ppt <- remove_slide(ppt, 1)
  }

  # Pulling out all of the layouts stored in the template
  lay_sum <- layout_summary(ppt)

  # Looping through each layout
  for (lidx in seq_len(nrow(lay_sum))) {
    # Pulling out the layout properties
    layout <- lay_sum[lidx, 1]
    master <- lay_sum[lidx, 2]
    lp <- layout_properties(x = ppt, layout = layout, master = master)

    # Adding a slide for the current layout
    ppt <- add_slide(x = ppt, layout = layout, master = master)
    size <- slide_size(ppt)
    fpar_ <- fpar(
      sprintf('layout ="%s", master = "%s"', layout, master),
      fp_t = fp_text(color = "orange", font.size = 20),
      fp_p = fp_par(text.align = "right", padding = 5)
    )
    ppt <- ph_with(
      x = ppt,
      value = fpar_,
      ph_label = "layout_ph",
      location = ph_location(
        left = 0,
        top = -0.5,
        width = size$width,
        height = 1,
        bg = "transparent",
        newlabel = "layout_ph"
      )
    )

    # Blank slides have nothing
    if (length(lp[, 1] > 0)) {
      # Now we go through each placholder
      for (pidx in seq_len(nrow(lp))) {
        textstr <- paste(
          "type=",
          lp$type[pidx],
          ", index=",
          lp$id[pidx],
          ", ph_label=",
          lp$ph_label[pidx]
        )
        ppt <- ph_with(
          x = ppt,
          value = textstr,
          location = ph_location_label(
            type = lp$type[pidx],
            ph_label = lp$ph_label[pidx]
          )
        )
      }
    }
  }

  if (!is.null(output_file)) {
    print(ppt, target = output_file)
  }

  ppt
}


#' @export
#' @title Slide content in a data.frame
#' @description Get content and positions of current slide
#' into a data.frame. Data for any tables, images, or paragraphs are
#' imported into the resulting data.frame.
#' @note
#' The column `id` of the result is not to be used by users.
#' This is a technical string id whose value will be used by office
#' when the document will be rendered. This is not related to argument
#' `index` required by functions [ph_with()].
#' @inheritParams length.rpptx
#' @param index slide index
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres, "Title and Content")
#' my_pres <- ph_with(my_pres, format(Sys.Date()),
#'   location = ph_location_type(type="dt"))
#' my_pres <- add_slide(my_pres, "Title and Content")
#' my_pres <- ph_with(my_pres, iris[1:2,],
#'   location = ph_location_type(type="body"))
#' slide_summary(my_pres)
#' slide_summary(my_pres, index = 1)
#' @family functions for reading presentation information
slide_summary <- function(x, index = NULL) {
  l_ <- length(x)
  if (l_ < 1) {
    cli::cli_abort("Presentation contains no slide.")
  }

  if (is.null(index)) {
    index <- x$cursor
  }

  if (!between(index, 1, l_)) {
    cli::cli_abort(
      c(
        "Slide index {.val {index}} is out of range.",
        "x" = "Presentation has {cli::no(l_)} slide{?s}."
      )
    )
  }

  slide <- x$slide$get_slide(index)

  nodes <- xml_find_all(slide$get(), as_xpath_content_sel("p:cSld/p:spTree/"))
  data <- read_xfrm(nodes, file = "slide", name = "")
  data$text <- sapply(nodes, xml_text)
  data[["offx"]] <- data[["offx"]] / 914400
  data[["offy"]] <- data[["offy"]] / 914400
  data[["cx"]] <- data[["cx"]] / 914400
  data[["cy"]] <- data[["cy"]] / 914400

  data$name <- NULL
  data$file <- NULL
  data$ph <- NULL
  data
}


#' @export
#' @title Color scheme of a PowerPoint file
#' @description Get the color scheme of a
#' 'PowerPoint' master layout into a data.frame.
#' @inheritParams length.rpptx
#' @examples
#' x <- read_pptx()
#' color_scheme ( x = x )
#' @family functions for reading presentation information
color_scheme <- function(x) {
  x$masterLayouts$get_color_scheme()
}

dev_png <- function(
  filename,
  width = 800,
  height = 700,
  units = "px",
  res = 150
) {
  gdtools::register_liberationsans()
  ragg::agg_png(
    filename = filename,
    width = width,
    height = height,
    units = units,
    res = res
  )
  par(family = "Liberation Sans")
}


test_that("layout summary", {
  x <- read_pptx()
  laysum <- layout_summary(x)
  expect_is(laysum, "data.frame")
  expect_true(all(c("layout", "master") %in% names(laysum)))
  expect_is(laysum$layout, "character")
  expect_is(laysum$master, "character")
})


test_that("layout summary - layout order (#596)", {
  file <- test_path("docs_dir", "test-layouts-ordering.pptx")
  x <- read_pptx(file)
  df <- layout_summary(x)
  order_exp <- c(
    "Title Slide",
    "Title and Content",
    "Section Header",
    "Two Content",
    "Comparison",
    "Title Only",
    "Blank",
    "layout_8",
    "layout_9",
    "layout_10",
    "layout_11"
  )
  expect_equal(df$layout, order_exp)
  df <- x$slideLayouts$get_metadata() # used inside layout_summary
  expect_true(all(get_file_index(df$filename) == 1:11))

  file <- test_path("docs_dir", "test-layouts-ordering-3-masters.pptx")
  x <- read_pptx(file)
  df <- layout_summary(x)
  la <- c(
    "Title Slide",
    "Title and Content",
    "Section Header",
    "Two Content",
    "Comparison",
    "Title Only",
    "Blank"
  )
  order_exp <- rep(la, 3)
  expect_equal(df$layout, order_exp)
  order_exp <- rep(paste0("Master_", 1:3), each = length(la))
  expect_equal(df$master, order_exp)
})


test_that("layout properties", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title", location = ph_location_type(type = "title"))

  laypr <- layout_properties(
    x,
    layout = "Title and Content",
    master = "Office Theme"
  )

  expect_is(laypr, "data.frame")
  expect_true(all(
    c(
      "master_name",
      "name",
      "type",
      "offx",
      "offy",
      "cx",
      "cy",
      "rotation"
    ) %in%
      names(laypr)
  ))
  expect_is(laypr$master_name, "character")
  expect_is(laypr$name, "character")
  expect_is(laypr$type, "character")
  expect_is(laypr$offx, "numeric")
  expect_is(laypr$offy, "numeric")
  expect_is(laypr$cx, "numeric")
  expect_is(laypr$cy, "numeric")
  expect_is(laypr$rotation, "numeric")
})


test_that("layout properties - all phs for multiple masters (#597)", {
  file <- test_path("docs_dir/test-three-identical-masters.pptx")
  x <- read_pptx(file)
  lap <- layout_properties(x)

  expect_true(all(table(lap$ph) == 3)) # 3 identical masters => each ph 3 times

  l_df <- split(lap, lap$master_name) # phs sorted by y coords
  is_y_sorted <- vapply(
    l_df,
    function(x) {
      all(diff(x$offy) >= 0)
    },
    logical(1)
  )
  expect_true(all(is_y_sorted))
})


test_that("plot layout properties (part 1)", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  skip_if_not_installed("gdtools")
  require(doconv)
  local_edition(3L)
  x <- read_pptx()

  png1 <- tempfile(fileext = ".png")
  dev_png(png1, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x,
    layout = "Title Slide",
    master = "Office Theme"
  )
  dev.off()

  png2 <- tempfile(fileext = ".png")
  dev_png(png2, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x,
    layout = "Title Slide",
    master = "Office Theme",
    labels = TRUE,
    type = FALSE,
    id = FALSE,
    title = FALSE
  )
  dev.off()

  png3 <- tempfile(fileext = ".png")
  dev_png(png3, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x,
    layout = "Title Slide",
    master = "Office Theme",
    legend = TRUE
  )
  dev.off()

  expect_snapshot_doc(
    name = "plot-titleslide-layout-default",
    x = png1,
    engine = "testthat"
  )
  expect_snapshot_doc(
    name = "plot-titleslide-layout-labels-only",
    x = png2,
    engine = "testthat"
  )
  expect_snapshot_doc(
    name = "plot-titleslide-layout-default-with-legend",
    x = png3,
    engine = "testthat"
  )

  # issue #604
  p <- test_path("docs_dir/test-content-order.pptx")
  x <- read_pptx(p)

  png4 <- tempfile(fileext = ".png")
  dev_png(png4, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x,
    layout = "Many Contents",
    master = "Office Theme"
  )
  dev.off()

  png5 <- tempfile(fileext = ".png")
  dev_png(png5, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x,
    layout = "Many Contents",
    master = "Office Theme",
    labels = TRUE,
    type = FALSE,
    id = FALSE,
    title = FALSE
  )
  dev.off()

  expect_snapshot_doc(
    name = "plot-content-order-default",
    x = png4,
    engine = "testthat"
  )
  expect_snapshot_doc(
    name = "plot-content-order-labels-only",
    x = png5,
    engine = "testthat"
  )
})


test_that("plot layout properties (part 2)", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  # issue #645: cex arg flexibility
  x <- read_pptx()
  expect_no_error(plot_layout_properties(x, "Title and Content", cex = 0))
  expect_no_error(plot_layout_properties(x, "Title and Content", cex = 1:3))
  expect_no_error(plot_layout_properties(
    x,
    "Title and Content",
    cex = as.list(1:3)
  ))
  expect_no_error(plot_layout_properties(
    x,
    "Title and Content",
    cex = list(t = 1, i = 2)
  ))
  expect_no_error(plot_layout_properties(
    x,
    "Title and Content",
    cex = c(t = 1, i = 2)
  ))

  # issue #645: plot default layout if there are no slides yet, else  show current slide's layout
  x <- read_pptx()
  expect_error(
    plot_layout_properties(x),
    "No `layout` selected and no slides in presentation",
    fixed = TRUE
  )
  x <- layout_default(x, "Two Content")
  expect_message(
    plot_layout_properties(x),
    'Showing default layout: "Two Content"',
    fixed = TRUE
  )
  x <- add_slide(x, "Title and Content")
  expect_message(
    plot_layout_properties(x),
    "Showing current slide's layout: \"Title and Content\"",
    fixed = TRUE
  )
  # issue 666: specify which slide's layout to plot by index
  expect_message(
    plot_layout_properties(x, slide_idx = 1),
    "Showing layout of slide 1: \"Title and Content\"",
    fixed = TRUE
  )
  expect_no_error(plot_layout_properties(x, "Title and Content", legend = TRUE))
})


par(family = "")

test_that("slide summary", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title", location = ph_location_type(type = "title"))

  sm <- slide_summary(x)

  expect_is(sm, "data.frame")
  expect_equal(nrow(sm), 2)
  expect_true(all(c("id", "type", "offx", "offy", "cx", "cy") %in% names(sm)))
  expect_is(sm$id, "character")
  expect_is(sm$type, "character")
  expect_true(is.double(sm$offx))
  expect_true(is.double(sm$offy))
  expect_true(is.double(sm$cx))
  expect_true(is.double(sm$cy))
})


test_that("plot layout properties: layout arg takes numeric index", {
  x <- read_pptx()
  las <- layout_summary(x)
  ii <- as.numeric(rownames(las))

  discarded_plot <- function(x, layout = NULL, master = NULL) {
    # avoid Rplots.pdf creation
    file <- tempfile(fileext = ".png")
    dev_png(file, width = 7, height = 6, res = 150, units = "in")
    plot_layout_properties(x, layout, master)
    dev.off()
  }

  for (idx in ii) {
    expect_no_error(discarded_plot(x, idx))
  }
  expect_no_error(discarded_plot(x, 1, "Office Theme"))
})


test_that("color scheme", {
  x <- read_pptx()
  cs <- color_scheme(x)

  expect_is(cs, "data.frame")
  expect_equal(ncol(cs), 4)
  expect_true(all(c("name", "type", "value", "theme") %in% names(cs)))
  expect_is(cs$name, "character")
  expect_is(cs$type, "character")
  expect_is(cs$value, "character")
  expect_is(cs$theme, "character")
})

test_that("layout summary", {
  x <- read_pptx()
  laysum <- layout_summary(x)
  expect_is(laysum, "data.frame")
  expect_true(all(c("layout", "master") %in% names(laysum)))
  expect_is(laysum$layout, "character")
  expect_is(laysum$master, "character")
})


test_that("layout summary - layout order (#596)", {
  file <- test_path("docs_dir", "test-layouts-ordering.pptx")
  x <- read_pptx(file)
  df <- layout_summary(x)
  order_exp <- c(
    "Title Slide", "Title and Content", "Section Header", "Two Content", "Comparison",
    "Title Only", "Blank", "layout_8", "layout_9", "layout_10", "layout_11"
  )
  expect_equal(df$layout, order_exp)
  df <- x$slideLayouts$get_metadata() # used inside layout_summary
  expect_true(all(get_file_index(df$filename) == 1:11))

  file <- test_path("docs_dir", "test-layouts-ordering-3-masters.pptx")
  x <- read_pptx(file)
  df <- layout_summary(x)
  la <- c("Title Slide", "Title and Content", "Section Header", "Two Content", "Comparison", "Title Only", "Blank")
  order_exp <- rep(la, 3)
  expect_equal(df$layout, order_exp)
  order_exp <- rep(paste0("Master_", 1:3), each = length(la))
  expect_equal(df$master, order_exp)
})


test_that("layout properties", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title", location = ph_location_type(type = "title"))

  laypr <- layout_properties(x, layout = "Title and Content", master = "Office Theme")

  expect_is(laypr, "data.frame")
  expect_true(all(c("master_name", "name", "type", "offx", "offy", "cx", "cy", "rotation") %in% names(laypr)))
  expect_is(laypr$master_name, "character")
  expect_is(laypr$name, "character")
  expect_is(laypr$type, "character")
  expect_is(laypr$offx, "numeric")
  expect_is(laypr$offy, "numeric")
  expect_is(laypr$cx, "numeric")
  expect_is(laypr$cy, "numeric")
  expect_is(laypr$rotation, "numeric")
})


test_that("layout properties - all phs for multiple masters (#597)", {
  file <- test_path("docs_dir/test-three-identical-masters.pptx")
  x <- read_pptx(file)
  lap <- layout_properties(x)

  expect_true(all(table(lap$ph) == 3)) # 3 identical masters => each ph 3 times

  l_df <- split(lap, lap$master_name) # phs sorted by y coords
  is_y_sorted <- vapply(l_df, function(x) {
    all(diff(x$offy) >= 0)
  }, logical(1))
  expect_true(all(is_y_sorted))
})


save_png <- function(code, width = 700, height = 700) {
  path <- tempfile(fileext = ".png")
  png(path, width = width, height = height, res = 150)
  on.exit(dev.off())
  code

  path
}


test_that("plot layout properties", {
  skip_if_not_installed("doconv")
  skip_if_not(doconv::msoffice_available())
  require(doconv)
  local_edition(3L)
  x <- read_pptx()

  png1 <- tempfile(fileext = ".png")
  png(png1, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x, layout = "Title Slide", master = "Office Theme"
  )
  dev.off()

  png2 <- tempfile(fileext = ".png")
  png(png2, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x, layout = "Title Slide", master = "Office Theme",
    labels = TRUE, type = FALSE, id = FALSE, title = FALSE
  )
  dev.off()

  png3 <- tempfile(fileext = ".png")
  png(png3, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x, layout = "Title Slide", master = "Office Theme", legend = TRUE
  )
  dev.off()

  expect_snapshot_doc(name = "plot-titleslide-layout-default", x = png1, engine = "testthat")
  expect_snapshot_doc(name = "plot-titleslide-layout-labels-only", x = png2, engine = "testthat")
  expect_snapshot_doc(name = "plot-titleslide-layout-default-with-legend", x = png3, engine = "testthat")

  # issue #604
  p <- test_path("docs_dir/test-content-order.pptx")
  x <- read_pptx(p)

  png4 <- tempfile(fileext = ".png")
  png(png4, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x, layout = "Many Contents", master = "Office Theme"
  )
  dev.off()

  png5 <- tempfile(fileext = ".png")
  png(png5, width = 7, height = 6, res = 150, units = "in")
  plot_layout_properties(
    x = x, layout = "Many Contents", master = "Office Theme",
    labels = TRUE, type = FALSE, id = FALSE, title = FALSE
  )
  dev.off()

  expect_snapshot_doc(name = "plot-content-order-default", x = png4, engine = "testthat")
  expect_snapshot_doc(name = "plot-content-order-labels-only", x = png5, engine = "testthat")
})


test_that("slide summary", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title", location = ph_location_type(type = "title"))

  sm <- slide_summary(x)

  expect_is(sm, "data.frame")
  expect_equal(nrow(sm), 2)
  expect_true(all(c("id", "type", "offx", "offy", "cx", "cy") %in% names(sm)))
  expect_is(sm$id, "character")
  expect_is(sm$type, "character")
  expect_true(is.double(sm$offx))
  expect_true(is.double(sm$offy))
  expect_true(is.double(sm$cx))
  expect_true(is.double(sm$cy))
})



test_that("plot layout properties: layout arg takes numeric index", {
  x <- read_pptx()
  las <- layout_summary(x)
  ii <- as.numeric(rownames(las))

  discarded_plot <- function(x, layout = NULL, master = NULL) { # avoid Rplots.pdf creation
    file <- tempfile(fileext = ".png")
    png(file, width = 7, height = 6, res = 150, units = "in")
      plot_layout_properties(x, layout, master)
    dev.off()
  }

  for (idx in ii) {
    expect_no_error(discarded_plot(x, idx))
  }
  expect_no_error(discarded_plot(x, 1, "Office Theme"))
})


test_that("color scheme", {
  x <- read_pptx()
  cs <- color_scheme(x)

  expect_is(cs, "data.frame")
  expect_equal(ncol(cs), 4)
  expect_true(all(c("name", "type", "value", "theme") %in% names(cs)))
  expect_is(cs$name, "character")
  expect_is(cs$type, "character")
  expect_is(cs$value, "character")
  expect_is(cs$theme, "character")
})

#' @export
#' @title Add a page break in a 'Word' document
#' @description add a page break into an rdocx object
#' @param x an rdocx object
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @examples
#' doc <- read_docx()
#' doc <- body_add_break(doc)
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_break <- function(x, pos = "after") {
  str <- runs_to_p_wml(run_pagebreak(), add_ns = TRUE)
  body_add_xml(x = x, str = str, pos = pos)
}

#' @export
#' @title Add an image in a 'Word' document
#' @description add an image into an rdocx object.
#' @inheritParams body_add_break
#' @param src image filename, the basename of the file must not contain any blank.
#' @param style paragraph style
#' @param width height in inches
#' @param height height in inches
#' @examples
#' doc <- read_docx()
#'
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#' if (file.exists(img.file)) {
#'   doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)
#' }
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_img <- function(x, src, style = NULL, width, height, pos = "after") {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", src)

  if (file_type %in% ".svg") {
    if (!requireNamespace("rsvg")) {
      stop("package 'rsvg' is required to convert svg file to rasters.")
    }

    file <- tempfile(fileext = ".png")
    rsvg::rsvg_png(src, file = file)
    src <- file
    file_type <- ".png"
  }

  new_src <- tempfile(fileext = file_type)
  file.copy(src, to = new_src)

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  ext_img <- external_img(new_src, width = width, height = height)
  xml_elt <- runs_to_p_wml(ext_img, add_ns = TRUE, style_id = style_id)

  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add an external docx in a 'Word' document
#' @description Add content of a docx into an rdocx object.
#'
#' The function is using a 'Microsoft Word' feature: when the
#' document will be edited, the content of the file will be
#' inserted in the main document.
#'
#' This feature is unlikely to work as expected if the
#' resulting document is edited by another software.
#'
#' The file is added when the method `print()` that
#' produces the final Word file is called, so don't remove
#' file defined with `src` before.
#' @inheritParams body_add_break
#' @param src docx filename, the path of the file must not contain
#' any '&' and the basename must not contain any space.
#' @examples
#' file1 <- tempfile(fileext = ".docx")
#' file2 <- tempfile(fileext = ".docx")
#' file3 <- tempfile(fileext = ".docx")
#' x <- read_docx()
#' x <- body_add_par(x, "hello world 1", style = "Normal")
#' print(x, target = file1)
#'
#' x <- read_docx()
#' x <- body_add_par(x, "hello world 2", style = "Normal")
#' print(x, target = file2)
#'
#' x <- read_docx(path = file1)
#' x <- body_add_break(x)
#' x <- body_add_docx(x, src = file2)
#' print(x, target = file3)
#' @export
#' @family functions for adding content
body_add_docx <- function(x, src, pos = "after") {

  if(grepl(" ", basename(src))){
    stop("The basename of the file to be inserted cannot contain spaces.")
  }

  xml_elt <- to_wml(block_pour_docx(file = src), add_ns = TRUE)
  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add a 'ggplot' in a 'Word' document
#' @description add a ggplot as a png image into an rdocx object.
#' @inheritParams body_add_break
#' @param value ggplot object
#' @param style paragraph style
#' @param width height in inches
#' @param height height in inches
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @param ... Arguments to be passed to png function.
#' @importFrom grDevices dev.off
#' @examples
#' if (require("ggplot2")) {
#'   doc <- read_docx()
#'
#'   gg_plot <- ggplot(data = iris) +
#'     geom_point(mapping = aes(Sepal.Length, Petal.Length))
#'
#'   if (capabilities(what = "png")) {
#'     doc <- body_add_gg(doc, value = gg_plot, style = "centered")
#'   }
#'
#'   print(doc, target = tempfile(fileext = ".docx"))
#' }
#' @family functions for adding content
#' @importFrom ragg agg_png
body_add_gg <- function(x, value, width = 6, height = 5, res = 300, style = "Normal", scale = 1, pos = "after", ...) {
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  stopifnot(inherits(value, "gg"))
  file <- tempfile(fileext = ".png")
  agg_png(filename = file, width = width, height = height, scaling = scale, units = "in", res = res, background = "transparent", ...)
  print(value)
  dev.off()
  on.exit(unlink(file))
  body_add_img(x, src = file, style = style, width = width, height = height, pos = pos)
}


#' @export
#' @title Add a list of blocks into a 'Word' document
#' @description add a list of blocks produced by \code{block_list} into
#' into an rdocx object.
#' @inheritParams body_add_break
#' @param blocks set of blocks to be used as footnote content returned by
#'   function [block_list()].
#' @examples
#' library(officer)
#'
#' img.file <- file.path(R.home("doc"), "html", "logo.jpg")
#'
#' bl <- block_list(
#'   fpar(ftext("hello", shortcuts$fp_bold(color = "red"))),
#'   fpar(
#'     ftext("hello world", shortcuts$fp_bold()),
#'     external_img(src = img.file, height = 1.06, width = 1.39),
#'     fp_p = fp_par(text.align = "center")
#'   )
#' )
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_blocks(doc_1, blocks = bl)
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_blocks <- function(x, blocks, pos = "after") {
  stopifnot(inherits(blocks, "block_list"))

  if (length(blocks) > 0) {
    pos_vector <- rep("after", length(blocks))
    pos_vector[1] <- pos
    for (i in seq_along(blocks)) {
      x <- body_add_xml(x,
        str = to_wml(blocks[[i]], add_ns = TRUE),
        pos = pos_vector[i]
      )
    }
  }

  x
}




#' @export
#' @title Add paragraphs of text in a 'Word' document
#' @description add a paragraph of text into an rdocx object
#' @param x a docx device
#' @param value a character
#' @param style paragraph style name
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "A title", style = "heading 1")
#' doc <- body_add_par(doc, "Hello world!", style = "Normal")
#' doc <- body_add_par(doc, "centered text", style = "centered")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_par <- function(x, value, style = NULL, pos = "after") {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr><w:r><w:t xml:space=\"preserve\">",
    htmlEscapeCopy(value), "</w:t></w:r></w:p>"
  )
  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add fpar in a 'Word' document
#' @description Add an \code{fpar} (a formatted paragraph)
#' into an rdocx object.
#' @param x a docx device
#' @param value a character
#' @param style paragraph style. If NULL, paragraph settings from `fpar` will be used. If not
#' NULL, it must be a paragraph style name (located in the template
#' provided as `read_docx(path = ...)`); in that case, paragraph settings from `fpar` will be
#' ignored.
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @examples
#' bold_face <- shortcuts$fp_bold(font.size = 30)
#' bold_redface <- update(bold_face, color = "red")
#' fpar_ <- fpar(
#'   ftext("Hello ", prop = bold_face),
#'   ftext("World", prop = bold_redface),
#'   ftext(", how are you?", prop = bold_face)
#' )
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, fpar_)
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#' # a way of using fpar to center an image in a Word doc ----
#' rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
#' img_in_par <- fpar(
#'   external_img(src = rlogo, height = 1.06 / 2, width = 1.39 / 2),
#'   hyperlink_ftext(
#'     href = "https://cran.r-project.org/index.html",
#'     text = "cran", prop = bold_redface
#'   ),
#'   fp_p = fp_par(text.align = "center")
#' )
#'
#' doc <- read_docx()
#' doc <- body_add_fpar(doc, img_in_par)
#' print(doc, target = tempfile(fileext = ".docx"))
#'
#' @seealso \code{\link{fpar}}
#' @family functions for adding content
body_add_fpar <- function(x, value, style = NULL, pos = "after") {
  if (!is.null(style)) {
    style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")
  } else {
    style_id <- NULL
  }

  xml_elt <- to_wml(value, add_ns = TRUE, style_id = style_id)

  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add table in a 'Word' document
#' @description Add a table into an rdocx object.
#' @param x a docx device
#' @param value a data.frame to add as a table
#' @param style table style
#' @param pos where to add the new element relative to the cursor,
#' one of after", "before", "on".
#' @param header display header if TRUE
#' @param alignment columns alignement, argument length must match with columns length,
#' values must be "l" (left), "r" (right) or "c" (center).
#' @param align_table table alignment within document, value must be "left", "center" or "right"
#' @param stylenames columns styles defined by [table_stylenames()]
#' @param first_row Specifies that the first column conditional formatting should be
#' applied. Details for this and other conditional formatting options can be found
#' at http://officeopenxml.com/WPtblLook.php.
#' @param last_row Specifies that the first column conditional formatting should be applied.
#' @param first_column Specifies that the first column conditional formatting should
#' be applied.
#' @param last_column Specifies that the first column conditional formatting should be
#' applied.
#' @param no_hband Specifies that the first column conditional formatting should be applied.
#' @param no_vband Specifies that the first column conditional formatting should be applied.
#' @examples
#' doc <- read_docx()
#' doc <- body_add_table(doc, iris, style = "table_template")
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_table <- function(x, value, style = NULL, pos = "after", header = TRUE,
                           alignment = NULL,
                           align_table = "center",
                           stylenames = table_stylenames(),
                           first_row = TRUE, first_column = FALSE,
                           last_row = FALSE, last_column = FALSE,
                           no_hband = FALSE, no_vband = TRUE) {
  pt <- prop_table(
    style = style, layout = table_layout(),
    width = table_width(), stylenames = stylenames,
    tcf = table_conditional_formatting(
      first_row = first_row, first_column = first_column,
      last_row = last_row, last_column = last_column,
      no_hband = no_hband, no_vband = no_vband
    ),
    align = align_table
  )

  bt <- block_table(x = value, header = header, properties = pt, alignment = alignment)
  xml_elt <- to_wml(bt, add_ns = TRUE, base_document = x)
  body_add_xml(x = x, str = xml_elt, pos = pos)
}

#' @export
#' @title Add table of content in a 'Word' document
#' @description Add a table of content into an rdocx object.
#' The TOC will be generated by Word, if the document is not
#' edited with Word (i.e. Libre Office) the TOC will not be generated.
#' @param x an rdocx object
#' @param level max title level of the table
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @param style optional. style in the document that will be used to build entries of the TOC.
#' @param separator optional. Some configurations need "," (i.e. from Canada) separator instead of ";"
#' @examples
#' doc <- read_docx()
#' doc <- body_add_toc(doc)
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_toc <- function(x, level = 3, pos = "after", style = NULL, separator = ";") {
  bt <- block_toc(level = level, style = style, separator = separator)
  out <- to_wml(bt, add_ns = TRUE)
  body_add_xml(x = x, str = out, pos = pos)
}




#' @export
#' @title Add plot in a 'Word' document
#' @description Add a plot as a png image into an rdocx object.
#' @inheritParams body_add_break
#' @param value plot instructions, see [plot_instr()].
#' @param style paragraph style
#' @param width height in inches
#' @param height height in inches
#' @param res resolution of the png image in ppi
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @param ... Arguments to be passed to png function.
#' @importFrom grDevices dev.off
#' @examples
#' doc <- read_docx()
#'
#' if (capabilities(what = "png")) {
#'   doc <- body_add_plot(doc,
#'     value = plot_instr(
#'       code = {
#'         barplot(1:5, col = 2:6)
#'       }
#'     ),
#'     style = "centered"
#'   )
#' }
#'
#' print(doc, target = tempfile(fileext = ".docx"))
#' @family functions for adding content
body_add_plot <- function(x, value, width = 6, height = 5, res = 300, style = "Normal", pos = "after", ...) {
  file <- tempfile(fileext = ".png")
  agg_png(filename = file, width = width, height = height, units = "in", res = res, background = "transparent", ...)
  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )
  on.exit(unlink(file))
  body_add_img(x, src = file, style = style, width = width, height = height, pos = pos)
}


#' @export
#' @title Add Word caption in a 'Word' document
#' @description Add a Word caption into an rdocx object.
#' @param x an rdocx object
#' @param value an object returned by [block_caption()]
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @family functions for adding content
#' @examples
#' doc <- read_docx()
#'
#' if (capabilities(what = "png")) {
#'   doc <- body_add_plot(doc,
#'     value = plot_instr(
#'       code = {
#'         barplot(1:5, col = 2:6)
#'       }
#'     ),
#'     style = "centered"
#'   )
#' }
#' run_num <- run_autonum(
#'   seq_id = "fig", pre_label = "Figure ",
#'   bkm = "barplot"
#' )
#' caption <- block_caption("a barplot",
#'   style = "Normal",
#'   autonum = run_num
#' )
#' doc <- body_add_caption(doc, caption)
#' print(doc, target = tempfile(fileext = ".docx"))
body_add_caption <- function(x, value, pos = "after") {
  stopifnot(inherits(value, "block_caption"))
  if (!value$style %in% x$styles$style_name) {
    stop("caption is using style ", shQuote(value$style), " that does not exist in the Word document.")
  }
  out <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml(x = x, str = out, pos = pos)
}




#' @export
#' @title Add an xml string as document element
#' @description Add an xml string as document element in the document. This function
#' is to be used to add custom openxml code.
#' @param x an rdocx object
#' @param str a wml string
#' @param pos where to add the new element relative to the cursor,
#' one of "after", "before", "on".
#' @keywords internal
body_add_xml <- function(x, str, pos = c("after", "before", "on")) {
  xml_elt <- as_xml_document(str)
  pos <- match.arg(pos)

  cursor_elt <- docx_current_block_xml(x)
  if (is.null(cursor_elt)) {
    xml_add_child(xml_find_first(x$doc_obj$get(), "/w:document/w:body"),
      xml_elt,
      .where = 0
    )
    .name <- xml_name(xml_elt)
    x$officer_cursor <- cursor_append(x$officer_cursor, .name)
  } else if (pos == "on") {
    xml_replace(cursor_elt, xml_elt)
    x$officer_cursor <- cursor_replace_nodename(x$officer_cursor, xml_name(xml_elt))
  } else if (pos == "after") {
    xml_add_sibling(cursor_elt, xml_elt, .where = pos)
    x$officer_cursor <- cursor_add_after(x$officer_cursor, xml_name(xml_elt))
  } else {
    xml_add_sibling(cursor_elt, xml_elt, .where = pos)
    x$officer_cursor <- cursor_add_before(x$officer_cursor, xml_name(xml_elt))
  }
  x
}

body_add_xml2 <- function(x, str) {
  x <- cursor_end(x)
  xml_elt <- as_xml_document(str)
  cursor_elt <- docx_current_block_xml(x)
  if (is.null(cursor_elt)) {
    xml_add_child(xml_find_first(x$doc_obj$get(), "/w:document/w:body"),
      xml_elt,
      .where = 0
    )
    x$officer_cursor <- cursor_append(x$officer_cursor, xml_name(xml_elt))
  } else {
    xml_add_sibling(cursor_elt, xml_elt, .where = "after")
    x$officer_cursor <- cursor_append(x$officer_cursor, xml_name(cursor_elt))
  }

  x
}

#' @export
#' @title Add bookmark in a 'Word' document
#' @description Add a bookmark at the cursor location. The bookmark
#' is added on the first run of text in the current paragraph.
#' @param x an rdocx object
#' @param id bookmark name
#' @examples
#'
#' # cursor_bookmark ----
#'
#' doc <- read_docx()
#' doc <- body_add_par(doc, "centered text", style = "centered")
#' doc <- body_bookmark(doc, "text_to_replace")
body_bookmark <- function(x, id) {
  cursor_elt <- docx_current_block_xml(x)
  ns_ <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\""
  new_id <- uuid_generate()
  id <- check_bookmark_id(id)

  bm_start_str <- sprintf("<w:bookmarkStart w:id=\"%s\" w:name=\"%s\" %s/>", new_id, id, ns_)
  bm_start_end <- sprintf("<w:bookmarkEnd %s w:id=\"%s\"/>", ns_, new_id)

  path_ <- paste0(xml_path(cursor_elt), "//w:r")

  node <- xml_find_first(x$doc_obj$get(), path_)
  xml_add_sibling(node, as_xml_document(bm_start_str), .where = "before")
  xml_add_sibling(node, as_xml_document(bm_start_end), .where = "after")

  x
}

#' @export
#' @title Remove an element in a 'Word' document
#' @description Remove element pointed by cursor from a
#' 'Word' document.
#' @param x an rdocx object
#' @examples
#' library(officer)
#'
#' str1 <- rep("Lorem ipsum dolor sit amet, consectetur adipiscing elit. ", 20)
#' str1 <- paste(str1, collapse = "")
#'
#' str2 <- "Drop that text"
#'
#' str3 <- rep("Aenean venenatis varius elit et fermentum vivamus vehicula. ", 20)
#' str3 <- paste(str3, collapse = "")
#'
#' my_doc <- read_docx()
#' my_doc <- body_add_par(my_doc, value = str1, style = "Normal")
#' my_doc <- body_add_par(my_doc, value = str2, style = "centered")
#' my_doc <- body_add_par(my_doc, value = str3, style = "Normal")
#'
#' new_doc_file <- print(my_doc,
#'   target = tempfile(fileext = ".docx")
#' )
#'
#' my_doc <- read_docx(path = new_doc_file)
#' my_doc <- cursor_reach(my_doc, keyword = "that text")
#' my_doc <- body_remove(my_doc)
#'
#' print(my_doc, target = tempfile(fileext = ".docx"))
body_remove <- function(x) {
  cursor_elt <- docx_current_block_xml(x)

  if (is.null(cursor_elt)) {
    warning("There is nothing left to remove in the document")
    return(x)
  }
  x$officer_cursor <- cursor_delete(x$officer_cursor, xml_name(cursor_elt))
  xml_remove(cursor_elt)
  x
}

#' @export
#' @title Add comment in a 'Word' document
#' @description Add a comment at the cursor location. The comment
#' is added on the first run of text in the current paragraph.
#' @param x an rdocx object
#' @param cmt a set of blocks to be used as comment content returned by
#'   function [block_list()].
#' @param author comment author.
#' @param date comment date
#' @param initials comment initials
#' @examples
#' doc <- read_docx()
#' doc <- body_add_par(doc, "Paragraph")
#' doc <- body_comment(doc, block_list("This is a comment."))
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_comments(read_docx(docx_file))
body_comment <- function(x,
                         cmt = ftext(""),
                         author = "",
                         date = "",
                         initials = "") {
  cursor_elt <- docx_current_block_xml(x)
  ns_ <- "xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\""

  open_tag <- wr_ns_yes

  blocks <- sapply(cmt, to_wml)
  blocks <- paste(blocks, collapse = "")

  id <- basename(tempfile(pattern = "comment"))

  cmt_xml <- paste0(
    sprintf(
      "<w:comment %s  w:id=\"%s\" w:author=\"%s\" w:date=\"%s\" w:initials=\"%s\">",
      ns_, id, author, date, initials
    ),
    blocks,
    "</w:comment>"
  )

  cmt_start_str <- sprintf("<w:commentRangeStart w:id=\"%s\" %s/>", id, ns_)
  cmt_start_end <- sprintf("<w:commentRangeEnd %s w:id=\"%s\"/>", ns_, id)

  path_ <- paste0(xml_path(cursor_elt), "//w:r")

  cmt_ref_xml <- paste0(
    open_tag,
    if (!is.null(x$pr)) rpr_wml(x$pr),
    "<w:commentReference w:id=\"",
    id,
    "\">",
    cmt_xml,
    "</w:commentReference>",
    "</w:r>"
  )

  path_ <- paste0(xml_path(cursor_elt), "//w:r")

  node <- xml_find_first(x$doc_obj$get(), path_)
  xml_add_sibling(node, as_xml_document(cmt_start_str), .where = "before")
  xml_add_sibling(node, as_xml_document(cmt_start_end), .where = "after")
  xml_add_sibling(node, as_xml_document(cmt_ref_xml), .where = "after")

  x
}

# body_add and methods -----
#' @export
#' @title Add content into a Word document
#' @description This function add objects into a Word document. Values are added
#' as new paragraphs or tables.
#'
#' This function is experimental and will replace the `body_add_*` functions
#' later. For now it is only to be used for successive additions and cannot
#' be used in conjunction with the `body_add_*` functions.
#' @param x an rdocx object
#' @param value object to add in the document. Supported objects
#' are vectors, data.frame, graphics, block of formatted paragraphs,
#' unordered list of formatted paragraphs,
#' pretty tables with package flextable, 'Microsoft' charts with package mschart.
#' @param ... further arguments passed to or from other methods. When
#' adding a `ggplot` object or [plot_instr], these arguments will be used
#' by png function. See method signatures to see what arguments can be used.
#' @examples
#' doc_1 <- read_docx()
#' doc_1 <- body_add(doc_1, "Table of content", style = "heading 1")
#' doc_1 <- body_add(doc_1, block_toc())
#' doc_1 <- body_add(doc_1, run_pagebreak())
#' doc_1 <- body_add(doc_1, "A title", style = "heading 1")
#' doc_1 <- body_add(doc_1, head(iris), style = "table_template")
#' doc_1 <- body_add(doc_1, "Another title", style = "heading 1")
#' doc_1 <- body_add(doc_1, letters, style = "Normal")
#' doc_1 <- body_add(
#'   doc_1,
#'   block_section(prop_section(type = "continuous"))
#' )
#' doc_1 <- body_add(doc_1, plot_instr(code = barplot(1:5, col = 2:6)))
#' doc_1 <- body_add(
#'   doc_1,
#'   block_section(prop_section(page_size = page_size(orient = "landscape")))
#' )
#' print(doc_1, target = tempfile(fileext = ".docx"))
#' # print(doc_1, target = "test.docx")
#' @section Illustrations:
#'
#' \if{html}{\figure{body_add_doc_1.png}{options: width=70\%}}
#' @keywords internal
body_add <- function(x, value, ...) {
  UseMethod("body_add", value)
}


#' @export
#' @describeIn body_add add a character vector.
#' @param style paragraph style name. These names are available with function [styles_info]
#' and are the names of the Word styles defined in the base document (see
#' argument `path` from [read_docx]).
body_add.character <- function(x, value, style = NULL, ...) {
  if (!is.null(style)) {
    style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")
  } else {
    style_id <- NULL
  }

  runs <- paste0(
    "<w:r><w:t xml:space=\"preserve\">",
    htmlEscapeCopy(value), "</w:t></w:r>"
  )

  xml_elt <- paste0(wp_ns_yes, "<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>", runs, "</w:p>")
  for (str in xml_elt) {
    x <- body_add_xml2(x = x, str = str)
  }
  x
}

#' @export
#' @describeIn body_add add a numeric vector.
#' @param format_fun a function to be used to format values.
body_add.numeric <- function(x, value, style = NULL, format_fun = formatC, ...) {
  value <- format_fun(value, ...)
  body_add(x, value = value, style = style, ...)
}

#' @export
#' @describeIn body_add add a factor vector.
body_add.factor <- function(x, value, style = NULL, format_fun = as.character, ...) {
  value <- format_fun(value)
  body_add(x, value = value, style = style, ...)
}



#' @export
#' @describeIn body_add add a [fpar] object. These objects enable
#' the creation of formatted paragraphs made of formatted chunks of text.
body_add.fpar <- function(x, value, style = NULL, ...) {
  img_src <- sapply(value$chunks, function(x) {
    if (inherits(x, "external_img")) {
      as.character(x)
    } else {
      NA_character_
    }
  })
  img_src <- unique(img_src[!is.na(img_src)])

  if (!is.null(style)) {
    style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")
  } else {
    style_id <- NULL
  }

  xml_elt <- to_wml(value, add_ns = TRUE, style_id = style_id)

  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @param header display header if TRUE
#' @param tcf conditional formatting settings defined by [table_conditional_formatting()]
#' @param alignment columns alignement, argument length must match with columns length,
#' values must be "l" (left), "r" (right) or "c" (center).
#' @describeIn body_add add a data.frame object with [block_table()].
body_add.data.frame <- function(x, value, style = NULL, header = TRUE,
                                tcf = table_conditional_formatting(),
                                alignment = NULL,
                                ...) {
  pt <- prop_table(
    style = style, layout = table_layout(),
    width = table_width(),
    tcf = tcf
  )

  bt <- block_table(x = value, header = header, properties = pt, alignment = alignment)
  xml_elt <- to_wml(bt, add_ns = TRUE, base_document = x)

  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add a [block_caption] object. These objects enable
#' the creation of set of formatted paragraphs made of formatted chunks of text.
body_add.block_caption <- function(x, value, ...) {

  if (!value$style %in% x$styles$style_name) {
    stop("caption is using style ", shQuote(value$style), " that does not exist in the Word document.")
  }

  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add a [block_list] object.
body_add.block_list <- function(x, value, ...) {
  if (length(value) > 0) {
    for (i in seq_along(value)) {
      x <- body_add(x, value = value[[i]])
    }
  }

  x
}

#' @export
#' @describeIn body_add add a table of content (a [block_toc] object).
body_add.block_toc <- function(x, value, ...) {
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add an image (a [external_img] object).
body_add.external_img <- function(x, value, style = "Normal", ...) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  file_type <- gsub("(.*)(\\.[a-zA-Z0-0]+)$", "\\2", value)

  if (file_type %in% ".svg") {
    if (!requireNamespace("rsvg")) {
      stop("package 'rsvg' is required to convert svg file to rasters.")
    }

    file <- tempfile(fileext = ".png")
    rsvg::rsvg_png(src, file = file)
    value <- external_img(
      src = file,
      width = attr(value, "dims")$width,
      height = attr(value, "dims")$height
    )
    file_type <- ".png"
  }

  new_src <- tempfile(fileext = file_type)
  file.copy(value, to = new_src)
  value[1] <- new_src

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>",
    to_wml(value, add_ns = FALSE),
    "</w:p>"
  )

  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @describeIn body_add add a [run_pagebreak] object.
body_add.run_pagebreak <- function(x, value, style = NULL, ...) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>",
    to_wml(value, add_ns = FALSE),
    "</w:p>"
  )

  body_add_xml2(x = x, str = xml_elt)
}

#' @export
#' @describeIn body_add add a [run_columnbreak] object.
body_add.run_columnbreak <- function(x, value, style = NULL, ...) {
  if (is.null(style)) {
    style <- x$default_styles$paragraph
  }

  style_id <- get_style_id(data = x$styles, style = style, type = "paragraph")

  xml_elt <- paste0(
    wp_ns_yes,
    "<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>",
    to_wml(value, add_ns = FALSE),
    "</w:p>"
  )

  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add add a ggplot object.
#' @param width height in inches
#' @param height height in inches
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
body_add.gg <- function(x, value, width = 6, height = 5, res = 300, style = "Normal", scale = 1, ...) {
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  file <- tempfile(fileext = ".png")
  agg_png(filename = file, width = width, height = height, units = "in", res = res, scaling = scale, background = "transparent", ...)
  print(value)
  dev.off()
  on.exit(unlink(file))

  value <- external_img(src = file, width = width, height = height)
  body_add(x, value, style = style)
}

#' @export
#' @describeIn body_add add a base plot with a [plot_instr] object.
body_add.plot_instr <- function(x, value, width = 6, height = 5, res = 300, style = "Normal", ...) {
  file <- tempfile(fileext = ".png")
  agg_png(filename = file, width = width, height = height, units = "in", res = res, scaling = 1, background = "transparent", ...)
  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )
  on.exit(unlink(file))

  value <- external_img(src = file, width = width, height = height)

  body_add(x, value, style = style)
}


#' @export
#' @describeIn body_add pour content of an external docx file with with a [block_pour_docx] object
body_add.block_pour_docx <- function(x, value, ...) {
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}


#' @export
#' @describeIn body_add ends a section with a [block_section] object
body_add.block_section <- function(x, value, ...) {
  xml_elt <- to_wml(value, add_ns = TRUE, base_document = x)
  body_add_xml2(x = x, str = xml_elt)
}

#' @importFrom xml2 xml_attr<- xml_name<- xml_text<- as_list as_xml_document read_xml
#' @importFrom xml2 write_xml xml_add_child xml_add_parent xml_add_sibling xml_attr xml_attrs xml_child
#' @importFrom xml2 xml_children xml_find_all xml_find_first xml_length xml_missing xml_name xml_ns
#' @importFrom xml2 xml_path xml_remove xml_replace xml_set_attr xml_set_attrs xml_text
#' @importFrom stats setNames

read_xfrm <- function(nodeset, file, name) {
  if (length(nodeset) < 1) {
    return(data.frame(
      stringsAsFactors = FALSE,
      type = character(0),
      id = character(0),
      ph_label = character(0),
      ph = character(0),
      file = character(0),
      offx = integer(0),
      offy = integer(0),
      cx = integer(0),
      cy = integer(0),
      rotation = integer(0),
      name = character(0),
      fld_id = character(0),
      fld_type = character(0)
    ))
  }

  ph <- xml_child(nodeset, "p:nvSpPr/p:nvPr/p:ph")
  type <- xml_attr(ph, "type")
  type[is.na(type)] <- "body"
  id <- xml_attr(xml_child(nodeset, "/p:cNvPr"), "id")
  label <- xml_attr(xml_child(nodeset, "/p:cNvPr"), "name")

  off <- xml_child(nodeset, "p:spPr/a:xfrm/a:off")
  ext <- xml_child(nodeset, "p:spPr/a:xfrm/a:ext")
  rot <- xml_child(nodeset, "p:spPr/a:xfrm")

  fld_id <- xml_attr(xml_child(nodeset, "/p:txBody/a:p/a:fld"), "id")
  fld_type <- xml_attr(xml_child(nodeset, "/p:txBody/a:p/a:fld"), "type")

  data.frame(
    stringsAsFactors = FALSE,
    type = type,
    id = id,
    ph_label = label,
    ph = as.character(ph),
    file = basename(file),
    offx = as.integer(xml_attr(off, "x")),
    offy = as.integer(xml_attr(off, "y")),
    cx = as.integer(xml_attr(ext, "cx")),
    cy = as.integer(xml_attr(ext, "cy")),
    rotation = as.integer(xml_attr(rot, "rot")),
    fld_id,
    fld_type,
    name = name
  )
}


fortify_master_xfrm <- function(master_xfrm) {
  master_xfrm <- as.data.frame(master_xfrm)
  has_type <- grepl("type=", master_xfrm$ph)
  master_xfrm <- master_xfrm[has_type, ]
  if (nrow(master_xfrm) > 0) {
    # see #597
    list_xfrm <- split(master_xfrm, master_xfrm$file)
    list_xfrm <- lapply(list_xfrm, function(x) {
      x[!duplicated(x$type), , drop = FALSE]
    })
    master_xfrm <- do.call("rbind", list_xfrm)
  }

  master_xfrm <- dplyr::rename(
    master_xfrm,
    offx_ref = "offx",
    offy_ref = "offy",
    cx_ref = "cx",
    cy_ref = "cy",
    fld_id_ref = "fld_id",
    fld_type_ref = "fld_type",
    master_name = "name"
  )
  master_xfrm$id <- NULL
  master_xfrm$ph <- NULL
  master_xfrm$ph_label <- NULL
  master_xfrm$rotation <- NULL

  master_xfrm
}


xfrmize <- function(slide_xfrm, master_xfrm) {
  x <- as.data.frame(slide_xfrm)

  master_ref <- unique(data.frame(
    file = master_xfrm$file,
    master_name = master_xfrm$name,
    stringsAsFactors = FALSE
  ))
  master_xfrm <- fortify_master_xfrm(master_xfrm)

  slide_key_id <- paste0(x$master_file, x$type)
  master_key_id <- paste0(master_xfrm$file, master_xfrm$type)

  slide_xfrm_no_match <- x[!slide_key_id %in% master_key_id, ]
  slide_xfrm_no_match <- merge(
    slide_xfrm_no_match,
    master_ref,
    by.x = "master_file",
    by.y = "file",
    all.x = TRUE,
    all.y = FALSE
  )

  x <- merge(
    x,
    master_xfrm,
    by.x = c("master_file", "type"),
    by.y = c("file", "type"),
    all = FALSE
  )
  x$offx <- ifelse(!is.finite(x$offx), x$offx_ref, x$offx)
  x$offy <- ifelse(!is.finite(x$offy), x$offy_ref, x$offy)
  x$cx <- ifelse(!is.finite(x$cx), x$cx_ref, x$cx)
  x$cy <- ifelse(!is.finite(x$cy), x$cy_ref, x$cy)
  x$offx_ref <- NULL
  x$offy_ref <- NULL
  x$cx_ref <- NULL
  x$cy_ref <- NULL
  x$fld_id_ref <- NULL
  x$fld_type_ref <- NULL

  x <- rbind(x, slide_xfrm_no_match, stringsAsFactors = FALSE)

  # intuitive sorting: top -> bottom, left -> right
  x$.i_master <- get_file_index(x$master_file)
  x$.i_layout <- get_file_index(x$file)
  x <- arrange(
    x,
    .data$.i_master,
    .data$.i_layout,
    .data$offy,
    .data$offx
  )
  x$.i_master <- NULL
  x$.i_layout <- NULL
  x <- filter(
    x,
    !(is.na(.data$offx) | is.na(.data$offy) | is.na(.data$cx) | is.na(.data$cy))
  )

  x$type_idx <- stats::ave(
    x$type,
    x$master_file,
    x$file,
    x$type,
    FUN = seq_along
  )
  x$type_idx <- as.numeric(x$type_idx) # NB: ave returns character

  x$id <- as.integer(x$id)

  rownames(x) <- NULL # prevent meaningless rownames
  x
}


read_theme_colors <- function(doc, theme) {
  nodes <- xml_find_all(doc, "//a:clrScheme/*")

  names_ <- xml_name(nodes)
  col_types_ <- xml_name(xml_children(nodes))
  vals <- xml_attr(xml_children(nodes), "val")
  last_colors_ <- xml_attr(xml_children(nodes), "lastClr")
  vals <- ifelse(
    col_types_ == "srgbClr",
    paste0("#", vals),
    paste0("#", last_colors_)
  )
  data.frame(
    stringsAsFactors = FALSE,
    name = names_,
    type = col_types_,
    value = vals,
    theme = theme
  )
}


characterise_df <- function(x) {
  names(x) <- htmlEscapeCopy(names(x))
  x <- lapply(x, function(x) {
    if (is.character(x)) {
      x
    } else if (is.factor(x)) {
      as.character(x)
    } else {
      gsub("(^ | $)+", "", format(x))
    }
  })
  data.frame(x, stringsAsFactors = FALSE, check.names = FALSE)
}


xpath_content_selector <- "*[self::p:cxnSp or self::p:sp or self::p:graphicFrame or self::p:grpSp or self::p:pic]"

as_xpath_content_sel <- function(prefix) {
  paste0(prefix, xpath_content_selector)
}


between <- function(x, left, right) {
  x >= left & x <= right
}


rbind_match_columns <- function(list_df) {
  col <- unique(unlist(lapply(list_df, colnames)))
  x <- Filter(function(x) nrow(x) > 0, list_df)
  x <- lapply(
    x,
    function(x, col) {
      x[, setdiff(col, colnames(x))] <- NA
      x
    },
    col = col
  )
  do.call(rbind, x)
}

set_row_span <- function(row_details) {
  row_details$first[!row_details$first & !row_details$row_merge] <- TRUE
  row_details$row_merge <- NULL

  row_details <- split(row_details, row_details$cell_id)

  row_details <- mapply(
    function(dat) {
      rowspan_values_at_breaks <- rle(cumsum(dat$first))$lengths
      rowspan_pos_at_breaks <- which(dat$first)
      dat$row_span <- 0L
      dat$row_span[rowspan_pos_at_breaks] <- rowspan_values_at_breaks
      dat
    },
    row_details,
    SIMPLIFY = FALSE
  )
  row_details <- rbind_match_columns(row_details)
  row_details$first <- NULL
  row_details
}


#' @importFrom grDevices col2rgb rgb
is.color = function(x) {
  # http://stackoverflow.com/a/13290832/3315962
  out = vapply(
    x,
    function(x) {
      tryCatch(is.matrix(col2rgb(x)), error = function(e) FALSE)
    },
    logical(1)
  )

  nout <- names(out)
  if (!is.null(nout) && anyNA(nout)) {
    out[is.na(nout)] = FALSE
  }

  out
}

correct_id <- function(doc, int_id) {
  all_uid <- xml_find_all(doc, "//*[@id]")
  for (z in seq_along(all_uid)) {
    if (!grepl("[^0-9]", xml_attr(all_uid[[z]], "id"))) {
      xml_attr(all_uid[[z]], "id") <- int_id
      int_id <- int_id + 1
    }
  }
  int_id
}


check_bookmark_id <- function(bkm) {
  if (!is.null(bkm)) {
    invalid_bkm <- is.character(bkm) &&
      length(bkm) == 1 &&
      nchar(bkm) > 0 &&
      grepl("[^:[:alnum:]_-]+", bkm)
    if (invalid_bkm) {
      stop(
        "bkm [",
        bkm,
        "] should only contain alphanumeric characters, ':', '-' and '_'.",
        call. = FALSE
      )
    }
  }
  bkm
}

is_windows <- function() {
  "windows" %in% .Platform$OS.type
}

is_doc_open <- function(file) {
  # The function checks if the `file` is open (a.k.a. is being edited).
  # This function is valid on Windows operating system only.
  suppressWarnings(file.exists(file) && !file.rename(from = file, to = file))
}


# Extract trailing numeric index in .xml filename
#
# Useful to for slideMaster and slideLayout .xml files.
#
# Examples:
#   files <- c("slideLayout1.xml", "slideLayout2.xml", "slideLayout10.xml")
#   get_file_index(files)
#
get_file_index <- function(file) {
  x <- sub(
    pattern = ".+?(\\d+).xml$",
    replacement = "\\1",
    x = basename(file),
    ignore.case = TRUE
  )
  as.numeric(x)
}


# Sort xml filenames by trailing numeric index
#
# Useful to for slideMaster and slideLayout xml files.
#
# Examples:
#   files <- c("slideLayout1.xml", "slideLayout2.xml", "slideLayout12.xml")
#   sort_vec_by_index(files)  # => order corresponding to trailing index
#   sort(files)           # => incorrect lexicographical ordering
#
sort_vec_by_index <- function(x) {
  indexes <- get_file_index(x)
  x[order(indexes)]
}


# Sort dataframe column by trailing index
#
# df: A dataframe
# ...: columsn to sort by, comma separated
#
# Examples:
# df <- data.frame(
#   a = paste0("file_", rep(3:1, each = 2), ".xml"),
#   b = paste0("file_", rep(3:1, 2), ".xml")
# )
# sort_dataframe_by_index(df, "a", "b")
# sort_dataframe_by_index(df, "b", "a")
#
sort_dataframe_by_index <- function(df, ...) {
  sort_columns <- c(...)
  l <- lapply(sort_columns, function(.col) {
    get_file_index(df[[.col]])
  })
  df[do.call(order, l), , drop = FALSE]
}


# rename dataframe columns
#
# Examples:
#   df_rename(mtcars, c("mpg", "cyl"), c("A", "B"))
#
df_rename <- function(df, old, new) {
  .nms <- names(df)
  .nms[match(old, .nms)] <- new
  stats::setNames(df, .nms)
}


# replacement for stopifnot() with nicer user feedback
stop_if_not_class <- function(x, class, arg = NULL) {
  check <- inherits(x, what = class)
  if (!check) {
    msg_arg <- ifelse(
      is.null(arg),
      "Incorrect input.",
      "Incorrect input for {.arg {arg}}"
    )
    cli::cli_abort(
      c(
        msg_arg,
        "x" = "Expected {.cls {class}} but got {.cls {class(x)[1]}}"
      ),
      call = NULL
    )
  }
}


stop_if_not_rpptx <- function(x, arg = NULL) {
  stop_if_not_class(x, "rpptx", arg)
}


stop_if_not_integerish <- function(x, arg = NULL) {
  check <- is_integerish(x)
  if (!check) {
    msg_arg <- ifelse(
      is.null(arg),
      "Incorrect input.",
      "Incorrect input for {.arg {arg}}"
    )
    cli::cli_abort(
      c(
        msg_arg,
        "x" = "Expected integerish values but got {.cls {class(x)[1]}}"
      ),
      call = NULL
    )
  }
}


#' Ensure valid slide indexes
#'
#' @param x An `rpptx` object.
#' @param idx Slide indexes.
#' @param arg Name of argument to use in error message (optional).
#' @param call Environment to display in error message. Defaults to caller env.
#'   Set `NULL` to suppress (see [cli::cli_abort]).
#' @keywords internal
stop_if_not_in_slide_range <- function(
  x,
  idx,
  arg = NULL,
  call = parent.frame()
) {
  stop_if_not_rpptx(x)
  stop_if_not_integerish(idx)

  n_slides <- length(x)
  idx_available <- seq_len(n_slides)
  idx_outside <- setdiff(idx, idx_available)
  n_outside <- length(idx_outside)

  if (n_outside == 0) {
    return(invisible(NULL))
  }
  argname <- ifelse(is.null(arg), "", "of {.arg {arg}} ")
  part_1 <- paste0(
    "{n_outside} index{?es} ",
    argname,
    "outside slide range: {.val {idx_outside}}"
  )
  part_2 <- ifelse(
    n_slides == 0,
    "Presentation has no slides!",
    "Slide indexes must be in the range [{min(idx_available)}..{max(idx_available)}]"
  )
  cli::cli_abort(c(part_1, "x" = part_2), call = call)
}


check_unit <- function(unit, choices, several.ok = FALSE) {
  if (!several.ok && length(unit) != 1) {
    cli::cli_abort(
      c(
        "{.arg unit} is not length 1.",
        "x" = "{.arg unit} must be {.emph a string}."
      )
    )
  }
  if (!unit %in% choices) {
    cli::cli_abort(
      c(
        "{.arg unit} should be one of {.or {choices}}.",
        "x" = "{.arg unit} was {.emph {unit}\"}."
      )
    )
  }

  unit
}


#' Update a vector or list of named defaults values
#'
#' Helper to update named default values. Takes care of most common cases and
#' returns sensible error messages.
#'
#' @param x A (named) vector or list to update the default values with.
#' @param default A *named* vector or list with default values.
#' @param argname Name of the arg. Used in error message, defaults to `"x"`.
#' @param default_if_null Returns defaults values if `x` is `NULL`? (default is `TRUE`)
#' @param partial Enable partial name matching= (default `TRUE`).
#' @param as_list Return a list (default `TRUE`) or a named vector?
#' @return List of update default values.
#' @noRd
#' @examples
#' defaults <- c(aa = 1, bb = 2, cc = 3)
#' update_named_defaults(c(a = 99, b = 0), defaults) # partial name match
#' update_named_defaults(c(a = 99, b = 0), defaults, as_list = FALSE) # return as vector
#' update_named_defaults(3:1, defaults) # match by position
#' update_named_defaults(2, defaults)
#' update_named_defaults(NULL, defaults, default_if_null = TRUE)
#'
update_named_defaults <- function(
  x,
  default,
  argname = "x",
  default_if_null = TRUE,
  partial = TRUE,
  as_list = TRUE
) {
  if (default_if_null && is.null(x)) {
    res <- as.list(default)
    if (!as_list) {
      res <- unlist(res)
    }
    return(res)
  }

  x <- as.list(x)
  default <- as.list(default)

  if (!is_named(default)) {
    cli::cli_abort(
      c(
        "Some default vector elements have no names",
        "x" = "{.arg default} must be a named vector"
      ),
      call = NULL
    )
  }
  x <- as.list(x)
  len_x <- length(x)
  len_default <- length(default)
  names_default <- names(default)
  if (len_x > len_default) {
    cli::cli_abort(
      c(
        "Length of {.arg {argname}} ({.val {len_x}}) exceeds length of {.arg default} ({.val {len_default}})",
        "x" = "Length of {.arg x} must be smaller or equal to the length of {.arg default}"
      ),
      call = NULL
    )
  }

  # unnamed case => convert to named case
  if (!is_named(x)) {
    if (len_x == 1) {
      x <- rep(x, len_default)
      len_x <- length(x)
    }
    if (len_x != len_default) {
      cli::cli_abort(
        c(
          "{.arg {argname}} has incorrect length ({len_x})",
          "x" = "If {.arg {argname}} has no names, it must be length 1 or the length of {.arg default} ({len_default})"
        ),
        call = NULL
      )
    }
    names(x) <- names_default
  }

  # named case
  if (partial) {
    # => partial name matching
    matched <- pmatch(names(x), names_default, duplicates.ok = TRUE)
  } else {
    # exact name matching
    matched <- match(names(x), names_default, nomatch = NA)
  }
  nms_new <- ifelse(is.na(matched), NA, names_default[matched])
  i_na <- is.na(nms_new)

  if (any(i_na)) {
    msg_partial <- ifelse(
      partial,
      "Partial name matching is supported",
      "Partial name matching is not enabled"
    )
    cli::cli_abort(
      c(
        "Found {sum(i_na)} unknown name{?s} in {.arg {argname}}: {.val {names(x)[i_na]}}",
        "x" = "{.arg {argname}} understands {.val {names_default}}",
        "i" = cli::col_silver(msg_partial)
      ),
      call = NULL
    )
  }
  # duplicate position
  ii_dupes <- duplicated(nms_new)
  if (any(ii_dupes)) {
    cli::cli_abort(
      c(
        "Duplicate entries in {.arg location}: {.val {unique(nms_new[ii_dupes])}}",
        "x" = "Each name in {.arg location} must be unique",
        "i" = cli::col_silver("Partial name matching is supported")
      ),
      call = NULL
    )
  }
  x <- setNames(x, nms_new)
  res <- utils::modifyList(x = default, val = as.list(x), keep.null = TRUE)
  if (!as_list) {
    res <- unlist(res)
  }
  res
}


# htmlEscapeCopy ----

htmlEscapeCopy <- local({
  .htmlSpecials <- list(
    `&` = '&amp;',
    `<` = '&lt;',
    `>` = '&gt;'
  )
  .htmlSpecialsPattern <- paste(names(.htmlSpecials), collapse = '|')
  .htmlSpecialsAttrib <- c(
    .htmlSpecials,
    `'` = '&#39;',
    `"` = '&quot;',
    `\r` = '&#13;',
    `\n` = '&#10;'
  )
  .htmlSpecialsPatternAttrib <- paste(
    names(.htmlSpecialsAttrib),
    collapse = '|'
  )
  function(text, attribute = FALSE) {
    pattern <- if (attribute) {
      .htmlSpecialsPatternAttrib
    } else {
      .htmlSpecialsPattern
    }
    text <- enc2utf8(as.character(text))
    # Short circuit in the common case that there's nothing to escape
    if (!any(grepl(pattern, text, useBytes = TRUE))) {
      return(text)
    }
    specials <- if (attribute) {
      .htmlSpecialsAttrib
    } else {
      .htmlSpecials
    }
    for (chr in names(specials)) {
      text <- gsub(chr, specials[[chr]], text, fixed = TRUE, useBytes = TRUE)
    }
    Encoding(text) <- "UTF-8"
    return(text)
  }
})


# metric units -----------------------------------------------
#
cm_to_inches <- function(x) {
  x / 2.54
}

mm_to_inches <- function(x) {
  x / 25.4
}

convin <- function(unit, x) {
  unit <- match.arg(unit, choices = c("in", "cm", "mm"), several.ok = FALSE)
  if (!identical("in", unit)) {
    x <- do.call(paste0(unit, "_to_inches"), list(x = x))
  }
  x
}


# from rlang pkg  ------------------------------------------

is_named <- function(x) {
  nms <- names(x)
  if (is.null(nms)) {
    return(FALSE)
  }
  if (any(detect_void_name(nms))) {
    return(FALSE)
  }
  TRUE
}


detect_void_name <- function(x) {
  x == "" | is.na(x)
}


# is_integerish(1)
# is_integerish(1.0)
# is_integerish(c(1.0, 2.0))
is_integerish <- function(x) {
  ii <- all(is.numeric(x) | is.integer(x))
  jj <- all(x == as.integer(x))
  ii && jj
}


# coalesce for R
`%||%` <- function(l, r) {
  if (is.null(l)) r else l
}


# file ops  ------------------------------------------

#' Opens a file locally
#'
#' Opening a file locally requires a compatible application to be installed
#' (e.g., MS Office or LibreOffice for .pptx or .docx files).
#'
#' @details
#' *NB:* Function is a small wrapper around [utils::browseURL()] to have a more
#' suitable function name.
#'
#' @param path File path.
#' @export
#' @examples
#' x <- read_pptx()
#' x <- add_slide(x, "Title Slide", ctrTitle = "My Title")
#' file <- print(x, tempfile(fileext = ".pptx"))
#' \dontrun{
#' open_file(file)}
open_file <- function(path) {
  path <- normalizePath(path, winslash = "/", mustWork = TRUE)
  utils::browseURL(path)
}

attr_chunk <- function( x ){
  if( !is.null(x) && length( x ) > 0){
    attribs <- paste0(names(x), "=", shQuote(x, type = "cmd"), collapse = " " )
    attribs <- paste0(" ", attribs)
  } else attribs <- ""
  attribs
}

# content_type -----
content_type <- R6Class(
  "content_type",
  public = list(

    initialize = function( package_dir ) {

      private$filename <- file.path(package_dir, "[Content_Types].xml")

      doc <- read_xml(x = private$filename )
      ns <- xml_ns(doc)

      node_template <- xml_find_first(doc, "d1:Override[@ContentType='application/vnd.openxmlformats-officedocument.presentationml.template.main+xml']")
      if (!inherits(node_template, "xml_missing")) {
        xml_attr(node_template, 'ContentType') <- "application/vnd.openxmlformats-officedocument.presentationml.presentation.main+xml"
      }
      node_template <- xml_find_first(doc, "d1:Override[@ContentType='application/vnd.openxmlformats-officedocument.wordprocessingml.template.main+xml']")
      if (!inherits(node_template, "xml_missing")) {
        xml_attr(node_template, 'ContentType') <- "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml"
      }

      extension <- xml_find_all(doc, "//*[contains(local-name(), 'Default')]/@Extension", ns = ns)
      extension <- xml_text( extension )
      content_type <- xml_find_all(doc, "//*[contains(local-name(), 'Default')]/@ContentType", ns = ns)
      content_type <- xml_text( content_type )
      names(content_type) <- extension
      private$default <- content_type

      partname <- xml_find_all(doc, "//*[contains(local-name(), 'Override')]/@PartName", ns = ns)
      partname <- xml_text(partname)
      content_type <- xml_find_all(doc, "//*[contains(local-name(), 'Override')]/@ContentType", ns = ns)
      content_type <- xml_text(content_type)
      names(content_type) <- partname
      private$override <- content_type

    },

    add_slide = function(partname){
      partname <- basename(partname)
      partname <- file.path("/ppt", "slides", partname )
      content_type <- setNames("application/vnd.openxmlformats-officedocument.presentationml.slide+xml", partname )
      override <- c( private$override, content_type )
      private$override <- override
      self
    },
    add_notesSlide = function(partname){
      partname <- basename(partname)
      partname <- file.path("/ppt", "notesSlides", partname )
      content_type <- setNames("application/vnd.openxmlformats-officedocument.presentationml.notesSlide+xml", partname )
      override <- c( private$override, content_type )
      private$override <- override
      self
    },
    add_override = function(value){
      if (!names(value) %in% names(private$override)){
        override <- c( private$override, value )
        private$override <- override
      }
      self
    },
    remove_slide = function(partname){
      id <- which( basename(names(private$override)) %in% basename(partname) )
      private$override <- private$override[-id]
      self
    },
    add_ext = function( extension, type ){
      if( !type %in% private$default && !extension %in% names(private$default) ){
        content_type <- setNames(type, extension )
        default <- c( private$default, content_type )
        private$default <- default
      }
      self
    },

    save = function() {
      self$add_ext(extension = "jpeg", type = "image/jpeg")
      self$add_ext(extension = "gif", type = "image/gif")
      self$add_ext(extension = "png", type = "image/png")
      self$add_ext(extension = "svg", type = "image/svg+xml")
      self$add_ext(extension = "bmp", type = "image/bmp")
      self$add_ext(extension = "emf", type = "image/x-emf")
      self$add_ext(extension = "wmf", type = "image/x-wmf")
      self$add_ext(extension = "tiff", type = "image/tiff")
      self$add_ext(extension = "jpg", type = "application/octet-stream")
      attribs <- attr_chunk(c(xmlns = "http://schemas.openxmlformats.org/package/2006/content-types"))
      out <- paste0("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
                    "\n<Types", attribs, ">")
      if(length(private$default) > 0 ){
        default <- sprintf("<Default Extension=\"%s\" ContentType=\"%s\"/>", names(private$default), private$default )
        default <- paste0(default, collapse = "")
        out <- paste0(out, default )
      }
      if(length(private$override) > 0 ){
        override <- sprintf("<Override PartName=\"%s\" ContentType=\"%s\"/>", names(private$override), private$override )
        override <- paste0(override, collapse = "")
        out <- paste0(out, override )
      }
      out <- paste0(out, "</Types>" )
      cat(out, file = private$filename)
      self

    }
  ),
  private = list(
    filename = NULL,
    default = NULL,
    override = NULL
  )
)

#' Detect and handle duplicate placeholder labels
#'
#' PowerPoint does not enforce unique placeholder labels in a layout.
#' Selecting a placeholder via its label using [ph_location_label] will throw
#' an error, if the label is not unique. [layout_dedupe_ph_labels] helps to detect,
#' rename, or delete duplicate placholder labels.
#'
#' @param x An `rpptx` object.
#' @param action Action to perform on duplicate placeholder labels. One of:
#' * `detect` (default) = show info on dupes only, make no changes
#' * `rename` = create unique labels. Labels are renamed by appending a sequential number
#'    separated by dot to duplicate labels. For example, `c("title", "title")` becomes `c("title.1", "title.2")`.
#' * `delete` = only keep one of the placeholders with a duplicate label
#' @param print_info Print action information (e.g. renamed placeholders) to console?
#'  Default is `FALSE`. Always `TRUE` for action `detect`.
#' @return A `rpptx` object (with modified placeholder labels).
#' @export
#' @examples
#' x <- read_pptx()
#' layout_dedupe_ph_labels(x)
#'
#' file <- system.file("doc_examples", "ph_dupes.pptx", package = "officer")
#' x <- read_pptx(file)
#' layout_dedupe_ph_labels(x)
#' layout_dedupe_ph_labels(x, "rename", print_info = TRUE)
#'
layout_dedupe_ph_labels <- function(x, action = "detect", print_info = FALSE) {
  if (!inherits(x, "rpptx")) {
    stop("'x' must be an 'rpptx' object", call. = FALSE)
  }
  action <- match.arg(action, c("detect", "rename", "delete"))
  layout_names <- x$slideLayouts$get_metadata()$filename
  xfrm_list <- lapply(layout_names, .dedupe_phs_in_layout, x = x, action = action)
  x <- reload_slidelayouts(x) # reinit slideLayouts to get processed ph labels [e.g. when calling x$slideLayouts$get_xfrm_data()]
  if (print_info | action == "detect") {
    .print_dedupe_info(x = x, xfrm_list = xfrm_list, action = action)
  }
  invisible(x)
}


# handle placeholder labels in a single layout
#
# layout_file: layout filename (e.g. "slideLayout1.xml").
# x: An `rpptx` object
#
# returns: Dataframe with placeholder info. Only needed for .print_dedupe_info()
.dedupe_phs_in_layout <- function(layout_file, x, action = "rename") {
  ph_label <- NULL
  if (!grepl("\\.xml$", layout_file, ignore.case = TRUE)) {
    stop("'layout_file' must be an .xml file", call. = FALSE)
  }
  action <- match.arg(action, c("detect", "rename", "delete"))
  layout <- x$slideLayouts$collection_get(layout_file)
  xfrm <- layout$xfrm()
  xfrm <- subset(xfrm, duplicated(ph_label) | duplicated(ph_label, fromLast = TRUE))
  if (nrow(xfrm) == 0) {
    return()
  }
  xfrm <- transform(xfrm, ph_label_new = make_strings_unique(ph_label), delete_flag = duplicated(ph_label)) # prepare once for all action types
  if (action == "detect") {
    return(xfrm) # no further action required
  } else if (action == "rename") {
    xfrm$delete_flag <- FALSE
  } else if (action == "delete") {
    xfrm$ph_label_new <- xfrm$ph_label
  }

  # rename label or delete ph shape
  layout_xml <- layout$get()
  for (i in 1L:nrow(xfrm)) {
    shape <- xml2::xml_find_first(layout_xml, sprintf("p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]", xfrm$id[i]))
    if (xfrm$delete_flag[i]) {
      xml2::xml_remove(shape)
    } else {
      nodes <- xml2::xml_find_first(shape, ".//p:cNvPr")
      xml2::xml_set_attr(nodes, "name", xfrm$ph_label_new[i])
    }
  }
  layout$save() # persist changes in slideout xml file
  xfrm
}


# reload slideLayouts (if layout XML in package_dir has changed)
reload_slidelayouts <- function(x) {
  x$slideLayouts$initialize(x$package_dir,
    master_metadata = x$masterLayouts$get_metadata(),
    master_xfrm = x$masterLayouts$xfrm()
  )
  x
}


# Create unique string by appending a sepatator and a number
# make_strings_unique(c("A", "B", "B", "C", "A"))
make_strings_unique <- function(x, sep = ".") {
  ii <- stats::ave(x, x, FUN = seq_along)
  paste0(x, sep, ii)
}


# helper mostly for testing
has_ph_dupes <- function(x) {
  if (!inherits(x, "rpptx")) {
    stop("'x' must be an 'rpptx' object", call. = FALSE)
  }
  xfrm <- x$slideLayouts$get_xfrm_data()
  dupes <- stats::aggregate(ph_label ~ master_name + name, data = xfrm, FUN = function(x) sum(duplicated(x)) > 0)
  any(dupes$ph_label)
}


# print info on what was done (if print_info = TRUE)
.print_dedupe_info <- function(x, xfrm_list, action) {
  .df_1 <- do.call(rbind, xfrm_list)
  if (is.null(.df_1)) {
    cat("No duplicate placeholder labels detected.")
    return(invisible(NULL))
  }
  .df_2 <- x$slideLayouts$get_xfrm_data()
  .df_2 <- unique(.df_2[, c("master_file", "master_name"), drop = FALSE])
  df <- merge(.df_1, .df_2, sort = FALSE)
  rownames(df) <- NULL
  df <- df[, c("master_name", "name", "ph_label", "ph_label_new", "delete_flag"), drop = FALSE]
  colnames(df)[2] <- "layout_name"
  if (action == "detect") {
    cat("Placeholders with duplicate labels:\n")
    cat(cli::col_grey("* 'ph_label_new' = new placeholder label for action = 'rename'\n"))
    cat(cli::col_grey("* 'delete_flag' = deleted placeholders for action = 'delete'\n"))
  } else if (action == "rename") {
    df$delete_flag <- NULL
    cat("Renamed duplicate placeholder labels:\n")
    cat(cli::col_grey("* 'ph_label_new' = new placeholder label\n"))
  } else if (action == "delete") {
    df <- df[df$delete_flag, , drop = FALSE]
    df$ph_label_new <- NULL
    cat("Removed placeholders with duplicate labels:\n")
  }
  print(df)
}

library(officer)

doc <- read_docx()
doc <- body_add_par(doc, "Placeholder one")
doc <- body_add_par(doc, "Placeholder two")

# Show text chunk at cursor
docx_show_chunk(doc) # Output is 'Placeholder two'

# Simple search-and-replace at current cursor, with regex turned off
doc <- body_replace_all_text(
  doc,
  old_value = "Placeholder",
  new_value = "new",
  only_at_cursor = TRUE,
  fixed = TRUE
)
docx_show_chunk(doc) # Output is 'new two'

# Do the same, but in the entire document and ignoring case
doc <- body_replace_all_text(
  doc,
  old_value = "placeholder",
  new_value = "new",
  only_at_cursor = FALSE,
  ignore.case = TRUE
)
doc <- cursor_backward(doc)
docx_show_chunk(doc) # Output is 'new one'

# Use regex : replace all words starting with "n" with the word "example"
doc <- body_replace_all_text(doc, "\\bn.*?\\b", "example")
docx_show_chunk(doc) # Output is 'example one'

#' @title PowerPoint table to matrix
#' @description Convert the data in an a 'PowerPoint' table
#' to a matrix or all data to a list of matrices.
#' @param x The rpptx object to convert (as created by
#'   \code{officer::read_pptx()})
#' @param ... Ignored
#' @param slide_id The slide number to load from (NA indicates first slide with
#'   a table, NULL indicates all slides and all tables)
#' @param id The table ID to load from (ignored it \code{is.null(slide_id)}, NA
#'   indicates to load the first table from the \code{slide_id})
#' @param span How should col_span/row_span values be handled?  \code{NA} means
#'   to leave the value as \code{NA}, and \code{"fill"} means to fill matrix
#'   cells with the value.
#' @return A matrix with the data, or if \code{slide_id=NULL}, a list of
#'   matrices
#' @examples
#' library(officer)
#' pptx_file <- system.file(package="officer", "doc_examples", "example.pptx")
#' z <- read_pptx(pptx_file)
#' as.matrix(z, slide_id = NULL)
#' @export
as.matrix.rpptx <- function(x, ..., slide_id=NA_integer_, id=NA_character_, span=c(NA_character_, "fill")) {
  span <- match.arg(span)
  d_summary_raw <- officer::pptx_summary(x)
  # filter to just table data
  d_summary_prep_table <- d_summary_raw[d_summary_raw$content_type %in% "table cell", ]
  stopifnot("No tables to extract"=nrow(d_summary_prep_table) > 0)
  d_summary_prep <-
    d_summary_prep_table[
      d_summary_prep_table$row_span > 0 & d_summary_prep_table$col_span > 0,
    ]
  stopifnot("All cells had col_span=0 or row_span=0"=nrow(d_summary_prep) > 0)
  if (is.null(slide_id)) {
    data_extract <- d_summary_prep
  } else {
    if (is.na(slide_id)) {
      slide_id <- min(d_summary_prep$slide_id)
      id <- min(d_summary_prep$id[d_summary_prep$slide_id == slide_id])
    }
    data_extract <- d_summary_prep[d_summary_prep$slide_id == slide_id & d_summary_prep$id == id, ]
    if (nrow(data_extract) == 0) {
      stop("No data in slide_id=%s, id=%s", as.character(slide_id), as.character(id))
    }
  }
  ret <- list()
  for (current_slide_id in unique(data_extract$slide_id)) {
    ret[[as.character(current_slide_id)]] <-
      as_matrix_rpptx_single_slide(data_extract[data_extract$slide_id == current_slide_id, ], span=span)
  }
  if (length(slide_id) == 1 && length(id) == 1) {
    # When the user is specific, return a matrix (otherwise, return the list of
    # lists of matrices)
    ret <- ret[[1]][[1]]
  }
  ret
}

as_matrix_rpptx_single_slide <- function(x, span=c(NA_character_, "fill")) {
  span <- match.arg(span)
  stopifnot(nrow(x) > 0)
  stopifnot(length(unique(x$slide_id)) == 1)
  ret <- list()
  for (current_id in unique(x$id)) {
    ret[[as.character(current_id)]] <-
      as_matrix_rpptx_single_table(x[x$id == current_id, ], span=span)
  }
  ret
}

as_matrix_rpptx_single_table <- function(x, span=c(NA_character_, "fill")) {
  span <- match.arg(span)
  stopifnot(nrow(x) > 0)
  stopifnot(length(unique(x$slide_id)) == 1)
  stopifnot(length(unique(x$id)) == 1)
  ret <- matrix(NA_character_, nrow=max(x$row_id), ncol=max(x$cell_id))
  for (current_data_idx in seq_len(nrow(x))) {
    if (span %in% "fill") {
      # fill the row/col span
      all_rows <- seq(x$row_id[current_data_idx], x$row_id[current_data_idx] + x$row_span[current_data_idx] - 1)
      all_cols <- seq(x$cell_id[current_data_idx], x$cell_id[current_data_idx] + x$col_span[current_data_idx] - 1)
    } else {
      # just the current cell
      all_rows <- x$row_id[current_data_idx]
      all_cols <- x$cell_id[current_data_idx]
    }
    for (current_col in all_cols) {
      # A loop or some other multiple assignment is needed (multiple indexing
      # does not work on rows and columns simultaneously for matrices)
      ret[all_rows, current_col] <- x$text[current_data_idx]
    }
  }
  ret
}

old_cli_width <- options(cli.width = 200)

if (require("withr")) {
  withr::defer(
    {
      options(old_cli_width)
      if (file.exists("Rplots.pdf")) file.remove("Rplots.pdf")
    },
    teardown_env()
  )
}

unfold_row_pml <- function(node, row_id, preserve = FALSE) {
  children_ <- xml_children(node)
  cell_nodes <- children_[sapply(children_, function(x) xml_name(x) == "tc")]

  if (preserve) {
    txt <- sapply(cell_nodes, function(x) {
      paras <- xml2::xml_text(xml2::xml_find_all(x, ".//a:r"))
      paste0(paras, collapse = "\n")
    })
  } else {
    txt <- sapply(cell_nodes, xml2::xml_text)
  }

  col_span <- sapply(cell_nodes, function(x) {
    as.integer(xml_attr(x, "gridSpan"))
  })
  h_merge <- sapply(cell_nodes, function(x) {
    as.integer(xml_attr(x, "hMerge"))
  }) %in%
    c(1)
  col_span[is.na(col_span)] <- 1L
  col_span[h_merge] <- 0L

  row_span <- lapply(cell_nodes, function(x) {
    row_span <- as.integer(xml_attr(x, "rowSpan"))
    v_merged <- as.integer(xml_attr(x, "vMerge"))
    data.frame(
      v_merged = v_merged,
      row_span = row_span,
      stringsAsFactors = FALSE
    )
  })
  row_span <- rbind_match_columns(row_span)
  row_span$row_merge <- !is.na(row_span$v_merged) | !is.na(row_span$row_span)
  row_span$first <- !is.na(row_span$row_span)
  row_span$row_span[!is.na(row_span$v_merged)] <- 0L
  row_span$row_span[!row_span$row_merge] <- 1L
  row_span <- row_span[, c("row_merge", "first", "row_span")]

  out <- data.frame(
    row_id = row_id,
    cell_id = seq_along(cell_nodes),
    text = txt,
    col_span = col_span,
    row_merge = row_span$row_merge,
    first = row_span$first,
    stringsAsFactors = FALSE
  )
  out
}

globalVariables(c(".", "src", "width", "height"))

pptxtable_as_tibble <- function(node, preserve = FALSE) {
  tbl <- xml_child(node, "/a:graphic/a:graphicData/a:tbl")
  xpath_ <- paste0(xml_path(tbl), "/a:tr")
  rows <- xml_find_all(node, xpath_)
  if (length(rows) < 1) {
    return(NULL)
  }
  row_details <- mapply(
    unfold_row_pml,
    rows,
    seq_along(rows),
    preserve = preserve,
    SIMPLIFY = FALSE
  )
  row_details <- rbind_match_columns(row_details)
  row_details <- set_row_span(row_details)
  row_details$text[
    row_details$col_span < 1 | row_details$row_span < 1
  ] <- NA_character_
  row_details
}


pptx_par_as_tibble <- function(node) {
  xpath_ <- paste0(
    xml_path(node),
    c("/p:txBody/a:p", "//*/p:txBody/a:p"), # standard and groupedshapes
    collapse = "|"
  )

  p_nodes <- xml_find_all(node, xpath_)
  data.frame(text = xml_text(p_nodes), stringsAsFactors = FALSE)
}


desc_pptx_img <- function(node, img_src) {
  blip <- xml_child(node, "/p:blipFill/a:blip")
  img_id <- xml_attr(blip, "embed")

  file_ <- img_src[img_id]
  stopifnot(is.character(file_), length(file_) == 1)
  data.frame(
    media_file = file.path("ppt/media/", basename(file_)),
    stringsAsFactors = FALSE
  )
}

#' @export
#' @title Extract media from a document object
#' @description Extract files from a `rpptx` object.
#' @param x an rpptx object
#' @param path media path, should be a relative path
#' @param target target file
#' @examples
#' example_pptx <- system.file(package = "officer",
#'   "doc_examples/example.pptx")
#' doc <- read_pptx(example_pptx)
#' content <- pptx_summary(doc)
#' image_row <- content[content$content_type %in% "image", ]
#' media_file <- image_row$media_file
#' png_file <- tempfile(fileext = ".png")
#' media_extract(doc, path = media_file, target = png_file)
media_extract <- function(x, path, target) {
  media <- file.path(x$package_dir, path)
  stopifnot(file.exists(media))
  file.copy(from = media, to = target)
}

#' @title PowerPoint content in a data.frame
#' @description Read content of a PowerPoint document and
#' return a dataset representing the document.
#' @param x an rpptx object
#' @inheritParams docx_summary
#' @examples
#' example_pptx <- system.file(package = "officer",
#'   "doc_examples/example.pptx")
#' doc <- read_pptx(example_pptx)
#' pptx_summary(doc)
#' pptx_summary(example_pptx)
#' @export
pptx_summary <- function(x, preserve = FALSE) {
  if (is.character(x)) {
    x = read_pptx(x)
  }
  list_content <- list()
  for (i in seq_len(length(x))) {
    slide <- x$slide$get_slide(i)
    str = as_xpath_content_sel("p:cSld/p:spTree/")
    nodes <- xml_find_all(slide$get(), str)

    if (length(nodes) < 1) {
      # when slide is empty
      list_content[[length(list_content) + 1]] <- data.frame(
        id = character(0),
        content_type = character(0),
        slide_id = integer(0),
        stringsAsFactors = FALSE
      )
      next
    }

    content <- mapply(
      function(node, slide_id) {
        is_table <- !inherits(
          xml_child(node, "/a:graphic/a:graphicData/a:tbl"),
          "xml_missing"
        )
        is_par <- !inherits(xml_child(node, "/p:txBody/a:p"), "xml_missing")
        is_img <- xml_name(node) == "pic"
        id <- xml_attr(xml_child(node, "/p:cNvPr"), "id")

        if (is_table) {
          ppt_tab <- pptxtable_as_tibble(node, preserve = preserve)
          ppt_tab$id <- id
          ppt_tab$content_type <- "table cell"
          ppt_tab$slide_id <- slide_id
          ppt_tab
        } else if (is_par) {
          ppt_tab <- pptx_par_as_tibble(node)
          ppt_tab$id <- id
          ppt_tab$content_type <- "paragraph"
          ppt_tab$slide_id <- slide_id
          ppt_tab
        } else if (is_img) {
          rel <- slide$relationship()
          images_ <- rel$get_images_path()
          img_id <- names(images_)
          images_ <- normalizePath(file.path(
            dirname(slide$file_name()),
            images_
          ))
          names(images_) <- img_id
          ppt_tab <- desc_pptx_img(node, images_)
          ppt_tab$id <- id
          ppt_tab$content_type <- "image"
          ppt_tab$slide_id <- slide_id
          ppt_tab
        } else {
          data.frame(
            id = id,
            content_type = "unknown",
            slide_id = slide_id,
            stringsAsFactors = FALSE
          )
        }
      },
      nodes,
      slide_id = i,
      SIMPLIFY = FALSE
    )

    list_content[[length(list_content) + 1]] <- rbind_match_columns(content)
  }

  rbind_match_columns(list_content)
}

# to_rtf ----
#' @export
#' @title Convert officer objects to RTF
#' @description Convert an object made with package officer
#' to RTF.
#' @param x object to convert to RTF. Supported
#' objects are:
#' - [ftext()]
#' - [external_img()]
#' - [run_autonum()]
#' - [run_columnbreak()]
#' - [run_linebreak()]
#' - [run_word_field()]
#' - [run_reference()]
#' - [run_pagebreak()]
#' - [hyperlink_ftext()]
#' - [block_list()]
#' - [fpar()]
#' @param ... Arguments to be passed to methods
#' @family functions for officer extensions
#' @return a string containing the RTF code
#' @keywords internal
to_rtf <- function(x, ...) {
  UseMethod("to_rtf")
}

## to_rtf blocks ----
#' @export
to_rtf.fpar <- function(x, ...) {
  chks <- fortify_fpar(x)
  rtf_chunks <- lapply(chks, to_rtf)
  rtf_chunks$collapse <- ""
  rtf_chunks <- do.call(paste0, rtf_chunks)
  style_index <- attr(x, "rtf_style_index")
  if (!is.null(style_index)) {
    # Canonical RTF emission for a styled paragraph (RTF spec, additive
    # model): \pard\plain \sN <style properties duplicated> text \par.
    # The duplicated properties make readers that ignore styles still
    # render the formatting, and prevent Word from recording spurious
    # "overrides" against the named style.
    style_props <- attr(x, "rtf_style_props")
    if (is.null(style_props)) style_props <- ""
    return(paste0(
      sprintf("{\\pard\\plain\\s%.0f", style_index),
      style_props,
      " ",
      rtf_chunks,
      "\\par}"
    ))
  }
  paste0("{", ppr_rtf(x$fp_p), rtf_chunks, "\\par}")
}

#' @export
to_rtf.default <- function(x, ...) {
  if (is.null(x)) {
    return("")
  }
  as.character(x)
}

#' @export
to_rtf.block_list <- function(x, ...) {
  rtf_blocks <- lapply(x, to_rtf)
  rtf_blocks$collapse <- ""
  do.call(paste0, rtf_blocks)
}


## to_rtf chunks ----

#' @export
to_rtf.ftext <- function(x, ...) {
  rtf_fp_strings <- ""
  value <- str_encode_to_rtf(x$value)
  if (!is.null(x$pr)) {
    rtf_fp_strings <- rpr_rtf(x$pr)
    if (!is.na(x$pr$shading.color)) {
      value <- paste0(
        "%ftshading:",
        x$pr$shading.color,
        "%",
        value,
        "\\highlight0"
      )
    }
  }

  paste0(rtf_fp_strings, value)
}

#' @export
to_rtf.run_tab <- function(x, ...) {
  "\\tab "
}


#' @export
to_rtf.external_img <- function(x, ...) {
  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  if (!grepl("\\.png$", x)) {
    imgpath <- tempfile(fileext = ".png")
    img <- magick::image_read(x)
    magick::image_write(img, path = imgpath)
  } else {
    imgpath <- x
  }

  # https://stat.ethz.ch/pipermail/r-help/2017-September/449253.html
  dat <- readBin(imgpath, what = "raw", size = 1, endian = "little", n = 1e+8)
  dat <- paste(dat, collapse = "")

  rtf_str <- sprintf(
    "{\\pict\\pngblip\\picwgoal%.0f\\pichgoal%.0f ",
    width * 1440,
    height * 1440
  )
  paste(rtf_str, dat, "\n}", sep = "")
}


#' @export
to_rtf.floating_external_img <- function(x, ...) {
  dims <- attr(x, "dims")
  width <- dims$width
  height <- dims$height

  pos <- attr(x, "pos")
  pos_x <- pos$x
  pos_y <- pos$y
  pos_h_from <- pos$h_from
  pos_v_from <- pos$v_from

  wrap <- attr(x, "wrap")
  wrap_type <- wrap$type
  wrap_side <- wrap$side
  wrap_dist_top <- wrap$dist_top
  wrap_dist_bottom <- wrap$dist_bottom
  wrap_dist_left <- wrap$dist_left
  wrap_dist_right <- wrap$dist_right

  if (!grepl("\\.png$", x)) {
    imgpath <- tempfile(fileext = ".png")
    img <- magick::image_read(x)
    magick::image_write(img, path = imgpath)
  } else {
    imgpath <- x
  }

  # Read image data
  dat <- readBin(imgpath, what = "raw", size = 1, endian = "little", n = 1e+8)
  dat <- paste(dat, collapse = "")

  # RTF units: twips (1/1440 inch)
  width_twips <- width * 1440
  height_twips <- height * 1440
  pos_x_twips <- pos_x * 1440
  pos_y_twips <- pos_y * 1440

  # EMU units for wrap distances (1 EMU = 1/914400 inch, so multiply inches by 914400)
  # But for RTF {\sp{\sn property}{\sv value}}, wrap distances are in EMUs
  dist_t_emus <- wrap_dist_top * 914400
  dist_b_emus <- wrap_dist_bottom * 914400
  dist_l_emus <- wrap_dist_left * 914400
  dist_r_emus <- wrap_dist_right * 914400

  # RTF wrap codes for \shpwr:
  # 1 = wrap around top and bottom
  # 2 = wrap around shape
  # 3 = none (as if shape isn't present)
  # 4 = wrap tightly
  # 5 = wrap text through shape
  wrap_code <- switch(
    wrap_type,
    "square" = "\\shpwr2",
    "tight" = "\\shpwr4",
    "topAndBottom" = "\\shpwr1",
    "through" = "\\shpwr5",
    "none" = "\\shpwr3",
    "\\shpwr2" # Default to square wrap
  )

  # RTF wrap side codes for \shpwrk:
  # 0 = wrap both sides
  # 1 = wrap left side only
  # 2 = wrap right side only
  # 3 = wrap only on largest side
  wrap_side_code <- switch(
    wrap_side,
    "bothSides" = "\\shpwrk0",
    "left" = "\\shpwrk1",
    "right" = "\\shpwrk2",
    "largest" = "\\shpwrk3",
    "\\shpwrk0" # Default
  )

  # RTF positioning codes
  # Horizontal positioning relative to:
  # \shpbxpage = page, \shpbxmargin = margin, \shpbxcolumn = column
  pos_h_code <- switch(
    pos_h_from,
    "page" = "\\shpbxpage",
    "margin" = "\\shpbxmargin",
    "column" = "\\shpbxcolumn",
    "character" = "\\shpbxcolumn",
    "\\shpbxmargin" # Default
  )

  # Vertical positioning relative to:
  # \shpbypage = page, \shpbymargin = margin, \shpbypara = paragraph
  pos_v_code <- switch(
    pos_v_from,
    "page" = "\\shpbypage",
    "margin" = "\\shpbymargin",
    "paragraph" = "\\shpbypara",
    "line" = "\\shpbypara",
    "\\shpbymargin" # Default
  )

  # posrelh property: 0=margin, 1=page, 2=column
  posrelh <- switch(
    pos_h_from,
    "margin" = 0,
    "page" = 1,
    "column" = 2,
    "character" = 2,
    0 # Default to margin
  )

  # posrelv property: 0=margin, 1=page, 2=paragraph
  posrelv <- switch(
    pos_v_from,
    "margin" = 0,
    "page" = 1,
    "paragraph" = 2,
    "line" = 2,
    0 # Default to margin
  )

  # Build the RTF shape structure following Word's format:
  # {\shp \shpleft... \shptop... \shpright... \shpbottom... \shpfhdr0
  #       \shpbx... \shpbxignore \shpby... \shpbyignore
  #       \shpwr... \shpwrk... \shpfblwtxt... \shpz... \shplid...
  #       {\*\shpinst {\sp{\sn prop}{\sv val}}... \par}}}

  rtf_str <- paste0(
    "{\\shp",
    sprintf(
      "\\shpleft%.0f\\shptop%.0f\\shpright%.0f\\shpbottom%.0f",
      pos_x_twips,
      pos_y_twips,
      pos_x_twips + width_twips,
      pos_y_twips + height_twips
    ),
    "\\shpfhdr0", # Not in header/footer
    pos_h_code,
    "\\shpbxignore", # Horizontal positioning
    pos_v_code,
    "\\shpbyignore", # Vertical positioning
    wrap_code,
    wrap_side_code, # Wrap type and side
    "\\shpfblwtxt0", # Image in front of text
    "\\shpz0", # Z-order
    "\\shplid1027" # Shape ID
  )

  # Start the shape properties ({\*\shpinst ...})
  rtf_str <- paste0(rtf_str, "{\\*\\shpinst")

  # Add shape properties following {\sp{\sn PropertyName}{\sv PropertyValue}} format
  # shapeType: 75 = picture frame
  rtf_str <- paste0(rtf_str, "{\\sp{\\sn shapeType}{\\sv 75}}")

  # pib: Binary picture data
  rtf_str <- paste0(
    rtf_str,
    sprintf(
      "{\\sp{\\sn pib}{\\sv {\\pict\\pngblip\\picwgoal%.0f\\pichgoal%.0f %s}}}",
      width_twips,
      height_twips,
      dat
    )
  )

  # Add wrap distances (in EMUs)
  if (dist_l_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dxWrapDistLeft}{\\sv %.0f}}", dist_l_emus)
    )
  }
  if (dist_r_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dxWrapDistRight}{\\sv %.0f}}", dist_r_emus)
    )
  }
  if (dist_t_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dyWrapDistTop}{\\sv %.0f}}", dist_t_emus)
    )
  }
  if (dist_b_emus > 0) {
    rtf_str <- paste0(
      rtf_str,
      sprintf("{\\sp{\\sn dyWrapDistBottom}{\\sv %.0f}}", dist_b_emus)
    )
  }

  # Add positioning properties (posrelh, posrelv are not in the Word example,
  # but keeping them won't hurt for compatibility with newer readers)
  # Note: These are commented out since the Word example doesn't include them
  # rtf_str <- paste0(rtf_str, sprintf("{\\sp{\\sn posrelh}{\\sv %d}}", posrelh))
  # rtf_str <- paste0(rtf_str, sprintf("{\\sp{\\sn posrelv}{\\sv %d}}", posrelv))

  # fBehindDocument: 0 = in front of text, 1 = behind text
  # Note: This is redundant with \shpfblwtxt0 above, but Word includes it
  # rtf_str <- paste0(rtf_str, "{\\sp{\\sn fBehindDocument}{\\sv 0}}")

  # Close the shape structure: \par}}} closes {\*\shpinst and {\shp
  paste0(rtf_str, "\\par}}}")
}


#' @export
to_rtf.block_toc <- function(x, ...) {
  # Field switches must be written with doubled backslashes inside
  # \fldinst (e.g. \\o, \\h, \\z, \\u). Single-backslash forms collide
  # with RTF control words (\o = overprint, \h = hidden, \u = Unicode
  # character), and Word will reject the document as malformed.
  field <- if (is.null(x$style) && is.null(x$seq_id)) {
    sprintf("TOC \\\\o \"1-%.0f\" \\\\h \\\\z \\\\u", x$level)
  } else if (!is.null(x$style)) {
    sprintf("TOC \\\\h \\\\z \\\\t \"%s\"", x$style)
  } else {
    sprintf("TOC \\\\h \\\\z \\\\c \"%s\"", x$seq_id)
  }
  paste0("{\\pard ", to_rtf(run_word_field(field = field)), "\\par}")
}

#' @export
to_rtf.run_word_field <- function(x, ...) {
  rtf_fp <- ""
  if (!is.null(x$pr)) {
    rtf_fp <- rpr_rtf(x$pr)
  }
  # The \fldrslt group is required by the RTF spec for a field group:
  # {\field {\*\fldinst <instructions>} {\fldrslt <result>}}. Word
  # tolerates its absence for simple fields like PAGE but rejects the
  # document for complex fields like TOC. An empty result is valid;
  # Word recalculates it when the field is updated.
  sprintf("%s{\\field{\\*\\fldinst %s}{\\fldrslt }}", rtf_fp, x$field)
}

#' @export
to_rtf.run_pagebreak <- function(x, ...) {
  "\\page"
}
#' @export
to_rtf.run_columnbreak <- function(x, ...) {
  "\\column "
}
#' @export
to_rtf.run_linebreak <- function(x, ...) {
  "\\line"
}

#' @export
to_rtf.hyperlink_ftext <- function(x, ...) {
  rtf_fp <- ""
  if (!is.null(x$pr)) {
    rtf_fp <- rpr_rtf(x$pr)
  }
  sprintf(
    "%s{\\field{\\*\\fldinst HYPERLINK \"%s\"}{\\fldrslt %s}}",
    rtf_fp,
    officer_url_encode(x$href),
    str_encode_to_rtf(x$value)
  )
}
rtf_bookmark <- function(id, str) {
  bm_start_str <- sprintf("{\\*\\bkmkstart %s}", id)
  bm_start_end <- sprintf("{\\*\\bkmkend %s}", id)
  paste0(bm_start_str, str, bm_start_end)
}

#' @export
to_rtf.run_autonum <- function(x, ...) {
  if (is.null(x$seq_id)) {
    return("")
  }

  run_str_pre <- to_rtf(ftext(x$pre_label, prop = x$pr))
  run_str_post <- to_rtf(ftext(x$post_label, prop = x$pr))

  seq_str <- paste0("SEQ ", x$seq_id, " \\\\* Arabic")
  if (!is.null(x$start_at) && is.numeric(x$start_at)) {
    seq_str <- paste(seq_str, "\\r", as.integer(x$start_at))
  }

  sqf <- run_word_field(field = seq_str, prop = x$pr)
  sf_str <- to_rtf(sqf)

  if (x$tnd > 0) {
    z <- paste0(
      to_rtf(run_word_field(
        field = paste0("STYLEREF ", x$tnd, " \\r"),
        prop = x$pr
      )),
      to_rtf(ftext(x$tns, prop = x$pr))
    )
    sf_str <- paste0(z, sf_str)
  }

  if (!is.null(x$bookmark)) {
    if (x$bookmark_all) {
      out <- paste0(
        rtf_bookmark(id = x$bookmark, str = paste0(run_str_pre, sf_str)),
        run_str_post
      )
    } else {
      sf_str <- rtf_bookmark(id = x$bookmark, sf_str)
      out <- paste0(run_str_pre, sf_str, run_str_post)
    }
  } else {
    out <- paste0(run_str_pre, sf_str, run_str_post)
  }

  out
}

#' @export
to_rtf.run_reference <- function(x, ...) {
  out <- to_rtf(run_word_field(
    field = paste0(" REF ", x$id, " \\\\h "),
    prop = x$pr
  ))
  out
}

## to_rtf sections ----

#' @export
to_rtf.block_section <- function(x, ...) {
  # API convention: block_section() configures the section that
  # *follows* (content added after it). In RTF that maps to
  # \sect (close previous section) then the property keywords,
  # which apply to the newly opened section. \sectd resets section
  # properties to defaults so that a prop_section() without columns
  # does not inherit \cols2 from a previous column section.
  paste0("\\sect\\sectd", to_rtf(x$property))
}

#' @export
to_rtf.prop_section <- function(x, ...) {
  rtf_type <- if (is.null(x$type)) {
    ""
  } else if (x$type %in% "nextPage") {
    "\\sbkpage"
  } else if (x$type %in% "oddPage") {
    "\\sbkodd"
  } else if (x$type %in% "evenPage") {
    "\\sbkeven"
  } else {
    "\\sbknone"
  }

  hfr_type <- guess_hfr_type(x)
  extra_rtf <- ""
  str_h <- ""
  str_f <- ""
  str_h_even <- ""
  str_f_even <- ""
  str_h_first <- ""
  str_f_first <- ""

  if ("none" %in% hfr_type) {
    hfr_str <- ""
  } else if ("default" %in% hfr_type) {
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\header ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footer ", to_rtf(x$footer_default), "}")
    }
    hfr_str <- paste0(str_h, str_f)
  } else if ("evenodd" %in% hfr_type) {
    extra_rtf <- "\\facingp"
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\headerl ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$header_even)) {
      str_h_even <- paste0("{\\headerr ", to_rtf(x$header_even), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footerl ", to_rtf(x$footer_default), "}")
    }
    if (!is.null(x$footer_even)) {
      str_f_even <- paste0("{\\footerr ", to_rtf(x$footer_even), "}")
    }
    hfr_str <- paste0(
      str_h,
      str_h_even,
      str_f,
      str_f_even
    )
  } else if ("evenoddfirst" %in% hfr_type) {
    extra_rtf <- "\\facingp\\titlepg"
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\headerl ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$header_even)) {
      str_h_even <- paste0("{\\headerr ", to_rtf(x$header_even), "}")
    }
    if (!is.null(x$header_first)) {
      str_h_first <- paste0("{\\headerf ", to_rtf(x$header_first), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footerl ", to_rtf(x$footer_default), "}")
    }
    if (!is.null(x$footer_even)) {
      str_f_even <- paste0("{\\footerr ", to_rtf(x$footer_even), "}")
    }
    if (!is.null(x$footer_first)) {
      str_f_first <- paste0("{\\footerf ", to_rtf(x$footer_first), "}")
    }
    hfr_str <- paste0(
      str_h,
      str_h_even,
      str_h_first,
      str_f,
      str_f_even,
      str_f_first
    )
  } else {
    extra_rtf <- "\\titlepg"
    if (!is.null(x$header_default)) {
      str_h <- paste0("{\\headerl ", to_rtf(x$header_default), "}")
    }
    if (!is.null(x$header_first)) {
      str_h_first <- paste0("{\\headerf ", to_rtf(x$header_first), "}")
    }
    if (!is.null(x$footer_default)) {
      str_f <- paste0("{\\footerl ", to_rtf(x$footer_default), "}")
    }
    if (!is.null(x$footer_first)) {
      str_f_first <- paste0("{\\footerf ", to_rtf(x$footer_first), "}")
    }
    hfr_str <- paste0(
      str_h,
      str_h_first,
      str_f,
      str_f_first
    )
  }

  rtf_str <- paste0(
    if (!is.null(x$page_margins)) to_rtf(x$page_margins),
    if (!is.null(x$page_size)) to_rtf(x$page_size),
    rtf_type,
    if (!is.null(x$section_columns)) to_rtf(x$section_columns),
    extra_rtf
  )

  # sections def and hdr/ftr content are separated in rtf
  attr(rtf_str, "hfr") <- hfr_str

  rtf_str
}

guess_hfr_type <- function(x) {
  has_default <- (!is.null(x$header_default) || !is.null(x$footer_default))
  has_even <- (!is.null(x$header_even) || !is.null(x$footer_even))
  has_first <- (!is.null(x$header_first) || !is.null(x$footer_first))

  if (has_default && !has_even && !has_first) {
    "default"
  } else if (has_default && has_even && !has_first) {
    "evenodd"
  } else if (has_default && has_even && has_first) {
    "evenoddfirst"
  } else if (has_default && !has_even && has_first) {
    "defaultfirst"
  } else {
    "none"
  }
}

#' @export
to_rtf.page_mar <- function(x, ...) {
  sprintf(
    "\\margb%.0f\\margt%.0f\\margr%.0f\\margl%.0f\\gutter%.0f\\headery%.0f\\footery%.0f",
    inch_to_tweep(x$bottom),
    inch_to_tweep(x$top),
    inch_to_tweep(x$right),
    inch_to_tweep(x$left),
    inch_to_tweep(x$gutter),
    inch_to_tweep(x$header),
    inch_to_tweep(x$footer)
  )
}

#' @export
to_rtf.page_size <- function(x, ...) {
  if (!is.null(x$width) && !is.null(x$height)) {
    out <- sprintf(
      "\\pghsxn%.0f\\pgwsxn%.0f%s",
      inch_to_tweep(x$height),
      inch_to_tweep(x$width),
      if ("landscape" %in% x$orient) {
        "\\lndscpsxn"
      } else {
        ""
      }
    )
  } else {
    out <- if ("landscape" %in% x$orient) {
      "\\lndscpsxn"
    } else {
      ""
    }
  }

  out
}

#' @export
to_rtf.section_columns <- function(x, ...) {
  widths <- x$widths * 20 * 72
  space <- x$space * 20 * 72

  # RTF needs one <\colno N \colw N> per column for unequal widths.
  # \colsr (per-column right space) duplicates the global \colsx when
  # spaces are uniform, but matches what Word/LibreOffice emit and
  # leaves room for per-column spacing later. The last column has no
  # trailing \colsr.
  columns_str_all_but_last <- sprintf(
    "\\colno%.0f\\colw%.0f\\colsr%.0f",
    seq(length(widths) - 1),
    widths[-length(widths)],
    rep(space, length(widths) - 1)
  )
  columns_str_last <- sprintf(
    "\\colno%.0f\\colw%.0f",
    length(widths),
    widths[length(widths)]
  )
  columns_str <- c(columns_str_all_but_last, columns_str_last)

  linebetcol <- ""
  if (x$sep) {
    linebetcol <- "\\linebetcol"
  }

  sprintf(
    "\\cols%.0f\\colsx%.0f%s%s",
    length(widths),
    space,
    linebetcol,
    paste(columns_str, collapse = "")
  )
}

# RTF management ----
font_table <- function(families) {
  data.frame(
    family = families,
    index = seq_along(families) - 1,
    match = sprintf("%%font:%s%%", families),
    stringsAsFactors = FALSE
  )
}
color_table <- function(colors) {
  data.frame(
    color = colors,
    index = seq_along(colors),
    match_ftcolor = sprintf("%%ftcolor:%s%%", colors),
    match_ftshading = sprintf("%%ftshading:%s%%", colors),
    match_ftlinecolor = sprintf("%%ftlinecolor:%s%%", colors),
    match_ftbgcolor = sprintf("%%ftbgcolor:%s%%", colors),
    stringsAsFactors = FALSE
  )
}

ppr_rtf <- function(x, include_pard = TRUE) {
  text_align_ <- ""
  if (!is.na(x$text.align)) {
    if ("left" %in% x$text.align) {
      text_align_ <- "\\ql"
    } else if ("center" %in% x$text.align) {
      text_align_ <- "\\qc"
    } else if ("right" %in% x$text.align) {
      text_align_ <- "\\qr"
    } else if ("justified" %in% x$text.align) {
      text_align_ <- "\\qj"
    }
  }

  tabs <- ""
  if (!is.null(x$tabs) && length(x$tabs) > 0 && !isFALSE(x$tabs)) {
    tabs <- paste0(sapply(x$tabs, rtf_fp_tab), collapse = "")
  }

  keep_with_next <- ""
  if (isTRUE(x$keep_with_next)) {
    keep_with_next <- "\\keepn"
  }
  leftright_padding <- ""
  if (!is.na(x$padding.left) && !is.na(x$padding.right)) {
    hang <- x$hanging %||% NA_real_
    first <- x$first_line %||% NA_real_
    if (!is.na(hang) && hang > 0) {
      fi <- sprintf("\\fi-%.0f", hang * 20)
    } else if (!is.na(first) && first > 0) {
      fi <- sprintf("\\fi%.0f", first * 20)
    } else {
      fi <- "\\fi0"
    }
    leftright_padding <- sprintf(
      "%s\\li%.0f\\ri%.0f",
      fi,
      x$padding.left * 20,
      x$padding.right * 20
    )
  }
  topbot_spacing <- ""
  if (!is.na(x$padding.bottom) && !is.na(x$padding.top)) {
    topbot_spacing <- sprintf(
      "\\sb%.0f\\sa%.0f",
      x$padding.bottom * 20,
      x$padding.top * 20
    )
  }

  line_spacing <- ""
  if (!is.na(x$line_spacing)) {
    line_spacing <- sprintf("\\sl%.0f\\slmult1", 240 * x$line_spacing)
  }

  out <- paste0(
    line_spacing,
    text_align_,
    tabs,
    keep_with_next,
    topbot_spacing,
    leftright_padding
  )
  if (nchar(out) > 0) {
    prefix <- if (isTRUE(include_pard)) "\\pard" else ""
    out <- paste0(prefix, out, " ")
  }
  out
}

rpr_rtf <- function(x) {
  out <- ""

  if (!is.na(x$font.family)) {
    out <- paste0(out, "%font:", x$font.family, "%")
  }

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, "\\i")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, "\\b")
    }
  }
  if (!is.na(x$underlined)) {
    if (x$underlined) {
      out <- paste0(out, "\\ul")
    }
  }
  if (!is.na(x$strike)) {
    if (x$strike) {
      out <- paste0(out, "\\strike")
    }
  }

  if (x$vertical.align == "superscript" && !is.na(x$font.size)) {
    out <- paste0(out, sprintf("\\super\\up%.0f", round(x$font.size / 2)))
  } else if (x$vertical.align == "subscript" && !is.na(x$font.size)) {
    out <- paste0(out, sprintf("\\sub\\dn%.0f", round(x$font.size / 2)))
  }

  if (!is.na(x$font.size)) {
    out <- paste0(
      out,
      rtf_fontsize(x$font.size)
    )
  }

  if (!is.na(x$color)) {
    out <- paste0(
      out,
      paste0("%ftcolor:", x$color, "%")
    )
  }

  if (nchar(out) > 0) {
    out <- paste0(out, " ")
  }
  out
}

border_rtf <- function(x, side) {
  x$style[x$style %in% "solid"] <- "single"
  if (!x$style %in% c("dotted", "dashed", "single")) {
    x$style <- "single"
  }
  if (x$width < 0.0001 || is_transparent(x$color)) {
    return("")
  }

  linestyle <- c(
    "dotted" = "\\brdrdot",
    "dashed" = "\\brdrdash",
    "single" = "\\brdrs"
  )

  style_ <- linestyle[x$style]
  width_ <- paste0("\\brdrw", round(x$width * 20))
  color_ <- paste0("%ftlinecolor:", x$color, "%")

  paste0(
    "\\clbrdr",
    substr(side, 1, 1),
    style_,
    width_,
    color_
  )
}

tcpr_rtf <- function(x) {
  if (colalpha(x$background.color) > 0) {
    background.color <- paste0("%ftbgcolor:", x$background.color, "%")
  } else {
    background.color <- ""
  }
  vertalign <- c(
    "center" = "\\clvertalc",
    "top" = "\\clvertalt",
    "bottom" = "\\clvertalb"
  )
  vertical.align <- vertalign[x$vertical.align]

  text.direction <- ""
  text.direction[x$text.direction %in% "btlr"] <- "\\stextflow2"
  text.direction[x$text.direction %in% "tbrl"] <- "\\stextflow3"

  bb <- border_rtf(x$border.bottom, side = "bottom")
  bt <- border_rtf(x$border.top, side = "top")
  bl <- border_rtf(x$border.left, side = "left")
  br <- border_rtf(x$border.right, side = "right")

  rowspan <- ""
  if (x$rowspan > 1) {
    rowspan <- "\\clmgf"
  } else if (x$rowspan < 1) {
    rowspan <- "\\clmrg"
  }
  colspan <- ""
  if (x$colspan < 1) {
    colspan <- "\\clvmrg"
  } else {
    colspan <- "\\clvmgf"
  }

  paste0(
    bb,
    bt,
    bl,
    br,
    text.direction,
    rowspan,
    colspan,
    vertical.align,
    background.color
  )
}

# main ----
#' @export
#' @title Create an RTF document object
#' @description Creation of the object representing an
#' RTF document which can then receive contents with
#' the [rtf_add()] function and be written to a file with
#' the `print(x, target="doc.rtf")` function.
#' @param def_sec a [prop_section] object used to define the default section
#' (page size, margins, header / footer, columns) applied to content added
#' before any explicit [block_section()].
#' @param normal_par an object generated by [fp_par()]
#' @param normal_chunk an object generated by [fp_text()]
#'
#' @section Built-in styles:
#' `rtf_doc()` registers four paragraph styles: `"Normal"` (built from
#' `normal_par` / `normal_chunk`) and `"heading 1"` / `"heading 2"` /
#' `"heading 3"` (bold, dark blue shades, with outline levels so the
#' paragraphs feed Word's navigation pane and TOC field). The heading
#' names mirror Word's own convention so a script that adds paragraphs
#' with `style = "heading 1"` works identically against Word output
#' (`body_add_par(..., style = "heading 1")`) and RTF output
#' (`rtf_add(..., style = "heading 1")`).
#'
#' To override a built-in look (size, color, spacing, outline level),
#' call [rtf_set_paragraph_style()] after `rtf_doc()` with the same
#' `style_id`.
#'
#' ```r
#' doc <- rtf_doc()
#' doc <- rtf_set_paragraph_style(
#'   doc,
#'   style_id = "heading 1",
#'   fp_p = fp_par(text.align = "center", padding = 12),
#'   fp_t = fp_text_lite(bold = TRUE, font.size = 24, color = "#000000"),
#'   outline_level = 1L
#' )
#' ```
#'
#' Use [rtf_styles_info()] to inspect the current style table.
#' @examples
#' rtf_doc(normal_par = fp_par(padding = 3))
#' @seealso [read_docx()], [print.rtf()], [rtf_add()]. See `?rtf_add` for a
#' complete multi-section example exercising `def_sec`, headers / footers
#' and several `block_section()` calls.
#' @return an object of class `rtf` representing an
#' empty RTF document.
rtf_doc <- function(
  def_sec = prop_section(),
  normal_par = fp_par(),
  normal_chunk = fp_text(font.family = "Arial", font.size = 11)
) {
  styles <- data.frame(
    style_id = "Normal",
    style_name = "Normal",
    type = "paragraph",
    rtf_index = 1L,
    based_on = NA_character_,
    rtf = rtf_par_style(fp_p = normal_par, fp_t = normal_chunk),
    stringsAsFactors = FALSE
  )

  x <- list(
    content = list(),
    default_section = def_sec,
    styles = styles,
    normal_par = normal_par,
    normal_chunk = normal_chunk
  )
  class(x) <- "rtf"

  # Built-in heading styles. Names align with Word's "heading 1".."heading 3"
  # convention. outline_level wires the style into Word's navigation pane
  # and TOC field.
  x <- rtf_set_paragraph_style(
    x, style_id = "heading 1", style_name = "Heading 1",
    fp_p = fp_par(padding.top = 12, padding.bottom = 6),
    fp_t = fp_text_lite(bold = TRUE, font.size = 18, color = "#1F4E79"),
    outline_level = 1L
  )
  x <- rtf_set_paragraph_style(
    x, style_id = "heading 2", style_name = "Heading 2",
    fp_p = fp_par(padding.top = 8, padding.bottom = 4),
    fp_t = fp_text_lite(bold = TRUE, font.size = 14, color = "#2E75B6"),
    outline_level = 2L
  )
  x <- rtf_set_paragraph_style(
    x, style_id = "heading 3", style_name = "Heading 3",
    fp_p = fp_par(padding.top = 6, padding.bottom = 3),
    fp_t = fp_text_lite(bold = TRUE, font.size = 12, color = "#5B9BD5"),
    outline_level = 3L
  )
  x
}

#' @export
#' @title Add content into an RTF document
#' @description This function add 'officer' objects into an RTF document.
#' Values are added as new paragraphs. See section 'Methods (by class)'
#' that list supported objects.
#'
#' @section Section model in RTF:
#' A `block_section` added with `rtf_add()` applies to the content that
#' **follows** the call: it opens a new section whose layout (orientation,
#' columns, margins, headers / footers) is inherited by every paragraph,
#' table or graphic added afterwards, until the next `block_section` (or
#' the end of the document).
#'
#' Typical pattern: declare the section, then add the content.
#'
#' ```r
#' doc <- rtf_doc() |>
#'   rtf_add(block_section(prop_section(
#'     page_size = page_size(orient = "landscape")
#'   ))) |>
#'   rtf_add(fpar("This paragraph is in landscape orientation."))
#' ```
#'
#' The first section of the document is configured via the `def_sec`
#' argument of [rtf_doc()] (a [prop_section] object, not a
#' `block_section`). It applies to every element added before the first
#' `rtf_add(block_section(...))` call.
#'
#' The Word output uses the opposite model: `body_end_block_section()`
#' applies to the content that *precedes* the call. See
#' [body_end_block_section()].
#'
#' @param x rtf object, created by [rtf_doc()].
#' @param value object to add in the document. Supported objects
#' are vectors, graphics, block of formatted paragraphs. Use package
#' 'flextable' to add tables.
#' @param ... further arguments passed to or from other methods. When
#' adding a `ggplot` object or [plot_instr], these arguments will be used
#' by png function. See section 'Methods' to see what arguments can be used.
#' @example inst/examples/example_rtf_add.R
#' @example inst/examples/example_rtf_sections.R
rtf_add <- function(x, value, ...) {
  UseMethod("rtf_add", value)
}


#' @export
#' @describeIn rtf_add add a new section definition
rtf_add.block_section <- function(x, value, ...) {
  x$content[[length(x$content) + 1]] <- value
  x
}

# Walk the based_on chain of a style, returning the style ids in root-to-leaf
# order. `parents` is a named vector that maps each style_id to its parent
# style_id (or NA for roots), used as a constant-time dictionary. `visited`
# carries ids already traversed, so the walk terminates when a cycle exists
# (a parent that loops back to a descendant).
walk_based_on_chain <- function(style_id, parents, visited = character()) {
  if (is.na(style_id) || style_id %in% visited) {
    return(character())
  }
  c(
    walk_based_on_chain(parents[[style_id]], parents, c(visited, style_id)),
    style_id
  )
}

# Resolve a user-facing style id to (rtf_index, rtf_props). Concatenates the
# fragments of the based_on chain so the returned props string fully describes
# the style (the additive model the RTF spec requires for styled paragraphs).
resolve_rtf_style <- function(x, style) {
  if (is.null(style)) {
    return(NULL)
  }
  if (!is.character(style) || length(style) != 1L) {
    stop("`style` must be NULL or a single string matching a style_id.")
  }
  idx <- match(style, x$styles$style_id)
  if (is.na(idx)) {
    stop(
      "style ", shQuote(style),
      " is not registered. Use rtf_styles_info() to list available styles ",
      "or rtf_set_paragraph_style() to add one."
    )
  }
  # Walk the based_on chain from root to leaf. styles$rtf is stored
  # without leading \pard (see rtf_set_paragraph_style), so we can just
  # concatenate the chain in order: parent props come first, child
  # props win on conflicts under the RTF additive model.
  parents <- setNames(x$styles$based_on, x$styles$style_id)
  chain <- walk_based_on_chain(style, parents)
  rtf_by_id <- setNames(x$styles$rtf, x$styles$style_id)
  props <- paste0(rtf_by_id[chain], collapse = "")
  list(index = x$styles$rtf_index[idx], props = props)
}

apply_rtf_style <- function(value, style) {
  if (is.null(style)) {
    return(value)
  }
  attr(value, "rtf_style_index") <- style$index
  attr(value, "rtf_style_props") <- style$props
  value
}

#' @export
#' @describeIn rtf_add add characters as new paragraphs
#' @param style style identifier (`style_id`) of a paragraph style registered
#' on the document. Defaults to `NULL` (use the document's normal style). See
#' [rtf_set_paragraph_style()] and [rtf_styles_info()].
rtf_add.character <- function(x, value, style = NULL, ...) {
  styled <- resolve_rtf_style(x, style)
  chunk_prop <- if (is.null(styled)) x$normal_chunk else NULL
  par_prop <- if (is.null(styled)) x$normal_par else fp_par()
  values <- lapply(
    X = str_encode_to_rtf(value),
    function(z) {
      apply_rtf_style(
        fpar(ftext(z, prop = chunk_prop), fp_p = par_prop),
        styled
      )
    }
  )

  x$content <- append(x$content, values)
  x
}

#' @export
#' @describeIn rtf_add add a factor vector as new paragraphs
rtf_add.factor <- function(x, value, style = NULL, ...) {
  styled <- resolve_rtf_style(x, style)
  chunk_prop <- if (is.null(styled)) x$normal_chunk else NULL
  par_prop <- if (is.null(styled)) x$normal_par else fp_par()
  values <- lapply(
    X = str_encode_to_rtf(as.character(value)),
    function(z) {
      apply_rtf_style(
        fpar(ftext(z, prop = chunk_prop), fp_p = par_prop),
        styled
      )
    }
  )

  x$content <- append(x$content, values)
  x
}

#' @export
#' @describeIn rtf_add add a double vector as new paragraphs
#' @param formatter function used to format the numerical values
rtf_add.double <- function(x, value, formatter = formatC, style = NULL, ...) {
  styled <- resolve_rtf_style(x, style)
  chunk_prop <- if (is.null(styled)) x$normal_chunk else NULL
  par_prop <- if (is.null(styled)) x$normal_par else fp_par()
  values <- lapply(
    X = value,
    function(z) {
      apply_rtf_style(
        fpar(ftext(formatC(z), prop = chunk_prop), fp_p = par_prop),
        styled
      )
    }
  )

  x$content <- append(x$content, values)
  x
}

#' @export
#' @describeIn rtf_add add an [fpar()]
rtf_add.fpar <- function(x, value, style = NULL, ...) {
  styled <- resolve_rtf_style(x, style)
  x$content[[length(x$content) + 1]] <- apply_rtf_style(value, styled)
  x
}

#' @export
#' @describeIn rtf_add add an [block_list()]
rtf_add.block_list <- function(x, value, style = NULL, ...) {
  styled <- resolve_rtf_style(x, style)
  items <- lapply(as.list(value), apply_rtf_style, style = styled)
  x$content <- append(x$content, items)
  x
}

#' @export
#' @describeIn rtf_add add a [block_toc()] (table of contents).
#' Word populates the TOC at open time from paragraphs styled with the
#' built-in heading styles (which carry an outline level). LibreOffice
#' will not render the TOC automatically.
rtf_add.block_toc <- function(x, value, ...) {
  x$content[[length(x$content) + 1]] <- value
  x
}


#' @export
#' @describeIn rtf_add add a ggplot2
#' @param width height in inches
#' @param height height in inches
#' @param res resolution of the png image in ppi
#' @param scale Multiplicative scaling factor, same as in ggsave
#' @param ppr [fp_par()] to apply to paragraph.
rtf_add.gg <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  ppr = fp_par(text.align = "center"),
  ...
) {
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      print(value)
    },
    finally = {
      dev.off()
    }
  )

  value <- fpar(
    external_img(
      src = file,
      width = width,
      height = height,
      guess_size = FALSE
    ),
    fp_p = ppr
  )

  x$content[[length(x$content) + 1]] <- value
  x
}
#' @export
#' @describeIn rtf_add add a [plot_instr()] object
rtf_add.plot_instr <- function(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  ppr = fp_par(text.align = "center"),
  ...
) {
  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )

  value <- fpar(
    external_img(
      src = file,
      width = width,
      height = height,
      guess_size = FALSE
    ),
    fp_p = ppr
  )

  x$content[[length(x$content) + 1]] <- value
  x
}

#' @title Write an 'RTF' File
#' @description Write the RTF object and its content to a file.
#' @param x an 'rtf' object created with [rtf_doc()]
#' @param target path to the RTF file to write
#' @param ... unused
#' @examples
#' # write a rdocx object in a rtf file ----
#' doc <- rtf_doc()
#' print(doc, target = tempfile(fileext = ".rtf"))
#' @export
#' @seealso [rtf_doc()]
print.rtf <- function(x, target = NULL, ...) {
  if (is.null(target)) {
    cat("rtf document with", length(x$content), "element(s)\n")
    return(invisible())
  }

  # Walk content elements, emit one RTF chunk per element and interleave
  # per-section header/footer groups so that each section carries its own
  # {\header}{\footer}. The default section's hfr applies to content added
  # before the first block_section; each subsequent block_section opens a
  # new section whose hfr is emitted just before the next \sect (or at
  # end of document for the last section). hfr groups must precede the
  # \sect that closes the section they describe (RTF spec).
  default_sect <- to_rtf(x$default_section)
  rtf_content <- character()
  current_hfr <- attr(default_sect, "hfr")
  for (elem in x$content) {
    if (inherits(elem, "block_section")) {
      props <- to_rtf(elem$property)
      rtf_content <- c(
        rtf_content,
        current_hfr,
        paste0("\\sect\\sectd", props)
      )
      current_hfr <- attr(props, "hfr")
    } else {
      rtf_content <- c(rtf_content, to_rtf(elem))
    }
  }
  trailing_hfr <- current_hfr

  rtf_ss <- rtf_stylesheet(x)
  txt_to_scan <- c(rtf_content, rtf_ss, trailing_hfr)
  m <- gregexec(pattern = "%font:[a-zA-Z ]+%", text = txt_to_scan)
  fonts <- unique(unlist(regmatches(txt_to_scan, m)))
  fonts <- gsub("(^%font:|%$)", "", fonts)
  family_table <- font_table(families = fonts)

  m <- gregexec(pattern = "\\%ftcolor\\:[a-zA-Z #0-9]+\\%", text = txt_to_scan)
  ftcolor <- unlist(regmatches(txt_to_scan, m))
  m <- gregexec(
    pattern = "\\%ftshading\\:[a-zA-Z #0-9]+\\%",
    text = txt_to_scan
  )
  ftshading <- unlist(regmatches(txt_to_scan, m))
  m <- gregexec(
    pattern = "\\%ftlinecolor\\:[a-zA-Z #0-9]+\\%",
    text = txt_to_scan
  )
  ftlinecolor <- unlist(regmatches(txt_to_scan, m))
  m <- gregexec(
    pattern = "\\%ftbgcolor\\:[a-zA-Z #0-9]+\\%",
    text = txt_to_scan
  )
  ftbgcolor <- unlist(regmatches(txt_to_scan, m))

  supported_highlight <- c(
    "black",
    "blue",
    "cyan",
    "green",
    "magenta",
    "red",
    "yellow",
    "darkblue",
    "darkcyan",
    "darkgreen",
    "darkmagenta",
    "darkred",
    "#8B8000",
    "darkgrey",
    "lightgray"
  )

  colors <- unique(c(
    ftcolor,
    ftshading,
    ftlinecolor,
    ftbgcolor,
    supported_highlight
  ))
  colors <- gsub(
    "(^%(ftcolor|ftshading|ftlinecolor|ftbgcolor):|%$)",
    "",
    colors
  )
  color_table <- color_table(colors = colors)

  header <- c(
    "{\\rtf1\\deff0",
    rtfize_font_table(family_table),
    rtfize_color_table(color_table)
  )

  all_strings <- c(
    header,
    rtf_ss,
    default_sect,
    rtf_content,
    trailing_hfr,
    "}"
  )

  all_strings <- fix_font_ref(all_strings, family_table)
  all_strings <- fix_font_color(all_strings, color_table)
  all_strings <- fix_line_color(all_strings, color_table)
  all_strings <- fix_bg_color(all_strings, color_table)
  all_strings <- fix_font_shading(all_strings, color_table)

  writeLines(all_strings, target, useBytes = TRUE)
  target
}

#' @export
#' @title Encode UTF8 string to RTF
#' @description Convert strings to RTF valid codes.
#' @param z character vector to be converted
#' @family functions for officer extensions
#' @return character vector of results encoded to RTF
#' @keywords internal
#' @examples
#' str_encode_to_rtf("Hello World")
str_encode_to_rtf <- function(z) {
  # https://stackoverflow.com/questions/1368020/how-to-output-unicode-string-to-rtf-using-c
  char_list <- strsplit(z, "")
  int_list <- lapply(z, utf8ToInt)
  rtf_strings <- mapply(
    function(charv, intv) {
      is_tobeprotected <- charv %in% c("\\", "{", "}")
      # is_std_ansi <- intv < 128
      # is_ext_ansi <- !is_std_ansi & intv < 256
      is_ext_ansi <- intv < 128
      is_unic2b <- !is_ext_ansi & intv < 32768
      is_unic4b <- !is_ext_ansi & !is_unic2b
      charv[is_unic2b] <- paste0("\\uc1\\u", intv[is_unic2b], "?")
      charv[is_unic4b] <- paste0("\\uc1\\u", intv[is_unic4b] - 65536, "?")
      paste0(charv, collapse = "")
    },
    char_list,
    int_list,
    SIMPLIFY = TRUE
  )
  as.character(rtf_strings)
}


# tools ----
rtf_fontsize <- function(x) {
  size <- formatC(x * 2, format = "f", digits = 0)
  z <- sprintf("\\fs%s", size)
  z[is.na(x)] <- ""
  z
}
rtf_color <- function(x, color_table, mark = "cf") {
  index <- color_table$index[match(x, color_table$color)]
  z <- sprintf("\\%s%.0f", mark, index)
  z[is.na(index)] <- ""
  z
}

rtfize_font_table <- function(x) {
  z <- sprintf("{\\f%.0f %s;}", x$index, x$family)
  z <- paste0(z, collapse = "")
  paste0("{\\fonttbl", z, "}")
}

rtf_color_code <- function(color) {
  color <- as.vector(col2rgb(color, alpha = FALSE))
  sprintf(
    "\\red%.0f\\green%.0f\\blue%.0f;",
    color[1],
    color[2],
    color[3]
  )
}

rtf_fp_tab <- function(x) {
  tab_style <- "\\tqdec"
  if (x$style %in% "left") {
    tab_style <- ""
  } else if (x$style %in% "center") {
    tab_style <- "\\tqc"
  } else if (x$style %in% "right") {
    tab_style <- "\\tqr"
  }

  sprintf(
    "\\tx%.0f%s",
    inch_to_tweep(x$pos),
    tab_style
  )
}

rtfize_color_table <- function(x) {
  z <- vapply(x$color, rtf_color_code, "")
  z <- paste0(z, collapse = "")
  paste0("{\\colortbl;", z, "}")
}

fix_font_ref <- function(x, font_tbl) {
  for (i in seq_len(nrow(font_tbl))) {
    m <- gregexec(pattern = font_tbl$match[i], text = x)
    regmatches(x, m) <- paste0("\\f", font_tbl$index[i])
  }
  x
}
fix_font_color <- function(x, color_tbl) {
  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftcolor[i], text = x)
    regmatches(x, m) <- paste0("\\cf", color_tbl$index[i])
  }
  x
}
fix_line_color <- function(x, color_tbl) {
  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftlinecolor[i], text = x)
    regmatches(x, m) <- paste0("\\brdrcf", color_tbl$index[i])
  }
  x
}
fix_bg_color <- function(x, color_tbl) {
  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftbgcolor[i], text = x)
    regmatches(x, m) <- paste0("\\clcbpat", color_tbl$index[i])
  }
  x
}
fix_font_shading <- function(x, color_tbl) {
  supported_highlight <- c(
    "black",
    "blue",
    "cyan",
    "green",
    "magenta",
    "red",
    "yellow",
    "darkblue",
    "darkcyan",
    "darkgreen",
    "darkmagenta",
    "darkred",
    "#8B8000",
    "darkgrey",
    "lightgray"
  )

  supported_highlight <-
    data.frame(
      colors = supported_highlight,
      index = seq_along(supported_highlight),
      match_ftshading = sprintf("%%ftshading:%s%%", supported_highlight)
    )

  for (i in seq_len(nrow(color_tbl))) {
    m <- gregexec(pattern = color_tbl$match_ftshading[i], text = x)
    matches <- Filter(function(x) x > -1, m)
    if (length(matches) > 0) {
      used_shading <- unique(unlist(regmatches(x, m)))
      if (
        !used_shading %in%
          c(supported_highlight$match_ftshading, "%ftshading:transparent%")
      ) {
        stop(
          "chunk highlight color (",
          used_shading,
          ") in RTF can only have values: ",
          paste0(shQuote(supported_highlight$colors), collapse = ", ")
        )
      }
    }

    regmatches(x, m) <- sprintf("\\highlight%.0f", color_tbl$index[i])
  }
  x <- gsub("%ftshading:transparent%", "", x, fixed = TRUE)
  x
}

rtf_par_style <- function(fp_p = fp_par(), fp_t = NULL, include_pard = FALSE) {
  if (!is.null(fp_t)) {
    fp_t_rtf <- rpr_rtf(fp_t)
  } else {
    fp_t_rtf <- ""
  }

  paste0(ppr_rtf(fp_p, include_pard = include_pard), fp_t_rtf)
}

#' @export
#' @title Add or replace a paragraph style in an RTF document
#' @description Add or replace a paragraph style in the document stylesheet
#' so that subsequent paragraphs can reference it via the `style` argument of
#' [rtf_add()]. The function mirrors [docx_set_paragraph_style()] for Word.
#' @param x an rtf object created by [rtf_doc()].
#' @param style_id user-facing identifier used as the key when a paragraph
#' references the style (`rtf_add(..., style = style_id)`).
#' @param style_name display label of the style as it appears in Word's style
#' menu. Defaults to `style_id`.
#' @param base_on the `style_id` of the parent style. Properties not defined
#' in the new style are inherited from the parent. Defaults to `"Normal"`.
#' @param fp_p paragraph formatting properties, see [fp_par()].
#' @param fp_t text formatting properties, see [fp_text()]. If `NULL` the
#' paragraph inherits text properties from the parent style.
#' @param outline_level integer between 1 and 9, or `NULL`. Sets the outline
#' level (`\\outlinelevel`) so paragraphs using the style feed Word's
#' navigation pane and TOC field. Level 1 is the topmost (used by Heading 1).
#' @return the rtf object with the style added or updated.
#' @seealso [docx_set_paragraph_style()], [rtf_doc()], [rtf_add()]
#' @examples
#' doc <- rtf_doc()
#' doc <- rtf_set_paragraph_style(
#'   doc,
#'   style_id = "Callout",
#'   fp_p = fp_par(text.align = "center", padding = 6),
#'   fp_t = fp_text_lite(bold = TRUE, color = "#1F4E79")
#' )
#' doc <- rtf_add(doc, "Heads up", style = "Callout")
#' print(doc, target = tempfile(fileext = ".rtf"))
rtf_set_paragraph_style <- function(
  x,
  style_id,
  style_name = style_id,
  base_on = "Normal",
  fp_p = fp_par(),
  fp_t = NULL,
  outline_level = NULL
) {
  stopifnot(is.character(style_id), length(style_id) == 1L, nzchar(style_id))
  stopifnot(is.character(style_name), length(style_name) == 1L)
  if (!is.null(base_on)) {
    if (!base_on %in% x$styles$style_id) {
      stop(
        "base_on = ", shQuote(base_on),
        " does not match any existing style_id. Existing: ",
        paste0(shQuote(x$styles$style_id), collapse = ", ")
      )
    }
  }
  outline_str <- ""
  if (!is.null(outline_level)) {
    stopifnot(
      is.numeric(outline_level), length(outline_level) == 1L,
      outline_level >= 1L, outline_level <= 9L
    )
    # RTF / OOXML outline levels are 0-indexed; Heading 1 → \outlinelevel0.
    outline_str <- sprintf("\\outlinelevel%.0f", as.integer(outline_level) - 1L)
  }
  # Place \outlinelevelN inside the paragraph properties section, before
  # run-level properties, otherwise some readers stop parsing the
  # property list. include_pard = FALSE because \pard inside a style
  # definition (or duplicated in a styled paragraph after \pard\plain)
  # terminates the style's property list in Word's parser.
  ppr <- ppr_rtf(fp_p, include_pard = FALSE)
  rpr <- if (!is.null(fp_t)) rpr_rtf(fp_t) else ""
  rtf_fragment <- paste0(ppr, outline_str, rpr)

  idx <- which(x$styles$style_id == style_id)
  if (length(idx) < 1L) {
    rtf_index <- max(x$styles$rtf_index, 0L) + 1L
    new_row <- data.frame(
      style_id = style_id,
      style_name = style_name,
      type = "paragraph",
      rtf_index = rtf_index,
      based_on = if (is.null(base_on)) NA_character_ else base_on,
      rtf = rtf_fragment,
      stringsAsFactors = FALSE
    )
    x$styles <- rbind(x$styles, new_row)
  } else {
    x$styles[idx, "style_name"] <- style_name
    x$styles[idx, "based_on"] <- if (is.null(base_on)) NA_character_ else base_on
    x$styles[idx, "rtf"] <- rtf_fragment
  }
  x
}

#' @export
#' @title Read paragraph styles defined on an RTF document
#' @description Return the data.frame of styles currently registered on
#' the document. Useful for debugging or for checking which built-in
#' styles are available before referencing them via `style = ...`.
#' @param x an rtf object created by [rtf_doc()].
#' @return a data.frame with one row per style.
#' @seealso [rtf_set_paragraph_style()], [styles_info()]
#' @examples
#' rtf_styles_info(rtf_doc())
rtf_styles_info <- function(x) {
  stopifnot(inherits(x, "rtf"))
  x$styles
}

rtf_stylesheet <- function(x) {
  styles <- x$styles
  out <- character(nrow(styles))
  for (i in seq_len(nrow(styles))) {
    based_on_str <- ""
    if (!is.na(styles$based_on[i])) {
      parent_idx <- styles$rtf_index[match(styles$based_on[i], styles$style_id)]
      if (!is.na(parent_idx)) {
        based_on_str <- sprintf("\\sbasedon%.0f", parent_idx)
      }
    }
    # RTF spec canonical order in a paragraph style definition:
    #   \sN <paragraph props> <character props> <metadata> Name;
    # \sbasedon belongs to the metadata trailer; placing it right
    # after \sN makes Word attribute the property keywords that
    # follow to the parent style. The fragment in `styles$rtf` is
    # already emitted without \pard (rtf_set_paragraph_style uses
    # ppr_rtf(include_pard = FALSE)).
    out[i] <- sprintf(
      "\n{\\s%.0f%s%s %s;}",
      styles$rtf_index[i],
      styles$rtf[i],
      based_on_str,
      styles$style_name[i]
    )
  }
  paste0(
    "{\\stylesheet ",
    paste0(out, collapse = ""),
    "}"
  )
}


# check for gregexec -----
if (!"gregexec" %in% getNamespaceExports("base")) {
  # copied from R source, grep.R
  gregexec <- function(
    pattern,
    text,
    ignore.case = FALSE,
    perl = FALSE,
    fixed = FALSE,
    useBytes = FALSE
  ) {
    if (is.factor(text) && nlevels(text) < length(text)) {
      out <- gregexec(
        pattern,
        c(levels(text), NA_character_),
        ignore.case,
        perl,
        fixed,
        useBytes
      )
      outna <- out[length(out)]
      out <- out[text]
      out[is.na(text)] <- outna
      return(out)
    }

    dat <- gregexpr(
      pattern = pattern,
      text = text,
      ignore.case = ignore.case,
      fixed = fixed,
      useBytes = useBytes,
      perl = perl
    )
    if (perl && !fixed) {
      ## Perl generates match data, so use that
      capt.attr <- c("capture.start", "capture.length", "capture.names")
      process <- function(x) {
        if (anyNA(x) || any(x < 0)) {
          y <- x
        } else {
          ## Interleave matches with captures
          y <- t(cbind(x, attr(x, "capture.start")))
          attributes(y)[names(attributes(x))] <- attributes(x)
          ml <- t(cbind(attr(x, "match.length"), attr(x, "capture.length")))
          nm <- attr(x, "capture.names")
          ## Remove empty names that `gregexpr` returns
          dimnames(ml) <- dimnames(y) <-
            if (any(nzchar(nm))) list(c("", nm), NULL)
          attr(y, "match.length") <- ml
          y
        }
        attributes(y)[capt.attr] <- NULL
        y
      }
      lapply(dat, process)
    } else {
      ## For TRE or fixed we must compute the match data ourselves
      m1 <- lapply(
        regmatches(text, dat),
        regexec,
        pattern = pattern,
        ignore.case = ignore.case,
        perl = perl,
        fixed = fixed,
        useBytes = useBytes
      )
      mlen <- lengths(m1)
      res <- vector("list", length(m1))
      im <- mlen > 0
      res[!im] <- dat[!im] # -1, NA
      res[im] <- Map(
        function(outer, inner) {
          tmp <- do.call(cbind, inner)
          attributes(tmp)[names(attributes(inner))] <- attributes(inner)
          attr(tmp, "match.length") <-
            do.call(cbind, lapply(inner, `attr`, "match.length"))
          # useBytes/index.type should be same for all so use outer vals
          attr(tmp, "useBytes") <- attr(outer, "useBytes")
          attr(tmp, "index.type") <- attr(outer, "index.type")
          tmp + rep(outer - 1L, each = nrow(tmp))
        },
        dat[im],
        m1[im]
      )
      res
    }
  }
}

test_that("rtf_doc() registers Normal and heading 1..3 with outline levels", {
  doc <- rtf_doc()
  info <- rtf_styles_info(doc)
  expect_equal(
    info$style_id,
    c("Normal", "heading 1", "heading 2", "heading 3")
  )
  expect_equal(info$rtf_index, 1:4)
  expect_equal(info$based_on, c(NA_character_, "Normal", "Normal", "Normal"))
  # Outline levels are stored 0-indexed in the rtf fragment (Heading 1 -> 0).
  expect_match(info$rtf[info$style_id == "heading 1"], "\\\\outlinelevel0")
  expect_match(info$rtf[info$style_id == "heading 2"], "\\\\outlinelevel1")
  expect_match(info$rtf[info$style_id == "heading 3"], "\\\\outlinelevel2")
})

test_that("rtf_styles_info() requires an rtf object", {
  expect_error(rtf_styles_info(list()), "inherits")
})

test_that("rtf_set_paragraph_style() inserts a new style and assigns rtf_index", {
  doc <- rtf_doc()
  before <- max(rtf_styles_info(doc)$rtf_index)
  doc <- rtf_set_paragraph_style(
    doc,
    style_id = "Callout",
    fp_p = fp_par(text.align = "center"),
    fp_t = fp_text_lite(bold = TRUE)
  )
  info <- rtf_styles_info(doc)
  row <- info[info$style_id == "Callout", , drop = FALSE]
  expect_equal(nrow(row), 1L)
  expect_equal(row$rtf_index, before + 1L)
  expect_equal(row$style_name, "Callout")
  expect_equal(row$based_on, "Normal")
})

test_that("rtf_set_paragraph_style() updates an existing style by style_id", {
  doc <- rtf_doc()
  original_index <- rtf_styles_info(doc)$rtf_index[
    rtf_styles_info(doc)$style_id == "heading 1"
  ]
  doc <- rtf_set_paragraph_style(
    doc,
    style_id = "heading 1",
    style_name = "Custom H1",
    fp_p = fp_par(text.align = "center"),
    fp_t = fp_text_lite(bold = TRUE, font.size = 30, color = "#000000"),
    outline_level = 1L
  )
  info <- rtf_styles_info(doc)
  # rtf_index is preserved on replace
  row <- info[info$style_id == "heading 1", , drop = FALSE]
  expect_equal(row$rtf_index, original_index)
  expect_equal(row$style_name, "Custom H1")
  # New properties are in the stored fragment
  expect_match(row$rtf, "\\\\fs60") # 30pt -> \fs60
})

test_that("rtf_set_paragraph_style() validates base_on against existing styles", {
  doc <- rtf_doc()
  expect_error(
    rtf_set_paragraph_style(doc, style_id = "X", base_on = "nope"),
    "does not match any existing style_id"
  )
})

test_that("rtf_set_paragraph_style() validates outline_level range", {
  doc <- rtf_doc()
  expect_error(
    rtf_set_paragraph_style(doc, "X", outline_level = 0L),
    "outline_level"
  )
  expect_error(
    rtf_set_paragraph_style(doc, "X", outline_level = 10L),
    "outline_level"
  )
})

test_that("rtf_add() with style references a registered style in the output", {
  doc <- rtf_doc()
  doc <- rtf_add(doc, "Chapter 1", style = "heading 1")
  doc <- rtf_add(doc, "Body text")
  target <- tempfile(fileext = ".rtf")
  print(doc, target = target)
  rtf <- paste(readLines(target, warn = FALSE), collapse = "\n")
  # styled paragraph emits \pard\plain\sN <duplicated style props>
  expect_match(rtf, "\\\\pard\\\\plain\\\\s2[^\n]*Chapter 1\\\\par")
  # unstyled paragraph keeps the historical emission (\pard with ppr keywords)
  expect_match(rtf, "\\\\pard\\\\sl240[^\n]*Body text")
})

test_that("rtf_add() errors on an unknown style id", {
  doc <- rtf_doc()
  expect_error(
    rtf_add(doc, "x", style = "nonexistent"),
    "is not registered"
  )
})

test_that("rtf_add(fpar, style) and rtf_add(block_list, style) carry the style", {
  doc <- rtf_doc()
  doc <- rtf_add(doc, fpar(ftext("Title")), style = "heading 2")
  doc <- rtf_add(
    doc,
    block_list(fpar(ftext("A")), fpar(ftext("B"))),
    style = "heading 3"
  )
  expect_equal(attr(doc$content[[1]], "rtf_style_index"), 3L) # heading 2 -> rtf_index 3
  expect_equal(attr(doc$content[[2]], "rtf_style_index"), 4L) # heading 3 -> rtf_index 4
  expect_equal(attr(doc$content[[3]], "rtf_style_index"), 4L)
})

test_that("walk_based_on_chain returns root-to-leaf order", {
  doc <- rtf_doc()
  parents <- setNames(doc$styles$based_on, doc$styles$style_id)
  expect_equal(
    walk_based_on_chain("heading 1", parents),
    c("Normal", "heading 1")
  )
  expect_equal(walk_based_on_chain("Normal", parents), "Normal")
})

test_that("rtf_add(block_toc()) emits a TOC field", {
  doc <- rtf_doc()
  doc <- rtf_add(doc, block_toc(level = 3))
  target <- tempfile(fileext = ".rtf")
  print(doc, target = target)
  rtf <- paste(readLines(target, warn = FALSE), collapse = "\n")
  # Field switches must be double-escaped in the RTF stream so Word's
  # parser doesn't consume them as RTF control words.
  expect_match(rtf, "\\\\fldinst TOC \\\\\\\\o \"1-3\"")
})

test_that("rtf_add(block_toc(style)) emits a style-keyed TOC field", {
  doc <- rtf_doc()
  doc <- rtf_add(doc, block_toc(style = "heading 1"))
  target <- tempfile(fileext = ".rtf")
  print(doc, target = target)
  rtf <- paste(readLines(target, warn = FALSE), collapse = "\n")
  expect_match(rtf, "\\\\fldinst TOC \\\\\\\\h \\\\\\\\z \\\\\\\\t \"heading 1\"")
})

test_that("walk_based_on_chain terminates on a cycle", {
  # Build a 3-style cycle: A -> B -> C -> A
  doc <- rtf_doc()
  doc <- rtf_set_paragraph_style(doc, "A", base_on = "Normal")
  doc <- rtf_set_paragraph_style(doc, "B", base_on = "A")
  doc <- rtf_set_paragraph_style(doc, "C", base_on = "B")
  # Force a cycle by repointing A onto C
  doc$styles$based_on[doc$styles$style_id == "A"] <- "C"
  parents <- setNames(doc$styles$based_on, doc$styles$style_id)
  chain <- walk_based_on_chain("A", parents)
  # A -> C -> B -> A (cycle): once A is hit again, recursion stops.
  # Normal is not reachable any more since A's parent was repointed.
  expect_setequal(chain, c("A", "B", "C"))
  # Each id appears exactly once: no infinite loop, no duplicates.
  expect_equal(length(chain), length(unique(chain)))
})

library(officer)

bl <- block_list(
  fpar("Comment multiple words."),
  fpar("Second line")
)

a_par <- fpar(
  "This paragraph contains",
  run_comment(
    cmt = bl,
    run = ftext("a comment."),
    author = "Author Me",
    date = "2023-06-01"
  )
)

doc <- read_docx()
doc <- body_add_fpar(doc, value = a_par, style = "Normal")

docx_file <- print(doc, target = tempfile(fileext = ".docx"))

docx_comments(read_docx(docx_file))

# tags with namespaces ----

wp_ns_yes <- "<w:p xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">"
wp_ns_no <- "<w:p>"

tbl_ns_yes <- "<w:tbl xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\">"
tbl_ns_no <- "<w:tbl>"

wr_ns_yes <- "<w:r xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">"
wr_ns_no <- "<w:r>"

ar_ns_yes <- "<a:r xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">"
ar_ns_no <- "<a:r>"

ap_ns_yes <- "<a:p xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">"
ap_ns_no <- "<a:p>"

psp_ns_yes <- "<p:sp xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">"
psp_ns_no <- "<p:sp>"

# utils -----
runs_to_p_wml <- function(..., add_ns = FALSE, style_id = NULL) {
  runs <- list(...)
  run_str <- lapply(runs, to_wml, add_ns = FALSE)
  run_str$collapse <- ""
  run_str <- do.call(paste0, run_str)
  open_tag <- wp_ns_no
  if (add_ns) {
    open_tag <- wp_ns_yes
  }

  if (!is.null(style_id)) {
    ppr <- paste0("<w:pPr><w:pStyle w:val=\"", style_id, "\"/></w:pPr>")
  } else {
    ppr <- "<w:pPr/>"
  }
  out <- paste0(open_tag, ppr, run_str, "</w:p>")
  out
}


a_xfrm_str <- function(left = 0, top = 0, width = 3, height = 3, rot = 0) {
  start_tag <- "<a:xfrm>"
  if (!is.null(rot) && !is.na(rot) && rot != 0) {
    start_tag <- sprintf("<a:xfrm rot=\"%.0f\">", -rot * 60000)
  }

  xfrm_str <- paste0(start_tag, "<a:off x=\"%.0f\" y=\"%.0f\"/><a:ext cx=\"%.0f\" cy=\"%.0f\"/></a:xfrm>")
  sprintf(
    xfrm_str,
    left * 914400, top * 914400,
    width * 914400, height * 914400
  )
}

p_xfrm_str <- function(left = 0, top = 0, width = 3, height = 3, rot = 0) {
  if (is.null(rot) || !is.finite(rot)) {
    rot <- 0
  }

  xfrm_str <- "<p:xfrm rot=\"%.0f\"><a:off x=\"%.0f\" y=\"%.0f\"/><a:ext cx=\"%.0f\" cy=\"%.0f\"/></p:xfrm>"
  sprintf(
    xfrm_str, -rot * 60000,
    left * 914400, top * 914400,
    width * 914400, height * 914400
  )
}


# colors ----

css_color <- function(color) {
  color <- as.vector(col2rgb(color, alpha = TRUE)) / c(1, 1, 1, 255)

  if (!(color[4] > 0)) {
    "transparent"
  } else {
    sprintf(
      "rgba(%.0f,%.0f,%.0f,%.2f)",
      color[1], color[2], color[3], color[4]
    )
  }
}
hex_color <- function(color) {
  color <- as.vector(col2rgb(color, alpha = TRUE)) / c(1, 1, 1, 255)
  sprintf(
    "%02X%02X%02X",
    color[1], color[2], color[3]
  )
}
colalpha <- function(x) {
  if ("transparent" %in% x) {
    return(0)
  }
  alpha <- as.vector(col2rgb(x, alpha = TRUE))[4] / 255
  alpha
}

is_transparent <- function(color) {
  !(colalpha(color) > 0)
}

solid_fill <- function(color) {
  sprintf(
    paste0(
      "<a:solidFill><a:srgbClr val=\"%s\">",
      "<a:alpha val=\"%.0f\"/>",
      "</a:srgbClr></a:solidFill>"
    ),
    hex_color(color),
    colalpha(color) * 100000
  )
}

solid_fill_pml <- function(bg) {
  bg_str <- ""
  if (!is.null(bg)) {
    bg_str <- solid_fill(bg)
  }
  bg_str
}

# geom ----

prst_geom_pml <- function(x) {
  geom_str <- ""
  if (!is.null(x)) {
    x <- check_set_geom(x)
    tagname <- paste0("a:prstGeom")
    geom_str <- sprintf("<%s prst=\"%s\"><a:avLst/></%s>", tagname, x, tagname)
  }
  geom_str
}

# line ----

ln_pml <- function(x) {
  ln_str <- ""
  if (!is.null(x)) {
    color_ <- ""
    if (is_transparent(x$color) || x$lwd < .001) {
      color_ <- "<a:noFill/>"
    } else {
      color_ <- solid_fill(x$color)
    }

    dash_ <- dash_pml(x$lty)
    join_ <- linejoin_pml(x$linejoin)
    head_ <- lineend_pml(x$headend, "head")
    tail_ <- lineend_pml(x$tailend, "tail")

    ln_str <- sprintf(
      paste0(
        "<a:ln w=\"%s\" cap=\"%s\" cmpd=\"%s\">",
        color_,
        dash_,
        join_,
        head_,
        tail_,
        "</a:ln>"
      ),
      12700 * x$lwd, x$lineend, x$linecmpd
    )
  }
  ln_str
}

lineend_pml <- function(x, side) {
  lineend_str <- ""
  if (!is.null(x)) {
    tagname <- paste0("a:", side, "End")
    lineend_str <- sprintf("<%s type=\"%s\" w=\"%s\" len=\"%s\"/>", tagname, x$type, x$width, x$length)
  }
  lineend_str
}

dash_pml <- function(x) {
  dash_str <- ""
  if (!is.null(x)) {
    dash_str <- sprintf("<a:prstDash val=\"%s\"/>", x)
  }
  dash_str
}

linejoin_pml <- function(x) {
  linejoin_str <- ""
  if (!is.null(x)) {
    linejoin_str <- sprintf("<a:%s/>", x)
  }
  linejoin_str
}

# border ----

border_pml <- function(x, side) {
  tagname <- paste0("a:ln", side)

  width_ <- sprintf("w=\"%.0f\"", x$width * 12700)

  if (is_transparent(x$color) || x$width < .001) {
    color_ <- "<a:noFill/>"
  } else {
    color_ <- solid_fill(x$color)
  }

  if (!x$style %in% c("dotted", "dashed", "solid")) {
    x$style <- "solid"
  }
  if ("dotted" %in% x$style) {
    x$style <- "sysDot"
  } else if ("dashed" %in% x$style) {
    x$style <- "sysDash"
  }

  style_ <- sprintf("<a:prstDash val=\"%s\"/>", x$style)

  paste0(
    "<", tagname,
    " ", "algn=\"ctr\" cmpd=\"sng\" cap=\"flat\"",
    " ", width_,
    ">", color_, style_,
    "</", tagname, ">"
  )
}


border_wml <- function(x, side) {
  tagname <- paste0("w:", side)
  x$style[x$style %in% "solid"] <- "single"
  x$style[x$style %in% "ridge"] <- "threeDEmboss"
  x$style[x$style %in% "groove"] <- "threeDEngrave"
  if (!x$style %in% border_styles) {
    x$style <- "single"
  }
  if (x$width < 0.0001 || is_transparent(x$color)) {
    x$style <- "none"
  }

  style_ <- sprintf("w:val=\"%s\"", x$style)
  width_ <- sprintf("w:sz=\"%.0f\"", x$width * 8)
  color_ <- sprintf("w:color=\"%s\"", hex_color(x$color))

  paste0(
    "<", tagname,
    " ", style_,
    " ", width_,
    " ", "w:space=\"0\"",
    " ", color_,
    "/>"
  )
}
border_css <- function(x, side) {
  color_ <- css_color(x$color)
  if (!(x$width > 0)) {
    color_ <- "transparent"
  }

  width_ <- sprintf("%.02fpt", x$width)

  x$style[x$style %in% "threeDEmboss"] <- "ridge"
  x$style[x$style %in% "threeDEngrave"] <- "groove"
  x$style[x$style %in% "nil"] <- "none"

  if (!x$style %in% c(
    "dotted", "dashed", "solid",
    "double", "inset", "outset",
    "ridge", "groove", "none"
  )) {
    x$style <- "solid"
  }
  paste0("border-", side, ": ", width_, " ", x$style, " ", color_, ";")
}

# ppr ----
ppr_pml <- function(x) {
  align <- " algn=\"r\""
  if (x$text.align == "left") {
    align <- " algn=\"l\""
  } else if (x$text.align == "center") {
    align <- " algn=\"ctr\""
  } else if (x$text.align == "justify") {
    align <- " algn=\"just\""
  }
  leftright_padding <- sprintf(" marL=\"%.0f\" marR=\"%.0f\"", x$padding.left * 12700, x$padding.right * 12700)
  top_padding <- sprintf("<a:spcBef><a:spcPts val=\"%.0f\" /></a:spcBef>", x$padding.top * 100)
  bottom_padding <- sprintf("<a:spcAft><a:spcPts val=\"%.0f\" /></a:spcAft>", x$padding.bottom * 100)

  line_spacing <- sprintf("<a:lnSpc><a:spcPct val=\"%.0f\"/></a:lnSpc>", x$line_spacing * 100000)

  paste0(
    "<a:pPr", align, leftright_padding, ">",
    line_spacing,
    top_padding, bottom_padding, "<a:buNone/>",
    "</a:pPr>"
  )
}

ppr_css <- function(x) {
  text.align <- sprintf("text-align:%s;", x$text.align)
  borders <- paste0(
    border_css(x$border.bottom, "bottom"),
    border_css(x$border.top, "top"),
    border_css(x$border.left, "left"),
    border_css(x$border.right, "right")
  )

  paddings <- sprintf(
    "padding-top:%.0fpt;padding-bottom:%.0fpt;padding-left:%.0fpt;padding-right:%.0fpt;",
    x$padding.top, x$padding.bottom, x$padding.left, x$padding.right
  )
  ls <- formatC(x$line_spacing, format = "f", digits = 2, decimal.mark = ".", drop0trailing = TRUE)
  line_spacing <- sprintf("line-height: %s;", ls)

  shading.color <- sprintf("background-color:%s;", css_color(x$shading.color))

  paste0(
    "margin:0pt;", text.align, borders, paddings, line_spacing,
    shading.color
  )
}

ppr_wml <- function(x) {
  if ("justify" %in% x$text.align) {
    x$text.align <- "both"
  }
  pstyle <- ""
  if (!is.null(x$word_style)) {
    word_style_id <- gsub("[^a-zA-Z0-9]", "", x$word_style)
    pstyle <- sprintf("<w:pStyle w:val=\"%s\"/>", word_style_id)
  }
  tabs <- ""
  if (!is.null(x$tabs)) {
    tabs <- to_wml(x$tabs)
  }
  text_align_ <- sprintf("<w:jc w:val=\"%s\"/>", x$text.align)
  keep_with_next <- ""
  if (x$keep_with_next) {
    keep_with_next <- "<w:keepNext/>"
  }
  borders_ <- paste0(
    "<w:pBdr>",
    border_wml(x$border.bottom, "bottom"),
    border_wml(x$border.top, "top"),
    border_wml(x$border.left, "left"),
    border_wml(x$border.right, "right"), "</w:pBdr>"
  )

  leftright_padding <- sprintf(
    "<w:ind w:left=\"%.0f\" w:right=\"%.0f\" w:firstLine=\"0\" w:firstLineChars=\"0\"/>",
    x$padding.left * 20, x$padding.right * 20
  )
  topbot_spacing <- sprintf(
    "<w:spacing w:after=\"%.0f\" w:before=\"%.0f\" w:line=\"%.0f\"/>",
    x$padding.bottom * 20, x$padding.top * 20, x$line_spacing * 240
  )
  shading_ <- ""
  if (!is_transparent(x$shading.color)) {
    shading_ <- sprintf(
      "<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"%s\"/>",
      hex_color(x$shading.color)
    )
  }

  paste0(
    "<w:pPr>",
    pstyle,
    text_align_,
    keep_with_next,
    borders_,
    tabs,
    topbot_spacing,
    leftright_padding,
    shading_,
    "</w:pPr>"
  )
}


# rpr ----

rpr_pml <- function(x) {
  if (is_transparent(x$color)) {
    return("")
  }

  out <- "<a:rPr cap=\"none\""

  if (!is.na(x$font.size)) {
    if (x$font.size > 0) {
      out <- paste0(out, sprintf(" sz=\"%.0f\"", x$font.size * 100))
    }
  }

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, " i=\"1\"")
    } else {
      out <- paste0(out, " i=\"0\"")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, " b=\"1\"")
    } else {
      out <- paste0(out, " b=\"0\"")
    }
  }

  if (!is.na(x$underlined)) {
    if (x$underlined) {
      out <- paste0(out, " u=\"sng\"")
    } else {
      out <- paste0(out, " u=\"none\"")
    }
  }

  if (x$vertical.align == "superscript") {
    out <- paste0(out, " baseline=\"40000\"")
  } else if (x$vertical.align == "subscript") {
    out <- paste0(out, " baseline=\"-40000\"")
  }
  out <- paste0(out, ">")

  if (!is.na(x$color)) {
    out <- paste0(out, solid_fill(x$color))

    if (!is_transparent(x$shading.color)) {
      shad <- sprintf(
        paste0(
          "<a:highlight><a:srgbClr val=\"%s\">",
          "<a:alpha val=\"%.0f\"/>",
          "</a:srgbClr></a:highlight>"
        ),
        hex_color(x$shading.color),
        colalpha(x$shading.color) * 100000
      )
      out <- paste0(out, shad)
    }
  }

  out <- paste0(
    out,
    if (!is.na(x$font.family)) sprintf("<a:latin typeface=\"%s\"/>", x$font.family),
    if (!is.na(x$cs.family)) sprintf("<a:cs typeface=\"%s\"/>", x$cs.family),
    if (!is.na(x$eastasia.family)) sprintf("<a:ea typeface=\"%s\"/>", x$eastasia.family),
    if (!is.na(x$hansi.family)) sprintf("<a:sym typeface=\"%s\"/>", x$hansi.family)
  )

  out <- paste0(out, "</a:rPr>")
  out
}

rpr_wml <- function(x) {
  out <- paste0(
    "<w:rPr><w:rFonts",
    if (!is.na(x$font.family)) paste0(" w:ascii=\"", x$font.family, "\""),
    if (!is.na(x$hansi.family)) paste0(" w:hAnsi=\"", x$hansi.family, "\""),
    if (!is.na(x$eastasia.family)) paste0(" w:eastAsia=\"", x$eastasia.family, "\""),
    if (!is.na(x$cs.family)) paste0(" w:cs=\"", x$cs.family, "\""),
    "/>"
  )

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, "<w:i w:val=\"true\"/>")
    } else {
      out <- paste0(out, "<w:i w:val=\"false\"/>")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, "<w:b w:val=\"true\"/>")
    } else {
      out <- paste0(out, "<w:b w:val=\"false\"/>")
    }
  }
  if (!is.na(x$underlined)) {
    if (x$underlined) {
      out <- paste0(out, "<w:u w:val=\"single\"/>")
    } else {
      out <- paste0(out, "<w:u w:val=\"none\"/>")
    }
  }

  if (x$vertical.align == "superscript") {
    out <- paste0(out, "<w:vertAlign w:val=\"superscript\"/>")
  } else if (x$vertical.align == "subscript") {
    out <- paste0(out, "<w:vertAlign w:val=\"subscript\"/>")
  }

  if (!is.na(x$font.size)) {
    out <- paste0(
      out,
      sprintf(
        "<w:sz w:val=\"%.0f\"/><w:szCs w:val=\"%.0f\"/>",
        x$font.size * 2, x$font.size * 2
      )
    )
  }

  if (!is.na(x$color)) {
    out <- paste0(
      out,
      sprintf("<w:color w:val=\"%s\"/>", hex_color(x$color))
    )
  }

  if (!is.na(x$shading.color)) {
    if (is.na(x$color)) {
      stop("shading.color must be used **with** a color.")
    }
    if (!is_transparent(x$shading.color)) {
      out <- paste0(
        out,
        sprintf("<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"%s\"/>", hex_color(x$shading.color))
      )
    }
    if (colalpha(x$color) < 1) {
      out <- paste0(
        out,
        sprintf(
          paste0(
            "<w14:textFill><w14:solidFill><w14:srgbClr val=\"%s\">",
            "<w14:alpha val=\"%.0f\"/>",
            "</w14:srgbClr></w14:solidFill></w14:textFill>"
          ),
          hex_color(x$color),
          colalpha(x$color) * 100000
        )
      )
    }
  }

  out <- paste0(out, "</w:rPr>")
  out
}

rpr_css <- function(x) {
  out <- ""

  if (!is.na(x$font.family)) out <- paste0(out, sprintf("font-family:'%s';", x$font.family))
  if (!is.na(x$color)) out <- paste0(out, sprintf("color:%s;", css_color(x$color)))
  if (!is.na(x$font.size)) out <- paste0(out, sprintf("font-size:%0.1fpt;", x$font.size))

  if (!is.na(x$italic)) {
    if (x$italic) {
      out <- paste0(out, "font-style:italic;")
    } else {
      out <- paste0(out, "font-style:normal;")
    }
  }

  if (!is.na(x$bold)) {
    if (x$bold) {
      out <- paste0(out, "font-weight:bold;")
    } else {
      out <- paste0(out, "font-weight:normal;")
    }
  }

  if (!is.na(x$bold)) {
    if (x$underlined) {
      out <- paste0(out, "text-decoration:underline;")
    } else {
      out <- paste0(out, "text-decoration:none;")
    }
  }

  if (!is.na(x$shading.color)) {
    if (!is_transparent(x$shading.color)) {
      out <- paste0(
        out,
        sprintf("background-color:%s;", css_color(x$shading.color))
      )
    } else {
      out <- paste0(out, "background-color:transparent;")
    }
  }

  if (x$vertical.align == "superscript") {
    out <- paste0(out, "vertical-align: super;")
  } else if (x$vertical.align == "subscript") {
    out <- paste0(out, "vertical-align: sub;")
  }


  out
}
# tcpr ----

tcpr_pml <- function(x) {
  text.direction <-
    if (x$text.direction %in% "btlr") {
      " vert=\"vert270\""
    } else if (x$text.direction %in% "tbrl") {
      " vert=\"vert\""
    } else {
      ""
    }

  vertical.align <-
    if (x$vertical.align %in% "center") {
      " anchor=\"ctr\""
    } else if (x$vertical.align %in% "top") {
      " anchor=\"t\""
    } else {
      " anchor=\"b\""
    }

  margins <- sprintf(
    " marB=\"%.0f\" marT=\"%.0f\" marR=\"%.0f\" marL=\"%.0f\"",
    x$margin.bottom * 12700, x$margin.top * 12700,
    x$margin.right * 12700, x$margin.left * 12700
  )

  background.color <- paste0(
    sprintf("<a:solidFill><a:srgbClr val=\"%s\">", hex_color(x$background.color)),
    sprintf("<a:alpha val=\"%.0f\"/>", colalpha(x$background.color) * 100000),
    "</a:srgbClr></a:solidFill>"
  )

  bb <- border_pml(x$border.bottom, side = "B")
  bt <- border_pml(x$border.top, side = "T")
  bl <- border_pml(x$border.left, side = "L")
  br <- border_pml(x$border.right, side = "R")

  pml_attrs <- paste0(text.direction, vertical.align, margins)
  paste0(
    "<a:tcPr", pml_attrs, ">", bl, br, bt, bb,
    background.color, "</a:tcPr>"
  )
}

tcpr_wml <- function(x) {
  background.color <- sprintf("<w:shd w:val=\"clear\" w:color=\"auto\" w:fill=\"%s\"/>", hex_color(x$background.color))
  vertical.align <- ifelse(x$vertical.align %in% c("center", "top"), sprintf("<w:vAlign w:val=\"%s\"/>", x$vertical.align), "<w:vAlign w:val=\"bottom\"/>")
  text.direction <- ifelse(
    x$text.direction %in% "btlr", "<w:textDirection w:val=\"btLr\"/>",
    ifelse(x$text.direction %in% "tbrl", "<w:textDirection w:val=\"tbRl\"/>", "")
  )

  bb <- border_wml(x$border.bottom, side = "bottom")
  bt <- border_wml(x$border.top, side = "top")
  bl <- border_wml(x$border.left, side = "left")
  br <- border_wml(x$border.right, side = "right")

  margin.bottom <- sprintf("<w:bottom w:w=\"%.0f\" w:type=\"dxa\"/>", x$margin.bottom * 20)
  margin.top <- sprintf("<w:top w:w=\"%.0f\" w:type=\"dxa\"/>", x$margin.top * 20)
  margin.left <- sprintf("<w:left w:w=\"%.0f\" w:type=\"dxa\"/>", x$margin.left * 20)
  margin.right <- sprintf("<w:right w:w=\"%.0f\" w:type=\"dxa\"/>", x$margin.right * 20)

  rowspan <- ""
  if (x$rowspan > 1) {
    rowspan <- paste0("<w:gridSpan w:val=\"", x$rowspan, "\"/>")
  }
  colspan <- ""
  if (x$colspan > 1) {
    colspan <- "<w:vMerge w:val=\"restart\"/>"
  } else if (x$colspan < 1) {
    colspan <- "<w:vMerge/>"
  }

  paste0(
    "<w:tcPr>", rowspan, colspan,
    "<w:tcBorders>", bb, bt, bl, br, "</w:tcBorders>",
    background.color,
    "<w:tcMar>", margin.top, margin.bottom, margin.left, margin.right, "</w:tcMar>",
    text.direction, vertical.align, "</w:tcPr>"
  )
}

css_px <- function(x, format = "%.0fpx") {
  ifelse(is.na(x), "inherit",
    ifelse(x < 0.001, "0", sprintf(format, x))
  )
}

tcpr_css <- function(x) {
  background.color <- ifelse(colalpha(x$background.color) > 0,
    sprintf("background-clip: padding-box;background-color:%s;", css_color(x$background.color)),
    "background-color:transparent;"
  )

  width <- ifelse(is.null(x$width) || is.na(x$width), "", sprintf("width:%s;", css_px(x$width * 72)))
  height <- ifelse(is.null(x$height) || is.na(x$height), "", sprintf("height:%s;", css_px(x$height * 72)))
  vertical.align <- ifelse(
    x$vertical.align %in% "center", "vertical-align:middle;",
    ifelse(x$vertical.align %in% "top", "vertical-align:top;", "vertical-align:bottom;")
  )

  bb <- border_css(x$border.bottom, side = "bottom")
  bt <- border_css(x$border.top, side = "top")
  bl <- border_css(x$border.left, side = "left")
  br <- border_css(x$border.right, side = "right")

  margin.bottom <- sprintf("margin-bottom:%s;", sprintf("%.0fpt", x$margin.bottom))
  margin.top <- sprintf("margin-top:%s;", sprintf("%.0fpt", x$margin.top))
  margin.left <- sprintf("margin-left:%s;", sprintf("%.0fpt", x$margin.left))
  margin.right <- sprintf("margin-right:%s;", sprintf("%.0fpt", x$margin.right))

  paste0(
    width, height, background.color, vertical.align, bb, bt, bl, br,
    margin.bottom, margin.top, margin.left, margin.right
  )
}

# openxml_document --------------------------------------------------------
# This class handle basic operations on a openxml document:
# - initialize
openxml_document <- R6Class(
  "openxml_document",
  public = list(
    initialize = function(dir) {
      private$reldir = dir
      private$rels_doc <- relationship$new()
    },

    feed = function(file) {
      private$filename <- file
      private$doc <- read_xml(file)

      private$rels_filename <- file.path(
        dirname(file),
        "_rels",
        paste0(basename(file), ".rels")
      )

      if (file.exists(private$rels_filename)) {
        private$rels_doc <- relationship$new()$feed_from_xml(
          private$rels_filename
        )
      } else {
        private$rels_doc <- relationship$new()
      }

      self
    },
    file_name = function() {
      private$filename
    },
    name = function() {
      basename(private$filename)
    },
    get = function() {
      private$doc
    },
    replace_xml = function(file) {
      # dont use `options = "NSCLEAN"` otherwise invalid document as a result
      private$doc <- read_xml(file, options = "NOBLANKS")
    },
    dir_name = function() {
      private$reldir
    },
    save = function() {
      # remove duplicate namespace definitions
      private$doc <- read_xml(as.character(private$doc), options = "NSCLEAN")
      write_xml(private$doc, file = private$filename)
      if (nrow(self$rel_df()) > 0) {
        private$rels_doc$write(private$rels_filename)
      }
      self
    },
    remove = function() {
      unlink(private$filename)
      if (file.exists(private$rels_filename)) {
        unlink(private$rels_filename)
      }
      private$filename
    },
    rel_df = function() {
      private$rels_doc$get_data()
    },
    relationship = function() {
      private$rels_doc
    }
  ),
  private = list(
    filename = NULL,
    rels_filename = NULL,
    doc = NULL,
    rels_doc = NULL,
    reldir = NULL
  )
)

href_list <- c(
  "https://github.com/davidgohel",
  "http://www.google.com?a=1&b=2",
  "http://www.google.com/folder1",
  "http://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=EDC&sectionNum=52164.1",
  "https://spec.tass.ru/avangard-na-shabolovke/?from=newsapp&utm_partner_id=149"
)
block_links <- lapply(href_list, function(link) {
  fpar(hyperlink_ftext(text = "link", href = link))
})
block_links <- do.call(block_list, block_links)

test_that("add hyperlink in docx", {
  x <- read_docx()
  x <- body_add_blocks(x, block_links)
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  expect_true(all(href_list %in% rel_df$target))
  subset_rel <- rel_df[rel_df$target %in% href_list, ]
  if (nrow(subset_rel) > 0) {
    expect_true(all(subset_rel$target %in% href_list))
    expect_true(all(subset_rel$target_mode %in% "External"))
    expect_match(subset_rel$type, "^http://schemas(.*)hyperlink$")

    body <- docx_body_xml(x)
    nodes_hyperlink <- xml_find_all(body, "//w:hyperlink")
    expect_length(nodes_hyperlink, 5)
    expect_true(all(xml_attr(nodes_hyperlink, "id") %in% subset_rel$id))
  }
})

test_that("add hyperlink in pptx", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content")
  x <- ph_with(x, block_links, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = filename)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()
  expect_true(all(href_list %in% rel_df$target))
  subset_rel <- rel_df[rel_df$target %in% href_list, ]
  if (nrow(subset_rel) > 0) {
    expect_true(all(subset_rel$target %in% href_list))
    expect_true(all(subset_rel$target_mode %in% "External"))
    expect_match(subset_rel$type, "^http://schemas(.*)hyperlink$")

    nodes_hlinkclick <- xml_find_all(
      slide$get(),
      "/p:sld/p:cSld/p:spTree/p:sp/p:txBody/a:p/a:r/a:rPr/a:hlinkClick"
    )
    expect_length(nodes_hlinkclick, 5)
    expect_true(all(xml_attr(nodes_hlinkclick, "id") %in% subset_rel$id))
  }
})

href_list <- c(
  "https://github.com/davidgohel",
  "http://www.google.com?a=1&b=2",
  "http://www.google.com/folder1",
  "http://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=EDC&sectionNum=52164.1",
  "https://spec.tass.ru/avangard-na-shabolovke/?from=newsapp&utm_partner_id=149"
)
block_links <- lapply(href_list, function(link) {
  fpar(hyperlink_ftext(text = "link", href = link))
})
block_links <- do.call(block_list, block_links)

test_that("add hyperlink in docx", {
  x <- read_docx()
  x <- body_add_blocks(x, block_links)
  filename <- print(x, target = tempfile(fileext = ".docx"))

  x <- read_docx(path = filename)
  rel_df <- x$doc_obj$rel_df()
  expect_true(all(href_list %in% rel_df$target))
  subset_rel <- rel_df[rel_df$target %in% href_list, ]
  if (nrow(subset_rel) > 0) {
    expect_true(all(subset_rel$target %in% href_list))
    expect_true(all(subset_rel$target_mode %in% "External"))
    expect_match(subset_rel$type, "^http://schemas(.*)hyperlink$")

    body <- docx_body_xml(x)
    nodes_hyperlink <- xml_find_all(body, "//w:hyperlink")
    expect_length(nodes_hyperlink, 5)
    expect_true(all(xml_attr(nodes_hyperlink, "id") %in% subset_rel$id))
  }
})

test_that("add hyperlink in pptx", {
  x <- read_pptx()
  x <- add_slide(x)
  x <- ph_with(x, block_links, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  x <- read_pptx(path = filename)
  slide <- x$slide$get_slide(x$cursor)

  rel_df <- slide$rel_df()
  expect_true(all(href_list %in% rel_df$target))
  subset_rel <- rel_df[rel_df$target %in% href_list, ]
  if (nrow(subset_rel) > 0) {
    expect_true(all(subset_rel$target %in% href_list))
    expect_true(all(subset_rel$target_mode %in% "External"))
    expect_match(subset_rel$type, "^http://schemas(.*)hyperlink$")

    nodes_hlinkclick <- xml_find_all(slide$get(), "/p:sld/p:cSld/p:spTree/p:sp/p:txBody/a:p/a:r/a:rPr/a:hlinkClick")
    expect_length(nodes_hlinkclick, 5)
    expect_true(all(xml_attr(nodes_hlinkclick, "id") %in% subset_rel$id))
  }
})

library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- add_slide(doc, "Title and Content")
doc <- add_slide(doc, "Title and Content")
doc <- on_slide(doc, index = 1)
doc <- ph_with(
  x = doc,
  "First title",
  location = ph_location_type(type = "title")
)
doc <- on_slide(doc, index = 3)
doc <- ph_with(
  x = doc,
  "Third title",
  location = ph_location_type(type = "title")
)

file <- tempfile(fileext = ".pptx")
print(doc, target = file)

#' @name officer-defunct
#' @title Defunct Functions in Package officer
#' @export
#' @rdname officer-defunct
#' @details `docx_reference_img()` is not usefull anymore.
#' @param ... unused arguments
#' @keywords internal
docx_reference_img <- function(...) {
  .Defunct(msg = "The `docx_reference_img()` function is no longer useful.")
}

#' @title Manipulate Microsoft Word and PowerPoint Documents with 'officer'
#'
#' @description
#' The officer package facilitates access to and manipulation of
#' 'Microsoft Word' and 'Microsoft PowerPoint' documents from R.
#' It also supports the writing of 'RTF' documents.
#'
#' Examples of usage are:
#'
#' * Create Word documents with tables, titles, TOC and graphics
#' * Importation of Word and PowerPoint files into data objects
#' * Write updated content back to a PowerPoint presentation
#' * Clinical reporting automation
#' * Production of reports from a shiny application
#'
#' To start with officer, read about [read_docx()], [read_pptx()]
#' or [rtf_doc()].
#'
#' The package is also providing several objects that can be printed
#' in 'R Markdown' documents for advanced Word or PowerPoint
#' reporting as [run_autonum()] and [block_caption()].
#'
#' @seealso The user documentation: \url{https://ardata-fr.github.io/officeverse/} and
#' manuals \url{https://davidgohel.github.io/officer/}
#' @docType package
#' @name officer
"_PACKAGE"

#' @export
#' @title Fast Append context to a Word document
#' @description
#' This function is used to append content to a Word document
#' in a fast way.
#'
#' It does not use the XML tree of the document
#' neither the cursor that is responsible for increasing the
#' performance of Word document generation when looping
#' over a large number of elements.
#'
#' This function must be used with the `write_elements_to_context()`
#' and `body_append_stop_context()` functions:
#'
#' 1. `body_append_start_context()` creates a context and returns
#'    a list with the context and the file connection.
#' 2. `write_elements_to_context()` writes the elements to the context
#'   file connection.
#' 3. `body_append_stop_context()` closes the file connection and
#'   replaces the XML in the document with the new XML.
#' @param x an rdocx object
#' @param additional_ns a named character vector of additional XML namespaces
#' to be added to the root node of the document. The names of the vector
#' are the namespace prefixes and the values are the namespace URIs.
#'
#' This argument is useful when the elements to be added to the document require
#' additional namespaces that are not already present in the document and not
#' part of the xml generated by `to_wml()`. Simple users are not expected
#' to use this argument. It is mainly intended for developers of officer extensions.
#' @return `body_append_start_context()` returns a list representing the context
#' that contains:
#' - `doc`: the original document object
#' - `file_con`: the file connection to the context
#' - `file_path`: the path to the context file
#' - `final_str`: the final XML string to be appended to the document
#' later when calling `body_append_stop_context()`.
#'
#' This object should not be modified by the user but instead
#' passed to `write_elements_to_context()` and `body_append_stop_context()`.
#'
#' `write_elements_to_context()` returns the context object.
#'
#' `body_append_stop_context()` returns the `rdocx` object with the
#' cursor position set to the end of the document.
#' @importFrom utils tail
#' @rdname body_append_context
#' @example inst/examples/example_body_append_context.R
#' @family functions for adding content
body_append_start_context <- function(x, additional_ns = character()) {
  if (!is.character(additional_ns)) {
    cli::cli_abort("`additional_ns` must be a named character vector.")
  }
  if (length(additional_ns) > 0 && is.null(names(additional_ns))) {
    cli::cli_abort("`additional_ns` must be a named character vector.")
  }
  if (!inherits(x, "rdocx")) {
    cli::cli_abort("`x` must be an rdocx object.")
  }

  current_xml_file <- tempfile(fileext = ".txt")
  xml_doc <- docx_body_xml(x)
  current_xml_str <- as.character(xml_doc)
  writeLines(current_xml_str, current_xml_file, useBytes = TRUE)
  current_xml_str_lines <- readLines(
    current_xml_file,
    encoding = "UTF-8",
    warn = FALSE
  )
  unlink(current_xml_file)

  section_start <- grepl("<w:sectPr", current_xml_str_lines)
  default_section_pos <- tail(which(section_start), n = 1)

  if (length(additional_ns) > 0) {
    add_ns_str <- sprintf(
      " xmlns:%s=\"%s\"",
      names(additional_ns),
      additional_ns
    )
    add_ns_str <- paste0(add_ns_str, collapse = "")
    current_xml_str_lines[2] <- gsub(
      "<w:document",
      paste0("<w:document", add_ns_str),
      current_xml_str_lines[2],
      fixed = TRUE
    )
  }

  output_file <- tempfile(fileext = ".txt")
  con <- file(output_file, open = "w", encoding = "UTF-8")
  cat(
    current_xml_str_lines[seq_len(default_section_pos - 1)],
    sep = "\n",
    file = con,
    append = TRUE
  )

  out <- list(
    doc = x,
    file_con = con,
    file_path = output_file,
    final_str = current_xml_str_lines[
      default_section_pos:length(current_xml_str_lines)
    ]
  )
  class(out) <- "rdocx_append_context"
  out
}

#' @rdname body_append_context
#' @param ... elements to be written to the context. These can be
#' paragraphs, tables, images, etc. The elements should have an associated
#' `to_wml()` method that converts them to WML format.
#' @param context the context object created by `body_append_start_context()`.
#' @export
write_elements_to_context <- function(context, ...) {
  objects <- list(...)
  for (obj in objects) {
    cat(
      to_wml(obj, add_ns = TRUE),
      sep = "\n",
      file = context$file_con,
      append = TRUE
    )
  }
  context
}

#' @rdname body_append_context
#' @export
body_append_stop_context <- function(context) {
  cat(context$final_str, sep = "\n", file = context$file_con, append = TRUE)
  close(context$file_con)
  context$doc$doc_obj$replace_xml(context$file_path)
  unlink(context$file_path)
  x <- context$doc
  x$officer_cursor <- officer_cursor(x$doc_obj$get())
  x
}

# dir_collection ---------------------------------------------------------
dir_collection <- R6Class(
  "dir_collection",
  public = list(
    initialize = function(package_dir, container) {
      dir_ <- file.path(package_dir, container$dir_name())
      private$package_dir <- package_dir
      filenames <- list.files(
        path = dir_,
        pattern = "\\.xml$",
        full.names = TRUE
      )
      filenames <- sort_vec_by_index(filenames) # see issue #596
      private$collection <- lapply(
        filenames,
        function(x, container) {
          container$clone()$feed(x)
        },
        container = container
      )
      names(private$collection) <- basename(filenames)
    },

    collection_get = function(name) {
      private$collection[[name]]
    },

    get_metadata = function() {
      dat <- lapply(private$collection, function(x) x$get_metadata())
      rbind_match_columns(dat)
    },
    names = function() {
      sapply(private$collection, function(x) x$name())
    },
    xfrm = function() {
      dat <- lapply(private$collection, function(x) x$xfrm())
      do.call(rbind, dat)
    }
  ),

  private = list(
    collection = NULL,
    package_dir = NULL
  )
)


# dir_layout ---------------------------------------------------------
dir_layout <- R6Class(
  "dir_layout",
  inherit = dir_collection,
  public = list(
    initialize = function(package_dir, master_metadata, master_xfrm) {
      super$initialize(package_dir, slide_layout$new("ppt/slideLayouts"))
      private$master_metadata <- master_metadata
      private$xfrm_data <- xfrmize(self$xfrm(), master_xfrm)
    },

    get_xfrm_data = function() {
      private$xfrm_data
    },

    get_metadata = function() {
      data_layouts <- super$get_metadata()
      data_masters <- private$master_metadata
      data_masters$master_file <- basename(data_masters$filename)
      data_masters$filename <- NULL
      data_layouts$master_file <- basename(data_layouts$master_file)
      out <- merge(data_layouts, data_masters, by = "master_file", all = FALSE)
      out$filename <- basename(out$filename)
      out
    }
  ),
  private = list(
    master_collection = NULL,
    master_metadata = NULL,
    xfrm_data = NULL
  )
)


# dir_slide ---------------------------------------------------------
dir_slide <- R6Class(
  "dir_slide",
  inherit = dir_collection,
  public = list(
    initialize = function(package_dir, xfrm_layout_data) {
      super$initialize(package_dir, slide$new("ppt/slides"))
      private$collection <- lapply(
        private$collection,
        function(x, ref) x$set_layout_xfrm(ref),
        ref = xfrm_layout_data
      )
      names(private$collection) <- sapply(private$collection, function(x) {
        x$name()
      })
      private$slides_list <- private$get_slide_list()
    },

    add_slide = function(slide_file, xfrm_layout_data) {
      slide <- slide$new("ppt/slides")
      slide$feed(slide_file)
      slide$set_layout_xfrm(xfrm_layout_data)

      collect <- private$collection
      new_elt <- list(slide)
      names(new_elt) <- basename(slide_file)
      collect <- append(collect, new_elt)

      sl_id <- as.integer(gsub(
        "(slide)([0-9]+)(\\.xml)$",
        "\\2",
        names(collect)
      ))
      private$collection <- collect[order(sl_id)]
      private$slides_list <- names(private$collection)

      self
    },
    slide_index = function(name) {
      which(private$slides_list %in% name)
    },

    remove_slide = function(index) {
      slide_obj <- private$collection[[index]]
      private$collection <- private$collection[-index]
      private$slides_list <- names(private$collection)
      slide_obj$remove()
    },

    save_slides = function() {
      lapply(private$collection, function(x) {
        x$save()
      })
      self
    },

    get_xfrm = function() {
      lapply(private$collection, function(x) x$get_xfrm())
    },

    get_slide = function(id) {
      l_ <- self$length()
      if (is.null(id) || !between(id, 1, l_)) {
        stop("unvalid id ", id, " (", l_, " slide(s))", call. = FALSE)
      }
      index <- which(names(private$collection) == private$slides_list[id])
      private$collection[[index]]
    },

    get_metadata = function() {
      super$get_metadata()
    },

    length = function() {
      length(private$collection)
    },

    get_new_slidename = function() {
      slide_dir <- file.path(private$package_dir, "ppt/slides")
      if (!file.exists(slide_dir)) {
        dir.create(
          file.path(slide_dir, "_rels"),
          showWarnings = FALSE,
          recursive = TRUE
        )
      }

      slide_files <- names(private$collection)
      slidename <- "slide1.xml"
      if (length(slide_files)) {
        slide_index <- as.integer(gsub(
          "^(slide)([0-9]+)(\\.xml)$",
          "\\2",
          slide_files
        ))
        slidename <- gsub(
          pattern = "[0-9]+",
          replacement = max(slide_index) + 1,
          slidename
        )
      }
      slidename
    }
  ),
  private = list(
    slides_list = NULL,

    get_slide_list = function() {
      slide_dir <- file.path(private$package_dir, "ppt/slides")
      slide_files <- list.files(slide_dir, pattern = "\\.xml$")
      slide_index <- seq_along(slide_files)
      if (length(slide_files)) {
        slide_files <- basename(slide_files)
        slide_index <- as.integer(gsub(
          "^(slide)([0-9]+)(\\.xml)$",
          "\\2",
          slide_files
        ))
        slide_files <- slide_files[order(slide_index)]
      }
      slide_files
    }
  )
)


# dir_master ---------------------------------------------------------

dir_master <- R6Class(
  "dir_master",
  inherit = dir_collection,
  public = list(
    get_metadata = function() {
      unames <- sapply(private$collection, function(x) x$name())
      ufnames <- sapply(private$collection, function(x) x$file_name())
      data.frame(
        stringsAsFactors = FALSE,
        master_name = unames,
        filename = ufnames
      )
    },
    get_color_scheme = function() {
      dat <- lapply(private$collection, function(x) x$colors())
      rbind_match_columns(dat)
    }
  )
)

# dir_notesMaster ---------------------------------------------------------

dir_notesMaster <- R6Class(
  "dir_notesMaster",
  inherit = dir_collection,
  public = list(
    get_metadata = function() {
      unames <- sapply(private$collection, function(x) x$name())
      ufnames <- sapply(private$collection, function(x) x$file_name())
      data.frame(
        stringsAsFactors = FALSE,
        master_name = unames,
        filename = ufnames
      )
    },

    add_notesMaster = function(file) {
      nmaster <- slide_master$new("ppt/notesMasters")
      nmaster$feed(file)

      collect <- private$collection
      new_elt <- list(nmaster)
      names(new_elt) <- basename(file)
      collect <- append(collect, new_elt)

      sl_id <- as.integer(gsub(
        "(notesMaster)([0-9]+)(\\.xml)$",
        "\\2",
        names(collect)
      ))
      private$collection <- collect[order(sl_id)]

      self
    }
  )
)


# dir_notesSlide ---------------------------------------------------------
dir_notesSlide <- R6Class(
  "dir_notesSlide",
  inherit = dir_collection,
  public = list(
    initialize = function(package_dir) {
      super$initialize(package_dir, notesSlide$new("ppt/notesSlides"))
    },

    add_slide = function(slide_file) {
      nslide <- notesSlide$new("ppt/notesSlides")
      nslide$feed(slide_file)

      collect <- private$collection
      new_elt <- list(nslide)
      names(new_elt) <- basename(slide_file)
      collect <- append(collect, new_elt)

      sl_id <- as.integer(gsub(
        "(notesSlide)([0-9]+)(\\.xml)$",
        "\\2",
        names(collect)
      ))
      private$collection <- collect[order(sl_id)]

      self
    },
    slide_index = function(name) {
      which(names(private$collection) %in% name)
    },

    remove_slide = function(index) {
      slide_obj <- private$collection[[index]]
      private$collection <- private$collection[-index]
      slide_obj$remove()
    },

    save_slides = function() {
      lapply(private$collection, function(x) {
        x$save()
      })
      self
    },

    get_slide = function(id) {
      l_ <- self$length()
      if (is.null(id) || !between(id, 1, l_)) {
        stop("unvalid id ", id, " (", l_, " slide(s))", call. = FALSE)
      }
      private$collection[[id]]
    },

    get_metadata = function() {
      super$get_metadata()
    },

    length = function() {
      length(private$collection)
    },

    get_new_slidename = function() {
      slide_dir <- file.path(private$package_dir, "ppt/notesSlides")
      if (!file.exists(slide_dir)) {
        dir.create(
          file.path(slide_dir, "_rels"),
          showWarnings = FALSE,
          recursive = TRUE
        )
      }

      slide_files <- names(private$collection)
      slidename <- "notesSlide1.xml"
      if (length(slide_files)) {
        slide_index <- as.integer(gsub(
          "^(notesSlide)([0-9]+)(\\.xml)$",
          "\\2",
          slide_files
        ))
        slidename <- gsub(
          pattern = "[0-9]+",
          replacement = max(slide_index) + 1,
          slidename
        )
      }
      slidename
    }
  )
)

test_that("defaul template", {
  x <- read_pptx()
  expect_equal(length(x), 0)
  expect_true(file.exists(x$package_dir))
})


test_that("console printing", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  layouts <- layout_summary(x)$layout

  # 1 slide
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  output_print <- capture.output({
    # capture layout dataframe
    output_cli <- cli::cli_fmt(print(x)) # capture cli info lines
  })
  expect_true(any(grepl("pptx document with 1 slide", output_cli)))
  output_shows_layout <- vapply(
    layouts,
    function(l) any(grepl(l, output_print)),
    logical(1)
  )
  expect_true(all(output_shows_layout))

  # 2 slides
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  output_print <- capture.output({
    output_cli <- cli::cli_fmt(print(x))
  })
  expect_true(any(grepl("pptx document with 2 slides", output_cli)))

  # 2 slides + default layout
  x <- layout_default(x, "Title and Content")
  output_print <- capture.output({
    output_cli <- cli::cli_fmt(print(x))
  })
  expect_true(any(grepl("(*) = Default layout", output_cli)))
  output_shows_layout <- vapply(
    layouts,
    function(l) any(grepl(l, output_print)),
    logical(1)
  )
  expect_true(all(output_shows_layout))

  # default layout removed again
  x <- layout_default(x, NULL)
  output_print <- capture.output({
    output_cli <- cli::cli_fmt(print(x))
  })
  expect_false(any(grepl("(*) = Default layout", output_cli)))
})


test_that("check extention and print document", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  print(x, target = "print.pptx")
  expect_true(file.exists("print.pptx"))

  expect_error(print(x, target = "print.pptxxx"))
})


test_that("check template", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  print(x, target = "template.pptx")

  expect_silent(x <- read_pptx(path = "template.pptx"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(
    x,
    "Hello world again",
    location = ph_location_type(type = "body")
  )
  print(x, target = "example.pptx")

  expect_equal(length(x), 2)
})


test_that("slide remove", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))
  print(x, target = "template.pptx")

  x <- read_pptx(path = "template.pptx")
  x <- remove_slide(x = x)
  expect_equal(length(x), 2)
  x <- remove_slide(x = x, index = 1)
  expect_equal(length(x), 1)

  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 2")
})

test_that("remove multiple slides at once", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 3", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 4", location = ph_location_type(type = "body"))

  # Remove slides 2 and 4
  x <- remove_slide(x, index = c(2, 4))
  expect_equal(length(x), 2)

  # Check remaining slides are 1 and 3
  sm1 <- slide_summary(x, index = 1)
  sm2 <- slide_summary(x, index = 2)
  expect_equal(sm1[1, ]$text, "Slide 1")
  expect_equal(sm2[1, ]$text, "Slide 3")
})

test_that("remove multiple slides - edge cases", {
  # Test removing all slides
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- remove_slide(x, index = c(1, 2))
  expect_equal(length(x), 0)

  # Test with duplicate indices (should only remove once)
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 3", location = ph_location_type(type = "body"))
  x <- remove_slide(x, index = c(2, 2, 2))
  expect_equal(length(x), 2)
  sm <- slide_summary(x, index = 2)
  expect_equal(sm[1, ]$text, "Slide 3")

  # Test with empty index vector (should return unchanged)
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- remove_slide(x, index = integer(0))
  expect_equal(length(x), 2)

  # Test removing slides in non-sequential order
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 3", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Slide 4", location = ph_location_type(type = "body"))
  x <- remove_slide(x, index = c(4, 1, 3))
  expect_equal(length(x), 1)
  sm <- slide_summary(x, index = 1)
  expect_equal(sm[1, ]$text, "Slide 2")
})

test_that("remove multiple slides - invalid indices", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- add_slide(x, "Title and Content", "Office Theme")

  # Test with out of range indices
  expect_error(remove_slide(x, index = c(1, 5)), "out of range")
  expect_error(remove_slide(x, index = c(0, 1)), "out of range")
  expect_error(remove_slide(x, index = c(-1, 2)), "out of range")

  # Test with all invalid indices
  expect_error(remove_slide(x, index = c(5, 6)), "out of range")
})

test_that("ph remove", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))
  print(x, target = "template.pptx")

  x <- read_pptx(path = "template.pptx")
  x <- ph_remove(x = x)

  expect_equal(nrow(slide_summary(x)), 0)
})


test_that("cursor is incremented as expected", {
  x <- read_pptx()
  for (i in 1:11) {
    x <- add_slide(x, "Title Slide", "Office Theme")
    x <- ph_with(x, i, location = ph_location_type(type = "ctrTitle"))
  }
  expect_equal(nrow(slide_summary(x, 11)), 1)
  expect_equal(x$slide$get_slide(11)$name(), "slide11.xml")
})

test_that("annotate base template", {
  expect_s3_class(try(annotate_base(), silent = TRUE), "rpptx")
})

test_that("no master do not generate an error", {
  x <- read_pptx("docs_dir/no_master.pptx")
  x <- add_slide(x, layout = "Page One", master = "Office Theme")
  x <- try(
    ph_with(
      x,
      i,
      value = "graphic title",
      location = ph_location_type(type = "body")
    ),
    silent = TRUE
  )
  expect_s3_class(x, "rpptx")
})


unlink("*.pptx")
unlink("*.emf")


test_that("slide_visible", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()
  expect_equal(slide_visible(x), logical(0)) # works with 0 slides

  path <- testthat::test_path("docs_dir", "test-slides-visible.pptx")
  x <- read_pptx(path)

  expect_equal(slide_visible(x), c(FALSE, TRUE, FALSE))
  x <- slide_visible(x, hide = 1:2)
  expect_s3_class(x, "rpptx")
  expect_equal(slide_visible(x), c(FALSE, FALSE, FALSE))
  x <- slide_visible(x, show = 1:2)
  expect_s3_class(x, "rpptx")
  expect_equal(slide_visible(x), c(TRUE, TRUE, FALSE))
  x <- slide_visible(x, hide = 1:2, show = 3)
  expect_s3_class(x, "rpptx")
  expect_equal(slide_visible(x), c(FALSE, FALSE, TRUE))

  expect_error(
    regex = "Overlap between indexes in `hide` and `show`",
    slide_visible(x, hide = 1:2, show = 1:2)
  )
  expect_error(
    regex = "2 indexes of `hide` outside slide range",
    slide_visible(x, hide = 1:5)
  )
  expect_error(
    regex = "1 index of `show` outside slide range",
    slide_visible(x, show = -1)
  )

  slide_visible(x) <- FALSE # hide all slides
  expect_false(any(slide_visible(x)))
  slide_visible(x) <- c(TRUE, FALSE, TRUE)
  expect_equal(slide_visible(x), c(TRUE, FALSE, TRUE))
  expect_warning(
    regexp = "Value is not length 1 or same length as number of slides",
    slide_visible(x) <- c(TRUE, FALSE) # warns that rhs values are recycled
  )
  expect_equal(slide_visible(x), c(TRUE, FALSE, TRUE))

  slide_visible(x)[2] <- TRUE
  expect_equal(slide_visible(x), c(TRUE, TRUE, TRUE))
  slide_visible(x)[c(1, 3)] <- FALSE
  expect_equal(slide_visible(x), c(FALSE, TRUE, FALSE))

  slide_visible(x) <- TRUE
  expect_warning(
    regexp = "number of items to replace is not a multiple of replacement length",
    slide_visible(x)[c(1, 2)] <- rep(FALSE, 4)
  )
  expect_equal(slide_visible(x), c(FALSE, FALSE, TRUE))

  expect_error(
    {
      slide_visible(x) <- rep(FALSE, 4)
    },
    regexp = "More values \\(4\\) than slides \\(3\\)"
  )

  # test that changes are written to file
  path <- testthat::test_path("docs_dir", "test-slides-visible.pptx")
  x <- read_pptx(path)

  slide_visible(x) <- FALSE
  path <- tempfile(fileext = ".pptx")
  print(x, path)
  x <- read_pptx(path)
  expect_equal(slide_visible(x), c(FALSE, FALSE, FALSE))

  slide_visible(x)[c(1, 3)] <- TRUE
  path <- tempfile(fileext = ".pptx")
  print(x, path)
  x <- read_pptx(path)
  expect_equal(slide_visible(x), c(TRUE, FALSE, TRUE))
})

test_that("add alt text to ggplot", {
  skip_if_not_installed("ggplot2")
  require(ggplot2)

  alt_text <- "Alt text added with 'alt_text='."

  gg <- ggplot(mtcars, aes(factor(cyl))) +
    geom_bar()

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(
    doc,
    value = gg,
    location = ph_location_type("body"),
    alt_text = alt_text
  )

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//p:nvPicPr//p:cNvPr"), "descr"),
    alt_text
  )

  alt_labs <- "Alt text added with 'ggplot2::labs(alt=)'."
  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(
    doc,
    value = gg + labs(alt = alt_labs),
    location = ph_location_type("body")
  )
  xmldoc <- doc$slide$get_slide(id = 2)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//p:nvPicPr//p:cNvPr"), "descr"),
    alt_labs
  )

  doc <- add_slide(doc, "Title and Content")
  doc <- ph_with(doc, value = gg, location = ph_location_type("body"))
  xmldoc <- doc$slide$get_slide(id = 3)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//p:nvPicPr//p:cNvPr"), "descr"),
    ""
  )
})

test_that("defaul template", {
  x <- read_pptx()
  expect_equal(length( x ), 0)
  expect_true(file.exists(x$package_dir))
})

test_that("console printing", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  expect_output(print(x), "pptx document with 1 ")
})

test_that("check extention and print document", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  print(x, target = "print.pptx")
  expect_true( file.exists("print.pptx") )

  expect_error(print(x, target = "print.pptxxx"))
})


test_that("check template", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  print(x, target = "template.pptx")

  expect_silent(x <- read_pptx(path = "template.pptx"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world again", location = ph_location_type(type = "body"))
  print(x, target = "example.pptx")

  expect_equal(length(x), 2)
})



test_that("slide remove", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))
  print(x, target = "template.pptx")

  x <- read_pptx(path = "template.pptx")
  x <- remove_slide(x = x)
  expect_equal(length(x), 2)
  x <- remove_slide(x = x, index = 1)
  expect_equal(length(x), 1)

  sm <- slide_summary(x)
  expect_equal(sm[1,]$text, "Hello world 2")
})

test_that("slide remove with rm_images", {
  img.file <- file.path(R.home(component = "doc"), "html", "logo.jpg")
  ext_img <- external_img(img.file)

  x <- read_pptx()
  x <- add_slide(x)
  x <- ph_with(x, ext_img, location = ph_location_type())
  filename <- print(x, target = tempfile(fileext = ".pptx"))

  z <- read_pptx(filename)
  z <- remove_slide(z, index = 1)
  file1 <- print(z, target = tempfile(fileext = ".pptx"))
  z <- read_pptx(filename)
  z <- remove_slide(z, index = 1, rm_images = TRUE)
  file2 <- print(z, target = tempfile(fileext = ".pptx"))

  expect_gt(file.size(file1), file.size(file2))
})

test_that("ph remove", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))
  print(x, target = "template.pptx")

  x <- read_pptx(path = "template.pptx")
  x <- ph_remove(x = x)

  expect_equal(nrow(slide_summary(x)), 0)
})


test_that("cursor is incremented as expected", {
  x <- read_pptx()
  for(i in 1:11){
    x <- add_slide(x, "Title Slide", "Office Theme")
    x <- ph_with(x, i, location = ph_location_type(type = "ctrTitle"))
  }
  expect_equal(nrow( slide_summary(x, 11) ), 1 )
  expect_equal(x$slide$get_slide(11)$name(), "slide11.xml" )

})

test_that("annotate base template", {
  expect_s3_class(try(annotate_base(), silent = TRUE), "rpptx")
})

test_that("no master do not generate an error", {
  x <- read_pptx("docs_dir/no_master.pptx")
  x <- add_slide(x, layout = "Page One", master = "Office Theme")
  x <- try(ph_with(x, i, value = "graphic title",
               location = ph_location_type(type="body")), silent = TRUE)
  expect_s3_class(x, "rpptx")
})



unlink("*.pptx")
unlink("*.emf")

test_that("add notesMaster", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  add_notesMaster(doc)

  doc$content_type$save()
  content_type <- read_xml(file.path(doc$package_dir, "[Content_Types].xml"))
  nodeset <- xml_find_all(content_type, "//d1:Override")
  ct_df <- data.frame(PartName = xml_attr(nodeset, "PartName"), ContentType = xml_attr(nodeset, "ContentType"))
  notesMasterFile <- ct_df$PartName[ct_df$ContentType == "application/vnd.openxmlformats-officedocument.presentationml.notesMaster+xml"]
  expect_length(notesMasterFile, 1)
  expect_true(file.exists(file.path(doc$package_dir, notesMasterFile)))
  expect_equal(basename(notesMasterFile), names(doc$notesMaster$names()))

  nslideMaster <- doc$notesMaster$collection_get(basename(notesMasterFile))
  rels <- nslideMaster$rel_df()
  theme_file <- rels$target[rels$type == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme"]
  expect_length(theme_file, 1)
  expect_true(file.exists(file.path(doc$package_dir, dirname(notesMasterFile), theme_file)))
  expect_true(any(grepl(basename(theme_file), ct_df$PartName[ct_df$ContentType == "application/vnd.openxmlformats-officedocument.theme+xml"])))

  xml <- doc$presentation$get()
  node <- xml_find_all(xml, "//p:notesMasterIdLst/p:notesMasterId")
  expect_length(node, 1)
  rid <- xml_attr(node, "id")
  pr_rel <- doc$presentation$rel_df()
  pr_rel <- pr_rel[pr_rel$id == rid, ]
  expect_equal(pr_rel$type, "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster")
  expect_equal(file.path("/ppt", pr_rel$target), notesMasterFile)
})


test_that("add notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  get_or_create_notesSlide(doc)

  doc$content_type$save()
  content_type <- read_xml(file.path(doc$package_dir, "[Content_Types].xml"))
  nodeset <- xml_find_all(content_type, "//d1:Override")
  ct_df <- data.frame(PartName = xml_attr(nodeset, "PartName"), ContentType = xml_attr(nodeset, "ContentType"))
  notesSlideFile <- ct_df$PartName[ct_df$ContentType == "application/vnd.openxmlformats-officedocument.presentationml.notesSlide+xml"]
  expect_length(notesSlideFile, 1)
  expect_true(file.exists(file.path(doc$package_dir, notesSlideFile)))

  current_slide <- doc$slide$get_slide(doc$cursor)
  rel_df <- current_slide$rel_df()
  notes_ref <- rel_df$target[rel_df$type == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesSlide"]
  expect_length(notes_ref, 1)
  expect_equal(basename(notes_ref), basename(notesSlideFile))

  idx <- doc$notesSlide$slide_index(basename(notesSlideFile))
  nslide <- doc$notesSlide$get_slide(idx)
  rel_df <- nslide$rel_df()
  master_rel <- rel_df$target[rel_df$type == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster"]
  expect_equal(basename(master_rel), names(doc$notesMaster$names()))
  slide_rel <- rel_df$target[rel_df$type == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide"]
  expect_equal(basename(slide_rel), current_slide$name())
})

test_that("add notes to notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, "I am a test!", notes_location_label("Notes Placeholder 4"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:cNvPr/@name='Notes Placeholder 4']"
  node_ <- xml_find_all(xml, xpath_)
  expect_false(inherits(node_, "xml_missing"))
  expect_equal(xml_text(node_), "I am a test!")
})

test_that("add notes to notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, "I am a test!", notes_location_type("body"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:nvPr/p:ph/@type='body']"
  node_ <- xml_find_all(xml, xpath_)
  expect_false(inherits(node_, "xml_missing"))
  expect_equal(xml_text(node_), "I am a test!")
})
test_that("add block_list to notesSlide", {
  value <- block_list(
    fpar(ftext("hello world")),
    fpar(
      ftext("hello"), " ",
      ftext("world")
    ),
    fpar(
      ftext("blah blah blah")
    )
  )

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, value, notes_location_type("body"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:nvPr/p:ph/@type='body']/p:txBody/a:p"
  nodes <- xml_find_all(xml, xpath_)
  expect_true(length(nodes) > 0)
  expect_equal(xml_text(nodes), c("hello world", "hello world", "blah blah blah"))
})

test_that("replace notes on notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, "I am a test!", notes_location_type("body"))
  doc <- set_notes(doc, "I am a new test!", notes_location_type("body"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:nvPr/p:ph/@type='body']"
  node_ <- xml_find_all(xml, xpath_)
  expect_false(inherits(node_, "xml_missing"))
  expect_equal(xml_text(node_), "I am a new test!")
})

test_that("add alt text to ggplot", {
  skip_if_not_installed("ggplot2")
  require(ggplot2)

  alt_text <- "Alt text added with 'alt_text='."

  gg <- ggplot(mtcars, aes(factor(cyl))) +
    geom_bar()

  doc <- read_pptx()
  doc <- add_slide(doc)
  doc <- ph_with(doc,
    value = gg,
    location = ph_location_type("body"),
    alt_text = alt_text
  )

  xmldoc <- doc$slide$get_slide(id = 1)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//p:nvPicPr//p:cNvPr"), "descr"),
    alt_text
  )

  alt_labs <- "Alt text added with 'ggplot2::labs(alt=)'."
  doc <- add_slide(doc)
  doc <- ph_with(doc,
    value = gg  + labs(alt = alt_labs),
    location = ph_location_type("body")
  )
  xmldoc <- doc$slide$get_slide(id = 2)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//p:nvPicPr//p:cNvPr"), "descr"),
    alt_labs
  )

  doc <- add_slide(doc)
  doc <- ph_with(doc,
                 value = gg,
                 location = ph_location_type("body")
  )
  xmldoc <- doc$slide$get_slide(id = 3)$get()
  expect_equal(
    xml_attr(xml_find_all(xmldoc, "//p:nvPicPr//p:cNvPr"), "descr"),
    ""
  )
})

test_that("ph_with_table", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", master = "Office Theme")
  x <- ph_with(x, value = iris, location = ph_location_type(type = "body"))
  xmldoc <- x$slide$get_slide(1)$get()
  xml_tab <- xml_find_all(
    xmldoc,
    xpath = "p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl"
  )
  expect_false(inherits(xml_tab, "xml_missing"))
})

test_that("pml structure", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", master = "Office Theme")
  x <- ph_with(x, value = mtcars, location = ph_location_type(type = "body"))
  xmldoc <- x$slide$get_slide(1)$get()

  xml_tr <- xml_find_all(
    xmldoc,
    xpath = "p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl/a:tr"
  )
  expect_length(xml_tr, nrow(mtcars) + 1)
  expect_equal(length(unique(xml_attr(xml_tr, "h"))), 1)
  expect_true(all(grepl("^[0-9]+$", xml_attr(xml_tr, "h"))))

  xml_tr <- xml_find_all(
    xmldoc,
    xpath = "p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl/a:tr/a:tc"
  )
  expect_length(xml_tr, (nrow(mtcars) + 1) * ncol(mtcars))
})

test_that("add notesMaster", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  add_notesMaster(doc)

  doc$content_type$save()
  content_type <- read_xml(file.path(doc$package_dir, "[Content_Types].xml"))
  nodeset <- xml_find_all(content_type, "//d1:Override")
  ct_df <- data.frame(
    PartName = xml_attr(nodeset, "PartName"),
    ContentType = xml_attr(nodeset, "ContentType")
  )
  notesMasterFile <- ct_df$PartName[
    ct_df$ContentType ==
      "application/vnd.openxmlformats-officedocument.presentationml.notesMaster+xml"
  ]
  expect_length(notesMasterFile, 1)
  expect_true(file.exists(file.path(doc$package_dir, notesMasterFile)))
  expect_equal(basename(notesMasterFile), names(doc$notesMaster$names()))

  nslideMaster <- doc$notesMaster$collection_get(basename(notesMasterFile))
  rels <- nslideMaster$rel_df()
  theme_file <- rels$target[
    rels$type ==
      "http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme"
  ]
  expect_length(theme_file, 1)
  expect_true(file.exists(file.path(
    doc$package_dir,
    dirname(notesMasterFile),
    theme_file
  )))
  expect_true(any(grepl(
    basename(theme_file),
    ct_df$PartName[
      ct_df$ContentType ==
        "application/vnd.openxmlformats-officedocument.theme+xml"
    ]
  )))

  xml <- doc$presentation$get()
  node <- xml_find_all(xml, "//p:notesMasterIdLst/p:notesMasterId")
  expect_length(node, 1)
  rid <- xml_attr(node, "id")
  pr_rel <- doc$presentation$rel_df()
  pr_rel <- pr_rel[pr_rel$id == rid, ]
  expect_equal(
    pr_rel$type,
    "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster"
  )
  expect_equal(file.path("/ppt", pr_rel$target), notesMasterFile)
})


test_that("add notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  get_or_create_notesSlide(doc)

  doc$content_type$save()
  content_type <- read_xml(file.path(doc$package_dir, "[Content_Types].xml"))
  nodeset <- xml_find_all(content_type, "//d1:Override")
  ct_df <- data.frame(
    PartName = xml_attr(nodeset, "PartName"),
    ContentType = xml_attr(nodeset, "ContentType")
  )
  notesSlideFile <- ct_df$PartName[
    ct_df$ContentType ==
      "application/vnd.openxmlformats-officedocument.presentationml.notesSlide+xml"
  ]
  expect_length(notesSlideFile, 1)
  expect_true(file.exists(file.path(doc$package_dir, notesSlideFile)))

  current_slide <- doc$slide$get_slide(doc$cursor)
  rel_df <- current_slide$rel_df()
  notes_ref <- rel_df$target[
    rel_df$type ==
      "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesSlide"
  ]
  expect_length(notes_ref, 1)
  expect_equal(basename(notes_ref), basename(notesSlideFile))

  idx <- doc$notesSlide$slide_index(basename(notesSlideFile))
  nslide <- doc$notesSlide$get_slide(idx)
  rel_df <- nslide$rel_df()
  master_rel <- rel_df$target[
    rel_df$type ==
      "http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesMaster"
  ]
  expect_equal(basename(master_rel), names(doc$notesMaster$names()))
  slide_rel <- rel_df$target[
    rel_df$type ==
      "http://schemas.openxmlformats.org/officeDocument/2006/relationships/slide"
  ]
  expect_equal(basename(slide_rel), current_slide$name())
})

test_that("add notes to notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(
    doc,
    "I am a test!",
    notes_location_label("Notes Placeholder 4")
  )

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:cNvPr/@name='Notes Placeholder 4']"
  node_ <- xml_find_all(xml, xpath_)
  expect_false(inherits(node_, "xml_missing"))
  expect_equal(xml_text(node_), "I am a test!")
})

test_that("add notes to notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, "I am a test!", notes_location_type("body"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:nvPr/p:ph/@type='body']"
  node_ <- xml_find_all(xml, xpath_)
  expect_false(inherits(node_, "xml_missing"))
  expect_equal(xml_text(node_), "I am a test!")
})
test_that("add block_list to notesSlide", {
  value <- block_list(
    fpar(ftext("hello world")),
    fpar(
      ftext("hello"),
      " ",
      ftext("world")
    ),
    fpar(
      ftext("blah blah blah")
    )
  )

  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, value, notes_location_type("body"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:nvPr/p:ph/@type='body']/p:txBody/a:p"
  nodes <- xml_find_all(xml, xpath_)
  expect_true(length(nodes) > 0)
  expect_equal(
    xml_text(nodes),
    c("hello world", "hello world", "blah blah blah")
  )
})

test_that("replace notes on notesSlide", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- set_notes(doc, "I am a test!", notes_location_type("body"))
  doc <- set_notes(doc, "I am a new test!", notes_location_type("body"))

  nslide <- doc$notesSlide$get_slide(1)
  xml <- nslide$get()
  xpath_ <- "//p:sp[p:nvSpPr/p:nvPr/p:ph/@type='body']"
  node_ <- xml_find_all(xml, xpath_)
  expect_false(inherits(node_, "xml_missing"))
  expect_equal(xml_text(node_), "I am a new test!")
})

test_that("ph_with_table", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", master = "Office Theme")
  x <- ph_with(x, value = iris, location = ph_location_type(type = "body"))
  xmldoc <- x$slide$get_slide(1)$get()
  xml_tab <- xml_find_all(xmldoc, xpath = "p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl")
  expect_false( inherits(xml_tab, "xml_missing") )
})

test_that("pml structure", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", master = "Office Theme")
  x <- ph_with(x, value = mtcars, location = ph_location_type(type = "body"))
  xmldoc <- x$slide$get_slide(1)$get()

  xml_tr <- xml_find_all(xmldoc, xpath = "p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl/a:tr")
  expect_length(xml_tr, nrow(mtcars) + 1)
  expect_equal(length(unique(xml_attr(xml_tr, "h")) ), 1)
  expect_true( all( grepl( "^[0-9]+$", xml_attr(xml_tr, "h") ) ) )

  xml_tr <- xml_find_all(xmldoc, xpath = "p:cSld/p:spTree/p:graphicFrame/a:graphic/a:graphicData/a:tbl/a:tr/a:tc")
  expect_length(xml_tr, (nrow(mtcars) + 1) * ncol(mtcars))

})

#' @importFrom cli
#'  cli_abort
#' @importFrom dplyr
#'  mutate case_when left_join arrange across all_of lag consecutive_id syms n
#'  select inner_join if_else filter relocate summarise group_by as_tibble first
#'  bind_rows
#' @importFrom tidyr
#'  complete
#' @importFrom xml2
#'  as_xml_document xml_has_attr xml_child xml_name xml_children xml_replace

# utils ----------
extract_runs_attr_str <- function(
  run_nodes,
  node_xpath,
  attr_name,
  default = NULL
) {
  nodes <- suppressWarnings({
    xml_child(run_nodes, node_xpath)
  })
  nodes_exist <- !is.na(xml_name(nodes))

  str_values <- rep("", length(run_nodes))
  tmp_results <- xml_attr(nodes, attr_name)
  str_values[nodes_exist] <- tmp_results[nodes_exist]
  if (!is.null(default)) {
    str_values[is.na(str_values)] <- default
  }

  str_values[str_values %in% ""] <- NA_character_
  str_values
}

extract_runs_attr_lgl <- function(
  run_nodes,
  node_xpath,
  attr_name,
  default = NULL,
  true_values = c("1", "on", "true", NA_character_)
) {
  nodes <- xml_child(run_nodes, node_xpath)
  nodes_exist <- !is.na(xml_name(nodes))

  lgl_values <- rep(default, length(run_nodes))
  lgl_values[nodes_exist] <- xml_attr(nodes[nodes_exist], attr_name) %in%
    true_values
  lgl_values
}

add_text_column <- function(x) {
  mutate(
    .data = x,
    text = case_when(
      !is.na(.data$run_content_text) ~ .data$run_content_text,
      .default = ""
    )
  )
}

#' @title Compute tables information
#' @description Joins and processes table-related information data frames
#'   to compute cell spans and merges.
#' @param tmp_infotbl_tablecells_paragraphs data.frame with paragraph
#'   information in table cells
#' @param tmp_infotbl_tablecells data.frame with table cells information
#' @param tmp_infotbl_tablerows data.frame with table rows information
#' @param tmp_infotbl_tables data.frame with tables information
#' @return A data.frame with complete table information including computed
#'   cell_id, row_span, and col_span values.
#' @noRd
infotbl_tables_compute <- function(
  tmp_infotbl_tablecells_paragraphs,
  tmp_infotbl_tablecells,
  tmp_infotbl_tablerows,
  tmp_infotbl_tables
) {
  infotbl_tables <- left_join(
    tmp_infotbl_tablecells_paragraphs,
    tmp_infotbl_tablecells,
    by = "cell_id"
  )
  infotbl_tables <- left_join(
    infotbl_tables,
    tmp_infotbl_tablerows,
    by = c("table_index", "row_id")
  )
  infotbl_tables <- left_join(
    infotbl_tables,
    tmp_infotbl_tables,
    by = c("table_index")
  )
  infotbl_tables <- arrange(
    .data = infotbl_tables,
    across(all_of(
      c("table_index", "row_id", "cell_id", "doc_index")
    ))
  )
  infotbl_tables <- mutate(
    .data = infotbl_tables,
    .by = all_of(c("table_index", "row_id")),
    add_span = as.integer(
      lag(.data$col_span, default = "1")
    ) -
      1L,
    cell_id = consecutive_id(.data$cell_id) + cumsum(.data$add_span),
    add_span = NULL
  )
  infotbl_tables <- tidyr::complete(
    infotbl_tables,
    !!!syms(c("table_index", "row_id", "cell_id")),
    fill = list(
      row_merge = FALSE,
      first = FALSE,
      col_span = "0",
      is_header = FALSE
    )
  )

  infotbl_tables <- mutate(
    .data = infotbl_tables,
    .by = all_of(c("table_index", "cell_id")),
    merge_group = cumsum(.data$first) + cumsum(!.data$row_merge)
  )

  infotbl_tables <- mutate(
    .data = infotbl_tables,
    .by = all_of(c("table_index", "cell_id", "merge_group")),
    row_span = n(),
    .after = all_of("col_span")
  )
  infotbl_tables <- mutate(
    .data = infotbl_tables,
    merge_group = NULL,
    row_span = case_when(
      .data$row_merge & !.data$first ~ 0L,
      .default = .data$row_span
    )
  )

  infotbl_tables
}

#' @title Augment data with non-text content
#' @description Adds information about images, footnotes, and hyperlinks
#'   to the data by joining with document relationships and processing
#'   footnote XML content.
#' @param data data.frame with document content including IDs for
#'   images (blip_id), footnotes (footnote_id), and hyperlinks (hyperlink_id)
#' @param x an rdocx object containing document relationships and footnotes
#' @return A data.frame with image paths, footnote text, and hyperlink URLs
#'   added, and temporary ID columns removed.
#' @noRd
augment_with_non_text_content <- function(data, x) {
  doc_rel <- x$doc_obj$rel_df()

  # add images
  doc_rel_images <- doc_rel[basename(doc_rel$type) %in% "image", ]
  doc_rel_images <- select(
    doc_rel_images,
    all_of(c("id", "target"))
  )
  names(doc_rel_images) <- c("blip_id", "image_path")
  doc_rel_images$image_path <- file.path(
    x$package_dir,
    "word",
    doc_rel_images$image_path
  )

  data <- left_join(
    data,
    doc_rel_images,
    by = "blip_id"
  )
  data$blip_id <- NULL

  # add footnotes
  ref_wml_str <- x$footnotes$encode_wml_str(additional_ns = list())
  ref_str <- rep(NA_character_, length(ref_wml_str))
  suppressWarnings({
    ref_str <- vapply(
      ref_wml_str,
      function(x) {
        str <- xml_text(xml2::as_xml_document(x))
        paste0(str, collapse = "\n")
      },
      FUN.VALUE = NA_character_
    )
    ref_str
  })
  ref_data <- data.frame(
    footnote_id = names(ref_str),
    footnote_text = unname(ref_str)
  )
  data <- left_join(
    data,
    ref_data,
    by = "footnote_id"
  )
  data$footnote_id <- NULL

  # add hyperlinks
  run_within_hl_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:hyperlink/w:r"
  )

  hl_nodes <- xml_child(run_within_hl_nodes, "parent::w:hyperlink")
  table_hyperlink <- data.frame(
    hyperlink_id = xml_attr(hl_nodes, "id"),
    link_to_bookmark = xml_attr(hl_nodes, "anchor"),
    run_index = as.integer(xml_attr(run_within_hl_nodes, "run_index"))
  )

  doc_rel_external_hyperlink <- doc_rel[
    basename(doc_rel$type) %in% "hyperlink",
  ]
  doc_rel_external_hyperlink <- select(
    doc_rel_external_hyperlink,
    all_of(c("id", "target"))
  )
  names(doc_rel_external_hyperlink) <- c("hyperlink_id", "link")

  table_to_anchor <- dplyr::anti_join(
    table_hyperlink,
    doc_rel_external_hyperlink,
    by = "hyperlink_id"
  )
  table_to_anchor$hyperlink_id <- NULL
  data <- left_join(
    data,
    table_to_anchor,
    by = c("run_index")
  )

  table_hyperlink <- inner_join(
    table_hyperlink,
    doc_rel_external_hyperlink,
    by = "hyperlink_id"
  )
  data <- left_join(
    data,
    select(
      table_hyperlink,
      all_of(c("hyperlink_id", "run_index", "link"))
    ),
    by = c("hyperlink_id", "run_index")
  )
  data$hyperlink_id <- NULL

  data
}

#' @title Augment data with style names
#' @description Adds style names (character, table, paragraph) to the data
#'   by joining with styles information and removes temporary style ID columns.
#' @param data data.frame with document content including style IDs
#'   (run_stylename, tab_style_name, par_style_name)
#' @param styles_data data.frame with style information containing
#'   style_type, style_id, and style_name columns
#' @return A data.frame with human-readable style names added
#'   (character_stylename, table_stylename, paragraph_stylename) and
#'   temporary style ID columns removed.
#' @noRd
augment_with_styles <- function(data, styles_data) {
  styles_character <- filter(
    styles_data,
    .data$style_type %in% "character"
  )
  styles_character <- select(
    styles_character,
    all_of(c("style_id", "style_name"))
  )
  names(styles_character) <- c("run_stylename", "character_stylename")
  data <- left_join(
    data,
    styles_character,
    by = "run_stylename"
  )

  styles_table <- filter(styles_data, .data$style_type %in% "table")
  styles_table <- select(
    styles_table,
    all_of(c("style_id", "style_name"))
  )
  names(styles_table) <- c("tab_style_name", "table_stylename")
  data <- left_join(
    data,
    styles_table,
    by = "tab_style_name"
  )

  styles_paragraph <- filter(
    styles_data,
    .data$style_type %in% "paragraph"
  )
  styles_paragraph <- select(
    styles_paragraph,
    all_of(c("style_id", "style_name"))
  )
  names(styles_paragraph) <- c("par_style_name", "paragraph_stylename")
  data <- left_join(
    data,
    styles_paragraph,
    by = "par_style_name"
  )
  data <- select(
    data,
    -all_of(c("par_style_name", "tab_style_name", "run_stylename"))
  )

  data
}

#' @title Join information tables for document content
#' @description Joins runs, paragraphs, and tables information data frames
#'   and processes the final document content structure.
#' @param infotbl_runs_contents data.frame with run content information
#'   (text, images, footnotes, etc.)
#' @param infotbl_runs data.frame with run formatting information
#'   (bold, italic, font, etc.)
#' @param infotbl_paragraphs data.frame with paragraph information
#'   (style, alignment, numbering, etc.)
#' @param infotbl_tables data.frame with table information
#'   (cell positions, spans, etc.)
#' @return A data.frame with all information joined and properly ordered,
#'   with consecutive run and content indices.
#' @noRd
infotbl_join <- function(
  infotbl_runs_contents,
  infotbl_runs,
  infotbl_paragraphs,
  infotbl_tables
) {
  data <- left_join(
    infotbl_runs_contents,
    infotbl_runs,
    by = "run_index"
  )
  data <- left_join(
    data,
    infotbl_paragraphs,
    by = "doc_index"
  )

  data <- left_join(
    data,
    infotbl_tables,
    by = "doc_index"
  )

  data <- relocate(
    .data = data,
    all_of(
      c(
        "doc_index",
        "run_index",
        "run_content_index",
        "table_index",
        "row_id",
        "cell_id"
      )
    )
  )
  data <- mutate(
    .data = data,
    doc_index = as.integer(.data$doc_index),
    run_index = as.integer(.data$run_index)
  )
  data <- arrange(
    .data = data,
    across(all_of(
      c("doc_index", "run_index", "run_content_index")
    ))
  )
  data <- mutate(
    .data = data,
    .by = all_of(c("doc_index", "run_index")),
    .keep = "all",
    run_content_index = consecutive_id(.data$run_content_index)
  )

  data
}


# extract tools ----
docx_tablecells_information <- function(tc_nodes) {
  col_span_str <- rep("1", length(tc_nodes))
  col_span_nodes <- xml_child(tc_nodes, "w:tcPr/w:gridSpan")
  has_col_span_nodes <- xml_name(col_span_nodes) %in% "gridSpan"
  col_span_str[has_col_span_nodes] <- xml_attr(
    col_span_nodes[has_col_span_nodes],
    "val"
  )

  vmerge_nodes <- xml_child(tc_nodes, "w:tcPr/w:vMerge")
  row_merge <- !is.na(xml_name(vmerge_nodes))
  first <- xml_attr(vmerge_nodes, "val") %in% "restart"

  tc_index <- xml_attr(tc_nodes, "tc_index")
  table_index_sel <- xml_child(tc_nodes, "parent::w:tr/parent::w:tbl")
  table_index <- xml_attr(table_index_sel, "table_index")
  tr_index_sel <- xml_child(tc_nodes, "parent::w:tr")
  tr_index <- xml_attr(tr_index_sel, "tr_index")

  data.frame(
    table_index = as.integer(table_index),
    row_id = as.integer(tr_index),
    cell_id = as.integer(tc_index),
    row_merge = row_merge,
    first = first,
    col_span = col_span_str
  )
}

docx_p_cells_information <- function(p_in_cell_nodes) {
  tc_index <- xml_child(p_in_cell_nodes, "parent::w:tc")
  tc_index <- xml_attr(tc_index, "tc_index")
  doc_index <- xml_attr(p_in_cell_nodes, "doc_index")

  data.frame(
    doc_index = as.integer(doc_index),
    cell_id = as.integer(tc_index)
  )
}
docx_p_information <- function(p_nodes) {
  ilvl_nodes <- xml_child(p_nodes, "w:pPr/w:numPr/w:ilvl")
  numId_nodes <- xml_child(p_nodes, "w:pPr/w:numPr/w:numId")
  pstyle_nodes <- xml_child(p_nodes, "w:pPr/w:pStyle")
  jc_nodes <- xml_child(p_nodes, "w:pPr/w:jc")
  kn_nodes <- xml_child(p_nodes, "w:pPr/w:keepNext")

  doc_index <- xml_attr(p_nodes, "doc_index")

  has_num_pr <- xml_name(xml_child(p_nodes, "w:pPr/w:numPr")) %in% "numPr"

  level_str <- rep(NA_character_, length(p_nodes))
  level_str[has_num_pr] <- xml_attr(ilvl_nodes[has_num_pr], "val")

  num_id_str <- rep(NA_character_, length(p_nodes))
  num_id_str[has_num_pr] <- xml_attr(numId_nodes[has_num_pr], "val")

  data.frame(
    doc_index = as.integer(doc_index),
    par_style_name = xml_attr(pstyle_nodes, "val"),
    keep_with_next = xml_name(kn_nodes) %in% "keepNext",
    align = xml_attr(jc_nodes, "val"),
    level = as.integer(level_str) + 1L,
    num_id = as.integer(num_id_str),
    stringsAsFactors = FALSE
  )
}

docx_tables_information <- function(tbl_nodes) {
  tbl_stylename <- xml_attr(tbl_nodes, "val")
  tblstyle_nodes <- xml_child(
    tbl_nodes,
    "w:tblPr/w:tblStyle"
  )
  data.frame(
    tab_style_name = xml_attr(tblstyle_nodes, "val"),
    table_index = as.integer(xml_attr(tbl_nodes, "table_index"))
  )
}

docx_runs_information <- function(run_nodes) {
  bold <- extract_runs_attr_lgl(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:b",
    attr_name = "val",
    default = FALSE
  )
  italic <- extract_runs_attr_lgl(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:i",
    attr_name = "val",
    default = FALSE
  )
  underline <- extract_runs_attr_lgl(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:u",
    attr_name = "val",
    true_values = c(
      "single",
      "words",
      "double",
      "thick",
      "dash",
      NA_character_
    ),
    default = FALSE
  )

  shd_nodes <- xml_child(run_nodes, "w:rPr/w:shd")
  shading <- xml_attr(shd_nodes, "val")
  shading[!is.na(shading)] <- sprintf("#%s", shading[!is.na(shading)])
  shading_color <- xml_attr(shd_nodes, "color")
  shading_color[!is.na(shading_color)] <- sprintf(
    "#%s",
    shading_color[!is.na(shading_color)]
  )
  shading_fill <- xml_attr(shd_nodes, "fill")
  shading_fill[!is.na(shading_fill)] <- sprintf(
    "#%s",
    shading_fill[!is.na(shading_fill)]
  )

  run_stylename <- extract_runs_attr_str(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:rStyle",
    attr_name = "val",
    default = NA_character_
  )
  hyperlink_id <- extract_runs_attr_str(
    run_nodes = run_nodes,
    node_xpath = "parent::w:hyperlink",
    attr_name = "id",
    default = NA_character_
  )

  run_index <- xml_attr(run_nodes, "run_index")

  rfonts_nodes <- xml_child(run_nodes, "w:rPr/w:rFonts")

  doc_index <- extract_runs_attr_str(
    run_nodes = run_nodes,
    node_xpath = "parent::w:p|parent::*/parent::w:p",
    attr_name = "doc_index",
    default = NA_character_
  )

  sz <- xml_attr(xml_child(run_nodes, "w:rPr/w:sz"), "val")
  sz <- as.integer(sz)

  sz_cs = xml_attr(xml_child(run_nodes, "w:rPr/w:szCs"), "val")
  sz_cs <- as.integer(sz_cs)

  color <- xml_attr(xml_child(run_nodes, "w:rPr/w:color"), "val")
  color[!is.na(color)] <- sprintf("#%s", color[!is.na(color)])

  data.frame(
    doc_index = as.integer(doc_index),
    run_index = run_index,
    run_stylename = run_stylename,
    hyperlink_id = hyperlink_id,
    sz = sz,
    sz_cs = sz_cs,
    font_family_ascii = xml_attr(rfonts_nodes, "ascii"),
    font_family_eastasia = xml_attr(rfonts_nodes, "eastAsia"),
    font_family_hansi = xml_attr(rfonts_nodes, "hAnsi"),
    font_family_cs = xml_attr(rfonts_nodes, "cs"),
    bold = bold,
    italic = italic,
    underline = underline,
    color = color,
    shading = shading,
    shading_color = shading_color,
    shading_fill = shading_fill
  )
}


docx_runs_content_information <- function(run_nodes) {
  run_content_nodes_names <- xml_name(run_nodes)

  suppressWarnings({
    blip_id_ <- xml_attr(
      xml_child(
        run_nodes,
        "wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip"
      ),
      "embed"
    )
    footnote_id_ <- xml_attr(run_nodes, "id")
    text_values <- xml_text(run_nodes)
  })
  field_code <- case_when(
    run_content_nodes_names %in% "instrText" ~ text_values,
    .default = NA_character_
  )
  data.frame(
    blip_id = case_when(
      run_content_nodes_names %in% "drawing" ~ blip_id_,
      .default = NA_character_
    ),
    footnote_id = case_when(
      run_content_nodes_names %in% "footnoteReference" ~ footnote_id_,
      .default = NA_character_
    ),
    field_code = field_code,
    run_content_text = case_when(
      run_content_nodes_names %in% "t" ~ text_values,
      run_content_nodes_names %in% "noBreakHyphen" ~ "\u002D",
      run_content_nodes_names %in% c("br", "cr") ~ "\n",
      run_content_nodes_names %in% "tab" ~ "\t",
      run_content_nodes_names %in% "instrText" ~ field_code,
      run_content_nodes_names %in% "softHyphen" ~ "",
      run_content_nodes_names %in% "sym" ~ xml_attr(run_nodes, "char"),
      .default = NA_character_
    ),
    run_content_index = xml_attr(run_nodes, "run_content_index"),
    run_index = xml_attr(xml_child(run_nodes, "parent::w:r"), "run_index")
  )
}

docx_tablerows_information <- function(tr_nodes) {
  tbl_header_nodes <- xml_child(tr_nodes, "w:trPr/w:tblHeader")
  first_row_nodes <- xml_child(tr_nodes, "w:trPr/w:cnfStyle[@w:firstRow='1']")
  is_header <- xml_name(first_row_nodes) %in%
    "cnfStyle" |
    xml_name(tbl_header_nodes) %in% "tblHeader"
  tr_index <- xml_attr(tr_nodes, "tr_index")

  table_index <- xml_child(tr_nodes, "parent::w:tbl")
  table_index <- xml_attr(table_index, "table_index")

  data.frame(
    table_index = as.integer(table_index),
    row_id = as.integer(tr_index),
    is_header = is_header
  )
}


# main ----------

#' @title Summarise data as paragraphs
#' @description Summarises detailed run-level data into paragraph-level
#'   summaries by collapsing text within paragraphs. Handles both regular
#'   paragraphs and paragraphs within table cells differently.
#' @param data data.frame with detailed run-level information including
#'   doc_index, table_index, paragraph_stylename, and other formatting details
#' @return A data.frame summarised at the paragraph level with collapsed text
#'   in the column mamed `text`. For tables, includes cell and row
#'   information; for regular paragraphs, includes only doc_index and style.
#' @noRd
summarise_as_paragraph <- function(data, preserve = FALSE) {
  finest_grp_by <- c(
    "doc_index",
    "content_type",
    "table_index",
    "row_id",
    "cell_id",
    "is_header",
    "row_span",
    "col_span",
    "table_stylename",
    "paragraph_stylename"
  )

  table_data <- filter(.data = data, !is.na(.data$table_index))
  table_data <- add_text_column(table_data)
  table_data <- summarise(
    .data = table_data,
    .by = all_of(finest_grp_by),
    text = paste0(.data$text, collapse = "")
  )

  if (preserve) {
    finest_grp_by <- c(
      "content_type",
      "table_index",
      "row_id",
      "cell_id",
      "is_header",
      "row_span",
      "col_span",
      "table_stylename",
      "paragraph_stylename"
    )
    table_data <- summarise(
      .data = table_data,
      doc_index = first(.data$doc_index),
      text = paste0(.data$text, collapse = "\n"),
      .by = all_of(finest_grp_by)
    )
  }

  par_data <- filter(.data = data, is.na(.data$table_index))
  par_data <- add_text_column(par_data)
  par_data <- summarise(
    .data = par_data,
    .by = all_of(
      c("doc_index", "content_type", "paragraph_stylename")
    ),
    text = paste0(.data$text, collapse = "")
  )

  dataset <- bind_rows(par_data, table_data)

  # patch for WordR
  names(dataset)[names(dataset) %in% "paragraph_stylename"] <- "style_name"

  dataset
}


#' @title Get Word content in a data.frame
#' @description read content of a Word document and
#' return a data.frame representing the document.
#' @note
#' Documents included with [body_add_docx()] will
#' not be accessible in the results.
#' @param x an rdocx object
#' @param preserve If `FALSE` (default), text in table cells is collapsed into a
#'   single line. If `TRUE`, line breaks in table cells are preserved as a "\\n"
#'   character. This feature is adapted from `docxtractr::docx_extract_tbl()`
#'   published under a [MIT
#'   licensed](https://github.com/hrbrmstr/docxtractr/blob/master/LICENSE) in
#'   the 'docxtractr' package by Bob Rudis.
#' @param remove_fields if TRUE, prevent field codes from appearing in the
#' returned data.frame.
#' @param detailed Should run-level information be included in the dataframe?
#'   Defaults to `FALSE`. If `TRUE`, the dataframe contains detailed information
#'   about each run (text formatting, images, hyperlinks, etc.) instead of
#'   collapsing content at the paragraph level. When `FALSE`, run-level
#'   information such as images, hyperlinks, and text formatting is not available
#'   since data is aggregated at the paragraph level.
#' @return A data.frame with the following columns depending on the value of `detailed`:
#'
#' When `detailed = FALSE` (default), the data.frame contains:
#'
#' - `doc_index`: Document element index (integer).
#' - `content_type`: Type of content: "paragraph" or "table cell" (character).
#' - `style_name`: Name of the paragraph style (character).
#' - `text`: Collapsed text content of the paragraph or cell (character).
#' - `table_index`: Index of the table (integer). `NA` for non-table content.
#' - `row_id`: Row position in table (integer). `NA` for non-table content.
#' - `cell_id`: Cell position in table row (integer). `NA` for non-table content.
#' - `is_header`: Whether the row is a table header (logical). `NA` for non-table content.
#' - `row_span`: Number of rows spanned by the cell (integer). `0` for merged cells. `NA` for non-table content.
#' - `col_span`: Number of columns spanned by the cell (character). `NA` for non-table content.
#' - `table_stylename`: Name of the table style (character). `NA` for non-table content.
#'
#' When `detailed = TRUE`, the data.frame contains additional run-level information:
#'
#' - `run_index`: Index of the run within the paragraph (integer).
#' - `run_content_index`: Index of content element within the run (integer).
#' - `run_content_text`: Text content of the run element (character).
#' - `image_path`: Path to embedded image stored in the temporary directory
#'   associated with the rdocx object (character).
#'   Images should be copied to a permanent location before closing the R
#'   session if needed.
#' - `field_code`: Field code content (character).
#' - `footnote_text`: Footnote text content (character).
#' - `link`: Hyperlink URL (character).
#' - `link_to_bookmark`: Internal bookmark anchor name for hyperlinks (character).
#' - `bookmark_start`: Names of the bookmarks starting on this paragraph
#' (values are concatenated with '|').
#' - `character_stylename`: Name of the character/run style (character).
#' - `sz`: Font size in half-points (integer).
#' - `sz_cs`: Complex script font size in half-points (integer).
#' - `font_family_ascii`: Font family for ASCII characters (character).
#' - `font_family_eastasia`: Font family for East Asian characters (character).
#' - `font_family_hansi`: Font family for high ANSI characters (character).
#' - `font_family_cs`: Font family for complex script characters (character).
#' - `bold`: Whether the run is bold (logical).
#' - `italic`: Whether the run is italic (logical).
#' - `underline`: Whether the run is underlined (logical).
#' - `color`: Text color in hexadecimal format (character).
#' - `shading`: Shading pattern (character).
#' - `shading_color`: Shading foreground color (character).
#' - `shading_fill`: Shading background fill color (character).
#' - `keep_with_next`: Whether paragraph should stay with next (logical).
#' - `align`: Paragraph alignment (character).
#' - `level`: Numbering level (integer). `NA` if not a numbered list.
#' - `num_id`: Numbering definition ID (integer). `NA` if not a numbered list.
#' @example inst/examples/example_docx_summary.R
#' @export
docx_summary <- function(
  x,
  preserve = FALSE,
  remove_fields = FALSE,
  detailed = FALSE
) {
  if (!requireNamespace("dplyr")) {
    cli::cli_abort(
      "package {.pkg dplyr} is required for running {.fn docx_summary}."
    )
  }
  if (!requireNamespace("tidyr")) {
    cli::cli_abort(
      "package {.pkg tidyr} is required for running {.fn docx_summary}."
    )
  }

  if (remove_fields) {
    # also remove complex field characters
    fldChar_nodes <- xml_find_all(x$doc_obj$get(), "//w:fldChar")
    xml_remove(fldChar_nodes)

    # remove instrText nodes
    instrText_nodes <- xml_find_all(x$doc_obj$get(), "//w:instrText")
    xml_remove(instrText_nodes)

    # remove fldData nodes
    fldData_nodes <- xml_find_all(x$doc_obj$get(), "//w:fldData")
    xml_remove(fldData_nodes)
  }

  # add info to xml -----
  ## p_nodes
  p_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:p"
  )
  xml_attr(p_nodes, "w:doc_index") <- seq_along(p_nodes)

  ## tbl_nodes
  tbl_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tbl"
  )
  xml_attr(tbl_nodes, "w:table_index") <- seq_along(tbl_nodes)

  ## tr_nodes
  tr_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tr"
  )
  xml_attr(tr_nodes, "w:tr_index") <- seq_along(tr_nodes)

  ## tc_nodes
  tc_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tc"
  )
  xml_attr(tc_nodes, "w:tc_index") <- seq_along(tc_nodes)

  ## run_nodes
  run_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:r"
  )
  xml_attr(run_nodes, "w:run_index") <- seq_along(run_nodes)

  ## run_content_nodes
  run_content_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:r/*[not(self::w:rPr)]"
  )
  xml_attr(run_content_nodes, "w:run_content_index") <- seq_along(
    run_content_nodes
  )

  ## bookmark_nodes
  bookmark_nodes <- xml_find_all(
    x$doc_obj$get(),
    "w:body/w:p/w:bookmarkStart"
  )

  data_bookmark <- data.frame(
    bookmark_start = xml_attr(bookmark_nodes, "name"),
    doc_index = xml_attr(xml_child(bookmark_nodes, "parent::w:p"), "doc_index")
  )
  data_bookmark <- summarise(
    .data = data_bookmark,
    bookmark_start = paste0(.data$bookmark_start, collapse = "|"),
    .by = all_of("doc_index")
  )
  data_bookmark$doc_index <- as.integer(data_bookmark$doc_index)

  ## p_in_cell_nodes
  p_in_cell_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tbl/w:tr/w:tc/w:p"
  )

  # info for runs: infotbl_runs -----
  infotbl_runs_contents <- docx_runs_content_information(run_content_nodes)
  infotbl_runs <- docx_runs_information(run_nodes)

  # info for tables: infotbl_tables -----
  tmp_infotbl_tables <- docx_tables_information(tbl_nodes)
  tmp_infotbl_tablecells <- docx_tablecells_information(tc_nodes)
  tmp_infotbl_tablecells_paragraphs <- docx_p_cells_information(p_in_cell_nodes)
  tmp_infotbl_tablerows <- docx_tablerows_information(tr_nodes)
  ## final table infotbl_tables
  infotbl_tables <- infotbl_tables_compute(
    tmp_infotbl_tablecells_paragraphs,
    tmp_infotbl_tablecells,
    tmp_infotbl_tablerows,
    tmp_infotbl_tables
  )

  # info for paragraphs: infotbl_paragraphs -----
  infotbl_paragraphs <- docx_p_information(p_nodes)
  infotbl_paragraphs <- left_join(
    infotbl_paragraphs,
    data_bookmark,
    by = "doc_index"
  )

  # final joins -------
  data <- infotbl_join(
    infotbl_runs_contents,
    infotbl_runs,
    infotbl_paragraphs,
    infotbl_tables
  )

  # add stylenames -----
  styles_data <- select(
    styles_info(x),
    all_of(c("style_type", "style_id", "style_name"))
  )
  data <- augment_with_styles(data, styles_data)

  # add non textual content -----
  data <- augment_with_non_text_content(data, x)

  data <- mutate(
    .data = data,
    .by = all_of(c("doc_index")),
    .keep = "all",
    run_index = consecutive_id(.data$run_index)
  )
  data <- mutate(
    .data = data,
    content_type = if_else(
      is.na(.data$table_index),
      "paragraph",
      "table cell"
    )
  )

  column_names <- c(
    "doc_index",
    "content_type",
    "run_index",
    "run_content_index",

    "run_content_text",
    "image_path",
    "field_code",
    "footnote_text",
    "link",
    "link_to_bookmark",
    "bookmark_start",
    "character_stylename",
    "sz",
    "sz_cs",
    "font_family_ascii",
    "font_family_eastasia",
    "font_family_hansi",
    "font_family_cs",
    "bold",
    "italic",
    "underline",
    "color",
    "shading",
    "shading_color",
    "shading_fill",

    "paragraph_stylename",
    "keep_with_next",
    "align",
    "level",
    "num_id",

    "table_index",
    "row_id",
    "cell_id",
    "col_span",
    "row_span",
    "is_header",
    "table_stylename"
  )

  data <- select(.data = data, all_of(column_names))

  if (!detailed) {
    data <- summarise_as_paragraph(data, preserve = preserve)
  }

  data
}

#' @title Write a ggplot Object to PNG File
#'
#' @description Renders a ggplot object to a PNG file using ragg for high-quality output.
#'
#' @param ggobj A ggplot object to render
#' @param width Numeric, width of the output image
#' @param height Numeric, height of the output image
#' @param res Numeric, resolution in DPI (default 200)
#' @param units Character, units for width and height ("in", "cm", "mm", "px") (default "in")
#' @param pointsize Integer, The default pointsize of the device in pt
#' @param scaling scaling factor to apply
#' @param path Character, output file path. If NULL, a temporary file is created (default NULL)
#'
#' @return Character, the path to the created PNG file
#'
#' @importFrom ragg agg_png
#' @importFrom grDevices dev.off
#' @examples
#' plot_in_png(
#'   code = {
#'     barplot(1:10)
#'   },
#'   width = 5,
#'   height = 4,
#'   res = 72,
#'   units = "in"
#' )
#' @export
#' @keywords internal
plot_in_png <- function(
  ggobj = NULL,
  code = NULL,
  width,
  height,
  res = 200,
  units = "in",
  pointsize = 11,
  scaling = 1,
  path = NULL
) {
  if (is.null(path)) {
    path <- tempfile(fileext = ".png")
  }

  if (!is.null(ggobj)) {
    code <- str2lang("print(ggobj)")
  }

  agg_png(
    filename = path,
    width = width,
    height = height,
    units = units,
    background = "transparent",
    res = res,
    pointsize = pointsize,
    scaling = scaling
  )
  tryCatch(
    {
      eval(code)
    },
    finally = dev.off()
  )
  path
}

#' @importFrom openssl base64_encode
#' @export
#' @title Encode Character Vector to Base64
#' @description
#' Encodes one or more elements of a character vector into Base64 format.
#' @param x A character vector. NA values are preserved.
#' @return A character vector of Base64-encoded strings, same length as `x`.
#' @examples
#' as_base64(letters)
#' as_base64(c("hello", NA, "world"))
#' @keywords internal
as_base64 <- function(x) {
  if (!is.character(x)) {
    stop("'x' must be a character vector.")
  }
  z <- vapply(
    x,
    function(elem) {
      if (is.na(elem)) {
        NA_character_
      } else {
        base64_encode(charToRaw(elem))
      }
    },
    NA_character_
  )
  unname(z)
}

#' @importFrom openssl base64_decode
#' @export
#' @title Decode Base64 Vector to Character
#' @description
#' Decodes one or more Base64-encoded elements back into their original character form.
#' @param x A character vector of Base64-encoded strings. NA values are preserved.
#' @return A character vector of decoded strings, same length as `x`.
#' @examples
#' z <- as_base64(c("hello", "world"))
#' from_base64(z)
#' @keywords internal
from_base64 <- function(x) {
  if (!is.character(x)) {
    stop("'x' must be a character vector of Base64 strings.")
  }
  z <- vapply(
    x,
    function(elem) {
      if (is.na(elem)) {
        NA_character_
      } else {
        raw <- tryCatch(
          {
            b64_str <- base64_decode(elem)
            if (all(nchar(b64_str) < 1)) {
              stop("empty result")
            }
            b64_str
          },
          error = function(e) {
            stop(
              "Failed to decode Base64 element: '",
              elem,
              "'.",
              call. = FALSE
            )
          }
        )
        rawToChar(raw)
      }
    },
    NA_character_
  )
  unname(z)
}


#' Convert Data URIs to PNG Files
#'
#' @description Decodes base64-encoded data URIs and writes them to PNG files.
#'
#' @param data_uri Character, a data URI character vector starting with "data:image/png;base64,"
#' @param output_files Character, paths to the output PNG files
#'
#' @return Character, the paths to the created PNG files
#'
#' @importFrom openssl base64_decode
#' @examples
#' rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
#' base64_str <- image_to_base64(rlogo)
#' base64_to_image(
#'   data_uri = base64_str,
#'   output_files = tempfile(fileext = ".jpeg")
#' )
#' @export
#' @keywords internal
base64_to_image <- function(data_uri, output_files) {
  for (i in seq_along(data_uri)) {
    base64_part <- sub("^data:image/[^;]+;base64,", "", data_uri[[i]])
    raw_data <- base64_decode(base64_part)
    writeBin(raw_data, output_files[[i]])
  }

  output_files
}

mime_type <- function(paths) {
  result <- character(length(paths))
  pattern <- "\\.(png|gif|jpg|jpeg|svg|tiff|pdf|webp)$"
  m <- regexpr(pattern = pattern, text = paths)
  result[attr(m, "match.length") > -1] <- regmatches(paths, m)
  result <- gsub("\\.jpg", ".jpeg", result)
  result <- gsub("\\.svg", ".svg+xml", result)
  result <- gsub("^\\.{1}", "", result)
  prefix <- ifelse(result %in% "pdf", "application", "image")
  result <- paste(prefix, result, sep = "/")
  result[attr(m, "match.length") < 0] <- NA_character_
  result
}

#' @importFrom openssl base64_encode
#' @export
#' @title Images to base64
#' @description encodes images into base64 strings.
#' @param filepaths file names.
#' @keywords internal
#' @examples
#' rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
#' base64_str <- image_to_base64(rlogo)
image_to_base64 <- function(filepaths) {
  mimes <- mime_type(paths = filepaths)
  if (anyNA(mimes)) {
    cli::cli_abort(
      paste0(
        "Unknown image(s) format: ",
        cli::ansi_collapse(
          basename(filepaths)[is.na(mimes)],
          trunc = 5
        )
      )
    )
  }

  if (!all(file.exists(filepaths))) {
    cli::cli_abort(
      paste0(
        "File(s) not found: ",
        cli::ansi_collapse(
          basename(filepaths)[!file.exists(filepaths)],
          trunc = 5
        )
      )
    )
  }

  base64_lst <- mapply(
    FUN = function(filepath, mime) {
      dat <- readBin(
        filepath,
        what = "raw",
        size = 1,
        endian = "little",
        n = file.info(filepath)$size
      )
      base64_str <- base64_encode(bin = dat)
      base64_str <- sprintf("data:%s;base64,%s", mime, base64_str)
      base64_str
    },
    filepath = filepaths,
    mime = mimes,
    SIMPLIFY = FALSE
  )
  unname(unlist(base64_lst))
}

#' @importFrom uuid UUIDgenerate
#' @export
#' @title generates unique identifiers
#' @description generates unique identifiers based
#' on [uuid::UUIDgenerate()].
#' @param n integer, number of unique identifiers to generate.
#' @param ... arguments sent to [uuid::UUIDgenerate()]
#' @keywords internal
#' @examples
#' uuid_generate(n = 5)
uuid_generate <- function(n = 1, ...) {
  UUIDgenerate(n = n, ...)
}

.url_special_chars <- list(
  `&` = "&amp;",
  `<` = "&lt;",
  `>` = "&gt;",
  `'` = "&#39;",
  `"` = "&quot;",
  ` ` = "&nbsp;"
)

#' @export
#' @title officer url encoder
#' @description encode url so that it can be easily
#' decoded when 'officer' write a file to the disk.
#' @param x a character vector of URL
#' @keywords internal
#' @examples
#' officer_url_encode("https://cran.r-project.org/")
officer_url_encode <- function(x) {
  for (chr in names(.url_special_chars)) {
    x <- gsub(chr, .url_special_chars[[chr]], x, fixed = TRUE, useBytes = TRUE)
  }
  Encoding(x) <- "UTF-8"
  x
}

officer_url_decode <- function(x) {
  for (chr in rev(names(.url_special_chars))) {
    x <- gsub(.url_special_chars[[chr]], chr, x, fixed = TRUE, useBytes = TRUE)
  }
  Encoding(x) <- "UTF-8"
  x
}

#' @export
#' @title Add a slide
#' @description Add a slide into a pptx presentation.
#' @param x an `rpptx` object.
#' @param layout slide layout name to use. Can be ommited of a default layout is set via [layout_default()].
#' @param master master layout name where `layout` is located. Only required in case of several masters if layout is
#'   not unique.
#' @param ... Key-value pairs of the form `"short form location" = object` passed to [phs_with]. See section
#'   `"Short forms"` in [phs_with] for details, available short forms and examples.
#' @param .dots List of key-value pairs of the form `list("short form location" = object)`. Alternative to `...`. See
#'   [phs_with] for details.
#' @example inst/examples/example_add_slide.R
#' @seealso [print.rpptx()], [read_pptx()], [layout_summary()], [plot_layout_properties()], [ph_with()], [phs_with()], [layout_default()]
#' @family slide_manipulation
add_slide <- function(x, layout = NULL, master = NULL, ..., .dots = NULL) {
  if (is.null(layout) && !has_layout_default(x)) {
    # inform user. Passing no layout will be defunct in a future verion
    .Deprecated(
      "",
      package = "officer",
      msg = paste(
        "Calling `add_slide()` without specifying a `layout` is deprecated.\n",
        "Please pass a `layout` or use `layout_default()` to set a default.\n",
        '=> I will now continue with the former `layout` default "Title and Content" for backwards compatibility...'
      )
    )
    layout <- "Title and Content"
  }

  if (is.null(layout) && has_layout_default(x)) {
    ld <- get_layout_default(x)
    layout <- ld$layout
    master <- ld$master
  }

  la <- get_layout(x, layout, master, layout_by_id = FALSE)

  dots_list <- list(...)
  if (length(dots_list) > 0 && !is_named(dots_list)) {
    cli::cli_abort(
      c(
        "Missing key in {.arg ...}",
        "x" = "{.arg ...} requires a key-value syntax, with a ph short-form location as the key",
        "i" = "Example: {.code add_slide(x, ..., title = 'My title', 'body[1]' = 'My body')}"
      )
    )
  }
  if (length(.dots) > 0 && !is_named(.dots)) {
    cli::cli_abort(
      c(
        "Missing names in {.arg .dots}",
        "x" = "{.arg .dots} must be a named list, with ph short-form locations as names",
        "i" = "Example: {.code add_slide(x, ..., .dots = list(title = 'My title', 'body[1]' = 'My body'))}"
      )
    )
  }

  new_slidename <- x$slide$get_new_slidename()

  xml_file <- file.path(x$package_dir, "ppt/slides", new_slidename)
  layout_obj <- x$slideLayouts$collection_get(la$layout_file)
  layout_obj$write_template(xml_file)

  # update presentation elements
  x$presentation$add_slide(target = file.path("slides", new_slidename))
  x$content_type$add_slide(partname = file.path("/ppt/slides", new_slidename))

  x$slide$add_slide(xml_file, x$slideLayouts$get_xfrm_data())
  x$cursor <- x$slide$length()

  # fill placeholders
  dots <- utils::modifyList(dots_list, .dots %||% list())
  if (length(dots) > 0) {
    x <- phs_with(x, .dots = dots)
  }
  x
}


#' @export
#' @title Change current slide
#' @description Change current slide index of an rpptx object.
#' @param x an rpptx object
#' @param index slide index
#' @example inst/examples/example_on_slide.R
#' @family slide_manipulation
#' @seealso [read_pptx()], [ph_with()]
on_slide <- function(x, index) {
  l_ <- length(x)
  if (l_ < 1) {
    cli::cli_abort("Presentation contains no slide.")
  }
  if (!between(index, 1, l_)) {
    cli::cli_abort(
      c(
        "Slide index {.val {index}} is out of range.",
        "x" = "Presentation has {cli::no(l_)} slide{?s}."
      )
    )
  }

  filename <- basename(x$presentation$slide_data()$target[index])
  location <- which(x$slide$get_metadata()$name %in% filename)

  x$cursor <- x$slide$slide_index(filename)
  x
}


#' @export
#' @title Remove slide(s)
#' @description Remove one or more slides from a pptx presentation.
#' @param x an rpptx object
#' @param index slide index or a vector of slide indices to remove,
#' default to current slide position.
#' @param rm_images unused anymore.
#' @note cursor is set on the last slide.
#' @example inst/examples/example_remove_slide.R
#' @family slide_manipulation
#' @seealso [read_pptx()], [ph_with()], [ph_remove()]
remove_slide <- function(x, index = NULL, rm_images = FALSE) {
  stop_if_not_rpptx(x)

  l_ <- length(x)
  if (l_ < 1) {
    cli::cli_abort("Presentation contains no slide to delete.")
  }

  if (is.null(index)) {
    index <- x$cursor
  }

  if (length(index) == 0) {
    return(x)
  }

  # Validate all indices
  indices <- unique(as.integer(index))
  invalid_indices <- indices[!between(indices, 1, l_)]
  if (length(invalid_indices) > 0) {
    n_invalid <- length(invalid_indices)
    cli::cli_abort(
      c(
        "{cli::qty(n_invalid)}Slide index{?es} {.val {invalid_indices}} {cli::qty(n_invalid)}{?is/are} out of range.",
        "x" = "Presentation has {cli::no(l_)} slide{?s}."
      )
    )
  }

  # Get slide mapping information using existing slide_data method
  slide_map <- x$presentation$slide_data()
  slide_map$filename <- basename(slide_map$target)

  indices_to_remove <- sort(unique(indices), decreasing = TRUE)
  slides_to_remove_df <- slide_map[indices_to_remove, , drop = FALSE]

  # Build file paths for deletion
  files_to_delete <- file.path(
    x$package_dir,
    "ppt",
    "slides",
    slides_to_remove_df$filename
  )
  rels_to_delete <- file.path(
    x$package_dir,
    "ppt",
    "slides",
    "_rels",
    paste0(slides_to_remove_df$filename, ".rels")
  )

  # Delete slide XML files if they exist
  unlink(files_to_delete[file.exists(files_to_delete)], force = TRUE)
  # Delete slide relationship files if they exist
  unlink(rels_to_delete[file.exists(rels_to_delete)], force = TRUE)

  # Remove slides from internal collections
  for (filename in slides_to_remove_df$filename) {
    slide_idx <- x$slide$slide_index(filename)
    x$slide$remove_slide(slide_idx)
    x$presentation$remove_slide(filename)
    x$content_type$remove_slide(partname = filename)
  }

  # Set cursor to last slide
  x$cursor <- x$slide$length()
  x
}


#' @export
#' @title Move a slide
#' @description Move a slide in a pptx presentation.
#' @inheritParams remove_slide
#' @param to new slide index.
#' @note cursor is set on the last slide.
#' @example inst/examples/example_move_slide.R
#' @family slide_manipulation
#' @seealso [read_pptx()]
move_slide <- function(x, index = NULL, to) {
  x$presentation$slide_data()

  if (is.null(index)) {
    index <- x$cursor
  }

  l_ <- length(x)

  if (l_ < 1) {
    cli::cli_abort("Presentation contains no slide.")
  }
  if (!between(index, 1, l_)) {
    cli::cli_abort(
      c(
        "Slide index {.val {index}} is out of range.",
        "x" = "Presentation has {cli::no(l_)} slide{?s}."
      )
    )
  }
  if (!between(to, 1, l_)) {
    cli::cli_abort(
      c(
        "Target position {.val {to}} is out of range.",
        "x" = "Presentation has {cli::no(l_)} slide{?s}."
      )
    )
  }

  x$presentation$move_slide(from = index, to = to)
  x$cursor <- to
  x
}


#' @title Correct pptx content references
#' @description Content references are not managed directly
#' but computed after the content is added. This function
#' loop over each slide and fix references (links and images)
#' if necessary.
#' @param x an rpptx object
#' @return an rpptx object
#' @noRd
pptx_fortify_slides <- function(x) {
  for (cursor_index in seq_len(x$slide$length())) {
    slide <- x$slide$get_slide(cursor_index)

    process_images(
      slide,
      slide$relationship(),
      x$package_dir,
      media_dir = "ppt/media",
      media_rel_dir = "../media"
    )

    process_links(slide, type = "pml")

    cnvpr <- xml_find_all(slide$get(), "//p:cNvPr")
    for (i in seq_along(cnvpr)) {
      xml_attr(cnvpr[[i]], "id") <- i
    }
  }

  x
}


#' @export
#' @title pptx tags for visual and non visual properties
#' @description Visual and non visual properties of a
#' shape can be returned by this function.
#' @param left,top,width,height place holder coordinates
#' in inches.
#' @param bg background color
#' @param rot rotation angle
#' @param label a label for the placeholder.
#' @param ph string containing xml code for ph tag
#' @param ln a [sp_line()] object specifying the outline style.
#' @param geom shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html
#' @return a character value
#' @family functions for officer extensions
#' @keywords internal
shape_properties_tags <- function(
  left = 0,
  top = 0,
  width = 3,
  height = 3,
  bg = "transparent",
  rot = 0,
  label = "",
  ph = "<p:ph/>",
  ln = sp_line(lwd = 0, linecmpd = "solid", lineend = "rnd"),
  geom = NULL
) {
  if (!is.null(bg) && !is.color(bg)) {
    cli::cli_abort("{.arg bg} must be a valid color, not {.val {bg}}.")
  }

  bg_str <- solid_fill_pml(bg)
  ln_str <- ln_pml(ln)
  geom_str <- prst_geom_pml(geom)

  xfrm_str <- a_xfrm_str(
    left = left,
    top = top,
    width = width,
    height = height,
    rot = rot
  )
  if (is.null(ph) || is.na(ph)) {
    ph <- "<p:ph/>"
  }

  randomid <- officer::uuid_generate()

  str <- "<p:nvSpPr><p:cNvPr id=\"%s\" name=\"%s\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr>%s</p:nvPr></p:nvSpPr><p:spPr>%s%s%s%s</p:spPr>"

  sprintf(str, randomid, label, ph, xfrm_str, geom_str, bg_str, ln_str)
}


# check if slide index exists
ensure_slide_index_exists <- function(x, slide_idx) {
  stop_if_not_rpptx(x)
  if (!is.numeric(slide_idx)) {
    cli::cli_abort(
      c(
        "{.arg slide_idx} must be {.cls numeric}",
        "x" = "You provided {.cls {class(slide_idx)[1]}} instead."
      ),
      call = NULL
    )
  }
  n <- length(x) # no of slides
  check <- slide_idx %in% seq_len(n)
  if (!check) {
    cli::cli_abort(
      c(
        "Slide index {.val {slide_idx}} is out of range.",
        "x" = "Presentation has {cli::no(n)} slide{?s}."
      ),
      call = NULL
    )
  }
}


# internal workhorse get/set slide visibility
# x : rpptx object
# slide_idx: id of slide
# value: Use TRUE / FALSE to set visibility.
.slide_visible <- function(x, slide_idx, value = NULL) {
  stop_if_not_rpptx(x)
  slide <- x$slide$get_slide(slide_idx)
  slide_xml <- slide$get()
  node <- xml2::xml_find_first(slide_xml, "/p:sld")
  if (is.null(value)) {
    value <- xml2::xml_attr(node, "show")
    value <- as.logical(as.numeric(value))
    ifelse(is.na(value), TRUE, value) # if show is not set, the slide is shown
  } else {
    stop_if_not_class(value, "logical", arg = "value")
    xml2::xml_set_attr(node, "show", value = as.numeric(value))
    slide$save()
    invisible(x)
  }
}


#' Get or set slide visibility
#'
#' PPTX slides can be visible or hidden. This function gets or sets the visibility of slides.
#' @param x An `rpptx` object.
#' @param value Boolean vector with slide visibilities.
#' @rdname slide-visible
#' @export
#' @example inst/examples/example_slide_visible.R
#' @return Boolean vector with slide visibilities or `rpptx` object if changes are made to the object.
`slide_visible<-` <- function(x, value) {
  stop_if_not_rpptx(x)
  stop_if_not_class(value, "logical", arg = "value")
  n_vals <- length(value)
  n_slides <- length(x)
  if (n_vals > n_slides) {
    cli::cli_abort(
      "More values ({.val {n_vals}}) than slides ({.val {n_slides}})"
    )
  }
  if (n_vals != 1 && n_vals != n_slides) {
    cli::cli_warn(
      "Value is not length 1 or same length as number of slides ({.val {n_slides}}). Recycling values."
    )
  }
  value <- rep(value, length.out = n_slides)
  for (i in seq_along(value)) {
    .slide_visible(x, i, value[i])
  }
  invisible(x)
}


#' @param hide,show Indexes of slides to hide or show.
#' @rdname slide-visible
#' @export
slide_visible <- function(x, hide = NULL, show = NULL) {
  stop_if_not_rpptx(x)
  idx_in_both <- intersect(as.integer(hide), as.integer(show))
  if (length(idx_in_both) > 1) {
    cli::cli_abort(
      "Overlap between indexes in {.arg hide} and {.arg show}: {.val {idx_in_both}}",
      "x" = "Indexes must be mutually exclusive."
    )
  }
  if (!is.null(hide)) {
    stop_if_not_integerish(hide, "hide")
    stop_if_not_in_slide_range(x, hide, arg = "hide")
    slide_visible(x)[hide] <- FALSE
  }
  if (!is.null(show)) {
    stop_if_not_integerish(show, "show")
    stop_if_not_in_slide_range(x, show, arg = "show")
    slide_visible(x)[show] <- TRUE
  }
  n_slides <- length(x)
  res <- vapply(
    seq_len(n_slides),
    function(idx) .slide_visible(x, idx),
    logical(1)
  )
  if (is.null(hide) && is.null(show)) {
    res
  } else {
    x
  }
}

test_that("check errors", {
  x <- read_pptx()
  expect_error(
    move_slide(x, index = 2, to = 1),
    "Presentation contains no slide"
  )

  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))

  x <- on_slide(x, index = 1)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 1")

  expect_error(move_slide(x, to = 4))
  expect_error(move_slide(x, index = 5, to = 4))
  x <- move_slide(x, to = 3)
  x <- on_slide(x, index = 3)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 1")
  x <- remove_slide(x, index = 3)
  x <- move_slide(x, index = 2, to = 1)
  x <- on_slide(x, index = 1)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 3")
})


unlink("*.pptx")

test_that("check errors", {
  x <- read_pptx()
  expect_error(move_slide(x, index = 2, to = 1), "presentation contains no slide")

  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))

  x <- on_slide(x, index = 1)
  sm <- slide_summary(x)
  expect_equal(sm[1,]$text, "Hello world 1")

  expect_error(move_slide(x, to = 4))
  expect_error(move_slide(x, index = 5, to = 4))
  x <- move_slide(x, to = 3)
  x <- on_slide(x, index = 3)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 1")
  x <- remove_slide(x, index = 3)
  x <- move_slide(x, index = 2, to = 1)
  x <- on_slide(x, index = 1)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 3")
})


unlink("*.pptx")

test_that("check slide selection", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))

  x <- on_slide(x, index = 1)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 1")

  x <- on_slide(x, index = 2)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 2")

  x <- on_slide(x, index = 3)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 3")
})


test_that("check errors", {
  x <- read_pptx()
  expect_error(slide_summary(x), "Presentation contains no slide")
  expect_error(remove_slide(x), "Presentation contains no slide to delete")
  expect_error(on_slide(x, index = 1), "Presentation contains no slide")

  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title", location = ph_location_type(type = "title"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title 2", location = ph_location_type(type = "title"))

  expect_error(on_slide(x, index = 3), "out of range")
  expect_error(remove_slide(x, index = 3), "Presentation has 2 slides")
  expect_error(slide_summary(x, index = 3), "Presentation has 2 slides")
})


test_that("get shape id", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(doc, "hello", location = ph_location_type(type = "body"))
  file <- print(doc, target = tempfile(fileext = ".pptx"))
  doc <- read_pptx(file)
  expect_equal(officer:::get_shape_id(doc, type = "body", id = 1), "2")
  expect_equal(
    officer:::get_shape_id(doc, ph_label = "Content Placeholder 2", id = 1),
    "2"
  )
  expect_error(officer:::get_shape_id(doc, type = "body", id = 4))
})


unlink("*.pptx")


test_that("ensure_slide_index_exists ", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()

  expect_error(ensure_slide_index_exists(x, 0), "Slide index 0 is out of range")
  expect_error(ensure_slide_index_exists(x, 2), "Slide index 2 is out of range")

  x <- add_slide(x, "Comparison")
  expect_no_error(ensure_slide_index_exists(x, 1))
  expect_error(ensure_slide_index_exists(x, 0), "Slide index 0 is out of range")
  expect_error(ensure_slide_index_exists(x, 2), "Slide index 2 is out of range")

  error_msg <- "`slide_idx` must be <numeric>"
  expect_error(ensure_slide_index_exists(x, "1"), error_msg)
  expect_error(ensure_slide_index_exists(x, NA), error_msg)
  expect_error(ensure_slide_index_exists(x, NULL), error_msg)
})

test_that("check slide selection", {
  x <- read_pptx()
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 1", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 2", location = ph_location_type(type = "body"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world 3", location = ph_location_type(type = "body"))

  x <-  on_slide(x, index = 1)
  sm <- slide_summary(x)
  expect_equal(sm[1,]$text, "Hello world 1")

  x <-  on_slide(x, index = 2)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 2")

  x <-  on_slide(x, index = 3)
  sm <- slide_summary(x)
  expect_equal(sm[1, ]$text, "Hello world 3")

})


test_that("check errors", {
  x <- read_pptx()
  expect_error(slide_summary(x), "presentation contains no slide")
  expect_error(remove_slide(x), "presentation contains no slide to delete")
  expect_error(on_slide(x, index = 1), "presentation contains no slide")

  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title", location = ph_location_type(type = "title"))
  x <- add_slide(x, "Title and Content", "Office Theme")
  x <- ph_with(x, "Hello world", location = ph_location_type(type = "body"))
  x <- ph_with(x, "my title 2", location = ph_location_type(type = "title"))

  expect_error(on_slide(x, index = 3), "unvalid index 3")
  expect_error(remove_slide(x, index = 3), "unvalid index 3")
  expect_error(slide_summary(x, index = 3), "unvalid index 3")

})


test_that("get shape id", {
  doc <- read_pptx()
  doc <- add_slide(doc, "Title and Content", "Office Theme")
  doc <- ph_with(doc, "hello", location = ph_location_type(type = "body"))
  file <- print(doc, target = tempfile(fileext = ".pptx"))
  doc <- read_pptx(file)
  expect_equal(officer:::get_shape_id(doc, type = "body", id = 1), "2")
  expect_equal(officer:::get_shape_id(doc, ph_label = "Content Placeholder 2", id = 1), "2")
  expect_error(officer:::get_shape_id(doc, type = "body", id = 4) )
})


unlink("*.pptx")


test_that("ensure_slide_index_exists ", {
  opts <- options(cli.num_colors = 1) # suppress colors for error message check
  on.exit(options(opts))

  x <- read_pptx()

  expect_error(ensure_slide_index_exists(x, 0), "Slide index 0 is out of range")
  expect_error(ensure_slide_index_exists(x, 2), "Slide index 2 is out of range")

  x <- add_slide(x, "Comparison")
  expect_no_error(ensure_slide_index_exists(x, 1))
  expect_error(ensure_slide_index_exists(x, 0), "Slide index 0 is out of range")
  expect_error(ensure_slide_index_exists(x, 2), "Slide index 2 is out of range")

  error_msg <- "`slide_idx` must be <numeric>"
  expect_error(ensure_slide_index_exists(x, "1"), error_msg)
  expect_error(ensure_slide_index_exists(x, NA), error_msg)
  expect_error(ensure_slide_index_exists(x, NULL), error_msg)
})

# openxml_document --------------------------------------------------------
# This class handle basic operations on a openxml document:
# - initialize
openxml_document <- R6Class(
  "openxml_document",
  public = list(

    initialize = function( dir ) {
      private$reldir = dir
      private$rels_doc <- relationship$new()
    },

    feed = function( file ) {
      private$filename <- file
      private$doc <- read_xml(file)

      private$rels_filename <- file.path( dirname(file), "_rels", paste0(basename(file), ".rels") )

      if( file.exists(private$rels_filename) )
        private$rels_doc <- relationship$new()$feed_from_xml(private$rels_filename)
      else private$rels_doc <- relationship$new()

      self
    },
    file_name = function(){
      private$filename
    },
    name = function(){
      basename(private$filename)
    },
    get = function(){
      private$doc
    },
    dir_name = function(){
      private$reldir
    },
    save = function() {
      # remove duplicate namespace definitions
      private$doc <- read_xml(as.character(private$doc), options = "NSCLEAN")
      write_xml(private$doc, file = private$filename)
      if( nrow(self$rel_df()) > 0 ){
        private$rels_doc$write(private$rels_filename)
      }
      self
    },
    remove = function() {
      unlink(private$filename)
      if( file.exists(private$rels_filename) )
        unlink(private$rels_filename)
      private$filename
    },
    rel_df = function(){
      private$rels_doc$get_data()
    },
    relationship = function(){
      private$rels_doc
    }

  ),
  private = list(

    filename = NULL,
    rels_filename = NULL,
    doc = NULL,
    rels_doc = NULL,
    reldir = NULL

  )
)

get_shape_id <- function(x, type = NULL, id = NULL, ph_label = NULL) {
  slsmry <- slide_summary(x)

  if (is.null(ph_label)) {
    sel <- which(slsmry$type %in% type)
    if (length(sel) < 1) {
      stop("no shape of type ", shQuote(type), " has been found")
    }
    sel <- sel[id]
    if (sum(is.finite(sel)) != 1) {
      stop(
        "no shape of type ",
        shQuote(type),
        " and with id ",
        id,
        " has been found"
      )
    }
  } else {
    sel <- which(slsmry$ph_label %in% ph_label)
    if (length(sel) < 1) {
      stop("no shape with label ", shQuote(ph_label), " has been found")
    }
    sel <- sel[id]
    if (sum(is.finite(sel)) != 1) {
      stop(
        "no shape with label ",
        shQuote(ph_label),
        "and with id ",
        id,
        " has been found"
      )
    }
  }
  slsmry$id[sel]
}


#' @export
#' @title Remove a shape
#' @description Remove a shape in a slide.
#' @param x an rpptx object
#' @param type placeholder type
#' @param id placeholder index (integer) for a duplicated type. This is to be used when a placeholder
#' type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
#' add onto the first, use `id = 1` and `id = 2` for the second one.
#' Values can be read from [slide_summary()].
#' @param ph_label label associated to the placeholder. Use column
#' `ph_label` of result returned by [slide_summary()].
#' If used, `type` and `id` are ignored.
#' @param id_chr deprecated.
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' dummy_fun <- function(doc) {
#'   doc <- add_slide(doc,
#'     layout = "Two Content",
#'     master = "Office Theme"
#'   )
#'   doc <- ph_with(
#'     x = doc, value = "Un titre",
#'     location = ph_location_type(type = "title")
#'   )
#'   doc <- ph_with(
#'     x = doc, value = "Un corps 1",
#'     location = ph_location_type(type = "body", id = 1)
#'   )
#'   doc <- ph_with(
#'     x = doc, value = "Un corps 2",
#'     location = ph_location_type(type = "body", id = 2)
#'   )
#'   doc
#' }
#' doc <- read_pptx()
#' for (i in 1:3) {
#'   doc <- dummy_fun(doc)
#' }
#'
#' doc <- on_slide(doc, index = 1)
#' doc <- ph_remove(x = doc, type = "title")
#'
#' doc <- on_slide(doc, index = 2)
#' doc <- ph_remove(x = doc, type = "body", id = 2)
#'
#' doc <- on_slide(doc, index = 3)
#' doc <- ph_remove(x = doc, type = "body", id = 1)
#'
#' print(doc, target = fileout)
#' @family functions for placeholders manipulation
#' @seealso [ph_with()]
ph_remove <- function(
  x,
  type = "body",
  id = 1,
  ph_label = NULL,
  id_chr = NULL
) {
  slide <- x$slide$get_slide(x$cursor)
  office_id <- get_shape_id(x, type = type, id = id, ph_label = ph_label)
  current_elt <- xml_find_first(
    slide$get(),
    sprintf("p:cSld/p:spTree/*[*/p:cNvPr[@id='%s']]", office_id)
  )
  xml_remove(current_elt)

  x
}


#' @export
#' @title Slide link to a placeholder
#' @description Add slide link to a placeholder in the current slide.
#' @inheritParams ph_remove
#' @param slide_index slide index to reach
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' loc_title <- ph_location_type(type = "title")
#' doc <- read_pptx()
#' doc <- add_slide(doc, "Title and Content")
#' doc <- ph_with(x = doc, "Un titre 1", location = loc_title)
#' doc <- add_slide(doc, "Title and Content")
#' doc <- ph_with(x = doc, "Un titre 2", location = loc_title)
#' doc <- on_slide(doc, 1)
#' slide_summary(doc) # read column ph_label here
#' doc <- ph_slidelink(x = doc, ph_label = "Title 1", slide_index = 2)
#'
#' print(doc, target = fileout)
#' @family functions for placeholders manipulation
#' @seealso [ph_with()]
ph_slidelink <- function(
  x,
  type = "body",
  id = 1,
  id_chr = NULL,
  ph_label = NULL,
  slide_index
) {
  slide <- x$slide$get_slide(x$cursor)
  office_id <- get_shape_id(x, type = type, id = id, ph_label = ph_label)
  current_elt <- xml_find_first(
    slide$get(),
    sprintf("p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]", office_id)
  )

  # declare slide ref in relationships
  slide_name <- x$slide$names()[slide_index]
  slide$reference_slide(slide_name)
  rel_df <- slide$rel_df()
  id <- rel_df[rel_df$target == slide_name, "id"]

  # add hlinkClick
  cnvpr <- xml_child(current_elt, "p:nvSpPr/p:cNvPr")
  str_ <- "<a:hlinkClick xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:id=\"%s\" action=\"ppaction://hlinksldjump\"/>"
  str_ <- sprintf(str_, id)
  xml_add_child(cnvpr, as_xml_document(str_))

  x
}


#' @export
#' @title Hyperlink a placeholder
#' @description Add hyperlink to a placeholder in the current slide.
#' @inheritParams ph_remove
#' @param href hyperlink (do not forget http or https prefix)
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' loc_manual <- ph_location(bg = "red", newlabel = "mytitle")
#' doc <- read_pptx()
#' doc <- add_slide(doc, "Title and Content")
#' doc <- ph_with(x = doc, "Un titre 1", location = loc_manual)
#' slide_summary(doc) # read column ph_label here
#' doc <- ph_hyperlink(
#'   x = doc, ph_label = "mytitle",
#'   href = "https://cran.r-project.org"
#' )
#'
#' print(doc, target = fileout)
#' @family functions for placeholders manipulation
#' @seealso [ph_with()]
ph_hyperlink <- function(
  x,
  type = "body",
  id = 1,
  id_chr = NULL,
  ph_label = NULL,
  href
) {
  slide <- x$slide$get_slide(x$cursor)
  office_id <- get_shape_id(x, type = type, id = id, ph_label = ph_label)
  current_elt <- xml_find_first(
    slide$get(),
    sprintf(
      "p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]|p:cSld/p:spTree/*[p:nvPicPr/p:cNvPr[@id='%s']]",
      office_id,
      office_id
    )
  )

  # add hlinkClick
  if (xml_name(current_elt) %in% "pic") {
    cnvpr <- xml_child(current_elt, "p:nvPicPr/p:cNvPr")
  } else {
    cnvpr <- xml_child(current_elt, "p:nvSpPr/p:cNvPr")
  }
  str_ <- "<a:hlinkClick xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:id=\"%s\"/>"
  str_ <- sprintf(str_, href)
  xml_add_child(cnvpr, as_xml_document(str_))
  x
}

wml_str <- function(str) {
  paste0(
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n",
    "<w:document xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">",
    str,
    "</w:document>"
  )
}

pml_str <- function(str) {
  paste0(
    "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n",
    "<a:document xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
    str,
    "</a:document>"
  )
}

has_css_color <- function(x, atname, color) {
  css <- format(x, type = "html")
  reg <- sprintf("%s:%s", atname, color)
  grepl(reg, css)
}

has_css_attr <- function(x, atname, value) {
  css <- format(x, type = "html")
  reg <- sprintf("%s:%s", atname, value)
  grepl(reg, css)
}

source("utils.R")


wml_is_italic <- function(doc_) {
  !inherits(xml_find_first(doc_, "/w:document/w:rPr/w:i"), "xml_missing")
}
wml_is_bold <- function(doc_) {
  !inherits(xml_find_first(doc_, "/w:document/w:rPr/w:b"), "xml_missing")
}
wml_is_underline <- function(doc_) {
  !inherits(xml_find_first(doc_, "/w:document/w:rPr/w:u"), "xml_missing")
}

test_that("wml - font size", {
  fp <- fp_text(font.size = 10)
  xml_ <- format(fp, type = "wml")

  doc <- read_xml(wml_str(xml_))

  sz <- xml_find_first(doc, "/w:document/w:rPr/w:sz")
  szCs <- xml_find_first(doc, "/w:document/w:rPr/w:szCs")
  expect_false(inherits(szCs, "xml_missing"))
  expect_false(inherits(sz, "xml_missing"))

  expect_equal(xml_attr(sz, "val"), xml_attr(szCs, "val"))
  expect_equal(xml_attr(sz, "val"), expected = "20")
})

test_that("wml - bold italic underlined", {
  fp_bold <- fp_text(bold = TRUE)
  fp_italic <- update(fp_bold, bold = FALSE, italic = TRUE)
  fp_bold_italic <- update(fp_bold, italic = TRUE)
  fp_underline <- fp_text(underlined = TRUE)
  fp_srtike <- fp_text(strike = TRUE)

  xml_bold_ <- format(fp_bold, type = "wml")
  xml_italic_ <- format(fp_italic, type = "wml")
  xml_bolditalic_ <- format(fp_bold_italic, type = "wml")
  xml_underline_ <- format(fp_underline, type = "wml")
  xml_strike_ <- format(fp_srtike, type = "wml")

  doc_bold_ <- read_xml(wml_str(xml_bold_))
  doc_italic_ <- read_xml(wml_str(xml_italic_))
  doc_bolditalic_ <- read_xml(wml_str(xml_bolditalic_))
  doc_underline_ <- read_xml(wml_str(xml_underline_))
  doc_strike_ <- read_xml(wml_str(xml_strike_))

  expect_equal(wml_is_bold(doc_bold_), TRUE)
  expect_equal(
    xml_attr(xml_find_first(doc_bold_, "/w:document/w:rPr/w:i"), "val"),
    "false"
  )
  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:b"), "val"),
    "false"
  )
  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:i"), "val"),
    "true"
  )

  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:b"), "val"),
    "false"
  )
  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:i"), "val"),
    "true"
  )
  expect_equal(wml_is_bold(doc_bolditalic_), TRUE)
  expect_equal(wml_is_italic(doc_bolditalic_), TRUE)

  expect_equal(
    xml_attr(xml_find_first(doc_bold_, "/w:document/w:rPr/w:u"), "val"),
    "none"
  )
  expect_equal(
    xml_attr(xml_find_first(doc_underline_, "/w:document/w:rPr/w:u"), "val"),
    "single"
  )
  expect_equal(
    xml_attr(xml_find_first(doc_strike_, "/w:document/w:rPr/w:strike"), "val"),
    "true"
  )
})


test_that("wml - font name", {
  fontname = "Arial"
  fp_ <- fp_text(font.family = fontname)

  xml_ <- format(fp_, type = "wml")
  doc_ <- read_xml(wml_str(xml_))

  node <- xml_find_first(doc_, "/w:document/w:rPr/w:rFonts")
  expect_false(inherits(node, "xml_missing"))

  expect_equal(xml_attr(node, "ascii"), fontname)
  expect_equal(xml_attr(node, "hAnsi"), fontname)
  expect_equal(xml_attr(node, "cs"), fontname)
})

test_that("wml - font color", {
  fp_ <- fp_text(color = grDevices::rgb(.8, .2, .1, .6))

  xml_ <- format(fp_, type = "wml")
  doc_ <- read_xml(wml_str(xml_))

  node <- xml_find_first(doc_, "/w:document/w:rPr/w:color")
  expect_false(inherits(node, "xml_missing"))

  expect_equal(xml_attr(node, "val"), "CC331A")

  node <- xml_find_first(doc_, "/w:document/w:rPr/w14:textFill")
  expect_false(inherits(node, "xml_missing"))
  expect_equal(
    xml_attr(xml_child(node, "w14:solidFill/w14:srgbClr"), "val"),
    "CC331A"
  )
  expect_equal(
    xml_attr(xml_child(node, "w14:solidFill/w14:srgbClr/w14:alpha"), "val"),
    "60000"
  )
})


test_that("wml - shading color", {
  fp_ <- fp_text(shading.color = rgb(1, 0, 1))

  xml_ <- format(fp_, type = "wml")
  doc_ <- read_xml(wml_str(xml_))

  node <- xml_find_first(doc_, "/w:document/w:rPr/w:shd")
  expect_false(inherits(node, "xml_missing"))

  expect_equal(xml_attr(node, "fill"), "FF00FF")
})


pml_has_true_attr <- function(doc_, what = "b") {
  rpr <- xml_find_first(doc_, "/a:document/a:rPr")
  val <- xml_attr(rpr, what)
  !is.na(val) && val == "1"
}
pml_attr <- function(doc_, what = "u") {
  rpr <- xml_find_first(doc_, "/a:document/a:rPr")
  xml_attr(rpr, what)
}


test_that("pml - font size", {
  fp <- fp_text(font.size = 10)
  xml_ <- format(fp, type = "pml")

  doc_ <- read_xml(pml_str(xml_))
  rpr <- xml_find_first(doc_, "/a:document/a:rPr")

  expect_equal(xml_attr(rpr, "sz"), "1000")
})

test_that("pml - bold italic underlined", {
  fp_bold <- fp_text(bold = TRUE)
  fp_italic <- update(fp_bold, bold = FALSE, italic = TRUE)
  fp_bold_italic <- update(fp_bold, italic = TRUE)
  fp_underline <- fp_text(underlined = TRUE)
  fp_srtike <- fp_text(strike = TRUE)

  xml_bold_ <- format(fp_bold, type = "pml")
  xml_italic_ <- format(fp_italic, type = "pml")
  xml_bolditalic_ <- format(fp_bold_italic, type = "pml")
  xml_underline_ <- format(fp_underline, type = "pml")
  xml_strike_ <- format(fp_srtike, type = "pml")

  doc_bold_ <- read_xml(pml_str(xml_bold_))
  doc_italic_ <- read_xml(pml_str(xml_italic_))
  doc_bolditalic_ <- read_xml(pml_str(xml_bolditalic_))
  doc_underline_ <- read_xml(pml_str(xml_underline_))
  doc_strike_ <- read_xml(pml_str(xml_strike_))

  expect_equal(pml_has_true_attr(doc_bold_, "b"), TRUE)
  expect_equal(pml_has_true_attr(doc_bold_, "i"), FALSE)
  expect_equal(pml_has_true_attr(doc_bold_, "u"), FALSE)

  expect_equal(pml_has_true_attr(doc_italic_, "b"), FALSE)
  expect_equal(pml_has_true_attr(doc_italic_, "i"), TRUE)
  expect_equal(pml_has_true_attr(doc_italic_, "u"), FALSE)

  expect_equal(pml_has_true_attr(doc_bolditalic_, "b"), TRUE)
  expect_equal(pml_has_true_attr(doc_bolditalic_, "i"), TRUE)
  expect_equal(pml_has_true_attr(doc_bolditalic_, "u"), FALSE)

  expect_equal(pml_has_true_attr(doc_underline_, "b"), FALSE)
  expect_equal(pml_has_true_attr(doc_underline_, "i"), FALSE)
  expect_equal(pml_attr(doc_underline_, "u"), "sng")

  expect_equal(pml_has_true_attr(doc_strike_, "b"), FALSE)
  expect_equal(pml_has_true_attr(doc_strike_, "i"), FALSE)
  expect_equal(pml_attr(doc_strike_, "strike"), "sngStrike")
})


test_that("pml - font name", {
  fontname = "Arial"
  fp_ <- fp_text(font.family = fontname)

  xml_ <- format(fp_, type = "pml")
  doc_ <- read_xml(pml_str(xml_))

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:latin")
  expect_false(inherits(node, "xml_missing"))
  expect_equal(xml_attr(node, "typeface"), fontname)

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:cs")
  expect_false(inherits(node, "xml_missing"))
  expect_equal(xml_attr(node, "typeface"), fontname)
})

test_that("pml - font color", {
  fp_ <- fp_text(color = rgb(1, 0, 0, .5))

  xml_ <- format(fp_, type = "pml")
  doc_ <- read_xml(pml_str(xml_))

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:solidFill/a:srgbClr")
  expect_false(inherits(node, "xml_missing"))
  expect_equal(xml_attr(node, "val"), "FF0000")

  node <- xml_find_first(
    doc_,
    "/a:document/a:rPr/a:solidFill/a:srgbClr/a:alpha"
  )
  expect_equal(xml_attr(node, "val"), "50196")
})


test_that("css", {
  fp <- fp_text(
    font.size = 10,
    color = "#00FFFF34",
    shading.color = "#00FFFFCC"
  )
  expect_true(has_css_color(fp, "color", "rgba\\(0,255,255,0.20\\)"))
  expect_true(has_css_attr(fp, "font-family", "'Arial'"))
  expect_true(has_css_attr(fp, "font-size", "10.0pt"))
  expect_true(has_css_attr(fp, "font-style", "normal"))
  expect_true(has_css_attr(fp, "font-weight", "normal"))
  expect_true(has_css_color(fp, "background-color", "rgba\\(0,255,255,0.80\\)"))

  fp <- fp_text(bold = TRUE, italic = TRUE, underlined = TRUE)
  expect_true(has_css_attr(fp, "font-style", "italic"))
  expect_true(has_css_attr(fp, "font-weight", "bold"))
  expect_true(has_css_attr(fp, "text-decoration", "underline"))

  fp <- fp_text(strike = TRUE)
  expect_true(has_css_attr(fp, "text-decoration", "line-through"))
})

source("utils.R")


wml_is_italic <- function(doc_){
  !inherits(xml_find_first(doc_, "/w:document/w:rPr/w:i"), "xml_missing")
}
wml_is_bold <- function(doc_){
  !inherits(xml_find_first(doc_, "/w:document/w:rPr/w:b"), "xml_missing")
}
wml_is_underline <- function(doc_){
  !inherits(xml_find_first(doc_, "/w:document/w:rPr/w:u"), "xml_missing")
}

test_that("wml - font size", {
  fp <- fp_text(font.size = 10)
  xml_ <- format(fp, type = "wml")

  doc <- read_xml( wml_str(xml_) )

  sz <- xml_find_first(doc, "/w:document/w:rPr/w:sz")
  szCs <- xml_find_first(doc, "/w:document/w:rPr/w:szCs")
  expect_false(inherits( szCs, "xml_missing") )
  expect_false(inherits( sz, "xml_missing") )

  expect_equal(xml_attr(sz, "val"), xml_attr(szCs, "val"))
  expect_equal(xml_attr(sz, "val"), expected = "20")

})

test_that("wml - bold italic underlined", {
  fp_bold <- fp_text(bold = TRUE)
  fp_italic <- update(fp_bold, bold = FALSE, italic = TRUE)
  fp_bold_italic <- update(fp_bold, italic = TRUE)
  fp_underline <- fp_text(underlined = TRUE )

  xml_bold_ <- format(fp_bold, type = "wml")
  xml_italic_ <- format(fp_italic, type = "wml")
  xml_bolditalic_ <- format(fp_bold_italic, type = "wml")
  xml_underline_ <- format(fp_underline, type = "wml")

  doc_bold_ <- read_xml( wml_str(xml_bold_) )
  doc_italic_ <- read_xml( wml_str(xml_italic_) )
  doc_bolditalic_ <- read_xml( wml_str(xml_bolditalic_) )
  doc_underline_ <- read_xml( wml_str(xml_underline_) )

  expect_equal(wml_is_bold(doc_bold_), TRUE)
  expect_equal(
    xml_attr(xml_find_first(doc_bold_, "/w:document/w:rPr/w:i"), "val"),
    "false")
  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:b"), "val"),
    "false")
  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:i"), "val"),
    "true")

  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:b"), "val"),
    "false")
  expect_equal(
    xml_attr(xml_find_first(doc_italic_, "/w:document/w:rPr/w:i"), "val"),
    "true")
  expect_equal(wml_is_bold(doc_bolditalic_), TRUE)
  expect_equal(wml_is_italic(doc_bolditalic_), TRUE)

  expect_equal(
    xml_attr(xml_find_first(doc_bold_, "/w:document/w:rPr/w:u"), "val"),
    "none")
  expect_equal(
    xml_attr(xml_find_first(doc_underline_, "/w:document/w:rPr/w:u"), "val"),
    "single")
})


test_that("wml - font name", {
  fontname = "Arial"
  fp_ <- fp_text(font.family = fontname)

  xml_ <- format(fp_, type = "wml")
  doc_ <- read_xml( wml_str(xml_) )

  node <- xml_find_first(doc_, "/w:document/w:rPr/w:rFonts")
  expect_false(inherits(node, "xml_missing"))

  expect_equal( xml_attr(node, "ascii"), fontname )
  expect_equal( xml_attr(node, "hAnsi"), fontname )
  expect_equal( xml_attr(node, "cs"), fontname )
})

test_that("wml - font color", {
  fp_ <- fp_text(color = grDevices::rgb(.8, .2, .1, .6))

  xml_ <- format(fp_, type = "wml")
  doc_ <- read_xml( wml_str(xml_) )

  node <- xml_find_first(doc_, "/w:document/w:rPr/w:color")
  expect_false(inherits(node, "xml_missing"))

  expect_equal( xml_attr(node, "val"), "CC331A" )

  node <- xml_find_first(doc_, "/w:document/w:rPr/w14:textFill")
  expect_false(inherits(node, "xml_missing"))
  expect_equal( xml_attr(xml_child(node, "w14:solidFill/w14:srgbClr"), "val"), "CC331A" )
  expect_equal( xml_attr(xml_child(node, "w14:solidFill/w14:srgbClr/w14:alpha"), "val"), "60000" )

})


test_that("wml - shading color", {
  fp_ <- fp_text(shading.color = rgb(1,0,1))

  xml_ <- format(fp_, type = "wml")
  doc_ <- read_xml( wml_str(xml_) )

  node <- xml_find_first(doc_, "/w:document/w:rPr/w:shd")
  expect_false(inherits(node, "xml_missing"))

  expect_equal( xml_attr(node, "fill"), "FF00FF" )
})



pml_has_true_attr <- function(doc_, what = "b"){
  rpr <- xml_find_first(doc_, "/a:document/a:rPr")
  val <- xml_attr(rpr, what)
  !is.na( val ) && val == "1"
}
pml_attr <- function(doc_, what = "u"){
  rpr <- xml_find_first(doc_, "/a:document/a:rPr")
  xml_attr(rpr, what)
}


test_that("pml - font size", {
  fp <- fp_text(font.size = 10)
  xml_ <- format(fp, type = "pml")

  doc_ <- read_xml( pml_str(xml_) )
  rpr <- xml_find_first(doc_, "/a:document/a:rPr")

  expect_equal(xml_attr(rpr, "sz"), "1000")
})

test_that("pml - bold italic underlined", {
  fp_bold <- fp_text(bold = TRUE)
  fp_italic <- update(fp_bold, bold = FALSE, italic = TRUE)
  fp_bold_italic <- update(fp_bold, italic = TRUE)
  fp_underline <- fp_text(underlined = TRUE )

  xml_bold_ <- format(fp_bold, type = "pml")
  xml_italic_ <- format(fp_italic, type = "pml")
  xml_bolditalic_ <- format(fp_bold_italic, type = "pml")
  xml_underline_ <- format(fp_underline, type = "pml")

  doc_bold_ <- read_xml( pml_str(xml_bold_) )
  doc_italic_ <- read_xml( pml_str(xml_italic_) )
  doc_bolditalic_ <- read_xml( pml_str(xml_bolditalic_) )
  doc_underline_ <- read_xml( pml_str(xml_underline_) )

  expect_equal(pml_has_true_attr(doc_bold_, "b"), TRUE)
  expect_equal(pml_has_true_attr(doc_bold_, "i"), FALSE)
  expect_equal(pml_has_true_attr(doc_bold_, "u"), FALSE)

  expect_equal(pml_has_true_attr(doc_italic_, "b"), FALSE)
  expect_equal(pml_has_true_attr(doc_italic_, "i"), TRUE)
  expect_equal(pml_has_true_attr(doc_italic_, "u"), FALSE)

  expect_equal(pml_has_true_attr(doc_bolditalic_, "b"), TRUE)
  expect_equal(pml_has_true_attr(doc_bolditalic_, "i"), TRUE)
  expect_equal(pml_has_true_attr(doc_bolditalic_, "u"), FALSE)

  expect_equal(pml_has_true_attr(doc_underline_, "b"), FALSE)
  expect_equal(pml_has_true_attr(doc_underline_, "i"), FALSE)
  expect_equal(pml_attr(doc_underline_, "u"), "sng")


})


test_that("pml - font name", {
  fontname = "Arial"
  fp_ <- fp_text(font.family = fontname)

  xml_ <- format(fp_, type = "pml")
  doc_ <- read_xml( pml_str(xml_) )

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:latin")
  expect_false(inherits(node, "xml_missing"))
  expect_equal( xml_attr(node, "typeface"), fontname )

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:cs")
  expect_false(inherits(node, "xml_missing"))
  expect_equal( xml_attr(node, "typeface"), fontname )

})

test_that("pml - font color", {
  fp_ <- fp_text(color= rgb(1, 0, 0, .5 ))

  xml_ <- format(fp_, type = "pml")
  doc_ <- read_xml( pml_str(xml_) )

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:solidFill/a:srgbClr")
  expect_false(inherits(node, "xml_missing"))
  expect_equal( xml_attr(node, "val"), "FF0000" )

  node <- xml_find_first(doc_, "/a:document/a:rPr/a:solidFill/a:srgbClr/a:alpha")
  expect_equal( xml_attr(node, "val"), "50196" )

})






test_that("css", {
  fp <- fp_text(font.size = 10, color = "#00FFFF34", shading.color = "#00FFFFCC")
  expect_true(has_css_color(fp, "color", "rgba\\(0,255,255,0.20\\)"))
  expect_true(has_css_attr(fp, "font-family", "'Arial'"))
  expect_true(has_css_attr(fp, "font-size", "10.0pt"))
  expect_true(has_css_attr(fp, "font-style", "normal"))
  expect_true(has_css_attr(fp, "font-weight", "normal"))
  expect_true(has_css_attr(fp, "text-decoration", "none"))
  expect_true(has_css_color(fp, "background-color", "rgba\\(0,255,255,0.80\\)"))

  fp <- fp_text(bold = TRUE, italic = TRUE, underlined = TRUE)
  expect_true(has_css_attr(fp, "font-style", "italic"))
  expect_true(has_css_attr(fp, "font-weight", "bold"))
  expect_true(has_css_attr(fp, "text-decoration", "underline"))
})

source("utils.R")

test_that("fp_cell", {
  expect_error(fp_cell(text.direction = "sdqsd"), "must be one of")
  expect_error(fp_cell(margin = -2), "must be a positive integer scalar")
  expect_error(fp_cell(border = fp_text()), "border must be a fp_border object")
  expect_error(
    fp_cell(border.bottom = fp_text()),
    "border.bottom must be a fp_border object"
  )
  expect_error(
    fp_cell(background.color = "#340F2A78d"),
    " must be a valid color"
  )
  x <- fp_cell()
  x <- update(
    x,
    margin = 3,
    background.color = "#340F2A78",
    border = fp_border(color = "red")
  )
  expect_equal(x$background.color, "#340F2A78")
  expect_equal(x$margin.bottom, 3)
  expect_equal(x$margin.left, 3)
  expect_equal(x$border.bottom, fp_border(color = "red"))
})


test_that("print fp_cell", {
  x <- fp_cell()
  expect_output(print(x))
})


pml_cell_node <- function(x) {
  xml_ <- pml_str(format(x, type = "pml"))
  doc <- read_xml(xml_)
  xml_find_first(doc, "//a:tcPr")
}


test_that("pml fp_border", {
  node <- pml_cell_node(fp_cell(background.color = "#00FF0099"))

  col <- xml_attr(xml_child(node, "a:solidFill/a:srgbClr"), "val")
  expect_equal(col, "00FF00")

  alpha <- xml_attr(xml_child(node, "a:solidFill/a:srgbClr/a:alpha"), "val")
  expect_equal(alpha, "60000")

  border_nodes <- xml_find_all(
    node,
    "//*[self::a:lnL or self::a:lnR or self::a:lnT or self::a:lnB]"
  )
  expect_length(border_nodes, 4)

  expect_equal(xml_attr(node, "marB"), "0")
  expect_equal(xml_attr(node, "marT"), "0")
  expect_equal(xml_attr(node, "marR"), "0")
  expect_equal(xml_attr(node, "marL"), "0")

  node <- pml_cell_node(fp_cell(margin = 2, margin.bottom = 4))
  expect_equal(as.integer(xml_attr(node, "marB")), 12700 * 4)
  expect_equal(as.integer(xml_attr(node, "marT")), 12700 * 2)
  expect_equal(as.integer(xml_attr(node, "marR")), 12700 * 2)
  expect_equal(as.integer(xml_attr(node, "marL")), 12700 * 2)

  node <- pml_cell_node(fp_cell(vertical.align = "top"))
  expect_equal(xml_attr(node, "anchor"), "t")
  node <- pml_cell_node(fp_cell(vertical.align = "center"))
  expect_equal(xml_attr(node, "anchor"), "ctr")
  node <- pml_cell_node(fp_cell(vertical.align = "bottom"))
  expect_equal(xml_attr(node, "anchor"), "b")

  x <- fp_cell(text.direction = "btlr")
  node <- pml_cell_node(x)
  expect_equal(xml_attr(node, "vert"), "vert270")
  x <- fp_cell(text.direction = "tbrl")
  node <- pml_cell_node(x)
  expect_equal(xml_attr(node, "vert"), "vert")
})


wml_cell_node <- function(x) {
  xml_ <- wml_str(format(x, type = "wml"))
  doc <- read_xml(xml_)
  xml_find_first(doc, "//w:tcPr")
}

test_that("wml fp_border", {
  node <- wml_cell_node(fp_cell(background.color = "#00FF0099", margin = 2))

  col <- xml_attr(xml_child(node, "w:shd"), "fill")
  expect_equal(col, "00FF00")

  margins <- xml_children(xml_child(node, "w:tcMar"))
  margins <- sapply(margins, function(x) xml_attr(x, "w"))
  margins <- as.integer(margins)
  expect_equal(margins, rep(40, 4))

  node <- wml_cell_node(fp_cell(margin = 2, margin.bottom = 0))
  margins <- xml_children(xml_child(node, "w:tcMar"))
  margins <- sapply(margins, function(x) xml_attr(x, "w"))
  margins <- as.integer(margins)
  expect_equal(margins, c(40, 0, 40, 40))

  node <- wml_cell_node(fp_cell(border = fp_border()))
  border_nodes <- xml_find_all(
    node,
    "w:tcBorders/*[self::w:bottom or self::w:top or self::w:left or self::w:right]"
  )
  expect_length(border_nodes, 4)

  node <- wml_cell_node(fp_cell(vertical.align = "top"))
  valign <- xml_attr(xml_child(node, "w:vAlign"), "val")
  expect_equal(valign, "top")
  node <- wml_cell_node(fp_cell(vertical.align = "center"))
  valign <- xml_attr(xml_child(node, "w:vAlign"), "val")
  expect_equal(valign, "center")
  node <- wml_cell_node(fp_cell(vertical.align = "bottom"))
  valign <- xml_attr(xml_child(node, "w:vAlign"), "val")
  expect_equal(valign, "bottom")

  x <- fp_cell(text.direction = "btlr")
  node <- wml_cell_node(x)
  td <- xml_attr(xml_child(node, "w:textDirection"), "val")
  expect_equal(td, "btLr")
  x <- fp_cell(text.direction = "tbrl")
  node <- wml_cell_node(x)
  td <- xml_attr(xml_child(node, "w:textDirection"), "val")
  expect_equal(td, "tbRl")
})


test_that("css fp_border", {
  x <- fp_cell(background.color = "#00FF0099", margin = 2)
  expect_true(has_css_color(x, "background-color", "rgba\\(0,255,0,0.60\\)"))

  expect_true(has_css_attr(x, "margin-top", "2pt"))
  expect_true(has_css_attr(x, "margin-bottom", "2pt"))
  expect_true(has_css_attr(x, "margin-left", "2pt"))
  expect_true(has_css_attr(x, "margin-right", "2pt"))

  x <- fp_cell(margin = 2, margin.bottom = 0)
  expect_true(has_css_attr(x, "margin-top", "2pt"))
  expect_true(has_css_attr(x, "margin-bottom", "0pt"))
  expect_true(has_css_attr(x, "margin-left", "2pt"))
  expect_true(has_css_attr(x, "margin-right", "2pt"))

  x <- fp_cell(vertical.align = "top")
  expect_true(has_css_attr(x, "vertical-align", "top"))
  x <- fp_cell(vertical.align = "center")
  expect_true(has_css_attr(x, "vertical-align", "middle"))
  x <- fp_cell(vertical.align = "bottom")
  expect_true(has_css_attr(x, "vertical-align", "bottom"))
})

source("utils.R")

test_that("fp_cell", {
  expect_error( fp_cell(text.direction = "sdqsd"), "must be one of" )
  expect_error( fp_cell(margin = -2), "must be a positive integer scalar" )
  expect_error( fp_cell(border = fp_text()), "border must be a fp_border object" )
  expect_error( fp_cell(border.bottom = fp_text()), "border.bottom must be a fp_border object" )
  expect_error( fp_cell(background.color = "#340F2A78d"), " must be a valid color" )
  x <- fp_cell()
  x <- update(x, margin = 3, background.color = "#340F2A78",
              border = fp_border(color = "red"))
  expect_equal(x$background.color, "#340F2A78")
  expect_equal(x$margin.bottom, 3)
  expect_equal(x$margin.left, 3)
  expect_equal(x$border.bottom, fp_border(color = "red"))
})



test_that("print fp_cell", {
  x <- fp_cell()
  expect_output(print(x))
})



pml_cell_node <- function(x){
  xml_ <- pml_str(format(x, type = "pml"))
  doc <- read_xml( xml_ )
  xml_find_first(doc, "//a:tcPr")
}



test_that("pml fp_border", {

  node <- pml_cell_node(fp_cell(background.color = "#00FF0099"))

  col <- xml_attr(xml_child(node, "a:solidFill/a:srgbClr"), "val")
  expect_equal(col, "00FF00")

  alpha <- xml_attr(xml_child(node, "a:solidFill/a:srgbClr/a:alpha"), "val")
  expect_equal(alpha, "60000")

  border_nodes <- xml_find_all(node, "//*[self::a:lnL or self::a:lnR or self::a:lnT or self::a:lnB]")
  expect_length(border_nodes, 4)

  expect_equal(xml_attr(node, "marB"), "0")
  expect_equal(xml_attr(node, "marT"), "0")
  expect_equal(xml_attr(node, "marR"), "0")
  expect_equal(xml_attr(node, "marL"), "0")

  node <- pml_cell_node(fp_cell(margin = 2, margin.bottom = 4))
  expect_equal(as.integer(xml_attr(node, "marB")), 12700 * 4)
  expect_equal(as.integer(xml_attr(node, "marT")), 12700 * 2)
  expect_equal(as.integer(xml_attr(node, "marR")), 12700 * 2)
  expect_equal(as.integer(xml_attr(node, "marL")), 12700 * 2)

  node <- pml_cell_node(fp_cell(vertical.align = "top"))
  expect_equal(xml_attr(node, "anchor"), "t")
  node <- pml_cell_node(fp_cell(vertical.align = "center"))
  expect_equal(xml_attr(node, "anchor"), "ctr")
  node <- pml_cell_node(fp_cell(vertical.align = "bottom"))
  expect_equal(xml_attr(node, "anchor"), "b")

  x <- fp_cell(text.direction = "btlr")
  node <- pml_cell_node(x)
  expect_equal(xml_attr(node, "vert"), "vert270")
  x <- fp_cell(text.direction = "tbrl")
  node <- pml_cell_node(x)
  expect_equal(xml_attr(node, "vert"), "vert")
})




wml_cell_node <- function(x){
  xml_ <- wml_str(format(x, type = "wml"))
  doc <- read_xml( xml_ )
  xml_find_first(doc, "//w:tcPr")
}

test_that("wml fp_border", {

  node <- wml_cell_node(fp_cell(background.color = "#00FF0099", margin = 2))

  col <- xml_attr(xml_child(node, "w:shd"), "fill")
  expect_equal(col, "00FF00")

  margins <- xml_children(xml_child(node, "w:tcMar"))
  margins <- sapply(margins, function(x) xml_attr(x, "w"))
  margins <- as.integer(margins)
  expect_equal( margins, rep(40, 4) )

  node <- wml_cell_node(fp_cell(margin = 2, margin.bottom = 0))
  margins <- xml_children(xml_child(node, "w:tcMar"))
  margins <- sapply(margins, function(x) xml_attr(x, "w"))
  margins <- as.integer(margins)
  expect_equal( margins, c(40, 0, 40, 40) )

  node <- wml_cell_node(fp_cell(border = fp_border()))
  border_nodes <- xml_find_all(node, "w:tcBorders/*[self::w:bottom or self::w:top or self::w:left or self::w:right]")
  expect_length(border_nodes, 4)

  node <- wml_cell_node(fp_cell(vertical.align = "top"))
  valign <- xml_attr(xml_child(node, "w:vAlign"), "val")
  expect_equal(valign, "top")
  node <- wml_cell_node(fp_cell(vertical.align = "center"))
  valign <- xml_attr(xml_child(node, "w:vAlign"), "val")
  expect_equal(valign, "center")
  node <- wml_cell_node(fp_cell(vertical.align = "bottom"))
  valign <- xml_attr(xml_child(node, "w:vAlign"), "val")
  expect_equal(valign, "bottom")

  x <- fp_cell(text.direction = "btlr")
  node <- wml_cell_node(x)
  td <- xml_attr(xml_child(node, "w:textDirection"), "val")
  expect_equal(td, "btLr")
  x <- fp_cell(text.direction = "tbrl")
  node <- wml_cell_node(x)
  td <- xml_attr(xml_child(node, "w:textDirection"), "val")
  expect_equal(td, "tbRl")
})


test_that("css fp_border", {

  x <- fp_cell(background.color = "#00FF0099", margin = 2)
  expect_true(has_css_color(x, "background-color", "rgba\\(0,255,0,0.60\\)"))

  expect_true(has_css_attr(x, "margin-top", "2pt"))
  expect_true(has_css_attr(x, "margin-bottom", "2pt"))
  expect_true(has_css_attr(x, "margin-left", "2pt"))
  expect_true(has_css_attr(x, "margin-right", "2pt"))

  x <- fp_cell(margin = 2, margin.bottom = 0)
  expect_true(has_css_attr(x, "margin-top", "2pt"))
  expect_true(has_css_attr(x, "margin-bottom", "0pt"))
  expect_true(has_css_attr(x, "margin-left", "2pt"))
  expect_true(has_css_attr(x, "margin-right", "2pt"))

  x <- fp_cell(vertical.align = "top")
  expect_true(has_css_attr(x, "vertical-align", "top"))
  x <- fp_cell(vertical.align = "center")
  expect_true(has_css_attr(x, "vertical-align", "middle"))
  x <- fp_cell(vertical.align = "bottom")
  expect_true(has_css_attr(x, "vertical-align", "bottom"))

})

library(officer)

doc <- read_docx()
doc <- body_add_par(doc, "a paragraph to replace", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
doc <- body_replace_text_at_bkm(doc, "text_to_replace", "new text")


# demo usage of bookmark and images ----
template <- system.file(package = "officer", "doc_examples/example.docx")

img.file <- file.path(R.home("doc"), "html", "logo.jpg")

doc <- read_docx(path = template)
doc <- headers_replace_img_at_bkm(
  x = doc,
  bookmark = "bmk_header",
  value = external_img(src = img.file, width = .53, height = .7)
)
doc <- footers_replace_img_at_bkm(
  x = doc,
  bookmark = "bmk_footer",
  value = external_img(src = img.file, width = .53, height = .7)
)
print(doc, target = tempfile(fileext = ".docx"))

# dir_collection ---------------------------------------------------------
dir_collection <- R6Class(
  "dir_collection",
  public = list(

    initialize = function( package_dir, container ) {
      dir_ <- file.path(package_dir, container$dir_name())
      private$package_dir <- package_dir
      filenames <- list.files(path = dir_, pattern = "\\.xml$", full.names = TRUE)
      filenames <- sort_vec_by_index(filenames)  # see issue #596
      private$collection <- lapply( filenames, function(x, container){
        container$clone()$feed(x)
      }, container = container)
      names(private$collection) <- basename(filenames)
    },

    collection_get = function(name){
      private$collection[[name]]
    },

    get_metadata = function(){
      dat <- lapply(private$collection, function(x) x$get_metadata())
      rbind_match_columns(dat)
    },
    names = function(){
      sapply(private$collection, function(x) x$name())
    },
    xfrm = function( ){
      dat <- lapply(private$collection, function(x) x$xfrm() )
      do.call(rbind, dat)
    }
  ),

  private = list(

    collection = NULL,
    package_dir = NULL

  )
)


# dir_layout ---------------------------------------------------------
dir_layout <- R6Class(
  "dir_layout",
  inherit = dir_collection,
  public = list(
    initialize = function( package_dir, master_metadata, master_xfrm ) {
      super$initialize(package_dir, slide_layout$new("ppt/slideLayouts"))
      private$master_metadata <- master_metadata
      private$xfrm_data <- xfrmize(self$xfrm(), master_xfrm)
    },

    get_xfrm_data = function(){
      private$xfrm_data
    },

    get_metadata = function( ){
      data_layouts <- super$get_metadata()
      data_masters <- private$master_metadata
      data_masters$master_file <- basename(data_masters$filename)
      data_masters$filename <- NULL
      data_layouts$master_file <- basename(data_layouts$master_file)
      out <- merge(data_layouts, data_masters, by = "master_file", all = FALSE)
      out$filename <- basename(out$filename)
      out
    }

  ),
  private = list(
    master_collection = NULL,
    master_metadata = NULL,
    xfrm_data = NULL
  )
)


# dir_slide ---------------------------------------------------------
dir_slide <- R6Class(
  "dir_slide",
  inherit = dir_collection,
  public = list(

    initialize = function( package_dir, xfrm_layout_data ) {
      super$initialize(package_dir, slide$new("ppt/slides"))
      private$collection <- lapply(private$collection, function(x, ref) x$set_layout_xfrm(ref), ref = xfrm_layout_data )
      names(private$collection) <- sapply(private$collection, function(x) x$name() )
      private$slides_list <- private$get_slide_list()

    },

    add_slide = function(slide_file, xfrm_layout_data){

      slide <- slide$new("ppt/slides")
      slide$feed(slide_file)
      slide$set_layout_xfrm(xfrm_layout_data)

      collect <- private$collection
      new_elt <- list(slide)
      names(new_elt) <- basename(slide_file)
      collect <- append(collect, new_elt)

      sl_id <- as.integer( gsub( "(slide)([0-9]+)(\\.xml)$", "\\2", names(collect) ) )
      private$collection <- collect[order(sl_id)]
      private$slides_list <- names(private$collection)

      self
    },
    slide_index = function( name ){
      which( private$slides_list %in% name )
    },

    remove_slide = function(index ){
      slide_obj <- private$collection[[index]]
      private$collection <- private$collection[-index]
      private$slides_list <- names(private$collection)
      slide_obj$remove()
    },

    save_slides = function(){
      lapply( private$collection, function(x){
        x$save()
      } )
      self
    },

    get_xfrm = function( ){
      lapply(private$collection, function(x) x$get_xfrm() )
    },


    get_slide = function(id){
      l_ <- self$length()
      if( is.null(id) || !between(id, 1, l_ ) ){
        stop("unvalid id ", id, " (", l_," slide(s))", call. = FALSE)
      }
      index <- which( names(private$collection) == private$slides_list[id])
      private$collection[[index]]
    },

    get_metadata = function(){
      super$get_metadata()
    },

    length = function(){
      length(private$collection)
    },

    get_new_slidename = function(){
      slide_dir <- file.path(private$package_dir, "ppt/slides")
      if( !file.exists(slide_dir)){
        dir.create(file.path(slide_dir, "_rels"), showWarnings = FALSE, recursive = TRUE)
      }

      slide_files <- names(private$collection)
      slidename <- "slide1.xml"
      if( length(slide_files)){
        slide_index <- as.integer(gsub("^(slide)([0-9]+)(\\.xml)$", "\\2", slide_files ))
        slidename <- gsub(pattern = "[0-9]+", replacement = max(slide_index) + 1, slidename)
      }
      slidename
    }
  ),
  private = list(
    slides_list = NULL,

    get_slide_list = function(){
      slide_dir <- file.path(private$package_dir, "ppt/slides")
      slide_files <- list.files(slide_dir, pattern = "\\.xml$")
      slide_index <- seq_along(slide_files)
      if( length(slide_files)){
        slide_files <- basename( slide_files )
        slide_index <- as.integer(gsub("^(slide)([0-9]+)(\\.xml)$", "\\2", slide_files ))
        slide_files <- slide_files[order(slide_index)]
      }
      slide_files
    }


  )
)


# dir_master ---------------------------------------------------------

dir_master <- R6Class(
  "dir_master",
  inherit = dir_collection,
  public = list(

    get_metadata = function( ){
      unames <- sapply(private$collection, function(x) x$name())
      ufnames <- sapply(private$collection, function(x) x$file_name())
      data.frame(stringsAsFactors = FALSE, master_name = unames, filename = ufnames)
    },
    get_color_scheme = function( ){
      dat <- lapply(private$collection, function(x) x$colors())
      rbind_match_columns(dat)
    }

  )
)

# dir_notesMaster ---------------------------------------------------------

dir_notesMaster <- R6Class(
  "dir_notesMaster",
  inherit = dir_collection,
  public = list(

    get_metadata = function( ){
      unames <- sapply(private$collection, function(x) x$name())
      ufnames <- sapply(private$collection, function(x) x$file_name())
      data.frame(stringsAsFactors = FALSE, master_name = unames, filename = ufnames)
    },

    add_notesMaster = function(file){

      nmaster <- slide_master$new("ppt/notesMasters")
      nmaster$feed(file)

      collect <- private$collection
      new_elt <- list(nmaster)
      names(new_elt) <- basename(file)
      collect <- append(collect, new_elt)

      sl_id <- as.integer( gsub( "(notesMaster)([0-9]+)(\\.xml)$", "\\2", names(collect) ) )
      private$collection <- collect[order(sl_id)]

      self
    }

  )
)


# dir_notesSlide ---------------------------------------------------------
dir_notesSlide <- R6Class(
  "dir_notesSlide",
  inherit = dir_collection,
  public = list(

    initialize = function( package_dir ) {
      super$initialize(package_dir, notesSlide$new("ppt/notesSlides"))


    },

    add_slide = function(slide_file){

      nslide <- notesSlide$new("ppt/notesSlides")
      nslide$feed(slide_file)

      collect <- private$collection
      new_elt <- list(nslide)
      names(new_elt) <- basename(slide_file)
      collect <- append(collect, new_elt)

      sl_id <- as.integer( gsub( "(notesSlide)([0-9]+)(\\.xml)$", "\\2", names(collect) ) )
      private$collection <- collect[order(sl_id)]

      self
    },
    slide_index = function( name ){
      which( names(private$collection) %in% name )
    },

    remove_slide = function(index ){
      slide_obj <- private$collection[[index]]
      private$collection <- private$collection[-index]
      slide_obj$remove()
    },

    save_slides = function(){
      lapply( private$collection, function(x){
        x$save()
      } )
      self
    },


    get_slide = function(id){
      l_ <- self$length()
      if( is.null(id) || !between(id, 1, l_ ) ){
        stop("unvalid id ", id, " (", l_," slide(s))", call. = FALSE)
      }
      private$collection[[id]]
    },

    get_metadata = function(){
      super$get_metadata()
    },

    length = function(){
      length(private$collection)
    },

    get_new_slidename = function(){
      slide_dir <- file.path(private$package_dir, "ppt/notesSlides")
      if( !file.exists(slide_dir)){
        dir.create(file.path(slide_dir, "_rels"), showWarnings = FALSE, recursive = TRUE)
      }

      slide_files <- names(private$collection)
      slidename <- "notesSlide1.xml"
      if( length(slide_files)){
        slide_index <- as.integer(gsub("^(notesSlide)([0-9]+)(\\.xml)$", "\\2", slide_files ))
        slidename <- gsub(pattern = "[0-9]+", replacement = max(slide_index) + 1, slidename)
      }
      slidename
    }
  )
)

library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, "Hello left", location = ph_location_left())
doc <- ph_with(doc, "Hello right", location = ph_location_right())
print(doc, target = tempfile(fileext = ".pptx"))

# PH_WITH -------

#' @export
#' @title Add objects on the current slide
#' @description add an object into a new shape in the current slide. This
#' function is able to add all supported outputs to a presentation. See
#' section **Methods (by class)** to see supported outputs.
#' @param x an rpptx object
#' @param value object to add as a new shape. Supported objects
#' are vectors, data.frame, graphics, block of formatted paragraphs,
#' unordered list of formatted paragraphs,
#' pretty tables with package flextable, editable graphics with
#' package rvg, 'Microsoft' charts with package mschart.
#' @param location a placeholder location object or a location short form. It will be used
#' to specify the location of the new shape. This location can be defined with a call to one
#' of the `ph_location_*` functions (see section `"see also"`). In `ph_with()`, several location
#' short forms can be used, as listed in section `"Short forms"`.
#' @param ... further arguments passed to or from other methods. When
#' adding a `ggplot` object or `plot_instr`, these arguments will be used
#' by the png function.
#'
#' @section Short forms:
#' The `location` argument of `ph_with()` either expects a location object as returned by the
#' `ph_location_*` functions or a corresponding location *short form* (string or numeric):
#'
#' | **Location function**                       | **Short form**             | **Description**                                      |
#' |---------------------------------------------|----------------------------|------------------------------------------------------|
#' | `ph_location_left()`                        | `"left"`                   | Keyword string                                       |
#' | `ph_location_right()`                       | `"right"`                  | Keyword string                                       |
#' | `ph_location_fullsize()`                    | `"fullsize"`               | Keyword string                                       |
#' | `ph_location_type("body", 1)`               | `"body [1]"`               | String: type + index in brackets (`1` if omitted)    |
#' | `ph_location_label("my_label")`             | `"my_label"`               | Any string not matching a keyword or type            |
#' | `ph_location_id(1)`                         | `1`                        | Length 1 integer                                     |
#' | `ph_location(0, 0, 4, 5)`                   | `c(0,0,4,5)`               | Length 4 numeric, optionally named, `c(top=0, left=0, ...)` |
#'
#' @example inst/examples/example_ph_with.R
#' @seealso Specify placeholder locations with [ph_location_type], [ph_location],
#' [ph_location_label], [ph_location_left], [ph_location_right],
#' [ph_location_fullsize], [ph_location_template]. [phs_with] is a sibling of
#' `ph_with` that fills multiple placeholders at once. Use [add_slide] to add new slides.
#' @section Illustrations:
#'
#' \if{html}{\figure{ph_with_doc_1.png}{options: style="width:80\%;"}}
ph_with <- function(x, value, location, ...) {
  location <- resolve_location(location)
  .ph_with(x, value, location, ...)
}


.ph_with <- function(x, value, location, ...) {
  UseMethod("ph_with", value)
}


#' @export
#' @describeIn ph_with add a character vector to a new shape on the
#' current slide, values will be added as paragraphs.
ph_with.character <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)
  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )

  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}

#' @export
#' @param format_fun format function for non character vectors
#' @describeIn ph_with add a numeric vector to a new shape on the
#' current slide, values will be be first formatted then
#' added as paragraphs.
ph_with.numeric <- function(x, value, location, format_fun = format, ...) {
  slide <- x$slide$get_slide(x$cursor)
  value <- format_fun(value, ...)
  location <- fortify_location(location, doc = x)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}

#' @export
#' @describeIn ph_with add a factor vector to a new shape on the
#' current slide, values will be be converted as character and then
#' added as paragraphs.
ph_with.factor <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)
  value <- as.character(value)
  location <- fortify_location(location, doc = x)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}
#' @export
#' @rdname ph_with
ph_with.logical <- ph_with.numeric


#' @export
#' @param date_format A format string for dates (default `"%Y-%m-%d"`).
#' See `format` arg in [strftime()] for details.
#' Set a global default via `options(officer.date_format = ...)`.
#' @describeIn ph_with add a `Date` object vector to a new shape on the
#' current slide, values will be be first converted to character.
ph_with.Date <- function(x, value, location, date_format = NULL, ...) {
  opt <- options()$officer.date_format
  if (!is.null(opt) && is.na(opt)) {
    # catch NA
    opt <- NULL
  }
  format <- date_format %||% opt %||% "%Y-%m-%d" # fallback to format.Date() default
  value_str <- format(value, format = format)
  ph_with(x, value = value_str, location, ...)
}


#' @export
#' @param level_list The list of levels for hierarchy structure as integer values.
#' If used the object is formated as an unordered list. If 1 and 2,
#' item 1 level will be 1, item 2 level will be 2.
#' @describeIn ph_with add a [block_list()] made
#' of [fpar()] to a new shape on the current slide.
ph_with.block_list <- function(
  x,
  value,
  location,
  level_list = integer(0),
  ...
) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)

  pars <- sapply(value, to_pml)

  if (length(level_list) > 0) {
    pars <- gsub("<a:buNone/>", "", pars, fixed = TRUE)

    level_values <- rep(1L, length(pars))
    level_values[] <- level_list
    level_values <- level_values - 1L

    lvl <- sprintf(" lvl=\"%.0f\"", level_values)
    lvl <- paste0("<a:pPr", ifelse(level_values > 0, lvl, ""), "/>")

    pars <- mapply(
      function(par, lvl) {
        paste0(par[1], lvl, par[2])
      },
      strsplit(pars, split = "<a:pPr(.*)</a:pPr>"),
      lvl,
      SIMPLIFY = FALSE
    )
    pars <- unlist(pars)
  }

  pars <- paste0(pars, collapse = "")

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )

  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}


#' @export
#' @describeIn ph_with add a [unordered_list()] made
#' of [fpar()] to a new shape on the current slide.
ph_with.unordered_list <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)
  location <- fortify_location(location, doc = x)

  p <- to_pml(value)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    p,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}


#' @export
#' @param header display header if TRUE
#' @param tcf conditional formatting settings defined by [table_conditional_formatting()]
#' @param alignment alignment for each columns, 'l' for left, 'r' for right
#' and 'c' for center. Default to NULL.
#' @describeIn ph_with add a data.frame to a new shape on the current slide with
#' function [block_table()]. Use package 'flextable' instead for more
#' advanced formattings.
ph_with.data.frame <- function(
  x,
  value,
  location,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
) {
  location <- fortify_location(location, doc = x)

  slide <- x$slide$get_slide(x$cursor)
  style_id <- x$table_styles$def[1]

  pt <- prop_table(
    style = style_id,
    layout = table_layout(),
    width = table_width(),
    tcf = tcf
  )

  bt <- block_table(
    x = value,
    header = header,
    properties = pt,
    alignment = alignment
  )

  xml_elt <- to_pml(
    bt,
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  value <- as_xml_document(xml_elt)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), value)
  x
}


#' @export
#' @describeIn ph_with add a ggplot object to a new shape on the
#' current slide. Use package 'rvg' for more advanced graphical features.
#' @param res resolution of the png image in ppi
#' @param alt_text Alt-text for screen-readers. Defaults to `""`. If `""` or `NULL`
#'    an alt text added with `ggplot2::labs(alt = ...)` will be used if any.
#' @param scale Multiplicative scaling factor, same as in ggsave
ph_with.gg <- function(
  x,
  value,
  location,
  res = 300,
  alt_text = "",
  scale = 1,
  ...
) {
  location_ <- fortify_location(location, doc = x)
  slide <- x$slide$get_slide(x$cursor)
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  slide <- x$slide$get_slide(x$cursor)
  width <- location_$width
  height <- location_$height

  stopifnot(inherits(value, "gg"))
  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  print(value)
  dev.off()
  on.exit(unlink(file))

  if (is.null(alt_text) || alt_text == "") {
    alt_text <- ggplot2::get_alt_text(value)
    if (is.null(alt_text)) alt_text <- ""
  }

  ext_img <- external_img(file, width = width, height = height, alt = alt_text)

  ph_with(x, ext_img, location = location)
}

#' @export
#' @describeIn ph_with add an R plot to a new shape on the
#' current slide. Use package 'rvg' for more advanced graphical features.
ph_with.plot_instr <- function(x, value, location, res = 300, ...) {
  location_ <- fortify_location(location, doc = x)
  slide <- x$slide$get_slide(x$cursor)
  slide <- x$slide$get_slide(x$cursor)
  width <- location_$width
  height <- location_$height

  file <- tempfile(fileext = ".png")

  dirname <- tempfile()
  dir.create(dirname)
  filename <- paste(dirname, "/plot%03d.png", sep = "")
  agg_png(
    filename = filename,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = 1,
    background = "transparent",
    ...
  )

  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )
  file <- list.files(dirname, full.names = TRUE)
  on.exit(unlink(dirname, recursive = TRUE, force = TRUE))

  if (length(file) > 1) {
    stop(
      length(file),
      " files have been produced. Multiple plot are not supported"
    )
  }

  ext_img <- external_img(file, width = width, height = height)
  ph_with(x, ext_img, location = location)
}


#' @export
#' @param use_loc_size if set to FALSE, external_img width and height will
#' be used.
#' @describeIn ph_with add a [external_img()] to a new shape
#' on the current slide.
#'
#' When value is a external_img object, image will be copied
#' into the PowerPoint presentation. The width and height
#' specified in call to [external_img()] will be
#' ignored, their values will be those of the location,
#' unless use_loc_size is set to FALSE.
ph_with.external_img <- function(x, value, location, use_loc_size = TRUE, ...) {
  location <- fortify_location(location, doc = x)

  slide <- x$slide$get_slide(x$cursor)

  if (!use_loc_size) {
    location$width <- attr(value, "dims")$width
    location$height <- attr(value, "dims")$height
  }
  width <- location$width
  height <- location$height

  xml_str <- to_pml(
    x = value,
    left = location$left,
    top = location$top,
    width = width,
    height = height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln
  )

  value <- as_xml_document(xml_str)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), value)
  x
}


#' @export
#' @describeIn ph_with add an [fpar()] to a new shape
#' on the current slide as a single paragraph in a [block_list()].
ph_with.fpar <- function(x, value, location, ...) {
  ph_with.block_list(x, value = block_list(value), location = location)

  x
}

#' @export
#' @describeIn ph_with add an [empty_content()] to a new shape
#' on the current slide.
ph_with.empty_content <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)
  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  if (is.na(location$fld_id)) {
    xml_elt <- paste0(psp_ns_yes, new_ph, "</p:sp>")
  } else {
    pars <- paste0(
      "<a:p><a:fld id=\"",
      location$fld_id,
      "\" type = \"",
      location$fld_type,
      "\"><a:rPr/><a:t>",
      x$cursor,
      "</a:t></a:fld></a:p>",
      collapse = ""
    )
    xml_elt <- sprintf(paste0(
      psp_ns_yes,
      new_ph,
      "<p:txBody><a:bodyPr/><a:lstStyle/>",
      pars,
      "</p:txBody></p:sp>"
    ))
  }
  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}


xml_to_slide <- function(slide, location, value, package_dir) {
  node <- xml_find_first(value, as_xpath_content_sel("//"))
  if (xml_name(node) == "grpSp") {
    node_sppr <- xml_child(node, "p:grpSpPr")
    node_xfrm <- xml_child(node, "p:grpSpPr/a:xfrm")
    node_name <- xml_child(node, "p:nvGrpSpPr/p:cNvPr")
  } else if (xml_name(node) == "graphicFrame") {
    node_xfrm <- xml_child(node, "p:xfrm")
    node_name <- xml_child(node, "p:nvGraphicFramePr/p:cNvPr")
    node_sppr <- xml_missing()
  } else if (xml_name(node) == "pic") {
    node_sppr <- xml_child(node, "p:spPr")
    node_xfrm <- xml_child(node, "p:spPr/a:xfrm")
    node_name <- xml_child(node, "p:nvPicPr/p:cNvPr")
  } else if (xml_name(node) == "sp") {
    node_sppr <- xml_child(node, "p:spPr")
    node_xfrm <- xml_child(node, "p:spPr/a:xfrm")
    node_name <- xml_child(node, "p:nvSpPr/p:cNvPr")
  } else {
    node_xfrm <- xml_missing()
    node_name <- xml_missing()
    node_sppr <- xml_missing()
  }

  if (!inherits(node_name, "xml_missing")) {
    xml_attr(node_name, "name") <- location$ph_label
  }

  if (!inherits(node_xfrm, "xml_missing")) {
    off <- xml_child(node_xfrm, "a:off")
    ext <- xml_child(node_xfrm, "a:ext")
    chOff <- xml_child(node_xfrm, "a:chOff")
    chExt <- xml_child(node_xfrm, "a:chExt")
    xml_attr(off, "x") <- sprintf("%.0f", location$left * 914400)
    xml_attr(off, "y") <- sprintf("%.0f", location$top * 914400)
    xml_attr(ext, "cx") <- sprintf("%.0f", location$width * 914400)
    xml_attr(ext, "cy") <- sprintf("%.0f", location$height * 914400)
    xml_attr(chOff, "x") <- sprintf("%.0f", location$left * 914400)
    xml_attr(chOff, "y") <- sprintf("%.0f", location$top * 914400)
    xml_attr(chExt, "cx") <- sprintf("%.0f", location$width * 914400)
    xml_attr(chExt, "cy") <- sprintf("%.0f", location$height * 914400)

    # add location$rotation to cNvPr:name
    if (
      !is.null(location$rotation) &&
        is.finite(location$rotation) &&
        is.numeric(location$rotation)
    ) {
      xml_attr(node_xfrm, "rot") <- sprintf("%.0f", -location$rotation * 60000)
    }
  }

  if (!inherits(node_sppr, "xml_missing")) {
    # add location$bg to SpPr
    if (!is.null(location$bg)) {
      bg_str <- solid_fill_pml(location$bg)
      xml_add_child(node_sppr, as_xml_document(bg_str))
    }
  }
  value
}

#' @export
#' @describeIn ph_with add an xml_document object to a new shape on the
#' current slide. This function is to be used to add custom openxml code.
ph_with.xml_document <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)

  xml_to_slide(slide, location, value, x$package_dir)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), value)
  x
}


# PHS_WITH -------

#' @title Fill multiple placeholders using key value syntax
#' @description A sibling of [officer::ph_with] that fills mutiple placeholders at once. Placeholder locations are
#'   specfied using the short form syntax. The location and corresponding object are passed as key value pairs
#'   (`phs_with("short form location" = object)`). Under the hood, [officer::ph_with] is called for each pair. Note
#'   that `phs_with` does not cover all options from the `ph_location_*` family and is also less customization. It is a
#'   covenience wrapper for the most common use cases. The implemented short forms are listed in section
#'   `"Short forms"`.
#' @param x A `rpptx` object.
#' @param ... Key-value pairs of the form `"short form location" = object`. If the short form is an integer or a string
#'   with blanks, you must wrap it in quotes or backticks.
#' @param .dots List of key-value pairs `"short form location" = object`. Alternative to `...`.
#' @param .slide_idx Numeric indexes of slides to process. `NULL` (default) processes the current slide only. Use
#'   keyword `all` for all slides.
#' @section Short forms: The following short forms are implemented and can be used as the parameter in the function
#'   call. The corresponding function from the `ph_location_*` family (called under the hood) is displayed on the
#'   right.
#'
#' | **Short form** | **Description**                                   | **Location function**           |
#' |----------------|---------------------------------------------------|---------------------------------|
#' | `"left"`       | Keyword string                                    | `ph_location_left()`            |
#' | `"right"`      | Keyword string                                    | `ph_location_right()`           |
#' | `"fullsize"`   | Keyword string                                    | `ph_location_fullsize()`        |
#' | `"body [1]"`   | String: type + index in brackets (`1` if omitted) | `ph_location_type("body", 1)`   |
#' | `"my_label"`   | Any string not matching a keyword or type         | `ph_location_label("my_label")` |
#' | `1`            | Length 1 integer                                  | `ph_location_id(1)`             |
#'
#' @export
#' @seealso [ph_with()], [add_slide()]
#' @example inst/examples/example_phs_with.R
phs_with <- function(x, ..., .dots = NULL, .slide_idx = NULL) {
  dots_list <- list(...)
  if (length(dots_list) > 0 && !is_named(dots_list)) {
    cli::cli_abort(
      c(
        "Missing key in {.arg ...}",
        "x" = "{.arg ...} requires a key-value syntax, with a ph short-form location as the key",
        "i" = "Example: {.code phs_with(x, title = 'My title', 'body[1]' = 'My body')}"
      )
    )
  }
  if (length(.dots) > 0 && !is_named(.dots)) {
    cli::cli_abort(
      c(
        "Missing names in {.arg .dots}",
        "x" = "{.arg .dots} must be a named list, with ph short-form locations as names",
        "i" = "Example: {.code phs_with(x, .dots = list(title = 'My title', 'body[1]' = 'My body'))}"
      )
    )
  }
  dots <- utils::modifyList(dots_list, .dots %||% list())
  if (length(dots) == 0) {
    return(x)
  }
  .slide_idx <- .slide_idx %||% x$cursor # default is current slide
  if (is.character(.slide_idx) && .slide_idx == "all") {
    .slide_idx <- seq_len(length(x))
  }
  stop_if_not_in_slide_range(x, .slide_idx)

  loc_strings <- as.list(names(dots))
  ii <- grepl("^\\d+$", loc_strings) # find integer short-forms
  loc_strings[ii] <- as.integer(loc_strings[ii])
  locations <- lapply(loc_strings, resolve_location)

  .old_cursor <- x$cursor
  for (slide_idx in .slide_idx) {
    x$cursor <- slide_idx # pw_with always uses the current slide
    for (i in seq_along(dots)) {
      x <- ph_with(x, dots[[i]], locations[[i]])
    }
  }
  x$cursor <- .old_cursor
  x
}

#' @export
#' @title Add a slide
#' @description Add a slide into a pptx presentation.
#' @param x an rpptx object
#' @param layout slide layout name to use
#' @param master master layout name where \code{layout} is located
#' @examples
#' my_pres <- read_pptx()
#' layout_summary(my_pres)
#' my_pres <- add_slide(my_pres,
#'   layout = "Two Content", master = "Office Theme"
#' )
#' @seealso [print.rpptx()], [read_pptx()], [plot_layout_properties()], [ph_with()], [layout_summary()]
#' @family functions slide manipulation
add_slide <- function(x, layout = "Title and Content", master = "Office Theme") {
  slide_info <- x$slideLayouts$get_metadata()
  slide_info <- slide_info[slide_info$name == layout & slide_info$master_name == master, ]

  if (nrow(slide_info) < 1) {
    stop("could not find layout named ", shQuote(layout), " in master named ", shQuote(master))
  } else if (nrow(slide_info) > 1) {
    stop(
      "found two layouts named ", shQuote(layout), " in master named ", shQuote(master),
      ". Layout names should not be duplicated."
    )
  }
  new_slidename <- x$slide$get_new_slidename()

  xml_file <- file.path(x$package_dir, "ppt/slides", new_slidename)
  layout_obj <- x$slideLayouts$collection_get(slide_info$filename)
  layout_obj$write_template(xml_file)

  # update presentation elements
  x$presentation$add_slide(target = file.path("slides", new_slidename))
  x$content_type$add_slide(partname = file.path("/ppt/slides", new_slidename))

  x$slide$add_slide(xml_file, x$slideLayouts$get_xfrm_data())

  x$cursor <- x$slide$length()
  x
}


#' @export
#' @title Change current slide
#' @description Change current slide index of an rpptx object.
#' @param x an rpptx object
#' @param index slide index
#' @examples
#' doc <- read_pptx()
#' doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
#' doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
#' doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
#' doc <- on_slide(doc, index = 1)
#' doc <- ph_with(
#'   x = doc, "First title",
#'   location = ph_location_type(type = "title")
#' )
#' doc <- on_slide(doc, index = 3)
#' doc <- ph_with(
#'   x = doc, "Third title",
#'   location = ph_location_type(type = "title")
#' )
#'
#' file <- tempfile(fileext = ".pptx")
#' print(doc, target = file)
#' @family functions slide manipulation
#' @seealso [read_pptx()], [ph_with()]
on_slide <- function(x, index) {
  l_ <- length(x)
  if (l_ < 1) {
    stop("presentation contains no slide", call. = FALSE)
  }
  if (!between(index, 1, l_)) {
    stop("unvalid index ", index, " (", l_, " slide(s))", call. = FALSE)
  }

  filename <- basename(x$presentation$slide_data()$target[index])
  location <- which(x$slide$get_metadata()$name %in% filename)

  x$cursor <- x$slide$slide_index(filename)

  x
}

#' @export
#' @title Remove a slide
#' @description Remove a slide from a pptx presentation.
#' @param x an rpptx object
#' @param index slide index, default to current slide position.
#' @param rm_images if TRUE (defaults to FALSE), images presented in
#' the slide to remove are also removed from the file.
#' @note cursor is set on the last slide.
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres)
#' my_pres <- remove_slide(my_pres)
#' @family functions slide manipulation
#' @seealso [read_pptx()], [ph_with()], [ph_remove()]
remove_slide <- function(x, index = NULL, rm_images = FALSE) {
  l_ <- length(x)
  if (l_ < 1) {
    stop("presentation contains no slide to delete", call. = FALSE)
  }

  if (is.null(index)) {
    index <- x$cursor
  }

  if (!between(index, 1, l_)) {
    stop("unvalid index ", index, " (", l_, " slide(s))", call. = FALSE)
  }
  filename <- basename(x$presentation$slide_data()$target[index])
  location <- which(x$slide$get_metadata()$name %in% filename)

  if (rm_images) {
    media_files <- pptx_summary(x)$media_file
    if (length(media_files)) {
      unlink(file.path(x$package_dir, media_files), force = TRUE)
    }
  }

  del_file <- x$slide$remove_slide(location)

  # update presentation elements
  x$presentation$remove_slide(del_file)
  x$content_type$remove_slide(partname = del_file)
  x$cursor <- x$slide$length()
  x
}



#' @export
#' @title Move a slide
#' @description Move a slide in a pptx presentation.
#' @inheritParams remove_slide
#' @param to new slide index.
#' @note cursor is set on the last slide.
#' @examples
#' x <- read_pptx()
#' x <- add_slide(x)
#' x <- ph_with(x, "Hello world 1", location = ph_location_type())
#' x <- add_slide(x)
#' x <- ph_with(x, "Hello world 2", location = ph_location_type())
#' x <- move_slide(x, index = 1, to = 2)
#' @family functions slide manipulation
#' @seealso [read_pptx()]
move_slide <- function(x, index = NULL, to) {
  x$presentation$slide_data()

  if (is.null(index)) {
    index <- x$cursor
  }

  l_ <- length(x)

  if (l_ < 1) {
    stop("presentation contains no slide", call. = FALSE)
  }
  if (!between(index, 1, l_)) {
    stop("unvalid index ", index, " (", l_, " slide(s))", call. = FALSE)
  }
  if (!between(to, 1, l_)) {
    stop("unvalid 'to' ", to, " (", l_, " slide(s))", call. = FALSE)
  }

  x$presentation$move_slide(from = index, to = to)
  x$cursor <- to
  x
}



#' @title Correct pptx content references
#' @description Content references are not managed directly
#' but computed after the content is added. This function
#' loop over each slide and fix references (links and images)
#' if necessary.
#' @param x an rpptx object
#' @return an rpptx object
#' @noRd
pptx_fortify_slides <- function(x) {
  for (cursor_index in seq_len(x$slide$length())) {
    slide <- x$slide$get_slide(cursor_index)

    process_images(slide, slide$relationship(), x$package_dir, media_dir = "ppt/media", media_rel_dir = "../media")

    process_links(slide, type = "pml")

    cnvpr <- xml_find_all(slide$get(), "//p:cNvPr")
    for (i in seq_along(cnvpr)) {
      xml_attr(cnvpr[[i]], "id") <- i
    }
  }

  x
}

#' @export
#' @title pptx tags for visual and non visual properties
#' @description Visual and non visual properties of a
#' shape can be returned by this function.
#' @param left,top,width,height place holder coordinates
#' in inches.
#' @param bg background color
#' @param rot rotation angle
#' @param label a label for the placeholder.
#' @param ph string containing xml code for ph tag
#' @param ln a [sp_line()] object specifying the outline style.
#' @param geom shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html
#' @return a character value
#' @family functions for officer extensions
#' @keywords internal
shape_properties_tags <- function(left = 0, top = 0, width = 3, height = 3,
                                  bg = "transparent", rot = 0, label = "", ph = "<p:ph/>",
                                  ln = sp_line(lwd = 0, linecmpd = "solid", lineend = "rnd"), geom = NULL) {
  if (!is.null(bg) && !is.color(bg)) {
    stop("bg must be a valid color.", call. = FALSE)
  }

  bg_str <- solid_fill_pml(bg)
  ln_str <- ln_pml(ln)
  geom_str <- prst_geom_pml(geom)

  xfrm_str <- a_xfrm_str(left = left, top = top, width = width, height = height, rot = rot)
  if (is.null(ph) || is.na(ph)) {
    ph <- "<p:ph/>"
  }

  randomid <- officer::uuid_generate()

  str <- "<p:nvSpPr><p:cNvPr id=\"%s\" name=\"%s\"/><p:cNvSpPr><a:spLocks noGrp=\"1\"/></p:cNvSpPr><p:nvPr>%s</p:nvPr></p:nvSpPr><p:spPr>%s%s%s%s</p:spPr>"

  sprintf(str, randomid, label, ph, xfrm_str, geom_str, bg_str, ln_str)
}


# check if slide index exists
ensure_slide_index_exists <- function(x, slide_idx) {
  stop_if_not_rpptx(x)
  if (!is.numeric(slide_idx)) {
    cli::cli_abort(
      c("{.arg slide_idx} must be {.cls numeric}",
        "x" = "You provided {.cls {class(slide_idx)[1]}} instead.")
      , call = NULL)
  }
  n <- length(x) # no of slides
  check <- slide_idx %in% seq_len(n)
  if (!check) {
    cli::cli_abort(
      c("Slide index {.val {slide_idx}} is out of range.",
        "x" = "Presentation has {cli::no(n)} slide{?s}."
      ), call = NULL
    )
  }
}

get_shape_id <- function(x, type = NULL, id = NULL, ph_label = NULL) {
  slsmry <- slide_summary(x)

  if (is.null(ph_label)) {
    sel <- which(slsmry$type %in% type)
    if (length(sel) < 1) stop("no shape of type ", shQuote(type), " has been found")
    sel <- sel[id]
    if (sum(is.finite(sel)) != 1) stop("no shape of type ", shQuote(type), " and with id ", id, " has been found")
  } else {
    sel <- which(slsmry$ph_label %in% ph_label)
    if (length(sel) < 1) stop("no shape with label ", shQuote(ph_label), " has been found")
    sel <- sel[id]
    if (sum(is.finite(sel)) != 1) stop("no shape with label ", shQuote(ph_label), "and with id ", id, " has been found")
  }
  slsmry$id[sel]
}




#' @export
#' @title Remove a shape
#' @description Remove a shape in a slide.
#' @param x an rpptx object
#' @param type placeholder type
#' @param id placeholder index (integer) for a duplicated type. This is to be used when a placeholder
#' type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
#' add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
#' Values can be read from \code{\link{slide_summary}}.
#' @param ph_label label associated to the placeholder. Use column
#' \code{ph_label} of result returned by \code{\link{slide_summary}}.
#' If used, `type` and `id` are ignored.
#' @param id_chr deprecated.
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' dummy_fun <- function(doc) {
#'   doc <- add_slide(doc,
#'     layout = "Two Content",
#'     master = "Office Theme"
#'   )
#'   doc <- ph_with(
#'     x = doc, value = "Un titre",
#'     location = ph_location_type(type = "title")
#'   )
#'   doc <- ph_with(
#'     x = doc, value = "Un corps 1",
#'     location = ph_location_type(type = "body", id = 1)
#'   )
#'   doc <- ph_with(
#'     x = doc, value = "Un corps 2",
#'     location = ph_location_type(type = "body", id = 2)
#'   )
#'   doc
#' }
#' doc <- read_pptx()
#' for (i in 1:3) {
#'   doc <- dummy_fun(doc)
#' }
#'
#' doc <- on_slide(doc, index = 1)
#' doc <- ph_remove(x = doc, type = "title")
#'
#' doc <- on_slide(doc, index = 2)
#' doc <- ph_remove(x = doc, type = "body", id = 2)
#'
#' doc <- on_slide(doc, index = 3)
#' doc <- ph_remove(x = doc, type = "body", id = 1)
#'
#' print(doc, target = fileout)
#' @family functions for placeholders manipulation
#' @seealso \code{\link{ph_with}}
ph_remove <- function(x, type = "body", id = 1, ph_label = NULL, id_chr = NULL) {
  slide <- x$slide$get_slide(x$cursor)
  office_id <- get_shape_id(x, type = type, id = id, ph_label = ph_label)
  current_elt <- xml_find_first(slide$get(), sprintf("p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]", office_id))
  xml_remove(current_elt)

  x
}



#' @export
#' @title Slide link to a placeholder
#' @description Add slide link to a placeholder in the current slide.
#' @inheritParams ph_remove
#' @param slide_index slide index to reach
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' loc_title <- ph_location_type(type = "title")
#' doc <- read_pptx()
#' doc <- add_slide(doc)
#' doc <- ph_with(x = doc, "Un titre 1", location = loc_title)
#' doc <- add_slide(doc)
#' doc <- ph_with(x = doc, "Un titre 2", location = loc_title)
#' doc <- on_slide(doc, 1)
#' slide_summary(doc) # read column ph_label here
#' doc <- ph_slidelink(x = doc, ph_label = "Title 1", slide_index = 2)
#'
#' print(doc, target = fileout)
#' @family functions for placeholders manipulation
#' @seealso \code{\link{ph_with}}
ph_slidelink <- function(x, type = "body", id = 1, id_chr = NULL, ph_label = NULL, slide_index) {
  slide <- x$slide$get_slide(x$cursor)
  office_id <- get_shape_id(x, type = type, id = id, ph_label = ph_label)
  current_elt <- xml_find_first(slide$get(), sprintf("p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]", office_id))

  # declare slide ref in relationships
  slide_name <- x$slide$names()[slide_index]
  slide$reference_slide(slide_name)
  rel_df <- slide$rel_df()
  id <- rel_df[rel_df$target == slide_name, "id"]

  # add hlinkClick
  cnvpr <- xml_child(current_elt, "p:nvSpPr/p:cNvPr")
  str_ <- "<a:hlinkClick xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:id=\"%s\" action=\"ppaction://hlinksldjump\"/>"
  str_ <- sprintf(str_, id)
  xml_add_child(cnvpr, as_xml_document(str_))

  x
}


#' @export
#' @title Hyperlink a placeholder
#' @description Add hyperlink to a placeholder in the current slide.
#' @inheritParams ph_remove
#' @param href hyperlink (do not forget http or https prefix)
#' @examples
#' fileout <- tempfile(fileext = ".pptx")
#' loc_manual <- ph_location(bg = "red", newlabel = "mytitle")
#' doc <- read_pptx()
#' doc <- add_slide(doc)
#' doc <- ph_with(x = doc, "Un titre 1", location = loc_manual)
#' slide_summary(doc) # read column ph_label here
#' doc <- ph_hyperlink(
#'   x = doc, ph_label = "mytitle",
#'   href = "https://cran.r-project.org"
#' )
#'
#' print(doc, target = fileout)
#' @family functions for placeholders manipulation
#' @seealso \code{\link{ph_with}}
ph_hyperlink <- function(x, type = "body", id = 1, id_chr = NULL, ph_label = NULL, href) {
  slide <- x$slide$get_slide(x$cursor)
  office_id <- get_shape_id(x, type = type, id = id, ph_label = ph_label)
  current_elt <- xml_find_first(slide$get(), sprintf("p:cSld/p:spTree/*[p:nvSpPr/p:cNvPr[@id='%s']]|p:cSld/p:spTree/*[p:nvPicPr/p:cNvPr[@id='%s']]", office_id, office_id))

  # add hlinkClick
  if (xml_name(current_elt) %in% "pic") {
    cnvpr <- xml_child(current_elt, "p:nvPicPr/p:cNvPr")
  } else {
    cnvpr <- xml_child(current_elt, "p:nvSpPr/p:cNvPr")
  }
  str_ <- "<a:hlinkClick xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" r:id=\"%s\"/>"
  str_ <- sprintf(str_, href)
  xml_add_child(cnvpr, as_xml_document(str_))
  x
}

#' @export
#' @title Create a 'Word' document object
#' @description `read_docx()` is the starting point for creating Word documents
#' from R. It creates an R object representing a Word document that can be
#' manipulated programmatically. When called without arguments, it creates an
#' empty document based on a default template. When provided with a path, it
#' reads an existing Word document (`.docx`) or template (`.dotx`) file.
#'
#' Once created, you can:
#' - Add content from R: Insert text, formatted paragraphs ([fpar()]),
#'   tables ([body_add_table()]), plots ([body_add_plot()]), images,
#'   page breaks, table of contents, and more using the `body_add_*()` family
#'   of functions.
#' - Read and inspect content: Use [docx_summary()] to extract and analyze
#'   the document's content, structure, and formatting as a data frame.
#' - Write to file: Save the document to disk using
#'   `print(x, target = "path/to/file.docx")`.
#'
#' @param path path to the docx file to use as base document.
#' `dotx` file are supported.
#' @return an object of class `rdocx`.
#' @section styles:
#'
#' The template file (specified via `path` or the default template) determines
#' the available paragraph styles, character styles, and table styles in your
#' document. These styles control the appearance of headings, body text, tables,
#' and other elements.
#'
#' When you use functions like `body_add_par(style = "heading 2")`, the style
#' name must exist in the template. You can:
#' - Use [styles_info()] to list all available styles in your document
#' - Create a custom template in Word with your organization's styles and branding
#' - Use the default template for standard documents
#'
#' The document layout (page size, margins, headers and footer content,
#' orientation) also comes from the
#' template and can be modified using [body_set_default_section()] or by
#' adding section breaks with [body_end_section_continuous()] and related functions.
#' @examples
#' library(officer)
#'
#' # This example demonstrates how to create
#' # an empty document -----
#'
#' ## Create a new Word document
#' doc <- read_docx()
#' ## Save the document
#' output_file <- print(doc, target = tempfile(fileext = ".docx"))
#'
#' @example inst/examples/example_read_docx_1.R
#' @example inst/examples/example_read_docx_2.R
#' @seealso Save a 'Word' document object to a file with [print.rdocx()],
#' add content with functions [body_add_par()], [body_add_plot()],
#' [body_add_table()], change settings with [docx_set_settings()], set
#' properties with [set_doc_properties()],  read 'Word' styles with
#' [styles_info()].
#' @section Illustrations:
#'
#' \if{html}{\figure{read_docx.png}{options: style="width:80\%;"}}
read_docx <- function(path = NULL) {
  if (!is.null(path) && !file.exists(path)) {
    stop("could not find file ", shQuote(path), call. = FALSE)
  }

  if (is.null(path)) {
    path <- system.file(package = "officer", "template/template.docx")
  }

  if (!grepl("\\.(docx|dotx)$", path, ignore.case = TRUE)) {
    stop("read_docx only support docx files", call. = FALSE)
  }

  package_dir <- tempfile()
  unpack_folder(file = path, folder = package_dir)

  obj <- structure(
    list(package_dir = package_dir),
    .Names = c("package_dir"),
    class = "rdocx"
  )

  obj$settings <- update_docx_settings_from_file(
    x = docx_settings(),
    file = file.path(package_dir, "word", "settings.xml")
  )

  obj$rel <- relationship$new()
  obj$rel$feed_from_xml(file.path(package_dir, "_rels", ".rels"))

  obj$doc_properties_custom <- read_custom_properties(package_dir)
  obj$doc_properties <- read_core_properties(package_dir)
  obj$app_properties <- read_app_properties(package_dir)
  obj$content_type <- content_type$new(package_dir)
  obj$doc_obj <- body_part$new(
    package_dir,
    main_file = "document.xml",
    cursor = "/w:document/w:body/*[1]",
    body_xpath = "/w:document/w:body"
  )
  obj$styles <- read_docx_styles(package_dir)
  obj$officer_cursor <- officer_cursor(obj$doc_obj$get())

  obj$headers <- update_hf_list(
    part_list = list(),
    type = "header",
    package_dir = package_dir
  )
  obj$footers <- update_hf_list(
    part_list = list(),
    type = "footer",
    package_dir = package_dir
  )

  if (!file.exists(file.path(package_dir, "word", "comments.xml"))) {
    file.copy(
      system.file(package = "officer", "template", "comments.xml"),
      file.path(package_dir, "word", "comments.xml"),
      copy.mode = FALSE
    )
    obj$content_type$add_override(
      setNames(
        "application/vnd.openxmlformats-officedocument.wordprocessingml.comments+xml",
        "/word/comments.xml"
      )
    )
  }

  obj$comments <- docx_part$new(
    package_dir,
    main_file = "comments.xml",
    cursor = "/w:comments/*[last()]",
    body_xpath = "/w:comments"
  )

  if (!file.exists(file.path(package_dir, "word", "footnotes.xml"))) {
    file.copy(
      system.file(package = "officer", "template", "footnotes.xml"),
      file.path(package_dir, "word", "footnotes.xml"),
      copy.mode = FALSE
    )
    obj$content_type$add_override(
      setNames(
        "application/vnd.openxmlformats-officedocument.wordprocessingml.footnotes+xml",
        "/word/footnotes.xml"
      )
    )
  }

  obj$footnotes <- footnotes_part$new(
    package_dir,
    main_file = "footnotes.xml",
    cursor = "/w:footnotes/*[last()]",
    body_xpath = "/w:footnotes"
  )

  default_refs <- obj$styles[obj$styles$is_default, ]
  obj$default_styles <- setNames(
    as.list(default_refs$style_name),
    default_refs$style_type
  )

  last_sect <- xml_find_first(
    obj$doc_obj$get(),
    "/w:document/w:body/w:sectPr[last()]"
  )
  obj$sect_dim <- section_dimensions(last_sect)

  obj <- cursor_end(obj)
  obj
}


#' @export
#' @title Body xml document
#' @description Get the body document as xml. This function
#' is not to be used by end users, it has been implemented
#' to allow other packages to work with officer.
#' @param x an rdocx object
#' @examples
#' doc <- read_docx()
#' docx_body_xml(doc)
#' @keywords internal
docx_body_xml <- function(x) {
  x$doc_obj$get()
}
#' @export
#' @title xml element on which cursor is
#' @description Get the current block element as xml. This function
#' is not to be used by end users, it has been implemented
#' to allow other packages to work with officer. If the
#' document is empty, this block will be set to NULL.
#' @param x an rdocx object
#' @examples
#' doc <- read_docx()
#' docx_current_block_xml(doc)
#' @keywords internal
docx_current_block_xml <- function(x) {
  ooxml_on_cursor(x$officer_cursor, x$doc_obj$get())
}

#' @export
#' @title Body xml document
#' @description Get the body document as xml. This function
#' is not to be used by end users, it has been implemented
#' to allow other packages to work with officer.
#' @param x an rdocx object
#' @examples
#' doc <- read_docx()
#' docx_body_relationship(doc)
#' @keywords internal
docx_body_relationship <- function(x) {
  x$doc_obj$relationship()
}

getncheck <- function(x, str) {
  child_ <- xml_child(x, str)
  expect_false(inherits(child_, "xml_missing"))
  child_
}


test_that("seqfield add ", {
  x <- read_docx()
  x <- body_add_fpar(
    x,
    fpar(
      "Time is: ",
      run_word_field(field = "TIME \u005C@ \"HH:mm:ss\" \u005C* MERGEFORMAT")
    ),
    style = "Normal"
  )

  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:fldChar[@w:fldCharType='begin']")
  getncheck(node, "w:r/w:fldChar[@w:fldCharType='end']")

  child_ <- getncheck(node, "w:r/w:instrText")
  expect_equal(xml_text(child_), "TIME \\@ \"HH:mm:ss\" \\* MERGEFORMAT")

  z <- fpar(
    "Figure: ",
    run_word_field(field = "SEQ Figure \u005C* roman"),
    " - This is a figure title"
  )
  x <- body_add_fpar(x, z, style = "centered", )

  node <- docx_current_block_xml(x)
  expect_equal(
    xml_text(node),
    "Figure: SEQ Figure \\* roman - This is a figure title"
  )
})

getncheck <- function(x, str) {
  child_ <- xml_child(x, str)
  expect_false(inherits(child_, "xml_missing"))
  child_
}



test_that("seqfield add ", {
  x <- read_docx()
  x <- body_add_fpar(x,
    fpar(
      "Time is: ",
      run_word_field(field = "TIME \u005C@ \"HH:mm:ss\" \u005C* MERGEFORMAT")
    ),
    style = "Normal"
  )

  node <- docx_current_block_xml(x)
  getncheck(node, "w:r/w:fldChar[@w:fldCharType='begin']")
  getncheck(node, "w:r/w:fldChar[@w:fldCharType='end']")

  child_ <- getncheck(node, "w:r/w:instrText")
  expect_equal(xml_text(child_), "TIME \\@ \"HH:mm:ss\" \\* MERGEFORMAT")

  z <- fpar(
    "Figure: ",
    run_word_field(field = "SEQ Figure \u005C* roman"),
    " - This is a figure title"
  )
  x <- body_add_fpar(x, z, style = "centered", )

  node <- docx_current_block_xml(x)
  expect_equal(xml_text(node), "Figure: SEQ Figure \\* roman - This is a figure title")
})

wml_str <- function(str){
  paste0( "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n",
          "<w:document xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">",
          str,
          "</w:document>"
  )
}

pml_str <- function(str){
  paste0( "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n",
          "<a:document xmlns:a=\"http://schemas.openxmlformats.org/drawingml/2006/main\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:p=\"http://schemas.openxmlformats.org/presentationml/2006/main\">",
          str,
          "</a:document>"
  )
}

has_css_color <- function(x, atname, color){
  css <- format(x, type = "html")
  reg <- sprintf("%s:%s", atname, color)
  grepl(reg, css)
}

has_css_attr <- function(x, atname, value){
  css <- format(x, type = "html")
  reg <- sprintf("%s:%s", atname, value)
  grepl(reg, css)
}

#' @export
#' @title Set 'Microsoft Word' Document Settings
#' @description
#' Set various settings of a 'Microsoft Word' document generated with 'officer'.
#' Options include:
#' - zoom factor (default view in Word),
#' - default tab stop,
#' - hyphenation zone,
#' - decimal symbol,
#' - list separator (see details below),
#' - compatibility mode,
#' - even and odd headers management (see details below),
#' - and auto hyphenation activation.
#'
#' @details
#' - `even_and_odd_headers`: If TRUE, 'Microsoft Word' will use different
#' headers for odd and even pages ("Different Odd & Even Pages" feature in Word).
#' This is useful for professional documents or reports that require alternating
#' page layouts.
#'
#' - `list_separator`: Sets the character used by 'Microsoft Word' to separate
#' items in lists (for example, when inserting tables or lists in Word). This
#' parameter affects how 'Microsoft Word' handles data import/export (CSV, etc.) and can be
#' adapted to language or local conventions (e.g., ";" for French, "," for English).
#'
#' @seealso [read_docx()]
#' @param x an rdocx object
#' @param zoom zoom factor, default is 1 (100%)
#' @param default_tab_stop default tab stop in inches, default is 0.5
#' @param hyphenation_zone hyphenation zone in inches, default is 0.25
#' @param decimal_symbol decimal symbol, default is "."
#' @param list_separator list separator, default is ";". Sets the separator
#' used by Word for lists (see details).
#' @param compatibility_mode compatibility mode, default is "15"
#' @param even_and_odd_headers whether to use different headers for even and
#' odd pages, default is FALSE. Enables the "Different Odd and Even Pages"
#' feature in 'Microsoft Word'.
#' @param auto_hyphenation whether to enable auto hyphenation, default is FALSE.
#' @param unit unit for `default_tab_stop` and `hyphenation_zone`, one of "in", "cm", "mm".
#' @example inst/examples/example_docx_set_settings.R
docx_set_settings <- function(
  x,
  zoom = 1,
  default_tab_stop = 0.5,
  hyphenation_zone = 0.25,
  decimal_symbol = ".",
  list_separator = ";",
  compatibility_mode = "15",
  even_and_odd_headers = FALSE,
  auto_hyphenation = FALSE,
  unit = "in"
) {
  settings <- docx_settings(
    zoom = zoom,
    default_tab_stop = convin(unit = unit, x = default_tab_stop),
    hyphenation_zone = convin(unit = unit, x = hyphenation_zone),
    decimal_symbol = decimal_symbol,
    list_separator = list_separator,
    even_and_odd_headers = even_and_odd_headers,
    auto_hyphenation = auto_hyphenation,
    compatibility_mode = compatibility_mode
  )
  x$settings <- settings
  x
}

docx_settings <- function(
  zoom = 1,
  default_tab_stop = 0.5,
  hyphenation_zone = 0.25,
  decimal_symbol = ".",
  list_separator = ";",
  compatibility_mode = "15",
  even_and_odd_headers = FALSE,
  auto_hyphenation = FALSE
) {
  x <- list(
    zoom = zoom,
    default_tab_stop = default_tab_stop,
    hyphenation_zone = hyphenation_zone,
    decimal_symbol = decimal_symbol,
    list_separator = list_separator,
    even_and_odd_headers = even_and_odd_headers,
    auto_hyphenation = auto_hyphenation,
    compatibility_mode = compatibility_mode
  )
  class(x) <- "docx_settings"
  x
}

update_docx_settings_from_file <- function(x, file) {
  if (is.null(file)) {
    cli::cli_abort(
      "File settings for Word document is null."
    )
  }
  if (!file.exists(file)) {
    cli::cli_abort(
      "File settings for Word document does not exists {.file {file}}."
    )
  }

  node_doc <- read_xml(file)

  node_zoom <- xml_child(node_doc, "w:zoom")
  if (!inherits(node_zoom, "xml_missing")) {
    x$zoom <- as.integer(xml_attr(node_zoom, "percent")) / 100
  }

  node_tab_stop <- xml_child(node_doc, "w:defaultTabStop")
  if (!inherits(node_tab_stop, "xml_missing")) {
    x$default_tab_stop <- as.integer(xml_attr(node_tab_stop, "val")) /
      1440
  }

  node_hyphenation_zone <- xml_child(node_doc, "w:hyphenationZone")
  if (!inherits(node_hyphenation_zone, "xml_missing")) {
    x$hyphenation_zone <- as.integer(xml_attr(
      node_hyphenation_zone,
      "val"
    )) /
      1440
  }

  node_auto_hyphenation <- xml_child(node_doc, "w:autoHyphenation")
  if (!inherits(node_auto_hyphenation, "xml_missing")) {
    x$auto_hyphenation <- TRUE
  }

  node_decimal_symbol <- xml_child(node_doc, "w:decimalSymbol")
  if (!inherits(node_decimal_symbol, "xml_missing")) {
    x$decimal_symbol <- xml_attr(node_decimal_symbol, "val")
  }

  node_list_separator <- xml_child(node_doc, "w:listSeparator")
  if (!inherits(node_list_separator, "xml_missing")) {
    x$list_separator <- xml_attr(node_list_separator, "val")
  }

  node_evenodd_headers <- xml_child(node_doc, "w:evenAndOddHeaders")
  if (!inherits(node_evenodd_headers, "xml_missing")) {
    x$even_and_odd_headers <- TRUE
  } else {
    x$even_and_odd_headers <- FALSE
  }

  x
}

write_docx_settings <- function(x) {
  settings <- x$settings
  file <- file.path(x$package_dir, "word", "settings.xml")
  doc <- read_xml(file)
  ns <- "http://schemas.openxmlformats.org/wordprocessingml/2006/main"

  upsert_simple <- function(doc, tag, attr, value) {
    node <- xml_child(doc, tag)
    new_xml <- sprintf(
      "<%s xmlns:w=\"%s\" w:%s=\"%s\"/>",
      tag,
      ns,
      attr,
      value
    )
    if (inherits(node, "xml_missing")) {
      xml_add_child(doc, read_xml(new_xml))
    } else {
      xml_replace(node, read_xml(new_xml))
    }
  }

  upsert_simple(
    doc,
    "w:zoom",
    "percent",
    sprintf("%.0f", settings$zoom * 100)
  )
  upsert_simple(
    doc,
    "w:defaultTabStop",
    "val",
    sprintf("%.0f", settings$default_tab_stop * 1440)
  )
  upsert_simple(
    doc,
    "w:hyphenationZone",
    "val",
    sprintf("%.0f", settings$hyphenation_zone * 1440)
  )
  upsert_simple(doc, "w:decimalSymbol", "val", settings$decimal_symbol)
  upsert_simple(doc, "w:listSeparator", "val", settings$list_separator)

  # autoHyphenation: add if TRUE, remove if FALSE
  node <- xml_child(doc, "w:autoHyphenation")
  if (settings$auto_hyphenation && inherits(node, "xml_missing")) {
    new_xml <- sprintf(
      "<w:autoHyphenation xmlns:w=\"%s\"/>",
      ns
    )
    xml_add_child(doc, read_xml(new_xml))
  } else if (!settings$auto_hyphenation && !inherits(node, "xml_missing")) {
    xml_remove(node)
  }

  # evenAndOddHeaders
  eoh_val <- if (settings$even_and_odd_headers) "1" else "0"
  upsert_simple(doc, "w:evenAndOddHeaders", "val", eoh_val)

  # compat / compatSetting
  compat_node <- xml_child(doc, "w:compat")
  compat_xml <- sprintf(
    paste0(
      "<w:compat xmlns:w=\"%s\">",
      "<w:compatSetting w:name=\"compatibilityMode\"",
      " w:uri=\"http://schemas.microsoft.com/office/word\"",
      " w:val=\"%s\"/></w:compat>"
    ),
    ns,
    settings$compatibility_mode
  )
  if (inherits(compat_node, "xml_missing")) {
    xml_add_child(doc, read_xml(compat_xml))
  } else {
    xml_replace(compat_node, read_xml(compat_xml))
  }

  # embedTrueTypeFonts: add if TRUE, leave existing if not set
  if (isTRUE(settings$embed_true_type_fonts)) {
    node <- xml_child(doc, "w:embedTrueTypeFonts")
    if (inherits(node, "xml_missing")) {
      new_xml <- sprintf(
        "<w:embedTrueTypeFonts xmlns:w=\"%s\"/>",
        ns
      )
      xml_add_child(doc, read_xml(new_xml))
    }
  }

  write_xml(doc, file)
  TRUE
}

#' @export
#' @title Create a 'Word' document object
#' @description read and import a docx file as an R object
#' representing the document. When no file is specified, it
#' uses a default empty file.
#'
#' Use then this object to add content to it and create Word files
#' from R.
#' @param path path to the docx file to use as base document.
#' `dotx` file are supported.
#' @return an object of class `rdocx`.
#' @section styles:
#'
#' `read_docx()` uses a Word file as the initial document.
#' This is the original Word document from which the document
#' layout, paragraph styles, or table styles come.
#'
#' You will be able to add formatted text, change the paragraph
#' style with the R api but also use the styles from the
#' original document.
#'
#' See `body_add_*` functions to add content.
#' @examples
#' library(officer)
#'
#' pinst <- plot_instr({
#'   z <- c(rnorm(100), rnorm(50, mean = 5))
#'   plot(density(z))
#' })
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, "This is a table", style = "heading 2")
#' doc_1 <- body_add_table(doc_1, value = mtcars, style = "table_template")
#' doc_1 <- body_add_par(doc_1, "This is a plot", style = "heading 2")
#' doc_1 <- body_add_plot(doc_1, pinst)
#' docx_file_1 <- print(doc_1, target = tempfile(fileext = ".docx"))
#'
#' template <- system.file(package = "officer",
#'   "doc_examples", "landscape.docx")
#' doc_2 <- read_docx(path = template)
#' doc_2 <- body_add_par(doc_2, "This is a table", style = "heading 2")
#' doc_2 <- body_add_table(doc_2, value = mtcars)
#' doc_2 <- body_add_par(doc_2, "This is a plot", style = "heading 2")
#' doc_2 <- body_add_plot(doc_2, pinst)
#' docx_file_2 <- print(doc_2, target = tempfile(fileext = ".docx"))
#'
#' @seealso [body_add_par], [body_add_plot], [body_add_table]
#' @section Illustrations:
#'
#' \if{html}{\figure{read_docx_doc_1.png}{options: width=80\%}}
#'
#' \if{html}{\figure{read_docx_doc_2.png}{options: width=80\%}}
read_docx <- function(path = NULL) {
  if (!is.null(path) && !file.exists(path)) {
    stop("could not find file ", shQuote(path), call. = FALSE)
  }

  if (is.null(path)) {
    path <- system.file(package = "officer", "template/template.docx")
  }

  if (!grepl("\\.(docx|dotx)$", path, ignore.case = TRUE)) {
    stop("read_docx only support docx files", call. = FALSE)
  }

  package_dir <- tempfile()
  unpack_folder(file = path, folder = package_dir)

  obj <- structure(list(package_dir = package_dir),
    .Names = c("package_dir"),
    class = "rdocx"
  )

  obj$settings <- update(
    object = docx_settings(),
    file = file.path(package_dir, "word", "settings.xml")
  )

  obj$rel <- relationship$new()
  obj$rel$feed_from_xml(file.path(package_dir, "_rels", ".rels"))

  obj$doc_properties_custom <- read_custom_properties(package_dir)
  obj$doc_properties <- read_core_properties(package_dir)
  obj$content_type <- content_type$new(package_dir)
  obj$doc_obj <- docx_part$new(package_dir,
    main_file = "document.xml",
    cursor = "/w:document/w:body/*[1]",
    body_xpath = "/w:document/w:body"
  )
  obj$styles <- read_docx_styles(package_dir)
  obj$officer_cursor <- officer_cursor(obj$doc_obj$get())

  obj$headers <- update_hf_list(part_list = list(), type = "header", package_dir = package_dir)
  obj$footers <- update_hf_list(part_list = list(), type = "footer", package_dir = package_dir)

  if (!file.exists(file.path(package_dir, "word", "comments.xml"))) {
    file.copy(
      system.file(package = "officer", "template", "comments.xml"),
      file.path(package_dir, "word", "comments.xml"),
      copy.mode = FALSE
    )
    obj$content_type$add_override(
      setNames("application/vnd.openxmlformats-officedocument.wordprocessingml.comments+xml", "/word/comments.xml")
    )
  }

  obj$comments <- docx_part$new(
    package_dir,
    main_file = "comments.xml",
    cursor = "/w:comments/*[last()]", body_xpath = "/w:comments"
  )

  if (!file.exists(file.path(package_dir, "word", "footnotes.xml"))) {
    file.copy(
      system.file(package = "officer", "template", "footnotes.xml"),
      file.path(package_dir, "word", "footnotes.xml"),
      copy.mode = FALSE
    )
    obj$content_type$add_override(
      setNames("application/vnd.openxmlformats-officedocument.wordprocessingml.footnotes+xml", "/word/footnotes.xml")
    )
  }

  obj$footnotes <- docx_part$new(
    package_dir,
    main_file = "footnotes.xml",
    cursor = "/w:footnotes/*[last()]", body_xpath = "/w:footnotes"
  )

  default_refs <- obj$styles[obj$styles$is_default, ]
  obj$default_styles <- setNames(as.list(default_refs$style_name), default_refs$style_type)

  last_sect <- xml_find_first(obj$doc_obj$get(), "/w:document/w:body/w:sectPr[last()]")
  obj$sect_dim <- section_dimensions(last_sect)

  obj <- cursor_end(obj)
  obj
}

#' @export
#' @describeIn read_docx write docx to a file. It returns the path of the result
#' file.
#' @param x an rdocx object
#' @param target path to the docx file to write
#' @param ... unused
print.rdocx <- function(x, target = NULL, ...) {
  if (is.null(target)) {
    cat("rdocx document with", length(x), "element(s)\n")
    cat("\n* styles:\n")

    style_names <- styles_info(x)
    style_sample <- style_names$style_type
    names(style_sample) <- style_names$style_name
    print(style_sample)

    if (length(x) > 1) {
      cursor_elt <- docx_current_block_xml(x)
      cat("\n* Content at cursor location:\n")
      print(node_content(cursor_elt, x))
    } else {
      cat("\n* empty document\n")
    }
    return(invisible())
  }

  if (!grepl(x = target, pattern = "\\.(docx)$", ignore.case = TRUE)) {
    stop(target, " should have '.docx' extension.")
  }

  if (is_windows() && is_doc_open(target)) {
    stop(target, " is open. To write to this document, please, close it.")
  }

  x <- process_sections(x)

  process_comments(x)
  process_footnotes(x)
  process_stylenames(x$doc_obj, x$styles)
  process_links(x$doc_obj, type = "wml")
  process_docx_poured(
    doc_obj = x$doc_obj,
    relationships = x$doc_obj$relationship(),
    content_type = x$content_type,
    package_dir = x$package_dir
  )
  process_images(x$doc_obj, x$doc_obj$relationship(), x$package_dir)
  process_images(x$footnotes, x$footnotes$relationship(), x$package_dir)

  x$headers <- update_hf_list(part_list = x$headers, type = "header", package_dir = x$package_dir)
  x$footers <- update_hf_list(part_list = x$footers, type = "footer", package_dir = x$package_dir)
  for (header in x$headers) process_links(header, type = "wml")
  for (footer in x$footers) process_links(footer, type = "wml")
  for (header in x$headers) process_images(header, header$relationship(), x$package_dir)
  for (footer in x$footers) process_images(footer, footer$relationship(), x$package_dir)

  int_id <- 1 # unique id identifier

  # make all id unique for document
  int_id <- correct_id(x$doc_obj$get(), int_id)
  # make all id unique for footnote
  int_id <- correct_id(x$footnotes$get(), int_id)
  # make all id unique for footnote
  int_id <- correct_id(x$comments$get(), int_id)
  # make all id unique for headers
  for (docpart in x[["headers"]]) {
    int_id <- correct_id(docpart$get(), int_id)
  }
  # make all id unique for footers
  for (docpart in x[["footers"]]) {
    int_id <- correct_id(docpart$get(), int_id)
  }

  body <- xml_find_first(x$doc_obj$get(), "w:body")

  # If body is not ending with an sectPr, create a continuous one append it
  if (!xml_name(xml_child(body, search = xml_length(body))) %in% "sectPr") {
    str <- paste0(
      "<w:sectPr xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" xmlns:wp=\"http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing\" xmlns:r=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships\" xmlns:w14=\"http://schemas.microsoft.com/office/word/2010/wordml\">",
      "<w:type w:val=\"continuous\"/></w:sectPr>"
    )
    xml_add_child(body, as_xml_document(str))
  }

  for (header in x$headers) {
    header$save()
  }
  for (footer in x$footers) {
    footer$save()
  }
  x$doc_obj$save()
  x$content_type$save()
  x$footnotes$save()
  x$comments$save()

  x$rel$write(file.path(x$package_dir, "_rels", ".rels"))
  write_docx_settings(x)

  # save doc properties
  if (nrow(x$doc_properties$data) > 0) {
    x$doc_properties["modified", "value"] <- format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ")
    x$doc_properties["lastModifiedBy", "value"] <- Sys.getenv("USER")
    write_core_properties(x$doc_properties, x$package_dir)
  }
  if (nrow(x$doc_properties_custom$data) > 0) {
    write_custom_properties(x$doc_properties_custom, x$package_dir)
  }
  invisible(pack_folder(folder = x$package_dir, target = target))
}



#' @export
#' @title Number of blocks inside an rdocx object
#' @description return the number of blocks inside an rdocx object.
#' This number also include the default section definition of a
#' Word document - default Word section is an uninvisible element.
#' @param x an rdocx object
#' @examples
#' # how many elements are there in an new document produced
#' # with the default template.
#' length( read_docx() )
#' @family functions for Word document informations
length.rdocx <- function( x ){
  xml_length(xml_child(x$doc_obj$get(), "w:body"))
}

#' @export
#' @title Read 'Word' styles
#' @description read Word styles and get results in
#' a data.frame.
#' @param x an rdocx object
#' @param type,is_default subsets for types (i.e. paragraph) and
#' default style (when `is_default` is TRUE or FALSE)
#' @examples
#' x <- read_docx()
#' styles_info(x)
#' styles_info(x, type = "paragraph", is_default = TRUE)
#' @family functions for Word document informations
styles_info <- function( x, type = c("paragraph", "character", "table", "numbering"),
                         is_default = c(TRUE, FALSE) ){
  styles <- x$styles
  styles <- styles[styles$style_type %in% type & styles$is_default %in% is_default,]
  styles
}

#' @export
#' @title Add or replace paragraph style in a Word document
#' @description The function lets you add or replace a Word paragraph style.
#' @param x an rdocx object
#' @param style_id a unique style identifier for Word.
#' @param style_name a unique label associated with the style identifier.
#' This label is the name of the style when Word edit the document.
#' @param base_on the style name used as base style
#' @param fp_p paragraph formatting properties, see [fp_par()].
#' @param fp_t default text formatting properties. This is used as
#' text formatting properties, see [fp_text()]. If NULL (default), the
#' paragraph will used the default text formatting properties (defined by
#' the `base_on` argument).
#' @examples
#' library(officer)
#'
#' doc <- read_docx()
#'
#' doc <- docx_set_paragraph_style(
#'   doc,
#'   style_id = "rightaligned",
#'   style_name = "Explicit label",
#'   fp_p = fp_par(text.align = "right", padding = 20),
#'   fp_t = fp_text_lite(
#'     bold = TRUE,
#'     shading.color = "#FD34F0",
#'     color = "white")
#' )
#'
#' doc <- body_add_par(doc,
#'   value = "This is a test",
#'   style = "Explicit label")
#'
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_file
docx_set_paragraph_style <- function(x, style_id, style_name, base_on = "Normal", fp_p = fp_par(), fp_t = NULL) {
  styles_file <- file.path(x$package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)

  if (grepl("[^a-zA-Z0-9\\-]+", style_id)) {
    stop("`style_id` should only contain '-', numbers and ascii characters.")
  }

  node_styles <- xml_find_first(doc, "/w:styles")

  fp_p$word_style <- NULL

  if (!is.null(fp_t)){
    fp_t_xml <- rpr_wml(fp_t)
  } else {
    fp_t_xml <- ""
  }
  base_on <- get_style_id(data = x$styles, style = base_on, type = "paragraph")

  xml_code <- paste0(
    sprintf("<w:style xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:type=\"paragraph\" w:customStyle=\"1\" w:styleId=\"%s\">", style_id),
    sprintf("<w:name w:val=\"%s\"/>", style_name),
    sprintf("<w:basedOn w:val=\"%s\"/>", base_on),
    ppr_wml(fp_p), fp_t_xml,
    "</w:style>"
  )

  node_style <- xml_child(node_styles, sprintf("w:style[@w:styleId='%s']", style_id))
  if (inherits(node_style, "xml_missing")) {
    xml_add_child(node_styles, as_xml_document(xml_code))
  } else {
    xml_replace(node_style, as_xml_document(xml_code))
  }

  write_xml(doc, file = styles_file)
  styles <- read_docx_styles(x$package_dir)
  x$styles <- styles

  x
}

#' @export
#' @title Add character style in a Word document
#' @description The function lets you add or modify Word character styles.
#' @param x an rdocx object
#' @param style_id a unique style identifier for Word.
#' @param style_name a unique label associated with the style identifier.
#' This label is the name of the style when Word edit the document.
#' @param base_on the character style name used as base style
#' @param fp_t Text formatting properties, see [fp_text()].
#' @examples
#' library(officer)
#' doc <- read_docx()
#'
#' doc <- docx_set_character_style(
#'   doc,
#'   style_id = "newcharstyle",
#'   style_name = "label for char style",
#'   base_on = "Default Paragraph Font",
#'   fp_text_lite(
#'     shading.color = "red",
#'     color = "white")
#' )
#' paragraph <- fpar(
#'   run_wordtext("hello",
#'     style_id = "newcharstyle"))
#'
#' doc <- body_add_fpar(doc, value = paragraph)
#' docx_file <- print(doc, target = tempfile(fileext = ".docx"))
#' docx_file
docx_set_character_style <- function(x, style_id, style_name, base_on, fp_t = fp_text_lite()) {
  styles_file <- file.path(x$package_dir, "word/styles.xml")
  doc <- read_xml(styles_file)
  node_styles <- xml_find_first(doc, "/w:styles")

  if (grepl("[^a-zA-Z0-9\\-]+", style_id)) {
    stop("`style_id` should only contain '-', numbers and ascii characters.")
  }

  base_on <- get_style_id(data = x$styles, style = base_on, type = "character")

  xml_code <- paste0(
    sprintf("<w:style xmlns:w=\"http://schemas.openxmlformats.org/wordprocessingml/2006/main\" w:type=\"character\" w:customStyle=\"1\" w:styleId=\"%s\">", style_id),
    sprintf("<w:name w:val=\"%s\"/>", style_name),
    sprintf("<w:basedOn w:val=\"%s\"/>", base_on),
    rpr_wml(fp_t),
    "</w:style>"
  )

  node_style <- xml_child(node_styles, sprintf("w:style[@w:styleId='%s']", style_id))
  if (inherits(node_style, "xml_missing")) {
    xml_add_child(node_styles, as_xml_document(xml_code))
  } else {
    xml_replace(node_style, as_xml_document(xml_code))
  }

  write_xml(doc, file = styles_file)
  styles <- read_docx_styles(x$package_dir)
  x$styles <- styles

  x
}


#' @export
#' @title Read document properties
#' @description Read Word or PowerPoint document properties
#' and get results in a data.frame.
#' @param x an `rdocx` or `rpptx` object
#' @examples
#' x <- read_docx()
#' doc_properties(x)
#' @return a data.frame
#' @family functions for Word document informations
#' @family functions for reading presentation information
doc_properties <- function(x) {
  if (inherits(x, "rdocx")) {
    cp <- x$doc_properties
  } else if (inherits(x, "rpptx") || inherits(x, "rxlsx")) {
    cp <- x$core_properties
  } else {
    stop("x should be a rpptx or a rdocx or a rxlsx object.")
  }

  properties_custom <- x$doc_properties_custom

  out_custom <- data.frame(
    tag = properties_custom[, "name"],
    value = properties_custom[, "value"],
    stringsAsFactors = FALSE
  )
  out <- data.frame(
    tag = cp[, "name"],
    value = cp[, "value"],
    stringsAsFactors = FALSE
  )
  out <- rbind(out, out_custom)
  row.names(out) <- NULL
  out
}

#' @export
#' @title Set document properties
#' @description set Word or PowerPoint document properties. These are not visible
#' in the document but are available as metadata of the document.
#'
#' Any character property can be added as a document property.
#' It provides an easy way to insert arbitrary fields. Given the challenges
#' that can be encountered with find-and-replace in word with officer, the
#' use of document fields and quick text fields provides a much more robust
#' approach to automatic document generation from R.
#' @note
#' The "last modified" and "last modified by" fields will be automatically be updated
#' when the file is written.
#' @param x an rdocx or rpptx object
#' @param title,subject,creator,description text fields
#' @param created a date object
#' @param ... named arguments (names are field names), each element is a single
#' character value specifying value associated with the corresponding field name.
#' @param values a named list (names are field names), each element is a single
#' character value specifying value associated with the corresponding field name.
#' If `values` is provided, argument `...` will be ignored.
#' @examples
#' x <- read_docx()
#' x <- set_doc_properties(x, title = "title",
#'   subject = "document subject", creator = "Me me me",
#'   description = "this document is empty",
#'   created = Sys.time(),
#'   yoyo = "yok yok",
#'   glop = "pas glop")
#' x <- doc_properties(x)
#' @family functions for Word document informations
set_doc_properties <- function( x, title = NULL, subject = NULL,
                                creator = NULL, description = NULL, created = NULL,
                                ..., values = NULL){

  if( inherits(x, "rdocx"))
    cp <- x$doc_properties
  else if( inherits(x, "rpptx")) cp <- x$core_properties
  else stop("x should be a rpptx or rdocx object.")

  if( !is.null(title) ) cp['title','value'] <- title
  if( !is.null(subject) ) cp['subject','value'] <- subject
  if( !is.null(creator) ) cp['creator','value'] <- creator
  if( !is.null(description) ) cp['description','value'] <- description
  if( !is.null(created) ) cp['created','value'] <- format( created, "%Y-%m-%dT%H:%M:%SZ")

  if (is.null(values)) {
    values <- list(...)
  }

  if (length(values) > 0) {
    x$content_type$add_override(
      setNames("application/vnd.openxmlformats-officedocument.custom-properties+xml",
               "/docProps/custom.xml")
    )
    x$rel$add(id = paste0("rId", x$rel$get_next_id()),
                type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/custom-properties",
                target = "docProps/custom.xml")

    custom_props <- x$doc_properties_custom
    for(i in seq_along(values)) {
      custom_props[names(values)[i], 'value'] <- enc2utf8(values[[i]])
    }
    x$doc_properties_custom <- custom_props
  }

  if( inherits(x, "rdocx"))
    x$doc_properties <- cp
  else x$core_properties <- cp

  x
}


#' @export
#' @title 'Word' page layout
#' @description Get page width, page height and margins (in inches). The return values
#' are those corresponding to the section where the cursor is.
#' @param x an \code{rdocx} object
#' @examples
#' docx_dim(read_docx())
#' @family functions for Word document informations
docx_dim <- function(x){
  cursor <- as.character(x$officer_cursor)
  if (is.na(cursor)) {
    next_section <- xml_find_first(x$doc_obj$get(), "/w:document/w:body/w:sectPr")
  } else {
    xpath_ <- paste0(
      file.path( cursor, "following-sibling::w:sectPr"),
      "|",
      file.path( cursor, "following-sibling::w:p/w:pPr/w:sectPr"),
      "|",
      "//w:sectPr"
    )
    next_section <- xml_find_first(x$doc_obj$get(), xpath_)
  }

  sd <- section_dimensions(next_section)
  sd$page <- sd$page / (20*72)
  sd$margins <- sd$margins / (20*72)
  sd
}


#' @export
#' @title List Word bookmarks
#' @description List bookmarks id that can be found in a
#' 'Word' document.
#' @param x an \code{rdocx} object
#' @examples
#' library(officer)
#'
#' doc_1 <- read_docx()
#' doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
#' doc_1 <- body_bookmark(doc_1, "text_to_replace_1")
#' doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
#' doc_1 <- body_bookmark(doc_1, "text_to_replace_2")
#'
#' docx_bookmarks(doc_1)
#'
#' docx_bookmarks(read_docx())
#' @family functions for Word document informations
docx_bookmarks <- function(x){
  stopifnot(inherits(x, "rdocx"))

  doc_ <- xml_find_all(x$doc_obj$get(), "//w:bookmarkStart[@w:name]")
  setdiff(xml_attr(doc_, "name"), "_GoBack")
}

#' @export
#' @title Replace styles in a 'Word' Document
#' @description Replace styles with others in a 'Word' document. This function
#' can be used for paragraph, run/character and table styles.
#' @param x an rdocx object
#' @param mapstyles a named list, names are the replacement style,
#' content (as a character vector) are the styles to be replaced.
#' Use [styles_info()] to display available styles.
#' @examples
#' # creating a sample docx so that we can illustrate how
#' # to change styles
#' doc_1 <- read_docx()
#'
#' doc_1 <- body_add_par(doc_1, "A title", style = "heading 1")
#' doc_1 <- body_add_par(doc_1, "Another title", style = "heading 2")
#' doc_1 <- body_add_par(doc_1, "Hello world!", style = "Normal")
#' file <- print(doc_1, target = tempfile(fileext = ".docx"))
#'
#' # now we can illustrate how
#' # to change styles with `change_styles`
#' doc_2 <- read_docx(path = file)
#' mapstyles <- list(
#'   "centered" = c("Normal", "heading 2"),
#'   "strong" = "Default Paragraph Font"
#' )
#' doc_2 <- change_styles(doc_2, mapstyles = mapstyles)
#' print(doc_2, target = tempfile(fileext = ".docx"))
change_styles <- function( x, mapstyles ){

  if( is.null(mapstyles) || length(mapstyles) < 1 ) return(x)

  table_styles <- styles_info(x, type = c("paragraph", "character", "table"))

  from_styles <- unique( as.character( unlist(mapstyles) ) )
  to_styles <- unique( names( mapstyles) )

  if( any( is.na( mfrom <- match( from_styles, table_styles$style_name ) ) ) ){
    stop("could not find style ", paste0( shQuote(from_styles[is.na(mfrom)]), collapse = ", " ), ".", call. = FALSE)
  }
  if( any( is.na( mto <- match( to_styles, table_styles$style_name ) ) ) ){
    stop("could not find style ", paste0( shQuote(to_styles[is.na(mto)]), collapse = ", " ), ".", call. = FALSE)
  }

  mapping <- mapply(function(from, to) {
    id_to <- which( table_styles$style_name %in% to )
    id_to <- table_styles$style_id[id_to]

    id_from <- which( table_styles$style_name %in% from )
    types <- substring(table_styles$style_type[id_from], first = 1, last = 1)
    types[types %in% "c"] <- "r"
    types[types %in% "t"] <- "tbl"
    id_from <- table_styles$style_id[id_from]

    data.frame( from = id_from, to = rep(id_to, length(from)), types = types, stringsAsFactors = FALSE )
  }, mapstyles, names(mapstyles), SIMPLIFY = FALSE)

  mapping <- do.call(rbind, mapping)
  row.names(mapping) <- NULL

  for(i in seq_len( nrow(mapping) )){
    all_nodes <- xml_find_all(x$doc_obj$get(), sprintf("//w:%sStyle[@w:val='%s']", mapping$types[i], mapping$from[i]))
    xml_attr(all_nodes, "w:val") <- rep(mapping$to[i], length(all_nodes) )
  }

  x
}



#' @export
#' @title Body xml document
#' @description Get the body document as xml. This function
#' is not to be used by end users, it has been implemented
#' to allow other packages to work with officer.
#' @param x an rdocx object
#' @examples
#' doc <- read_docx()
#' docx_body_xml(doc)
#' @keywords internal
docx_body_xml <- function( x ){
  x$doc_obj$get()
}
#' @export
#' @title xml element on which cursor is
#' @description Get the current block element as xml. This function
#' is not to be used by end users, it has been implemented
#' to allow other packages to work with officer. If the
#' document is empty, this block will be set to NULL.
#' @param x an rdocx object
#' @examples
#' doc <- read_docx()
#' docx_current_block_xml(doc)
#' @keywords internal
docx_current_block_xml <- function( x ){
  ooxml_on_cursor(x$officer_cursor, x$doc_obj$get())
}

#' @export
#' @title Body xml document
#' @description Get the body document as xml. This function
#' is not to be used by end users, it has been implemented
#' to allow other packages to work with officer.
#' @param x an rdocx object
#' @examples
#' doc <- read_docx()
#' docx_body_relationship(doc)
#' @keywords internal
docx_body_relationship <- function( x ){
  x$doc_obj$relationship()
}

test_that("read docx properties", {
  doc <- read_docx(
    path = system.file(package = "officer", "doc_examples/example.docx")
  )
  properties <- doc_properties(doc)
  expect_equal(properties$value[properties$tag %in% "title"], "document title")
  expect_equal(
    properties$value[properties$tag %in% "subject"],
    "document subject"
  )
  expect_equal(properties$value[properties$tag %in% "creator"], "author")
  expect_equal(
    properties$value[properties$tag %in% "description"],
    "these are comments"
  )
  expect_equal(
    properties$value[properties$tag %in% "created"],
    "2017-04-26T13:10:00Z"
  )
})

test_that("read pptx properties", {
  doc <- read_pptx(
    path = system.file(package = "officer", "doc_examples/example.pptx")
  )
  properties <- doc_properties(doc)
  expect_equal(properties$value[properties$tag %in% "title"], "document title")
  expect_equal(
    properties$value[properties$tag %in% "subject"],
    "document subject"
  )
  expect_equal(properties$value[properties$tag %in% "creator"], "author")
  expect_equal(
    properties$value[properties$tag %in% "description"],
    "these are comments"
  )
  expect_equal(
    properties$value[properties$tag %in% "created"],
    "2017-04-27T11:29:40Z"
  )
})


test_that("set docx properties", {
  doc <- read_docx()
  time_now <- Sys.time()
  filename <- tempfile(fileext = ".docx")
  doc <- set_doc_properties(
    doc,
    title = "title",
    subject = "document subject",
    creator = "Me me me",
    description = "this document is not empty",
    created = time_now
  )
  dooc <- print(doc, target = filename)

  doc <- read_docx(path = filename)
  properties <- doc_properties(doc)

  expect_equal(properties$value[properties$tag %in% "title"], "title")
  expect_equal(
    properties$value[properties$tag %in% "subject"],
    "document subject"
  )
  expect_equal(properties$value[properties$tag %in% "creator"], "Me me me")
  expect_equal(
    properties$value[properties$tag %in% "description"],
    "this document is not empty"
  )
  expect_equal(
    properties$value[properties$tag %in% "created"],
    format(time_now, "%Y-%m-%dT%H:%M:%SZ")
  )
})

test_that("set custom properties", {
  filename <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- set_doc_properties(
    doc,
    coco = "coucou",
    zozo = "zuzu",
    isbn = NULL,
    bozo = NA_character_
  )
  print(doc, target = filename)

  doc <- read_docx(path = filename)
  properties <- doc_properties(doc)

  expect_equal(properties$value[properties$tag %in% "coco"], "coucou")
  expect_equal(properties$value[properties$tag %in% "zozo"], "zuzu")
  expect_equal(properties$value[properties$tag %in% "isbn"], "")
  expect_equal(properties$value[properties$tag %in% "bozo"], "")

  filename <- tempfile(fileext = ".pptx")
  doc <- read_pptx()
  doc <- set_doc_properties(doc, coco = "coucou", zozo = "zuzu")
  print(doc, target = filename)

  doc <- read_pptx(path = filename)
  properties <- doc_properties(doc)

  expect_equal(properties$value[properties$tag %in% "coco"], "coucou")
  expect_equal(properties$value[properties$tag %in% "zozo"], "zuzu")

  doc <- read_docx()
  filename <- tempfile(fileext = ".docx")
  doc <- set_doc_properties(doc, values = list(coco = "coucou", zozo = "zuzu"))
  print(doc, target = filename)

  doc <- read_docx(path = filename)
  properties <- doc_properties(doc)

  expect_equal(properties$value[properties$tag %in% "coco"], "coucou")
  expect_equal(properties$value[properties$tag %in% "zozo"], "zuzu")

  doc <- read_pptx()
  filename <- tempfile(fileext = ".pptx")
  doc <- set_doc_properties(doc, values = list(coco = "coucou", zozo = "zuzu"))
  print(doc, target = filename)

  doc <- read_pptx(path = filename)
  properties <- doc_properties(doc)

  expect_equal(properties$value[properties$tag %in% "coco"], "coucou")
  expect_equal(properties$value[properties$tag %in% "zozo"], "zuzu")
})


test_that("set pptx properties", {
  doc <- read_pptx()
  time_now <- Sys.time()
  filename <- tempfile(fileext = ".pptx")
  doc <- set_doc_properties(
    doc,
    title = "title",
    subject = "document subject",
    creator = "Me me me",
    description = "this document is not empty",
    created = time_now
  )
  print(doc, target = filename)

  doc <- read_pptx(path = filename)
  properties <- doc_properties(doc)

  expect_equal(properties$value[properties$tag %in% "title"], "title")
  expect_equal(
    properties$value[properties$tag %in% "subject"],
    "document subject"
  )
  expect_equal(properties$value[properties$tag %in% "creator"], "Me me me")
  expect_equal(
    properties$value[properties$tag %in% "description"],
    "this document is not empty"
  )
  expect_equal(
    properties$value[properties$tag %in% "created"],
    format(time_now, "%Y-%m-%dT%H:%M:%SZ")
  )
})

test_that("read docx properties", {
  doc <- read_docx(path = system.file(package = "officer", "doc_examples/example.docx"))
  properties <- doc_properties(doc)
  expect_equal( properties$value[properties$tag %in% "title"], "document title" )
  expect_equal( properties$value[properties$tag %in% "subject"], "document subject" )
  expect_equal( properties$value[properties$tag %in% "creator"], "author" )
  expect_equal( properties$value[properties$tag %in% "description"], "these are comments" )
  expect_equal( properties$value[properties$tag %in% "created"], "2017-04-26T13:10:00Z" )

})

test_that("read pptx properties", {
  doc <- read_pptx(path = system.file(package = "officer", "doc_examples/example.pptx"))
  properties <- doc_properties(doc)
  expect_equal( properties$value[properties$tag %in% "title"], "document title" )
  expect_equal( properties$value[properties$tag %in% "subject"], "document subject" )
  expect_equal( properties$value[properties$tag %in% "creator"], "author" )
  expect_equal( properties$value[properties$tag %in% "description"], "these are comments" )
  expect_equal( properties$value[properties$tag %in% "created"], "2017-04-27T11:29:40Z" )

})


test_that("set docx properties", {
  doc <- read_docx()
  time_now <- Sys.time()
  filename <- tempfile(fileext = ".docx")
  doc <- set_doc_properties(doc, title = "title",
                       subject = "document subject", creator = "Me me me",
                       description = "this document is not empty",
                       created = time_now )
  dooc <- print(doc, target = filename )

  doc <- read_docx(path = filename)
  properties <- doc_properties(doc)

  expect_equal( properties$value[properties$tag %in% "title"], "title" )
  expect_equal( properties$value[properties$tag %in% "subject"], "document subject" )
  expect_equal( properties$value[properties$tag %in% "creator"], "Me me me" )
  expect_equal( properties$value[properties$tag %in% "description"], "this document is not empty" )
  expect_equal( properties$value[properties$tag %in% "created"], format( time_now, "%Y-%m-%dT%H:%M:%SZ") )

})

test_that("set custom properties", {
  filename <- tempfile(fileext = ".docx")
  doc <- read_docx()
  doc <- set_doc_properties(doc, coco = "coucou", zozo = "zuzu")
  print(doc, target = filename )

  doc <- read_docx(path = filename)
  properties <- doc_properties(doc)

  expect_equal( properties$value[properties$tag %in% "coco"], "coucou" )
  expect_equal( properties$value[properties$tag %in% "zozo"], "zuzu" )

  filename <- tempfile(fileext = ".pptx")
  doc <- read_pptx()
  doc <- set_doc_properties(doc, coco = "coucou", zozo = "zuzu")
  print(doc, target = filename )

  doc <- read_pptx(path = filename)
  properties <- doc_properties(doc)

  expect_equal( properties$value[properties$tag %in% "coco"], "coucou" )
  expect_equal( properties$value[properties$tag %in% "zozo"], "zuzu" )

  doc <- read_docx()
  filename <- tempfile(fileext = ".docx")
  doc <- set_doc_properties(doc, values = list(coco = "coucou", zozo = "zuzu"))
  print(doc, target = filename )

  doc <- read_docx(path = filename)
  properties <- doc_properties(doc)

  expect_equal( properties$value[properties$tag %in% "coco"], "coucou" )
  expect_equal( properties$value[properties$tag %in% "zozo"], "zuzu" )

  doc <- read_pptx()
  filename <- tempfile(fileext = ".pptx")
  doc <- set_doc_properties(doc, values = list(coco = "coucou", zozo = "zuzu"))
  print(doc, target = filename )

  doc <- read_pptx(path = filename)
  properties <- doc_properties(doc)

  expect_equal( properties$value[properties$tag %in% "coco"], "coucou" )
  expect_equal( properties$value[properties$tag %in% "zozo"], "zuzu" )
})



test_that("set pptx properties", {
  doc <- read_pptx()
  time_now <- Sys.time()
  filename <- tempfile(fileext = ".pptx")
  doc <- set_doc_properties(doc, title = "title",
                       subject = "document subject", creator = "Me me me",
                       description = "this document is not empty",
                       created = time_now )
  print(doc, target = filename )

  doc <- read_pptx(path = filename)
  properties <- doc_properties(doc)

  expect_equal( properties$value[properties$tag %in% "title"], "title" )
  expect_equal( properties$value[properties$tag %in% "subject"], "document subject" )
  expect_equal( properties$value[properties$tag %in% "creator"], "Me me me" )
  expect_equal( properties$value[properties$tag %in% "description"], "this document is not empty" )
  expect_equal( properties$value[properties$tag %in% "created"], format( time_now, "%Y-%m-%dT%H:%M:%SZ") )

})

test_that("styles_info is returning a df", {
  x <- read_docx()
  df <- styles_info(x)

  expect_is(df, "data.frame")
  expect_true(all(
    c("style_type", "style_id", "style_name", "is_custom", "is_default") %in%
      names(df)
  ))
  expect_is(df$style_type, "character")
  expect_is(df$style_id, "character")
  expect_is(df$style_name, "character")
  expect_is(df$is_custom, "logical")
  expect_is(df$is_default, "logical")
})

test_that("set par style", {
  doc <- read_docx()
  doc <- docx_set_paragraph_style(
    doc,
    style_id = "rightaligned",
    style_name = "Explicit label",
    fp_p = fp_par(text.align = "right", padding = 20),
    fp_t = fp_text_lite(
      bold = TRUE,
      shading.color = "#FD34F0",
      color = "white"
    )
  )
  doc <- body_add_par(
    doc,
    value = "This is a test",
    style = "Explicit label"
  )

  current_node <- docx_current_block_xml(doc)
  expect_equal(
    xml_attr(xml_child(current_node, "w:pPr/w:pStyle"), "val"),
    "rightaligned"
  )

  df <- styles_info(doc, type = "paragraph")
  df <- df[df$style_id %in% "rightaligned", ]
  expect_true(nrow(df) == 1L)
  if (nrow(df) == 1L) {
    expect_equal(df$shading.color, "FD34F0")
    expect_equal(df$color, "FFFFFF")
    expect_equal(df$align, "right")
  }
})

test_that("set char style", {
  doc <- read_docx()
  doc <- docx_set_character_style(
    doc,
    style_id = "bold",
    base_on = "Default Paragraph Font",
    style_name = "Bold test",
    fp_t = fp_text_lite(bold = TRUE)
  )
  doc <- body_add_fpar(
    doc,
    fpar(run_wordtext(text = "coco", style_id = "bold")),
    style = "Normal"
  )

  current_node <- docx_current_block_xml(doc)
  expect_equal(
    xml_attr(xml_child(current_node, "w:pPr/w:pStyle"), "val"),
    "Normal"
  )

  df <- styles_info(doc, type = "character")
  df <- df[df$style_id %in% "bold", ]
  expect_true(nrow(df) == 1L)
  if (nrow(df) == 1L) {
    expect_equal(df$bold, "true")
    expect_equal(df$style_name, "Bold test")
  }
})

test_that("change_styles changes paragraph styles", {
  # Create a document with different paragraph styles
  doc <- read_docx()
  doc <- body_add_par(doc, "Paragraph with Normal style", style = "Normal")
  doc <- body_add_par(doc, "Paragraph with heading 2", style = "heading 2")
  doc <- body_add_par(doc, "Another Normal paragraph", style = "Normal")

  # Save and reload to ensure styles are properly set
  tf <- tempfile(fileext = ".docx")
  print(doc, target = tf)
  doc <- read_docx(path = tf)

  # Verify original styles
  summary_before <- docx_summary(doc)
  expect_equal(
    summary_before$style_name,
    c("Normal", "heading 2", "Normal")
  )

  # Change styles: Normal and heading 2 -> centered
  mapstyles <- list("centered" = c("Normal", "heading 2"))
  doc <- change_styles(doc, mapstyles = mapstyles)

  # Save and reload to verify changes
  tf2 <- tempfile(fileext = ".docx")
  print(doc, target = tf2)
  doc_changed <- read_docx(path = tf2)

  summary_after <- docx_summary(doc_changed)
  expect_equal(
    summary_after$style_name,
    c("centered", "centered", "centered")
  )

  unlink(tf)
  unlink(tf2)
})

test_that("change_styles changes character styles", {
  # Create a document with character styles
  doc <- read_docx()
  doc <- docx_set_character_style(
    doc,
    style_id = "mystyle",
    base_on = "Default Paragraph Font",
    style_name = "My Style",
    fp_t = fp_text_lite(bold = TRUE)
  )
  doc <- body_add_fpar(
    doc,
    fpar(
      run_wordtext(text = "default text", style_id = "mystyle"),
      ftext(" and styled text", prop = fp_text_lite(bold = TRUE))
    ),
    style = "Normal"
  )

  # Save and reload
  tf <- tempfile(fileext = ".docx")
  print(doc, target = tf)
  doc <- read_docx(path = tf)

  # Change character style
  mapstyles <- list("strong" = "My Style")
  doc <- change_styles(doc, mapstyles = mapstyles)

  all_nodes <- xml2::xml_find_all(
    docx_body_xml(doc),
    "//w:rStyle[@w:val='strong']"
  )
  expect_true(length(all_nodes) > 0)

  unlink(tf)
})

test_that("change_styles handles multiple mappings", {
  doc <- read_docx()
  doc <- body_add_par(doc, "Normal text", style = "Normal")
  doc <- body_add_par(doc, "Heading 1", style = "heading 1")
  doc <- body_add_par(doc, "Heading 2", style = "heading 2")

  tf <- tempfile(fileext = ".docx")
  print(doc, target = tf)
  doc <- read_docx(path = tf)

  # Map multiple styles
  mapstyles <- list(
    "centered" = c("Normal", "heading 2"),
    "heading 1" = "heading 1"
  )
  doc <- change_styles(doc, mapstyles = mapstyles)

  tf2 <- tempfile(fileext = ".docx")
  print(doc, target = tf2)
  doc_changed <- read_docx(path = tf2)

  summary_after <- docx_summary(doc_changed)
  expect_equal(
    summary_after$style_name,
    c("centered", "heading 1", "centered")
  )

  unlink(tf)
  unlink(tf2)
})

test_that("change_styles returns unchanged doc when mapstyles is NULL", {
  doc <- read_docx()
  doc <- body_add_par(doc, "Test paragraph", style = "Normal")

  doc_unchanged <- change_styles(doc, mapstyles = NULL)
  expect_identical(doc, doc_unchanged)

  doc_unchanged2 <- change_styles(doc, mapstyles = list())
  expect_identical(doc, doc_unchanged2)
})

test_that("change_styles errors when style not found", {
  doc <- read_docx()
  doc <- body_add_par(doc, "Test", style = "Normal")

  # Error when 'from' style doesn't exist
  mapstyles <- list("centered" = "NonExistentStyle")
  expect_error(
    change_styles(doc, mapstyles = mapstyles),
    "could not find style.*NonExistentStyle"
  )

  # Error when 'to' style doesn't exist
  mapstyles <- list("NonExistentTarget" = "Normal")
  expect_error(
    change_styles(doc, mapstyles = mapstyles),
    "could not find style.*NonExistentTarget"
  )
})

test_that("styles_info is returning a df", {
  x <- read_docx()
  df <- styles_info(x)

  expect_is(df, "data.frame")
  expect_true(all(c("style_type", "style_id", "style_name", "is_custom", "is_default") %in% names(df)))
  expect_is(df$style_type, "character")
  expect_is(df$style_id, "character")
  expect_is(df$style_name, "character")
  expect_is(df$is_custom, "logical")
  expect_is(df$is_default, "logical")
})

test_that("set par style", {

  doc <- read_docx()
  doc <- docx_set_paragraph_style(
    doc,
    style_id = "rightaligned",
    style_name = "Explicit label",
    fp_p = fp_par(text.align = "right", padding = 20),
    fp_t = fp_text_lite(
      bold = TRUE,
      shading.color = "#FD34F0",
      color = "white")
  )
  doc <- body_add_par(
    doc, value = "This is a test",
    style = "Explicit label")

  current_node <- docx_current_block_xml(doc)
  expect_equal(
    xml_attr(xml_child(current_node, "w:pPr/w:pStyle"), "val"),
    "rightaligned")

  df <- styles_info(doc, type = "paragraph")
  df <- df[df$style_id %in% "rightaligned",]
  expect_true(nrow(df) == 1L)
  if (nrow(df) == 1L) {
    expect_equal(df$shading.color, "FD34F0")
    expect_equal(df$color, "FFFFFF")
    expect_equal(df$align, "right")
  }

})

test_that("set char style", {

  doc <- read_docx()
  doc <- docx_set_character_style(
    doc,
    style_id = "bold",
    base_on = "Default Paragraph Font",
    style_name = "Bold test",
    fp_t = fp_text_lite(bold = TRUE)
  )
  doc <- body_add_fpar(
    doc, fpar(run_wordtext(text = "coco", style_id = "bold")),
    style = "Normal")

  current_node <- docx_current_block_xml(doc)
  expect_equal(
    xml_attr(xml_child(current_node, "w:pPr/w:pStyle"), "val"),
    "Normal")

  df <- styles_info(doc, type = "character")
  df <- df[df$style_id %in% "bold",]
  expect_true(nrow(df) == 1L)
  if (nrow(df) == 1L) {
    expect_equal(df$bold, "true")
    expect_equal(df$style_name, "Bold test")
  }

})

#' @export
#' @title Embed Fonts in a Word Document
#' @description Copy TrueType or OpenType font files from the local
#' system into a Word document. The font data is stored inside the
#' `.docx` archive so that the document renders correctly when opened
#' on a system where the font is not installed.
#'
#' When only `font_family` is provided (without `regular`), the function
#' looks up font file paths on the local system via [gdtools::sys_fonts()]
#' and copies them into the document. A message shows the equivalent
#' explicit call for reproducibility.
#'
#' @section Font licensing:
#' Embedding a font in a document redistributes it. You must ensure that
#' the font license permits embedding. Fonts under the SIL Open Font
#' License (e.g. Liberation, Google Fonts) generally allow it. Many
#' commercial fonts restrict or prohibit embedding. Check the license
#' of the font before using this function.
#' @param x an rdocx object
#' @param font_family font family name as it will appear in the document.
#' When `regular` is not provided, this name is used to look up font
#' files via [gdtools::sys_fonts()].
#' @param regular path to the regular .ttf/.otf font file. If `NULL`,
#' font files are detected automatically via gdtools.
#' @param bold path to the bold .ttf/.otf font file (optional)
#' @param italic path to the italic .ttf/.otf font file (optional)
#' @param bold_italic path to the bold-italic .ttf/.otf font file (optional)
#' @return the rdocx object with embedded fonts
#' @seealso [gdtools::sys_fonts()], [gdtools::register_gfont()],
#' [docx_set_settings()]
#' @examplesIf require("gdtools", quietly = TRUE, character.only = TRUE)
#' # automatic detection (requires gdtools)
#' gdtools::register_liberationsans()
#' doc <- read_docx()
#' doc <- docx_embed_font(doc, font_family = "Liberation Sans")
#'
#' # # explicit paths (no gdtools needed)
#' # doc <- docx_embed_font(
#' #   doc,
#' #   font_family = "My Font",
#' #   regular = "path/to/font-regular.ttf",
#' #   bold = "path/to/font-bold.ttf"
#' # )
docx_embed_font <- function(
  x,
  font_family,
  regular = NULL,
  bold = NULL,
  italic = NULL,
  bold_italic = NULL
) {
  if (!inherits(x, "rdocx")) {
    cli::cli_abort("{.arg x} must be an rdocx object.")
  }

  if (is.null(regular)) {
    detected <- detect_font_files(font_family)
    regular <- detected$regular
    bold <- detected$bold %||% bold
    italic <- detected$italic %||% italic
    bold_italic <- detected$bold_italic %||% bold_italic
  }

  if (!file.exists(regular)) {
    cli::cli_abort("Font file {.file {regular}} does not exist.")
  }

  styles <- list(
    embedRegular = regular,
    embedBold = bold,
    embedItalic = italic,
    embedBoldItalic = bold_italic
  )
  styles <- Filter(Negate(is.null), styles)

  for (path in styles) {
    if (!file.exists(path)) {
      cli::cli_abort("Font file {.file {path}} does not exist.")
    }
  }

  package_dir <- x$package_dir
  fonts_dir <- file.path(package_dir, "word", "fonts")
  if (!dir.exists(fonts_dir)) {
    dir.create(fonts_dir, recursive = TRUE)
  }

  # determine next font file number
  existing <- list.files(fonts_dir, pattern = "^font\\d+\\.odttf$")
  if (length(existing) > 0) {
    nums <- as.integer(gsub("\\D", "", existing))
    next_num <- max(nums) + 1L
  } else {
    next_num <- 1L
  }

  # update fontTable.xml.rels
  rels_file <- file.path(package_dir, "word", "_rels", "fontTable.xml.rels")
  if (file.exists(rels_file)) {
    rels_doc <- read_xml(rels_file)
    existing_rels <- xml_find_all(
      rels_doc,
      "d1:Relationship",
      ns = xml_ns(rels_doc)
    )
    existing_ids <- xml_attr(existing_rels, "Id")
    if (length(existing_ids) > 0) {
      next_rid <- max(as.integer(gsub("\\D", "", existing_ids))) + 1L
    } else {
      next_rid <- 1L
    }
  } else {
    rels_dir <- file.path(package_dir, "word", "_rels")
    if (!dir.exists(rels_dir)) {
      dir.create(rels_dir, recursive = TRUE)
    }
    rels_doc <- read_xml(paste0(
      '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
      '<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships"/>'
    ))
    next_rid <- 1L
  }

  # update fontTable.xml
  font_table_file <- file.path(package_dir, "word", "fontTable.xml")
  font_table_doc <- read_xml(font_table_file)
  ns_w <- "http://schemas.openxmlformats.org/wordprocessingml/2006/main"
  ns_r <- "http://schemas.openxmlformats.org/officeDocument/2006/relationships"

  font_node <- xml_find_first(
    font_table_doc,
    sprintf("w:font[@w:name='%s']", font_family),
    ns = c(w = ns_w)
  )
  if (inherits(font_node, "xml_missing")) {
    font_xml <- sprintf(
      '<w:font xmlns:w="%s" xmlns:r="%s" w:name="%s"/>',
      ns_w,
      ns_r,
      font_family
    )
    xml_add_child(font_table_doc, read_xml(font_xml))
    font_node <- xml_find_first(
      font_table_doc,
      sprintf("w:font[@w:name='%s']", font_family),
      ns = c(w = ns_w)
    )
  }

  # process each style
  rel_type <- "http://schemas.openxmlformats.org/officeDocument/2006/relationships/font"
  ns_pkg <- "http://schemas.openxmlformats.org/package/2006/relationships"

  for (embed_tag in names(styles)) {
    font_path <- styles[[embed_tag]]
    guid <- format_guid(uuid_generate())
    key <- guid_to_xor_key(guid)

    # obfuscate
    font_bytes <- readBin(font_path, "raw", file.info(font_path)$size)
    font_bytes[1:16] <- xor(font_bytes[1:16], key)
    font_bytes[17:32] <- xor(font_bytes[17:32], key)

    # write odttf
    odttf_name <- sprintf("font%d.odttf", next_num)
    writeBin(font_bytes, file.path(fonts_dir, odttf_name))
    next_num <- next_num + 1L

    # add relationship
    rid <- sprintf("rId%d", next_rid)
    rel_xml <- sprintf(
      '<Relationship xmlns="%s" Id="%s" Type="%s" Target="fonts/%s"/>',
      ns_pkg,
      rid,
      rel_type,
      odttf_name
    )
    xml_add_child(rels_doc, read_xml(rel_xml))
    next_rid <- next_rid + 1L

    # remove existing embed node of same type if present
    old_embed <- xml_child(font_node, paste0("w:", embed_tag))
    if (!inherits(old_embed, "xml_missing")) {
      xml_remove(old_embed)
    }

    # add embed node
    embed_xml <- sprintf(
      '<w:%s xmlns:w="%s" xmlns:r="%s" r:id="%s" w:fontKey="%s"/>',
      embed_tag,
      ns_w,
      ns_r,
      rid,
      guid
    )
    xml_add_child(font_node, read_xml(embed_xml))
  }

  write_xml(rels_doc, rels_file)
  write_xml(font_table_doc, font_table_file)

  # content type for odttf
  x$content_type$add_ext(
    "odttf",
    "application/vnd.openxmlformats-officedocument.obfuscatedFont"
  )

  # ensure embedTrueTypeFonts in settings
  x$settings$embed_true_type_fonts <- TRUE

  x
}

# Detect font files from gdtools::sys_fonts() for a given family
detect_font_files <- function(font_family) {
  if (!requireNamespace("gdtools", quietly = TRUE)) {
    cli::cli_abort(c(
      "Package {.pkg gdtools} is required to auto-detect font files.",
      "i" = "Install it with {.code install.packages(\"gdtools\")}",
      "i" = "Or provide explicit file paths via {.arg regular}, {.arg bold}, etc."
    ))
  }

  sysfonts <- gdtools::sys_fonts()
  matches <- sysfonts[sysfonts$family == font_family, ]

  if (nrow(matches) == 0L) {
    cli::cli_abort(c(
      "Font family {.val {font_family}} not found in system fonts.",
      "i" = "Register it first with {.code gdtools::register_gfont()} or {.code gdtools::register_liberationsans()}.",
      "i" = "Check available families with {.code unique(gdtools::sys_fonts()$family)}."
    ))
  }

  style_map <- list(
    regular = "Regular",
    bold = "Bold",
    italic = "Italic",
    bold_italic = "Bold Italic"
  )

  result <- list()
  for (nm in names(style_map)) {
    row <- matches[matches$style == style_map[[nm]], ]
    if (nrow(row) > 0L) {
      result[[nm]] <- row$path[1L]
    }
  }

  if (is.null(result$regular)) {
    cli::cli_abort(c(
      "No {.val Regular} style found for font family {.val {font_family}}.",
      "i" = "Provide the path explicitly via {.arg regular}."
    ))
  }

  # print equivalent explicit call
  arg_names <- c(
    regular = "regular",
    bold = "bold",
    italic = "italic",
    bold_italic = "bold_italic"
  )
  present <- Filter(Negate(is.null), result)
  code_args <- vapply(
    names(present),
    function(nm) {
      sprintf("  %s = \"%s\"", arg_names[[nm]], present[[nm]])
    },
    character(1L)
  )
  code <- paste0(
    "docx_embed_font(\n  x,\n",
    sprintf("  font_family = \"%s\",\n", font_family),
    paste(code_args, collapse = ",\n"),
    "\n)"
  )
  cli::cli_inform(
    "Embedding {.val {font_family}} with detected font files:"
  )
  message(code)

  result
}

# Format a UUID string as a Word fontKey GUID
format_guid <- function(uuid_str) {
  toupper(paste0("{", uuid_str, "}"))
}

# Convert a GUID string like {XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX}
# to the 16-byte XOR key used for OOXML font obfuscation
guid_to_xor_key <- function(guid_str) {
  hex <- gsub("[{}-]", "", guid_str)
  bytes <- as.raw(strtoi(substring(hex, seq(1, 31, 2), seq(2, 32, 2)), 16L))
  # Windows GUID memory layout: groups 1-3 little-endian, groups 4-5 big-endian
  guid_mem <- c(rev(bytes[1:4]), rev(bytes[5:6]), rev(bytes[7:8]), bytes[9:16])
  rev(as.raw(guid_mem))
}

#' @export
#' @title transform an xml string with images references
#' @description This function is useless now, as the processing of images
#' is automated when using [print.rdocx()].
#' @param x an rdocx object
#' @param str wml string
#' @family functions for officer extensions
#' @keywords internal
wml_link_images <- function(x, str) {
  str
}

# by capturing the path, we are making 'unique' new image names.
#' @importFrom openssl sha1
fake_newname <- function(filename) {
  which_files <- grepl("\\.[a-zA-Z0-9]+$", filename)
  file_type <- gsub("(.*)(\\.[a-zA-Z0-9]+)$", "\\2", filename[which_files])
  dest_basename <- sapply(filename[which_files], function(z) {
    as.character(sha1(file(z)))
  })
  dest_basename <- paste0(dest_basename, file_type)
  x <- filename
  x[which_files] <- dest_basename
  x
}

process_images <- function(
  doc_obj,
  relationships,
  package_dir,
  media_dir = "word/media",
  media_rel_dir = "media"
) {
  hl_nodes <- xml_find_all(
    doc_obj$get(),
    "//a:blip[@r:embed]|//asvg:svgBlip[@r:embed]",
    ns = c(
      "a" = "http://schemas.openxmlformats.org/drawingml/2006/main",
      "asvg" = "http://schemas.microsoft.com/office/drawing/2016/SVG/main",
      "r" = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
    )
  )
  which_to_add <- hl_nodes[!grepl("^rId[0-9]+$", xml_attr(hl_nodes, "embed"))]
  hl_ref <- unique(xml_attr(which_to_add, "embed"))
  for (i in seq_along(hl_ref)) {
    dest_basename <- fake_newname(hl_ref[i])
    img_path <- file.path(package_dir, media_dir)
    if (!file.exists(file.path(img_path, dest_basename))) {
      dir.create(img_path, recursive = TRUE, showWarnings = FALSE)
      file.copy(from = hl_ref[i], to = file.path(img_path, dest_basename))
    }
    if (
      !file.path(media_rel_dir, dest_basename) %in%
        relationships$get_data()$target
    ) {
      rid <- sprintf("rId%.0f", relationships$get_next_id())
      relationships$add(
        id = rid,
        type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image",
        target = file.path(media_rel_dir, dest_basename)
      )
    } else {
      reldf <- relationships$get_data()
      rid <- reldf$id[basename(reldf$target) %in% dest_basename]
    }
    which_match_id <- grepl(
      dest_basename,
      fake_newname(xml_attr(which_to_add, "embed")),
      fixed = TRUE
    )
    xml_attr(which_to_add[which_match_id], "r:embed") <- rep(
      rid,
      sum(which_match_id)
    )
  }
}

process_docx_poured <- function(
  doc_obj,
  relationships,
  content_type,
  package_dir,
  media_dir = "word"
) {
  hl_nodes <- xml_find_all(
    doc_obj$get(),
    "//w:altChunk[@r:id]",
    ns = c(
      "w" = "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
      "r" = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
    )
  )

  which_to_add <- hl_nodes[!grepl("^rId[0-9]+$", xml_attr(hl_nodes, "id"))]
  hl_ref <- unique(xml_attr(which_to_add, "id"))
  for (i in seq_along(hl_ref)) {
    rid <- sprintf("rId%.0f", relationships$get_next_id())

    file.copy(
      from = hl_ref[i],
      to = file.path(package_dir, media_dir, basename(hl_ref[i]))
    )

    relationships$add(
      id = rid,
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/aFChunk",
      target = basename(hl_ref[i])
    )
    content_type$add_override(
      setNames(
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml",
        paste0("/", media_dir, "/", basename(hl_ref[i]))
      )
    )

    which_match_id <- grepl(
      hl_ref[i],
      xml_attr(which_to_add, "id"),
      fixed = TRUE
    )
    xml_attr(which_to_add[which_match_id], "r:id") <- rep(
      rid,
      sum(which_match_id)
    )
  }
}

#' @importFrom xml2 xml_remove as_xml_document xml_parent xml_child
process_links <- function(doc_obj, type = "wml") {
  rel <- doc_obj$relationship()
  if ("wml" %in% type) {
    hl_nodes <- xml_find_all(doc_obj$get(), "//w:hyperlink[@r:id]")
  } else {
    hl_nodes <- xml_find_all(doc_obj$get(), "//a:hlinkClick[@r:id]")
  }
  which_to_add <- hl_nodes[!grepl("^rId[0-9]+$", xml_attr(hl_nodes, "id"))]
  hl_ref <- unique(xml_attr(which_to_add, "id"))
  for (i in seq_along(hl_ref)) {
    rid <- sprintf("rId%.0f", rel$get_next_id())

    rel$add(
      id = rid,
      type = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink",
      target = officer_url_decode(hl_ref[i]),
      target_mode = "External"
    )

    which_match_id <- grepl(
      hl_ref[i],
      xml_attr(which_to_add, "id"),
      fixed = TRUE
    )
    xml_attr(which_to_add[which_match_id], "r:id") <- rep(
      rid,
      sum(which_match_id)
    )
  }
}

process_list_markers <- function(xml_str, package_dir) {
  marker_pattern <- "officer-list-(bullet|decimal)-[a-f0-9-]+"
  has_markers <- grepl(marker_pattern, xml_str)
  if (!any(has_markers)) {
    return(xml_str)
  }

  all_markers <- regmatches(
    xml_str[has_markers],
    gregexpr(marker_pattern, xml_str[has_markers])
  )
  unique_markers <- unique(unlist(all_markers))

  numbering_file <- file.path(package_dir, "word", "numbering.xml")
  if (!file.exists(numbering_file)) {
    cli::cli_abort("numbering.xml not found in {.file {package_dir}}")
  }
  numbering_doc <- read_xml(numbering_file)
  ns_w <- "http://schemas.openxmlformats.org/wordprocessingml/2006/main"

  abstract_nums <- xml_find_all(
    numbering_doc,
    "w:abstractNum",
    ns = c(w = ns_w)
  )
  nums <- xml_find_all(numbering_doc, "w:num", ns = c(w = ns_w))
  next_abstract_id <- if (length(abstract_nums) > 0) {
    max(as.integer(xml_attr(abstract_nums, "abstractNumId"))) + 1L
  } else {
    0L
  }
  next_num_id <- if (length(nums) > 0) {
    max(as.integer(xml_attr(nums, "numId"))) + 1L
  } else {
    1L
  }

  for (marker in unique_markers) {
    list_type <- sub("officer-list-(bullet|decimal)-.*", "\\1", marker)
    abstract_xml <- build_abstract_num_xml(
      next_abstract_id,
      list_type,
      ns_w
    )
    num_xml <- sprintf(
      paste0(
        '<w:num xmlns:w="%s" w:numId="%d">',
        '<w:abstractNumId w:val="%d"/></w:num>'
      ),
      ns_w,
      next_num_id,
      next_abstract_id
    )

    xml_add_child(numbering_doc, read_xml(abstract_xml))
    xml_add_child(numbering_doc, read_xml(num_xml))

    xml_str <- gsub(
      marker,
      as.character(next_num_id),
      xml_str,
      fixed = TRUE
    )

    next_abstract_id <- next_abstract_id + 1L
    next_num_id <- next_num_id + 1L
  }

  write_xml(numbering_doc, numbering_file)
  xml_str
}

build_abstract_num_xml <- function(abstract_num_id, list_type, ns_w) {
  levels <- vapply(
    0:8,
    function(ilvl) {
      indent_left <- 360L * (ilvl + 1L)
      if (list_type == "bullet") {
        bullets <- c(
          "\u2022",
          "\u25E6",
          "\u2013",
          "\u2022",
          "\u25E6",
          "\u2013",
          "\u2022",
          "\u25E6",
          "\u2013"
        )
        sprintf(
          paste0(
            '<w:lvl w:ilvl="%d">',
            '<w:numFmt w:val="bullet"/>',
            '<w:lvlText w:val="%s"/>',
            '<w:lvlJc w:val="left"/>',
            '<w:pPr><w:ind w:left="%d" w:hanging="360"/></w:pPr>',
            '</w:lvl>'
          ),
          ilvl,
          bullets[ilvl + 1L],
          indent_left
        )
      } else {
        sprintf(
          paste0(
            '<w:lvl w:ilvl="%d">',
            '<w:start w:val="1"/>',
            '<w:numFmt w:val="decimal"/>',
            '<w:lvlText w:val="%%%d."/>',
            '<w:lvlJc w:val="left"/>',
            '<w:pPr><w:ind w:left="%d" w:hanging="360"/></w:pPr>',
            '</w:lvl>'
          ),
          ilvl,
          ilvl + 1L,
          indent_left
        )
      }
    },
    character(1L)
  )

  sprintf(
    paste0(
      '<w:abstractNum xmlns:w="%s" w:abstractNumId="%d">',
      '<w:multiLevelType w:val="multilevel"/>%s',
      '</w:abstractNum>'
    ),
    ns_w,
    abstract_num_id,
    paste0(levels, collapse = "")
  )
}

update_hf_list <- function(part_list = list(), type = "header", package_dir) {
  files <- list.files(
    path = file.path(package_dir, "word"),
    pattern = sprintf("^%s[0-9]*.xml$", type)
  )
  files <- files[!basename(files) %in% names(part_list)]
  if (type %in% "header") {
    cursor <- "/w:hdr/*[1]"
    body_xpath <- "/w:hdr"
  } else {
    cursor <- "/w:ftr/*[1]"
    body_xpath <- "/w:ftr"
  }

  new_list <- lapply(files, function(x) {
    docx_part$new(
      path = package_dir,
      main_file = x,
      cursor = cursor,
      body_xpath = body_xpath
    )
  })
  names(new_list) <- basename(files)
  append(part_list, new_list)
}

#' @export
#' @title Remove unused media from a document
#' @description The function will scan the media
#' directory and delete images that are not used
#' anymore. This function is to be used when images
#' have been replaced many times.
#' @param x `rdocx` or `rpptx` object
#' @param warn_user TRUE to make sure users are warned when
#' using this function that will be un-exported in a next
#' version
#' @keywords internal
sanitize_images <- function(x, warn_user = TRUE) {
  if (warn_user) {
    cli::cli_warn(
      c(
        "!" = "Function {.fn sanitize_images} will be removed soon. You should remove calls to {.fn sanitize_images}.",
        "i" = "Instead the function is run before saving the document to a file."
      )
    )
  }

  if (inherits(x, "rdocx")) {
    image_files <- c()
    all_docs <- append(x$headers, x$footers)
    all_docs[[length(all_docs) + 1]] <- x$doc_obj
    all_docs[[length(all_docs) + 1]] <- x$footnotes

    for (doc_part in all_docs) {
      suppressWarnings({
        blip_nodes <- xml_find_all(
          doc_part$get(),
          "//a:blip[contains(@r:embed, 'rId')]|//asvg:svgBlip[contains(@r:embed, 'rId')]",
          ns = c(
            "a" = "http://schemas.openxmlformats.org/drawingml/2006/main",
            "asvg" = "http://schemas.microsoft.com/office/drawing/2016/SVG/main",
            "r" = "http://schemas.openxmlformats.org/officeDocument/2006/relationships"
          )
        )
      })

      embed_list <- xml_attr(blip_nodes, "embed")
      embed_data <- filter(
        .data = doc_part$rel_df(),
        basename(.data$type) %in% "image",
        .data$id %in% embed_list
      )
      embed_data <- embed_data$target
      image_files[[length(image_files) + 1]] <- embed_data
    }

    image_files <- do.call(c, image_files)
    image_files <- unique(image_files)

    base_doc <- file.path(x$package_dir, "word")
    existing_img <- list.files(
      file.path(base_doc, "media"),
      pattern = "\\.(png|jpg|jpeg|eps|emf|svg)$",
      ignore.case = TRUE,
      recursive = TRUE,
      full.names = TRUE
    )

    existing_img <- gsub(paste0(base_doc, "/"), "", existing_img, fixed = TRUE)
    unlink(
      file.path(base_doc, setdiff(existing_img, image_files)),
      force = TRUE
    )

    # these iterations will remove removed images from relationships
    for (doc_part in all_docs) {
      rel <- doc_part$relationship()
      rel_data <- rel$get_data()
      rel_data <- rel_data[basename(rel_data$type) %in% "image", ]
      rel_data <- rel_data[!file.exists(file.path(base_doc, rel_data$target)), ]
      if (nrow(rel_data) > 0) {
        rel$remove(rel_data$target)
        doc_part$save()
      }
    }
  } else if (inherits(x, "rpptx")) {
    rel_files <- list.files(
      x$package_dir,
      pattern = "\\.xml.rels$",
      recursive = TRUE,
      full.names = TRUE
    )

    image_files <- lapply(rel_files, function(x) {
      zz <- read_xml(x)
      rels <- xml_children(zz)
      rels <- rels[
        xml_attr(rels, "Type") %in%
          "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"
      ]
      xml_attr(rels, "Target")
    })
    image_files <- unique(unlist(image_files))
    base_doc <- file.path(x$package_dir, "ppt")
    existing_img <- list.files(
      file.path(base_doc, "media"),
      pattern = "\\.(png|jpg|jpeg|eps|emf|svg)$",
      ignore.case = TRUE,
      recursive = TRUE,
      full.names = TRUE
    )
    existing_img <- gsub(
      paste0(base_doc, "/"),
      "../",
      existing_img,
      fixed = TRUE
    )
    unlink(
      file.path(base_doc, setdiff(existing_img, image_files)),
      force = TRUE
    )
  }
  x
}

#' @importFrom R6 R6Class
relationship <- R6Class(
  "relationship",
  public = list(
    initialize = function(
      id = character(0),
      type = character(0),
      target = character(0)
    ) {
      private$id <- id
      private$type <- type
      private$target <- target
      private$target_mode <- rep(NA_character_, length(target))
      private$ext_src <- character(length(id))
    },
    feed_from_xml = function(path) {
      doc <- read_xml(x = path)
      children <- xml_children(doc)
      ns <- xml_ns(doc)

      private$id <- c(
        private$id,
        vapply(children, xml_attr, NA_character_, attr = "Id", ns)
      )
      private$type <- c(
        private$type,
        vapply(children, xml_attr, NA_character_, attr = "Type", ns)
      )
      private$target <- c(
        private$target,
        vapply(children, xml_attr, NA_character_, attr = "Target", ns)
      )
      private$target_mode <- c(
        private$target_mode,
        vapply(children, xml_attr, NA_character_, attr = "TargetMode", ns)
      )
      private$ext_src <- c(private$ext_src, character(length(children)))
      self
    },
    write = function(path) {
      if (length(private$id)) {
        str <- paste0(
          "<Relationship Id=\"",
          private$id,
          "\" Type=\"",
          private$type,
          "\" Target=\"",
          htmlEscapeCopy(private$target),
          ifelse(
            is.na(private$target_mode),
            "",
            "\" TargetMode=\"External"
          ),
          "\"/>"
        )
      } else {
        str <- character(length = 0)
      }

      str <- c(
        "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>",
        "\n<Relationships  xmlns=\"http://schemas.openxmlformats.org/package/2006/relationships\">",
        str,
        "</Relationships>"
      )
      dir.create(dirname(path), showWarnings = FALSE, recursive = FALSE)
      writeLines(str, con = path, useBytes = TRUE)
      self
    },
    get_next_id = function() {
      max(c(0, private$get_int_id()), na.rm = TRUE) + 1
    },
    get_data = function() {
      data <- data.frame(
        id = character(0),
        int_id = integer(0),
        type = character(0),
        target = character(0),
        target_mode = character(0),
        ext_src = character(0),
        stringsAsFactors = FALSE
      )
      if (length(private$id)) {
        chr_stripped_id <- gsub("rId([0-9]+)", "\\1", private$id)
        test_id_as_num <- grepl("^rId[0-9]+$", private$id)
        if (!all(test_id_as_num)) {
          wrong_targets <- paste0(
            "invalid id(s) for {shQuote(private$target[!test_id_as_num])}"
          )
          names(wrong_targets) <- rep("*", length(wrong_targets))
          cli::cli_warn(
            c(
              "x" = "relationship file contains invalid id(s).",
              wrong_targets
            )
          )
        }
        data <- data.frame(
          id = private$id,
          int_id = as.integer(chr_stripped_id),
          type = private$type,
          target = private$target,
          target_mode = private$target_mode,
          ext_src = private$ext_src,
          stringsAsFactors = FALSE
        )

        data[order(data$id), ]
      }
      data
    },
    get_images_path = function() {
      is_img <- basename(private$type) %in% "image"
      targets <- private$target[is_img]
      names(targets) <- private$id[is_img]
      targets
    },
    add_img = function(src, root_target) {
      src <- setdiff(src, private$ext_src)
      if (!length(src)) {
        return(self)
      }

      if (any(grepl(" ", basename(src)))) {
        stop(
          paste(src, collapse = ", "),
          ": images with blanks in their basenames are not supported, please rename the file(s).",
          call. = FALSE
        )
      }

      last_id <- max(c(0, private$get_int_id()), na.rm = TRUE)

      id <- paste0("rId", seq_along(src) + last_id)
      type <- rep(
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/image",
        length(src)
      )
      target <- file.path(root_target, basename(src))

      private$id <- c(private$id, id)
      private$type <- c(private$type, type)
      private$target <- c(private$target, target)
      private$target_mode <- c(private$target_mode, rep(NA, length(id)))
      private$ext_src <- c(private$ext_src, src)

      self
    },
    add_drawing = function(src, root_target) {
      src <- setdiff(src, private$ext_src)
      if (!length(src)) {
        return(self)
      }
      last_id <- max(c(0, private$get_int_id()), na.rm = TRUE)

      id <- paste0("rId", seq_along(src) + last_id)
      type <- rep(
        "http://schemas.openxmlformats.org/officeDocument/2006/relationships/drawing",
        length(src)
      )
      target <- file.path(root_target, basename(src))

      private$id <- unlist(c(private$id, id))
      private$type <- unlist(c(private$type, type))
      private$target <- unlist(c(private$target, target))
      private$target_mode <- unlist(c(private$target_mode, rep(NA, length(id))))
      private$ext_src <- unlist(c(private$ext_src, rep(NA, length(id))))

      self
    },
    add = function(id, type, target, target_mode = NA) {
      Encoding(target) <- "UTF-8"
      if (!target %in% private$target) {
        private$id <- c(private$id, id)
        private$type <- c(private$type, type)
        private$target <- c(private$target, target)
        private$target_mode <- c(private$target_mode, target_mode)
        private$ext_src <- c(private$ext_src, "")
      }
      self
    },
    remove = function(target) {
      id <- which(basename(private$target) %in% basename(target))
      private$id <- private$id[-id]
      private$type <- private$type[-id]
      private$target <- private$target[-id]
      private$target_mode <- private$target_mode[-id]
      private$ext_src <- private$ext_src[-id]
      self
    },
    show = function() {
      print(self$get_data())
    }
  ),
  private = list(
    id = NA,
    type = NA,
    target = NA,
    target_mode = NA,
    ext_src = NA,
    get_int_id = function() {
      if (length(private$id) > 0) {
        as.integer(gsub("rId([0-9]+)", "\\1", private$id))
      } else {
        integer(0)
      }
    }
  )
)

#' @export
#' @title Number of slides
#' @description Function `length` will return the number of slides.
#' @param x an rpptx object
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres, "Title and Content")
#' my_pres <- add_slide(my_pres, "Title and Content")
#' length(my_pres)
#' @family functions for reading presentation information
length.rpptx <- function(x) {
  x$slide$length()
}


#' @export
#' @title Slides width and height
#' @description Get the width and height of slides in inches as
#' a named vector.
#' @inheritParams length.rpptx
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres,
#'   layout = "Two Content", master = "Office Theme")
#' slide_size(my_pres)
#' @family functions for reading presentation information
slide_size <- function(x) {
  pres <- x$presentation$get()
  dimensions <- xml_attrs(xml_find_first(pres, "p:sldSz"))
  dimensions <- as.list(as.integer(dimensions[c("cx", "cy")]) / 914400)
  names(dimensions) <- c("width", "height")
  dimensions
}


#' @export
#' @title Presentation layouts summary
#' @description Get information about slide layouts and
#' master layouts into a data.frame. This function returns
#' a data.frame containing all layout and master names.
#' @inheritParams length.rpptx
#' @examples
#' my_pres <- read_pptx()
#' layout_summary ( x = my_pres )
#' @family functions for reading presentation information
layout_summary <- function(x) {
  data <- x$slideLayouts$get_metadata()
  data.frame(
    layout = data$name,
    master = data$master_name,
    stringsAsFactors = FALSE
  )
}


#' @export
#' @title Slide layout properties
#' @description Detailed information about the placeholders on the slide layouts (label, position, etc.).
#' See *Value* section below for more info.
#' @param x an `rpptx` object
#' @param layout slide layout name. If `NULL`, returns all layouts.
#' @param master master layout name where `layout` is located. If `NULL`, returns all masters.
#' @returns Returns a data frame with one row per placeholder and the following columns:
#' * `master_name`: Name of master (a `.pptx` file may have more than one)
#' * `name`: Name of layout
#' * `type`: Placeholder type
#' * `type_idx`: Running index for phs of the same type. Ordering by ph position
#'    (top -> bottom, left -> right)
#' * `id`: A unique placeholder id (assigned by PowerPoint automatically, starts at 2, potentially non-consecutive)
#' * `ph_label`: Placeholder label (can be set by the user in PowerPoint)
#' * `ph`: Placholder XML fragment (usually not needed)
#' * `offx`,`offy`: placeholder's distance from left and top edge (in inch)
#' * `cx`,`cy`: width and height of placeholder (in inch)
#' * `rotation`: rotation in degrees
#' * `fld_id` is generally stored as a hexadecimal or GUID value
#' * `fld_type`: a unique identifier for a particular field
#'
#' @example inst/examples/example_layout_properties.R
#' @family functions for reading presentation information
layout_properties <- function(x, layout = NULL, master = NULL) {
  data <- x$slideLayouts$get_xfrm_data()
  if (!is.null(layout) && !is.null(master)) {
    data <- data[data$name == layout & data$master_name %in% master, ]
  } else if (is.null(layout) && !is.null(master)) {
    data <- data[data$master_name %in% master, ]
  } else if (!is.null(layout) && is.null(master)) {
    data <- data[data$name == layout, ]
  }
  data <- data[, c(
    "master_name",
    "name",
    "type",
    "type_idx",
    "id",
    "ph_label",
    "ph",
    "offx",
    "offy",
    "cx",
    "cy",
    "rotation",
    "fld_id",
    "fld_type"
  )]
  data[["offx"]] <- data[["offx"]] / 914400
  data[["offy"]] <- data[["offy"]] / 914400
  data[["cx"]] <- data[["cx"]] / 914400
  data[["cy"]] <- data[["cy"]] / 914400
  data[["rotation"]] <- data[["rotation"]] / 60000

  rownames(data) <- NULL
  data
}


#' @export
#' @title Slide layout properties plot
#' @description Plot slide layout properties into corresponding placeholders.
#'  This can be useful to help visualize placeholders locations and identifiers.
#'  *All* information in the plot stems from the [layout_properties()] output.
#'  See *Details* section for more info.
#' @details
#' The plot contains all relevant information to reference a placeholder via the `ph_location_*`
#' function family:
#'
#' * `label`: ph label (red, center) to be used in [ph_location_label()].
#'    _NB_: The label can be assigned by the user in PowerPoint.
#' * `type[idx]`: ph type + type index in brackets (blue, upper left) to be used in [ph_location_type()].
#'    _NB_: The index is consecutive and is sorted by ph position (top -> bottom, left -> right).
#' * `id`: ph id (green, upper right) to be used in `ph_location_id()` (forthcoming).
#'    _NB_: The id is set by PowerPoint automatically and lack a meaningful order.
#'
#' @param x an `rpptx` object
#' @param layout slide layout name or numeric index (row index from [layout_summary()]. If `NULL` (default), it plots
#'   the current slide's layout or the default layout (if set and there are not slides yet).
#' @param master master layout name where `layout` is located. Can be omitted if layout is unambiguous.
#' @param slide_idx Numeric slide index (default `NULL`) to specify which slide’s layout should be plotted.
#' @param title if `TRUE` (default), adds a title with the layout and master name (latter in square brackets) at the top.
#' @param labels if `TRUE` (default), adds placeholder labels (centered in *red*).
#' @param type if `TRUE` (default), adds the placeholder type and its index (in square brackets)
#'   in the upper left corner (in *blue*).
#' @param id if `TRUE` (default), adds the placeholder's unique `id` (see column `id` from
#'  [layout_properties()]) in the upper right corner (in *green*).
#' @param cex List or vector to specify font size for `labels`, `type`, and `id`. Default is
#'   `c(labels = 0.5, type = 0.5, id = 0.5)`. See [graphics::text()] for details on how `cex` works.
#'   Matching by position and partial name matching is supported. A single numeric value will apply to all
#'   three parameters.
#' @param legend Add a legend to the plot (default `FALSE`).
#' @importFrom graphics plot rect text box
#' @family functions for reading presentation information
#' @example inst/examples/example_plot_layout_properties.R
#'
plot_layout_properties <- function(
  x,
  layout = NULL,
  master = NULL,
  slide_idx = NULL,
  labels = TRUE,
  title = TRUE,
  type = TRUE,
  id = TRUE,
  cex = c(labels = 0.5, type = 0.5, id = 0.5),
  legend = FALSE
) {
  stop_if_not_rpptx(x, "x")
  loffset <- ifelse(legend, 1, 0) # make space for legend at top
  old_par <- par(mar = c(2, 2, 1.5 + loffset, 0))
  on.exit(par(old_par))

  .cex <- update_named_defaults(
    cex,
    default = list(labels = 0.5, type = 0.5, id = 0.5),
    default_if_null = TRUE,
    argname = "cex"
  )
  if (.cex$labels <= 0) {
    labels <- FALSE
  }
  if (.cex$type <= 0) {
    type <- FALSE
  }
  if (.cex$id <= 0) {
    id <- FALSE
  }

  if (is.null(layout) && is.null(master) && length(x) == 0) {
    # fail or use default layout if set
    if (!has_layout_default(x)) {
      cli::cli_abort(
        c(
          "No {.arg layout} selected and no slides in presentation.",
          "x" = "Pass a layout name or index (see {.fn layout_summary})"
        )
      )
    }
    .ld <- get_layout_default(x)
    la <- get_layout(x, layout = .ld$layout, master = .ld$master)
    cli::cli_inform(c("i" = "Showing default layout: {.val {la$layout_name}}"))
  } else if (
    is.null(layout) && is.null(master) && length(x) > 0 && is.null(slide_idx)
  ) {
    # use current slides layout as default
    la <- get_layout_for_current_slide(x)
    cli::cli_inform(c(
      "i" = "Showing current slide's layout: {.val {la$layout_name}}"
    ))
  } else if (
    is.null(layout) && is.null(master) && length(x) > 0 && !is.null(slide_idx)
  ) {
    # use layout for selected slide
    la <- get_slide_layout(x, slide_idx = slide_idx)
    cli::cli_inform(c(
      "i" = "Showing layout of slide {.val {slide_idx}}: {.val {la$layout_name}}"
    ))
  } else {
    la <- get_layout(x, layout, master, layout_by_id = TRUE)
  }

  dat <- layout_properties(x, layout = la$layout_name, master = la$master_name)

  s <- slide_size(x)
  h <- s$height
  w <- s$width
  offx <- offy <- cx <- cy <- NULL # avoid R CMD CHECK problem
  list2env(dat[, c("offx", "offy", "cx", "cy")], environment()) # make available inside functions

  plot(
    x = c(0, w),
    y = -c(0, h),
    asp = 1,
    type = "n",
    axes = FALSE,
    xlab = NA,
    ylab = NA
  )
  rect(xleft = 0, xright = w, ybottom = 0, ytop = -h, border = "darkgrey")
  rect(xleft = offx, xright = offx + cx, ybottom = -offy, ytop = -(offy + cy))
  mtext("y [inch]", side = 2, line = 0, cex = 0.9, col = "darkgrey")
  mtext("x [inch]", side = 1, line = 0, cex = 0.9, col = "darkgrey")

  if (title) {
    title(
      main = paste0(la$layout_name, " [", la$master_name, "]"),
      line = 0 + loffset
    )
  }
  if (labels) {
    # centered
    text(
      x = offx + cx / 2,
      y = -(offy + cy / 2),
      labels = dat$ph_label,
      cex = .cex$labels,
      col = "red",
      adj = c(0.5, 1)
    ) # adj-vert: avoid interference with type/id in small phs
  }
  if (type) {
    # upper left corner
    .type_info <- paste0(dat$type, " [", dat$type_idx, "]") # type + index in brackets
    text(
      x = offx,
      y = -offy,
      labels = .type_info,
      cex = .cex$type,
      col = "blue",
      adj = c(-0.1, 1.2)
    )
  }
  if (id) {
    # upper right corner
    text(
      x = offx + cx,
      y = -offy,
      labels = dat$id,
      cex = .cex$id,
      col = "darkgreen",
      adj = c(1.3, 1.2)
    )
  }
  if (legend) {
    legend(
      x = w / 2,
      y = 0,
      x.intersp = 0.4,
      xjust = 0.5,
      yjust = 0,
      legend = c("type [type_idx]", "ph_label", "id"),
      fill = c("blue", "red", "darkgreen"),
      bty = "n",
      pt.cex = 1.2,
      cex = 0.7,
      text.width = NA,
      text.col = c("blue", "red", "darkgreen"),
      horiz = TRUE,
      xpd = TRUE
    )
  }
}


#' @export
#' @title Placeholder parameters annotation
#' @description generates a slide from each layout in the base document to
#' identify the placeholder indexes, types, names, master names and layout names.
#'
#' This is to be used when need to know what parameters should be used with
#' `ph_location*` calls. The parameters are printed in their corresponding shapes.
#'
#' Note that if there are duplicated `ph_label`, you should not use `ph_location_label()`.
#' Hint: You can dedupe labels using [layout_dedupe_ph_labels()].
#'
#' @param path path to the pptx file to use as base document or NULL to use the officer default
#' @param output_file filename to store the annotated powerpoint file or NULL to suppress generation
#' @return rpptx object of the annotated PowerPoint file
#' @examples
#' # To generate an anotation of the default base document with officer:
#' annotate_base(output_file = tempfile(fileext = ".pptx"))
#'
#' # To generate an annotation of the base document 'mydoc.pptx' and place the
#' # annotated output in 'mydoc_annotate.pptx'
#' # annotate_base(path = 'mydoc.pptx', output_file='mydoc_annotate.pptx')
#'
#' @family functions for reading presentation information
annotate_base <- function(path = NULL, output_file = 'annotated_layout.pptx') {
  ppt <- read_pptx(path = path)
  while (length(ppt) > 0) {
    ppt <- remove_slide(ppt, 1)
  }

  # Pulling out all of the layouts stored in the template
  lay_sum <- layout_summary(ppt)

  # Looping through each layout
  for (lidx in seq_len(nrow(lay_sum))) {
    # Pulling out the layout properties
    layout <- lay_sum[lidx, 1]
    master <- lay_sum[lidx, 2]
    lp <- layout_properties(x = ppt, layout = layout, master = master)

    # Adding a slide for the current layout
    ppt <- add_slide(x = ppt, layout = layout, master = master)
    size <- slide_size(ppt)
    fpar_ <- fpar(
      sprintf('layout ="%s", master = "%s"', layout, master),
      fp_t = fp_text(color = "orange", font.size = 20),
      fp_p = fp_par(text.align = "right", padding = 5)
    )
    ppt <- ph_with(
      x = ppt,
      value = fpar_,
      ph_label = "layout_ph",
      location = ph_location(
        left = 0,
        top = -0.5,
        width = size$width,
        height = 1,
        bg = "transparent",
        newlabel = "layout_ph"
      )
    )

    # Blank slides have nothing
    if (nrow(lp) > 0) {
      # Now we go through each placholder
      for (pidx in seq_len(nrow(lp))) {
        textstr <- paste(
          "type=",
          lp$type[pidx],
          ", index=",
          lp$id[pidx],
          ", ph_label=",
          lp$ph_label[pidx]
        )
        ppt <- ph_with(
          x = ppt,
          value = textstr,
          location = ph_location_label(
            type = lp$type[pidx],
            ph_label = lp$ph_label[pidx]
          )
        )
      }
    }
  }

  if (!is.null(output_file)) {
    print(ppt, target = output_file)
  }

  ppt
}


#' @export
#' @title Slide content in a data.frame
#' @description Get content and positions of current slide
#' into a data.frame. Data for any tables, images, or paragraphs are
#' imported into the resulting data.frame.
#' @note
#' The column `id` of the result is not to be used by users.
#' This is a technical string id whose value will be used by office
#' when the document will be rendered. This is not related to argument
#' `index` required by functions [ph_with()].
#' @inheritParams length.rpptx
#' @param index slide index
#' @examples
#' my_pres <- read_pptx()
#' my_pres <- add_slide(my_pres, "Title and Content")
#' my_pres <- ph_with(my_pres, format(Sys.Date()),
#'   location = ph_location_type(type="dt"))
#' my_pres <- add_slide(my_pres, "Title and Content")
#' my_pres <- ph_with(my_pres, iris[1:2,],
#'   location = ph_location_type(type="body"))
#' slide_summary(my_pres)
#' slide_summary(my_pres, index = 1)
#' @family functions for reading presentation information
slide_summary <- function(x, index = NULL) {
  l_ <- length(x)
  if (l_ < 1) {
    cli::cli_abort("Presentation contains no slide.")
  }

  if (is.null(index)) {
    index <- x$cursor
  }

  if (!between(index, 1, l_)) {
    cli::cli_abort(
      c(
        "Slide index {.val {index}} is out of range.",
        "x" = "Presentation has {cli::no(l_)} slide{?s}."
      )
    )
  }

  slide <- x$slide$get_slide(index)

  nodes <- xml_find_all(slide$get(), as_xpath_content_sel("p:cSld/p:spTree/"))
  data <- read_xfrm(nodes, file = "slide", name = "")
  data$text <- sapply(nodes, xml_text)
  data[["offx"]] <- data[["offx"]] / 914400
  data[["offy"]] <- data[["offy"]] / 914400
  data[["cx"]] <- data[["cx"]] / 914400
  data[["cy"]] <- data[["cy"]] / 914400

  data$name <- NULL
  data$file <- NULL
  data$ph <- NULL
  data
}


#' @export
#' @title Color scheme of a PowerPoint file
#' @description Get the color scheme of a
#' 'PowerPoint' master layout into a data.frame.
#' @inheritParams length.rpptx
#' @examples
#' x <- read_pptx()
#' color_scheme ( x = x )
#' @family functions for reading presentation information
color_scheme <- function(x) {
  x$masterLayouts$get_color_scheme()
}

unfold_row_pml <- function(node, row_id, preserve = FALSE) {
  children_ <- xml_children(node)
  cell_nodes <- children_[vapply(
    children_,
    function(x) xml_name(x) == "tc",
    logical(1)
  )]

  if (preserve) {
    txt <- vapply(
      cell_nodes,
      function(x) {
        paras <- xml2::xml_text(xml2::xml_find_all(x, ".//a:r"))
        paste0(paras, collapse = "\n")
      },
      character(1)
    )
  } else {
    txt <- vapply(cell_nodes, xml2::xml_text, character(1))
  }

  col_span <- vapply(
    cell_nodes,
    function(x) {
      as.integer(xml_attr(x, "gridSpan"))
    },
    integer(1)
  )
  h_merge <- vapply(
    cell_nodes,
    function(x) {
      as.integer(xml_attr(x, "hMerge"))
    },
    integer(1)
  ) %in%
    c(1)
  col_span[is.na(col_span)] <- 1L
  col_span[h_merge] <- 0L

  row_span <- lapply(cell_nodes, function(x) {
    row_span <- as.integer(xml_attr(x, "rowSpan"))
    v_merged <- as.integer(xml_attr(x, "vMerge"))
    data.frame(
      v_merged = v_merged,
      row_span = row_span,
      stringsAsFactors = FALSE
    )
  })
  row_span <- rbind_match_columns(row_span)
  row_span$row_merge <- !is.na(row_span$v_merged) | !is.na(row_span$row_span)
  row_span$first <- !is.na(row_span$row_span)
  row_span$row_span[!is.na(row_span$v_merged)] <- 0L
  row_span$row_span[!row_span$row_merge] <- 1L
  row_span <- row_span[, c("row_merge", "first", "row_span")]

  out <- data.frame(
    row_id = row_id,
    cell_id = seq_along(cell_nodes),
    text = txt,
    col_span = col_span,
    row_merge = row_span$row_merge,
    first = row_span$first,
    stringsAsFactors = FALSE
  )
  out
}

globalVariables(c(".", "src", "width", "height"))

pptxtable_as_tibble <- function(node, preserve = FALSE) {
  tbl <- xml_child(node, "/a:graphic/a:graphicData/a:tbl")
  xpath_ <- paste0(xml_path(tbl), "/a:tr")
  rows <- xml_find_all(node, xpath_)
  if (length(rows) < 1) {
    return(NULL)
  }
  row_details <- mapply(
    unfold_row_pml,
    rows,
    seq_along(rows),
    preserve = preserve,
    SIMPLIFY = FALSE
  )
  row_details <- rbind_match_columns(row_details)
  row_details <- set_row_span(row_details)
  row_details$text[
    row_details$col_span < 1 | row_details$row_span < 1
  ] <- NA_character_
  row_details
}


pptx_par_as_tibble <- function(node) {
  xpath_ <- paste0(
    xml_path(node),
    c("/p:txBody/a:p", "//*/p:txBody/a:p"), # standard and groupedshapes
    collapse = "|"
  )

  p_nodes <- xml_find_all(node, xpath_)
  data.frame(text = xml_text(p_nodes), stringsAsFactors = FALSE)
}


desc_pptx_img <- function(node, img_src) {
  blip <- xml_child(node, "/p:blipFill/a:blip")
  img_id <- xml_attr(blip, "embed")

  file_ <- img_src[img_id]
  stopifnot(is.character(file_), length(file_) == 1)
  data.frame(
    media_file = file.path("ppt/media/", basename(file_)),
    stringsAsFactors = FALSE
  )
}

#' @export
#' @title Extract media from a document object
#' @description Extract files from a `rpptx` object.
#' @param x an rpptx object
#' @param path media path, should be a relative path
#' @param target target file
#' @examples
#' example_pptx <- system.file(package = "officer",
#'   "doc_examples/example.pptx")
#' doc <- read_pptx(example_pptx)
#' content <- pptx_summary(doc)
#' image_row <- content[content$content_type %in% "image", ]
#' media_file <- image_row$media_file
#' png_file <- tempfile(fileext = ".png")
#' media_extract(doc, path = media_file, target = png_file)
media_extract <- function(x, path, target) {
  media <- file.path(x$package_dir, path)
  stopifnot(file.exists(media))
  file.copy(from = media, to = target)
}

#' @title PowerPoint content in a data.frame
#' @description Read content of a PowerPoint document and
#' return a dataset representing the document.
#' @param x an rpptx object
#' @inheritParams docx_summary
#' @examples
#' example_pptx <- system.file(package = "officer",
#'   "doc_examples/example.pptx")
#' doc <- read_pptx(example_pptx)
#' pptx_summary(doc)
#' pptx_summary(example_pptx)
#' @export
pptx_summary <- function(x, preserve = FALSE) {
  if (is.character(x)) {
    x = read_pptx(x)
  }
  list_content <- list()
  for (i in seq_len(length(x))) {
    slide <- x$slide$get_slide(i)
    str = as_xpath_content_sel("p:cSld/p:spTree/")
    nodes <- xml_find_all(slide$get(), str)

    if (length(nodes) < 1) {
      # when slide is empty
      list_content[[length(list_content) + 1]] <- data.frame(
        id = character(0),
        content_type = character(0),
        slide_id = integer(0),
        stringsAsFactors = FALSE
      )
      next
    }

    content <- mapply(
      function(node, slide_id) {
        is_table <- !inherits(
          xml_child(node, "/a:graphic/a:graphicData/a:tbl"),
          "xml_missing"
        )
        is_par <- !inherits(xml_child(node, "/p:txBody/a:p"), "xml_missing")
        is_img <- xml_name(node) == "pic"
        id <- xml_attr(xml_child(node, "/p:cNvPr"), "id")

        if (is_table) {
          ppt_tab <- pptxtable_as_tibble(node, preserve = preserve)
          ppt_tab$id <- id
          ppt_tab$content_type <- "table cell"
          ppt_tab$slide_id <- slide_id
          ppt_tab
        } else if (is_par) {
          ppt_tab <- pptx_par_as_tibble(node)
          ppt_tab$id <- id
          ppt_tab$content_type <- "paragraph"
          ppt_tab$slide_id <- slide_id
          ppt_tab
        } else if (is_img) {
          rel <- slide$relationship()
          images_ <- rel$get_images_path()
          img_id <- names(images_)
          images_ <- normalizePath(file.path(
            dirname(slide$file_name()),
            images_
          ))
          names(images_) <- img_id
          ppt_tab <- desc_pptx_img(node, images_)
          ppt_tab$id <- id
          ppt_tab$content_type <- "image"
          ppt_tab$slide_id <- slide_id
          ppt_tab
        } else {
          data.frame(
            id = id,
            content_type = "unknown",
            slide_id = slide_id,
            stringsAsFactors = FALSE
          )
        }
      },
      nodes,
      slide_id = i,
      SIMPLIFY = FALSE
    )

    list_content[[length(list_content) + 1]] <- rbind_match_columns(content)
  }

  rbind_match_columns(list_content)
}

#' @importFrom cli cli_abort
#' @importFrom dplyr mutate case_when left_join arrange across all_of lag
#' @importFrom dplyr consecutive_id syms n select inner_join if_else filter
#' @importFrom dplyr relocate summarise group_by as_tibble first bind_rows
#' @importFrom tidyr complete
#' @importFrom xml2 as_xml_document xml_has_attr xml_child xml_name xml_children xml_replace

# utils ----------
extract_runs_attr_str <- function(
  run_nodes,
  node_xpath,
  attr_name,
  default = NULL
) {
  nodes <- suppressWarnings({
    xml_child(run_nodes, node_xpath)
  })
  nodes_exist <- !is.na(xml_name(nodes))

  str_values <- rep("", length(run_nodes))
  tmp_results <- xml_attr(nodes, attr_name)
  str_values[nodes_exist] <- tmp_results[nodes_exist]
  if (!is.null(default)) {
    str_values[is.na(str_values)] <- default
  }

  str_values[str_values %in% ""] <- NA_character_
  str_values
}

extract_runs_attr_lgl <- function(
  run_nodes,
  node_xpath,
  attr_name,
  default = NULL,
  true_values = c("1", "on", "true", NA_character_)
) {
  nodes <- xml_child(run_nodes, node_xpath)
  nodes_exist <- !is.na(xml_name(nodes))

  lgl_values <- rep(default, length(run_nodes))
  lgl_values[nodes_exist] <- xml_attr(nodes[nodes_exist], attr_name) %in%
    true_values
  lgl_values
}

add_text_column <- function(x) {
  mutate(
    .data = x,
    text = case_when(
      !is.na(.data$run_content_text) ~ .data$run_content_text,
      .default = ""
    )
  )
}

#' @title Compute tables information
#' @description Joins and processes table-related information data frames
#'   to compute cell spans and merges.
#' @param tmp_infotbl_tablecells_paragraphs data.frame with paragraph
#'   information in table cells
#' @param tmp_infotbl_tablecells data.frame with table cells information
#' @param tmp_infotbl_tablerows data.frame with table rows information
#' @param tmp_infotbl_tables data.frame with tables information
#' @return A data.frame with complete table information including computed
#'   cell_id, row_span, and col_span values.
#' @noRd
infotbl_tables_compute <- function(
  tmp_infotbl_tablecells_paragraphs,
  tmp_infotbl_tablecells,
  tmp_infotbl_tablerows,
  tmp_infotbl_tables
) {
  infotbl_tables <- left_join(
    tmp_infotbl_tablecells_paragraphs,
    tmp_infotbl_tablecells,
    by = "cell_id"
  )
  infotbl_tables <- left_join(
    infotbl_tables,
    tmp_infotbl_tablerows,
    by = c("table_index", "row_id")
  )
  infotbl_tables <- left_join(
    infotbl_tables,
    tmp_infotbl_tables,
    by = c("table_index")
  )
  infotbl_tables <- arrange(
    .data = infotbl_tables,
    across(all_of(
      c("table_index", "row_id", "cell_id", "doc_index")
    ))
  )
  infotbl_tables <- mutate(
    .data = infotbl_tables,
    .by = all_of(c("table_index", "row_id")),
    add_span = as.integer(
      lag(.data$col_span, default = "1")
    ) -
      1L,
    cell_id = consecutive_id(.data$cell_id) + cumsum(.data$add_span),
    add_span = NULL
  )
  infotbl_tables <- tidyr::complete(
    infotbl_tables,
    !!!syms(c("table_index", "row_id", "cell_id")),
    fill = list(
      row_merge = FALSE,
      first = FALSE,
      col_span = "0",
      is_header = FALSE
    )
  )

  infotbl_tables <- mutate(
    .data = infotbl_tables,
    .by = all_of(c("table_index", "cell_id")),
    merge_group = cumsum(.data$first) + cumsum(!.data$row_merge)
  )

  infotbl_tables <- mutate(
    .data = infotbl_tables,
    .by = all_of(c("table_index", "cell_id", "merge_group")),
    row_span = n(),
    .after = all_of("col_span")
  )
  infotbl_tables <- mutate(
    .data = infotbl_tables,
    merge_group = NULL,
    row_span = case_when(
      .data$row_merge & !.data$first ~ 0L,
      .default = .data$row_span
    )
  )

  infotbl_tables
}

#' @title Augment data with non-text content
#' @description Adds information about images, footnotes, and hyperlinks
#'   to the data by joining with document relationships and processing
#'   footnote XML content.
#' @param data data.frame with document content including IDs for
#'   images (blip_id), footnotes (footnote_id), and hyperlinks (hyperlink_id)
#' @param x an rdocx object containing document relationships and footnotes
#' @return A data.frame with image paths, footnote text, and hyperlink URLs
#'   added, and temporary ID columns removed.
#' @noRd
augment_with_non_text_content <- function(data, x) {
  doc_rel <- x$doc_obj$rel_df()

  # add images
  doc_rel_images <- doc_rel[basename(doc_rel$type) %in% "image", ]
  doc_rel_images <- select(
    doc_rel_images,
    all_of(c("id", "target"))
  )
  names(doc_rel_images) <- c("blip_id", "image_path")
  doc_rel_images$image_path <- file.path(
    x$package_dir,
    "word",
    doc_rel_images$image_path
  )

  data <- left_join(
    data,
    doc_rel_images,
    by = "blip_id"
  )
  data$blip_id <- NULL

  # add footnotes
  ref_wml_str <- x$footnotes$encode_wml_str(additional_ns = list())
  ref_str <- rep(NA_character_, length(ref_wml_str))
  suppressWarnings({
    ref_str <- vapply(
      ref_wml_str,
      function(x) {
        str <- xml_text(xml2::as_xml_document(x))
        paste0(str, collapse = "\n")
      },
      FUN.VALUE = NA_character_
    )
    ref_str
  })
  ref_data <- data.frame(
    footnote_id = names(ref_str),
    footnote_text = unname(ref_str)
  )
  data <- left_join(
    data,
    ref_data,
    by = "footnote_id"
  )
  data$footnote_id <- NULL

  # add hyperlinks
  run_within_hl_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:hyperlink/w:r"
  )

  hl_nodes <- xml_child(run_within_hl_nodes, "parent::w:hyperlink")
  table_hyperlink <- data.frame(
    hyperlink_id = xml_attr(hl_nodes, "id"),
    link_to_bookmark = xml_attr(hl_nodes, "anchor"),
    run_index = as.integer(xml_attr(run_within_hl_nodes, "run_index"))
  )

  doc_rel_external_hyperlink <- doc_rel[
    basename(doc_rel$type) %in% "hyperlink",
  ]
  doc_rel_external_hyperlink <- select(
    doc_rel_external_hyperlink,
    all_of(c("id", "target"))
  )
  names(doc_rel_external_hyperlink) <- c("hyperlink_id", "link")

  table_to_anchor <- dplyr::anti_join(
    table_hyperlink,
    doc_rel_external_hyperlink,
    by = "hyperlink_id"
  )
  table_to_anchor$hyperlink_id <- NULL
  data <- left_join(
    data,
    table_to_anchor,
    by = c("run_index")
  )

  table_hyperlink <- inner_join(
    table_hyperlink,
    doc_rel_external_hyperlink,
    by = "hyperlink_id"
  )
  data <- left_join(
    data,
    select(
      table_hyperlink,
      all_of(c("hyperlink_id", "run_index", "link"))
    ),
    by = c("hyperlink_id", "run_index")
  )
  data$hyperlink_id <- NULL

  data
}

#' @title Augment data with style names
#' @description Adds style names (character, table, paragraph) to the data
#'   by joining with styles information and removes temporary style ID columns.
#' @param data data.frame with document content including style IDs
#'   (run_stylename, tab_style_name, par_style_name)
#' @param styles_data data.frame with style information containing
#'   style_type, style_id, and style_name columns
#' @return A data.frame with human-readable style names added
#'   (character_stylename, table_stylename, paragraph_stylename) and
#'   temporary style ID columns removed.
#' @noRd
augment_with_styles <- function(data, styles_data) {
  styles_character <- filter(
    styles_data,
    .data$style_type %in% "character"
  )
  styles_character <- select(
    styles_character,
    all_of(c("style_id", "style_name"))
  )
  names(styles_character) <- c("run_stylename", "character_stylename")
  data <- left_join(
    data,
    styles_character,
    by = "run_stylename"
  )

  styles_table <- filter(styles_data, .data$style_type %in% "table")
  styles_table <- select(
    styles_table,
    all_of(c("style_id", "style_name"))
  )
  names(styles_table) <- c("tab_style_name", "table_stylename")
  data <- left_join(
    data,
    styles_table,
    by = "tab_style_name"
  )

  styles_paragraph <- filter(
    styles_data,
    .data$style_type %in% "paragraph"
  )
  styles_paragraph <- select(
    styles_paragraph,
    all_of(c("style_id", "style_name"))
  )
  names(styles_paragraph) <- c("par_style_name", "paragraph_stylename")
  data <- left_join(
    data,
    styles_paragraph,
    by = "par_style_name"
  )
  data <- select(
    data,
    -all_of(c("par_style_name", "tab_style_name", "run_stylename"))
  )

  data
}

#' @title Join information tables for document content
#' @description Joins runs, paragraphs, and tables information data frames
#'   and processes the final document content structure.
#' @param infotbl_runs_contents data.frame with run content information
#'   (text, images, footnotes, etc.)
#' @param infotbl_runs data.frame with run formatting information
#'   (bold, italic, font, etc.)
#' @param infotbl_paragraphs data.frame with paragraph information
#'   (style, alignment, numbering, etc.)
#' @param infotbl_tables data.frame with table information
#'   (cell positions, spans, etc.)
#' @return A data.frame with all information joined and properly ordered,
#'   with consecutive run and content indices.
#' @noRd
infotbl_join <- function(
  infotbl_runs_contents,
  infotbl_runs,
  infotbl_paragraphs,
  infotbl_tables
) {
  data <- left_join(
    infotbl_runs_contents,
    infotbl_runs,
    by = "run_index"
  )
  data <- left_join(
    data,
    infotbl_paragraphs,
    by = "doc_index"
  )

  data <- left_join(
    data,
    infotbl_tables,
    by = "doc_index"
  )

  data <- relocate(
    .data = data,
    all_of(
      c(
        "doc_index",
        "run_index",
        "run_content_index",
        "table_index",
        "row_id",
        "cell_id"
      )
    )
  )
  data <- mutate(
    .data = data,
    doc_index = as.integer(.data$doc_index),
    run_index = as.integer(.data$run_index)
  )
  data <- arrange(
    .data = data,
    across(all_of(
      c("doc_index", "run_index", "run_content_index")
    ))
  )
  data <- mutate(
    .data = data,
    .by = all_of(c("doc_index", "run_index")),
    .keep = "all",
    run_content_index = consecutive_id(.data$run_content_index)
  )

  data
}


# extract tools ----
docx_tablecells_information <- function(tc_nodes) {
  col_span_str <- rep("1", length(tc_nodes))
  col_span_nodes <- xml_child(tc_nodes, "w:tcPr/w:gridSpan")
  has_col_span_nodes <- xml_name(col_span_nodes) %in% "gridSpan"
  col_span_str[has_col_span_nodes] <- xml_attr(
    col_span_nodes[has_col_span_nodes],
    "val"
  )

  vmerge_nodes <- xml_child(tc_nodes, "w:tcPr/w:vMerge")
  row_merge <- !is.na(xml_name(vmerge_nodes))
  first <- xml_attr(vmerge_nodes, "val") %in% "restart"

  tc_index <- xml_attr(tc_nodes, "tc_index")
  table_index_sel <- xml_child(tc_nodes, "parent::w:tr/parent::w:tbl")
  table_index <- xml_attr(table_index_sel, "table_index")
  tr_index_sel <- xml_child(tc_nodes, "parent::w:tr")
  tr_index <- xml_attr(tr_index_sel, "tr_index")

  data.frame(
    table_index = as.integer(table_index),
    row_id = as.integer(tr_index),
    cell_id = as.integer(tc_index),
    row_merge = row_merge,
    first = first,
    col_span = col_span_str
  )
}

docx_p_cells_information <- function(p_in_cell_nodes) {
  tc_index <- xml_child(p_in_cell_nodes, "parent::w:tc")
  tc_index <- xml_attr(tc_index, "tc_index")
  doc_index <- xml_attr(p_in_cell_nodes, "doc_index")

  data.frame(
    doc_index = as.integer(doc_index),
    cell_id = as.integer(tc_index)
  )
}
docx_p_information <- function(p_nodes) {
  ilvl_nodes <- xml_child(p_nodes, "w:pPr/w:numPr/w:ilvl")
  numId_nodes <- xml_child(p_nodes, "w:pPr/w:numPr/w:numId")
  pstyle_nodes <- xml_child(p_nodes, "w:pPr/w:pStyle")
  jc_nodes <- xml_child(p_nodes, "w:pPr/w:jc")
  kn_nodes <- xml_child(p_nodes, "w:pPr/w:keepNext")

  doc_index <- xml_attr(p_nodes, "doc_index")

  has_num_pr <- xml_name(xml_child(p_nodes, "w:pPr/w:numPr")) %in% "numPr"

  level_str <- rep(NA_character_, length(p_nodes))
  level_str[has_num_pr] <- xml_attr(ilvl_nodes[has_num_pr], "val")

  num_id_str <- rep(NA_character_, length(p_nodes))
  num_id_str[has_num_pr] <- xml_attr(numId_nodes[has_num_pr], "val")

  data.frame(
    doc_index = as.integer(doc_index),
    par_style_name = xml_attr(pstyle_nodes, "val"),
    keep_with_next = xml_name(kn_nodes) %in% "keepNext",
    align = xml_attr(jc_nodes, "val"),
    level = as.integer(level_str) + 1L,
    num_id = as.integer(num_id_str),
    stringsAsFactors = FALSE
  )
}

docx_tables_information <- function(tbl_nodes) {
  tbl_stylename <- xml_attr(tbl_nodes, "val")
  tblstyle_nodes <- xml_child(
    tbl_nodes,
    "w:tblPr/w:tblStyle"
  )
  data.frame(
    tab_style_name = xml_attr(tblstyle_nodes, "val"),
    table_index = as.integer(xml_attr(tbl_nodes, "table_index"))
  )
}

docx_runs_information <- function(run_nodes) {
  bold <- extract_runs_attr_lgl(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:b",
    attr_name = "val",
    default = FALSE
  )
  italic <- extract_runs_attr_lgl(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:i",
    attr_name = "val",
    default = FALSE
  )
  underline <- extract_runs_attr_lgl(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:u",
    attr_name = "val",
    true_values = c(
      "single",
      "words",
      "double",
      "thick",
      "dash",
      NA_character_
    ),
    default = FALSE
  )

  shd_nodes <- xml_child(run_nodes, "w:rPr/w:shd")
  shading <- xml_attr(shd_nodes, "val")
  shading[!is.na(shading)] <- sprintf("#%s", shading[!is.na(shading)])
  shading_color <- xml_attr(shd_nodes, "color")
  shading_color[!is.na(shading_color)] <- sprintf(
    "#%s",
    shading_color[!is.na(shading_color)]
  )
  shading_fill <- xml_attr(shd_nodes, "fill")
  shading_fill[!is.na(shading_fill)] <- sprintf(
    "#%s",
    shading_fill[!is.na(shading_fill)]
  )

  run_stylename <- extract_runs_attr_str(
    run_nodes = run_nodes,
    node_xpath = "w:rPr/w:rStyle",
    attr_name = "val",
    default = NA_character_
  )
  hyperlink_id <- extract_runs_attr_str(
    run_nodes = run_nodes,
    node_xpath = "parent::w:hyperlink",
    attr_name = "id",
    default = NA_character_
  )

  run_index <- xml_attr(run_nodes, "run_index")

  rfonts_nodes <- xml_child(run_nodes, "w:rPr/w:rFonts")

  doc_index <- extract_runs_attr_str(
    run_nodes = run_nodes,
    node_xpath = "parent::w:p|parent::*/parent::w:p",
    attr_name = "doc_index",
    default = NA_character_
  )

  sz <- xml_attr(xml_child(run_nodes, "w:rPr/w:sz"), "val")
  sz <- as.integer(sz)

  sz_cs = xml_attr(xml_child(run_nodes, "w:rPr/w:szCs"), "val")
  sz_cs <- as.integer(sz_cs)

  color <- xml_attr(xml_child(run_nodes, "w:rPr/w:color"), "val")
  color[!is.na(color)] <- sprintf("#%s", color[!is.na(color)])

  data.frame(
    doc_index = as.integer(doc_index),
    run_index = run_index,
    run_stylename = run_stylename,
    hyperlink_id = hyperlink_id,
    sz = sz,
    sz_cs = sz_cs,
    font_family_ascii = xml_attr(rfonts_nodes, "ascii"),
    font_family_eastasia = xml_attr(rfonts_nodes, "eastAsia"),
    font_family_hansi = xml_attr(rfonts_nodes, "hAnsi"),
    font_family_cs = xml_attr(rfonts_nodes, "cs"),
    bold = bold,
    italic = italic,
    underline = underline,
    color = color,
    shading = shading,
    shading_color = shading_color,
    shading_fill = shading_fill
  )
}


docx_runs_content_information <- function(run_nodes) {
  run_content_nodes_names <- xml_name(run_nodes)

  suppressWarnings({
    blip_id_ <- xml_attr(
      xml_child(
        run_nodes,
        "wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip"
      ),
      "embed"
    )
    footnote_id_ <- xml_attr(run_nodes, "id")
    text_values <- xml_text(run_nodes)
  })
  field_code <- case_when(
    run_content_nodes_names %in% "instrText" ~ text_values,
    .default = NA_character_
  )
  data.frame(
    blip_id = case_when(
      run_content_nodes_names %in% "drawing" ~ blip_id_,
      .default = NA_character_
    ),
    footnote_id = case_when(
      run_content_nodes_names %in% "footnoteReference" ~ footnote_id_,
      .default = NA_character_
    ),
    field_code = field_code,
    run_content_text = case_when(
      run_content_nodes_names %in% "t" ~ text_values,
      run_content_nodes_names %in% "noBreakHyphen" ~ "\u002D",
      run_content_nodes_names %in% c("br", "cr") ~ "\n",
      run_content_nodes_names %in% "tab" ~ "\t",
      run_content_nodes_names %in% "instrText" ~ field_code,
      run_content_nodes_names %in% "softHyphen" ~ "",
      run_content_nodes_names %in% "sym" ~ xml_attr(run_nodes, "char"),
      .default = NA_character_
    ),
    run_content_index = xml_attr(run_nodes, "run_content_index"),
    run_index = xml_attr(xml_child(run_nodes, "parent::w:r"), "run_index")
  )
}

docx_tablerows_information <- function(tr_nodes) {
  tbl_header_nodes <- xml_child(tr_nodes, "w:trPr/w:tblHeader")
  first_row_nodes <- xml_child(tr_nodes, "w:trPr/w:cnfStyle[@w:firstRow='1']")
  is_header <- xml_name(first_row_nodes) %in%
    "cnfStyle" |
    xml_name(tbl_header_nodes) %in% "tblHeader"
  tr_index <- xml_attr(tr_nodes, "tr_index")

  table_index <- xml_child(tr_nodes, "parent::w:tbl")
  table_index <- xml_attr(table_index, "table_index")

  data.frame(
    table_index = as.integer(table_index),
    row_id = as.integer(tr_index),
    is_header = is_header
  )
}


# main ----------

#' @title Summarise data as paragraphs
#' @description Summarises detailed run-level data into paragraph-level
#'   summaries by collapsing text within paragraphs. Handles both regular
#'   paragraphs and paragraphs within table cells differently.
#' @param data data.frame with detailed run-level information including
#'   doc_index, table_index, paragraph_stylename, and other formatting details
#' @return A data.frame summarised at the paragraph level with collapsed text
#'   in the column mamed `text`. For tables, includes cell and row
#'   information; for regular paragraphs, includes only doc_index and style.
#' @noRd
summarise_as_paragraph <- function(data, preserve = FALSE) {
  finest_grp_by <- c(
    "doc_index",
    "content_type",
    "table_index",
    "row_id",
    "cell_id",
    "is_header",
    "row_span",
    "col_span",
    "table_stylename",
    "paragraph_stylename"
  )

  table_data <- filter(.data = data, !is.na(.data$table_index))
  table_data <- add_text_column(table_data)
  table_data <- summarise(
    .data = table_data,
    .by = all_of(finest_grp_by),
    text = paste0(.data$text, collapse = "")
  )

  if (preserve) {
    finest_grp_by <- c(
      "content_type",
      "table_index",
      "row_id",
      "cell_id",
      "is_header",
      "row_span",
      "col_span",
      "table_stylename",
      "paragraph_stylename"
    )
    table_data <- summarise(
      .data = table_data,
      doc_index = first(.data$doc_index),
      text = paste0(.data$text, collapse = "\n"),
      .by = all_of(finest_grp_by)
    )
  }

  par_data <- filter(.data = data, is.na(.data$table_index))
  par_data <- add_text_column(par_data)
  par_data <- summarise(
    .data = par_data,
    .by = all_of(
      c("doc_index", "content_type", "paragraph_stylename")
    ),
    text = paste0(.data$text, collapse = "")
  )

  dataset <- bind_rows(par_data, table_data)

  # patch for WordR
  names(dataset)[names(dataset) %in% "paragraph_stylename"] <- "style_name"

  dataset
}


#' @title Get Word content in a data.frame
#' @description read content of a Word document and
#' return a data.frame representing the document.
#' @note
#' Documents included with [body_add_docx()] will
#' not be accessible in the results.
#' @param x an rdocx object
#' @param preserve If `FALSE` (default), text in table cells is collapsed into a
#'   single line. If `TRUE`, line breaks in table cells are preserved as a "\\n"
#'   character. This feature is adapted from `docxtractr::docx_extract_tbl()`
#'   published under a [MIT
#'   licensed](https://github.com/hrbrmstr/docxtractr/blob/master/LICENSE) in
#'   the 'docxtractr' package by Bob Rudis.
#' @param remove_fields if TRUE, prevent field codes from appearing in the
#' returned data.frame.
#' @param detailed Should run-level information be included in the dataframe?
#'   Defaults to `FALSE`. If `TRUE`, the dataframe contains detailed information
#'   about each run (text formatting, images, hyperlinks, etc.) instead of
#'   collapsing content at the paragraph level. When `FALSE`, run-level
#'   information such as images, hyperlinks, and text formatting is not available
#'   since data is aggregated at the paragraph level.
#' @return A data.frame with the following columns depending on the value of `detailed`:
#'
#' When `detailed = FALSE` (default), the data.frame contains:
#'
#' - `doc_index`: Document element index (integer).
#' - `content_type`: Type of content: "paragraph" or "table cell" (character).
#' - `style_name`: Name of the paragraph style (character).
#' - `text`: Collapsed text content of the paragraph or cell (character).
#' - `table_index`: Index of the table (integer). `NA` for non-table content.
#' - `row_id`: Row position in table (integer). `NA` for non-table content.
#' - `cell_id`: Cell position in table row (integer). `NA` for non-table content.
#' - `is_header`: Whether the row is a table header (logical). `NA` for non-table content.
#' - `row_span`: Number of rows spanned by the cell (integer). `0` for merged cells. `NA` for non-table content.
#' - `col_span`: Number of columns spanned by the cell (character). `NA` for non-table content.
#' - `table_stylename`: Name of the table style (character). `NA` for non-table content.
#'
#' When `detailed = TRUE`, the data.frame contains additional run-level information:
#'
#' - `run_index`: Index of the run within the paragraph (integer).
#' - `run_content_index`: Index of content element within the run (integer).
#' - `run_content_text`: Text content of the run element (character).
#' - `image_path`: Path to embedded image stored in the temporary directory
#'   associated with the rdocx object (character).
#'   Images should be copied to a permanent location before closing the R
#'   session if needed.
#' - `field_code`: Field code content (character).
#' - `footnote_text`: Footnote text content (character).
#' - `link`: Hyperlink URL (character).
#' - `link_to_bookmark`: Internal bookmark anchor name for hyperlinks (character).
#' - `bookmark_start`: Names of the bookmarks starting on this paragraph
#' (values are concatenated with '|').
#' - `character_stylename`: Name of the character/run style (character).
#' - `sz`: Font size in half-points (integer).
#' - `sz_cs`: Complex script font size in half-points (integer).
#' - `font_family_ascii`: Font family for ASCII characters (character).
#' - `font_family_eastasia`: Font family for East Asian characters (character).
#' - `font_family_hansi`: Font family for high ANSI characters (character).
#' - `font_family_cs`: Font family for complex script characters (character).
#' - `bold`: Whether the run is bold (logical).
#' - `italic`: Whether the run is italic (logical).
#' - `underline`: Whether the run is underlined (logical).
#' - `color`: Text color in hexadecimal format (character).
#' - `shading`: Shading pattern (character).
#' - `shading_color`: Shading foreground color (character).
#' - `shading_fill`: Shading background fill color (character).
#' - `keep_with_next`: Whether paragraph should stay with next (logical).
#' - `align`: Paragraph alignment (character).
#' - `level`: Numbering level (integer). `NA` if not a numbered list.
#' - `num_id`: Numbering definition ID (integer). `NA` if not a numbered list.
#' @example inst/examples/example_docx_summary.R
#' @export
docx_summary <- function(
  x,
  preserve = FALSE,
  remove_fields = FALSE,
  detailed = FALSE
) {
  if (!requireNamespace("dplyr")) {
    cli::cli_abort(
      "package {.pkg dplyr} is required for running {.fn docx_summary}."
    )
  }
  if (!requireNamespace("tidyr")) {
    cli::cli_abort(
      "package {.pkg tidyr} is required for running {.fn docx_summary}."
    )
  }

  if (remove_fields) {
    # also remove complex field characters
    fldChar_nodes <- xml_find_all(x$doc_obj$get(), "//w:fldChar")
    xml_remove(fldChar_nodes)

    # remove instrText nodes
    instrText_nodes <- xml_find_all(x$doc_obj$get(), "//w:instrText")
    xml_remove(instrText_nodes)

    # remove fldData nodes
    fldData_nodes <- xml_find_all(x$doc_obj$get(), "//w:fldData")
    xml_remove(fldData_nodes)
  }

  # add info to xml -----
  ## p_nodes
  p_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:p"
  )
  xml_attr(p_nodes, "w:doc_index") <- seq_along(p_nodes)

  ## tbl_nodes
  tbl_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tbl"
  )
  xml_attr(tbl_nodes, "w:table_index") <- seq_along(tbl_nodes)

  ## tr_nodes
  tr_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tr"
  )
  xml_attr(tr_nodes, "w:tr_index") <- seq_along(tr_nodes)

  ## tc_nodes
  tc_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tc"
  )
  xml_attr(tc_nodes, "w:tc_index") <- seq_along(tc_nodes)

  ## run_nodes
  run_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:r"
  )
  xml_attr(run_nodes, "w:run_index") <- seq_along(run_nodes)

  ## run_content_nodes
  run_content_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:r/*[not(self::w:rPr)]"
  )
  xml_attr(run_content_nodes, "w:run_content_index") <- seq_along(
    run_content_nodes
  )

  ## bookmark_nodes
  bookmark_nodes <- xml_find_all(
    x$doc_obj$get(),
    "w:body/w:p/w:bookmarkStart"
  )

  data_bookmark <- data.frame(
    bookmark_start = xml_attr(bookmark_nodes, "name"),
    doc_index = xml_attr(xml_child(bookmark_nodes, "parent::w:p"), "doc_index")
  )
  data_bookmark <- summarise(
    .data = data_bookmark,
    bookmark_start = paste0(.data$bookmark_start, collapse = "|"),
    .by = all_of("doc_index")
  )
  data_bookmark$doc_index <- as.integer(data_bookmark$doc_index)

  ## p_in_cell_nodes
  p_in_cell_nodes <- xml_find_all(
    x$doc_obj$get(),
    "//w:tbl/w:tr/w:tc/w:p"
  )

  # info for runs: infotbl_runs -----
  infotbl_runs_contents <- docx_runs_content_information(run_content_nodes)
  infotbl_runs <- docx_runs_information(run_nodes)

  # info for tables: infotbl_tables -----
  tmp_infotbl_tables <- docx_tables_information(tbl_nodes)
  tmp_infotbl_tablecells <- docx_tablecells_information(tc_nodes)
  tmp_infotbl_tablecells_paragraphs <- docx_p_cells_information(p_in_cell_nodes)
  tmp_infotbl_tablerows <- docx_tablerows_information(tr_nodes)
  ## final table infotbl_tables
  infotbl_tables <- infotbl_tables_compute(
    tmp_infotbl_tablecells_paragraphs,
    tmp_infotbl_tablecells,
    tmp_infotbl_tablerows,
    tmp_infotbl_tables
  )

  # info for paragraphs: infotbl_paragraphs -----
  infotbl_paragraphs <- docx_p_information(p_nodes)
  infotbl_paragraphs <- left_join(
    infotbl_paragraphs,
    data_bookmark,
    by = "doc_index"
  )

  # final joins -------
  data <- infotbl_join(
    infotbl_runs_contents,
    infotbl_runs,
    infotbl_paragraphs,
    infotbl_tables
  )

  # add stylenames -----
  styles_data <- select(
    styles_info(x),
    all_of(c("style_type", "style_id", "style_name"))
  )
  data <- augment_with_styles(data, styles_data)

  # add non textual content -----
  data <- augment_with_non_text_content(data, x)

  data <- mutate(
    .data = data,
    .by = all_of(c("doc_index")),
    .keep = "all",
    run_index = consecutive_id(.data$run_index)
  )
  data <- mutate(
    .data = data,
    content_type = if_else(
      is.na(.data$table_index),
      "paragraph",
      "table cell"
    )
  )

  column_names <- c(
    "doc_index",
    "content_type",
    "run_index",
    "run_content_index",

    "run_content_text",
    "image_path",
    "field_code",
    "footnote_text",
    "link",
    "link_to_bookmark",
    "bookmark_start",
    "character_stylename",
    "sz",
    "sz_cs",
    "font_family_ascii",
    "font_family_eastasia",
    "font_family_hansi",
    "font_family_cs",
    "bold",
    "italic",
    "underline",
    "color",
    "shading",
    "shading_color",
    "shading_fill",

    "paragraph_stylename",
    "keep_with_next",
    "align",
    "level",
    "num_id",

    "table_index",
    "row_id",
    "cell_id",
    "col_span",
    "row_span",
    "is_header",
    "table_stylename"
  )

  data <- select(.data = data, all_of(column_names))

  if (!detailed) {
    data <- summarise_as_paragraph(data, preserve = preserve)
  }

  data
}

# Example 1: Setting page layout properties ----
# This example demonstrates how to configure the default section
# properties for page orientation, type, and margins

# Define custom section properties
# - Landscape orientation for wide tables
# - Continuous section (no page break)
# - Custom margins: top=1.5", bottom=0.75", left/right=2"
default_sect_properties <- prop_section(
  page_size = page_size(orient = "landscape"),
  type = "continuous",
  page_margins = page_mar(bottom = .75, top = 1.5, right = 2, left = 2)
)

# Create a new document
doc_1 <- read_docx()

# Add a wide table that benefits from landscape orientation
doc_1 <- body_add_par(doc_1, "Motor Trend Car Road Tests", style = "heading 1")
doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")

# Add some text content
doc_1 <- body_add_par(doc_1, "Sample Text Content", style = "heading 2")
doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 40), collapse = " "))

# Apply the section properties to the entire document
# This must be called at the end, after all content is added
doc_1 <- body_set_default_section(doc_1, default_sect_properties)

# Save the document
print(doc_1, target = tempfile(fileext = ".docx"))


# Example 2: Adding headers and footers ----
# This example shows how to create a document with:
# - A header containing the R logo
# - A footer with the current date and page numbers

# Get the path to R logo
img_path <- file.path(R.home("doc"), "html", "logo.jpg")

# Create header content with the R logo
# The logo is positioned on the right side with specific dimensions
header_content <- block_list(
  fpar(
    external_img(src = img_path, height = 0.5, width = 0.5),
    fp_p = fp_par(text.align = "right")
  )
)

# Create footer content with date and page numbers
# Format: "Document generated on: [Date] | Page [X]"
footer_content <- block_list(
  fpar(
    "Document generated on: ",
    run_word_field(field = "Date \\@ \"MMMM d, yyyy\""),
    " | Page ",
    run_word_field(field = "PAGE"),
    fp_p = fp_par(text.align = "center")
  )
)

# Define section properties that include header and footer
# The header and footer will appear on all pages
sect_with_hf <- prop_section(
  page_size = page_size(orient = "portrait", width = 8.3, height = 11.7),
  page_margins = page_mar(
    bottom = 1,
    top = 1,
    right = 1,
    left = 1,
    header = 0.5,
    footer = 0.5
  ),
  type = "continuous",
  header_default = header_content,
  footer_default = footer_content
)

# Create a new document with content
doc_2 <- read_docx()

# Add a title page
doc_2 <- body_add_par(doc_2, "Annual Report 2024", style = "heading 1")
doc_2 <- body_add_par(
  doc_2,
  "Company Performance Analysis",
  style = "heading 2"
)

# Add some sections with content
doc_2 <- body_add_par(doc_2, "Executive Summary", style = "heading 2")
doc_2 <- body_add_par(
  doc_2,
  "This report provides a comprehensive analysis of company performance metrics."
)

# Add a table
doc_2 <- body_add_par(doc_2, "Key Metrics", style = "heading 2")
summary_data <- data.frame(
  Metric = c("Revenue", "Profit", "Growth"),
  Q1 = c(1.2, 0.3, 12),
  Q2 = c(1.5, 0.4, 15),
  Q3 = c(1.8, 0.5, 18),
  Q4 = c(2.1, 0.6, 20)
)
doc_2 <- body_add_table(doc_2, value = summary_data, style = "table_template")

# Add a plot
doc_2 <- body_add_par(doc_2, "Revenue Trend", style = "heading 2")
revenue_plot <- plot_instr({
  quarters <- paste0("Q", 1:4)
  revenue <- c(1.2, 1.5, 1.8, 2.1)
  barplot(
    revenue,
    names.arg = quarters,
    col = "#4472C4",
    border = NA,
    main = "Quarterly Revenue (Millions)",
    ylab = "Revenue ($M)",
    xlab = "Quarter"
  )
})
doc_2 <- body_add_plot(doc_2, revenue_plot, width = 5, height = 4)

# Apply section properties with header and footer
# The header (with R logo) and footer (with date and page number)
# will appear on all pages
doc_2 <- body_set_default_section(doc_2, sect_with_hf)

# Save the document
output_file <- tempfile(fileext = ".docx")
print(doc_2, target = output_file)

# PH_WITH -------

#' @export
#' @title Add objects on the current slide
#' @description add an object into a new shape in the current slide. This
#' function is able to add all supported outputs to a presentation. See
#' section **Methods (by class)** to see supported outputs.
#' @param x an rpptx object
#' @param value object to add as a new shape. Supported objects
#' are vectors, data.frame, graphics, block of formatted paragraphs,
#' unordered list of formatted paragraphs,
#' pretty tables with package flextable, editable graphics with
#' package rvg, 'Microsoft' charts with package mschart.
#' @param location a placeholder location object or a location short form. It will be used
#' to specify the location of the new shape. This location can be defined with a call to one
#' of the `ph_location_*` functions (see section `"see also"`). In `ph_with()`, several location
#' short forms can be used, as listed in section `"Short forms"`.
#' @param ... further arguments passed to or from other methods. When
#' adding a `ggplot` object or `plot_instr`, these arguments will be used
#' by the png function.
#'
#' @section Short forms:
#' The `location` argument of `ph_with()` either expects a location object as returned by the
#' `ph_location_*` functions or a corresponding location *short form* (string or numeric):
#'
#' | **Location function**                       | **Short form**             | **Description**                                      |
#' |---------------------------------------------|----------------------------|------------------------------------------------------|
#' | `ph_location_left()`                        | `"left"`                   | Keyword string                                       |
#' | `ph_location_right()`                       | `"right"`                  | Keyword string                                       |
#' | `ph_location_fullsize()`                    | `"fullsize"`               | Keyword string                                       |
#' | `ph_location_type("body", 1)`               | `"body [1]"`               | String: type + index in brackets (`1` if omitted)    |
#' | `ph_location_label("my_label")`             | `"my_label"`               | Any string not matching a keyword or type            |
#' | `ph_location_id(1)`                         | `1`                        | Length 1 integer                                     |
#' | `ph_location(0, 0, 4, 5)`                   | `c(0,0,4,5)`               | Length 4 numeric, optionally named, `c(top=0, left=0, ...)` |
#'
#' @example inst/examples/example_ph_with.R
#' @seealso Specify placeholder locations with [ph_location_type], [ph_location],
#' [ph_location_label], [ph_location_left], [ph_location_right],
#' [ph_location_fullsize], [ph_location_template]. [phs_with] is a sibling of
#' `ph_with` that fills multiple placeholders at once. Use [add_slide] to add new slides.
#' @section Illustrations:
#'
#' \if{html}{\figure{ph_with_doc_1.png}{options: style="width:80\%;"}}
ph_with <- function(x, value, location, ...) {
  location <- resolve_location(location)
  .ph_with(x, value, location, ...)
}


.ph_with <- function(x, value, location, ...) {
  UseMethod("ph_with", value)
}


#' @export
#' @describeIn ph_with add a character vector to a new shape on the
#' current slide, values will be added as paragraphs.
ph_with.character <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)
  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )

  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}

#' @export
#' @param format_fun format function for non character vectors
#' @describeIn ph_with add a numeric vector to a new shape on the
#' current slide, values will be be first formatted then
#' added as paragraphs.
ph_with.numeric <- function(x, value, location, format_fun = format, ...) {
  slide <- x$slide$get_slide(x$cursor)
  value <- format_fun(value, ...)
  location <- fortify_location(location, doc = x)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}

#' @export
#' @describeIn ph_with add a factor vector to a new shape on the
#' current slide, values will be be converted as character and then
#' added as paragraphs.
ph_with.factor <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)
  value <- as.character(value)
  location <- fortify_location(location, doc = x)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  pars <- paste0(
    "<a:p><a:r><a:rPr/><a:t>",
    htmlEscapeCopy(value),
    "</a:t></a:r></a:p>",
    collapse = ""
  )
  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}
#' @export
#' @rdname ph_with
ph_with.logical <- ph_with.numeric


#' @export
#' @param date_format A format string for dates (default `"%Y-%m-%d"`).
#' See `format` arg in [strftime()] for details.
#' Set a global default via `options(officer.date_format = ...)`.
#' @describeIn ph_with add a `Date` object vector to a new shape on the
#' current slide, values will be be first converted to character.
ph_with.Date <- function(x, value, location, date_format = NULL, ...) {
  opt <- options()$officer.date_format
  if (!is.null(opt) && is.na(opt)) {
    # catch NA
    opt <- NULL
  }
  format <- date_format %||% opt %||% "%Y-%m-%d" # fallback to format.Date() default
  value_str <- format(value, format = format)
  ph_with(x, value = value_str, location, ...)
}


#' @export
#' @param level_list The list of levels for hierarchy structure as integer values.
#' If used the object is formated as an unordered list. If 1 and 2,
#' item 1 level will be 1, item 2 level will be 2.
#' @describeIn ph_with add a [block_list()] made
#' of [fpar()] to a new shape on the current slide.
ph_with.block_list <- function(
  x,
  value,
  location,
  level_list = integer(0),
  ...
) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)

  pars <- vapply(value, to_pml, character(1))

  if (length(level_list) > 0) {
    pars <- gsub("<a:buNone/>", "", pars, fixed = TRUE)

    level_values <- rep(1L, length(pars))
    level_values[] <- level_list
    level_values <- level_values - 1L

    lvl <- sprintf(" lvl=\"%.0f\"", level_values)
    lvl <- paste0("<a:pPr", ifelse(level_values > 0, lvl, ""), "/>")

    pars <- mapply(
      function(par, lvl) {
        paste0(par[1], lvl, par[2])
      },
      strsplit(pars, split = "<a:pPr(.*)</a:pPr>"),
      lvl,
      SIMPLIFY = FALSE
    )
    pars <- unlist(pars)
  }

  pars <- paste0(pars, collapse = "")

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    pars,
    "</p:txBody></p:sp>"
  )

  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}


#' @export
#' @describeIn ph_with add a [unordered_list()] made
#' of [fpar()] to a new shape on the current slide.
ph_with.unordered_list <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)
  location <- fortify_location(location, doc = x)

  p <- to_pml(value)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    p,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}

#' @export
#' @describeIn ph_with add a [block_list_items()] (bullet or numbered list)
#' to a new shape on the current slide.
ph_with.block_list_items <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)
  location <- fortify_location(location, doc = x)

  p <- to_pml(value)

  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  xml_elt <- paste0(
    psp_ns_yes,
    new_ph,
    "<p:txBody><a:bodyPr/><a:lstStyle/>",
    p,
    "</p:txBody></p:sp>"
  )
  node <- as_xml_document(xml_elt)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}

#' @export
#' @param header display header if TRUE
#' @param tcf conditional formatting settings defined by [table_conditional_formatting()]
#' @param alignment alignment for each columns, 'l' for left, 'r' for right
#' and 'c' for center. Default to NULL.
#' @describeIn ph_with add a data.frame to a new shape on the current slide with
#' function [block_table()]. Use package 'flextable' instead for more
#' advanced formattings.
ph_with.data.frame <- function(
  x,
  value,
  location,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
) {
  location <- fortify_location(location, doc = x)

  slide <- x$slide$get_slide(x$cursor)
  style_id <- x$table_styles$def[1]

  pt <- prop_table(
    style = style_id,
    layout = table_layout(),
    width = table_width(),
    tcf = tcf
  )

  bt <- block_table(
    x = value,
    header = header,
    properties = pt,
    alignment = alignment
  )

  xml_elt <- to_pml(
    bt,
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  value <- as_xml_document(xml_elt)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), value)
  x
}


#' @export
#' @describeIn ph_with add a ggplot object to a new shape on the
#' current slide. Use package 'rvg' for more advanced graphical features.
#' @param res resolution of the png image in ppi
#' @param alt_text Alt-text for screen-readers. Defaults to `""`. If `""` or `NULL`
#'    an alt text added with `ggplot2::labs(alt = ...)` will be used if any.
#' @param scale Multiplicative scaling factor, same as in ggsave
ph_with.gg <- function(
  x,
  value,
  location,
  res = 300,
  alt_text = "",
  scale = 1,
  ...
) {
  location_ <- fortify_location(location, doc = x)
  slide <- x$slide$get_slide(x$cursor)
  if (!requireNamespace("ggplot2")) {
    stop("package ggplot2 is required to use this function")
  }

  slide <- x$slide$get_slide(x$cursor)
  width <- location_$width
  height <- location_$height

  stopifnot(inherits(value, "gg"))
  file <- tempfile(fileext = ".png")
  agg_png(
    filename = file,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = scale,
    background = "transparent",
    ...
  )
  print(value)
  dev.off()
  on.exit(unlink(file))

  if (is.null(alt_text) || alt_text == "") {
    alt_text <- ggplot2::get_alt_text(value)
    if (is.null(alt_text)) alt_text <- ""
  }

  ext_img <- external_img(file, width = width, height = height, alt = alt_text)

  ph_with(x, ext_img, location = location)
}

#' @export
#' @describeIn ph_with add an R plot to a new shape on the
#' current slide. Use package 'rvg' for more advanced graphical features.
ph_with.plot_instr <- function(x, value, location, res = 300, ...) {
  location_ <- fortify_location(location, doc = x)
  slide <- x$slide$get_slide(x$cursor)
  slide <- x$slide$get_slide(x$cursor)
  width <- location_$width
  height <- location_$height

  file <- tempfile(fileext = ".png")

  dirname <- tempfile()
  dir.create(dirname)
  filename <- file.path(dirname, "plot%03d.png")
  agg_png(
    filename = filename,
    width = width,
    height = height,
    units = "in",
    res = res,
    scaling = 1,
    background = "transparent",
    ...
  )

  tryCatch(
    {
      eval(value$code)
    },
    finally = {
      dev.off()
    }
  )
  file <- list.files(dirname, full.names = TRUE)
  on.exit(unlink(dirname, recursive = TRUE, force = TRUE))

  if (length(file) > 1) {
    stop(
      length(file),
      " files have been produced. Multiple plot are not supported"
    )
  }

  ext_img <- external_img(file, width = width, height = height)
  ph_with(x, ext_img, location = location)
}


#' @export
#' @param use_loc_size if set to FALSE, external_img width and height will
#' be used.
#' @describeIn ph_with add a [external_img()] to a new shape
#' on the current slide.
#'
#' When value is a external_img object, image will be copied
#' into the PowerPoint presentation. The width and height
#' specified in call to [external_img()] will be
#' ignored, their values will be those of the location,
#' unless use_loc_size is set to FALSE.
ph_with.external_img <- function(x, value, location, use_loc_size = TRUE, ...) {
  location <- fortify_location(location, doc = x)

  slide <- x$slide$get_slide(x$cursor)

  if (!use_loc_size) {
    location$width <- attr(value, "dims")$width
    location$height <- attr(value, "dims")$height
  }
  width <- location$width
  height <- location$height

  xml_str <- to_pml(
    x = value,
    left = location$left,
    top = location$top,
    width = width,
    height = height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln
  )

  value <- as_xml_document(xml_str)
  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), value)
  x
}


#' @export
#' @describeIn ph_with add an [fpar()] to a new shape
#' on the current slide as a single paragraph in a [block_list()].
ph_with.fpar <- function(x, value, location, ...) {
  ph_with.block_list(x, value = block_list(value), location = location)

  x
}

#' @export
#' @describeIn ph_with add an [empty_content()] to a new shape
#' on the current slide.
ph_with.empty_content <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)
  new_ph <- shape_properties_tags(
    left = location$left,
    top = location$top,
    width = location$width,
    height = location$height,
    label = location$ph_label,
    ph = location$ph,
    rot = location$rotation,
    bg = location$bg,
    ln = location$ln,
    geom = location$geom
  )

  if (is.na(location$fld_id)) {
    xml_elt <- paste0(psp_ns_yes, new_ph, "</p:sp>")
  } else {
    pars <- paste0(
      "<a:p><a:fld id=\"",
      location$fld_id,
      "\" type = \"",
      location$fld_type,
      "\"><a:rPr/><a:t>",
      x$cursor,
      "</a:t></a:fld></a:p>",
      collapse = ""
    )
    xml_elt <- sprintf(paste0(
      psp_ns_yes,
      new_ph,
      "<p:txBody><a:bodyPr/><a:lstStyle/>",
      pars,
      "</p:txBody></p:sp>"
    ))
  }
  node <- as_xml_document(xml_elt)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), node)
  x
}


xml_to_slide <- function(slide, location, value, package_dir) {
  node <- xml_find_first(value, as_xpath_content_sel("//"))
  if (xml_name(node) == "grpSp") {
    node_sppr <- xml_child(node, "p:grpSpPr")
    node_xfrm <- xml_child(node, "p:grpSpPr/a:xfrm")
    node_name <- xml_child(node, "p:nvGrpSpPr/p:cNvPr")
  } else if (xml_name(node) == "graphicFrame") {
    node_xfrm <- xml_child(node, "p:xfrm")
    node_name <- xml_child(node, "p:nvGraphicFramePr/p:cNvPr")
    node_sppr <- xml_missing()
  } else if (xml_name(node) == "pic") {
    node_sppr <- xml_child(node, "p:spPr")
    node_xfrm <- xml_child(node, "p:spPr/a:xfrm")
    node_name <- xml_child(node, "p:nvPicPr/p:cNvPr")
  } else if (xml_name(node) == "sp") {
    node_sppr <- xml_child(node, "p:spPr")
    node_xfrm <- xml_child(node, "p:spPr/a:xfrm")
    node_name <- xml_child(node, "p:nvSpPr/p:cNvPr")
  } else {
    node_xfrm <- xml_missing()
    node_name <- xml_missing()
    node_sppr <- xml_missing()
  }

  if (!inherits(node_name, "xml_missing")) {
    xml_attr(node_name, "name") <- location$ph_label
  }

  if (!inherits(node_xfrm, "xml_missing")) {
    off <- xml_child(node_xfrm, "a:off")
    ext <- xml_child(node_xfrm, "a:ext")
    chOff <- xml_child(node_xfrm, "a:chOff")
    chExt <- xml_child(node_xfrm, "a:chExt")
    xml_attr(off, "x") <- sprintf("%.0f", location$left * 914400)
    xml_attr(off, "y") <- sprintf("%.0f", location$top * 914400)
    xml_attr(ext, "cx") <- sprintf("%.0f", location$width * 914400)
    xml_attr(ext, "cy") <- sprintf("%.0f", location$height * 914400)
    xml_attr(chOff, "x") <- sprintf("%.0f", location$left * 914400)
    xml_attr(chOff, "y") <- sprintf("%.0f", location$top * 914400)
    xml_attr(chExt, "cx") <- sprintf("%.0f", location$width * 914400)
    xml_attr(chExt, "cy") <- sprintf("%.0f", location$height * 914400)

    # add location$rotation to cNvPr:name
    if (
      !is.null(location$rotation) &&
        is.finite(location$rotation) &&
        is.numeric(location$rotation)
    ) {
      xml_attr(node_xfrm, "rot") <- sprintf("%.0f", -location$rotation * 60000)
    }
  }

  if (!inherits(node_sppr, "xml_missing")) {
    # add location$bg to SpPr
    if (!is.null(location$bg)) {
      bg_str <- solid_fill_pml(location$bg)
      xml_add_child(node_sppr, as_xml_document(bg_str))
    }
  }
  value
}

#' @export
#' @describeIn ph_with add an xml_document object to a new shape on the
#' current slide. This function is to be used to add custom openxml code.
ph_with.xml_document <- function(x, value, location, ...) {
  slide <- x$slide$get_slide(x$cursor)

  location <- fortify_location(location, doc = x)

  xml_to_slide(slide, location, value, x$package_dir)

  xml_add_child(xml_find_first(slide$get(), "//p:spTree"), value)
  x
}


# PHS_WITH -------

#' @title Fill multiple placeholders using key value syntax
#' @description A sibling of [officer::ph_with] that fills mutiple placeholders at once. Placeholder locations are
#'   specfied using the short form syntax. The location and corresponding object are passed as key value pairs
#'   (`phs_with("short form location" = object)`). Under the hood, [officer::ph_with] is called for each pair. Note
#'   that `phs_with` does not cover all options from the `ph_location_*` family and is also less customization. It is a
#'   covenience wrapper for the most common use cases. The implemented short forms are listed in section
#'   `"Short forms"`.
#' @param x A `rpptx` object.
#' @param ... Key-value pairs of the form `"short form location" = object`. If the short form is an integer or a string
#'   with blanks, you must wrap it in quotes or backticks.
#' @param .dots List of key-value pairs `"short form location" = object`. Alternative to `...`.
#' @param .slide_idx Numeric indexes of slides to process. `NULL` (default) processes the current slide only. Use
#'   keyword `all` for all slides.
#' @section Short forms: The following short forms are implemented and can be used as the parameter in the function
#'   call. The corresponding function from the `ph_location_*` family (called under the hood) is displayed on the
#'   right.
#'
#' | **Short form** | **Description**                                   | **Location function**           |
#' |----------------|---------------------------------------------------|---------------------------------|
#' | `"left"`       | Keyword string                                    | `ph_location_left()`            |
#' | `"right"`      | Keyword string                                    | `ph_location_right()`           |
#' | `"fullsize"`   | Keyword string                                    | `ph_location_fullsize()`        |
#' | `"body [1]"`   | String: type + index in brackets (`1` if omitted) | `ph_location_type("body", 1)`   |
#' | `"my_label"`   | Any string not matching a keyword or type         | `ph_location_label("my_label")` |
#' | `1`            | Length 1 integer                                  | `ph_location_id(1)`             |
#'
#' @export
#' @seealso [ph_with()], [add_slide()]
#' @example inst/examples/example_phs_with.R
phs_with <- function(x, ..., .dots = NULL, .slide_idx = NULL) {
  dots_list <- list(...)
  if (length(dots_list) > 0 && !is_named(dots_list)) {
    cli::cli_abort(
      c(
        "Missing key in {.arg ...}",
        "x" = "{.arg ...} requires a key-value syntax, with a ph short-form location as the key",
        "i" = "Example: {.code phs_with(x, title = 'My title', 'body[1]' = 'My body')}"
      )
    )
  }
  if (length(.dots) > 0 && !is_named(.dots)) {
    cli::cli_abort(
      c(
        "Missing names in {.arg .dots}",
        "x" = "{.arg .dots} must be a named list, with ph short-form locations as names",
        "i" = "Example: {.code phs_with(x, .dots = list(title = 'My title', 'body[1]' = 'My body'))}"
      )
    )
  }
  dots <- utils::modifyList(dots_list, .dots %||% list())
  if (length(dots) == 0) {
    return(x)
  }
  .slide_idx <- .slide_idx %||% x$cursor # default is current slide
  if (is.character(.slide_idx) && .slide_idx == "all") {
    .slide_idx <- seq_len(length(x))
  }
  stop_if_not_in_slide_range(x, .slide_idx)

  loc_strings <- as.list(names(dots))
  ii <- grepl("^\\d+$", loc_strings) # find integer short-forms
  loc_strings[ii] <- as.integer(loc_strings[ii])
  locations <- lapply(loc_strings, resolve_location)

  .old_cursor <- x$cursor
  for (slide_idx in .slide_idx) {
    x$cursor <- slide_idx # pw_with always uses the current slide
    for (i in seq_along(dots)) {
      x <- ph_with(x, dots[[i]], locations[[i]])
    }
  }
  x$cursor <- .old_cursor
  x
}

test_that("opts_current_table works as expected", {
  skip_if_not_installed("knitr")
  skip_if_not_installed("rmarkdown")
  skip_if_not(rmarkdown::pandoc_available())

  new_dir <- tempfile()
  dir.create(new_dir, showWarnings = FALSE)
  rmd_fp <- file.path(new_dir, "knitr-utils.Rmd")
  generated_rds <- tempfile(fileext = ".RDS")
  file.copy("docs_dir/knitr-utils.Rmd", to = rmd_fp)
  rmarkdown::render(
    input = rmd_fp,
    envir = parent.frame(),
    params = list(
      rds_path = generated_rds
    ),
    quiet = TRUE
  )
  x <- readRDS(generated_rds)
  expect_equal(x$cap.pre, "Tableau ")
  expect_equal(x$cap.style, "Normal")
  expect_equal(x$cap.sep, " : ")
  expect_equal(x$alt.title, "alt title")
  expect_equal(x$alt.description, "alt description")
  expect_equal(x$cap, "Coucou")
})

---
title: "Untitled"
output: md_document
params:
  rds_path: temp.RDS
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = FALSE)
```

```{r tab.topcaption=FALSE, tab.cap.pre = "Tableau", tab.cap.sep = " :", tab.cap = "Coucou", label="plop", tab.cap.style = "Normal", tab.alt.title = "alt title", tab.alt.description = "alt description"}
saveRDS(
  object = officer::opts_current_table(),
  params$rds_path
)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{sheet_select}
\alias{sheet_select}
\title{Select sheet}
\usage{
sheet_select(x, sheet)
}
\arguments{
\item{x}{rxlsx object}

\item{sheet}{sheet name}
}
\description{
Set a particular sheet selected when workbook will be
edited.
}
\examples{
my_ws <- read_xlsx()
my_pres <- add_sheet(my_ws, label = "new sheet")
my_pres <- sheet_select(my_ws, sheet = "new sheet")
print(my_ws, target = tempfile(fileext = ".xlsx") )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_gg}
\alias{body_add_gg}
\title{Add a 'ggplot' in a 'Word' document}
\usage{
body_add_gg(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  pos = "after",
  ...
)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{ggplot object}

\item{width}{height in inches}

\item{height}{height in inches}

\item{res}{resolution of the png image in ppi}

\item{style}{paragraph style}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}

\item{...}{Arguments to be passed to png function.}
}
\description{
add a ggplot as a png image into an rdocx object.
}
\examples{
if (require("ggplot2")) {
  doc <- read_docx()

  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))

  if (capabilities(what = "png")) {
    doc <- body_add_gg(doc, value = gg_plot, style = "centered")
  }

  print(doc, target = tempfile(fileext = ".docx"))
}
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pack_folder.R
\name{unpack_folder}
\alias{unpack_folder}
\title{Extract files from a zip file}
\usage{
unpack_folder(file, folder)
}
\arguments{
\item{file}{path of the archive to unzip}

\item{folder}{folder to create}
}
\description{
Extract files from a zip file to a folder. The
function returns the complete path to destination folder.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_fullsize}
\alias{ph_location_fullsize}
\title{Location of a full size element}
\usage{
ph_location_fullsize(newlabel = "", ...)
}
\arguments{
\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will return the location corresponding
to a full size display.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, "Hello world", location = ph_location_fullsize())
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_text}
\alias{fp_text}
\alias{fp_text_lite}
\alias{format.fp_text}
\alias{print.fp_text}
\alias{update.fp_text}
\title{Text formatting properties}
\usage{
fp_text(
  color = "black",
  font.size = 10,
  bold = FALSE,
  italic = FALSE,
  underlined = FALSE,
  strike = FALSE,
  font.family = "Arial",
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  vertical.align = "baseline",
  shading.color = "transparent"
)

fp_text_lite(
  color = NA,
  font.size = NA,
  font.family = NA,
  cs.family = NA,
  eastasia.family = NA,
  hansi.family = NA,
  bold = NA,
  italic = NA,
  underlined = NA,
  strike = NA,
  vertical.align = "baseline",
  shading.color = NA
)

\method{format}{fp_text}(x, type = "wml", ...)

\method{print}{fp_text}(x, ...)

\method{update}{fp_text}(
  object,
  color,
  font.size,
  bold,
  italic,
  underlined,
  strike,
  font.family,
  cs.family,
  eastasia.family,
  hansi.family,
  vertical.align,
  shading.color,
  ...
)
}
\arguments{
\item{color}{font color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{font.size}{font size (in point) - 0 or positive integer value.}

\item{bold}{is bold}

\item{italic}{is italic}

\item{underlined}{is underlined}

\item{strike}{is strikethrough}

\item{font.family}{single character value. Specifies the font to
be used to format characters in the Unicode range (U+0000-U+007F).}

\item{cs.family}{optional font to be used to format
characters in a complex script Unicode range. For example, Arabic
text might be displayed using the "Arial Unicode MS" font.}

\item{eastasia.family}{optional font to be used to
format characters in an East Asian Unicode range. For example,
Japanese text might be displayed using the "MS Mincho" font.}

\item{hansi.family}{optional. Specifies the font to be used to format
characters in a Unicode range which does not fall into one of the
other categories.}

\item{vertical.align}{single character value specifying font vertical alignments.
Expected value is one of the following : default \code{'baseline'}
or \code{'subscript'} or \code{'superscript'}}

\item{shading.color}{shading color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{x}{\code{fp_text} object}

\item{type}{output type - one of 'wml', 'pml', 'html', 'rtf'.}

\item{...}{further arguments - not used}

\item{object}{\code{fp_text} object to modify}

\item{format}{format type, wml for MS word, pml for
MS PowerPoint and html.}
}
\value{
a \code{fp_text} object
}
\description{
Create an \code{fp_text} object that describes
text formatting properties.

Function \code{fp_text_lite()} is generating properties
with only entries for the parameters users provided. The
undefined properties will inherit from the default settings.
}
\examples{
fp_text()
fp_text(color = "red")
fp_text(bold = TRUE, shading.color = "yellow")
print(fp_text(color = "red", font.size = 12))
}
\seealso{
\code{\link[=ftext]{ftext()}}, \code{\link[=fpar]{fpar()}}

Other functions for defining formatting properties: 
\code{\link{fp_border}()},
\code{\link{fp_cell}()},
\code{\link{fp_par}()},
\code{\link{fp_tab}()},
\code{\link{fp_tabs}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_fullsize}
\alias{ph_location_fullsize}
\title{Location of a full size element}
\usage{
ph_location_fullsize(newlabel = "", ...)
}
\arguments{
\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will return the location corresponding
to a full size display.
}
\examples{
doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(doc, "Hello world", location = ph_location_fullsize() )
print(doc, target = tempfile(fileext = ".pptx") )
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_write.R
\name{print.rdocx}
\alias{print.rdocx}
\title{Write a 'Word' File}
\usage{
\method{print}{rdocx}(
  x,
  target = NULL,
  copy_header_refs = FALSE,
  copy_footer_refs = FALSE,
  preview = FALSE,
  ...
)
}
\arguments{
\item{x}{an \code{rdocx} object created with \code{\link[=read_docx]{read_docx()}}}

\item{target}{path to the \code{.docx} file to write. The file will be created
or overwritten if it already exists. If \code{NULL} and \code{preview = FALSE}, the
function returns \code{NULL} without writing a file.}

\item{copy_header_refs, copy_footer_refs}{logical, default is FALSE.
If TRUE, copy the references to the header and footer in each section
of the body of the document. This parameter is experimental and may change
in a future version.}

\item{preview}{Save \code{x} to a temporary file and open it (default \code{FALSE}).
When \code{TRUE}, the document is saved to a temporary location and opened with
the system's default application for \code{.docx} files, useful for quick previewing
during development.}

\item{...}{unused}
}
\value{
The full path to the created \code{.docx} file (invisibly). This allows
chaining operations or capturing the output path for further use.
}
\description{
\code{print.rdocx()} is the essential output function for creating Word files
with officer. It takes an \code{rdocx} object (created with \code{\link[=read_docx]{read_docx()}} and
populated with content) and writes it to disk as a \code{.docx} file.

This function performs all necessary post-processing operations before
writing the file.

The function is typically called at the end of your document creation
workflow, after all content has been added with \verb{body_add_*()} functions.
}
\examples{
library(officer)

# This example demonstrates how to create
# an small document -----

## Create a new Word document
doc <- read_docx()
doc <- body_add_par(doc, "hello world")
## Save the document
output_file <- print(doc, target = tempfile(fileext = ".docx"))

# preview mode: save to temp file and open locally ----
## Not run:
# print(doc, preview = TRUE)
}
\seealso{
Create a 'Word' document object with \code{\link[=read_docx]{read_docx()}}, add content with
functions \code{\link[=body_add_par]{body_add_par()}}, \code{\link[=body_add_plot]{body_add_plot()}},
\code{\link[=body_add_table]{body_add_table()}}, change settings with \code{\link[=docx_set_settings]{docx_set_settings()}}, set
properties with \code{\link[=set_doc_properties]{set_doc_properties()}},  read 'Word' styles with
\code{\link[=styles_info]{styles_info()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{docx_body_relationship}
\alias{docx_body_relationship}
\title{Body xml document}
\usage{
docx_body_relationship(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
Get the body document as xml. This function
is not to be used by end users, it has been implemented
to allow other packages to work with officer.
}
\examples{
doc <- read_docx()
docx_body_relationship(doc)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_text}
\alias{fp_text}
\alias{fp_text_lite}
\alias{format.fp_text}
\alias{print.fp_text}
\alias{update.fp_text}
\title{Text formatting properties}
\usage{
fp_text(
  color = "black",
  font.size = 10,
  bold = FALSE,
  italic = FALSE,
  underlined = FALSE,
  font.family = "Arial",
  cs.family = NULL,
  eastasia.family = NULL,
  hansi.family = NULL,
  vertical.align = "baseline",
  shading.color = "transparent"
)

fp_text_lite(
  color = NA,
  font.size = NA,
  font.family = NA,
  cs.family = NA,
  eastasia.family = NA,
  hansi.family = NA,
  bold = NA,
  italic = NA,
  underlined = NA,
  vertical.align = "baseline",
  shading.color = NA
)

\method{format}{fp_text}(x, type = "wml", ...)

\method{print}{fp_text}(x, ...)

\method{update}{fp_text}(
  object,
  color,
  font.size,
  bold,
  italic,
  underlined,
  font.family,
  cs.family,
  eastasia.family,
  hansi.family,
  vertical.align,
  shading.color,
  ...
)
}
\arguments{
\item{color}{font color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{font.size}{font size (in point) - 0 or positive integer value.}

\item{bold}{is bold}

\item{italic}{is italic}

\item{underlined}{is underlined}

\item{font.family}{single character value. Specifies the font to
be used to format characters in the Unicode range (U+0000-U+007F).}

\item{cs.family}{optional font to be used to format
characters in a complex script Unicode range. For example, Arabic
text might be displayed using the "Arial Unicode MS" font.}

\item{eastasia.family}{optional font to be used to
format characters in an East Asian Unicode range. For example,
Japanese text might be displayed using the "MS Mincho" font.}

\item{hansi.family}{optional. Specifies the font to be used to format
characters in a Unicode range which does not fall into one of the
other categories.}

\item{vertical.align}{single character value specifying font vertical alignments.
Expected value is one of the following : default \code{'baseline'}
or \code{'subscript'} or \code{'superscript'}}

\item{shading.color}{shading color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{x}{\code{fp_text} object}

\item{type}{output type - one of 'wml', 'pml', 'html', 'rtf'.}

\item{...}{further arguments - not used}

\item{object}{\code{fp_text} object to modify}

\item{format}{format type, wml for MS word, pml for
MS PowerPoint and html.}
}
\value{
a \code{fp_text} object
}
\description{
Create a \code{fp_text} object that describes
text formatting properties.

Function \code{fp_text_lite()} is generating properties
with only entries for the parameters users provided. The
undefined properties will inherit from the default settings.
}
\examples{
fp_text()
fp_text(color = "red")
fp_text(bold = TRUE, shading.color = "yellow")
print(fp_text(color = "red", font.size = 12))
}
\seealso{
\link{ftext}, \link{fpar}

Other functions for defining formatting properties: 
\code{\link{fp_border}()},
\code{\link{fp_cell}()},
\code{\link{fp_par}()},
\code{\link{fp_tab}()},
\code{\link{fp_tabs}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{rtf_add}
\alias{rtf_add}
\alias{rtf_add.block_section}
\alias{rtf_add.character}
\alias{rtf_add.factor}
\alias{rtf_add.double}
\alias{rtf_add.fpar}
\alias{rtf_add.block_list}
\alias{rtf_add.block_toc}
\alias{rtf_add.gg}
\alias{rtf_add.plot_instr}
\title{Add content into an RTF document}
\usage{
rtf_add(x, value, ...)

\method{rtf_add}{block_section}(x, value, ...)

\method{rtf_add}{character}(x, value, style = NULL, ...)

\method{rtf_add}{factor}(x, value, style = NULL, ...)

\method{rtf_add}{double}(x, value, formatter = formatC, style = NULL, ...)

\method{rtf_add}{fpar}(x, value, style = NULL, ...)

\method{rtf_add}{block_list}(x, value, style = NULL, ...)

\method{rtf_add}{block_toc}(x, value, ...)

\method{rtf_add}{gg}(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  ppr = fp_par(text.align = "center"),
  ...
)

\method{rtf_add}{plot_instr}(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  ppr = fp_par(text.align = "center"),
  ...
)
}
\arguments{
\item{x}{rtf object, created by \code{\link[=rtf_doc]{rtf_doc()}}.}

\item{value}{object to add in the document. Supported objects
are vectors, graphics, block of formatted paragraphs. Use package
'flextable' to add tables.}

\item{...}{further arguments passed to or from other methods. When
adding a \code{ggplot} object or \link{plot_instr}, these arguments will be used
by png function. See section 'Methods' to see what arguments can be used.}

\item{style}{style identifier (\code{style_id}) of a paragraph style registered
on the document. Defaults to \code{NULL} (use the document's normal style). See
\code{\link[=rtf_set_paragraph_style]{rtf_set_paragraph_style()}} and \code{\link[=rtf_styles_info]{rtf_styles_info()}}.}

\item{formatter}{function used to format the numerical values}

\item{width}{height in inches}

\item{height}{height in inches}

\item{res}{resolution of the png image in ppi}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{ppr}{\code{\link[=fp_par]{fp_par()}} to apply to paragraph.}
}
\description{
This function add 'officer' objects into an RTF document.
Values are added as new paragraphs. See section 'Methods (by class)'
that list supported objects.
}
\section{Methods (by class)}{
\itemize{
\item \code{rtf_add(block_section)}: add a new section definition

\item \code{rtf_add(character)}: add characters as new paragraphs

\item \code{rtf_add(factor)}: add a factor vector as new paragraphs

\item \code{rtf_add(double)}: add a double vector as new paragraphs

\item \code{rtf_add(fpar)}: add an \code{\link[=fpar]{fpar()}}

\item \code{rtf_add(block_list)}: add an \code{\link[=block_list]{block_list()}}

\item \code{rtf_add(block_toc)}: add a \code{\link[=block_toc]{block_toc()}} (table of contents).
Word populates the TOC at open time from paragraphs styled with the
built-in heading styles (which carry an outline level). LibreOffice
will not render the TOC automatically.

\item \code{rtf_add(gg)}: add a ggplot2

\item \code{rtf_add(plot_instr)}: add a \code{\link[=plot_instr]{plot_instr()}} object

}}
\section{Section model in RTF}{

A \code{block_section} added with \code{rtf_add()} applies to the content that
\strong{follows} the call: it opens a new section whose layout (orientation,
columns, margins, headers / footers) is inherited by every paragraph,
table or graphic added afterwards, until the next \code{block_section} (or
the end of the document).

Typical pattern: declare the section, then add the content.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{doc <- rtf_doc() |>
  rtf_add(block_section(prop_section(
    page_size = page_size(orient = "landscape")
  ))) |>
  rtf_add(fpar("This paragraph is in landscape orientation."))
}\if{html}{\out{</div>}}

The first section of the document is configured via the \code{def_sec}
argument of \code{\link[=rtf_doc]{rtf_doc()}} (a \link{prop_section} object, not a
\code{block_section}). It applies to every element added before the first
\code{rtf_add(block_section(...))} call.

The Word output uses the opposite model: \code{body_end_block_section()}
applies to the content that \emph{precedes} the call. See
\code{\link[=body_end_block_section]{body_end_block_section()}}.
}

\examples{
## Simple RTF example ----

library(officer)

def_text <- fp_text_lite(color = "#006699", bold = TRUE)
center_par <- fp_par(text.align = "center", padding = 3)

doc <- rtf_doc(
  normal_par = fp_par(line_spacing = 1.4, padding = 3)
)

doc <- rtf_add(
  x = doc,
  value = fpar(
    ftext("how are you?", prop = def_text),
    fp_p = fp_par(text.align = "center")
  )
)

a_paragraph <- fpar(
  ftext("Here is a date: ", prop = def_text),
  run_word_field(field = "Date \\\\@ \"MMMM d yyyy\""),
  fp_p = center_par
)
doc <- rtf_add(
  x = doc,
  value = block_list(
    a_paragraph,
    a_paragraph,
    a_paragraph
  )
)

if (require("ggplot2")) {
  gg <- gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))
  doc <- rtf_add(doc, gg, width = 3, height = 4, ppr = center_par)
}
anyplot <- plot_instr(code = {
  barplot(1:5, col = 2:6)
})
doc <- rtf_add(doc, anyplot, width = 5, height = 4, ppr = center_par)

print(doc, target = tempfile(fileext = ".rtf"))


## RTF example with sections ----

library(officer)

quick_section_header <- function(label) {
  block_list(
    fpar(
      ftext(label, fp_text_lite(bold = TRUE, color = "#006699"))
    )
  )
}

quick_section_footer <- function(label) {
  block_list(
    fpar(
      "Page ",
      run_word_field(field = "PAGE  \\\\* MERGEFORMAT")
    )
  )
}

quick_hello_world <- function(doc) {
  rtf_add(
    doc,
    fpar(
      ftext(
        "Hello World"
      ),
      fp_p = fp_par(text.align = "left")
    )
  )
}

three_cols_section <- block_section(
  prop_section(
    type = "continuous",
    section_columns = section_columns(widths = c(1.7, 1.7, 1.7), space = 0.25),
    header_default = quick_section_header("Three columns section"),
    footer_default = quick_section_footer("Three columns section")
  )
)

doc <- rtf_doc(
  def_sec = prop_section(
    header_default = quick_section_header("Default section"),
    footer_default = quick_section_footer("Default section")
  ),
  normal_par = fp_par(padding = 3)
)
doc <- rtf_add(doc, block_toc(level = 3))

doc <- rtf_add(doc, "Sections demo", style = "heading 1")
doc <- rtf_add(doc, "Default section", style = "heading 2")
doc <- quick_hello_world(doc)

if (require("ggplot2")) {
  gg_iris <- ggplot(iris, aes(Sepal.Length, Petal.Length, colour = Species)) +
    geom_point() +
    theme_minimal()
  doc <- rtf_add(doc, gg_iris, width = 4, height = 3)
}

doc <- rtf_add(doc, "Three columns", style = "heading 2")

doc <- rtf_add(doc, three_cols_section)

titles_list <- block_list(
  fpar(ftext("Left Title"), fp_p = fp_par(text.align = "left")),
  fpar(
    run_columnbreak(),
    ftext("Centered Title"),
    fp_p = fp_par(text.align = "center")
  ),
  fpar(
    run_columnbreak(),
    ftext("Right Title"),
    fp_p = fp_par(text.align = "right")
  )
)
doc <- rtf_add(doc, titles_list)
doc <- rtf_add(doc, block_section(prop_section()))
doc <- rtf_add(doc, fpar(run_linebreak()))
doc <- quick_hello_world(doc)

landscape_section <- block_section(prop_section(
  type = "nextPage",
  page_size = page_size(orient = "landscape"),
  header_default = quick_section_header("Landscape section"),
  footer_default = quick_section_footer("Landscape section")
))
doc <- rtf_add(doc, landscape_section)
doc <- rtf_add(doc, "Landscape orientation", style = "heading 2")
doc <- quick_hello_world(doc)

doc <- rtf_add(
  doc,
  block_section(
    prop_section(
      type = "nextPage",
      header_default = quick_section_header("Final section"),
      footer_default = quick_section_footer("Final section")
    )
  )
)
doc <- rtf_add(doc, "Back to portrait", style = "heading 2")
doc <- quick_hello_world(doc)
print(doc, target = tempfile(fileext = ".rtf"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_layout_helper.R
\name{layout_default}
\alias{layout_default}
\title{Default layout for new slides}
\usage{
layout_default(x, layout = NULL, master = NULL, as_list = FALSE)
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{layout}{Layout name. If \code{NULL} (default), removes the default layout.}

\item{master}{Name of master. Only required if layout name is not unique across masters.}

\item{as_list}{If \code{TRUE}, return a list with layout and master instead of the \code{rpptx} object.}
}
\value{
The \code{rpptx} object.
}
\description{
Set or remove the default layout used when calling \code{add_slide()}.
}
\examples{
# set and remove the default layout
x <- read_pptx()
layout_default(x) # no defaults
x <- layout_default(x, "Title and Content") # set default
layout_default(x)
x <- add_slide(x) # new slide with default layout
x <- layout_default(x, NULL) # remove default
layout_default(x) # no defaults

# use when repeatedly adding slides with same layout
x <- read_pptx()
x <- layout_default(x, "Title and Content")
x <- add_slide(x, title = "1. Slide", body = "Some content")
x <- add_slide(x, title = "2. Slide", body = "Some more content")
x <- add_slide(x, title = "3. Slide", body = "Even more content")
}
\seealso{
\code{\link[=add_slide]{add_slide()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defunct.R
\name{officer-defunct}
\alias{officer-defunct}
\alias{docx_reference_img}
\title{Defunct Functions in Package officer}
\usage{
docx_reference_img(...)
}
\arguments{
\item{...}{unused arguments}
}
\description{
Defunct Functions in Package officer
}
\details{
\code{docx_reference_img()} is not usefull anymore.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_right}
\alias{ph_location_right}
\title{Location of a right body element}
\usage{
ph_location_right(newlabel = NULL, ...)
}
\arguments{
\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will return the location corresponding
to a right bounding box. The function assume the layout 'Two Content'
is existing. This is an helper function, if you don't have a layout
named 'Two Content', use \code{\link[=ph_location_type]{ph_location_type()}} and set arguments
to your specific needs.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, "Hello left", location = ph_location_left())
doc <- ph_with(doc, "Hello right", location = ph_location_right())
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location:
\code{\link[=ph_location]{ph_location()}},
\code{\link[=ph_location_fullsize]{ph_location_fullsize()}},
\code{\link[=ph_location_id]{ph_location_id()}},
\code{\link[=ph_location_label]{ph_location_label()}},
\code{\link[=ph_location_left]{ph_location_left()}},
\code{\link[=ph_location_template]{ph_location_template()}},
\code{\link[=ph_location_type]{ph_location_type()}}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_list}
\alias{block_list}
\title{List of blocks}
\usage{
block_list(...)
}
\arguments{
\item{...}{a list of blocks. When output is only for
Word, objects of class \code{\link[=external_img]{external_img()}} can
also be used in fpar construction to mix text and images
in a single paragraph. Supported objects are:
\code{\link[=block_caption]{block_caption()}}, \code{\link[=block_pour_docx]{block_pour_docx()}}, \code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}}, \code{\link[=block_toc]{block_toc()}}, \code{\link[=fpar]{fpar()}}, \code{\link[=plot_instr]{plot_instr()}}.}
}
\description{
A list of blocks can be used to gather
several blocks (paragraphs, tables, ...) into a single
object. The result can be added into a Word document or a
PowerPoint presentation.
}
\examples{
# block list ------

img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)


## This can be only be used in a MS word output as pptx does
## not support paragraphs made of text and images.
## (actually it can be used but image will not appear in the
## pptx output)
value <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(ftext("hello", fpt_blue_bold), " ",
       ftext("world", fpt_red_italic)),
  fpar(
    ftext("hello world", fpt_red_italic),
          external_img(
            src = img.file, height = 1.06, width = 1.39)))
value

doc <- read_docx()
doc <- body_add(doc, value)
print(doc, target = tempfile(fileext = ".docx"))


value <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(ftext("hello", fpt_blue_bold), " ",
       ftext("world", fpt_red_italic)),
  fpar(
    ftext("blah blah blah", fpt_red_italic)))
value

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, value, location = ph_location_type(type = "body"))
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=body_add_blocks]{body_add_blocks()}}, \code{\link[=fpar]{fpar()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_gg}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{styles_info}
\alias{styles_info}
\title{Read 'Word' styles}
\usage{
styles_info(
  x,
  type = c("paragraph", "character", "table", "numbering"),
  is_default = c(TRUE, FALSE)
)
}
\arguments{
\item{x}{an rdocx object}

\item{type, is_default}{subsets for types (i.e. paragraph) and
default style (when \code{is_default} is TRUE or FALSE)}
}
\description{
read Word styles and get results in
a data.frame.
}
\examples{
x <- read_docx()
styles_info(x)
styles_info(x, type = "paragraph", is_default = TRUE)
}
\seealso{
Other functions for Word document informations: 
\code{\link{doc_properties}()},
\code{\link{docx_bookmarks}()},
\code{\link{docx_dim}()},
\code{\link{length.rdocx}()},
\code{\link{set_doc_properties}()}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_type}
\alias{ph_location_type}
\title{Location of a placeholder based on a type}
\usage{
ph_location_type(
  type = "body",
  type_idx = NULL,
  position_right = TRUE,
  position_top = TRUE,
  newlabel = NULL,
  id = NULL,
  ...
)
}
\arguments{
\item{type}{placeholder type to look for in the slide layout, one
of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.}

\item{type_idx}{Type index of the placeholder. If there is more than one
placeholder of a type (e.g., \code{body}), the type index can be supplied to uniquely
identify a ph. The index is a running number starting at 1. It is assigned by
placeholder position  (top -> bottom, left -> right). See \code{\link[=plot_layout_properties]{plot_layout_properties()}}
for details. If \code{idx} argument is used, \code{position_right} and \code{position_top}
are ignored.}

\item{position_right}{the parameter is used when a selection with above
parameters does not provide a unique position (for example
layout 'Two Content' contains two element of type 'body').
If \code{TRUE}, the element the most on the right side will be selected,
otherwise the element the most on the left side will be selected.}

\item{position_top}{same than \code{position_right} but applied
to top versus bottom.}

\item{newlabel}{a label to associate with the placeholder.}

\item{id}{(\strong{DEPRECATED, use \code{type_idx} instead}) Index of the placeholder.
If two body placeholder, there can be two different index: 1 and 2 for the
first and second body placeholders defined in the layout. If this argument
is used, \code{position_right} and \code{position_top} will be ignored.}

\item{...}{unused arguments}
}
\description{
The function will use the type name of the placeholder (e.g. body, title),
the layout name and few other criterias to find the corresponding location.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

# ph_location_type demo ----

loc_title <- ph_location_type(type = "title")
loc_footer <- ph_location_type(type = "ftr")
loc_dt <- ph_location_type(type = "dt")
loc_slidenum <- ph_location_type(type = "sldNum")
loc_body <- ph_location_type(type = "body")


doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(x = doc, "Un titre", location = loc_title)
doc <- ph_with(x = doc, "pied de page", location = loc_footer)
doc <- ph_with(x = doc, format(Sys.Date()), location = loc_dt)
doc <- ph_with(x = doc, "slide 1", location = loc_slidenum)
doc <- ph_with(x = doc, letters[1:10], location = loc_body)

loc_subtitle <- ph_location_type(type = "subTitle")
loc_ctrtitle <- ph_location_type(type = "ctrTitle")
doc <- add_slide(doc, layout = "Title Slide")
doc <- ph_with(x = doc, "Un sous titre", location = loc_subtitle)
doc <- ph_with(x = doc, "Un titre", location = loc_ctrtitle)

fileout <- tempfile(fileext = ".pptx")
print(doc, target = fileout)
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{unordered_list}
\alias{unordered_list}
\title{Unordered list}
\usage{
unordered_list(str_list = character(0), level_list = integer(0), style = NULL)
}
\arguments{
\item{str_list}{list of strings to be included in the object}

\item{level_list}{list of levels for hierarchy structure. Use
0 for 'no bullet', 1 for level 1, 2 for level 2 and so on.}

\item{style}{text style, a \code{fp_text} object list or a
single \code{fp_text} objects. Use \code{fp_text(font.size = 0, ...)} to
inherit from default sizes of the presentation.}
}
\description{
unordered list of text for PowerPoint
presentations. Each text is associated with
a hierarchy level.
}
\examples{
unordered_list(
  level_list = c(1, 2, 2, 3, 3, 1),
  str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
  style = fp_text(color = "red", font.size = 0)
)

unordered_list(
  level_list = c(1, 2, 1),
  str_list = c("Level1", "Level2", "Level1"),
  style = list(
    fp_text(color = "red", font.size = 0),
    fp_text(color = "pink", font.size = 0),
    fp_text(color = "orange", font.size = 0)
  )
)
}
\seealso{
\code{\link[=ph_with]{ph_with()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_gg}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_properties.R
\name{sp_lineend}
\alias{sp_lineend}
\alias{print.sp_lineend}
\alias{update.sp_lineend}
\title{Line end properties}
\usage{
sp_lineend(type = "none", width = "med", length = "med")

\method{print}{sp_lineend}(x, ...)

\method{update}{sp_lineend}(object, type, width, length, ...)
}
\arguments{
\item{type}{single character value specifying the line end type.
Expected value is one of the following : default \code{'none'}
or \code{'triangle'} or \code{'stealth'} or \code{'diamond'}
or \code{'oval'} or \code{'arrow'}}

\item{width}{single character value specifying the line end width
Expected value is one of the following : default \code{'sm'}
or \code{'med'} or \code{'lg'}}

\item{length}{single character value specifying the line end length
Expected value is one of the following : default \code{'sm'}
or \code{'med'} or \code{'lg'}}

\item{x, object}{\code{sp_lineend} object}

\item{...}{further arguments - not used}
}
\value{
a \code{sp_lineend} object
}
\description{
Create a \code{sp_lineend} object that describes
line end properties.
}
\examples{
library(officer)

sp_lineend()
sp_lineend(type = "triangle")
sp_lineend(type = "arrow", width = "lg", length = "lg")
print(sp_lineend(type = "triangle", width = "lg"))
obj <- sp_lineend(type = "triangle", width = "lg")
update(obj, type = "arrow")
}
\seealso{
\link{sp_line}

Other functions for defining shape properties: 
\code{\link{sp_line}()}
}
\concept{functions for defining shape properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{from_base64}
\alias{from_base64}
\title{Decode Base64 Vector to Character}
\usage{
from_base64(x)
}
\arguments{
\item{x}{A character vector of Base64-encoded strings. NA values are preserved.}
}
\value{
A character vector of decoded strings, same length as \code{x}.
}
\description{
Decodes one or more Base64-encoded elements back into their original character form.
}
\examples{
z <- as_base64(c("hello", "world"))
from_base64(z)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defunct.R
\name{officer-defunct}
\alias{officer-defunct}
\alias{slip_in_seqfield}
\alias{slip_in_column_break}
\alias{slip_in_xml}
\alias{slip_in_text}
\alias{slip_in_footnote}
\title{Defunct Functions in Package officer}
\usage{
slip_in_seqfield(...)

slip_in_column_break(...)

slip_in_xml(...)

slip_in_text(...)

slip_in_footnote(...)
}
\arguments{
\item{...}{unused arguments}
}
\description{
Defunct Functions in Package officer
}
\details{
\code{slip_in_seqfield()} is replaced by \code{run_word_field()}.

\code{slip_in_column_break()} is replaced by \code{run_columnbreak()}.

\code{slip_in_xml()} is replaced by \code{fpar()}.

\code{slip_in_text()} is replaced by \code{fpar()}.

\code{slip_in_footnote()} is replaced by \code{run_footnote()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_set_default_section}
\alias{body_set_default_section}
\title{Define Default Section}
\usage{
body_set_default_section(x, value)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{a \link{prop_section} object}
}
\description{
Define default section of the document. You can
define section propeerties (page size, orientation, ...) with a \link{prop_section} object.
}
\section{Illustrations}{


\if{html}{\figure{body_set_default_section_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
# Example 1: Setting page layout properties ----
# This example demonstrates how to configure the default section
# properties for page orientation, type, and margins

# Define custom section properties
# - Landscape orientation for wide tables
# - Continuous section (no page break)
# - Custom margins: top=1.5", bottom=0.75", left/right=2"
default_sect_properties <- prop_section(
  page_size = page_size(orient = "landscape"),
  type = "continuous",
  page_margins = page_mar(bottom = .75, top = 1.5, right = 2, left = 2)
)

# Create a new document
doc_1 <- read_docx()

# Add a wide table that benefits from landscape orientation
doc_1 <- body_add_par(doc_1, "Motor Trend Car Road Tests", style = "heading 1")
doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")

# Add some text content
doc_1 <- body_add_par(doc_1, "Sample Text Content", style = "heading 2")
doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 40), collapse = " "))

# Apply the section properties to the entire document
# This must be called at the end, after all content is added
doc_1 <- body_set_default_section(doc_1, default_sect_properties)

# Save the document
print(doc_1, target = tempfile(fileext = ".docx"))


# Example 2: Adding headers and footers ----
# This example shows how to create a document with:
# - A header containing the R logo
# - A footer with the current date and page numbers

# Get the path to R logo
img_path <- file.path(R.home("doc"), "html", "logo.jpg")

# Create header content with the R logo
# The logo is positioned on the right side with specific dimensions
header_content <- block_list(
  fpar(
    external_img(src = img_path, height = 0.5, width = 0.5),
    fp_p = fp_par(text.align = "right")
  )
)

# Create footer content with date and page numbers
# Format: "Document generated on: [Date] | Page [X]"
footer_content <- block_list(
  fpar(
    "Document generated on: ",
    run_word_field(field = "Date \\\\@ \"MMMM d, yyyy\""),
    " | Page ",
    run_word_field(field = "PAGE"),
    fp_p = fp_par(text.align = "center")
  )
)

# Define section properties that include header and footer
# The header and footer will appear on all pages
sect_with_hf <- prop_section(
  page_size = page_size(orient = "portrait", width = 8.3, height = 11.7),
  page_margins = page_mar(
    bottom = 1,
    top = 1,
    right = 1,
    left = 1,
    header = 0.5,
    footer = 0.5
  ),
  type = "continuous",
  header_default = header_content,
  footer_default = footer_content
)

# Create a new document with content
doc_2 <- read_docx()

# Add a title page
doc_2 <- body_add_par(doc_2, "Annual Report 2024", style = "heading 1")
doc_2 <- body_add_par(
  doc_2,
  "Company Performance Analysis",
  style = "heading 2"
)

# Add some sections with content
doc_2 <- body_add_par(doc_2, "Executive Summary", style = "heading 2")
doc_2 <- body_add_par(
  doc_2,
  "This report provides a comprehensive analysis of company performance metrics."
)

# Add a table
doc_2 <- body_add_par(doc_2, "Key Metrics", style = "heading 2")
summary_data <- data.frame(
  Metric = c("Revenue", "Profit", "Growth"),
  Q1 = c(1.2, 0.3, 12),
  Q2 = c(1.5, 0.4, 15),
  Q3 = c(1.8, 0.5, 18),
  Q4 = c(2.1, 0.6, 20)
)
doc_2 <- body_add_table(doc_2, value = summary_data, style = "table_template")

# Add a plot
doc_2 <- body_add_par(doc_2, "Revenue Trend", style = "heading 2")
revenue_plot <- plot_instr({
  quarters <- paste0("Q", 1:4)
  revenue <- c(1.2, 1.5, 1.8, 2.1)
  barplot(
    revenue,
    names.arg = quarters,
    col = "#4472C4",
    border = NA,
    main = "Quarterly Revenue (Millions)",
    ylab = "Revenue ($M)",
    xlab = "Quarter"
  )
})
doc_2 <- body_add_plot(doc_2, revenue_plot, width = 5, height = 4)

# Apply section properties with header and footer
# The header (with R logo) and footer (with date and page number)
# will appear on all pages
doc_2 <- body_set_default_section(doc_2, sect_with_hf)

# Save the document
output_file <- tempfile(fileext = ".docx")
print(doc_2, target = output_file)
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_block_section}()},
\code{\link{body_end_section_columns}()},
\code{\link{body_end_section_columns_landscape}()},
\code{\link{body_end_section_continuous}()},
\code{\link{body_end_section_landscape}()},
\code{\link{body_end_section_portrait}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_list}
\alias{block_list}
\title{List of blocks}
\usage{
block_list(...)
}
\arguments{
\item{...}{a list of blocks. When output is only for
Word, objects of class \code{\link[=external_img]{external_img()}} can
also be used in fpar construction to mix text and images
in a single paragraph. Supported objects are:
\code{\link[=block_caption]{block_caption()}}, \code{\link[=block_pour_docx]{block_pour_docx()}}, \code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}}, \code{\link[=block_toc]{block_toc()}}, \code{\link[=fpar]{fpar()}}, \code{\link[=plot_instr]{plot_instr()}}.}
}
\description{
A list of blocks can be used to gather
several blocks (paragraphs, tables, ...) into a single
object. The result can be added into a Word document or a
PowerPoint presentation.
}
\examples{
# block list ------

img.file <- file.path( R.home("doc"), "html", "logo.jpg" )
fpt_blue_bold <- fp_text(color = "#006699", bold = TRUE)
fpt_red_italic <- fp_text(color = "#C32900", italic = TRUE)


## This can be only be used in a MS word output as pptx does
## not support paragraphs made of text and images.
## (actually it can be used but image will not appear in the
## pptx output)
value <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(ftext("hello", fpt_blue_bold), " ",
       ftext("world", fpt_red_italic)),
  fpar(
    ftext("hello world", fpt_red_italic),
          external_img(
            src = img.file, height = 1.06, width = 1.39)))
value

doc <- read_docx()
doc <- body_add(doc, value)
print(doc, target = tempfile(fileext = ".docx"))


value <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(ftext("hello", fpt_blue_bold), " ",
       ftext("world", fpt_red_italic)),
  fpar(
    ftext("blah blah blah", fpt_red_italic)))
value

doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(doc, value, location = ph_location_type(type = "body"))
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=body_add_blocks]{body_add_blocks()}}, \code{\link[=fpar]{fpar()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{rtf_doc}
\alias{rtf_doc}
\title{Create an RTF document object}
\usage{
rtf_doc(
  def_sec = prop_section(),
  normal_par = fp_par(),
  normal_chunk = fp_text(font.family = "Arial", font.size = 11)
)
}
\arguments{
\item{def_sec}{a \link{prop_section} object used to define the default section
(page size, margins, header / footer, columns) applied to content added
before any explicit \code{\link[=block_section]{block_section()}}.}

\item{normal_par}{an object generated by \code{\link[=fp_par]{fp_par()}}}

\item{normal_chunk}{an object generated by \code{\link[=fp_text]{fp_text()}}}
}
\value{
an object of class \code{rtf} representing an
empty RTF document.
}
\description{
Creation of the object representing an
RTF document which can then receive contents with
the \code{\link[=rtf_add]{rtf_add()}} function and be written to a file with
the \code{print(x, target="doc.rtf")} function.
}
\section{Built-in styles}{

\code{rtf_doc()} registers four paragraph styles: \code{"Normal"} (built from
\code{normal_par} / \code{normal_chunk}) and \code{"heading 1"} / \code{"heading 2"} /
\code{"heading 3"} (bold, dark blue shades, with outline levels so the
paragraphs feed Word's navigation pane and TOC field). The heading
names mirror Word's own convention so a script that adds paragraphs
with \code{style = "heading 1"} works identically against Word output
(\code{body_add_par(..., style = "heading 1")}) and RTF output
(\code{rtf_add(..., style = "heading 1")}).

To override a built-in look (size, color, spacing, outline level),
call \code{\link[=rtf_set_paragraph_style]{rtf_set_paragraph_style()}} after \code{rtf_doc()} with the same
\code{style_id}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{doc <- rtf_doc()
doc <- rtf_set_paragraph_style(
  doc,
  style_id = "heading 1",
  fp_p = fp_par(text.align = "center", padding = 12),
  fp_t = fp_text_lite(bold = TRUE, font.size = 24, color = "#000000"),
  outline_level = 1L
)
}\if{html}{\out{</div>}}

Use \code{\link[=rtf_styles_info]{rtf_styles_info()}} to inspect the current style table.
}

\examples{
rtf_doc(normal_par = fp_par(padding = 3))
}
\seealso{
\code{\link[=read_docx]{read_docx()}}, \code{\link[=print.rtf]{print.rtf()}}, \code{\link[=rtf_add]{rtf_add()}}. See \code{?rtf_add} for a
complete multi-section example exercising \code{def_sec}, headers / footers
and several \code{block_section()} calls.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{base64_to_image}
\alias{base64_to_image}
\title{Convert Data URIs to PNG Files}
\usage{
base64_to_image(data_uri, output_files)
}
\arguments{
\item{data_uri}{Character, a data URI character vector starting with "data:image/png;base64,"}

\item{output_files}{Character, paths to the output PNG files}
}
\value{
Character, the paths to the created PNG files
}
\description{
Decodes base64-encoded data URIs and writes them to PNG files.
}
\examples{
rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
base64_str <- image_to_base64(rlogo)
base64_to_image(
  data_uri = base64_str,
  output_files = tempfile(fileext = ".jpeg")
)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{add_sheet}
\alias{add_sheet}
\title{Add a sheet}
\usage{
add_sheet(x, label)
}
\arguments{
\item{x}{rxlsx object}

\item{label}{sheet label}
}
\description{
Add a sheet into an xlsx worksheet.
}
\details{
\code{read_xlsx()} returns a workbook that already contains one default
sheet shipped with the template (named \code{"Sheet1"} or \code{"Feuil1"}
depending on the locale). \code{add_sheet()} is purely additive: the
default sheet is kept as-is. Remove it explicitly with
\code{\link[=sheet_remove]{sheet_remove()}} if it is not wanted.
}
\examples{
my_ws <- read_xlsx()
my_pres <- add_sheet(my_ws, label = "new sheet")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_fpar}
\alias{body_add_fpar}
\title{Add fpar in a 'Word' document}
\usage{
body_add_fpar(x, value, style = NULL, pos = "after")
}
\arguments{
\item{x}{a docx device}

\item{value}{a character}

\item{style}{paragraph style. If NULL, paragraph settings from \code{fpar} will be used. If not
NULL, it must be a paragraph style name (located in the template
provided as \code{read_docx(path = ...)}); in that case, paragraph settings from \code{fpar} will be
ignored.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add an \code{\link[=fpar]{fpar()}} (a formatted paragraph)
into an rdocx object.
}
\examples{
bold_face <- shortcuts$fp_bold(font.size = 30)
bold_redface <- update(bold_face, color = "red")
fpar_ <- fpar(
  ftext("Hello ", prop = bold_face),
  ftext("World", prop = bold_redface),
  ftext(", how are you?", prop = bold_face)
)
doc <- read_docx()
doc <- body_add_fpar(doc, fpar_)

print(doc, target = tempfile(fileext = ".docx"))

# a way of using fpar to center an image in a Word doc ----
rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
img_in_par <- fpar(
  external_img(src = rlogo, height = 1.06 / 2, width = 1.39 / 2),
  hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "cran", prop = bold_redface
  ),
  fp_p = fp_par(text.align = "center")
)

doc <- read_docx()
doc <- body_add_fpar(doc, img_in_par)
print(doc, target = tempfile(fileext = ".docx"))

}
\seealso{
\code{\link[=fpar]{fpar()}}

Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()},
\code{\link{body_append_start_context}()},
\code{\link{body_import_docx}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{body_replace_text_at_bkm}
\alias{body_replace_text_at_bkm}
\alias{body_replace_img_at_bkm}
\alias{headers_replace_text_at_bkm}
\alias{headers_replace_img_at_bkm}
\alias{footers_replace_text_at_bkm}
\alias{footers_replace_img_at_bkm}
\title{Replace text at a bookmark location}
\usage{
body_replace_text_at_bkm(x, bookmark, value)

body_replace_img_at_bkm(x, bookmark, value)

headers_replace_text_at_bkm(x, bookmark, value)

headers_replace_img_at_bkm(x, bookmark, value)

footers_replace_text_at_bkm(x, bookmark, value)

footers_replace_img_at_bkm(x, bookmark, value)
}
\arguments{
\item{x}{a docx device}

\item{bookmark}{bookmark id}

\item{value}{the replacement string, of type character}
}
\description{
Replace text content enclosed in a bookmark
with different text. A bookmark will be considered as valid if enclosing words
within a paragraph; i.e., a bookmark along two or more paragraphs is invalid,
a bookmark set on a whole paragraph is also invalid, but bookmarking few words
inside a paragraph is valid.
}
\examples{
library(officer)

doc <- read_docx()
doc <- body_add_par(doc, "a paragraph to replace", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
doc <- body_replace_text_at_bkm(doc, "text_to_replace", "new text")


# demo usage of bookmark and images ----
template <- system.file(package = "officer", "doc_examples/example.docx")

img.file <- file.path(R.home("doc"), "html", "logo.jpg")

doc <- read_docx(path = template)
doc <- headers_replace_img_at_bkm(
  x = doc,
  bookmark = "bmk_header",
  value = external_img(src = img.file, width = .53, height = .7)
)
doc <- footers_replace_img_at_bkm(
  x = doc,
  bookmark = "bmk_footer",
  value = external_img(src = img.file, width = .53, height = .7)
)
print(doc, target = tempfile(fileext = ".docx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_properties.R
\name{sp_lineend}
\alias{sp_lineend}
\alias{print.sp_lineend}
\alias{update.sp_lineend}
\title{Line end properties}
\usage{
sp_lineend(type = "none", width = "med", length = "med")

\method{print}{sp_lineend}(x, ...)

\method{update}{sp_lineend}(object, type, width, length, ...)
}
\arguments{
\item{type}{single character value specifying the line end type.
Expected value is one of the following : default \code{'none'}
or \code{'triangle'} or \code{'stealth'} or \code{'diamond'}
or \code{'oval'} or \code{'arrow'}}

\item{width}{single character value specifying the line end width
Expected value is one of the following : default \code{'sm'}
or \code{'med'} or \code{'lg'}}

\item{length}{single character value specifying the line end length
Expected value is one of the following : default \code{'sm'}
or \code{'med'} or \code{'lg'}}

\item{x, object}{\code{sp_lineend} object}

\item{...}{further arguments - not used}
}
\value{
a \code{sp_lineend} object
}
\description{
Create a \code{sp_lineend} object that describes
line end properties.
}
\examples{
sp_lineend()
sp_lineend(type = "triangle")
sp_lineend(type = "arrow", width = "lg", length = "lg")
print( sp_lineend (type="triangle", width = "lg") )
obj <- sp_lineend (type="triangle", width = "lg")
update( obj, type = "arrow" )
}
\seealso{
\link{sp_line}

Other functions for defining shape properties: 
\code{\link{sp_line}()}
}
\concept{functions for defining shape properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_cursor.R
\name{cursor_begin}
\alias{cursor_begin}
\alias{cursor_bookmark}
\alias{cursor_end}
\alias{cursor_reach_index}
\alias{cursor_reach}
\alias{cursor_reach_test}
\alias{cursor_forward}
\alias{cursor_backward}
\title{Set cursor in a 'Word' document}
\usage{
cursor_begin(x)

cursor_bookmark(x, id)

cursor_end(x)

cursor_reach_index(x, index)

cursor_reach(x, keyword, fixed = FALSE)

cursor_reach_test(x, keyword)

cursor_forward(x)

cursor_backward(x)
}
\arguments{
\item{x}{a docx device}

\item{id}{bookmark id}

\item{index}{element index in the document}

\item{keyword}{keyword to look for as a regular expression}

\item{fixed}{logical. If TRUE, pattern is a string to be matched as is.}
}
\description{
A set of functions is available to manipulate
the position of a virtual cursor. This cursor will be used when
inserting, deleting or updating elements in the document.
}
\section{cursor_begin}{

Set the cursor at the beginning of the document, on the first element
of the document (usually a paragraph or a table).
}

\section{cursor_bookmark}{

Set the cursor at a bookmark that has previously been set.
}

\section{cursor_end}{

Set the cursor at the end of the document, on the last element
of the document.
}

\section{cursor_reach_index}{

Set the cursor at a specific index position in the document.
}

\section{cursor_reach}{

Set the cursor on the first element of the document
that contains text specified in argument \code{keyword}.
The argument \code{keyword} is a regexpr pattern.
}

\section{cursor_reach_test}{

Test if an expression has a match in the document
that contains text specified in argument \code{keyword}.
The argument \code{keyword} is a regexpr pattern.
}

\section{cursor_forward}{

Move the cursor forward, it increments the cursor in the document.
}

\section{cursor_backward}{

Move the cursor backward, it decrements the cursor in the document.
}

\examples{
library(officer)

# create a template ----
doc <- read_docx()
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "Hello text to replace")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "Hello text to replace")
doc <- body_add_par(doc, "blah blah blah")
template_file <- print(
  x = doc,
  target = tempfile(fileext = ".docx")
)

# replace all pars containing "to replace" ----
doc <- read_docx(path = template_file)
while (cursor_reach_test(doc, "to replace")) {
  doc <- cursor_reach(doc, "to replace")

  doc <- body_add_fpar(
    x = doc,
    pos = "on",
    value = fpar(
      "Here is a link: ",
      hyperlink_ftext(
        text = "yopyop",
        href = "https://cran.r-project.org/"
      )
    )
  )
}

doc <- cursor_end(doc)
doc <- body_add_par(doc, "Yap yap yap yap...")

result_file <- print(
  x = doc,
  target = tempfile(fileext = ".docx")
)

# cursor_bookmark ----

doc <- read_docx()
doc <- body_add_par(doc, "centered text", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
doc <- body_add_par(doc, "A title", style = "heading 1")
doc <- body_add_par(doc, "Hello world!", style = "Normal")
doc <- cursor_bookmark(doc, "text_to_replace")
doc <- body_add_table(doc, value = iris, style = "table_template")

print(doc, target = tempfile(fileext = ".docx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_type}
\alias{ph_location_type}
\title{Location of a placeholder based on a type}
\usage{
ph_location_type(
  type = "body",
  type_idx = NULL,
  position_right = TRUE,
  position_top = TRUE,
  newlabel = NULL,
  id = NULL,
  ...
)
}
\arguments{
\item{type}{placeholder type to look for in the slide layout, one
of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.}

\item{type_idx}{Type index of the placeholder. If there is more than one
placeholder of a type (e.g., \code{body}), the type index can be supplied to uniquely
identify a ph. The index is a running number starting at 1. It is assigned by
placeholder position  (top -> bottom, left -> right). See \code{\link[=plot_layout_properties]{plot_layout_properties()}}
for details. If \code{idx} argument is used, \code{position_right} and \code{position_top}
are ignored.}

\item{position_right}{the parameter is used when a selection with above
parameters does not provide a unique position (for example
layout 'Two Content' contains two element of type 'body').
If \code{TRUE}, the element the most on the right side will be selected,
otherwise the element the most on the left side will be selected.}

\item{position_top}{same than \code{position_right} but applied
to top versus bottom.}

\item{newlabel}{a label to associate with the placeholder.}

\item{id}{(\strong{DEPRECATED, use \code{type_idx} instead}) Index of the placeholder.
If two body placeholder, there can be two different index: 1 and 2 for the
first and second body placeholders defined in the layout. If this argument
is used, \code{position_right} and \code{position_top} will be ignored.}

\item{...}{unused arguments}
}
\description{
The function will use the type name of the placeholder (e.g. body, title),
the layout name and few other criterias to find the corresponding location.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

# ph_location_type demo ----

loc_title <- ph_location_type(type = "title")
loc_footer <- ph_location_type(type = "ftr")
loc_dt <- ph_location_type(type = "dt")
loc_slidenum <- ph_location_type(type = "sldNum")
loc_body <- ph_location_type(type = "body")


doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(x = doc, "Un titre", location = loc_title)
doc <- ph_with(x = doc, "pied de page", location = loc_footer)
doc <- ph_with(x = doc, format(Sys.Date()), location = loc_dt)
doc <- ph_with(x = doc, "slide 1", location = loc_slidenum)
doc <- ph_with(x = doc, letters[1:10], location = loc_body)

loc_subtitle <- ph_location_type(type = "subTitle")
loc_ctrtitle <- ph_location_type(type = "ctrTitle")
doc <- add_slide(doc, layout = "Title Slide")
doc <- ph_with(x = doc, "Un sous titre", location = loc_subtitle)
doc <- ph_with(x = doc, "Un titre", location = loc_ctrtitle)

fileout <- tempfile(fileext = ".pptx")
print(doc, target = fileout)
}
\seealso{
Other functions for placeholder location:
\code{\link[=ph_location]{ph_location()}},
\code{\link[=ph_location_fullsize]{ph_location_fullsize()}},
\code{\link[=ph_location_id]{ph_location_id()}},
\code{\link[=ph_location_label]{ph_location_label()}},
\code{\link[=ph_location_left]{ph_location_left()}},
\code{\link[=ph_location_right]{ph_location_right()}},
\code{\link[=ph_location_template]{ph_location_template()}}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shorcuts.R
\name{shortcuts}
\alias{shortcuts}
\title{shortcuts for formatting properties}
\usage{
shortcuts
}
\description{
Shortcuts for \code{\link[=fp_text]{fp_text()}}, \code{\link[=fp_par]{fp_par()}}, \code{\link[=fp_cell]{fp_cell()}} and \code{\link[=fp_border]{fp_border()}}.
}
\examples{
shortcuts$fp_bold()
shortcuts$fp_italic()
shortcuts$b_null()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{unordered_list}
\alias{unordered_list}
\title{Unordered list}
\usage{
unordered_list(str_list = character(0), level_list = integer(0), style = NULL)
}
\arguments{
\item{str_list}{list of strings to be included in the object}

\item{level_list}{list of levels for hierarchy structure. Use
0 for 'no bullet', 1 for level 1, 2 for level 2 and so on.}

\item{style}{text style, a \code{fp_text} object list or a
single \code{fp_text} objects. Use \code{fp_text(font.size = 0, ...)} to
inherit from default sizes of the presentation.}
}
\description{
unordered list of text for PowerPoint
presentations. Each text is associated with
a hierarchy level.

Consider using \code{\link[=block_list_items]{block_list_items()}} instead, which supports
rich text via \code{\link[=fpar]{fpar()}}, works in both Word and PowerPoint,
and supports numbered lists.
}
\examples{
unordered_list(
  level_list = c(1, 2, 2, 3, 3, 1),
  str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
  style = fp_text(color = "red", font.size = 0)
)

unordered_list(
  level_list = c(1, 2, 1),
  str_list = c("Level1", "Level2", "Level1"),
  style = list(
    fp_text(color = "red", font.size = 0),
    fp_text(color = "pink", font.size = 0),
    fp_text(color = "orange", font.size = 0)
  )
)
}
\seealso{
\code{\link[=ph_with]{ph_with()}}

Other block functions for reporting:
\code{\link[=block_caption]{block_caption()}},
\code{\link[=block_gg]{block_gg()}},
\code{\link[=block_list]{block_list()}},
\code{\link[=block_list_items]{block_list_items()}},
\code{\link[=block_pour_docx]{block_pour_docx()}},
\code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}},
\code{\link[=block_toc]{block_toc()}},
\code{\link[=fpar]{fpar()}},
\code{\link[=list_item]{list_item()}},
\code{\link[=plot_instr]{plot_instr()}}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{slide_visible<-}
\alias{slide_visible<-}
\alias{slide_visible}
\title{Get or set slide visibility}
\usage{
slide_visible(x) <- value

slide_visible(x, hide = NULL, show = NULL)
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{value}{Boolean vector with slide visibilities.}

\item{hide, show}{Indexes of slides to hide or show.}
}
\value{
Boolean vector with slide visibilities or \code{rpptx} object if changes are made to the object.
}
\description{
PPTX slides can be visible or hidden. This function gets or sets the visibility of slides.
}
\examples{
path <- system.file("doc_examples/example.pptx", package = "officer")
x <- read_pptx(path)

slide_visible(x) # get slide visibilities

x <- slide_visible(x, hide = 1:2) # hide slides 1 and 2
x <- slide_visible(x, show = 1:2) # make slides 1 and 2 visible
x <- slide_visible(x, show = 1:2, hide = 3)

slide_visible(x) <- FALSE # hide all slides
slide_visible(x) <- c(TRUE, FALSE, TRUE) # set each slide separately
slide_visible(x) <- c(TRUE, FALSE) # warns that rhs values are recycled

slide_visible(x)[2] <- TRUE # set 2nd slide to visible
slide_visible(x)[c(1, 3)] <- FALSE # 1st and 3rd slide
slide_visible(x)[c(1, 3)] <- c(FALSE, FALSE) # identical
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_set_default_section}
\alias{body_set_default_section}
\title{Define Default Section}
\usage{
body_set_default_section(x, value)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{a \link{prop_section} object}
}
\description{
Define default section of the document. You can
define section propeerties (page size, orientation, ...) with a \link{prop_section} object.
}
\section{Illustrations}{


\if{html}{\figure{body_set_default_section_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
# Example 1: Setting page layout properties ----
# This example demonstrates how to configure the default section
# properties for page orientation, type, and margins

# Define custom section properties
# - Landscape orientation for wide tables
# - Continuous section (no page break)
# - Custom margins: top=1.5", bottom=0.75", left/right=2"
default_sect_properties <- prop_section(
  page_size = page_size(orient = "landscape"),
  type = "continuous",
  page_margins = page_mar(bottom = .75, top = 1.5, right = 2, left = 2)
)

# Create a new document
doc_1 <- read_docx()

# Add a wide table that benefits from landscape orientation
doc_1 <- body_add_par(doc_1, "Motor Trend Car Road Tests", style = "heading 1")
doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")

# Add some text content
doc_1 <- body_add_par(doc_1, "Sample Text Content", style = "heading 2")
doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 40), collapse = " "))

# Apply the section properties to the entire document
# This must be called at the end, after all content is added
doc_1 <- body_set_default_section(doc_1, default_sect_properties)

# Save the document
print(doc_1, target = tempfile(fileext = ".docx"))


# Example 2: Adding headers and footers ----
# This example shows how to create a document with:
# - A header containing the R logo
# - A footer with the current date and page numbers

# Get the path to R logo
img_path <- file.path(R.home("doc"), "html", "logo.jpg")

# Create header content with the R logo
# The logo is positioned on the right side with specific dimensions
header_content <- block_list(
  fpar(
    external_img(src = img_path, height = 0.5, width = 0.5),
    fp_p = fp_par(text.align = "right")
  )
)

# Create footer content with date and page numbers
# Format: "Document generated on: [Date] | Page [X]"
footer_content <- block_list(
  fpar(
    "Document generated on: ",
    run_word_field(field = "Date \\\\@ \"MMMM d, yyyy\""),
    " | Page ",
    run_word_field(field = "PAGE"),
    fp_p = fp_par(text.align = "center")
  )
)

# Define section properties that include header and footer
# The header and footer will appear on all pages
sect_with_hf <- prop_section(
  page_size = page_size(orient = "portrait", width = 8.3, height = 11.7),
  page_margins = page_mar(
    bottom = 1,
    top = 1,
    right = 1,
    left = 1,
    header = 0.5,
    footer = 0.5
  ),
  type = "continuous",
  header_default = header_content,
  footer_default = footer_content
)

# Create a new document with content
doc_2 <- read_docx()

# Add a title page
doc_2 <- body_add_par(doc_2, "Annual Report 2024", style = "heading 1")
doc_2 <- body_add_par(
  doc_2,
  "Company Performance Analysis",
  style = "heading 2"
)

# Add some sections with content
doc_2 <- body_add_par(doc_2, "Executive Summary", style = "heading 2")
doc_2 <- body_add_par(
  doc_2,
  "This report provides a comprehensive analysis of company performance metrics."
)

# Add a table
doc_2 <- body_add_par(doc_2, "Key Metrics", style = "heading 2")
summary_data <- data.frame(
  Metric = c("Revenue", "Profit", "Growth"),
  Q1 = c(1.2, 0.3, 12),
  Q2 = c(1.5, 0.4, 15),
  Q3 = c(1.8, 0.5, 18),
  Q4 = c(2.1, 0.6, 20)
)
doc_2 <- body_add_table(doc_2, value = summary_data, style = "table_template")

# Add a plot
doc_2 <- body_add_par(doc_2, "Revenue Trend", style = "heading 2")
revenue_plot <- plot_instr({
  quarters <- paste0("Q", 1:4)
  revenue <- c(1.2, 1.5, 1.8, 2.1)
  barplot(
    revenue,
    names.arg = quarters,
    col = "#4472C4",
    border = NA,
    main = "Quarterly Revenue (Millions)",
    ylab = "Revenue ($M)",
    xlab = "Quarter"
  )
})
doc_2 <- body_add_plot(doc_2, revenue_plot, width = 5, height = 4)

# Apply section properties with header and footer
# The header (with R logo) and footer (with date and page number)
# will appear on all pages
doc_2 <- body_set_default_section(doc_2, sect_with_hf)

# Save the document
output_file <- tempfile(fileext = ".docx")
print(doc_2, target = output_file)
}
\seealso{
Other functions for Word sections:
\code{\link[=body_end_block_section]{body_end_block_section()}},
\code{\link[=body_end_section_columns]{body_end_section_columns()}},
\code{\link[=body_end_section_columns_landscape]{body_end_section_columns_landscape()}},
\code{\link[=body_end_section_continuous]{body_end_section_continuous()}},
\code{\link[=body_end_section_landscape]{body_end_section_landscape()}},
\code{\link[=body_end_section_portrait]{body_end_section_portrait()}}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_remove}
\alias{body_remove}
\title{Remove an element in a 'Word' document}
\usage{
body_remove(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
Remove element pointed by cursor from a
'Word' document.
}
\examples{
library(officer)

str1 <- rep("Lorem ipsum dolor sit amet, consectetur adipiscing elit. ", 20)
str1 <- paste(str1, collapse = "")

str2 <- "Drop that text"

str3 <- rep("Aenean venenatis varius elit et fermentum vivamus vehicula. ", 20)
str3 <- paste(str3, collapse = "")

my_doc <- read_docx()
my_doc <- body_add_par(my_doc, value = str1, style = "Normal")
my_doc <- body_add_par(my_doc, value = str2, style = "centered")
my_doc <- body_add_par(my_doc, value = str3, style = "Normal")

new_doc_file <- print(my_doc,
  target = tempfile(fileext = ".docx")
)

my_doc <- read_docx(path = new_doc_file)
my_doc <- cursor_reach(my_doc, keyword = "that text")
my_doc <- body_remove(my_doc)

print(my_doc, target = tempfile(fileext = ".docx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_rename_layout.R
\name{layout_rename_ph_labels}
\alias{layout_rename_ph_labels}
\alias{layout_rename_ph_labels<-}
\title{Change ph labels in a layout}
\usage{
layout_rename_ph_labels(x, layout, master = NULL, ..., .dots = NULL)

layout_rename_ph_labels(x, layout, master = NULL, id = NULL) <- value
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{layout}{Layout name or index. Index is the row index of \code{\link[=layout_summary]{layout_summary()}}.}

\item{master}{Name of master. Only required if the layout name is not unique across masters.}

\item{...}{Comma separated list of key-value pairs to rename phs. Either reference a ph via its label
(\code{"old label"} = \code{"new label"}) or its unique id (\code{"id"} = \code{"new label"}).}

\item{.dots}{Provide a named list or vector of key-value pairs to rename phs
(\verb{list("old label"} = \code{"new label"}).}

\item{id}{Unique placeholder id (see column \code{id} in \code{\link[=layout_properties]{layout_properties()}} or \code{\link[=plot_layout_properties]{plot_layout_properties()}}).}

\item{value}{Not relevant for user. A pure technical necessity for rhs assignments.}
}
\value{
Vector of renamed ph labels.
}
\description{
There are two versions of the function. The first takes a set of key-value pairs to rename the
ph labels. The second uses a right hand side (rhs) assignment to specify the new ph labels.
See section \emph{Details}. \cr\cr
\emph{NB:} You can also rename ph labels directly in PowerPoint. Open the master template view
(\code{Alt} + \code{F10}) and go to \code{Home} > \code{Arrange} > \verb{Selection Pane}.
}
\details{
\itemize{
\item Note the difference between the terms \code{id} and \code{index}. Both can be found in the output of
\code{\link[=layout_properties]{layout_properties()}}. The unique ph \code{id} is found in column \code{id}. The \code{index} refers to the
index of the data frame row.
\item In a right hand side (rhs) label assignment (\verb{<- new_labels}), there are two ways to
optionally specify a subset of phs to rename. In both cases, the length of the rhs vector
(the new labels) must match the length of the id or index:
\enumerate{
\item use the \code{id} argument to specify ph ids to rename: \code{layout_rename_ph_labels(..., id = 2:3) <- new_labels}
\item use an \code{index} in squared brackets: \code{layout_rename_ph_labels(...)[1:2] <- new_labels}
}
}
}
\examples{
x <- read_pptx()

# INFO -------------

# Returns layout's ph_labels by default in same order as layout_properties()
layout_rename_ph_labels(x, "Comparison")
layout_properties(x, "Comparison")$ph_label


# BASICS -----------
#
# HINT: run `plot_layout_properties(x, "Comparison")` to see how labels change

# rename using key-value pairs: 'old label' = 'new label' or 'id' = 'new label'
layout_rename_ph_labels(x, "Comparison", "Title 1" = "LABEL MATCHED") # label matching
layout_rename_ph_labels(x, "Comparison", "3" = "ID MATCHED") # id matching
layout_rename_ph_labels(x, "Comparison", "Date Placeholder 6" = "DATE", "8" = "FOOTER") # label, id

# rename using a named list and the .dots arg
renames <- list("Content Placeholder 3" = "CONTENT_1", "6" = "CONTENT_2")
layout_rename_ph_labels(x, "Comparison", .dots = renames)

# rename via rhs assignment and optional index (not id!)
layout_rename_ph_labels(x, "Comparison") <- LETTERS[1:8]
layout_rename_ph_labels(x, "Comparison")[1:3] <- paste("CHANGED", 1:3)

# rename via rhs assignment and ph id (not index)
layout_rename_ph_labels(x, "Comparison", id = c(2, 4)) <- paste("ID =", c(2, 4))


# MORE ------------

# make all labels lower case
labels <- layout_rename_ph_labels(x, "Comparison")
layout_rename_ph_labels(x, "Comparison") <- tolower(labels)

# rename all labels to type [type_idx]
lp <- layout_properties(x, "Comparison")
layout_rename_ph_labels(x, "Comparison") <- paste0(lp$type, " [", lp$type_idx, "]")

# rename duplicated placeholders (see also `layout_dedupe_ph_labels()`)
file <- system.file("doc_examples", "ph_dupes.pptx", package = "officer")
x <- read_pptx(file)
lp <- layout_properties(x, "2-dupes")
idx <- which(lp$ph_label == "Content 7") # exists twice
layout_rename_ph_labels(x, "2-dupes")[idx] <- paste("DUPLICATE", seq_along(idx))

# warning: in case of duped labels only the first occurrence is renamed
x <- read_pptx(file)
layout_rename_ph_labels(x, "2-dupes", "Content 7" = "new label")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_blocks}
\alias{body_add_blocks}
\title{Add a list of blocks into a 'Word' document}
\usage{
body_add_blocks(x, blocks, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{blocks}{set of blocks to be used as footnote content returned by
function \code{\link[=block_list]{block_list()}}.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
add a list of blocks produced by \code{block_list} into
into an rdocx object.
}
\examples{
library(officer)

img.file <- file.path(R.home("doc"), "html", "logo.jpg")

bl <- block_list(
  fpar(ftext("hello", shortcuts$fp_bold(color = "red"))),
  fpar(
    ftext("hello world", shortcuts$fp_bold()),
    external_img(src = img.file, height = 1.06, width = 1.39),
    fp_p = fp_par(text.align = "center")
  )
)

doc_1 <- read_docx()
doc_1 <- body_add_blocks(doc_1, blocks = bl)
print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_section_landscape}
\alias{body_end_section_landscape}
\title{Add landscape section}
\usage{
body_end_section_landscape(x, w = 16838/1440, h = 11906/1440)
}
\arguments{
\item{x}{an rdocx object}

\item{w, h}{page width, page height (in inches)}
}
\description{
A section with landscape orientation is added to the document.
}
\examples{
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 5)
str1 <- paste(str1, collapse = " ")

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_end_section_landscape(doc_1)

print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_block_section}()},
\code{\link{body_end_section_columns}()},
\code{\link{body_end_section_columns_landscape}()},
\code{\link{body_end_section_continuous}()},
\code{\link{body_end_section_portrait}()},
\code{\link{body_set_default_section}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{fpar}
\alias{fpar}
\alias{update.fpar}
\title{Formatted paragraph}
\usage{
fpar(
  ...,
  fp_p = fp_par(word_style = NA_character_),
  fp_t = fp_text_lite(),
  values = NULL
)

\method{update}{fpar}(object, fp_p = NULL, fp_t = NULL, ...)
}
\arguments{
\item{...}{cot objects (\code{\link[=ftext]{ftext()}}, \code{\link[=external_img]{external_img()}})}

\item{fp_p}{paragraph formatting properties, see \code{\link[=fp_par]{fp_par()}}}

\item{fp_t}{default text formatting properties. This is used as
text formatting properties when simple text is provided as argument,
see \code{\link[=fp_text]{fp_text()}}.}

\item{values}{a list of cot objects. If provided, argument \code{...} will be ignored.}

\item{object}{fpar object}
}
\description{
Create a paragraph representation by concatenating
formatted text or images. The result can be inserted in a Word document
or a PowerPoint presentation and can also be inserted in a \code{\link[=block_list]{block_list()}}
call.

All its arguments will be concatenated to create a paragraph where chunks of
text and images are associated with formatting properties.

\code{fpar()} supports \code{\link[=ftext]{ftext()}}, \code{\link[=external_img]{external_img()}}, \verb{run_*()} functions
(i.e. \code{\link[=run_autonum]{run_autonum()}}, \code{\link[=run_word_field]{run_word_field()}}) when output is Word, and simple strings.

Default text and paragraph formatting properties can also be modified
with function \code{update()}.
}
\examples{
fpar(ftext("hello", shortcuts$fp_bold()))

# mix text and image -----
img.file <- file.path( R.home("doc"), "html", "logo.jpg" )

bold_face <- shortcuts$fp_bold(font.size = 12)
bold_redface <- update(bold_face, color = "red")
fpar_1 <- fpar(
  "Hello World, ",
  ftext("how ", prop = bold_redface ),
  external_img(src = img.file, height = 1.06/2, width = 1.39/2),
  ftext(" you?", prop = bold_face ) )
fpar_1

img_in_par <- fpar(
  external_img(src = img.file, height = 1.06/2, width = 1.39/2),
  fp_p = fp_par(text.align = "center") )
}
\seealso{
\code{\link[=block_list]{block_list()}}, \code{\link[=body_add_fpar]{body_add_fpar()}}, \code{\link[=ph_with]{ph_with()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_gg}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{body_replace_gg_at_bkm}
\alias{body_replace_gg_at_bkm}
\alias{body_replace_plot_at_bkm}
\title{Add plots at bookmark location in a 'Word' document}
\usage{
body_replace_gg_at_bkm(
  x,
  bookmark,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  keep = FALSE,
  ...
)

body_replace_plot_at_bkm(
  x,
  bookmark,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  keep = FALSE,
  ...
)
}
\arguments{
\item{x}{an rdocx object}

\item{bookmark}{bookmark id}

\item{value}{a ggplot object for body_replace_gg_at_bkm() or a set plot instructions
body_replace_plot_at_bkm(), see plot_instr().}

\item{width}{height in inches}

\item{height}{height in inches}

\item{res}{resolution of the png image in ppi}

\item{style}{paragraph style}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{keep}{Should the bookmark be preserved? Defaults to \code{FALSE},
i.e.the bookmark will be lost as a side effect.}

\item{...}{Arguments to be passed to png function.}
}
\description{
Use these functions if you want to replace a paragraph containing
a bookmark with a 'ggplot' or a base plot.
}
\examples{
if (require("ggplot2")) {
  gg_plot <- ggplot(data = iris) +
    geom_point(mapping = aes(Sepal.Length, Petal.Length))

  doc <- read_docx()
  doc <- body_add_par(doc, "insert_plot_here")
  doc <- body_bookmark(doc, "plot")
  doc <- body_replace_gg_at_bkm(doc, bookmark = "plot", value = gg_plot)
  print(doc, target = tempfile(fileext = ".docx"))
}
doc <- read_docx()
doc <- body_add_par(doc, "insert_plot_here")
doc <- body_bookmark(doc, "plot")
if (capabilities(what = "png")) {
  doc <- body_replace_plot_at_bkm(
    doc, bookmark = "plot",
    value = plot_instr(
      code = {
        barplot(1:5, col = 2:6)
      }
    )
  )
}
print(doc, target = tempfile(fileext = ".docx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_layout_helper.R
\name{get_layout}
\alias{get_layout}
\title{Layout selection helper}
\usage{
get_layout(
  x,
  layout = NULL,
  master = NULL,
  layout_by_id = TRUE,
  get_first = FALSE
)
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{layout}{Layout name or index. Index refers to the row index of the \code{\link[=layout_summary]{layout_summary()}}
output.}

\item{master}{Name of master. Only required if layout name is not unique across masters.}

\item{layout_by_id}{Allow layout index instead of name? (default is \code{TRUE})}

\item{get_first}{If layout exists in multiple master, return first occurence (default \code{FALSE}).}
}
\value{
A \verb{<layout_info>} object, i.e. a list with the entries \code{index}, \code{layout_name},
\code{layout_file}, \code{master_name}, \code{master_file}, and \code{slide_layout}.
}
\description{
Select a layout by name or index. The master name is inferred and only required
for disambiguation in case the layout name is not unique across masters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_fpar}
\alias{body_add_fpar}
\title{Add fpar in a 'Word' document}
\usage{
body_add_fpar(x, value, style = NULL, pos = "after")
}
\arguments{
\item{x}{a docx device}

\item{value}{a character}

\item{style}{paragraph style. If NULL, paragraph settings from \code{fpar} will be used. If not
NULL, it must be a paragraph style name (located in the template
provided as \code{read_docx(path = ...)}); in that case, paragraph settings from \code{fpar} will be
ignored.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add an \code{fpar} (a formatted paragraph)
into an rdocx object.
}
\examples{
bold_face <- shortcuts$fp_bold(font.size = 30)
bold_redface <- update(bold_face, color = "red")
fpar_ <- fpar(
  ftext("Hello ", prop = bold_face),
  ftext("World", prop = bold_redface),
  ftext(", how are you?", prop = bold_face)
)
doc <- read_docx()
doc <- body_add_fpar(doc, fpar_)

print(doc, target = tempfile(fileext = ".docx"))

# a way of using fpar to center an image in a Word doc ----
rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
img_in_par <- fpar(
  external_img(src = rlogo, height = 1.06 / 2, width = 1.39 / 2),
  hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "cran", prop = bold_redface
  ),
  fp_p = fp_par(text.align = "center")
)

doc <- read_docx()
doc <- body_add_fpar(doc, img_in_par)
print(doc, target = tempfile(fileext = ".docx"))

}
\seealso{
\code{\link{fpar}}

Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_tab}
\alias{run_tab}
\title{Tab for 'Word'}
\usage{
run_tab()
}
\description{
Object representing a tab
in a Word document. The result must be used
within a call to \link{fpar}. It will only have
effects in Word output.

Tabulation marks settings can be defined with \code{\link[=fp_tabs]{fp_tabs()}} in
paragraph settings defined with \code{\link[=fp_par]{fp_par()}}.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
z <- fp_tabs(
  fp_tab(pos = 0.5, style = "decimal"),
  fp_tab(pos = 1.5, style = "decimal")
)
par1 <- fpar(
  run_tab(), ftext("88."),
  run_tab(), ftext("987.45"),
  fp_p = fp_par(
    tabs = z
  )
)
par2 <- fpar(
  run_tab(), ftext("8."),
  run_tab(), ftext("670987.45"),
  fp_p = fp_par(
    tabs = z
  )
)
x <- read_docx()
x <- body_add(x, par1)
x <- body_add(x, par2)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_section_portrait}
\alias{body_end_section_portrait}
\title{Add portrait section}
\usage{
body_end_section_portrait(x, w = 21/2.54, h = 29.7/2.54)
}
\arguments{
\item{x}{an rdocx object}

\item{w, h}{page width, page height (in inches)}
}
\description{
A section with portrait orientation is added to the document.
}
\examples{
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 5)
str1 <- paste(str1, collapse = " ")

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_end_section_portrait(doc_1)
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_block_section}()},
\code{\link{body_end_section_columns}()},
\code{\link{body_end_section_columns_landscape}()},
\code{\link{body_end_section_continuous}()},
\code{\link{body_end_section_landscape}()},
\code{\link{body_set_default_section}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{page_mar}
\alias{page_mar}
\title{Page margins object}
\usage{
page_mar(
  bottom = 1417/1440,
  top = 1417/1440,
  right = 1417/1440,
  left = 1417/1440,
  header = 708/1440,
  footer = 708/1440,
  gutter = 0/1440
)
}
\arguments{
\item{bottom, top}{distance (in inches) between the bottom/top of the text
margin and the bottom/top of the page. The text is placed at the greater of
the value of this attribute and the extent of the header/footer text. A
negative value indicates that the content should be measured from the
bottom/top of the page regardless of the footer/header, and so will overlap
the footer/header. For example, \verb{header=-0.5, bottom=1} means that the footer
must start one inch from the bottom of the page and the main document text
must start a half inch from the bottom of the page. In this case, the text
and footer overlap since bottom is negative.}

\item{left, right}{distance (in inches) from the left/right edge of the page
to the left/right edge of the text.}

\item{header}{distance (in inches) from the top edge of the page to the top
edge of the header.}

\item{footer}{distance (in inches) from the bottom edge of the page to the
bottom edge of the footer.}

\item{gutter}{page gutter (in inches).}
}
\description{
Define margins for each page of a section.

The function creates a representation of the dimensions of a page. The
dimensions are defined by length, width and orientation. If the orientation
is in landscape mode then the length becomes the width and the width becomes
the length.
}
\examples{
page_mar()
}
\seealso{
Other functions for section definition: 
\code{\link{page_size}()},
\code{\link{prop_section}()},
\code{\link{section_columns}()}
}
\concept{functions for section definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_template}
\alias{ph_location_template}
\title{Location for a placeholder based on a template}
\usage{
ph_location_template(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  type = NULL,
  id = 1,
  ...
)
}
\arguments{
\item{left, top, width, height}{place holder coordinates
in inches.}

\item{newlabel}{a label for the placeholder. See section details.}

\item{type}{placeholder type to look for in the slide layout, one
of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.
It will be used as a template placeholder.}

\item{id}{index of the placeholder template. If two body placeholder, there can be
two different index: 1 and 2 for the first and second body placeholders defined
in the layout.}

\item{...}{unused arguments}
}
\description{
The function will return a list that complies with
expected format for argument \code{location} of function
\code{\link[=ph_with]{ph_with()}}. A placeholder will be used as template
and its positions will be updated with values \code{left}, \code{top}, \code{width}, \code{height}.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, "Title", location = ph_location_type(type = "title"))
doc <- ph_with(
  doc,
  "Hello world",
  location = ph_location_template(top = 4, type = "title")
)
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{set_autonum_bookmark}
\alias{set_autonum_bookmark}
\title{Update bookmark of an autonumber run}
\usage{
set_autonum_bookmark(x, bkm = NULL)
}
\arguments{
\item{x}{an object of class \code{\link[=run_autonum]{run_autonum()}}}

\item{bkm}{bookmark id to associate with autonumber run.
Value can only be made of alpha numeric characters, ':', -' and '_'.}
}
\description{
This function lets recycling a object
made by \code{\link[=run_autonum]{run_autonum()}} by changing the bookmark value. This
is useful to avoid calling \code{run_autonum()} several times
because of many tables.
}
\examples{
z <- run_autonum(
  seq_id = "tab", pre_label = "Table ",
  bkm = "anytable"
)
set_autonum_bookmark(z, bkm = "anothertable")
}
\seealso{
\code{\link[=run_autonum]{run_autonum()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{doc_properties}
\alias{doc_properties}
\title{Read document properties}
\usage{
doc_properties(x)
}
\arguments{
\item{x}{an \code{rdocx} or \code{rpptx} object}
}
\value{
a data.frame
}
\description{
Read Word or PowerPoint document properties
and get results in a data.frame.
}
\examples{
x <- read_docx()
doc_properties(x)
}
\seealso{
Other functions for Word document informations: 
\code{\link{docx_bookmarks}()},
\code{\link{docx_dim}()},
\code{\link{length.rdocx}()},
\code{\link{set_doc_properties}()},
\code{\link{styles_info}()}

Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for Word document informations}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{body_replace_text_at_bkm}
\alias{body_replace_text_at_bkm}
\alias{body_replace_img_at_bkm}
\alias{headers_replace_text_at_bkm}
\alias{headers_replace_img_at_bkm}
\alias{footers_replace_text_at_bkm}
\alias{footers_replace_img_at_bkm}
\title{Replace text at a bookmark location}
\usage{
body_replace_text_at_bkm(x, bookmark, value)

body_replace_img_at_bkm(x, bookmark, value)

headers_replace_text_at_bkm(x, bookmark, value)

headers_replace_img_at_bkm(x, bookmark, value)

footers_replace_text_at_bkm(x, bookmark, value)

footers_replace_img_at_bkm(x, bookmark, value)
}
\arguments{
\item{x}{a docx device}

\item{bookmark}{bookmark id}

\item{value}{the replacement string, of type character}
}
\description{
Replace text content enclosed in a bookmark
with different text. A bookmark will be considered as valid if enclosing words
within a paragraph; i.e., a bookmark along two or more paragraphs is invalid,
a bookmark set on a whole paragraph is also invalid, but bookmarking few words
inside a paragraph is valid.
}
\examples{
doc <- read_docx()
doc <- body_add_par(doc, "a paragraph to replace", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
doc <- body_replace_text_at_bkm(doc, "text_to_replace", "new text")


# demo usage of bookmark and images ----
template <- system.file(package = "officer", "doc_examples/example.docx")

img.file <- file.path( R.home("doc"), "html", "logo.jpg" )

doc <- read_docx(path = template)
doc <- headers_replace_img_at_bkm(x = doc, bookmark = "bmk_header",
                                  value = external_img(src = img.file, width = .53, height = .7))
doc <- footers_replace_img_at_bkm(x = doc, bookmark = "bmk_footer",
                                  value = external_img(src = img.file, width = .53, height = .7))
print(doc, target = tempfile(fileext = ".docx"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{section_columns}
\alias{section_columns}
\title{Section columns}
\usage{
section_columns(widths = c(2.5, 2.5), space = 0.25, sep = FALSE)
}
\arguments{
\item{widths}{columns widths in inches. If 3 values, 3 columns
will be produced.}

\item{space}{space in inches between columns.}

\item{sep}{if TRUE a line is separating columns.}
}
\description{
The function creates a representation of the columns of a section.
}
\examples{
section_columns()
}
\seealso{
Other functions for section definition: 
\code{\link{page_mar}()},
\code{\link{page_size}()},
\code{\link{prop_section}()}
}
\concept{functions for section definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_import_docx}
\alias{body_import_docx}
\title{Import an external docx in a 'Word' document}
\usage{
body_import_docx(
  x,
  src,
  par_style_mapping = list(),
  run_style_mapping = list(),
  tbl_style_mapping = list(),
  prepend_chunks_on_styles = list()
)
}
\arguments{
\item{x}{an rdocx object}

\item{src}{path to docx file to import}

\item{par_style_mapping, run_style_mapping, tbl_style_mapping}{Named lists describing how to remap styles from the source document (\code{src})
to styles available in the target document \code{x}. For each list entry, the name
of the element is the target style (in \code{x}), and the value is a character
vector of style names from the source document that should be replaced by
this target style.
\itemize{
\item \code{par_style_mapping}: applies to paragraph styles.
\item \code{run_style_mapping}: applies to character (run) styles.
\item \code{tbl_style_mapping}: applies to table styles.
}

Examples:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{par_style_mapping = list(
  "Normal"    = c("List Paragraph", "Body Text"),
  "heading 1" = "Heading 1"
)
run_style_mapping = list(
  "Emphasis"  = c("Emphasis", "Italic")
)
tbl_style_mapping = list(
  "Normal Table" = c("Light Shading")
)
}\if{html}{\out{</div>}}

Use \code{\link[=styles_info]{styles_info()}} to inspect available styles and verify their names.}

\item{prepend_chunks_on_styles}{A named list of run chunks to prepend
to runs with specific styles. The names of the list are paragraph style names
and the values run chunks to prepend. The first motivation for this argument
is to allow prepending of runs in paragraphs with a defined style, for example
to add a \code{\link[=run_autonum]{run_autonum()}} with all image captions.}
}
\description{
Import body content and footnotes of a Word document into an rdocx object.

The function is similar to \code{\link[=body_add_docx]{body_add_docx()}} but instead of adding
the content as an external object, the document is read and all its
content is appended to the target document.
}
\details{
The following operations are performed when importing a document:
\itemize{
\item Numberings are copied from the source document to the target document.
\item Styles are not copied. If styles in the source document do not exist
in the target document, the style specified in the \code{par_style_mapping},
\code{run_style_mapping} and \code{tbl_style_mapping} arguments will be used instead.
If no mapping is provided, the default style will be used and a warning is emitted.
}
}
\examples{
library(officer)

# example file from the package
file_input <- system.file(
  package = "officer",
  "doc_examples/example.docx"
)

# create a new rdocx document
x <- read_docx()

# import content from file_input
x <- body_import_docx(
  x = x,
  src = file_input,
  # style mapping for paragraphs and tables
  par_style_mapping = list(
    "Normal" = c("List Paragraph")
  ),
  tbl_style_mapping = list(
    "Normal Table" = "Light Shading"
  )
)

# Create temporary file
tf <- tempfile(fileext = ".docx")
# write to file
print(x, target = tf)
}
\seealso{
Other functions for adding content:
\code{\link[=body_add_blocks]{body_add_blocks()}},
\code{\link[=body_add_break]{body_add_break()}},
\code{\link[=body_add_caption]{body_add_caption()}},
\code{\link[=body_add_docx]{body_add_docx()}},
\code{\link[=body_add_fpar]{body_add_fpar()}},
\code{\link[=body_add_gg]{body_add_gg()}},
\code{\link[=body_add_img]{body_add_img()}},
\code{\link[=body_add_list]{body_add_list()}},
\code{\link[=body_add_par]{body_add_par()}},
\code{\link[=body_add_plot]{body_add_plot()}},
\code{\link[=body_add_table]{body_add_table()}},
\code{\link[=body_add_toc]{body_add_toc()}},
\code{\link[=body_append_start_context]{body_append_start_context()}}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_manipulate.R
\name{ph_remove}
\alias{ph_remove}
\title{Remove a shape}
\usage{
ph_remove(x, type = "body", id = 1, ph_label = NULL, id_chr = NULL)
}
\arguments{
\item{x}{an rpptx object}

\item{type}{placeholder type}

\item{id}{placeholder index (integer) for a duplicated type. This is to be used when a placeholder
type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
Values can be read from \code{\link[=slide_summary]{slide_summary()}}.}

\item{ph_label}{label associated to the placeholder. Use column
\code{ph_label} of result returned by \code{\link[=slide_summary]{slide_summary()}}.
If used, \code{type} and \code{id} are ignored.}

\item{id_chr}{deprecated.}
}
\description{
Remove a shape in a slide.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
dummy_fun <- function(doc) {
  doc <- add_slide(doc,
    layout = "Two Content",
    master = "Office Theme"
  )
  doc <- ph_with(
    x = doc, value = "Un titre",
    location = ph_location_type(type = "title")
  )
  doc <- ph_with(
    x = doc, value = "Un corps 1",
    location = ph_location_type(type = "body", id = 1)
  )
  doc <- ph_with(
    x = doc, value = "Un corps 2",
    location = ph_location_type(type = "body", id = 2)
  )
  doc
}
doc <- read_pptx()
for (i in 1:3) {
  doc <- dummy_fun(doc)
}

doc <- on_slide(doc, index = 1)
doc <- ph_remove(x = doc, type = "title")

doc <- on_slide(doc, index = 2)
doc <- ph_remove(x = doc, type = "body", id = 2)

doc <- on_slide(doc, index = 3)
doc <- ph_remove(x = doc, type = "body", id = 1)

print(doc, target = fileout)
}
\seealso{
\code{\link[=ph_with]{ph_with()}}

Other functions for placeholders manipulation: 
\code{\link{ph_hyperlink}()},
\code{\link{ph_slidelink}()}
}
\concept{functions for placeholders manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_section_columns}
\alias{body_end_section_columns}
\title{Add multi columns section}
\usage{
body_end_section_columns(x, widths = c(2.5, 2.5), space = 0.25, sep = FALSE)
}
\arguments{
\item{x}{an rdocx object}

\item{widths}{columns widths in inches. If 3 values, 3 columns
will be produced.}

\item{space}{space in inches between columns.}

\item{sep}{if TRUE a line is separating columns.}
}
\description{
A section with multiple columns is added to the
document.

You may prefer to use \code{\link[=body_end_block_section]{body_end_block_section()}} that is
more flexible.
}
\examples{
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 5)
str1 <- paste(str1, collapse = " ")

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_end_section_columns(doc_1)
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_block_section}()},
\code{\link{body_end_section_columns_landscape}()},
\code{\link{body_end_section_continuous}()},
\code{\link{body_end_section_landscape}()},
\code{\link{body_end_section_portrait}()},
\code{\link{body_set_default_section}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_cursor.R
\name{cursor_begin}
\alias{cursor_begin}
\alias{cursor_bookmark}
\alias{cursor_end}
\alias{cursor_reach}
\alias{cursor_reach_test}
\alias{cursor_forward}
\alias{cursor_backward}
\title{Set cursor in a 'Word' document}
\usage{
cursor_begin(x)

cursor_bookmark(x, id)

cursor_end(x)

cursor_reach(x, keyword, fixed = FALSE)

cursor_reach_test(x, keyword)

cursor_forward(x)

cursor_backward(x)
}
\arguments{
\item{x}{a docx device}

\item{id}{bookmark id}

\item{keyword}{keyword to look for as a regular expression}

\item{fixed}{logical. If TRUE, pattern is a string to be matched as is.}
}
\description{
A set of functions is available to manipulate
the position of a virtual cursor. This cursor will be used when
inserting, deleting or updating elements in the document.
}
\section{cursor_begin}{

Set the cursor at the beginning of the document, on the first element
of the document (usually a paragraph or a table).
}

\section{cursor_bookmark}{

Set the cursor at a bookmark that has previously been set.
}

\section{cursor_end}{

Set the cursor at the end of the document, on the last element
of the document.
}

\section{cursor_reach}{

Set the cursor on the first element of the document
that contains text specified in argument \code{keyword}.
The argument \code{keyword} is a regexpr pattern.
}

\section{cursor_reach_test}{

Test if an expression has a match in the document
that contains text specified in argument \code{keyword}.
The argument \code{keyword} is a regexpr pattern.
}

\section{cursor_forward}{

Move the cursor forward, it increments the cursor in the document.
}

\section{cursor_backward}{

Move the cursor backward, it decrements the cursor in the document.
}

\examples{
library(officer)

# create a template ----
doc <- read_docx()
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "Hello text to replace")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "blah blah blah")
doc <- body_add_par(doc, "Hello text to replace")
doc <- body_add_par(doc, "blah blah blah")
template_file <- print(
  x = doc,
  target = tempfile(fileext = ".docx")
)

# replace all pars containing "to replace" ----
doc <- read_docx(path = template_file)
while (cursor_reach_test(doc, "to replace")) {
  doc <- cursor_reach(doc, "to replace")

  doc <- body_add_fpar(
    x = doc,
    pos = "on",
    value = fpar(
      "Here is a link: ",
      hyperlink_ftext(
        text = "yopyop",
        href = "https://cran.r-project.org/"
      )
    )
  )
}

doc <- cursor_end(doc)
doc <- body_add_par(doc, "Yap yap yap yap...")

result_file <- print(
  x = doc,
  target = tempfile(fileext = ".docx")
)

# cursor_bookmark ----

doc <- read_docx()
doc <- body_add_par(doc, "centered text", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
doc <- body_add_par(doc, "A title", style = "heading 1")
doc <- body_add_par(doc, "Hello world!", style = "Normal")
doc <- cursor_bookmark(doc, "text_to_replace")
doc <- body_add_table(doc, value = iris, style = "table_template")

print(doc, target = tempfile(fileext = ".docx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_section}
\alias{block_section}
\title{Section for 'Word'}
\usage{
block_section(property)
}
\arguments{
\item{property}{section properties defined with function \link{prop_section}}
}
\description{
Create a representation of a section.

A section affects preceding paragraphs or tables; i.e.
a section starts at the end of the previous section (or the beginning of
the document if no preceding section exists), and stops where the
section is declared.

When a new landscape section is needed, it is recommended to add a block_section
with \code{type = "continuous"}, to add the content to be appened in the new section
and finally to add a block_section with \code{page_size = page_size(orient = "landscape")}.
}
\examples{
ps <- prop_section(
  page_size = page_size(orient = "landscape"),
  page_margins = page_mar(top = 2),
  type = "continuous"
)
block_section(ps)
}
\seealso{
Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_gg}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{set_doc_properties}
\alias{set_doc_properties}
\title{Set document properties}
\usage{
set_doc_properties(
  x,
  title = NULL,
  subject = NULL,
  creator = NULL,
  description = NULL,
  created = NULL,
  hyperlink_base = NULL,
  ...,
  values = NULL
)
}
\arguments{
\item{x}{an rdocx or rpptx object}

\item{title, subject, creator, description}{text fields}

\item{created}{a date object}

\item{hyperlink_base}{a string specifying the base URL for relative hyperlinks
in the document (only for rdocx).}

\item{...}{named arguments (names are field names), each element is a single
character value specifying value associated with the corresponding field name.
These pairs of \emph{key-value} are added as custom properties. If a value is
\code{NULL} or \code{NA}, the corresponding field is set to '' in the document properties.}

\item{values}{a named list (names are field names), each element is a single
character value specifying value associated with the corresponding field name.
If \code{values} is provided, argument \code{...} will be ignored.}
}
\description{
set Word or PowerPoint document properties. These are not visible
in the document but are available as metadata of the document.

Any character property can be added as a document property.
It provides an easy way to insert arbitrary fields. Given the challenges
that can be encountered with find-and-replace in word with officer, the
use of document fields and quick text fields provides a much more robust
approach to automatic document generation from R.
}
\note{
The "last modified" and "last modified by" fields will be automatically be updated
when the file is written.
}
\examples{
x <- read_docx()
x <- set_doc_properties(x, title = "title",
  subject = "document subject", creator = "Me me me",
  description = "this document is empty",
  created = Sys.time(),
  yoyo = "yok yok",
  glop = "pas glop")
x <- doc_properties(x)
}
\seealso{
Other functions for Word document informations: 
\code{\link{doc_properties}()},
\code{\link{docx_bookmarks}()},
\code{\link{docx_dim}()},
\code{\link{length.rdocx}()},
\code{\link{styles_info}()}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shorcuts.R
\name{shortcuts}
\alias{shortcuts}
\title{shortcuts for formatting properties}
\usage{
shortcuts
}
\description{
Shortcuts for \code{fp_text}, \code{fp_par},
\code{fp_cell} and \code{fp_border}.
}
\examples{
shortcuts$fp_bold()
shortcuts$fp_italic()
shortcuts$b_null()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{str_encode_to_rtf}
\alias{str_encode_to_rtf}
\title{Encode UTF8 string to RTF}
\usage{
str_encode_to_rtf(z)
}
\arguments{
\item{z}{character vector to be converted}
}
\value{
character vector of results encoded to RTF
}
\description{
Convert strings to RTF valid codes.
}
\examples{
str_encode_to_rtf("Hello World")
}
\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_comment}
\alias{body_comment}
\title{Add comment in a 'Word' document}
\usage{
body_comment(x, cmt = ftext(""), author = "", date = "", initials = "")
}
\arguments{
\item{x}{an rdocx object}

\item{cmt}{a set of blocks to be used as comment content returned by
function \code{\link[=block_list]{block_list()}}.}

\item{author}{comment author.}

\item{date}{comment date}

\item{initials}{comment initials}
}
\description{
Add a comment at the cursor location. The comment
is added on the first run of text in the current paragraph.
}
\examples{
doc <- read_docx()
doc <- body_add_par(doc, "Paragraph")
doc <- body_comment(doc, block_list("This is a comment."))
docx_file <- print(doc, target = tempfile(fileext = ".docx"))
docx_comments(read_docx(docx_file))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{layout_properties}
\alias{layout_properties}
\title{Slide layout properties}
\usage{
layout_properties(x, layout = NULL, master = NULL)
}
\arguments{
\item{x}{an \code{rpptx} object}

\item{layout}{slide layout name. If \code{NULL}, returns all layouts.}

\item{master}{master layout name where \code{layout} is located. If \code{NULL}, returns all masters.}
}
\value{
Returns a data frame with one row per placeholder and the following columns:
\itemize{
\item \code{master_name}: Name of master (a \code{.pptx} file may have more than one)
\item \code{name}: Name of layout
\item \code{type}: Placeholder type
\item \code{type_idx}: Running index for phs of the same type. Ordering by ph position
(top -> bottom, left -> right)
\item \code{id}: A unique placeholder id (assigned by PowerPoint automatically, starts at 2, potentially non-consecutive)
\item \code{ph_label}: Placeholder label (can be set by the user in PowerPoint)
\item \code{ph}: Placholder XML fragment (usually not needed)
\item \code{offx},\code{offy}: placeholder's distance from left and top edge (in inch)
\item \code{cx},\code{cy}: width and height of placeholder (in inch)
\item \code{rotation}: rotation in degrees
\item \code{fld_id} is generally stored as a hexadecimal or GUID value
\item \code{fld_type}: a unique identifier for a particular field
}
}
\description{
Detailed information about the placeholders on the slide layouts (label, position, etc.).
See \emph{Value} section below for more info.
}
\examples{
library(officer)

x <- read_pptx()
layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
layout_properties(x = x, master = "Office Theme")
layout_properties(x = x, layout = "Two Content")
layout_properties(x = x)
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{docx_body_xml}
\alias{docx_body_xml}
\title{Body xml document}
\usage{
docx_body_xml(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
Get the body document as xml. This function
is not to be used by end users, it has been implemented
to allow other packages to work with officer.
}
\examples{
doc <- read_docx()
docx_body_xml(doc)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pack_folder.R
\name{pack_folder}
\alias{pack_folder}
\title{compress a folder}
\usage{
pack_folder(folder, target)
}
\arguments{
\item{folder}{folder to compress}

\item{target}{path of the archive to create}
}
\description{
compress a folder to a target file. The
function returns the complete path to target file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_toc}
\alias{block_toc}
\title{Table of content for 'Word'}
\usage{
block_toc(level = 3, style = NULL, seq_id = NULL, separator = ";")
}
\arguments{
\item{level}{max title level of the table}

\item{style}{optional. If not NULL, its value is used as style in the
document that will be used to build entries of the TOC.}

\item{seq_id}{optional. If not NULL, its value is used as sequence
identifier in the document that will be used to build entries of the
TOC. See also \code{\link[=run_autonum]{run_autonum()}} to specify a sequence identifier.}

\item{separator}{unused, no effect}
}
\description{
Create a representation of a table
of content for Word documents.
}
\examples{
block_toc(level = 2)
block_toc(style = "Table Caption")
}
\seealso{
Other block functions for reporting:
\code{\link[=block_caption]{block_caption()}},
\code{\link[=block_gg]{block_gg()}},
\code{\link[=block_list]{block_list()}},
\code{\link[=block_list_items]{block_list_items()}},
\code{\link[=block_pour_docx]{block_pour_docx()}},
\code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}},
\code{\link[=fpar]{fpar()}},
\code{\link[=list_item]{list_item()}},
\code{\link[=plot_instr]{plot_instr()}},
\code{\link[=unordered_list]{unordered_list()}}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{plot_layout_properties}
\alias{plot_layout_properties}
\title{Slide layout properties plot}
\usage{
plot_layout_properties(
  x,
  layout = NULL,
  master = NULL,
  slide_idx = NULL,
  labels = TRUE,
  title = TRUE,
  type = TRUE,
  id = TRUE,
  cex = c(labels = 0.5, type = 0.5, id = 0.5),
  legend = FALSE
)
}
\arguments{
\item{x}{an \code{rpptx} object}

\item{layout}{slide layout name or numeric index (row index from \code{\link[=layout_summary]{layout_summary()}}. If \code{NULL} (default), it plots
the current slide's layout or the default layout (if set and there are not slides yet).}

\item{master}{master layout name where \code{layout} is located. Can be omitted if layout is unambiguous.}

\item{slide_idx}{Numeric slide index (default \code{NULL}) to specify which slide’s layout should be plotted.}

\item{labels}{if \code{TRUE} (default), adds placeholder labels (centered in \emph{red}).}

\item{title}{if \code{TRUE} (default), adds a title with the layout and master name (latter in square brackets) at the top.}

\item{type}{if \code{TRUE} (default), adds the placeholder type and its index (in square brackets)
in the upper left corner (in \emph{blue}).}

\item{id}{if \code{TRUE} (default), adds the placeholder's unique \code{id} (see column \code{id} from
\code{\link[=layout_properties]{layout_properties()}}) in the upper right corner (in \emph{green}).}

\item{cex}{List or vector to specify font size for \code{labels}, \code{type}, and \code{id}. Default is
\code{c(labels = .5, type = .5, id = .5)}. See \code{\link[graphics:text]{graphics::text()}} for details on how \code{cex} works.
Matching by position and partial name matching is supported. A single numeric value will apply to all
three parameters.}

\item{legend}{Add a legend to the plot (default \code{FALSE}).}
}
\description{
Plot slide layout properties into corresponding placeholders.
This can be useful to help visualize placeholders locations and identifiers.
\emph{All} information in the plot stems from the \code{\link[=layout_properties]{layout_properties()}} output.
See \emph{Details} section for more info.
}
\details{
The plot contains all relevant information to reference a placeholder via the \verb{ph_location_*}
function family:
\itemize{
\item \code{label}: ph label (red, center) to be used in \code{\link[=ph_location_label]{ph_location_label()}}.
\emph{NB}: The label can be assigned by the user in PowerPoint.
\item \code{type[idx]}: ph type + type index in brackets (blue, upper left) to be used in \code{\link[=ph_location_type]{ph_location_type()}}.
\emph{NB}: The index is consecutive and is sorted by ph position (top -> bottom, left -> right).
\item \code{id}: ph id (green, upper right) to be used in \code{ph_location_id()} (forthcoming).
\emph{NB}: The id is set by PowerPoint automatically and lack a meaningful order.
}
}
\examples{
x <- read_pptx()

# select layout explicitly
plot_layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
plot_layout_properties(x = x, layout = "Title Slide") # no master needed if layout name unique
plot_layout_properties(x = x, layout = 1) # use layout index instead of name

# plot default layout if one is set
x <- layout_default(x, "Title and Content")
plot_layout_properties(x)

# plot current slide's layout (default if no layout is passed)
x <- add_slide(x, "Title Slide")
plot_layout_properties(x)

# specify which slide's layout to plot by index
plot_layout_properties(x, slide_idx = 1)

# change appearance: what to show, font size, legend etc.
plot_layout_properties(
  x,
  layout = "Two Content",
  title = FALSE,
  type = FALSE,
  id = FALSE
)
plot_layout_properties(x, layout = 4, cex = c(labels = .8, id = .7, type = .7))
plot_layout_properties(x, 1, legend = TRUE)
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{fpar}
\alias{fpar}
\alias{update.fpar}
\title{Formatted paragraph}
\usage{
fpar(..., fp_p = fp_par(), fp_t = fp_text_lite(), values = NULL)

\method{update}{fpar}(object, fp_p = NULL, fp_t = NULL, ...)
}
\arguments{
\item{...}{cot objects (\code{\link[=ftext]{ftext()}}, \code{\link[=external_img]{external_img()}})}

\item{fp_p}{paragraph formatting properties, see \code{\link[=fp_par]{fp_par()}}}

\item{fp_t}{default text formatting properties. This is used as
text formatting properties when simple text is provided as argument,
see \code{\link[=fp_text]{fp_text()}}.}

\item{values}{a list of cot objects. If provided, argument \code{...} will be ignored.}

\item{object}{fpar object}
}
\description{
Create a paragraph representation by concatenating
formatted text or images. The result can be inserted in a Word document
or a PowerPoint presentation and can also be inserted in a \code{\link[=block_list]{block_list()}}
call.

All its arguments will be concatenated to create a paragraph where chunks of
text and images are associated with formatting properties.

\code{fpar} supports \code{\link[=ftext]{ftext()}}, \code{\link[=external_img]{external_img()}}, \code{run_*} functions
(i.e. \code{\link[=run_autonum]{run_autonum()}}, \code{\link[=run_word_field]{run_word_field()}}) when output is Word, and simple strings.

Default text and paragraph formatting properties can also be modified
with function \code{update()}.
}
\examples{
fpar(ftext("hello", shortcuts$fp_bold()))

# mix text and image -----
img.file <- file.path( R.home("doc"), "html", "logo.jpg" )

bold_face <- shortcuts$fp_bold(font.size = 12)
bold_redface <- update(bold_face, color = "red")
fpar_1 <- fpar(
  "Hello World, ",
  ftext("how ", prop = bold_redface ),
  external_img(src = img.file, height = 1.06/2, width = 1.39/2),
  ftext(" you?", prop = bold_face ) )
fpar_1

img_in_par <- fpar(
  external_img(src = img.file, height = 1.06/2, width = 1.39/2),
  fp_p = fp_par(text.align = "center") )
}
\seealso{
\code{\link[=block_list]{block_list()}}, \code{\link[=body_add_fpar]{body_add_fpar()}}, \code{\link[=ph_with]{ph_with()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{hyperlink_ftext}
\alias{hyperlink_ftext}
\title{Formatted chunk of text with hyperlink}
\usage{
hyperlink_ftext(text, prop = NULL, href)
}
\arguments{
\item{text}{text value, a single character value}

\item{prop}{formatting text properties returned by \link{fp_text}. It also can be NULL in
which case, no formatting is defined (the default is applied).}

\item{href}{URL value}
}
\description{
Format a chunk of text with text formatting properties (bold, color, ...),
the chunk is associated with an hyperlink.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
ft <- fp_text(font.size = 12, bold = TRUE)
hyperlink_ftext(
  href = "https://cran.r-project.org/index.html",
  text = "some text", prop = ft
)
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_tab}
\alias{fp_tab}
\title{Tabulation mark properties object}
\usage{
fp_tab(pos, style = "decimal")
}
\arguments{
\item{pos}{Specifies the position of the tab stop (in inches).}

\item{style}{style of the tab. Possible values are:
"decimal", "left", "right" or "center".}
}
\description{
create a tabulation mark properties setting object for Word
or RTF. Results can be used as arguments of \code{\link[=fp_tabs]{fp_tabs()}}.

Once tabulation marks settings are defined, tabulation marks can
be added with \code{\link[=run_tab]{run_tab()}} inside a call to \code{\link[=fpar]{fpar()}} or
with \verb{\\t} within 'flextable' content.
}
\examples{
fp_tab(pos = 0.4, style = "decimal")
fp_tab(pos = 1, style = "right")
}
\seealso{
Other functions for defining formatting properties: 
\code{\link{fp_border}()},
\code{\link{fp_cell}()},
\code{\link{fp_par}()},
\code{\link{fp_tabs}()},
\code{\link{fp_text}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{layout_summary}
\alias{layout_summary}
\title{Presentation layouts summary}
\usage{
layout_summary(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Get information about slide layouts and
master layouts into a data.frame. This function returns
a data.frame containing all layout and master names.
}
\examples{
my_pres <- read_pptx()
layout_summary ( x = my_pres )
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_id}
\alias{ph_location_id}
\title{Location of a placeholder based on its id}
\usage{
ph_location_id(id, newlabel = NULL, ...)
}
\arguments{
\item{id}{placeholder id.}

\item{newlabel}{a new label to associate with the placeholder.}

\item{...}{not used.}
}
\description{
Each placeholder has an id (a low integer value). The ids are unique across a single
layout. The function uses the placeholder's id to reference it. Different from a ph label,
the id is auto-assigned by PowerPoint and cannot be modified by the user.
Use \code{\link[=layout_properties]{layout_properties()}} (column \code{id}) and \code{\link[=plot_layout_properties]{plot_layout_properties()}} (upper right
corner, in green) to find a placeholder's id.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Comparison")
plot_layout_properties(doc, "Comparison")

doc <- ph_with(doc, "The Title", location = ph_location_id(id = 2)) # title
doc <- ph_with(doc, "Left Header", location = ph_location_id(id = 3)) # left header
doc <- ph_with(doc, "Left Content", location = ph_location_id(id = 4)) # left content
doc <- ph_with(doc, "The Footer", location = ph_location_id(id = 8)) # footer

file <- tempfile(fileext = ".pptx")
print(doc, file)

## file.show(file) # may not work on your system
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_layout_helper.R
\name{get_layout}
\alias{get_layout}
\title{Layout selection helper}
\usage{
get_layout(x, layout, master = NULL)
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{layout}{Layout name or index. Index refers to the row index of the \code{\link[=layout_summary]{layout_summary()}}
output.}

\item{master}{Name of master. Only required if layout name is not unique across masters.}
}
\value{
A \verb{<layout_info>} object, i.e. a list with the entries \code{index}, \code{layout_name},
\code{layout_file}, \code{master_name}, \code{master_file}, and \code{slide_layout}.
}
\description{
Select a layout by name or index. The master name is inferred and only required
for disambiguation in case the layout name is not unique across masters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr_utils.R
\name{get_reference_value}
\alias{get_reference_value}
\title{Get the document being used as a template}
\usage{
get_reference_value(format = NULL)
}
\arguments{
\item{format}{document format, one of 'pptx', 'docx' or 'html'}
}
\value{
a name file
}
\description{
Get filename of the document
being used as a template in an R Markdown document
rendered as HTML, PowerPoint presentation or Word document. It requires
packages rmarkdown >= 1.10.14 and knitr.
}
\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_tab}
\alias{run_tab}
\title{Tab for 'Word'}
\usage{
run_tab()
}
\description{
Object representing a tab
in a Word document. The result must be used
within a call to \link{fpar}. It will only have
effects in Word output.

Tabulation marks settings can be defined with \code{\link[=fp_tabs]{fp_tabs()}} in
paragraph settings defined with \code{\link[=fp_par]{fp_par()}}.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
z <- fp_tabs(
  fp_tab(pos = 0.5, style = "decimal"),
  fp_tab(pos = 1.5, style = "decimal")
)
par1 <- fpar(
  run_tab(), ftext("88."),
  run_tab(), ftext("987.45"),
  fp_p = fp_par(
    tabs = z
  )
)
par2 <- fpar(
  run_tab(), ftext("8."),
  run_tab(), ftext("670987.45"),
  fp_p = fp_par(
    tabs = z
  )
)
x <- read_docx()
x <- body_add(x, par1)
x <- body_add(x, par2)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_notes.R
\name{notes_location_type}
\alias{notes_location_type}
\title{Location of a placeholder for notes}
\usage{
notes_location_type(type = "body", ...)
}
\arguments{
\item{type}{placeholder label of the used notes master}

\item{...}{unused arguments}
}
\description{
The function will use the type name of the placeholder (e.g. body, hdr),
to find the corresponding location.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{page_size}
\alias{page_size}
\title{Page size object}
\usage{
page_size(
  width = 21/2.54,
  height = 29.7/2.54,
  orient = "portrait",
  unit = "in"
)
}
\arguments{
\item{width, height}{page width, page height (in inches).}

\item{orient}{page orientation, either 'landscape', either 'portrait'.}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}
}
\description{
The function creates a representation of the dimensions of a page. The
dimensions are defined by length, width and orientation. If the orientation is
in landscape mode then the length becomes the width and the width becomes the length.
}
\examples{
page_size(orient = "landscape")
}
\seealso{
Other functions for section definition: 
\code{\link{page_mar}()},
\code{\link{prop_section}()},
\code{\link{section_columns}()}
}
\concept{functions for section definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_break}
\alias{body_add_break}
\title{Add a page break in a 'Word' document}
\usage{
body_add_break(x, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add a page break into an rdocx object
}
\examples{
doc <- read_docx()
doc <- body_add_break(doc)
print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()},
\code{\link{body_append_start_context}()},
\code{\link{body_import_docx}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_template}
\alias{ph_location_template}
\title{Location for a placeholder based on a template}
\usage{
ph_location_template(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  type = NULL,
  id = 1,
  ...
)
}
\arguments{
\item{left, top, width, height}{place holder coordinates
in inches.}

\item{newlabel}{a label for the placeholder. See section details.}

\item{type}{placeholder type to look for in the slide layout, one
of 'body', 'title', 'ctrTitle', 'subTitle', 'dt', 'ftr', 'sldNum'.
It will be used as a template placeholder.}

\item{id}{index of the placeholder template. If two body placeholder, there can be
two different index: 1 and 2 for the first and second body placeholders defined
in the layout.}

\item{...}{unused arguments}
}
\description{
The function will return a list that complies with
expected format for argument \code{location} of function
\code{ph_with}. A placeholder will be used as template
and its positions will be updated with values \code{left}, \code{top}, \code{width}, \code{height}.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the Selection Pane of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(doc, "Title",
  location = ph_location_type(type = "title") )
doc <- ph_with(doc, "Hello world",
    location = ph_location_template(top = 4, type = "title") )
print(doc, target = tempfile(fileext = ".pptx") )
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{ftext}
\alias{ftext}
\title{Formatted chunk of text}
\usage{
ftext(text, prop = NULL)
}
\arguments{
\item{text}{text value, a single character value}

\item{prop}{formatting text properties returned by \link{fp_text}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Format a chunk of text with text formatting properties (bold, color, ...).
The function allows you to create pieces of text formatted the way you want.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
ftext("hello", fp_text())

properties1 <- fp_text(color = "red")
properties2 <- fp_text(bold = TRUE, shading.color = "yellow")
ftext1 <- ftext("hello", properties1)
ftext2 <- ftext("World", properties2)
paragraph <- fpar(ftext1, " ", ftext2)

x <- read_docx()
x <- body_add(x, paragraph)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
\link{fp_text}

Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml.R
\name{solid_fill}
\alias{solid_fill}
\title{Solid fill DrawingML fragment}
\usage{
solid_fill(color)
}
\arguments{
\item{color}{a color value: a name (\code{"red"}), a hex string with or
without leading \verb{#} (\code{"#FF0000"} or \code{"FF0000"}), or \code{"transparent"}.
Anything accepted by \code{\link[grDevices:col2rgb]{grDevices::col2rgb()}} works; the alpha channel
is honored.}
}
\value{
a character string with the OOXML fragment, namespaced \verb{a:}.
}
\description{
Build an \verb{<a:solidFill>} block with an sRGB color and an
alpha channel. Used inside DrawingML shape properties wrappers such
as \verb{<a:spPr>}, \verb{<a:ln>}, \verb{<a:rPr>}, and chartEx-specific wrappers
such as \verb{<cx:spPr>}.

This is a low-level helper intended for packages that emit OOXML
fragments (e.g. \code{mschart}, \code{rvg}). End users typically rely on
higher-level objects (\code{\link[=fp_text]{fp_text()}}, \code{\link[=fp_border]{fp_border()}}, \code{\link[=sp_line]{sp_line()}}) that
embed the \verb{<a:solidFill>} block themselves.
}
\examples{
solid_fill("red")
solid_fill("#3366CC")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_append_context.R
\name{body_append_start_context}
\alias{body_append_start_context}
\alias{write_elements_to_context}
\alias{body_append_stop_context}
\title{Fast Append context to a Word document}
\usage{
body_append_start_context(x, additional_ns = character())

write_elements_to_context(context, ...)

body_append_stop_context(context)
}
\arguments{
\item{x}{an rdocx object}

\item{additional_ns}{a named character vector of additional XML namespaces
to be added to the root node of the document. The names of the vector
are the namespace prefixes and the values are the namespace URIs.

This argument is useful when the elements to be added to the document require
additional namespaces that are not already present in the document and not
part of the xml generated by \code{to_wml()}. Simple users are not expected
to use this argument. It is mainly intended for developers of officer extensions.}

\item{context}{the context object created by \code{body_append_start_context()}.}

\item{...}{elements to be written to the context. These can be
paragraphs, tables, images, etc. The elements should have an associated
\code{to_wml()} method that converts them to WML format.}
}
\value{
\code{body_append_start_context()} returns a list representing the context
that contains:
\itemize{
\item \code{doc}: the original document object
\item \code{file_con}: the file connection to the context
\item \code{file_path}: the path to the context file
\item \code{final_str}: the final XML string to be appended to the document
later when calling \code{body_append_stop_context()}.
}

This object should not be modified by the user but instead
passed to \code{write_elements_to_context()} and \code{body_append_stop_context()}.

\code{write_elements_to_context()} returns the context object.

\code{body_append_stop_context()} returns the \code{rdocx} object with the
cursor position set to the end of the document.
}
\description{
This function is used to append content to a Word document
in a fast way.

It does not use the XML tree of the document
neither the cursor that is responsible for increasing the
performance of Word document generation when looping
over a large number of elements.

This function must be used with the \code{write_elements_to_context()}
and \code{body_append_stop_context()} functions:
\enumerate{
\item \code{body_append_start_context()} creates a context and returns
a list with the context and the file connection.
\item \code{write_elements_to_context()} writes the elements to the context
file connection.
\item \code{body_append_stop_context()} closes the file connection and
replaces the XML in the document with the new XML.
}
}
\examples{
library(officer)

doc <- read_docx()
doc <- body_add_par(doc, value = "blah blah blah", style = "Normal")

z <- body_append_start_context(doc)

for (i in seq_len(50)) {
  write_elements_to_context(
    context = z,
    fpar(
      "Hello World, ",
      i,
      fp_p = fp_par(word_style = "heading 1")
    ),
    fpar(run_pagebreak())
  )
}
doc <- body_append_stop_context(z)


print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content:
\code{\link[=body_add_blocks]{body_add_blocks()}},
\code{\link[=body_add_break]{body_add_break()}},
\code{\link[=body_add_caption]{body_add_caption()}},
\code{\link[=body_add_docx]{body_add_docx()}},
\code{\link[=body_add_fpar]{body_add_fpar()}},
\code{\link[=body_add_gg]{body_add_gg()}},
\code{\link[=body_add_img]{body_add_img()}},
\code{\link[=body_add_list]{body_add_list()}},
\code{\link[=body_add_par]{body_add_par()}},
\code{\link[=body_add_plot]{body_add_plot()}},
\code{\link[=body_add_table]{body_add_table()}},
\code{\link[=body_add_toc]{body_add_toc()}},
\code{\link[=body_import_docx]{body_import_docx()}}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{docx_dim}
\alias{docx_dim}
\title{'Word' page layout}
\usage{
docx_dim(x)
}
\arguments{
\item{x}{an \code{rdocx} object}
}
\description{
Get page width, page height and margins (in inches). The return values
are those corresponding to the section where the cursor is.
}
\examples{
docx_dim(read_docx())
}
\seealso{
Other functions for Word document informations: 
\code{\link{doc_properties}()},
\code{\link{docx_bookmarks}()},
\code{\link{length.rdocx}()},
\code{\link{set_doc_properties}()},
\code{\link{styles_info}()}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{slip_in_footnote}
\alias{slip_in_footnote}
\title{append a footnote}
\usage{
slip_in_footnote(x, ...)
}
\arguments{
\item{x}{an rdocx object}

\item{...}{unused}
}
\description{
this function is deprecated.
}
\note{
This function is now deprecated because it is not
efficient and make users write complex code. Use instead \code{\link[=fpar]{fpar()}} to build
formatted paragraphs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_manipulate.R
\name{ph_remove}
\alias{ph_remove}
\title{Remove a shape}
\usage{
ph_remove(x, type = "body", id = 1, ph_label = NULL, id_chr = NULL)
}
\arguments{
\item{x}{an rpptx object}

\item{type}{placeholder type}

\item{id}{placeholder index (integer) for a duplicated type. This is to be used when a placeholder
type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
Values can be read from \code{\link{slide_summary}}.}

\item{ph_label}{label associated to the placeholder. Use column
\code{ph_label} of result returned by \code{\link{slide_summary}}.
If used, \code{type} and \code{id} are ignored.}

\item{id_chr}{deprecated.}
}
\description{
Remove a shape in a slide.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
dummy_fun <- function(doc) {
  doc <- add_slide(doc,
    layout = "Two Content",
    master = "Office Theme"
  )
  doc <- ph_with(
    x = doc, value = "Un titre",
    location = ph_location_type(type = "title")
  )
  doc <- ph_with(
    x = doc, value = "Un corps 1",
    location = ph_location_type(type = "body", id = 1)
  )
  doc <- ph_with(
    x = doc, value = "Un corps 2",
    location = ph_location_type(type = "body", id = 2)
  )
  doc
}
doc <- read_pptx()
for (i in 1:3) {
  doc <- dummy_fun(doc)
}

doc <- on_slide(doc, index = 1)
doc <- ph_remove(x = doc, type = "title")

doc <- on_slide(doc, index = 2)
doc <- ph_remove(x = doc, type = "body", id = 2)

doc <- on_slide(doc, index = 3)
doc <- ph_remove(x = doc, type = "body", id = 1)

print(doc, target = fileout)
}
\seealso{
\code{\link{ph_with}}

Other functions for placeholders manipulation: 
\code{\link{ph_hyperlink}()},
\code{\link{ph_slidelink}()}
}
\concept{functions for placeholders manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_caption}
\alias{block_caption}
\title{Caption block}
\usage{
block_caption(label, style = NULL, autonum = NULL)
}
\arguments{
\item{label}{a scalar character representing label to display}

\item{style}{paragraph style name}

\item{autonum}{an object generated with function \link{run_autonum}}
}
\description{
Create a representation of a
caption that can be used for cross reference.
}
\examples{
library(officer)

run_num <- run_autonum(seq_id = "tab", pre_label = "tab. ",
  bkm = "mtcars_table")
caption <- block_caption("mtcars table",
  style = "Normal",
  autonum = run_num
)

doc_1 <- read_docx()
doc_1 <- body_add(doc_1, "A title", style = "heading 1")
doc_1 <- body_add(doc_1, "Hello world!", style = "Normal")
doc_1 <- body_add(doc_1, caption)
doc_1 <- body_add(doc_1, mtcars, style = "table_template")

print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other block functions for reporting: 
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fortify_pptx.R
\name{pptx_summary}
\alias{pptx_summary}
\title{PowerPoint content in a data.frame}
\usage{
pptx_summary(x, preserve = FALSE)
}
\arguments{
\item{x}{an rpptx object}

\item{preserve}{If \code{FALSE} (default), text in table cells is collapsed into a
single line. If \code{TRUE}, line breaks in table cells are preserved as a "\\n"
character. This feature is adapted from \code{docxtractr::docx_extract_tbl()}
published under a \href{https://github.com/hrbrmstr/docxtractr/blob/master/LICENSE}{MIT licensed} in
the 'docxtractr' package by Bob Rudis.}
}
\description{
Read content of a PowerPoint document and
return a dataset representing the document.
}
\examples{
example_pptx <- system.file(package = "officer",
  "doc_examples/example.pptx")
doc <- read_pptx(example_pptx)
pptx_summary(doc)
pptx_summary(example_pptx)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_img}
\alias{body_add_img}
\title{Add an image in a 'Word' document}
\usage{
body_add_img(x, src, style = NULL, width, height, pos = "after", unit = "in")
}
\arguments{
\item{x}{an rdocx object}

\item{src}{image filename, the basename of the file must not contain any blank.}

\item{style}{paragraph style}

\item{width, height}{image size in units expressed by the unit argument.
Defaults to "in"ches.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}
}
\description{
Add an image into an rdocx object.
}
\examples{
library(officer)

doc <- read_docx()

img.file <- file.path(R.home("doc"), "html", "logo.jpg")
if (file.exists(img.file)) {
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)

  # Set the unit in which the width and height arguments are expressed
  doc <- body_add_img(
    x = doc,
    src = img.file,
    height = 2.69,
    width = 3.53,
    unit = "cm"
  )
}

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()},
\code{\link{body_append_start_context}()},
\code{\link{body_import_docx}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{image_to_base64}
\alias{image_to_base64}
\title{Images to base64}
\usage{
image_to_base64(filepaths)
}
\arguments{
\item{filepaths}{file names.}
}
\description{
encodes images into base64 strings.
}
\examples{
rlogo <- file.path(R.home("doc"), "html", "logo.jpg")
base64_str <- image_to_base64(rlogo)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_section}
\alias{block_section}
\title{Section for 'Word'}
\usage{
block_section(property)
}
\arguments{
\item{property}{section properties defined with function \link{prop_section}}
}
\description{
Create a representation of a section.

A section affects preceding paragraphs or tables; i.e.
a section starts at the end of the previous section (or the beginning of
the document if no preceding section exists), and stops where the
section is declared.

When a new landscape section is needed, it is recommended to add a block_section
with \code{type = "continuous"}, to add the content to be appened in the new section
and finally to add a block_section with \code{page_size = page_size(orient = "landscape")}.
}
\examples{
ps <- prop_section(
  page_size = page_size(orient = "landscape"),
  page_margins = page_mar(top = 2),
  type = "continuous"
)
block_section(ps)
}
\seealso{
Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_comment}
\alias{run_comment}
\title{Comment for 'Word'}
\usage{
run_comment(
  cmt,
  run = ftext(""),
  author = "",
  date = "",
  initials = "",
  prop = NULL
)
}
\arguments{
\item{cmt}{a set of blocks to be used as comment content returned by
function \code{\link[=block_list]{block_list()}}.
the "run functions for reporting".}

\item{run}{a run object, made with a call to one of}

\item{author}{comment author.}

\item{date}{comment date}

\item{initials}{comment initials}

\item{prop}{formatting text properties returned by
\code{\link[=fp_text_lite]{fp_text_lite()}} or \code{\link[=fp_text]{fp_text()}}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Add a comment on a run object.
}
\examples{
fp_bold <- fp_text_lite(bold = TRUE)
fp_red <- fp_text_lite(color = "red")

bl <- block_list(
  fpar(ftext("Comment multiple words.", fp_bold)),
  fpar(
    ftext("Second line.", fp_red)
  )
)

comment1 <- run_comment(
  cmt = bl,
  run = ftext("with a comment"),
  author = "Author Me",
  date = Sys.Date(),
  initials = "AM"
)
par1 <- fpar("A paragraph ", comment1)

bl <- block_list(
  fpar(ftext("Comment a paragraph."))
)

comment2 <- run_comment(
  cmt = bl, run = ftext("A commented paragraph"),
  author = "Author You",
  date = Sys.Date(),
  initials = "AY"
)
par2 <- fpar(comment2)

doc <- read_docx()
doc <- body_add_fpar(doc, value = par1, style = "Normal")
doc <- body_add_fpar(doc, value = par2, style = "Normal")

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{page_mar}
\alias{page_mar}
\title{Page margins object}
\usage{
page_mar(
  bottom = 1417/1440,
  top = 1417/1440,
  right = 1417/1440,
  left = 1417/1440,
  header = 708/1440,
  footer = 708/1440,
  gutter = 0/1440
)
}
\arguments{
\item{bottom, top}{distance (in inches) between the bottom/top of the text
margin and the bottom/top of the page. The text is placed at the greater of
the value of this attribute and the extent of the header/footer text. A
negative value indicates that the content should be measured from the
bottom/top of the page regardless of the footer/header, and so will overlap
the footer/header. For example, \verb{header=-0.5, bottom=1} means that the footer
must start one inch from the bottom of the page and the main document text
must start a half inch from the bottom of the page. In this case, the text
and footer overlap since bottom is negative.}

\item{left, right}{distance (in inches) from the left/right edge of the page
to the left/right edge of the text.}

\item{header}{distance (in inches) from the top edge of the page to the top
edge of the header.}

\item{footer}{distance (in inches) from the bottom edge of the page to the
bottom edge of the footer.}

\item{gutter}{page gutter (in inches).}
}
\description{
Define margins for each page of a section.

The function creates a representation of the dimensions of a page. The
dimensions are defined by length, width and orientation. If the orientation
is in landscape mode then the length becomes the width and the width becomes
the length.
}
\examples{
page_mar()
}
\seealso{
Other functions for section definition:
\code{\link[=page_size]{page_size()}},
\code{\link[=prop_section]{prop_section()}},
\code{\link[=section_columns]{section_columns()}}
}
\concept{functions for section definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_manipulate.R
\name{ph_hyperlink}
\alias{ph_hyperlink}
\title{Hyperlink a placeholder}
\usage{
ph_hyperlink(x, type = "body", id = 1, id_chr = NULL, ph_label = NULL, href)
}
\arguments{
\item{x}{an rpptx object}

\item{type}{placeholder type}

\item{id}{placeholder index (integer) for a duplicated type. This is to be used when a placeholder
type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
Values can be read from \code{\link[=slide_summary]{slide_summary()}}.}

\item{id_chr}{deprecated.}

\item{ph_label}{label associated to the placeholder. Use column
\code{ph_label} of result returned by \code{\link[=slide_summary]{slide_summary()}}.
If used, \code{type} and \code{id} are ignored.}

\item{href}{hyperlink (do not forget http or https prefix)}
}
\description{
Add hyperlink to a placeholder in the current slide.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
loc_manual <- ph_location(bg = "red", newlabel = "mytitle")
doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(x = doc, "Un titre 1", location = loc_manual)
slide_summary(doc) # read column ph_label here
doc <- ph_hyperlink(
  x = doc, ph_label = "mytitle",
  href = "https://cran.r-project.org"
)

print(doc, target = fileout)
}
\seealso{
\code{\link[=ph_with]{ph_with()}}

Other functions for placeholders manipulation: 
\code{\link{ph_remove}()},
\code{\link{ph_slidelink}()}
}
\concept{functions for placeholders manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{length.rpptx}
\alias{length.rpptx}
\title{Number of slides}
\usage{
\method{length}{rpptx}(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Function \code{length} will return the number of slides.
}
\examples{
my_pres <- read_pptx()
my_pres <- add_slide(my_pres, "Title and Content")
my_pres <- add_slide(my_pres, "Title and Content")
length(my_pres)
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{layout_properties}
\alias{layout_properties}
\title{Slide layout properties}
\usage{
layout_properties(x, layout = NULL, master = NULL)
}
\arguments{
\item{x}{an \code{rpptx} object}

\item{layout}{slide layout name. If \code{NULL}, returns all layouts.}

\item{master}{master layout name where \code{layout} is located. If \code{NULL}, returns all masters.}
}
\value{
Returns a data frame with one row per placeholder and the following columns:
\itemize{
\item \code{master_name}: Name of master (a \code{.pptx} file may have more than one)
\item \code{name}: Name of layout
\item \code{type}: Placeholder type
\item \code{type_idx}: Running index for phs of the same type. Ordering by ph position
(top -> bottom, left -> right)
\item \code{id}: A unique placeholder id (assigned by PowerPoint automatically, starts at 2, potentially non-consecutive)
\item \code{ph_label}: Placeholder label (can be set by the user in PowerPoint)
\item \code{ph}: Placholder XML fragment (usually not needed)
\item \code{offx},\code{offy}: placeholder's distance from left and top edge (in inch)
\item \code{cx},\code{cy}: width and height of placeholder (in inch)
\item \code{rotation}: rotation in degrees
\item \code{fld_id} is generally stored as a hexadecimal or GUID value
\item \code{fld_type}: a unique identifier for a particular field
}
}
\description{
Detailed information about the placeholders on the slide layouts (label, position, etc.).
See \emph{Value} section below for more info.
}
\examples{
x <- read_pptx()
layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
layout_properties(x = x, master = "Office Theme")
layout_properties(x = x, layout = "Two Content")
layout_properties(x = x)
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{doc_properties}
\alias{doc_properties}
\title{Read document properties}
\usage{
doc_properties(x)
}
\arguments{
\item{x}{an \code{rdocx} or \code{rpptx} object}
}
\value{
a data.frame
}
\description{
Read Word or PowerPoint document properties
and get results in a data.frame.
}
\examples{
x <- read_docx()
doc_properties(x)
}
\seealso{
Other functions for Word document informations:
\code{\link[=docx_bookmarks]{docx_bookmarks()}},
\code{\link[=docx_dim]{docx_dim()}},
\code{\link[=length.rdocx]{length.rdocx()}},
\code{\link[=set_doc_properties]{set_doc_properties()}},
\code{\link[=styles_info]{styles_info()}}

Other functions for reading presentation information:
\code{\link[=annotate_base]{annotate_base()}},
\code{\link[=color_scheme]{color_scheme()}},
\code{\link[=layout_properties]{layout_properties()}},
\code{\link[=layout_summary]{layout_summary()}},
\code{\link[=length.rpptx]{length.rpptx()}},
\code{\link[=plot_layout_properties]{plot_layout_properties()}},
\code{\link[=slide_size]{slide_size()}},
\code{\link[=slide_summary]{slide_summary()}}
}
\concept{functions for Word document informations}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_wordtext}
\alias{run_wordtext}
\title{Word chunk of text with a style}
\usage{
run_wordtext(text, style_id = NULL)
}
\arguments{
\item{text}{text value, a single character value}

\item{style_id}{'Word' unique style identifier associated with the style to use.}
}
\description{
Format a chunk of text associated with a 'Word' character style.
The style is defined with its unique identifer.
}
\examples{
run1 <- run_wordtext("hello", "DefaultParagraphFont")
paragraph <- fpar(run1)

x <- read_docx()
x <- body_add_fpar(x, paragraph)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
\code{\link[=ftext]{ftext()}}

Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_pour_docx}
\alias{block_pour_docx}
\title{External Word document placeholder}
\usage{
block_pour_docx(file)
}
\arguments{
\item{file}{external docx file path}
}
\description{
Pour the content of a docx file in the resulting docx
from an 'R Markdown' document.
}
\examples{
library(officer)
docx <- tempfile(fileext = ".docx")
doc <- read_docx()
doc <- body_add(doc, iris[1:20,], style = "table_template")
print(doc, target = docx)

target <- tempfile(fileext = ".docx")
doc_1 <- read_docx()
doc_1 <- body_add(doc_1, block_pour_docx(docx))
print(doc_1, target = target)
}
\seealso{
Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_list}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_manipulate.R
\name{ph_slidelink}
\alias{ph_slidelink}
\title{Slide link to a placeholder}
\usage{
ph_slidelink(
  x,
  type = "body",
  id = 1,
  id_chr = NULL,
  ph_label = NULL,
  slide_index
)
}
\arguments{
\item{x}{an rpptx object}

\item{type}{placeholder type}

\item{id}{placeholder index (integer) for a duplicated type. This is to be used when a placeholder
type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
Values can be read from \code{\link[=slide_summary]{slide_summary()}}.}

\item{id_chr}{deprecated.}

\item{ph_label}{label associated to the placeholder. Use column
\code{ph_label} of result returned by \code{\link[=slide_summary]{slide_summary()}}.
If used, \code{type} and \code{id} are ignored.}

\item{slide_index}{slide index to reach}
}
\description{
Add slide link to a placeholder in the current slide.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
loc_title <- ph_location_type(type = "title")
doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(x = doc, "Un titre 1", location = loc_title)
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(x = doc, "Un titre 2", location = loc_title)
doc <- on_slide(doc, 1)
slide_summary(doc) # read column ph_label here
doc <- ph_slidelink(x = doc, ph_label = "Title 1", slide_index = 2)

print(doc, target = fileout)
}
\seealso{
\code{\link[=ph_with]{ph_with()}}

Other functions for placeholders manipulation: 
\code{\link{ph_hyperlink}()},
\code{\link{ph_remove}()}
}
\concept{functions for placeholders manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_pagebreak}
\alias{run_pagebreak}
\title{Page break for 'Word'}
\usage{
run_pagebreak()
}
\description{
Object representing a page break for a Word document.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
fp_t <- fp_text(font.size = 12, bold = TRUE)
an_fpar <- fpar("let's add a break page", run_pagebreak(), ftext("and blah blah!", fp_t))

x <- read_docx()
x <- body_add(x, an_fpar)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_properties.R
\name{sp_line}
\alias{sp_line}
\alias{print.sp_line}
\alias{update.sp_line}
\title{Line properties}
\usage{
sp_line(
  color = "transparent",
  lwd = 1,
  lty = "solid",
  linecmpd = "sng",
  lineend = "rnd",
  linejoin = "round",
  headend = sp_lineend(type = "none"),
  tailend = sp_lineend(type = "none")
)

\method{print}{sp_line}(x, ...)

\method{update}{sp_line}(
  object,
  color,
  lwd,
  lty,
  linecmpd,
  lineend,
  linejoin,
  headend,
  tailend,
  ...
)
}
\arguments{
\item{color}{line color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{lwd}{line width (in point) - 0 or positive integer value.}

\item{lty}{single character value specifying the line type.
Expected value is one of the following : default \code{'solid'}
or \code{'dot'} or \code{'dash'} or \code{'lgDash'}
or \code{'dashDot'} or \code{'lgDashDot'} or \code{'lgDashDotDot'}
or \code{'sysDash'} or \code{'sysDot'} or \code{'sysDashDot'}
or \code{'sysDashDotDot'}.}

\item{linecmpd}{single character value specifying the compound line type.
Expected value is one of the following : default \code{'sng'}
or \code{'dbl'} or \code{'tri'} or \code{'thinThick'}
or \code{'thickThin'}}

\item{lineend}{single character value specifying the line end style
Expected value is one of the following : default \code{'rnd'}
or \code{'sq'} or \code{'flat'}}

\item{linejoin}{single character value specifying the line join style
Expected value is one of the following : default \code{'round'}
or \code{'bevel'} or \code{'miter'}}

\item{headend}{a \code{sp_lineend} object specifying line head end style}

\item{tailend}{a \code{sp_lineend} object specifying line tail end style}

\item{x, object}{\code{sp_line} object}

\item{...}{further arguments - not used}
}
\value{
a \code{sp_line} object
}
\description{
Create a \code{sp_line} object that describes
line properties.
}
\examples{
library(officer)

sp_line()
sp_line(color = "red", lwd = 2)
sp_line(lty = "dot", linecmpd = "dbl")
print(sp_line(color = "red", lwd = 2))
obj <- sp_line(color = "red", lwd = 2)
update(obj, linecmpd = "dbl")
}
\seealso{
\link{sp_lineend}

Other functions for defining shape properties: 
\code{\link{sp_lineend}()}
}
\concept{functions for defining shape properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{set_doc_properties}
\alias{set_doc_properties}
\title{Set document properties}
\usage{
set_doc_properties(
  x,
  title = NULL,
  subject = NULL,
  creator = NULL,
  description = NULL,
  created = NULL,
  hyperlink_base = NULL,
  ...,
  values = NULL
)
}
\arguments{
\item{x}{an rdocx or rpptx object}

\item{title, subject, creator, description}{text fields}

\item{created}{a date object}

\item{hyperlink_base}{a string specifying the base URL for relative hyperlinks
in the document (only for rdocx).}

\item{...}{named arguments (names are field names), each element is a single
character value specifying value associated with the corresponding field name.
These pairs of \emph{key-value} are added as custom properties. If a value is
\code{NULL} or \code{NA}, the corresponding field is set to '' in the document properties.}

\item{values}{a named list (names are field names), each element is a single
character value specifying value associated with the corresponding field name.
If \code{values} is provided, argument \code{...} will be ignored.}
}
\description{
set Word or PowerPoint document properties. These are not visible
in the document but are available as metadata of the document.

Any character property can be added as a document property.
It provides an easy way to insert arbitrary fields. Given the challenges
that can be encountered with find-and-replace in word with officer, the
use of document fields and quick text fields provides a much more robust
approach to automatic document generation from R.
}
\note{
The "last modified" and "last modified by" fields will be automatically be updated
when the file is written.
}
\examples{
x <- read_docx()
x <- set_doc_properties(x, title = "title",
  subject = "document subject", creator = "Me me me",
  description = "this document is empty",
  created = Sys.time(),
  yoyo = "yok yok",
  glop = "pas glop")
x <- doc_properties(x)
}
\seealso{
Other functions for Word document informations:
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=docx_bookmarks]{docx_bookmarks()}},
\code{\link[=docx_dim]{docx_dim()}},
\code{\link[=length.rdocx]{length.rdocx()}},
\code{\link[=styles_info]{styles_info()}}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_bookmark}
\alias{run_bookmark}
\title{Bookmark for 'Word'}
\usage{
run_bookmark(bkm, run)
}
\arguments{
\item{bkm}{bookmark id to associate with run. Value can only be
made of alpha numeric characters, '-' and '_'.}

\item{run}{a run object, made with a call to one of
the "run functions for reporting".}
}
\description{
Add a bookmark on a run object.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
ft <- fp_text(font.size = 12, bold = TRUE)
run_bookmark("par1", ftext("some text", ft))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{str_encode_to_rtf}
\alias{str_encode_to_rtf}
\title{Encode UTF8 string to RTF}
\usage{
str_encode_to_rtf(z)
}
\arguments{
\item{z}{character vector to be converted}
}
\value{
character vector of results encoded to RTF
}
\description{
Convert strings to RTF valid codes.
}
\examples{
str_encode_to_rtf("Hello World")
}
\seealso{
Other functions for officer extensions:
\code{\link[=fortify_location]{fortify_location()}},
\code{\link[=get_reference_value]{get_reference_value()}},
\code{\link[=opts_current_table]{opts_current_table()}},
\code{\link[=shape_properties_tags]{shape_properties_tags()}},
\code{\link[=to_html]{to_html()}},
\code{\link[=to_pml]{to_pml()}},
\code{\link[=to_rtf]{to_rtf()}},
\code{\link[=to_wml]{to_wml()}},
\code{\link[=wml_link_images]{wml_link_images()}}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{table_width}
\alias{table_width}
\title{Preferred width for a table}
\usage{
table_width(width = 1, unit = "pct")
}
\arguments{
\item{width}{value of the preferred width of the table.}

\item{unit}{unit of the width. Possible values are 'in' (inches) and 'pct' (percent)}
}
\description{
Define the preferred width for a table.
}
\section{Word}{

All widths in a table are considered preferred because widths of
columns can conflict and the table layout rules can require a
preference to be overridden.
}

\seealso{
Other functions for table definition: 
\code{\link{prop_table}()},
\code{\link{table_colwidths}()},
\code{\link{table_conditional_formatting}()},
\code{\link{table_layout}()},
\code{\link{table_stylenames}()}
}
\concept{functions for table definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_xml}
\alias{body_add_xml}
\title{Add an xml string as document element}
\usage{
body_add_xml(x, str, pos = c("after", "before", "on"))
}
\arguments{
\item{x}{an rdocx object}

\item{str}{a wml string}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add an xml string as document element in the document. This function
is to be used to add custom openxml code.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr_utils.R
\name{get_reference_value}
\alias{get_reference_value}
\title{Get the document being used as a template}
\usage{
get_reference_value(format = NULL)
}
\arguments{
\item{format}{document format, one of 'pptx', 'docx' or 'html'}
}
\value{
a name file
}
\description{
Get filename of the document
being used as a template in an R Markdown document
rendered as HTML, PowerPoint presentation or Word document. It requires
packages rmarkdown >= 1.10.14 and knitr.
}
\seealso{
Other functions for officer extensions: 
\code{\link{docx_reference_img}()},
\code{\link{fortify_location}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_comments.R
\name{docx_comments}
\alias{docx_comments}
\title{Get comments in a Word document as a data.frame}
\usage{
docx_comments(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
return a data.frame representing the comments in a Word document.
}
\details{
Each row of the returned data frame contains data for one comment. The
columns contain the following information:
\itemize{
\item "comment_id" - unique comment id
\item "author" - name of the comment author
\item "initials" - initials of the comment author
\item "date" - timestamp of the comment
\item "text" - a list column of characters containing the comment text. Elements can
be vectors of length > 1 if a comment contains multiple paragraphs,
blocks or runs or of length 0 if the comment is empty.
\item "para_id" - a list column of characters containing the parent paragraph IDs.
Elememts can be vectors of length > 1 if a comment spans multiple paragraphs
or of length 0 if the comment has no parent paragraph.
\item "commented_text" - a list column of characters containing the
commented text. Elements can be vectors of length > 1 if a comment
spans multiple paragraphs or runs or of length 0 if the commented text is empty.
}
}
\examples{
library(officer)

bl <- block_list(
  fpar("Comment multiple words."),
  fpar("Second line")
)

a_par <- fpar(
  "This paragraph contains",
  run_comment(
    cmt = bl,
    run = ftext("a comment."),
    author = "Author Me",
    date = "2023-06-01"
  )
)

doc <- read_docx()
doc <- body_add_fpar(doc, value = a_par, style = "Normal")

docx_file <- print(doc, target = tempfile(fileext = ".docx"))

docx_comments(read_docx(docx_file))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{rtf_set_paragraph_style}
\alias{rtf_set_paragraph_style}
\title{Add or replace a paragraph style in an RTF document}
\usage{
rtf_set_paragraph_style(
  x,
  style_id,
  style_name = style_id,
  base_on = "Normal",
  fp_p = fp_par(),
  fp_t = NULL,
  outline_level = NULL
)
}
\arguments{
\item{x}{an rtf object created by \code{\link[=rtf_doc]{rtf_doc()}}.}

\item{style_id}{user-facing identifier used as the key when a paragraph
references the style (\code{rtf_add(..., style = style_id)}).}

\item{style_name}{display label of the style as it appears in Word's style
menu. Defaults to \code{style_id}.}

\item{base_on}{the \code{style_id} of the parent style. Properties not defined
in the new style are inherited from the parent. Defaults to \code{"Normal"}.}

\item{fp_p}{paragraph formatting properties, see \code{\link[=fp_par]{fp_par()}}.}

\item{fp_t}{text formatting properties, see \code{\link[=fp_text]{fp_text()}}. If \code{NULL} the
paragraph inherits text properties from the parent style.}

\item{outline_level}{integer between 1 and 9, or \code{NULL}. Sets the outline
level (\verb{\\\\outlinelevel}) so paragraphs using the style feed Word's
navigation pane and TOC field. Level 1 is the topmost (used by Heading 1).}
}
\value{
the rtf object with the style added or updated.
}
\description{
Add or replace a paragraph style in the document stylesheet
so that subsequent paragraphs can reference it via the \code{style} argument of
\code{\link[=rtf_add]{rtf_add()}}. The function mirrors \code{\link[=docx_set_paragraph_style]{docx_set_paragraph_style()}} for Word.
}
\examples{
doc <- rtf_doc()
doc <- rtf_set_paragraph_style(
  doc,
  style_id = "Callout",
  fp_p = fp_par(text.align = "center", padding = 6),
  fp_t = fp_text_lite(bold = TRUE, color = "#1F4E79")
)
doc <- rtf_add(doc, "Heads up", style = "Callout")
print(doc, target = tempfile(fileext = ".rtf"))
}
\seealso{
\code{\link[=docx_set_paragraph_style]{docx_set_paragraph_style()}}, \code{\link[=rtf_doc]{rtf_doc()}}, \code{\link[=rtf_add]{rtf_add()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{prop_section}
\alias{prop_section}
\title{Section properties}
\usage{
prop_section(
  page_size = NULL,
  page_margins = NULL,
  type = "continuous",
  section_columns = NULL,
  header_default = NULL,
  header_even = NULL,
  header_first = NULL,
  footer_default = NULL,
  footer_even = NULL,
  footer_first = NULL
)
}
\arguments{
\item{page_size}{page dimensions, an object generated with function \link{page_size}.}

\item{page_margins}{page margins, an object generated with function \link{page_mar}.
It is recommended to use the same margins object across all sections to prevent
unintended page breaks.}

\item{type}{Section type. It defines how the contents of the section will be
placed relative to the previous section. Available types are "continuous"
(begins the section on the next paragraph), "evenPage" (begins on the next
even-numbered page), "nextColumn" (begins on the next column on the page),
"nextPage" (begins on the following page), "oddPage" (begins on the next
odd-numbered page).}

\item{section_columns}{section columns, an object generated with function
\link{section_columns}. Use NULL (default value) for no content.}

\item{header_default}{content as a \code{\link[=block_list]{block_list()}} for the default page header.
Use NULL (default value) for no content.}

\item{header_even}{content as a \code{\link[=block_list]{block_list()}} for the even page header.
Use NULL (default value) for no content.}

\item{header_first}{content as a \code{\link[=block_list]{block_list()}} for the first page header.
Use NULL (default value) for no content.}

\item{footer_default}{content as a \code{\link[=block_list]{block_list()}} for the default page footer.
Use NULL (default value) for no content.}

\item{footer_even}{content as a \code{\link[=block_list]{block_list()}} for the even page footer. Use
NULL (default value) for no content.}

\item{footer_first}{content as a \code{\link[=block_list]{block_list()}} for the default page footer.
Use NULL (default value) for no content.}
}
\description{
A section is a grouping of blocks (ie. paragraphs and tables)
that have a set of properties that define pages on which the text will appear.

A Section properties object stores information about page composition,
such as page size, page orientation, borders and margins.

\strong{Important}: When creating multiple sections in a document, it is strongly
recommended to use the \strong{same \code{page_margins} object} for all sections to avoid
unwanted page breaks. Changing page margins between sections can cause Word to
insert automatic page breaks, even when using \code{type = "continuous"}. To ensure
consistent behavior, create a single \code{page_mar()} object and reuse it across
all \code{prop_section()} calls. See the examples in \code{\link[=body_end_block_section]{body_end_block_section()}} which
demonstrate this best practice.
}
\section{Illustrations}{


\if{html}{\figure{prop_section_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
library(officer)

# Example 1: Mixing different section layouts ----
# This example demonstrates how to create a document with multiple sections,
# each with different page orientations and column layouts

# Define a landscape section with single column
# This is useful for wide tables or charts
landscape_one_column <- block_section(
  prop_section(
    page_size = page_size(orient = "landscape"),
    type = "continuous"
  )
)

# Define a landscape section with two columns
# Useful for text-heavy content in landscape mode (e.g., newsletters)
landscape_two_columns <- block_section(
  prop_section(
    page_size = page_size(orient = "landscape"),
    type = "continuous",
    section_columns = section_columns(widths = c(4.75, 4.75))
  )
)

# Create a new document
doc_1 <- read_docx()

# Section 1: Landscape single column for wide table ----
# Add a title for the first section
doc_1 <- body_add_par(doc_1, "Wide Table Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "This table is displayed in landscape orientation to accommodate all columns."
)

# Add a wide table with multiple columns
doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")

# End the landscape single-column section
doc_1 <- body_end_block_section(doc_1, value = landscape_one_column)

# Section 2: Landscape two columns for text content ----
# Add a title for the two-column section
doc_1 <- body_add_par(doc_1, "Two-Column Text Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "The following text flows across two columns in landscape orientation."
)

# Add text content that will flow across two columns
doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 50), collapse = " "))

# End the landscape two-column section
doc_1 <- body_end_block_section(doc_1, value = landscape_two_columns)

# Section 3: Return to portrait orientation ----
# After ending the previous sections, we're back to portrait (default)
doc_1 <- body_add_par(doc_1, "Portrait Table Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "This section returns to portrait orientation with a taller table."
)

# Add a longer table in portrait orientation
doc_1 <- body_add_table(doc_1, value = mtcars[1:25, ], style = "table_template")

# Save the document
output_file_1 <- tempfile(fileext = ".docx")
print(doc_1, target = output_file_1)


# Example 2: Different headers and footers (first, even, odd pages) ----
# This example demonstrates the complete header/footer system with:
# - Different header/footer for the first page
# - Different header/footer for even pages (left-side pages in duplex printing)
# - Default header/footer for odd pages (right-side pages)

# Create sample text to generate multiple pages
lorem_text <- paste(
  rep("Purus lectus eros metus turpis mattis platea praesent sed. ", 50),
  collapse = ""
)

# Define content for FIRST page header
# Typically used for title pages or cover pages
header_first <- block_list(
  fpar(
    ftext(
      "First Page Header - Title Page",
      fp_text_lite(bold = TRUE, color = "#4472C4", font.size = 14)
    ),
    fp_p = fp_par(
      text.align = "center",
      padding.bottom = 12,
      border.bottom = fp_border(color = "#4472C4", width = 2)
    )
  )
)

# Define content for EVEN pages header (left-side pages when printed)
# In duplex printing, this appears on the left side
header_even <- block_list(
  fpar(
    ftext("Chapter Title", fp_text_lite(italic = TRUE, font.size = 10)),
    fp_p = fp_par(text.align = "left")
  )
)

# Define content for DEFAULT pages header (odd pages/right-side)
# In duplex printing, this appears on the right side
header_default <- block_list(
  fpar(
    ftext("Document Title", fp_text_lite(italic = TRUE, font.size = 10)),
    fp_p = fp_par(text.align = "right")
  )
)

# Define content for FIRST page footer
footer_first <- block_list(
  fpar(
    ftext(
      "Company Name - Confidential",
      fp_text_lite(font.size = 9, color = "#666666")
    ),
    fp_p = fp_par(text.align = "center")
  )
)

# Define content for EVEN pages footer (includes page number on left)
footer_even <- block_list(
  fpar(
    run_word_field(field = "PAGE", prop = fp_text_lite(font.size = 9)),
    ftext(" | Document Name", fp_text_lite(font.size = 9)),
    fp_p = fp_par(
      text.align = "left",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Define content for DEFAULT pages footer (includes page number on right)
footer_default <- block_list(
  fpar(
    ftext("Document Name | ", fp_text_lite(font.size = 9)),
    run_word_field(field = "PAGE", prop = fp_text_lite(font.size = 9)),
    fp_p = fp_par(
      text.align = "right",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Create section properties with all header/footer variants
# When all three are defined (first, even, default), Word will use:
# - header_first/footer_first for page 1
# - header_even/footer_even for pages 2, 4, 6, etc.
# - header_default/footer_default for pages 3, 5, 7, etc.
section_with_all_hf <- prop_section(
  header_default = header_default,
  footer_default = footer_default,
  header_first = header_first,
  footer_first = footer_first,
  header_even = header_even,
  footer_even = footer_even
)

# Create a new document
doc_2 <- read_docx()

# Add enough content to create multiple pages
# This will demonstrate how the different headers/footers appear
for (i in 1:20) {
  doc_2 <- body_add_par(doc_2, paste0("Paragraph ", i, ": ", lorem_text))
}

# Apply the section properties with all header/footer configurations
doc_2 <- body_set_default_section(doc_2, value = section_with_all_hf)

# Save the document
# Open this document and scroll through pages to see:
# - Page 1: Special first page header/footer
# - Page 2: Even page header/footer with page number on left
# - Page 3: Odd page (default) header/footer with page number on right
# - And so on...
output_file_2 <- tempfile(fileext = ".docx")
print(doc_2, target = output_file_2)
}
\seealso{
\link{block_section}

Other functions for section definition: 
\code{\link{page_mar}()},
\code{\link{page_size}()},
\code{\link{section_columns}()}
}
\concept{functions for section definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_break}
\alias{body_add_break}
\title{Add a page break in a 'Word' document}
\usage{
body_add_break(x, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
add a page break into an rdocx object
}
\examples{
doc <- read_docx()
doc <- body_add_break(doc)
print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{add_slide}
\alias{add_slide}
\title{Add a slide}
\usage{
add_slide(x, layout = NULL, master = NULL, ..., .dots = NULL)
}
\arguments{
\item{x}{an \code{rpptx} object.}

\item{layout}{slide layout name to use. Can be ommited of a default layout is set via \code{\link[=layout_default]{layout_default()}}.}

\item{master}{master layout name where \code{layout} is located. Only required in case of several masters if layout is
not unique.}

\item{...}{Key-value pairs of the form \code{"short form location" = object} passed to \link{phs_with}. See section
\code{"Short forms"} in \link{phs_with} for details, available short forms and examples.}

\item{.dots}{List of key-value pairs of the form \code{list("short form location" = object)}. Alternative to \code{...}. See
\link{phs_with} for details.}
}
\description{
Add a slide into a pptx presentation.
}
\examples{
x <- read_pptx()

layout_summary(x) # available layouts

x <- add_slide(x, layout = "Two Content")

x <- layout_default(x, "Title Slide") # set default layout for `add_slide()`
x <- add_slide(x) # uses default layout

# use `...` to fill placeholders when adding slide
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)
}
\seealso{
\code{\link[=print.rpptx]{print.rpptx()}}, \code{\link[=read_pptx]{read_pptx()}}, \code{\link[=layout_summary]{layout_summary()}}, \code{\link[=plot_layout_properties]{plot_layout_properties()}}, \code{\link[=ph_with]{ph_with()}}, \code{\link[=phs_with]{phs_with()}}, \code{\link[=layout_default]{layout_default()}}

Other functions to manipulate slides:
\code{\link{move_slide}()},
\code{\link{on_slide}()},
\code{\link{remove_slide}()},
\code{\link{set_notes}()}
}
\concept{slide_manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{ftext}
\alias{ftext}
\title{Formatted chunk of text}
\usage{
ftext(text, prop = NULL)
}
\arguments{
\item{text}{text value, a single character value}

\item{prop}{formatting text properties returned by \link{fp_text}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Format a chunk of text with text formatting properties (bold, color, ...).
The function allows you to create pieces of text formatted the way you want.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
ftext("hello", fp_text())

properties1 <- fp_text(color = "red")
properties2 <- fp_text(bold = TRUE, shading.color = "yellow")
ftext1 <- ftext("hello", properties1)
ftext2 <- ftext("World", properties2)
paragraph <- fpar(ftext1, " ", ftext2)

x <- read_docx()
x <- body_add(x, paragraph)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
\link{fp_text}

Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fortify_docx.R
\name{docx_summary}
\alias{docx_summary}
\title{Get Word content in a data.frame}
\usage{
docx_summary(x, preserve = FALSE, remove_fields = FALSE, detailed = FALSE)
}
\arguments{
\item{x}{an rdocx object}

\item{preserve}{If \code{FALSE} (default), text in table cells is collapsed into a
single line. If \code{TRUE}, line breaks in table cells are preserved as a "\\n"
character. This feature is adapted from \code{docxtractr::docx_extract_tbl()}
published under a \href{https://github.com/hrbrmstr/docxtractr/blob/master/LICENSE}{MIT licensed} in
the 'docxtractr' package by Bob Rudis.}

\item{remove_fields}{if TRUE, prevent field codes from appearing in the
returned data.frame.}

\item{detailed}{Should run-level information be included in the dataframe?
Defaults to \code{FALSE}. If \code{TRUE}, the dataframe contains detailed information
about each run (text formatting, images, hyperlinks, etc.) instead of
collapsing content at the paragraph level. When \code{FALSE}, run-level
information such as images, hyperlinks, and text formatting is not available
since data is aggregated at the paragraph level.}
}
\value{
A data.frame with the following columns depending on the value of \code{detailed}:

When \code{detailed = FALSE} (default), the data.frame contains:
\itemize{
\item \code{doc_index}: Document element index (integer).
\item \code{content_type}: Type of content: "paragraph" or "table cell" (character).
\item \code{style_name}: Name of the paragraph style (character).
\item \code{text}: Collapsed text content of the paragraph or cell (character).
\item \code{table_index}: Index of the table (integer). \code{NA} for non-table content.
\item \code{row_id}: Row position in table (integer). \code{NA} for non-table content.
\item \code{cell_id}: Cell position in table row (integer). \code{NA} for non-table content.
\item \code{is_header}: Whether the row is a table header (logical). \code{NA} for non-table content.
\item \code{row_span}: Number of rows spanned by the cell (integer). \code{0} for merged cells. \code{NA} for non-table content.
\item \code{col_span}: Number of columns spanned by the cell (character). \code{NA} for non-table content.
\item \code{table_stylename}: Name of the table style (character). \code{NA} for non-table content.
}

When \code{detailed = TRUE}, the data.frame contains additional run-level information:
\itemize{
\item \code{run_index}: Index of the run within the paragraph (integer).
\item \code{run_content_index}: Index of content element within the run (integer).
\item \code{run_content_text}: Text content of the run element (character).
\item \code{image_path}: Path to embedded image stored in the temporary directory
associated with the rdocx object (character).
Images should be copied to a permanent location before closing the R
session if needed.
\item \code{field_code}: Field code content (character).
\item \code{footnote_text}: Footnote text content (character).
\item \code{link}: Hyperlink URL (character).
\item \code{link_to_bookmark}: Internal bookmark anchor name for hyperlinks (character).
\item \code{bookmark_start}: Names of the bookmarks starting on this paragraph
(values are concatenated with '|').
\item \code{character_stylename}: Name of the character/run style (character).
\item \code{sz}: Font size in half-points (integer).
\item \code{sz_cs}: Complex script font size in half-points (integer).
\item \code{font_family_ascii}: Font family for ASCII characters (character).
\item \code{font_family_eastasia}: Font family for East Asian characters (character).
\item \code{font_family_hansi}: Font family for high ANSI characters (character).
\item \code{font_family_cs}: Font family for complex script characters (character).
\item \code{bold}: Whether the run is bold (logical).
\item \code{italic}: Whether the run is italic (logical).
\item \code{underline}: Whether the run is underlined (logical).
\item \code{color}: Text color in hexadecimal format (character).
\item \code{shading}: Shading pattern (character).
\item \code{shading_color}: Shading foreground color (character).
\item \code{shading_fill}: Shading background fill color (character).
\item \code{keep_with_next}: Whether paragraph should stay with next (logical).
\item \code{align}: Paragraph alignment (character).
\item \code{level}: Numbering level (integer). \code{NA} if not a numbered list.
\item \code{num_id}: Numbering definition ID (integer). \code{NA} if not a numbered list.
}
}
\description{
read content of a Word document and
return a data.frame representing the document.
}
\note{
Documents included with \code{\link[=body_add_docx]{body_add_docx()}} will
not be accessible in the results.
}
\examples{
library(officer)

example_docx <- system.file(
  package = "officer",
  "doc_examples/example.docx"
)
doc <- read_docx(example_docx)

docx_summary(doc)

docx_summary(doc, detailed = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{plot_layout_properties}
\alias{plot_layout_properties}
\title{Slide layout properties plot}
\usage{
plot_layout_properties(
  x,
  layout = NULL,
  master = NULL,
  slide_idx = NULL,
  labels = TRUE,
  title = TRUE,
  type = TRUE,
  id = TRUE,
  cex = c(labels = 0.5, type = 0.5, id = 0.5),
  legend = FALSE
)
}
\arguments{
\item{x}{an \code{rpptx} object}

\item{layout}{slide layout name or numeric index (row index from \code{\link[=layout_summary]{layout_summary()}}. If \code{NULL} (default), it plots
the current slide's layout or the default layout (if set and there are not slides yet).}

\item{master}{master layout name where \code{layout} is located. Can be omitted if layout is unambiguous.}

\item{slide_idx}{Numeric slide index (default \code{NULL}) to specify which slide’s layout should be plotted.}

\item{labels}{if \code{TRUE} (default), adds placeholder labels (centered in \emph{red}).}

\item{title}{if \code{TRUE} (default), adds a title with the layout and master name (latter in square brackets) at the top.}

\item{type}{if \code{TRUE} (default), adds the placeholder type and its index (in square brackets)
in the upper left corner (in \emph{blue}).}

\item{id}{if \code{TRUE} (default), adds the placeholder's unique \code{id} (see column \code{id} from
\code{\link[=layout_properties]{layout_properties()}}) in the upper right corner (in \emph{green}).}

\item{cex}{List or vector to specify font size for \code{labels}, \code{type}, and \code{id}. Default is
\code{c(labels = 0.5, type = 0.5, id = 0.5)}. See \code{\link[graphics:text]{graphics::text()}} for details on how \code{cex} works.
Matching by position and partial name matching is supported. A single numeric value will apply to all
three parameters.}

\item{legend}{Add a legend to the plot (default \code{FALSE}).}
}
\description{
Plot slide layout properties into corresponding placeholders.
This can be useful to help visualize placeholders locations and identifiers.
\emph{All} information in the plot stems from the \code{\link[=layout_properties]{layout_properties()}} output.
See \emph{Details} section for more info.
}
\details{
The plot contains all relevant information to reference a placeholder via the \verb{ph_location_*}
function family:
\itemize{
\item \code{label}: ph label (red, center) to be used in \code{\link[=ph_location_label]{ph_location_label()}}.
\emph{NB}: The label can be assigned by the user in PowerPoint.
\item \code{type[idx]}: ph type + type index in brackets (blue, upper left) to be used in \code{\link[=ph_location_type]{ph_location_type()}}.
\emph{NB}: The index is consecutive and is sorted by ph position (top -> bottom, left -> right).
\item \code{id}: ph id (green, upper right) to be used in \code{ph_location_id()} (forthcoming).
\emph{NB}: The id is set by PowerPoint automatically and lack a meaningful order.
}
}
\examples{
x <- read_pptx()

# select layout explicitly
plot_layout_properties(x = x, layout = "Title Slide", master = "Office Theme")
plot_layout_properties(x = x, layout = "Title Slide") # no master needed if layout name unique
plot_layout_properties(x = x, layout = 1) # use layout index instead of name

# plot default layout if one is set
x <- layout_default(x, "Title and Content")
plot_layout_properties(x)

# plot current slide's layout (default if no layout is passed)
x <- add_slide(x, "Title Slide")
plot_layout_properties(x)

# specify which slide's layout to plot by index
plot_layout_properties(x, slide_idx = 1)

# change appearance: what to show, font size, legend etc.
plot_layout_properties(
  x,
  layout = "Two Content",
  title = FALSE,
  type = FALSE,
  id = FALSE
)
plot_layout_properties(x, layout = 4, cex = c(labels = .8, id = .7, type = .7))
plot_layout_properties(x, 1, legend = TRUE)
}
\seealso{
Other functions for reading presentation information:
\code{\link[=annotate_base]{annotate_base()}},
\code{\link[=color_scheme]{color_scheme()}},
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=layout_properties]{layout_properties()}},
\code{\link[=layout_summary]{layout_summary()}},
\code{\link[=length.rpptx]{length.rpptx()}},
\code{\link[=slide_size]{slide_size()}},
\code{\link[=slide_summary]{slide_summary()}}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{to_html}
\alias{to_html}
\title{Convert officer objects to HTML}
\usage{
to_html(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{Arguments to be passed to methods}
}
\description{
Convert an object made with package officer
to HTML. The result is a string.
}
\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{plot_instr}
\alias{plot_instr}
\title{Wrap plot instructions for png plotting in Powerpoint or Word}
\usage{
plot_instr(code)
}
\arguments{
\item{code}{plotting instructions}
}
\description{
A simple wrapper to capture
plot instructions that will be executed and copied in a document. It produces
an object of class 'plot_instr' with a corresponding method \code{\link[=ph_with]{ph_with()}} and
\code{\link[=body_add_plot]{body_add_plot()}}.

The function enable usage of any R plot with argument \code{code}. Wrap your code
between curly bracket if more than a single expression.
}
\examples{
# plot_instr demo ----

anyplot <- plot_instr(code = {
  barplot(1:5, col = 2:6)
  })

doc <- read_docx()
doc <- body_add(doc, anyplot, width = 5, height = 4)
print(doc, target = tempfile(fileext = ".docx"))


doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(
  doc, anyplot,
  location = ph_location_fullsize(),
  bg = "#00000066", pointsize = 12)
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=body_add_plot]{body_add_plot()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_gg}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{docx_bookmarks}
\alias{docx_bookmarks}
\title{List Word bookmarks}
\usage{
docx_bookmarks(x)
}
\arguments{
\item{x}{an \code{rdocx} object}
}
\description{
List bookmarks id that can be found in a
'Word' document.
}
\examples{
library(officer)

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
doc_1 <- body_bookmark(doc_1, "text_to_replace_1")
doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
doc_1 <- body_bookmark(doc_1, "text_to_replace_2")

docx_bookmarks(doc_1)

docx_bookmarks(read_docx())
}
\seealso{
Other functions for Word document informations: 
\code{\link{doc_properties}()},
\code{\link{docx_dim}()},
\code{\link{length.rdocx}()},
\code{\link{set_doc_properties}()},
\code{\link{styles_info}()}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_list}
\alias{body_add_list}
\title{Add a list of items into a 'Word' document}
\usage{
body_add_list(x, items, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{items}{a \code{\link[=block_list_items]{block_list_items()}} object.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add a bullet or numbered list produced by
\code{\link[=block_list_items]{block_list_items()}} into an rdocx object.
}
\examples{
library(officer)

items <- block_list_items(
  list_item(fpar(ftext("Item 1", fp_text(color = "red"))), level = 1),
  list_item(fpar("Sub-item"), level = 2),
  list_item(fpar("Item 2"), level = 1),
  list_type = "bullet"
)

doc <- read_docx()
doc <- body_add_list(doc, items = items)
print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content:
\code{\link[=body_add_blocks]{body_add_blocks()}},
\code{\link[=body_add_break]{body_add_break()}},
\code{\link[=body_add_caption]{body_add_caption()}},
\code{\link[=body_add_docx]{body_add_docx()}},
\code{\link[=body_add_fpar]{body_add_fpar()}},
\code{\link[=body_add_gg]{body_add_gg()}},
\code{\link[=body_add_img]{body_add_img()}},
\code{\link[=body_add_par]{body_add_par()}},
\code{\link[=body_add_plot]{body_add_plot()}},
\code{\link[=body_add_table]{body_add_table()}},
\code{\link[=body_add_toc]{body_add_toc()}},
\code{\link[=body_append_start_context]{body_append_start_context()}},
\code{\link[=body_import_docx]{body_import_docx()}}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_docx}
\alias{body_add_docx}
\title{Add an external docx in a 'Word' document}
\usage{
body_add_docx(x, src, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{src}{docx filename, the path of the file must not contain
any '&' and the basename must not contain any space.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add content of a docx into an rdocx object.

The function is using a 'Microsoft Word' feature: when the
document will be edited, the content of the file will be
inserted in the main document.

This feature is unlikely to work as expected if the
resulting document is edited by another software.

The file is added when the method \code{print()} that
produces the final Word file is called, so don't remove
file defined with \code{src} before.
}
\examples{
file1 <- tempfile(fileext = ".docx")
file2 <- tempfile(fileext = ".docx")
file3 <- tempfile(fileext = ".docx")
x <- read_docx()
x <- body_add_par(x, "hello world 1", style = "Normal")
print(x, target = file1)

x <- read_docx()
x <- body_add_par(x, "hello world 2", style = "Normal")
print(x, target = file2)

x <- read_docx(path = file1)
x <- body_add_break(x)
x <- body_add_docx(x, src = file2)
print(x, target = file3)
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{hyperlink_ftext}
\alias{hyperlink_ftext}
\title{Formatted chunk of text with hyperlink}
\usage{
hyperlink_ftext(text, prop = NULL, href)
}
\arguments{
\item{text}{text value, a single character value}

\item{prop}{formatting text properties returned by \link{fp_text}. It also can be NULL in
which case, no formatting is defined (the default is applied).}

\item{href}{URL value}
}
\description{
Format a chunk of text with text formatting properties (bold, color, ...),
the chunk is associated with an hyperlink.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
ft <- fp_text(font.size = 12, bold = TRUE)
hyperlink_ftext(
  href = "https://cran.r-project.org/index.html",
  text = "some text", prop = ft
)
}
\seealso{
Other run functions for reporting:
\code{\link[=external_img]{external_img()}},
\code{\link[=floating_external_img]{floating_external_img()}},
\code{\link[=ftext]{ftext()}},
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_bookmark]{run_bookmark()}},
\code{\link[=run_columnbreak]{run_columnbreak()}},
\code{\link[=run_comment]{run_comment()}},
\code{\link[=run_footnote]{run_footnote()}},
\code{\link[=run_footnoteref]{run_footnoteref()}},
\code{\link[=run_linebreak]{run_linebreak()}},
\code{\link[=run_pagebreak]{run_pagebreak()}},
\code{\link[=run_reference]{run_reference()}},
\code{\link[=run_tab]{run_tab()}},
\code{\link[=run_word_field]{run_word_field()}},
\code{\link[=run_wordtext]{run_wordtext()}}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_matrix.R
\name{as.matrix.rpptx}
\alias{as.matrix.rpptx}
\title{PowerPoint table to matrix}
\usage{
\method{as.matrix}{rpptx}(
  x,
  ...,
  slide_id = NA_integer_,
  id = NA_character_,
  span = c(NA_character_, "fill")
)
}
\arguments{
\item{x}{The rpptx object to convert (as created by \code{\link[=read_pptx]{read_pptx()}})}

\item{...}{Ignored}

\item{slide_id}{The slide number to load from (NA indicates first slide with
a table, NULL indicates all slides and all tables)}

\item{id}{The table ID to load from (ignored it \code{is.null(slide_id)}, NA
indicates to load the first table from the \code{slide_id})}

\item{span}{How should col_span/row_span values be handled? \code{NA} means
to leave the value as \code{NA}, and \code{"fill"} means to fill matrix
cells with the value.}
}
\value{
A matrix with the data, or if \code{slide_id=NULL}, a list of
matrices
}
\description{
Convert the data in an a 'PowerPoint' table
to a matrix or all data to a list of matrices.
}
\examples{
library(officer)
pptx_file <- system.file(package="officer", "doc_examples", "example.pptx")
z <- read_pptx(pptx_file)
as.matrix(z, slide_id = NULL)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fortify_pptx.R
\name{pptx_summary}
\alias{pptx_summary}
\title{PowerPoint content in a data.frame}
\usage{
pptx_summary(x, preserve = FALSE)
}
\arguments{
\item{x}{an rpptx object}

\item{preserve}{If \code{FALSE} (default), text in table cells is collapsed into a
single line. If \code{TRUE}, line breaks in table cells are preserved as a "\\n"
character. This feature is adapted from \code{docxtractr::docx_extract_tbl()}
published under a \href{https://github.com/hrbrmstr/docxtractr/blob/master/LICENSE}{MIT licensed} in
the \code{{docxtractr}} package by Bob Rudis.}
}
\description{
Read content of a PowerPoint document and
return a dataset representing the document.
}
\examples{
example_pptx <- system.file(package = "officer",
  "doc_examples/example.pptx")
doc <- read_pptx(example_pptx)
pptx_summary(doc)
pptx_summary(example_pptx)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_id}
\alias{ph_location_id}
\title{Location of a placeholder based on its id}
\usage{
ph_location_id(id, newlabel = NULL, ...)
}
\arguments{
\item{id}{placeholder id.}

\item{newlabel}{a new label to associate with the placeholder.}

\item{...}{not used.}
}
\description{
Each placeholder has an id (a low integer value). The ids are unique across a single
layout. The function uses the placeholder's id to reference it. Different from a ph label,
the id is auto-assigned by PowerPoint and cannot be modified by the user.
Use \code{\link[=layout_properties]{layout_properties()}} (column \code{id}) and \code{\link[=plot_layout_properties]{plot_layout_properties()}} (upper right
corner, in green) to find a placeholder's id.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Comparison")
plot_layout_properties(doc, "Comparison")

doc <- ph_with(doc, "The Title", location = ph_location_id(id = 2)) # title
doc <- ph_with(doc, "Left Header", location = ph_location_id(id = 3)) # left header
doc <- ph_with(doc, "Left Content", location = ph_location_id(id = 4)) # left content
doc <- ph_with(doc, "The Footer", location = ph_location_id(id = 8)) # footer

file <- tempfile(fileext = ".pptx")
print(doc, file)

## file.show(file) # may not work on your system
}
\seealso{
Other functions for placeholder location:
\code{\link[=ph_location]{ph_location()}},
\code{\link[=ph_location_fullsize]{ph_location_fullsize()}},
\code{\link[=ph_location_label]{ph_location_label()}},
\code{\link[=ph_location_left]{ph_location_left()}},
\code{\link[=ph_location_right]{ph_location_right()}},
\code{\link[=ph_location_template]{ph_location_template()}},
\code{\link[=ph_location_type]{ph_location_type()}}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{resolve_location}
\alias{resolve_location}
\title{Resolve short form location}
\usage{
resolve_location(x)
}
\description{
Convert short form location format, a numeric or a string (e.g. \code{"body [1]"}),
into its corresponding location object. Under the hood, we parse the short form location and
call the corresponding function from the \verb{ph_location_*} family. Note that short forms may not
cover all function from the \verb{ph_location_*} and offer less customization.
}
\section{Short forms}{

The following location short forms are implemented. The corresponding call of the function from the
\verb{ph_location_*} family is displayed on the right.\tabular{lll}{
   \strong{Short form} \tab \strong{Description} \tab \strong{Location function} \cr
   \code{"left"} \tab Keyword string \tab \code{ph_location_left()} \cr
   \code{"right"} \tab Keyword string \tab \code{ph_location_right()} \cr
   \code{"fullsize"} \tab Keyword string \tab \code{ph_location_fullsize()} \cr
   \code{"body [1]"} \tab String: type + index in brackets (\code{1} if omitted) \tab \code{ph_location_type("body", 1)} \cr
   \code{"my_label"} \tab Any string not matching a keyword or type \tab \code{ph_location_label("my_label")} \cr
   \code{2} \tab Length 1 integer \tab \code{ph_location_id(2)} \cr
   \code{c(0,0,4,5)} \tab Length 4 numeric, optionally named, \code{c(top=0, left=0, ...)} \tab \code{ph_location(0, 0, 4, 5)} \cr
}
}

\examples{
library(officer)

resolve_location("left")
resolve_location("right")
resolve_location("fullsize")
resolve_location("body")
resolve_location("body [1]")
resolve_location("<some label>")
resolve_location(2)
resolve_location(c(0, 0, 4, 5))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_bookmark}
\alias{body_bookmark}
\title{Add bookmark in a 'Word' document}
\usage{
body_bookmark(x, id)
}
\arguments{
\item{x}{an rdocx object}

\item{id}{bookmark name}
}
\description{
Add a bookmark at the cursor location. The bookmark
is added on the first run of text in the current paragraph.
}
\examples{

# cursor_bookmark ----

doc <- read_docx()
doc <- body_add_par(doc, "centered text", style = "centered")
doc <- body_bookmark(doc, "text_to_replace")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_section_continuous}
\alias{body_end_section_continuous}
\title{Add continuous section}
\usage{
body_end_section_continuous(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
Section break starts the new section on the same page. This
type of section break is often used to change the number of columns
without starting a new page.
}
\examples{
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 5)
str1 <- paste(str1, collapse = " ")
str2 <- "Aenean venenatis varius elit et fermentum vivamus vehicula."
str2 <- rep(str2, 5)
str2 <- paste(str2, collapse = " ")

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = "Default section", style = "heading 1")
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_add_par(doc_1, value = str2, style = "Normal")
doc_1 <- body_end_section_continuous(doc_1)

print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_block_section}()},
\code{\link{body_end_section_columns}()},
\code{\link{body_end_section_columns_landscape}()},
\code{\link{body_end_section_landscape}()},
\code{\link{body_end_section_portrait}()},
\code{\link{body_set_default_section}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_autonum}
\alias{run_autonum}
\title{Auto number}
\usage{
run_autonum(
  seq_id = "table",
  pre_label = "Table ",
  post_label = ": ",
  bkm = NULL,
  bkm_all = FALSE,
  prop = NULL,
  start_at = NULL,
  tnd = 0,
  tns = "-"
)
}
\arguments{
\item{seq_id}{sequence identifier}

\item{pre_label, post_label}{text to add before and after number}

\item{bkm}{bookmark id to associate with autonumber run. If NULL, no bookmark
is added. Value can only be made of alpha numeric characters, ':', -' and '_'.}

\item{bkm_all}{if TRUE, the bookmark will be set on the whole string, if
FALSE, the bookmark will be set on the number only. Default to FALSE.
As an effect when a reference to this bookmark is used, the text can
be like "Table 1" or "1" (pre_label is not included in the referenced
text).}

\item{prop}{formatting text properties returned by \link{fp_text}.}

\item{start_at}{If not NULL, it must be a positive integer, it
specifies the new number to use, at which number the auto
numbering will restart.}

\item{tnd}{\emph{title number depth}, a positive integer (only applies if positive)
that specify the depth (or heading of level \emph{depth}) to use for prefixing
the caption number with this last reference number. For example, setting
\code{tnd=2} will generate numbered captions like '4.3-2' (figure 2 of
chapter 4.3).}

\item{tns}{separator to use between title number and table
number. Default is "-".}
}
\description{
Create an autonumbered chunk, i.e. a string
representation of a sequence, each item will be numbered.
These runs can also be bookmarked and be used later for
cross references.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
run_autonum()
run_autonum(seq_id = "fig", pre_label = "fig. ")
run_autonum(seq_id = "tab", pre_label = "Table ", bkm = "anytable")
run_autonum(
  seq_id = "tab", pre_label = "Table ", bkm = "anytable",
  tnd = 2, tns = " "
)
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}

Other Word computed fields: 
\code{\link{run_reference}()},
\code{\link{run_word_field}()}
}
\concept{Word computed fields}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{image_to_base64}
\alias{image_to_base64}
\title{Images to base64}
\usage{
image_to_base64(filepaths)
}
\arguments{
\item{filepaths}{file names.}
}
\description{
encodes images into base64 strings.
}
\examples{
rlogo <- file.path( R.home("doc"), "html", "logo.jpg")
image_to_base64(rlogo)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{as_base64}
\alias{as_base64}
\title{Encode Character Vector to Base64}
\usage{
as_base64(x)
}
\arguments{
\item{x}{A character vector. NA values are preserved.}
}
\value{
A character vector of Base64-encoded strings, same length as \code{x}.
}
\description{
Encodes one or more elements of a character vector into Base64 format.
}
\examples{
as_base64(letters)
as_base64(c("hello", NA, "world"))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_dedupe_layout.R
\name{layout_dedupe_ph_labels}
\alias{layout_dedupe_ph_labels}
\title{Detect and handle duplicate placeholder labels}
\usage{
layout_dedupe_ph_labels(x, action = "detect", print_info = FALSE)
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{action}{Action to perform on duplicate placeholder labels. One of:
\itemize{
\item \code{detect} (default) = show info on dupes only, make no changes
\item \code{rename} = create unique labels. Labels are renamed by appending a sequential number
separated by dot to duplicate labels. For example, \code{c("title", "title")} becomes \code{c("title.1", "title.2")}.
\item \code{delete} = only keep one of the placeholders with a duplicate label
}}

\item{print_info}{Print action information (e.g. renamed placeholders) to console?
Default is \code{FALSE}. Always \code{TRUE} for action \code{detect}.}
}
\value{
A \code{rpptx} object (with modified placeholder labels).
}
\description{
PowerPoint does not enforce unique placeholder labels in a layout.
Selecting a placeholder via its label using \link{ph_location_label} will throw
an error, if the label is not unique. \link{layout_dedupe_ph_labels} helps to detect,
rename, or delete duplicate placholder labels.
}
\examples{
x <- read_pptx()
layout_dedupe_ph_labels(x)

file <- system.file("doc_examples", "ph_dupes.pptx", package = "officer")
x <- read_pptx(file)
layout_dedupe_ph_labels(x)
layout_dedupe_ph_labels(x, "rename", print_info = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{xlsx_styles}
\alias{xlsx_styles}
\title{Style manager for xlsx workbooks}
\description{
R6 class that manages fonts, fills, borders, and
cell formats in an xlsx workbook's \code{styles.xml}. Used internally
by \code{\link[=sheet_write_data]{sheet_write_data()}} and available for extensions.
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-xlsx_styles-initialize}{\code{xlsx_styles$new()}}
    \item \href{#method-xlsx_styles-get_font_id}{\code{xlsx_styles$get_font_id()}}
    \item \href{#method-xlsx_styles-get_fill_id}{\code{xlsx_styles$get_fill_id()}}
    \item \href{#method-xlsx_styles-get_border_id}{\code{xlsx_styles$get_border_id()}}
    \item \href{#method-xlsx_styles-get_style_id}{\code{xlsx_styles$get_style_id()}}
    \item \href{#method-xlsx_styles-get_xf_id}{\code{xlsx_styles$get_xf_id()}}
    \item \href{#method-xlsx_styles-save}{\code{xlsx_styles$save()}}
    \item \href{#method-xlsx_styles-clone}{\code{xlsx_styles$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-initialize}{}}}
\subsection{\code{xlsx_styles$new()}}{
  Initialize styles from an xlsx package directory.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$new(package_dir)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{package_dir}}{path to the unpacked xlsx directory}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-get_font_id"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-get_font_id}{}}}
\subsection{\code{xlsx_styles$get_font_id()}}{
  Get or create a font index.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$get_font_id(
  name = "Calibri",
  size = 11,
  bold = FALSE,
  italic = FALSE,
  underline = FALSE,
  color = "000000"
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{name}}{font family name}
      \item{\code{size}}{font size in points}
      \item{\code{bold}}{logical, bold}
      \item{\code{italic}}{logical, italic}
      \item{\code{underline}}{logical, underline}
      \item{\code{color}}{hex color string (6 chars, without \verb{#})}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-get_fill_id"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-get_fill_id}{}}}
\subsection{\code{xlsx_styles$get_fill_id()}}{
  Get or create a fill index.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$get_fill_id(bg_color = "FFFFFF")}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{bg_color}}{hex color string for fill background}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-get_border_id"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-get_border_id}{}}}
\subsection{\code{xlsx_styles$get_border_id()}}{
  Get or create a border index.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$get_border_id(
  top_style = NULL,
  top_color = NULL,
  bottom_style = NULL,
  bottom_color = NULL,
  left_style = NULL,
  left_color = NULL,
  right_style = NULL,
  right_color = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{top_style}}{border style for top side}
      \item{\code{top_color}}{border color for top side}
      \item{\code{bottom_style}}{border style for bottom side}
      \item{\code{bottom_color}}{border color for bottom side}
      \item{\code{left_style}}{border style for left side}
      \item{\code{left_color}}{border color for left side}
      \item{\code{right_style}}{border style for right side}
      \item{\code{right_color}}{border color for right side}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-get_style_id"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-get_style_id}{}}}
\subsection{\code{xlsx_styles$get_style_id()}}{
  Get or create a cell format (xf) index.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$get_style_id(
  font_id = 0L,
  fill_id = 0L,
  border_id = 0L,
  num_fmt_id = 0L,
  halign = NA,
  valign = NA,
  text_rotation = 0L,
  wrap_text = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{font_id}}{integer, font index (0-based)}
      \item{\code{fill_id}}{integer, fill index (0-based)}
      \item{\code{border_id}}{integer, border index (0-based)}
      \item{\code{num_fmt_id}}{integer, number format id}
      \item{\code{halign}}{horizontal alignment}
      \item{\code{valign}}{vertical alignment}
      \item{\code{text_rotation}}{text rotation angle (0-180)}
      \item{\code{wrap_text}}{logical, enable text wrapping}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-get_xf_id"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-get_xf_id}{}}}
\subsection{\code{xlsx_styles$get_xf_id()}}{
  Get or create a cell format index for a number format.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$get_xf_id(num_fmt_id)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{num_fmt_id}}{integer, number format id}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-save"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-save}{}}}
\subsection{\code{xlsx_styles$save()}}{
  Save styles.xml to disk.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$save()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_styles-clone"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_styles-clone}{}}}
\subsection{\code{xlsx_styles$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_styles$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_comment}
\alias{run_comment}
\title{Comment for 'Word'}
\usage{
run_comment(
  cmt,
  run = ftext(""),
  author = "",
  date = "",
  initials = "",
  prop = NULL
)
}
\arguments{
\item{cmt}{a set of blocks to be used as comment content returned by
function \code{\link[=block_list]{block_list()}}.
the "run functions for reporting".}

\item{run}{a run object, made with a call to one of}

\item{author}{comment author.}

\item{date}{comment date}

\item{initials}{comment initials}

\item{prop}{formatting text properties returned by
\code{\link[=fp_text_lite]{fp_text_lite()}} or \code{\link[=fp_text]{fp_text()}}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Add a comment on a run object.
}
\examples{
fp_bold <- fp_text_lite(bold = TRUE)
fp_red <- fp_text_lite(color = "red")

bl <- block_list(
  fpar(ftext("Comment multiple words.", fp_bold)),
  fpar(
    ftext("Second line.", fp_red)
  )
)

comment1 <- run_comment(
  cmt = bl,
  run = ftext("with a comment"),
  author = "Author Me",
  date = Sys.Date(),
  initials = "AM"
)
par1 <- fpar("A paragraph ", comment1)

bl <- block_list(
  fpar(ftext("Comment a paragraph."))
)

comment2 <- run_comment(
  cmt = bl, run = ftext("A commented paragraph"),
  author = "Author You",
  date = Sys.Date(),
  initials = "AY"
)
par2 <- fpar(comment2)

doc <- read_docx()
doc <- body_add_fpar(doc, value = par1, style = "Normal")
doc <- body_add_fpar(doc, value = par2, style = "Normal")

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{to_rtf}
\alias{to_rtf}
\title{Convert officer objects to RTF}
\usage{
to_rtf(x, ...)
}
\arguments{
\item{x}{object to convert to RTF. Supported
objects are:
\itemize{
\item \code{\link[=ftext]{ftext()}}
\item \code{\link[=external_img]{external_img()}}
\item \code{\link[=run_autonum]{run_autonum()}}
\item \code{\link[=run_columnbreak]{run_columnbreak()}}
\item \code{\link[=run_linebreak]{run_linebreak()}}
\item \code{\link[=run_word_field]{run_word_field()}}
\item \code{\link[=run_reference]{run_reference()}}
\item \code{\link[=run_pagebreak]{run_pagebreak()}}
\item \code{\link[=hyperlink_ftext]{hyperlink_ftext()}}
\item \code{\link[=block_list]{block_list()}}
\item \code{\link[=fpar]{fpar()}}
}}

\item{...}{Arguments to be passed to methods}
}
\value{
a string containing the RTF code
}
\description{
Convert an object made with package officer
to RTF.
}
\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_manipulate.R
\name{ph_hyperlink}
\alias{ph_hyperlink}
\title{Hyperlink a placeholder}
\usage{
ph_hyperlink(x, type = "body", id = 1, id_chr = NULL, ph_label = NULL, href)
}
\arguments{
\item{x}{an rpptx object}

\item{type}{placeholder type}

\item{id}{placeholder index (integer) for a duplicated type. This is to be used when a placeholder
type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
Values can be read from \code{\link{slide_summary}}.}

\item{id_chr}{deprecated.}

\item{ph_label}{label associated to the placeholder. Use column
\code{ph_label} of result returned by \code{\link{slide_summary}}.
If used, \code{type} and \code{id} are ignored.}

\item{href}{hyperlink (do not forget http or https prefix)}
}
\description{
Add hyperlink to a placeholder in the current slide.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
loc_manual <- ph_location(bg = "red", newlabel = "mytitle")
doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(x = doc, "Un titre 1", location = loc_manual)
slide_summary(doc) # read column ph_label here
doc <- ph_hyperlink(
  x = doc, ph_label = "mytitle",
  href = "https://cran.r-project.org"
)

print(doc, target = fileout)
}
\seealso{
\code{\link{ph_with}}

Other functions for placeholders manipulation: 
\code{\link{ph_remove}()},
\code{\link{ph_slidelink}()}
}
\concept{functions for placeholders manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_footnoteref}
\alias{run_footnoteref}
\title{Word footnote reference}
\usage{
run_footnoteref(prop = NULL)
}
\arguments{
\item{prop}{formatting text properties returned by
\code{\link[=fp_text_lite]{fp_text_lite()}} or \code{\link[=fp_text]{fp_text()}}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Wraps a footnote reference in an object that can then be inserted
as a run/chunk with \code{\link[=fpar]{fpar()}} or within an R Markdown document.
}
\examples{
run_footnoteref()
to_wml(run_footnoteref())
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{rtf_styles_info}
\alias{rtf_styles_info}
\title{Read paragraph styles defined on an RTF document}
\usage{
rtf_styles_info(x)
}
\arguments{
\item{x}{an rtf object created by \code{\link[=rtf_doc]{rtf_doc()}}.}
}
\value{
a data.frame with one row per style.
}
\description{
Return the data.frame of styles currently registered on
the document. Useful for debugging or for checking which built-in
styles are available before referencing them via \code{style = ...}.
}
\examples{
rtf_styles_info(rtf_doc())
}
\seealso{
\code{\link[=rtf_set_paragraph_style]{rtf_set_paragraph_style()}}, \code{\link[=styles_info]{styles_info()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_notes.R
\name{notes_location_label}
\alias{notes_location_label}
\title{Location of a named placeholder for notes}
\usage{
notes_location_label(ph_label, ...)
}
\arguments{
\item{ph_label}{placeholder label of the used notes master}

\item{...}{unused arguments}
}
\description{
The function will use the label of a placeholder
to find the corresponding location in the slide notes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{floating_external_img}
\alias{floating_external_img}
\title{Floating external image}
\usage{
floating_external_img(
  src,
  width = 0.5,
  height = 0.2,
  pos_x = 0,
  pos_y = 0,
  pos_h_from = "margin",
  pos_v_from = "margin",
  wrap_type = "square",
  wrap_side = "bothSides",
  wrap_dist_top = 0,
  wrap_dist_bottom = 0,
  wrap_dist_left = 0.125,
  wrap_dist_right = 0.125,
  unit = "in",
  guess_size = FALSE,
  alt = ""
)
}
\arguments{
\item{src}{image file path}

\item{width, height}{size of the image file. It can be ignored
if parameter \code{guess_size=TRUE}, see parameter \code{guess_size}.}

\item{pos_x, pos_y}{horizontal and vertical position of the image relative to the anchor point}

\item{pos_h_from}{horizontal positioning reference point, one of "margin", "page", "column", "character"}

\item{pos_v_from}{vertical positioning reference point, one of "margin", "page", "paragraph", "line"}

\item{wrap_type}{text wrapping type, one of "square", "topAndBottom", "through", "tight", "none"}

\item{wrap_side}{which side text wraps around, one of "bothSides", "left", "right", "largest"}

\item{wrap_dist_top, wrap_dist_bottom, wrap_dist_left, wrap_dist_right}{distance between image and text (in inches)}

\item{unit}{unit for width, height, pos_x and pos_y, one of "in", "cm", "mm".}

\item{guess_size}{If package 'magick' is installed, this option
can be used (set it to \code{TRUE}). The images will be read and
width and height will be guessed.}

\item{alt}{alternative text for images}
}
\description{
Wraps an image in an object that can be embedded as a floating
image in a 'Word' document. Unlike \code{\link[=external_img]{external_img()}}, which creates inline images,
this function creates floating images that can be positioned anywhere on the page
and allow text wrapping around them.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
library(officer)
srcfile <- file.path(R.home("doc"), "html", "logo.jpg")
floatimg <- floating_external_img(
  src = srcfile, height = 1.06 / 2, width = 1.39 / 2,
  pos_x = 0, pos_y = 0,
  pos_h_from = "margin", pos_v_from = "margin"
)

text <- paste0(
  " is a floating image in a ",
  paste0(rep("very ", 30), collapse = ""),
  " long text!"
)

# docx example ----
x <- read_docx()
fp_t <- fp_text(font.size = 20, color = "red")
an_fpar <- fpar(floatimg, ftext(text, fp_t))
x <- body_add_fpar(x, an_fpar)
print(x, target = tempfile(fileext = ".docx"))

# rtf example ----
rtf_doc <- rtf_doc()
rtf_doc <- rtf_add(rtf_doc, an_fpar)
print(rtf_doc, target = tempfile(fileext = ".rtf"))
}
\seealso{
\link{external_img}, \link{body_add}, \link{fpar}, \link{rtf_doc}, \link{rtf_add}

Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{uuid_generate}
\alias{uuid_generate}
\title{generates unique identifiers}
\usage{
uuid_generate(n = 1, ...)
}
\arguments{
\item{n}{integer, number of unique identifiers to generate.}

\item{...}{arguments sent to \code{\link[uuid:UUIDgenerate]{uuid::UUIDgenerate()}}}
}
\description{
generates unique identifiers based
on \code{\link[uuid:UUIDgenerate]{uuid::UUIDgenerate()}}.
}
\examples{
uuid_generate(n = 5)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{docx_dim}
\alias{docx_dim}
\title{'Word' page layout}
\usage{
docx_dim(x)
}
\arguments{
\item{x}{an \code{rdocx} object}
}
\description{
Get page width, page height and margins (in inches). The return values
are those corresponding to the section where the cursor is.
}
\examples{
docx_dim(read_docx())
}
\seealso{
Other functions for Word document informations:
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=docx_bookmarks]{docx_bookmarks()}},
\code{\link[=length.rdocx]{length.rdocx()}},
\code{\link[=set_doc_properties]{set_doc_properties()}},
\code{\link[=styles_info]{styles_info()}}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{change_styles}
\alias{change_styles}
\title{Replace styles in a 'Word' Document}
\usage{
change_styles(x, mapstyles)
}
\arguments{
\item{x}{an rdocx object}

\item{mapstyles}{a named list, names are the replacement style,
content (as a character vector) are the styles to be replaced.
Use \code{\link[=styles_info]{styles_info()}} to display available styles.}
}
\description{
Replace styles with others in a 'Word' document. This function
can be used for paragraph, run/character and table styles.
}
\examples{
# creating a sample docx so that we can illustrate how
# to change styles
doc_1 <- read_docx()

doc_1 <- body_add_par(doc_1, "A title", style = "heading 1")
doc_1 <- body_add_par(doc_1, "Another title", style = "heading 2")
doc_1 <- body_add_par(doc_1, "Hello world!", style = "Normal")
file <- print(doc_1, target = tempfile(fileext = ".docx"))

# now we can illustrate how
# to change styles with `change_styles`
doc_2 <- read_docx(path = file)
mapstyles <- list(
  "centered" = c("Normal", "heading 2"),
  "strong" = "Default Paragraph Font"
)
doc_2 <- change_styles(doc_2, mapstyles = mapstyles)
print(doc_2, target = tempfile(fileext = ".docx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{on_slide}
\alias{on_slide}
\title{Change current slide}
\usage{
on_slide(x, index)
}
\arguments{
\item{x}{an rpptx object}

\item{index}{slide index}
}
\description{
Change current slide index of an rpptx object.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- add_slide(doc, "Title and Content")
doc <- add_slide(doc, "Title and Content")
doc <- on_slide(doc, index = 1)
doc <- ph_with(
  x = doc,
  "First title",
  location = ph_location_type(type = "title")
)
doc <- on_slide(doc, index = 3)
doc <- ph_with(
  x = doc,
  "Third title",
  location = ph_location_type(type = "title")
)

file <- tempfile(fileext = ".pptx")
print(doc, target = file)
}
\seealso{
\code{\link[=read_pptx]{read_pptx()}}, \code{\link[=ph_with]{ph_with()}}

Other functions to manipulate slides:
\code{\link[=add_slide]{add_slide()}},
\code{\link[=move_slide]{move_slide()}},
\code{\link[=remove_slide]{remove_slide()}},
\code{\link[=set_notes]{set_notes()}}
}
\concept{slide_manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{fortify_location}
\alias{fortify_location}
\title{Eval a location on the current slide}
\usage{
fortify_location(x, doc, ...)
}
\arguments{
\item{x}{a location for a placeholder.}

\item{doc}{an rpptx object}

\item{...}{unused arguments}
}
\description{
Eval a shape location against the current slide.
This function is to be used to add custom openxml code. A
list is returned, it contains informations width, height, left
and top positions and other informations necessary to add a
content on a slide.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
fortify_location(ph_location_fullsize(), doc)
}
\seealso{
\code{\link[=ph_location]{ph_location()}}, \code{\link[=ph_with]{ph_with()}}

Other functions for officer extensions: 
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{color_scheme}
\alias{color_scheme}
\title{Color scheme of a PowerPoint file}
\usage{
color_scheme(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Get the color scheme of a
'PowerPoint' master layout into a data.frame.
}
\examples{
x <- read_pptx()
color_scheme ( x = x )
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fortify_pptx.R
\name{media_extract}
\alias{media_extract}
\title{Extract media from a document object}
\usage{
media_extract(x, path, target)
}
\arguments{
\item{x}{an rpptx object}

\item{path}{media path, should be a relative path}

\item{target}{target file}
}
\description{
Extract files from a \code{rpptx} object.
}
\examples{
example_pptx <- system.file(package = "officer",
  "doc_examples/example.pptx")
doc <- read_pptx(example_pptx)
content <- pptx_summary(doc)
image_row <- content[content$content_type \%in\% "image", ]
media_file <- image_row$media_file
png_file <- tempfile(fileext = ".png")
media_extract(doc, path = media_file, target = png_file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_caption}
\alias{body_add_caption}
\title{Add Word caption in a 'Word' document}
\usage{
body_add_caption(x, value, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{value}{an object returned by \code{\link[=block_caption]{block_caption()}}}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add a Word caption into an rdocx object.
}
\examples{
doc <- read_docx()

if (capabilities(what = "png")) {
  doc <- body_add_plot(doc,
    value = plot_instr(
      code = {
        barplot(1:5, col = 2:6)
      }
    ),
    style = "centered"
  )
}
run_num <- run_autonum(
  seq_id = "fig", pre_label = "Figure ",
  bkm = "barplot"
)
caption <- block_caption("a barplot",
  style = "Normal",
  autonum = run_num
)
doc <- body_add_caption(doc, caption)
print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()},
\code{\link{body_append_start_context}()},
\code{\link{body_import_docx}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_table}
\alias{block_table}
\title{Table block}
\usage{
block_table(x, header = TRUE, properties = prop_table(), alignment = NULL)
}
\arguments{
\item{x}{a data.frame to add as a table}

\item{header}{display header if TRUE}

\item{properties}{table properties, see \code{\link[=prop_table]{prop_table()}}.
Table properties are not handled identically between Word and PowerPoint
output format. They are fully supported with Word but for PowerPoint (which
does not handle as many things as Word for tables), only conditional
formatting properties are supported.}

\item{alignment}{alignment for each columns, 'l' for left, 'r' for right
and 'c' for center. Default to NULL.}
}
\description{
Create a representation of a table
}
\examples{
block_table(x = head(iris))

block_table(x = mtcars, header = TRUE,
  properties = prop_table(
    tcf = table_conditional_formatting(
      first_row = TRUE, first_column = TRUE)
  ))
}
\seealso{
\code{\link[=prop_table]{prop_table()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_gg}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_section_columns_landscape}
\alias{body_end_section_columns_landscape}
\title{Add a landscape multi columns section}
\usage{
body_end_section_columns_landscape(
  x,
  widths = c(2.5, 2.5),
  space = 0.25,
  sep = FALSE,
  w = 21/2.54,
  h = 29.7/2.54
)
}
\arguments{
\item{x}{an rdocx object}

\item{widths}{columns widths in inches. If 3 values, 3 columns
will be produced.}

\item{space}{space in inches between columns.}

\item{sep}{if TRUE a line is separating columns.}

\item{w, h}{page width, page height (in inches)}
}
\description{
A landscape section with multiple columns is added to the document.
}
\examples{
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 5)
str1 <- paste(str1, collapse = " ")

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_end_section_columns_landscape(doc_1, widths = c(6, 2))
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_block_section}()},
\code{\link{body_end_section_columns}()},
\code{\link{body_end_section_continuous}()},
\code{\link{body_end_section_landscape}()},
\code{\link{body_end_section_portrait}()},
\code{\link{body_set_default_section}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_img}
\alias{body_add_img}
\title{Add an image in a 'Word' document}
\usage{
body_add_img(x, src, style = NULL, width, height, pos = "after", unit = "in")
}
\arguments{
\item{x}{an rdocx object}

\item{src}{image filename, the basename of the file must not contain any blank.}

\item{style}{paragraph style}

\item{width, height}{image size in units expressed by the unit argument.
Defaults to "in"ches.}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}
}
\description{
Add an image into an rdocx object.
}
\examples{
library(officer)

doc <- read_docx()

img.file <- file.path(R.home("doc"), "html", "logo.jpg")
if (file.exists(img.file)) {
  doc <- body_add_img(x = doc, src = img.file, height = 1.06, width = 1.39)

  # Set the unit in which the width and height arguments are expressed
  doc <- body_add_img(
    x = doc,
    src = img.file,
    height = 2.69,
    width = 3.53,
    unit = "cm"
  )
}

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content:
\code{\link[=body_add_blocks]{body_add_blocks()}},
\code{\link[=body_add_break]{body_add_break()}},
\code{\link[=body_add_caption]{body_add_caption()}},
\code{\link[=body_add_docx]{body_add_docx()}},
\code{\link[=body_add_fpar]{body_add_fpar()}},
\code{\link[=body_add_gg]{body_add_gg()}},
\code{\link[=body_add_list]{body_add_list()}},
\code{\link[=body_add_par]{body_add_par()}},
\code{\link[=body_add_plot]{body_add_plot()}},
\code{\link[=body_add_table]{body_add_table()}},
\code{\link[=body_add_toc]{body_add_toc()}},
\code{\link[=body_append_start_context]{body_append_start_context()}},
\code{\link[=body_import_docx]{body_import_docx()}}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_reference}
\alias{run_reference}
\title{Cross reference}
\usage{
run_reference(id, prop = NULL)
}
\arguments{
\item{id}{reference id, a string}

\item{prop}{formatting text properties returned by \link{fp_text}.}
}
\description{
Create a representation of a reference
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
run_reference("a_ref")
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{floating_external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}

Other Word computed fields: 
\code{\link{run_autonum}()},
\code{\link{run_word_field}()}
}
\concept{Word computed fields}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_manipulate.R
\name{ph_slidelink}
\alias{ph_slidelink}
\title{Slide link to a placeholder}
\usage{
ph_slidelink(
  x,
  type = "body",
  id = 1,
  id_chr = NULL,
  ph_label = NULL,
  slide_index
)
}
\arguments{
\item{x}{an rpptx object}

\item{type}{placeholder type}

\item{id}{placeholder index (integer) for a duplicated type. This is to be used when a placeholder
type is not unique in the layout of the current slide, e.g. two placeholders with type 'body'. To
add onto the first, use \code{id = 1} and \code{id = 2} for the second one.
Values can be read from \code{\link{slide_summary}}.}

\item{id_chr}{deprecated.}

\item{ph_label}{label associated to the placeholder. Use column
\code{ph_label} of result returned by \code{\link{slide_summary}}.
If used, \code{type} and \code{id} are ignored.}

\item{slide_index}{slide index to reach}
}
\description{
Add slide link to a placeholder in the current slide.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
loc_title <- ph_location_type(type = "title")
doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(x = doc, "Un titre 1", location = loc_title)
doc <- add_slide(doc)
doc <- ph_with(x = doc, "Un titre 2", location = loc_title)
doc <- on_slide(doc, 1)
slide_summary(doc) # read column ph_label here
doc <- ph_slidelink(x = doc, ph_label = "Title 1", slide_index = 2)

print(doc, target = fileout)
}
\seealso{
\code{\link{ph_with}}

Other functions for placeholders manipulation: 
\code{\link{ph_hyperlink}()},
\code{\link{ph_remove}()}
}
\concept{functions for placeholders manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_label}
\alias{ph_location_label}
\title{Location of a named placeholder}
\usage{
ph_location_label(ph_label, newlabel = NULL, ...)
}
\arguments{
\item{ph_label}{placeholder label of the used layout. It can be read in PowerPoint or
with function \code{\link[=layout_properties]{layout_properties()}} in column \code{ph_label}.}

\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will use the label of a placeholder
to find the corresponding location.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

# ph_location_label demo ----

doc <- read_pptx()
doc <- add_slide(doc, layout = "Title and Content")

# all ph_label can be read here
layout_properties(doc, layout = "Title and Content")

doc <- ph_with(
  doc,
  head(iris),
  location = ph_location_label(ph_label = "Content Placeholder 2")
)
doc <- ph_with(
  doc,
  format(Sys.Date()),
  location = ph_location_label(ph_label = "Date Placeholder 3")
)
doc <- ph_with(
  doc,
  "This is a title",
  location = ph_location_label(ph_label = "Title 1")
)

print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_left}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_pagebreak}
\alias{run_pagebreak}
\title{Page break for 'Word'}
\usage{
run_pagebreak()
}
\description{
Object representing a page break for a Word document.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
fp_t <- fp_text(font.size = 12, bold = TRUE)
an_fpar <- fpar("let's add a break page", run_pagebreak(), ftext("and blah blah!", fp_t))

x <- read_docx()
x <- body_add(x, an_fpar)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shape_properties.R
\name{sp_line}
\alias{sp_line}
\alias{print.sp_line}
\alias{update.sp_line}
\title{Line properties}
\usage{
sp_line(
  color = "transparent",
  lwd = 1,
  lty = "solid",
  linecmpd = "sng",
  lineend = "rnd",
  linejoin = "round",
  headend = sp_lineend(type = "none"),
  tailend = sp_lineend(type = "none")
)

\method{print}{sp_line}(x, ...)

\method{update}{sp_line}(
  object,
  color,
  lwd,
  lty,
  linecmpd,
  lineend,
  linejoin,
  headend,
  tailend,
  ...
)
}
\arguments{
\item{color}{line color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{lwd}{line width (in point) - 0 or positive integer value.}

\item{lty}{single character value specifying the line type.
Expected value is one of the following : default \code{'solid'}
or \code{'dot'} or \code{'dash'} or \code{'lgDash'}
or \code{'dashDot'} or \code{'lgDashDot'} or \code{'lgDashDotDot'}
or \code{'sysDash'} or \code{'sysDot'} or \code{'sysDashDot'}
or \code{'sysDashDotDot'}.}

\item{linecmpd}{single character value specifying the compound line type.
Expected value is one of the following : default \code{'sng'}
or \code{'dbl'} or \code{'tri'} or \code{'thinThick'}
or \code{'thickThin'}}

\item{lineend}{single character value specifying the line end style
Expected value is one of the following : default \code{'rnd'}
or \code{'sq'} or \code{'flat'}}

\item{linejoin}{single character value specifying the line join style
Expected value is one of the following : default \code{'round'}
or \code{'bevel'} or \code{'miter'}}

\item{headend}{a \code{sp_lineend} object specifying line head end style}

\item{tailend}{a \code{sp_lineend} object specifying line tail end style}

\item{x, object}{\code{sp_line} object}

\item{...}{further arguments - not used}
}
\value{
a \code{sp_line} object
}
\description{
Create a \code{sp_line} object that describes
line properties.
}
\examples{
sp_line()
sp_line(color = "red", lwd = 2)
sp_line(lty = "dot", linecmpd = "dbl")
print( sp_line (color="red", lwd = 2) )
obj <- sp_line (color="red", lwd = 2)
update( obj, linecmpd = "dbl" )
}
\seealso{
\link{sp_lineend}

Other functions for defining shape properties: 
\code{\link{sp_lineend}()}
}
\concept{functions for defining shape properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{docx_show_chunk}
\alias{docx_show_chunk}
\title{Show underlying text tag structure}
\usage{
docx_show_chunk(x)
}
\arguments{
\item{x}{a docx device}
}
\description{
Show the structure of text tags at the current cursor. This is
most useful when trying to troubleshoot search-and-replace functionality
using \code{\link[=body_replace_all_text]{body_replace_all_text()}}.
}
\seealso{
\code{\link[=body_replace_all_text]{body_replace_all_text()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{external_img}
\alias{external_img}
\title{External image}
\usage{
external_img(
  src,
  width = 0.5,
  height = 0.2,
  unit = "in",
  guess_size = FALSE,
  alt = ""
)
}
\arguments{
\item{src}{image file path}

\item{width, height}{size of the image file. It can be ignored
if parameter \code{guess_size=TRUE}, see parameter \code{guess_size}.}

\item{unit}{unit for width and height, one of "in", "cm", "mm".}

\item{guess_size}{If package 'magick' is installed, this option
can be used (set it to \code{TRUE}). The images will be read and
width and height will be guessed.}

\item{alt}{alternative text for images}
}
\description{
Wraps an image in an object that can then be embedded
in a PowerPoint slide or within a Word paragraph.

The image is added as a shape in PowerPoint (it is not possible to mix text and
images in a PowerPoint form). With a Word document, the image will be
added inside a paragraph.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
# wrap r logo with external_img ----
srcfile <- file.path(R.home("doc"), "html", "logo.jpg")
extimg <- external_img(
  src = srcfile, height = 1.06 / 2,
  width = 1.39 / 2
)

# pptx example ----
doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(
  x = doc, value = extimg,
  location = ph_location_type(type = "body"),
  use_loc_size = FALSE
)
print(doc, target = tempfile(fileext = ".pptx"))

fp_t <- fp_text(font.size = 20, color = "red")
an_fpar <- fpar(extimg, ftext(" is cool!", fp_t))

# docx example ----
x <- read_docx()
x <- body_add(x, an_fpar)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
\link{ph_with}, \link{body_add}, \link{fpar}

Other run functions for reporting: 
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr_utils.R
\name{opts_current_table}
\alias{opts_current_table}
\title{Table options in a 'knitr' context}
\usage{
opts_current_table()
}
\value{
a list
}
\description{
Get options for table rendering in a 'knitr'
context. It should not be used by the end user, but
its documentation should be read as a place where
table options are documented when 'knitr' is used.

The function is a utility to facilitate the retrieval of table
options supported by the 'flextable', 'officedown' and 'officer' packages.

These options should be set with \code{knitr::opts_chunk$set()}.
The names and expected values are listed in the following sections.
}
\section{knitr chunk options for table captions}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   caption id/bookmark \tab tab.id \tab NULL \cr
   caption \tab tab.cap \tab NULL \cr
   display table caption on top of the table or not \tab tab.topcaption \tab TRUE \cr
   caption table sequence identifier. \tab tab.lp \tab "tab:" \cr
}
}

\section{knitr chunk options for Word table captions}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   Word stylename to use for table captions. \tab tab.cap.style \tab NULL \cr
   prefix for numbering chunk (default to   "Table "). \tab tab.cap.pre \tab Table \cr
   suffix for numbering chunk (default to   ": "). \tab tab.cap.sep \tab " :" \cr
   title number depth \tab tab.cap.tnd \tab 0 \cr
   separator to use between title number and table number. \tab tab.cap.tns \tab "-" \cr
   caption prefix formatting properties \tab tab.cap.fp_text \tab fp_text_lite(bold = TRUE) \cr
}
}

\section{knitr chunk options for Word tables}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   the Word stylename to use for tables \tab tab.style \tab NULL \cr
   autofit' or 'fixed' algorithm. \tab tab.layout \tab "autofit" \cr
   value of the preferred width of the table in percent (base 1). \tab tab.width \tab 1 \cr
   Alternative title text \tab tab.alt.title \tab NULL \cr
   Alternative description text \tab tab.alt.description \tab NULL \cr
}
}

\section{knitr chunk options for data.frame with officedown}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   apply or remove formatting from the first row in the table \tab first_row \tab TRUE \cr
   apply or remove formatting from the first column in the table \tab first_column \tab FALSE \cr
   apply or remove formatting from the last row in the table \tab last_row \tab FALSE \cr
   apply or remove formatting from the last column in the table \tab last_column \tab FALSE \cr
   don't display odd and even rows \tab no_hband \tab TRUE \cr
   don't display odd and even columns \tab no_vband \tab TRUE \cr
}
}

\section{returned elements}{

\itemize{
\item cap.style (default: NULL)
\item cap.pre (default: "Table ")
\item cap.sep (default: ":")
\item cap.tnd (default: 0)
\item cap.tns (default: "-")
\item cap.fp_text (default: \code{fp_text_lite(bold = TRUE)})
\item id (default: NULL)
\item cap (default: NULL)
\item alt.title (default: NULL)
\item alt.description (default: NULL)
\item topcaption (default: TRUE)
\item style (default: NULL)
\item tab.lp (default: "tab:")
\item table_layout (default: "autofit")
\item table_width (default: 1)
\item first_row (default: TRUE)
\item first_column (default: FALSE)
\item last_row (default: FALSE)
\item last_column (default: FALSE)
\item no_hband (default: TRUE)
\item no_vband (default: TRUE)
}
}

\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{table_stylenames}
\alias{table_stylenames}
\title{Paragraph styles for columns}
\usage{
table_stylenames(stylenames = list())
}
\arguments{
\item{stylenames}{a named character vector, names are column names, values are
paragraph styles associated with each column. If a column is not
specified, default value 'Normal' is used.
Another form is as a named list, the list names are the styles
and the contents are column names to be formatted with the
corresponding style.}
}
\description{
The function defines the paragraph styles for columns.
}
\examples{
library(officer)

stylenames <- c(
  vs = "centered", am = "centered",
  gear = "centered", carb = "centered"
)

doc_1 <- read_docx()
doc_1 <- body_add_table(doc_1,
  value = mtcars, style = "table_template",
  stylenames = table_stylenames(stylenames = stylenames)
)

print(doc_1, target = tempfile(fileext = ".docx"))


stylenames <- list(
  "centered" = c("vs", "am", "gear", "carb")
)

doc_2 <- read_docx()
doc_2 <- body_add_table(doc_2,
  value = mtcars, style = "table_template",
  stylenames = table_stylenames(stylenames = stylenames)
)

print(doc_2, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for table definition: 
\code{\link{prop_table}()},
\code{\link{table_colwidths}()},
\code{\link{table_conditional_formatting}()},
\code{\link{table_layout}()},
\code{\link{table_width}()}
}
\concept{functions for table definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{shape_properties_tags}
\alias{shape_properties_tags}
\title{pptx tags for visual and non visual properties}
\usage{
shape_properties_tags(
  left = 0,
  top = 0,
  width = 3,
  height = 3,
  bg = "transparent",
  rot = 0,
  label = "",
  ph = "<p:ph/>",
  ln = sp_line(lwd = 0, linecmpd = "solid", lineend = "rnd"),
  geom = NULL
)
}
\arguments{
\item{left, top, width, height}{place holder coordinates
in inches.}

\item{bg}{background color}

\item{rot}{rotation angle}

\item{label}{a label for the placeholder.}

\item{ph}{string containing xml code for ph tag}

\item{ln}{a \code{\link[=sp_line]{sp_line()}} object specifying the outline style.}

\item{geom}{shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html}
}
\value{
a character value
}
\description{
Visual and non visual properties of a
shape can be returned by this function.
}
\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{annotate_base}
\alias{annotate_base}
\title{Placeholder parameters annotation}
\usage{
annotate_base(path = NULL, output_file = "annotated_layout.pptx")
}
\arguments{
\item{path}{path to the pptx file to use as base document or NULL to use the officer default}

\item{output_file}{filename to store the annotated powerpoint file or NULL to suppress generation}
}
\value{
rpptx object of the annotated PowerPoint file
}
\description{
generates a slide from each layout in the base document to
identify the placeholder indexes, types, names, master names and layout names.

This is to be used when need to know what parameters should be used with
\verb{ph_location*} calls. The parameters are printed in their corresponding shapes.

Note that if there are duplicated \code{ph_label}, you should not use \code{ph_location_label()}.
Hint: You can dedupe labels using \code{\link[=layout_dedupe_ph_labels]{layout_dedupe_ph_labels()}}.
}
\examples{
# To generate an anotation of the default base document with officer:
annotate_base(output_file = tempfile(fileext = ".pptx"))

# To generate an annotation of the base document 'mydoc.pptx' and place the
# annotated output in 'mydoc_annotate.pptx'
# annotate_base(path = 'mydoc.pptx', output_file='mydoc_annotate.pptx')

}
\seealso{
Other functions for reading presentation information: 
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/officer.R
\docType{package}
\name{officer}
\alias{officer-package}
\alias{officer}
\title{Manipulate Microsoft Word and PowerPoint Documents with 'officer'}
\description{
The officer package facilitates access to and manipulation of
'Microsoft Word' and 'Microsoft PowerPoint' documents from R.
It also supports the writing of 'RTF' documents.

Examples of usage are:
\itemize{
\item Create Word documents with tables, titles, TOC and graphics
\item Importation of Word and PowerPoint files into data objects
\item Write updated content back to a PowerPoint presentation
\item Clinical reporting automation
\item Production of reports from a shiny application
}

To start with officer, read about \code{\link[=read_docx]{read_docx()}}, \code{\link[=read_pptx]{read_pptx()}}
or \code{\link[=rtf_doc]{rtf_doc()}}.

The package is also providing several objects that can be printed
in 'R Markdown' documents for advanced Word or PowerPoint
reporting as \code{\link[=run_autonum]{run_autonum()}} and \code{\link[=block_caption]{block_caption()}}.
}
\seealso{
The user documentation: \url{https://ardata-fr.github.io/officeverse/} and
manuals \url{https://davidgohel.github.io/officer/}
}
\author{
\strong{Maintainer}: David Gohel \email{david.gohel@ardata.fr}

Authors:
\itemize{
  \item Stefan Moog \email{moogs@gmx.de}
  \item Mark Heckmann \email{heckmann.mark@gmail.com} (\href{https://orcid.org/0000-0002-0736-7417}{ORCID})
}

Other contributors:
\itemize{
  \item ArData [copyright holder]
  \item Frank Hangler \email{frank@plotandscatter.com} (function body_replace_all_text) [contributor]
  \item Liz Sander \email{lsander@civisanalytics.com} (several documentation fixes) [contributor]
  \item Anton Victorson \email{anton@victorson.se} (fixes xml structures) [contributor]
  \item Jon Calder \email{jonmcalder@gmail.com} (update vignettes) [contributor]
  \item John Harrold \email{john.m.harrold@gmail.com} (function annotate_base) [contributor]
  \item John Muschelli \email{muschellij2@gmail.com} (google doc compatibility) [contributor]
  \item Bill Denney \email{wdenney@humanpredictions.com} (\href{https://orcid.org/0000-0002-5759-428X}{ORCID}) (function as.matrix.rpptx) [contributor]
  \item Nikolai Beck \email{beck.nikolai@gmail.com} (set speaker notes for .pptx documents) [contributor]
  \item Greg Leleu \email{gregoire.leleu@gmail.com} (fields functionality in ppt) [contributor]
  \item Majid Eismann [contributor]
  \item Hongyuan Jia \email{hongyuanjia@cqust.edu.cn} (\href{https://orcid.org/0000-0002-0075-8183}{ORCID}) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{length.rpptx}
\alias{length.rpptx}
\title{Number of slides}
\usage{
\method{length}{rpptx}(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Function \code{length} will return the number of slides.
}
\examples{
my_pres <- read_pptx()
my_pres <- add_slide(my_pres, "Title and Content")
my_pres <- add_slide(my_pres, "Title and Content")
length(my_pres)
}
\seealso{
Other functions for reading presentation information:
\code{\link[=annotate_base]{annotate_base()}},
\code{\link[=color_scheme]{color_scheme()}},
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=layout_properties]{layout_properties()}},
\code{\link[=layout_summary]{layout_summary()}},
\code{\link[=plot_layout_properties]{plot_layout_properties()}},
\code{\link[=slide_size]{slide_size()}},
\code{\link[=slide_summary]{slide_summary()}}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_left}
\alias{ph_location_left}
\title{Location of a left body element}
\usage{
ph_location_left(newlabel = NULL, ...)
}
\arguments{
\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will return the location corresponding
to a left bounding box. The function assume the layout 'Two Content'
is existing. This is an helper function, if you don't have a layout
named 'Two Content', use \code{\link[=ph_location_type]{ph_location_type()}} and set arguments
to your specific needs.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(doc, "Hello left", location = ph_location_left())
doc <- ph_with(doc, "Hello right", location = ph_location_right())
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{table_conditional_formatting}
\alias{table_conditional_formatting}
\title{Table conditional formatting}
\usage{
table_conditional_formatting(
  first_row = TRUE,
  first_column = FALSE,
  last_row = FALSE,
  last_column = FALSE,
  no_hband = FALSE,
  no_vband = TRUE
)
}
\arguments{
\item{first_row, last_row}{apply or remove formatting from the first or last row in the table.}

\item{first_column, last_column}{apply or remove formatting from the first or last column in the table.}

\item{no_hband, no_vband}{don't display odd and even rows or columns with
alternating shading for ease of reading.}
}
\description{
Tables can be conditionally formatted based on few properties as
whether the content is in the first row, last row, first column, or last
column, or whether the rows or columns are to be banded.
}
\note{
You must define a format for first_row, first_column and other properties
if you need to use them. The format is defined in a docx template.
}
\examples{
table_conditional_formatting(first_row = TRUE, first_column = TRUE)
}
\seealso{
Other functions for table definition: 
\code{\link{prop_table}()},
\code{\link{table_colwidths}()},
\code{\link{table_layout}()},
\code{\link{table_stylenames}()},
\code{\link{table_width}()}
}
\concept{functions for table definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_wordtext}
\alias{run_wordtext}
\title{Word chunk of text with a style}
\usage{
run_wordtext(text, style_id = NULL)
}
\arguments{
\item{text}{text value, a single character value}

\item{style_id}{'Word' unique style identifier associated with the style to use.}
}
\description{
Format a chunk of text associated with a 'Word' character style.
The style is defined with its unique identifer.
}
\examples{
run1 <- run_wordtext("hello", "DefaultParagraphFont")
paragraph <- fpar(run1)

x <- read_docx()
x <- body_add_fpar(x, paragraph)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
\code{\link[=ftext]{ftext()}}

Other run functions for reporting:
\code{\link[=external_img]{external_img()}},
\code{\link[=floating_external_img]{floating_external_img()}},
\code{\link[=ftext]{ftext()}},
\code{\link[=hyperlink_ftext]{hyperlink_ftext()}},
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_bookmark]{run_bookmark()}},
\code{\link[=run_columnbreak]{run_columnbreak()}},
\code{\link[=run_comment]{run_comment()}},
\code{\link[=run_footnote]{run_footnote()}},
\code{\link[=run_footnoteref]{run_footnoteref()}},
\code{\link[=run_linebreak]{run_linebreak()}},
\code{\link[=run_pagebreak]{run_pagebreak()}},
\code{\link[=run_reference]{run_reference()}},
\code{\link[=run_tab]{run_tab()}},
\code{\link[=run_word_field]{run_word_field()}}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_plot}
\alias{body_add_plot}
\title{Add plot in a 'Word' document}
\usage{
body_add_plot(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  pos = "after",
  unit = "in",
  ...
)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{plot instructions, see \code{\link[=plot_instr]{plot_instr()}}.}

\item{width, height}{plot size in units expressed by the unit argument.
Defaults to a width of 6 and a height of 5 "in"ches.}

\item{res}{resolution of the png image in ppi}

\item{style}{paragraph style}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}

\item{...}{Arguments to be passed to png function.}
}
\description{
Add a plot as a png image into an rdocx object.
}
\examples{
doc <- read_docx()

if (capabilities(what = "png")) {
  p <- plot_instr(
      code = {
        barplot(1:5, col = 2:6)
      }
    )

  doc <- body_add_plot(doc, value = p, style = "centered")

  # Set the unit in which the width and height arguments are expressed
  doc <- body_add_plot(doc, value = p, style = "centered", unit = "cm")
}

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()},
\code{\link{body_append_start_context}()},
\code{\link{body_import_docx}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_comments.R
\name{docx_comments}
\alias{docx_comments}
\title{Get comments in a Word document as a data.frame}
\usage{
docx_comments(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
return a data.frame representing the comments in a Word document.
}
\details{
Each row of the returned data frame contains data for one comment. The
columns contain the following information:
\itemize{
\item "comment_id" - unique comment id
\item "author" - name of the comment author
\item "initials" - initials of the comment author
\item "date" - timestamp of the comment
\item "text" - a list column of characters containing the comment text. Elements can
be vectors of length > 1 if a comment contains multiple paragraphs,
blocks or runs or of length 0 if the comment is empty.
\item "para_id" - a list column of characters containing the parent paragraph IDs.
Elememts can be vectors of length > 1 if a comment spans multiple paragraphs
or of length 0 if the comment has no parent paragraph.
\item "commented_text" - a list column of characters containing the
commented text. Elememts can be vectors of length > 1 if a comment
spans multiple paragraphs or runs or of length 0 if the commented text is empty.
}
}
\examples{
bl <- block_list(
  fpar("Comment multiple words."),
  fpar("Second line")
)

a_par <- fpar(
  "This paragraph contains",
  run_comment(
    cmt = bl,
    run = ftext("a comment."),
    author = "Author Me",
    date = "2023-06-01"
  )
)

doc <- read_docx()
doc <- body_add_fpar(doc, value = a_par, style = "Normal")

docx_file <- print(doc, target = tempfile(fileext = ".docx"))

docx_comments(read_docx(docx_file))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_tabs}
\alias{fp_tabs}
\title{Tabs properties object}
\usage{
fp_tabs(...)
}
\arguments{
\item{...}{\link{fp_tab} objects}
}
\description{
create a set of tabulation mark properties object for Word or RTF.
Results can be used as arguments \code{tabs} of \code{\link[=fp_par]{fp_par()}} and will only have
effects in Word or RTF outputs.

Once a set of tabulation marks settings is defined, tabulation marks can
be added with \code{\link[=run_tab]{run_tab()}} inside a call to \code{\link[=fpar]{fpar()}} or
with \verb{\\t} within 'flextable' content.
}
\examples{
z <- fp_tabs(
  fp_tab(pos = 0.4, style = "decimal"),
  fp_tab(pos = 1, style = "decimal")
)
fpar(
  run_tab(), ftext("88."),
  run_tab(), ftext("987.45"),
  fp_p = fp_par(
    tabs = z
  )
)
}
\seealso{
Other functions for defining formatting properties: 
\code{\link{fp_border}()},
\code{\link{fp_cell}()},
\code{\link{fp_par}()},
\code{\link{fp_tab}()},
\code{\link{fp_text}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{read_xlsx}
\alias{read_xlsx}
\alias{length.rxlsx}
\alias{print.rxlsx}
\title{Create an 'Excel' document object}
\usage{
read_xlsx(path = NULL)

\method{length}{rxlsx}(x)

\method{print}{rxlsx}(x, target = NULL, ...)
}
\arguments{
\item{path}{path to the xlsx file to use as base document.}

\item{x}{an rxlsx object}

\item{target}{path to the xlsx file to write}

\item{...}{unused}
}
\description{
Read and import an xlsx file as an R object
representing the document. This function is experimental.
}
\examples{
read_xlsx()
x <- read_xlsx()
print(x, target = tempfile(fileext = ".xlsx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{ooxml_chart_uris}
\alias{ooxml_chart_uris}
\title{URIs for chart relationships in xlsx drawings}
\usage{
ooxml_chart_uris(kind = c("drawingml", "chartex"))
}
\arguments{
\item{kind}{chart family. \code{"drawingml"} covers classic charts
(bar, line, pie, scatter, area, bubble, radar, stock).
\code{"chartex"} covers the newer family (boxplot, funnel, histogram,
pareto, sunburst, treemap, waterfall).}
}
\value{
a named list with elements \code{rel_type} and \code{graphic_uri}.
}
\description{
Returns the pair of URIs needed by
\link{xlsx_drawing} to embed a chart into a sheet, keeping the
relationship type and graphic URI consistent so Excel opens the
file without repair.
}
\examples{
ooxml_chart_uris("drawingml")
ooxml_chart_uris("chartex")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{stop_if_not_in_slide_range}
\alias{stop_if_not_in_slide_range}
\title{Ensure valid slide indexes}
\usage{
stop_if_not_in_slide_range(x, idx, arg = NULL, call = parent.frame())
}
\arguments{
\item{x}{An \code{rpptx} object.}

\item{idx}{Slide indexes.}

\item{arg}{Name of argument to use in error message (optional).}

\item{call}{Environment to display in error message. Defaults to caller env.
Set \code{NULL} to suppress (see \link[cli:cli_abort]{cli::cli_abort}).}
}
\description{
Ensure valid slide indexes
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_pptx.R
\name{read_pptx}
\alias{read_pptx}
\title{Create a 'PowerPoint' document object}
\usage{
read_pptx(path = NULL)
}
\arguments{
\item{path}{path to the pptx file to use as base document. \code{potx} file are supported.}
}
\description{
Read and import a pptx file as an R object
representing the document.

The function is called \code{read_pptx} because it allows you to initialize an
object of class \code{rpptx} from an existing PowerPoint file. Content will be
added to the existing presentation. By default, an empty document is used.
}
\section{master layouts and slide layouts}{

\code{read_pptx()} uses a PowerPoint file as the initial document. This is the original
PowerPoint document where all slide layouts, placeholders for shapes and
styles come from. Major points to be aware of are:
\itemize{
\item Slide layouts are relative to a master layout. A document can contain one or
more master layouts; a master layout can contain one or more slide layouts.
\item A slide layout inherits design properties from its master layout but some
properties can be overwritten.
\item Designs and formatting properties of layouts and shapes (placeholders in a
layout) are defined within the initial document. There is no R function to
modify these values - they must be defined in the initial document.
}
}

\examples{
read_pptx()
}
\seealso{
\code{\link[=print.rpptx]{print.rpptx()}}, \code{\link[=add_slide]{add_slide()}}, \code{\link[=plot_layout_properties]{plot_layout_properties()}}, \code{\link[=ph_with]{ph_with()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add}
\alias{body_add}
\alias{body_add.character}
\alias{body_add.numeric}
\alias{body_add.factor}
\alias{body_add.fpar}
\alias{body_add.data.frame}
\alias{body_add.block_caption}
\alias{body_add.block_list}
\alias{body_add.block_toc}
\alias{body_add.external_img}
\alias{body_add.run_pagebreak}
\alias{body_add.run_columnbreak}
\alias{body_add.gg}
\alias{body_add.plot_instr}
\alias{body_add.block_pour_docx}
\alias{body_add.block_section}
\title{Add content into a Word document}
\usage{
body_add(x, value, ...)

\method{body_add}{character}(x, value, style = NULL, ...)

\method{body_add}{numeric}(x, value, style = NULL, format_fun = formatC, ...)

\method{body_add}{factor}(x, value, style = NULL, format_fun = as.character, ...)

\method{body_add}{fpar}(x, value, style = NULL, ...)

\method{body_add}{data.frame}(
  x,
  value,
  style = NULL,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
)

\method{body_add}{block_caption}(x, value, ...)

\method{body_add}{block_list}(x, value, ...)

\method{body_add}{block_toc}(x, value, ...)

\method{body_add}{external_img}(x, value, style = "Normal", ...)

\method{body_add}{run_pagebreak}(x, value, style = NULL, ...)

\method{body_add}{run_columnbreak}(x, value, style = NULL, ...)

\method{body_add}{gg}(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  unit = "in",
  ...
)

\method{body_add}{plot_instr}(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  unit = "in",
  ...
)

\method{body_add}{block_pour_docx}(x, value, ...)

\method{body_add}{block_section}(x, value, ...)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{object to add in the document. Supported objects
are vectors, data.frame, graphics, block of formatted paragraphs,
unordered list of formatted paragraphs,
pretty tables with package flextable, 'Microsoft' charts with package mschart.}

\item{...}{further arguments passed to or from other methods. When
adding a \code{ggplot} object or \link{plot_instr}, these arguments will be used
by png function. See method signatures to see what arguments can be used.}

\item{style}{paragraph style name. These names are available with function \link{styles_info}
and are the names of the Word styles defined in the base document (see
argument \code{path} from \link{read_docx}).}

\item{format_fun}{a function to be used to format values.}

\item{header}{display header if TRUE}

\item{tcf}{conditional formatting settings defined by \code{\link[=table_conditional_formatting]{table_conditional_formatting()}}}

\item{alignment}{columns alignement, argument length must match with columns length,
values must be "l" (left), "r" (right) or "c" (center).}

\item{width, height}{plot size in units expressed by the unit argument.
Defaults to a width of 6 and a height of 5 "in"ches.}

\item{res}{resolution of the png image in ppi}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}
}
\description{
This function adds objects into a Word document. Values are added
as new paragraphs or tables.

This function is experimental and will replace the \verb{body_add_*} functions
later. For now it is only to be used for successive additions and cannot
be used in conjunction with the \verb{body_add_*} functions.
}
\section{Methods (by class)}{
\itemize{
\item \code{body_add(character)}: add a character vector.

\item \code{body_add(numeric)}: add a numeric vector.

\item \code{body_add(factor)}: add a factor vector.

\item \code{body_add(fpar)}: add a \link{fpar} object. These objects enable
the creation of formatted paragraphs made of formatted chunks of text.

\item \code{body_add(data.frame)}: add a data.frame object with \code{\link[=block_table]{block_table()}}.

\item \code{body_add(block_caption)}: add a \link{block_caption} object. These objects enable
the creation of set of formatted paragraphs made of formatted chunks of text.

\item \code{body_add(block_list)}: add a \link{block_list} object. Use this function
to add a list of block elements (e.g. paragraphs, images, tables)
into a Word document in a more efficient way than with usual
\verb{body_add_*} functions. This function will add several elements
in a faster way because the cursor is not calculated for
each iteraction over the elements, as a consequence the
function only append elements at the end of the document
and does not allow to insert elements at a specific position.

\item \code{body_add(block_toc)}: add a table of content (a \link{block_toc} object).

\item \code{body_add(external_img)}: add an image (a \link{external_img} object).

\item \code{body_add(run_pagebreak)}: add a \link{run_pagebreak} object.

\item \code{body_add(run_columnbreak)}: add a \link{run_columnbreak} object.

\item \code{body_add(gg)}: add a ggplot object.

\item \code{body_add(plot_instr)}: add a base plot with a \link{plot_instr} object.

\item \code{body_add(block_pour_docx)}: pour content of an external docx file with with a \link{block_pour_docx} object

\item \code{body_add(block_section)}: ends a section with a \link{block_section} object

}}
\section{Illustrations}{


\if{html}{\figure{body_add_doc_1.png}{options: style="width:70\%;"}}
}

\examples{
# This example demonstrates the versatility of body_add()
# by showing how to add various content types to a Word document

# Create a new Word document
x <- read_docx()

# Add a heading for the table of contents section
x <- body_add(x, "Table of content", style = "heading 1")

# Insert an automatic table of contents
# This will list all headings in the document
x <- body_add(x, block_toc())

# Insert a page break to start content on a new page
x <- body_add(x, run_pagebreak())

# Add a main section heading
x <- body_add(x, "Iris Dataset Sample", style = "heading 1")

# Add a data.frame as a table
# The first 6 rows of the iris dataset will be displayed
x <- body_add(x, head(iris), style = "table_template")

# Add another section heading
x <- body_add(x, "Alphabetic List", style = "heading 1")

# Add a character vector as paragraphs
# Each letter will appear as a separate paragraph with "Normal" style
x <- body_add(x, letters, style = "Normal")

# Insert a continuous section break
# Content continues on the same page but section properties can change
x <- body_add(x, block_section(prop_section(type = "continuous")))

# Add a plot created with R base graphics
# plot_instr() captures the plotting code to generate the image
x <- body_add(x, plot_instr(code = barplot(1:5, col = 2:6)))

# Change to landscape orientation for the next section
# This creates a new section with landscape page layout
x <- body_add(
  x,
  block_section(prop_section(page_size = page_size(orient = "landscape")))
)

# Save the document to a file
output_file <- tempfile(fileext = ".docx")
print(x, target = output_file)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/empty_content.R
\name{empty_content}
\alias{empty_content}
\title{Empty block for 'PowerPoint'}
\usage{
empty_content()
}
\description{
Create an empty object to include as an empty placeholder shape in a
presentation. This comes in handy when presentation are updated
through R, but a user still wants to add some comments in this new content.

Empty content also works with layout fields (slide number and date) to preserve them:
they are included on the slide and keep being updated by PowerPoint, i.e. update to the
when the slide number when the slide moves in the deck, update to the date.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
doc <- read_pptx()
doc <- add_slide(doc, layout = "Two Content",
  master = "Office Theme")
doc <- ph_with(x = doc, value = empty_content(),
 location = ph_location_type(type = "title") )

doc <- add_slide(doc, "Title and Content")
# add slide number as a computer field
doc <- ph_with(
  x = doc, value = empty_content(),
  location = ph_location_type(type = "sldNum"))

print(doc, target = fileout )
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=body_add_blocks]{body_add_blocks()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{move_slide}
\alias{move_slide}
\title{Move a slide}
\usage{
move_slide(x, index = NULL, to)
}
\arguments{
\item{x}{an rpptx object}

\item{index}{slide index, default to current slide position.}

\item{to}{new slide index.}
}
\description{
Move a slide in a pptx presentation.
}
\note{
cursor is set on the last slide.
}
\examples{
x <- read_pptx()
x <- add_slide(x)
x <- ph_with(x, "Hello world 1", location = ph_location_type())
x <- add_slide(x)
x <- ph_with(x, "Hello world 2", location = ph_location_type())
x <- move_slide(x, index = 1, to = 2)
}
\seealso{
\code{\link[=read_pptx]{read_pptx()}}

Other functions slide manipulation: 
\code{\link{add_slide}()},
\code{\link{on_slide}()},
\code{\link{remove_slide}()},
\code{\link{set_notes}()}
}
\concept{functions slide manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_gg}
\alias{block_gg}
\title{'ggplot' block}
\usage{
block_gg(
  value,
  fp_p = fp_par(),
  width = 6,
  height = 5,
  res = 300,
  scale = 1,
  unit = "in"
)
}
\arguments{
\item{value}{'ggplot' object}

\item{fp_p}{paragraph formatting properties, see \code{\link[=fp_par]{fp_par()}}}

\item{width, height}{image size in units expressed by the unit argument.
Defaults to "in"ches.}

\item{res}{resolution of the png image in ppi}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}
}
\description{
A simple wrapper to add a 'ggplot' object as a png in a document.
It produces an object of class 'block_gg' with a corresponding method \code{\link[=to_wml]{to_wml()}}
that can be used to convert the object to a WordML string.
}
\examples{
library(officer)
if (require("ggplot2")) {
  set.seed(2)
  doc <- read_docx()

  z <- body_append_start_context(doc)

  for (i in seq_len(3)) {
    df <- data.frame(x = runif(10), y = runif(10))
    gg <- ggplot(df, aes(x = x, y = y)) + geom_point()

    write_elements_to_context(
      context = z,
      block_gg(
        value = gg
      )
    )
  }

  doc <- body_append_stop_context(z)

  print(doc, target = tempfile(fileext = ".docx"))
}
}
\seealso{
Other block functions for reporting:
\code{\link[=block_caption]{block_caption()}},
\code{\link[=block_list]{block_list()}},
\code{\link[=block_list_items]{block_list_items()}},
\code{\link[=block_pour_docx]{block_pour_docx()}},
\code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}},
\code{\link[=block_toc]{block_toc()}},
\code{\link[=fpar]{fpar()}},
\code{\link[=list_item]{list_item()}},
\code{\link[=plot_instr]{plot_instr()}},
\code{\link[=unordered_list]{unordered_list()}}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{body_replace_all_text}
\alias{body_replace_all_text}
\alias{headers_replace_all_text}
\alias{footers_replace_all_text}
\title{Replace text anywhere in the document}
\usage{
body_replace_all_text(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
)

headers_replace_all_text(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
)

footers_replace_all_text(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
)
}
\arguments{
\item{x}{a docx device}

\item{old_value}{the value to replace}

\item{new_value}{the value to replace it with}

\item{only_at_cursor}{if \code{TRUE}, only search-and-replace at the current
cursor; if \code{FALSE} (default), search-and-replace in the entire document
(this can be slow on large documents!)}

\item{warn}{warn if \code{old_value} could not be found.}

\item{...}{optional arguments to grepl/gsub (e.g. \code{fixed=TRUE})}
}
\description{
Replace text anywhere in the document, or at a cursor.

Replace all occurrences of old_value with new_value. This method
uses \code{\link[=grepl]{grepl()}}/\code{\link[=gsub]{gsub()}} for pattern matching; you may
supply arguments as required (and therefore use \code{\link[=regex]{regex()}} features)
using the optional \code{...} argument.

Note that by default, grepl/gsub will use \code{fixed=FALSE}, which means
that \code{old_value} and \code{new_value} will be interepreted as regular
expressions.

\strong{Chunking of text}

Note that the behind-the-scenes representation of text in a Word document is
frequently not what you might expect! Sometimes a paragraph of text is broken
up (or "chunked") into several "runs," as a result of style changes, pauses
in text entry, later revisions and edits, etc. If you have not styled the
text, and have entered it in an "all-at-once" fashion, e.g. by pasting it or
by outputing it programmatically into your Word document, then this will
likely not be a problem. If you are working with a manually-edited document,
however, this can lead to unexpected failures to find text.

You can use the officer function \code{\link[=docx_show_chunk]{docx_show_chunk()}} to
show how the paragraph of text at the current cursor has been chunked into
runs, and what text is in each chunk. This can help troubleshoot unexpected
failures to find text.
}
\section{header_replace_all_text}{

Replacements will be performed in each header of all sections.


Replacements will be performed in each footer of all sections.
}

\examples{
library(officer)

doc <- read_docx()
doc <- body_add_par(doc, "Placeholder one")
doc <- body_add_par(doc, "Placeholder two")

# Show text chunk at cursor
docx_show_chunk(doc) # Output is 'Placeholder two'

# Simple search-and-replace at current cursor, with regex turned off
doc <- body_replace_all_text(
  doc,
  old_value = "Placeholder",
  new_value = "new",
  only_at_cursor = TRUE,
  fixed = TRUE
)
docx_show_chunk(doc) # Output is 'new two'

# Do the same, but in the entire document and ignoring case
doc <- body_replace_all_text(
  doc,
  old_value = "placeholder",
  new_value = "new",
  only_at_cursor = FALSE,
  ignore.case = TRUE
)
doc <- cursor_backward(doc)
docx_show_chunk(doc) # Output is 'new one'

# Use regex : replace all words starting with "n" with the word "example"
doc <- body_replace_all_text(doc, "\\\\bn.*?\\\\b", "example")
docx_show_chunk(doc) # Output is 'example one'
}
\seealso{
\code{\link[=grepl]{grepl()}}, \code{\link[=regex]{regex()}}, \code{\link[=docx_show_chunk]{docx_show_chunk()}}
}
\author{
Frank Hangler, \email{frank@plotandscatter.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{read_docx}
\alias{read_docx}
\title{Create a 'Word' document object}
\usage{
read_docx(path = NULL)
}
\arguments{
\item{path}{path to the docx file to use as base document.
\code{dotx} file are supported.}
}
\value{
an object of class \code{rdocx}.
}
\description{
\code{read_docx()} is the starting point for creating Word documents
from R. It creates an R object representing a Word document that can be
manipulated programmatically. When called without arguments, it creates an
empty document based on a default template. When provided with a path, it
reads an existing Word document (\code{.docx}) or template (\code{.dotx}) file.

Once created, you can:
\itemize{
\item Add content from R: Insert text, formatted paragraphs (\code{\link[=fpar]{fpar()}}),
tables (\code{\link[=body_add_table]{body_add_table()}}), plots (\code{\link[=body_add_plot]{body_add_plot()}}), images,
page breaks, table of contents, and more using the \verb{body_add_*()} family
of functions.
\item Read and inspect content: Use \code{\link[=docx_summary]{docx_summary()}} to extract and analyze
the document's content, structure, and formatting as a data frame.
\item Write to file: Save the document to disk using
\code{print(x, target = "path/to/file.docx")}.
}
}
\section{styles}{


The template file (specified via \code{path} or the default template) determines
the available paragraph styles, character styles, and table styles in your
document. These styles control the appearance of headings, body text, tables,
and other elements.

When you use functions like \code{body_add_par(style = "heading 2")}, the style
name must exist in the template. You can:
\itemize{
\item Use \code{\link[=styles_info]{styles_info()}} to list all available styles in your document
\item Create a custom template in Word with your organization's styles and branding
\item Use the default template for standard documents
}

The document layout (page size, margins, headers and footer content,
orientation) also comes from the
template and can be modified using \code{\link[=body_set_default_section]{body_set_default_section()}} or by
adding section breaks with \code{\link[=body_end_section_continuous]{body_end_section_continuous()}} and related functions.
}

\section{Illustrations}{


\if{html}{\figure{read_docx.png}{options: style="width:80\%;"}}
}

\examples{
library(officer)

# This example demonstrates how to create
# an empty document -----

## Create a new Word document
doc <- read_docx()
## Save the document
output_file <- print(doc, target = tempfile(fileext = ".docx"))

# This example demonstrates how to create a document
# with text, formatted paragraphs, tables, and plots ----
# organized in sections

# Create a new Word document
doc <- read_docx()

# Add main title
doc <- body_add_par(doc, "Annual Sales Report", style = "heading 1")

# Add introduction with formatted text using fpar
intro_text <- fpar(
  "This report presents the ",
  ftext(
    "quarterly sales analysis",
    fp_text_lite(bold = TRUE, color = "#C32900")
  ),
  " for the fiscal year. The following sections provide detailed insights ",
  "into our performance metrics and trends."
)
doc <- body_add_fpar(doc, intro_text)

## Section 1: Sales Data ----
doc <- body_add_par(doc, "Sales Performance", style = "heading 2")

# Add descriptive text
doc <- body_add_par(
  doc,
  "The table below summarizes sales data across different product categories:"
)

# Create and add a data frame as a table
sales_data <- data.frame(
  Quarter = c("Q1", "Q2", "Q3", "Q4"),
  Revenue = c(125000, 142000, 156000, 178000),
  Units = c(1250, 1420, 1560, 1780),
  Growth = c("5\%", "13.6\%", "9.9\%", "14.1\%"),
  stringsAsFactors = FALSE
)
doc <- body_add_table(doc, value = sales_data, style = "table_template")

# Add commentary with multiple formatted text elements
comment <- fpar(
  "Key finding: ",
  ftext(
    "Q4 showed the strongest performance",
    fp_text_lite(bold = TRUE, font.size = 11)
  ),
  " with a ",
  ftext("14.1\% growth rate", fp_text_lite(color = "#006699", bold = TRUE)),
  " compared to the previous quarter."
)
doc <- body_add_fpar(doc, comment)

## Section 2: Visualizations ----
doc <- body_add_par(doc, "Revenue Trends", style = "heading 2")

# Add explanatory text
doc <- body_add_par(
  doc,
  "Figure 1 illustrates the quarterly revenue progression throughout the year."
)

# Create a plot showing revenue trends
revenue_plot <- plot_instr({
  quarters <- c("Q1", "Q2", "Q3", "Q4")
  revenue <- c(125000, 142000, 156000, 178000)
  barplot(
    revenue,
    names.arg = quarters,
    col = "#4472C4",
    border = NA,
    main = "Quarterly Revenue",
    ylab = "Revenue ($)",
    xlab = "Quarter",
    ylim = c(0, 200000)
  )
  grid(nx = NA, ny = NULL, col = "gray90", lty = 1)
})
doc <- body_add_plot(doc, revenue_plot, width = 6, height = 4)

# Add another section with a different plot
doc <- body_add_par(doc, "Sales Distribution Analysis", style = "heading 2")

# Add context for the second plot
analysis_intro <- fpar(
  "The distribution analysis below shows the ",
  ftext("variability in daily sales", fp_text_lite(italic = TRUE)),
  " across all quarters. This helps identify patterns and outliers in our sales data."
)
doc <- body_add_fpar(doc, analysis_intro)

# Create a density plot
distribution_plot <- plot_instr({
  # Simulate daily sales data
  set.seed(123)
  daily_sales <- c(
    rnorm(90, mean = 1400, sd = 200), # Q1-Q3
    rnorm(30, mean = 2000, sd = 250) # Q4 (higher mean)
  )
  plot(
    density(daily_sales),
    main = "Distribution of Daily Sales",
    xlab = "Daily Sales (Units)",
    ylab = "Density",
    col = "#C32900",
    lwd = 2
  )
  polygon(density(daily_sales), col = rgb(0.76, 0.16, 0, 0.2), border = NA)
  abline(v = mean(daily_sales), col = "#006699", lwd = 2, lty = 2)
  legend("topright", legend = "Mean", col = "#006699", lty = 2, lwd = 2)
})
doc <- body_add_plot(doc, distribution_plot, width = 6, height = 4)

# Add concluding remarks
doc <- body_add_par(doc, "Conclusion", style = "heading 2")
conclusion <- fpar(
  "The analysis demonstrates ",
  ftext("consistent growth", fp_text_lite(bold = TRUE, color = "#006699")),
  " throughout the year, with particularly strong performance in Q4. ",
  "This trend suggests effective market strategies and increasing customer demand."
)
doc <- body_add_fpar(doc, conclusion)

# Save the document
comprehensive_file <- print(doc, target = tempfile(fileext = ".docx"))
# Using a custom template ----
# This example shows how to start from an existing template
# instead of creating a blank document

# Get the path to a landscape template included in the package
template <- system.file(package = "officer", "doc_examples", "landscape.docx")

# Create a document based on the template
# The document will inherit the template's styles and page settings
doc_2 <- read_docx(path = template)

# Add a section with a table
doc_2 <- body_add_par(doc_2, "Motor Trend Car Data", style = "heading 2")
doc_2 <- body_add_table(doc_2, value = head(mtcars))

# Add a section with a plot
doc_2 <- body_add_par(doc_2, "Sales Distribution", style = "heading 2")
doc_2 <- body_add_plot(doc_2, distribution_plot)

# Save the document
docx_file_output <- print(doc_2, target = tempfile(fileext = ".docx"))
}
\seealso{
Save a 'Word' document object to a file with \code{\link[=print.rdocx]{print.rdocx()}},
add content with functions \code{\link[=body_add_par]{body_add_par()}}, \code{\link[=body_add_plot]{body_add_plot()}},
\code{\link[=body_add_table]{body_add_table()}}, change settings with \code{\link[=docx_set_settings]{docx_set_settings()}}, set
properties with \code{\link[=set_doc_properties]{set_doc_properties()}},  read 'Word' styles with
\code{\link[=styles_info]{styles_info()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_with_methods.R
\name{ph_with}
\alias{ph_with}
\alias{ph_with.character}
\alias{ph_with.numeric}
\alias{ph_with.factor}
\alias{ph_with.logical}
\alias{ph_with.Date}
\alias{ph_with.block_list}
\alias{ph_with.unordered_list}
\alias{ph_with.data.frame}
\alias{ph_with.gg}
\alias{ph_with.plot_instr}
\alias{ph_with.external_img}
\alias{ph_with.fpar}
\alias{ph_with.empty_content}
\alias{ph_with.xml_document}
\title{Add objects on the current slide}
\usage{
ph_with(x, value, location, ...)

ph_with.character(x, value, location, ...)

ph_with.numeric(x, value, location, format_fun = format, ...)

ph_with.factor(x, value, location, ...)

ph_with.logical(x, value, location, format_fun = format, ...)

ph_with.Date(x, value, location, date_format = NULL, ...)

ph_with.block_list(x, value, location, level_list = integer(0), ...)

ph_with.unordered_list(x, value, location, ...)

ph_with.data.frame(
  x,
  value,
  location,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
)

ph_with.gg(x, value, location, res = 300, alt_text = "", scale = 1, ...)

ph_with.plot_instr(x, value, location, res = 300, ...)

ph_with.external_img(x, value, location, use_loc_size = TRUE, ...)

ph_with.fpar(x, value, location, ...)

ph_with.empty_content(x, value, location, ...)

ph_with.xml_document(x, value, location, ...)
}
\arguments{
\item{x}{an rpptx object}

\item{value}{object to add as a new shape. Supported objects
are vectors, data.frame, graphics, block of formatted paragraphs,
unordered list of formatted paragraphs,
pretty tables with package flextable, editable graphics with
package rvg, 'Microsoft' charts with package mschart.}

\item{location}{a placeholder location object or a location short form. It will be used
to specify the location of the new shape. This location can be defined with a call to one
of the \verb{ph_location_*} functions (see section \code{"see also"}). In \code{ph_with()}, several location
short forms can be used, as listed in section \code{"Short forms"}.}

\item{...}{further arguments passed to or from other methods. When
adding a \code{ggplot} object or \code{plot_instr}, these arguments will be used
by the png function.}

\item{format_fun}{format function for non character vectors}

\item{date_format}{A format string for dates (default \code{"\%Y-\%m-\%d"}).
See \code{format} arg in \code{\link[=strftime]{strftime()}} for details.
Set a global default via \code{options(officer.date_format = ...)}.}

\item{level_list}{The list of levels for hierarchy structure as integer values.
If used the object is formated as an unordered list. If 1 and 2,
item 1 level will be 1, item 2 level will be 2.}

\item{header}{display header if TRUE}

\item{tcf}{conditional formatting settings defined by \code{\link[=table_conditional_formatting]{table_conditional_formatting()}}}

\item{alignment}{alignment for each columns, 'l' for left, 'r' for right
and 'c' for center. Default to NULL.}

\item{res}{resolution of the png image in ppi}

\item{alt_text}{Alt-text for screen-readers. Defaults to \code{""}. If \code{""} or \code{NULL}
an alt text added with \code{ggplot2::labs(alt = ...)} will be used if any.}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{use_loc_size}{if set to FALSE, external_img width and height will
be used.}
}
\description{
add an object into a new shape in the current slide. This
function is able to add all supported outputs to a presentation. See
section \strong{Methods (by class)} to see supported outputs.
}
\section{Functions}{
\itemize{
\item \code{ph_with.character()}: add a character vector to a new shape on the
current slide, values will be added as paragraphs.

\item \code{ph_with.numeric()}: add a numeric vector to a new shape on the
current slide, values will be be first formatted then
added as paragraphs.

\item \code{ph_with.factor()}: add a factor vector to a new shape on the
current slide, values will be be converted as character and then
added as paragraphs.

\item \code{ph_with.Date()}: add a \code{Date} object vector to a new shape on the
current slide, values will be be first converted to character.

\item \code{ph_with.block_list()}: add a \code{\link[=block_list]{block_list()}} made
of \code{\link[=fpar]{fpar()}} to a new shape on the current slide.

\item \code{ph_with.unordered_list()}: add a \code{\link[=unordered_list]{unordered_list()}} made
of \code{\link[=fpar]{fpar()}} to a new shape on the current slide.

\item \code{ph_with.data.frame()}: add a data.frame to a new shape on the current slide with
function \code{\link[=block_table]{block_table()}}. Use package 'flextable' instead for more
advanced formattings.

\item \code{ph_with.gg()}: add a ggplot object to a new shape on the
current slide. Use package 'rvg' for more advanced graphical features.

\item \code{ph_with.plot_instr()}: add an R plot to a new shape on the
current slide. Use package 'rvg' for more advanced graphical features.

\item \code{ph_with.external_img()}: add a \code{\link[=external_img]{external_img()}} to a new shape
on the current slide.

When value is a external_img object, image will be copied
into the PowerPoint presentation. The width and height
specified in call to \code{\link[=external_img]{external_img()}} will be
ignored, their values will be those of the location,
unless use_loc_size is set to FALSE.

\item \code{ph_with.fpar()}: add an \code{\link[=fpar]{fpar()}} to a new shape
on the current slide as a single paragraph in a \code{\link[=block_list]{block_list()}}.

\item \code{ph_with.empty_content()}: add an \code{\link[=empty_content]{empty_content()}} to a new shape
on the current slide.

\item \code{ph_with.xml_document()}: add an xml_document object to a new shape on the
current slide. This function is to be used to add custom openxml code.

}}
\section{Short forms}{

The \code{location} argument of \code{ph_with()} either expects a location object as returned by the
\verb{ph_location_*} functions or a corresponding location \emph{short form} (string or numeric):\tabular{lll}{
   \strong{Location function} \tab \strong{Short form} \tab \strong{Description} \cr
   \code{ph_location_left()} \tab \code{"left"} \tab Keyword string \cr
   \code{ph_location_right()} \tab \code{"right"} \tab Keyword string \cr
   \code{ph_location_fullsize()} \tab \code{"fullsize"} \tab Keyword string \cr
   \code{ph_location_type("body", 1)} \tab \code{"body [1]"} \tab String: type + index in brackets (\code{1} if omitted) \cr
   \code{ph_location_label("my_label")} \tab \code{"my_label"} \tab Any string not matching a keyword or type \cr
   \code{ph_location_id(1)} \tab \code{1} \tab Length 1 integer \cr
   \code{ph_location(0, 0, 4, 5)} \tab \code{c(0,0,4,5)} \tab Length 4 numeric, optionally named, \code{c(top=0, left=0, ...)} \cr
}
}

\section{Illustrations}{


\if{html}{\figure{ph_with_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
# this name will be used to print the file
# change it to "youfile.pptx" to write the pptx
# file in your working directory.
fileout <- tempfile(fileext = ".pptx")

doc_1 <- read_pptx()
sz <- slide_size(doc_1)

# add text and a table ----
doc_1 <- add_slide(doc_1, layout = "Two Content", master = "Office Theme")
doc_1 <- ph_with(
  x = doc_1,
  value = c("Table cars"),
  location = ph_location_type(type = "title")
)
doc_1 <- ph_with(
  x = doc_1,
  value = names(cars),
  location = ph_location_left()
)
doc_1 <- ph_with(
  x = doc_1,
  value = cars,
  location = ph_location_right()
)
doc_1 <- ph_with(
  x = doc_1,
  value = Sys.Date(),
  location = ph_location_type("dt")
)

# add a base plot ----
anyplot <- plot_instr(code = {
  col <- c(
    "#440154FF",
    "#443A83FF",
    "#31688EFF",
    "#21908CFF",
    "#35B779FF",
    "#8FD744FF",
    "#FDE725FF"
  )
  barplot(1:7, col = col, yaxt = "n")
})

doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  doc_1,
  anyplot,
  location = ph_location_fullsize(),
  bg = "#006699"
)

# add a ggplot2 plot ----
if (require("ggplot2")) {
  doc_1 <- add_slide(doc_1, "Title and Content")
  gg_plot <- ggplot(data = iris) +
    geom_point(
      mapping = aes(Sepal.Length, Petal.Length),
      size = 3
    ) +
    theme_minimal()
  doc_1 <- ph_with(
    x = doc_1,
    value = gg_plot,
    location = ph_location_type(type = "body"),
    bg = "transparent"
  )
  doc_1 <- ph_with(
    x = doc_1,
    value = "graphic title",
    location = ph_location_type(type = "title")
  )
}

# add a external images ----
doc_1 <- add_slide(doc_1, layout = "Title and Content", master = "Office Theme")
doc_1 <- ph_with(
  x = doc_1,
  value = empty_content(),
  location = ph_location(
    left = 0,
    top = 0,
    width = sz$width,
    height = sz$height,
    bg = "black"
  )
)

svg_file <- file.path(R.home(component = "doc"), "html/Rlogo.svg")
if (require("rsvg")) {
  doc_1 <- ph_with(
    x = doc_1,
    value = "External images",
    location = ph_location_type(type = "title")
  )
  doc_1 <- ph_with(
    x = doc_1,
    external_img(svg_file, 100 / 72, 76 / 72),
    location = ph_location_right(),
    use_loc_size = FALSE
  )
  doc_1 <- ph_with(
    x = doc_1,
    external_img(svg_file),
    location = ph_location_left(),
    use_loc_size = TRUE
  )
}

# add a block_list ----
dummy_text <- readLines(system.file(
  package = "officer",
  "doc_examples/text.txt"
))
fp_1 <- fp_text(bold = TRUE, color = "pink", font.size = 0)
fp_2 <- fp_text(bold = TRUE, font.size = 0)
fp_3 <- fp_text(italic = TRUE, color = "red", font.size = 0)
bl <- block_list(
  fpar(ftext("hello world", fp_1)),
  fpar(
    ftext("hello", fp_2),
    ftext("hello", fp_3)
  ),
  dummy_text
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = bl,
  location = ph_location_type(type = "body")
)


# fpar ------
fpt <- fp_text(
  bold = TRUE,
  font.family = "Bradley Hand",
  font.size = 150,
  color = "#F5595B"
)
hw <- fpar(
  ftext("hello ", fpt),
  hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "cran",
    prop = fpt
  )
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = hw,
  location = ph_location_type(type = "body")
)
# unordered_list ----
ul <- unordered_list(
  level_list = c(1, 2, 2, 3, 3, 1),
  str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
  style = fp_text(color = "red", font.size = 0)
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = ul,
  location = ph_location_type()
)

print(doc_1, target = fileout)


# Example using short-form locations ----
x <- read_pptx()
x <- add_slide(x, "Title Slide")
x <- ph_with(x, "A title", "Title 1") # label
x <- ph_with(x, "A subtitle", 3) # id
x <- ph_with(x, "A left text", "left") # keyword
x <- ph_with(x, Sys.Date(), "dt[1]") # type + index
x <- ph_with(x, "More content", c(5, .5, 5, 2)) # numeric vector (left, top, width, heigh)
# \dontrun{
# print(x, preview = TRUE) # opens file locally
# }
}
\seealso{
Specify placeholder locations with \link{ph_location_type}, \link{ph_location},
\link{ph_location_label}, \link{ph_location_left}, \link{ph_location_right},
\link{ph_location_fullsize}, \link{ph_location_template}. \link{phs_with} is a sibling of
\code{ph_with} that fills multiple placeholders at once. Use \link{add_slide} to add new slides.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{plot_instr}
\alias{plot_instr}
\title{Wrap plot instructions for png plotting in Powerpoint or Word}
\usage{
plot_instr(code)
}
\arguments{
\item{code}{plotting instructions}
}
\description{
A simple wrapper to capture
plot instructions that will be executed and copied in a document. It produces
an object of class 'plot_instr' with a corresponding method \code{\link[=ph_with]{ph_with()}} and
\code{\link[=body_add_plot]{body_add_plot()}}.

The function enable usage of any R plot with argument \code{code}. Wrap your code
between curly bracket if more than a single expression.
}
\examples{
# plot_instr demo ----

anyplot <- plot_instr(code = {
  barplot(1:5, col = 2:6)
  })

doc <- read_docx()
doc <- body_add(doc, anyplot, width = 5, height = 4)
print(doc, target = tempfile(fileext = ".docx"))


doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(
  doc, anyplot,
  location = ph_location_fullsize(),
  bg = "#00000066", pointsize = 12)
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=body_add_plot]{body_add_plot()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_table}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_list_items}
\alias{block_list_items}
\title{List of items for Word and PowerPoint}
\usage{
block_list_items(..., list_type = "bullet")
}
\arguments{
\item{...}{\code{\link[=list_item]{list_item()}} objects}

\item{list_type}{\code{"bullet"} for an unordered list or \code{"decimal"}
for a numbered list}
}
\description{
Create a bullet or numbered list from \code{\link[=list_item]{list_item()}}
elements. Supports rich text via \code{\link[=fpar]{fpar()}} and multi-level nesting.
Works in both Word and PowerPoint documents.
}
\examples{
items <- block_list_items(
  list_item(fpar(
    ftext("Item 1", fp_text(color = "red"))
  ), level = 1),
  list_item(fpar("Sub-item"), level = 2),
  list_item(fpar("Item 2"), level = 1),
  list_type = "bullet"
)
items
}
\seealso{
\code{\link[=list_item]{list_item()}}, \code{\link[=body_add]{body_add()}}, \code{\link[=ph_with]{ph_with()}}

Other block functions for reporting:
\code{\link[=block_caption]{block_caption()}},
\code{\link[=block_gg]{block_gg()}},
\code{\link[=block_list]{block_list()}},
\code{\link[=block_pour_docx]{block_pour_docx()}},
\code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}},
\code{\link[=block_toc]{block_toc()}},
\code{\link[=fpar]{fpar()}},
\code{\link[=list_item]{list_item()}},
\code{\link[=plot_instr]{plot_instr()}},
\code{\link[=unordered_list]{unordered_list()}}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_bookmark}
\alias{run_bookmark}
\title{Bookmark for 'Word'}
\usage{
run_bookmark(bkm, run)
}
\arguments{
\item{bkm}{bookmark id to associate with run. Value can only be
made of alpha numeric characters, '-' and '_'.}

\item{run}{a run object, made with a call to one of
the "run functions for reporting".}
}
\description{
Add a bookmark on a run object.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
ft <- fp_text(font.size = 12, bold = TRUE)
run_bookmark("par1", ftext("some text", ft))
}
\seealso{
Other run functions for reporting:
\code{\link[=external_img]{external_img()}},
\code{\link[=floating_external_img]{floating_external_img()}},
\code{\link[=ftext]{ftext()}},
\code{\link[=hyperlink_ftext]{hyperlink_ftext()}},
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_columnbreak]{run_columnbreak()}},
\code{\link[=run_comment]{run_comment()}},
\code{\link[=run_footnote]{run_footnote()}},
\code{\link[=run_footnoteref]{run_footnoteref()}},
\code{\link[=run_linebreak]{run_linebreak()}},
\code{\link[=run_pagebreak]{run_pagebreak()}},
\code{\link[=run_reference]{run_reference()}},
\code{\link[=run_tab]{run_tab()}},
\code{\link[=run_word_field]{run_word_field()}},
\code{\link[=run_wordtext]{run_wordtext()}}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{slide_size}
\alias{slide_size}
\title{Slides width and height}
\usage{
slide_size(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Get the width and height of slides in inches as
a named vector.
}
\examples{
my_pres <- read_pptx()
my_pres <- add_slide(my_pres,
  layout = "Two Content", master = "Office Theme")
slide_size(my_pres)
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{table_layout}
\alias{table_layout}
\title{Algorithm for table layout}
\usage{
table_layout(type = "autofit")
}
\arguments{
\item{type}{'autofit' or 'fixed' algorithm. Default to 'autofit'.}
}
\description{
When a table is displayed in a document, it can
either be displayed using a fixed width or autofit layout algorithm:
\itemize{
\item fixed: uses fixed widths for columns. The width of the table is not
changed regardless of the contents of the cells.
\item autofit: uses the contents of each cell and the table width to
determine the final column widths.
}
}
\seealso{
Other functions for table definition: 
\code{\link{prop_table}()},
\code{\link{table_colwidths}()},
\code{\link{table_conditional_formatting}()},
\code{\link{table_stylenames}()},
\code{\link{table_width}()}
}
\concept{functions for table definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_matrix.R
\name{as.matrix.rpptx}
\alias{as.matrix.rpptx}
\title{PowerPoint table to matrix}
\usage{
\method{as.matrix}{rpptx}(
  x,
  ...,
  slide_id = NA_integer_,
  id = NA_character_,
  span = c(NA_character_, "fill")
)
}
\arguments{
\item{x}{The rpptx object to convert (as created by
\code{officer::read_pptx()})}

\item{...}{Ignored}

\item{slide_id}{The slide number to load from (NA indicates first slide with
a table, NULL indicates all slides and all tables)}

\item{id}{The table ID to load from (ignored it \code{is.null(slide_id)}, NA
indicates to load the first table from the \code{slide_id})}

\item{span}{How should col_span/row_span values be handled?  \code{NA} means
to leave the value as \code{NA}, and \code{"fill"} means to fill matrix
cells with the value.}
}
\value{
A matrix with the data, or if \code{slide_id=NULL}, a list of
matrices
}
\description{
Convert the data in an a 'PowerPoint' table
to a matrix or all data to a list of matrices.
}
\examples{
library(officer)
pptx_file <- system.file(package="officer", "doc_examples", "example.pptx")
z <- read_pptx(pptx_file)
as.matrix(z, slide_id = NULL)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_styles.R
\name{docx_set_paragraph_style}
\alias{docx_set_paragraph_style}
\title{Add or replace paragraph style in a Word document}
\usage{
docx_set_paragraph_style(
  x,
  style_id,
  style_name,
  base_on = "Normal",
  fp_p = fp_par(),
  fp_t = NULL
)
}
\arguments{
\item{x}{an rdocx object}

\item{style_id}{a unique style identifier for Word.}

\item{style_name}{a unique label associated with the style identifier.
This label is the name of the style when Word edit the document.}

\item{base_on}{the style name used as base style}

\item{fp_p}{paragraph formatting properties, see \code{\link[=fp_par]{fp_par()}}.}

\item{fp_t}{default text formatting properties. This is used as
text formatting properties, see \code{\link[=fp_text]{fp_text()}}. If NULL (default), the
paragraph will used the default text formatting properties (defined by
the \code{base_on} argument).}
}
\description{
The function lets you add or replace a Word paragraph style.
}
\examples{
library(officer)

doc <- read_docx()

doc <- docx_set_paragraph_style(
  doc,
  style_id = "rightaligned",
  style_name = "Explicit label",
  fp_p = fp_par(text.align = "right", padding = 20),
  fp_t = fp_text_lite(
    bold = TRUE,
    shading.color = "#FD34F0",
    color = "white")
)

doc <- body_add_par(doc,
  value = "This is a test",
  style = "Explicit label")

docx_file <- print(doc, target = tempfile(fileext = ".docx"))
docx_file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_block_section}
\alias{body_end_block_section}
\title{Add any section}
\usage{
body_end_block_section(x, value)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{a \link{block_section} object}
}
\description{
Add a section to the document. You can
define any section with a \link{block_section} object. All other
\verb{body_end_section_*} are specialized, this one is highly flexible
but it's up to the user to define the section properties.
}
\section{Illustrations}{


\if{html}{\figure{body_end_block_section_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
library(officer)
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 20)
str1 <- paste(str1, collapse = " ")

ps <- prop_section(
  page_size = page_size(orient = "landscape"),
  page_margins = page_mar(top = 2),
  type = "continuous"
)

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")

doc_1 <- body_end_block_section(doc_1, block_section(ps))

doc_1 <- body_add_par(doc_1, value = str1, style = "centered")

print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections: 
\code{\link{body_end_section_columns}()},
\code{\link{body_end_section_columns_landscape}()},
\code{\link{body_end_section_continuous}()},
\code{\link{body_end_section_landscape}()},
\code{\link{body_end_section_portrait}()},
\code{\link{body_set_default_section}()}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_autonum}
\alias{run_autonum}
\title{Auto number}
\usage{
run_autonum(
  seq_id = "table",
  pre_label = "Table ",
  post_label = ": ",
  bkm = NULL,
  bkm_all = FALSE,
  prop = NULL,
  start_at = NULL,
  tnd = 0,
  tns = "-"
)
}
\arguments{
\item{seq_id}{sequence identifier}

\item{pre_label, post_label}{text to add before and after number}

\item{bkm}{bookmark id to associate with autonumber run. If NULL, no bookmark
is added. Value can only be made of alpha numeric characters, ':', -' and '_'.}

\item{bkm_all}{if TRUE, the bookmark will be set on the whole string, if
FALSE, the bookmark will be set on the number only. Default to FALSE.
As an effect when a reference to this bookmark is used, the text can
be like "Table 1" or "1" (pre_label is not included in the referenced
text).}

\item{prop}{formatting text properties returned by \link{fp_text}.}

\item{start_at}{If not NULL, it must be a positive integer, it
specifies the new number to use, at which number the auto
numbering will restart.}

\item{tnd}{\emph{title number depth}, a positive integer (only applies if positive)
that specify the depth (or heading of level \emph{depth}) to use for prefixing
the caption number with this last reference number. For example, setting
\code{tnd=2} will generate numbered captions like '4.3-2' (figure 2 of
chapter 4.3).}

\item{tns}{separator to use between title number and table
number. Default is "-".}
}
\description{
Create an autonumbered chunk, i.e. a string
representation of a sequence, each item will be numbered.
These runs can also be bookmarked and be used later for
cross references.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
run_autonum()
run_autonum(seq_id = "fig", pre_label = "fig. ")
run_autonum(seq_id = "tab", pre_label = "Table ", bkm = "anytable")
run_autonum(
  seq_id = "tab", pre_label = "Table ", bkm = "anytable",
  tnd = 2, tns = " "
)
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}

Other Word computed fields: 
\code{\link{run_reference}()},
\code{\link{run_word_field}()}
}
\concept{Word computed fields}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{table_colwidths}
\alias{table_colwidths}
\title{Column widths of a table}
\usage{
table_colwidths(widths = NULL)
}
\arguments{
\item{widths}{Column widths expressed in inches.}
}
\description{
The function defines the size of each column of a table.
}
\seealso{
Other functions for table definition:
\code{\link[=prop_table]{prop_table()}},
\code{\link[=table_conditional_formatting]{table_conditional_formatting()}},
\code{\link[=table_layout]{table_layout()}},
\code{\link[=table_stylenames]{table_stylenames()}},
\code{\link[=table_width]{table_width()}}
}
\concept{functions for table definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{print.rtf}
\alias{print.rtf}
\title{Write an 'RTF' document to a file}
\usage{
\method{print}{rtf}(x, target = NULL, ...)
}
\arguments{
\item{x}{an 'rtf' object created with \code{\link[=rtf_doc]{rtf_doc()}}}

\item{target}{path to the RTF file to write}

\item{...}{unused}
}
\description{
Write the RTF object and its content to a file.
}
\examples{
# write a rdocx object in a rtf file ----
doc <- rtf_doc()
print(doc, target = tempfile(fileext = ".rtf"))
}
\seealso{
\code{\link[=rtf_doc]{rtf_doc()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_word_field}
\alias{run_word_field}
\alias{run_seqfield}
\title{'Word' computed field}
\usage{
run_word_field(field, prop = NULL, seqfield = NULL)

run_seqfield(field, prop = NULL, seqfield = NULL)
}
\arguments{
\item{field}{Value for a "Word Computed Field" as a string.}

\item{prop}{formatting text properties returned by \link{fp_text}.}

\item{seqfield}{deprecated in favor of \code{field}.}
}
\description{
Create a 'Word' computed field.
}
\note{
In the previous version, this function was called \code{run_seqfield}
but the name was wrong and should have been \code{run_word_field}.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
run_word_field(field = "PAGE  \\\\* MERGEFORMAT")
run_word_field(field = "Date \\\\@ \"MMMM d yyyy\"")
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnote}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_wordtext}()}

Other Word computed fields: 
\code{\link{run_autonum}()},
\code{\link{run_reference}()}
}
\concept{Word computed fields}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{prop_section}
\alias{prop_section}
\title{Section properties}
\usage{
prop_section(
  page_size = NULL,
  page_margins = NULL,
  type = "continuous",
  section_columns = NULL,
  header_default = NULL,
  header_even = NULL,
  header_first = NULL,
  footer_default = NULL,
  footer_even = NULL,
  footer_first = NULL
)
}
\arguments{
\item{page_size}{page dimensions, an object generated with function \link{page_size}.}

\item{page_margins}{page margins, an object generated with function \link{page_mar}.
It is recommended to use the same margins object across all sections to prevent
unintended page breaks.}

\item{type}{Section type. It defines how the contents of the section will be
placed relative to the previous section. Available types are "continuous"
(begins the section on the next paragraph), "evenPage" (begins on the next
even-numbered page), "nextColumn" (begins on the next column on the page),
"nextPage" (begins on the following page), "oddPage" (begins on the next
odd-numbered page).}

\item{section_columns}{section columns, an object generated with function
\link{section_columns}. Use NULL (default value) for no content.}

\item{header_default}{content as a \code{\link[=block_list]{block_list()}} for the default page header.
Use NULL (default value) for no content.}

\item{header_even}{content as a \code{\link[=block_list]{block_list()}} for the even page header.
Use NULL (default value) for no content.}

\item{header_first}{content as a \code{\link[=block_list]{block_list()}} for the first page header.
Use NULL (default value) for no content.}

\item{footer_default}{content as a \code{\link[=block_list]{block_list()}} for the default page footer.
Use NULL (default value) for no content.}

\item{footer_even}{content as a \code{\link[=block_list]{block_list()}} for the even page footer. Use
NULL (default value) for no content.}

\item{footer_first}{content as a \code{\link[=block_list]{block_list()}} for the default page footer.
Use NULL (default value) for no content.}
}
\description{
A section is a grouping of blocks (ie. paragraphs and tables)
that have a set of properties that define pages on which the text will appear.

A Section properties object stores information about page composition,
such as page size, page orientation, borders and margins.

\strong{Important}: When creating multiple sections in a document, it is strongly
recommended to use the \strong{same \code{page_margins} object} for all sections to avoid
unwanted page breaks. Changing page margins between sections can cause Word to
insert automatic page breaks, even when using \code{type = "continuous"}. To ensure
consistent behavior, create a single \code{page_mar()} object and reuse it across
all \code{prop_section()} calls. See the examples in \code{\link[=body_end_block_section]{body_end_block_section()}} which
demonstrate this best practice.
}
\section{Illustrations}{


\if{html}{\figure{prop_section_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
library(officer)

# Example 1: Mixing different section layouts ----
# This example demonstrates how to create a document with multiple sections,
# each with different page orientations and column layouts

# Define a landscape section with single column
# This is useful for wide tables or charts
landscape_one_column <- block_section(
  prop_section(
    page_size = page_size(orient = "landscape"),
    type = "continuous"
  )
)

# Define a landscape section with two columns
# Useful for text-heavy content in landscape mode (e.g., newsletters)
landscape_two_columns <- block_section(
  prop_section(
    page_size = page_size(orient = "landscape"),
    type = "continuous",
    section_columns = section_columns(widths = c(4.75, 4.75))
  )
)

# Create a new document
doc_1 <- read_docx()

# Section 1: Landscape single column for wide table ----
# Add a title for the first section
doc_1 <- body_add_par(doc_1, "Wide Table Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "This table is displayed in landscape orientation to accommodate all columns."
)

# Add a wide table with multiple columns
doc_1 <- body_add_table(doc_1, value = mtcars[1:10, ], style = "table_template")

# End the landscape single-column section
doc_1 <- body_end_block_section(doc_1, value = landscape_one_column)

# Section 2: Landscape two columns for text content ----
# Add a title for the two-column section
doc_1 <- body_add_par(doc_1, "Two-Column Text Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "The following text flows across two columns in landscape orientation."
)

# Add text content that will flow across two columns
doc_1 <- body_add_par(doc_1, value = paste(rep(letters, 50), collapse = " "))

# End the landscape two-column section
doc_1 <- body_end_block_section(doc_1, value = landscape_two_columns)

# Section 3: Return to portrait orientation ----
# After ending the previous sections, we're back to portrait (default)
doc_1 <- body_add_par(doc_1, "Portrait Table Section", style = "heading 1")
doc_1 <- body_add_par(
  doc_1,
  "This section returns to portrait orientation with a taller table."
)

# Add a longer table in portrait orientation
doc_1 <- body_add_table(doc_1, value = mtcars[1:25, ], style = "table_template")

# Save the document
output_file_1 <- tempfile(fileext = ".docx")
print(doc_1, target = output_file_1)


# Example 2: Different headers and footers (first, even, odd pages) ----
# This example demonstrates the complete header/footer system with:
# - Different header/footer for the first page
# - Different header/footer for even pages (left-side pages in duplex printing)
# - Default header/footer for odd pages (right-side pages)

# Create sample text to generate multiple pages
lorem_text <- paste(
  rep("Purus lectus eros metus turpis mattis platea praesent sed. ", 50),
  collapse = ""
)

# Define content for FIRST page header
# Typically used for title pages or cover pages
header_first <- block_list(
  fpar(
    ftext(
      "First Page Header - Title Page",
      fp_text_lite(bold = TRUE, color = "#4472C4", font.size = 14)
    ),
    fp_p = fp_par(
      text.align = "center",
      padding.bottom = 12,
      border.bottom = fp_border(color = "#4472C4", width = 2)
    )
  )
)

# Define content for EVEN pages header (left-side pages when printed)
# In duplex printing, this appears on the left side
header_even <- block_list(
  fpar(
    ftext("Chapter Title", fp_text_lite(italic = TRUE, font.size = 10)),
    fp_p = fp_par(text.align = "left")
  )
)

# Define content for DEFAULT pages header (odd pages/right-side)
# In duplex printing, this appears on the right side
header_default <- block_list(
  fpar(
    ftext("Document Title", fp_text_lite(italic = TRUE, font.size = 10)),
    fp_p = fp_par(text.align = "right")
  )
)

# Define content for FIRST page footer
footer_first <- block_list(
  fpar(
    ftext(
      "Company Name - Confidential",
      fp_text_lite(font.size = 9, color = "#666666")
    ),
    fp_p = fp_par(text.align = "center")
  )
)

# Define content for EVEN pages footer (includes page number on left)
footer_even <- block_list(
  fpar(
    run_word_field(field = "PAGE", prop = fp_text_lite(font.size = 9)),
    ftext(" | Document Name", fp_text_lite(font.size = 9)),
    fp_p = fp_par(
      text.align = "left",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Define content for DEFAULT pages footer (includes page number on right)
footer_default <- block_list(
  fpar(
    ftext("Document Name | ", fp_text_lite(font.size = 9)),
    run_word_field(field = "PAGE", prop = fp_text_lite(font.size = 9)),
    fp_p = fp_par(
      text.align = "right",
      padding.top = 6,
      border.top = fp_border(color = "#CCCCCC", width = 1)
    )
  )
)

# Create section properties with all header/footer variants
# When all three are defined (first, even, default), Word will use:
# - header_first/footer_first for page 1
# - header_even/footer_even for pages 2, 4, 6, etc.
# - header_default/footer_default for pages 3, 5, 7, etc.
section_with_all_hf <- prop_section(
  header_default = header_default,
  footer_default = footer_default,
  header_first = header_first,
  footer_first = footer_first,
  header_even = header_even,
  footer_even = footer_even
)

# Create a new document
doc_2 <- read_docx()

# Add enough content to create multiple pages
# This will demonstrate how the different headers/footers appear
for (i in 1:20) {
  doc_2 <- body_add_par(doc_2, paste0("Paragraph ", i, ": ", lorem_text))
}

# Apply the section properties with all header/footer configurations
doc_2 <- body_set_default_section(doc_2, value = section_with_all_hf)

# Save the document
# Open this document and scroll through pages to see:
# - Page 1: Special first page header/footer
# - Page 2: Even page header/footer with page number on left
# - Page 3: Odd page (default) header/footer with page number on right
# - And so on...
output_file_2 <- tempfile(fileext = ".docx")
print(doc_2, target = output_file_2)
}
\seealso{
\link{block_section}

Other functions for section definition:
\code{\link[=page_mar]{page_mar()}},
\code{\link[=page_size]{page_size()}},
\code{\link[=section_columns]{section_columns()}}
}
\concept{functions for section definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{add_slide}
\alias{add_slide}
\title{Add a slide}
\usage{
add_slide(x, layout = NULL, master = NULL, ..., .dots = NULL)
}
\arguments{
\item{x}{an \code{rpptx} object.}

\item{layout}{slide layout name to use. Can be ommited of a default layout is set via \code{\link[=layout_default]{layout_default()}}.}

\item{master}{master layout name where \code{layout} is located. Only required in case of several masters if layout is
not unique.}

\item{...}{Key-value pairs of the form \code{"short form location" = object} passed to \link{phs_with}. See section
\code{"Short forms"} in \link{phs_with} for details, available short forms and examples.}

\item{.dots}{List of key-value pairs of the form \code{list("short form location" = object)}. Alternative to \code{...}. See
\link{phs_with} for details.}
}
\description{
Add a slide into a pptx presentation.
}
\examples{
x <- read_pptx()

layout_summary(x) # available layouts

x <- add_slide(x, layout = "Two Content")

x <- layout_default(x, "Title Slide") # set default layout for `add_slide()`
x <- add_slide(x) # uses default layout

# use `...` to fill placeholders when adding slide
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)
}
\seealso{
\code{\link[=print.rpptx]{print.rpptx()}}, \code{\link[=read_pptx]{read_pptx()}}, \code{\link[=layout_summary]{layout_summary()}}, \code{\link[=plot_layout_properties]{plot_layout_properties()}}, \code{\link[=ph_with]{ph_with()}}, \code{\link[=phs_with]{phs_with()}}, \code{\link[=layout_default]{layout_default()}}

Other functions to manipulate slides:
\code{\link[=move_slide]{move_slide()}},
\code{\link[=on_slide]{on_slide()}},
\code{\link[=remove_slide]{remove_slide()}},
\code{\link[=set_notes]{set_notes()}}
}
\concept{slide_manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location}
\alias{ph_location}
\title{Location for a placeholder from scratch}
\usage{
ph_location(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  newlabel = "",
  bg = NULL,
  rotation = NULL,
  ln = NULL,
  geom = NULL,
  ...
)
}
\arguments{
\item{left, top, width, height}{place holder coordinates
in inches.}

\item{newlabel}{a label for the placeholder. See section details.}

\item{bg}{background color}

\item{rotation}{rotation angle}

\item{ln}{a \code{\link[=sp_line]{sp_line()}} object specifying the outline style.}

\item{geom}{shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html}

\item{...}{unused arguments}
}
\description{
The function will return a list that complies with
expected format for argument \code{location} of function \code{\link[=ph_with]{ph_with()}}.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
doc <- ph_with(
  doc,
  "Hello world",
  location = ph_location(width = 4, height = 3, newlabel = "hello")
)
print(doc, target = tempfile(fileext = ".pptx"))

# Set geometry and outline
doc <- read_pptx()
doc <- add_slide(doc, "Title and Content")
loc <- ph_location(
  left = 1,
  top = 1,
  width = 4,
  height = 3,
  bg = "steelblue",
  ln = sp_line(color = "red", lwd = 2.5),
  geom = "trapezoid"
)
doc <- ph_with(doc, "", loc = loc)
print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location:
\code{\link[=ph_location_fullsize]{ph_location_fullsize()}},
\code{\link[=ph_location_id]{ph_location_id()}},
\code{\link[=ph_location_label]{ph_location_label()}},
\code{\link[=ph_location_left]{ph_location_left()}},
\code{\link[=ph_location_right]{ph_location_right()}},
\code{\link[=ph_location_template]{ph_location_template()}},
\code{\link[=ph_location_type]{ph_location_type()}}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/post-proc.R
\name{docx_reference_img}
\alias{docx_reference_img}
\title{add images into an rdocx object}
\usage{
docx_reference_img(x, src)
}
\arguments{
\item{x}{an rdocx object}

\item{src}{a vector of character containing image filenames.}
}
\description{
reference images into a Word document.
This function is now useless as the processing of images
is automated when using \code{\link[=print.rdocx]{print.rdocx()}}.
}
\seealso{
Other functions for officer extensions: 
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{sheet_add_drawing}
\alias{sheet_add_drawing}
\title{Add a drawing to an Excel sheet}
\usage{
sheet_add_drawing(x, value, sheet, ...)
}
\arguments{
\item{x}{rxlsx object created by \code{\link[=read_xlsx]{read_xlsx()}}}

\item{value}{object to add (dispatched to the appropriate method)}

\item{sheet}{sheet name (must already exist, see \code{\link[=add_sheet]{add_sheet()}})}

\item{...}{method-specific arguments. Common across \code{external_img}
and \code{gg} methods: \code{left} / \code{top} (fixed-position top-left in
inches, default \verb{(1, 1)}), \code{width} / \code{height} (size in inches),
\code{anchor} (cell reference like \code{"B2"} or \code{"B2:H20"}, see
description), \code{edit_as} (one of \code{"twoCell"}, \code{"oneCell"},
\code{"absolute"}; how Excel treats the drawing when rows/columns are
resized — only meaningful with a cell range \code{anchor}). The \code{gg}
method also accepts \code{res} (ppi, default 300), \code{alt_text}
(auto-detected via \code{ggplot2::get_alt_text()} if empty) and
\code{scale} (passed to \code{\link[ragg:agg_png]{ragg::agg_png()}}).}
}
\value{
the rxlsx object (invisibly)
}
\description{
Add a graphical element into a sheet of an xlsx workbook.
This is a generic function dispatching on \code{value}. Supported inputs
include:
\itemize{
\item \code{\link[=external_img]{external_img()}}: an image file (PNG, JPEG, GIF, ...) copied into
the workbook and placed on the sheet.
\item \code{gg}: a ggplot object rendered to PNG via \code{ragg::agg_png()} and
embedded via the \code{external_img} path. Extra arguments: \code{res}
(default 300 ppi), \code{alt_text} (auto-detected via
\code{ggplot2::get_alt_text()} when empty), \code{scale} (same semantics as
\code{ggplot2::ggsave()}).
}

The drawing can be positioned in three ways, matching Excel's own
"Format Picture" options:
\itemize{
\item Fixed position (default): pass \code{left} / \code{top} and
\code{width} / \code{height} in inches. The drawing stays put when rows or
columns are inserted/resized.
\item Move and size with cells (Excel's default): pass
\code{anchor = "B2:H20"} (a cell range). The drawing spans those two
cells and follows them.
\item Move but don't size with cells: pass \code{anchor = "B2"}
(a single cell). The drawing anchors to that cell, keeping the
size set by \code{width} / \code{height}.
}

Additional methods for \code{ms_chart} and \code{dml} are provided by the
extension packages 'mschart' and 'rvg'.

Use \code{\link[=sheet_write_data]{sheet_write_data()}} to write data into the sheet before or
after adding a drawing.
}
\examples{
img <- system.file("extdata", "example.png", package = "officer")
if (nzchar(img) && file.exists(img)) {
  x <- read_xlsx()
  x <- add_sheet(x, label = "pics")
  x <- sheet_add_drawing(
    x, sheet = "pics",
    value = external_img(img, width = 2, height = 2)
  )
  print(x, target = tempfile(fileext = ".xlsx"))
}

if (requireNamespace("ggplot2", quietly = TRUE)) {
  gg <- ggplot2::ggplot(iris, ggplot2::aes(Sepal.Length, Sepal.Width)) +
    ggplot2::geom_point()
  x <- read_xlsx()
  x <- add_sheet(x, label = "plots")
  x <- sheet_add_drawing(x, value = gg, sheet = "plots",
                         width = 4, height = 3)
  print(x, target = tempfile(fileext = ".xlsx"))
}
}
\seealso{
\code{\link[=read_xlsx]{read_xlsx()}}, \code{\link[=add_sheet]{add_sheet()}}, \code{\link[=sheet_write_data]{sheet_write_data()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/post-proc.R
\name{sanitize_images}
\alias{sanitize_images}
\title{Remove unused media from a document}
\usage{
sanitize_images(x, warn_user = TRUE)
}
\arguments{
\item{x}{\code{rdocx} or \code{rpptx} object}

\item{warn_user}{TRUE to make sure users are warned when
using this function that will be un-exported in a next
version}
}
\description{
The function will scan the media
directory and delete images that are not used
anymore. This function is to be used when images
have been replaced many times.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{to_wml}
\alias{to_wml}
\title{Convert officer objects to WordprocessingML}
\usage{
to_wml(x, add_ns = FALSE, ...)
}
\arguments{
\item{x}{object}

\item{add_ns}{should namespace be added to the top tag}

\item{...}{Arguments to be passed to methods}
}
\description{
Convert an object made with package officer
to WordprocessingML. The result is a string.
}
\seealso{
Other functions for officer extensions:
\code{\link[=fortify_location]{fortify_location()}},
\code{\link[=get_reference_value]{get_reference_value()}},
\code{\link[=opts_current_table]{opts_current_table()}},
\code{\link[=shape_properties_tags]{shape_properties_tags()}},
\code{\link[=str_encode_to_rtf]{str_encode_to_rtf()}},
\code{\link[=to_html]{to_html()}},
\code{\link[=to_pml]{to_pml()}},
\code{\link[=to_rtf]{to_rtf()}},
\code{\link[=wml_link_images]{wml_link_images()}}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_caption}
\alias{body_add_caption}
\title{Add Word caption in a 'Word' document}
\usage{
body_add_caption(x, value, pos = "after")
}
\arguments{
\item{x}{an rdocx object}

\item{value}{an object returned by \code{\link[=block_caption]{block_caption()}}}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}
}
\description{
Add a Word caption into an rdocx object.
}
\examples{
doc <- read_docx()

if (capabilities(what = "png")) {
  doc <- body_add_plot(doc,
    value = plot_instr(
      code = {
        barplot(1:5, col = 2:6)
      }
    ),
    style = "centered"
  )
}
run_num <- run_autonum(
  seq_id = "fig", pre_label = "Figure ",
  bkm = "barplot"
)
caption <- block_caption("a barplot",
  style = "Normal",
  autonum = run_num
)
doc <- body_add_caption(doc, caption)
print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_footnote}
\alias{run_footnote}
\title{Footnote for 'Word'}
\usage{
run_footnote(x, prop = NULL)
}
\arguments{
\item{x}{a set of blocks to be used as footnote content returned by
function \code{\link[=block_list]{block_list()}}.}

\item{prop}{formatting text properties returned by
\code{\link[=fp_text_lite]{fp_text_lite()}} or \code{\link[=fp_text]{fp_text()}}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Wraps a footnote in an object that can then be inserted
as a run/chunk with \code{\link[=fpar]{fpar()}} or within an R Markdown document.
}
\examples{
library(officer)

fp_bold <- fp_text_lite(bold = TRUE)
fp_refnote <- fp_text_lite(vertical.align = "superscript")

img.file <- file.path(R.home("doc"), "html", "logo.jpg")
bl <- block_list(
  fpar(ftext("hello", fp_bold)),
  fpar(
    ftext("hello world", fp_bold),
    external_img(src = img.file, height = 1.06, width = 1.39)
  )
)

a_par <- fpar(
  "this paragraph contains a note ",
  run_footnote(x = bl, prop = fp_refnote),
  "."
)

doc <- read_docx()
doc <- body_add_fpar(doc, value = a_par, style = "Normal")

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other run functions for reporting: 
\code{\link{external_img}()},
\code{\link{ftext}()},
\code{\link{hyperlink_ftext}()},
\code{\link{run_autonum}()},
\code{\link{run_bookmark}()},
\code{\link{run_columnbreak}()},
\code{\link{run_comment}()},
\code{\link{run_footnoteref}()},
\code{\link{run_linebreak}()},
\code{\link{run_pagebreak}()},
\code{\link{run_reference}()},
\code{\link{run_tab}()},
\code{\link{run_word_field}()},
\code{\link{run_wordtext}()}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{block_table}
\alias{block_table}
\title{Table block}
\usage{
block_table(x, header = TRUE, properties = prop_table(), alignment = NULL)
}
\arguments{
\item{x}{a data.frame to add as a table}

\item{header}{display header if TRUE}

\item{properties}{table properties, see \code{\link[=prop_table]{prop_table()}}.
Table properties are not handled identically between Word and PowerPoint
output format. They are fully supported with Word but for PowerPoint (which
does not handle as many things as Word for tables), only conditional
formatting properties are supported.}

\item{alignment}{alignment for each columns, 'l' for left, 'r' for right
and 'c' for center. Default to NULL.}
}
\description{
Create a representation of a table
}
\examples{
block_table(x = head(iris))

block_table(x = mtcars, header = TRUE,
  properties = prop_table(
    tcf = table_conditional_formatting(
      first_row = TRUE, first_column = TRUE)
  ))
}
\seealso{
\code{\link[=prop_table]{prop_table()}}

Other block functions for reporting: 
\code{\link{block_caption}()},
\code{\link{block_list}()},
\code{\link{block_pour_docx}()},
\code{\link{block_section}()},
\code{\link{block_toc}()},
\code{\link{fpar}()},
\code{\link{plot_instr}()},
\code{\link{unordered_list}()}
}
\concept{block functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{docx_bookmarks}
\alias{docx_bookmarks}
\title{List Word bookmarks}
\usage{
docx_bookmarks(x)
}
\arguments{
\item{x}{an \code{rdocx} object}
}
\description{
List bookmarks id that can be found in a
'Word' document.
}
\examples{
library(officer)

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
doc_1 <- body_bookmark(doc_1, "text_to_replace_1")
doc_1 <- body_add_par(doc_1, "centered text", style = "centered")
doc_1 <- body_bookmark(doc_1, "text_to_replace_2")

docx_bookmarks(doc_1)

docx_bookmarks(read_docx())
}
\seealso{
Other functions for Word document informations:
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=docx_dim]{docx_dim()}},
\code{\link[=length.rdocx]{length.rdocx()}},
\code{\link[=set_doc_properties]{set_doc_properties()}},
\code{\link[=styles_info]{styles_info()}}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_par}
\alias{fp_par}
\alias{print.fp_par}
\alias{update.fp_par}
\title{Paragraph formatting properties}
\usage{
fp_par(
  text.align = "left",
  padding = 0,
  line_spacing = 1,
  border = fp_border(width = 0),
  padding.bottom,
  padding.top,
  padding.left,
  padding.right,
  border.bottom,
  border.left,
  border.top,
  border.right,
  shading.color = "transparent",
  keep_with_next = FALSE,
  tabs = NULL,
  word_style = "Normal"
)

\method{print}{fp_par}(x, ...)

\method{update}{fp_par}(
  object,
  text.align,
  padding,
  border,
  padding.bottom,
  padding.top,
  padding.left,
  padding.right,
  border.bottom,
  border.left,
  border.top,
  border.right,
  shading.color,
  keep_with_next,
  word_style,
  ...
)
}
\arguments{
\item{text.align}{text alignment - a single character value, expected value
is one of 'left', 'right', 'center', 'justify'.}

\item{padding}{paragraph paddings - 0 or positive integer value. Argument \code{padding} overwrites
arguments \code{padding.bottom}, \code{padding.top}, \code{padding.left}, \code{padding.right}.}

\item{line_spacing}{line spacing, 1 is single line spacing, 2 is double line spacing.}

\item{border}{shortcut for all borders.}

\item{padding.bottom, padding.top, padding.left, padding.right}{paragraph paddings - 0 or positive integer value.}

\item{border.bottom, border.left, border.top, border.right}{\code{\link{fp_border}} for
borders. overwrite other border properties.}

\item{shading.color}{shading color - a single character value specifying
a valid color (e.g. "#000000" or "black").}

\item{keep_with_next}{a scalar logical. Specifies that the paragraph (or at least part of it) should be rendered
on the same page as the next paragraph when possible.}

\item{tabs}{NULL (default) for no tabulation marks setting
or an object returned by \code{\link[=fp_tabs]{fp_tabs()}}. Note this can only have effect with Word
or RTF outputs.}

\item{word_style}{Word paragraph style name}

\item{x, object}{\code{fp_par} object}

\item{...}{further arguments - not used}
}
\value{
a \code{fp_par} object
}
\description{
Create a \code{fp_par} object that describes
paragraph formatting properties.
}
\examples{
fp_par(text.align = "center", padding = 5)
obj <- fp_par(text.align = "center", padding = 1)
update(obj, padding.bottom = 5)
}
\seealso{
\link{fpar}

Other functions for defining formatting properties: 
\code{\link{fp_border}()},
\code{\link{fp_cell}()},
\code{\link{fp_tab}()},
\code{\link{fp_tabs}()},
\code{\link{fp_text}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_toc}
\alias{body_add_toc}
\title{Add table of content in a 'Word' document}
\usage{
body_add_toc(x, level = 3, pos = "after", style = NULL, separator = ";")
}
\arguments{
\item{x}{an rdocx object}

\item{level}{max title level of the table}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}

\item{style}{optional. style in the document that will be used to build entries of the TOC.}

\item{separator}{optional. Some configurations need "," (i.e. from Canada) separator instead of ";"}
}
\description{
Add a table of content into an rdocx object.
The TOC will be generated by Word, if the document is not
edited with Word (i.e. Libre Office) the TOC will not be generated.
}
\examples{
doc <- read_docx()
doc <- body_add_toc(doc)

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_table}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_utils_funs.R
\name{length.rdocx}
\alias{length.rdocx}
\title{Number of blocks inside an rdocx object}
\usage{
\method{length}{rdocx}(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
Return the number of blocks inside an rdocx object.
This number also include the default section definition of a
Word document - default Word section is an uninvisible element.
}
\examples{
# how many elements are there in an new document produced
# with the default template.
length(read_docx())
}
\seealso{
Other functions for Word document informations:
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=docx_bookmarks]{docx_bookmarks()}},
\code{\link[=docx_dim]{docx_dim()}},
\code{\link[=set_doc_properties]{set_doc_properties()}},
\code{\link[=styles_info]{styles_info()}}
}
\concept{functions for Word document informations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{open_file}
\alias{open_file}
\title{Opens a file locally}
\usage{
open_file(path)
}
\arguments{
\item{path}{File path.}
}
\description{
Opening a file locally requires a compatible application to be installed
(e.g., MS Office or LibreOffice for .pptx or .docx files).
}
\details{
\emph{NB:} Function is a small wrapper around \code{\link[utils:browseURL]{utils::browseURL()}} to have a more
suitable function name.
}
\examples{
x <- read_pptx()
x <- add_slide(x, "Title Slide", ctrTitle = "My Title")
file <- print(x, tempfile(fileext = ".pptx"))
\dontrun{
open_file(file)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{docx_show_chunk}
\alias{docx_show_chunk}
\title{Show underlying text tag structure}
\usage{
docx_show_chunk(x)
}
\arguments{
\item{x}{a docx device}
}
\description{
Show the structure of text tags at the current cursor. This is
most useful when trying to troubleshoot search-and-replace functionality
using \code{\link{body_replace_all_text}}.
}
\examples{
doc <- read_docx()
doc <- body_add_par(doc, "Placeholder one")
doc <- body_add_par(doc, "Placeholder two")

# Show text chunk at cursor
docx_show_chunk(doc)  # Output is 'Placeholder two'
}
\seealso{
\code{\link{body_replace_all_text}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_notes.R
\name{set_notes}
\alias{set_notes}
\alias{set_notes.character}
\alias{set_notes.block_list}
\title{Set notes for current slide}
\usage{
set_notes(x, value, location, ...)

\method{set_notes}{character}(x, value, location, ...)

\method{set_notes}{block_list}(x, value, location, ...)
}
\arguments{
\item{x}{an rpptx object}

\item{value}{text to be added to notes}

\item{location}{a placeholder location object.
It will be used to specify the location of the new shape. This location
can be defined with a call to one of the notes_ph functions. See
section \code{"see also"}.}

\item{...}{further arguments passed to or from other methods.}
}
\description{
Set speaker notes for the current slide in a pptx presentation.
}
\section{Methods (by class)}{
\itemize{
\item \code{set_notes(character)}: add a character vector to a place holder in the notes on the
current slide, values will be added as paragraphs.

\item \code{set_notes(block_list)}: add a \code{\link[=block_list]{block_list()}} to a place holder in the notes on the
current slide.

}}
\examples{
# this name will be used to print the file
# change it to "youfile.pptx" to write the pptx
# file in your working directory.
fileout <- tempfile(fileext = ".pptx")
fpt_blue_bold <- fp_text_lite(color = "#006699", bold = TRUE)
doc <- read_pptx()
# add a slide with some text ----
doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
doc <- ph_with(x = doc, value = "Slide Title 1",
   location = ph_location_type(type = "title") )
# set speaker notes for the slide ----
doc <- set_notes(doc, value = "This text will only be visible for the speaker.",
   location = notes_location_type("body"))

# add a slide with some text ----
doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
doc <- ph_with(x = doc, value = "Slide Title 2",
   location = ph_location_type(type = "title") )
bl <- block_list(
  fpar(ftext("hello world", fpt_blue_bold)),
  fpar(ftext("Turlututu chapeau pointu", fpt_blue_bold))
)
doc <- set_notes(doc, value = bl,
   location = notes_location_type("body"))

print(doc, target = fileout)

}
\seealso{
\code{\link[=print.rpptx]{print.rpptx()}}, \code{\link[=read_pptx]{read_pptx()}}, \code{\link[=add_slide]{add_slide()}}, \code{\link[=notes_location_label]{notes_location_label()}}, \code{\link[=notes_location_type]{notes_location_type()}}

Other functions slide manipulation: 
\code{\link{add_slide}()},
\code{\link{move_slide}()},
\code{\link{on_slide}()},
\code{\link{remove_slide}()}
}
\concept{functions slide manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr_utils.R
\name{opts_current_table}
\alias{opts_current_table}
\title{Table options in a 'knitr' context}
\usage{
opts_current_table()
}
\value{
a list
}
\description{
Get options for table rendering in a 'knitr'
context. It should not be used by the end user, but
its documentation should be read as a place where
table options are documented when 'knitr' is used.

The function is a utility to facilitate the retrieval of table
options supported by the 'flextable', 'officedown' and 'officer' packages.

These options should be set with \code{knitr::opts_chunk$set()}.
The names and expected values are listed in the following sections.
}
\section{knitr chunk options for table captions}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   caption id/bookmark \tab tab.id \tab NULL \cr
   caption \tab tab.cap \tab NULL \cr
   display table caption on top of the table or not \tab tab.topcaption \tab TRUE \cr
   caption table sequence identifier. \tab tab.lp \tab "tab:" \cr
}
}

\section{knitr chunk options for Word table captions}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   Word stylename to use for table captions. \tab tab.cap.style \tab NULL \cr
   prefix for numbering chunk (default to   "Table "). \tab tab.cap.pre \tab Table \cr
   suffix for numbering chunk (default to   ": "). \tab tab.cap.sep \tab " :" \cr
   title number depth \tab tab.cap.tnd \tab 0 \cr
   separator to use between title number and table number. \tab tab.cap.tns \tab "-" \cr
   caption prefix formatting properties \tab tab.cap.fp_text \tab fp_text_lite(bold = TRUE) \cr
}
}

\section{knitr chunk options for Word tables}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   the Word stylename to use for tables \tab tab.style \tab NULL \cr
   autofit' or 'fixed' algorithm. \tab tab.layout \tab "autofit" \cr
   value of the preferred width of the table in percent (base 1). \tab tab.width \tab 1 \cr
   Alternative title text \tab tab.alt.title \tab NULL \cr
   Alternative description text \tab tab.alt.description \tab NULL \cr
}
}

\section{knitr chunk options for data.frame with officedown}{
\tabular{lcc}{
   \strong{label} \tab \strong{name} \tab \strong{value} \cr
   apply or remove formatting from the first row in the table \tab first_row \tab TRUE \cr
   apply or remove formatting from the first column in the table \tab first_column \tab FALSE \cr
   apply or remove formatting from the last row in the table \tab last_row \tab FALSE \cr
   apply or remove formatting from the last column in the table \tab last_column \tab FALSE \cr
   don't display odd and even rows \tab no_hband \tab TRUE \cr
   don't display odd and even columns \tab no_vband \tab TRUE \cr
}
}

\section{returned elements}{

\itemize{
\item cap.style (default: NULL)
\item cap.pre (default: "Table ")
\item cap.sep (default: ":")
\item cap.tnd (default: 0)
\item cap.tns (default: "-")
\item cap.fp_text (default: \code{fp_text_lite(bold = TRUE)})
\item id (default: NULL)
\item cap (default: NULL)
\item alt.title (default: NULL)
\item alt.description (default: NULL)
\item topcaption (default: TRUE)
\item style (default: NULL)
\item tab.lp (default: "tab:")
\item table_layout (default: "autofit")
\item table_width (default: 1)
\item first_row (default: TRUE)
\item first_column (default: FALSE)
\item last_row (default: FALSE)
\item last_column (default: FALSE)
\item no_hband (default: TRUE)
\item no_vband (default: TRUE)
}
}

\seealso{
Other functions for officer extensions: 
\code{\link{docx_reference_img}()},
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_pml}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_settings.R
\name{docx_set_settings}
\alias{docx_set_settings}
\title{Set 'Microsoft Word' Document Settings}
\usage{
docx_set_settings(
  x,
  zoom = 1,
  default_tab_stop = 0.5,
  hyphenation_zone = 0.25,
  decimal_symbol = ".",
  list_separator = ";",
  compatibility_mode = "15",
  even_and_odd_headers = FALSE,
  auto_hyphenation = FALSE,
  unit = "in"
)
}
\arguments{
\item{x}{an rdocx object}

\item{zoom}{zoom factor, default is 1 (100\%)}

\item{default_tab_stop}{default tab stop in inches, default is 0.5}

\item{hyphenation_zone}{hyphenation zone in inches, default is 0.25}

\item{decimal_symbol}{decimal symbol, default is "."}

\item{list_separator}{list separator, default is ";". Sets the separator
used by Word for lists (see details).}

\item{compatibility_mode}{compatibility mode, default is "15"}

\item{even_and_odd_headers}{whether to use different headers for even and
odd pages, default is FALSE. Enables the "Different Odd and Even Pages"
feature in 'Microsoft Word'.}

\item{auto_hyphenation}{whether to enable auto hyphenation, default is FALSE.}

\item{unit}{unit for \code{default_tab_stop} and \code{hyphenation_zone}, one of "in", "cm", "mm".}
}
\description{
Set various settings of a 'Microsoft Word' document generated with 'officer'.
Options include:
\itemize{
\item zoom factor (default view in Word),
\item default tab stop,
\item hyphenation zone,
\item decimal symbol,
\item list separator (see details below),
\item compatibility mode,
\item even and odd headers management (see details below),
\item and auto hyphenation activation.
}
}
\details{
\itemize{
\item \code{even_and_odd_headers}: If TRUE, 'Microsoft Word' will use different
headers for odd and even pages ("Different Odd & Even Pages" feature in Word).
This is useful for professional documents or reports that require alternating
page layouts.
\item \code{list_separator}: Sets the character used by 'Microsoft Word' to separate
items in lists (for example, when inserting tables or lists in Word). This
parameter affects how 'Microsoft Word' handles data import/export (CSV, etc.) and can be
adapted to language or local conventions (e.g., ";" for French, "," for English).
}
}
\examples{
library(officer)

txt_lorem <- rep(
  "Purus lectus eros metus turpis mattis platea praesent sed. ",
  50
)
txt_lorem <- paste0(txt_lorem, collapse = "")

header_first <- block_list(fpar(ftext("text for first page header")))
header_even <- block_list(fpar(ftext("text for even page header")))
header_default <- block_list(fpar(ftext("text for default page header")))
footer_first <- block_list(fpar(ftext("text for first page footer")))
footer_even <- block_list(fpar(ftext("text for even page footer")))
footer_default <- block_list(fpar(ftext("text for default page footer")))

ps <- prop_section(
  header_default = header_default,
  footer_default = footer_default,
  header_first = header_first,
  footer_first = footer_first,
  header_even = header_even,
  footer_even = footer_even
)

x <- read_docx()

x <- docx_set_settings(
  x = x,
  zoom = 2,
  list_separator = ",",
  even_and_odd_headers = TRUE
)

for (i in 1:20) {
  x <- body_add_par(x, value = txt_lorem)
}
x <- body_set_default_section(
  x,
  value = ps
)
print(x, target = tempfile(fileext = ".docx"))
}
\seealso{
\code{\link[=read_docx]{read_docx()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{plot_in_png}
\alias{plot_in_png}
\title{Write a ggplot Object to PNG File}
\usage{
plot_in_png(
  ggobj = NULL,
  code = NULL,
  width,
  height,
  res = 200,
  units = "in",
  pointsize = 11,
  scaling = 1,
  path = NULL
)
}
\arguments{
\item{ggobj}{A ggplot object to render}

\item{width}{Numeric, width of the output image}

\item{height}{Numeric, height of the output image}

\item{res}{Numeric, resolution in DPI (default 200)}

\item{units}{Character, units for width and height ("in", "cm", "mm", "px") (default "in")}

\item{pointsize}{Integer, The default pointsize of the device in pt}

\item{scaling}{scaling factor to apply}

\item{path}{Character, output file path. If NULL, a temporary file is created (default NULL)}
}
\value{
Character, the path to the created PNG file
}
\description{
Renders a ggplot object to a PNG file using ragg for high-quality output.
}
\examples{
plot_in_png(
  code = {
    barplot(1:10)
  },
  width = 5,
  height = 4,
  res = 72,
  units = "in"
)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{annotate_base}
\alias{annotate_base}
\title{Placeholder parameters annotation}
\usage{
annotate_base(path = NULL, output_file = "annotated_layout.pptx")
}
\arguments{
\item{path}{path to the pptx file to use as base document or NULL to use the officer default}

\item{output_file}{filename to store the annotated powerpoint file or NULL to suppress generation}
}
\value{
rpptx object of the annotated PowerPoint file
}
\description{
generates a slide from each layout in the base document to
identify the placeholder indexes, types, names, master names and layout names.

This is to be used when need to know what parameters should be used with
\code{ph_location*} calls. The parameters are printed in their corresponding shapes.

Note that if there are duplicated \code{ph_label}, you should not use \code{ph_location_label}.
Hint: You can dedupe labels using \code{\link{layout_dedupe_ph_labels}}.
}
\examples{
# To generate an anotation of the default base document with officer:
annotate_base(output_file = tempfile(fileext = ".pptx"))

# To generate an annotation of the base document 'mydoc.pptx' and place the
# annotated output in 'mydoc_annotate.pptx'
# annotate_base(path = 'mydoc.pptx', output_file='mydoc_annotate.pptx')

}
\seealso{
Other functions for reading presentation information: 
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()},
\code{\link{slide_summary}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_left}
\alias{ph_location_left}
\title{Location of a left body element}
\usage{
ph_location_left(newlabel = NULL, ...)
}
\arguments{
\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will return the location corresponding
to a left bounding box. The function assume the layout 'Two Content'
is existing. This is an helper function, if you don't have a layout
named 'Two Content', use \code{\link[=ph_location_type]{ph_location_type()}} and set arguments
to your specific needs.
}
\examples{
doc <- read_pptx()
doc <- add_slide(doc)
doc <- ph_with(doc, "Hello left", location = ph_location_left() )
doc <- ph_with(doc, "Hello right", location = ph_location_right() )
print(doc, target = tempfile(fileext = ".pptx") )
}
\seealso{
Other functions for placeholder location: 
\code{\link{ph_location}()},
\code{\link{ph_location_fullsize}()},
\code{\link{ph_location_id}()},
\code{\link{ph_location_label}()},
\code{\link{ph_location_right}()},
\code{\link{ph_location_template}()},
\code{\link{ph_location_type}()}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rtf.R
\name{to_rtf}
\alias{to_rtf}
\title{Convert officer objects to RTF}
\usage{
to_rtf(x, ...)
}
\arguments{
\item{x}{object to convert to RTF. Supported
objects are:
\itemize{
\item \code{\link[=ftext]{ftext()}}
\item \code{\link[=external_img]{external_img()}}
\item \code{\link[=run_autonum]{run_autonum()}}
\item \code{\link[=run_columnbreak]{run_columnbreak()}}
\item \code{\link[=run_linebreak]{run_linebreak()}}
\item \code{\link[=run_word_field]{run_word_field()}}
\item \code{\link[=run_reference]{run_reference()}}
\item \code{\link[=run_pagebreak]{run_pagebreak()}}
\item \code{\link[=hyperlink_ftext]{hyperlink_ftext()}}
\item \code{\link[=block_list]{block_list()}}
\item \code{\link[=fpar]{fpar()}}
}}

\item{...}{Arguments to be passed to methods}
}
\value{
a string containing the RTF code
}
\description{
Convert an object made with package officer
to RTF.
}
\seealso{
Other functions for officer extensions:
\code{\link[=fortify_location]{fortify_location()}},
\code{\link[=get_reference_value]{get_reference_value()}},
\code{\link[=opts_current_table]{opts_current_table()}},
\code{\link[=shape_properties_tags]{shape_properties_tags()}},
\code{\link[=str_encode_to_rtf]{str_encode_to_rtf()}},
\code{\link[=to_html]{to_html()}},
\code{\link[=to_pml]{to_pml()}},
\code{\link[=to_wml]{to_wml()}},
\code{\link[=wml_link_images]{wml_link_images()}}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_footnoteref}
\alias{run_footnoteref}
\title{Word footnote reference}
\usage{
run_footnoteref(prop = NULL)
}
\arguments{
\item{prop}{formatting text properties returned by
\code{\link[=fp_text_lite]{fp_text_lite()}} or \code{\link[=fp_text]{fp_text()}}. It also can be NULL in
which case, no formatting is defined (the default is applied).}
}
\description{
Wraps a footnote reference in an object that can then be inserted
as a run/chunk with \code{\link[=fpar]{fpar()}} or within an R Markdown document.
}
\examples{
run_footnoteref()
to_wml(run_footnoteref())
}
\seealso{
Other run functions for reporting:
\code{\link[=external_img]{external_img()}},
\code{\link[=floating_external_img]{floating_external_img()}},
\code{\link[=ftext]{ftext()}},
\code{\link[=hyperlink_ftext]{hyperlink_ftext()}},
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_bookmark]{run_bookmark()}},
\code{\link[=run_columnbreak]{run_columnbreak()}},
\code{\link[=run_comment]{run_comment()}},
\code{\link[=run_footnote]{run_footnote()}},
\code{\link[=run_linebreak]{run_linebreak()}},
\code{\link[=run_pagebreak]{run_pagebreak()}},
\code{\link[=run_reference]{run_reference()}},
\code{\link[=run_tab]{run_tab()}},
\code{\link[=run_word_field]{run_word_field()}},
\code{\link[=run_wordtext]{run_wordtext()}}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{prop_table}
\alias{prop_table}
\title{Table properties}
\usage{
prop_table(
  style = NA_character_,
  layout = table_layout(),
  width = table_width(),
  stylenames = table_stylenames(),
  colwidths = table_colwidths(),
  tcf = table_conditional_formatting(),
  align = "center",
  word_title = NULL,
  word_description = NULL
)
}
\arguments{
\item{style}{table style to be used to format table}

\item{layout}{layout defined by \code{\link[=table_layout]{table_layout()}},}

\item{width}{table width in the document defined by \code{\link[=table_width]{table_width()}}}

\item{stylenames}{columns styles defined by \code{\link[=table_stylenames]{table_stylenames()}}}

\item{colwidths}{column widths defined by \code{\link[=table_colwidths]{table_colwidths()}}}

\item{tcf}{conditional formatting settings defined by \code{\link[=table_conditional_formatting]{table_conditional_formatting()}}}

\item{align}{table alignment (one of left, center or right)}

\item{word_title}{alternative text for Word table (used as title of the table)}

\item{word_description}{alternative text for Word table (used as description of the table)}
}
\description{
Define table properties such as fixed or autofit layout,
table width in the document, eventually column widths.
}
\examples{
prop_table()
to_wml(prop_table())
}
\seealso{
Other functions for table definition: 
\code{\link{table_colwidths}()},
\code{\link{table_conditional_formatting}()},
\code{\link{table_layout}()},
\code{\link{table_stylenames}()},
\code{\link{table_width}()}
}
\concept{functions for table definition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{floating_external_img}
\alias{floating_external_img}
\title{Floating external image}
\usage{
floating_external_img(
  src,
  width = 0.5,
  height = 0.2,
  pos_x = 0,
  pos_y = 0,
  pos_h_from = "margin",
  pos_v_from = "margin",
  wrap_type = "square",
  wrap_side = "bothSides",
  wrap_dist_top = 0,
  wrap_dist_bottom = 0,
  wrap_dist_left = 0.125,
  wrap_dist_right = 0.125,
  unit = "in",
  guess_size = FALSE,
  alt = ""
)
}
\arguments{
\item{src}{image file path}

\item{width, height}{size of the image file. It can be ignored
if parameter \code{guess_size=TRUE}, see parameter \code{guess_size}.}

\item{pos_x, pos_y}{horizontal and vertical position of the image relative to the anchor point}

\item{pos_h_from}{horizontal positioning reference point, one of "margin", "page", "column", "character"}

\item{pos_v_from}{vertical positioning reference point, one of "margin", "page", "paragraph", "line"}

\item{wrap_type}{text wrapping type, one of "square", "topAndBottom", "through", "tight", "none"}

\item{wrap_side}{which side text wraps around, one of "bothSides", "left", "right", "largest"}

\item{wrap_dist_top, wrap_dist_bottom, wrap_dist_left, wrap_dist_right}{distance between image and text (in inches)}

\item{unit}{unit for width, height, pos_x and pos_y, one of "in", "cm", "mm".}

\item{guess_size}{If package 'magick' is installed, this option
can be used (set it to \code{TRUE}). The images will be read and
width and height will be guessed.}

\item{alt}{alternative text for images}
}
\description{
Wraps an image in an object that can be embedded as a floating
image in a 'Word' document. Unlike \code{\link[=external_img]{external_img()}}, which creates inline images,
this function creates floating images that can be positioned anywhere on the page
and allow text wrapping around them.
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
library(officer)
srcfile <- file.path(R.home("doc"), "html", "logo.jpg")
floatimg <- floating_external_img(
  src = srcfile, height = 1.06 / 2, width = 1.39 / 2,
  pos_x = 0, pos_y = 0,
  pos_h_from = "margin", pos_v_from = "margin"
)

text <- paste0(
  " is a floating image in a ",
  paste0(rep("very ", 30), collapse = ""),
  " long text!"
)

# docx example ----
x <- read_docx()
fp_t <- fp_text(font.size = 20, color = "red")
an_fpar <- fpar(floatimg, ftext(text, fp_t))
x <- body_add_fpar(x, an_fpar)
print(x, target = tempfile(fileext = ".docx"))

# rtf example ----
rtf_doc <- rtf_doc()
rtf_doc <- rtf_add(rtf_doc, an_fpar)
print(rtf_doc, target = tempfile(fileext = ".rtf"))
}
\seealso{
\link{external_img}, \link{body_add}, \link{fpar}, \link{rtf_doc}, \link{rtf_add}

Other run functions for reporting:
\code{\link[=external_img]{external_img()}},
\code{\link[=ftext]{ftext()}},
\code{\link[=hyperlink_ftext]{hyperlink_ftext()}},
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_bookmark]{run_bookmark()}},
\code{\link[=run_columnbreak]{run_columnbreak()}},
\code{\link[=run_comment]{run_comment()}},
\code{\link[=run_footnote]{run_footnote()}},
\code{\link[=run_footnoteref]{run_footnoteref()}},
\code{\link[=run_linebreak]{run_linebreak()}},
\code{\link[=run_pagebreak]{run_pagebreak()}},
\code{\link[=run_reference]{run_reference()}},
\code{\link[=run_tab]{run_tab()}},
\code{\link[=run_word_field]{run_word_field()}},
\code{\link[=run_wordtext]{run_wordtext()}}
}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{docx_current_block_xml}
\alias{docx_current_block_xml}
\title{xml element on which cursor is}
\usage{
docx_current_block_xml(x)
}
\arguments{
\item{x}{an rdocx object}
}
\description{
Get the current block element as xml. This function
is not to be used by end users, it has been implemented
to allow other packages to work with officer. If the
document is empty, this block will be set to NULL.
}
\examples{
doc <- read_docx()
docx_current_block_xml(doc)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{fortify_location}
\alias{fortify_location}
\title{Eval a location on the current slide}
\usage{
fortify_location(x, doc, ...)
}
\arguments{
\item{x}{a location for a placeholder.}

\item{doc}{an rpptx object}

\item{...}{unused arguments}
}
\description{
Eval a shape location against the current slide.
This function is to be used to add custom openxml code. A
list is returned, it contains informations width, height, left
and top positions and other informations necessary to add a
content on a slide.
}
\examples{
library(officer)

doc <- read_pptx()
doc <- add_slide(doc, layout = "Title and Content", master = "Office Theme")
fortify_location(ph_location_fullsize(), doc)
}
\seealso{
\code{\link[=ph_location]{ph_location()}}, \code{\link[=ph_with]{ph_with()}}

Other functions for officer extensions:
\code{\link[=get_reference_value]{get_reference_value()}},
\code{\link[=opts_current_table]{opts_current_table()}},
\code{\link[=shape_properties_tags]{shape_properties_tags()}},
\code{\link[=str_encode_to_rtf]{str_encode_to_rtf()}},
\code{\link[=to_html]{to_html()}},
\code{\link[=to_pml]{to_pml()}},
\code{\link[=to_rtf]{to_rtf()}},
\code{\link[=to_wml]{to_wml()}},
\code{\link[=wml_link_images]{wml_link_images()}}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{color_scheme}
\alias{color_scheme}
\title{Color scheme of a PowerPoint file}
\usage{
color_scheme(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Get the color scheme of a
'PowerPoint' master layout into a data.frame.
}
\examples{
x <- read_pptx()
color_scheme ( x = x )
}
\seealso{
Other functions for reading presentation information:
\code{\link[=annotate_base]{annotate_base()}},
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=layout_properties]{layout_properties()}},
\code{\link[=layout_summary]{layout_summary()}},
\code{\link[=length.rpptx]{length.rpptx()}},
\code{\link[=plot_layout_properties]{plot_layout_properties()}},
\code{\link[=slide_size]{slide_size()}},
\code{\link[=slide_summary]{slide_summary()}}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{docx_set_character_style}
\alias{docx_set_character_style}
\title{Add character style in a Word document}
\usage{
docx_set_character_style(
  x,
  style_id,
  style_name,
  base_on,
  fp_t = fp_text_lite()
)
}
\arguments{
\item{x}{an rdocx object}

\item{style_id}{a unique style identifier for Word.}

\item{style_name}{a unique label associated with the style identifier.
This label is the name of the style when Word edit the document.}

\item{base_on}{the character style name used as base style}

\item{fp_t}{Text formatting properties, see \code{\link[=fp_text]{fp_text()}}.}
}
\description{
The function lets you add or modify Word character styles.
}
\examples{
library(officer)
doc <- read_docx()

doc <- docx_set_character_style(
  doc,
  style_id = "newcharstyle",
  style_name = "label for char style",
  base_on = "Default Paragraph Font",
  fp_text_lite(
    shading.color = "red",
    color = "white")
)
paragraph <- fpar(
  run_wordtext("hello",
    style_id = "newcharstyle"))

doc <- body_add_fpar(doc, value = paragraph)
docx_file <- print(doc, target = tempfile(fileext = ".docx"))
docx_file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{sheet_write_data}
\alias{sheet_write_data}
\title{Write data to a sheet}
\usage{
sheet_write_data(x, value, sheet, start_row = 1L, start_col = 1L, ...)
}
\arguments{
\item{x}{rxlsx object}

\item{value}{a \code{data.frame}, \code{character} vector, \code{\link[=fpar]{fpar()}} or
\code{\link[=block_list]{block_list()}}}

\item{sheet}{sheet name (must already exist)}

\item{start_row}{row index where the header / first cell will be
written (default 1)}

\item{start_col}{column index where the first column / first cell
will be written (default 1)}

\item{...}{method-specific arguments. In particular,
\code{direction = "vertical"} (default) or \code{"horizontal"} is honoured by
the \code{character} and \code{block_list} methods.}
}
\description{
Write a content into a sheet of an xlsx workbook.
Multiple calls can write to different positions on the same sheet.

This is a generic function dispatching on \code{value}. Supported inputs:
\itemize{
\item \code{data.frame}: written as a table (header in row \code{start_row}, data
starting at \code{start_row + 1}).
\item \code{character}: each element in its own cell. The \code{character} method
accepts a \code{direction} argument (\code{"vertical"} -- default -- or
\code{"horizontal"}) to stack elements in a column or a row.
\item \code{\link[=fpar]{fpar()}}: richtext paragraph written into a single cell at
\verb{(start_row, start_col)}. Font, size, colour, bold, italic,
underline, strikethrough and sub/superscript chunks are honoured.
\item \code{\link[=block_list]{block_list()}}: one cell per \code{fpar} item. Also accepts
\code{direction = "vertical"} (default) or \code{"horizontal"}.
}
}
\examples{
# sheet_write_data() is an S3 generic dispatching on `value`:
#   - data.frame
#   - character
#   - fpar
#   - block_list

library(officer)

wb <- read_xlsx()
# drop the template's default sheet so the workbook contains only `demo`
wb <- sheet_remove(wb, sheet = sheet_names(wb)[1])
wb <- add_sheet(wb, label = "demo")

# --- A1:C4  data.frame (tabular write) ----
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  value = head(iris, 3)[, 1:3],
  start_row = 1,
  start_col = 1
)

# --- A6:A8  character, vertical default ----
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  value = c("first", "second", "third"),
  start_row = 6,
  start_col = 1
)

# --- A10:C10  character, horizontal ----
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  value = c("one", "two", "three"),
  direction = "horizontal",
  start_row = 10,
  start_col = 1
)

# --- A12  fpar with rich text in a single cell ----
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  start_row = 12,
  start_col = 1,
  value = fpar(
    ftext("bold ", fp_text(bold = TRUE, color = "red", font.size = 12)),
    ftext("italic ", fp_text(italic = TRUE, color = "blue")),
    ftext("under ", fp_text(underlined = TRUE)),
    ftext("strike ", fp_text(strike = TRUE, color = "#888888")),
    ftext("H", fp_text()),
    ftext("2", fp_text(vertical.align = "subscript")),
    ftext("O  (m", fp_text()),
    ftext("2", fp_text(vertical.align = "superscript")),
    ftext(")", fp_text())
  )
)

# --- A14:A16  block_list of fpars, stacked vertically ----
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  start_row = 14,
  start_col = 1,
  value = block_list(
    fpar(ftext(
      "header row",
      fp_text(bold = TRUE, font.size = 14, color = "#006699")
    )),
    fpar(ftext("middle row", fp_text(italic = TRUE, color = "#222222"))),
    fpar(ftext(
      "final row",
      fp_text(bold = TRUE, color = "darkgreen", font.size = 11)
    ))
  )
)

# --- B14:D14  block_list horizontal, 3 fpars in a row ----
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  start_row = 14,
  start_col = 2,
  direction = "horizontal",
  value = block_list(
    fpar(ftext("left", fp_text(bold = TRUE))),
    fpar(ftext("middle", fp_text(color = "red"))),
    fpar(ftext("right", fp_text(italic = TRUE)))
  )
)

# --- Commentary column (E) via character ---
wb <- sheet_write_data(
  wb,
  sheet = "demo",
  start_row = 1,
  start_col = 5,
  value = c(
    "data.frame -> tabular (iris cols 1-3)",
    "",
    "",
    "character(3) vertical -> A6:A8",
    "",
    "",
    "",
    "character(3) horizontal -> A10:C10",
    "",
    "",
    "fpar with 9 ftext chunks in A12",
    "",
    "",
    "block_list(3) vertical -> A14:A16",
    "",
    "",
    "(B14:D14) block_list horizontal"
  )
)

out <- tempfile(fileext = ".xlsx")
print(wb, target = out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_cell}
\alias{fp_cell}
\alias{format.fp_cell}
\alias{print.fp_cell}
\alias{update.fp_cell}
\title{Cell formatting properties}
\usage{
fp_cell(
  border = fp_border(width = 0),
  border.bottom,
  border.left,
  border.top,
  border.right,
  vertical.align = "center",
  margin = 0,
  margin.bottom,
  margin.top,
  margin.left,
  margin.right,
  background.color = "transparent",
  text.direction = "lrtb",
  rowspan = 1,
  colspan = 1
)

\method{format}{fp_cell}(x, type = "wml", ...)

\method{print}{fp_cell}(x, ...)

\method{update}{fp_cell}(
  object,
  border,
  border.bottom,
  border.left,
  border.top,
  border.right,
  vertical.align,
  margin = 0,
  margin.bottom,
  margin.top,
  margin.left,
  margin.right,
  background.color,
  text.direction,
  rowspan = 1,
  colspan = 1,
  ...
)
}
\arguments{
\item{border}{shortcut for all borders.}

\item{border.bottom, border.left, border.top, border.right}{\code{\link{fp_border}} for borders.}

\item{vertical.align}{cell content vertical alignment - a single character value,
expected value is one of "center" or "top" or "bottom"}

\item{margin}{shortcut for all margins.}

\item{margin.bottom, margin.top, margin.left, margin.right}{cell margins - 0 or positive integer value.}

\item{background.color}{cell background color - a single character value specifying a
valid color (e.g. "#000000" or "black").}

\item{text.direction}{cell text rotation - a single character value, expected
value is one of "lrtb", "tbrl", "btlr".}

\item{rowspan}{specify how many rows the cell is spanned over}

\item{colspan}{specify how many columns the cell is spanned over}

\item{x, object}{\code{fp_cell} object}

\item{type}{output type - one of 'wml', 'pml', 'html', 'rtf'.}

\item{...}{further arguments - not used}
}
\description{
Create a \code{fp_cell} object that describes cell formatting properties.
}
\examples{
obj <- fp_cell(margin = 1)
update(obj, margin.bottom = 5)
}
\seealso{
Other functions for defining formatting properties: 
\code{\link{fp_border}()},
\code{\link{fp_par}()},
\code{\link{fp_tab}()},
\code{\link{fp_tabs}()},
\code{\link{fp_text}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_table}
\alias{body_add_table}
\title{Add table in a 'Word' document}
\usage{
body_add_table(
  x,
  value,
  style = NULL,
  pos = "after",
  header = TRUE,
  alignment = NULL,
  align_table = "center",
  stylenames = table_stylenames(),
  first_row = TRUE,
  first_column = FALSE,
  last_row = FALSE,
  last_column = FALSE,
  no_hband = FALSE,
  no_vband = TRUE
)
}
\arguments{
\item{x}{a docx device}

\item{value}{a data.frame to add as a table}

\item{style}{table style}

\item{pos}{where to add the new element relative to the cursor,
one of after", "before", "on".}

\item{header}{display header if TRUE}

\item{alignment}{columns alignement, argument length must match with columns length,
values must be "l" (left), "r" (right) or "c" (center).}

\item{align_table}{table alignment within document, value must be "left", "center" or "right"}

\item{stylenames}{columns styles defined by \code{\link[=table_stylenames]{table_stylenames()}}}

\item{first_row}{Specifies that the first column conditional formatting should be
applied. Details for this and other conditional formatting options can be found
at http://officeopenxml.com/WPtblLook.php.}

\item{first_column}{Specifies that the first column conditional formatting should
be applied.}

\item{last_row}{Specifies that the first column conditional formatting should be applied.}

\item{last_column}{Specifies that the first column conditional formatting should be
applied.}

\item{no_hband}{Specifies that the first column conditional formatting should be applied.}

\item{no_vband}{Specifies that the first column conditional formatting should be applied.}
}
\description{
Add a table into an rdocx object.
}
\examples{
doc <- read_docx()
doc <- body_add_table(doc, iris, style = "table_template")

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_par}()},
\code{\link{body_add_plot}()},
\code{\link{body_add_toc}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{run_reference}
\alias{run_reference}
\title{Cross reference}
\usage{
run_reference(id, prop = NULL)
}
\arguments{
\item{id}{reference id, a string}

\item{prop}{formatting text properties returned by \link{fp_text}.}
}
\description{
Create a representation of a reference
}
\section{usage}{

You can use this function in conjunction with \link{fpar} to create paragraphs
consisting of differently formatted text parts. You can also use this
function as an \emph{r chunk} in an R Markdown document made with package
officedown.
}

\examples{
run_reference("a_ref")
}
\seealso{
Other run functions for reporting:
\code{\link[=external_img]{external_img()}},
\code{\link[=floating_external_img]{floating_external_img()}},
\code{\link[=ftext]{ftext()}},
\code{\link[=hyperlink_ftext]{hyperlink_ftext()}},
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_bookmark]{run_bookmark()}},
\code{\link[=run_columnbreak]{run_columnbreak()}},
\code{\link[=run_comment]{run_comment()}},
\code{\link[=run_footnote]{run_footnote()}},
\code{\link[=run_footnoteref]{run_footnoteref()}},
\code{\link[=run_linebreak]{run_linebreak()}},
\code{\link[=run_pagebreak]{run_pagebreak()}},
\code{\link[=run_tab]{run_tab()}},
\code{\link[=run_word_field]{run_word_field()}},
\code{\link[=run_wordtext]{run_wordtext()}}

Other Word computed fields:
\code{\link[=run_autonum]{run_autonum()}},
\code{\link[=run_word_field]{run_word_field()}}
}
\concept{Word computed fields}
\concept{run functions for reporting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{remove_slide}
\alias{remove_slide}
\title{Remove slide(s)}
\usage{
remove_slide(x, index = NULL, rm_images = FALSE)
}
\arguments{
\item{x}{an rpptx object}

\item{index}{slide index or a vector of slide indices to remove,
default to current slide position.}

\item{rm_images}{unused anymore.}
}
\description{
Remove one or more slides from a pptx presentation.
}
\note{
cursor is set on the last slide.
}
\examples{
library(officer)

x <- read_pptx()
x <- add_slide(x, "Title and Content")
x <- remove_slide(x)

# Remove multiple slides at once
x <- read_pptx()
x <- add_slide(x, "Title and Content")
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)
x <- add_slide(
  x,
  layout = "Two Content",
  `Title 1` = "A title",
  dt = "Jan. 26, 2025",
  `body[2]` = "Body 2",
  left = "Left side",
  `6` = "Footer"
)
x <- add_slide(x, "Title and Content")
x <- remove_slide(x, index = c(2, 4))
pptx_file <- print(x, target = tempfile(fileext = ".pptx"))
pptx_file
}
\seealso{
\code{\link[=read_pptx]{read_pptx()}}, \code{\link[=ph_with]{ph_with()}}, \code{\link[=ph_remove]{ph_remove()}}

Other functions to manipulate slides:
\code{\link[=add_slide]{add_slide()}},
\code{\link[=move_slide]{move_slide()}},
\code{\link[=on_slide]{on_slide()}},
\code{\link[=set_notes]{set_notes()}}
}
\concept{slide_manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/empty_content.R
\name{empty_content}
\alias{empty_content}
\title{Empty block for 'PowerPoint'}
\usage{
empty_content()
}
\description{
Create an empty object to include as an empty placeholder shape in a
presentation. This comes in handy when presentation are updated
through R, but a user still wants to add some comments in this new content.

Empty content also works with layout fields (slide number and date) to preserve them:
they are included on the slide and keep being updated by PowerPoint, i.e. update to the
when the slide number when the slide moves in the deck, update to the date.
}
\examples{
fileout <- tempfile(fileext = ".pptx")
doc <- read_pptx()
doc <- add_slide(doc, layout = "Two Content",
  master = "Office Theme")
doc <- ph_with(x = doc, value = empty_content(),
 location = ph_location_type(type = "title") )

doc <- add_slide(doc)
# add slide number as a computer field
doc <- ph_with(
  x = doc, value = empty_content(),
  location = ph_location_type(type = "sldNum"))

print(doc, target = fileout )
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=body_add_blocks]{body_add_blocks()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ph_location.R
\name{ph_location_label}
\alias{ph_location_label}
\title{Location of a named placeholder}
\usage{
ph_location_label(ph_label, newlabel = NULL, ...)
}
\arguments{
\item{ph_label}{placeholder label of the used layout. It can be read in PowerPoint or
with function \code{\link[=layout_properties]{layout_properties()}} in column \code{ph_label}.}

\item{newlabel}{a label to associate with the placeholder.}

\item{...}{unused arguments}
}
\description{
The function will use the label of a placeholder
to find the corresponding location.
}
\details{
The location of the bounding box associated to a placeholder
within a slide is specified with the left top coordinate,
the width and the height. These are defined in inches:

\describe{
\item{left}{left coordinate of the bounding box}
\item{top}{top coordinate of the bounding box}
\item{width}{width of the bounding box}
\item{height}{height of the bounding box}
}

In addition to these attributes, a label can be
associated with the shape. Shapes, text boxes, images and other objects
will be identified with that label in the \emph{Selection Pane} of PowerPoint.
This label can then be reused by other functions such as \code{ph_location_label()}.
It can be set with argument \code{newlabel}.
}
\examples{
library(officer)

# ph_location_label demo ----

doc <- read_pptx()
doc <- add_slide(doc, layout = "Title and Content")

# all ph_label can be read here
layout_properties(doc, layout = "Title and Content")

doc <- ph_with(
  doc,
  head(iris),
  location = ph_location_label(ph_label = "Content Placeholder 2")
)
doc <- ph_with(
  doc,
  format(Sys.Date()),
  location = ph_location_label(ph_label = "Date Placeholder 3")
)
doc <- ph_with(
  doc,
  "This is a title",
  location = ph_location_label(ph_label = "Title 1")
)

print(doc, target = tempfile(fileext = ".pptx"))
}
\seealso{
Other functions for placeholder location:
\code{\link[=ph_location]{ph_location()}},
\code{\link[=ph_location_fullsize]{ph_location_fullsize()}},
\code{\link[=ph_location_id]{ph_location_id()}},
\code{\link[=ph_location_left]{ph_location_left()}},
\code{\link[=ph_location_right]{ph_location_right()}},
\code{\link[=ph_location_template]{ph_location_template()}},
\code{\link[=ph_location_type]{ph_location_type()}}
}
\concept{functions for placeholder location}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_replace.R
\name{body_replace_all_text}
\alias{body_replace_all_text}
\alias{headers_replace_all_text}
\alias{footers_replace_all_text}
\title{Replace text anywhere in the document}
\usage{
body_replace_all_text(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
)

headers_replace_all_text(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
)

footers_replace_all_text(
  x,
  old_value,
  new_value,
  only_at_cursor = FALSE,
  warn = TRUE,
  ...
)
}
\arguments{
\item{x}{a docx device}

\item{old_value}{the value to replace}

\item{new_value}{the value to replace it with}

\item{only_at_cursor}{if \code{TRUE}, only search-and-replace at the current
cursor; if \code{FALSE} (default), search-and-replace in the entire document
(this can be slow on large documents!)}

\item{warn}{warn if \code{old_value} could not be found.}

\item{...}{optional arguments to grepl/gsub (e.g. \code{fixed=TRUE})}
}
\description{
Replace text anywhere in the document, or at a cursor.

Replace all occurrences of old_value with new_value. This method
uses \code{\link{grepl}}/\code{\link{gsub}} for pattern matching; you may
supply arguments as required (and therefore use \code{\link{regex}} features)
using the optional \code{...} argument.

Note that by default, grepl/gsub will use \code{fixed=FALSE}, which means
that \code{old_value} and \code{new_value} will be interepreted as regular
expressions.

\strong{Chunking of text}

Note that the behind-the-scenes representation of text in a Word document is
frequently not what you might expect! Sometimes a paragraph of text is broken
up (or "chunked") into several "runs," as a result of style changes, pauses
in text entry, later revisions and edits, etc. If you have not styled the
text, and have entered it in an "all-at-once" fashion, e.g. by pasting it or
by outputing it programmatically into your Word document, then this will
likely not be a problem. If you are working with a manually-edited document,
however, this can lead to unexpected failures to find text.

You can use the officer function \code{\link{docx_show_chunk}} to
show how the paragraph of text at the current cursor has been chunked into
runs, and what text is in each chunk. This can help troubleshoot unexpected
failures to find text.
}
\section{header_replace_all_text}{

Replacements will be performed in each header of all sections.


Replacements will be performed in each footer of all sections.
}

\examples{
doc <- read_docx()
doc <- body_add_par(doc, "Placeholder one")
doc <- body_add_par(doc, "Placeholder two")

# Show text chunk at cursor
docx_show_chunk(doc)  # Output is 'Placeholder two'

# Simple search-and-replace at current cursor, with regex turned off
doc <- body_replace_all_text(doc, old_value = "Placeholder",
  new_value = "new", only_at_cursor = TRUE, fixed = TRUE)
docx_show_chunk(doc)  # Output is 'new two'

# Do the same, but in the entire document and ignoring case
doc <- body_replace_all_text(doc, old_value = "placeholder",
  new_value = "new", only_at_cursor=FALSE, ignore.case = TRUE)
doc <- cursor_backward(doc)
docx_show_chunk(doc) # Output is 'new one'

# Use regex : replace all words starting with "n" with the word "example"
doc <- body_replace_all_text(doc, "\\\\bn.*?\\\\b", "example")
docx_show_chunk(doc) # Output is 'example one'
}
\seealso{
\code{\link{grep}}, \code{\link{regex}}, \code{\link{docx_show_chunk}}
}
\author{
Frank Hangler, \email{frank@plotandscatter.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{read_docx}
\alias{read_docx}
\alias{print.rdocx}
\title{Create a 'Word' document object}
\usage{
read_docx(path = NULL)

\method{print}{rdocx}(x, target = NULL, ...)
}
\arguments{
\item{path}{path to the docx file to use as base document.
\code{dotx} file are supported.}

\item{x}{an rdocx object}

\item{target}{path to the docx file to write}

\item{...}{unused}
}
\value{
an object of class \code{rdocx}.
}
\description{
read and import a docx file as an R object
representing the document. When no file is specified, it
uses a default empty file.

Use then this object to add content to it and create Word files
from R.
}
\section{Functions}{
\itemize{
\item \code{print(rdocx)}: write docx to a file. It returns the path of the result
file.

}}
\section{styles}{


\code{read_docx()} uses a Word file as the initial document.
This is the original Word document from which the document
layout, paragraph styles, or table styles come.

You will be able to add formatted text, change the paragraph
style with the R api but also use the styles from the
original document.

See \verb{body_add_*} functions to add content.
}

\section{Illustrations}{


\if{html}{\figure{read_docx_doc_1.png}{options: width=80\%}}

\if{html}{\figure{read_docx_doc_2.png}{options: width=80\%}}
}

\examples{
library(officer)

pinst <- plot_instr({
  z <- c(rnorm(100), rnorm(50, mean = 5))
  plot(density(z))
})

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, "This is a table", style = "heading 2")
doc_1 <- body_add_table(doc_1, value = mtcars, style = "table_template")
doc_1 <- body_add_par(doc_1, "This is a plot", style = "heading 2")
doc_1 <- body_add_plot(doc_1, pinst)
docx_file_1 <- print(doc_1, target = tempfile(fileext = ".docx"))

template <- system.file(package = "officer",
  "doc_examples", "landscape.docx")
doc_2 <- read_docx(path = template)
doc_2 <- body_add_par(doc_2, "This is a table", style = "heading 2")
doc_2 <- body_add_table(doc_2, value = mtcars)
doc_2 <- body_add_par(doc_2, "This is a plot", style = "heading 2")
doc_2 <- body_add_plot(doc_2, pinst)
docx_file_2 <- print(doc_2, target = tempfile(fileext = ".docx"))

}
\seealso{
\link{body_add_par}, \link{body_add_plot}, \link{body_add_table}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_embed_font.R
\name{docx_embed_font}
\alias{docx_embed_font}
\title{Embed Fonts in a Word Document}
\usage{
docx_embed_font(
  x,
  font_family,
  regular = NULL,
  bold = NULL,
  italic = NULL,
  bold_italic = NULL
)
}
\arguments{
\item{x}{an rdocx object}

\item{font_family}{font family name as it will appear in the document.
When \code{regular} is not provided, this name is used to look up font
files via \code{\link[gdtools:sys_fonts]{gdtools::sys_fonts()}}.}

\item{regular}{path to the regular .ttf/.otf font file. If \code{NULL},
font files are detected automatically via gdtools.}

\item{bold}{path to the bold .ttf/.otf font file (optional)}

\item{italic}{path to the italic .ttf/.otf font file (optional)}

\item{bold_italic}{path to the bold-italic .ttf/.otf font file (optional)}
}
\value{
the rdocx object with embedded fonts
}
\description{
Copy TrueType or OpenType font files from the local
system into a Word document. The font data is stored inside the
\code{.docx} archive so that the document renders correctly when opened
on a system where the font is not installed.

When only \code{font_family} is provided (without \code{regular}), the function
looks up font file paths on the local system via \code{\link[gdtools:sys_fonts]{gdtools::sys_fonts()}}
and copies them into the document. A message shows the equivalent
explicit call for reproducibility.
}
\section{Font licensing}{

Embedding a font in a document redistributes it. You must ensure that
the font license permits embedding. Fonts under the SIL Open Font
License (e.g. Liberation, Google Fonts) generally allow it. Many
commercial fonts restrict or prohibit embedding. Check the license
of the font before using this function.
}

\examples{
\dontshow{if (require("gdtools", quietly = TRUE, character.only = TRUE)) withAutoprint(\{ # examplesIf}
# automatic detection (requires gdtools)
gdtools::register_liberationsans()
doc <- read_docx()
doc <- docx_embed_font(doc, font_family = "Liberation Sans")

# # explicit paths (no gdtools needed)
# doc <- docx_embed_font(
#   doc,
#   font_family = "My Font",
#   regular = "path/to/font-regular.ttf",
#   bold = "path/to/font-bold.ttf"
# )
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[gdtools:sys_fonts]{gdtools::sys_fonts()}}, \code{\link[gdtools:register_gfont]{gdtools::register_gfont()}},
\code{\link[=docx_set_settings]{docx_set_settings()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_run_objects.R
\name{to_pml}
\alias{to_pml}
\title{Convert officer objects to PresentationML}
\usage{
to_pml(x, add_ns = FALSE, ...)
}
\arguments{
\item{x}{object}

\item{add_ns}{should namespace be added to the top tag}

\item{...}{Arguments to be passed to methods}
}
\description{
Convert an object made with package officer
to PresentationML. The result is a string.
}
\seealso{
Other functions for officer extensions: 
\code{\link{docx_reference_img}()},
\code{\link{fortify_location}()},
\code{\link{get_reference_value}()},
\code{\link{opts_current_table}()},
\code{\link{shape_properties_tags}()},
\code{\link{str_encode_to_rtf}()},
\code{\link{to_html}()},
\code{\link{to_rtf}()},
\code{\link{to_wml}()},
\code{\link{wml_link_images}()}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{xlsx_drawing}
\alias{xlsx_drawing}
\title{Drawing manager for xlsx sheets}
\description{
R6 class that manages a drawing file for an xlsx
sheet. Used internally by \code{\link[=sheet_add_drawing]{sheet_add_drawing()}} methods.
}
\keyword{internal}
\section{Super class}{
\code{openxml_document} -> \code{xlsx_drawing}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-xlsx_drawing-initialize}{\code{xlsx_drawing$new()}}
    \item \href{#method-xlsx_drawing-add_chart_anchor}{\code{xlsx_drawing$add_chart_anchor()}}
    \item \href{#method-xlsx_drawing-add_chart_rel}{\code{xlsx_drawing$add_chart_rel()}}
    \item \href{#method-xlsx_drawing-add_image_rel}{\code{xlsx_drawing$add_image_rel()}}
    \item \href{#method-xlsx_drawing-add_image_anchor}{\code{xlsx_drawing$add_image_anchor()}}
    \item \href{#method-xlsx_drawing-clone}{\code{xlsx_drawing$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><code>openxml_document$dir_name()</code></li>
  <li><code>openxml_document$feed()</code></li>
  <li><code>openxml_document$file_name()</code></li>
  <li><code>openxml_document$get()</code></li>
  <li><code>openxml_document$name()</code></li>
  <li><code>openxml_document$rel_df()</code></li>
  <li><code>openxml_document$relationship()</code></li>
  <li><code>openxml_document$remove()</code></li>
  <li><code>openxml_document$replace_xml()</code></li>
  <li><code>openxml_document$save()</code></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_drawing-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_drawing-initialize}{}}}
\subsection{\code{xlsx_drawing$new()}}{
  Create or reuse a drawing for a sheet.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_drawing$new(package_dir, sheet_obj, content_type)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{package_dir}}{path to the unpacked xlsx directory}
      \item{\code{sheet_obj}}{a sheet R6 object}
      \item{\code{content_type}}{a content_type R6 object}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_drawing-add_chart_anchor"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_drawing-add_chart_anchor}{}}}
\subsection{\code{xlsx_drawing$add_chart_anchor()}}{
  Place a chart on the sheet. The chart can be
positioned in three ways, matching Excel's own "Format
Picture" options:
\itemize{
\item Move and size with cells (Excel's default): supply
\code{from} and \code{to} as cell references (e.g. \code{"B2"} and
\code{"H20"}). The chart spans those two cells and follows them
when rows/columns are inserted or resized.
\item Move but don't size with cells: supply \code{from} only.
The chart anchors to that cell with the size set by
\code{width}/\code{height}.
\item Fixed position: leave \code{from} and \code{to} \code{NULL}. The
chart is placed at \code{left}/\code{top} with size \code{width}/\code{height},
independent of cell layout.
}
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_drawing$add_chart_anchor(
  chart_rid,
  left = 1,
  top = 1,
  width = 6,
  height = 4,
  from = NULL,
  to = NULL,
  edit_as = c("twoCell", "oneCell", "absolute"),
  graphic_uri = ooxml_chart_uris("drawingml")$graphic_uri,
  chart_inner = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{chart_rid}}{relationship id of the chart (returned by
\code{add_chart_rel()}). Ignored when \code{chart_inner} is supplied
explicitly.}
      \item{\code{left, top}}{top-left position in inches (fixed-position mode)}
      \item{\code{width, height}}{size in inches}
      \item{\code{from, to}}{Excel cell references like \code{"C4"} (1-based,
case-insensitive)}
      \item{\code{edit_as}}{how Excel should treat the drawing when rows or
columns are resized: \code{"twoCell"} (move and size with cells, the
default), \code{"oneCell"} (move only) or \code{"absolute"} (neither). Only
meaningful when both \code{from} and \code{to} are supplied.}
      \item{\code{graphic_uri}}{chart family URI; defaults to the classic
chart family. Use \code{\link[=ooxml_chart_uris]{ooxml_chart_uris()}} to obtain the value
for the chartEx family.}
      \item{\code{chart_inner}}{advanced. XML fragment referencing the chart
part. Leave \code{NULL} to let the method build the right reference
from \code{graphic_uri}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_drawing-add_chart_rel"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_drawing-add_chart_rel}{}}}
\subsection{\code{xlsx_drawing$add_chart_rel()}}{
  Register a chart part with the drawing. Returns
the relationship id, to be passed to \code{add_chart_anchor()}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_drawing$add_chart_rel(
  chart_basename,
  rel_type = ooxml_chart_uris("drawingml")$rel_type
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{chart_basename}}{filename of the chart XML}
      \item{\code{rel_type}}{chart family relationship type; defaults to the
classic chart family. Use \code{\link[=ooxml_chart_uris]{ooxml_chart_uris()}} to obtain the
value for the chartEx family.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_drawing-add_image_rel"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_drawing-add_image_rel}{}}}
\subsection{\code{xlsx_drawing$add_image_rel()}}{
  Add a relationship from the drawing to a media image.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_drawing$add_image_rel(image_basename)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{image_basename}}{filename (without directory) of the image
sitting in \verb{xl/media/}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_drawing-add_image_anchor"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_drawing-add_image_anchor}{}}}
\subsection{\code{xlsx_drawing$add_image_anchor()}}{
  Place an image on the sheet. Positioning works
the same way as \code{add_chart_anchor()}: supply \code{from} and \code{to}
to move and size with cells, \code{from} only to move (but not
resize) with cells, or leave both \code{NULL} for a fixed position.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_drawing$add_image_anchor(
  image_rid,
  left = 1,
  top = 1,
  width = 2,
  height = 2,
  from = NULL,
  to = NULL,
  edit_as = c("twoCell", "oneCell", "absolute"),
  alt = ""
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{image_rid}}{relationship id of the image (returned by
\code{add_image_rel()})}
      \item{\code{left, top}}{top-left position in inches (fixed-position mode)}
      \item{\code{width, height}}{size in inches}
      \item{\code{from, to}}{Excel cell references like \code{"C4"} (1-based,
case-insensitive)}
      \item{\code{edit_as}}{how Excel should treat the drawing when rows or
columns are resized: \code{"twoCell"} (move and size with cells, the
default), \code{"oneCell"} (move only) or \code{"absolute"} (neither). Only
meaningful when both \code{from} and \code{to} are supplied.}
      \item{\code{alt}}{alternative text}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-xlsx_drawing-clone"></a>}}
\if{latex}{\out{\hypertarget{method-xlsx_drawing-clone}{}}}
\subsection{\code{xlsx_drawing$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{xlsx_drawing$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_docx.R
\name{docx_set_paragraph_style}
\alias{docx_set_paragraph_style}
\title{Add or replace paragraph style in a Word document}
\usage{
docx_set_paragraph_style(
  x,
  style_id,
  style_name,
  base_on = "Normal",
  fp_p = fp_par(),
  fp_t = NULL
)
}
\arguments{
\item{x}{an rdocx object}

\item{style_id}{a unique style identifier for Word.}

\item{style_name}{a unique label associated with the style identifier.
This label is the name of the style when Word edit the document.}

\item{base_on}{the style name used as base style}

\item{fp_p}{paragraph formatting properties, see \code{\link[=fp_par]{fp_par()}}.}

\item{fp_t}{default text formatting properties. This is used as
text formatting properties, see \code{\link[=fp_text]{fp_text()}}. If NULL (default), the
paragraph will used the default text formatting properties (defined by
the \code{base_on} argument).}
}
\description{
The function lets you add or replace a Word paragraph style.
}
\examples{
library(officer)

doc <- read_docx()

doc <- docx_set_paragraph_style(
  doc,
  style_id = "rightaligned",
  style_name = "Explicit label",
  fp_p = fp_par(text.align = "right", padding = 20),
  fp_t = fp_text_lite(
    bold = TRUE,
    shading.color = "#FD34F0",
    color = "white")
)

doc <- body_add_par(doc,
  value = "This is a test",
  style = "Explicit label")

docx_file <- print(doc, target = tempfile(fileext = ".docx"))
docx_file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dev-utils.R
\name{officer_url_encode}
\alias{officer_url_encode}
\title{officer url encoder}
\usage{
officer_url_encode(x)
}
\arguments{
\item{x}{a character vector of URL}
}
\description{
encode url so that it can be easily
decoded when 'officer' write a file to the disk.
}
\examples{
officer_url_encode("https://cran.r-project.org/")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_slide_manip.R
\name{shape_properties_tags}
\alias{shape_properties_tags}
\title{pptx tags for visual and non visual properties}
\usage{
shape_properties_tags(
  left = 0,
  top = 0,
  width = 3,
  height = 3,
  bg = "transparent",
  rot = 0,
  label = "",
  ph = "<p:ph/>",
  ln = sp_line(lwd = 0, linecmpd = "solid", lineend = "rnd"),
  geom = NULL
)
}
\arguments{
\item{left, top, width, height}{place holder coordinates
in inches.}

\item{bg}{background color}

\item{rot}{rotation angle}

\item{label}{a label for the placeholder.}

\item{ph}{string containing xml code for ph tag}

\item{ln}{a \code{\link[=sp_line]{sp_line()}} object specifying the outline style.}

\item{geom}{shape geometry, see http://www.datypic.com/sc/ooxml/t-a_ST_ShapeType.html}
}
\value{
a character value
}
\description{
Visual and non visual properties of a
shape can be returned by this function.
}
\seealso{
Other functions for officer extensions:
\code{\link[=fortify_location]{fortify_location()}},
\code{\link[=get_reference_value]{get_reference_value()}},
\code{\link[=opts_current_table]{opts_current_table()}},
\code{\link[=str_encode_to_rtf]{str_encode_to_rtf()}},
\code{\link[=to_html]{to_html()}},
\code{\link[=to_pml]{to_pml()}},
\code{\link[=to_rtf]{to_rtf()}},
\code{\link[=to_wml]{to_wml()}},
\code{\link[=wml_link_images]{wml_link_images()}}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{slide_summary}
\alias{slide_summary}
\title{Slide content in a data.frame}
\usage{
slide_summary(x, index = NULL)
}
\arguments{
\item{x}{an rpptx object}

\item{index}{slide index}
}
\description{
Get content and positions of current slide
into a data.frame. Data for any tables, images, or paragraphs are
imported into the resulting data.frame.
}
\note{
The column \code{id} of the result is not to be used by users.
This is a technical string id whose value will be used by office
when the document will be rendered. This is not related to argument
\code{index} required by functions \code{ph_with}.
}
\examples{
my_pres <- read_pptx()
my_pres <- add_slide(my_pres)
my_pres <- ph_with(my_pres, format(Sys.Date()),
  location = ph_location_type(type="dt"))
my_pres <- add_slide(my_pres)
my_pres <- ph_with(my_pres, iris[1:2,],
  location = ph_location_type(type="body"))
slide_summary(my_pres)
slide_summary(my_pres, index = 1)
}
\seealso{
Other functions for reading presentation information: 
\code{\link{annotate_base}()},
\code{\link{color_scheme}()},
\code{\link{doc_properties}()},
\code{\link{layout_properties}()},
\code{\link{layout_summary}()},
\code{\link{length.rpptx}()},
\code{\link{plot_layout_properties}()},
\code{\link{slide_size}()}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add_plot}
\alias{body_add_plot}
\title{Add plot in a 'Word' document}
\usage{
body_add_plot(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  pos = "after",
  unit = "in",
  ...
)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{plot instructions, see \code{\link[=plot_instr]{plot_instr()}}.}

\item{width, height}{plot size in units expressed by the unit argument.
Defaults to a width of 6 and a height of 5 "in"ches.}

\item{res}{resolution of the png image in ppi}

\item{style}{paragraph style}

\item{pos}{where to add the new element relative to the cursor,
one of "after", "before", "on".}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}

\item{...}{Arguments to be passed to png function.}
}
\description{
Add a plot as a png image into an rdocx object.
}
\examples{
doc <- read_docx()

if (capabilities(what = "png")) {
  p <- plot_instr(
      code = {
        barplot(1:5, col = 2:6)
      }
    )

  doc <- body_add_plot(doc, value = p, style = "centered")

  # Set the unit in which the width and height arguments are expressed
  doc <- body_add_plot(doc, value = p, style = "centered", unit = "cm")
}

print(doc, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for adding content: 
\code{\link{body_add_blocks}()},
\code{\link{body_add_break}()},
\code{\link{body_add_caption}()},
\code{\link{body_add_docx}()},
\code{\link{body_add_fpar}()},
\code{\link{body_add_gg}()},
\code{\link{body_add_img}()},
\code{\link{body_add_list}()},
\code{\link{body_add_par}()},
\code{\link{body_add_table}()},
\code{\link{body_add_toc}()},
\code{\link{body_append_start_context}()},
\code{\link{body_import_docx}()}
}
\concept{functions for adding content}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/formatting_properties.R
\name{fp_border}
\alias{fp_border}
\alias{update.fp_border}
\title{Border properties object}
\usage{
fp_border(color = "black", style = "solid", width = 1)

\method{update}{fp_border}(object, color, style, width, ...)
}
\arguments{
\item{color}{border color - single character value (e.g. "#000000" or "black")}

\item{style}{border style - single character value : See Details for supported
border styles.}

\item{width}{border width - an integer value : 0>= value}

\item{object}{fp_border object}

\item{...}{further arguments - not used}
}
\description{
create a border properties object.
}
\details{
For Word output the following border styles are supported:
\itemize{
\item "none" or "nil" - No Border
\item "solid" or "single" - Single Line Border
\item "thick" - Single Line Border
\item "double" - Double Line Border
\item "dotted" - Dotted Line Border
\item "dashed" - Dashed Line Border
\item "dotDash" - Dot Dash Line Border
\item "dotDotDash" - Dot Dot Dash Line Border
\item "triple" - Triple Line Border
\item "thinThickSmallGap" - Thin, Thick Line Border
\item "thickThinSmallGap" - Thick, Thin Line Border
\item "thinThickThinSmallGap" - Thin, Thick, Thin Line Border
\item "thinThickMediumGap" - Thin, Thick Line Border
\item "thickThinMediumGap" - Thick, Thin Line Border
\item "thinThickThinMediumGap" - Thin, Thick, Thin Line Border
\item "thinThickLargeGap" - Thin, Thick Line Border
\item "thickThinLargeGap" - Thick, Thin Line Border
\item "thinThickThinLargeGap" - Thin, Thick, Thin Line Border
\item "wave" - Wavy Line Border
\item "doubleWave" - Double Wave Line Border
\item "dashSmallGap" - Dashed Line Border
\item "dashDotStroked" - Dash Dot Strokes Line Border
\item "threeDEmboss" or "ridge" - 3D Embossed Line Border
\item "threeDEngrave" or "groove" - 3D Engraved Line Border
\item "outset" - Outset Line Border
\item "inset" - Inset Line Border
}

For HTML output only a limited amount of border styles are supported:
\itemize{
\item "none" or "nil" - No Border
\item "solid" or "single" - Single Line Border
\item "double" - Double Line Border
\item "dotted" - Dotted Line Border
\item "dashed" - Dashed Line Border
\item "threeDEmboss" or "ridge" - 3D Embossed Line Border
\item "threeDEngrave" or "groove" - 3D Engraved Line Border
\item "outset" - Outset Line Border
\item "inset" - Inset Line Border
}

Non-supported Word border styles will default to "solid".
}
\examples{
fp_border()
fp_border(color = "orange", style = "solid", width = 1)
fp_border(color = "gray", style = "dotted", width = 1)

# modify object ------
border <- fp_border()
update(border, style = "dotted", width = 3)
}
\seealso{
Other functions for defining formatting properties: 
\code{\link{fp_cell}()},
\code{\link{fp_par}()},
\code{\link{fp_tab}()},
\code{\link{fp_tabs}()},
\code{\link{fp_text}()}
}
\concept{functions for defining formatting properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_pptx.R
\name{print.rpptx}
\alias{print.rpptx}
\title{Write a 'PowerPoint' file.}
\usage{
\method{print}{rpptx}(x, target = NULL, preview = FALSE, ...)
}
\arguments{
\item{x}{an \code{rpptx} object.}

\item{target}{path to the .pptx file to write. If \code{target} is \code{NULL} (default), the \code{rpptx} object is printed to
the console.}

\item{preview}{Save \code{x} to a temporary file and open it (default \code{FALSE}).}

\item{...}{unused.}
}
\value{
If preview is \code{TRUE}, returns the temp file path invisibly.
}
\description{
Create a 'PowerPoint' file from an \code{rpptx} object (created by \code{\link[=read_pptx]{read_pptx()}}).
}
\examples{
# write an rpptx object to a .pptx file ----
file <- tempfile(fileext = ".pptx")
x <- read_pptx() # empty presentation, has no slides yet
print(x, target = file)

# preview mode: save to temp file and open locally ----
\dontrun{
print(x, preview = TRUE)
}
}
\seealso{
\code{\link[=read_pptx]{read_pptx()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_with_methods.R
\name{phs_with}
\alias{phs_with}
\title{Fill multiple placeholders using key value syntax}
\usage{
phs_with(x, ..., .dots = NULL, .slide_idx = NULL)
}
\arguments{
\item{x}{A \code{rpptx} object.}

\item{...}{Key-value pairs of the form \code{"short form location" = object}. If the short form is an integer or a string
with blanks, you must wrap it in quotes or backticks.}

\item{.dots}{List of key-value pairs \code{"short form location" = object}. Alternative to \code{...}.}

\item{.slide_idx}{Numeric indexes of slides to process. \code{NULL} (default) processes the current slide only. Use
keyword \code{all} for all slides.}
}
\description{
A sibling of \link{ph_with} that fills mutiple placeholders at once. Placeholder locations are
specfied using the short form syntax. The location and corresponding object are passed as key value pairs
(\code{phs_with("short form location" = object)}). Under the hood, \link{ph_with} is called for each pair. Note
that \code{phs_with} does not cover all options from the \verb{ph_location_*} family and is also less customization. It is a
covenience wrapper for the most common use cases. The implemented short forms are listed in section
\code{"Short forms"}.
}
\section{Short forms}{
 The following short forms are implemented and can be used as the parameter in the function
call. The corresponding function from the \verb{ph_location_*} family (called under the hood) is displayed on the
right.\tabular{lll}{
   \strong{Short form} \tab \strong{Description} \tab \strong{Location function} \cr
   \code{"left"} \tab Keyword string \tab \code{ph_location_left()} \cr
   \code{"right"} \tab Keyword string \tab \code{ph_location_right()} \cr
   \code{"fullsize"} \tab Keyword string \tab \code{ph_location_fullsize()} \cr
   \code{"body [1]"} \tab String: type + index in brackets (\code{1} if omitted) \tab \code{ph_location_type("body", 1)} \cr
   \code{"my_label"} \tab Any string not matching a keyword or type \tab \code{ph_location_label("my_label")} \cr
   \code{1} \tab Length 1 integer \tab \code{ph_location_id(1)} \cr
}
}

\examples{
library(officer)

# use key-value format to fill phs
x <- read_pptx()
x <- add_slide(x, "Two Content")
x <- phs_with(
  x,
  `Title 1` = "A title", # ph label
  dt = Sys.Date(), # ph type
  `body[2]` = "Body 2", # ph type + type index
  left = "Left side", # ph keyword
  `6` = "Footer" # ph index
)

# reuse ph content via the .dots arg
x <- read_pptx()
my_ph_list <- list(`6` = "Footer", dt = Sys.Date())
x <- add_slide(x, "Two Content")
x <- phs_with(
  x,
  `Title 1` = "Title A",
  `body[2]` = "Body A",
  .dots = my_ph_list
)
x <- add_slide(x, "Two Content")
x <- phs_with(
  x,
  `Title 1` = "Title B",
  `body[2]` = "Body B",
  .dots = my_ph_list
)

# use the .slide_idx arg to select which slide(s) to process
x <- read_pptx()
x <- add_slide(x, "Two Content")
x <- add_slide(x, "Two Content")
x <- phs_with(x, `6` = "Footer", dt = Sys.Date(), .slide_idx = 1:2)

# run to open temp pptx file locally
# \dontrun{
# print(x, preview = TRUE)
# }
}
\seealso{
\code{\link[=ph_with]{ph_with()}}, \code{\link[=add_slide]{add_slide()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_add.R
\name{body_add}
\alias{body_add}
\alias{body_add.character}
\alias{body_add.numeric}
\alias{body_add.factor}
\alias{body_add.fpar}
\alias{body_add.data.frame}
\alias{body_add.block_caption}
\alias{body_add.block_list}
\alias{body_add.block_list_items}
\alias{body_add.block_toc}
\alias{body_add.external_img}
\alias{body_add.run_pagebreak}
\alias{body_add.run_columnbreak}
\alias{body_add.gg}
\alias{body_add.plot_instr}
\alias{body_add.block_pour_docx}
\alias{body_add.block_section}
\title{Add content into a Word document}
\usage{
body_add(x, value, ...)

\method{body_add}{character}(x, value, style = NULL, ...)

\method{body_add}{numeric}(x, value, style = NULL, format_fun = formatC, ...)

\method{body_add}{factor}(x, value, style = NULL, format_fun = as.character, ...)

\method{body_add}{fpar}(x, value, style = NULL, ...)

\method{body_add}{data.frame}(
  x,
  value,
  style = NULL,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
)

\method{body_add}{block_caption}(x, value, ...)

\method{body_add}{block_list}(x, value, ...)

\method{body_add}{block_list_items}(x, value, ...)

\method{body_add}{block_toc}(x, value, ...)

\method{body_add}{external_img}(x, value, style = "Normal", ...)

\method{body_add}{run_pagebreak}(x, value, style = NULL, ...)

\method{body_add}{run_columnbreak}(x, value, style = NULL, ...)

\method{body_add}{gg}(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  scale = 1,
  unit = "in",
  ...
)

\method{body_add}{plot_instr}(
  x,
  value,
  width = 6,
  height = 5,
  res = 300,
  style = "Normal",
  unit = "in",
  ...
)

\method{body_add}{block_pour_docx}(x, value, ...)

\method{body_add}{block_section}(x, value, ...)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{object to add in the document. Supported objects
are vectors, data.frame, graphics, block of formatted paragraphs,
unordered list of formatted paragraphs,
pretty tables with package flextable, 'Microsoft' charts with package mschart.}

\item{...}{further arguments passed to or from other methods. When
adding a \code{ggplot} object or \link{plot_instr}, these arguments will be used
by png function. See method signatures to see what arguments can be used.}

\item{style}{paragraph style name. These names are available with function \link{styles_info}
and are the names of the Word styles defined in the base document (see
argument \code{path} from \link{read_docx}).}

\item{format_fun}{a function to be used to format values.}

\item{header}{display header if TRUE}

\item{tcf}{conditional formatting settings defined by \code{\link[=table_conditional_formatting]{table_conditional_formatting()}}}

\item{alignment}{columns alignement, argument length must match with columns length,
values must be "l" (left), "r" (right) or "c" (center).}

\item{width, height}{plot size in units expressed by the unit argument.
Defaults to a width of 6 and a height of 5 "in"ches.}

\item{res}{resolution of the png image in ppi}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{unit}{One of the following units in which the width and height
arguments are expressed: "in", "cm" or "mm".}
}
\description{
This function adds objects into a Word document. Values are added
as new paragraphs or tables.

This function is experimental and will replace the \verb{body_add_*} functions
later. For now it is only to be used for successive additions and cannot
be used in conjunction with the \verb{body_add_*} functions.
}
\section{Methods (by class)}{
\itemize{
\item \code{body_add(character)}: add a character vector.

\item \code{body_add(numeric)}: add a numeric vector.

\item \code{body_add(factor)}: add a factor vector.

\item \code{body_add(fpar)}: add a \link{fpar} object. These objects enable
the creation of formatted paragraphs made of formatted chunks of text.

\item \code{body_add(data.frame)}: add a data.frame object with \code{\link[=block_table]{block_table()}}.

\item \code{body_add(block_caption)}: add a \link{block_caption} object. These objects enable
the creation of set of formatted paragraphs made of formatted chunks of text.

\item \code{body_add(block_list)}: add a \link{block_list} object. Use this function
to add a list of block elements (e.g. paragraphs, images, tables)
into a Word document in a more efficient way than with usual
\verb{body_add_*} functions. This function will add several elements
in a faster way because the cursor is not calculated for
each iteraction over the elements, as a consequence the
function only append elements at the end of the document
and does not allow to insert elements at a specific position.

\item \code{body_add(block_list_items)}: add a \link{block_list_items} object (bullet or numbered list).

\item \code{body_add(block_toc)}: add a table of content (a \link{block_toc} object).

\item \code{body_add(external_img)}: add an image (a \link{external_img} object).

\item \code{body_add(run_pagebreak)}: add a \link{run_pagebreak} object.

\item \code{body_add(run_columnbreak)}: add a \link{run_columnbreak} object.

\item \code{body_add(gg)}: add a ggplot object.

\item \code{body_add(plot_instr)}: add a base plot with a \link{plot_instr} object.

\item \code{body_add(block_pour_docx)}: pour content of an external docx file with with a \link{block_pour_docx} object

\item \code{body_add(block_section)}: ends a section with a \link{block_section} object

}}
\section{Illustrations}{


\if{html}{\figure{body_add_doc_1.png}{options: style="width:70\%;"}}
}

\examples{
# This example demonstrates the versatility of body_add()
# by showing how to add various content types to a Word document

# Create a new Word document
x <- read_docx()

# Add a heading for the table of contents section
x <- body_add(x, "Table of content", style = "heading 1")

# Insert an automatic table of contents
# This will list all headings in the document
x <- body_add(x, block_toc())

# Insert a page break to start content on a new page
x <- body_add(x, run_pagebreak())

# Add a main section heading
x <- body_add(x, "Iris Dataset Sample", style = "heading 1")

# Add a data.frame as a table
# The first 6 rows of the iris dataset will be displayed
x <- body_add(x, head(iris), style = "table_template")

# Add another section heading
x <- body_add(x, "Alphabetic List", style = "heading 1")

# Add a character vector as paragraphs
# Each letter will appear as a separate paragraph with "Normal" style
x <- body_add(x, letters, style = "Normal")

# Insert a continuous section break
# Content continues on the same page but section properties can change
x <- body_add(x, block_section(prop_section(type = "continuous")))

# Add a plot created with R base graphics
# plot_instr() captures the plotting code to generate the image
x <- body_add(x, plot_instr(code = barplot(1:5, col = 2:6)))

# Change to landscape orientation for the next section
# This creates a new section with landscape page layout
x <- body_add(
  x,
  block_section(prop_section(page_size = page_size(orient = "landscape")))
)

# Save the document to a file
output_file <- tempfile(fileext = ".docx")
print(x, target = output_file)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/post-proc.R
\name{wml_link_images}
\alias{wml_link_images}
\title{transform an xml string with images references}
\usage{
wml_link_images(x, str)
}
\arguments{
\item{x}{an rdocx object}

\item{str}{wml string}
}
\description{
This function is useless now, as the processing of images
is automated when using \code{\link[=print.rdocx]{print.rdocx()}}.
}
\seealso{
Other functions for officer extensions:
\code{\link[=fortify_location]{fortify_location()}},
\code{\link[=get_reference_value]{get_reference_value()}},
\code{\link[=opts_current_table]{opts_current_table()}},
\code{\link[=shape_properties_tags]{shape_properties_tags()}},
\code{\link[=str_encode_to_rtf]{str_encode_to_rtf()}},
\code{\link[=to_html]{to_html()}},
\code{\link[=to_pml]{to_pml()}},
\code{\link[=to_rtf]{to_rtf()}},
\code{\link[=to_wml]{to_wml()}}
}
\concept{functions for officer extensions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ppt_ph_with_methods.R
\name{ph_with}
\alias{ph_with}
\alias{ph_with.character}
\alias{ph_with.numeric}
\alias{ph_with.factor}
\alias{ph_with.logical}
\alias{ph_with.Date}
\alias{ph_with.block_list}
\alias{ph_with.unordered_list}
\alias{ph_with.block_list_items}
\alias{ph_with.data.frame}
\alias{ph_with.gg}
\alias{ph_with.plot_instr}
\alias{ph_with.external_img}
\alias{ph_with.fpar}
\alias{ph_with.empty_content}
\alias{ph_with.xml_document}
\title{Add objects on the current slide}
\usage{
ph_with(x, value, location, ...)

ph_with.character(x, value, location, ...)

ph_with.numeric(x, value, location, format_fun = format, ...)

ph_with.factor(x, value, location, ...)

ph_with.logical(x, value, location, format_fun = format, ...)

ph_with.Date(x, value, location, date_format = NULL, ...)

ph_with.block_list(x, value, location, level_list = integer(0), ...)

ph_with.unordered_list(x, value, location, ...)

ph_with.block_list_items(x, value, location, ...)

ph_with.data.frame(
  x,
  value,
  location,
  header = TRUE,
  tcf = table_conditional_formatting(),
  alignment = NULL,
  ...
)

ph_with.gg(x, value, location, res = 300, alt_text = "", scale = 1, ...)

ph_with.plot_instr(x, value, location, res = 300, ...)

ph_with.external_img(x, value, location, use_loc_size = TRUE, ...)

ph_with.fpar(x, value, location, ...)

ph_with.empty_content(x, value, location, ...)

ph_with.xml_document(x, value, location, ...)
}
\arguments{
\item{x}{an rpptx object}

\item{value}{object to add as a new shape. Supported objects
are vectors, data.frame, graphics, block of formatted paragraphs,
unordered list of formatted paragraphs,
pretty tables with package flextable, editable graphics with
package rvg, 'Microsoft' charts with package mschart.}

\item{location}{a placeholder location object or a location short form. It will be used
to specify the location of the new shape. This location can be defined with a call to one
of the \verb{ph_location_*} functions (see section \code{"see also"}). In \code{ph_with()}, several location
short forms can be used, as listed in section \code{"Short forms"}.}

\item{...}{further arguments passed to or from other methods. When
adding a \code{ggplot} object or \code{plot_instr}, these arguments will be used
by the png function.}

\item{format_fun}{format function for non character vectors}

\item{date_format}{A format string for dates (default \code{"\%Y-\%m-\%d"}).
See \code{format} arg in \code{\link[=strftime]{strftime()}} for details.
Set a global default via \code{options(officer.date_format = ...)}.}

\item{level_list}{The list of levels for hierarchy structure as integer values.
If used the object is formated as an unordered list. If 1 and 2,
item 1 level will be 1, item 2 level will be 2.}

\item{header}{display header if TRUE}

\item{tcf}{conditional formatting settings defined by \code{\link[=table_conditional_formatting]{table_conditional_formatting()}}}

\item{alignment}{alignment for each columns, 'l' for left, 'r' for right
and 'c' for center. Default to NULL.}

\item{res}{resolution of the png image in ppi}

\item{alt_text}{Alt-text for screen-readers. Defaults to \code{""}. If \code{""} or \code{NULL}
an alt text added with \code{ggplot2::labs(alt = ...)} will be used if any.}

\item{scale}{Multiplicative scaling factor, same as in ggsave}

\item{use_loc_size}{if set to FALSE, external_img width and height will
be used.}
}
\description{
add an object into a new shape in the current slide. This
function is able to add all supported outputs to a presentation. See
section \strong{Methods (by class)} to see supported outputs.
}
\section{Functions}{
\itemize{
\item \code{ph_with.character()}: add a character vector to a new shape on the
current slide, values will be added as paragraphs.

\item \code{ph_with.numeric()}: add a numeric vector to a new shape on the
current slide, values will be be first formatted then
added as paragraphs.

\item \code{ph_with.factor()}: add a factor vector to a new shape on the
current slide, values will be be converted as character and then
added as paragraphs.

\item \code{ph_with.Date()}: add a \code{Date} object vector to a new shape on the
current slide, values will be be first converted to character.

\item \code{ph_with.block_list()}: add a \code{\link[=block_list]{block_list()}} made
of \code{\link[=fpar]{fpar()}} to a new shape on the current slide.

\item \code{ph_with.unordered_list()}: add a \code{\link[=unordered_list]{unordered_list()}} made
of \code{\link[=fpar]{fpar()}} to a new shape on the current slide.

\item \code{ph_with.block_list_items()}: add a \code{\link[=block_list_items]{block_list_items()}} (bullet or numbered list)
to a new shape on the current slide.

\item \code{ph_with.data.frame()}: add a data.frame to a new shape on the current slide with
function \code{\link[=block_table]{block_table()}}. Use package 'flextable' instead for more
advanced formattings.

\item \code{ph_with.gg()}: add a ggplot object to a new shape on the
current slide. Use package 'rvg' for more advanced graphical features.

\item \code{ph_with.plot_instr()}: add an R plot to a new shape on the
current slide. Use package 'rvg' for more advanced graphical features.

\item \code{ph_with.external_img()}: add a \code{\link[=external_img]{external_img()}} to a new shape
on the current slide.

When value is a external_img object, image will be copied
into the PowerPoint presentation. The width and height
specified in call to \code{\link[=external_img]{external_img()}} will be
ignored, their values will be those of the location,
unless use_loc_size is set to FALSE.

\item \code{ph_with.fpar()}: add an \code{\link[=fpar]{fpar()}} to a new shape
on the current slide as a single paragraph in a \code{\link[=block_list]{block_list()}}.

\item \code{ph_with.empty_content()}: add an \code{\link[=empty_content]{empty_content()}} to a new shape
on the current slide.

\item \code{ph_with.xml_document()}: add an xml_document object to a new shape on the
current slide. This function is to be used to add custom openxml code.

}}
\section{Short forms}{

The \code{location} argument of \code{ph_with()} either expects a location object as returned by the
\verb{ph_location_*} functions or a corresponding location \emph{short form} (string or numeric):\tabular{lll}{
   \strong{Location function} \tab \strong{Short form} \tab \strong{Description} \cr
   \code{ph_location_left()} \tab \code{"left"} \tab Keyword string \cr
   \code{ph_location_right()} \tab \code{"right"} \tab Keyword string \cr
   \code{ph_location_fullsize()} \tab \code{"fullsize"} \tab Keyword string \cr
   \code{ph_location_type("body", 1)} \tab \code{"body [1]"} \tab String: type + index in brackets (\code{1} if omitted) \cr
   \code{ph_location_label("my_label")} \tab \code{"my_label"} \tab Any string not matching a keyword or type \cr
   \code{ph_location_id(1)} \tab \code{1} \tab Length 1 integer \cr
   \code{ph_location(0, 0, 4, 5)} \tab \code{c(0,0,4,5)} \tab Length 4 numeric, optionally named, \code{c(top=0, left=0, ...)} \cr
}
}

\section{Illustrations}{


\if{html}{\figure{ph_with_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
# this name will be used to print the file
# change it to "youfile.pptx" to write the pptx
# file in your working directory.
fileout <- tempfile(fileext = ".pptx")

doc_1 <- read_pptx()
sz <- slide_size(doc_1)

# add text and a table ----
doc_1 <- add_slide(doc_1, layout = "Two Content", master = "Office Theme")
doc_1 <- ph_with(
  x = doc_1,
  value = c("Table cars"),
  location = ph_location_type(type = "title")
)
doc_1 <- ph_with(
  x = doc_1,
  value = names(cars),
  location = ph_location_left()
)
doc_1 <- ph_with(
  x = doc_1,
  value = cars,
  location = ph_location_right()
)
doc_1 <- ph_with(
  x = doc_1,
  value = Sys.Date(),
  location = ph_location_type("dt")
)

# add a base plot ----
anyplot <- plot_instr(code = {
  col <- c(
    "#440154FF",
    "#443A83FF",
    "#31688EFF",
    "#21908CFF",
    "#35B779FF",
    "#8FD744FF",
    "#FDE725FF"
  )
  barplot(1:7, col = col, yaxt = "n")
})

doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  doc_1,
  anyplot,
  location = ph_location_fullsize(),
  bg = "#006699"
)

# add a ggplot2 plot ----
if (require("ggplot2")) {
  doc_1 <- add_slide(doc_1, "Title and Content")
  gg_plot <- ggplot(data = iris) +
    geom_point(
      mapping = aes(Sepal.Length, Petal.Length),
      size = 3
    ) +
    theme_minimal()
  doc_1 <- ph_with(
    x = doc_1,
    value = gg_plot,
    location = ph_location_type(type = "body"),
    bg = "transparent"
  )
  doc_1 <- ph_with(
    x = doc_1,
    value = "graphic title",
    location = ph_location_type(type = "title")
  )
}

# add a external images ----
doc_1 <- add_slide(doc_1, layout = "Title and Content", master = "Office Theme")
doc_1 <- ph_with(
  x = doc_1,
  value = empty_content(),
  location = ph_location(
    left = 0,
    top = 0,
    width = sz$width,
    height = sz$height,
    bg = "black"
  )
)

svg_file <- file.path(R.home(component = "doc"), "html/Rlogo.svg")
if (require("rsvg")) {
  doc_1 <- ph_with(
    x = doc_1,
    value = "External images",
    location = ph_location_type(type = "title")
  )
  doc_1 <- ph_with(
    x = doc_1,
    external_img(svg_file, 100 / 72, 76 / 72),
    location = ph_location_right(),
    use_loc_size = FALSE
  )
  doc_1 <- ph_with(
    x = doc_1,
    external_img(svg_file),
    location = ph_location_left(),
    use_loc_size = TRUE
  )
}

# add a block_list ----
dummy_text <- readLines(system.file(
  package = "officer",
  "doc_examples/text.txt"
))
fp_1 <- fp_text(bold = TRUE, color = "pink", font.size = 0)
fp_2 <- fp_text(bold = TRUE, font.size = 0)
fp_3 <- fp_text(italic = TRUE, color = "red", font.size = 0)
bl <- block_list(
  fpar(ftext("hello world", fp_1)),
  fpar(
    ftext("hello", fp_2),
    ftext("hello", fp_3)
  ),
  dummy_text
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = bl,
  location = ph_location_type(type = "body")
)


# fpar ------
fpt <- fp_text(
  bold = TRUE,
  font.family = "Bradley Hand",
  font.size = 150,
  color = "#F5595B"
)
hw <- fpar(
  ftext("hello ", fpt),
  hyperlink_ftext(
    href = "https://cran.r-project.org/index.html",
    text = "cran",
    prop = fpt
  )
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = hw,
  location = ph_location_type(type = "body")
)
# unordered_list ----
ul <- unordered_list(
  level_list = c(1, 2, 2, 3, 3, 1),
  str_list = c("Level1", "Level2", "Level2", "Level3", "Level3", "Level1"),
  style = fp_text(color = "red", font.size = 0)
)
doc_1 <- add_slide(doc_1, "Title and Content")
doc_1 <- ph_with(
  x = doc_1,
  value = ul,
  location = ph_location_type()
)

print(doc_1, target = fileout)


# Example using short-form locations ----
x <- read_pptx()
x <- add_slide(x, "Title Slide")
x <- ph_with(x, "A title", "Title 1") # label
x <- ph_with(x, "A subtitle", 3) # id
x <- ph_with(x, "A left text", "left") # keyword
x <- ph_with(x, Sys.Date(), "dt[1]") # type + index
x <- ph_with(x, "More content", c(5, .5, 5, 2)) # numeric vector (left, top, width, heigh)
# \dontrun{
# print(x, preview = TRUE) # opens file locally
# }
}
\seealso{
Specify placeholder locations with \link{ph_location_type}, \link{ph_location},
\link{ph_location_label}, \link{ph_location_left}, \link{ph_location_right},
\link{ph_location_fullsize}, \link{ph_location_template}. \link{phs_with} is a sibling of
\code{ph_with} that fills multiple placeholders at once. Use \link{add_slide} to add new slides.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pptx_informations.R
\name{slide_size}
\alias{slide_size}
\title{Slides width and height}
\usage{
slide_size(x)
}
\arguments{
\item{x}{an rpptx object}
}
\description{
Get the width and height of slides in inches as
a named vector.
}
\examples{
my_pres <- read_pptx()
my_pres <- add_slide(my_pres,
  layout = "Two Content", master = "Office Theme")
slide_size(my_pres)
}
\seealso{
Other functions for reading presentation information:
\code{\link[=annotate_base]{annotate_base()}},
\code{\link[=color_scheme]{color_scheme()}},
\code{\link[=doc_properties]{doc_properties()}},
\code{\link[=layout_properties]{layout_properties()}},
\code{\link[=layout_summary]{layout_summary()}},
\code{\link[=length.rpptx]{length.rpptx()}},
\code{\link[=plot_layout_properties]{plot_layout_properties()}},
\code{\link[=slide_summary]{slide_summary()}}
}
\concept{functions for reading presentation information}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{sheet_names}
\alias{sheet_names}
\title{Sheet names of an xlsx workbook}
\usage{
sheet_names(x)
}
\arguments{
\item{x}{an rxlsx object (created by \code{\link[=read_xlsx]{read_xlsx()}}).}
}
\value{
A character vector of sheet names.
}
\description{
Return the sheet names of an \code{rxlsx} object, in the
order they appear in the workbook.
}
\examples{
wb <- read_xlsx()
sheet_names(wb)

wb <- add_sheet(wb, label = "new sheet")
sheet_names(wb)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_xlsx.R
\name{sheet_remove}
\alias{sheet_remove}
\title{Remove a sheet}
\usage{
sheet_remove(x, sheet)
}
\arguments{
\item{x}{rxlsx object}

\item{sheet}{name of the sheet to remove}
}
\value{
the rxlsx object (invisibly)
}
\description{
Remove a sheet from an xlsx workbook, deleting the
worksheet XML, its relationship file and the content-type override.
}
\examples{
wb <- read_xlsx()
wb <- add_sheet(wb, "kept")
default_name <- sheet_names(wb)[1]
wb <- sheet_remove(wb, sheet = default_name)
print(wb, target = tempfile(fileext = ".xlsx"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docx_section.R
\name{body_end_block_section}
\alias{body_end_block_section}
\title{Add any section}
\usage{
body_end_block_section(x, value)
}
\arguments{
\item{x}{an rdocx object}

\item{value}{a \link{block_section} object}
}
\description{
Add a section to the document. You can
define any section with a \link{block_section} object. All other
\verb{body_end_section_*} are specialized, this one is highly flexible
but it's up to the user to define the section properties.
}
\section{Section model in Word}{

A \code{block_section} added with \code{body_end_block_section()} applies to the
content that \strong{precedes} the call: it closes the section that holds the
previous paragraphs / tables and inherits any layout (orientation, columns,
margins, headers / footers) defined by the \code{block_section}. The function
name reflects this: it marks the \emph{end} of a section.

Typical pattern: add the content, then close it with the section that
should layout it.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{doc <- read_docx() |>
  body_add_par("This paragraph is in landscape orientation.") |>
  body_end_block_section(block_section(prop_section(
    page_size = page_size(orient = "landscape")
  )))
}\if{html}{\out{</div>}}

The default section of the document (defined by the template or by
\code{\link[=body_set_default_section]{body_set_default_section()}}) closes any content added after the last
\code{body_end_block_section()} call.

The RTF output uses the opposite model: \code{rtf_add(block_section(...))}
applies to the content that \emph{follows} the call. See \code{\link[=rtf_add]{rtf_add()}}.
}

\section{Illustrations}{


\if{html}{\figure{body_end_block_section_doc_1.png}{options: style="width:80\%;"}}
}

\examples{
library(officer)
str1 <- "Lorem ipsum dolor sit amet, consectetur adipiscing elit."
str1 <- rep(str1, 20)
str1 <- paste(str1, collapse = " ")

ps <- prop_section(
  page_size = page_size(orient = "landscape"),
  page_margins = page_mar(top = 2),
  type = "continuous"
)

doc_1 <- read_docx()
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")
doc_1 <- body_add_par(doc_1, value = str1, style = "Normal")

doc_1 <- body_end_block_section(doc_1, block_section(ps))

doc_1 <- body_add_par(doc_1, value = str1, style = "centered")

print(doc_1, target = tempfile(fileext = ".docx"))
}
\seealso{
Other functions for Word sections:
\code{\link[=body_end_section_columns]{body_end_section_columns()}},
\code{\link[=body_end_section_columns_landscape]{body_end_section_columns_landscape()}},
\code{\link[=body_end_section_continuous]{body_end_section_continuous()}},
\code{\link[=body_end_section_landscape]{body_end_section_landscape()}},
\code{\link[=body_end_section_portrait]{body_end_section_portrait()}},
\code{\link[=body_set_default_section]{body_set_default_section()}}
}
\concept{functions for Word sections}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ooxml_block_objects.R
\name{list_item}
\alias{list_item}
\title{Create a list item}
\usage{
list_item(x, level = 1L)
}
\arguments{
\item{x}{an \code{\link[=fpar]{fpar()}} object or a character string
(automatically converted to \code{\link[=fpar]{fpar()}})}

\item{level}{hierarchy level, integer starting at 1}
}
\description{
Wrap an \code{\link[=fpar]{fpar()}} with a hierarchy level for use
inside \code{\link[=block_list_items]{block_list_items()}}.
}
\seealso{
\code{\link[=block_list_items]{block_list_items()}}

Other block functions for reporting:
\code{\link[=block_caption]{block_caption()}},
\code{\link[=block_gg]{block_gg()}},
\code{\link[=block_list]{block_list()}},
\code{\link[=block_list_items]{block_list_items()}},
\code{\link[=block_pour_docx]{block_pour_docx()}},
\code{\link[=block_section]{block_section()}},
\code{\link[=block_table]{block_table()}},
\code{\link[=block_toc]{block_toc()}},
\code{\link[=fpar]{fpar()}},
\code{\link[=plot_instr]{plot_instr()}},
\code{\link[=unordered_list]{unordered_list()}}
}
\concept{block functions for reporting}
